Package: GeoModels
Version: 2.2.3
Date: 2026-03-31
Depends: R (>= 4.1.0), fields, mapproj, shape,progressr,future.apply
Suggests: numDeriv, memuse
Authors@R: c(person("Moreno", "Bevilacqua", role = c("aut", "cre","cph"),
                     email = "moreno.bevilacqua89@gmail.com"),
             person("Víctor", "Morales-Oñate", role = "ctb"),
             person("Francisco", "Cuevas-Pacheco", role = "ctb"),
             person("Christian", "Caamaño-Carrillo", role = "ctb"))
Description: Functions for Gaussian and Non Gaussian (bivariate) spatial and spatio-temporal data analysis are provided for a) (fast) simulation of random fields,  b) inference  for random fields using standard likelihood and a likelihood approximation  method called  weighted composite likelihood based on pairs and b) prediction using (local) best linear unbiased prediction. Weighted composite likelihood can be very efficient for estimating massive datasets. Both regression and spatial (temporal) dependence analysis can be jointly performed. Flexible covariance models for spatial and spatial-temporal data on Euclidean domains and spheres are provided. There are also many useful functions for plotting and performing diagnostic analysis. Different non Gaussian random fields can be considered in the analysis. Among them, random fields with marginal distributions such as Skew-Gaussian, Student-t, Tukey-h, Sin-Arcsin, Two-piece, Weibull, Gamma, Log-Gaussian, Binomial, Negative Binomial  and Poisson. See the URL for the papers associated with this package, as for instance, Bevilacqua and Gaetan (2015) <doi:10.1007/s11222-014-9460-6>, Bevilacqua et al. (2016) <doi:10.1007/s13253-016-0256-3>, Vallejos et al. (2020) <doi:10.1007/978-3-030-56681-4>, Bevilacqua et. al (2020) <doi:10.1002/env.2632>, Bevilacqua et. al (2021) <doi:10.1111/sjos.12447>, Bevilacqua et al. (2022) <doi:10.1016/j.jmva.2022.104949>, Morales-Navarrete et al. (2023) <doi:10.1080/01621459.2022.2140053>, and a large class of examples and tutorials.
Title: Procedures for Gaussian and Non Gaussian Geostatistical (Large)
        Data Analysis
License: GPL (>= 3)
Encoding: UTF-8
Imports: methods, spam, scatterplot3d, dotCall64, FastGP, plotrix,
        pracma, pbivnorm, sn, sp, nabor, hypergeo, VGAM, foreach,
        future, doFuture,minqa,withr
URL: https://vmoprojs.github.io/GeoModels-page/
BugReports: https://github.com/vmoprojs/GeoModels/issues
NeedsCompilation: yes
Packaged: 2026-04-06 13:51:22 UTC; morenobevilacqua
Author: Moreno Bevilacqua [aut, cre, cph],
  Víctor Morales-Oñate [ctb],
  Francisco Cuevas-Pacheco [ctb],
  Christian Caamaño-Carrillo [ctb]
Maintainer: Moreno Bevilacqua <moreno.bevilacqua89@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-06 14:20:08 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-08 22:33:49 UTC; unix
Archs: GeoModels.so.dSYM

Package: GeoModels
Version: 2.2.8
Date: 2026-08-15
Depends: R (>= 4.1.0), fields, progressr, future.apply
Suggests: numDeriv, memuse, mapproj, scatterplot3d, sp
Authors@R: c(person("Moreno", "Bevilacqua", role = c("aut", "cre","cph"),
                     email = "moreno.bevilacqua89@gmail.com"),
             person("Víctor", "Morales-Oñate", role = "ctb"),
             person("Francisco", "Cuevas-Pacheco", role = "ctb"),
             person("Christian", "Caamaño-Carrillo", role = "ctb"))
Description: Functions for Gaussian and Non Gaussian (bivariate) spatial and spatio-temporal data analysis are provided for a) (fast) simulation of random fields,  b) inference  for random fields using standard likelihood and a likelihood approximation  method called  weighted composite likelihood based on pairs and b) prediction using (local) best linear unbiased prediction. Weighted composite likelihood can be very efficient for estimating massive datasets. Both regression and spatial (temporal) dependence analysis can be jointly performed. Flexible covariance models for spatial and spatial-temporal data on Euclidean domains and spheres are provided. There are also many useful functions for plotting and performing diagnostic analysis. Different non Gaussian random fields can be considered in the analysis. Among them, random fields with marginal distributions such as Skew-Gaussian, Student-t, Tukey-h, Sin-Arcsin, Two-piece, Weibull, Gamma, Log-Gaussian, Binomial, Negative Binomial  and Poisson. See the URL for the papers associated with this package, as for instance, Bevilacqua and Gaetan (2015) <doi:10.1007/s11222-014-9460-6>, Bevilacqua et al. (2016) <doi:10.1007/s13253-016-0256-3>, Vallejos et al. (2020) <doi:10.1007/978-3-030-56681-4>, Bevilacqua et. al (2020) <doi:10.1002/env.2632>, Bevilacqua et. al (2021) <doi:10.1111/sjos.12447>, Bevilacqua et al. (2022) <doi:10.1016/j.jmva.2022.104949>, Morales-Navarrete et al. (2023) <doi:10.1080/01621459.2022.2140053>, and a large class of examples and tutorials.
Title: Procedures for Gaussian and Non Gaussian Geostatistical (Large)
        Data Analysis
License: GPL (>= 3)
Encoding: UTF-8
Imports: methods, spam, dotCall64, FastGP, pbivnorm, sn, nabor,
        hypergeo, VGAM, future, minqa
URL: https://vmoprojs.github.io/GeoModels-page/
BugReports: https://github.com/vmoprojs/GeoModels/issues
NeedsCompilation: yes
Author: Moreno Bevilacqua [aut, cre, cph],
  Víctor Morales-Oñate [ctb],
  Francisco Cuevas-Pacheco [ctb],
  Christian Caamaño-Carrillo [ctb]
Maintainer: Moreno Bevilacqua <moreno.bevilacqua89@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-21 05:46:27 UTC
Packaged: 2026-08-20 21:30:50 UTC; morenobevilacqua

export(CheckBiv,CheckDistance,CheckSph,CheckST,	CkCorrModel,CkInput,CkLikelihood,CkModel,	CkType,CompIndLik2,CompLik,CompLik2,	CorrelationPar,	CorrParam, corrsas, GeoAniso,
GeoKrigWeights,GeoKriglocWeights,GeoCorrFct_Cop,GeoDistances,	GeoCorrFct,GeoCovariogram,GeoCovDisplay,GeoCovmatrix,GeoCV,GeoFit,GeoFit2,GeoTestIsotropy,	GeoKrig,GeoKrigloc,GeoNA,GeoNeighborhood,GeoNeighIndex,GeoOutlier,GeoPit,	GeoQQ,GeoResiduals,GeoScatterplot,GeoSim,GeoSimapprox,GeoSimcond,GeoSimCopula,GeoScores,GeoTests,GeoVarestbootstrap,GeoVariogram,GeoVariogramDir,Lik,MatDecomp,NuisParam,NuisParam2,MargParam,GeoDoScores,StartParam,WlsStart,MatSqrt,MatInv,MatLogDet,GeoWLS,sp2Geo,GeoNeighbSelect,GeoTestsupp_space)
useDynLib(GeoModels, .registration = TRUE)
S3method(print,GeoFit) 
S3method(print,GeoCovmatrix) 
S3method(summary, GeoFit)
S3method(print,GeoWLS)
S3method(plot,GeoVariogram)
S3method(plot,GeoVariogramDir)
S3method(plot,GeoCorrFct)
importFrom("grDevices", dev.cur,gray.colors)
importFrom("graphics", abline, lines, par, persp,plot,points,image,plot.default,box,
axis,plot.new,plot.window,title)
importFrom("graphics", "hist","legend")
importFrom("stats", setNames, median, nlminb, dist, fft, quantile, qqnorm, qbinom,qpois,qnbinom,  pweibull, pgamma, plnorm, nlm, optim, optimize, integrate,approxfun,pchisq,pnorm,rt,qnorm,dnorm,dpois,rbeta,dbeta,dnbinom,dbinom,dweibull,dgamma,dlnorm,pbinom, plogis, pnbinom, ppois,rbinom,sd,
pbeta,dt,dlogis,qlnorm,qgamma,qweibull, qbeta,qlogis,lm, rnorm, runif, rgamma,rf,pt,qt,qunif,uniroot, var,cor)
importFrom("methods","is", "new", "slot<-","as")
importFrom("utils", "setTxtProgressBar", "txtProgressBar","capture.output")
importFrom("utils", "head","object.size")
importFrom("withr", with_seed)
importFrom("sp", "coordinates","is.projected")
importFrom("minqa","bobyqa")
importFrom("plotrix", "draw.circle")
importFrom("FastGP","rcppeigen_get_chol","rcppeigen_invert_matrix")
importFrom("VGAM","qzinegbin","qzipois")
import("fields", "mapproj","shape")
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(future.apply, future_lapply)
importFrom(future,multisession)
importFrom(future,plan)
importFrom(future,sequential)
importFrom(doFuture,"%dofuture%")
importFrom(progressr,handlers,progressor)

OLD versions of GeographicLib
=============================

This is where you will find pre-2.0 versions of GeographicLib.  For
newer versions of the C++ library, go to
https://sourceforge.net/projects/geographiclib/files/distrib-C++

For information about the implementations in other languages, see

> https://geographiclib.sourceforge.io/doc/library.html#languages

Programmers should download the latest `.tar.gz` or `.zip` file.  The
`.exe` files provide Windows installers for the utilities (and the
library compiled with Visual Studio 14 2015).

For installation instructions, see

> https://geographiclib.sourceforge.io/C++/doc/install.html

Test data for GeographicLib
===========================

*NOTE:* It's not necessary to download these files to use
GeographicLib.

`GeodTest.dat.gz` and `GeodTest-short.dat.gz` provide test data for
geodesics.  See

> https://geographiclib.sourceforge.io/C++/doc/geodesic.html#testgeod

`TMCoords.dat.gz` provides test data for the transverse Mercator
projection.  See

> https://geographiclib.sourceforge.io/C++/doc/transversemercator.html#testmerc

`GeoidHeights.dat.gz` provides test data for geoid heights.  See

> https://geographiclib.sourceforge.io/C++/doc/geoid.html#testgeoid

Geoid models for use with GeographicLib
=======================================

These files give the height of geoid above the ellipsoid on a
rectangular grid.  It's necessary to download and install at least one
of these files if you want to use the `Geoid` class in GeographicLib
or the `geoid_height` function in the [geographiclib-octave](
https://github.com/geographiclib/geographiclib-octave#readme) package.

The data is provided either as a compressed tar (extension `.tar.bz2`), a
zip file, or a windows installer (extension `.exe`).  If you're unsure
which one to install, start with

> egm96-5

For more information, see

> https://geographiclib.sourceforge.io/C++/doc/geoid.html#geoidinst

Gravity models for use with GeographicLib
=========================================

These files give the data for various gravity models.  It's
necessary to download and install at least one of these files if you
want to use the `GravityModel` class in GeographicLib.

The data is provided either as a compressed tar (extension `.tar.bz2`), a
zip file, or a windows installer (extension `.exe`).  If you're unsure
which one to install, start with

> egm96

For more information, see

> https://geographiclib.sourceforge.io/C++/doc/gravity.html#gravityinst

# Source code for C++ version of GeographicLib

Programmers should download the latest `.tar.gz` or `.zip` file.  The
`.exe` files provide Windows installers for the utilities (and the
library compiled with Visual Studio 14 2015).

The algorithms are documented in

* C. F. F. Karney,
  [Algorithms for geodesics](https://doi.org/10.1007/s00190-012-0578-z),
  J. Geodesy 87, 43-55 (2013);
  [Addenda](https://geographiclib.sourceforge.io/misc/geod-addenda.html).

Other links:

* Installation instructions:
  https://geographiclib.sourceforge.io/C++/doc/install.html
* Library documentation: https://geographiclib.sourceforge.io/C++/doc
* Git repository: https://github.com/geographiclib/geographiclib
* Git mirror site: https://sourceforge.net/p/geographiclib/code
* Source distribution (2.0 and later):
  https://sourceforge.net/projects/geographiclib/files/distrib-C++
* GeographicLib: https://geographiclib.sourceforge.io
* GeographicLib in various languages:
  https://geographiclib.sourceforge.io/doc/library.html#languages
* Old versions of Geographiclib:
  https://sourceforge.net/projects/geographiclib/files/distrib
* Author: Charles Karney, <karney@alum.mit.edu>

Magnetic models for use with GeographicLib
==========================================

These files give the data for various geomagnetic models.  It's
necessary to download and install at least one of these files if you
want to use the `MagneticModel` class in GeographicLib.

The data is provided either as a compressed tar (extension `.tar.bz2`), a
zip file, or a windows installer (extension `.exe`).  If you're unsure
which one to install, start with

> wmm2025

For more information, see

> https://geographiclib.sourceforge.io/C++/doc/magnetic.html#magneticinst

## Internal utilities shared by the parametric-bootstrap test functions.
## They centralize runtime plumbing only; each public test keeps its own
## statistical model, fitting strategy, and result construction.

.GeoTestPreserveRuntime <- function(gc_on_exit = FALSE) {
  old_plan <- if (requireNamespace("future", quietly = TRUE)) future::plan() else NULL
  old_handlers <- if (requireNamespace("progressr", quietly = TRUE)) progressr::handlers() else NULL
  old_progress_global <- if (requireNamespace("progressr", quietly = TRUE)) {
    progressr::handlers(global = NA)
  } else {
    NULL
  }

  function() {
    if (!is.null(old_plan)) try(future::plan(old_plan), silent = TRUE)
    if (requireNamespace("progressr", quietly = TRUE)) {
      try(progressr::handlers(old_handlers), silent = TRUE)
      if (!is.null(old_progress_global)) {
        ## Do not wrap global handler changes in try()/tryCatch(); progressr
        ## must call base::globalCallingHandlers() without handlers on stack.
        progressr::handlers(global = old_progress_global)
      }
    }
    if (isTRUE(gc_on_exit)) gc(verbose = FALSE, full = TRUE)
    invisible(NULL)
  }
}

.GeoTestSetSeed <- function(seed) {
  if (is.null(seed)) return(function() invisible(NULL))
  if (!is.numeric(seed) || length(seed) != 1L || !is.finite(seed)) {
    stop("seed must be NULL or a single finite numeric value", call. = FALSE)
  }

  had_seed <- exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  old_seed <- if (had_seed) get(".Random.seed", envir = .GlobalEnv, inherits = FALSE) else NULL
  set.seed(as.integer(seed))

  function() {
    if (had_seed) {
      assign(".Random.seed", old_seed, envir = .GlobalEnv)
    } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      rm(".Random.seed", envir = .GlobalEnv)
    }
    invisible(NULL)
  }
}

.GeoTestConfigureProgress <- function(progress, warn_missing = TRUE) {
  use_progressr <- isTRUE(progress) && requireNamespace("progressr", quietly = TRUE)
  if (isTRUE(progress) && !use_progressr && isTRUE(warn_missing)) {
    warning(
      "progress=TRUE but progressr is not available; progress bars disabled.",
      call. = FALSE
    )
  }
  if (use_progressr) {
    progressr::handlers(global = TRUE)
    progressr::handlers(progressr::handler_txtprogressbar(clear = TRUE))
  } else if (requireNamespace("progressr", quietly = TRUE)) {
    progressr::handlers("void")
  }
  use_progressr
}

.GeoTestCoreConfig <- function(parallel, ncores, B, strict_positive = TRUE) {
  ## strict_positive is retained for internal-call compatibility; all public
  ## ncores values now share the same positive-integer validation.
  ncores <- .GeoResolveWorkers(parallel, ncores, n_jobs = B)
  list(parallel = isTRUE(parallel) && ncores > 1L, ncores = ncores)
}

.GeoTestNormalizeBounds <- function(b, which = c("lower", "upper"),
                                    start_names, require_list = TRUE) {
  which <- match.arg(which)
  if (is.null(b)) b <- list()
  if (isTRUE(require_list) && !is.list(b)) {
    stop(which, " must be a named list", call. = FALSE)
  }

  b_names <- names(b)
  if (length(b) > 0L && (is.null(b_names) || any(b_names == ""))) {
    stop(which, " must be a named list with names matching start", call. = FALSE)
  }
  unknown <- setdiff(b_names, start_names)
  if (length(unknown) > 0L) {
    stop(which, " contains names not in start: ", paste(unknown, collapse = ", "), call. = FALSE)
  }

  base_val <- if (which == "lower") -Inf else Inf
  out <- setNames(rep(base_val, length(start_names)), start_names)
  if (length(b) > 0L) out[b_names] <- unlist(b, use.names = FALSE)
  as.list(out)
}

.GeoTestLogLikValue <- function(fit_obj) {
  if (inherits(fit_obj, "try-error") || is.null(fit_obj)) return(NA_real_)
  val <- if (!is.null(fit_obj$logCompLik)) fit_obj$logCompLik else fit_obj$logLik
  val <- as.numeric(val)
  if (length(val) != 1L || !is.finite(val)) NA_real_ else val
}

.GeoTestSamePairs <- function(fit0, fit1, type, require_indices = FALSE) {
  if (!identical(type, "Pairwise")) return(TRUE)
  if (isTRUE(require_indices) &&
      (is.null(fit0$rowidx) || is.null(fit0$colidx) ||
       is.null(fit1$rowidx) || is.null(fit1$colidx))) {
    return(FALSE)
  }
  identical(fit0$rowidx, fit1$rowidx) && identical(fit0$colidx, fit1$colidx)
}

.GeoTestGetRep <- function(sim, b) {
  x <- sim$data
  if (is.list(x)) return(x[[b]])
  if (is.null(dim(x))) return(x)
  x[, b]
}

.GeoTestSequentialApply <- function(sim, B, FUN, progress = FALSE) {
  run <- function() {
    pb <- if (isTRUE(progress)) progressr::progressor(along = seq_len(B)) else NULL
    out <- vector("list", B)
    for (b in seq_len(B)) {
      out[[b]] <- FUN(.GeoTestGetRep(sim, b))
      if (!is.null(pb)) pb(sprintf("b=%d", b))
    }
    out
  }
  if (isTRUE(progress)) progressr::with_progress(run()) else run()
}

.GeoTestSaveReplicates <- function(sim, B, prefix) {
  tag <- paste0(format(Sys.time(), "%Y%m%d_%H%M%S"), "_", Sys.getpid())
  vapply(seq_len(B), function(b) {
    path <- file.path(tempdir(), sprintf("%s_%s_%05d.rds", prefix, tag, b))
    saveRDS(.GeoTestGetRep(sim, b), path, compress = FALSE)
    path
  }, character(1))
}

.GeoTestParallelFileApply <- function(data_files, FUN, ncores,
                                      progressor = NULL,
                                      future_packages = NULL,
                                      globals = NULL,
                                      progress_every = FALSE,
                                      error_value = NULL) {
  old_plan <- future::plan()
  on.exit(try(future::plan(old_plan), silent = TRUE), add = TRUE)
  future::plan(future::multisession, workers = ncores)

  if (!is.null(globals)) {
    old_max_size <- getOption("future.globals.maxSize")
    on.exit(options(future.globals.maxSize = old_max_size), add = TRUE)
    required_size <- sum(vapply(
      globals, function(x) as.numeric(utils::object.size(x)), numeric(1)
    ))
    options(future.globals.maxSize = max(500 * 1024^2, 1.5 * required_size))
  }

  ## Use a bounded queue of individual futures instead of future_lapply().
  ## This lets the main R process detect each completed bootstrap refit and
  ## advance the package-standard progress bar immediately.  In particular,
  ## progress updates are not buffered until an entire future.apply chunk or
  ## bootstrap batch has finished.
  n_jobs <- length(data_files)
  if (n_jobs == 0L) return(vector("list", 0L))
  workers <- max(1L, min(as.integer(ncores), n_jobs))
  ans <- vector("list", n_jobs)
  active <- list()
  next_job <- 1L

  launch_one <- function(b) {
    path <- data_files[[b]]
    future::future(
      {
        FUN(readRDS(path))
      },
      globals = list(FUN = FUN, path = path),
      packages = future_packages,
      seed = TRUE
    )
  }

  launch_available <- function() {
    while (length(active) < workers && next_job <= n_jobs) {
      b <- next_job
      next_job <<- next_job + 1L
      active[[as.character(b)]] <<- launch_one(b)
    }
    invisible(NULL)
  }

  launch_available()
  while (length(active) > 0L) {
    ready <- vapply(active, future::resolved, logical(1))
    if (!any(ready)) {
      Sys.sleep(0.05)
      next
    }

    ids <- names(active)[ready]
    for (id in ids) {
      b <- as.integer(id)
      out <- tryCatch(
        future::value(active[[id]]),
        error = function(e) {
          if (!is.null(error_value)) return(error_value)
          list(
            valid = FALSE,
            lrt = NA_real_,
            ratio = NA_real_,
            angle = NA_real_,
            boundary_selected = FALSE,
            h1_attempts = 0L,
            h1_valid_attempts = 0L,
            error = conditionMessage(e)
          )
        }
      )
      ans[[b]] <- out
      active[[id]] <- NULL

      if (!is.null(progressor)) {
        ## Most bootstrap tests return a structured list with valid/lrt/ratio,
        ## but GeoTestIndependence() and GeoTestsupp_space() return named
        ## numeric vectors.  Never use $ on an arbitrary worker result.
        progress_ok <- isTRUE(progress_every)
        if (!progress_ok && is.list(out)) {
          progress_ok <- isTRUE(out$valid) &&
            length(out$lrt) == 1L && is.finite(out$lrt) &&
            length(out$ratio) == 1L && is.finite(out$ratio)
        }
        if (progress_ok) progressor(sprintf("b=%d", b))
      }
      launch_available()
    }
  }

  ans
}

.GeoVarestbootstrap_coordt_use <- function(coordt, spacetime = NULL,
                                                coordx_dyn = NULL) {
  ## Older GeoFit objects may encode the absence of a temporal dimension as
  ## a scalar (typically zero).  Passing that scalar back to GeoFit makes a
  ## purely spatial refit look space-time and fails CkInput().  Respect an
  ## explicit spacetime flag; for older objects without the flag, infer a
  ## temporal fit only from dynamic coordinates or more than one time value.
  if (!is.null(coordx_dyn)) return(coordt)
  if (!is.null(spacetime)) {
    if (!isTRUE(spacetime)) return(NULL)
    return(coordt)
  }
  if (is.numeric(coordt) && length(coordt) > 1L) coordt else NULL
}

.GeoVarestbootstrap_refit_worker <- function(ids, mode = c("tb", "file"),
                                               fit_spec, coords, target_names,
                                               data_files = NULL, sim_spec = NULL,
                                               sim_seeds = NULL, L = NULL,
                                               rng_kind = NULL, progressor = NULL) {
  .dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
  on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)

  mode <- match.arg(mode)
  ids <- as.integer(ids)
  ctx <- .GeoVarest_worker_context(fit_spec, coords)
  on.exit(.GeoVarest_release_fit_context(fit_spec), add = TRUE)

  use_stream <- identical(mode, "tb") && !identical(
    getOption("GeoModels.GeoVarest_tb_stream", TRUE), FALSE
  )
  sim_data <- NULL
  if (identical(mode, "tb") && !use_stream) {
    sim_data <- .GeoVarest_tb_simulate_batch(
      ids, sim_spec, sim_seeds, L, rng_kind = rng_kind
    )
  }

  out <- vector("list", length(ids))
  names(out) <- as.character(ids)

  for (ii in seq_along(ids)) {
    k <- ids[[ii]]
    current_file <- if (identical(mode, "file")) data_files[[k]] else NULL

    out[[ii]] <- tryCatch({
      current_data <- if (identical(mode, "file")) {
        readRDS(current_file)
      } else if (use_stream) {
        .GeoVarest_tb_simulate_one(k, sim_spec, sim_seeds, L, rng_kind = rng_kind)
      } else {
        sim_data[[ii]]
      }

      fit_result <- .GeoVarest_run_fit(
        current_data, coords, fit_spec, fit_spec$param,
        ctx$pair_info, ctx$fit_context, score = FALSE
      )
      .GeoVarest_extract_refit(fit_result, target_names)
    }, error = function(e) {
      .GeoVarest_failed_result(target_names, e)
    }, finally = {
      if (!is.null(current_file)) unlink(current_file, force = TRUE)
    })

    gc_every <- if (identical(mode, "tb")) 5L else 25L
    if (ii %% gc_every == 0L) gc(verbose = FALSE)
    if (!is.null(progressor)) progressor(sprintf("k=%d", k))
  }

  if (!is.null(sim_data)) rm(sim_data)
  gc(verbose = FALSE, full = identical(mode, "file"))
  out
}


GeoVarestbootstrap <- function(fit, K = 100, sparse = FALSE,
                               optimizer = NULL, lower = NULL, upper = NULL,
                               method = "cholesky", alpha = 0.95,
                               L = 10000, parallel = FALSE, ncores = 6,
                               progress = TRUE, seed = NULL,
                               min_success_rate = 0.8) {
  .dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
  on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)


  .GeoVarest_reject_full(fit, "GeoVarestbootstrap")
  ctl <- .GeoVarest_validate_controls(
    K, parallel, ncores, min_success_rate, "GeoVarestbootstrap"
  )
  K <- ctl$K
  ncores <- ctl$ncores
  min_success_rate <- ctl$min_success_rate

  ## ------------------------------------------------------------------
  ## Optional RNG seed, restoring user RNG state on exit
  ## ------------------------------------------------------------------
  restore_rng <- .GeoVarest_preserve_seed(seed)
  on.exit(restore_rng(), add = TRUE)

  ## ------------------------------------------------------------------
  ## Setup cleanup handlers
  ## ------------------------------------------------------------------
  future_plan_original <- future::plan()

  cleanup_resources <- function() {
    tryCatch({
      future::plan(future_plan_original)
    }, error = function(e) {
      tryCatch(
        future::plan(future::sequential),
        error = function(e2) invisible(NULL)
      )
    })
    gc(verbose = FALSE, full = TRUE)
  }
  on.exit(cleanup_resources(), add = TRUE)

  ## ------------------------------------------------------------------
  ## Progress handlers
  ## ------------------------------------------------------------------
  if (!is.logical(progress)) stop("progress must be logical (TRUE or FALSE)")
  restore_progress_handlers <- .GeoProgressHandlersPush(
    handler = if (progress) "txtprogressbar" else "void",
    global = if (progress) TRUE else NULL
  )
  on.exit(restore_progress_handlers(), add = TRUE)

  ## ------------------------------------------------------------------
  ## 1. Check input
  ## ------------------------------------------------------------------
  if (!(method %in% c("cholesky", "TB", "CE"))) {
    stop("The method of simulation is not correct")
  }

  coordt_use <- .GeoVarestbootstrap_coordt_use(
    fit$coordt, fit$spacetime, fit$coordx_dyn
  )

  .GeoValidate3D(
    coordx = fit$coordx, coordy = fit$coordy, coordz = fit$coordz,
    coordx_dyn = fit$coordx_dyn,
    distance = if (is.null(fit$distance)) "Eucl" else fit$distance,
    grid = isTRUE(fit$grid),
    approximate = !identical(method, "cholesky"), method = method,
    context = "GeoVarestbootstrap"
  )

  fit_dimension <- .GeoCoordinateDimension(
    fit$coordx, fit$coordy, fit$coordz, fit$coordx_dyn
  )
  if (identical(fit_dimension, 3L) && identical(method, "TB")) {
    if (isTRUE(fit$bivariate) || isTRUE(fit$spacetime)) {
      stop("GeoVarestbootstrap: three-dimensional TB simulation is supported only for purely spatial univariate fits.",
           call. = FALSE)
    }
    if (!is.null(fit$anisopars)) {
      stop("GeoVarestbootstrap: three-dimensional TB simulation does not support anisopars; transform the coordinates explicitly.",
           call. = FALSE)
    }
  }

  if (!(is.numeric(alpha) && length(alpha) == 1 && alpha > 0 && alpha < 1)) {
    stop("alpha must be a single numeric in (0,1)")
  }

  if (is.null(optimizer)) {
    optimizer <- fit$optimizer
    lower     <- fit$lower
    upper     <- fit$upper
  }

if (is.null(fit$thin_method)) {
  fit$thin_method <- "bernoulli"
}

thin_method_canonical <- function(x) {
  x <- tolower(as.character(x))
  if (x %in% c("bernoulli")) {
    return("bernoulli")
  }
  if (x %in% c("fixedbudget", "fixed_budget", "targetbalanced")) {
    return("FixedBudget")
  }
  if (x %in% c("match")) {
    return("match")
  }
  stop("Unsupported thin_method: ", x, call. = FALSE)
}

fit$thin_method <- thin_method_canonical(fit$thin_method)
  if (is.null(fit$p_neighb)) {
    fit$p_neighb <- 1
  }
  fit_memdist <- TRUE
  fit_weighted <- if (is.null(fit$weighted)) FALSE else isTRUE(fit$weighted)
  selected_pair_index <- .GeoVarest_selected_pairs(fit)
  ## Local reference only: it points to the two index vectors already stored
  ## in fit and is removed before returning the updated object.
  fit$pair_index <- selected_pair_index

  model <- fit$model
  cat("Parametric bootstrap can be time consuming ...\n")

  ## ------------------------------------------------------------------
  ## 2. Data-generating versus estimation model
  ## ------------------------------------------------------------------
  model_sim <- as.character(model)[1L]
  model_est <- .GeoVarest_estimation_model(fit, "GeoVarestbootstrap")
  boot_par <- .GeoVarest_bootstrap_parameters(fit)

  .GeoVarest_preflight_simulation(
    fit, method, model_sim, boot_par$sim_param,
    context = "GeoVarestbootstrap"
  )

  ## ------------------------------------------------------------------
  ## 3. Geometry and design matrix
  ## ------------------------------------------------------------------
  geometry <- .GeoVarest_prepare_geometry(fit, coordt_use)
  dimat <- geometry$dimat
  X_use <- geometry$X
  coords <- geometry$coords

  selected_pair_info <- .GeoVarest_prepare_pair_info(
    selected_pair_index,
    coords,
    list(
      type = fit$type,
      bivariate = fit$bivariate,
      spacetime = fit$spacetime,
      coordx_dyn = fit$coordx_dyn,
      distance = fit$distance,
      radius = fit$radius
    )
  )
  fit_context_main <- if (as.character(fit$type)[1L] %in% c("Pairwise", "Independence")) new.env(parent = emptyenv()) else NULL
  on.exit(.GeoVarest_release_fit_context(list(type = fit$type)), add = TRUE)

  ## For TB without a copula, stream each simulation directly into its
  ## bootstrap refit.  No temporary field files are created.
  tb_stream_mode <- identical(method, "TB") && is.null(fit$copula)

  ## ------------------------------------------------------------------
  ## 3.1. Memory pre-check
  ## ------------------------------------------------------------------
  estimated_size_mb <- (K * fit$numtime * fit$numcoord * 8) / (1024^2)
  if (!tb_stream_mode && estimated_size_mb > 500) {
    warning(sprintf(
      "Estimated simulated dataset size: %.1f MB (>500MB). Consider reducing K or using sparse methods.",
      estimated_size_mb
    ))
  }

  ## ------------------------------------------------------------------
  ## 4. Simulation
  ## ------------------------------------------------------------------
  if (!tb_stream_mode) {
    sim_args <- list(
      coordx     = coords,
      coordt     = coordt_use,
      coordx_dyn = fit$coordx_dyn,
      anisopars  = boot_par$anisopars,
      corrmodel  = fit$corrmodel,
      model      = model_sim,
      param      = boot_par$sim_param,
      grid       = fit$grid,
      X          = fit$X,
      n          = fit$n,
      distance   = fit$distance,
      radius     = fit$radius,
      nrep       = K,
      progress   = progress
    )

    if (is.null(fit$copula)) {
      if (method == "cholesky") {
        data_sim_full <- do.call(GeoSim, c(sim_args, list(sparse = sparse, method = method)))
      } else if (method %in% c("TB", "CE")) {
        ## GeoSimapprox may parallelize over nrep = K. This phase is
        ## completed before the bootstrap refits start, so the refit
        ## parallelism below is not nested.
        old_plan_after_sim <- NULL

        if (isTRUE(parallel) && requireNamespace("future", quietly = TRUE)) {
          old_plan_after_sim <- future::plan()
        }

        data_sim_full <- do.call(
          GeoSimapprox,
          c(
            sim_args,
            list(
              method   = method,
              L        = L,
              parallel = parallel,
              ncores   = ncores
            )
          )
        )

        if (!is.null(old_plan_after_sim)) {
          try(future::plan(old_plan_after_sim), silent = TRUE)
        }
      } else {
        stop("Unsupported method for simulation")
      }
    } else {
      if (method %in% c("cholesky", "TB")) {
        copula_args <- list(copula = fit$copula, sparse = sparse, method = method)
        if (identical(method, "TB")) {
          copula_args <- c(copula_args, list(L = L, parallel = parallel, ncores = ncores))
        }
        data_sim_full <- do.call(
          GeoSimCopula,
          c(sim_args, copula_args)
        )
      } else {
        stop("Copula simulation supports method = 'cholesky' or method = 'TB'.")
      }
    }

    data_sim <- data_sim_full$data
    rm(data_sim_full)
    gc(verbose = FALSE, full = TRUE)

  }

  ## ------------------------------------------------------------------
  ## 5. Setting parallelization
  ## ------------------------------------------------------------------
  ncores <- .GeoResolveWorkers(parallel, ncores, n_jobs = K)
  parallel <- isTRUE(parallel) && ncores > 1L

  ## ------------------------------------------------------------------
  ## 6. Parameter names and sequential refit helper
  ## ------------------------------------------------------------------
  target_names <- names(fit$param)
  if (is.null(target_names) || length(target_names) != length(fit$param)) {
    target_names <- names(unlist(fit$param, use.names = TRUE))
  }
  if (is.null(target_names) || length(target_names) != length(fit$param)) {
    target_names <- paste0("par", seq_along(fit$param))
  }

  main_fit_spec <- .GeoVarest_make_fit_spec(
    fit, boot_par, model_est, X_use, optimizer, lower, upper,
    selected_pair_index, coordt_use, memdist = fit_memdist,
    weighted = fit_weighted, include_start = TRUE
  )

  num_params <- length(fit$param)

  estimate_fun <- function(k, current_data) {
    result <- tryCatch({
      capture.output({
        fit_result <- .GeoVarest_run_fit(
          current_data, coords, main_fit_spec, main_fit_spec$param,
          selected_pair_info, fit_context_main, score = FALSE
        )
      }, file = nullfile())

      out <- .GeoVarest_extract_refit(fit_result, target_names)
      rm(fit_result)
      if (k %% 5L == 0L) gc(verbose = FALSE)
      out
    }, error = function(e) {
      if (k <= 3) {
        cat("Warning: iteration", k, "failed:", conditionMessage(e), "\n")
      }
      rep(NA_real_, num_params + 1L)
    })

    result
  }

  ## ------------------------------------------------------------------
  ## 9. Bootstrap
  ## ------------------------------------------------------------------
  param_cols <- seq_len(num_params)

  if (tb_stream_mode) {
    sim_spec <- list(
      coordx     = coords,
      coordt     = coordt_use,
      coordx_dyn = fit$coordx_dyn,
      anisopars  = boot_par$anisopars,
      corrmodel  = fit$corrmodel,
      model      = model_sim,
      param      = boot_par$sim_param,
      grid       = fit$grid,
      X          = fit$X,
      n          = fit$n,
      distance   = fit$distance,
      radius     = fit$radius
    )

    fit_spec <- .GeoVarest_make_fit_spec(
      fit, boot_par, model_est, X_use, optimizer, lower, upper,
      selected_pair_index, coordt_use, memdist = fit_memdist,
      weighted = fit_weighted, include_start = TRUE
    )

    sim_seeds <- sample.int(.Machine$integer.max, K, replace = FALSE)
    ## Preserve the caller's RNG kind across future multisession workers.
    ## future.seed=TRUE switches workers to L'Ecuyer-CMRG; without this
    ## explicit metadata the same scalar replicate seed generates different
    ## TB bootstrap fields in sequential and parallel execution.
    sim_rng_kind <- RNGkind()
    worker_fun <- if (requireNamespace("GeoModels", quietly = TRUE) &&
                      exists(".GeoVarestbootstrap_refit_worker",
                             envir = asNamespace("GeoModels"), inherits = FALSE)) {
      getFromNamespace(".GeoVarestbootstrap_refit_worker", "GeoModels")
    } else {
      .GeoVarestbootstrap_refit_worker
    }

    workers <- if (isTRUE(parallel)) ncores else 1L
    use_outer_stream <- isTRUE(parallel) && workers > 1L &&
      requireNamespace("future", quietly = TRUE) &&
      requireNamespace("future.apply", quietly = TRUE)

    if (isTRUE(parallel) && !use_outer_stream) {
      warning(
        "Parallel TB bootstrap requested but future/future.apply is unavailable; using sequential streaming.",
        call. = FALSE
      )
    }

    if (use_outer_stream) {
      future_backend <- .GeoVarest_set_future_plan(workers)
      chunks <- split(seq_len(K), rep(seq_len(workers), length.out = K))

      payload_size <- as.numeric(utils::object.size(list(
        sim_spec = sim_spec,
        fit_spec = fit_spec,
        coords = coords,
        target_names = target_names,
        sim_seeds = sim_seeds,
        sim_rng_kind = sim_rng_kind
      )))
      old_limit <- getOption("future.globals.maxSize")
      on.exit(options(future.globals.maxSize = old_limit), add = TRUE)
      options(future.globals.maxSize = max(500 * 1024^2, 1.5 * payload_size))

      cat("Performing", K, "TB simulations and bootstrap refits on",
          workers, "cores...\n")

      run_chunks <- function(progressor = NULL) {
        future.apply::future_lapply(
          X = chunks,
          FUN = worker_fun,
          mode = "tb",
          sim_spec = sim_spec,
          fit_spec = fit_spec,
          coords = coords,
          target_names = target_names,
          sim_seeds = sim_seeds,
          L = L,
          rng_kind = sim_rng_kind,
          progressor = progressor,
          future.seed = TRUE,
          future.scheduling = 1,
          future.packages = "GeoModels",
          future.globals = FALSE
        )
      }

      chunk_results <- if (progress) {
        progressr::with_progress({
          pb <- progressr::progressor(steps = K)
          pb_worker <- if (as.numeric(utils::object.size(pb)) <= 1024^2) pb else NULL
          run_chunks(pb_worker)
        })
      } else {
        run_chunks(NULL)
      }

      res_list <- do.call(c, chunk_results)
      if (length(res_list) > 0L && !is.null(names(res_list))) {
        ids <- suppressWarnings(as.integer(names(res_list)))
        if (all(is.finite(ids))) res_list <- res_list[order(ids)]
      }
    } else {
      cat("Performing", K, "TB simulations and bootstrap refits sequentially...\n")
      if (progress) {
        res_list <- progressr::with_progress({
          pb <- progressr::progressor(steps = K)
          unname(worker_fun(
            ids = seq_len(K),
            mode = "tb",
            sim_spec = sim_spec,
            fit_spec = fit_spec,
            coords = coords,
            target_names = target_names,
            sim_seeds = sim_seeds,
            L = L,
            rng_kind = sim_rng_kind,
            progressor = pb
          ))
        })
      } else {
        res_list <- unname(worker_fun(
          ids = seq_len(K),
          mode = "tb",
          sim_spec = sim_spec,
          fit_spec = fit_spec,
          coords = coords,
          target_names = target_names,
          sim_seeds = sim_seeds,
          L = L,
          rng_kind = sim_rng_kind,
          progressor = NULL
        ))
      }
    }

    for (k in seq_len(min(3L, length(res_list)))) {
      err <- attr(res_list[[k]], "error")
      if (!is.null(err)) cat("Warning: iteration", k, "failed:", err, "\n")
    }

    res_raw <- do.call(rbind, res_list)
    if (is.null(res_raw)) {
      res_raw <- matrix(NA_real_, nrow = 0L, ncol = num_params + 1L)
    }

    if (exists("chunk_results", inherits = FALSE)) rm(chunk_results)
    rm(res_list, sim_spec, fit_spec, sim_seeds)
    gc(verbose = FALSE, full = TRUE)

  } else if (!parallel) {
    cat("Performing", K, "estimations sequentially...\n")
    res_list <- vector("list", K)

    if (progress) {
      progressr::with_progress({
        pb <- progressr::progressor(along = seq_len(K))
        for (k in seq_len(K)) {
          pb(sprintf("k=%d", k))
          res_list[[k]] <- suppressWarnings(estimate_fun(k, data_sim[[k]]))
        }
      })
    } else {
      for (k in seq_len(K)) {
        res_list[[k]] <- suppressWarnings(estimate_fun(k, data_sim[[k]]))
      }
    }

    res_raw <- do.call(rbind, res_list)
    if (is.null(res_raw)) {
      res_raw <- matrix(NA_real_, nrow = 0, ncol = num_params + 1L)
    }

  } else {
    cat("Performing", K, "estimations using", ncores, "cores...\n")

    ## Save one dataset per file. Workers receive only paths and a compact
    ## fit specification; no local GeoFit/extraction closures are exported.
    temp_root <- tempfile("GeoVarestbootstrap_")
    dir.create(temp_root, recursive = TRUE, showWarnings = FALSE)
    on.exit(unlink(temp_root, recursive = TRUE, force = TRUE), add = TRUE)
    data_files <- file.path(temp_root, sprintf("boot_%05d.rds", seq_len(K)))
    for (k in seq_len(K)) {
      saveRDS(data_sim[[k]], data_files[[k]], compress = FALSE)
    }

    rm(data_sim)
    gc(verbose = FALSE, full = TRUE)

    fit_spec <- .GeoVarest_make_fit_spec(
      fit, boot_par, model_est, X_use, optimizer, lower, upper,
      selected_pair_index, coordt_use, memdist = fit_memdist,
      weighted = fit_weighted, include_start = TRUE
    )

    worker_fun <- if (requireNamespace("GeoModels", quietly = TRUE) &&
                      exists(".GeoVarestbootstrap_refit_worker",
                             envir = asNamespace("GeoModels"),
                             inherits = FALSE)) {
      getFromNamespace(".GeoVarestbootstrap_refit_worker", "GeoModels")
    } else {
      .GeoVarestbootstrap_refit_worker
    }

    future::plan(
      future::multisession,
      workers = ncores,
      rscript_libs = .GeoVarest_future_libpaths()
    )
    chunks <- split(seq_len(K), rep(seq_len(ncores), length.out = K))

    payload_size <- as.numeric(utils::object.size(list(
      data_files = data_files,
      fit_spec = fit_spec,
      coords = coords,
      target_names = target_names
    )))
    old_limit <- getOption("future.globals.maxSize")
    on.exit(options(future.globals.maxSize = old_limit), add = TRUE)
    options(future.globals.maxSize = max(500 * 1024^2, 1.5 * payload_size))

    run_refits <- function(progressor = NULL) {
      future.apply::future_lapply(
        X = chunks,
        FUN = worker_fun,
        mode = "file",
        data_files = data_files,
        fit_spec = fit_spec,
        coords = coords,
        target_names = target_names,
        progressor = progressor,
        future.seed = TRUE,
        future.packages = "GeoModels",
        future.globals = FALSE
      )
    }

    chunk_results <- if (progress) {
      progressr::with_progress({
        pb_fit <- progressr::progressor(steps = K)
        pb_worker <- if (as.numeric(utils::object.size(pb_fit)) <= 1024^2) {
          pb_fit
        } else {
          NULL
        }
        run_refits(pb_worker)
      })
    } else {
      run_refits(NULL)
    }

    res_list <- do.call(c, chunk_results)
    if (length(res_list) > 0L && !is.null(names(res_list))) {
      ids <- suppressWarnings(as.integer(names(res_list)))
      if (all(is.finite(ids))) res_list <- res_list[order(ids)]
    }

    for (k in seq_len(min(3L, length(res_list)))) {
      err <- attr(res_list[[k]], "error")
      if (!is.null(err)) {
        cat("Warning: iteration", k, "failed:", err, "\n")
      }
    }

    res_raw <- do.call(rbind, res_list)
    if (is.null(res_raw)) {
      res_raw <- matrix(NA_real_, nrow = 0L, ncol = num_params + 1L)
    }

    rm(res_list, chunk_results, fit_spec)
    gc(verbose = FALSE, full = TRUE)
  }

  ## Filter rows with all NA parameter estimates.
  if (nrow(res_raw) > 0) {
    valid <- !apply(
      res_raw[, param_cols, drop = FALSE],
      1L,
      function(r) all(is.na(r))
    )
    res <- res_raw[valid, , drop = FALSE]
  } else {
    res <- res_raw
  }

  if (ncol(res) == num_params + 1L) {
    colnames(res) <- c(target_names, "logCompLik")
  }

  ## ------------------------------------------------------------------
  ## Clean memory
  ## ------------------------------------------------------------------
  if (exists("data_sim", inherits = FALSE)) rm(data_sim)
  if (exists("res_list", inherits = FALSE)) rm(res_list)
  if (exists("res_raw", inherits = FALSE)) rm(res_raw)
  gc(verbose = FALSE, full = TRUE)

    ## ------------------------------------------------------------------
  ## 10. Post-processing
  ## ------------------------------------------------------------------
  n_successful <- nrow(res)
  .GeoVarest_check_success(
    n_successful, K, min_success_rate, "GeoVarestbootstrap"
  )

  cat("Successful bootstrap iterations:", n_successful, "out of", K, "\n")

  param_mat <- res[, param_cols, drop = FALSE]
  invG      <- stats::var(param_mat)
  G         <- try(solve(invG), silent = TRUE)
  if (!is.matrix(G)) {
    warning("Bootstrap estimated Godambe matrix is singular")
  }

  stderr <- sqrt(pmax(0, diag(invG)))
  names(stderr) <- target_names

  numparam <- length(fit$param)
  penalty <- NA_real_
  claic <- clbic <- NA_real_
  fit$varimat <- NULL
  fit$Jmat <- NULL

  if (!is.null(fit$sensmat)) {
    H <- as.matrix(fit$sensmat)
    storage.mode(H) <- "double"
    if (all(dim(H) == c(numparam, numparam)) && all(is.finite(H))) {
      H <- (H + t(H)) / 2
      penalty <- sum(diag(H %*% invG))
      claic   <- -2 * fit$logCompLik + 2 * penalty
      clbic   <- -2 * fit$logCompLik + log(dimat) * penalty
      ## Under Var(theta_hat)=H^{-1} J H^{-1}, recover J for
      ## compatibility with score-bootstrap output.
      fit$varimat <- H %*% invG %*% H
      fit$Jmat <- fit$varimat
    } else {
      warning(
        "GeoVarestbootstrap: fit$sensmat is incompatible with the bootstrap covariance; CLIC/CLBIC were not computed.",
        call. = FALSE
      )
    }
  }

  ## ------------------------------------------------------------------
  ## 10.1 Bootstrap percentile confidence intervals
  ## ------------------------------------------------------------------
  theta_hat <- as.numeric(fit$param)
  names(theta_hat) <- target_names
  probs <- c((1 - alpha) / 2, 1 - (1 - alpha) / 2)

  q_boot <- t(apply(param_mat, 2L, stats::quantile, probs = probs, na.rm = TRUE))
  colnames(q_boot) <- c("Lower_perc", "Upper_perc")
  fit$conf.int.bootstrap_percentile <- t(q_boot)
  rownames(fit$conf.int.bootstrap_percentile) <- c("Lower", "Upper")

  ## ------------------------------------------------------------------
  ## 11. Final output
  ## ------------------------------------------------------------------
  fit$claic     <- claic
  fit$clic      <- claic
  fit$clbic     <- clbic
  fit$clic_penalty <- penalty
  fit$stderr    <- stderr
  fit$varcov    <- invG
  fit$godambe   <- if (is.matrix(G)) (G + t(G)) / 2 else NULL
  fit$estimates <- res

  z_alpha <- stats::qnorm(1 - (1 - alpha) / 2)
  aa      <- z_alpha * stderr
  pp      <- as.numeric(fit$param)

  fit$conf.int <- rbind(pp - aa, pp + aa)
  colnames(fit$conf.int) <- target_names
  rownames(fit$conf.int) <- c("Lower", "Upper")

  fit$pvalues <- 2 * stats::pnorm(-abs(pp / stderr))
  names(fit$pvalues) <- target_names
  fit$pvalues_type <- "Wald p-values based on parametric-bootstrap variance"

  fit$bootstrap_type <- "parametric"
  fit$bootstrap_K <- K
  fit$bootstrap_successful <- n_successful
  fit$bootstrap_success_rate <- n_successful / K
  fit$bootstrap_min_success_rate <- min_success_rate
  fit$bootstrap_failures <- data.frame(
    iteration = which(!valid),
    error = rep(NA_character_, sum(!valid)),
    stringsAsFactors = FALSE
  )
  fit$pair_index <- NULL
  fit$bootstrap_pair_graph <- if (is.null(selected_pair_index)) "deterministic" else "reused_original"
  fit$bootstrap_thin_method <- fit$thin_method
  fit$bootstrap_p_neighb <- fit$p_neighb
  fit$bootstrap_seed <- seed
  fit$bootstrap_thinning_seed_rule <- if (is.null(selected_pair_index)) {
    "not applicable"
  } else {
    "not used; the original selected-pair graph is reused"
  }
  fit$bootstrap_simulation_method <- method
  fit$bootstrap_memdist <- fit_memdist
  fit$bootstrap_weighted <- fit_weighted
  fit$bootstrap_L <- if (identical(method, "TB")) L else NULL
  fit$bootstrap_workflow <- if (tb_stream_mode) "one-pass TB simulation/refit" else "standard"
  if (tb_stream_mode) {
    fit$bootstrap_execution_mode <- if (isTRUE(use_outer_stream)) {
      "parallel one-pass TB simulation and full refitting"
    } else {
      "sequential one-pass TB simulation and full refitting"
    }
    fit$bootstrap_parallel_workers <- if (isTRUE(use_outer_stream)) workers else 1L
    fit$bootstrap_simulation_seed_rule <-
      "explicit per-replication seeds generated from the optional seed argument"
  }

  return(fit)
}

####################################################
### File name: GeoKrig.r
####################################################

GeoKrig= function(estobj=NULL,data, coordx=NULL, coordy=NULL,coordz=NULL, coordt=NULL, coordx_dyn=NULL, corrmodel, distance="Eucl", grid=FALSE, loc, 
                method="cholesky", model="Gaussian", n=1,nloc=NULL, mse=FALSE,  param, anisopars=NULL,
               radius=1, sparse=FALSE,time=NULL,  type_krig="Simple",weigthed=TRUE,
               which=1, copula=NULL,X=NULL,Xloc=NULL,Mloc=NULL,spobj=NULL,spdata=NULL,varcov=NULL,progress=FALSE,
               check.duplicates=FALSE)

{
##################################################################################################################
##################################################################################################################

############## internal functions are defined in KrigUtils.R ############################


######################################
########## START #####################
######################################
call <- match.call()

###############################
## checking if there is a  GeoFit object
###############################
if(!is.null(estobj)){
   if(!inherits(estobj,"GeoFit"))
               stop("need  a 'GeoFit' object as input\n")

data=estobj$data
################################################################
if(!estobj$grid){  #not regular grid 
 if(!estobj$bivariate){  if(is.null(estobj$coordx_dyn)) coordx=cbind(estobj$coordx,estobj$coordy,estobj$coordz) ####spatial and spacetime fixed loc
                          else                          coordx=NULL
                      } ## spatial (temporal) non regular case
 else  {    if(is.null(estobj$coordx_dyn))  { coordx=estobj$coordx[1:estobj$ns[1]]    # bivariate not dynamic    
                                              coordy=estobj$coordy[1:estobj$ns[2]] 
                                            }  
            else {coordx=NULL}                                      # bivariate  dynamic  
       }
 }                  # regular grid
else  { coordx=estobj$coordx; coordy=estobj$coordy; coordz=estobj$coordz} # grid(nunca)

   if(isTRUE(estobj$spacetime)) coordt=estobj$coordt
   else coordt=NULL

   coordx_dyn=estobj$coordx_dyn
   corrmodel=estobj$corrmodel
   model=estobj$model
   distance=estobj$distance
   grid=estobj$grid
   n=estobj$n
   param=append(estobj$param,estobj$fixed)
   radius=estobj$radius
   copula=estobj$copula
   anisopars=estobj$anisopars
   varcov=estobj$varcov
   if(!is.null(estobj$sparse)) sparse=estobj$sparse
   if(!is.null(anisopars))  {param$angle=NULL; param$ratio=NULL}

   X=estobj$X
}
##################################
###### end check geofit object####
##################################
model_requested <- gsub("[[:blank:]]", "", as.character(model)[1L])
model <- .GeoPredictionCanonicalModel(model_requested)
param_input <- param
n_input <- n
nloc_input <- nloc
if(is.null(CkModel(model))) stop("The name of the  model  is not correct\n")
if(!is.character(corrmodel) || is.null(CkCorrModel(corrmodel))) stop("the name of the correlation model is wrong")
corrmodel=gsub("[[:blank:]]", "",corrmodel)
if (!is.null(copula)) {
  copula <- as.character(copula)[1L]
  if (is.na(copula) || !nzchar(gsub("[[:blank:]]", "", copula))) {
    copula <- NULL
  } else {
    copula <- gsub("[[:blank:]]", "", copula)
  }
}
if(!is.null(copula) && !(copula %in% c("Gaussian", "Clayton", "SkewGaussian")))
  stop("GeoKrig currently implements copula-based linear prediction for 'Gaussian', 'Clayton', and 'SkewGaussian'.", call.=FALSE)
distance=gsub("[[:blank:]]", "",distance)
method=.GeoKrigMatchMethod(method, sparse=sparse, context="GeoKrig")
type_key=tolower(gsub("[[:blank:]]", "",type_krig))
if(!(type_key %in% c("simple","universal"))) stop("type_krig must be 'Simple' or 'Universal'")
type_krig=if(type_key=="simple") "Simple" else "Universal"
if(!is.logical(weigthed) || length(weigthed)!=1L || is.na(weigthed))
  stop("weigthed must be TRUE or FALSE")
if(!is.logical(progress) || length(progress)!=1L || is.na(progress))
  stop("progress must be TRUE or FALSE")
if(!isTRUE(weigthed))
  warning("weigthed is retained for backward compatibility and has no effect in standard kriging.", call.=FALSE)


########################################################
####### extracting sp objects if necessary 
########################################################
bivariate<-CheckBiv(CkCorrModel(corrmodel))
spacetime<-CheckST(CkCorrModel(corrmodel))
space=!spacetime&&!bivariate
.GeoKrigValidatePredictionModel(model, copula = copula, bivariate = bivariate, context = "GeoKrig")
model_code_copula <- as.integer(CkModel(model))
copula_cont <- !bivariate && !is.null(copula) &&
  copula %in% c("Gaussian", "Clayton", "SkewGaussian") &&
  (model_code_copula %in% .GeoGaussianCopulaContinuousModels ||
   (copula %in% c("Clayton", "SkewGaussian") &&
    model_code_copula %in% .GeoGaussianCopulaDiscreteModels))
if(!is.null(spobj)) {
   if(space||bivariate){
        a=sp2Geo(spobj,spdata); coordx=a$coords 
       if(!a$pj) {if(distance!="Chor") distance="Geod"}
    }
   if(spacetime){
        a=sp2Geo(spobj,spdata); coordx=a$coords ; coordt=a$coordt 
        if(!a$pj) {if(distance!="Chor") distance="Geod"}
     }
   if (!is.null(a$Y)) data <- a$Y
   if (!is.null(a$X)) X <- a$X
}

if(spacetime && !is.null(coordx_dyn)) {
  coordx_dyn <- .GeoNormalizeDynamicCoordinates(
    coordx_dyn, coordt = coordt, context = "GeoKrig"
  )
  .GeoValidateDynamicData(data, coordx_dyn, context = "GeoKrig")
  .GeoValidateDynamicRows(X, coordx_dyn, name = "X", context = "GeoKrig")
}
if(bivariate) {
  biv_info <- .GeoBivariateCoordinateInfo(
    coordx = coordx, coordy = coordy, coordz = coordz,
    coordx_dyn = coordx_dyn, grid = grid, context = "GeoKrig"
  )
  if(biv_info$dynamic) coordx_dyn <- biv_info$blocks
  data <- .GeoBivariateDataVector(data, biv_info$ns, context = "GeoKrig")
  if(!is.null(X)) X <- .GeoBivariateDesign(X, biv_info$ns, context = "GeoKrig")
}
if(!bivariate && is.null(coordx_dyn)) {
  coords_input <- .GeoKrigCoordinateMatrix(
    coordx = coordx, coordy = coordy, coordz = coordz, grid = grid,
    context = "GeoKrig"
  )
  data <- .GeoKrigNormalizeData(
    data, nspace = nrow(coords_input),
    coordt = if(spacetime) coordt else NULL,
    grid = grid, context = "GeoKrig"
  )
}

.GeoCheckDuplicateLocations(
  coordx = coordx, coordy = coordy, coordz = coordz, coordt = coordt,
  coordx_dyn = coordx_dyn, spacetime = spacetime, bivariate = bivariate,
  grid = grid, context = "GeoKrig", check.duplicates = check.duplicates
)
.dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)

######################################
############ some checks##############
######################################
if (is.vector(loc)) {
  loc <- matrix(loc, nrow = 1)
} else if (!is.matrix(loc)) {
  loc <- as.matrix(loc)
}

if(!(ncol(loc)==2||ncol(loc)==3))   stop("loc parameter must be a matrix N x 2 or N x 3")
if(!is.numeric(loc) || !nrow(loc) || any(!is.finite(loc)))
  stop("loc must be a finite numeric N x 2 or N x 3 matrix")
if(!is.null(anisopars) && !identical(distance, "Eucl"))
  stop("GeoKrig: anisopars requires distance = 'Eucl'.")


.GeoValidate3D(
  coordx = coordx, coordy = coordy, coordz = coordz,
  coordx_dyn = coordx_dyn, distance = distance, grid = grid,
  context = "GeoKrig"
)
.GeoValidate3D(
  coordx = loc, distance = distance, grid = FALSE,
  context = "GeoKrig prediction locations"
)
.GeoValidateCoordinateCompatibility(
  coordx = coordx, coordy = coordy, coordz = coordz,
  coordx_dyn = coordx_dyn, loc = loc, context = "GeoKrig"
)

if (!is.null(X) && !is.matrix(X) && !is.list(X)) X <- as.matrix(X)

## Mean specification.  A vector param$mean is an external mean at the
## observations; mean, mean1, ... are regression coefficients otherwise.
Mobs <- .GeoMean_external(param)
if (!is.null(Mobs) && !is.null(X)) {
  stop("A fixed observation mean and X cannot be used together.\n")
}
if (!is.null(Mloc) && !is.null(Xloc)) {
  stop("Use either Mloc or Xloc, not both.\n")
}
if (!is.null(Mobs) && is.null(Mloc)) {
  stop("Mloc is required when the observation mean is supplied as a vector.\n")
}
if (!is.null(X) && is.null(Xloc) && is.null(Mloc)) {
  stop("Xloc or Mloc is required when X is supplied.\n")
}
if (is.null(X) && !is.null(Xloc)) {
  stop("X must be supplied when Xloc is used.\n")
}
if (is.null(X) && is.null(Mobs) &&
    length(setdiff(.GeoMean_supplied_names(param), "mean"))) {
  stop("Parameters mean1, mean2, ... require an explicit design matrix X; without X the mean is intercept-only.\n")
}
if(identical(type_krig, "Universal") && (!is.null(Mobs) || !is.null(Mloc))) {
  warning("External observation/prediction means are treated as known; ",
          "type_krig is therefore set to 'Simple'.", call. = FALSE)
  type_krig <- "Simple"
}
if(identical(type_krig, "Universal") && isTRUE(mse) && is.null(varcov)) {
  stop("Universal kriging with mse = TRUE requires varcov (the covariance matrix of the estimated parameters).")
}
if(spacetime) {
  if(is.null(time)) stop("At least one temporal instant is needed for space-time kriging")
  if(!is.numeric(time) || !length(time) || any(!is.finite(time)))
    stop("time must be a non-empty finite numeric vector")
  time=as.numeric(time)
}
loc_orig=loc
if(!is.null(anisopars)) {  loc=GeoAniso(loc,c(anisopars$angle,anisopars$ratio))}

if(!is.null(Mloc)) {
    if(!is.numeric(Mloc)) stop("Mloc must be numeric")
    Mloc <- as.numeric(Mloc)
    if(any(!is.finite(Mloc))) stop("Mloc contains non-finite values")
}
######################################
############ end some checks##########
######################################



#### locations points to predict ###
if(is.null(time)) time=0; numloc = nrow(loc); tloc = length(time);
if(!tloc)  tloc = 1;

## Normalize observation/prediction size parameters before covariance construction.
## A location-specific n can never be averaged to invent nloc.
count_sizes <- NULL
count_model_names <- c("Binary", "Bernoulli", "Binomial",
                       "Geom", "Geometric", "BinomialNeg", "BinomialNegZINB",
                       "Poisson", "PoissonZIP", "PoissonGamma",
                       "PoissonGammaZIP", "PoissonGammaZIP1")
n_cov <- n
if (!bivariate && model %in% count_model_names) {
  count_sizes <- .GeoKrigCountSizes(
    model, n = n, nloc = nloc, nobs = length(data),
    npred = numloc * tloc, context = "GeoKrig"
  )
  ## GeoCovmatrix distinguishes a site-specific Binomial trial vector from
  ## common field-size parameters.  Keep the expanded observation/prediction
  ## vectors in count_sizes for Corr_c_bin, but pass only the representation
  ## accepted by the covariance constructor.
  if (identical(model, "Binomial")) {
    n_cov <- count_sizes$obs
  } else if (model %in% c("Binary", "Bernoulli", "Geom", "Geometric")) {
    n_cov <- 1L
  } else if (model %in% c("BinomialNeg", "BinomialNegZINB")) {
    ## .GeoKrigCountSizes() has already validated that n is one common
    ## positive integer r (and that nloc, if supplied, agrees with it).
    n_cov <- as.integer(round(n[1L]))
  }
}

if(ncol(loc) == 2L) {
  locx = loc[,1]; locy = loc[,2]; locz = double(nrow(loc))
} else {
  locx = loc[,1]; locy = loc[,2]; locz = loc[,3]
}



type="Standard"

######################################
############ standard kriging ########
######################################
if(type %in% c("Standard","standard")) {

################################################
######################################################## 
############ optimal prediction cases ##################   
######################################################## 
logGausstemp=SinhAsinhtemp=Tukeyh2temp=Tukeyhtemp=FALSE 

######################################################## 
######################################################## 
## Remove an external observation mean before covariance construction.
## The covariance code must never interpret its entries as mean1, mean2, ... .
if (!is.null(Mobs) && !copula_cont) param$mean <- 0
Mtemp <- Mobs
   if(model %in% c("Weibull","Gamma","LogLogistic","LogGaussian") && !copula_cont)          # non-copula multiplicative kriging uses a standardized covariance matrix
{
     paramtemp=param; sel=substr(names(param),1,4)=="mean";  ## selecting mean values
     meantemp=names(param[sel])             ## saving mean parameters
     param=param[!sel];                     ## not mean parameters
     param$mean=0;        ## mean must be =0 when calling covariance matrix
     Xtemp=X;X=NULL                         ## saving X and setting X=NULL
}
#####################################################
##### computing covariance matrix  ##################
#####################################################

covmatrix = GeoCovmatrix(coordx=coordx, coordy=coordy,coordz=coordz, coordt=coordt, coordx_dyn=coordx_dyn,
            corrmodel=corrmodel, distance= distance,grid=grid,model=model,n=n_cov,
            param=param, anisopars=anisopars, radius=radius,sparse=sparse,copula=copula,X=X)

#####################################################
## Normalize legacy/no-copula GeoCovmatrix objects.  Some fitted objects may
## store an absent copula as NULL, NA, or an empty string; downstream kriging
## logic must treat all of these representations identically.
if (!is.null(covmatrix$copula)) {
  cov_copula <- as.character(covmatrix$copula)[1L]
  if (is.na(cov_copula) || !nzchar(gsub("[[:blank:]]", "", cov_copula)))
    covmatrix$copula <- NULL
}
covmatrix$param=unlist(covmatrix$param)
if(bivariate) tloc=1
spacetime_dyn=FALSE; if(!is.null(covmatrix$coordx_dyn)) spacetime_dyn=TRUE
########dimat is dimension  ######
if(!spacetime_dyn) dimat=covmatrix$numcoord*covmatrix$numtime
if(spacetime_dyn)  dimat =sum(covmatrix$ns)
dimat2=numloc*tloc

## GeoCovmatrix needs the external observation locations for Gaussian-copula
## covariance construction, but downstream parameter extraction expects a
## scalar regression coefficient.  MM retains the external mean vector.
if (copula_cont && !is.null(Mobs)) param$mean <- 0

MM <- Mtemp
if(!is.null(MM)) param$mean <- 0
###############
if(model %in% c("Weibull","Gamma","LogLogistic","LogGaussian") && !copula_cont) {
          if(is.null(Xtemp)) X=matrix(1, nrow=dimat, ncol=1L)
          else               X=Xtemp
          param=paramtemp
          if(!is.null(Mtemp)) param$mean=0
          covmatrix$namesnuis=unique(c(meantemp,covmatrix$namesnuis))   
}
else { if(!is.null(X)) X=covmatrix$X 
           else        X=matrix(1,nrow=dimat,ncol=1)

     }

###############
if (!bivariate) {
  X <- .GeoMean_design(X, n = dimat, name = "X")
  num_betas <- ncol(X)
  if (!is.null(MM) && length(MM) != dimat)
    stop("The fixed observation mean must have one value per observation.\n")
  if (!is.null(Mloc)) {
    if (length(Mloc) == 1L) Mloc <- rep(Mloc, dimat2)
    if (length(Mloc) != dimat2)
      stop("Mloc must have one value per prediction point/time combination.\n")
  }
  if (is.null(Mloc)) {
    Xloc <- .GeoMean_design(Xloc, n = dimat2, p = num_betas, name = "Xloc")
  } else {
    Xloc <- NULL
  }
} else {
  X <- .GeoBivariateDesign(
    if(is.null(X)) covmatrix$X else X,
    covmatrix$ns, context = "GeoKrig"
  )
  num_betas <- ncol(X)
  invisible(.GeoBivariateValidateMeanParameters(
    param, num_betas, context = "GeoKrig"
  ))
  if (!is.null(Mloc)) {
    Mloc <- as.numeric(Mloc)
    if (length(Mloc) == 1L) Mloc <- rep(Mloc, dimat2)
    if (length(Mloc) != dimat2 || any(!is.finite(Mloc)))
      stop("GeoKrig: Mloc must contain one finite value per prediction location.")
    Xloc <- NULL
  } else {
    Xloc <- .GeoMean_design(Xloc, n = dimat2, p = num_betas, name = "Xloc")
  }
}
XX <- X
NS=0
if(spacetime||bivariate) { NS=cumsum(covmatrix$ns); NS=c(0,NS)[-(length(covmatrix$ns)+1)] }
###########
nuisance = if (bivariate) covmatrix$param[covmatrix$namesnuis] else param[covmatrix$namesnuis]; nuisance=Filter(Negate(is.null),nuisance)
## n is an argument rather than a named nuisance parameter for count models,
## but the generic copula moment/quantile engines need the common size value.
if (!isTRUE(bivariate) && !is.null(copula) &&
    copula %in% c("Clayton", "SkewGaussian") &&
    as.integer(CkModel(model)) %in% c(11L, 16L)) {
  nn <- if (as.integer(CkModel(model)) == 11L && !is.null(count_sizes))
    unique(as.numeric(c(count_sizes$obs, count_sizes$pred))) else as.numeric(n_cov)[1L]
  if (length(nn) != 1L || !is.finite(nn))
    stop("GeoKrig: copula-based Binomial prediction currently requires one common n across observations and prediction locations.", call. = FALSE)
  nuisance["n"] <- nn[1L]
}
if(!bivariate) {
  sel <- names(nuisance) %in% .GeoMean_names(num_betas)
  betas <- .GeoMean_beta(nuisance, num_betas)
}
if(bivariate) {
  sel <- substr(names(nuisance),1,4)=="mean"
  beta_biv <- .GeoBivariateValidateMeanParameters(
    nuisance, num_betas, context = "GeoKrig"
  )
  betas1 <- beta_biv$beta1
  betas2 <- beta_biv$beta2
}
## Count cross-covariance kernels use the same alphabetically ordered
## nuisance vector as GeoCovmatrix().  Keeping one ordering is essential for
## models such as PoissonGamma/PoissonGammaZIP where `shape` is not adjacent
## to `nugget` in NuisParam().
other_nuis_named <- nuisance[!sel]
other_nuis <- as.numeric(other_nuis_named[order(names(other_nuis_named))])

## Marginal location predictors used by copula covariance and
## prediction.  These are eta = X beta (or externally supplied values), not
## necessarily the marginal means on the observed scale.
eta_obs_cop <- eta_loc_cop <- NULL
if (copula_cont) {
  eta_obs_cop <- if (is.null(MM)) as.numeric(X %*% betas) else as.numeric(MM)
  eta_loc_cop <- if (!is.null(Mloc)) as.numeric(Mloc) else as.numeric(Xloc %*% betas)
  if (length(eta_obs_cop) != dimat || length(eta_loc_cop) != dimat2 ||
      any(!is.finite(c(eta_obs_cop, eta_loc_cop))))
    stop("GeoKrig: inconsistent copula marginal locations.", call. = FALSE)
}

############
cmodel=corrmodel
cdistance=distance
corrmodel = CkCorrModel(covmatrix$corrmodel)
distance = CheckDistance(covmatrix$distance)
corrparam = covmatrix$param[covmatrix$namescorr]# selecting the correlation parameters
if(bivariate) {if(!(which==1 || which==2)) {stop("which  parameter must be 1 or 2")}}
pred = NULL
varpred=vv=vv2=NULL
k = 0

###### observed coordinates in covariance ordering
ccc <- .GeoKrigObservedCoordinates(covmatrix, dimat, context = "GeoKrig")

###### spatial anisotropy case
## Apply anisotropy only after the final prediction-coordinate matrix
## has been constructed for spatial, grid, space-time, and dynamic inputs.
if(!is.null(anisopars)) {
    ccc=GeoAniso(ccc,c(anisopars$angle,anisopars$ratio))
}
###############################################################

if(bivariate) dataT <- .GeoBivariateDataVector(data, covmatrix$ns, context = "GeoKrig")
else if(spacetime && spacetime_dyn) dataT <- as.numeric(unlist(data, use.names = FALSE))
else dataT <- as.numeric(data)

####################################################################
############### blocked global kriging ###############################
####################################################################
## The high-volume observed-scale copula path and the ordinary Gaussian
## baseline are evaluated in prediction-location blocks.  The observed
## covariance matrix is factorized once and the factorization is reused for
## every block.  The full weights matrix is still retained because it is part
## of the historical GeoKrig return object; therefore blocking reduces peak
## cross-covariance memory without changing the returned object.
blocked_continuous <- !bivariate && (copula_cont ||
  (is.null(covmatrix$copula) && as.integer(covmatrix$model) == 1L))

if (blocked_continuous) {
  if (copula_cont) {
    moments_obs_cop <- .GeoCopulaMarginalMoments(
      covmatrix$model, nuisance, eta_obs_cop
    )
    moments_loc_cop <- .GeoCopulaMarginalMoments(
      covmatrix$model, nuisance, eta_loc_cop
    )
    mu <- as.numeric(moments_obs_cop$mean)
    muloc <- as.numeric(moments_loc_cop$mean)
    vvar_all <- as.numeric(moments_loc_cop$variance)
    eta_range_cop <- range(c(eta_obs_cop, eta_loc_cop), finite = TRUE)
  } else {
    mu <- if (is.null(MM)) as.numeric(X %*% betas) else as.numeric(MM)
    muloc <- if (!is.null(Mloc)) as.numeric(Mloc) else as.numeric(Xloc %*% betas)
    vvar_all <- rep(as.numeric(covmatrix$param["sill"]), dimat2)
    eta_range_cop <- NULL
  }

  if (length(mu) != dimat || length(muloc) != dimat2 ||
      length(vvar_all) != dimat2 ||
      any(!is.finite(c(mu, muloc, vvar_all)))) {
    stop("GeoKrig: inconsistent blocked marginal moments.", call. = FALSE)
  }

  if (identical(type_krig, "Universal")) {
    if (copula_cont) {
      d_obs <- .GeoCopulaMarginalMeanDerivative(
        covmatrix$model, nuisance, eta_obs_cop
      )
      d_loc <- .GeoCopulaMarginalMeanDerivative(
        covmatrix$model, nuisance, eta_loc_cop
      )
      X_system <- X * d_obs
      Xloc_system <- Xloc * d_loc
    } else {
      X_system <- X
      Xloc_system <- Xloc
    }
    mean_names_system <- .GeoMean_names(ncol(X_system))
  } else {
    X_system <- Xloc_system <- mean_names_system <- NULL
  }

  krig_factor <- .GeoKrigFactor(
    covmatrix, method = method, context = "GeoKrig"
  )
  krig_blocks <- .GeoKrigLocationBlocks(
    dimat, numloc, tloc, context = "GeoKrig"
  )
  n_krig_blocks <- length(krig_blocks)
  if (isTRUE(progress) && n_krig_blocks > 1L) {
    restore_progress_handlers <- .GeoProgressHandlersPush(
      handler = "txtprogressbar", global = TRUE
    )
    on.exit(restore_progress_handlers(), add = TRUE)
    krig_pb <- progressr::progressor(
      steps = n_krig_blocks, on_exit = FALSE, auto_finish = FALSE
    )
    ## Finish before restoring the temporary global handler, including on
    ## errors or user interrupts.
    on.exit(krig_pb(type = "finish"), add = TRUE, after = FALSE)
  } else {
    krig_pb <- function(...) {}
  }

  ## Historical API: GeoKrig returns all kriging weights.  Keep that result
  ## exactly, but never materialize the complete cross-covariance matrix.
  krig_weights <- matrix(NA_real_, nrow = dimat, ncol = dimat2)
  pp <- numeric(dimat2)
  vv <- if (isTRUE(mse)) numeric(dimat2) else NULL
  datas <- as.matrix(as.numeric(dataT) - as.numeric(mu))
  nugget_scale <- 1 - as.numeric(covmatrix$param["nugget"])
  if (!is.finite(nugget_scale))
    stop("GeoKrig: invalid nugget parameter.", call. = FALSE)

  loc_matrix <- cbind(locx, locy, locz)
  block_id <- 0L
  for (loc_idx in krig_blocks) {
    block_id <- block_id + 1L
    first_task <- (loc_idx[1L] - 1L) * tloc + 1L
    last_task <- loc_idx[length(loc_idx)] * tloc
    task_idx <- seq.int(first_task, last_task)
    ntarget_block <- length(task_idx)

    rho_block <- .GeoKrigCorrelationBlock(
      covmatrix = covmatrix, observed_coords = ccc,
      corrmodel = corrmodel, corrparam = corrparam, distance = distance,
      loc_block = loc_matrix[loc_idx, , drop = FALSE],
      time = time, tloc = tloc, NS = NS, which = which,
      context = "GeoKrig"
    )
    latent_block <- nugget_scale * rho_block

    if (copula_cont) {
      eta_obs_pair <- rep(eta_obs_cop, times = ntarget_block)
      eta_loc_pair <- rep(eta_loc_cop[task_idx], each = dimat)
      CC_block <- .GeoCopulaCrossCovFast(
        covmatrix$copula, latent_block, covmatrix$model, nuisance,
        eta1 = eta_obs_pair, eta2 = eta_loc_pair,
        eta_range = eta_range_cop
      )
      dim(CC_block) <- c(dimat, ntarget_block)
    } else {
      CC_block <- matrix(
        latent_block * as.numeric(covmatrix$param["sill"]),
        nrow = dimat, ncol = ntarget_block
      )
    }

    Xloc_block <- if (identical(type_krig, "Universal")) {
      Xloc_system[task_idx, , drop = FALSE]
    } else NULL

    krig_system_block <- .GeoKrigSystem(
      covmatrix = covmatrix, CC = CC_block, method = method,
      type_krig = type_krig, X = X_system, Xloc = Xloc_block,
      target_variance = vvar_all[task_idx], mse = mse, varcov = varcov,
      mean_names = mean_names_system, factor = krig_factor,
      context = "GeoKrig"
    )
    wb <- krig_system_block$weights
    krig_weights[, task_idx] <- wb
    pp[task_idx] <- muloc[task_idx] + as.numeric(crossprod(datas, wb))
    if (isTRUE(mse)) vv[task_idx] <- krig_system_block$mse

    rm(rho_block, latent_block, CC_block, wb, krig_system_block)
    if (copula_cont) rm(eta_obs_pair, eta_loc_pair)
    krig_pb(sprintf("block %d/%d", block_id, n_krig_blocks))
  }

  cvv <- c(vv)
  if (mse) {
    if (anyNA(cvv)) cvv[is.na(cvv)] <- 1e-30
    selp <- cvv <= 0
    if (any(selp)) cvv[selp] <- 1e-30
  }
  if (spacetime) {
    pred <- matrix(t(pp), nrow = tloc, ncol = numloc)
    varpred <- if (isTRUE(mse)) matrix(cvv, nrow = tloc, ncol = numloc) else NULL
  } else {
    pred <- c(pp)
    varpred <- if (isTRUE(mse)) cvv else NULL
  }
}

####################################################################
############### computing  correlation vector ######################
####################################################################
if(!blocked_continuous && covmatrix$model %in% c(1,10,59,18,21,12,26,24,27,38,29,39,28,9,34,40,20,22,25,33,42,50))    
{
## gaussian=1
## skew gaussian=10
## skew Laplace=59
## student =12
## skewstudent =18
## gamma=21
## weibull=26
## loglogistic=24
## loggaussian=22
## twopieceStudentT=27
## beta=28
## twopieceGaussian=29
## twopieceTukeyh=38
## twopiecebimodal=39
## tukeygh=9
## tukey=34
## tukeyhh=40
## sihasin=20

if((type=="Standard"||type=="standard")) {

cc<-dotCall64::.C64(
"Corr_c",
SIGNATURE=c(
"double","double","double","double","double","integer",
"integer","double","double","double","integer","integer",
"integer","integer","integer","integer","double","integer",
"integer","double","integer","integer","double"
),
 corri = dotCall64::vector_dc("double",dimat * dimat2),
ccc[,1],ccc[,2],ccc[,3],covmatrix$coordt,corrmodel, 0,
locx,locy,locz,covmatrix$numcoord,numloc,tloc,
covmatrix$ns,NS,covmatrix$numtime,corrparam,
covmatrix$spacetime,covmatrix$bivariate,time,distance,which-1,covmatrix$radius,
INTENT=c("w",rep("r",22)),NAOK=TRUE,PACKAGE="GeoModels"
)

####  transforming gaussian correlations depending on the (non)Gaussian  model
if(bivariate){  corri=cc$corri  }

else {
        rho=cc$corri; cc=(1-as.numeric(covmatrix$param["nugget"]))*rho

if (copula_cont) {
  eta_obs_pair <- rep(eta_obs_cop, times = dimat2)
  eta_loc_pair <- rep(eta_loc_cop, each = dimat)
  corri <- .GeoCopulaCrossCovFast(
    covmatrix$copula, cc, covmatrix$model, nuisance,
    eta1 = eta_obs_pair, eta2 = eta_loc_pair
  )
  moments_obs_cop <- .GeoCopulaMarginalMoments(
    covmatrix$model, nuisance, eta_obs_cop
  )
  moments_loc_cop <- .GeoCopulaMarginalMoments(
    covmatrix$model, nuisance, eta_loc_cop
  )
} else {

################ Gaussian ##################################
if(covmatrix$model==1)   {  vv=as.numeric(covmatrix$param['sill']); 
          if(is.null(covmatrix$copula)) corri=cc 
          else {  if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
} 
#################### skew gaussian #########################################
if(covmatrix$model==10) {    
vv=as.numeric(covmatrix$param['sill'])
if(is.null(covmatrix$copula)){
    corr2=rho^2; sk=as.numeric(covmatrix$param['skew']);sk2=sk^2; 
    corri=(2*sk2)*(sqrt(1-corr2) + rho*asin(rho)-1)/(pi*vv+sk2*(pi-2)) + (cc*vv)/(vv+sk2*(1-2/pi))
                        }
    else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
}
if(covmatrix$model==59) {    
  vv = as.numeric(covmatrix$param['sill'])
  if(is.null(covmatrix$copula)){
    sk = as.numeric(covmatrix$param['skew'])
    corri = cc^2
  }
  else {
    if(covmatrix$copula=="Gaussian")
      corri = gaussian_copula_cov(cc,covmatrix$model, nuisance)
  }
}
##################### student T ########################################
if(covmatrix$model==12) 
{
vv=as.numeric(covmatrix$param['sill']); nu=1/as.numeric(covmatrix$param['df'])
if(is.null(covmatrix$copula)){
   cc2 <- cc^2;gamma_num <- gamma((nu - 1)/2)^2;gamma_den <- gamma(nu/2)^2;Fvals <- hypergeo::hypergeo(0.5, 0.5, nu/2, cc2)
   corri <- ((nu - 2) * gamma_num * Re(Fvals) * cc) / (2 * gamma_den)
}
else {
      if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)
     }
}
########################## tukeyh ###################################
if(covmatrix$model==34) 
{   vv=as.numeric(covmatrix$param['sill']); h=as.numeric(covmatrix$param['tail'])
    if(h>0){ 
            if(is.null(covmatrix$copula)){ corri=(cc*(1-2*h)^(1.5))/((1-h)^2-(h*cc)^2)^(1.5) }
            else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
            }
    else{ if(is.null(covmatrix$copula))  corri=cc
          else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
        }
}
######################### tukeyh2###################################
if(covmatrix$model==40)
{ vv=as.numeric(covmatrix$param['sill']);
  tail1=as.numeric(covmatrix$param['tail1']);tail2=as.numeric(covmatrix$param['tail2'])
  if(is.null(covmatrix$copula)){
    mom_hh <- .GeoTukeyh2StandardMoments(tail1, tail2)
    mm <- mom_hh$mean
    vv1 <- mom_hh$variance
    corri <- .GeoTukeyh2Correlation(cc, tail1, tail2)
  }
  else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
}
####################### sas ######################################
if (covmatrix$model == 20) {
  vv <- as.numeric(covmatrix$param['sill'])
  tail <- as.numeric(covmatrix$param['tail'])
  skew <- as.numeric(covmatrix$param['skew'])
  if(is.null(covmatrix$copula)){
    mom_sas <- .GeoSASStandardMoments(skew, tail)
    mm <- mom_sas$mean
    vv1 <- mom_sas$variance
    corri <- corrsas(cc, skew, tail)
  }
  else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
}
######################## skew student T#####################################
if(covmatrix$model==18) 
{
                  vv=as.numeric(covmatrix$param['sill']); nu=1/as.numeric(covmatrix$param['df']);sk=as.numeric(covmatrix$param['skew'])
            if(is.null(covmatrix$copula)){ 
                  sk2=sk*sk;l=nu/2; f=(nu-1)/2; w=sqrt(1-sk2);y=rho;
                  CorSkew=(2*sk2/(pi*w*w+sk2*(pi-2)))*(sqrt(1-y*y)+y*asin(y)-1)+w*w*cc/(w*w+sk2*(1-2/pi)) ;
                  corri=(pi*(nu-2)*gamma(f)^2/(2*(pi*gamma(l)^2-sk2*(nu-2)*gamma(f)^2)))*(Re(hypergeo::hypergeo(0.5,0.5,l,y*y))*((1-2*sk2/pi)*CorSkew+2*sk2/pi)-2*sk2/pi);
              }
                else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
}
######################## # two piece Student T#####################################
         if(covmatrix$model==27) {  
             nu=as.numeric(1/covmatrix$param['df']);
             sk=as.numeric(covmatrix$param['skew']);
             vv=as.numeric(covmatrix$param['sill'])
            if(is.null(covmatrix$copula)){ 
             corr2=cc^2;sk2=sk^2
             a1=Re(hypergeo::hypergeo(0.5,0.5,nu/2,corr2))
             a2=cc*asin(cc) + (1-corr2)^(0.5)
             ll=qnorm((1-sk)/2)
             p11=pbivnorm::pbivnorm(ll,ll, rho = rho, recycle = TRUE)
             a3=3*sk2 + 2*sk + 4*p11 - 1
             KK=( nu*(nu-2)*gamma((nu-1)/2)^2) / (nu*pi*gamma(nu/2)^2*(3*sk2+1)-4*sk2*nu*(nu-2)*gamma((nu-1)/2)^2 )
             corri= KK*(a1*a2*a3-4*sk2);
         }
          else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
                      }
########################### two piece Tukey h ##################################
         if(covmatrix$model==38) {  
                        tail=as.numeric(covmatrix$param['tail']);sk=as.numeric(covmatrix$param['skew']);vv=as.numeric(covmatrix$param['sill'])
 if(is.null(covmatrix$copula)){ 
                        corr2=cc^2;sk2=sk^2;
                        gg2=(1-(1-corr2)*tail)^2
                        xx=corr2/gg2
                        A=(asin(sqrt(xx))*sqrt(xx)+sqrt(1-xx))/(1-xx)^(1.5)
                        ll=qnorm((1-sk)/2)
                        p11=pbivnorm::pbivnorm(ll,ll, rho = rho, recycle = TRUE)
                        a3=3*sk2 + 2*sk + 4*p11 - 1
                        mm=8*sk2/(pi*(1-tail)^2);
                        ff=(1+3*sk2)/(1-2*tail)^(1.5)
                        M=(2*(1-corr2)^(3/2))/(pi*gg2)
                        corri=  (M*A*a3-mm)/( ff- mm)
                    }
                     else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
                      }
############################################################
         if(covmatrix$model==39) {  # bimodal
                                  nu=as.numeric(nuisance['df']); sk=as.numeric(nuisance['skew'])
                                  vv=as.numeric(covmatrix$param['sill']); delta=as.numeric(nuisance['shape'])
                                  alpha=2*(delta+1)/nu; nn=2^(1-alpha/2)
                                  ll=qnorm((1-sk)/2)
                                  p11=pbivnorm::pbivnorm(ll,ll, rho = rho, recycle = TRUE)
                                  corr2=cc^2;sk2=sk^2
                                  a1=Re(hypergeo::hypergeo(-1/alpha ,-1/alpha,nu/2,corr2))
                                  a3=3*sk2 + 2*sk + 4*p11 - 1
                                  MM=(2^(2/alpha)*(gamma(nu/2 + 1/alpha))^2)
                                  vari=2^(2/alpha)*(gamma(nu/2 + 2/alpha))*gamma(nu/2)* (1+3*sk2) - sk2*2^(2/alpha+2)*gamma(nu/2+1/alpha)^2
                                  corri= MM*(a1*a3-4*sk2)/vari
                      }
############################################################
        if(covmatrix$model==29) 
{  # two piece Gaussian
             if(is.null(covmatrix$copula)){ 
                          corr2=sqrt(1-cc^2)
                          vv=as.numeric(covmatrix$param['sill'])
                          sk=as.numeric(nuisance['skew']); sk2=sk^2
                          ll=qnorm((1-sk)/2)
                          p11=pbivnorm::pbivnorm(ll,ll, rho = rho, recycle = TRUE)
                          KK=3*sk2+2*sk+ 4*p11 - 1
                          corri=(2*((corr2 + cc*asin(cc))*KK)- 8*sk2)/(3*pi*sk2  -  8*sk2   +pi   )
                      }
                      else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
}
############################################################
       if(covmatrix$model==28)
{

         shape1=as.numeric(covmatrix$param['shape1']);shape2=as.numeric(covmatrix$param['shape2']);

             if(is.null(covmatrix$copula)){ 
          corr2=cc^2       
         cc1=0.5*(shape1+shape2); vv=shape1*shape2/((cc1+1)*(shape1+shape2)^2)
         idx=which(abs(corr2)>1e-10);corr22=corr2[idx]; nu2=shape1/2;alpha2=shape2/2
         A <- 0; res <- 0; ss <- 0; kk <- 0
         while(kk<=100){
             p1=2*(lgamma(cc1+kk)-lgamma(cc1)+lgamma(nu2+1+kk)-lgamma(nu2+1));p2=lgamma(kk+1)+(lgamma(nu2+kk)-lgamma(nu2))+2*(lgamma(cc1+1+kk)-lgamma(cc1+1))
             b1=p1-p2; b2=log(hypergeo::genhypergeo(U=c(cc1+kk,cc1+kk,alpha2), L=c(cc1+kk+1,cc1+kk+1), polynomial=TRUE,maxiter=1000, z=corr22)); b3=kk*log(corr22)
             sum=exp(b1+b2+b3); res=res+sum
             if (all(sum<1e-6)){ break} else{ A=res}
         kk=kk+1
        }
         cc[idx]=A; corri=shape1*(cc1 + 1 ) * ((1-corr2)^(cc1) *cc -1)/shape2
         corri[-idx]=0
    }
       else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
}
############################################################    

if(covmatrix$model==21) 
{ # gamma
                        sh=as.numeric(covmatrix$param["shape"])
                         if(is.null(covmatrix$copula)){  corri=cc^2 }
                         else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
}
if(covmatrix$model==26) 
{  # weibull
                        sh=as.numeric(covmatrix$param['shape']);
                        if(is.null(covmatrix$copula)){ 
                        corri <- .GeoWeibullCorr(cc, sh)
                         }
                        else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}           
}
if(covmatrix$model==24) 
{  # loglogistic
                        sh=as.numeric(covmatrix$param['shape'])
                             if(is.null(covmatrix$copula)){ 
                            cc2 <- cc^2 ; num <- pi * sin(2 * pi / sh);den <- 2 * sh * (sin(pi / sh))^2 - pi * sin(2 * pi / sh);factor <- num / den
                            F1 <- hypergeo::hypergeo(-1/sh, -1/sh, 1, cc2)
                            F2 <- hypergeo::hypergeo( 1/sh,  1/sh, 1, cc2)
                            corri <- factor * (Re(F1) * Re(F2) - 1)
                        }
                        else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
}
if(covmatrix$model==22)
 {  # loggaussian
                        ss=as.numeric(covmatrix$param['sill']);
                               if(is.null(covmatrix$copula)){  corri= (exp(ss*cc)-1)/(exp(ss)-1)}
                                     else { if(covmatrix$copula=="Gaussian") corri= gaussian_copula_cov(cc,covmatrix$model, nuisance)}
}
############################################################
} ## end legacy/non-copula marginal-correlation transformations
 }        
############################################################
if(bivariate)  {          if(which==1)    vvar=covmatrix$param["sill_1"]+covmatrix$param["nugget_1"]
                          if(which==2)    vvar=covmatrix$param["sill_2"]+covmatrix$param["nugget_2"]
               }
else    {
    ##### computing target marginal variance
    if (copula_cont) {
      vvar <- as.numeric(moments_loc_cop$variance)
      M <- 0
    } else {
    ##### legacy/non-copula mean and variances for each model
     if(covmatrix$model==1)   {vvar= vv;                M=0 }
     if(covmatrix$model==10)  {vvar= vv+sk^2*(1-2/pi); M=sk*sqrt(2/pi)}
     if(covmatrix$model==59)  {vvar = vv * (1 - 2*sk + 2*sk^2) / (sk^2 * (1-sk)^2)
                               M= sqrt(vv) * (1/sk - 1/(1-sk))
                              }
     if(covmatrix$model==12)  {vvar= vv*nu/(nu-2) ;     M=0 }
     if(covmatrix$model==34) {vvar= vv*(1-2*h)^(-1.5); M=0 }
     if(covmatrix$model==40) { vvar= vv* vv1 ; M=sqrt(vv)*mm}
     if(covmatrix$model==20) { vvar= vv* vv1; M=sqrt(vv)*mm}
   ##############################################################
     if(covmatrix$model==18)  {
                               D1=(nu-1)*0.5; D2=nu*0.5;
                               mm=sqrt(nu)*gamma(D1)*sk/(sqrt(pi)*gamma(D2));
                               vvar=vv*(nu/(nu-2)-mm^2); M=sqrt(vv)*mm
                              }
     if(covmatrix$model==27)  {
                              ttemp=gamma(0.5*(nu-1))/gamma(0.5*nu)
                              vvar= vv*((nu/(nu-2))*(1+3*sk2) - 4*sk2*(nu/pi)*ttemp^2); M= - sqrt(vv)*2*sk*sqrt(nu/pi)*ttemp
                              }
     if(covmatrix$model==28)  {
                              ssup=as.numeric((covmatrix$param["max"]-covmatrix$param["min"]))
                              vvar=(ssup^2)*(shape1*shape2/((cc1+1)*(shape1+shape2)^2))
                              M=ssup*shape1/(shape1+shape2)+as.numeric(covmatrix$param["min"])
                              muloc=0;mu=0
                              }
     if(covmatrix$model==29)  {vvar= vv*((1+3*sk2) - 8*sk2/pi ) ;M=-sqrt(vv)*2*sk*sqrt(2/pi)}
     if(covmatrix$model==38)  {vvar= vv*(ff - mm);M= -sqrt(vv)*2*sk*sqrt(2/pi)/(1-tail)}
     if(covmatrix$model==39)  {vvar= vv*vari/(nn^(2/alpha)*gamma(nu/2)^2);M=-sqrt(vv)*sk*2^(1/alpha+1)*gamma(nu/2+1/alpha)/(nn^(1/alpha)*gamma(nu*0.5)) }
     if(covmatrix$model==21)  {vvar= 2/sh }
     if(covmatrix$model==24)  {vvar= 2*sh*sin(pi/sh)^2/(pi*sin(2*pi/sh))-1 }
     if(covmatrix$model==26)  {vvar= gamma(1+2/sh)/gamma(1+1/sh)^2-1   }
     if(covmatrix$model==22)  {vvar= exp(ss)-1   }

     if(SinhAsinhtemp||Tukeyh2temp||Tukeyhtemp||logGausstemp) vvar=1
    } ## end legacy target variance

 }
###############################################################  

################################################################
############setting mean for the external mean case ################
################################################################
   if(!bivariate){
    if (copula_cont) {
      mu <- as.numeric(moments_obs_cop$mean)
      muloc <- as.numeric(moments_loc_cop$mean)
    } else {
      mu <- if (is.null(MM)) as.numeric(X %*% betas) else as.numeric(MM)
      muloc <- if (!is.null(Mloc)) as.numeric(Mloc) else as.numeric(Xloc %*% betas)
    }
                        }          
    else {
      i1 <- seq_len(covmatrix$ns[1L])
      i2 <- covmatrix$ns[1L] + seq_len(covmatrix$ns[2L])
      X11 <- X[i1, , drop = FALSE]
      X22 <- X[i2, , drop = FALSE]
      mu <- as.numeric(c(X11 %*% betas1, X22 %*% betas2))
      muloc <- .GeoBivariatePredictionMean(
        nuisance, Xloc, dimat2, which, num_betas, Mloc,
        context = "GeoKrig"
      )
    }
#### updating mean
        if(!bivariate && !copula_cont){
          #### additive model on the real line
         if(covmatrix$model %in% c(10,18,29,27,38,39,28, 34,40,20,59))
                                { muloc=muloc + M;   mu=mu + M }     
          ### multiplicative model on the positive real line
          if( (covmatrix$model %in% c(21,26,24,22)))  {emuloc=exp(muloc);emu=exp(mu) }
         }
         else{}

##################################################################
##########computing kriging weights##################################
##################################################################
if(bivariate) {
  ## Corr_c calls the bivariate covariance model directly; its output already
  ## contains sill, nugget and cross-covariance scaling.
  CC <- matrix(corri, nrow = dimat, ncol = dimat2)
} else if(is.null(covmatrix$copula)) {
  CC <- matrix(corri*vvar, nrow=dimat, ncol=dimat2)
} else if(covmatrix$copula %in% c("Gaussian", "Clayton", "SkewGaussian")) {
  CC <- matrix(corri, nrow=dimat, ncol=dimat2)
}
if(identical(type_krig, "Universal")) {
  if(bivariate) {
    X_system <- .GeoBivariateBlockDesign(X, covmatrix$ns)
    Xloc_system <- .GeoBivariateTargetBlockDesign(
      Xloc, nloc = dimat2, p = num_betas, which = as.integer(which)
    )
    mean_names_system <- colnames(X_system)
  } else {
    if (copula_cont) {
      d_obs <- .GeoCopulaMarginalMeanDerivative(covmatrix$model, nuisance, eta_obs_cop)
      d_loc <- .GeoCopulaMarginalMeanDerivative(covmatrix$model, nuisance, eta_loc_cop)
      X_system <- X * d_obs
      Xloc_system <- Xloc * d_loc
    } else {
      X_system <- X
      Xloc_system <- Xloc
    }
    mean_names_system <- .GeoMean_names(ncol(X_system))
  }
} else {
  X_system <- Xloc_system <- mean_names_system <- NULL
}
krig_system <- .GeoKrigSystem(
  covmatrix = covmatrix, CC = CC, method = method,
  type_krig = type_krig, X = X_system, Xloc = Xloc_system,
  target_variance = vvar, mse = mse, varcov = varcov,
  mean_names = mean_names_system, context = "GeoKrig"
)
krig_weights = krig_system$weights
BB = NULL

##################################################################
################# simple or universal kriging #################################
##################################################################
if(type_krig %in% c("Simple","Universal"))  {

if(!bivariate)
{  

if (copula_cont) {
  datas <- as.matrix(as.numeric(dataT) - as.numeric(mu))
  pp <- as.numeric(muloc) + as.numeric(crossprod(datas, krig_weights))
} else {

if(type_krig %in% c("Simple","Universal")) {
               ############################ optimal linear predictors #######################
               if(covmatrix$model %in% c(1,12,27,38,29,10,18,39,37,28,9, 34,40,20,59))
                {                        
                      datas=as.matrix(c(dataT)-c(mu))
                      pp = c(muloc)      +  crossprod(datas,krig_weights)
                }
             }
               ###################################################
               #### gamma weibull loglogistic loggaussian 
if(covmatrix$model %in% c(21,24,26,22))
                      {       
                 ones=rep(1,length(c(dataT))); one=rep(1,length(c(muloc)))    
                 datas=as.matrix(c(dataT)/emu-ones)
                 pp = c(emuloc) * ( one + crossprod(datas,krig_weights))  
                      }
} ## end legacy/non-copula predictor
}     
else {
  dat <- as.numeric(dataT) - as.numeric(mu)
  pp <- as.numeric(muloc) + as.numeric(crossprod(dat, krig_weights))
}
#### MSE COMPUTATION ##################

  if(mse) {
    if (copula_cont) {
      vv <- krig_system$mse
    } else {
      # Gaussian, StudentT, skew-Gaussian and two-piece linear kriging
      if(covmatrix$model %in% c(1,12,27,38,29,10,18,39,28,40,34,9,20,59))
        vv <- krig_system$mse

      # gamma, Weibull, log-logistic and log-Gaussian
      if(covmatrix$model %in% c(21,26,24,22))  {
        vv <- krig_system$mse
        vv <- emuloc^2 * vv
      }
    }
  }     


  }  

######################################################
####formatting data ##################################
######################################################
cvv=c(vv)

##### setting almost zero when zero mse
if(mse) { if(anyNA(cvv)) 
          {
              selna=is.na(cvv)
              cvv[selna]=1e-30
          }
          selp=(cvv<=0); 
          if(any(selp)) {cvv[selp]=1e-30}
    }
####   formatting data  
if(spacetime||bivariate) {
            pred=matrix(t(pp),nrow=tloc,ncol=numloc);
            varpred=if(isTRUE(mse)) matrix(cvv,nrow=tloc,ncol=numloc) else NULL;
            }
else {pred=c(pp);varpred=if(isTRUE(mse)) cvv else NULL}

}

} 

####################################################################################################################################
###################### binomial  negative binomial and poisson (inflated) #####################################
####################################################################################################################################
if(!copula_cont && covmatrix$model %in% c(2,11,14,19,30,36,16,43,44,45,46,47,57,58))
{

     if(type=="Standard"||type=="standard") {

      if(!bivariate){
               mu <- if (is.null(MM)) as.numeric(X %*% betas) else as.numeric(MM)
               mu0 <- if (!is.null(Mloc)) as.numeric(Mloc) else as.numeric(Xloc %*% betas)
      }
     if(bivariate)  {mu  = c(X11%*%betas1,X22%*%betas2)}
     ## Count-size vectors in observation and prediction ordering.
     n_obs_count <- if (is.null(count_sizes)) rep.int(1L, dimat) else count_sizes$obs
     n_pred_count <- if (is.null(count_sizes)) rep.int(1L, dimat2) else count_sizes$pred
     KK <- n_pred_count

## covariance between observed and prediction sites
if(covmatrix$model %in% c(2,11,14,16,19,30,36,43,44,45,46,47,57,58))
{ 
cop=0  
if(!is.null(covmatrix$copula)){ if(covmatrix$copula=="Gaussian") cop=1  }

ccorr = dotCall64::.C64(
  "Corr_c_bin",
  SIGNATURE = c(
    rep("double", 5),
    "integer", "integer", "double", "double", "double",
    rep("integer", 9),
    "double", "double",           # mean_obs, mean_loc
    "double", "double",           # other_nuis, corrparam
    "integer", "integer",         # spacetime, bivariate
    "double",                     # time
    "integer", "integer",         # distance, which
    "double",                     # radius
    "integer"                     # cop
  ),
  corri = dotCall64::vector_dc("double", dimat * dimat2),
  ccc[,1], ccc[,2], ccc[,3], covmatrix$coordt,
  corrmodel, as.integer(0),
  locx, locy, locz,
  covmatrix$numcoord, numloc,
  covmatrix$model, tloc,
  as.integer(KK), as.integer(n_obs_count), covmatrix$ns, NS, covmatrix$numtime,
  as.double(c(mu)),              # mean_obs
  as.double(c(mu0)),             # mean_loc
  as.double(other_nuis), as.double(corrparam),
  as.integer(covmatrix$spacetime), as.integer(bivariate),
  as.double(time), as.integer(distance), as.integer(which - 1),
  as.double(covmatrix$radius), as.integer(cop),
  INTENT = c("w", rep("r", 29)),
  PACKAGE = "GeoModels",
  VERBOSE = 0,
  NAOK = TRUE
)
}

if(cop==1) { 
nuisance$n=n_obs_count[1L]
corri= gaussian_copula_cov(ccorr$corri,covmatrix$model, nuisance)
     }
else {corri=ccorr$corri}

##################inverse of cov matrix##############################################
        if(!bivariate) cc = matrix(corri,nrow=dimat,ncol=dimat2)
        else           {}
        krig_weights = .GeoKrigSolve(covmatrix, cc, method=method,
                                         context="GeoKrig")
        BB = NULL
######################################################################################

       if(type_krig %in% c("Simple","Universal"))   {

      ##########################################################
       if(covmatrix$model==30||covmatrix$model==36){  ### poisson
        p0=exp(mu0); pmu=exp(mu)
            if(!bivariate) { 
            datas=c(dataT)-c(pmu)

            pp = c(p0) + crossprod(datas, krig_weights)
             }
            else{}
           if(mse)  vvar=p0
          }
      ##########################################################
      if(covmatrix$model==46||covmatrix$model==47){  ### poisson gamma
         p0=exp(mu0); pmu=exp(mu)
         if(!bivariate) { 
               datas=c(dataT)-c(pmu)
               pp = c(p0) + crossprod(datas, krig_weights)
          }
        else{}
        if(mse)  vvar=p0*(1+p0/covmatrix$param['shape'])
          }
      ##########################################################
       if(covmatrix$model==43||covmatrix$model==44){  ### poisson inflated
        p=pnorm(covmatrix$param['pmu'])
        p0=exp(mu0); pmu=exp(mu)
            if(!bivariate)  { 
             datas=c(dataT)-(1-p)*c(pmu)
             pp = (1-p)*c(p0) + crossprod(datas,krig_weights)
             }
            else{}
           if(mse)  vvar=(1-p)*p0*(1+p*p0)
        }
       ##########################################################
       if(covmatrix$model==2||covmatrix$model==11){  ### binomial
        p0=pnorm(mu0); pmu=pnorm(mu)
            if(!bivariate) { 
               datas=c(dataT)-n_obs_count*c(pmu)
               pp = n_pred_count*c(p0) + crossprod(datas,  krig_weights) 
            }
            else{}
           if(mse) vvar=KK*p0*(1-p0)
          }
      ###########################################################
           if(covmatrix$model==19){  ### binomial2
            p0=pnorm(mu0); pmu=pnorm(mu)
            if(!bivariate)   
              { datas=c(dataT)-c(n_obs_count*pmu)
                pp = n_pred_count*c(p0) + crossprod(datas,krig_weights)
              }
            else{}
          if(mse)    vvar=n_pred_count*p0*(1-p0)
          }
        ##########################################################
       if(covmatrix$model==14||covmatrix$model==16){    ###geometric or negative binomial
         p0=pnorm(mu0); pmu=pnorm(mu)
            if(!bivariate)
            {   k1=c(p0);k2=c(pmu);
                aa=n_pred_count*(1-k1)/k1
                datas=c(dataT)-n_obs_count*(1-k2)/k2
                pp = aa + crossprod(datas ,krig_weights)
          }
            else{}
            if(mse) vvar=aa/k1
          }
            ##########################################################
         if(covmatrix$model==57||covmatrix$model==58){  ### poisson gamma inflated
         p=pnorm(covmatrix$param['pmu'])
         p0=exp(mu0); pmu=exp(mu)
         if(!bivariate) { 
               datas=c(dataT)-(1-p)*c(pmu)
               pp = (1-p)*c(p0) + crossprod(datas, krig_weights)
          }
        else{}
        if(mse)  vvar=p0*(1+ (p0*(1+covmatrix$param['shape']*p)/((1-p)*covmatrix$param['shape'])) )
        }
        ##########################################################
        if(covmatrix$model==45){    ###inflated negative binomial
            p0=pnorm(mu0); pmu=pnorm(mu)
            p=as.numeric(pnorm(covmatrix$param['pmu']))
            if(!bivariate) { k1=c(p0);k2=c(pmu);
                             datas=c(dataT)-(1-p)*n_obs_count*(1-k2)/k2
                             pp = (1-p)*n_pred_count*(1-k1)/k1 + crossprod(datas,krig_weights)
                           }
            else{}
            if(mse) vvar=n_pred_count*(1-k1)*(1-p)*(1+n_pred_count*p*(1-k1))/k1^2
          }
        if(mse){
          ## Compute only the diagonal kriging variance; never allocate an
          ## Npred x Npred matrix.  Universal kriging additionally propagates
          ## uncertainty in the estimated regression coefficients.
          vv <- as.numeric(vvar) - colSums(cc * krig_weights)
          if (identical(type_krig, "Universal")) {
            d_obs <- .GeoKrigCountMeanDerivative(
              covmatrix$model, covmatrix$param, mu, n_obs_count
            )
            d_loc <- .GeoKrigCountMeanDerivative(
              covmatrix$model, covmatrix$param, mu0, n_pred_count
            )
            X_system_count <- X * d_obs
            Xloc_system_count <- Xloc * d_loc
            mean_names_count <- .GeoMean_names(ncol(X_system_count))
            var_mean_count <- .GeoKrigMeanVarcov(
              varcov, mean_names = mean_names_count, context = "GeoKrig"
            )
            trend_gap_count <- Xloc_system_count - t(krig_weights) %*% X_system_count
            vv <- vv + rowSums((trend_gap_count %*% var_mean_count) * trend_gap_count)
          }
        }
        }
##############################################################
if(!mse) vv <- NULL
cvv=c(vv)
##### setting almost zero when zero mse
if(mse) {
   if(anyNA(cvv)) 
     {selna=is.na(cvv)
     cvv[selna]=1e-30
    }
    selp=(cvv<=0); 
    if(any(selp)) {cvv[selp]=1e-30}
    }
####   formatting data  
if(spacetime||bivariate) {
            pred=matrix(t(pp),nrow=tloc,ncol=numloc);
            varpred=if(isTRUE(mse)) matrix(cvv,nrow=tloc,ncol=numloc) else NULL;
            }
          else {pred=c(pp);varpred=if(isTRUE(mse)) cvv else NULL}
    }}

if(tloc==1)  {pred=c(pred);varpred=c(varpred)}
if(!isTRUE(mse)) varpred <- NULL

    # Return the objects list:
   GeoKrig= list(     bivariate=bivariate,
                   coordx = covmatrix$coordx,coordy = covmatrix$coordy,coordz = covmatrix$coordz,
                   coordt = covmatrix$coordt,coordx_dyn=covmatrix$coordx_dyn,
                   covmatrix=covmatrix$covmatrix,corrmodel = corrmodel,copula=copula,data=data,distance = distance,
                   grid=covmatrix$grid,loc=loc_orig,nozero=covmatrix$nozero,ns=covmatrix$ns,X=XX,
                   numcoord = covmatrix$numcoord,numloc= numloc,numtime = covmatrix$numtime,numt = tloc,
                   model=model,fit_model=model_requested,param = param_input,pred=pred,n=n_input,nloc=nloc_input,
                   radius=radius,spacetime = covmatrix$spacetime,time=time,type=type,type_krig=type_krig,method=method,weigthed=weigthed,varcov=varcov,
                   weights=krig_weights,mse=varpred)
    structure(c(GeoKrig , call = call), class = c("GeoKrig"))
}

}

sp2Geo <- function(spobj, spdata = NULL)
{
  if (!requireNamespace("sp", quietly = TRUE)) {
    stop("Package 'sp' is required to use sp2Geo().", call. = FALSE)
  }

  # spobj: sp or spacetime object
  # spdata: name of dependent variable;
  # one of the columns in the @data slot must be the dependent variable

  X <- NULL
  Y <- NULL
  coordt <- NULL

  cls <- class(spobj)[1]

  if (!(cls %in% c("SpatialPoints", "SpatialPointsDataFrame", "STFDF", "STIDF")))
    stop("It is not a sp or st object ")

  if (cls == "SpatialPoints") {
    coords <- sp::coordinates(spobj)
    pj <- sp::is.projected(spobj)
    if (is.na(pj)) warning("coordinate reference system is not defined")
  }

  if (cls == "SpatialPointsDataFrame") {
    if (is.null(spdata)) stop("Dependent variable name must be declared")
    if (!(spdata %in% names(spobj@data))) stop("Dependent variable name is not in the data slot")

    coords <- sp::coordinates(spobj)
    X <- spobj@data[, names(spobj@data) != spdata, drop = FALSE]
    Y <- spobj@data[, spdata]
    pj <- sp::is.projected(spobj)
    if (is.na(pj)) warning("coordinate reference system is not defined")
  }

  if (cls == "STFDF") {
    if (is.null(spdata)) stop("Dependent variable name must be declared")
    if (!(spdata %in% names(spobj@data))) stop("Dependent variable name is not in the data slot")

    coords <- sp::coordinates(spobj@sp)
    pj <- sp::is.projected(spobj@sp)
    coordt <- seq_along(spobj@time)
    Y <- spobj@data[, spdata]
    X <- spobj@data[, names(spobj@data) != spdata, drop = FALSE]
    if (is.na(pj)) warning("coordinate reference system is not defined")
  }

  if (cls == "STIDF") {
    if (is.null(spdata)) stop("Dependent variable name must be declared")

    spobj <- as(spobj, "STFDF")

    if (!(spdata %in% names(spobj@data))) stop("Dependent variable name is not in the data slot")

    coords <- sp::coordinates(spobj@sp)
    pj <- sp::is.projected(spobj@sp)
    coordt <- seq_along(spobj@time)
    Y <- spobj@data[, spdata]
    X <- spobj@data[, names(spobj@data) != spdata, drop = FALSE]
    if (is.na(pj)) warning("coordinate reference system is not defined")
  }

  return(list(coords = coords,
              coordt = coordt,
              X = X,
              Y = Y,
              pj = pj,
              projected = pj))
}

####################################################
### File name: GeoFit2.R
####################################################

# GeoFit2 is a univariate starting-value wrapper around GeoFit.
#
# It first tries to estimate the marginal parameters under spatial
# independence and then calls the canonical GeoFit implementation with
# the updated starting values.  If an independence likelihood is not
# available for a model, the supplied starting values are used unchanged.
#

GeoFit2 <- function(data, coordx = NULL, coordy = NULL, coordz = NULL, coordt = NULL,
                    coordx_dyn = NULL, copula = NULL, corrmodel = NULL,
                    distance = "Eucl", fixed = NULL, anisopars = NULL,
                    est.aniso = c(FALSE, FALSE), grid = FALSE,
                    likelihood = "Marginal", lower = NULL, maxdist = Inf,
                    neighb = NULL, p_neighb = 1, maxtime = Inf,
                    memdist = TRUE, method = "cholesky", model = "Gaussian",
                    n = 1, onlyvar = FALSE, optimizer = "Nelder-Mead",
                    radius = 1, score = FALSE, sensitivity = FALSE,
                    sparse = FALSE, start = NULL,
                    thin_method = "bernoulli", type = "Pairwise",
                    upper = NULL, varest = FALSE, weighted = FALSE,
                    X = NULL, spobj = NULL, spdata = NULL,
                    independence_start = TRUE,
                    independence_optimizer = "Nelder-Mead",
                    warn_independence_failure = TRUE, check.duplicates = FALSE)
{
    call <- match.call()

    if (!isTRUE(memdist)) {
        warning(
            "'memdist = FALSE' is deprecated and ignored; GeoFit2 now always precomputes the selected pair structure.",
            call. = FALSE
        )
    }
    memdist <- TRUE

    start_was_null <- is.null(start)
    if (start_was_null) start <- list()

    # Collect only arguments accepted by GeoFit.  All values are forced
    # here, so do.call() does not depend on expressions in the caller.
    geofit_arg_names <- setdiff(
        names(formals(GeoFit2)),
        c(
            "independence_start",
            "independence_optimizer",
            "warn_independence_failure"
        )
    )

    fit_args <- mget(
        geofit_arg_names,
        envir = environment(),
        inherits = FALSE
    )

    start_original <- if (start_was_null) NULL else start
    start_used <- start
    independence_fit <- NULL
    independence_message <- NULL

    model_clean <- gsub("[[:blank:]]", "", model)
    type_clean <- gsub("[[:blank:]]", "", type)
    corrmodel_clean <- if (is.null(corrmodel)) NULL else
        gsub("[[:blank:]]", "", corrmodel)

    is_bivariate <- FALSE
    if (!is.null(corrmodel_clean)) {
        is_bivariate <- tryCatch(
            isTRUE(CheckBiv(CkCorrModel(corrmodel_clean))),
            error = function(e) FALSE
        )
    }

    if (is_bivariate) {
        stop(
            "GeoFit2 does not support bivariate correlation models. ",
            "Use GeoFit() directly for bivariate fitting.",
            call. = FALSE
        )
    }

    # Mean coefficients estimated by the preliminary independence fit are
    # useful also when the caller omitted them from start: GeoFit/WlsStart
    # will keep them in the estimated-parameter set.  The mapping follows the
    # columns of the user-supplied design matrix exactly; no intercept column
    # is added when X is supplied.
    mean_names <- if (is.null(X)) {
        "mean"
    } else if (is.list(X) && !is.data.frame(X)) {
        p <- unique(vapply(X, function(z) ncol(as.matrix(z)), integer(1L)))
        if (length(p) != 1L || is.na(p) || p < 1L) {
            stop(
                "All blocks of X must have the same positive number of columns.",
                call. = FALSE
            )
        }
        .GeoMean_names(p)
    } else {
        p <- ncol(as.matrix(X))
        if (is.null(p) || is.na(p) || p < 1L) {
            stop("X must contain at least one column.", call. = FALSE)
        }
        .GeoMean_names(p)
    }

    can_initialize <-
        isTRUE(independence_start) &&
        type_clean != "Independence"

    if (can_initialize) {

        ind_args <- fit_args

        # The preliminary fit estimates marginal parameters only.
        ind_args$likelihood <- "Marginal"
        ind_args$type <- "Independence"
        ind_args$optimizer <- independence_optimizer
        ind_args$onlyvar <- FALSE
        ind_args$score <- FALSE
        ind_args$sensitivity <- FALSE
        ind_args$varest <- FALSE
        ind_args$weighted <- FALSE
        ind_args$anisopars <- NULL
        ind_args$est.aniso <- c(FALSE, FALSE)
        ind_args$neighb <- NULL
        ind_args$maxdist <- Inf
        ind_args$maxtime <- Inf
        ind_args$p_neighb <- 1
        ind_args$thin_method <- "bernoulli"

        independence_fit <- tryCatch(
            do.call(GeoFit, ind_args),
            error = function(e) {
                independence_message <<- conditionMessage(e)
                NULL
            }
        )

        if (!is.null(independence_fit)) {
            estimated_marginal <- unlist(
                independence_fit$param,
                use.names = TRUE
            )

            start_names <- names(start_used)
            fixed_names <- names(unlist(fixed, use.names = TRUE))
            if (is.null(start_names)) start_names <- character(0L)
            if (is.null(fixed_names)) fixed_names <- character(0L)

            # Update marginal parameters already supplied in start and add
            # estimated regression coefficients that were intentionally
            # omitted from start.  Fixed parameters always have precedence.
            update_names <- if (start_was_null) {
                ## With start=NULL, use every eligible marginal estimate from
                ## the independence fit. Dependence parameters will keep the
                ## conservative automatic values prepared by GeoFit/WlsStart.
                names(estimated_marginal)
            } else {
                unique(c(
                    intersect(start_names, names(estimated_marginal)),
                    intersect(mean_names, names(estimated_marginal))
                ))
            }
            update_names <- setdiff(update_names, fixed_names)

            # In GeoModels, sill is the total marginal variance and nugget is
            # a correlation attenuation parameter.  Independence estimates
            # may initialize sill, but nugget must remain exactly as supplied
            # by the user and is never inferred from the marginal variance.
            update_names <- update_names[
                !grepl("^nugget($|_)", update_names)
            ]

            finite_update <- vapply(
                estimated_marginal[update_names],
                function(z) length(z) == 1L && is.finite(z),
                logical(1L)
            )
            update_names <- update_names[finite_update]

            if (length(update_names) > 0L) {
                start_used[update_names] <- as.list(
                    estimated_marginal[update_names]
                )
            } else {
                independence_message <- paste0(
                    "The independence fit succeeded for model '",
                    model_clean,
                    "', but it estimated no eligible parameter for the final start."
                )
            }
        }

        if (is.null(independence_fit) &&
            isTRUE(warn_independence_failure) &&
            !is.null(independence_message)) {
            warning(
                "Independence initialization was not available; ",
                "the supplied starting values are being used. ",
                "Reason: ", independence_message,
                call. = FALSE
            )
        }
    }

    fit_args$start <- start_used
    if (!is.null(independence_fit) && isTRUE(independence_fit$duplicate_locations_checked)) {
        fit_args$check.duplicates <- FALSE
    }
    fitted <- do.call(GeoFit, fit_args)

    # Keep the usual GeoFit class and add transparent initialization details.
    fitted$start_original <- start_original
    fitted$start_used <- start_used
    fitted$independence_fit <- independence_fit
    fitted$independence_start_message <- independence_message
    fitted$call <- call

    structure(fitted, class = "GeoFit")
}

####################################################
### Internal utilities for global and local kriging
####################################################

.GeoKrigMatchMethod <- function(method = "cholesky", sparse = FALSE,
                                context = "Kriging") {
  if (!is.character(method) || length(method) != 1L || is.na(method)) {
    stop(context, ": method must be 'cholesky' or 'svd'.", call. = FALSE)
  }
  method <- tolower(gsub("[[:blank:]]", "", method))
  if (!(method %in% c("cholesky", "svd"))) {
    stop(context, ": method must be 'cholesky' or 'svd'.", call. = FALSE)
  }
  if (isTRUE(sparse) && identical(method, "svd")) {
    stop(context, ": method = 'svd' is not available with sparse = TRUE; ",
         "use method = 'cholesky'.", call. = FALSE)
  }
  method
}

.GeoKrigCoordinateMatrix <- function(coordx = NULL, coordy = NULL,
                                     coordz = NULL, grid = FALSE,
                                     context = "Kriging") {
  if (is.null(coordx)) {
    stop(context, ": observed coordinates are missing.", call. = FALSE)
  }

  if ((is.matrix(coordx) || is.data.frame(coordx)) &&
      is.null(coordy) && is.null(coordz)) {
    coords <- as.matrix(coordx)
  } else {
    if (is.null(coordy)) {
      stop(context, ": coordy is required when coordx is not a coordinate matrix.",
           call. = FALSE)
    }
    if (!is.numeric(coordx) || !is.numeric(coordy) ||
        any(!is.finite(coordx)) || any(!is.finite(coordy))) {
      stop(context, ": coordx and coordy must be finite numeric vectors.",
           call. = FALSE)
    }
    if (!is.null(coordz) &&
        (!is.numeric(coordz) || any(!is.finite(coordz)))) {
      stop(context, ": coordz must be a finite numeric vector.", call. = FALSE)
    }

    if (isTRUE(grid)) {
      coords <- if (is.null(coordz)) {
        as.matrix(expand.grid(coordx, coordy))
      } else {
        as.matrix(expand.grid(coordx, coordy, coordz))
      }
    } else {
      if (length(coordx) != length(coordy)) {
        stop(context, ": coordx and coordy must have the same length.",
             call. = FALSE)
      }
      if (!is.null(coordz) && length(coordz) != length(coordx)) {
        stop(context, ": coordz must have the same length as coordx and coordy.",
             call. = FALSE)
      }
      coords <- if (is.null(coordz)) {
        cbind(coordx, coordy)
      } else {
        cbind(coordx, coordy, coordz)
      }
    }
  }

  if (!is.matrix(coords) || !is.numeric(coords) ||
      nrow(coords) < 1L || !(ncol(coords) %in% c(2L, 3L)) ||
      any(!is.finite(coords))) {
    stop(context, ": coordinates must be a finite numeric N x 2 or N x 3 matrix.",
         call. = FALSE)
  }
  storage.mode(coords) <- "double"
  unname(coords)
}

.GeoKrigNormalizeData <- function(data, nspace, coordt = NULL,
                                  grid = FALSE,
                                  context = "Kriging") {
  if (is.null(data)) stop(context, ": data are missing.", call. = FALSE)
  if (!is.numeric(data) || any(!is.finite(data))) {
    stop(context, ": data must be finite and numeric.", call. = FALSE)
  }
  if (!is.numeric(nspace) || length(nspace) != 1L ||
      !is.finite(nspace) || nspace < 1L || nspace != as.integer(nspace)) {
    stop(context, ": invalid number of spatial observations.", call. = FALSE)
  }
  nspace <- as.integer(nspace)

  if (is.null(coordt)) {
    if (length(data) != nspace) {
      stop(context, ": data must contain one value per spatial coordinate (",
           nspace, " values expected).", call. = FALSE)
    }
    ## For regular grids this is the order of expand.grid(coordx, coordy):
    ## the first coordinate varies fastest.  It is also the order returned by
    ## GeoSim for a spatial grid.
    return(unname(as.numeric(data)))
  }

  if (!is.numeric(coordt) || !length(coordt) || any(!is.finite(coordt))) {
    stop(context, ": coordt must be a non-empty finite numeric vector.",
         call. = FALSE)
  }
  nt <- length(coordt)
  total <- nspace * nt

  if (is.matrix(data)) {
    dd <- dim(data)
    if (identical(dd, c(nt, nspace))) {
      ## Rows are times, columns are spatial sites.
      return(unname(as.numeric(t(data))))
    }
    if (identical(dd, c(nspace, nt))) {
      ## Columns are times; column-major vectorisation is already time-major.
      return(unname(as.numeric(data)))
    }
  }

  if (length(dim(data)) >= 3L) {
    dd <- dim(data)
    if (!isTRUE(grid) || utils::tail(dd, 1L) != nt || prod(utils::head(dd, -1L)) != nspace) {
      stop(context, ": a regular-grid space-time array must have the last ",
           "dimension equal to length(coordt) and the remaining dimensions ",
           "equal to the spatial grid size.", call. = FALSE)
    }
    return(unname(as.numeric(data)))
  }

  if (is.vector(data) && length(data) == total) {
    return(unname(as.numeric(data)))
  }

  stop(context, ": data have an incompatible format for fixed-location ",
       "space-time observations; expected a length-", total,
       " vector, a ", nt, " x ", nspace, " matrix, a ", nspace,
       " x ", nt, " matrix, or a compatible regular-grid array.",
       call. = FALSE)
}

.GeoKrigObservedCoordinates <- function(covmatrix, dimat,
                                        context = "Kriging") {
  if (!is.null(covmatrix$coordx_dyn)) {
    coords <- do.call(rbind, lapply(covmatrix$coordx_dyn, as.matrix))
  } else {
    base <- if (is.null(covmatrix$coordz)) {
      cbind(covmatrix$coordx, covmatrix$coordy)
    } else {
      cbind(covmatrix$coordx, covmatrix$coordy, covmatrix$coordz)
    }
    repeats <- if (isTRUE(covmatrix$spacetime) || isTRUE(covmatrix$bivariate)) {
      as.integer(covmatrix$numtime)
    } else {
      1L
    }
    coords <- base[rep(seq_len(nrow(base)), times = repeats), , drop = FALSE]
  }

  coords <- as.matrix(coords)
  if (ncol(coords) == 2L) coords <- cbind(coords, 0)
  if (ncol(coords) != 3L || nrow(coords) != dimat ||
      !is.numeric(coords) || any(!is.finite(coords))) {
    stop(context, ": internal observed-coordinate ordering is inconsistent ",
         "with the covariance matrix.", call. = FALSE)
  }
  storage.mode(coords) <- "double"
  unname(coords)
}

.GeoKrigFactor <- function(covmatrix, method = "cholesky",
                           context = "Kriging") {
  method <- .GeoKrigMatchMethod(method, sparse = isTRUE(covmatrix$sparse),
                                context = context)
  C <- covmatrix$covmatrix
  nobs <- nrow(C)

  if (isTRUE(covmatrix$sparse)) {
    Cspam <- if (spam::is.spam(C)) C else spam::as.spam(C)
    U <- tryCatch(spam::chol.spam(Cspam), error = function(e) NULL)
    if (is.null(U)) {
      stop(context, ": covariance matrix is not positive definite.",
           call. = FALSE)
    }
    return(structure(
      list(method = "cholesky", sparse = TRUE, factor = U, nobs = nobs),
      class = "GeoKrigFactor"
    ))
  }

  C <- as.matrix(C)
  if (identical(method, "cholesky")) {
    U <- tryCatch(FastGP::rcppeigen_get_chol(C), error = function(e) NULL)
    if (is.null(U)) {
      stop(context, ": covariance matrix is not positive definite.",
           call. = FALSE)
    }
    return(structure(
      list(method = "cholesky", sparse = FALSE, factor = U, nobs = nobs),
      class = "GeoKrigFactor"
    ))
  }

  dec <- tryCatch(svd(C), error = function(e) NULL)
  if (is.null(dec) || !length(dec$d) || any(!is.finite(dec$d))) {
    stop(context, ": SVD decomposition of the covariance matrix failed.",
         call. = FALSE)
  }
  tol <- max(dim(C)) * max(dec$d) * .Machine$double.eps
  if (any(dec$d <= tol)) {
    stop(context, ": covariance matrix is numerically singular under SVD ",
         "(smallest singular value = ", format(min(dec$d)), ").",
         call. = FALSE)
  }
  structure(
    list(method = "svd", sparse = FALSE, factor = dec, nobs = nobs),
    class = "GeoKrigFactor"
  )
}

.GeoKrigSolveFactor <- function(factor, rhs, context = "Kriging") {
  if (!inherits(factor, "GeoKrigFactor")) {
    stop(context, ": invalid cached kriging factorization.", call. = FALSE)
  }
  rhs <- as.matrix(rhs)
  if (!is.numeric(rhs) || any(!is.finite(rhs))) {
    stop(context, ": the right-hand side of the kriging system must be finite.",
         call. = FALSE)
  }
  if (nrow(rhs) != factor$nobs) {
    stop(context, ": incompatible kriging-system dimensions.", call. = FALSE)
  }

  if (isTRUE(factor$sparse)) {
    tmp <- spam::forwardsolve(factor$factor, rhs)
    return(as.matrix(spam::backsolve(factor$factor, tmp)))
  }

  if (identical(factor$method, "cholesky")) {
    tmp <- forwardsolve(factor$factor, rhs)
    return(as.matrix(forwardsolve(factor$factor, tmp, transpose = TRUE)))
  }

  dec <- factor$factor
  as.matrix(dec$v %*% ((crossprod(dec$u, rhs)) / dec$d))
}

.GeoKrigSolve <- function(covmatrix, rhs, method = "cholesky",
                          context = "Kriging") {
  factor <- .GeoKrigFactor(covmatrix, method = method, context = context)
  .GeoKrigSolveFactor(factor, rhs, context = context)
}

.GeoKrigDenseSPDSolve <- function(C, rhs, context = "Kriging") {
  C <- as.matrix(C)
  rhs <- as.matrix(rhs)
  n <- nrow(C)
  nrhs <- ncol(rhs)

  if (!is.numeric(C) || n < 1L || ncol(C) != n || any(!is.finite(C))) {
    stop(context, ": covariance matrix must be a finite square numeric matrix.",
         call. = FALSE)
  }
  if (!is.numeric(rhs) || nrhs < 1L || nrow(rhs) != n ||
      any(!is.finite(rhs))) {
    stop(context, ": incompatible or non-finite kriging right-hand side.",
         call. = FALSE)
  }

  ## dposv performs Cholesky factorization and all triangular solves in one
  ## native LAPACK call.  Private copies are passed because LAPACK overwrites
  ## both the covariance matrix and the right-hand sides.  This is especially
  ## useful for GeoKrigloc, where thousands of small dense systems would
  ## otherwise cross the R/native boundary several times per neighborhood.
  z <- dotCall64::.C64(
    "GeoDenseSPDSolve",
    SIGNATURE = c("double", "double", "integer", "integer", "integer"),
    A = as.double(C),
    B = as.double(rhs),
    as.integer(n), as.integer(nrhs), info = as.integer(0L),
    INTENT = c("rw", "rw", "r", "r", "w"),
    PACKAGE = "GeoModels", VERBOSE = 0, NAOK = TRUE
  )
  info <- as.integer(z$info)[1L]
  if (!is.finite(info) || info != 0L) {
    if (is.finite(info) && info > 0L) {
      stop(context, ": covariance matrix is not positive definite.",
           call. = FALSE)
    }
    stop(context, ": native dense Cholesky solve failed (LAPACK info = ",
         info, ").", call. = FALSE)
  }
  out <- matrix(z$B, nrow = n, ncol = nrhs)
  if (any(!is.finite(out))) {
    stop(context, ": native dense Cholesky solve returned non-finite values.",
         call. = FALSE)
  }
  out
}

.GeoKrigLocationBlocks <- function(nobs, nloc, tloc = 1L,
                                   context = "Kriging") {
  nobs <- as.integer(nobs)
  nloc <- as.integer(nloc)
  tloc <- as.integer(tloc)
  if (any(!is.finite(c(nobs, nloc, tloc))) || nobs < 1L || nloc < 1L ||
      tloc < 1L) {
    stop(context, ": invalid dimensions for blocked prediction.", call. = FALSE)
  }

  ## The option is intentionally internal.  It controls the approximate size
  ## of one dense cross-covariance matrix, not the total memory footprint.
  max_bytes <- getOption("GeoModels.krig.block_bytes", 32 * 1024^2)
  max_bytes <- suppressWarnings(as.numeric(max_bytes)[1L])
  if (!is.finite(max_bytes) || max_bytes <= 0) max_bytes <- 32 * 1024^2

  bytes_per_location <- 8 * as.double(nobs) * as.double(tloc)
  block_nloc <- max(1L, min(nloc, floor(max_bytes / bytes_per_location)))
  starts <- seq.int(1L, nloc, by = block_nloc)
  lapply(starts, function(i) seq.int(i, min(nloc, i + block_nloc - 1L)))
}

.GeoKrigCorrelationBlock <- function(covmatrix, observed_coords,
                                     corrmodel, corrparam, distance,
                                     loc_block, time, tloc, NS,
                                     which = 1L, context = "Kriging") {
  loc_block <- as.matrix(loc_block)
  if (!is.numeric(loc_block) || nrow(loc_block) < 1L ||
      !(ncol(loc_block) %in% c(2L, 3L)) || any(!is.finite(loc_block))) {
    stop(context, ": invalid prediction-coordinate block.", call. = FALSE)
  }
  if (ncol(loc_block) == 2L) loc_block <- cbind(loc_block, 0)
  observed_coords <- as.matrix(observed_coords)
  if (ncol(observed_coords) == 2L) observed_coords <- cbind(observed_coords, 0)

  nobs <- nrow(observed_coords)
  nloc_block <- nrow(loc_block)
  ntarget <- nloc_block * as.integer(tloc)
  out <- dotCall64::.C64(
    "Corr_c",
    SIGNATURE = c(
      "double", "double", "double", "double", "double", "integer",
      "integer", "double", "double", "double", "integer", "integer",
      "integer", "integer", "integer", "integer", "double", "integer",
      "integer", "double", "integer", "integer", "double"
    ),
    corri = dotCall64::vector_dc("double", nobs * ntarget),
    observed_coords[, 1L], observed_coords[, 2L], observed_coords[, 3L],
    covmatrix$coordt, corrmodel, 0,
    loc_block[, 1L], loc_block[, 2L], loc_block[, 3L],
    covmatrix$numcoord, nloc_block, as.integer(tloc),
    covmatrix$ns, NS, covmatrix$numtime, corrparam,
    covmatrix$spacetime, covmatrix$bivariate, time, distance,
    as.integer(which) - 1L, covmatrix$radius,
    INTENT = c("w", rep("r", 22)), NAOK = TRUE, PACKAGE = "GeoModels"
  )$corri
  if (length(out) != nobs * ntarget || any(!is.finite(out))) {
    stop(context, ": non-finite cross-correlation block.", call. = FALSE)
  }
  out
}

.GeoKrigMeanVarcov <- function(varcov, mean_names,
                                context = "Kriging") {
  if (is.null(varcov)) {
    stop(context, ": type_krig = 'Universal' with mse = TRUE requires ",
         "varcov.  For a GeoFit object, compute/update its parameter ",
         "covariance matrix before kriging.", call. = FALSE)
  }

  V <- as.matrix(varcov)
  if (!is.numeric(V) || nrow(V) != ncol(V) || any(!is.finite(V))) {
    stop(context, ": varcov must be a finite square numeric matrix.",
         call. = FALSE)
  }
  rn <- rownames(V)
  cn <- colnames(V)
  if (is.null(rn) || is.null(cn) || anyDuplicated(rn) || anyDuplicated(cn)) {
    stop(context, ": varcov must have unique row and column names.",
         call. = FALSE)
  }
  if (!setequal(rn, cn)) {
    stop(context, ": row and column names of varcov must identify the same ",
         "parameters.", call. = FALSE)
  }

  out <- matrix(0, nrow = length(mean_names), ncol = length(mean_names),
                dimnames = list(mean_names, mean_names))
  available <- mean_names[mean_names %in% rn & mean_names %in% cn]
  if (length(available)) {
    block <- V[available, available, drop = FALSE]
    asym <- max(abs(block - t(block)))
    tol <- 1e-8 * max(1, max(abs(block)))
    if (!is.finite(asym) || asym > tol) {
      stop(context, ": the mean-parameter block of varcov is not symmetric.",
           call. = FALSE)
    }
    out[available, available] <- (block + t(block)) / 2
  }
  out
}


.GeoKrigSystem <- function(covmatrix, CC, method = "cholesky",
                           type_krig = "Simple", X = NULL, Xloc = NULL,
                           target_variance = NULL, mse = FALSE,
                           varcov = NULL, mean_names = NULL, factor = NULL,
                           weights = NULL, context = "Kriging") {
  type_key <- tolower(gsub("[[:blank:]]", "", type_krig))
  if (!(type_key %in% c("simple", "universal"))) {
    stop(context, ": type_krig must be 'Simple' or 'Universal'.",
         call. = FALSE)
  }

  CC <- as.matrix(CC)
  nobs <- nrow(CC)
  npred <- ncol(CC)
  if (nobs != nrow(covmatrix$covmatrix)) {
    stop(context, ": cross-covariance matrix has the wrong number of rows.",
         call. = FALSE)
  }

  ## Both Simple and Universal use the mean coefficients supplied by the
  ## fitted model (ML, REML, composite likelihood, or fixed by the user).
  ## Universal differs only in the prediction-MSE correction below.
  if (is.null(weights)) {
    weights <- if (is.null(factor)) {
      .GeoKrigSolve(covmatrix, CC, method = method, context = context)
    } else {
      .GeoKrigSolveFactor(factor, CC, context = context)
    }
  } else {
    weights <- as.matrix(weights)
    if (!is.numeric(weights) || any(!is.finite(weights)) ||
        nrow(weights) != nobs || ncol(weights) != npred) {
      stop(context, ": invalid precomputed kriging weights.", call. = FALSE)
    }
  }
  quad <- colSums(CC * weights)

  mse_value <- NULL
  if (isTRUE(mse)) {
    if (is.null(target_variance)) {
      stop(context, ": target variance is required to compute the MSE.",
           call. = FALSE)
    }
    mse_value <- rep(as.numeric(target_variance), length.out = npred) - quad

    if (identical(type_key, "universal")) {
      X <- as.matrix(X)
      Xloc <- as.matrix(Xloc)
      if (!is.numeric(X) || any(!is.finite(X)) || nrow(X) != nobs ||
          ncol(X) < 1L) {
        stop(context, ": X must be a finite design matrix with one row per ",
             "observation for universal-MSE correction.", call. = FALSE)
      }
      if (!is.numeric(Xloc) || any(!is.finite(Xloc)) ||
          nrow(Xloc) != npred || ncol(Xloc) != ncol(X)) {
        stop(context, ": Xloc must have one row per prediction task and the ",
             "same number of columns as X for universal-MSE correction.",
             call. = FALSE)
      }
      if (is.null(mean_names)) {
        mean_names <- colnames(X)
        if (is.null(mean_names) || any(!nzchar(mean_names))) {
          mean_names <- .GeoMean_names(ncol(X))
        }
      }
      if (length(mean_names) != ncol(X) || anyDuplicated(mean_names)) {
        stop(context, ": invalid mean-parameter names for universal kriging.",
             call. = FALSE)
      }

      var_mean <- .GeoKrigMeanVarcov(
        varcov, mean_names = mean_names, context = context
      )
      trend_gap <- Xloc - t(weights) %*% X
      correction <- rowSums((trend_gap %*% var_mean) * trend_gap)
      mse_value <- mse_value + correction
    }
  }

  list(weights = weights, mse = mse_value,
       type = if (identical(type_key, "universal")) "Universal" else "Simple")
}

################################################################################
## Prediction model/counter helpers                                            ##
################################################################################
.GeoPredictionCanonicalModel <- function(model) {
  if (!is.character(model) || length(model) != 1L || is.na(model)) return(model)
  model <- gsub("[[:blank:]]", "", model)
  aliases <- c(
    Gauss = "Gaussian", LogGauss = "LogGaussian", SkewGauss = "SkewGaussian",
    TwoPieceGauss = "TwoPieceGaussian", tukeyh = "Tukeyh", tukeyh2 = "Tukeyh2",
    poisson = "Poisson", poissongamma = "PoissonGamma",
    Binomiallogistic = "BinomialLogistic",
    Gaussian_misp_StudentT = "StudentT",
    Gaussian_misp_Poisson = "Poisson",
    Gaussian_misp_SkewStudentT = "SkewStudentT",
    Gaussian_misp_Tukeygh = "Tukeygh",
    Gaussian_misp_PoissonZIP = "PoissonZIP",
    Gaussian_misp_PoissonGamma = "PoissonGamma",
    Gaussian_misp_Binomial = "Binomial",
    Gaussian_misp_BinomialNeg = "BinomialNeg"
  )
  if (model %in% names(aliases)) unname(aliases[[model]]) else model
}

.GeoKrigSupportedPlainModels <- c(
  "Gaussian", "SkewGaussian", "SkewLaplace", "StudentT", "SkewStudentT",
  "Gamma", "Weibull", "LogLogistic", "LogGaussian", "TwoPieceStudentT",
  "Beta", "TwoPieceGaussian", "TwoPieceTukeyh", "TwoPieceBimodal",
  "Tukeygh", "Tukeyh", "Tukeyh2", "SinhAsinh",
  "Binary", "Bernoulli", "Binomial", "Geom", "Geometric", "BinomialNeg",
  "Poisson", "PoissonZIP", "BinomialNegZINB",
  "PoissonGamma", "PoissonGammaZIP", "PoissonGammaZIP1"
)

.GeoKrigValidatePredictionModel <- function(model, copula = NULL, bivariate = FALSE,
                                            context = "GeoKrig") {
  if (isTRUE(bivariate) && !identical(model, "Gaussian")) {
    stop(context, ": bivariate prediction is currently implemented only for model = 'Gaussian'.",
         call. = FALSE)
  }
  if (is.null(copula)) {
    if (!(model %in% .GeoKrigSupportedPlainModels)) {
      stop(context, ": prediction is not implemented for model = '", model,
           "' without a copula.", call. = FALSE)
    }
  } else {
    model_code <- as.integer(CkModel(model))
    supported_codes <- .GeoGaussianCopulaContinuousModels
    if (identical(copula, "Gaussian") || identical(context, "GeoKrig"))
      supported_codes <- c(supported_codes, .GeoGaussianCopulaDiscreteModels)
    if (!(model_code %in% supported_codes)) {
      if (model_code %in% .GeoGaussianCopulaDiscreteModels &&
          copula %in% c("Clayton", "SkewGaussian") && !identical(context, "GeoKrig")) {
        stop(context, ": ", copula,
             " copula with Poisson/Binomial/BinomialNeg is currently available in global GeoKrig/GeoCV only; local GeoKrigloc is not yet implemented.",
             call. = FALSE)
      }
      stop(context, ": copula-based prediction is not implemented for copula = '",
           copula, "' and model = '", model, "'.", call. = FALSE)
    }
  }
  invisible(TRUE)
}

.GeoKrigPositiveIntegerVector <- function(x, name, n = NULL) {
  if (!is.numeric(x) || !length(x) || any(!is.finite(x)) ||
      any(x < 1) || any(abs(x - round(x)) > sqrt(.Machine$double.eps))) {
    stop(name, " must contain positive integers.", call. = FALSE)
  }
  x <- as.integer(round(x))
  if (!is.null(n)) {
    if (length(x) == 1L) x <- rep.int(x, n)
    if (length(x) != n) {
      stop(name, " must be scalar or contain one value per corresponding location.",
           call. = FALSE)
    }
  }
  x
}

.GeoKrigCountSizes <- function(model, n, nloc, nobs, npred,
                               context = "GeoKrig") {
  model <- .GeoPredictionCanonicalModel(model)
  ## Models whose response mean/variance depends on a size/trial parameter.
  binomial_models <- c("Binomial", "Binomial2")
  negbin_models <- c("BinomialNeg", "BinomialNegZINB")
  if (model %in% c("Binary", "Bernoulli")) {
    if(!is.numeric(n) || length(n) != 1L || !is.finite(n) || as.integer(round(n)) != 1L)
      stop(context, ": Binary/Bernoulli is the one-trial field; n must equal 1.", call.=FALSE)
    if(!is.null(nloc) && any(as.numeric(nloc) != 1))
      stop(context, ": Binary/Bernoulli prediction requires nloc=1 when nloc is supplied.", call.=FALSE)
    return(list(obs = rep.int(1L, nobs), pred = rep.int(1L, npred)))
  }
  if (model %in% c("Geom", "Geometric")) {
    if(!is.numeric(n) || length(n) != 1L || !is.finite(n) || as.integer(round(n)) != 1L)
      stop(context, ": Geometric is the r = 1 Negative-Binomial field; n must equal 1.", call.=FALSE)
    if(!is.null(nloc) && any(as.numeric(nloc) != 1))
      stop(context, ": Geometric is the r = 1 Negative-Binomial field; nloc, when supplied, must equal 1.", call.=FALSE)
    return(list(obs = rep.int(1L, nobs), pred = rep.int(1L, npred)))
  }
  if (!(model %in% c(binomial_models, negbin_models))) {
    ## Corr_c_bin still expects integer vectors; ones are neutral for Poisson-type models.
    return(list(obs = rep.int(1L, nobs), pred = rep.int(1L, npred)))
  }

  if(model %in% negbin_models) {
    r <- .GeoKrigPositiveIntegerVector(n, paste0(context, ": n"), n = NULL)
    if(length(r) != 1L)
      stop(context, ": for BinomialNeg, n is the common number r of successes and must be a positive integer scalar.", call.=FALSE)
    if(!is.null(nloc)) {
      rloc <- .GeoKrigPositiveIntegerVector(nloc, paste0(context, ": nloc"), n = NULL)
      if(any(rloc != r[1L]))
        stop(context, ": for BinomialNeg, r is a common field parameter; nloc cannot differ from n.", call.=FALSE)
    }
    return(list(obs = rep.int(r[1L], nobs), pred = rep.int(r[1L], npred)))
  }

  obs <- .GeoKrigPositiveIntegerVector(n, paste0(context, ": n"), n = nobs)
  if (is.null(nloc)) {
    if (length(n) != 1L) {
      stop(context, ": nloc is required when Binomial n is location-specific.", call. = FALSE)
    }
    pred <- rep.int(as.integer(round(n[1L])), npred)
  } else {
    pred <- .GeoKrigPositiveIntegerVector(nloc, paste0(context, ": nloc"), n = npred)
  }
  list(obs = obs, pred = pred)
}

.GeoKrigCountMeanDerivative <- function(model_code, param, eta, size) {
  eta <- as.numeric(eta)
  size <- rep(as.numeric(size), length.out = length(eta))
  code <- as.integer(model_code)
  if (code == 30L) return(exp(eta))                         # Poisson
  if (code %in% c(46L, 57L, 58L)) {                        # Poisson-Gamma (+ ZIP)
    d <- exp(eta)
    if (code %in% c(57L, 58L)) d <- (1 - pnorm(as.numeric(param["pmu"]))) * d
    return(d)
  }
  if (code == 43L) {                                       # Poisson ZIP
    return((1 - pnorm(as.numeric(param["pmu"]))) * exp(eta))
  }
  if (code %in% c(2L, 11L, 19L)) {                         # Binary/Binomial
    return(size * dnorm(eta))
  }
  if (code %in% c(14L, 16L)) {                             # Geometric/NegBin
    p <- pnorm(eta)
    return(-size * dnorm(eta) / p^2)
  }
  if (code == 45L) {                                       # zero-inflated NegBin
    p <- pnorm(eta)
    z <- pnorm(as.numeric(param["pmu"]))
    return(-(1 - z) * size * dnorm(eta) / p^2)
  }
  stop("GeoKrig: no marginal-mean derivative is implemented for this count model.",
       call. = FALSE)
}

###########################################################################
## Sparse observation-pair correlation cache for local kriging           ##
###########################################################################

.GeoLocalPairOccurrenceCount <- function(indices) {
  if (!is.list(indices) || !length(indices)) return(0)
  sum(vapply(indices, function(idx) {
    k <- length(idx)
    if (k < 2L) 0 else as.double(k) * as.double(k - 1L) / 2
  }, numeric(1L)))
}

.GeoPairCacheOption <- function(auto = FALSE) {
  opt <- getOption("GeoModels.local_pair_cache", NULL)
  if (is.null(opt) || identical(opt, "auto")) return(isTRUE(auto))
  if (!is.logical(opt) || length(opt) != 1L || is.na(opt)) {
    stop("Option 'GeoModels.local_pair_cache' must be TRUE, FALSE, 'auto', or NULL.",
         call. = FALSE)
  }
  isTRUE(opt)
}

.GeoPairCacheNear <- function(x, values) {
  x <- suppressWarnings(as.numeric(x)[1L])
  is.finite(x) && any(abs(x - values) <= 100 * .Machine$double.eps * max(1, abs(x)))
}

.GeoPairCacheSpatialAuto <- function(corrmodel, param, weights = FALSE) {
  if (isTRUE(weights)) return(TRUE)
  code <- as.integer(CkCorrModel(corrmodel))
  if (!length(code) || is.na(code)) return(FALSE)

  ## Always expensive: Kummer / hypergeometric / hole-effect families.
  if (code %in% c(21L, 22L, 23L, 24L, 25L, 26L, 27L, 29L, 30L)) return(TRUE)

  ## General Matern is expensive except the closed half-integer cases.
  if (code == 14L) {
    return(!.GeoPairCacheNear(param[["smooth"]], c(0.5, 1.5, 2.5, 3.5)))
  }

  ## GenWend and its Matern reparameterizations share CorFunW_gen().
  ## smooth 0:3 use the cheap closed forms measured in the local benchmark.
  if (code %in% c(6L, 7L, 19L)) {
    return(!.GeoPairCacheNear(param[["smooth"]], 0:3))
  }
  FALSE
}

.GeoPairCacheSpacetimeAuto <- function(corrmodel, param, weights = FALSE) {
  if (isTRUE(weights)) return(TRUE)
  code <- as.integer(CkCorrModel(corrmodel))
  if (!length(code) || is.na(code)) return(FALSE)

  ## Space-time families containing a general Matern or generalized Wendland
  ## evaluation.  Their pair correlation depends only on fixed (h_ij, u_ij)
  ## during one local-kriging call, so it is safe to reuse across targets.
  code %in% c(61L, 62L, 78L, 85L, 86L, 87L, 88L, 89L, 95L)
}

.GeoPairCacheEnabled <- function(corrmodel, param, weights = FALSE,
                                 spacetime = FALSE) {
  auto <- if (isTRUE(spacetime)) {
    .GeoPairCacheSpacetimeAuto(corrmodel, param, weights = weights)
  } else {
    .GeoPairCacheSpatialAuto(corrmodel, param, weights = weights)
  }
  .GeoPairCacheOption(auto)
}

.GeoPairCacheThresholdOK <- function(indices) {
  total_pairs <- .GeoLocalPairOccurrenceCount(indices)
  min_pairs <- suppressWarnings(as.numeric(
    getOption("GeoModels.local_pair_cache_min_pairs", 1e6)
  )[1L])
  if (!is.finite(min_pairs) || min_pairs < 0) min_pairs <- 1e6
  is.finite(total_pairs) && total_pairs >= min_pairs
}

.GeoValidatePairCacheGeometry <- function(coords, times = NULL,
                                          context = "local kriging") {
  coords <- as.matrix(coords)
  if (!is.numeric(coords) || nrow(coords) < 1L ||
      !(ncol(coords) %in% c(2L, 3L)) || any(!is.finite(coords))) {
    stop(context, ": invalid coordinates for pair-cache construction.",
         call. = FALSE)
  }
  storage.mode(coords) <- "double"
  if (!is.null(times)) {
    times <- as.numeric(times)
    if (length(times) != nrow(coords) || any(!is.finite(times))) {
      stop(context, ": invalid times for pair-cache construction.", call. = FALSE)
    }
  }
  list(coords = coords, times = times)
}

.GeoBuildSpatialPairCache <- function(indices, coords, corrmodel_code,
                                      corrparam, distance_code, radius,
                                      context = "local spatial kriging") {
  if (!.GeoPairCacheThresholdOK(indices)) return(NULL)
  geom <- .GeoValidatePairCacheGeometry(coords, context = context)
  corrparam <- as.double(corrparam)
  if (!length(corrparam) || any(!is.finite(corrparam))) {
    stop(context, ": invalid correlation parameters for pair cache.",
         call. = FALSE)
  }

  cache <- .Call(
    "GeoBuildSpatialPairCache",
    indices, geom$coords, as.integer(corrmodel_code), corrparam,
    as.integer(distance_code), as.double(radius),
    PACKAGE = "GeoModels"
  )
  .GeoValidatePairCache(cache, context)
}

.GeoBuildSpacetimePairCache <- function(indices, coords, times,
                                        corrmodel_code, corrparam,
                                        distance_code, radius,
                                        context = "local space-time kriging") {
  if (!.GeoPairCacheThresholdOK(indices)) return(NULL)
  geom <- .GeoValidatePairCacheGeometry(coords, times, context)
  corrparam <- as.double(corrparam)
  if (!length(corrparam) || any(!is.finite(corrparam))) {
    stop(context, ": invalid correlation parameters for pair cache.",
         call. = FALSE)
  }

  cache <- .Call(
    "GeoBuildSpacetimePairCache",
    indices, geom$coords, geom$times, as.integer(corrmodel_code), corrparam,
    as.integer(distance_code), as.double(radius),
    PACKAGE = "GeoModels"
  )
  .GeoValidatePairCache(cache, context)
}

.GeoValidatePairCache <- function(cache, context) {
  if (!is.list(cache) || is.null(cache$row_ptr) || is.null(cache$col_idx) ||
      is.null(cache$rho) || length(cache$col_idx) != length(cache$rho)) {
    stop(context, ": native pair-cache construction returned invalid data.",
         call. = FALSE)
  }
  cache
}

.GeoPairCachePacked <- function(indices, cache,
                                context = "local kriging") {
  if (is.null(cache)) return(NULL)
  idx <- as.integer(indices)
  if (length(idx) < 2L) return(numeric(0L))
  out <- .Call(
    "GeoPairCachePacked",
    idx, cache$row_ptr, cache$col_idx, cache$rho,
    PACKAGE = "GeoModels"
  )
  if (length(out) != as.double(length(idx)) * as.double(length(idx) - 1L) / 2 ||
      any(!is.finite(out))) {
    stop(context, ": invalid pair-cache lookup result.", call. = FALSE)
  }
  as.numeric(out)
}

.GeoSpacetimeObservationGeometry <- function(coords, coordx_dyn, coordt,
                                             context = "local space-time kriging") {
  coordt <- as.numeric(coordt)
  if (!length(coordt) || any(!is.finite(coordt))) {
    stop(context, ": invalid observation times.", call. = FALSE)
  }
  if (!is.null(coordx_dyn)) {
    if (!is.list(coordx_dyn) || length(coordx_dyn) != length(coordt)) {
      stop(context, ": dynamic coordinates and times are inconsistent.", call. = FALSE)
    }
    sizes <- vapply(coordx_dyn, nrow, integer(1L))
    flat <- do.call(rbind, lapply(coordx_dyn, as.matrix))
    times <- rep(coordt, sizes)
  } else {
    coords <- as.matrix(coords)
    ns <- nrow(coords)
    flat <- coords[rep(seq_len(ns), times = length(coordt)), , drop = FALSE]
    times <- rep(coordt, each = ns)
  }
  geom <- .GeoValidatePairCacheGeometry(flat, times, context)
  list(coords = geom$coords, times = geom$times)
}

.GeoPairCrossCorrelation <- function(indices, coords, times = NULL, loc,
                                     pred_time = 0, corrmodel_code, corrparam,
                                     distance_code, radius,
                                     context = "local kriging") {
  geom <- .GeoValidatePairCacheGeometry(coords, times, context)
  loc <- as.numeric(loc)
  if (length(loc) != ncol(geom$coords) || any(!is.finite(loc))) {
    stop(context, ": invalid prediction coordinates for pair correlation.",
         call. = FALSE)
  }
  out <- .Call(
    "GeoPairCrossCorrelation",
    as.integer(indices), geom$coords, geom$times, loc, as.double(pred_time),
    as.integer(corrmodel_code), as.double(corrparam),
    as.integer(distance_code), as.double(radius), PACKAGE = "GeoModels"
  )
  if (length(out) != length(indices) || any(!is.finite(out))) {
    stop(context, ": invalid indexed cross-correlation result.", call. = FALSE)
  }
  as.numeric(out)
}

.GeoPairCorrelationParam <- function(corrmodel, param,
                                     context = "local kriging") {
  code <- as.integer(CkCorrModel(corrmodel))
  nm <- CorrelationPar(code)
  if (!length(nm)) {
    stop(context, ": correlation parameters are not defined.", call. = FALSE)
  }
  vals <- unlist(param[nm], use.names = FALSE)
  vals <- as.double(vals)
  if (length(vals) != length(nm) || any(!is.finite(vals))) {
    stop(context, ": missing or invalid correlation parameters.", call. = FALSE)
  }
  vals
}

.GeoGaussianPairCovariance <- function(indices, cache, param,
                                       context = "local kriging") {
  idx <- as.integer(indices)
  k <- length(idx)
  if (k < 1L) return(matrix(numeric(0L), 0L, 0L))
  packed <- .GeoPairCachePacked(idx, cache, context = context)
  R <- diag(1, k)
  if (k > 1L) {
    R[lower.tri(R)] <- packed
    R <- R + t(R) - diag(k)
  }
  nugget <- suppressWarnings(as.numeric(param[["nugget"]])[1L])
  sill <- suppressWarnings(as.numeric(param[["sill"]])[1L])
  if (!is.finite(nugget)) nugget <- 0
  if (!is.finite(sill) || sill <= 0) {
    stop(context, ": invalid Gaussian sill for pair-cache covariance.",
         call. = FALSE)
  }
  sill * ((1 - nugget) * R + nugget * diag(k))
}

.GeoSpatialCrossCorrelation <- function(coords_obs, loc,
                                        corrmodel_code, corrparam,
                                        distance_code, radius,
                                        context = "local spatial kriging") {
  coords_obs <- as.matrix(coords_obs)
  loc <- as.numeric(loc)
  if (ncol(coords_obs) == 2L) coords_obs <- cbind(coords_obs, 0)
  if (length(loc) == 2L) loc <- c(loc, 0)
  if (ncol(coords_obs) != 3L || length(loc) != 3L) {
    stop(context, ": incompatible coordinates for cross-correlation.",
         call. = FALSE)
  }
  storage.mode(coords_obs) <- "double"
  nobs <- nrow(coords_obs)
  ans <- dotCall64::.C64(
    "Corr_c",
    SIGNATURE = c(
      "double", "double", "double", "double", "double", "integer",
      "integer", "double", "double", "double", "integer", "integer",
      "integer", "integer", "integer", "integer", "double", "integer",
      "integer", "double", "integer", "integer", "double"
    ),
    corri = dotCall64::vector_dc("double", nobs),
    coords_obs[, 1L], coords_obs[, 2L], coords_obs[, 3L], 0,
    as.integer(corrmodel_code), 0L,
    as.double(loc[1L]), as.double(loc[2L]), as.double(loc[3L]),
    as.integer(nobs), 1L, 1L, as.integer(nobs), 0L, 1L,
    as.double(corrparam), 0L, 0L, 0,
    as.integer(distance_code), 0L, as.double(radius),
    INTENT = c("w", rep("r", 22)), NAOK = TRUE, PACKAGE = "GeoModels"
  )$corri
  if (length(ans) != nobs || any(!is.finite(ans))) {
    stop(context, ": non-finite spatial cross-correlation.", call. = FALSE)
  }
  as.numeric(ans)
}

####################################################
### File name: GeoTest.r
####################################################

### Procedures are in alphabetical order.

### Statistical hypothesis testing for nested models
GeoTests <- function(object1, object2, ..., statistic) {

  ## ------------------------------------------------------------------
  ## Internal utilities
  ## ------------------------------------------------------------------
  valid_statistics <- c("Wald", "Wilks", "WilksS")
  composite_likelihoods <- c("Marginal", "Conditional")

  same_value <- function(x, y) {
    isTRUE(all.equal(x, y, check.attributes = FALSE))
  }

  as_named_numeric <- function(x) {
    if (is.null(x)) {
      return(setNames(numeric(0), character(0)))
    }

    out <- unlist(x, use.names = TRUE)
    storage.mode(out) <- "double"
    out
  }

  canonical_model <- function(x) {
    x <- as.character(x)[1L]

    if (x %in% c("Gauss", "Gaussian")) {
      return("Gaussian")
    }
    if (x %in% c("SkewGauss", "SkewGaussian")) {
      return("SkewGaussian")
    }

    x
  }

  marginal_nesting_rule <- function(model1, model2) {
    unrestricted <- canonical_model(model1$model)
    restricted <- canonical_model(model2$model)

    if (identical(unrestricted, restricted)) {
      return(list(
        same_model = TRUE,
        null = setNames(numeric(0), character(0)),
        boundary = character(0)
      ))
    }

    rules <- list(
      "SkewGaussian->Gaussian" = list(
        null = c(skew = 0),
        boundary = character(0)
      ),
      "Tukeyh->Gaussian" = list(
        null = c(tail = 0),
        boundary = "tail = 0 in the Tukey-h model"
      ),
      "SinhAsinh->Gaussian" = list(
        null = c(skew = 0, tail = 1),
        boundary = character(0)
      )
    )

    key <- paste0(unrestricted, "->", restricted)
    rule <- rules[[key]]

    if (is.null(rule)) {
      stop(
        "The marginal models are not identical and no exact nesting relation ",
        "is implemented for ", unrestricted, " -> ", restricted,
        ". The first model must be the unrestricted model.",
        call. = FALSE
      )
    }

    list(
      same_model = FALSE,
      null = rule$null,
      boundary = rule$boundary
    )
  }

  ensure_named_matrix <- function(A, par_names, label) {
    if (is.null(A)) {
      stop(label, " is missing", call. = FALSE)
    }

    A <- as.matrix(A)
    storage.mode(A) <- "double"

    if (!all(dim(A) == c(length(par_names), length(par_names)))) {
      stop(label, " has incompatible dimensions", call. = FALSE)
    }

    if (is.null(rownames(A)) || is.null(colnames(A))) {
      rownames(A) <- colnames(A) <- par_names
    }

    if (!all(par_names %in% rownames(A)) ||
        !all(par_names %in% colnames(A))) {
      stop(label, " does not contain all estimated parameters", call. = FALSE)
    }

    A <- A[par_names, par_names, drop = FALSE]
    (A + t(A)) / 2
  }

  safe_solve <- function(A, label) {
    A <- as.matrix(A)
    storage.mode(A) <- "double"
    A <- (A + t(A)) / 2

    out <- tryCatch(
      solve(A),
      error = function(e) NULL
    )

    if (is.null(out) || any(!is.finite(out))) {
      stop(label, " is singular or numerically non-invertible", call. = FALSE)
    }

    (out + t(out)) / 2
  }

  finite_loglik <- function(model, label) {
    value <- if (!is.null(model$logCompLik)) {
      model$logCompLik
    } else {
      model$logLik
    }

    value <- as.numeric(value)

    if (length(value) != 1L || !is.finite(value)) {
      stop(label, " does not contain a finite log-likelihood value", call. = FALSE)
    }

    value
  }

  is_stochastic_thinning <- function(model) {
    p <- if (is.null(model$p_neighb)) 1 else as.numeric(model$p_neighb)
    method <- if (is.null(model$thin_method)) {
      "bernoulli"
    } else {
      tolower(as.character(model$thin_method))
    }

    method %in% c(
      "bernoulli", "fixedbudget", "fixed_budget",
      "targetbalanced", "match"
    ) && isTRUE(p < 1)
  }

  canonical_coordt <- function(model) {
    x <- model$coordt

    ## For purely spatial and bivariate fits, GeoModels may represent the
    ## absence of a time coordinate as NULL, numeric(0), or a scalar
    ## placeholder such as 0. These representations are equivalent.
    if (!isTRUE(model$spacetime) &&
        (is.null(x) || length(x) <= 1L)) {
      return(NULL)
    }

    unname(x)
  }

  regression_parameter_names <- function(model) {
    if (isTRUE(model$bivariate)) {
      p <- as.integer(model$numbetas)

      if (length(p) != 2L || any(!is.finite(p)) || any(p < 1L)) {
        return(character(0))
      }

      block_names <- function(prefix, size) {
        if (size == 1L) {
          prefix
        } else {
          c(prefix, paste0(prefix, seq_len(size - 1L)))
        }
      }

      return(c(
        block_names("mean_1", p[1L]),
        block_names("mean_2", p[2L])
      ))
    }

    p <- if (!is.null(model$numbetas)) {
      as.integer(model$numbetas[1L])
    } else if (!is.null(model$X)) {
      ncol(as.matrix(model$X))
    } else {
      1L
    }

    if (length(p) != 1L || !is.finite(p) || p < 1L) {
      return(character(0))
    }

    if (p == 1L) {
      "mean"
    } else {
      c("mean", paste0("mean", seq_len(p - 1L)))
    }
  }

  regression_restrictions <- function(model1, model2, par1, par2, fixed1, fixed2) {
    X1 <- model1$X
    X2 <- model2$X

    if (same_value(X1, X2)) {
      return(list(
        different = FALSE,
        R = matrix(0, nrow = 0L, ncol = length(par1),
                   dimnames = list(NULL, names(par1))),
        r = numeric(0),
        reg1 = regression_parameter_names(model1),
        reg2 = regression_parameter_names(model2)
      ))
    }

    if (isTRUE(model1$bivariate) || isTRUE(model2$bivariate)) {
      stop(
        "Different design matrices are currently supported only for ",
        "univariate fits",
        call. = FALSE
      )
    }

    if (is.list(X1) || is.list(X2) || is.null(X1) || is.null(X2)) {
      stop(
        "The two fits have different design matrices that cannot be compared",
        call. = FALSE
      )
    }

    X1 <- as.matrix(X1)
    X2 <- as.matrix(X2)
    storage.mode(X1) <- storage.mode(X2) <- "double"

    if (nrow(X1) != nrow(X2)) {
      stop(
        "The unrestricted and restricted design matrices have different ",
        "numbers of rows",
        call. = FALSE
      )
    }

    p1 <- ncol(X1)
    p2 <- ncol(X2)

    if (p2 >= p1) {
      stop(
        "The restricted design matrix must contain fewer regression ",
        "parameters than the unrestricted design matrix",
        call. = FALSE
      )
    }

    reg1 <- regression_parameter_names(model1)
    reg2 <- regression_parameter_names(model2)

    if (length(reg1) != p1 || length(reg2) != p2) {
      stop(
        "The number of regression parameter names is inconsistent with X",
        call. = FALSE
      )
    }

    ## With different design matrices, the current implementation assumes
    ## that all regression coefficients in both models are estimated.
    if (!all(reg1 %in% names(par1)) || !all(reg2 %in% names(par2)) ||
        any(reg1 %in% names(fixed1)) || any(reg2 %in% names(fixed2))) {
      stop(
        "Regression nesting with different design matrices requires all ",
        "regression coefficients to be estimated in both models",
        call. = FALSE
      )
    }

    tol <- 1e-8
    rank1 <- qr(X1, tol = tol)$rank
    rank2 <- qr(X2, tol = tol)$rank

    if (rank1 != p1 || rank2 != p2) {
      stop(
        "The regression design matrices must have full column rank",
        call. = FALSE
      )
    }

    ## Find A such that X2 = X1 A. The restricted regression is nested if
    ## the columns of X2 belong to the column space of X1.
    A <- tryCatch(
      qr.solve(X1, X2, tol = tol),
      error = function(e) NULL
    )

    if (is.null(A)) {
      stop(
        "Unable to verify nesting of the regression design matrices",
        call. = FALSE
      )
    }

    A <- as.matrix(A)
    if (nrow(A) != p1) {
      A <- matrix(A, nrow = p1, ncol = p2)
    }

    residual <- X2 - X1 %*% A
    relative_error <- sqrt(sum(residual^2)) /
      max(1, sqrt(sum(X2^2)))

    if (!is.finite(relative_error) || relative_error > tol) {
      stop(
        "The restricted regression is not nested in the unrestricted ",
        "regression: col(X_restricted) is not contained in ",
        "col(X_unrestricted)",
        call. = FALSE
      )
    }

    if (qr(A, tol = tol)$rank != p2) {
      stop(
        "The restricted regression parametrization is rank deficient",
        call. = FALSE
      )
    }

    ## Under H0, beta_unrestricted belongs to col(A). An orthonormal basis
    ## of the complement of col(A) gives C beta_unrestricted = 0.
    QA <- qr.Q(qr(A, tol = tol), complete = TRUE)
    complement <- seq.int(p2 + 1L, p1)
    C <- t(QA[, complement, drop = FALSE])

    R <- matrix(
      0,
      nrow = nrow(C),
      ncol = length(par1),
      dimnames = list(
        paste0("regression_constraint_", seq_len(nrow(C))),
        names(par1)
      )
    )
    R[, reg1] <- C

    list(
      different = TRUE,
      R = R,
      r = setNames(rep(0, nrow(C)), rownames(R)),
      reg1 = reg1,
      reg2 = reg2
    )
  }

  check_common_fit_structure <- function(model1, model2) {
    ## The correlation family must always be identical. Individual
    ## correlation parameters may instead be fixed in the restricted fit.
    fields <- c(
      "corrmodel", "likelihood", "type", "copula",
      "distance", "grid", "n", "neighb", "maxdist", "maxtime",
      "p_neighb", "thin_method", "weighted", "bivariate",
      "spacetime", "radius", "local", "ns", "numpairs"
    )

    bad <- fields[!vapply(
      fields,
      function(nm) same_value(model1[[nm]], model2[[nm]]),
      logical(1)
    )]

    if (length(bad) > 0L) {
      stop(
        "The fitted models are not directly comparable; different fields: ",
        paste(bad, collapse = ", "),
        call. = FALSE
      )
    }

    data_fields <- c(
      "data", "coordx", "coordy", "coordz", "coordx_dyn"
    )

    bad_data <- data_fields[!vapply(
      data_fields,
      function(nm) same_value(model1[[nm]], model2[[nm]]),
      logical(1)
    )]

    if (!same_value(
      canonical_coordt(model1),
      canonical_coordt(model2)
    )) {
      bad_data <- c(bad_data, "coordt")
    }

    if (length(bad_data) > 0L) {
      stop(
        "The fitted models must use the same data and coordinates; ",
        "different fields: ", paste(bad_data, collapse = ", "),
        call. = FALSE
      )
    }
  }

  selector_restrictions <- function(parameter_names, null_values, par_names) {
    if (length(parameter_names) == 0L) {
      return(list(
        R = matrix(0, nrow = 0L, ncol = length(par_names),
                   dimnames = list(NULL, par_names)),
        r = numeric(0)
      ))
    }

    R <- matrix(
      0,
      nrow = length(parameter_names),
      ncol = length(par_names),
      dimnames = list(parameter_names, par_names)
    )

    R[cbind(seq_along(parameter_names), match(parameter_names, par_names))] <- 1

    list(
      R = R,
      r = unname(null_values[parameter_names])
    )
  }

  boundary_reasons <- function(parameter_names, null_values, marginal_rule) {
    reasons <- marginal_rule$boundary

    if (length(parameter_names) > 0L) {
      for (nm in parameter_names) {
        value <- as.numeric(null_values[nm])

        if (grepl("^nugget", nm) && isTRUE(all.equal(value, 0))) {
          reasons <- c(reasons, paste0(nm, " = 0"))
        }

        if (identical(nm, "sep") &&
            (isTRUE(all.equal(value, 0)) || isTRUE(all.equal(value, 1)))) {
          reasons <- c(reasons, paste0(nm, " = ", value))
        }

        if (grepl("^pcol", nm) && isTRUE(all.equal(abs(value), 1))) {
          reasons <- c(reasons, paste0(nm, " = ", value))
        }
      }
    }

    unique(reasons[nzchar(reasons)])
  }

  nesting_information <- function(model1, model2) {
    par1 <- as_named_numeric(model1$param)
    par2 <- as_named_numeric(model2$param)
    fixed1 <- as_named_numeric(model1$fixed)
    fixed2 <- as_named_numeric(model2$fixed)

    if (is.null(names(par1)) || anyDuplicated(names(par1))) {
      stop("The unrestricted model has invalid parameter names", call. = FALSE)
    }

    if (is.null(names(par2)) || anyDuplicated(names(par2))) {
      stop("The restricted model has invalid parameter names", call. = FALSE)
    }

    if (length(par1) <= length(par2)) {
      stop(
        "The first model must be the unrestricted model and must have more ",
        "estimated parameters than the second model",
        call. = FALSE
      )
    }

    marginal_rule <- marginal_nesting_rule(model1, model2)

    regression <- regression_restrictions(
      model1, model2, par1, par2, fixed1, fixed2
    )

    if (regression$different) {
      nonreg1 <- setdiff(names(par1), regression$reg1)
      nonreg2 <- setdiff(names(par2), regression$reg2)

      if (!all(nonreg2 %in% nonreg1)) {
        stop(
          "Models are not nested: the restricted model estimates non-regression ",
          "parameters not estimated by the unrestricted model: ",
          paste(setdiff(nonreg2, nonreg1), collapse = ", "),
          call. = FALSE
        )
      }

      tested_parameters <- setdiff(nonreg1, nonreg2)
    } else {
      if (!all(names(par2) %in% names(par1))) {
        stop(
          "Models are not nested: the restricted model estimates parameters ",
          "not estimated by the unrestricted model: ",
          paste(setdiff(names(par2), names(par1)), collapse = ", "),
          call. = FALSE
        )
      }

      tested_parameters <- setdiff(names(par1), names(par2))
    }

    if (!marginal_rule$same_model) {
      required <- names(marginal_rule$null)
      missing_required <- setdiff(required, tested_parameters)

      if (length(missing_required) > 0L) {
        stop(
          "The unrestricted marginal model must estimate the parameter(s) ",
          "defining the reduction to the restricted model: ",
          paste(missing_required, collapse = ", "),
          call. = FALSE
        )
      }
    }

    null_values <- fixed2[tested_parameters]

    rule_names <- intersect(tested_parameters, names(marginal_rule$null))
    if (length(rule_names) > 0L) {
      both <- rule_names[rule_names %in% names(fixed2)]

      if (length(both) > 0L &&
          !same_value(fixed2[both], marginal_rule$null[both])) {
        stop(
          "The fixed values in the restricted fit are inconsistent with ",
          "the marginal nesting relation",
          call. = FALSE
        )
      }

      null_values[rule_names] <- marginal_rule$null[rule_names]
    }

    missing_null <- tested_parameters[
      !tested_parameters %in% names(null_values) |
        is.na(null_values[tested_parameters])
    ]

    if (length(missing_null) > 0L) {
      stop(
        "The restricted model must fix every additional non-regression ",
        "parameter of the unrestricted model. Missing null value(s): ",
        paste(missing_null, collapse = ", "),
        call. = FALSE
      )
    }

    null_values <- null_values[tested_parameters]

    if (any(!is.finite(null_values))) {
      stop(
        "The null parameter values must be finite. Limiting models such as ",
        "StudentT -> Gaussian are not supported by the standard tests",
        call. = FALSE
      )
    }

    common_fixed <- intersect(names(fixed1), names(fixed2))

    if (length(common_fixed) > 0L &&
        !same_value(fixed1[common_fixed], fixed2[common_fixed])) {
      stop(
        "Common fixed parameters must have identical values in the two models",
        call. = FALSE
      )
    }

    selector <- selector_restrictions(
      tested_parameters,
      null_values,
      names(par1)
    )

    R <- rbind(regression$R, selector$R)
    r <- c(regression$r, selector$r)

    if (nrow(R) < 1L) {
      stop("Models are not nested or have the same number of parameters", call. = FALSE)
    }

    df_expected <- length(par1) - length(par2)

    if (nrow(R) != df_expected) {
      stop(
        "The number of identified restrictions (", nrow(R),
        ") does not agree with the difference in estimated parameters (",
        df_expected, ")",
        call. = FALSE
      )
    }

    rank_R <- qr(t(R), tol = 1e-8)$rank
    if (rank_R != nrow(R)) {
      stop("The null restrictions are linearly dependent", call. = FALSE)
    }

    reasons <- boundary_reasons(
      tested_parameters,
      null_values,
      marginal_rule
    )

    list(
      par1 = par1,
      par2 = par2,
      R = R,
      r = as.numeric(r),
      df = nrow(R),
      tested = tested_parameters,
      null_values = null_values,
      boundary = reasons
    )
  }

  check_same_composite_pairs <- function(model1, model2) {
    have_pairs1 <- !is.null(model1$rowidx) && !is.null(model1$colidx)
    have_pairs2 <- !is.null(model2$rowidx) && !is.null(model2$colidx)

    if (have_pairs1 && have_pairs2) {
      same_pairs <- identical(model1$rowidx, model2$rowidx) &&
        identical(model1$colidx, model2$colidx)

      if (!same_pairs) {
        stop(
          "Composite likelihood-ratio tests require the same retained pairs ",
          "in the unrestricted and restricted fits",
          call. = FALSE
        )
      }

      return(invisible(TRUE))
    }

    if (is_stochastic_thinning(model1) || is_stochastic_thinning(model2)) {
      stop(
        "The fits use stochastic thinning, but the retained pair indices are ",
        "not available to verify that the same pairs were used",
        call. = FALSE
      )
    }

    invisible(TRUE)
  }

  compute_statistic <- function(model1, model2, statistic, nesting) {
    likelihood <- as.character(model1$likelihood)
    is_composite <- likelihood %in% composite_likelihoods
    is_full <- identical(likelihood, "Full")

    if (!is_composite && !is_full) {
      stop("Unsupported likelihood type: ", likelihood, call. = FALSE)
    }

    if (statistic == "Wilks" && !is_full) {
      stop("'Wilks' is available only for full likelihood fits", call. = FALSE)
    }

    if (statistic == "WilksS" && !is_composite) {
      stop("'WilksS' is available only for composite likelihood fits", call. = FALSE)
    }

    R_null <- nesting$R
    theta <- as.numeric(R_null %*% nesting$par1 - nesting$r)
    df <- nesting$df

    if (statistic == "Wald") {
      V <- ensure_named_matrix(
        model1$varcov,
        names(nesting$par1),
        "The covariance matrix of the unrestricted model"
      )

      V_null <- R_null %*% V %*% t(R_null)
      V_null_inv <- safe_solve(
        V_null,
        "The covariance matrix for the null restrictions"
      )

      W <- as.numeric(crossprod(theta, V_null_inv %*% theta))
      nu <- df
    }

    if (statistic == "Wilks") {
      W <- 2 * (
        finite_loglik(model1, "The unrestricted model") -
          finite_loglik(model2, "The restricted model")
      )
      nu <- df
    }

    if (statistic == "WilksS") {
      check_same_composite_pairs(model1, model2)

      W_raw <- 2 * (
        finite_loglik(model1, "The unrestricted model") -
          finite_loglik(model2, "The restricted model")
      )

      H <- ensure_named_matrix(
        model1$sensmat,
        names(nesting$par1),
        "The sensitivity matrix of the unrestricted model"
      )

      V <- ensure_named_matrix(
        model1$varcov,
        names(nesting$par1),
        "The Godambe covariance matrix of the unrestricted model"
      )

      ## For general linear restrictions R theta = r, the efficient
      ## sensitivity and Godambe covariance of the restrictions are
      ## H_eff = (R H^{-1} R')^{-1} and V_eff = R V R'.
      H_inv <- safe_solve(H, "The sensitivity matrix")
      H_restriction_inv <- R_null %*% H_inv %*% t(R_null)
      H_eff <- safe_solve(
        H_restriction_inv,
        "The efficient sensitivity matrix for the null restrictions"
      )

      V_eff <- R_null %*% V %*% t(R_null)
      V_eff <- (V_eff + t(V_eff)) / 2

      ## H_eff %*% V_eff has positive real eigenvalues. The symmetric
      ## similar matrix below is numerically more stable.
      chol_H <- tryCatch(chol(H_eff), error = function(e) NULL)

      if (is.null(chol_H)) {
        stop(
          "The efficient sensitivity matrix is not positive definite",
          call. = FALSE
        )
      }

      lambda_mat <- chol_H %*% V_eff %*% t(chol_H)
      lambda_mat <- (lambda_mat + t(lambda_mat)) / 2
      lambda <- eigen(lambda_mat, symmetric = TRUE, only.values = TRUE)$values

      tol <- max(1, max(abs(lambda))) * sqrt(.Machine$double.eps)

      if (any(lambda < -tol) || any(!is.finite(lambda))) {
        stop("Invalid eigenvalues in the WilksS adjustment", call. = FALSE)
      }

      lambda[lambda < 0] <- 0

      if (!all(lambda > 0)) {
        stop(
          "Non-positive eigenvalues in the WilksS adjustment; ",
          "the tested restrictions may be weakly identified",
          call. = FALSE
        )
      }

      sum_lambda <- sum(lambda)
      sum_lambda2 <- sum(lambda^2)

      nu <- sum_lambda^2 / sum_lambda2
      scale_factor <- sum_lambda2 / sum_lambda
      W <- W_raw / scale_factor
    }

    tol_W <- 100 * .Machine$double.eps *
      max(1, abs(finite_loglik(model1, "The unrestricted model")))

    if (W < -tol_W) {
      stop(
        "The test statistic is negative. Check optimizer convergence and ",
        "the ordering of unrestricted and restricted models",
        call. = FALSE
      )
    }

    W <- max(0, as.numeric(W))

    if (!is.finite(W) || !is.finite(nu) || nu <= 0) {
      stop("The test statistic or its degrees of freedom is invalid", call. = FALSE)
    }

    list(W = W, nu = as.numeric(nu))
  }

  ## ------------------------------------------------------------------
  ## Input checks and model collection
  ## ------------------------------------------------------------------
  if (missing(object1) || missing(object2)) {
    stop("Models one and two must be specified", call. = FALSE)
  }

  if (missing(statistic) || !is.character(statistic) || length(statistic) != 1L) {
    stop("statistic must be one of: Wald, Wilks, WilksS", call. = FALSE)
  }

  if (!statistic %in% valid_statistics) {
    stop(
      "Unknown statistic. Available tests are: ",
      paste(valid_statistics, collapse = ", "),
      call. = FALSE
    )
  }

  model_calls <- c(
    list(substitute(object1), substitute(object2)),
    as.list(substitute(list(...)))[-1L]
  )

  model_names <- vapply(
    model_calls,
    function(x) paste(deparse(x), collapse = ""),
    character(1)
  )

  models <- lapply(model_calls, eval, envir = parent.frame())

  if (!all(vapply(models, inherits, logical(1), what = "GeoFit"))) {
    stop("GeoTests can be used only with 'GeoFit' objects", call. = FALSE)
  }

  nummod <- length(models)
  numparam <- vapply(models, function(x) length(as_named_numeric(x$param)), integer(1))
  df <- nu <- W <- pvalue <- rep(NA_real_, nummod - 1L)

  ## ------------------------------------------------------------------
  ## Adjacent nested-model comparisons
  ## ------------------------------------------------------------------
  for (i in 2:nummod) {
    unrestricted <- models[[i - 1L]]
    restricted <- models[[i]]

    check_common_fit_structure(unrestricted, restricted)
    nesting <- nesting_information(unrestricted, restricted)

    if (length(nesting$boundary) > 0L) {
      warning(
        "Some null values lie on the boundary of the parameter space (",
        paste(nesting$boundary, collapse = "; "),
        "). The usual chi-square reference distribution may not be valid.",
        call. = FALSE
      )
    }

    stat <- compute_statistic(
      unrestricted,
      restricted,
      statistic,
      nesting
    )

    j <- i - 1L
    df[j] <- nesting$df
    nu[j] <- stat$nu
    W[j] <- stat$W
    pvalue[j] <- stats::pchisq(W[j], df = nu[j], lower.tail = FALSE)
  }

  ## ------------------------------------------------------------------
  ## Output table
  ## ------------------------------------------------------------------
  table <- data.frame(
    "Num.Par" = numparam,
    "Diff.Par" = c(NA_real_, df),
    "Df" = c(NA_real_, nu),
    "Chisq" = c(NA_real_, W),
    "Pr(>chisq)" = c(NA_real_, pvalue),
    check.names = FALSE
  )

  rownames(table) <- model_names

  structure(
    table,
    heading = "Statistical Hypothesis Test Table\n",
    class = c("data.frame")
  )
}

####################################################
### Internal utilities for mean specifications
####################################################

.GeoMean_names <- function(p, prefix = "mean") {
  if (length(p) != 1L || is.na(p) || p < 1L || p != as.integer(p)) {
    stop("The number of mean coefficients must be a positive integer.", call. = FALSE)
  }
  p <- as.integer(p)
  if (p == 1L) return(prefix)
  c(prefix, paste0(prefix, seq_len(p - 1L)))
}


.GeoMean_supplied_names <- function(param, prefix = "mean") {
  if (is.null(names(param))) return(character(0))
  pat <- paste0("^", prefix, "[0-9]*$")
  unique(names(param)[grepl(pat, names(param))])
}

.GeoMean_zero <- function(param, p, prefix = "mean") {
  for (nm in .GeoMean_names(p, prefix)) param[[nm]] <- 0
  param
}

.GeoMean_intercept_param <- function(param, value = 0, prefix = "mean") {
  for (nm in .GeoMean_supplied_names(param, prefix)) param[[nm]] <- NULL
  param[[prefix]] <- value
  param
}

.GeoMean_external <- function(param, n = NULL, name = "mean") {
  ## `fixed` and other parameter containers are not guaranteed to remain
  ## lists.  Several fitted objects store them as named atomic vectors, for
  ## which param[[name]] throws "subscript out of bounds" when the requested
  ## name is absent.  Absence of `mean` means that no external mean was
  ## supplied and must therefore return NULL.
  if (is.null(param) || !length(param) || is.null(names(param)) ||
      !(name %in% names(param))) {
    return(NULL)
  }

  value <- if (is.list(param)) {
    param[[name]]
  } else {
    param[names(param) == name]
  }
  if (is.null(value) || !length(value)) return(NULL)

  out <- as.numeric(unlist(value, use.names = FALSE))
  if (length(out) <= 1L) return(NULL)
  if (any(!is.finite(out))) {
    stop("The external mean vector contains non-finite values.", call. = FALSE)
  }
  if (!is.null(n) && length(out) != n) {
    stop("The external mean vector must have one value per observation.", call. = FALSE)
  }
  out
}

.GeoMean_beta <- function(param, p, prefix = "mean") {
  nm <- .GeoMean_names(p, prefix)
  if (is.null(names(param))) {
    stop("Mean coefficients must be supplied as named parameters.", call. = FALSE)
  }
  missing_nm <- nm[!nm %in% names(param)]
  if (length(missing_nm)) {
    stop("Missing mean coefficient(s): ", paste(missing_nm, collapse = ", "), call. = FALSE)
  }
  vals <- lapply(nm, function(z) param[[z]])
  bad <- vapply(vals, function(z) length(z) != 1L || !is.numeric(z) || !is.finite(z), logical(1L))
  if (any(bad)) {
    stop("Each mean coefficient must be a finite numeric scalar.", call. = FALSE)
  }
  out <- as.numeric(unlist(vals, use.names = FALSE))
  names(out) <- nm
  out
}

.GeoMean_design <- function(X, n, p = NULL, name = "X", allow_null = TRUE) {
  if (length(n) != 1L || is.na(n) || n < 1L || n != as.integer(n)) {
    stop("Invalid number of rows requested for ", name, ".", call. = FALSE)
  }
  n <- as.integer(n)

  if (is.null(X)) {
    if (!allow_null) stop(name, " is required.", call. = FALSE)
    if (!is.null(p) && p != 1L) {
      stop(name, " is required when more than one mean coefficient is used.", call. = FALSE)
    }
    return(matrix(1, nrow = n, ncol = 1L))
  }

  if (is.data.frame(X)) X <- as.matrix(X)
  if (is.list(X) && !is.matrix(X)) X <- do.call(rbind, lapply(X, as.matrix))

  if (is.vector(X) && !is.list(X)) {
    if (!is.null(p) && p == 1L && length(X) == n) {
      X <- matrix(X, nrow = n, ncol = 1L)
    } else if (n == 1L && (is.null(p) || length(X) == p)) {
      X <- matrix(X, nrow = 1L)
    } else {
      stop(name, " supplied as a vector is ambiguous; use an explicit matrix.", call. = FALSE)
    }
  } else {
    X <- as.matrix(X)
  }

  if (!is.numeric(X) || any(!is.finite(X))) {
    stop(name, " must be a finite numeric matrix.", call. = FALSE)
  }
  if (nrow(X) != n) {
    stop(name, " must have ", n, " rows.", call. = FALSE)
  }
  if (!is.null(p) && ncol(X) != p) {
    stop(name, " must have ", p, " columns, one for each mean coefficient.", call. = FALSE)
  }
  unname(X)
}

.GeoMean_vector <- function(param, X = NULL, n, external = NULL,
                            prefix = "mean", xname = "X") {
  if (!is.null(external)) {
    external <- as.numeric(external)
    if (length(external) == 1L) external <- rep(external, n)
    if (length(external) != n || any(!is.finite(external))) {
      stop("The fixed mean must contain one finite value per location.", call. = FALSE)
    }
    return(external)
  }

  if (is.null(X)) {
    X <- matrix(1, nrow = n, ncol = 1L)
    p <- 1L
  } else {
    X <- .GeoMean_design(X, n = n, name = xname)
    p <- ncol(X)
  }
  beta <- .GeoMean_beta(param, p = p, prefix = prefix)
  as.numeric(X %*% beta)
}

.GeoMean_validate_fit <- function(start, fixed, X, nobs) {
  start_mean <- .GeoMean_supplied_names(start)
  fixed_mean <- .GeoMean_supplied_names(fixed)
  duplicated_mean <- intersect(start_mean, fixed_mean)
  if (length(duplicated_mean)) {
    stop(
      "A mean coefficient cannot be supplied in both start and fixed: ",
      paste(duplicated_mean, collapse = ", "), ".",
      call. = FALSE
    )
  }

  external <- .GeoMean_external(fixed, n = nobs)
  if (!is.null(external)) {
    if (!is.null(X)) {
      stop("A site-specific fixed mean and X cannot be used together.", call. = FALSE)
    }
    if (length(start_mean)) {
      stop(
        "A site-specific fixed mean cannot be combined with estimated mean coefficients in start.",
        call. = FALSE
      )
    }
    extra_fixed <- setdiff(fixed_mean, "mean")
    if (length(extra_fixed)) {
      stop(
        "A site-specific fixed mean cannot be combined with additional mean coefficients: ",
        paste(extra_fixed, collapse = ", "), ".",
        call. = FALSE
      )
    }
    return(list(X = NULL, external = external))
  }

  supplied <- unique(c(start_mean, fixed_mean))

  if (!is.null(X)) {
    X <- .GeoMean_design(X, n = nobs, name = "X")
    expected <- .GeoMean_names(ncol(X))
    unexpected <- setdiff(supplied, expected)
    if (length(unexpected)) {
      stop(
        "For ncol(X) = ", ncol(X), ", regression coefficients can only be named: ",
        paste(expected, collapse = ", "), ". Unexpected: ",
        paste(unexpected, collapse = ", "), ".",
        call. = FALSE
      )
    }
    if (length(supplied)) {
      all_values <- c(start, fixed)
      vals <- lapply(supplied, function(nm) all_values[[nm]])
      bad <- vapply(
        vals,
        function(z) length(z) != 1L || !is.numeric(z) || !is.finite(z),
        logical(1L)
      )
      if (any(bad)) {
        stop("Supplied mean coefficients must be finite numeric scalars.", call. = FALSE)
      }
    }
  } else {
    unexpected <- setdiff(supplied, "mean")
    if (length(unexpected)) {
      stop(
        "Multiple mean coefficients require an explicit design matrix X; ",
        "without X the model is intercept-only and uses parameter 'mean'.",
        call. = FALSE
      )
    }
    if (length(supplied)) {
      all_values <- c(start, fixed)
      value <- all_values[["mean"]]
      if (length(value) != 1L || !is.numeric(value) || !is.finite(value)) {
        stop("The intercept parameter 'mean' must be a finite numeric scalar.", call. = FALSE)
      }
    }
  }

  list(X = X, external = NULL)
}

GeoNeighIndex <- function(coordx = NULL, coordy = NULL, coordz = NULL, coordt = NULL,
                          coordx_dyn = NULL,
                          distance = "Eucl", neighb = 4, maxdist = NULL,
                          maxtime = 1, radius = 1, bivariate = FALSE,
                          p_neighb = 1, thin_method = "bernoulli", check.duplicates = FALSE)
{
  ##############################################################
  ## Checks
  ##############################################################

  if (!is.numeric(neighb) ||
      !(length(neighb) %in% c(1L, 3L)) ||
      any(!is.finite(neighb)) ||
      any(neighb <= 0) ||
      any(neighb %% 1 != 0)) {
    stop("neighb must be a positive integer, or a length-3 positive integer vector in the bivariate case.")
  }

  if (!bivariate && length(neighb) != 1L) {
    stop("neighb can have length 3 only when bivariate = TRUE.")
  }

  K <- as.integer(neighb)

  .GeoCheckDuplicateLocations(
    coordx = coordx, coordy = coordy, coordz = coordz, coordt = coordt,
    coordx_dyn = coordx_dyn,
    spacetime = !isTRUE(bivariate) && !is.null(coordt),
    bivariate = isTRUE(bivariate), grid = FALSE,
    context = "GeoNeighIndex", check.duplicates = check.duplicates
  )

  ## Canonical thinning-method names.
  ## Accepted inputs include "bernoulli", "Bernoulli",
  ## "fixedbudget", and "FixedBudget". The old aliases
  ## "targetbalanced" and "TargetBalanced" are also accepted
  ## for backward compatibility and mapped to "FixedBudget".
  ## Internally we use only "Bernoulli" and "FixedBudget".
  if (!is.character(thin_method) || length(thin_method) != 1L || is.na(thin_method)) {
    stop("thin_method must be a character scalar.")
  }

  thin_method <- tolower(thin_method)

  thin_method <- match.arg(
    thin_method,
    c("bernoulli", "fixedbudget", "targetbalanced")
  )

  thin_method <- switch(
    thin_method,
    bernoulli = "Bernoulli",
    fixedbudget = "FixedBudget",
    targetbalanced = "FixedBudget"
  )
 
  indices <- function(nn_idx, nn_dists, drop_first = FALSE, x_equals_y = FALSE) {

    if (drop_first) {
      n <- nrow(nn_idx)
      kout <- max(0L, ncol(nn_idx) - 1L)

      ## Fast normal case: nabor returns the query point itself first.
      ## With duplicated coordinates/ties this ordering is not guaranteed,
      ## so remove the actual target index rather than blindly dropping col 1.
      if (isTRUE(x_equals_y) &&
          length(nn_idx) &&
          all(nn_idx[, 1L] == seq_len(n))) {
        nn_idx   <- nn_idx[, -1L, drop = FALSE]
        nn_dists <- nn_dists[, -1L, drop = FALSE]
      } else if (isTRUE(x_equals_y)) {
        out_idx <- matrix(0L, nrow = n, ncol = kout)
        out_dst <- matrix(NA_real_, nrow = n, ncol = kout)
        if (kout > 0L) {
          for (ii in seq_len(n)) {
            keep <- which(nn_idx[ii, ] > 0L & nn_idx[ii, ] != ii)
            if (length(keep)) {
              keep <- keep[seq_len(min(kout, length(keep)))]
              out_idx[ii, seq_along(keep)] <- nn_idx[ii, keep]
              out_dst[ii, seq_along(keep)] <- nn_dists[ii, keep]
            }
          }
        }
        nn_idx <- out_idx
        nn_dists <- out_dst
      } else {
        nn_idx   <- nn_idx[, -1L, drop = FALSE]
        nn_dists <- nn_dists[, -1L, drop = FALSE]
      }
    }

    k <- ncol(nn_idx)

    rowidx <- rep(seq_len(nrow(nn_idx)), times = k)
    colidx <- as.vector(nn_idx)
    lags   <- as.vector(nn_dists)

    list(
      rowidx = rowidx,
      colidx = colidx,
      lags = lags
    )
  }

  ##############################################################
  ## Spatial nearest-neighbor helper
  ##############################################################

  nn2Geo <- function(x, y, K = 1, distance = 0, maxdist = NULL, radius = 1,
                     drop_first = FALSE, x_equals_y = FALSE) {

    if (!is.matrix(x) || !is.matrix(y)) {
      stop("x and y must be matrices")
    }

    n_x <- nrow(x)
    n_y <- nrow(y)

    K <- as.integer(K)
    K <- min(K, n_x, n_y)

    sphere_xyz <- function(coords) {
      lon <- coords[, 1L] * pi / 180
      lat <- coords[, 2L] * pi / 180
      clat <- cos(lat)
      cbind(clat * cos(lon), clat * sin(lon), sin(lat))
    }

    if (distance %in% c(1, 2)) {
      if (!requireNamespace("fields", quietly = TRUE)) {
        stop("Package 'fields' is required by GeoNeighIndex() for geodesic or chordal distance.",
             call. = FALSE)
      }
      ## Select spherical neighbours in 3D unit-sphere coordinates.  Chordal
      ## distance is monotone in great-circle distance, so this produces the
      ## correct NN ordering globally (including dateline/polar cases), unlike
      ## longitude/latitude or a planar projection.
      x_nn <- sphere_xyz(x)
      y_nn <- sphere_xyz(y)

      if (is.null(maxdist)) {
        nearest <- nabor::knn(x_nn, y_nn, k = K)
      } else {
        if (!is.finite(radius) || radius <= 0) stop("radius must be positive")
        radius_nn <- if (distance == 2) {
          2 * sin(0.5 * pmin(pi, maxdist / radius))
        } else {
          pmin(2, maxdist / radius)
        }
        nearest <- nabor::knn(x_nn, y_nn, k = K, radius = radius_nn)
      }
    } else if (is.null(maxdist)) {
      nearest <- nabor::knn(x, y, k = K)
    } else {
      nearest <- nabor::knn(x, y, k = K, radius = maxdist)
    }

    if (distance %in% c(1, 2)) {

      dists_mat <- matrix(NA_real_, nrow = n_y, ncol = K)

      for (i in seq_len(n_y)) {

        idx <- nearest$nn.idx[i, ]
        valid <- idx > 0

        if (!any(valid)) {
          next
        }

        di <- fields::rdist.earth.vec(
          x1 = matrix(y[i, ], nrow = 1),
          x2 = x[idx[valid], , drop = FALSE],
          miles = FALSE,
          R = 1
        )

        dists_mat[i, seq_len(length(di))] <- di
      }

      if (isTRUE(x_equals_y)) {
        dists_mat[, 1] <- 0
      }

      if (distance == 2) {
        nearest$nn.dists <- radius * dists_mat
      } else {
        nearest$nn.dists <- 2 * radius * sin(0.5 * dists_mat)
      }
    }

    sol <- indices(
      nearest$nn.idx,
      nearest$nn.dists,
      drop_first = drop_first,
      x_equals_y = x_equals_y
    )

    if (!is.null(maxdist)) {

      sel <- sol$colidx > 0

      return(list(
        lags = sol$lags[sel],
        rowidx = sol$rowidx[sel],
        colidx = sol$colidx[sel]
      ))

    } else {

      return(list(
        lags = sol$lags,
        rowidx = sol$rowidx,
        colidx = sol$colidx
      ))
    }
  }

  ##############################################################
  ## Space-time index
  ##############################################################

  spacetime_index <- function(coords, coordx_dyn = NULL, N, K = 4,
                              coordt = NULL, numtime,
                              maxtime = 1, maxdist = NULL,
                              distance = "Eucl", radius = 1) {

    if (is.null(coordt)) {
      stop("coordt is required for spacetime_index().")
    }

    if (numtime <= 1L) {
      stop("numtime must be >= 2.")
    }

    if (!is.numeric(K) || length(K) != 1L || !is.finite(K) || K < 1) {
      stop("K must be a positive integer.")
    }

    K <- as.integer(K)

    if (!is.numeric(maxtime) || length(maxtime) != 1L ||
        is.na(maxtime) || maxtime < 0) {
      stop("maxtime must be a non-negative numeric scalar.")
    }

    if (is.null(coordx_dyn)) {

      ns <- rep.int(as.integer(N), numtime)
      offset <- as.integer(N) * (0:(numtime - 1L))

      getX <- function(t) coords

    } else {

      if (!is.list(coordx_dyn) || length(coordx_dyn) != numtime) {
        stop("coordx_dyn must be a list of length = numtime.")
      }

      ns <- vapply(coordx_dyn, nrow, integer(1))

      if (any(ns <= 0L)) {
        stop("Each coordx_dyn[[t]] must have at least 1 row.")
      }

      offset <- c(0L, cumsum(ns))[seq_len(numtime)]
      offset <- as.integer(offset)

      getX <- function(t) coordx_dyn[[t]]
    }

    ct <- as.numeric(coordt)

    if (length(ct) != numtime) {
      stop("length(coordt) must equal numtime.")
    }

    if (any(!is.finite(ct))) {
      stop("coordt must contain finite values.")
    }

    glob <- function(idx_local, t) {
      as.integer(idx_local + offset[t])
    }

    ##############################################################
    ## Temporal-pair helper
    ##
    ## maxtime is interpreted as a temporal-distance threshold in
    ## the scale of coordt. Thus all pairs with temporal lag
    ## 0 < |coordt[to] - coordt[from]| <= maxtime are retained.
    ## For regularly spaced time points coordt = 1, ..., n and
    ## maxtime = q, this coincides with the old behavior based on
    ## index lags 1, ..., q.
    ##############################################################

    make_time_pairs <- function(ct, maxtime) {

      nt <- length(ct)

      if (nt <= 1L || maxtime <= 0) {
        return(data.frame(
          from = integer(0),
          to = integer(0),
          lag_time = numeric(0)
        ))
      }

      out <- vector("list", nt - 1L)
      out_i <- 1L

      for (from in seq_len(nt - 1L)) {

        to <- (from + 1L):nt
        lag_time <- abs(ct[to] - ct[from])
        ttol <- 100 * .Machine$double.eps *
          pmax(1, abs(ct[to]), abs(ct[from]))

        keep <- is.finite(lag_time) & lag_time > 0 & lag_time <= maxtime + ttol

        if (!any(keep)) {
          next
        }

        out[[out_i]] <- data.frame(
          from = rep.int(from, sum(keep)),
          to = to[keep],
          lag_time = lag_time[keep]
        )

        out_i <- out_i + 1L
      }

      if (out_i == 1L) {
        return(data.frame(
          from = integer(0),
          to = integer(0),
          lag_time = numeric(0)
        ))
      }

      do.call(rbind, out[seq_len(out_i - 1L)])
    }

    knn_dir <- function(ref, query, K,
                        drop_self = FALSE,
                        x_equals_y = FALSE) {

      inf <- nn2Geo(
        ref,
        query,
        if (drop_self) K + 1L else K,
        distance,
        maxdist,
        radius,
        drop_first = drop_self,
        x_equals_y = x_equals_y
      )

      if (length(inf$rowidx) == 0L) {
        return(list(
          tar = integer(0),
          nei = integer(0),
          d = numeric(0)
        ))
      }

      tar <- inf$rowidx
      nei <- inf$colidx
      d   <- inf$lags

      ok <- is.finite(nei) &
        is.finite(tar) &
        nei > 0 &
        tar > 0 &
        is.finite(d)

      list(
        tar = tar[ok],
        nei = nei[ok],
        d = d[ok]
      )
    }

    same_location_pairs <- function(ref, query) {
      if (ncol(ref) != ncol(query) || !nrow(ref) || !nrow(query)) {
        return(list(tar = integer(0L), nei = integer(0L)))
      }

      scale <- suppressWarnings(max(abs(c(ref, query)), na.rm = TRUE))
      if (!is.finite(scale) || scale < 1) scale <- 1
      tol <- 100 * .Machine$double.eps * scale

      available <- seq_len(nrow(ref))
      tar <- integer(0L)
      nei <- integer(0L)

      for (i in seq_len(nrow(query))) {
        if (!length(available)) break
        delta <- abs(sweep(ref[available, , drop = FALSE], 2L,
                           query[i, ], FUN = "-"))
        dmax <- apply(delta, 1L, max)
        jpos <- which.min(dmax)
        if (length(jpos) && is.finite(dmax[jpos]) && dmax[jpos] <= tol) {
          tar <- c(tar, i)
          nei <- c(nei, available[jpos])
          available <- available[-jpos]
        }
      }

      list(tar = tar, nei = nei)
    }

    out <- list()
    out_i <- 1L

    ##############################################################
    ## 1) Spatial within-time
    ##############################################################

    for (t in seq_len(numtime)) {

      Xt <- getX(t)

      inf <- knn_dir(
        ref = Xt,
        query = Xt,
        K = K,
        drop_self = TRUE,
        x_equals_y = TRUE
      )

      if (length(inf$tar) > 0L) {

        out[[out_i]] <- cbind(
          glob(inf$tar, t),
          glob(inf$nei, t),
          rep.int(0, length(inf$tar)),
          as.numeric(inf$d)
        )

        out_i <- out_i + 1L
      }
    }

    time_pairs <- make_time_pairs(ct, maxtime)

    if (nrow(time_pairs) == 0L) {

      if (!length(out)) return(matrix(numeric(0L), nrow = 0L, ncol = 4L))
      sol <- do.call(rbind, out)

      ok <- is.finite(sol[, 1]) &
        is.finite(sol[, 2]) &
        sol[, 1] > 0 &
        sol[, 2] > 0

      return(sol[ok, , drop = FALSE])
    }

    ##############################################################
    ## 2) Pure temporal same-site
    ##############################################################

    for (rr in seq_len(nrow(time_pairs))) {

      t_from <- time_pairs$from[rr]
      t_to   <- time_pairs$to[rr]

      if (is.null(coordx_dyn)) {
        bb <- min(ns[t_from], ns[t_to])
        if (bb <= 0L) next
        same <- list(tar = seq_len(bb), nei = seq_len(bb))
      } else {
        same <- same_location_pairs(getX(t_from), getX(t_to))
        bb <- length(same$tar)
        if (bb <= 0L) next
      }

      out[[out_i]] <- cbind(
        glob(same$tar, t_to),
        glob(same$nei, t_from),
        rep.int(time_pairs$lag_time[rr], bb),
        rep.int(0, bb)
      )

      out_i <- out_i + 1L
    }

    ##############################################################
    ## 3) Spatio-temporal cross-time
    ##############################################################

    for (rr in seq_len(nrow(time_pairs))) {

      t_from <- time_pairs$from[rr]
      t_to   <- time_pairs$to[rr]

      Xn <- getX(t_from)
      Xt <- getX(t_to)

      inf <- knn_dir(
        ref = Xn,
        query = Xt,
        K = K,
        drop_self = FALSE,
        x_equals_y = FALSE
      )

      if (length(inf$tar) > 0L) {

        sel <- is.finite(inf$d) & (inf$d > 0)

        if (any(sel)) {

          out[[out_i]] <- cbind(
            glob(inf$tar[sel], t_to),
            glob(inf$nei[sel], t_from),
            rep.int(time_pairs$lag_time[rr], sum(sel)),
            as.numeric(inf$d[sel])
          )

          out_i <- out_i + 1L
        }
      }
    }

    if (!length(out)) return(matrix(numeric(0L), nrow = 0L, ncol = 4L))
    sol <- do.call(rbind, out)

    ok <- is.finite(sol[, 1]) &
      is.finite(sol[, 2]) &
      sol[, 1] > 0 &
      sol[, 2] > 0

    sol[ok, , drop = FALSE]
  }

  ##############################################################
  ## Bivariate index
  ##############################################################

  bivariate_index <- function(coords, coordx_dyn = NULL, N, K = 4,
                              maxdist, distance, radius) {

    ## Build every bivariate pair block with the same five-column schema.
    ## In particular, an empty block must be a 0 x 5 matrix: base::cbind()
    ## would otherwise drop zero-length vectors and retain only scalar
    ## columns, causing rbind() failures when no positive-lag cross pair
    ## exists (for example, coincident grids with neighb = 1).
    pair_block <- function(rowidx, colidx, first, second, lags) {

      n <- length(rowidx)

      if (length(colidx) != n || length(lags) != n) {
        stop("Internal error: inconsistent bivariate pair-block lengths.")
      }

      if (n == 0L) {
        out <- matrix(numeric(0L), nrow = 0L, ncol = 5L)
        colnames(out) <- c("rowidx", "colidx", "first", "second", "lag")
        return(out)
      }

      out <- cbind(
        as.integer(rowidx),
        as.integer(colidx),
        rep.int(as.integer(first), n),
        rep.int(as.integer(second), n),
        as.numeric(lags)
      )

      colnames(out) <- c("rowidx", "colidx", "first", "second", "lag")
      out
    }

    if (length(K) == 3L) {
      K1 <- K[1]
      K2 <- K[2]
      K3 <- K[3]
    } else {
      K1 <- K2 <- K3 <- K
    }

    if (length(maxdist) == 3L) {
      maxdist1 <- maxdist[1]
      maxdist2 <- maxdist[2]
      maxdist3 <- maxdist[3]
    } else {
      maxdist1 <- maxdist2 <- maxdist3 <- maxdist
    }

    auto_tol <- function(A, B) {

      scaleA <- suppressWarnings(max(abs(range(A, finite = TRUE))))
      scaleB <- suppressWarnings(max(abs(range(B, finite = TRUE))))

      scale <- max(scaleA, scaleB)

      if (!is.finite(scale) || scale <= 0) {
        scale <- 1
      }

      100 * .Machine$double.eps * scale
    }

    ##############################################################
    ## CASE 1: coordx_dyn NULL
    ##############################################################

    if (is.null(coordx_dyn)) {

      n_half <- as.integer(N / 2)
      cm <- coords[1:n_half, , drop = FALSE]

      inf1 <- nn2Geo(
        cm,
        cm,
        K1 + 1,
        distance,
        maxdist1,
        radius,
        drop_first = TRUE,
        x_equals_y = TRUE
      )

      aa1 <- pair_block(
        inf1$rowidx,
        inf1$colidx,
        0L,
        0L,
        inf1$lags
      )

      inf2 <- nn2Geo(
        cm,
        cm,
        K3 + 1,
        distance,
        maxdist3,
        radius,
        drop_first = TRUE,
        x_equals_y = TRUE
      )

      aa2 <- pair_block(
        inf2$rowidx + n_half,
        inf2$colidx + n_half,
        1L,
        1L,
        inf2$lags
      )

      inf21 <- nn2Geo(
        cm,
        cm,
        K2 + 1,
        distance,
        maxdist2,
        radius,
        drop_first = TRUE,
        x_equals_y = TRUE
      )

      sel <- is.finite(inf21$lags) &
        (inf21$lags > 0) &
        (inf21$colidx > 0) &
        (inf21$rowidx > 0)

      aa_cross <- pair_block(
        inf21$rowidx[sel],
        inf21$colidx[sel] + n_half,
        0L,
        1L,
        inf21$lags[sel]
      )

      ii <- seq_len(n_half)

      aa_coll <- pair_block(
        ii,
        ii + n_half,
        0L,
        1L,
        rep.int(0, length(ii))
      )

      return(rbind(
        aa1,
        aa2,
        aa_cross,
        aa_coll
      ))
    }

    ##############################################################
    ## CASE 2: coordx_dyn list length 2
    ##############################################################

    if (!is.list(coordx_dyn) || length(coordx_dyn) != 2L) {
      stop("For bivariate case with coordx_dyn, it must be a list of length 2: list(var1_coords, var2_coords).")
    }

    X1 <- coordx_dyn[[1]]
    X2 <- coordx_dyn[[2]]

    n1 <- nrow(X1)

    inf1 <- nn2Geo(
      X1,
      X1,
      K1 + 1,
      distance,
      maxdist1,
      radius,
      drop_first = TRUE,
      x_equals_y = TRUE
    )

    aa1 <- pair_block(
      inf1$rowidx,
      inf1$colidx,
      0L,
      0L,
      inf1$lags
    )

    inf2 <- nn2Geo(
      X2,
      X2,
      K3 + 1,
      distance,
      maxdist3,
      radius,
      drop_first = TRUE,
      x_equals_y = TRUE
    )

    aa2 <- pair_block(
      inf2$rowidx + n1,
      inf2$colidx + n1,
      1L,
      1L,
      inf2$lags
    )

    inf_cross <- nn2Geo(
      X2,
      X1,
      K2,
      distance,
      maxdist2,
      radius,
      drop_first = FALSE,
      x_equals_y = FALSE
    )

    sel <- is.finite(inf_cross$lags) &
      (inf_cross$lags > 0) &
      (inf_cross$colidx > 0) &
      (inf_cross$rowidx > 0)

    aa_cross <- pair_block(
      inf_cross$rowidx[sel],
      inf_cross$colidx[sel] + n1,
      0L,
      1L,
      inf_cross$lags[sel]
    )

    tol <- auto_tol(X1, X2)

    same_grid <- (nrow(X1) == nrow(X2)) &&
      isTRUE(all(abs(X1 - X2) <= tol))

    if (same_grid) {

      ii <- seq_len(n1)

      aa_coll <- pair_block(
        ii,
        ii + n1,
        0L,
        1L,
        rep.int(0, length(ii))
      )

    } else {

      nn12 <- nn2Geo(
        X2,
        X1,
        K = 1,
        distance,
        maxdist2,
        radius,
        drop_first = FALSE,
        x_equals_y = FALSE
      )

      l12 <- nn12$lags

      l12[is.finite(l12) & (l12 <= tol)] <- 0

      selc <- is.finite(nn12$rowidx) &
        is.finite(nn12$colidx) &
        (nn12$rowidx > 0) &
        (nn12$colidx > 0) &
        is.finite(l12) &
        (l12 == 0)

      aa_coll <- pair_block(
        nn12$rowidx[selc],
        nn12$colidx[selc] + n1,
        0L,
        1L,
        l12[selc]
      )
    }

    rbind(
      aa1,
      aa2,
      aa_cross,
      aa_coll
    )
  }

  ##############################################################
  ## Calibrated Bernoulli probabilities
  ##############################################################

  calibrated_bernoulli_prob <- function(w, p) {

    n <- length(w)

    if (n == 0L) {
      return(numeric(0))
    }

    if (!is.numeric(p) ||
        length(p) != 1L ||
        !is.finite(p) ||
        p <= 0 ||
        p > 1) {
      stop("p_neighb must be in (0,1].")
    }

    if (p >= 1) {
      return(rep.int(1, n))
    }

    w <- as.numeric(w)

    good <- is.finite(w) & w > 0

    if (!any(good)) {
      w[] <- 1
    } else {
      w[!good] <- min(w[good])
    }

    target <- p * n

    f <- function(cst) {
      sum(pmin(1, cst * w)) - target
    }

    hi <- 1 / max(w)

    while (f(hi) < 0) {
      hi <- hi * 2
    }

    cstar <- stats::uniroot(
      f,
      interval = c(0, hi),
      tol = 1e-12
    )$root

    pi <- pmin(1, cstar * w)
    pi[!is.finite(pi)] <- 0

    pmin(1, pmax(0, pi))
  }

  ##############################################################
  ## Subset helper
  ##############################################################

  subset_gb_keep <- function(gb, keep_idx) {

    gb$colidx <- gb$colidx[keep_idx]
    gb$rowidx <- gb$rowidx[keep_idx]
    gb$lags   <- gb$lags[keep_idx]

    if (!is.null(gb$lagt)) {
      gb$lagt <- gb$lagt[keep_idx]
    }

    if (!is.null(gb$first)) {
      gb$first <- gb$first[keep_idx]
    }

    if (!is.null(gb$second)) {
      gb$second <- gb$second[keep_idx]
    }

    gb
  }

  ##############################################################
  ## Method II helpers: FixedBudget thinning
  ##
  ## The option name remains "FixedBudget" for compatibility with
  ## existing scripts. The implementation is target-wise/stratum-wise:
  ## the global budget is exact, but the allocation is made across
  ## natural strata. In the purely spatial case the stratum is the
  ## target location. In the space-time and bivariate cases the stratum
  ## also includes the available pair type descriptors, so that distinct
  ## temporal lags or variable-pair types are not mixed.
  ##
  ## This version avoids split(), which can be memory-expensive when the
  ## number of target locations is very large. It sorts edges by stratum
  ## and by a random tie-breaker, then keeps the first k_g randomized
  ## ranks within each stratum.
  ##############################################################

  make_stratum_id <- function(gb) {

    n_edges <- length(gb$rowidx)

    if (n_edges == 0L) {
      return(numeric(0))
    }

    if (is.null(gb$rowidx) || length(gb$rowidx) != n_edges) {
      stop("gb$rowidx must exist and have one entry per candidate edge.")
    }

    stratum_id <- as.double(gb$rowidx)

    if (any(!is.finite(stratum_id))) {
      stop("rowidx must be finite when constructing FixedBudget strata.")
    }

    ## Fold an additional discrete descriptor into the stratum id.
    ## factor() is used so that numeric descriptors such as temporal lags
    ## are converted to compact integer codes.
    fold_in <- function(base, extra, name) {

      if (length(extra) != length(base)) {
        stop(name, " must have one entry per candidate edge.")
      }

      if (any(!is.finite(extra))) {
        stop(name, " must be finite when constructing FixedBudget strata.")
      }

      code <- as.integer(factor(extra))
      L <- max(code, na.rm = TRUE)

      base * (L + 1L) + code
    }

    if (!is.null(gb$lagt)) {
      stratum_id <- fold_in(stratum_id, gb$lagt, "lagt")
    }

    if (!is.null(gb$first)) {
      stratum_id <- fold_in(stratum_id, gb$first, "first")
    }

    if (!is.null(gb$second)) {
      stratum_id <- fold_in(stratum_id, gb$second, "second")
    }

    stratum_id
  }

  fixed_budget_keep <- function(stratum_id, p, K_target) {

    n_edges <- length(stratum_id)

    if (n_edges <= 0L || K_target <= 0L) {
      return(integer(0))
    }

    if (!is.numeric(p) || length(p) != 1L || !is.finite(p) ||
        p <= 0 || p > 1) {
      stop("p must be in (0,1].")
    }

    K_target <- as.integer(round(K_target))
    K_target <- max(0L, min(K_target, as.integer(n_edges)))

    if (K_target <= 0L) {
      return(integer(0))
    }

    if (K_target >= n_edges || p >= 1) {
      return(seq_len(n_edges))
    }

    if (length(stratum_id) != n_edges || any(!is.finite(stratum_id))) {
      stop("stratum_id must be finite and have one entry per candidate edge.")
    }

    ## Random tie-breaker: sorting by (stratum_id, u) gives a uniformly
    ## random ordering of the candidate edges within each stratum.
    u <- stats::runif(n_edges)

    ord <- order(stratum_id, u)
    str_sorted <- stratum_id[ord]

    new_stratum <- c(TRUE, str_sorted[-1L] != str_sorted[-n_edges])
    stratum_index <- cumsum(new_stratum)
    stratum_start <- which(new_stratum)

    n_strata <- length(stratum_start)

    if (n_strata == 0L) {
      return(integer(0))
    }

    stratum_size <- diff(c(stratum_start, n_edges + 1L))

    rank_in_stratum <- seq_len(n_edges) -
      stratum_start[stratum_index] + 1L

    quota_raw <- p * stratum_size
    quota_base <- floor(quota_raw + 1e-12)
    quota_base <- pmin(quota_base, stratum_size)

    ## Exact global budget by randomized residual allocation.
    residual <- K_target - sum(quota_base)

    if (residual > 0L) {

      eligible <- which(quota_base < stratum_size)

      if (length(eligible) > 0L) {

        residual <- min(as.integer(residual), length(eligible))

        frac <- quota_raw[eligible] -
          floor(quota_raw[eligible] + 1e-12)

        frac[!is.finite(frac) | frac < 0] <- 0

        if (sum(frac) <= 0) {
          prob <- rep.int(1 / length(eligible), length(eligible))
        } else {
          prob <- frac / sum(frac)
        }

        add_strata <- sample(
          eligible,
          size = residual,
          replace = FALSE,
          prob = prob
        )

        quota_base[add_strata] <- quota_base[add_strata] + 1L
      }
    }

    ## Numerical safeguards. These should almost never be active, but they
    ## keep the retained size exact in irregular cases.
    deficit <- K_target - sum(quota_base)

    if (deficit > 0L) {

      eligible <- which(quota_base < stratum_size)

      if (length(eligible) > 0L) {

        add_strata <- sample(
          eligible,
          size = min(as.integer(deficit), length(eligible)),
          replace = FALSE
        )

        quota_base[add_strata] <- quota_base[add_strata] + 1L
      }
    }

    surplus <- sum(quota_base) - K_target

    if (surplus > 0L) {

      eligible <- which(quota_base > 0L)

      if (length(eligible) > 0L) {

        drop_strata <- sample(
          eligible,
          size = min(as.integer(surplus), length(eligible)),
          replace = FALSE
        )

        quota_base[drop_strata] <- quota_base[drop_strata] - 1L
      }
    }

    keep_sorted <- rank_in_stratum <= quota_base[stratum_index]
    keep_idx <- sort(ord[keep_sorted])

    if (length(keep_idx) != K_target) {
      stop(
        "FixedBudget retained ", length(keep_idx),
        " pairs instead of K_target = ", K_target, "."
      )
    }

    keep_idx
  }

  ##############################################################
  ## Start
  ##############################################################

  spatial <- TRUE
  spacetime <- FALSE

  if (!is.null(coordx_dyn)) {
    if (!is.list(coordx_dyn)) {
      stop("coordx_dyn must be a list")
    }
  }

  if (!is.null(coordt)) {
    if (is.numeric(coordt) && is.numeric(maxtime)) {
      if (length(coordt) > 1L && length(maxtime) >= 1L) {
        spacetime <- TRUE
      }
    }
  }

  distance <- CheckDistance(distance)
  spatial <- !spacetime && !bivariate

  ##############################################################
  ## Coordinate parsing
  ##############################################################

  if (is.null(coordx_dyn)) {

    if (is.null(coordx)) {
      stop("coordx must be provided when coordx_dyn is NULL.")
    }

    if (is.null(coordy)) {
      coords <- as.matrix(coordx)
    } else {
      coords <- if (is.null(coordz)) {
        cbind(coordx, coordy)
      } else {
        cbind(coordx, coordy, coordz)
      }
      coords <- as.matrix(coords)
    }

    storage.mode(coords) <- "double"

    if (!(ncol(coords) %in% c(2L, 3L))) {
      stop("coordx must be a coordinate matrix with 2 or 3 columns, or coordx/coordy/(coordz) must be coordinate vectors.")
    }

    if (bivariate && (nrow(coords) %% 2L != 0L)) {
      stop("For bivariate = TRUE with common coordinates, coordx must have an even number of rows.")
    }

    .GeoValidate3D(
      coordx = coords,
      distance = c("Eucl", "Chor", "Geod")[distance + 1L],
      context = "GeoNeighIndex"
    )

    numcoord <- nrow(coords)

  } else {

    coordx_dyn <- .GeoNormalizeDynamicCoordinates(
      coordx_dyn,
      coordt = if (!bivariate) coordt else NULL,
      context = "GeoNeighIndex"
    )

    if (!bivariate) {
      .GeoValidate3D(
        coordx_dyn = coordx_dyn,
        distance = c("Eucl", "Chor", "Geod")[distance + 1L],
        context = "GeoNeighIndex"
      )
      coords <- NULL
      numcoord <- nrow(coordx_dyn[[1L]])
    } else {
      coords <- 1
      numcoord <- 1
    }
  }

  ##############################################################
  ## Build pair indices
  ##############################################################

  if (spatial) {

    K_sp <- as.integer(K[1])

    sol <- nn2Geo(
      coords,
      coords,
      K_sp + 1L,
      distance,
      maxdist,
      radius,
      drop_first = TRUE,
      x_equals_y = TRUE
    )

    gb <- list(
      colidx = sol$colidx,
      rowidx = sol$rowidx,
      lags = sol$lags,
      maxdist = maxdist,
      neighb = neighb
    )
  }

  if (spacetime) {

    K_st <- as.integer(K[1])
    numtime <- length(coordt)

    sol <- spacetime_index(
      coords,
      coordx_dyn,
      numcoord,
      K_st,
      coordt,
      numtime,
      maxtime,
      maxdist,
      distance,
      radius
    )

    gb <- list(
      colidx = sol[, 2],
      rowidx = sol[, 1],
      lags = sol[, 4],
      lagt = sol[, 3]
    )
  }

  if (bivariate) {

    sol <- bivariate_index(
      coords,
      coordx_dyn,
      numcoord,
      K,
      maxdist,
      distance,
      radius
    )

    gb <- list(
      colidx = sol[, 2],
      rowidx = sol[, 1],
      lags = sol[, 5],
      first = sol[, 3],
      second = sol[, 4],
      maxdist = maxdist,
      neighb = neighb
    )
  }

  ##############################################################
  ## p_neighb checks
  ##############################################################

  if (!is.numeric(p_neighb) ||
      length(p_neighb) != 1L ||
      !is.finite(p_neighb)) {
    stop("p_neighb must be a finite numeric scalar.")
  }

  if (p_neighb <= 0 || p_neighb > 1) {
    stop("p_neighb must be in (0,1].")
  }

  thinning_on <- p_neighb < 1

  ##############################################################
  ## Defensive cleaning
  ##############################################################

  if (length(gb$rowidx) > 0L) {

    valid <- (gb$rowidx > 0L) &
      (gb$colidx > 0L) &
      is.finite(gb$rowidx) &
      is.finite(gb$colidx)

    if (!all(valid)) {

      gb$rowidx <- gb$rowidx[valid]
      gb$colidx <- gb$colidx[valid]
      gb$lags   <- gb$lags[valid]

      if (!is.null(gb$lagt)) {
        gb$lagt <- gb$lagt[valid]
      }

      if (!is.null(gb$first)) {
        gb$first <- gb$first[valid]
      }

      if (!is.null(gb$second)) {
        gb$second <- gb$second[valid]
      }
    }
  }

  ##############################################################
  ## No thinning if p_neighb = 1
  ##############################################################

  if (!thinning_on) {
    return(gb)
  }

  ##############################################################
  ## Method I: calibrated Bernoulli thinning
  ## thin_method = "bernoulli"
  ##############################################################

  if (thin_method == "Bernoulli") {

    n_all <- length(gb$rowidx)

    if (n_all == 0L) {
      return(gb)
    }

    w <- rep.int(1, n_all)

    p_ij <- calibrated_bernoulli_prob(
      w = w,
      p = p_neighb
    )

    B <- stats::rbinom(
      n_all,
      size = 1L,
      prob = p_ij
    )

    keep_idx <- which(B == 1L)

    gb <- subset_gb_keep(gb, keep_idx)

    return(gb)
  }

  ##############################################################
  ## Method II: FixedBudget thinning
  ##
  ## The option name is unchanged, but the retained set is now built
  ## target-wise. In the regular directed m-NN case and when
  ## p_neighb*m is an integer, this retains exactly p_neighb*m
  ## candidate edges for every target location. The total retained
  ## size is always exactly K_target = round(p_neighb * n_all).
  ##############################################################

  if (thin_method == "FixedBudget") {

    n_all <- length(gb$rowidx)

    if (n_all == 0L) {
      return(gb)
    }

    K_target <- as.integer(round(p_neighb * n_all))
    K_target <- max(0L, min(K_target, n_all))

    stratum_id <- make_stratum_id(gb)

    keep_idx <- fixed_budget_keep(
      stratum_id = stratum_id,
      p = p_neighb,
      K_target = K_target
    )

    gb <- subset_gb_keep(gb, keep_idx)

    return(gb)
  }

  return(gb)
}

####################################################
### File name: GeoVariogram.r
####################################################

.GeoVariogram_bin_values <- function(lags, values, breaks) {
  nbin <- length(breaks) - 1L
  out_count <- integer(max(0L, nbin))
  out_sum <- numeric(max(0L, nbin))
  if (nbin <= 0L || !length(lags)) {
    return(list(count = out_count, sum = out_sum,
                mean = rep(NA_real_, max(0L, nbin))))
  }

  lags <- as.numeric(lags)
  values <- as.numeric(values)
  ok <- is.finite(lags) & is.finite(values)
  if (!any(ok)) {
    return(list(count = out_count, sum = out_sum,
                mean = rep(NA_real_, nbin)))
  }
  lags <- lags[ok]
  values <- values[ok]

  ib <- findInterval(lags, breaks, rightmost.closed = TRUE)
  ok <- ib >= 1L & ib <= nbin
  if (!any(ok)) {
    return(list(count = out_count, sum = out_sum,
                mean = rep(NA_real_, nbin)))
  }
  ib <- ib[ok]
  values <- values[ok]

  out_count <- tabulate(ib, nbins = nbin)
  rr <- rowsum(values, group = ib, reorder = FALSE, na.rm = TRUE)
  ii <- suppressWarnings(as.integer(rownames(rr)))
  out_sum[ii] <- rr[, 1L]
  out_mean <- rep(NA_real_, nbin)
  nz <- out_count > 0L
  out_mean[nz] <- out_sum[nz] / out_count[nz]

  list(count = out_count, sum = out_sum, mean = out_mean)
}

.GeoVariogram_bin_surface <- function(lags, lagt, values,
                                      spatial_breaks, temporal_breaks) {
  nh <- length(spatial_breaks) - 1L
  nt <- length(temporal_breaks) - 1L
  count <- matrix(0L, nrow = max(0L, nh), ncol = max(0L, nt))
  sums <- matrix(0, nrow = max(0L, nh), ncol = max(0L, nt))
  means <- matrix(NA_real_, nrow = max(0L, nh), ncol = max(0L, nt))

  if (nh <= 0L || nt <= 0L || !length(lags)) {
    return(list(count = count, sum = sums, mean = means))
  }

  lags <- as.numeric(lags)
  lagt <- as.numeric(lagt)
  values <- as.numeric(values)
  ok <- is.finite(lags) & is.finite(lagt) & is.finite(values)
  if (!any(ok)) return(list(count = count, sum = sums, mean = means))

  hb <- findInterval(lags[ok], spatial_breaks, rightmost.closed = TRUE)
  tb <- findInterval(lagt[ok], temporal_breaks, rightmost.closed = TRUE)
  values <- values[ok]
  ok2 <- hb >= 1L & hb <= nh & tb >= 1L & tb <= nt
  if (!any(ok2)) return(list(count = count, sum = sums, mean = means))

  hb <- hb[ok2]
  tb <- tb[ok2]
  values <- values[ok2]
  grp <- (hb - 1L) * nt + tb

  cc <- tabulate(grp, nbins = nh * nt)
  ss <- numeric(nh * nt)
  rr <- rowsum(values, group = grp, reorder = FALSE, na.rm = TRUE)
  ii <- suppressWarnings(as.integer(rownames(rr)))
  ss[ii] <- rr[, 1L]

  count <- matrix(cc, nrow = nh, ncol = nt, byrow = TRUE)
  sums <- matrix(ss, nrow = nh, ncol = nt, byrow = TRUE)
  nz <- count > 0L
  means[nz] <- sums[nz] / count[nz]
  list(count = count, sum = sums, mean = means)
}

.GeoVariogram_make_breaks <- function(lags, numbins) {
  lags <- as.numeric(lags)
  lags <- lags[is.finite(lags)]
  if (!length(lags)) return(numeric(0L))

  lo <- min(lags)
  hi <- max(lags)
  if (hi <= lo) {
    delta <- max(1, abs(lo)) * sqrt(.Machine$double.eps)
    lo <- max(0, lo - delta)
    hi <- hi + delta
  }
  seq(lo, hi, length.out = as.integer(numbins))
}

.GeoVariogram_time_bins <- function(coordt, maxtime = NULL, numbins_t = NULL) {
  coordt <- as.numeric(coordt)
  if (length(coordt) < 2L || any(!is.finite(coordt))) {
    stop("At least two finite temporal coordinates are required")
  }

  tt <- sort(coordt)
  dtt <- diff(tt)
  tol_t <- sqrt(.Machine$double.eps) * max(1, max(abs(tt)))
  if (any(dtt <= tol_t)) {
    stop("Temporal coordinates must be distinct (up to numerical tolerance)")
  }

  max_available <- tt[length(tt)] - tt[1L]
  if (is.null(maxtime)) {
    maxlag <- max_available
  } else {
    if (!is.numeric(maxtime) || length(maxtime) != 1L ||
        !is.finite(maxtime) || maxtime <= 0) {
      stop("maxtime must be a positive finite numeric scalar")
    }
    maxlag <- min(as.numeric(maxtime), max_available)
  }
  if (!(maxlag > 0)) {
    stop("maxtime does not include any positive temporal lag")
  }

  dt0 <- stats::median(dtt)
  tol_d <- sqrt(.Machine$double.eps) * max(1, abs(dt0), max(abs(dtt)))
  regular <- all(abs(dtt - dt0) <= tol_d)
  if (maxlag + tol_d < min(dtt)) {
    stop("maxtime does not include any positive temporal lag")
  }

  if (regular) {
    ## For regular time coordinates the attainable positive lags are integer
    ## multiples of the common spacing.  Construct them directly: unlike the
    ## historical outer(coordt, coordt, '-') implementation this is O(T), not
    ## O(T^2), and it does not split numerically equal decimal lags.
    nlag <- floor((maxlag + tol_d) / dt0)
    if (nlag < 1L) {
      stop("maxtime does not include any positive temporal lag")
    }
    centers <- dt0 * seq_len(nlag)
    centers <- centers[centers <= maxlag + tol_d]
    if (!length(centers)) {
      stop("maxtime does not include any positive temporal lag")
    }

    if (length(centers) == 1L) {
      upper <- max(maxlag, centers[1L] + 0.5 * dt0)
      breaks <- c(0, upper)
    } else {
      mids <- (centers[-1L] + centers[-length(centers)]) / 2
      upper <- max(maxlag, centers[length(centers)] + 0.5 * dt0)
      breaks <- c(0, mids, upper)
    }
  } else {
    ## Irregular times can generate O(T^2) distinct lags.  Do not enumerate
    ## them.  Use a controlled number of temporal bins instead.
    if (is.null(numbins_t)) numbins_t <- min(12L, length(coordt) - 1L)
    if (!is.numeric(numbins_t) || length(numbins_t) != 1L ||
        !is.finite(numbins_t) || numbins_t < 1 || numbins_t %% 1 != 0) {
      stop("numbins_t must be a positive integer")
    }
    ntbin <- min(as.integer(numbins_t), length(coordt) - 1L)
    breaks <- seq(0, maxlag, length.out = ntbin + 1L)
    centers <- breaks[-length(breaks)] + diff(breaks) / 2
  }

  list(breaks = as.numeric(breaks), centers = as.numeric(centers),
       regular = regular, maxlag = maxlag)
}

.GeoVariogram_validate_data <- function(data, context = "GeoVariogram") {
  if (is.list(data)) {
    if (!all(vapply(data, is.numeric, logical(1L)))) {
      stop(context, ": data must be numeric", call. = FALSE)
    }
    z <- unlist(data, use.names = FALSE)
  } else {
    if (!is.numeric(data)) stop(context, ": data must be numeric", call. = FALSE)
    z <- as.numeric(data)
  }
  if (any(is.infinite(z))) {
    stop(context, ": data may contain NA/NaN but not Inf or -Inf", call. = FALSE)
  }
  invisible(TRUE)
}

.GeoVariogram_validate_dynamic_data <- function(data, coordx_dyn,
                                                 context = "GeoVariogram") {
  if (!is.list(data) || length(data) != length(coordx_dyn)) {
    stop(context, ": data must be a list with one component per temporal instant when coordx_dyn is used.",
         call. = FALSE)
  }
  sizes <- vapply(coordx_dyn, nrow, integer(1L))
  for (i in seq_along(data)) {
    if (!is.numeric(data[[i]]) || any(is.infinite(data[[i]]))) {
      stop(context, ": data[[", i, "]] must be numeric and may contain NA/NaN but not Inf or -Inf.",
           call. = FALSE)
    }
    if (length(data[[i]]) != sizes[i]) {
      stop(context, ": data[[", i, "]] must contain one value per row of coordx_dyn[[", i, "]].",
           call. = FALSE)
    }
  }
  invisible(sizes)
}

.GeoVariogram_normalize_grid <- function(data, coordx, coordy = NULL, coordz = NULL,
                                         coordt = NULL, bivariate = FALSE) {
  if (is.matrix(coordx)) {
    coords <- as.matrix(coordx)
  } else {
    if (is.null(coordy)) {
      stop("grid=TRUE requires coordy when coordx is not an explicit coordinate matrix")
    }
    axes <- list(coordx, coordy)
    if (!is.null(coordz)) axes[[3L]] <- coordz
    coords <- as.matrix(do.call(expand.grid, axes))
  }
  storage.mode(coords) <- "double"
  nspace <- nrow(coords)

  if (isTRUE(bivariate)) {
    if (is.list(data)) {
      if (length(data) != 2L ||
          any(vapply(data, length, integer(1L)) != nspace)) {
        stop("For grid=TRUE and bivariate=TRUE, data must contain two fields on the full spatial grid")
      }
      data_out <- lapply(data, as.numeric)
    } else if (is.matrix(data) &&
               ((nrow(data) == 2L && ncol(data) == nspace) ||
                (ncol(data) == 2L && nrow(data) == nspace))) {
      data_out <- data
    } else if (!is.null(dim(data)) && tail(dim(data), 1L) == 2L &&
               prod(head(dim(data), -1L)) == nspace) {
      zz <- matrix(as.numeric(data), nrow = nspace, ncol = 2L)
      data_out <- t(zz)
    } else if (length(data) == 2L * nspace) {
      data_out <- as.numeric(data)
    } else {
      stop("Bivariate grid data are not compatible with the supplied spatial axes")
    }
  } else if (!is.null(coordt) && length(coordt) > 1L) {
    nt <- length(coordt)
    if (is.matrix(data) && nrow(data) == nt && ncol(data) == nspace) {
      data_out <- data
    } else if (is.matrix(data) && ncol(data) == nt && nrow(data) == nspace) {
      data_out <- t(data)
    } else if (!is.null(dim(data)) && tail(dim(data), 1L) == nt &&
               prod(head(dim(data), -1L)) == nspace) {
      zz <- matrix(as.numeric(data), nrow = nspace, ncol = nt)
      data_out <- t(zz)
    } else {
      stop("Space-time grid data are not compatible with the supplied spatial axes and coordt")
    }
  } else {
    if (length(data) != nspace) {
      stop("Grid data length does not match the number of spatial grid locations")
    }
    data_out <- as.numeric(data)
  }

  list(data = data_out, coordx = coords, coordy = NULL, coordz = NULL)
}

.GeoVariogram_deduplicate_pairs <- function(idx, directed = FALSE) {
  if (isTRUE(directed) || is.null(idx$rowidx) || !length(idx$rowidx)) return(idx)
  rr <- as.integer(idx$rowidx)
  cc <- as.integer(idx$colidx)
  key <- paste(pmin(rr, cc), pmax(rr, cc), sep = ":")
  keep <- !duplicated(key)
  n <- length(keep)
  for (nm in names(idx)) {
    x <- idx[[nm]]
    if (length(x) == n) idx[[nm]] <- x[keep]
  }
  idx
}

.GeoVariogram_exact_zero_tol <- function(lags) {
  z <- as.numeric(lags)
  z <- z[is.finite(z)]
  scale <- if (length(z)) max(1, max(abs(z))) else 1
  100 * .Machine$double.eps * scale
}

.GeoVariogram_st_data_vector <- function(data, coordt, ns, dynamic = FALSE) {
  if (dynamic) {
    if (!is.list(data) || length(data) != length(coordt)) {
      stop("Dynamic space-time data must be a list with one vector per time")
    }
    out <- unlist(data, use.names = FALSE)
  } else if (is.matrix(data)) {
    nt <- length(coordt)
    if (nrow(data) == nt) {
      out <- c(t(data))
    } else if (ncol(data) == nt) {
      out <- as.numeric(data)
    } else {
      stop("Fixed-site space-time data must have one row (or legacy column) per temporal coordinate")
    }
  } else {
    out <- as.numeric(data)
  }

  if (length(out) != sum(ns)) {
    stop("The space-time data length is not compatible with the spatial/temporal layout")
  }
  out
}

.GeoVariogram_biv_data_vector <- function(data, n1, n2) {
  if (is.list(data)) {
    if (length(data) != 2L || length(data[[1L]]) != n1 || length(data[[2L]]) != n2) {
      stop("Bivariate list data must contain one vector for each variable")
    }
    return(c(as.numeric(data[[1L]]), as.numeric(data[[2L]])))
  }

  if (is.matrix(data)) {
    if (nrow(data) == 2L && ncol(data) == n1 && n1 == n2) {
      return(c(as.numeric(data[1L, ]), as.numeric(data[2L, ])))
    }
    if (ncol(data) == 2L && nrow(data) == n1 && n1 == n2) {
      return(c(as.numeric(data[, 1L]), as.numeric(data[, 2L])))
    }
  }

  out <- as.numeric(data)
  if (length(out) != n1 + n2) {
    stop("The bivariate data length is not compatible with the two spatial supports")
  }
  out
}

.GeoVariogram_neighbor_coords <- function(coords, distance) {
  coords <- as.matrix(coords)
  storage.mode(coords) <- "double"
  key <- tolower(gsub("[[:blank:]]", "", as.character(distance)[1L]))
  if (key != "eucl" && ncol(coords) == 3L) coords <- coords[, 1:2, drop = FALSE]
  coords
}

.GeoVariogram_safe_max <- function(x, default = 1) {
  x <- as.numeric(x)
  x <- x[is.finite(x)]
  if (!length(x)) return(default)
  max(x)
}

GeoVariogram <- function(data, coordx, coordy=NULL, coordz=NULL, coordt=NULL, coordx_dyn=NULL,
                         cloud=FALSE, distance="Eucl",
                         grid=FALSE, maxdist=NULL, neighb=NULL, maxtime=NULL, numbins=NULL,
                         radius=1, type='variogram', bivariate=FALSE, subsample=1, subsample_t=1,
                         numbins_t=NULL, directed=FALSE)
{
  call <- match.call()
  grid_requested <- isTRUE(grid)
  corrmodel <- 'exponential'

  if (missing(coordx)) {
    if (is.null(coordx_dyn))
      stop("argument 'coordx' is missing: provide 'coordx' or 'coordx_dyn'.")
    if (!is.list(coordx_dyn) || length(coordx_dyn) < 1L)
      stop("'coordx_dyn' must be a non-empty list when 'coordx' is omitted.")
    base <- coordx_dyn[[1L]]
    if (is.data.frame(base)) base <- as.matrix(base)
    if (!is.matrix(base))
      stop("coordx_dyn[[1]] must be a matrix/data.frame (n x d).")
    coordx <- base
  }

  if (is.data.frame(coordx)) coordx <- as.matrix(coordx)

  if (grid_requested && !is.null(coordx_dyn)) {
    stop("grid=TRUE cannot be combined with coordx_dyn")
  }
  if (grid_requested) {
    gg <- .GeoVariogram_normalize_grid(
      data = data, coordx = coordx, coordy = coordy, coordz = coordz,
      coordt = coordt, bivariate = bivariate
    )
    data <- gg$data
    coordx <- gg$coordx
    coordy <- gg$coordy
    coordz <- gg$coordz
    ## All subsequent validation/computation uses the canonical explicit
    ## coordinate representation.  grid_requested is retained in the output.
    grid <- FALSE
  }

  dynamic_st <- !isTRUE(bivariate) && !is.null(coordx_dyn) && !is.null(coordt)
  if (dynamic_st) {
    coordx_dyn <- .GeoNormalizeDynamicCoordinates(
      coordx_dyn, coordt = coordt, context = "GeoVariogram"
    )
    .GeoVariogram_validate_dynamic_data(
      data, coordx_dyn, context = "GeoVariogram"
    )
    coordx <- coordx_dyn[[1L]]
    coordy <- NULL
    coordz <- NULL
  } else {
    .GeoVariogram_validate_data(data, context = "GeoVariogram")
  }

  if (!is.numeric(subsample) || length(subsample) != 1L ||
      !is.finite(subsample) || subsample <= 0 || subsample > 1)
    stop("subsample must be a numeric value in (0, 1]")
  if (!is.numeric(subsample_t) || length(subsample_t) != 1L ||
      !is.finite(subsample_t) || subsample_t <= 0 || subsample_t > 1)
    stop("subsample_t must be a numeric value in (0, 1]")
  if (!is.logical(cloud) || length(cloud) != 1L || is.na(cloud))
    stop("cloud must be TRUE or FALSE")
  if (!is.logical(bivariate) || length(bivariate) != 1L || is.na(bivariate))
    stop("bivariate must be TRUE or FALSE")
  if (!is.logical(directed) || length(directed) != 1L || is.na(directed))
    stop("directed must be TRUE or FALSE")

  if (!is.null(maxdist)) {
    allowed_maxdist_lengths <- if (bivariate) c(1L, 3L) else 1L
    if (!is.numeric(maxdist) || !(length(maxdist) %in% allowed_maxdist_lengths) ||
        any(!is.finite(maxdist)) || any(maxdist <= 0)) {
      stop(if (bivariate)
        "maxdist must be a positive finite numeric scalar or a length-3 vector in bivariate mode"
      else "maxdist must be a positive finite numeric scalar")
    }
    maxdist <- as.numeric(maxdist)
  }

  if (!is.null(numbins_t)) {
    if (!is.numeric(numbins_t) || length(numbins_t) != 1L ||
        !is.finite(numbins_t) || numbins_t < 1 || numbins_t %% 1 != 0) {
      stop("numbins_t must be a positive integer")
    }
    numbins_t <- as.integer(numbins_t)
  }

  if (is.null(type)) type <- "variogram"
  if (!is.character(type) || length(type) != 1L ||
      tolower(type) != "variogram")
    stop("type must be 'variogram'")
  type <- "variogram"
  model <- "Gaussian"

  if (is.null(numbins)) numbins <- 13L
  if (!is.numeric(numbins) || length(numbins) != 1L || !is.finite(numbins) ||
      numbins < 2 || numbins %% 1 != 0)
    stop("numbins must be an integer >= 2")
  numbins <- as.integer(numbins)

  if (!is.numeric(radius) || length(radius) != 1L || !is.finite(radius) || radius <= 0)
    stop("radius must be a positive finite numeric scalar")

  if (bivariate) coordt <- c(0, 1)
  if (!is.null(coordt) && is.numeric(coordt) && length(coordt) > 1L)
    corrmodel <- "gneiting"

  if (bivariate) subsample_t <- 1

  ## Temporal subsampling follows the documented T x N fixed-site ordering,
  ## while accepting the historical N x T orientation for compatibility.
  if (!bivariate && !is.null(coordt) && length(coordt) > 1L && subsample_t < 1) {
    nt0 <- length(coordt)
    mt <- min(nt0, max(2L, floor(subsample_t * nt0)))
    idx_t <- sort(sample.int(nt0, mt, replace = FALSE))

    if (is.list(data) && length(data) == nt0) {
      coordt <- coordt[idx_t]
      data <- data[idx_t]
      if (dynamic_st) {
        coordx_dyn <- coordx_dyn[idx_t]
        coordx <- coordx_dyn[[1L]]
      }
    } else if (is.matrix(data) && nrow(data) == nt0) {
      coordt <- coordt[idx_t]
      data <- data[idx_t, , drop = FALSE]
    } else if (is.matrix(data) && ncol(data) == nt0) {
      coordt <- coordt[idx_t]
      data <- data[, idx_t, drop = FALSE]
    } else if (is.array(data) && length(dim(data)) >= 3L && tail(dim(data), 1L) == nt0) {
      subs <- lapply(dim(data), seq_len)
      subs[[length(subs)]] <- idx_t
      data <- do.call(`[`, c(list(data), subs, list(drop = FALSE)))
      coordt <- coordt[idx_t]
    } else {
      warning("subsample_t was requested but the data layout is not compatible with coordt; ignoring subsample_t")
    }
  }

  .subset_data_spatial <- function(z, idx0, n0) {
    if (is.matrix(z)) {
      if (nrow(z) == n0) return(z[idx0, , drop = FALSE])
      if (ncol(z) == n0) return(z[, idx0, drop = FALSE])
      stop("Spatial subsampling cannot identify the location dimension of data")
    }
    if (is.list(z)) {
      if (!all(vapply(z, function(d) length(d) == n0, logical(1))))
        stop("Spatial subsampling requires one value per location in every data block")
      return(lapply(z, function(d) d[idx0]))
    }
    if (length(z) != n0)
      stop("Spatial subsampling requires one data value per location")
    z[idx0]
  }

  ## For a regular grid, sample actual grid locations rather than coordinate axes.
  if (subsample < 1 && isTRUE(grid) && is.null(coordx_dyn)) {
    if (is.matrix(coordx)) {
      full_coords <- as.matrix(coordx)
    } else {
      if (is.null(coordy)) stop("grid=TRUE requires coordy when coordx is not a coordinate matrix")
      full_coords <- if (is.null(coordz)) {
        as.matrix(expand.grid(coordx, coordy))
      } else {
        as.matrix(expand.grid(coordx, coordy, coordz))
      }
    }
    n0 <- nrow(full_coords)
    m <- min(n0, max(2L, floor(subsample * n0)))
    idx0 <- sort(sample.int(n0, m, replace = FALSE))

    if (bivariate) {
      data <- .subset_data_spatial(data, idx0, n0)
    } else if (!is.null(coordt) && length(coordt) > 1L) {
      nt <- length(coordt)
      if (is.array(data) && length(dim(data)) >= 3L && tail(dim(data), 1L) == nt &&
          prod(head(dim(data), -1L)) == n0) {
        z <- matrix(as.numeric(data), nrow = n0, ncol = nt)
        data <- t(z[idx0, , drop = FALSE])
      } else {
        data <- .subset_data_spatial(data, idx0, n0)
      }
    } else {
      z <- as.numeric(data)
      if (length(z) != n0) stop("Grid data length does not match the number of grid locations")
      data <- z[idx0]
    }

    coordx <- full_coords[idx0, , drop = FALSE]
    coordy <- NULL
    coordz <- NULL
    grid <- FALSE
  } else if (subsample < 1 && dynamic_st) {
    sampled <- lapply(seq_along(coordx_dyn), function(i) {
      n0 <- nrow(coordx_dyn[[i]])
      m <- min(n0, max(1L, floor(subsample * n0)))
      idx0 <- sort(sample.int(n0, m, replace = FALSE))
      list(coords = coordx_dyn[[i]][idx0, , drop = FALSE], data = data[[i]][idx0])
    })
    coordx_dyn <- lapply(sampled, `[[`, "coords")
    data <- lapply(sampled, `[[`, "data")
    coordx <- coordx_dyn[[1L]]
  } else if (subsample < 1 && bivariate && !is.null(coordx_dyn)) {
    if (!is.list(coordx_dyn) || length(coordx_dyn) != 2L ||
        nrow(coordx_dyn[[1L]]) != nrow(coordx_dyn[[2L]])) {
      stop("Bivariate spatial subsampling requires two supports with the same number of locations")
    }
    n0 <- nrow(coordx_dyn[[1L]])
    m <- min(n0, max(2L, floor(subsample * n0)))
    idx0 <- sort(sample.int(n0, m, replace = FALSE))
    coordx_dyn <- lapply(coordx_dyn, function(x) as.matrix(x)[idx0, , drop = FALSE])
    data <- .subset_data_spatial(data, idx0, n0)
    coordx <- coordx_dyn[[1L]]
    coordy <- NULL
    coordz <- NULL
  } else if (subsample < 1) {
    if (is.matrix(coordx)) {
      n0 <- nrow(coordx)
      m <- min(n0, max(2L, floor(subsample * n0)))
      idx0 <- sort(sample.int(n0, m, replace = FALSE))
      coordx <- coordx[idx0, , drop = FALSE]
      if (!is.null(coordy) && length(coordy) == n0) coordy <- coordy[idx0]
      if (!is.null(coordz) && length(coordz) == n0) coordz <- coordz[idx0]
      data <- .subset_data_spatial(data, idx0, n0)
    } else {
      n0 <- length(coordx)
      m <- min(n0, max(2L, floor(subsample * n0)))
      idx0 <- sort(sample.int(n0, m, replace = FALSE))
      coordx <- coordx[idx0]
      if (!is.null(coordy) && length(coordy) == n0) coordy <- coordy[idx0]
      if (!is.null(coordz) && length(coordz) == n0) coordz <- coordz[idx0]
      data <- .subset_data_spatial(data, idx0, n0)
    }
  }

  checkinput <- CkInput(coordx, coordy, coordz, coordt, coordx_dyn, corrmodel, data,
                        distance, "Fitting", NULL, grid, "None", maxdist, maxtime,
                        model, NULL, "Nelder-Mead", NULL, radius, NULL, NULL, NULL,
                        "GeoWLS", FALSE, FALSE, NULL, NULL)
  if (!is.null(checkinput$error)) stop(checkinput$error)

  if (bivariate) corrmodel <- "Bi_matern_sep"

  initparam <- StartParam(coordx, coordy, coordz, coordt, coordx_dyn, corrmodel, data,
                          distance, "Fitting", NULL, grid, "None", maxdist, neighb,
                          maxtime, model, 1, NULL, NULL, FALSE, radius, NULL, NULL,
                          NULL, "GeoWLS", "GeoWLS", FALSE, NULL, NULL, FALSE, FALSE)
  if (!is.null(initparam$error)) stop(initparam$error)

  ## GeoWLS setup always uses SetGlobalVar(), including neighborhood calls.
  ## Register the matching cleanup once so errors after setup cannot leak C state.
  on.exit({
    try(dotCall64::.C64("DeleteGlobalVar", SIGNATURE = character(0),
                        INTENT = character(0), PACKAGE = "GeoModels",
                        VERBOSE = 0, NAOK = TRUE), silent = TRUE)
  }, add = TRUE)

  coordx <- initparam$coordx
  coordy <- initparam$coordy
  coordz <- initparam$coordz
  coordt <- initparam$coordt
  if (is.null(coordz)) coordz <- double(length(coordx))

  spacetime_dyn <- if (isTRUE(initparam$spacetime)) !is.null(coordx_dyn) else FALSE
  bins <- bint <- centers <- centert <- NULL
  lenbins <- lenbinst <- lenbint <- NULL
  variograms <- variogramst <- variogramt <- NULL
  regular <- NULL
  time_breaks <- NULL

  if (cloud && (isTRUE(initparam$spacetime) || isTRUE(initparam$bivariate))) {
    stop("cloud=TRUE is supported for univariate spatial variograms only")
  }

  ##########################################################################
  ## Bivariate
  ##########################################################################
  if (isTRUE(initparam$bivariate)) {
    n_var <- initparam$numtime
    if (n_var != 2L) stop("GeoVariogram currently supports exactly two variables in bivariate mode")
    ns <- as.integer(initparam$ns)
    if (length(ns) < 2L) stop("Invalid bivariate spatial layout")
    n1 <- ns[1L]
    n2 <- ns[2L]

    if (!is.null(coordx_dyn)) {
      if (length(coordx_dyn) != 2L) stop("Bivariate coordx_dyn must contain two spatial supports")
      X1 <- as.matrix(coordx_dyn[[1L]])
      X2 <- as.matrix(coordx_dyn[[2L]])
      scale0 <- suppressWarnings(max(abs(c(X1, X2)), na.rm = TRUE))
      if (!is.finite(scale0) || scale0 < 1) scale0 <- 1
      tol <- 100 * .Machine$double.eps * scale0
      aligned <- identical(dim(X1), dim(X2)) && all(abs(X1 - X2) <= tol)
      if (!aligned) {
        stop("A classical bivariate cross-semivariogram requires aligned spatial supports; the two coordx_dyn grids are not aligned")
      }
    }

    data_biv <- .GeoVariogram_biv_data_vector(data, n1, n2)
    numvario <- numbins - 1L

    if (is.null(neighb)) {
      NS <- c(0L, n1)
      moments_marg <- double(2L * numvario)
      lenbins_marg <- integer(2L * numvario)
      moments_cross <- double(numvario)
      lenbins_cross <- integer(numvario)

      EV <- dotCall64::.C64(
        "Binned_Variogram_biv2",
        bins = double(numbins), coordx = as.double(coordx), coordy = as.double(coordy),
        coordz = as.double(coordz), coordt = as.double(coordt), data = as.double(data_biv),
        lenbins_cross = lenbins_cross, moments_cross = moments_cross,
        numbins = as.integer(numbins), lenbins_marg = lenbins_marg,
        moments_marg = moments_marg, ns = ns, NS = as.integer(NS),
        SIGNATURE = c("double","double","double","double","double","double",
                      "integer","double","integer","integer","double","integer","integer"),
        INTENT = c("rw","r","r","r","r","r","rw","rw","r","rw","rw","r","r"),
        NAOK = TRUE, PACKAGE = "GeoModels", VERBOSE = 0
      )

      bins <- EV$bins
      l11 <- EV$lenbins_marg[seq_len(numvario)]
      l22 <- EV$lenbins_marg[numvario + seq_len(numvario)]
      l12 <- EV$lenbins_cross[seq_len(numvario)]
      m11 <- EV$moments_marg[seq_len(numvario)]
      m22 <- EV$moments_marg[numvario + seq_len(numvario)]
      m12 <- EV$moments_cross[seq_len(numvario)]
    } else {
      if (is.null(coordx_dyn)) {
        base <- cbind(coordx[seq_len(n1)], coordy[seq_len(n1)], coordz[seq_len(n1)])
        pair_coords <- rbind(base, base)
        pair_coords <- .GeoVariogram_neighbor_coords(pair_coords, distance)
        idx <- GeoNeighIndex(coordx = pair_coords, distance = distance,
                             neighb = neighb, maxdist = maxdist, radius = radius,
                             bivariate = TRUE)
      } else {
        idx <- GeoNeighIndex(coordx_dyn = coordx_dyn, distance = distance,
                             neighb = neighb, maxdist = maxdist, radius = radius,
                             bivariate = TRUE)
      }

      breaks <- .GeoVariogram_make_breaks(idx$lags, numbins)
      if (!length(breaks)) {
        bins <- numeric(0L)
        centers <- numeric(0L)
        lenbins <- matrix(integer(0L), nrow = 2L)
        lenbinst <- integer(0L)
        variograms <- matrix(numeric(0L), nrow = 2L)
        variogramst <- numeric(0L)
      } else {
        bins <- breaks
        row <- as.integer(idx$rowidx)
        col <- as.integer(idx$colidx)
        first <- as.integer(idx$first)
        second <- as.integer(idx$second)
        lagv <- as.numeric(idx$lags)

        z1 <- data_biv[seq_len(n1)]
        z2 <- data_biv[n1 + seq_len(n2)]

        sel11 <- which(first == 0L & second == 0L)
        sel22 <- which(first == 1L & second == 1L)
        sel12 <- which(first != second)

        unique_within <- function(ii) {
          if (isTRUE(directed) || !length(ii)) return(ii)
          key <- paste(pmin(row[ii], col[ii]), pmax(row[ii], col[ii]), sep = ":")
          ii[!duplicated(key)]
        }
        sel11 <- unique_within(sel11)
        sel22 <- unique_within(sel22)

        b11 <- .GeoVariogram_bin_values(
          lagv[sel11], 0.5 * (data_biv[row[sel11]] - data_biv[col[sel11]])^2, breaks)
        b22 <- .GeoVariogram_bin_values(
          lagv[sel22], 0.5 * (data_biv[row[sel22]] - data_biv[col[sel22]])^2, breaks)

        cross_lag <- numeric(0L)
        cross_val <- numeric(0L)
        if (length(sel12)) {
          rr <- row[sel12]
          cc <- col[sel12]
          ff <- first[sel12]
          ii <- ifelse(ff == 0L, rr, cc)
          jj <- ifelse(ff == 0L, cc - n1, rr - n1)
          zero_tol <- .GeoVariogram_exact_zero_tol(lagv[sel12])
          ok <- n1 == n2 & ii >= 1L & ii <= n1 & jj >= 1L & jj <= n2 &
            ii != jj & is.finite(lagv[sel12]) & lagv[sel12] > zero_tol
          if (any(ok)) {
            ii <- ii[ok]
            jj <- jj[ok]
            kk <- sel12[ok]
            if (!isTRUE(directed)) {
              key <- paste(pmin(ii, jj), pmax(ii, jj), sep = ":")
              keep <- !duplicated(key)
              ii <- ii[keep]
              jj <- jj[keep]
              kk <- kk[keep]
            }
            cross_lag <- lagv[kk]
            cross_val <- 0.5 * (z1[ii] - z1[jj]) * (z2[ii] - z2[jj])
          }
        }
        b12 <- .GeoVariogram_bin_values(cross_lag, cross_val, breaks)

        l11 <- b11$count; l22 <- b22$count; l12 <- b12$count
        m11 <- b11$sum; m22 <- b22$sum; m12 <- b12$sum
      }
    }

    if (is.null(centers)) {
      centers_all <- bins[-length(bins)] + diff(bins) / 2
      keep <- (l11 + l22 + l12) > 0L
      centers <- centers_all[keep]
      l11 <- l11[keep]; l22 <- l22[keep]; l12 <- l12[keep]
      m11 <- m11[keep]; m22 <- m22[keep]; m12 <- m12[keep]

      v11 <- rep(NA_real_, length(l11)); v22 <- v11; v12 <- v11
      ii <- l11 > 0L; v11[ii] <- m11[ii] / l11[ii]
      ii <- l22 > 0L; v22[ii] <- m22[ii] / l22[ii]
      ii <- l12 > 0L; v12[ii] <- m12[ii] / l12[ii]
      lenbins <- rbind(l11, l22)
      lenbinst <- l12
      variograms <- rbind(v11, v22)
      variogramst <- v12
    }
  }

  ##########################################################################
  ## Space-time
  ##########################################################################
  if (isTRUE(initparam$spacetime)) {
    ns <- as.integer(initparam$ns)
    NS <- as.integer(c(0L, cumsum(ns))[-(length(ns) + 1L)])
    data_st <- .GeoVariogram_st_data_vector(data, coordt, ns, dynamic = spacetime_dyn)
    ti <- .GeoVariogram_time_bins(coordt, maxtime,
                                    numbins_t = if (is.null(numbins_t)) numbins - 1L else numbins_t)
    time_breaks <- ti$breaks
    centert_all <- ti$centers
    regular <- ti$regular
    ntbin <- length(time_breaks) - 1L
    nhbin <- numbins - 1L

    if (is.null(neighb)) {
      EV <- dotCall64::.C64(
        if (spacetime_dyn) "Binned_Variogram_st2_dyn" else "Binned_Variogram_st2",
        bins = double(numbins), bint = as.double(time_breaks),
        coordx = as.double(coordx), coordy = as.double(coordy), coordz = as.double(coordz),
        coordt = as.double(coordt), data = as.double(data_st),
        lenbins = integer(nhbin), lenbinst = integer(nhbin * ntbin),
        lenbint = integer(ntbin), moments = double(nhbin),
        momentst = double(nhbin * ntbin), momentt = double(ntbin),
        numbins = as.integer(numbins), numbint = as.integer(length(time_breaks)),
        ns = ns, NS = NS,
        SIGNATURE = c(rep("double", 7), rep("integer", 3), rep("double", 3), rep("integer", 4)),
        INTENT = c("rw","rw",rep("r",5),rep("rw",3),rep("rw",3),rep("r",4)),
        NAOK = TRUE, PACKAGE = "GeoModels", VERBOSE = 0
      )

      bins <- EV$bins
      spatial_count <- EV$lenbins
      spatial_sum <- EV$moments
      temporal_count <- EV$lenbint
      temporal_sum <- EV$momentt
      st_count <- matrix(EV$lenbinst, nrow = nhbin, ncol = ntbin, byrow = TRUE)
      st_sum <- matrix(EV$momentst, nrow = nhbin, ncol = ntbin, byrow = TRUE)
    } else {
      if (spacetime_dyn) {
        idx <- GeoNeighIndex(coordx_dyn = coordx_dyn, coordt = coordt,
                             distance = distance, neighb = neighb,
                             maxdist = maxdist, maxtime = ti$maxlag, radius = radius)
      } else {
        nsite <- ns[1L]
        base <- cbind(coordx[seq_len(nsite)], coordy[seq_len(nsite)], coordz[seq_len(nsite)])
        base <- .GeoVariogram_neighbor_coords(base, distance)
        idx <- GeoNeighIndex(coordx = base, coordt = coordt,
                             distance = distance, neighb = neighb,
                             maxdist = maxdist, maxtime = ti$maxlag, radius = radius)
      }

      idx <- .GeoVariogram_deduplicate_pairs(idx, directed = directed)
      lagv <- as.numeric(idx$lags)
      lagt <- as.numeric(idx$lagt)
      rr <- as.integer(idx$rowidx)
      cc <- as.integer(idx$colidx)
      ok_idx <- rr >= 1L & rr <= length(data_st) & cc >= 1L & cc <= length(data_st)
      lagv <- lagv[ok_idx]; lagt <- lagt[ok_idx]
      rr <- rr[ok_idx]; cc <- cc[ok_idx]
      values <- 0.5 * (data_st[rr] - data_st[cc])^2

      positive_spatial <- lagv[is.finite(lagv) & lagv > 0]
      bins <- .GeoVariogram_make_breaks(positive_spatial, numbins)
      if (!length(bins)) {
        bins <- seq(0, 1, length.out = numbins)
      }
      nhbin <- length(bins) - 1L

      sel_s <- is.finite(lagt) & abs(lagt) <= sqrt(.Machine$double.eps) & lagv > 0
      zero_tol <- .GeoVariogram_exact_zero_tol(lagv)
      sel_t <- is.finite(lagv) & abs(lagv) <= zero_tol & lagt > 0
      ## The temporal margin is gamma(0,u): collocated pairs are not mixed
      ## into the positive-distance space-time surface.
      sel_st <- lagv > zero_tol & lagt > 0

      bs <- .GeoVariogram_bin_values(lagv[sel_s], values[sel_s], bins)
      bt <- .GeoVariogram_bin_values(lagt[sel_t], values[sel_t], time_breaks)
      bst <- .GeoVariogram_bin_surface(lagv[sel_st], lagt[sel_st], values[sel_st],
                                       bins, time_breaks)
      spatial_count <- bs$count; spatial_sum <- bs$sum
      temporal_count <- bt$count; temporal_sum <- bt$sum
      st_count <- bst$count; st_sum <- bst$sum
    }

    centers_all <- bins[-length(bins)] + diff(bins) / 2
    spatial_mean <- rep(NA_real_, length(spatial_count))
    ii <- spatial_count > 0L
    spatial_mean[ii] <- spatial_sum[ii] / spatial_count[ii]
    temporal_mean <- rep(NA_real_, length(temporal_count))
    ii <- temporal_count > 0L
    temporal_mean[ii] <- temporal_sum[ii] / temporal_count[ii]
    st_mean <- matrix(NA_real_, nrow = nrow(st_count), ncol = ncol(st_count))
    ii <- st_count > 0L
    st_mean[ii] <- st_sum[ii] / st_count[ii]

    keep_s <- spatial_count > 0L | rowSums(st_count) > 0L
    keep_t <- temporal_count > 0L | colSums(st_count) > 0L
    centers <- centers_all[keep_s]
    centert <- centert_all[keep_t]
    spatial_count <- spatial_count[keep_s]
    spatial_mean <- spatial_mean[keep_s]
    temporal_count <- temporal_count[keep_t]
    temporal_mean <- temporal_mean[keep_t]
    st_count <- st_count[keep_s, keep_t, drop = FALSE]
    st_mean <- st_mean[keep_s, keep_t, drop = FALSE]

    lenbins <- spatial_count
    variograms <- spatial_mean
    lenbint <- temporal_count
    variogramt <- temporal_mean
    lenbinst <- as.integer(as.vector(t(st_count)))
    variogramst <- as.numeric(as.vector(t(st_mean)))
    bint <- centert
  }

  ##########################################################################
  ## Purely spatial
  ##########################################################################
  if (!isTRUE(initparam$bivariate) && !isTRUE(initparam$spacetime)) {
    nsite <- length(coordx)
    coords <- cbind(coordx, coordy, coordz)
    data_sp <- as.numeric(data)
    if (length(data_sp) != nsite) stop("The number of spatial data values differs from the number of locations")

    if (cloud) {
      if (!is.null(neighb)) {
        nncoords <- .GeoVariogram_neighbor_coords(coords, distance)
        idx <- GeoNeighIndex(coordx = nncoords, distance = distance, neighb = neighb,
                             maxdist = maxdist, radius = radius)
        idx <- .GeoVariogram_deduplicate_pairs(idx, directed = directed)
        rr <- as.integer(idx$rowidx); cc <- as.integer(idx$colidx)
        lagv <- as.numeric(idx$lags)
        val <- 0.5 * (data_sp[rr] - data_sp[cc])^2
        ok <- is.finite(lagv) & is.finite(val)
        bins <- centers <- lagv[ok]
        variograms <- val[ok]
        lenbins <- rep.int(1L, length(variograms))
      } else if (!is.null(maxdist) && length(maxdist) == 1L && is.finite(maxdist)) {
        pp <- .GeoScatterplot_spatial_pairs_maxdist(
          data_sp, coords, maxdist = maxdist, bins = c(0, maxdist),
          distance = distance, radius = radius
        )
        bins <- centers <- pp$lags
        variograms <- 0.5 * (pp$v1 - pp$v2)^2
        lenbins <- rep.int(1L, length(variograms))
      } else {
        nmax <- as.double(nsite) * (nsite - 1) / 2
        if (!is.finite(nmax) || nmax > .Machine$integer.max) {
          stop("The requested variogram cloud is too large for the native interface; use maxdist, neighb, or subsample")
        }
        nmax <- as.integer(nmax)
        EV <- dotCall64::.C64(
          "Cloud_Variogram2",
          bins = double(nmax), coordx = as.double(coordx), coordy = as.double(coordy),
          coordz = as.double(coordz), coordt = 0, data = data_sp,
          lenbins = integer(nmax), moments = double(nmax), numbins = nmax,
          SIGNATURE = c("double","double","double","double","double","double","integer","double","integer"),
          INTENT = c("rw","r","r","r","r","r","rw","rw","r"),
          NAOK = TRUE, PACKAGE = "GeoModels", VERBOSE = 0
        )
        keep <- EV$lenbins > 0L
        bins <- centers <- EV$bins[keep]
        lenbins <- EV$lenbins[keep]
        variograms <- EV$moments[keep]
      }
    } else if (is.null(neighb)) {
      EV <- dotCall64::.C64(
        "Binned_Variogram2",
        bins = double(numbins), coordx = as.double(coordx), coordy = as.double(coordy),
        coordz = as.double(coordz), coordt = 0, data = data_sp,
        lenbins = integer(numbins - 1L), moments = double(numbins - 1L),
        numbins = as.integer(numbins),
        SIGNATURE = c("double","double","double","double","double","double","integer","double","integer"),
        INTENT = c("rw","r","r","r","r","r","rw","rw","r"),
        NAOK = TRUE, PACKAGE = "GeoModels", VERBOSE = 0
      )
      bins <- EV$bins
      centers_all <- bins[-length(bins)] + diff(bins) / 2
      keep <- EV$lenbins > 0L
      centers <- centers_all[keep]
      lenbins <- EV$lenbins[keep]
      variograms <- EV$moments[keep] / lenbins
    } else {
      nncoords <- .GeoVariogram_neighbor_coords(coords, distance)
      idx <- GeoNeighIndex(coordx = nncoords, distance = distance, neighb = neighb,
                           maxdist = maxdist, radius = radius)
      idx <- .GeoVariogram_deduplicate_pairs(idx, directed = directed)
      lagv <- as.numeric(idx$lags)
      rr <- as.integer(idx$rowidx); cc <- as.integer(idx$colidx)
      breaks <- .GeoVariogram_make_breaks(lagv, numbins)
      if (!length(breaks)) {
        bins <- centers <- variograms <- numeric(0L)
        lenbins <- integer(0L)
      } else {
        bv <- .GeoVariogram_bin_values(
          lagv, 0.5 * (data_sp[rr] - data_sp[cc])^2, breaks
        )
        bins <- breaks
        centers_all <- bins[-length(bins)] + diff(bins) / 2
        keep <- bv$count > 0L
        centers <- centers_all[keep]
        lenbins <- bv$count[keep]
        variograms <- bv$mean[keep]
      }
    }
  }

  ans <- list(
    bins = bins,
    bint = bint,
    bivariate = isTRUE(initparam$bivariate),
    cloud = cloud,
    centers = centers,
    centert = centert,
    distance = distance,
    radius = radius,
    grid = grid_requested,
    neighb = neighb,
    directed = directed,
    lenbins = lenbins,
    lenbinst = lenbinst,
    lenbint = lenbint,
    maxdist = maxdist,
    maxtime = maxtime,
    numbins_t = numbins_t,
    regular = regular,
    time.breaks = time_breaks,
    spacetime_dyn = spacetime_dyn,
    temporal.margin = if (isTRUE(spacetime_dyn)) "collocated" else if (isTRUE(initparam$spacetime)) "same-site" else NULL,
    subsample = subsample,
    subsample_t = subsample_t,
    variograms = variograms,
    variogramst = variogramst,
    variogramt = variogramt,
    type = type
  )

  structure(c(ans, call = call), class = "GeoVariogram")
}

####################################################
### Plot method
####################################################
plot.GeoVariogram <- function(x, ...) {
  if (!inherits(x, "GeoVariogram"))
    stop("Enter an object obtained from the function GeoVariogram")

  opar <- par(no.readonly = TRUE)
  on.exit(par(opar))
  dots <- list(...)
  xlab <- if (!is.null(dots$xlab)) dots$xlab else "Distance"
  ylab <- if (!is.null(dots$ylab)) dots$ylab else "Semi-Variogram"
  dots$xlab <- NULL
  dots$ylab <- NULL

  plot_with_defaults <- function(defaults) {
    do.call(plot.default, utils::modifyList(defaults, dots))
  }

  safe_ylim <- function(z, include_zero = TRUE, symmetric = FALSE) {
    z <- as.numeric(z)
    z <- z[is.finite(z)]
    if (!length(z)) return(c(0, 1))
    if (symmetric) {
      a <- max(abs(z))
      if (!is.finite(a) || a <= 0) a <- 1
      return(c(-a, a))
    }
    lo <- if (include_zero) min(0, z) else min(z)
    hi <- max(z)
    if (!is.finite(hi) || hi <= lo) hi <- lo + 1
    c(lo, hi)
  }

  ispatim <- !is.null(x$centert) || !is.null(x$bint)
  bivariate <- isTRUE(x$bivariate)

  if (bivariate) {
    par(mfrow = c(2, 2))
    xmax <- .GeoVariogram_safe_max(x$centers)

    plot_with_defaults(list(
      x = x$centers, y = x$variograms[1L, ], main = "First semi-variogram",
      ylim = safe_ylim(x$variograms[1L, ]), xlim = c(0, xmax),
      xlab = xlab, ylab = ylab))

    plot_with_defaults(list(
      x = x$centers, y = x$variogramst, main = "Cross semi-variogram",
      ylim = safe_ylim(x$variogramst, symmetric = TRUE), xlim = c(0, xmax),
      xlab = xlab, ylab = ylab))

    plot_with_defaults(list(
      x = x$centers, y = x$variograms[2L, ], main = "Second semi-variogram",
      ylim = safe_ylim(x$variograms[2L, ]), xlim = c(0, xmax),
      xlab = xlab, ylab = ylab))

    plot.new()
    return(invisible())
  }

  if (ispatim) {
    par(mfrow = c(2, 2))
    temporal_lags <- if (!is.null(x$centert)) x$centert else x$bint
    nx <- length(x$centers)
    nt <- length(temporal_lags)
    if (length(x$variogramst) != nx * nt)
      stop("The space-time variogram surface is not rectangular")

    cross <- matrix(x$variogramst, nrow = nx, ncol = nt, byrow = TRUE)
    evario <- rbind(c(0, x$variogramt), cbind(x$variograms, cross))
    gx <- c(0, x$centers)
    gt <- c(0, temporal_lags)
    egrid <- expand.grid(gx, gt)
    zz <- as.numeric(evario)
    ok <- is.finite(zz)

    if (any(ok) && requireNamespace("scatterplot3d", quietly = TRUE)) {
      scatterplot3d::scatterplot3d(
        egrid[ok, 1L], egrid[ok, 2L], zz[ok], type = "h", highlight.3d = TRUE,
        cex.axis = .7, cex.lab = .7, main = "Empirical space-time semi-variogram",
        xlab = "Distance", ylab = "Time", zlab = ylab,
        mar = c(2, 2, 2, 2), mgp = c(0, 0, 0)
      )
    } else {
      plot.new(); title("Empirical space-time semi-variogram")
      if (any(ok)) {
        graphics::text(0.5, 0.5, "Install 'scatterplot3d' for the 3D empirical panel")
      }
    }

    par(mai = c(.2, .2, .2, .2))
    persp(x = gx, y = gt, z = evario,
          xlab = "h", ylab = "u", zlab = expression(gamma(h, u)),
          ltheta = 90, shade = 0.75, ticktype = "detailed", phi = 30,
          theta = 30, main = "Smoothed space-time semi-variogram",
          cex.axis = .8, cex.lab = .8)

    par(mai = c(.5, .5, .5, .5), mgp = c(1.6, .6, 0))
    if (any(is.finite(x$variogramt))) {
      plot_with_defaults(list(
        x = temporal_lags, y = x$variogramt,
        xlab = "t", ylab = expression(gamma(t)),
        ylim = safe_ylim(x$variogramt),
        xlim = c(0, .GeoVariogram_safe_max(temporal_lags)),
        main = "Marginal temporal semi-variogram"))
    } else {
      plot.new()
      title(main = "Marginal temporal semi-variogram")
      if (isTRUE(x$spacetime_dyn)) {
        text(0.5, 0.56, "Temporal marginal unavailable", font = 2, cex = 0.9)
        text(0.5, 0.43, "(no repeated spatial locations)", cex = 0.72)
      } else {
        graphics::text(0.5, 0.5, "No temporal marginal pairs available", cex = 0.85)
      }
      box()
    }

    plot_with_defaults(list(
      x = x$centers, y = x$variograms,
      xlab = "h", ylab = expression(gamma(h)),
      ylim = safe_ylim(x$variograms),
      xlim = c(0, .GeoVariogram_safe_max(x$centers)),
      main = "Marginal spatial semi-variogram"))
    return(invisible())
  }

  plot_with_defaults(list(x = x$centers, y = x$variograms,
                          xlab = xlab, ylab = ylab))
  invisible()
}

.onAttach <- function(lib, pkg) {
  pkg.info <- drop(read.dcf(file = system.file("DESCRIPTION", package = "GeoModels"),
                            fields = c("Title", "Version", "Date")))
  # Startup message
  packageStartupMessage(paste("--------------------------------------------------------------\n",
                              pkg.info["Title"], "\n",
                              "For an Introduction to GeoModels, go to https://vmoprojs.github.io/GeoModels-page/\n",
                              paste("GeoModels version ", pkg.info["Version"],
                                    " (built on ", pkg.info["Date"], ") is now loaded\n", sep = ""),
                              "--------------------------------------------------------------\n"))

}
.GeoModels_env <- new.env(parent = emptyenv())
.onLoad <- function(libname, pkgname) {
  .GeoModels_env$II_global_cache <- new.env(hash = TRUE)
  .GeoModels_env$ak_cache <- list()
}

GeoNA <- function(data, coordx = NULL, coordy = NULL, coordz = NULL, coordt = NULL, 
                  coordx_dyn = NULL, grid = FALSE, X = NULL, setting = "spatial") {
  

#' Validate function inputs
validate_inputs <- function(data, coordx, coordy, coordz, coordt, coordx_dyn, setting) {
  
  # Check setting
  if (!setting %in% c("spatial", "spacetime", "bivariate")) {
    stop("'setting' must be one of: 'spatial', 'spacetime', 'bivariate'")
  }
  
  # Check required inputs for spacetime
  if (setting == "spacetime" && is.null(coordt)) {
    stop("'coordt' is required for spacetime setting")
  }
  
  # Check coordx is provided
  if (is.null(coordx) && is.null(coordx_dyn)) {
    stop("Either 'coordx' or 'coordx_dyn' must be provided")
  }
  
  # Check bivariate specific requirements
  if (setting == "bivariate") {
    if (!is.null(coordx_dyn)) {
      if (!is.list(coordx_dyn) || length(coordx_dyn) != 2) {
        stop("For bivariate with coordx_dyn, coordx_dyn must be a list of length 2")
      }
      if (!is.list(data) || length(data) != 2) {
        stop("For bivariate with coordx_dyn, data must be a list of length 2")
      }
    } else {
      if (!is.matrix(data) || nrow(data) != 2) {
        stop("For bivariate with fixed coordinates, data must be a 2 x N matrix")
      }
    }
  }
}

#' Build coordinate matrix from input vectors
build_coordinate_matrix <- function(coordx, coordy, coordz, grid, coordx_dyn) {
  
  # Skip if using dynamic coordinates
  if (!is.null(coordx_dyn)) {
    return(NULL)
  }
  
  # Handle 1D case
  if (is.null(coordy)) {
    return(as.matrix(coordx))
  }
  
  # Handle 2D/3D cases
  if (grid) {
    # Create grid from coordinate vectors
    if (is.null(coordz)) {
      coords <- as.matrix(expand.grid(coordx, coordy))
    } else {
      coords <- as.matrix(expand.grid(coordx, coordy, coordz))
    }
  } else {
    # Use coordinates as-is
    if (is.null(coordz)) {
      coords <- cbind(coordx, coordy)
    } else {
      coords <- cbind(coordx, coordy, coordz)
    }
  }
  
  return(coords)
}

#' Identify missing/invalid values
identify_missing <- function(x) {
  is.na(x) | is.nan(x) | is.infinite(x)
}

#' Process spatial data
process_spatial <- function(data, coords, X) {
  
  # Find missing values
  missing_mask <- identify_missing(data)
  valid_mask <- !missing_mask
  
  # Calculate removal percentage
  perc <- sum(missing_mask) / length(data)
  
  # Remove missing values if any exist
  if (perc > 0) {
    data <- data[valid_mask]
    coords <- coords[valid_mask, , drop = FALSE]
    if (!is.null(X)) {
      X <- X[valid_mask, , drop = FALSE]
    }
  }
  
  return(list(
    data = data,
    coords = coords,
    coordx_dyn = NULL,
    X = X,
    perc = perc
  ))
}

#' Process spacetime data
process_spacetime <- function(data, coords, coordt, coordx_dyn, X) {
  
  if (!is.null(coordx_dyn)) {
    # Dynamic coordinates case
    return(process_spacetime_dynamic(data, coordx_dyn, X))
  } else {
    # Fixed coordinates case
    return(process_spacetime_fixed(data, coords, X))
  }
}

#' Process spacetime data with dynamic coordinates
process_spacetime_dynamic <- function(data, coordx_dyn, X) {
  
  if (!is.list(data)) {
    stop("For spacetime with dynamic coordinates, data must be a list")
  }
  
  cleaned_data <- list()
  cleaned_coords <- list()
  cleaned_X <- if (!is.null(X)) list() else NULL
  total_removed <- 0
  total_points <- 0
  
  for (t in seq_along(data)) {
    # Find missing values for time t
    missing_mask <- identify_missing(data[[t]])
    valid_mask <- !missing_mask
    
    # Update totals for percentage calculation
    total_removed <- total_removed + sum(missing_mask)
    total_points <- total_points + length(data[[t]])
    
    # Clean data for time t
    cleaned_data[[t]] <- data[[t]][valid_mask]
    cleaned_coords[[t]] <- coordx_dyn[[t]][valid_mask, , drop = FALSE]
    
    if (!is.null(X)) {
      cleaned_X[[t]] <- X[[t]][valid_mask, , drop = FALSE]
    }
  }
  
  perc <- total_removed / total_points
  
  return(list(
    data = cleaned_data,
    coords = NULL,
    coordx_dyn = cleaned_coords,
    X = cleaned_X,
    perc = perc
  ))
}

#' Process spacetime data with fixed coordinates
process_spacetime_fixed <- function(data, coords, X) {
  
  if (!is.matrix(data)) {
    stop("For spacetime with fixed coordinates, data must be a matrix (T x N)")
  }
  nt <- nrow(data)
  nsite <- ncol(data)

  if (nrow(coords) != nsite) {
    stop("For fixed spatio-temporal data, coords must have one row per spatial site")
  }
  if (!is.null(X)) {
    X <- .GeoMean_design(X, n = nt * nsite, name = "X")
  }
  
  # Find locations with any missing values across time
  missing_mask <- apply(data, 2, function(col) any(identify_missing(col)))
  valid_mask <- !missing_mask
  
  perc <- sum(missing_mask) / nsite
  
  # Remove locations with missing values.  The design matrix follows the
  # package observation order: all sites at time 1, then all sites at time 2,
  # and so on (the same order as c(t(data))).
  if (perc > 0) {
    keep_rows <- rep(valid_mask, times = nt)
    data <- data[, valid_mask, drop = FALSE]
    coords <- coords[valid_mask, , drop = FALSE]
    if (!is.null(X)) {
      X <- X[keep_rows, , drop = FALSE]
    }
  }
  
  return(list(
    data = data,
    coords = coords,
    coordx_dyn = NULL,
    X = X,
    perc = perc
  ))
}

#' Process bivariate data
process_bivariate <- function(data, coords, coordx_dyn, X) {
  
  if (!is.null(coordx_dyn)) {
    # Dynamic coordinates case
    return(process_bivariate_dynamic(data, coordx_dyn, X))
  } else {
    # Fixed coordinates case
    return(process_bivariate_fixed(data, coords, X))
  }
}

#' Process bivariate data with dynamic coordinates
process_bivariate_dynamic <- function(data, coordx_dyn, X) {
  
  cleaned_data <- list()
  cleaned_coords <- list()
  cleaned_X <- if (!is.null(X)) list() else NULL
  perc <- numeric(2)
  
  for (i in 1:2) {
    # Validate dimensions
    if (length(data[[i]]) != nrow(coordx_dyn[[i]])) {
      stop(paste0("Length mismatch between data[[", i, "]] and coordx_dyn[[", i, "]]"))
    }
    
    # Find valid observations
    missing_mask <- identify_missing(data[[i]])
    valid_mask <- !missing_mask
    
    perc[i] <- sum(missing_mask) / length(data[[i]])
    
    # Clean data for variable i
    cleaned_data[[i]] <- data[[i]][valid_mask]
    cleaned_coords[[i]] <- coordx_dyn[[i]][valid_mask, , drop = FALSE]
    
    if (!is.null(X)) {
      if (!is.matrix(X[[i]]) || nrow(X[[i]]) != length(data[[i]])) {
        stop(paste0("X[[", i, "]] dimension mismatch with data[[", i, "]]"))
      }
      cleaned_X[[i]] <- X[[i]][valid_mask, , drop = FALSE]
    }
  }
  
  return(list(
    data = cleaned_data,
    coords = NULL,
    coordx_dyn = cleaned_coords,
    X = cleaned_X,
    perc = perc
  ))
}

#' Process bivariate data with fixed coordinates
process_bivariate_fixed <- function(data, coords, X) {
  
  # Find locations valid for both variables
  valid_mask1 <- !identify_missing(data[1, ])
  valid_mask2 <- !identify_missing(data[2, ])
  valid_mask <- valid_mask1 & valid_mask2
  
  perc <- 1 - sum(valid_mask) / length(valid_mask)
  
  # Remove locations with missing values in either variable
  if (perc > 0) {
    data <- data[, valid_mask, drop = FALSE]
    coords <- coords[valid_mask, , drop = FALSE]
    if (!is.null(X)) {
      X <- X[valid_mask, , drop = FALSE]
    }
  }
  
  return(list(
    data = data,
    coords = coords,
    coordx_dyn = NULL,
    X = X,
    perc = perc
  ))
}

# Example usage and tests
if (FALSE) {
  # Test spatial case
  set.seed(123)
  data_spatial <- c(1, 2, NA, 4, Inf, 6)
  coords_x <- 1:6
  coords_y <- 1:6
  
  result_spatial <- GeoNA(data_spatial, coords_x, coords_y, setting = "spatial")
  cat("Spatial test - Removed percentage:", result_spatial$perc, "\n")
  
  # Test spacetime case with fixed coordinates
  data_st <- matrix(c(1, 2, NA, 4, 5, 6, 7, 8), nrow = 2)
  coords_st <- cbind(1:4, 1:4)
  
  result_st <- GeoNA(data_st, coords_st[,1], coords_st[,2], 
                     coordt = 1:2, setting = "spacetime")
  cat("Spacetime test - Removed percentage:", result_st$perc, "\n")
  
  # Test bivariate case
  data_biv <- rbind(c(1, 2, NA, 4), c(5, 6, 7, Inf))
  coords_biv <- cbind(1:4, 1:4)
  
  result_biv <- GeoNA(data_biv, coords_biv[,1], coords_biv[,2], 
                      setting = "bivariate")
  cat("Bivariate test - Removed percentage:", result_biv$perc, "\n")
}





  # Input validation
  validate_inputs(data, coordx, coordy, coordz, coordt, coordx_dyn, setting)
  
  # Build coordinate matrix for static coordinates
  coords <- build_coordinate_matrix(coordx, coordy, coordz, grid, coordx_dyn)
  
  # Process data based on setting
  result <- switch(setting,
    "spatial" = process_spatial(data, coords, X),
    "spacetime" = process_spacetime(data, coords, coordt, coordx_dyn, X),
    "bivariate" = process_bivariate(data, coords, coordx_dyn, X),
    stop("Invalid setting. Must be 'spatial', 'spacetime', or 'bivariate'")
  )
  
  # Return standardized output
  return(list(
    coordx = if (is.null(coordx_dyn)) result$coords else coordx,
    coordy = coordy,
    coordz = coordz,
    coordt = coordt,
    coordx_dyn = result$coordx_dyn,
    data = result$data,
    grid = grid,
    perc = result$perc,
    setting = setting,
    X = result$X
  ))
}

####################################################
### File name: GeoCovVariogram.r
####################################################

### Compute and plot the (estimated) covariance function and the variogram
### from  GeoFit object

## Select the empirical temporal profile used by GeoCovariogram for a
## dynamic space-time variogram.  The true temporal margin is gamma(0,u).
## When it is unavailable and fix.lags requests h=0, use the first retained
## positive-distance ST bin with finite values.  This helper changes only the
## diagnostic profile; it never alters the GeoVariogram object itself.
.GeoCovariogramDynamicTemporalProfile <- function(vario, evario, lags,
                                                   fix.lags, dynamic=FALSE) {
  idx <- fix.lags
  value <- if(!is.null(idx) && length(lags) >= idx) lags[idx] else NA_real_
  out <- list(index=idx, value=value, fallback=FALSE,
              lower=NA_real_, upper=NA_real_, first.bin=FALSE)

  if(!isTRUE(dynamic) || is.null(idx) || idx != 1L ||
     !is.matrix(evario) || nrow(evario) <= 1L || ncol(evario) <= 1L)
    return(out)

  ## If collocated dynamic pairs exist, retain the exact h=0 profile.
  if(any(is.finite(evario[1L,-1L]))) return(out)

  positive_surface <- evario[-1L,-1L,drop=FALSE]
  candidates <- which(rowSums(is.finite(positive_surface)) > 0L)
  if(!length(candidates)) return(out)

  out$index <- candidates[1L] + 1L
  if(length(lags) >= out$index) out$value <- lags[out$index]
  out$fallback <- TRUE

  ## GeoVariogram may omit empty retained centers.  Recover the original
  ## bin interval by matching the retained center to the full midpoint grid.
  br <- as.numeric(vario$bins)
  if(length(br) >= 2L && all(is.finite(br)) && is.finite(out$value)) {
    mids <- br[-length(br)] + diff(br)/2
    if(length(mids)) {
      jj <- which.min(abs(mids - out$value))
      out$lower <- br[jj]
      out$upper <- br[jj + 1L]
      out$first.bin <- identical(as.integer(jj),1L)
    }
  }
  out
}

GeoCovariogram <- function(fitted, distance="Eucl", answer.cov=FALSE, answer.vario=FALSE,
                         answer.range=FALSE, fix.lags=NULL, fix.lagt=NULL,
                        show.cov=FALSE, show.vario=TRUE,  show.range=FALSE,
                        add.cov=FALSE, add.vario=FALSE, pract.range=95,
                        vario=NULL,invisible=FALSE,...)
  {
    result <- NULL

    CheckDistance<- function(distance)
    {
        CheckDistance <- NULL
        CheckDistance <- switch(distance,
                                eucl=0,
                                Eucl=0,
                                chor=1,  
                                Chor=1,
                                geod=2,
                                Geod=2,
                                proj=3,
                                Proj=3)
        return(CheckDistance)
    }
CorrelationFct <- function(bivariate, corr_model, lags, lag_t, num_lags, num_lag_t, mu, model, nuisance, param, N) {
  if (!bivariate) {
    result <- dotCall64::.C64(
      "VectCorrelation",
      SIGNATURE = c("double", "integer", "double", "integer", "integer", "double", "integer", "double", "double", "double", "integer"),
      INTENT = c("w", rep("r", 10)),
      corr = dotCall64::vector_dc("double", num_lags * num_lag_t),
      corr_model = corr_model,
      lags = lags,num_lags = num_lags,num_lag_t = num_lag_t,mu = mu,
      model = model,nuisance = nuisance,param = param,lag_t = lag_t,N = N,PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
    
    cc <- result$corr
  } else {
    result <- dotCall64::.C64(
      "VectCorrelation_biv",
      SIGNATURE = c("double", "double", "integer", "double", "integer", "integer", "double", "integer", "double", "double", "double", "integer"),
      INTENT = c("w", "w", rep("r", 10)),
      corr = dotCall64::vector_dc("double", num_lags * 4),vario = dotCall64::vector_dc("double", num_lags * 4),corr_model = corr_model,
      lags = lags,num_lags = num_lags,num_lag_t = num_lag_t,mu = mu,
      model = model,nuisance = nuisance,param = param,lag_t = lag_t,N = N,PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
    
    cc <- c(result$corr, result$vario)
  }
  cc
}


.GeoCovariogram_sqerr <- function(empirical, theoretical) {
  empirical <- as.numeric(empirical)
  theoretical <- as.numeric(theoretical)
  if(length(empirical) != length(theoretical))
    stop("Empirical and theoretical semivariograms have incompatible dimensions.",
         call. = FALSE)
  ok <- is.finite(empirical) & is.finite(theoretical)
  if(!any(ok)) return(NA_real_)
  sum((empirical[ok] - theoretical[ok])^2)
}

.GeoCovariogram_nearest_index <- function(grid, value) {
  grid <- as.numeric(grid)
  if(!length(grid) || !is.finite(value)) return(NA_integer_)
  which.min(abs(grid - value))
}

.GeoCovariogram_finite_max <- function(x, default = 0) {
  x <- as.numeric(x)
  x <- x[is.finite(x)]
  if(!length(x)) return(as.numeric(default)[1L])
  max(x)
}

.GeoCovariogram_practical_range <- function(lag, variogram, covariance, fraction) {
  lag <- as.numeric(lag)
  variogram <- as.numeric(variogram)
  covariance <- as.numeric(covariance)
  ok <- is.finite(lag) & is.finite(variogram) & is.finite(covariance)
  if(sum(ok) < 2L) return(NA_real_)
  lag <- lag[ok]; variogram <- variogram[ok]; covariance <- covariance[ok]
  oo <- order(lag)
  lag <- lag[oo]; variogram <- variogram[oo]; covariance <- covariance[oo]
  variance <- stats::median(variogram + covariance, na.rm = TRUE)
  if(!is.finite(variance) || variance <= 0) return(NA_real_)
  target <- fraction * variance
  jj <- which(variogram >= target)[1L]
  if(is.na(jj)) return(NA_real_)
  if(jj <= 1L) return(lag[1L])
  y <- variogram[c(jj - 1L, jj)]
  x <- lag[c(jj - 1L, jj)]
  if(!all(is.finite(y)) || diff(y) == 0) return(x[2L])
  stats::approx(x = y, y = x, xout = target, ties = "ordered")$y
}


if(show.range || answer.range)  {
    # Practical range in the Gaussian case:
    PracRangeNorm <- function(corrmodel, lags, lagt, nuisance, numlags, numlagt, param, pract.range)
    { 
        return(nuisance["sill"]*(1-nuisance["nugget"])*CorrelationFct(bivariate,corrmodel, lags, lagt, numlags, numlagt, mu,CkModel(fitted$model),nuisance, param,fitted_n)
            -(nuisance["sill"]*(1-pract.range)))
    
    }
}
###########################################################     
#############  starting ###################################
########################################################### 

    if(!inherits(fitted,"GeoFit") && !inherits(fitted,"GeoWLS"))
        stop("Enter an object obtained from GeoFit or WLeastSquare")
    if(inherits(fitted, "GeoFit") && identical(fitted$residual_type, "Pearson")) {
        stop(
            "GeoCovariogram() cannot compare a Pearson-residual variogram with the covariance model of the original discrete response. Use GeoCovariogram() on the original fit, or inspect the residual variogram without a fitted response-scale curve.",
            call. = FALSE
        )
    }

    dyn=FALSE
    OLS=NULL
    isvario <- !is.null(vario) # is empirical variogram is passed?
    bivariate <- fitted$bivariate
    if(bivariate) fitted$numtime=1
    ispatim <- fitted$spacetime
    dyn<- is.list(fitted$coordx_dyn)

    ## Continuous fitted objects do not need a trial/count parameter, but the
    ## shared native correlation interface requires one integer scalar.  Older
    ## GeoWLS objects may therefore have `n = NULL`; use the same default as
    ## GeoCorrFct() so GeoCovariogram remains backward compatible.
    fitted_n <- fitted$n
    if(is.null(fitted_n) || !length(fitted_n)) fitted_n <- 1L

    
    if(!isvario) stop("an object vario is needed\n")


if (fitted$model %in% c("Weibull", "Poisson","PoissonGamma", "Binomial", "Gamma", 
        "LogLogistic", "BinomialNeg", "Bernoulli", "Geometric", 
        "Gaussian_misp_Poisson", "PoissonZIP", "Gaussian_misp_PoissonZIP", 
        "Gaussian_misp_PoissonGamma","PoissonGammaZIP","PoissonGammaZIP1",
        "BinomialNegZINB", "PoissonZIP1", "Gaussian_misp_PoissonZIP1", 
        "BinomialNegZINB1", "Beta2", "Kumaraswamy2", "Beta", 
        "Kumaraswamy")) { fitted$param$sill=1}


do_plot <- isTRUE(show.cov) || isTRUE(show.vario)
if(do_plot) {
  opar <- par(no.readonly = TRUE)
  on.exit(par(opar))

  if(!ispatim && !bivariate && show.cov && show.vario) par(mfrow=c(1,2))
  if(show.vario && ispatim) par(mfrow=c(1,2))
  if(show.vario && ispatim && !is.null(fix.lags) && !is.null(fix.lagt)) par(mfrow=c(2,2))
  if(show.cov && ispatim && !show.vario) {
    nprof <- as.integer(!is.null(fix.lags)) + as.integer(!is.null(fix.lagt))
    if(nprof == 0L) par(mfrow=c(1,1))
    if(nprof == 1L) par(mfrow=c(1,2))
    if(nprof == 2L) par(mfrow=c(2,2))
  }
  if(bivariate && !dyn && (show.vario || show.cov)) par(mfrow=c(2,2))
  if(bivariate&&dyn) par(mfrow=c(1,2))
}

fitted$param=unlist(fitted$param)
fitted$fixed=unlist(fitted$fixed)

    ## Normalize the optional copula field.  Older GeoFit objects may encode
    ## "no copula" as NULL, NA, or an empty string.
    fit_copula <- fitted$copula
    if(!is.null(fit_copula)) {
      fit_copula <- as.character(fit_copula)[1L]
      if(is.na(fit_copula) || !nzchar(trimws(fit_copula))) fit_copula <- NULL
    }
    fitted$copula <- fit_copula
    
    # START ---- check input --- #

    if( !inherits(fitted,"GeoFit")  & !inherits(fitted,"GeoWLS") )
        stop("Enter an object obtained GeoFit or WLeastSquare")
  
    if(isvario & !inherits(vario,"GeoVariogram") )
        stop("Enter an object obtained from the function GeoVariogram")

    vario_bivariate <- isTRUE(vario$bivariate)
    vario_spacetime <- !is.null(vario$bint) || !is.null(vario$centert)
    if(vario_bivariate != isTRUE(bivariate))
        stop("The GeoVariogram and fitted objects disagree on bivariate structure.", call. = FALSE)
    if(vario_spacetime != isTRUE(ispatim))
        stop("The GeoVariogram and fitted objects disagree on space-time structure.", call. = FALSE)

    ## When omitted, inherit the metric stored by the fit.  The empirical lag
    ## values themselves come from `vario`, so explicit disagreement is a
    ## useful diagnostic even though the correlation function only sees lags.
    if(missing(distance) && !is.null(fitted$distance)) distance <- fitted$distance
    if(!is.null(fitted$distance) && length(fitted$distance) &&
       !identical(tolower(as.character(distance)[1L]),
                  tolower(as.character(fitted$distance)[1L])))
        warning("`distance` differs from the metric stored in the fitted object; ",
                "the GeoVariogram and fit should use the same distance metric.",
                call. = FALSE)

    if(!is.numeric(pract.range) || length(pract.range) != 1L ||
       is.na(pract.range) || !is.finite(pract.range))
        stop("Enter a finite numeric scalar for pract.range.", call. = FALSE)
    if(pract.range < 0 || pract.range > 100)
        stop("pract.range must lie between 0 and 100.", call. = FALSE)
    pract.range <- pract.range / 100
    # Set the type of process:
    model <- CkModel(fitted$model)
    gaussian <- model==1
    skewgausssian<- model==10
    skewlaplace<-model==59
    gamma<- model==21
    skewstudentT<- model==18||model==37
    studentT<- model==12||model==35
    weibull<- model==26
    twopieceT<- model==27
    twopieceTukeyh<- model==38 
    twopieceGauss<- model==29
    twopiecebimodal<- model==39
    loggauss<- model==22
    binary <- model==2
    binomial <- model==11
    binomial2 <- model==19
    geom <- model==14
    binomialneg<-model==16
    binomialnegZINB<-model==45
    tukeyh<- model ==34
    tukeyh2<- model ==40
    sas<- model ==20
    poisson<- model==30
    poissongamma<- model==46
    ## Zero-inflated count models.  Codes 53 and 58 are the one-nugget
    ## versions and must follow the same observed-scale covariance path.
    poissonZIP<- model %in% c(43,44,53)
    poissongammaZIP<- model %in% c(57,58)
    loglogistic <- model==24
    tukeygh<- model==9||model==41
    Gaussian_misp_Binomial<-model==51
    Gaussian_misp_Poisson<-model==36
    Gaussian_misp_PoissonGamma<- model==47
    Gaussian_misp_BinomialNeg<-model==52

    copula_cont <- !bivariate && !is.null(fit_copula) &&
      fit_copula %in% c("Gaussian", "Clayton", "SkewGaussian") &&
      as.integer(model) %in% .GeoGaussianCopulaContinuousModels

    ## Clayton-like and skew-Gaussian covariance is currently defined here
    ## for the audited continuous copula margins only.  Gaussian-copula
    ## discrete margins retain the legacy GeoCovariogram path below.
    if(!is.null(fit_copula) && fit_copula %in% c("Clayton", "SkewGaussian") &&
       !copula_cont)
      stop("GeoCovariogram supports copula='", fit_copula,
           "' for the continuous copula margins only.", call. = FALSE)

    zero <- 0;slow=1e-3;
    if(copula_cont || gaussian||skewgausssian||skewlaplace||gamma||loggauss||binomial||Gaussian_misp_Binomial||Gaussian_misp_BinomialNeg||Gaussian_misp_Poisson||binomialneg||binomialnegZINB||geom||tukeyh||tukeyh2||sas||twopiecebimodal||skewstudentT
            ||twopieceGauss||twopieceTukeyh||twopieceT) slow=1e-6
    else slow=1e-5 
    # lags associated to empirical variogram estimation
    if(isvario){
    lags <- c(0,vario$centers);numlags <- length(lags)
    if(ispatim) {
     if(is.null(vario$centert)) lagt <-c(0,vario$bint) 
     else  lagt <-c(0,vario$centert)
    }
    else {lagt=0}
    numlagt <- length(lagt)
    }

    # Check profile indices.  They index the empirical lag grids (including
    # the zero-lag row/column), while the fitted curve is matched by lag value.
    .GeoCovariogram_check_profile_index <- function(x, upper, label) {
      if(is.null(x)) return(NULL)
      if(!is.numeric(x) || length(x) != 1L || is.na(x) || !is.finite(x) ||
         x < 1 || x > upper || x != as.integer(x))
        stop(label, " must be a single integer in 1:", upper, ".", call. = FALSE)
      as.integer(x)
    }
    fix.lags <- .GeoCovariogram_check_profile_index(fix.lags, numlags, "fix.lags")
    fix.lagt <- .GeoCovariogram_check_profile_index(fix.lagt, numlagt, "fix.lagt")
    # END ---- check input --- #
    # set range (for practical range)
    if(tolower(fitted$corrmodel)=="matern"){
        lower <- 1e-10
        upper <- 1e20}
    else{
        lower <- 0  
        upper <- 1e100}

    num_betas <- fitted$numbetas
    if(is.null(num_betas) || !length(num_betas)) num_betas <- 1L
    mu=0;nuisance=0
    mm=c(mean=0)
    ## selecting nuisance, correlation parameters and a representative mean
      if(!bivariate){
        allpar <- c(fitted$fixed, fitted$param)
        param <- allpar[CorrelationPar(CkCorrModel(fitted$corrmodel))]
        nuisance <- allpar[NuisParam2(fitted$model,FALSE,num_betas, copula = fit_copula)]
        mean_names <- .GeoMean_names(num_betas)
        nuisance <- nuisance[!(names(nuisance) %in% mean_names)]

        nobs_mean <- length(unlist(fitted$data, use.names = FALSE))
        external_mean <- .GeoMean_external(fitted$fixed, n = nobs_mean)
        if(!is.null(external_mean)) {
          mean_reference <- mean(external_mean)
        } else {
          beta <- .GeoMean_beta(allpar, num_betas)
          Xfit <- fitted$X
          ## Backward compatibility for GeoWLS objects created before fix16:
          ## their intercept-only design could be stored as the scalar 1.
          ## Treat that legacy sentinel as the default intercept design rather
          ## than passing an ambiguous vector to .GeoMean_design().
          if(inherits(fitted, "GeoWLS") &&
             is.atomic(Xfit) && is.null(dim(Xfit)) && length(Xfit)==1L &&
             length(num_betas)==1L && num_betas==1L) Xfit <- NULL
          Xobs <- .GeoMean_design(Xfit, n = nobs_mean, p = num_betas, name = "fitted$X")
          mean_reference <- mean(as.numeric(Xobs %*% beta))
        }
        mm <- c(mean = mean_reference)
        }
      if(bivariate){
        param <- c(fitted$fixed,fitted$param)[CorrelationPar(CkCorrModel(fitted$corrmodel))]
        nuisance <- c(fitted$fixed,fitted$param)[NuisParam2(fitted$model,FALSE,num_betas)]
    }
    
     #############################################
     # computing the spatio-temporal distances where to compute the fitted model
     #############################################
    if(isvario) {

             if(!invisible) {   #### standard case
                              lags_m <- seq(slow,max(vario$centers),length.out =150)
                              if(ispatim)  lagt_m <-seq(slow,max(vario$bint),length.out =150)
                              else         lagt_m<-0
                            }
             else           {   ### invisible case
                                if (ispatim) {lags_m=c(slow,vario$centers)
                                              lagt_m <-c(slow,vario$bint)}
                                else         {lags_m=vario$centers
                                             lagt_m<-0}
                            }    
     }
    else{
        mmm <- double(2)
        type_dist <- CheckDistance(distance)
        p=.C("Maxima_Minima_dist",mmm=as.double(mmm),as.double(fitted$coordx),as.double(fitted$coordy),as.double(fitted$coordz),
        as.integer(fitted$numcoord),as.integer(type_dist),as.double(fitted$radius),
        PACKAGE='GeoModels', NAOK=TRUE)
        lags_m <- seq(slow,p$mmm[2],length=150)
        if (ispatim) {
            tt <- double(2)
            lagt_m <- seq(slow,max(c(dist(fitted$coordt))) ,length=150)
        }
        else lagt_m<-0
    }
    numlags_m <- length(lags_m)
    numlagt_m <- length(lagt_m)
if(!bivariate) {
    ## A covariogram indexed only by lag cannot retain a location-varying
    ## mean.  Use the average fitted linear predictor as representative.
    mu <- rep(as.numeric(mm["mean"]), numlags_m * numlagt_m)
  } else{}  
   ###############
    corrmodel <- CkCorrModel(fitted$corrmodel)

     nui=nuisance
     nui['sill']=1;
if(!(binomial||geom||binomialneg||binomialnegZINB||Gaussian_misp_Binomial||
       Gaussian_misp_BinomialNeg||Gaussian_misp_Poisson||Gaussian_misp_PoissonGamma)) nui['nugget']=0
else                                     nui['nugget']=nuisance['nugget']

    if(fitted$model=="Gaussian_misp_Binomial") fitted$model="Binomial"
    if(fitted$model=="Gaussian_misp_BinomialNeg") fitted$model="BinomialNeg"
    if(fitted$model=="Gaussian_misp_Poisson") fitted$model="Poisson"
    if(fitted$model=="Gaussian_misp_PoissonGamma") fitted$model="PoissonGamma"
  
    ## Copula covariances are functions of the correlation of the underlying
    ## Gaussian RF, not of a margin-specific transformed correlation.  In
    ## particular, using the Beta/Kumaraswamy model id here would apply an
    ## additional transformation before the copula covariance engine and make
    ## GeoCovariogram inconsistent with GeoCovmatrix/GeoKrig.
    ## For the zero-inflated models GeoCovariogram needs the correlation of
    ## the underlying latent Gaussian field and constructs the observed-scale
    ## covariance below.  VectCorrelation has no dedicated latent-only branch
    ## for all ZIP model codes (notably 53, 57 and 58), so request the Gaussian
    ## correlation explicitly rather than passing an unsupported ZIP code.
    correlation_model <- if(copula_cont || poissonZIP || poissongammaZIP)
                           CkModel("Gaussian") else CkModel(fitted$model)
    correlation <- CorrelationFct(bivariate,corrmodel, lags_m, lagt_m, numlags_m, numlagt_m,mu,
                                     correlation_model, nui,param,fitted_n)
        #correlation <- CorrelationFct(bivariate,corrmodel, lags_m, lagt_m, numlags_m, numlagt_m,mu,
         #                            CkModel(fitted$model), nui,param,fitted$n)

iscopula <- identical(fit_copula, "Gaussian")

    ## For the supported continuous copula margins, compute the theoretical
    ## covariance directly on the observed scale with exactly the same engine
    ## used by GeoCovmatrix() and GeoKrig().  A lag-only curve cannot retain a
    ## location-varying marginal predictor, so eta_ref is the representative
    ## mean predictor computed above.
    copula_variance <- NULL
    if(copula_cont) {
      eta_ref <- as.numeric(mm["mean"])[1L]
      if(!is.finite(eta_ref)) eta_ref <- 0
      nugget_cop <- suppressWarnings(as.numeric(nuisance["nugget"]))
      if(!length(nugget_cop) || !is.finite(nugget_cop[1L])) nugget_cop <- 0
      nugget_cop <- nugget_cop[1L]
      if(nugget_cop < 0 || nugget_cop >= 1)
        stop("GeoCovariogram requires a copula nugget in [0,1).", call. = FALSE)

      rho_cop <- (1 - nugget_cop) * as.numeric(correlation)
      params_cop <- c(mm, nuisance)
      covariance <- .GeoCopulaCrossCovFast(
        fit_copula, rho_cop, model, params_cop,
        eta1 = rep(eta_ref, length(rho_cop)),
        eta2 = rep(eta_ref, length(rho_cop))
      )
      copula_variance <- as.numeric(
        .GeoCopulaMarginalMoments(model, params_cop, eta_ref)$variance
      )[1L]
      if(!is.finite(copula_variance) || copula_variance <= 0)
        stop("Invalid marginal variance in copula covariogram.", call. = FALSE)
      variogram <- copula_variance - covariance
    }

##########################################
########### starting legacy cases ########
##########################################
if(!copula_cont) {
    if(gaussian){
    llm=length(lags_m)
        if(bivariate){  covariance11 <- correlation[(0*llm+1):(1*llm)]
                        covariance12 <- correlation[(1*llm+1):(2*llm)]
                        covariance22 <- correlation[(3*llm+1):(4*llm)]
                        variogram11  <- correlation[(4*llm+1):(5*llm)]
                        variogram12  <- correlation[(5*llm+1):(6*llm)]
                        variogram22  <- correlation[(7*llm+1):(8*llm)]
                           }
        else { 

        #covariance <- nuisance["nugget"]+nuisance["sill"]*correlation
        covariance <- as.numeric(nuisance["sill"])*correlation*(1-as.numeric(nuisance["nugget"]))
        #variogram <- nuisance["nugget"]+nuisance["sill"]*(1-correlation)
        variogram <-as.numeric(nuisance["sill"])*(1-correlation*(1-as.numeric(nuisance["nugget"])))
       
        }
    }
##########################################
 if(skewgausssian) {    
            if(bivariate) {}
              else {
        if(!iscopula){
              correlation1=(1-as.numeric(nuisance['nugget']) )*correlation  
              vv=as.numeric(nuisance['sill']);sk=as.numeric(nuisance['skew']);sk2=sk^2;corr2=correlation^2;  
              cc=(2*sk2)*(sqrt(1-corr2) + correlation*asin(correlation)-1)/(pi*vv+sk2*(pi-2)) + (correlation1*vv)/(vv+sk2*(1-2/pi))
              vs=(vv+sk2*(1-2/pi))
              covariance=vs*cc;variogram=vs*(1-cc) }
              else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)   )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
              }
          }
    }
##########################################
 if(skewlaplace) {    
            if(bivariate) {}
              else {
        if(!iscopula){
              correlation1=(1-as.numeric(nuisance['nugget']) )*correlation  
              vv=as.numeric(nuisance['sill']);sk=as.numeric(nuisance['skew']);sk2=sk*sk;corr2=correlation1^2;  
              cc=corr2
              vs= vv*(1-2*sk+2*sk2)/(sk2*(1-sk)^2)
              covariance=vs*cc;variogram=vs*(1-cc) }
              else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)   )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
              }
          }
    }
##########################################
   if(twopieceT)        { if(bivariate) {}
                        else {
                        if(!iscopula){
                              correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))
                              nu=1/as.numeric(nuisance['df']); sk=as.numeric(nuisance['skew']);sill=as.numeric(nuisance['sill'])
                              sk2=sk^2
                              vs=sill* ((nu/(nu-2))*(1+3*sk2) - 4*sk2*(nu/pi)*(gamma(0.5*(nu-1))/gamma(0.5*nu))^2)
                              corr2=correlation1^2;sk2=sk^2
                                  a1=Re(hypergeo::hypergeo(0.5,0.5,nu/2,corr2))
                                  a2=correlation1 *asin(correlation1) + (1-corr2)^(0.5)
                                  ll=qnorm((1-sk)/2)
                                  p11=pbivnorm::pbivnorm(ll,ll, rho = correlation, recycle = TRUE)
                                  a3=3*sk2 + 2*sk + 4*p11 - 1
                                  KK=( nu*(nu-2)*gamma((nu-1)/2)^2) / (nu*pi*gamma(nu/2)^2*(3*sk2+1)-4*sk2*nu*(nu-2)*gamma((nu-1)/2)^2 )
                                  cc= KK*(a1*a2*a3-4*sk2);
                              ##
                              covariance=vs*cc;variogram=vs*(1-cc)  
                          }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)  )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
              }

                          }
                  } 
  ##########################################
   if(twopieceTukeyh)        { if(bivariate) {}
                        else {
                          if(!iscopula){     
                              correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))
                              tail=as.numeric(nuisance['tail']); sk=as.numeric(nuisance['skew']);
                              sill=as.numeric(nuisance['sill'])
                              corr2=correlation1^2;sk2=sk^2
                              gg2=(1-(1-corr2)*tail)^2
                              xx=corr2/gg2
                              A=(asin(sqrt(xx))*sqrt(xx)+sqrt(1-xx))/(1-xx)^(1.5)
                              ll=qnorm((1-sk)/2)
                              p11=pbivnorm::pbivnorm(ll,ll, rho = correlation, recycle = TRUE)
                              a3=3*sk2 + 2*sk + 4*p11 - 1
                              mm=8*sk2/(pi*(1-tail)^2); 
                              ff=(1+3*sk2)/(1-2*tail)^(1.5)
                              M=(2*(1-corr2)^(3/2))/(pi*gg2)
                              cc=  (M*A*a3-mm)/( ff- mm)
                              vs=sill*(ff- mm) 
                              covariance=vs*cc;variogram=vs*(1-cc) 
                                  }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)   )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                   }

                          }
                  } 
##########################################
   if(twopieceGauss)        { 
                        if(bivariate) {}
                        else {     
             if(!iscopula){     
                        correlation1=correlation*(1-as.numeric(nuisance['nugget']))
                        corr2=sqrt(1-correlation1^(2))
                        sk=as.numeric(nuisance['skew']); sk2=sk^2
                        ll=qnorm((1-sk)/2)
                        p11=pbivnorm::pbivnorm(ll,ll, rho = correlation, recycle = TRUE)
                        KK=3*sk2+2*sk+ 4*p11 - 1
                        cc=(2*((corr2 + correlation1*asin(correlation1))*KK)- 8*sk2)/(3*pi*sk2  -  8*sk2   +pi   )
                        vs= as.numeric(nuisance['sill'])*(1+3*sk2-8*sk2/pi)
                        covariance=vs*cc;variogram=vs*(1-cc) 
                          }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)  )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
##########################################
 if(twopiecebimodal)        { if(bivariate) {}
                        else {      
                            if(!iscopula){                         
                                  correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))
                                  nu=as.numeric(nuisance['df']); sk=as.numeric(nuisance['skew'])
                                  delta=as.numeric(nuisance['shape'])
                                  alpha=2*(delta+1)/nu
                                  nn=2^(1-alpha/2)
                                  ll=qnorm((1-sk)/2)
                                  p11=pbivnorm::pbivnorm(ll,ll, rho = correlation, recycle = TRUE)
                                  corr2=correlation1^2;sk2=sk^2
                                  a1=Re(hypergeo::hypergeo(-1/alpha ,-1/alpha,nu/2,corr2))
                                  a3=3*sk2 + 2*sk + 4*p11 - 1
                                  MM=(2^(2/alpha)*(gamma(nu/2 + 1/alpha))^2) 
                                  vari=2^(2/alpha)*(gamma(nu/2 + 2/alpha))*gamma(nu/2)* (1+3*sk2) - sk2*2^(2/alpha+2)*gamma(nu/2+1/alpha)^2 
                                  cc= MM*(a1*a3-4*sk2)/vari
                                  vs=as.numeric(nuisance['sill'])*vari/(nn^(2/alpha)*gamma(nu/2)^2)
                                  covariance=vs*cc; variogram=vs*(1-cc)
                                         }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)   )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
##########################################

   if(studentT)        { if(bivariate) {}
                        else {
                                if(!iscopula){   
                              correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))
                              nu=1/as.numeric(nuisance['df']);sill=as.numeric(nuisance['sill'])
      
                              vs=sill*(nu)/(nu-2)
                              cc=((nu-2)*gamma((nu-1)/2)^2*Re(hypergeo::hypergeo(0.5,0.5 ,nu/2 ,correlation^2))*correlation1)/(2*gamma(nu/2)^2)
                              covariance=vs*cc;variogram=vs*(1-cc) 
                                            }
                                else
              {

                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)  )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
##########################################  
   if(skewstudentT)        { if(bivariate) {}
                        else {
                                if(!iscopula){   
                               correlation1=correlation*(1-nuisance['nugget'] )
                               nu=as.numeric(1/nuisance['df']); sk=as.numeric(nuisance['skew'])
                               sill=as.numeric(nuisance['sill'])
                               skew2=sk*sk;l=nu/2; f=(nu-1)/2; w=sqrt(1-skew2);y=correlation;
                               CorSkew=(2*skew2/(pi*w*w+skew2*(pi-2)))*(sqrt(1-y*y)+y*asin(y)-1)+w*w*correlation1/(w*w+skew2*(1-2/pi)) ;
                               cc=(pi*(nu-2)*gamma(f)^2/(2*(pi*gamma(l)^2-skew2*(nu-2)*gamma(f)^2)))*(Re(hypergeo::hypergeo(0.5,0.5,l,y*y))
                                *((1-2*skew2/pi)*CorSkew+2*skew2/pi)-2*skew2/pi);
                               mm=sqrt(nu)*gamma(f)*sk/(sqrt(pi)*gamma(l));
                               vs=sill*(nu/(nu-2)-mm*mm);
                               covariance=vs*cc;variogram=vs*(1-cc)
                                            }
                                else
              {
               
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)  )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
##########################################
  if(tukeyh)        { if(bivariate) {}
                        else {
                                if(!iscopula){   
                              correlation=correlation*(1-nuisance['nugget'] )
                              h=as.numeric(nuisance['tail'])
                              sill=as.numeric(nuisance['sill'])
                              vs=  (1-2*h)^(-1.5)     
                              cc=correlation*(1-2*h)^(1.5)/ ((1-h)^2-(h*correlation)^2)^(1.5)
                              covariance=sill*vs*cc;variogram=sill*vs*(1-cc)  
                                         }
                                else
              {

                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)  )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }

if (tukeyh2) { 
  if (bivariate) {
    # (bivariate case not handled here)
  } else {
    if (!iscopula) {
      nugget <- as.numeric(nuisance['nugget'])
      hr <- as.numeric(nuisance['tail1'])
      hl <- as.numeric(nuisance['tail2'])
      sill <- as.numeric(nuisance['sill'])
      mom_hh <- .GeoTukeyh2StandardMoments(hr, hl)
      rho_lat <- (1 - nugget) * as.numeric(correlation)
      cc <- .GeoTukeyh2Correlation(rho_lat, hr, hl)
      covariance <- sill * mom_hh$variance * cc
      variogram <- sill * mom_hh$variance * (1 - cc)
    } else {
      cova <- gaussian_copula_cov((1 - as.numeric(nuisance['nugget'])) * correlation,
                                  model, c(mm, nuisance))
      vv <- variance_disp(model,nuisance)
      covariance <- cova
      variogram <- vv - cova
    }
  }
}

##########################################
  if(tukeygh)        { if(bivariate) {}
                        else {
                     if(!iscopula){   
                              correlation=correlation*(1-nuisance['nugget'] )
                              rho=correlation
                              tail=as.numeric(nuisance['tail'])
                              sill=as.numeric(nuisance['sill'])
                              eta=as.numeric(nuisance['skew'])
                              rho2=rho*rho; eta2=eta*eta; tail2=tail*tail;

                        if(tail>1e-05&&abs(eta)>1e-05){
                               rho2=rho*rho; eta2=eta*eta; tail2=tail*tail;
                               u=1-tail; a=1+rho;
                               mu=(exp(eta2/(2*u))-1)/(eta*sqrt(u));
                               vs=(exp(2*eta2/(1-2*tail))-2*exp(eta2/(2*(1-2*tail)))+1)/(eta2*sqrt(1-2*tail))-mu*mu;
                               A1=exp(a*eta2/(1-tail*a));
                               A2=2*exp(0.5*eta2*  (1-tail*(1-rho2))  / (u*u- tail2*rho2)  );
                               A3=eta2*sqrt(u*u- rho2*tail*tail);
                               cc=((A1-A2+1)/A3-mu*mu)/vs;
                               covariance=sill*vs*cc;variogram=sill*vs*(1-cc)
                            } 
                        if(tail<=1e-05&&abs(eta)>1e-05){
                              vs=( -exp(eta^2)+exp(eta^2*2))*eta^(-2)
                              cc= (( -exp(eta^2)+exp(eta^2*(1+rho)))*eta^(-2))/vs
                              covariance=sill*vs*cc;variogram=sill*vs*(1-cc) 
                            } 
                        if(tail>1e-05&&abs(eta)<=1e-05){
                              vs=  (1-2*tail)^(-1.5)     
                              cc=(-rho/((1+h*(tail-1))*(-1+tail+tail*rho)*(1+tail*(-2+tail-tail*rho^2))^0.5))/vs
                              covariance=sill*vs*cc;variogram=sill*vs*(1-cc) 
                            } 
                        if(tail<=1e-05&&abs(eta)<=1e-05){
                            covariance=sill*rho;variogram=sill*(1-rho)
                            }  
                                
                     }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)  )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
######################################################
  if(sas) { if(bivariate) {}
                        else {
             if(!iscopula){    
nugget <- as.numeric(nuisance['nugget'])  # in [0,1]
d      <- as.numeric(nuisance['tail'])
e      <- as.numeric(nuisance['skew'])
sill   <- as.numeric(nuisance['sill'])
mom_sas <- .GeoSASStandardMoments(e, d)
rho_lat <- (1 - nugget) * as.numeric(correlation)
cc <- corrsas(rho_lat, e, d)
sigma2_marg <- sill * mom_sas$variance
covariance <- sigma2_marg * cc
variogram  <- sigma2_marg * (1 - cc)

                        }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)  )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
##########################################
 if(gamma)        { if(bivariate) {}
                        else {
                if(!iscopula){ 
                              correlation=correlation*(1-as.numeric(nuisance['nugget'] ))
                              vs=2*exp(mm)^2/as.numeric(nuisance['shape'])
                              cc=correlation^2
                              covariance=vs*cc;variogram=vs*(1-cc)  
                        }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)   )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
##########################################
 if(weibull)        { if(bivariate) {} 
                        else {
                if(!iscopula){        
                            ssh=as.numeric(nuisance['shape'])
                        correlation=correlation*(1-as.numeric(nuisance['nugget'] ))  
                        vs=exp(mm)^2*(gamma(1+2/ssh)/gamma(1+1/ssh)^2-1)
                        cc=.GeoWeibullCorr(correlation, ssh)
                        covariance=vs*cc;variogram=vs*(1-cc) 
                      }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)  )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
##########################################
  if(loglogistic)    { if(bivariate) {}  
                      else { 
              if(!iscopula){   
                     correlation=correlation*(1-as.numeric(nuisance['nugget'] ))
                     sh=as.numeric(nuisance["shape"])
                     vs=exp(mm)^2*(2*sh*sin(pi/sh)^2/(pi*sin(2*pi/sh))-1)
                     cc=((pi*sin(2*pi/sh))/(2*sh*(sin(pi/sh))^2-pi*sin(2*pi/sh)))*
                                    (Re(hypergeo::hypergeo(-1/sh, -1/sh, 1,correlation^2))*
                                     Re(hypergeo::hypergeo( 1/sh,  1/sh, 1,correlation^2)) -1)
                      covariance=vs*cc;variogram=vs*(1-cc)  

                }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)  )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
##########################################
  if(loggauss)    { if(bivariate) {}  
                      else {  
             if(!iscopula){    
                    vvar=as.numeric(nuisance["sill"])
                    yy=vvar*correlation*(1-as.numeric(nuisance["nugget"]))
                    cc<-(exp(yy)-1)/(exp(vvar)-1)
                    eta_ref <- as.numeric(mm["mean"])[1L]
                    vs<-exp(2*eta_ref)*(exp(vvar)-1)
                    covariance=vs*cc;variogram=vs*(1-cc)   
                    
                     }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)   )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
   #                   
   if(binary||binomial||binomial2||geom||binomialneg||binomialnegZINB||Gaussian_misp_Binomial||Gaussian_misp_BinomialNeg) {
                    if(bivariate) {}
                    else {      
                    if(!iscopula){      
                           pp=pnorm(mu)
                           if(binary||binomial||binomial2) vv=min(fitted$n)*pp*(1-pp)
                           if(geom)             vv=(1-pp)/pp^2;
                           if(binomialneg)      vv=(fitted$n)*(1-pp)/pp^2;
                           if(binomialnegZINB) { 
                                     pg=pnorm(as.numeric(nuisance['pmu']))
                                     vv=fitted$n*(1-pp)*(1-pg)*(1+fitted$n*pg*(1-pp)) /pp^2
                                               }
                           if(Gaussian_misp_Binomial||Gaussian_misp_BinomialNeg) vv=1                    
                           covariance=vv*correlation
                           variogram=vv*(1-correlation)
                          }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)  )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
###########################################################################              
     if(poisson||Gaussian_misp_Poisson) {
                    if(bivariate) {}
                    else {   
                if(!iscopula){  
                           correlation=(1-as.numeric(nuisance['nugget']))*correlation   
                           corr2=correlation^2    
                           vv=exp(mu);
                           z=2*vv/(1-corr2)
                           cc=corr2*(1-(besselI(z,0,expon.scaled = TRUE)+besselI(z,1,expon.scaled = TRUE)))
                           if(Gaussian_misp_Poisson) vv=1   
                           covariance=vv*cc
                           variogram=vv*(1-cc)
                    }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)   )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
########################################################################### 
   if(poissongamma||Gaussian_misp_PoissonGamma) {

                    if(bivariate) {}
                    else {
                             if(!iscopula){
                       ## Correct stationary PEM / Poisson-Gamma correlation
                       ## (Corollary 5.1).  Keep GeoCovariogram on the same
                       ## compiled kernel used by GeoCorrFct/GeoCovmatrix,
                       ## rather than the obsolete R-level hypergeometric
                       ## representation previously used here.
                       nug=as.numeric(nuisance['nugget'])
                       a=as.numeric(nuisance['shape'])
                       mean_pg=exp(mu)

                       ## VectCorrelation() returns the latent correlation for
                       ## model 46/47.  The nugget acts on that latent field;
                       ## at the same space-time point the response correlation
                       ## is exactly one.
                       if(ispatim) {
                         zero_lag=rep(lags_m,times=numlagt_m)==0 &
                                  rep(lagt_m,each=numlags_m)==0
                       }
                       else zero_lag=rep(lags_m==0,times=numlagt_m)
                       latent_corr=(1-nug)*correlation + nug*zero_lag

                       cc=.Call("GeoCorrPGsVecCall",
                         as.double(latent_corr),as.double(mean_pg[1L]),
                         as.double(a),PACKAGE="GeoModels")
                       if(any(zero_lag)) cc[zero_lag]=1

                       ## Var{M(s)} = mu * (1 + mu/kappa).
                       vv=mean_pg*(1+mean_pg/a)
                       covariance=vv*cc
                       variogram=vv*(1-cc)
                           }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)   )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  } 
              }
########################################################################### 
     if(poissonZIP) {
                    if(bivariate) {}
               else {
                  if(!iscopula){
                       ## Y(s)=I(s)X(s), where X is the renewal-Poisson count
                       ## field and I is the non-structural-zero indicator.
                       ## Hence
                       ## Cov(Y_i,Y_j)=p11 Cov(X_i,X_j)
                       ##              + mu_i mu_j {p11-(1-p)^2}.
                       ## The previous code used the ZIP variance, rather than
                       ## the Poisson mean, inside the Bessel correlation and
                       ## did not assign `covariance` at all.
                       p=pnorm(as.numeric(nuisance['pmu']))
                       q=1-p
                       MM=exp(mu)

                       if(ispatim) {
                         zero_lag=rep(lags_m,times=numlagt_m)==0 &
                                  rep(lagt_m,each=numlags_m)==0
                       } else zero_lag=rep(lags_m==0,times=numlagt_m)

                       ## PoissonZIP1 (model 53) uses one shared nugget; the
                       ## standard PoissonZIP/misspecified version uses two.
                       if(model==53) {
                         nug_count=as.numeric(nuisance['nugget'])
                         nug_zero=nug_count
                       } else {
                         nug_count=as.numeric(nuisance['nugget1'])
                         nug_zero=as.numeric(nuisance['nugget2'])
                       }
                       rho_count=(1-nug_count)*correlation+nug_count*zero_lag
                       rho_zero=(1-nug_zero)*correlation+nug_zero*zero_lag

                       p11=numeric(length(correlation))
                       p11[zero_lag]=q
                       jj=!zero_lag
                       if(any(jj))
                         p11[jj]=1-2*p+pbivnorm::pbivnorm(
                           as.numeric(nuisance['pmu']),
                           as.numeric(nuisance['pmu']),
                           rho=rho_zero[jj],recycle=TRUE)

                       ## Stationary underlying Poisson correlation.  At zero
                       ## lag impose the exact endpoint and avoid 1-rho^2=0.
                       cc_count=numeric(length(correlation))
                       cc_count[zero_lag]=1
                       if(any(jj)) {
                         corr2=rho_count[jj]^2
                         z=2*MM[jj]/(1-corr2)
                         cc_count[jj]=corr2*(1-(
                           besselI(z,0,expon.scaled=TRUE)+
                           besselI(z,1,expon.scaled=TRUE)))
                       }
                       cov_count=MM*cc_count
                       covariance=p11*cov_count+MM^2*(p11-q^2)
                       vv=q*MM*(1+p*MM)
                       variogram=vv-covariance
            }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)   )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  }
              }
       if(poissongammaZIP) {
                    if(bivariate) {}
                    else {
                    if(!iscopula){
                       ## Zero-inflated PEM / Poisson-Gamma covariance.  Use
                       ## the corrected stationary PEM kernel for the count
                       ## component, then apply the exact zero-inflation
                       ## covariance identity.  This removes the obsolete
                       ## R-level hypergeometric expression from this branch.
                       p=pnorm(as.numeric(nuisance['pmu']))
                       q=1-p
                       MM=exp(mu)
                       a=as.numeric(nuisance['shape'])

                       if(ispatim) {
                         zero_lag=rep(lags_m,times=numlagt_m)==0 &
                                  rep(lagt_m,each=numlags_m)==0
                       } else zero_lag=rep(lags_m==0,times=numlagt_m)

                       ## PoissonGammaZIP1 (model 58) has a shared nugget;
                       ## model 57 has separate count/zero-process nuggets.
                       if(model==58) {
                         nug_count=as.numeric(nuisance['nugget'])
                         nug_zero=nug_count
                       } else {
                         nug_count=as.numeric(nuisance['nugget1'])
                         nug_zero=as.numeric(nuisance['nugget2'])
                       }
                       rho_count=(1-nug_count)*correlation+nug_count*zero_lag
                       rho_zero=(1-nug_zero)*correlation+nug_zero*zero_lag

                       p11=numeric(length(correlation))
                       p11[zero_lag]=q
                       jj=!zero_lag
                       if(any(jj))
                         p11[jj]=1-2*p+pbivnorm::pbivnorm(
                           as.numeric(nuisance['pmu']),
                           as.numeric(nuisance['pmu']),
                           rho=rho_zero[jj],recycle=TRUE)

                       cc_count=.Call("GeoCorrPGsVecCall",
                         as.double(rho_count),as.double(MM[1L]),
                         as.double(a),PACKAGE="GeoModels")
                       if(any(zero_lag)) cc_count[zero_lag]=1
                       var_count=MM*(1+MM/a)
                       cov_count=var_count*cc_count

                       covariance=p11*cov_count+MM^2*(p11-q^2)
                       vv=q*MM*(1+MM/a+p*MM)
                       variogram=vv-covariance
                 }
                                else
              {
                cova=gaussian_copula_cov((1-as.numeric(nuisance['nugget']))*correlation, model,c(mm,nuisance)   )
                vv=variance_disp(model,nuisance)
                covariance=cova
                variogram=vv-cova
                        }
                  }
              }



}

##########################################
############ starting graphics############
##########################################
      vario.main <- "Spatial semivariogram"
      vario.ylab <- "SemiVariogram"
        if(ispatim){
            dim(covariance) <- c(numlags_m, numlagt_m)
            dim(variogram) <- c(numlags_m, numlagt_m)
            vario.main <- "Space-time semivariogram"
            vario.zlab <- "SemiVariogram" }
        # Compute the practical range on the observed marginal scale.  For
        # Gaussian models retain the exact correlation-root calculation.  For
        # non-Gaussian models use the fitted covariance/variogram themselves,
        # so the definition is consistent with GeoCovmatrix/GeoKrig.
        Range <- NA_real_
        if(show.range || answer.range){
            if(bivariate) {
              Range <- NA_real_
            } else if(gaussian && !copula_cont) {
              Range <- tryCatch(
                uniroot(PracRangeNorm, c(lower, upper), corrmodel=corrmodel,
                        nuisance=nuisance, numlags=1, numlagt=1, lagt=0,
                        param=param, pract.range=pract.range)$root,
                error = function(e) NA_real_
              )
            } else {
              if(ispatim) {
                jt0 <- .GeoCovariogram_nearest_index(lagt_m, 0)
                vcurve <- variogram[, jt0]
                ccurve <- covariance[, jt0]
              } else {
                vcurve <- variogram
                ccurve <- covariance
              }
              Range <- .GeoCovariogram_practical_range(
                lags_m, vcurve, ccurve, pract.range
              )
            }
        }

        ## Sum-of-squares diagnostic.  Compute it independently of plotting so
        ## invisible=TRUE is useful with show.vario=FALSE as documented.
        OLS <- NULL
        if(isTRUE(invisible)) {
          if(bivariate) {
            OLS <- .GeoCovariogram_sqerr(vario$variograms[1,], variogram11) +
                   .GeoCovariogram_sqerr(vario$variograms[2,], variogram22)
            if(!dyn)
              OLS <- OLS + .GeoCovariogram_sqerr(vario$variogramst, variogram12)
          } else if(ispatim) {
            nbint_ols <- if(is.null(vario$centert)) length(vario$bint) else length(vario$centert)
            empirical_st <- matrix(vario$variogramst,
                                   nrow=length(vario$centers),
                                   ncol=nbint_ols, byrow=TRUE)
            empirical_st <- rbind(c(0,vario$variogramt),
                                  cbind(vario$variograms, empirical_st))
            OLS <- .GeoCovariogram_sqerr(empirical_st, variogram)
          } else {
            OLS <- .GeoCovariogram_sqerr(vario$variograms, variogram)
          }
        }



if(show.cov){

#####################################
#### bivariate case covariance ######
##################################### 
        if(bivariate&&!dyn){
            #par(mfrow=c(2,2))
       plot(lags_m, covariance11, type='l', ylim=c(min(covariance11),
                     max(covariance11)), main="First covariance",
                     xlab="Distance", ylab="Covariance",...)
       plot(lags_m, covariance12, type='l', ylim=c(min(covariance12),
                     max(covariance12)), main="Cross covariance",
                     xlab="Distance", ylab="Covariance",...)      
       plot(lags_m, covariance12, type='l', ylim=c(min(covariance12),
                     max(covariance12)), main="Cross covariance",
                     xlab="Distance", ylab="Covariance",...)   
       plot(lags_m, covariance22, type='l', ylim=c(min(covariance22),
                     max(covariance22)), main="Second covariance",
                     xlab="Distance", ylab="Covariance",...)          
         }
     if(bivariate&&dyn){
            #par(mfrow=c(2,2))
       plot(lags_m, covariance11, type='l', ylim=c(min(covariance11),
                     max(covariance11)), main="First covariance",
                     xlab="Distance", ylab="Covariance",...)
       plot(lags_m, covariance22, type='l', ylim=c(min(covariance22),
                     max(covariance22)), main="Second covariance",
                     xlab="Distance", ylab="Covariance",...)          
         }

        
#######################################
#### space time case  covariance ######
#######################################
if(ispatim){
            # build the covariance matrix:
            plagt <- !is.null(fix.lags)
            plags <- !is.null(fix.lagt)
            numplot <- 1+plags+plagt
          
            par(mai=c(.2,.2,.2,.2))
            persp(lags_m, lagt_m, covariance, xlab="Distance", ylab="Time",
                  zlab="Covariance", ltheta=90,
                  shade=0.75, ticktype="detailed", phi=30,
                  theta=30,main="Fitted space-time covariance",
                  cex.axis=.8, cex.lab=.8)
            if(plagt){
                fixed_space <- lags[fix.lags]
                ii_fit <- .GeoCovariogram_nearest_index(lags_m, fixed_space)
                par(mai=c(.5,.5,.5,.5),mgp=c(1.6,.6,0))
                plot(lagt_m, covariance[ii_fit,], xlab="Time",
                     ylab="Covariance", type="l",cex.axis=.8,cex.lab=.8,
                     main=paste0("Space-time cov: temporal profile (h=",
                                 signif(fixed_space, 4), ")"),...)}
            if(plags){
                fixed_time <- lagt[fix.lagt]
                jj_fit <- .GeoCovariogram_nearest_index(lagt_m, fixed_time)
                par(mai=c(.5,.5,.5,.5),mgp=c(1.6,.6,0))
                plot(lags_m, covariance[,jj_fit], xlab="Distance",
                     ylab="Covariance", type="l",cex.axis=.8,cex.lab=.8,
                     main=paste0("Space-time cov: spatial profile (u=",
                                 signif(fixed_time, 4), ")"),...)
            }
        }
#######################################
#### spatial case  covariance #########
#######################################
       if(!ispatim && !bivariate){# spatial case:
            if(add.cov & dev.cur()!=1){
                lines(lags_m, covariance,...)
                if(show.range && is.finite(Range)) abline(v=Range)}
            else{
                plot(lags_m, covariance, type='l', #ylim=c(0,max(covariance)), 
                     main="Spatial covariance",
                     xlab="Distance", ylab="Covariance",...)
                if(show.range && is.finite(Range)) abline(v=Range)}}
         }
########################### 
########################### 
########################### 
    # display the semivariogram function
    if(show.vario){

#####################################
#### bivariate case semivariogram ###
##################################### 
 if(bivariate){
      if(bivariate&&!dyn){
       plot(vario$centers,vario$variograms[1,], main="First semivariogram",ylim=c(0,max(vario$variograms[1,])),
           xlim=c(0,max(vario$centers)),
                     xlab="Distance", ylab="SemiVariogram",...)
       lines(lags_m, variogram11, type='l',...)
       if(min(vario$variogramst)>0) {ll1=0;ll=max(vario$variogramst)}
       if(min(vario$variogramst)<0) {ll1=min(vario$variogramst);ll=-min(vario$variogramst)}
       plot(vario$centers,vario$variogramst, main="Cross semivariogram",ylim=c(ll1,ll),
         xlim=c(0,max(vario$centers)),
                     xlab="Distance", ylab="SemiVariogram",...)
       lines(lags_m, variogram12, type='l',...)
       plot(vario$centers,vario$variogramst, main="Cross semivariogram",ylim=c(ll1,ll),
         xlim=c(0,max(vario$centers)),
                     xlab="Distance", ylab="SemiVariogram",...)
       lines(lags_m, variogram12, type='l',...)
       plot(vario$centers,vario$variograms[2,], main="Second semivariogram",ylim=c(0,max(vario$variograms[2,])),
         xlim=c(0,max(vario$centers)),
                     xlab="Distance", ylab="SemiVariogram",...)
       lines(lags_m, variogram22, type='l',...)  

   }
 
   if(bivariate&&dyn){
       plot(vario$centers,vario$variograms[1,], main="First semivariogram",ylim=c(0,max(vario$variograms[1,])),
                     xlab="Distance", ylab="SemiVariogram",...)
       lines(lags_m, variogram11, type='l',...)
       plot(vario$centers,vario$variograms[2,], main="Second semivariogram",ylim=c(0,max(vario$variograms[2,])),
                     xlab="Distance", ylab="SemiVariogram",...)
       lines(lags_m, variogram22, type='l',...)  


        }
}
#######################################
#### space time case  semivariogram ###
#######################################

 if(ispatim){

 # spatio-temporal case:
            plagt <- !is.null(fix.lags)
            plags <- !is.null(fix.lagt)
            nrowp <- 1
            ncolp <- 1
            if(isvario){
                ncolp <- ncolp+1
                if(plags || plagt) nrowp <- nrowp+1}
            else ncolp <- ncolp+plags+plagt
            sup <- 0
            tup <- 0

            ## Profile bookkeeping.  For dynamic supports the empirical
            ## temporal margin at h=0 may be unavailable because no sites are
            ## collocated across distinct times.  In that special case, when
            ## fix.lags selects the zero-lag entry, GeoCovariogram falls back
            ## to the first positive-distance ST bin containing finite pairs.
            ## The fitted curve is evaluated at that bin center, while the
            ## title reports the actual empirical distance interval.
            profile_space_idx <- fix.lags
            profile_space_value <- if(plagt) lags[fix.lags] else NA_real_
            profile_space_fallback <- FALSE
            profile_space_lower <- NA_real_
            profile_space_upper <- NA_real_
            profile_space_first_bin <- FALSE

            if(isvario){
                nbins <- length(vario$centers)
                if(is.null(vario$centert)) nbint <- length(vario$bint)
                else nbint <- length(vario$centert)

                ## GeoVariogram stores the positive-distance/positive-time
                ## surface row-wise, with the spatial and temporal margins
                ## held separately.  Rebuild the rectangular empirical grid
                ## explicitly.  For dynamic supports gamma(0,u) is available
                ## only when locations are actually collocated across times;
                ## unavailable cells therefore remain NA rather than being
                ## silently represented as zero.
                est_cross <- matrix(vario$variogramst,
                                    nrow=nbins,ncol=nbint,byrow=TRUE)
                evario <- rbind(c(0,vario$variogramt),
                                cbind(vario$variograms,est_cross))
                evario.grid <- as.matrix(expand.grid(lags,lagt))
                if(nrow(evario) != length(lags) || ncol(evario) != length(lagt) ||
                   nrow(evario.grid) != length(evario))
                  stop("GeoCovariogram: inconsistent empirical space-time variogram grid.",
                       call.=FALSE)

                empirical_z <- as.numeric(evario)
                empirical_ok <- is.finite(empirical_z)
                if(any(empirical_ok) && requireNamespace("scatterplot3d", quietly=TRUE)) {
                  ## Vertical stems are useful for a fixed spatial support but
                  ## become a dense wall for dynamic supports.  Plot finite
                  ## binned values as points in the dynamic case instead.
                  scatterplot3d::scatterplot3d(
                    evario.grid[empirical_ok,1], evario.grid[empirical_ok,2],
                    empirical_z[empirical_ok],
                    type=if(dyn) "p" else "h",
                    pch=if(dyn) 20 else 1,
                    highlight.3d=!dyn,cex.axis=.7,cex.lab=.7,
                    main=if(dyn) "Empirical dynamic ST semivariogram" else paste("Empirical",vario.main),
                    xlab="Distance",ylab="Time",zlab=vario.zlab,
                    mar=c(2,2,2,2),mgp=c(0,0,0))
                } else {
                  plot.new()
                  title(if(dyn) "Empirical dynamic ST semivariogram" else paste("Empirical",vario.main))
                  if(any(empirical_ok)) {
                    graphics::text(0.5,0.5,"Install 'scatterplot3d' for the 3D empirical panel")
                  }
                }
                if(plagt) {
                  profile_info <- .GeoCovariogramDynamicTemporalProfile(
                    vario=vario,evario=evario,lags=lags,fix.lags=fix.lags,
                    dynamic=dyn)
                  profile_space_idx <- profile_info$index
                  profile_space_value <- profile_info$value
                  profile_space_fallback <- profile_info$fallback
                  profile_space_lower <- profile_info$lower
                  profile_space_upper <- profile_info$upper
                  profile_space_first_bin <- profile_info$first.bin
                }
                if(plagt) tup <- .GeoCovariogram_finite_max(evario[profile_space_idx,],0)
                if(plags) sup <- .GeoCovariogram_finite_max(evario[,fix.lagt],0)
            }

            par(mai=c(.2,.2,.2,.2))
            if(dyn) {
              persp(lags_m, lagt_m, variogram, xlab="Distance",
                    ylab="Time", zlab=vario.zlab, ltheta=90,
                    shade=0.35, col="grey85", border="grey45",
                    ticktype="detailed", phi=30, theta=30,
                    main="Fitted ST semivariogram", cex.axis=.8,cex.lab=.8)
            } else {
              persp(lags_m, lagt_m, variogram, xlab="Distance",
                    ylab="Time", zlab=vario.zlab, ltheta=90,
                    shade=0.75, ticktype="detailed", phi=30,
                    theta=30,main=vario.main, cex.axis=.8,cex.lab=.8)
            }

            vvv=nuisance["sill"]
            if(copula_cont) vvv <- copula_variance
            ########
            if(!copula_cont && gamma)         vvv=2*exp(mm["mean"])^2/as.numeric(nuisance["shape"])
            if(!copula_cont && weibull)       vvv=exp(mm["mean"])^2*(gamma(1+2/as.numeric(nuisance["shape"]))/gamma(1+1/as.numeric(nuisance["shape"]))^2-1)
            if(!copula_cont && loglogistic)   vvv=exp(mm["mean"])^2*
                               (2*as.numeric(nuisance['shape'])*sin(pi/nuisance['shape'])^2/(pi*sin(2*pi/nuisance['shape']))-1)
            if(!copula_cont && loggauss)      vvv=variance_disp(model,c(mm,nuisance))
            if(!copula_cont && binomial)      vvv=fitted$n*pnorm(mm['mean'])*(1-pnorm(mm['mean']))
            if(!copula_cont && poisson)       vvv=exp(mm['mean'])
            #if(poissongamma)  vvv=exp(mm['mean']*(1+1/nuisance["shape"]))
            if(!copula_cont && poissongamma)  {MM=exp(mm['mean']); vvv=MM*(1+MM/as.numeric(nuisance["shape"]))
        }


            if(!copula_cont && poissonZIP){        p=pnorm(nuisance['pmu']); MM=exp(mm['mean']); 
                              vvv=(1-p)*MM*(1+p*MM)}

            if(!copula_cont && poissongammaZIP){     p=pnorm(nuisance['pmu']); MM=exp(mm['mean']); kk=as.numeric(nuisance["shape"]);
                              vvv=(1-p)*MM*(1+p*MM+(MM/kk))}

            if(!copula_cont && geom)          vvv= (1-pnorm(mm['mean']))/pnorm(mm['mean'])^2
            if(!copula_cont && binomialneg)   vvv= fitted$n*(1-pnorm(mm['mean']))/pnorm(mm['mean'])^2
            if(!copula_cont && binomialnegZINB) {
                                  MM=pnorm(mm['mean']);pg=pnorm(nuisance['pmu'])
                                }
            if(!copula_cont && skewgausssian) vvv=variance_disp(model,c(mm,nuisance))
            if(!copula_cont && skewlaplace)   vvv=variance_disp(model,c(mm,nuisance))
            if(!copula_cont && studentT)      vvv=variance_disp(model,c(mm,nuisance))
            if(!copula_cont && tukeyh)        vvv=variance_disp(model,c(mm,nuisance))
            if(!copula_cont && sas)           { 
                              d <- as.numeric(nuisance['tail']); e <- as.numeric(nuisance['skew'])
                              MM <- sinh(e/d) * exp(0.25) * ( besselK(.25,(d+1)/(2*d)) + besselK(.25,(1-d)/(2*d)) ) / sqrt(8*pi)
                              vv_std <- cosh(2*e/d) * exp(0.25) * ( besselK(.25,(d+2)/(2*d)) + besselK(.25,(2-d)/(2*d)) ) / sqrt(32*pi) - 0.5 - MM^2
                              vvv <- as.numeric(nuisance['sill']) * vv_std   # <-- sill*vv
               }
            if(!copula_cont && tukeyh2)  {        hr=as.numeric(nuisance["tail1"]);hl=as.numeric(nuisance["tail2"]);
                                  mom_hh <- .GeoTukeyh2StandardMoments(hr, hl)
                                  vvv=as.numeric(nuisance["sill"]) * mom_hh$variance
                         }
            ########
            if(plagt){
                fixed_space <- profile_space_value
                ii_fit <- .GeoCovariogram_nearest_index(lags_m, fixed_space)
                par(mai=c(.5,.5,.3,.3),mgp=c(1.6,.6,0))

                temporal_title <- if(dyn && profile_space_fallback) {
                  if(is.finite(profile_space_upper)) {
                    ## Exact h=0 pairs are not included in the positive-
                    ## distance ST surface; they belong to the true temporal
                    ## margin when available.
                    if(profile_space_first_bin)
                      paste0("Temporal profile (0<h<",
                             signif(profile_space_upper,4), ")")
                    else
                      paste0("Temporal profile (",signif(profile_space_lower,4),
                             "<=h<",signif(profile_space_upper,4), ")")
                  } else {
                    paste0("Temporal profile (h~",signif(fixed_space,4),")")
                  }
                } else if(dyn) {
                  paste0("Temporal profile (h=",signif(fixed_space,4),")")
                } else {
                  paste0(vario.ylab, ": temporal profile (h=",
                         signif(fixed_space,4), ")")
                }

                plot(lagt_m, variogram[ii_fit,], xlab="Time",cex.axis=.8,cex.lab=.8,
                     ylab=vario.ylab, type="l", ylim=c(0,max(vvv,tup,na.rm=TRUE)),
                     main=temporal_title,...)
                if(isvario) {
                  emp_x <- lagt[-1]
                  emp_y <- evario[profile_space_idx,-1]
                  emp_ok <- is.finite(emp_x) & is.finite(emp_y)
                  if(any(emp_ok)) {
                    points(emp_x[emp_ok],emp_y[emp_ok],...)
                  } else if(dyn) {
                    msg <- if(abs(fixed_space) <= sqrt(.Machine$double.eps))
                      "No collocated dynamic pairs for empirical h=0 profile"
                    else "No empirical pairs in this dynamic profile"
                    mtext(msg,side=3,line=-1.15,cex=.62)
                  }
                }
               }

            if(plags){
                fixed_time <- lagt[fix.lagt]
                jj_fit <- .GeoCovariogram_nearest_index(lagt_m, fixed_time)
                par(mai=c(.5,.5,.3,.3),mgp=c(1.6,.6,0))
                plot(lags_m, variogram[,jj_fit], xlab="Distance",cex.axis=.8,cex.lab=.8,
                     ylab=vario.ylab, type="l", ylim=c(0,max(vvv,sup,na.rm=TRUE)),
                     main=if(dyn) paste0("Spatial profile (u=",signif(fixed_time,4),")")
                          else paste0(vario.ylab, ": spatial profile (u=",
                                      signif(fixed_time, 4), ")"),...)
                if(isvario) {
                  emp_x <- lags[-1]
                  emp_y <- evario[-1,fix.lagt]
                  emp_ok <- is.finite(emp_x) & is.finite(emp_y)
                  if(any(emp_ok)) points(emp_x[emp_ok],emp_y[emp_ok],...)
                  else if(dyn) mtext("No empirical pairs in this dynamic profile",
                                     side=3,line=-1.15,cex=.62)
                }
                }
  #################################

 } ####### end space time
          
################################
#### spaatial semivariogram ###
#################################
    if(!ispatim && !bivariate){

   #################################
  #################################
            if(add.vario & dev.cur()!=1){
                points(vario$centers, vario$variograms,...)
                lines(lags_m, variogram,...)
                if(show.range && is.finite(Range)) abline(v=Range)}
            else{
                bnds <- range(variogram)
                bnds[1] <- min(bnds[1], min(vario$variograms))
                bnds[2] <- max(bnds[2], max(vario$variograms))
        
                plot(lags_m, variogram, type='l',  
                     main=vario.main,xlab="Distance",
                     ylab=vario.ylab,...)
                points(vario$centers, vario$variograms,...)
                if(show.range && is.finite(Range)) abline(v=Range)}
        }
 #################################       
}
    if(do_plot && ispatim) par(mai=c(1.02 ,0.85 ,0.85 ,0.45),mgp=c(3,1,0))
    # return the estimated covariance function
    if(answer.cov) {
      if(!bivariate) {
        result <- list(lags=lags_m,lagt=lagt_m, covariance=covariance)
      } else if(gaussian) {
        result <- list(lags=lags_m,lagt=lagt_m,
                       covariance11=covariance11, covariance12=covariance12,
                       covariance22=covariance22)
      }
    }
    # return the estimated variogram function
    if(answer.vario) {
        if(!is.list(result)) {
            if(!bivariate) result <- list(lags=lags_m,lagt=lagt_m, variogram=variogram)
            if(bivariate && gaussian)
              result <- list(lags=lags_m,lagt=lagt_m, variogram11=variogram11,
                             variogram12=variogram12,variogram22=variogram22)
        } else {
            if(!bivariate) result$variogram <- variogram
            if(bivariate && gaussian) {
              result$variogram11 <- variogram11
              result$variogram12 <- variogram12
              result$variogram22 <- variogram22
            }
        }
    }
    if(answer.range) {
      if(!is.list(result)) result <- list(range = Range)
      else result$range <- Range
    }
   ########################################################## 
if(!is.null(result)) {
        invisible(result)
    } else if(!is.null(OLS)) {
        invisible(OLS)
    } else {
        invisible()
    }
  }

####################################################
### File name: GeoDoScores.r
####################################################

GeoDoScores <- function(data, method = "cholesky", matrix) {

  if (!inherits(matrix, "GeoCovmatrix")) {
    stop("A GeoCovmatrix object is needed as input\n")
  }

  method <- tolower(gsub("[[:blank:]]", "", as.character(method)[1L]))
  if (!(method %in% c("cholesky", "svd"))) {
    stop("method must be 'cholesky' or 'svd'", call. = FALSE)
  }

  ## The data length is the authoritative number of observations.  This also
  ## covers dynamic space-time layouts, where length(coordx) is not nsite.
  dime <- length(data)
  varcov <- matrix$covmatrix
  if (nrow(varcov) != length(data)) {
    stop("The dimension of the covariance matrix and/or the vector data are not correct\n")
  }
  dimnames(varcov) <- NULL
  data <- as.numeric(data)

  ## Return Q b and diag(Q), where Q is the inverse covariance matrix.  The
  ## Cholesky and SVD paths deliberately use their corresponding decomposition;
  ## in particular an svd object must never be passed to forwardsolve().
  getInv2 <- function(covmatrix, b) {
    if (!covmatrix$sparse) {
      dec <- MatDecomp(covmatrix$covmatrix, method)
      if (is.logical(dec)) {
        stop("Covariance matrix is not positive definite")
      }

      if (identical(method, "cholesky")) {
        vec <- forwardsolve(dec, b)
        Invc <- forwardsolve(dec, vec, transpose = TRUE)
        ## Q = U^{-T} U^{-1}.  GeoDoScores needs only diag(Q).
        I <- diag(nrow(covmatrix$covmatrix))
        Uinv <- forwardsolve(dec, I)
        diagInv <- colSums(Uinv^2)
      } else {
        ## For A = U D V^T, A^{-1} = V D^{-1} U^T.
        invd <- 1 / dec$d
        Invc <- as.numeric(dec$v %*% (invd * crossprod(dec$u, b)))
        diagInv <- rowSums(sweep(dec$v * dec$u, 2L, dec$d, "/"))
      }
    } else {
      if (!identical(method, "cholesky")) {
        stop("method='svd' is not available for sparse GeoCovmatrix objects", call. = FALSE)
      }

      cc <- covmatrix$covmatrix
      if (!spam::is.spam(cc)) cc <- spam::as.spam(cc)

      U <- tryCatch(
        spam::chol.spam(cc),
        error = function(e) stop("Covariance matrix is not positive definite")
      )
      vec <- spam::forwardsolve(U, b)
      Invc <- spam::backsolve(U, vec)
      Inv <- spam::solve.spam(U)
      diagInv <- diag(Inv)
    }

    diagInv <- as.numeric(diagInv)
    if (length(diagInv) != length(b) || any(!is.finite(diagInv)) || any(diagInv <= 0)) {
      stop("Invalid diagonal of the inverse covariance matrix", call. = FALSE)
    }
    list(a = as.numeric(Invc), c = diagInv)
  }

  param <- matrix$param
  external_mean <- matrix$mean_external
  if (is.null(external_mean)) external_mean <- .GeoMean_external(param, n = dime)
  if (!is.null(external_mean)) {
    external_mean <- as.numeric(external_mean)
    if (length(external_mean) != dime || any(!is.finite(external_mean)))
      stop("The external mean stored in the covariance object is invalid.")
    if (!is.null(matrix$X))
      stop("A fixed observation mean and X cannot be used together.")
    MM <- external_mean
  } else {
    matrix$X <- .GeoMean_design(matrix$X, n = dime, name = "X")
    beta <- .GeoMean_beta(param, ncol(matrix$X))
    MM <- as.numeric(matrix$X %*% beta)
  }

  namesnuis <- matrix$namesnuis
  nuisance <- param[namesnuis]
  model <- matrix$model
  data <- switch(
    as.character(model),
    # Gaussian, StudentT, Tukey, Logistic
    "1" = , "35" = , "12" = , "34" = , "25" = data - MM,
    # SkewGaussian
    "10" = {
      kk <- as.numeric(param['skew'])
      data - (MM + kk * sqrt(2/pi))
    },
    # Binomial
    "11" = data - matrix$n * pnorm(MM),
    # Poisson
    "30" = , "36" = data - exp(MM),
    # Poisson inflated
    "43" = , "44" = {
      p <- pnorm(as.numeric(param['pmu']))
      data - (1 - p) * exp(MM)
    },
    # Two piece t models
    "27" = {
      kk <- as.numeric(param['skew'])
      dd <- as.numeric(param['df'])
      ss <- as.numeric(param['sill'])
      adjustment <- (2 * kk * sqrt(ss * dd) * gamma((dd - 1) / 2)) /
                   (gamma(dd / 2) * sqrt(pi))
      data - (MM - adjustment)
    },
    # Two piece gaussian
    "29" = {
      kk <- as.numeric(param['skew'])
      ss <- as.numeric(param['sill'])
      data - (MM - 2 * kk * sqrt(2 * ss / pi))
    },
    # Two piece tukeyh
    "38" = {
      kk <- as.numeric(param['skew'])
      ss <- as.numeric(param['sill'])
      tt <- as.numeric(param['tail'])
      data - (MM - 2 * kk * sqrt(2 * ss / pi) / (1 - tt))
    },
    # Tukeyh2
    "40" = {
      ss <- as.numeric(param['sill'])
      t1 <- as.numeric(param['tail1'])
      t2 <- as.numeric(param['tail2'])
      data - (MM + sqrt(ss) * (t1 - t2) / (sqrt(2 * pi) * (1 - t1) * (1 - t2)))
    },
    # Binomial negative
    "16" = data - matrix$n * (1 - pnorm(MM)) / pnorm(MM),
    # Binomial negative inflated
    "45" = {
      p <- pnorm(as.numeric(param['pmu']))
      data - (1 - p) * matrix$n * (1 - pnorm(MM)) / pnorm(MM)
    },
    # SAS
    "20" = {
      ss <- as.numeric(param['sill'])
      kk <- as.numeric(param['skew'])
      tt <- as.numeric(param['tail'])
      bessel_term <- besselK(0.25, (tt + 1) / (2 * tt)) +
                     besselK(0.25, (1 - tt) / (2 * tt))
      adjustment <- sqrt(ss) * sinh(kk / tt) * exp(0.25) * bessel_term / sqrt(8 * pi)
      data - (MM + adjustment)
    },
    data
  )

  ## LOO identities for Q = Sigma^{-1}:
  ##   e_i       = (Q r)_i / Q_ii
  ##   Var_i_LOO = 1 / Q_ii
  ##   z_i       = e_i / sqrt(Var_i_LOO) = (Q r)_i / sqrt(Q_ii)
  cc <- getInv2(matrix, data)
  temp <- cc$a
  vv <- cc$c
  sd_loo <- 1 / sqrt(vv)
  residual_loo <- temp / vv
  zstd <- temp / sqrt(vv)

  MAD <- median(abs(residual_loo))
  RMSE <- sqrt(mean(residual_loo^2))
  MAE <- mean(abs(residual_loo))
  LSCORE <- mean(0.5 * (log(2 * pi / vv) + zstd^2))
  CRPS <- mean(
    sd_loo * (zstd * (2 * pnorm(zstd) - 1) +
                2 * dnorm(zstd) - 1 / sqrt(pi))
  )

  list(
    RMSE   = RMSE,
    LSCORE = LSCORE,
    MAD    = MAD,
    CRPS   = CRPS,
    MAE    = MAE
  )
}

####################################################
### File name: GeoResiduals.r
####################################################

GeoResiduals<-function(fit)

{
call <- match.call()
if(!inherits(fit,"GeoFit"))  stop("A GeoFit object is needed as input\n")
######
extmean=FALSE
if(!fit$bivariate) {
    mmext <- .GeoMean_external(fit$fixed)
    if(!is.null(mmext)) {
        extmean=TRUE; fit$fixed$mean=0
    }
}  ## external fixed mean
###
nm=names(fit$param); nf=names(fit$fixed)
fit$param=unlist(fit$param); fit$fixed=unlist(fit$fixed)
model=fit$model        #type of model
if(fit$bivariate && !(model %in% c("Gaussian", "Gauss"))) {
    stop("GeoResiduals currently supports bivariate residuals only for Gaussian models.",
         call. = FALSE)
}
num_betas=fit$numbetas  #number of mean parameters


temp_X=fit$X

if(!fit$bivariate)
{

res1 <- NULL

## extracting mean parameters
namescorr <- CorrParam(fit$corrmodel) 
namesnuis <- NuisParam(fit$model,fit$bivariate,num_betas, copula = fit$copula)
param <- c(fit$param, fit$fixed)
namesparam<- names(param)
paramcorr <- param[namescorr]
nuisance <- param[namesnuis]
sel <- names(nuisance) %in% .GeoMean_names(num_betas)


copula=fit$copula


#################################
#### extracting data ########
#################################
if(is.list(fit$coordx_dyn)) dd=unlist(fit$data)
else dd=c(t(fit$data))

##### if X is NULL, use an intercept-only design matrix with the correct length
##### This is important for purely spatial objects, where coordt can be NULL.
if(is.null(fit$X)) {
    fit$X <- matrix(1, nrow = length(dd), ncol = 1L)
} else {
    fit$X <- .GeoMean_design(fit$X, n = length(dd), name = "X")
}

#################################
#### computing mean ########
#################################
if(extmean) {
    if(length(mmext) != length(dd))
        stop("the fixed mean vector does not match the number of observations\n")
    mu <- as.numeric(mmext)
} else {
    beta2 <- .GeoMean_beta(param, ncol(fit$X))
    mu <- as.numeric(fit$X %*% beta2)
}


############################################################
#if(!is.null(copula)){    #### copula models
#if(copula=="Clayton"||copula=="Gaussian"){
if(model=="Beta2")
             {
              mm=c(1/(1+exp(-mu)))
              sh=as.numeric(param['shape'])
              pmin=as.numeric(param['min']); pmax=as.numeric(param['max'])
              if(!is.finite(pmin) || !is.finite(pmax) || pmax <= pmin)
                  stop("Beta2 residuals require finite min < max", call. = FALSE)
              zz=(dd-pmin)/(pmax-pmin)
              res1=pbeta(zz,mm*sh,(1-mm)*sh)
             }
if(model=="Kumaraswamy2") {
             mm=c(1/(1+exp(-mu)))
             sh=as.numeric(param['shape'])
             pmin=as.numeric(param['min']); pmax=as.numeric(param['max'])
             if(!is.finite(pmin) || !is.finite(pmax) || pmax <= pmin)
                 stop("Kumaraswamy2 residuals require finite min < max", call. = FALSE)
             zz=(dd-pmin)/(pmax-pmin)
             shape2=log(0.5)/log1p(-(mm^sh))
             if(any(!is.finite(shape2)) || any(shape2 <= 0))
                 stop("Kumaraswamy2 produced invalid fitted shape parameters", call. = FALSE)
             res1=numeric(length(zz))
             res1[zz >= 1] = 1
             inside=which(zz > 0 & zz < 1)
             if(length(inside))
                 res1[inside]=1-(1-zz[inside]^sh)^shape2[inside]
             }
          #}
#}
#else {           #### non-copula models
###  positive multiplicative models
if(model %in% c("Gamma","Weibull","LogLogistic","LogGaussian"))
{  
res1=dd/exp(c(mu))
}
### additive  models  on the real line
if(model %in% c("Gaussian","Gauss","SkewGaussian","Logistic", "SkewLaplace",
               "Tukeyh","Tukeyh2","SinhAsinh","Tukeygh","Gaussian_misp_Tukeygh",
               "StudentT",  "Gaussian_misp_StudentT","Gaussian_misp_SkewStudentT","SkewStudentT",
               "TwoPieceGaussian","TwoPieceTukeyh","TwoPieceGauss","TwoPieceStudentT","TwoPieceBimodal"))
{res1=(dd-c(mu))/sqrt(as.numeric(param['sill']))}


if(model=="Gaussian_misp_Binomial")
{
    aa=pnorm(c(mu))
    hh=fit$n*aa
    res1=(dd-hh)/sqrt(hh*(1-aa))
}
if(model=="Gaussian_misp_BinomialNeg")
{
    aa=pnorm(c(mu))
    hh=fit$n*(1-aa)/aa
    res1=(dd-hh)/sqrt(hh/aa)
}
if(model=="Gaussian_misp_Poisson")
{
    aa=exp(c(mu))
    res1=(dd-aa)/sqrt(aa)
}

## Pearson residuals for the discrete marginal models described in the paper.
## The formulas use their fitted marginal means and variances; they are
## deterministic and are intended mainly for residual variograms and
## standardized diagnostic plots.
.GeoResidual_n <- function(n, nobs, model_name) {
    if(length(n) == 1L) n <- rep(as.numeric(n), nobs)
    if(length(n) != nobs || any(!is.finite(n)) || any(n <= 0) || any(n != round(n)))
        stop(sprintf("For %s residuals, n must contain positive integers and have length 1 or the number of observations.", model_name),
             call. = FALSE)
    as.numeric(n)
}
.GeoResidual_pearson <- function(obs, fitted_mean, fitted_var) {
    if(any(!is.finite(fitted_mean)) || any(!is.finite(fitted_var)) ||
       any(fitted_var < 0))
        stop("Non-finite fitted marginal moments encountered while computing residuals.",
             call. = FALSE)
    den <- sqrt(pmax(fitted_var, .Machine$double.eps))
    (obs - fitted_mean) / den
}

if(model %in% c("Binary", "Bernoulli")) {
    prob <- pnorm(c(mu))
    res1 <- .GeoResidual_pearson(dd, prob, prob * (1 - prob))
}
if(model == "Binomial") {
    nn <- .GeoResidual_n(fit$n, length(dd), model)
    prob <- pnorm(c(mu))
    fitted_mean <- nn * prob
    res1 <- .GeoResidual_pearson(dd, fitted_mean, nn * prob * (1 - prob))
}
if(model == "BinomialLogistic") {
    nn <- .GeoResidual_n(fit$n, length(dd), model)
    prob <- plogis(c(mu))
    fitted_mean <- nn * prob
    res1 <- .GeoResidual_pearson(dd, fitted_mean, nn * prob * (1 - prob))
}
if(model %in% c("Geom", "Geometric", "BinomialNeg")) {
    nn <- .GeoResidual_n(if(model %in% c("Geom", "Geometric")) 1 else fit$n,
                         length(dd), model)
    prob <- pnorm(c(mu))
    fitted_mean <- nn * (1 - prob) / prob
    fitted_var <- nn * (1 - prob) / (prob^2)
    res1 <- .GeoResidual_pearson(dd, fitted_mean, fitted_var)
}
if(model == "Poisson") {
    lambda <- exp(c(mu))
    res1 <- .GeoResidual_pearson(dd, lambda, lambda)
}
if(model == "PoissonGamma") {
    fitted_mean <- exp(c(mu))
    shape <- as.numeric(param['shape'])
    if(length(shape) != 1L || !is.finite(shape) || shape <= 0)
        stop("PoissonGamma residuals require a positive finite shape parameter.",
             call. = FALSE)
    fitted_var <- fitted_mean * (1 + fitted_mean / shape)
    res1 <- .GeoResidual_pearson(dd, fitted_mean, fitted_var)
}
#########
#}

if(is.null(res1)) stop("residuals are not implemented for this model in GeoResiduals()\n")

fit$X <- matrix(1, nrow = length(dd), ncol = 1)

#### updating  object
mm=0
names(mm)="mean"
nuis_update=c(mm,nuisance[!sel])
fit$param=c(nuis_update,paramcorr)
fit$numbetas=1
fit$X <- matrix(1, nrow = length(dd), ncol = 1)

#if(!is.null(copula)){
#############################################
#if(copula=="Clayton"||copula=="Gaussian"){
if(model %in% c("Beta2")) {fit$param['shape']=2;fit$param['mean']=0; fit$param['max']=1; fit$param['min']=0}
if(model %in% c("Kumaraswamy2")) {fit$param['shape']=1;fit$param['mean']=0; fit$param['max']=1; fit$param['min']=0}
#}
#}else{
#####################################################
if(model %in% c("Gaussian","Logistic","Tukeyh","Tukeyh2","Tukeygh","SinhAsinh", "Gaussian_misp_StudentT","Gaussian_misp_Tukeygh",
         "StudentT","TwoPieceGauss","TwoPieceStudentT","TwoPieceGaussian","TwoPieceTukeyh","TwoPieceBimodal"))
{fit$param['sill']=1;fit$param['mean']=0}

if(model %in% c("SkewGaussian")) 
{
param['mean']=0;
 fit$param['skew']=as.numeric(param['skew'])/sqrt(as.numeric(param['sill']))
 fit$param['sill']=1
}


if(model %in% c("SkewLaplace"))

{
param['mean']=0;
fit$param['sill']=1
}




if(model %in% c("Gaussian_misp_SkewStudentT","SkewStudentT")) 
{param['mean']=0;
 fit$param['skew']=as.numeric(param['skew'])
 fit$param['df']= fit$param['df']
 fit$param['sill']=1
}
##
#}

## names of estimated and fixed parameters

obj=as.list(fit$param)
fit$param=obj[nm]
fit$fixed=obj[nf]

.drop_empty_na <- function(z) {
    if(length(z)==0) return(z)
    keep <- !vapply(z, function(v) length(v)==0L || all(is.na(v)), logical(1))
    z[keep]
}

fit$param=.drop_empty_na(fit$param)
fit$fixed=.drop_empty_na(fit$fixed)


###adding mean and variance if missing for some reason
if(model %in% c("Gaussian","SkewGaussian","SkewLaplace","Logistic","Tukeyh","Tukeyh2","Tukeygh","SinhAsinh",
 "Gaussian_misp_StudentT","Gaussian_misp_Tukeygh","StudentT","TwoPieceGauss",
 "TwoPieceStudentT","TwoPieceGaussian","TwoPieceTukeyh",
 "TwoPieceBimodal","Gaussian_misp_SkewStudentT","SkewStudentT")){
if(!sum(names(fit$param)=="mean")) fit$param["mean"]=0
if(!sum(names(fit$param)=="sill")) fit$param["sill"]=1

   sel=substr(names(unlist(fit$fixed)),1,4)=="mean"
            if(sum(sel)>=1) fit$fixed=fit$fixed[!sel]
}

if (model %in% c("Weibull", "Poisson", "PoissonGamma", "Binomial", "Gamma",  "LogGaussian", 
        "LogLogistic", "BinomialNeg", "Binary", "Bernoulli", "Geom", "Geometric", 
        "Gaussian_misp_Poisson", "PoissonZIP", "Gaussian_misp_PoissonZIP", 
        "BinomialNegZINB", "PoissonZIP1", "Gaussian_misp_PoissonZIP1", 
        "BinomialNegZINB1", "Beta2", "Kumaraswamy2", "Beta", 
        "Kumaraswamy")) {  
            if(!sum(names(fit$param)=="mean")) fit$param["mean"]=0              
            sel=substr(names(unlist(fit$fixed)),1,4)=="mean"
            if(sum(sel)>=1) fit$fixed=fit$fixed[!sel]
}


### formatting data
if(fit$spacetime)     ## spacetime
{if(!is.list(fit$coordx_dyn)) 
          data_res=matrix(res1,nrow=nrow(fit$data),ncol=ncol(fit$data),byrow=TRUE)
 else{ 
     ns=unlist(lapply(fit$coordx_dyn,nrow))
     sim_temp=list()
                    for(k in 1:length(fit$coordt))
                       { if(k==1) {indx=1:(sum(ns[1:k]))}
                         if(k>1)    {indx=(sum(ns[1:(k-1)])+1):(sum(ns[1:k]))}
                         sim_temp[[k]]=c(res1)[indx] }
    data_res=sim_temp  
 }
}   
else   {data_res=as.vector(res1)}   ## space
}
#####################################################################
#####################################################################
if(fit$bivariate)
{
 X=fit$X
 ns=fit$ns

namescorr <- CorrParam(fit$corrmodel) 
namesnuis <- NuisParam(fit$model,fit$bivariate,num_betas)
param <- c(fit$param, fit$fixed)
namesparam<- names(param)
paramcov <- param[namescorr]
nuisance <- param[namesnuis]
###
if(is.list(fit$coordx_dyn)) dd=unlist(fit$data)
else dd=c(t(fit$data))

dd1=dd[1:ns[1]]
dd2=dd[(ns[1]+1):(ns[2]+ns[1])]
###

###
 sel1=substr(names(nuisance),1,6)=="mean_1"
 beta1=as.numeric(nuisance[sel1])
 sel2=substr(names(nuisance),1,6)=="mean_2"
 beta2=as.numeric(nuisance[sel2])

if(!is.null(X)){
 X1=as.matrix(X[1:ns[1],]);
 X2=as.matrix(X[(ns[1]+1):(ns[2]+ns[1]),]); 
}
else {
     X1=matrix(1,nrow=ns[1],ncol=1)
     X2=matrix(1,nrow=ns[2],ncol=1)
}
mu1=X1%*%beta1
mu2=X2%*%beta2

if(model %in% c("Gaussian","Gauss")){
       res1=(dd1-c(mu1))/sqrt(as.numeric(paramcov['sill_1']))
       res2=(dd2-c(mu2))/sqrt(as.numeric(paramcov['sill_2']))
        }



if(!is.list(fit$coordx_dyn)) 
           {data_res=rbind(res1,res2)}
else {data_res=list(); data_res[[1]]=res1;data_res[[2]]=res2}
}



#####################################################################
#####################################################################

fit$data=data_res
## Preserve the complete fitted object.  Diagnostic metadata such as
## neighbourhood indices, thinning design, anisotropy, pair counts and
## convergence information must survive the residual transformation.
fit$param=as.list(fit$param)
fit$fixed=as.list(fit$fixed)
fit$X=NULL
fit$residual_type=if(model %in% c("Binary", "Bernoulli", "Binomial", "BinomialLogistic",
                                  "Geom", "Geometric", "BinomialNeg", "Poisson",
                                  "PoissonGamma"))
                    "Pearson" else "model-standardized"
fit$call=call
class(fit)="GeoFit"
fit
#########################
}

####################################################
### Internal utilities for bivariate Gaussian paths
####################################################

.GeoBivariateCoordinateInfo <- function(coordx = NULL, coordy = NULL,
                                        coordz = NULL, coordx_dyn = NULL,
                                        grid = FALSE,
                                        context = "Bivariate model") {
  fail <- function(...) stop(context, ": ", ..., call. = FALSE)

  normalize_block <- function(z, label) {
    if (is.data.frame(z)) z <- as.matrix(z)
    z <- as.matrix(z)
    if (!is.numeric(z) || any(!is.finite(z))) {
      fail(label, " must be a finite numeric coordinate matrix.")
    }
    if (!nrow(z)) fail(label, " must contain at least one location.")
    if (!(ncol(z) %in% c(2L, 3L))) {
      fail(label, " must have two or three coordinate columns.")
    }
    unname(z)
  }

  if (!is.null(coordx_dyn)) {
    if (!is.list(coordx_dyn) || length(coordx_dyn) != 2L) {
      fail("coordx_dyn must be list(coords_variable_1, coords_variable_2).")
    }
    blocks <- list(
      normalize_block(coordx_dyn[[1L]], "coordx_dyn[[1]]"),
      normalize_block(coordx_dyn[[2L]], "coordx_dyn[[2]]")
    )
    if (ncol(blocks[[1L]]) != ncol(blocks[[2L]])) {
      fail("the two coordinate blocks must have the same spatial dimension.")
    }
    ns <- as.integer(vapply(blocks, nrow, integer(1L)))
    return(list(dynamic = TRUE, blocks = blocks,
                coords = do.call(rbind, blocks), ns = ns,
                dimension = ncol(blocks[[1L]])))
  }

  if (isTRUE(grid)) {
    if (is.null(coordx) || is.null(coordy)) {
      fail("coordx and coordy are required when grid = TRUE.")
    }
    coords <- if (is.null(coordz)) {
      as.matrix(expand.grid(coordx, coordy))
    } else {
      as.matrix(expand.grid(coordx, coordy, coordz))
    }
  } else if (!is.null(coordy)) {
    coords <- as.matrix(cbind(coordx, coordy, coordz))
  } else {
    coords <- as.matrix(coordx)
  }

  coords <- normalize_block(coords, "coordinates")
  ns <- rep.int(as.integer(nrow(coords)), 2L)
  list(dynamic = FALSE, blocks = list(coords, coords), coords = coords,
       ns = ns, dimension = ncol(coords))
}


.GeoBivariateDataVector <- function(data, ns, context = "Bivariate model",
                                      allow_na = FALSE) {
  fail <- function(...) stop(context, ": ", ..., call. = FALSE)
  ns <- as.integer(ns)
  if (length(ns) != 2L || any(is.na(ns)) || any(ns < 1L)) {
    fail("invalid bivariate block sizes.")
  }

  if (is.list(data) && !is.data.frame(data)) {
    if (length(data) != 2L) fail("data must be a list of length two.")
    out1 <- as.numeric(data[[1L]])
    out2 <- as.numeric(data[[2L]])
  } else if (is.matrix(data) || is.data.frame(data)) {
    dm <- as.matrix(data)
    if (nrow(dm) == 2L && ns[1L] == ns[2L] && ncol(dm) == ns[1L]) {
      out1 <- as.numeric(dm[1L, ])
      out2 <- as.numeric(dm[2L, ])
    } else if (ncol(dm) == 2L && ns[1L] == ns[2L] && nrow(dm) == ns[1L]) {
      out1 <- as.numeric(dm[, 1L])
      out2 <- as.numeric(dm[, 2L])
    } else if (length(dm) == sum(ns)) {
      vv <- as.numeric(dm)
      out1 <- vv[seq_len(ns[1L])]
      out2 <- vv[ns[1L] + seq_len(ns[2L])]
    } else {
      fail("data dimensions do not match the two coordinate blocks.")
    }
  } else {
    vv <- as.numeric(data)
    if (length(vv) != sum(ns)) {
      fail("data must contain ", sum(ns), " observations in variable-block order.")
    }
    out1 <- vv[seq_len(ns[1L])]
    out2 <- vv[ns[1L] + seq_len(ns[2L])]
  }

  if (length(out1) != ns[1L] || length(out2) != ns[2L]) {
    fail("data lengths do not match the two coordinate blocks.")
  }
  out <- c(out1, out2)
  if (!is.numeric(out) ||
      (!isTRUE(allow_na) && any(!is.finite(out))) ||
      (isTRUE(allow_na) && any(is.infinite(out)))) {
    fail(if (isTRUE(allow_na))
           "data must be numeric and cannot contain infinite values."
         else "data must be finite and numeric.")
  }
  unname(out)
}


.GeoBivariateDesign <- function(X, ns, context = "Bivariate model",
                                allow_null = TRUE) {
  fail <- function(...) stop(context, ": ", ..., call. = FALSE)
  ns <- as.integer(ns)
  if (length(ns) != 2L || any(is.na(ns)) || any(ns < 1L)) {
    fail("invalid bivariate block sizes.")
  }

  if (is.null(X)) {
    if (!allow_null) fail("X is required.")
    return(matrix(1, nrow = sum(ns), ncol = 1L))
  }

  if (is.list(X) && !is.data.frame(X) && !is.matrix(X)) {
    if (length(X) != 2L) fail("X must be a list of length two.")
    X1 <- as.matrix(X[[1L]])
    X2 <- as.matrix(X[[2L]])
    if (nrow(X1) != ns[1L] || nrow(X2) != ns[2L]) {
      fail("the rows of X[[1]] and X[[2]] must match their coordinate blocks.")
    }
    if (ncol(X1) != ncol(X2)) {
      fail("X[[1]] and X[[2]] must have the same number of columns.")
    }
    X <- rbind(X1, X2)
  } else {
    X <- as.matrix(X)
    if (nrow(X) == sum(ns)) {
      # already stacked by variable
    } else if (ns[1L] == ns[2L] && nrow(X) == ns[1L]) {
      X <- rbind(X, X)
    } else {
      fail("X must have ", sum(ns), " rows, or one shared block of ",
           ns[1L], " rows when both variables use the same locations.")
    }
  }

  if (!is.numeric(X) || any(!is.finite(X)) || ncol(X) < 1L) {
    fail("X must be a finite numeric matrix.")
  }
  unname(X)
}


.GeoBivariateValidateFitMeans <- function(start, fixed, X, ns,
                                           context = "Bivariate model") {
  p <- if(is.null(X)) 1L else ncol(.GeoBivariateDesign(X, ns, context=context))
  expected <- c(.GeoMean_names(p, "mean_1"),
                .GeoMean_names(p, "mean_2"))
  start_names <- unique(c(.GeoMean_supplied_names(start, "mean_1"),
                          .GeoMean_supplied_names(start, "mean_2")))
  fixed_names <- unique(c(.GeoMean_supplied_names(fixed, "mean_1"),
                          .GeoMean_supplied_names(fixed, "mean_2")))

  overlap <- intersect(start_names, fixed_names)
  if(length(overlap)) {
    stop(context, ": a mean coefficient cannot be supplied in both start and fixed: ",
         paste(overlap, collapse=", "), ".", call.=FALSE)
  }
  supplied <- unique(c(start_names, fixed_names))
  unexpected <- setdiff(supplied, expected)
  if(length(unexpected)) {
    stop(context, ": for ", p, " design column(s), mean coefficients can only be named ",
         paste(expected, collapse=", "), ". Unexpected: ",
         paste(unexpected, collapse=", "), ".", call.=FALSE)
  }

  values <- c(as.list(start), as.list(fixed))
  for(nm in supplied) {
    value <- values[[nm]]
    if(length(value) != 1L || !is.numeric(value) || !is.finite(value)) {
      stop(context, ": mean coefficient '", nm,
           "' must be a finite numeric scalar.", call.=FALSE)
    }
  }
  invisible(expected)
}


.GeoBivariateBetas <- function(param, p, variable,
                               context = "Bivariate model") {
  if (!(variable %in% c(1L, 2L))) {
    stop(context, ": variable must be 1 or 2.", call. = FALSE)
  }
  .GeoMean_beta(param, p = p, prefix = paste0("mean_", variable))
}


.GeoBivariateValidateMeanParameters <- function(param, p,
                                                   context = "Bivariate model") {
  expected <- c(.GeoMean_names(p, "mean_1"),
                .GeoMean_names(p, "mean_2"))
  supplied <- unique(c(
    .GeoMean_supplied_names(param, "mean_1"),
    .GeoMean_supplied_names(param, "mean_2")
  ))
  unexpected <- setdiff(supplied, expected)

  ## Intercept-only bivariate models historically default both means to zero.
  ## Preserve that convention, but require every coefficient when X has more
  ## than one column.
  param_use <- as.list(param)
  if (p == 1L) {
    if (!("mean_1" %in% names(param_use))) param_use$mean_1 <- 0
    if (!("mean_2" %in% names(param_use))) param_use$mean_2 <- 0
  }
  missing <- setdiff(expected, names(param_use))

  if (length(missing) || length(unexpected)) {
    msg <- paste0(context, ": the mean specification must contain ",
                  paste(expected, collapse = ", "), ".")
    if (length(missing))
      msg <- paste0(msg, " Missing: ", paste(missing, collapse = ", "), ".")
    if (length(unexpected))
      msg <- paste0(msg, " Unexpected: ", paste(unexpected, collapse = ", "), ".")
    stop(msg, call. = FALSE)
  }
  list(
    beta1 = .GeoBivariateBetas(param_use, p, 1L, context),
    beta2 = .GeoBivariateBetas(param_use, p, 2L, context)
  )
}


.GeoBivariateObservedMean <- function(param, X, ns,
                                      context = "Bivariate model") {
  X <- .GeoBivariateDesign(X, ns, context = context)
  p <- ncol(X)
  beta <- .GeoBivariateValidateMeanParameters(param, p, context)
  b1 <- beta$beta1
  b2 <- beta$beta2
  i1 <- seq_len(ns[1L])
  i2 <- ns[1L] + seq_len(ns[2L])
  as.numeric(c(X[i1, , drop = FALSE] %*% b1,
               X[i2, , drop = FALSE] %*% b2))
}


.GeoBivariatePredictionMean <- function(param, Xloc, nloc, which,
                                        p, Mloc = NULL,
                                        context = "Bivariate model") {
  if (!is.null(Mloc)) {
    Mloc <- as.numeric(Mloc)
    if (length(Mloc) == 1L) Mloc <- rep(Mloc, nloc)
    if (length(Mloc) != nloc || any(!is.finite(Mloc))) {
      stop(context, ": Mloc must contain one finite value per prediction location.",
           call. = FALSE)
    }
    return(Mloc)
  }
  Xloc <- .GeoMean_design(Xloc, n = nloc, p = p, name = "Xloc")
  beta <- .GeoBivariateValidateMeanParameters(param, p, context)
  beta_target <- if (as.integer(which) == 1L) beta$beta1 else beta$beta2
  as.numeric(Xloc %*% beta_target)
}


.GeoBivariateBlockDesign <- function(X, ns) {
  X <- .GeoBivariateDesign(X, ns)
  p <- ncol(X)
  out <- matrix(0, nrow = sum(ns), ncol = 2L * p)
  i1 <- seq_len(ns[1L])
  i2 <- ns[1L] + seq_len(ns[2L])
  out[i1, seq_len(p)] <- X[i1, , drop = FALSE]
  out[i2, p + seq_len(p)] <- X[i2, , drop = FALSE]
  colnames(out) <- c(.GeoMean_names(p, "mean_1"),
                     .GeoMean_names(p, "mean_2"))
  out
}


.GeoBivariateTargetBlockDesign <- function(Xloc, nloc, p, which) {
  Xloc <- .GeoMean_design(Xloc, n = nloc, p = p, name = "Xloc")
  out <- matrix(0, nrow = nloc, ncol = 2L * p)
  if (which == 1L) out[, seq_len(p)] <- Xloc
  else out[, p + seq_len(p)] <- Xloc
  colnames(out) <- c(.GeoMean_names(p, "mean_1"),
                     .GeoMean_names(p, "mean_2"))
  out
}

####################################################
### File name: GeoSimapprox.r
####################################################

# Simulate approximate spatial and spatio-temporal random felds:
GeoSimapprox <- function(coordx=NULL, coordy=NULL,coordz=NULL,coordt=NULL, coordx_dyn=NULL,corrmodel, distance="Eucl", grid=FALSE,
 max.ext=1,method="TB", L=10000,model='Gaussian', parallel=FALSE,ncores=6,
     n=1, param, anisopars=NULL,radius=6371,X=NULL,spobj=NULL,nrep=1,progress=TRUE,check.duplicates=FALSE)
{
    call <- match.call()
####################################################################
############  starting internal function ###########################
####################################################################
ddim<-function(coordx,coordy,coordz,coordt,grid=FALSE)
{
  dimt <- if(is.null(coordt)) 1L else length(coordt)
  if(isTRUE(grid)) {
    dims <- length(coordx) * max(1L, length(coordy)) * max(1L, length(coordz))
  } else if(!is.null(dim(coordx))) {
    dims <- nrow(as.matrix(coordx))
  } else {
    dims <- length(coordx)
  }
  dims * dimt
}
###############################
RFfct1<- function(numcoord,numtime,spacetime,bivariate,dime,nuisance,simd,X,ns)
    {
        if(!bivariate) {
          external_mean <- .GeoMean_external(nuisance, n = dime)
          if(!is.null(external_mean)) {
            sim <- external_mean + as.numeric(simd)
          } else {
            X <- .GeoMean_design(X, n = dime, name = "X")
            beta <- .GeoMean_beta(nuisance, ncol(X))
            sim <- as.numeric(X %*% beta) + as.numeric(simd)
          }
        }
        if(bivariate) {
          X <- .GeoBivariateDesign(X, ns, context = "GeoSimapprox")
          mean_vec <- .GeoBivariateObservedMean(
            nuisance, X, ns, context = "GeoSimapprox"
          )
          sim <- mean_vec + as.numeric(simd)
        }

        if(!spacetime&&!bivariate) sim <- c(sim)
        else if(bivariate) {
          ## The approximate bivariate kernel uses a common set of sites.
          sim <- matrix(sim, nrow=2L, ncol=ns[1L], byrow=TRUE)
        } else sim <- matrix(sim, nrow=numtime, ncol=numcoord,byrow=TRUE)
        return(sim)
    }
#########################################################

######################################################################
############## main  internal function  ##############################
######################################################################
simu_approx=function(numxgrid,numygrid,coordx,coordy,coordz,coords,coordt,method,corrmodel,param,M,L,bivariate,spacetime,parallel,ncores)
{
##### spatial case 


if(!spacetime){    
   ## Turning Bands
 if(method=="TB") { 
        if(bivariate)
        {
          tau1 <- as.numeric(param$nugget_1)
          tau2 <- as.numeric(param$nugget_2)
          sill1 <- as.numeric(param$sill_1)
          sill2 <- as.numeric(param$sill_2)
          if(any(lengths(list(tau1,tau2,sill1,sill2)) != 1L) ||
             any(!is.finite(c(tau1,tau2,sill1,sill2))) ||
             tau1 < 0 || tau1 >= 1 || tau2 < 0 || tau2 >= 1 ||
             sill1 <= 0 || sill2 <= 0)
            stop("Bivariate TB simulation requires positive sills and nugget parameters in [0,1).", call.=FALSE)
          raw <- tbm2d(coords,coordt, param, corrmodel,L,bivariate)
          if(is.null(raw)) return(NULL)
          raw <- as.matrix(raw)
          raw[,1] <- sqrt(sill1) * (sqrt(1-tau1) * raw[,1] + sqrt(tau1) * rnorm(nrow(raw)))
          raw[,2] <- sqrt(sill2) * (sqrt(1-tau2) * raw[,2] + sqrt(tau2) * rnorm(nrow(raw)))
          simu <- raw
        }
        else
        {   tau2=param$nugget
            vvv=param$sill
            ##----------###
            param$nugget=0
            param$sill=1
            ##----------###

         simu=tbm2d_uni(coords, coordt, param, corrmodel, L, bivariate,parallel,ncores)$simu


         if(is.null(simu)) {print("TB simulation method fails\n");return(simu)}
         simu=sqrt(vvv)*(sqrt(1-tau2)*simu+sqrt(tau2)*rnorm(length(simu)))
        }
           simu=c(simu)
    }     


 ## Circulant embeeding           for regular grid          
  if(method=="CE") simu= c(SimCE(numxgrid,numygrid,coordx,coordy,coordz,corrmodel,param,mean.val=0, max.ext)$X)        
}
##### spacetime case ######
if(spacetime)  {
 # if(method=="Vecchia") print("ciao") 
  if(method=="CE") simu=CE_Space_Time(
      coords=coords, time.seq=coordt, param=param,
      corrmodel=corrmodel, distance=distance, max.ext=max.ext
    )
    }
return(simu)
}
####################################################################
####################################################################
############# END internal functions ###############################
####################################################################
####################################################################
    M=NULL
    corrmodel <- .GeoSimulationCanonicalCorrModel(corrmodel)
    if(is.null(CkCorrModel(corrmodel))) stop("The name of the correlation model is not correct\n")
    if(!(method=="TB"||method=="CE")) stop("The method of simulation is not correct\n")
    model_requested <- gsub("[[:blank:]]", "",model)
    model <- .GeoSimulationCanonicalModel(model_requested)
    n <- .GeoSimulationEffectiveN(model_requested, model, n)
    nrep <- .GeoValidateNrep(nrep)
    .GeoValidateDirectSimulationModel(model, simulator = "GeoSimapprox")
    distance=gsub("[[:blank:]]", "",distance)
    method=gsub("[[:blank:]]", "",method)
    if(method == "TB") L <- .GeoValidateTBControls(L)
    numxgrid=numygrid=NULL
    spacetime_dyn <- !is.null(coordx_dyn)
    coordx_dyn_output <- coordx_dyn
    grid_output <- grid

param <- lapply(param, unname) 
param_input <- param

# Enforce the finite-moment Tukey-hh parameter domain at the public
# simulation entry point.  Keep this explicit even though the common
# marginal validator performs the same check: GeoSim/GeoSimapprox must
# never proceed with h_l or h_r >= 1/2.
if (identical(model, "Tukeyh2")) {
    for (nm in c("tail1", "tail2")) {
        value <- suppressWarnings(as.numeric(param[[nm]]))
        if (length(value) != 1L || !is.finite(value) || value < 0 || value >= 0.5)
            stop(sprintf("param$%s must lie in [0, 0.5) for Tukeyh2 simulation.", nm),
                 call. = FALSE)
    }
}

param <- .GeoValidateLatentIntegerParameters(model, param)
param <- .GeoValidateSimulationMarginal(model, param)

### some check for parallel #######
n.cores <- .GeoResolveWorkers(parallel, ncores)
parallel <- isTRUE(parallel) && n.cores > 1L
##################################
## Flag consumed by simu_approx()/tbm2d_uni() calls below: whether the
## TB simulation should parallelize internally over spatial chunks.
## Its actual value is decided later (right before running the
## replicates), based on whether outer (nrep) or inner (points)
## parallelism is the better fit -- only one of the two is ever active.
parallel_inner <- FALSE
##################################

##############################################################################
###### extracting sp object informations if necessary              ###########
##############################################################################
bivariate<-CheckBiv(CkCorrModel(corrmodel))
spacetime<-CheckST(CkCorrModel(corrmodel))
space=!spacetime&&!bivariate
if(bivariate && model != "Gaussian") {
  stop("GeoSimapprox currently supports bivariate approximate simulation only for model = 'Gaussian'; use GeoSim() for supported bivariate simulation.",
       call. = FALSE)
}
if(method == "TB") .GeoValidateTBCorrelationModel(corrmodel, bivariate = bivariate)
biv_info <- NULL
if(!is.null(spobj)) {
   if(space||bivariate){
        a=sp2Geo(spobj,NULL); coordx=a$coords 
       if(!a$pj) {if(distance!="Chor") distance="Geod"}
    }
   if(spacetime){
        a=sp2Geo(spobj,NULL); coordx=a$coords ; coordt=a$coordt 
        if(!a$pj) {if(distance!="Chor") distance="Geod"}
     }
}
if(method == "TB") .GeoValidateTBDistance(distance, context = "GeoSimapprox")
if(bivariate) {
  biv_info <- .GeoBivariateCoordinateInfo(
    coordx=coordx, coordy=coordy, coordz=coordz,
    coordx_dyn=coordx_dyn, grid=grid, context="GeoSimapprox"
  )
  if(biv_info$dynamic) {
    same_support <- identical(dim(biv_info$blocks[[1L]]), dim(biv_info$blocks[[2L]])) &&
      isTRUE(all.equal(biv_info$blocks[[1L]], biv_info$blocks[[2L]],
                       tolerance=sqrt(.Machine$double.eps), check.attributes=FALSE))
    if(!same_support)
      stop("GeoSimapprox: approximate bivariate simulation requires common locations for the two variables; use GeoSim for heterotopic coordinates.", call.=FALSE)
  }
  ## tbm2d simulates both components on one common coordinate matrix.
  coordx <- biv_info$blocks[[1L]]
  coordy <- coordz <- NULL
  coordx_dyn <- NULL
  spacetime_dyn <- FALSE
  grid <- FALSE
}
if(method == "CE" && bivariate) {
  stop("GeoSimapprox does not support bivariate simulation with method = 'CE'; use method = 'TB' or GeoSim().",
       call. = FALSE)
}
if(spacetime && method == "TB") {
  stop("GeoSimapprox does not implement turning-bands simulation for spatio-temporal models; use method = 'CE' for supported fixed-location separable models or use GeoSim().",
       call. = FALSE)
}
if(method == "CE" && spacetime_dyn) {
  stop("CE spatio-temporal simulation requires fixed spatial locations; use GeoSim() for dynamic locations.",
       call. = FALSE)
}
.GeoCheckDuplicateLocations(
  coordx = coordx, coordy = coordy, coordz = coordz, coordt = coordt,
  coordx_dyn = coordx_dyn, spacetime = spacetime, bivariate = bivariate,
  grid = grid, context = "GeoSimapprox", check.duplicates = check.duplicates
)
###############################################################
###############################################################

    if(spacetime_dyn) {
      coords <- do.call(rbind, lapply(coordx_dyn, as.matrix))
    } else if(grid) {
      xgrid=coordx;ygrid=coordy
      numxgrid=length(xgrid);numygrid=length(ygrid)
      if(is.null(coordz)) {
        coords=as.matrix(expand.grid(xgrid,ygrid))
      } else {
        zgrid=coordz;numzgrid=length(zgrid)
        coords=as.matrix(expand.grid(xgrid,ygrid,zgrid))
      }
    } else {
      coords=coordx
      if(!is.null(coordy)) coords=cbind(coordx,coordy,coordz)
      coords <- as.matrix(coords)
    }
    coords_orig=coords


    .GeoValidate3D(
      coordx = if (is.null(coordx_dyn)) coords else NULL,
      coordx_dyn = coordx_dyn, distance = distance, grid = grid,
      approximate = TRUE, method = method, context = "GeoSimapprox"
    )

####
    spatial_dimension <- ncol(coords)
    if(method == "TB" && identical(spatial_dimension, 3L)) {
      if(!isTRUE(space) || isTRUE(bivariate) || isTRUE(spacetime)) {
        stop("GeoSimapprox supports three-dimensional TB simulation only for purely spatial univariate fields.",
             call. = FALSE)
      }
      if(!is.null(anisopars)) {
        stop("GeoSimapprox does not support anisopars for three-dimensional TB simulation; transform the coordinates explicitly.",
             call. = FALSE)
      }
    }

    numtime <- if(is.null(coordt)) 1L else length(coordt)
    if(spacetime_dyn) numtime=1L
    numcoord=nrow(coords)
    dime <- if(spacetime_dyn) nrow(coords) else numcoord*numtime
    if(bivariate) numtime=2
    if(!bivariate && model %in% c("Binomial", "BinomialLogistic")) {
      n <- .GeoValidateIntegerVector(n, "n", nobs = dime, minimum = 1L)
    }
    if(!bivariate && model %in% c("BinomialNeg", "BinomialNegZINB")) {
      n <- .GeoValidateIntegerVector(n, "n", nobs = NULL, minimum = 1L)
      if(length(n) != 1L)
        stop("For direct Negative-Binomial simulation, n is the common number r of successes and must be a positive integer scalar.", call.=FALSE)
    }
  ####

    if(method == "CE" && space) {
      if(!isTRUE(grid))
        stop("The spatial CE method works only for a regular two-dimensional grid.", call.=FALSE)
      if(tolower(distance) != "eucl")
        stop("The spatial CE method requires distance = 'Eucl'.", call.=FALSE)
      if(!is.null(anisopars))
        stop("The spatial CE method does not support anisopars; transform the grid explicitly or use method = 'TB'.", call.=FALSE)
      if(is.null(coordy) || !is.null(coordz))
        stop("The spatial CE method requires separate two-dimensional grid axes coordx and coordy.", call.=FALSE)
    }

    if(!(is.null(anisopars))) coords=GeoAniso(coords,c(anisopars$angle,anisopars$ratio))
    ##############################################################################
    ##############################################################################

    if(spacetime)
       {if(!(corrmodel %in% c("Matern_Matern","GenWend_GenWend","GenWend_Matern_GenWend_Matern")))
          stop("GeoSimapprox(method='CE') does not implement this spatio-temporal correlation model.", call.=FALSE)
        }
    if(bivariate) {
         if(!(corrmodel %in% c("Bi_matern","Bi_Matern"))) stop("Not implemented for this correlation model  \n")}
   ################################################################################
    unname(coordt);
    if(is.null(coordx_dyn)){
    unname(coordx);unname(coordy);;unname(coordz)}
  
  ################################################################################
  ################ setting parameters for each model #############################
  ################################################################################
     if(!bivariate) {
       nobs_mean <- dime
       external_mean <- .GeoMean_external(param, n = nobs_mean)
       if(!is.null(external_mean) && !is.null(X))
         stop("A site-specific mean vector and X cannot be used together.\n")
       if(!is.null(X)) {
         X <- .GeoMean_design(X, n = nobs_mean, name = "X")
         num_betas <- ncol(X)
         invisible(.GeoMean_beta(param, num_betas))
       } else {
         supplied_mean <- .GeoMean_supplied_names(param)
         if(is.null(external_mean) && length(setdiff(supplied_mean, "mean")))
           stop("Multiple mean coefficients require an explicit design matrix X.\n")
         num_betas <- 1L
         if(is.null(external_mean)) invisible(.GeoMean_beta(param, 1L))
       }
       param_check <- param
       if(!is.null(external_mean)) param_check$mean <- 0
       if(!all(names(param_check) %in% c(CorrParam(corrmodel),
                                         NuisParam(model,bivariate,num_betas=num_betas))))
         stop("only nuisance and correlation parameters must be included in param\n")
     } else {
       ns_biv <- biv_info$ns
       X <- .GeoBivariateDesign(X, ns_biv, context="GeoSimapprox")
       p_mean_biv <- ncol(X)
       invisible(.GeoBivariateValidateMeanParameters(
         param, p_mean_biv, context="GeoSimapprox"
       ))
       mean_biv <- .GeoBivariateObservedMean(
         param, X, ns_biv, context="GeoSimapprox"
       )
       num_betas <- c(p_mean_biv, p_mean_biv)
       allowed <- c(CorrParam(corrmodel),
                    NuisParam(model, bivariate=TRUE, num_betas=num_betas))
       if(!all(names(param) %in% allowed))
         stop("only nuisance and correlation parameters must be included in param\n")
     }
    
    if(!bivariate) { if(is.null(param$sill))  param$sill=1 }
if(!bivariate) { if(is.null(param$sill)) param$sill=1}
else   { if(is.null(param$sill_1)) param$sill_1=1
         if(is.null(param$sill_2)) param$sill_2=1}


#################################
    if(model %in% c("SkewGaussian","SkewGauss","Beta",'Kumaraswamy','Kumaraswamy2','LogGaussian',#"Binomial","BinomialNeg","BinomialNegZINB",
                    "StudentT","SkewStudentT","Poisson","TwoPieceTukeyh","PoissonZIP","PoissonGamma","PoissonGammaZIP","PoissonWeibull",
                     "TwoPieceBimodal", "TwoPieceStudentT","TwoPieceGaussian","TwoPieceGauss","Tukeyh","Tukeyh2","Tukeygh","SinhAsinh",
                    "Gamma","Weibull","LogLogistic","Logistic","SkewLaplace","BinomialLogistic"))
       {
        if(spacetime_dyn && is.list(X)) X <- do.call(rbind, lapply(X, as.matrix))

  if(!bivariate){
           mm <- .GeoMean_vector(param, X = X, n = nobs_mean, external = external_mean)
           param <- .GeoMean_zero(param, num_betas)
        

        if((model %in% c("SkewGaussian","SkewGauss","TwoPieceGaussian","Logistic","SkewLaplace",
          "TwoPieceGauss","Gamma","Weibull","LogLogistic","Poisson","PoissonZIP","Tukeyh","Tukeyh2","PoissonGamma",
          "PoissonGammaZIP","PoissonWeibull",
          'LogGaussian',"TwoPieceTukeyh","TwoPieceBimodal", "Tukeygh","SinhAsinh",
                    "StudentT","SkewStudentT","TwoPieceStudentT")))  {vv<-param$sill; param$sill=1}

        if(model%in% c("SkewGaussian","SkewGauss","SkewStudentT","TwoPieceTukeyh","TwoPieceBimodal",
               "TwoPieceStudentT","TwoPieceGaussian","TwoPieceGauss"))
               { sk<-param$skew
                 if(model%in% c("TwoPieceTukeyh")) tl<-param$tail
                 if(model%in% c("TwoPieceBimodal")) bimo<-param$shape
               }
        }
        else {
            mm <- mean_biv
            param <- .GeoMean_zero(param, p_mean_biv, "mean_1")
            param <- .GeoMean_zero(param, p_mean_biv, "mean_2")
            vv1<-param$sill_1;param$sill_1=1-param$nugget_1;
            vv2<-param$sill_2;param$sill_2=1-param$nugget_2;vv=c(vv1,vv2)
            sk1<-param$skew_1;sk2<-param$skew_2;sk=c(sk1,sk2)
        }
      }
#################################
  if(model %in% c("Tukeygh","SinhAsinh"))  {
         if(!bivariate){
          param$mean=0
          sk<-param$skew; tl<-param$tail}
         else {
            mm <- mean_biv
            param <- .GeoMean_zero(param, p_mean_biv, "mean_1")
            param <- .GeoMean_zero(param, p_mean_biv, "mean_2")
            vv1<-param$sill_1;param$sill_1=1;vv2<-param$sill_2;param$sill_2=1;vv=c(vv1,vv2)
            sk1<-param$skew_1;sk2<-param$skew_2;sk=c(sk1,sk2)
            tl1<-param$tail_1;tl2<-param$tail_2;tl=c(tl1,tl2)
        }}

    if(model %in% c("Tukeyh"))  {
         if(!bivariate){
          param$mean=0
          tl<-param$tail}
         else {
            mm <- mean_biv
            param <- .GeoMean_zero(param, p_mean_biv, "mean_1")
            param <- .GeoMean_zero(param, p_mean_biv, "mean_2")
            vv1<-param$sill_1;param$sill_1=1;vv2<-param$sill_2;param$sill_2=1;vv=c(vv1,vv2)
            tl1<-param$tail_1;tl2<-param$tail_2;sk=c(tl1,tl2)
        }}

    if(model %in% c("Tukeyh2"))  {
         if(!bivariate){
          param$mean=0
          t1l<-param$tail1
          t2l<-param$tail2
           }
     }
#################################
   if(model %in% c("Wrapped"))  {
        if(!bivariate){
            mm <- 2 * atan(.GeoMean_vector(param, X = X, n = nobs_mean,
                                           external = external_mean)) + pi
            param <- .GeoMean_zero(param, num_betas)}
        else {
            mm <- 2*atan(mean_biv)+pi
            param <- .GeoMean_zero(param, p_mean_biv, "mean_1")
            param <- .GeoMean_zero(param, p_mean_biv, "mean_2")
        }}

###################################################
############# start  do_one_sim ###################
###################################################

do_one_sim <- function(LL, pb = NULL) {
  if (!is.null(pb)) pb(sprintf("LL=%g", LL))
    spec <- .GeoSimulationLatentSpec(model, param, n)
    k <- spec$k
    npoi <- spec$npoi
    param <- spec$param
     #if(model %in% c("Beta")) {  k=round(param$shape1)+round(param$shape2)
  ################################################################################
  ################################################################################
   ns <- if(bivariate) biv_info$ns else NULL
   if(spacetime_dyn) {
        coords=NULL
       if(bivariate) coordt=c(1,2)
       coords=do.call(rbind,args=c(coordx_dyn))
       ns=lengths(coordx_dyn)/2
       if(ncol(coords)==2) {coordx <- coords[,1]; coordy <- coords[,2];coordz <- NULL}
         if(ncol(coords)==3) {coordx <- coords[,1]; coordy <- coords[,2];coordz <- coords[,3]}
       dime=sum(ns)
   }
   else { dime=ddim(coordx,coordy,coordz,coordt,grid=grid)
          if(bivariate) ns <- biv_info$ns
        }
   if(bivariate || model %in% c("BinomialNeg", "BinomialNegZINB")) {
     dd <- NULL
   } else {
     dd <- array(0, dim = c(dime, 1L, k))
   }
 #########################################

#########################################################
KK=1;ssp=double(dime)
if(model%in% c("SkewGaussian","StudentT","SkewStudentT","TwoPieceTukeyh",
               "TwoPieceStudentT","TwoPieceGaussian"))
{
      simD=simu_approx(numxgrid,numygrid,coordx,coordy,coordz,coords,coordt,method,corrmodel,param,M,L,bivariate,spacetime,parallel_inner,n.cores)
      if(!spacetime&&!bivariate) simDD <- c(simD)
      else simDD <- matrix(simD, nrow=numtime, ncol=numcoord,byrow=TRUE)
      param$nugget=0 #ojo
}

if(model %in% c("BinomialNeg","BinomialNegZINB")) {
  rvec <- rep.int(as.integer(n[1L]), dime)
  counts <- integer(dime)
  finished <- logical(dime)
  nb_vals <- integer(dime)
  trial <- 0L
}
if(model %in% c("Poisson", "PoissonZIP", "PoissonGamma", "PoissonGammaZIP", "PoissonWeibull")) {
  poi_counts <- integer(dime)
}

while(KK<=npoi) {
for(i in 1:k) {

################# here the approximated simulation  ##################################################
simd=simu_approx(numxgrid,numygrid,coordx,coordy,coordz,coords,coordt,method,corrmodel,param,M,L,bivariate,spacetime,parallel_inner,n.cores)
######################################################################################################

namesnuis<-NuisParam(model, bivariate,num_betas=num_betas,copula=NULL)
nuisance<-param[namesnuis]
if(i==1&&(model=="SkewGaussian"||model=="SkewGauss")&&bivariate) param["pcol"]=0
####################################
sim<- RFfct1(numcoord,numtime,spacetime,bivariate,
                  dime,nuisance,simd,X,ns)
     
####################################
####### starting cases #############
####################################
if(model %in% c("Binomial", "BinomialNeg","BinomialNegZINB")) {

        simdim <- dim(sim)
        sim <- as.numeric(sim>0)
        dim(sim) <- simdim
         }
####################################
if(model %in% c("Weibull","SkewGaussian","SkewGauss","Binomial","BinomialLogistic","Poisson","PoissonGamma","PoissonWeibull","PoissonZIP","PoissonGammaZIP","Beta","Kumaraswamy","Kumaraswamy2",
              "LogGaussian","TwoPieceTukeyh",
                "Gamma","LogLogistic","Logistic","SkewLaplace","StudentT",
                "SkewStudentT","TwoPieceStudentT","TwoPieceGaussian","TwoPieceGauss","TwoPieceBimodal")) {
       if(!bivariate) dd[,,i]=t(sim)
       if(bivariate)  dd[,,i]=t(sim)
     }
 ####################################
if(model %in% c("BinomialNeg","BinomialNegZINB")){
                 trial <- trial + 1L
                 vec <- c(t(sim))
                 idx <- !finished
                 if(any(idx)) {
                   counts[idx] <- counts[idx] + vec[idx]
                   new_finished <- (counts >= rvec) & !finished
                   if(any(new_finished)) {
                     nb_vals[new_finished] <- trial - rvec[new_finished]
                     finished <- finished | new_finished
                   }
                 }
                 if(all(finished)) break
               }
    }
 ####################################
  if(model %in% c("Poisson","PoissonZIP"))   {
   pois1=0.5*(dd[,,1]^2+dd[,,2]^2)
   ssp=ssp+c(pois1)
   active <- ssp <= c(exp(mm))
   poi_counts <- poi_counts + as.integer(active)
   if(!any(active)) break  ## stopping rule
 }
  ####################################
 if(model %in% c("PoissonGamma","PoissonGammaZIP"))   {
   if(KK==1) sim3 <- .GeoSimulationSquareSums(dd, 3:k)
   #################################
   pois1=0.5*(dd[,,1]^2+dd[,,2]^2)
   ssp=ssp+c(pois1)
   threshold <- c(exp(mm)*sim3/(k-2))
   active <- ssp <= threshold
   poi_counts <- poi_counts + as.integer(active)
   if(!any(active)) break  ## stopping rule
 }
if(model %in% c("PoissonWeibull"))   {
  if(KK==1) sim3 <- .GeoSimulationSquareSums(dd, 3:k)
   #################################
   pois1=0.5*(dd[,,1]^2+dd[,,2]^2)
   ssp=ssp+c(pois1)
   threshold <- c(exp(mm)*(sim3/2)^(1/param$shape)/(gamma(1+1/param$shape)))
   active <- ssp <= threshold
   poi_counts <- poi_counts + as.integer(active)
   if(!any(active)) break  ## stopping rule
 }

 KK=KK+1
}
####### end for #########################
 ###############################################################################################
 #### simulation for discrete random field based on indipendent copies  of GRF ######
 ###############################################################################################
 if(model %in% c("Binomial","BinomialLogistic","Poisson","PoissonGamma","PoissonWeibull","PoissonGammaZIP","PoissonZIP","BinomialNeg","BinomialNegZINB"))   {
   if(model %in% c("Poisson","PoissonGamma","PoissonWeibull"))   {sim=poi_counts;byrow=TRUE}
    if(model %in% c("PoissonZIP"))   {
      a=simu_approx(numxgrid,numygrid,coordx,coordy,coordz,coords,coordt,method,corrmodel,param,M,L,bivariate,spacetime,parallel_inner,n.cores)
     ###
      a[a<as.numeric(param$pmu)]=0;a[a!=0]=1
      sim=a*poi_counts;
      byrow=TRUE
      }
########################################
   if(model == "Binomial") {
                  sim <- .GeoSimulationBinomialCounts(dd, n)
                  byrow <- TRUE }
########################################
if(model == "BinomialLogistic") {
                  sim <- .GeoSimulationBinomialLogisticCounts(dd, n, mm)
                  byrow <- TRUE }
#######################################
   if(model %in% c("BinomialNeg"))   {
          if(any(!finished)) stop("Some locations did not reach n successes during BinomialNeg simulation.", call.=FALSE)
          sim <- nb_vals
          byrow=TRUE
          }
  if(model %in% c("BinomialNegZINB"))   {
          if(any(!finished)) stop("Some locations did not reach n successes during BinomialNegZINB simulation.", call.=FALSE)
          sim <- nb_vals
      a=simu_approx(numxgrid,numygrid,coordx,coordy,coordz,coords,coordt,method,corrmodel,param,M,L,bivariate,spacetime,parallel_inner,n.cores)
     ###
          a[a<as.numeric(param$pmu)]=0;a[a!=0]=1
          sim=a*sim
          byrow=TRUE
          }
#############################################
############### formatting data #############
#############################################

    sim <- .GeoSimulationFormat(
      sim, grid = grid, space = !spacetime && !bivariate,
      numxgrid = numxgrid, numygrid = numygrid, numtime = numtime,
      numcoord = numcoord, byrow = byrow
    )
}
#########################################################################################################
#### Common transformation of latent Gaussian fields for univariate models.
#########################################################################################################
if(!bivariate) {
  transformed <- .GeoSimulationTransformUnivariate(
    model = model, sim = sim, dd = dd, simDD = get0("simDD", ifnotfound = NULL, inherits = TRUE),
    mm = get0("mm", ifnotfound = 0, inherits = TRUE),
    vv = get0("vv", ifnotfound = 1, inherits = TRUE),
    sk = get0("sk", ifnotfound = 0, inherits = TRUE),
    tl = get0("tl", ifnotfound = 0, inherits = TRUE),
    t1l = get0("t1l", ifnotfound = 0, inherits = TRUE),
    t2l = get0("t2l", ifnotfound = 0, inherits = TRUE),
    bimo = get0("bimo", ifnotfound = NULL, inherits = TRUE),
    param = param, k = k, spacetime = spacetime
  )
  if(transformed$handled) {
    sim <- transformed$sim
    if(transformed$needs_format) {
      sim <- .GeoSimulationFormat(
        sim, grid = grid, space = !spacetime, numxgrid = numxgrid, numygrid = numygrid,
        numtime = numtime, numcoord = numcoord, byrow = transformed$byrow
      )
    }
  }
} else {
  ## GeoSimapprox rejects non-Gaussian bivariate marginals above.
  sim <- .GeoSimulationFormat(
    c(sim), grid = grid, space = FALSE, numxgrid = numxgrid, numygrid = numygrid,
    numtime = numtime, numcoord = numcoord, byrow = FALSE
  )
}
##################################################################
###########. formatting data for space time dynamic case. #########
    if(spacetime_dyn) sim <- .GeoSimulationSplitDynamic(sim, ns)
##################################################################
##################################################################
##################################################################

    sim 
  }

##################################################################################
##################### end do_one_sim #############################################
##################################################################################

## Decide ONCE which level of parallelism to use -- outer (across the
## nrep replicates, via future_lapply) or inner (within a single TB
## simulation, across spatial chunks, inside tbm2d_uni). The two are
## never active at the same time to avoid nested-parallelism overhead
## and oversubscription.
##
## For TB, the scheduler uses both N * L (work in one field) and
## N * L * nrep (aggregate work), plus the number of independent
## replicates.  This replaces the old rules based on N * nrep and the
## overly conservative N * L >= 2.5e9 inner threshold.
force_inner <- isTRUE(getOption("GeoModels.force_tb_inner_parallel", FALSE))

if (identical(method, "TB")) {
  tb_parallel_mode <- .GeoTBParallelMode(
    dime = dime,
    L = L,
    nrep = nrep,
    parallel = parallel,
    ncores = n.cores,
    force_inner = force_inner
  )
  use_outer <- identical(tb_parallel_mode, "outer")
  use_inner <- identical(tb_parallel_mode, "inner")
} else {
  ## Preserve the pre-existing scheduling rule for non-TB approximate
  ## simulators (e.g. CE).  The new calibration is specific to TB.
  cost_outer <- as.numeric(dime) * as.numeric(nrep)
  use_outer <- isTRUE(parallel) && nrep > 1 && cost_outer > 1000000
  use_inner <- FALSE
}

parallel_inner <- use_inner   # read by the simu_approx() calls inside do_one_sim

if (use_outer) {

         cat("Performing", nrep, "simulation using",method ," method"," ...\n")
  # Sopprimi i warning
  old_warn <- getOption("warn")
  options(warn = -1)
  on.exit(options(warn = old_warn), add = TRUE)

  old_plan_outer <- future::plan()
  on.exit(try(future::plan(old_plan_outer), silent = TRUE), add = TRUE)
  future::plan(future::multisession, workers = n.cores)

  run_outer <- function(pb = NULL) {
    future.apply::future_lapply(
      seq_len(nrep),
      function(LL) {
        if (!is.null(pb)) pb(sprintf("Simulation %d of %d", LL, nrep))
        do_one_sim(LL)
      },
      future.seed = TRUE
    )
  }

  ## Do not open an empty with_progress() context when progress=FALSE.
  ## This is important when GeoSimapprox() is called inside another
  ## with_progress() block (for example by GeoVarest): an unnecessary
  ## nested context invalidates the outer progressor.
  if (isTRUE(progress)) {
    SIM <- progressr::with_progress({
      pb <- progressr::progressor(steps = nrep)
      run_outer(pb)
    })
  } else {
    SIM <- run_outer(NULL)
  }

} else {

  if (use_inner) {
    ## The plan is opened ONCE here, before looping over replicates, and
    ## reused by every call to tbm2d_uni() (which no longer opens/closes
    ## its own plan). This avoids paying worker start/stop overhead on
    ## every single replicate.
    cat("Performing", nrep, "simulation using",method ,
        "method (parallel over spatial chunks) ...\n")

    ## GeoVarest opens one multisession plan for the complete simulation
    ## phase and reuses it across all K one-replicate calls.  Stand-alone
    ## GeoSimapprox calls still manage and restore their own plan.
    external_plan <- isTRUE(getOption("GeoModels.tb_inner_plan_active", FALSE))
    if (!external_plan) {
      old_plan_inner <- future::plan()
      future::plan(future::multisession, workers = n.cores)
      on.exit(try(future::plan(old_plan_inner), silent = TRUE), add = TRUE)
    }
  }

  ## GeoVarest can provide one deterministic seed per requested replicate.
  ## This internal hook allows several TB replicates to share the expensive
  ## GeoSimapprox setup while preserving exactly the same RNG stream that
  ## would be obtained from separate nrep=1 calls.  It is deliberately an
  ## option rather than a public argument: stand-alone GeoSimapprox semantics
  ## are unchanged.
  rep_seeds_internal <- getOption("GeoModels.GeoSimapprox_rep_seeds", NULL)
  if (!is.null(rep_seeds_internal)) {
    rep_seeds_internal <- suppressWarnings(as.numeric(rep_seeds_internal))
    if (length(rep_seeds_internal) != nrep ||
        any(!is.finite(rep_seeds_internal)) ||
        any(rep_seeds_internal < 0) ||
        any(rep_seeds_internal > .Machine$integer.max)) {
      stop("Internal GeoSimapprox replicate seeds are invalid.", call. = FALSE)
    }
    rep_seeds_internal <- as.integer(rep_seeds_internal)
  }

  with_rep_seed <- function(seed_value, expr) {
    if (is.null(seed_value)) return(force(expr))

    ## Preserve and restore the caller RNG state using base R.
    old_seed <- if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
    } else NULL

    on.exit({
      if (is.null(old_seed)) {
        if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
          rm(".Random.seed", envir = .GlobalEnv)
        }
      } else {
        assign(".Random.seed", old_seed, envir = .GlobalEnv)
      }
    }, add = TRUE)

    set.seed(seed_value)
    force(expr)
  }

  run_sequential_replicates <- function(pb = NULL) {
    results <- vector("list", nrep)
    for (LL in seq_len(nrep)) {
      if (!is.null(pb)) pb(sprintf("Simulation %d of %d", LL, nrep))
      if (is.null(rep_seeds_internal)) {
        results[[LL]] <- do_one_sim(LL)
      } else {
        results[[LL]] <- with_rep_seed(
          rep_seeds_internal[[LL]],
          do_one_sim(LL)
        )
      }
    }
    results
  }

  ## For nrep=1, and whenever progress=FALSE, evaluate directly.  Avoiding
  ## an empty nested with_progress() context keeps an outer GeoVarest
  ## progressor alive throughout its complete simulation phase.
  if (isTRUE(progress) && nrep > 1L) {
    SIM <- progressr::with_progress({
      pb <- progressr::progressor(steps = nrep)
      run_sequential_replicates(pb)
    })
  } else {
    SIM <- run_sequential_replicates(NULL)
  }
}



if (nrep == 1) SIM <- SIM[[1]]

################################################
################################################
################################################



    #######################################
    if(bivariate)   numtime=1
    # Delete the global variables:
    # Return the objects list:
    GeoSim <- list(bivariate = bivariate,
    coordx = coords_orig[,1],
    coordy = coords_orig[,2],
    coordz = if(ncol(coords_orig) >= 3L) coords_orig[,3] else NULL,
    coordt = coordt,
    coordx_dyn =coordx_dyn_output,
    corrmodel = corrmodel,
    data = SIM,
    distance = distance,
    grid = grid_output,
    model = model_requested,
    method=method,
    n=n,
    numcoord = numcoord,
    numtime = numtime,
    param = param_input,
    radius = radius,
    nrep=nrep,
    spacetime = spacetime,
    X=X)
#}
##############################################
    structure(c(GeoSim, call = call), class = c("GeoSim"))
}

GeoNeighbSelect <- function(
  data, coordx, coordy=NULL, coordz=NULL, coordt=NULL, coordx_dyn=NULL,
  copula=NULL, corrmodel=NULL, distance="Eucl", fixed=NULL, anisopars=NULL,
  est.aniso=c(FALSE,FALSE), grid=FALSE, likelihood='Marginal',
  lower=NULL, neighb=c(1,2,3,4,5), p_neighb=1,
  maxtime=Inf, memdist=TRUE, model='Gaussian', n=1, ncores=6,
  optimizer='Nelder-Mead', parallel=FALSE, bivariate=FALSE, radius=1, start=NULL,
  type='Pairwise', upper=NULL, weighted=FALSE,
  X=NULL, spobj=NULL, spdata=NULL, vario=NULL, progress=TRUE, check.duplicates=FALSE
)
{
  if (!isTRUE(memdist)) {
    warning(
      "'memdist = FALSE' is deprecated and ignored; GeoNeighbSelect now always precomputes the selected pair structure.",
      call. = FALSE
    )
  }
  memdist <- TRUE

  if (!is.numeric(neighb) || length(neighb) == 0L)
    stop("neighb must be a non-empty numeric vector")

  if (anyNA(neighb) || any(!is.finite(neighb)))
    stop("neighb must contain only finite, non-missing values")

  if (any(neighb <= 0))
    stop("neighb must contain only positive values")

  tol <- sqrt(.Machine$double.eps)
  if (any(abs(neighb - round(neighb)) > tol))
    stop("neighb must be a positive integer numeric vector")

  neighb <- as.integer(round(neighb))

  if (!is.numeric(p_neighb) || length(p_neighb) != 1L || !is.finite(p_neighb))
    stop("p_neighb must be a finite numeric scalar.")
  if (p_neighb <= 0 || p_neighb > 1)
    stop("p_neighb must be in (0,1].")
  if (p_neighb < 1) {
    warning(
      "With 'p_neighb < 1', neighborhood selection depends on the stochastic thinning realization; use a fixed seed or p_neighb = 1 for a more stable comparison.",
      call. = FALSE
    )
  }

  tau <- 0.15

  corrmodel_valid <- try(CkCorrModel(corrmodel), silent = TRUE)
  if (inherits(corrmodel_valid, "try-error") || is.null(corrmodel_valid))
    stop("correlation model is not valid\n")

  bivariate <- CheckBiv(corrmodel_valid)
  spacetime <- CheckST(corrmodel_valid)
  space <- !spacetime && !bivariate

  ## Validate the observation design once before fitting all candidate
  ## neighborhoods. Subsequent GeoFit calls reuse subsets/identical locations
  ## and must not rescan a potentially massive coordinate set.
  dup_coordx <- coordx; dup_coordy <- coordy; dup_coordz <- coordz
  dup_coordt <- coordt; dup_coordx_dyn <- coordx_dyn
  if (!is.null(spobj)) {
    aa_dup <- sp2Geo(spobj, spdata)
    dup_coordx <- aa_dup$coords
    if (spacetime) dup_coordt <- aa_dup$coordt
  }
  .GeoCheckDuplicateLocations(
    coordx = dup_coordx, coordy = dup_coordy, coordz = dup_coordz,
    coordt = dup_coordt, coordx_dyn = dup_coordx_dyn,
    spacetime = spacetime, bivariate = bivariate, grid = grid,
    context = "GeoNeighbSelect", check.duplicates = check.duplicates
  )
  .dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
  on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)

  if (!is.null(vario)) {
    if (!inherits(vario, "GeoVariogram"))
      stop("A GeoVariogram object is needed as input for vario\n")
    vario_bivariate <- isTRUE(vario$bivariate)
    vario_spacetime <- !is.null(vario$bint)
    ok_type <- (bivariate && vario_bivariate && !vario_spacetime) ||
      (spacetime && vario_spacetime && !vario_bivariate) ||
      (space && !vario_spacetime && !vario_bivariate)
    if (!ok_type)
      stop("The GeoVariogram object is not of the same type of the correlation model\n")
    semiv <- vario
  } else {
    stop("A GeoVariogram object is needed as input for vario\n")
  }

  if (is.null(start) || length(start) == 0L)
    stop("start parameter must be provided\n")
  start_names <- names(start)
  if (is.null(start_names) || any(!nzchar(start_names)) || anyDuplicated(start_names))
    stop("start must be a uniquely named list.\n")

  # Build one empirical layout and use the same ordering for every candidate fit.
  build_variogram_layout <- function() {
    spatial_lags <- as.numeric(semiv$centers)
    if (!length(spatial_lags) || anyNA(spatial_lags) || any(!is.finite(spatial_lags)))
      stop("The GeoVariogram object does not contain valid spatial bin centers.\n")

    if (bivariate) {
      marginal <- semiv$variograms
      cross <- as.numeric(semiv$variogramst)
      if (is.null(dim(marginal)) || nrow(marginal) != 2L ||
          ncol(marginal) != length(spatial_lags) ||
          length(cross) != length(spatial_lags)) {
        stop(
          "The bivariate GeoVariogram object must contain two marginal variograms and one cross-variogram evaluated at the same spatial bin centers.\n"
        )
      }
      empirical <- rbind(
        as.numeric(marginal[1L, ]),
        cross,
        as.numeric(marginal[2L, ])
      )
      return(list(
        kind = "bivariate",
        x = spatial_lags,
        t = NULL,
        empirical = empirical
      ))
    }

    if (spacetime) {
      temporal_lags <- semiv$centert
      if (is.null(temporal_lags) || length(temporal_lags) == 0L)
        temporal_lags <- semiv$bint
      temporal_lags <- as.numeric(temporal_lags)

      nx <- length(spatial_lags)
      nt <- length(temporal_lags)
      spatial_vario <- as.numeric(semiv$variograms)
      temporal_vario <- as.numeric(semiv$variogramt)
      cross_vario <- as.numeric(semiv$variogramst)

      if (!nt || anyNA(temporal_lags) || any(!is.finite(temporal_lags)) ||
          length(spatial_vario) != nx || length(temporal_vario) != nt ||
          length(cross_vario) != nx * nt) {
        stop(
          "The space-time GeoVariogram object must contain compatible spatial and temporal margins and a complete rectangular space-time variogram surface.\n"
        )
      }

      cross_matrix <- matrix(cross_vario, nrow = nx, ncol = nt, byrow = TRUE)
      empirical <- rbind(
        c(0, temporal_vario),
        cbind(spatial_vario, cross_matrix)
      )
      return(list(
        kind = "spacetime",
        x = c(0, spatial_lags),
        t = c(0, temporal_lags),
        empirical = empirical
      ))
    }

    empirical <- as.numeric(semiv$variograms)
    if (length(empirical) != length(spatial_lags))
      stop("The GeoVariogram values and spatial bin centers have incompatible lengths.\n")

    list(
      kind = "spatial",
      x = spatial_lags,
      t = NULL,
      empirical = empirical
    )
  }

  variogram_layout <- build_variogram_layout()

  format_theoretical_variogram <- function(cc) {
    if (variogram_layout$kind == "bivariate") {
      cc <- as.matrix(cc)
      if (!identical(dim(cc), dim(variogram_layout$empirical)))
        return(NULL)
      return(cc)
    }

    if (variogram_layout$kind == "spacetime") {
      nr <- length(variogram_layout$x)
      nc <- length(variogram_layout$t)
      if (length(cc) != nr * nc)
        return(NULL)
      return(matrix(as.numeric(cc), nrow = nr, ncol = nc))
    }

    cc <- as.numeric(cc)
    if (length(cc) != length(variogram_layout$empirical))
      return(NULL)
    cc
  }

  variogram_sse <- function(theoretical) {
    empirical <- variogram_layout$empirical
    if (is.null(theoretical) || !identical(dim(empirical), dim(theoretical))) {
      return(Inf)
    }

    empirical <- as.numeric(empirical)
    theoretical <- as.numeric(theoretical)
    empirical_ok <- is.finite(empirical)
    if (!any(empirical_ok) || any(!is.finite(theoretical[empirical_ok])))
      return(Inf)

    sum((empirical[empirical_ok] - theoretical[empirical_ok])^2)
  }

  align_estimates <- function(param) {
    ans <- rep(NA_real_, length(start_names))
    names(ans) <- start_names
    values <- unlist(param, use.names = TRUE)

    if (!length(values))
      return(ans)

    if (!is.null(names(values)) && any(nzchar(names(values)))) {
      common <- intersect(start_names, names(values))
      ans[common] <- suppressWarnings(as.numeric(values[common]))
    } else if (length(values) == length(ans)) {
      ans[] <- suppressWarnings(as.numeric(values))
    }
    ans
  }

  if (spacetime) {
    if (!is.numeric(maxtime) || length(maxtime) == 0L)
      stop("maxtime must be a non-empty numeric vector")
    candidate_grid <- expand.grid(
      neighb = neighb, maxtime = maxtime,
      KEEP.OUT.ATTRS = FALSE, stringsAsFactors = FALSE
    )
  } else {
    candidate_grid <- data.frame(
      neighb = neighb, maxtime = rep(Inf, length(neighb))
    )
  }

  total_iter <- nrow(candidate_grid)
  estimates <- matrix(
    NA_real_, nrow = total_iter, ncol = length(start_names),
    dimnames = list(NULL, start_names)
  )
  res <- rep(NA_real_, total_iter)

  common_params <- list(
    data = data, coordx = coordx, coordy = coordy, coordz = coordz, coordt = coordt,
    coordx_dyn = coordx_dyn, copula = copula, corrmodel = corrmodel, distance = distance,
    fixed = fixed, anisopars = anisopars, est.aniso = est.aniso, grid = grid,
    likelihood = likelihood, lower = lower, memdist = memdist, model = model, n = n,
    optimizer = optimizer, radius = radius, start = start, type = type, upper = upper,
    weighted = weighted, X = X, spobj = spobj, spdata = spdata,
    p_neighb = p_neighb
  )

  compute_fit <- function(neighb_val, maxtime_val = Inf) {
    params <- common_params
    params$neighb <- neighb_val
    params$maxtime <- maxtime_val

    aa <- try(suppressWarnings(do.call(GeoFit, params)), silent = TRUE)
    if (inherits(aa, "try-error") || is.null(aa)) {
      return(list(estimates = align_estimates(NULL), res = Inf))
    }

    estimates_val <- align_estimates(aa$param)
    valid_loglik <- is.numeric(aa$logCompLik) && length(aa$logCompLik) == 1L &&
      is.finite(aa$logCompLik)
    if (!valid_loglik)
      return(list(estimates = estimates_val, res = Inf))

    corr_params <- c(aa$param, aa$fixed)
    corr_call <- if (is.null(copula)) {
      try(
        suppressWarnings(
          GeoCorrFct(
            variogram_layout$x, t = variogram_layout$t,
            corrmodel = corrmodel, model = model, distance = distance,
            param = corr_params, radius = radius, n = n,
            covariance = TRUE, variogram = TRUE
          )
        ),
        silent = TRUE
      )
    } else {
      try(
        suppressWarnings(
          GeoCorrFct_Cop(
            variogram_layout$x, t = variogram_layout$t,
            corrmodel = corrmodel, model = model, distance = distance,
            copula = copula, param = corr_params, radius = radius, n = n,
            covariance = TRUE, variogram = TRUE
          )
        ),
        silent = TRUE
      )
    }

    if (inherits(corr_call, "try-error") || is.null(corr_call$corr))
      return(list(estimates = estimates_val, res = Inf))

    theoretical <- format_theoretical_variogram(corr_call$corr)
    list(estimates = estimates_val, res = variogram_sse(theoretical))
  }

  task_label <- function(i) {
    if (spacetime) {
      sprintf(
        "neighb=%g, maxtime=%g",
        candidate_grid$neighb[i], candidate_grid$maxtime[i]
      )
    } else {
      sprintf("neighb=%g", candidate_grid$neighb[i])
    }
  }

  run_candidate <- function(i) {
    out <- compute_fit(candidate_grid$neighb[i], candidate_grid$maxtime[i])
    c(out$estimates, res = out$res)
  }

  if (progress) {
    restore_progress_handlers <- .GeoProgressHandlersPush(
      handler = "txtprogressbar", global = TRUE
    )
    on.exit(restore_progress_handlers(), add = TRUE)
  }

  resolved_workers <- .GeoResolveWorkers(parallel, ncores, n_jobs = total_iter)
  if (resolved_workers <= 1L) {
    if (progress) {
      pb <- progressr::progressor(
        along = seq_len(total_iter), on_exit = FALSE, auto_finish = FALSE
      )
      ## The serial branch owns this progressor directly.  Finish it before
      ## restoring the temporary global handler on every exit path.
      on.exit(pb(type = "finish"), add = TRUE, after = FALSE)
    }

    n_params <- length(start_names)
    for (i in seq_len(total_iter)) {
      if (progress)
        pb(task_label(i))
      out <- run_candidate(i)
      estimates[i, ] <- out[seq_len(n_params)]
      res[i] <- unname(out[n_params + 1L])
    }
  } else {
    n.cores <- resolved_workers

    cat("Performing", total_iter, "estimations using", n.cores, "cores...\n")

    old_plan <- future::plan()
    on.exit(future::plan(old_plan), add = TRUE)
    future::plan(future::multisession, workers = n.cores)

    if (progress) {
      results <- progressr::with_progress({
        pb <- progressr::progressor(along = seq_len(total_iter))
        future.apply::future_lapply(
          seq_len(total_iter),
          function(i) {
            out <- run_candidate(i)
            pb(task_label(i))
            out
          },
          future.seed = TRUE, future.stdout = FALSE, future.conditions = "none"
        )
      })
    } else {
      results <- future.apply::future_lapply(
        seq_len(total_iter), run_candidate,
        future.seed = TRUE, future.stdout = FALSE, future.conditions = "none"
      )
    }

    n_params <- length(start_names)
    estimates <- do.call(rbind, lapply(results, function(x) x[seq_len(n_params)]))
    colnames(estimates) <- start_names
    res <- vapply(results, function(x) unname(x[n_params + 1L]), numeric(1))
  }

  finite_ok <- is.finite(res)
  if (!any(finite_ok)) {
    stop(
      "All candidate neighborhood fits failed or produced a non-finite variogram criterion.",
      call. = FALSE
    )
  }

  indexmin <- which.min(res)
  bestK <- candidate_grid$neighb[indexmin]
  best_T <- if (spacetime) candidate_grid$maxtime[indexmin] else NULL
  best_estimates <- estimates[indexmin, ]

  minres <- min(res[finite_ok])
  idx_ok <- which(finite_ok & res <= (1 + tau) * minres)
  sugg_neighb <- min(candidate_grid$neighb[idx_ok])

  if (spacetime) {
    idx_m <- idx_ok[candidate_grid$neighb[idx_ok] == sugg_neighb]
    idx_s <- idx_m[which.min(res[idx_m])]
    sugg_time <- candidate_grid$maxtime[idx_s]
  } else {
    sugg_time <- NULL
  }

  list(
    best_neighb = as.numeric(bestK),
    best_maxtime = best_T,
    best_estimates = best_estimates,
    res = unname(res),
    estimates = estimates,
    sugg_neighb = as.numeric(sugg_neighb),
    sugg_time = sugg_time
  )
}

########################################
## Parametric bootstrap test for absence of spatial dependence
########################################
GeoTestIndependence <- function(data, coordx,
                                start, fixed = NULL,
                                corrmodel = "Matern",
                                model = "Gaussian",
                                optimizer = "bobyqa",
                                optimizer_ind = "Nelder-Mead",
                                lower = NULL, upper = NULL,
                                X = NULL, n = 1,
                                distance = "Eucl", radius = 1,
                                anisopars = NULL,
                                est.aniso = c(FALSE, FALSE),
                                sparse = FALSE,
                                B = 1000,
                                parallel = FALSE,
                                ncores = 6,
                                progress = TRUE,
                                seed = NULL) {

  ## This implementation compares:
  ## H0: independent observations at distinct sites;
  ## H1: a spatially dependent model with nugget < 1.
  ##
  ## H0 is fitted with the marginal independence likelihood, whereas H1 is
  ## fitted with the full likelihood. Parametric bootstrap calibration is used
  ## because nugget = 1 is a boundary point and the correlation parameters are
  ## not identified under H0.

  ## ====== preserve runtime settings and optional RNG seed ======
  cleanup_resources <- .GeoTestPreserveRuntime(gc_on_exit = FALSE)
  on.exit(cleanup_resources(), add = TRUE)
  restore_seed <- .GeoTestSetSeed(seed)
  on.exit(restore_seed(), add = TRUE)

  ## ====== input validation ======
  if (!is.matrix(coordx)) coordx <- as.matrix(coordx)
  if (!is.numeric(data)) data <- as.numeric(data)

  if (!is.numeric(coordx) || any(!is.finite(coordx))) {
    stop("coordx must contain only finite numeric values", call. = FALSE)
  }
  if (any(!is.finite(data))) {
    stop("data must contain only finite numeric values", call. = FALSE)
  }
  if (nrow(coordx) != length(data)) {
    stop(
      "coordx rows (", nrow(coordx), ") must match data length (",
      length(data), ")",
      call. = FALSE
    )
  }
  if (nrow(coordx) < 2L) {
    stop("At least two spatial observations are required", call. = FALSE)
  }

  if (!is.null(X)) {
    X <- as.matrix(X)
    if (nrow(X) != length(data)) {
      stop("nrow(X) must match data length", call. = FALSE)
    }
    if (!is.numeric(X) || any(!is.finite(X))) {
      stop("X must contain only finite numeric values", call. = FALSE)
    }
  }

  if (!is.list(start) || length(start) < 1L ||
      is.null(names(start)) || any(names(start) == "")) {
    stop("start must be a non-empty named list", call. = FALSE)
  }

  if (is.null(fixed)) fixed <- list()
  if (!is.list(fixed)) {
    stop("fixed must be NULL or a named list", call. = FALSE)
  }
  if (length(fixed) > 0L &&
      (is.null(names(fixed)) || any(names(fixed) == ""))) {
    stop("fixed must be NULL or a named list", call. = FALSE)
  }

  if (anyDuplicated(names(start)) || anyDuplicated(names(fixed))) {
    stop("start and fixed cannot contain duplicated parameter names",
         call. = FALSE)
  }

  overlap <- intersect(names(start), names(fixed))
  if (length(overlap) > 0L) {
    stop(
      "Parameters cannot be present in both start and fixed: ",
      paste(overlap, collapse = ", "),
      call. = FALSE
    )
  }

  if (!is.character(corrmodel) || length(corrmodel) != 1L) {
    stop("corrmodel must be a single character string", call. = FALSE)
  }
  if (!is.character(model) || length(model) != 1L) {
    stop("model must be a single character string", call. = FALSE)
  }
  if (!is.character(optimizer) || length(optimizer) != 1L) {
    stop("optimizer must be a single character string", call. = FALSE)
  }
  if (!is.character(optimizer_ind) || length(optimizer_ind) != 1L) {
    stop("optimizer_ind must be a single character string", call. = FALSE)
  }

  if (!is.numeric(B) || length(B) != 1L || !is.finite(B) || B < 1) {
    stop("B must be a positive integer", call. = FALSE)
  }
  B <- as.integer(B)

  if (!is.logical(parallel) || length(parallel) != 1L) {
    stop("parallel must be logical (TRUE or FALSE)", call. = FALSE)
  }
  if (!is.logical(progress) || length(progress) != 1L) {
    stop("progress must be logical (TRUE or FALSE)", call. = FALSE)
  }
  if (!is.logical(sparse) || length(sparse) != 1L) {
    stop("sparse must be logical (TRUE or FALSE)", call. = FALSE)
  }
  if (!is.numeric(radius) || length(radius) != 1L ||
      !is.finite(radius) || radius <= 0) {
    stop("radius must be a positive finite number", call. = FALSE)
  }
  if (!is.logical(est.aniso) || length(est.aniso) != 2L) {
    stop("est.aniso must be a logical vector of length two", call. = FALSE)
  }
  if (!is.null(anisopars) && !is.list(anisopars)) {
    stop("anisopars must be NULL or a list", call. = FALSE)
  }

  if (!is.null(lower) &&
      (!is.list(lower) || is.null(names(lower)) || any(names(lower) == ""))) {
    stop("lower must be NULL or a named list", call. = FALSE)
  }
  if (!is.null(upper) &&
      (!is.list(upper) || is.null(names(upper)) || any(names(upper) == ""))) {
    stop("upper must be NULL or a named list", call. = FALSE)
  }

  ## ====== progress handlers ======
  use_progressr <- .GeoTestConfigureProgress(progress)

  ## ====== core handling ======
  core_cfg <- .GeoTestCoreConfig(parallel, ncores, B, strict_positive = TRUE)
  parallel <- core_cfg$parallel
  ncores <- core_cfg$ncores

  ## ====== minimum distance, used only for exact null simulation ======
  ## Use the same metric and radius as the fitted spatial model. Full
  ## likelihood fitting already requires dense calculations, so constructing
  ## this distance matrix does not change the intended computational regime.
  dmat <- GeoModels::GeoDistances(
    coordx,
    distance = distance,
    radius = radius
  )
  diag(dmat) <- Inf
  d_min <- min(dmat)
  rm(dmat)

  if (!is.finite(d_min) || d_min <= 0) {
    stop(
      "The minimum inter-site distance is not positive; check duplicated coordinates.",
      call. = FALSE
    )
  }

  ## ====== construct H1 parameterization ======
  corr_names <- CorrParam(corrmodel)
  if (!length(corr_names)) {
    stop("Unable to determine the correlation parameters", call. = FALSE)
  }

  start1 <- start
  fixed1 <- fixed

  ## nugget must be estimated under H1. If supplied in fixed, promote it to
  ## start and move it away from the lower boundary when necessary.
  if (!("nugget" %in% names(start1))) {
    if (!("nugget" %in% names(fixed1))) {
      stop(
        "nugget must be supplied in start or fixed; it is estimated under H1.",
        call. = FALSE
      )
    }

    nugget_start <- as.numeric(fixed1$nugget)
    fixed1$nugget <- NULL

    if (!is.finite(nugget_start)) nugget_start <- 0.1
    nugget_start <- min(max(nugget_start, 0.05), 0.95)
    start1$nugget <- nugget_start
  }

  ## Current implementation is for univariate models with one nugget
  ## proportion. More complicated nugget parameterizations require a separate
  ## definition of the independence boundary.
  other_nuggets <- grep("^nugget", c(names(start1), names(fixed1)),
                         value = TRUE)
  other_nuggets <- setdiff(unique(other_nuggets), "nugget")
  if (length(other_nuggets) > 0L) {
    stop(
      "GeoTestIndependence currently supports models with a single parameter named 'nugget'.",
      call. = FALSE
    )
  }

  ## Validate that user-supplied bounds refer only to estimated H1 parameters.
  if (!is.null(lower)) {
    unknown <- setdiff(names(lower), names(start1))
    if (length(unknown) > 0L) {
      stop(
        "lower contains names not estimated under H1: ",
        paste(unknown, collapse = ", "),
        call. = FALSE
      )
    }
  }
  if (!is.null(upper)) {
    unknown <- setdiff(names(upper), names(start1))
    if (length(unknown) > 0L) {
      stop(
        "upper contains names not estimated under H1: ",
        paste(unknown, collapse = ", "),
        call. = FALSE
      )
    }
  }

  ## Bounded optimizers used by GeoFit require finite boxes.  In particular,
  ## bobyqa cannot work with -Inf lower bounds.  Preserve all finite bounds
  ## supplied by the user and replace only non-finite endpoints by broad,
  ## data-adaptive finite values.
  dat_sd <- suppressWarnings(stats::sd(data))
  if (!is.finite(dat_sd) || dat_sd <= 0) dat_sd <- 1
  dat_max <- suppressWarnings(max(abs(data)))
  if (!is.finite(dat_max)) dat_max <- 1
  BIG_M <- max(1e3, 100 * dat_sd, 100 * dat_max)
  POS_EPS <- 1e-8

  complete_bounds <- function(b, parameter_names, default) {
    out <- setNames(rep(default, length(parameter_names)), parameter_names)
    if (!is.null(b) && length(b) > 0L) {
      out[names(b)] <- unlist(b, use.names = FALSE)
    }
    as.list(out)
  }

  make_finite_bounds <- function(lows, ups, starts) {
    positive_names <- intersect(
      c("sill", "scale", "scale_s", "scale_t"),
      names(starts)
    )

    for (nm in names(starts)) {
      lo <- as.numeric(lows[[nm]])
      up <- as.numeric(ups[[nm]])

      if (length(lo) != 1L || is.na(lo)) lo <- -Inf
      if (length(up) != 1L || is.na(up)) up <- Inf

      if (!is.finite(lo)) {
        lo <- if (nm %in% positive_names) POS_EPS else -BIG_M
      }
      if (!is.finite(up)) up <- BIG_M

      if (nm %in% positive_names && lo <= 0) lo <- POS_EPS

      if (lo >= up) {
        stop(
          "Invalid bounds for parameter '", nm,
          "': lower must be strictly smaller than upper",
          call. = FALSE
        )
      }

      lows[[nm]] <- lo
      ups[[nm]] <- up
    }

    list(lower = lows, upper = ups)
  }

  lower1 <- complete_bounds(lower, names(start1), -Inf)
  upper1 <- complete_bounds(upper, names(start1), Inf)

  nugget_eps <- 1e-8
  lower_nugget_user <- if (!is.null(lower1$nugget)) {
    as.numeric(lower1$nugget)
  } else {
    0
  }
  upper_nugget_user <- if (!is.null(upper1$nugget)) {
    as.numeric(upper1$nugget)
  } else {
    Inf
  }

  if (!is.finite(lower_nugget_user)) lower_nugget_user <- 0
  if (lower_nugget_user < 0 || lower_nugget_user >= 1) {
    stop("The lower bound for nugget must lie in [0, 1)", call. = FALSE)
  }

  ## The alternative must approach nugget = 1; otherwise H0 is not in the
  ## closure of H1. GeoModels requires nugget < 1, hence 1 - nugget_eps.
  if (is.finite(upper_nugget_user) &&
      upper_nugget_user < 1 - 1e-6) {
    stop(
      "The upper bound for nugget must be 1 (or omitted) so that H0 is in the closure of H1.",
      call. = FALSE
    )
  }

  lower1$nugget <- lower_nugget_user
  upper1$nugget <- 1 - nugget_eps

  fb1 <- make_finite_bounds(lower1, upper1, start1)
  lower1 <- fb1$lower
  upper1 <- fb1$upper

  clamp_value <- function(x, lo = -Inf, up = Inf) {
    x <- as.numeric(x)
    lo <- as.numeric(lo)
    up <- as.numeric(up)

    if (!is.finite(x)) {
      if (is.finite(lo) && is.finite(up)) {
        x <- (lo + up) / 2
      } else if (is.finite(lo)) {
        x <- lo + max(1, abs(lo))
      } else if (is.finite(up)) {
        x <- up - max(1, abs(up))
      } else {
        x <- 0
      }
    }

    if (is.finite(lo) && x <= lo) {
      x <- lo + max(1e-10, 1e-8 * max(1, abs(lo)))
    }
    if (is.finite(up) && x >= up) {
      x <- up - max(1e-10, 1e-8 * max(1, abs(up)))
    }
    x
  }

  clamp_start <- function(starts, lows, ups) {
    for (nm in names(starts)) {
      lo <- if (!is.null(lows[[nm]])) lows[[nm]] else -Inf
      up <- if (!is.null(ups[[nm]])) ups[[nm]] else Inf
      starts[[nm]] <- clamp_value(starts[[nm]], lo, up)
    }
    starts
  }

  start1 <- clamp_start(start1, lower1, upper1)

  ## ====== construct H0 parameterization ======
  dependence_names <- unique(c(corr_names, "nugget"))

  start0 <- start1[setdiff(names(start1), dependence_names)]
  fixed0 <- fixed1[setdiff(names(fixed1), dependence_names)]

  if (length(start0) < 1L) {
    stop(
      "At least one marginal parameter must be estimated under independence.",
      call. = FALSE
    )
  }

  lower0 <- lower1[names(start0)]
  upper0 <- upper1[names(start0)]

  fixed0_arg <- if (length(fixed0)) fixed0 else NULL
  fixed1_arg <- if (length(fixed1)) fixed1 else NULL

  ## ====== fit helpers ======
  do_fit0 <- function(data_, starts = start0,
                      optimizer0 = optimizer_ind) {
    ## The independence likelihood is a low-dimensional marginal fit.
    ## In particular, the current CompIndLik2 implementation can return a
    ## non-finite objective with bounded optimizers such as bobyqa even when
    ## the same fit is regular with Nelder-Mead.  Therefore H0 has its own
    ## optimizer, with Nelder-Mead as the stable default.
    args0 <- list(
      data = data_,
      coordx = coordx,
      corrmodel = NULL,
      model = model,
      start = starts,
      fixed = fixed0_arg,
      optimizer = optimizer0,
      likelihood = "Marginal",
      type = "Independence",
      X = X,
      n = n,
      distance = distance,
      radius = radius
    )

    bounded0 <- optimizer0 %in% c(
      "L-BFGS-B", "nlminb", "nmkb", "multinlminb",
      "multiNelder-Mead", "bobyqa", "sbplx"
    )
    if (bounded0) {
      args0$lower <- lower0
      args0$upper <- upper0
    }

    do.call(GeoFit, args0)
  }

  fit_independence <- function(data_, starts = start0) {
    optimizers0 <- unique(c(optimizer_ind, "Nelder-Mead"))
    candidates <- list()
    used <- character(0)

    for (opt0 in optimizers0) {
      ff <- suppressWarnings(
        try(do_fit0(data_, starts = starts, optimizer0 = opt0),
            silent = TRUE)
      )
      ll <- loglik_value(ff)
      if (is.finite(ll)) {
        candidates[[length(candidates) + 1L]] <- ff
        used <- c(used, opt0)
      }
    }

    if (!length(candidates)) {
      return(list(fit = NULL, loglik = NA_real_, optimizer = NA_character_))
    }

    vals <- vapply(candidates, loglik_value, numeric(1))
    ii <- which.max(vals)
    list(fit = candidates[[ii]], loglik = vals[ii], optimizer = used[ii])
  }

  do_fit1 <- function(data_, starts = start1) {
    GeoFit(
      data = data_,
      coordx = coordx,
      corrmodel = corrmodel,
      model = model,
      start = starts,
      fixed = fixed1_arg,
      optimizer = optimizer,
      lower = lower1,
      upper = upper1,
      likelihood = "Full",
      type = "Standard",
      X = X,
      n = n,
      distance = distance,
      radius = radius,
      anisopars = anisopars,
      est.aniso = est.aniso,
      sparse = sparse
    )
  }

  loglik_value <- .GeoTestLogLikValue

  update_marginal_start <- function(template, fit0_obj) {
    out <- template
    est <- unlist(fit0_obj$param, use.names = TRUE)
    common <- intersect(names(out), names(est))
    out[common] <- as.list(est[common])
    clamp_start(out, lower1, upper1)
  }

  make_h1_candidates <- function(primary, fit0_obj = NULL,
                                 include_fallbacks = FALSE) {
    candidates <- list(clamp_start(primary, lower1, upper1))

    if (!include_fallbacks) return(candidates)

    base <- primary
    if (!is.null(fit0_obj)) {
      base <- update_marginal_start(base, fit0_obj)
    }

    nugget_grid <- c(0.95, 0.75, 0.50, 0.25)
    nugget_grid <- nugget_grid[
      nugget_grid > as.numeric(lower1$nugget) &
        nugget_grid < as.numeric(upper1$nugget)
    ]

    for (ng in nugget_grid) {
      ss <- base
      ss$nugget <- ng
      candidates[[length(candidates) + 1L]] <-
        clamp_start(ss, lower1, upper1)
    }

    ## Remove exact duplicate starts.
    keys <- vapply(candidates, function(x) {
      paste(names(x), signif(unlist(x, use.names = FALSE), 12),
            sep = "=", collapse = "|")
    }, character(1))
    candidates[!duplicated(keys)]
  }

  fit_spatial <- function(data_, start_primary, fit0_ref, ll0_ref) {
    fit_candidates <- list()

    first <- suppressWarnings(
      try(do_fit1(data_, start_primary), silent = TRUE)
    )
    if (is.finite(loglik_value(first))) {
      fit_candidates[[1L]] <- first
    }

    best_ll <- if (length(fit_candidates)) {
      max(vapply(fit_candidates, loglik_value, numeric(1)))
    } else {
      -Inf
    }

    tol <- 1e-8 * max(1, abs(ll0_ref), abs(best_ll))
    need_fallback <- !is.finite(best_ll) || best_ll < ll0_ref - tol
    fallback_used <- FALSE

    if (need_fallback) {
      fallback_used <- TRUE
      starts <- make_h1_candidates(
        start_primary,
        fit0_obj = fit0_ref,
        include_fallbacks = TRUE
      )

      ## The first start has already been attempted.
      if (length(starts) > 1L) {
        for (ii in 2:length(starts)) {
          ff <- suppressWarnings(
            try(do_fit1(data_, starts[[ii]]), silent = TRUE)
          )
          if (is.finite(loglik_value(ff))) {
            fit_candidates[[length(fit_candidates) + 1L]] <- ff
          }
        }
      }
    }

    if (!length(fit_candidates)) {
      return(list(
        fit = NULL,
        loglik = NA_real_,
        fallback = fallback_used,
        boundary = FALSE,
        valid = FALSE
      ))
    }

    vals <- vapply(fit_candidates, loglik_value, numeric(1))
    idx <- which.max(vals)
    best_fit <- fit_candidates[[idx]]
    best_ll <- vals[idx]
    tol <- 1e-8 * max(1, abs(ll0_ref), abs(best_ll))

    ## Since H0 belongs to the closure of H1, the effective unrestricted
    ## supremum cannot be below ll0. If the best spatial fit does not improve
    ## ll0 after the adaptive retries, the MLE is treated as a boundary
    ## solution and the LR statistic is zero.
    boundary <- best_ll <= ll0_ref + tol

    list(
      fit = best_fit,
      loglik = best_ll,
      fallback = fallback_used,
      boundary = boundary,
      valid = TRUE
    )
  }

  ## ====== observed fits ======
  message("Testing H0: absence of spatial dependence (independence)")

  ind_obs <- fit_independence(data, starts = start0)
  fit0 <- ind_obs$fit
  ll0 <- ind_obs$loglik
  optimizer_ind_used <- ind_obs$optimizer

  if (!is.finite(ll0) || is.null(fit0)) {
    stop(
      "The observed fit under independence failed with both optimizer_ind='",
      optimizer_ind, "' and the Nelder-Mead fallback.",
      call. = FALSE
    )
  }

  spatial_obs <- fit_spatial(
    data_ = data,
    start_primary = start1,
    fit0_ref = fit0,
    ll0_ref = ll0
  )

  if (!isTRUE(spatial_obs$valid) || is.null(spatial_obs$fit)) {
    stop("The observed spatial fit failed", call. = FALSE)
  }

  fit1 <- spatial_obs$fit
  ll1 <- spatial_obs$loglik

  if (isTRUE(spatial_obs$boundary)) {
    Lambda_obs <- 0
  } else {
    Lambda_obs <- 2 * (ll1 - ll0)
  }

  if (!is.finite(Lambda_obs) || Lambda_obs < 0) {
    stop("Invalid observed likelihood-ratio statistic", call. = FALSE)
  }

  par1 <- unlist(c(fit1$param, fit1$fixed), use.names = TRUE)
  nugget_hat <- if ("nugget" %in% names(par1)) {
    as.numeric(par1["nugget"])
  } else {
    NA_real_
  }

  ## A zero LR statistic implies a bootstrap p-value equal to one because all
  ## bootstrap LR statistics are nonnegative.
  if (Lambda_obs == 0) {
    cat("The unrestricted fit is on the independence boundary; bootstrap not needed.\n")

    return(invisible(list(
      lambda_obs = 0,
      pvalue = 1,
      nugget_hat = nugget_hat,
      boundary_observed = TRUE,
      optimizer_independence = optimizer_ind_used,
      optimizer_spatial = optimizer,
      seed = seed,
      B_requested = B,
      bootstrap_successful = 0L,
      bootstrap_failed = 0L,
      bootstrap_success_rate = NA_real_,
      fallback_observed = spatial_obs$fallback,
      fallback_bootstrap = 0L,
      boundary_bootstrap = 0L,
      B_rep = numeric(0),
      fit_H0 = fit0,
      fit_H1 = fit1
    )))
  }

  ## ====== parameters for exact independent null simulation ======
  ## GeoSim does not accept nugget = 1. Exact independence at the observed
  ## sites is therefore generated with an auxiliary compactly supported
  ## GenWend correlation whose support is strictly below d_min. This changes
  ## only the simulation device, not the null marginal model.
  par0_all <- unlist(c(fit0$param, fit0$fixed), use.names = TRUE)
  marginal_names <- unique(c(names(start0), names(fixed0)))
  missing_marginal <- setdiff(marginal_names, names(par0_all))
  if (length(missing_marginal) > 0L) {
    stop(
      "Unable to recover fitted marginal parameters under H0: ",
      paste(missing_marginal, collapse = ", "),
      call. = FALSE
    )
  }

  param_H0_marginal <- as.list(par0_all[marginal_names])
  sim_scale <- 0.5 * d_min
  param_H0_sim <- c(
    param_H0_marginal,
    list(
      nugget = 0,
      scale = sim_scale,
      smooth = 0.5,
      power2 = 4
    )
  )

  message("Running parametric bootstrap under independence (B = ", B, ") ...")

  data_sim <- GeoModels::GeoSim(
    coordx = coordx,
    corrmodel = "GenWend",
    distance = distance,
    model = model,
    n = n,
    param = param_H0_sim,
    radius = radius,
    sparse = TRUE,
    X = X,
    nrep = B,
    progress = FALSE
  )

  ## Use observed estimates as bootstrap starting values.
  start0_boot <- as.list(fit0$param)
  start1_boot <- as.list(fit1$param)
  start1_boot <- clamp_start(start1_boot, lower1, upper1)

  estimate_fun <- function(Zb) {
    ind_b <- fit_independence(Zb, starts = start0_boot)
    f0 <- ind_b$fit
    ll0b <- ind_b$loglik

    if (!is.finite(ll0b) || is.null(f0)) {
      return(c(lambda = NA_real_, fallback = 0, boundary = 0))
    }

    spatial <- fit_spatial(
      data_ = Zb,
      start_primary = start1_boot,
      fit0_ref = f0,
      ll0_ref = ll0b
    )

    if (!isTRUE(spatial$valid) || is.null(spatial$fit)) {
      return(c(
        lambda = NA_real_,
        fallback = as.numeric(spatial$fallback),
        boundary = 0
      ))
    }

    if (isTRUE(spatial$boundary)) {
      lambda <- 0
    } else {
      lambda <- 2 * (spatial$loglik - ll0b)
    }

    if (!is.finite(lambda) || lambda < 0) {
      return(c(
        lambda = NA_real_,
        fallback = as.numeric(spatial$fallback),
        boundary = as.numeric(spatial$boundary)
      ))
    }

    c(
      lambda = lambda,
      fallback = as.numeric(spatial$fallback),
      boundary = as.numeric(spatial$boundary)
    )
  }

  ## ====== sequential or parallel bootstrap ======
  if (!parallel) {
    message(sprintf("Performing %d fits under H0 and H1 sequentially...", B))
    result_list <- .GeoTestSequentialApply(
      data_sim, B, estimate_fun, progress = use_progressr
    )
  } else if (!requireNamespace("future", quietly = TRUE) ||
             !requireNamespace("future.apply", quietly = TRUE)) {
    warning(
      "Parallel bootstrap requested but future/future.apply is unavailable; using sequential evaluation.",
      call. = FALSE
    )
    result_list <- .GeoTestSequentialApply(
      data_sim, B, estimate_fun, progress = FALSE
    )
  } else {
    message(sprintf(
      "Performing %d fits under H0 and H1 using %d cores...",
      B, ncores
    ))
    data_files <- .GeoTestSaveReplicates(data_sim, B, "independence")
    on.exit(unlink(data_files), add = TRUE)
    rm(data_sim)
    gc(verbose = FALSE, full = TRUE)
    pb_boot <- if (use_progressr) {
      progressr::progressor(steps = B)
    } else {
      NULL
    }
    result_list <- .GeoTestParallelFileApply(
      data_files = data_files,
      FUN = estimate_fun,
      ncores = ncores,
      progressor = pb_boot,
      globals = list(coordx = coordx, estimate_fun = estimate_fun),
      progress_every = TRUE,
      error_value = c(lambda = NA_real_, fallback = 0, boundary = 0)
    )
  }

  result_mat <- do.call(rbind, result_list)
  Lambda_all <- result_mat[, "lambda"]
  fallback_all <- result_mat[, "fallback"]
  boundary_all <- result_mat[, "boundary"]

  valid <- is.finite(Lambda_all)
  Lambda_boot <- Lambda_all[valid]

  n_valid <- length(Lambda_boot)
  failed <- B - n_valid
  success_rate <- n_valid / B
  fallback_boot <- sum(fallback_all > 0, na.rm = TRUE)
  boundary_boot <- sum(boundary_all[valid] > 0, na.rm = TRUE)

  if (failed > 0L) {
    warning(sprintf(
      "%d/%d bootstrap replications failed (%.1f%%)",
      failed, B, 100 * failed / B
    ), call. = FALSE)
  }

  min_required <- min(20L, B)
  if (n_valid < min_required) {
    stop(sprintf(
      "Only %d/%d bootstrap replications were successful; at least %d are required.",
      n_valid, B, min_required
    ), call. = FALSE)
  }

  if (success_rate < 0.8) {
    warning(sprintf(
      "Only %.1f%% of bootstrap replications were successful; results may be unreliable.",
      100 * success_rate
    ), call. = FALSE)
  }

  if (B < 99L) {
    warning(
      "B < 99 gives a coarse bootstrap p-value; use a larger B for final analyses.",
      call. = FALSE
    )
  }

  pval <- (1 + sum(Lambda_boot >= Lambda_obs)) / (n_valid + 1)

  res <- list(
    lambda_obs = Lambda_obs,
    pvalue = pval,
    nugget_hat = nugget_hat,
    boundary_observed = spatial_obs$boundary,
    optimizer_independence = optimizer_ind_used,
    optimizer_spatial = optimizer,
    seed = seed,
    B_requested = B,
    bootstrap_successful = n_valid,
    bootstrap_failed = failed,
    bootstrap_success_rate = success_rate,
    fallback_observed = spatial_obs$fallback,
    fallback_bootstrap = fallback_boot,
    boundary_bootstrap = boundary_boot,
    simulation_support = sim_scale,
    B_rep = Lambda_boot,
    fit_H0 = fit0,
    fit_H1 = fit1
  )

  cat("\n=== FINAL RESULTS ===\n")
  cat("Null hypothesis                    = spatial independence\n")
  cat("Observed LR statistic              =", round(Lambda_obs, 5), "\n")
  cat("Optimizer under H0                 =", optimizer_ind_used, "\n")
  cat("Optimizer under H1                 =", optimizer, "\n")
  cat("Estimated nugget under H1          =", round(nugget_hat, 5), "\n")
  cat("Valid bootstrap reps               =", n_valid, "/", B, "\n")
  cat("Boundary bootstrap solutions       =", boundary_boot, "\n")
  cat("Adaptive H1 fallbacks              =", fallback_boot,
      "bootstrap reps\n")
  cat("p-value                            =", round(pval, 5), "\n\n")

  conclusion <- if (pval < 0.05) {
    "Reject H0 => significant spatial dependence detected"
  } else {
    "Do not reject H0 => insufficient evidence of spatial dependence"
  }
  cat("Conclusion:", conclusion, "\n\n")

  invisible(res)
}

####################################################
### File name: GeoDistances.R
####################################################

GeoDistances <- function(coordx = NULL, coordy = NULL, coordz = NULL,
                         distance = c("Eucl", "Chor", "Geod"),
                         radius = 1) {

  distance <- match.arg(distance)

  ##########################################################################
  ## Build and validate the coordinate matrix
  ##########################################################################

  matrix_input <-
    !is.null(coordx) &&
    is.null(coordy) &&
    is.null(coordz) &&
    (is.matrix(coordx) || is.data.frame(coordx))

  if (matrix_input) {

    coords <- as.matrix(coordx)

  } else {

    if (is.matrix(coordx) || is.data.frame(coordx)) {
      stop(
        "When 'coordx' is a matrix or data frame, 'coordy' and 'coordz' ",
        "must be NULL.",
        call. = FALSE
      )
    }

    if (is.null(coordx) || is.null(coordy)) {
      stop(
        "Provide either a coordinate matrix in 'coordx' or vectors ",
        "'coordx' and 'coordy'.",
        call. = FALSE
      )
    }

    if (length(coordx) != length(coordy)) {
      stop(
        "'coordx' and 'coordy' must have the same length.",
        call. = FALSE
      )
    }

    if (!is.null(coordz) && length(coordz) != length(coordx)) {
      stop(
        "'coordz' must have the same length as 'coordx' and 'coordy'.",
        call. = FALSE
      )
    }

    coords <- if (is.null(coordz)) {
      cbind(coordx, coordy)
    } else {
      cbind(coordx, coordy, coordz)
    }
  }

  if (!is.numeric(coords)) {
    stop("Coordinates must be numeric.", call. = FALSE)
  }

  if (length(dim(coords)) != 2L) {
    stop("Coordinates must be a two-dimensional matrix.", call. = FALSE)
  }

  n <- nrow(coords)
  p <- ncol(coords)

  if (n < 1L) {
    stop("At least one coordinate is required.", call. = FALSE)
  }

  if (!(p %in% c(2L, 3L))) {
    stop("Coordinates must have two or three columns.", call. = FALSE)
  }

  if (any(!is.finite(coords))) {
    stop("Coordinates must contain only finite values.", call. = FALSE)
  }

  ##########################################################################
  ## Metric-specific checks
  ##########################################################################

  if (!is.numeric(radius) ||
      length(radius) != 1L ||
      !is.finite(radius) ||
      radius <= 0) {
    stop("'radius' must be a finite positive scalar.", call. = FALSE)
  }

  if (distance %in% c("Geod", "Chor")) {

    if (p != 2L) {
      stop(
        "'Geod' and 'Chor' require exactly two columns: ",
        "longitude and latitude in decimal degrees.",
        call. = FALSE
      )
    }

    latitude <- coords[, 2L]

    if (any(latitude < -90 | latitude > 90)) {
      stop(
        "Latitude values must lie between -90 and 90 degrees.",
        call. = FALSE
      )
    }
  }

  ##########################################################################
  ## Memory and compiled-code index checks
  ##########################################################################

  ## geo_distances() currently uses signed int indices i + j*n in C.
  max_n_c <- floor(sqrt(.Machine$integer.max))

  if (n > max_n_c) {
    stop(
      sprintf(
        paste0(
          "The compiled routine uses 32-bit integer matrix indices and ",
          "cannot safely construct a full distance matrix for n > %d."
        ),
        max_n_c
      ),
      call. = FALSE
    )
  }

  n2 <- as.double(n) * as.double(n)
  output_gb <- 8 * n2 / 1024^3

  ## During the call and matrix construction, more than one copy may exist.
  if (output_gb > 4) {
    warning(
      sprintf(
        paste0(
          "The n x n distance matrix alone requires approximately %.2f GB; ",
          "peak memory usage can be substantially larger."
        ),
        output_gb
      ),
      call. = FALSE
    )
  }

  ##########################################################################
  ## Call the compiled routine
  ##########################################################################

  type_dist <- switch(
    distance,
    Eucl = 0L,
    Chor = 1L,
    Geod = 2L
  )

  out <- numeric(n2)

  res <- dotCall64::.C64(
    "geo_distances",
    SIGNATURE = c(
      "double", "integer", "integer",
      "integer", "double", "double"
    ),
    INTENT = c("r", "r", "r", "r", "r", "w"),
    coords = as.double(coords),
    ncoords = as.integer(n),
    p = as.integer(p),
    type_dist = as.integer(type_dist),
    radius = as.double(radius),
    out = out,
    NAOK = FALSE,
    PACKAGE = "GeoModels",
    VERBOSE = 0
  )$out

  if (length(res) != n2 || any(!is.finite(res))) {
    stop(
      "The compiled distance routine returned an invalid result.",
      call. = FALSE
    )
  }

  ans <- matrix(res, nrow = n, ncol = n, byrow = FALSE)

  ## Preserve observation labels when available.
  rn <- rownames(coords)
  if (!is.null(rn)) {
    dimnames(ans) <- list(rn, rn)
  }

  ans
}

GeoPit <- function(object, type = c("Uniform", "Gaussian"), data_to_pred = NULL)
{
  type <- match.arg(type)

  valid_class <- inherits(object, "GeoFit") ||
    inherits(object, "GeoKrig") ||
    inherits(object, "GeoKrigloc")

  if (!valid_class) {
    stop("A GeoFit, GeoKrig, or GeoKrigloc object is required.")
  }

  if (isTRUE(object$bivariate)) {
    stop("The spatial bivariate case is not implemented in GeoPit().")
  }

  model <- .GeoMarginalCanonicalModel(object$model)
  EST <- inherits(object, "GeoFit")
  PRED <- inherits(object, "GeoKrig") || inherits(object, "GeoKrigloc")

  if (EST && identical(object$residual_type, "Pearson")) {
    stop(
      "GeoPit() is not defined for Pearson-residual objects. Use the original discrete GeoFit object for PIT diagnostics.",
      call. = FALSE
    )
  }

  if (PRED && !(model %in% c("Gaussian", "Gauss"))) {
    stop(
      "Predictive PIT from GeoKrig/GeoKrigloc is currently supported only for Gaussian models, because non-Gaussian linear kriging does not provide a full conditional predictive distribution.",
      call. = FALSE
    )
  }

  pp <- c(unlist(object$param), unlist(object$fixed))

  getpar <- function(name, default = NA_real_) {
    val <- suppressWarnings(as.numeric(pp[name][1L]))
    if (length(val) == 0L || !is.finite(val)) default else val
  }

  require_par <- function(name) {
    val <- getpar(name)
    if (!is.finite(val)) {
      stop("Parameter '", name, "' is missing in the fitted object.")
    }
    val
  }

  scaled_to_unit <- function(x, lower, upper) {
    if (!is.finite(lower) || !is.finite(upper) || upper <= lower) {
      stop("The support parameters must satisfy min < max.")
    }
    (x - lower) / (upper - lower)
  }

  if (EST) {
    if (is.list(object$data)) {
      dd <- unlist(object$data, use.names = FALSE)
    } else {
      dd <- c(t(object$data))
    }
    dd <- as.numeric(dd)

    ## A site-specific external mean vector takes precedence over X beta.
    external_mean <- .GeoMean_external(object$fixed, n = length(dd))

    if (!is.null(external_mean)) {
      allmean <- external_mean
    } else if (!is.null(object$X)) {
      X <- object$X
      if (is.list(X)) {
        X <- as.matrix(do.call(rbind, X))
      } else {
        X <- as.matrix(X)
      }

      if (nrow(X) != length(dd)) {
        stop("The number of rows of X does not match the number of observations.")
      }

      beta <- .GeoMean_beta(pp, ncol(X))
      allmean <- as.numeric(X %*% beta)
    } else {
      allmean <- rep(getpar("mean", 0), length(dd))
    }

    ## Models without a sill parameter use the structural value one.
    allsill <- rep(getpar("sill", 1), length(dd))
  } else {
    target <- data_to_pred
    if (is.null(target) && !is.null(object$data_to_pred)) {
      target <- object$data_to_pred
    }
    if (is.null(target)) {
      stop(
        "For a GeoKrig/GeoKrigloc object, supply the observed prediction targets through data_to_pred.",
        call. = FALSE
      )
    }
    dd <- as.numeric(c(target))
    allmean <- as.numeric(c(object$pred))
    allsill <- as.numeric(c(object$mse))
    if (!length(allsill)) {
      stop("Predictive PIT requires mse=TRUE in GeoKrig/GeoKrigloc.", call. = FALSE)
    }
  }

  dd <- as.numeric(dd)
  allmean <- as.numeric(allmean)
  allsill <- as.numeric(allsill)

  if (length(allmean) == 1L) allmean <- rep(allmean, length(dd))
  if (length(allsill) == 1L) allsill <- rep(allsill, length(dd))

  if (length(allmean) != length(dd)) {
    stop("The length of the mean vector does not match the data length.")
  }
  if (length(allsill) != length(dd)) {
    stop("The length of the variance vector does not match the data length.")
  }
  if (any(!is.finite(allsill)) || any(allsill <= 0)) {
    stop("All variance values used by GeoPit() must be positive and finite.")
  }

  data <- NULL

  #######################################
  ## Gaussian marginal models
  if (model %in% c("Gaussian", "Gaussian_misp_Binomial",
                   "Gaussian_misp_Poisson", "Gaussian_misp_BinomialNeg")) {
    data <- pnorm(dd, mean = allmean, sd = sqrt(allsill))
  }

  #######################################
  ## Positive-support marginal models
  if (model == "Weibull") {
    shape <- require_par("shape")
    if (shape <= 0) stop("The Weibull shape parameter must be positive.")
    data <- pweibull(
      dd,
      shape = shape,
      scale = exp(allmean) / gamma(1 + 1 / shape)
    )
  }

  if (model == "Gamma") {
    shape <- require_par("shape")
    if (shape <= 0) stop("The Gamma shape parameter must be positive.")
    data <- pgamma(
      dd,
      shape = shape / 2,
      rate = shape / (2 * exp(allmean))
    )
  }

  ## GeoModels LogGaussian parametrization:
  ## Y = exp(MM) * exp(sqrt(VV) * Z) / exp(VV / 2).
  if (model == "LogGaussian") {
    data <- suppressWarnings(
      pnorm((log(dd) - allmean + allsill / 2) / sqrt(allsill))
    )
  }

  if (model == "LogLogistic") {
    shape <- require_par("shape")
    if (shape <= 1) {
      stop("The LogLogistic shape parameter must be greater than one.")
    }
    cc <- gamma(1 + 1 / shape) * gamma(1 - 1 / shape)
    scale <- exp(allmean) / cc
    data <- .GeoMarginalPLogLogistic(dd, shape = shape, scale = scale)
  }

  #######################################
  ## Bounded-support marginal models
  if (model == "Beta") {
    shape1 <- require_par("shape1")
    shape2 <- require_par("shape2")
    pmin <- require_par("min")
    pmax <- require_par("max")
    if (shape1 <= 0 || shape2 <= 0) {
      stop("The Beta shape parameters must be positive.")
    }
    zz <- scaled_to_unit(dd, pmin, pmax)
    data <- pbeta(zz, shape1 = shape1, shape2 = shape2)
  }

  if (model == "Beta2") {
    shape <- require_par("shape")
    pmin <- require_par("min")
    pmax <- require_par("max")
    if (shape <= 0) stop("The Beta2 shape parameter must be positive.")
    mu <- plogis(allmean)
    zz <- scaled_to_unit(dd, pmin, pmax)
    data <- pbeta(
      zz,
      shape1 = mu * shape,
      shape2 = (1 - mu) * shape
    )
  }

  if (model == "Kumaraswamy") {
    shape1 <- require_par("shape1")
    shape2 <- require_par("shape2")
    pmin <- require_par("min")
    pmax <- require_par("max")
    if (shape1 <= 0 || shape2 <= 0) {
      stop("The Kumaraswamy shape parameters must be positive.")
    }
    zz <- scaled_to_unit(dd, pmin, pmax)
    data <- numeric(length(zz))
    data[zz >= 1] <- 1
    inside <- which(zz > 0 & zz < 1)
    if (length(inside)) {
      data[inside] <- 1 - (1 - zz[inside]^shape1)^shape2
    }
  }

  if (model == "Kumaraswamy2") {
    shape <- require_par("shape")
    pmin <- require_par("min")
    pmax <- require_par("max")
    if (shape <= 0) {
      stop("The Kumaraswamy2 shape parameter must be positive.")
    }
    mu <- plogis(allmean)
    zz <- scaled_to_unit(dd, pmin, pmax)
    second_shape <- log(0.5) / log1p(-mu^shape)
    if (any(!is.finite(second_shape)) || any(second_shape <= 0)) {
      stop("Kumaraswamy2 produced invalid location-specific shape parameters.")
    }
    data <- numeric(length(zz))
    data[zz >= 1] <- 1
    inside <- which(zz > 0 & zz < 1)
    if (length(inside)) {
      data[inside] <- 1 - (1 - zz[inside]^shape)^second_shape[inside]
    }
  }

  #######################################
  ## Real-valued non-Gaussian marginal models
  if (model == "SkewGaussian") {
    skew <- require_par("skew")
    sill <- getpar("sill", 1)
    omega <- sqrt((skew^2 + allsill) / allsill)
    alpha <- skew / sqrt(sill)
    data <- sn::psn(
      (dd - allmean) / sqrt(allsill),
      xi = 0,
      omega = omega,
      alpha = alpha
    )
  }

  if (model %in% c("StudentT", "Gaussian_misp_StudentT")) {
    inv_df <- require_par("df")
    if (inv_df <= 0) stop("The inverse degrees-of-freedom parameter must be positive.")
    df <- 1 / inv_df
    data <- pt((dd - allmean) / sqrt(allsill), df = df)
  }

  if (model %in% c("SkewStudentT", "Gaussian_misp_SkewStudentT")) {
    inv_df <- require_par("df")
    if (inv_df <= 0) stop("The inverse degrees-of-freedom parameter must be positive.")
    alpha <- require_par("skew")
    nu <- 1 / inv_df
    data <- sn::pst(
      (dd - allmean) / sqrt(allsill),
      xi = 0,
      omega = 1,
      alpha = alpha,
      nu = nu
    )
  }

  if (model == "Logistic") {
    data <- plogis((dd - allmean) / sqrt(allsill))
  }

  if (model == "SkewLaplace") {
    skew <- require_par("skew")
    if (skew <= 0 || skew >= 1) {
      stop("The SkewLaplace skew parameter must lie in (0,1).")
    }
    z <- (dd - allmean) / sqrt(allsill)
    data <- .GeoMarginalPSkewLaplace(z, skew)
  }

  if (model == "SinhAsinh") {
    tail <- require_par("tail")
    skew <- require_par("skew")
    if (tail <= 0) stop("The SinhAsinh tail parameter must be positive.")
    x <- (dd - allmean) / sqrt(allsill)
    data <- pnorm(sinh(tail * asinh(x) - skew))
  }

  if (model == "Tukeyh") {
    tail <- require_par("tail")
    if (tail < 0 || tail >= 0.5) stop("The Tukeyh tail parameter must lie in [0,0.5).")
    x <- (dd - allmean) / sqrt(allsill)
    data <- .GeoMarginalPTukeyH(x, tail)
  }

  if (model == "Tukeyh2") {
    tail1 <- require_par("tail1")
    tail2 <- require_par("tail2")
    if (tail1 < 0 || tail1 >= 0.5 || tail2 < 0 || tail2 >= 0.5)
      stop("The Tukeyh2 tail1 and tail2 parameters must lie in [0,0.5).")
    x <- (dd - allmean) / sqrt(allsill)
    data <- .GeoMarginalPTukeyH2(x, tail1, tail2)
  }

  if (model == "TwoPieceGaussian") {
    skew <- require_par("skew")
    x <- (dd - allmean) / sqrt(allsill)
    data <- .GeoMarginalPTwoPieceGaussian(x, skew)
  }

  if (model == "TwoPieceStudentT") {
    skew <- require_par("skew")
    inv_df <- require_par("df")
    if (inv_df <= 0) stop("The inverse degrees-of-freedom parameter must be positive.")
    df <- 1 / inv_df
    x <- (dd - allmean) / sqrt(allsill)
    data <- .GeoMarginalPTwoPieceStudentT(x, skew, df)
  }

  if (model == "TwoPieceTukeyh") {
    skew <- require_par("skew")
    tail <- require_par("tail")
    x <- (dd - allmean) / sqrt(allsill)
    data <- .GeoMarginalPTwoPieceTukeyH(x, skew, tail)
  }

  if (model == "TwoPieceBimodal") {
    skew <- require_par("skew")
    df <- require_par("df")
    delta <- require_par("shape")
    alpha <- 2 * (delta + 1) / df
    nn <- 2^(1 - alpha / 2)
    x <- (dd - allmean) / sqrt(allsill)
    data <- .GeoMarginalSplitCDF(
      x,
      fun_left = function(z) {
        0.5 * (1 + skew) *
          (1 - pgamma((-z / (1 + skew))^alpha,
                      shape = df, scale = 1 / nn))
      },
      fun_right = function(z) {
        0.5 * (1 + skew) +
          0.5 * (1 - skew) *
          pgamma((z / (1 - skew))^alpha,
                 shape = df, scale = 1 / nn)
      }
    )
  }

  #######################################
  ## Discrete marginal models
  if (model == "Binomial") {
    data <- pbinom(dd, size = object$n, prob = pnorm(allmean))
  }

  if (model == "BinomialNeg") {
    data <- pnbinom(dd, size = object$n, prob = pnorm(allmean))
  }

  if (model == "Poisson") {
    data <- ppois(dd, lambda = exp(allmean))
  }

  if (model == "PoissonGamma") {
    shape <- require_par("shape")
    if (shape <= 0) {
      stop("The PoissonGamma shape parameter must be positive.")
    }
    data <- pnbinom(dd, size = shape, mu = exp(allmean))
  }

  if (is.null(data)) {
    stop("PIT is not implemented for model '", model, "' in GeoPit().")
  }

  ## Numerical protection. The raw Uniform PIT is left on [0, 1], while
  ## the Gaussian transform is clipped only to avoid infinite qnorm values.
  data <- pmin(1, pmax(0, data))

  if (type == "Gaussian") {
    eps <- .Machine$double.eps
    data <- qnorm(pmin(1 - eps, pmax(eps, data)))
  }

  object$data <- data
  object
}

####################################################
### File name: GeoSimCopula.r
####################################################


# Simulate spatial and spatio-temporal random felds:
GeoSimCopula <- function(coordx=NULL, coordy=NULL, coordz=NULL,coordt=NULL, coordx_dyn=NULL,corrmodel, distance="Eucl", grid=FALSE,
     method="cholesky",model='Gaussian', n=1, param,anisopars=NULL, radius=1, sparse=FALSE,
     copula="Gaussian",X=NULL,spobj=NULL,nrep=1,progress=FALSE,check.duplicates=FALSE,
     L=10000,parallel=FALSE,ncores=6)
{
call <- match.call()

    corrmodel <- .GeoSimulationCanonicalCorrModel(corrmodel)
    if(is.null(CkCorrModel(corrmodel))) stop("The name of the correlation model is not correct\n")
    model_requested <- gsub("[[:blank:]]", "",model)
    model <- .GeoSimulationCanonicalModel(model_requested)
    if(is.null(CkModel(model))) stop("The name of the model is not correct\n")
    .GeoValidateCopulaSimulationModel(model)
    n <- .GeoSimulationEffectiveN(model_requested, model, n)
    nrep <- .GeoValidateNrep(nrep)
    distance=gsub("[[:blank:]]", "",distance)
    method <- .GeoValidateCopulaSimulationMethod(method)
    copula=gsub("[[:blank:]]", "",as.character(copula)[1L])
    param <- lapply(param, unname)
    param_input <- param
    .GeoValidateCopulaMarginal(model, param)

##############################################################################
###### extracting sp object informations if necessary              ###########
##############################################################################
bivariate<-CheckBiv(CkCorrModel(corrmodel))
spacetime<-CheckST(CkCorrModel(corrmodel))
space=!spacetime&&!bivariate
if(bivariate) {
  stop("GeoSimCopula currently supports only univariate correlation models.",
       call. = FALSE)
}

.GeoCopula_flatten <- function(z, spacetime=FALSE, grid=FALSE) {
  if(is.list(z)) return(unlist(z, use.names=FALSE))
  if(is.matrix(z) && isTRUE(spacetime) && !isTRUE(grid))
    return(as.numeric(t(z)))
  as.numeric(z)
}
.GeoCopula_restore <- function(z, template, spacetime=FALSE, grid=FALSE) {
  z <- as.numeric(z)
  if(is.list(template)) {
    lens <- lengths(template)
    if(sum(lens) != length(z))
      stop("GeoSimCopula: simulated data length does not match dynamic coordinate blocks.",
           call. = FALSE)
    ends <- cumsum(lens)
    starts <- c(1L, utils::head(ends, -1L) + 1L)
    return(Map(function(a, b) z[a:b], starts, ends))
  }
  if(is.matrix(template) && isTRUE(spacetime) && !isTRUE(grid))
    return(matrix(z, nrow=nrow(template), ncol=ncol(template), byrow=TRUE))
  if(!is.null(dim(template))) return(array(z, dim=dim(template)))
  z
}
if(!is.null(spobj)) {
   if(space||bivariate){
        a=sp2Geo(spobj,NULL); coordx=a$coords 
       if(!a$pj) {if(distance!="Chor") distance="Geod"}
    }
   if(spacetime){
        a=sp2Geo(spobj,NULL); coordx=a$coords ; coordt=a$coordt 
        if(!a$pj) {if(distance!="Chor") distance="Geod"}
     }
}
if(identical(method, "TB")) {
  L <- .GeoValidateTBControls(L)
  .GeoValidateTBDistance(distance, context = "GeoSimCopula")
  if(isTRUE(spacetime)) {
    stop("GeoSimCopula(method='TB') does not support spatio-temporal correlation models; use method = 'cholesky' or 'svd'.",
         call. = FALSE)
  }
  .GeoValidateTBCorrelationModel(corrmodel, bivariate = FALSE)
}
###############################################################
###############################################################

if((copula!="Clayton")&&(copula!="Gaussian")&&(copula!="AMH")&&(copula!="SkewGaussian")) stop("the type of copula is wrong")

## Use exact parameter-name matching. In R, param$nu may partially match
## param$nugget when nu is absent, which can silently corrupt copula checks.
has_param <- function(nm) !is.null(names(param)) && nm %in% names(param)
get_param <- function(nm) {
  if (!has_param(nm)) return(NULL)
  param[[nm, exact = TRUE]]
}

if(copula=="Clayton" && !has_param("nu")) stop("Clayton copula needs a nu parameter")
if(copula=="AMH" && !has_param("nu")) stop("AMH copula needs a nu parameter")
if(copula=="AMH") {
  nu_amh <- as.numeric(get_param("nu"))
  if(length(nu_amh) != 1L || !is.finite(nu_amh))
    stop("AMH nu must be a finite scalar.", call.=FALSE)
}
if(copula=="SkewGaussian" && !has_param("nu")) stop("Skew Gaussian copula needs a nu parameter")
if(!has_param("nugget")) stop("GeoSimCopula requires an explicit param$nugget.", call.=FALSE)

nugget_cop <- as.numeric(get_param("nugget"))
if(length(nugget_cop) != 1L || !is.finite(nugget_cop) ||
   nugget_cop < 0 || nugget_cop >= 1)
  stop("GeoSimCopula requires param$nugget to be a finite scalar in [0,1).", call.=FALSE)

#### corr parameters
paramcorr=param[CorrParam(corrmodel)]

.GeoCheckDuplicateLocations(
  coordx = coordx, coordy = coordy, coordz = coordz, coordt = coordt,
  coordx_dyn = coordx_dyn, spacetime = spacetime, bivariate = FALSE,
  grid = grid, context = "GeoSimCopula", check.duplicates = check.duplicates
)
.dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)

## Keep copula construction separate from the numerical latent-field engine.
## Exact methods use GeoSim(), while turning bands uses GeoSimapprox().
.GeoCopula_latent_sim <- function(latent_model, latent_param, n_latent = 1L) {
  common_args <- list(
    coordx = coordx, coordy = coordy, coordz = coordz, coordt = coordt,
    coordx_dyn = coordx_dyn, corrmodel = corrmodel, distance = distance,
    grid = grid, model = latent_model, n = n_latent, param = latent_param,
    anisopars = anisopars, radius = radius, nrep = 1L,
    progress = FALSE, check.duplicates = FALSE
  )

  if(identical(method, "TB")) {
    return(do.call(
      GeoSimapprox,
      c(common_args, list(method = "TB", L = L,
                          parallel = parallel, ncores = ncores))
    ))
  }

  do.call(
    GeoSim,
    c(common_args, list(method = method, sparse = sparse))
  )
}

SIM=list()
###################################################
### starting number of replicates
###################################################
for(rep_idx in seq_len(nrep)){



####Gaussian copula #############################################
if(copula=="Gaussian")
{
param1=c(list(mean=0,sill=1,nugget=get_param("nugget")),paramcorr)

sim <- .GeoCopula_latent_sim("Gaussian", param1, n_latent = 1L)
latent_template <- sim$data
latent <- .GeoCopula_flatten(latent_template, sim$spacetime, sim$grid)
unif <- pnorm(latent, mean=0, sd=1)

}

####skewGaussian copula #############################################
if(copula=="SkewGaussian")
{

nu <- as.numeric(get_param("nu"))
if(length(nu) != 1L || !is.finite(nu) || abs(nu) >= 1)
  stop("nu parameter must be finite and between -1 and 1")
alpha <- nu / sqrt(1 - nu^2)

## The skew-Gaussian copula is generated from
##   Z(s) = alpha * abs(G1(s)) + G2(s),
## where G1 and G2 are independent Gaussian RFs sharing exactly the
## same correlation function.  Simulating the two latent fields
## explicitly also applies the nugget to both of them, consistently
## with the copula likelihood and its latent-field definition.
param_latent <- c(list(mean=0, sill=1, nugget=get_param("nugget")), paramcorr)
sim_g1 <- .GeoCopula_latent_sim("Gaussian", param_latent, n_latent = 1L)
sim_g2 <- .GeoCopula_latent_sim("Gaussian", param_latent, n_latent = 1L)
latent_template <- sim_g2$data
g1 <- .GeoCopula_flatten(sim_g1$data, sim_g1$spacetime, sim_g1$grid)
g2 <- .GeoCopula_flatten(sim_g2$data, sim_g2$spacetime, sim_g2$grid)
latent <- alpha * abs(g1) + g2
omega <- sqrt(alpha^2 + 1)
unif <- sn::psn(latent, xi=0, omega=omega, alpha=alpha)
sim <- sim_g2
}


#if (copula == "SkewGaussian") {
#  delta <- param$nu  
#  print("h")
#  print(delta)
#  eps   <- 1e-6;
#  alpha <- delta / sqrt(1 - delta^2)  
#  print(alpha) 
#  skew_eta <- alpha
#  param1 <- c(list(mean   = 0,sill   = 1,nugget = param$nugget,skew   = skew_eta),paramcorr)
#  sim <- GeoSim(
#    coordx = coordx, coordy = coordy,coordz = coordz, coordt = coordt,
#    coordx_dyn = coordx_dyn,corrmodel = corrmodel,distance  = distance, grid = grid,
#    method = method, model = "SkewGaussian",
#    n = 1, param = param1,anisopars = anisopars, radius = radius,sparse = sparse, nrep = 1
#  )
#  omega <- sqrt(1 + alpha^2)
#  unif  <- sn::psn(sim$data, xi = 0, omega = omega, alpha = alpha)
#}

####beta copula #############################################
if(copula=="Clayton")
{
nu_clayton <- as.numeric(get_param("nu"))
if(length(nu_clayton) != 1L || !is.finite(nu_clayton) || nu_clayton <= 0 ||
   abs(nu_clayton - round(nu_clayton)) > sqrt(.Machine$double.eps)) {
  stop("GeoSimCopula: the constructive Clayton simulator requires nu to be a positive integer; nu is not rounded automatically.",
       call. = FALSE)
}
pp <- as.integer(round(nu_clayton))
param1=c(list(shape1=pp,shape2=2,mean=0,min=0,max=1,nugget=get_param("nugget")),paramcorr)
sim <- .GeoCopula_latent_sim("Beta", param1, n_latent = 1L)
latent_template <- sim$data
latent <- .GeoCopula_flatten(latent_template, sim$spacetime, sim$grid)
unif <- latent^(pp/2)
}
####AMH copula #############################################
if(copula=="AMH")
{

param1=c(list(mean=0,nugget=as.numeric(get_param("nugget")),shape=2),paramcorr)
sim <- .GeoCopula_latent_sim("Gamma", param1, n_latent = 1L)
param2=c(list(mean=as.numeric(get_param("nu")),nugget=as.numeric(get_param("nugget"))),paramcorr)
sim2 <- .GeoCopula_latent_sim("BinomialNeg", param2, n_latent = 1L)
latent_template <- sim$data
latent1 <- .GeoCopula_flatten(sim$data, sim$spacetime, sim$grid)
latent2 <- .GeoCopula_flatten(sim2$data, sim2$spacetime, sim2$grid)
if(length(latent1) != length(latent2))
  stop("GeoSimCopula: incompatible latent simulations in the AMH construction.", call.=FALSE)
kk <- latent1/(latent2+1)
pp <- pnorm(as.numeric(get_param("nu")))
unif <- pp/(exp(kk)+pp-1)
}
####################################################################

## Keep inverse-CDF evaluations away from exact 0 and 1.
eps_u <- .Machine$double.eps
unif <- pmin(pmax(as.numeric(unif), eps_u), 1 - eps_u)

if(!sim$bivariate){
           ## length(data) is valid for fixed and dynamic space-time layouts.
           nobs_mean <- length(unif)
           external_mean <- .GeoMean_external(param, n = nobs_mean)
           if(!is.null(external_mean) && !is.null(X))
             stop("A site-specific mean vector and X cannot be used together.\n")
           if(!is.null(X)) {
             if(is.list(X)) X <- do.call(rbind, lapply(X, as.matrix))
             X <- .GeoMean_design(X, n = nobs_mean, name = "X")
             invisible(.GeoMean_beta(param, ncol(X)))
           } else if(is.null(external_mean) &&
                     length(setdiff(.GeoMean_supplied_names(param), "mean"))) {
             stop("Multiple mean coefficients require an explicit design matrix X.\n")
           }
           mm <- .GeoMean_vector(param, X = X, n = nobs_mean,
                                 external = external_mean)

           # Keep param unchanged across replications.  In particular, do not
           # reset mean, mean1, ... inside the nrep loop.
    }       
##############################
##############################
#######     models     #######
##############################
simcop <- NULL

####################################
############ discrete  RF ##########
####################################
.GeoCopula_size <- function(size, nobs, model_name) {
  size <- as.numeric(size)
  if(length(size) == 1L) size <- rep(size, nobs)
  if(length(size) != nobs || any(!is.finite(size)) ||
     any(size <= 0) || any(size != round(size))) {
    stop(sprintf("For %s, n must contain positive integers and have length 1 or the number of observations.", model_name),
         call. = FALSE)
  }
  size
}
if(model=="Binomial") {
 nn <- .GeoCopula_size(n, length(unif), model_requested)
 simcop=qbinom(unif, size=nn, prob=pnorm(mm))
}
if(model=="BinomialNeg") {
 nn <- .GeoCopula_size(n, length(unif), model_requested)
 simcop=qnbinom(unif, size=nn, prob=pnorm(mm))
}
if(model=="BinomialNegZINB") {
 nn <- .GeoCopula_size(n, length(unif), model_requested)
 simcop=qzinegbin(unif, size=nn, #prob = pnorm(mm),                # require package VGAM
        munb=   nn*(1-pnorm(mm))/pnorm(mm),#,n/pnorm(mm)-n,
        pstr0 = as.numeric(param$pmu))
}
if(model=="Poisson") {
 simcop=qpois(unif, lambda=exp(mm))
}
if(model=="PoissonZIP") {
 simcop=qzipois(unif, lambda=exp(mm), pstr0 = as.numeric(param$pmu) ) # require package VGAM
}
####################################
############ positive real  RF #####
####################################
if(model=="Gamma") 
         {
p2=as.numeric(param$shape)
if(length(p2) != 1L || !is.finite(p2) || p2 <= 0)
  stop("Gamma shape must be a finite positive scalar.", call.=FALSE)
simcop=exp(mm)*qgamma(unif, shape=p2/2, scale=2/p2)
         }
############
if(model=="Weibull") 
         {
p2=as.numeric(param$shape)
if(length(p2) != 1L || !is.finite(p2) || p2 <= 0)
  stop("Weibull shape must be a finite positive scalar.", call.=FALSE)
simcop= exp(mm)*qweibull(unif,shape=p2,scale=1/(gamma(1+1/p2 )))

         }
########################
if(model=="LogLogistic") 
         {

p2=as.numeric(param$shape)
if(length(p2) != 1L || !is.finite(p2) || p2 <= 1)
  stop("LogLogistic shape must be finite and greater than 1 under the mean parametrization used by GeoSimCopula.", call.=FALSE)
cc=gamma(1+1/p2)*gamma(1-1/p2)
simcop=.GeoMarginalQLogLogistic(unif,shape = p2,scale=exp(mm)/cc)            
         }
##############################

if(model=="LogGaussian")
{ 
    vv=as.numeric(param$sill)
   if(length(vv) != 1L || !is.finite(vv) || vv <= 0)
     stop("LogGaussian sill must be a finite positive scalar.", call.=FALSE)
   simcop = qlnorm(unif, mm-vv/2, sqrt(vv))
     #simcop = qlnorm(unif, 1+mm, sqrt(vv))
}
##############################
##############################
############ Real RF #########
##############################


if(model=="Gaussian") {
         vv <- as.numeric(param$sill)
         if(length(vv) != 1L || !is.finite(vv) || vv <= 0)
           stop("Gaussian sill must be a finite positive scalar.", call.=FALSE)
         simcop=qnorm(unif,mean=mm,sd=sqrt(vv))
         }
if(model=="Logistic") 
         {
         vv <- as.numeric(param$sill)
         if(length(vv) != 1L || !is.finite(vv) || vv <= 0)
           stop("Logistic sill must be a finite positive scalar.", call.=FALSE)
         simcop=qlogis(unif,location=mm,scale=sqrt(vv))
         }
if(model=="StudentT") {
vv=as.numeric(param$sill)
dd=as.numeric(param$df)
if(length(vv) != 1L || !is.finite(vv) || vv <= 0)
  stop("StudentT sill must be a finite positive scalar.", call.=FALSE)
if(length(dd) != 1L || !is.finite(dd) || dd <= 0)
  stop("StudentT df is the reciprocal degrees-of-freedom parameter and must be finite and positive.", call.=FALSE)
simcop=mm+sqrt(vv)*qt(unif,df=1/dd)

         }
############
if(model=="SkewGaussian") 
         {
vv=as.numeric(param$sill)  
sk=as.numeric(param$skew)  
omega=as.numeric(sqrt((sk^2 + vv)/vv)) 
alpha=as.numeric(sk/(vv^0.5))         
simcop=mm+sqrt(vv)*sn::qsn(unif,xi=0,omega= as.numeric(omega),alpha= as.numeric(alpha))
         }
#######################################  
if(model=="SkewStudentT")
{
  vv=as.numeric(param$sill) 
  delta <- as.numeric(param$skew)
  if(length(delta) != 1L || !is.finite(delta) || abs(delta) >= 1)
    stop("For SkewStudentT, skew must lie strictly between -1 and 1.", call.=FALSE)
  alpha <- delta/sqrt(1-delta^2)
  eps_df <- as.numeric(param$df)
  if(length(eps_df) != 1L || !is.finite(eps_df) || eps_df <= 0)
    stop("SkewStudentT df is the reciprocal degrees-of-freedom parameter and must be finite and positive.", call.=FALSE)
  dd <- 1 / eps_df
  simcop=mm+sqrt(vv)*sn::qst(unif, xi=0, omega=1, alpha=alpha, nu=dd)
}
#######################################
if(model=="SinhAsinh")
{
vv=as.numeric(param$sill) 
tail = as.numeric(param$tail) 
skew = as.numeric(param$skew)
simcop=mm+sqrt(vv)*sinh(1/tail * asinh(qnorm(unif))+skew/tail)
}
#######################################   OK
if(model=="Tukeyh")
{
vv=as.numeric(param$sill) 
tail = as.numeric(param$tail) 
uu=qnorm(unif)
simcop=mm+sqrt(vv)*uu*exp(0.5*tail*uu^2);
}
#######################################  OK
if(model=="Tukeyh2")
{
tail1 =as.numeric(param$tail1);tail2 = as.numeric(param$tail2)
vv=as.numeric(param$sill) 
simcop =mm+sqrt(vv)*.GeoMarginalQTukeyH2(unif, tail1, tail2)
}

#######################################  OK
if(model=="SkewLaplace")
{  

# Quantile della Skew/Asymmetric Laplace
sk=as.numeric(param$skew) 
vv=as.numeric(param$sill)
if(length(vv) != 1L || !is.finite(vv) || vv <= 0)
  stop("SkewLaplace sill must be a finite positive scalar.", call.=FALSE)
if(length(sk) != 1L || !is.finite(sk) || sk <= 0 || sk >= 1)
  stop("SkewLaplace skew must lie strictly between 0 and 1.", call.=FALSE)
simcop =mm+sqrt(vv)*.GeoMarginalQSkewLaplace(unif, sk)
}
#######################################  OK
if(model=="Tukeygh")
{
tail = as.numeric(param$tail);skew =  as.numeric(param$skew) 
vv=as.numeric(param$sill) 
uu=qnorm(unif)
if(abs(skew) <= sqrt(.Machine$double.eps)) {
  core <- uu*exp(0.5*tail*uu^2)
} else {
  core <- expm1(skew*uu)*exp(0.5*tail*uu^2)/skew
}
simcop=mm+sqrt(vv)*core
}
#######################################   OK
if(model=="TwoPieceGaussian")
{
skew = as.numeric(param$skew) 
vv=as.numeric(param$sill) 
simcop =mm+sqrt(vv)*.GeoMarginalQTwoPieceGaussian(unif, skew)
}

#######################################  
if(model=="TwoPieceStudentT")
{
vv=as.numeric(param$sill) 
skew = as.numeric(param$skew)
eps_df <- as.numeric(param$df)
if(length(eps_df) != 1L || !is.finite(eps_df) || eps_df <= 0)
  stop("TwoPieceStudentT df is the reciprocal degrees-of-freedom parameter and must be finite and positive.", call.=FALSE)
df   = 1/eps_df
simcop=mm+sqrt(vv)*.GeoMarginalQTwoPieceStudentT(unif, skew, df)
}
#######################################   OK
if(model=="TwoPieceTukeyh")
{
vv=as.numeric(param$sill) 
skew = as.numeric(param$skew)
tail= as.numeric(param$tail)
simcop =mm+sqrt(vv)*.GeoMarginalQTwoPieceTukeyH(unif, skew, tail)
}




####################################
############ bounded  real RF    ###
####################################
if(model=="Kumaraswamy") 
{
p1=as.numeric(param$shape1);p2=as.numeric(param$shape2)
pmin=as.numeric(param$min);pmax=as.numeric(param$max);
if(any(!is.finite(c(p1,p2))) || any(c(p1,p2) <= 0)) stop("Kumaraswamy shape parameters must be finite and positive.", call.=FALSE)
if(!is.finite(pmin) || !is.finite(pmax) || pmin >= pmax) stop("Kumaraswamy requires finite min < max.", call.=FALSE)
simcop=pmin + (pmax-pmin)*(1-(1-unif)^(1/p2))^(1/p1)
}
############
if(model=="Kumaraswamy2") 
{ # parametrization using beta median  regression
mm=1/(1+exp(-mm))
p2=as.numeric(param$shape)
pmin=as.numeric(param$min);pmax=as.numeric(param$max);
if(!is.finite(p2) || p2 <= 0) stop("Kumaraswamy2 shape must be finite and positive.", call.=FALSE)
if(!is.finite(pmin) || !is.finite(pmax) || pmin >= pmax) stop("Kumaraswamy2 requires finite min < max.", call.=FALSE)
shape2 <- log(0.5)/log1p(-mm^p2)
simcop=pmin + (pmax-pmin)*(1-(1-unif)^(1/shape2))^(1/p2)
}
############
if(model=="Beta") 
{ # parametrization using beta regression
pmin=as.numeric(param$min);pmax=as.numeric(param$max);
sh1=as.numeric(param$shape1);sh2=as.numeric(param$shape2)
if(any(!is.finite(c(sh1,sh2))) || any(c(sh1,sh2) <= 0)) stop("Beta shape parameters must be finite and positive.", call.=FALSE)
if(!is.finite(pmin) || !is.finite(pmax) || pmin >= pmax) stop("Beta requires finite min < max.", call.=FALSE)
simcop=pmin + (pmax-pmin)*qbeta(unif,shape1=sh1,shape2=sh2)
}
############
if(model=="Beta2") 
{ # parametrization using beta mean  regression

mm=1/(1+exp(-mm))
p2=as.numeric(param$shape)
pmin=as.numeric(param$min);pmax=as.numeric(param$max);
if(!is.finite(p2) || p2 <= 0) stop("Beta2 shape must be finite and positive.", call.=FALSE)
if(!is.finite(pmin) || !is.finite(pmax) || pmin >= pmax) stop("Beta2 requires finite min < max.", call.=FALSE)
simcop=pmin + (pmax-pmin)*qbeta(unif,shape1=mm*p2,shape2=(1-mm)*p2)
}





############
if(is.null(simcop))
  stop(sprintf("GeoSimCopula does not implement marginal model '%s'.", model_requested),
       call. = FALSE)
simcop <- .GeoCopula_restore(simcop, latent_template, sim$spacetime, sim$grid)
##############################




SIM[[rep_idx]]=simcop
}
#### end numrep ##############
##############################
##############################

 if(nrep==1) SIM=SIM[[1]] 

    GeoSim_Copula <- list(bivariate = sim$bivariate,
    coordx = sim$coordx,
    coordy = sim$coordy,
    coordt = sim$coordt,
    coordz=sim$coordz,
    coordx_dyn =sim$coordx_dyn,
    corrmodel = corrmodel,
    data = SIM,
    distance = sim$distance,
    grid = sim$grid,
    model = model_requested,
    method=method,
    n=n,
    numcoord = sim$numcoord,
    numtime = sim$numtime,
    param = param_input,
    radius = radius,
    spacetime = sim$spacetime,
    sparse=if(is.null(sim$sparse)) FALSE else sim$sparse,
    copula=copula,
    nrep=nrep,
    X=X)
#}
##############################################
  structure(c(GeoSim_Copula, call = call), class = c("GeoSim_Copula"))
}

###########################################################################
##  Internal covariance helpers used by GeoKriglocWeights               ##
###########################################################################

###########################################################################
##  Shared dense spatial covariance helpers                              ##
###########################################################################

.GeoKriglocPrepareSpatialCovariance <- function(neigh, coordx_global,
                                                 X_global, param, corrmodel,
                                                 distance, n, model,
                                                 radius, copula, M = NULL) {
  if (is.null(neigh$indices)) return(NULL)
  nonempty <- which(vapply(neigh$indices, length, integer(1L)) >= 2L)
  if (!length(nonempty)) return(NULL)

  k0 <- nonempty[1L]
  idx0 <- as.integer(neigh$indices[[k0]])
  take <- idx0[seq_len(min(2L, length(idx0)))]
  coords0 <- as.matrix(coordx_global[take, , drop = FALSE])
  X0 <- if (!is.null(X_global)) {
    as.matrix(X_global[take, , drop = FALSE])
  } else NULL

  param0 <- param
  ## External observation means are location-specific and are supplied to the
  ## direct covariance assembler at run time.  The one-time template only
  ## needs a valid scalar mean so GeoCovmatrix can normalize metadata exactly
  ## as in the historical path.
  if (!is.null(M)) param0$mean <- 0

  n0 <- n
  if (length(n) > 1L && length(take) && length(n) >= max(take)) {
    n0 <- n[take]
  }

  template <- GeoCovmatrix(
    coordx = coords0, corrmodel = corrmodel, distance = distance,
    model = model, n = n0, param = param0, anisopars = NULL,
    radius = radius, sparse = FALSE, copula = copula, X = X0
  )
  if (!is.null(template$copula)) {
    cp <- as.character(template$copula)[1L]
    if (is.na(cp) || !nzchar(gsub("[[:blank:]]", "", cp))) {
      template$copula <- NULL
    }
  }
  template$param <- unlist(template$param)
  template$local_cov_radius <- if (CheckSph(CkCorrModel(corrmodel))) 1 else radius
  template
}

.GeoKriglocLocalParamVector <- function(template, param) {
  out <- template$param
  z <- unlist(param)
  if (length(z) && !is.null(names(z))) {
    keep <- intersect(names(z), names(out))
    if (length(keep)) out[keep] <- z[keep]
  }
  out
}

.GeoKriglocSpatialPackedCorrelation <- function(coords, corrmodel_code,
                                                    corrparam, distance_code,
                                                    radius,
                                                    context = "GeoKrigloc") {
  coords <- as.matrix(coords)
  if (!is.numeric(coords) || nrow(coords) < 1L ||
      !(ncol(coords) %in% c(2L, 3L)) || any(!is.finite(coords))) {
    stop(context, ": invalid coordinates for packed correlation assembly.",
         call. = FALSE)
  }
  if (ncol(coords) == 2L) coords <- cbind(coords, 0)
  storage.mode(coords) <- "double"
  nobs <- nrow(coords)
  npairs <- as.double(nobs) * as.double(nobs - 1L) / 2
  if (npairs <= 0) return(numeric(0L))

  ans <- dotCall64::.C64(
    "Corr_spatial_packed",
    SIGNATURE = c("double", "double", "double", "double",
                  "integer", "integer", "double", "integer", "double"),
    rho = dotCall64::vector_dc("double", npairs),
    coords[, 1L], coords[, 2L], coords[, 3L],
    as.integer(corrmodel_code), as.integer(nobs), as.double(corrparam),
    as.integer(distance_code), as.double(radius),
    INTENT = c("w", rep("r", 8)),
    NAOK = TRUE, PACKAGE = "GeoModels"
  )$rho

  if (length(ans) != npairs || any(!is.finite(ans))) {
    stop(context, ": non-finite packed spatial correlations.", call. = FALSE)
  }
  as.numeric(ans)
}

.GeoKriglocMultiplicativeStdVariance <- function(model_code, nuisance,
                                                       context = "GeoKrigloc") {
  code <- as.integer(model_code)[1L]
  if (code == 21L) {
    sh <- as.numeric(nuisance["shape"])[1L]
    out <- 2 / sh
  } else if (code == 26L) {
    sh <- as.numeric(nuisance["shape"])[1L]
    out <- gamma(1 + 2 / sh) / gamma(1 + 1 / sh)^2 - 1
  } else if (code == 24L) {
    sh <- as.numeric(nuisance["shape"])[1L]
    out <- 2 * sh * sin(pi / sh)^2 / (pi * sin(2 * pi / sh)) - 1
  } else if (code == 22L) {
    ss <- as.numeric(nuisance["sill"])[1L]
    out <- exp(ss) - 1
  } else {
    stop(context, ": unsupported positive multiplicative model.", call. = FALSE)
  }
  if (!length(out) || !is.finite(out) || out <= 0) {
    stop(context, ": invalid standardized multiplicative variance.", call. = FALSE)
  }
  as.numeric(out)
}

.GeoKriglocMultiplicativeCorr <- function(latent, model_code, nuisance,
                                           context = "GeoKrigloc") {
  code <- as.integer(model_code)[1L]
  latent <- as.numeric(latent)
  if (code == 21L) {
    out <- latent^2
  } else if (code == 26L) {
    sh <- as.numeric(nuisance["shape"])[1L]
    out <- .GeoWeibullCorr(latent, sh)
  } else if (code == 24L) {
    sh <- as.numeric(nuisance["shape"])[1L]
    cc2 <- latent^2
    factor <- pi * sin(2 * pi / sh) /
      (2 * sh * sin(pi / sh)^2 - pi * sin(2 * pi / sh))
    F1 <- hypergeo::hypergeo(-1 / sh, -1 / sh, 1, cc2)
    F2 <- hypergeo::hypergeo( 1 / sh,  1 / sh, 1, cc2)
    out <- factor * (Re(F1) * Re(F2) - 1)
  } else if (code == 22L) {
    ss <- as.numeric(nuisance["sill"])[1L]
    out <- (exp(ss * latent) - 1) / (exp(ss) - 1)
  } else {
    stop(context, ": unsupported positive multiplicative model.", call. = FALSE)
  }
  if (length(out) != length(latent) || any(!is.finite(out))) {
    stop(context, ": non-finite multiplicative correlation.", call. = FALSE)
  }
  as.numeric(out)
}

.GeoKriglocSpatialDenseCovariance <- function(coords_obs, template,
                                               param_vec, nuisance,
                                               eta_obs, copula_cont,
                                               plain_multiplicative = FALSE,
                                               corrmodel_code,
                                               distance_code,
                                               raw_corr = NULL,
                                               context = "GeoKrigloc") {
  coords_obs <- as.matrix(coords_obs)
  if (!is.numeric(coords_obs) || nrow(coords_obs) < 1L ||
      !(ncol(coords_obs) %in% c(2L, 3L)) || any(!is.finite(coords_obs))) {
    stop(context, ": invalid local coordinates for direct covariance assembly.",
         call. = FALSE)
  }
  if (ncol(coords_obs) == 2L) coords3 <- cbind(coords_obs, 0) else coords3 <- coords_obs
  storage.mode(coords3) <- "double"
  dime <- nrow(coords3)
  npairs <- as.double(dime) * as.double(dime - 1L) / 2

  ## For fix26 the strict lower triangle can be supplied from an adaptive
  ## global latent-correlation cache.  Otherwise compute it locally.  The
  ## native packed kernel evaluates each unordered pair only once; the legacy
  ## Corr_c(n x n) route remains available behind an internal regression
  ## option so numerical equivalence can be tested directly.
  if (is.null(raw_corr)) {
    use_packed <- isTRUE(getOption("GeoModels.krigloc.packed_correlation", TRUE))
    if (use_packed) {
      raw_corr <- .GeoKriglocSpatialPackedCorrelation(
        coords = coords3,
        corrmodel_code = corrmodel_code,
        corrparam = param_vec[template$namescorr],
        distance_code = distance_code,
        radius = template$local_cov_radius,
        context = context
      )
    } else {
      rho_full <- dotCall64::.C64(
        "Corr_c",
        SIGNATURE = c(
          "double", "double", "double", "double", "double", "integer",
          "integer", "double", "double", "double", "integer", "integer",
          "integer", "integer", "integer", "integer", "double", "integer",
          "integer", "double", "integer", "integer", "double"
        ),
        corri = dotCall64::vector_dc("double", dime * dime),
        coords3[, 1L], coords3[, 2L], coords3[, 3L], 0,
        as.integer(corrmodel_code), 0L,
        coords3[, 1L], coords3[, 2L], coords3[, 3L],
        as.integer(dime), as.integer(dime), 1L,
        as.integer(dime), 0L, 1L, param_vec[template$namescorr],
        0L, 0L, 0, as.integer(distance_code), 0L,
        as.numeric(template$local_cov_radius),
        INTENT = c("w", rep("r", 22)), NAOK = TRUE, PACKAGE = "GeoModels"
      )$corri
      if (length(rho_full) != dime * dime || any(!is.finite(rho_full))) {
        stop(context, ": non-finite local correlation matrix.", call. = FALSE)
      }
      rho_full <- matrix(rho_full, nrow = dime, ncol = dime)
      raw_corr <- if (npairs > 0) rho_full[lower.tri(rho_full)] else numeric(0L)
    }
  }
  if (length(raw_corr) != npairs || any(!is.finite(raw_corr))) {
    stop(context, ": invalid packed local correlation vector.", call. = FALSE)
  }

  nugget_scale <- 1 - as.numeric(nuisance["nugget"])
  if (!length(nugget_scale) || !is.finite(nugget_scale[1L])) nugget_scale <- 1
  nugget_scale <- nugget_scale[1L]

  if (isTRUE(copula_cont)) {
    moments <- .GeoCopulaMarginalMoments(template$model, nuisance, eta_obs)
    vv <- as.numeric(moments$variance)
    if (length(vv) != dime || any(!is.finite(vv))) {
      stop(context, ": invalid local marginal variances.", call. = FALSE)
    }
    return(.GeoDenseCopulaCovarianceBlocked(
      raw_corr = raw_corr, dime = dime, copula = template$copula,
      model_type = template$model, params = nuisance, eta_obs = eta_obs,
      variances = vv, nugget_scale = nugget_scale
    ))
  }

  if (isTRUE(plain_multiplicative)) {
    stdvar <- .GeoKriglocMultiplicativeStdVariance(
      template$model, nuisance, context = context
    )
    pair_corr <- .GeoKriglocMultiplicativeCorr(
      nugget_scale * raw_corr, template$model, nuisance, context = context
    )
    C <- diag(stdvar, nrow = dime, ncol = dime)
    if (length(pair_corr)) {
      pair_cov <- stdvar * pair_corr
      C[lower.tri(C)] <- pair_cov
      C <- t(C)
      C[lower.tri(C)] <- pair_cov
    }
    return(C)
  }

  if (as.integer(template$model) %in% c(20L, 34L, 40L)) {
    moments <- .GeoPlainTransformedMarginalMoments(template$model, nuisance, eta_obs)
    vv <- as.numeric(moments$variance)
    if (length(vv) == 1L) vv <- rep(vv, dime)
    if (length(vv) != dime || any(!is.finite(vv)) || any(vv <= 0))
      stop(context, ": invalid transformed marginal variances.", call. = FALSE)
    pair_cov <- .GeoPlainTransformedCrossCov(
      nugget_scale * raw_corr, template$model, nuisance
    )
    C <- diag(vv, nrow = dime, ncol = dime)
    if (length(pair_cov)) {
      C[lower.tri(C)] <- pair_cov
      C <- t(C)
      C[lower.tri(C)] <- pair_cov
    }
    return(C)
  }

  ## Gaussian direct path: construct the correlation matrix first and apply
  ## the marginal sill.
  sill <- as.numeric(nuisance["sill"])
  if (!length(sill) || !is.finite(sill[1L])) {
    stop(context, ": invalid Gaussian sill parameter.", call. = FALSE)
  }
  corr <- as.numeric(raw_corr) * nugget_scale
  C <- diag(dime)
  if (length(corr)) {
    C[lower.tri(C)] <- corr
    C <- t(C)
    C[lower.tri(C)] <- corr
  }
  C * sill[1L]
}


## Fast compact spatial Gaussian weights using the shared observation-pair
## correlation cache.  The cross-correlation to the prediction location is
## still evaluated directly because those pairs are location-specific.
.GeoKriglocWeightsCachedSpatial <- function(idx, i, vars) {
  coords_obs <- as.matrix(vars$coords_global[idx, , drop = FALSE])
  raw_corr <- .GeoPairCachePacked(
    idx, vars$pair_cache, context = "GeoKriglocWeights"
  )
  nuisance <- vars$cache_param_vec[vars$cov_template$namesnuis]
  nuisance <- nuisance[!is.na(names(nuisance))]
  C <- .GeoKriglocSpatialDenseCovariance(
    coords_obs = coords_obs,
    template = vars$cov_template,
    param_vec = vars$cache_param_vec,
    nuisance = nuisance,
    eta_obs = rep(0, length(idx)),
    copula_cont = FALSE,
    plain_multiplicative = FALSE,
    corrmodel_code = vars$corrmodel_code,
    distance_code = vars$distance_code,
    raw_corr = raw_corr,
    context = "GeoKriglocWeights"
  )

  cross_corr <- .GeoSpatialCrossCorrelation(
    coords_obs, vars$loc_metric[i, ],
    corrmodel_code = vars$corrmodel_code,
    corrparam = vars$cache_param_vec[vars$cov_template$namescorr],
    distance_code = vars$distance_code,
    radius = vars$cov_template$local_cov_radius,
    context = "GeoKriglocWeights"
  )
  nugget <- as.numeric(nuisance["nugget"])[1L]
  sill <- as.numeric(nuisance["sill"])[1L]
  if (!is.finite(nugget)) nugget <- 0
  if (!is.finite(sill))
    stop("GeoKriglocWeights: invalid sill in cached spatial core.", call. = FALSE)
  CC <- matrix((1 - nugget) * sill * cross_corr, ncol = 1L)
  as.numeric(.GeoKrigDenseSPDSolve(C, CC, context = "GeoKriglocWeights"))
}

## Compact Gaussian space-time weights from the shared pair cache.  Global
## observation coordinates/times are flattened once in neighborhood-index
## order, so static and dynamic space-time data use the same kernel.
.GeoKriglocWeightsCachedSpacetime <- function(idx, loc, pred_time, vars) {
  C <- .GeoGaussianPairCovariance(
    idx, vars$pair_cache, vars$param, context = "GeoKriglocWeights"
  )
  cross_corr <- .GeoPairCrossCorrelation(
    idx, vars$st_geometry$coords, vars$st_geometry$times,
    loc = loc, pred_time = pred_time,
    corrmodel_code = vars$corrmodel_code,
    corrparam = vars$corrparam,
    distance_code = vars$distance_code,
    radius = vars$radius,
    context = "GeoKriglocWeights"
  )
  nugget <- suppressWarnings(as.numeric(vars$param[["nugget"]])[1L])
  sill <- suppressWarnings(as.numeric(vars$param[["sill"]])[1L])
  if (!is.finite(nugget)) nugget <- 0
  if (!is.finite(sill) || sill <= 0)
    stop("GeoKriglocWeights: invalid sill in cached space-time core.", call. = FALSE)
  CC <- matrix((1 - nugget) * sill * cross_corr, ncol = 1L)
  as.numeric(.GeoKrigDenseSPDSolve(C, CC, context = "GeoKriglocWeights"))
}

## -------------------------------------------------------------------------
## Internal worker shared by serial and parallel spatial/bivariate weights.
## The neighborhood payload is intentionally compact: only indices and small
## per-task metadata are retained; coordinates/design rows are reconstructed
## from the global objects once inside the worker.
## -------------------------------------------------------------------------
.GeoKriglocWeightsTask <- function(i, vars) {
  idx <- vars$neigh$indices[[i]]
  if (is.null(idx) || !length(idx)) return(NULL)
  idx <- as.integer(idx)

  if (identical(vars$mode, "space") && isTRUE(vars$compact) &&
      !is.null(vars$pair_cache) && !is.null(vars$cov_template)) {
    return(list(
      weights = .GeoKriglocWeightsCachedSpatial(idx, i, vars),
      neighbor_indices = idx
    ))
  }

  param_local <- vars$param
  X_neigh <- NULL
  coordx_arg <- NULL
  coordx_dyn_arg <- NULL

  if (identical(vars$mode, "space")) {
    coordx_arg <- vars$coords_global[idx, , drop = FALSE]
    if (!is.null(vars$X_global))
      X_neigh <- vars$X_global[idx, , drop = FALSE]
    if (!is.null(vars$M_global))
      param_local$mean <- vars$M_global[idx]
  } else {
    idx_by_var <- vars$neigh$indices_by_variable[[i]]
    if (is.null(idx_by_var) || length(idx_by_var) != 2L ||
        any(vapply(idx_by_var, length, integer(1L)) == 0L)) return(NULL)

    if (!is.null(vars$coordx_dyn_global)) {
      coordx_dyn_arg <- Map(
        function(z, ii) z[ii, , drop = FALSE],
        vars$coordx_dyn_global, idx_by_var
      )
    } else {
      coordx_dyn_arg <- lapply(
        idx_by_var,
        function(ii) vars$coords_global[ii, , drop = FALSE]
      )
    }
    if (!is.null(vars$X_global))
      X_neigh <- vars$X_global[idx, , drop = FALSE]
  }

  xloc_row <- if (!is.null(vars$Xloc)) vars$Xloc[i, , drop = FALSE] else NULL

  ans <- GeoKrigWeights(
    coordx = coordx_arg,
    coordx_dyn = coordx_dyn_arg,
    loc = matrix(vars$loc_metric[i, ], nrow = 1L),
    corrmodel = vars$corrmodel,
    distance = vars$distance,
    n = vars$n,
    X = X_neigh,
    Xloc = xloc_row,
    Mloc = if (!is.null(vars$Mloc)) vars$Mloc[i] else NULL,
    sparse = vars$sparse,
    model = vars$model,
    param = param_local,
    anisopars = NULL,
    radius = vars$radius,
    copula = vars$copula,
    method = vars$method,
    which = vars$which
  )

  if (isTRUE(vars$compact)) {
    return(list(
      weights = as.numeric(ans$weights),
      neighbor_indices = idx
    ))
  }
  ans$neighbor_indices <- idx
  ans
}

.GeoKriglocWeightsRunTasks <- function(n_tasks, vars, parallel, workers) {
  ids <- seq_len(n_tasks)
  if (!isTRUE(parallel) || workers <= 1L) {
    return(lapply(ids, .GeoKriglocWeightsTask, vars = vars))
  }
  future.apply::future_lapply(
    ids, .GeoKriglocWeightsTask, vars = vars,
    future.seed = TRUE, future.globals = TRUE
  )
}

GeoKriglocWeights <- function(coordx=NULL, coordy=NULL, coordz=NULL, coordt=NULL, coordx_dyn=NULL,
                             corrmodel, distance="Eucl", grid=FALSE, loc, neighb=NULL,
                             maxdist=NULL, maxtime=NULL, method="cholesky", model="Gaussian",
                             n=1, nloc=NULL, param, anisopars=NULL,
                             radius=1, sparse=FALSE, time=NULL, which=1,
                             copula=NULL, X=NULL, Xloc=NULL, Mloc=NULL,
                             parallel=FALSE, ncores=6, compact=FALSE)  
{
  ###########################################################################
  ##  Validazione input
  ###########################################################################
  call = match.call()
  
  if(is.null(CkModel(model))) stop("The name of the model is not correct\n")
  if(!is.character(corrmodel) || is.null(CkCorrModel(corrmodel))) stop("the name of the correlation model is wrong")
  
  corrmodel <- gsub("[[:blank:]]", "", corrmodel)
  model <- .GeoPredictionCanonicalModel(gsub("[[:blank:]]", "", model))
  distance <- gsub("[[:blank:]]", "", distance)
  method <- gsub("[[:blank:]]", "", method)

  bivariate = CheckBiv(CkCorrModel(corrmodel))
  spacetime = CheckST(CkCorrModel(corrmodel))
  space = !spacetime && !bivariate

  if (!identical(model, "Gaussian"))
    stop("GeoKriglocWeights currently returns validated local weights only for model = 'Gaussian'.", call. = FALSE)
  if (!is.null(copula))
    stop("GeoKriglocWeights currently requires copula = NULL.", call. = FALSE)

  if (spacetime && !is.null(coordx_dyn)) {
    coordx_dyn <- .GeoNormalizeDynamicCoordinates(
      coordx_dyn, coordt = coordt, context = "GeoKriglocWeights"
    )
    .GeoValidateDynamicRows(X, coordx_dyn, name = "X",
                            context = "GeoKriglocWeights")
  }

  biv_info <- NULL
  if (bivariate) {
    biv_info <- .GeoBivariateCoordinateInfo(
      coordx = coordx, coordy = coordy, coordz = coordz,
      coordx_dyn = coordx_dyn, grid = grid,
      context = "GeoKriglocWeights"
    )
    if (!is.null(X)) {
      X <- .GeoBivariateDesign(X, biv_info$ns,
                               context = "GeoKriglocWeights")
    }
    if (biv_info$dynamic) {
      coordx_dyn <- biv_info$blocks
      coordx <- coordy <- coordz <- NULL
    }
  }

  # coordinate osservate
  if(is.null(coordx_dyn)){
    if(is.null(coordy)) {
      coords = as.matrix(coordx)
    } else {
      if(!grid) {
        coords = cbind(coordx, coordy)
        if(!is.null(coordz)) coords = cbind(coords, coordz)
      } else {
        if(!is.null(coordz)) {
          coords = as.matrix(expand.grid(coordx, coordy, coordz))
        } else {
          coords = as.matrix(expand.grid(coordx, coordy))
        }
      }
    }
  } else {
    coords = NULL
  }

  if(is.vector(loc)) loc = matrix(loc, nrow=1)
  loc = as.matrix(loc)

  .GeoValidate3D(
    coordx = if (is.null(coordx_dyn)) coords else NULL,
    coordx_dyn = coordx_dyn, distance = distance, grid = grid,
    context = "GeoKriglocWeights"
  )
  .GeoValidate3D(
    coordx = loc, distance = distance, grid = FALSE,
    context = "GeoKriglocWeights prediction locations"
  )
  .GeoValidateCoordinateCompatibility(
    coordx = if (is.null(coordx_dyn)) coords else NULL,
    coordx_dyn = coordx_dyn, loc = loc, context = "GeoKriglocWeights"
  )
  Nloc = nrow(loc)
  
  if(is.null(time)) time = 0
  Tloc = length(time)
  if(bivariate) Tloc = 1
  
  ## Mean specification for local weights.  The weights themselves do not
  ## depend on the mean in the Gaussian case, but X/Xloc and M/Mloc must be
  ## aligned with the same local tasks used by GeoKrigloc().
  if (!bivariate) {
    nobs_mean <- if (!is.null(coordx_dyn)) {
      sum(vapply(coordx_dyn, nrow, integer(1L)))
    } else if (spacetime) {
      nrow(coords) * length(coordt)
    } else {
      nrow(coords)
    }
    M <- .GeoMean_external(param, n = nobs_mean)
    if (!is.null(M) && !is.null(X))
      stop("A fixed observation mean and X cannot be used together.")
    if (!is.null(Mloc) && !is.null(Xloc))
      stop("Use either Mloc or Xloc, not both.")
    if (!is.null(M) && is.null(Mloc))
      stop("Mloc is required when the observation mean is supplied as a vector.")
    if (!is.null(X)) {
      X <- .GeoMean_design(X, n = nobs_mean, name = "X")
      p_mean <- ncol(X)
      invisible(.GeoMean_beta(param, p_mean))
      if (is.null(Xloc) && is.null(Mloc))
        stop("Xloc or Mloc is required when X is supplied.")
    } else {
      p_mean <- 1L
      if (!is.null(Xloc)) stop("X must be supplied when Xloc is used.")
      if (is.null(M)) {
        if (length(setdiff(.GeoMean_supplied_names(param), "mean")))
          stop("Parameters mean1, mean2, ... require an explicit design matrix X; without X the mean is intercept-only.")
        invisible(.GeoMean_beta(param, 1L))
      }
    }

    total_mean_tasks <- if (space) Nloc else Nloc * Tloc
    if (!is.null(Mloc)) {
      if (!is.numeric(Mloc) || any(!is.finite(Mloc)))
        stop("Mloc must be finite and numeric.")
      Mloc <- as.numeric(Mloc)
      if (length(Mloc) == 1L) Mloc <- rep(Mloc, total_mean_tasks)
      if (length(Mloc) != total_mean_tasks)
        stop("Mloc must have one value per local prediction task.")
      Xloc <- NULL
    } else if (!is.null(Xloc)) {
      Xloc <- .GeoMean_design(Xloc, n = total_mean_tasks, p = p_mean, name = "Xloc")
    }
  } else {
    M <- NULL
    p_mean <- if (is.null(X)) 1L else ncol(X)
    invisible(.GeoBivariateValidateMeanParameters(
      param, p_mean, context = "GeoKriglocWeights"
    ))
    if (!is.null(Mloc) && !is.null(Xloc))
      stop("Use either Mloc or Xloc, not both.")
    if (!is.null(X) && is.null(Xloc) && is.null(Mloc))
      stop("Xloc or Mloc is required when X is supplied.")
    if (is.null(X) && !is.null(Xloc))
      stop("X must be supplied when Xloc is used.")
    if (!is.null(Mloc)) {
      if (!is.numeric(Mloc) || any(!is.finite(Mloc)))
        stop("Mloc must be finite and numeric.")
      Mloc <- as.numeric(Mloc)
      if (length(Mloc) == 1L) Mloc <- rep(Mloc, Nloc)
      if (length(Mloc) != Nloc)
        stop("Mloc must have one value per prediction location.")
      Xloc <- NULL
    } else if (!is.null(Xloc)) {
      Xloc <- .GeoMean_design(Xloc, n = Nloc, p = p_mean, name = "Xloc")
    }
  }

  ###########################################################################
  ##  SUGG. 1: anisotropia anche per la selezione dei vicini
  ###########################################################################
  coords_aniso <- coords
  coordx_dyn_aniso <- coordx_dyn
  loc_aniso    <- loc
  if (!is.null(anisopars)) {
    if (!is.null(coords)) coords_aniso <- GeoAniso(coords, c(anisopars$angle, anisopars$ratio))
    if (!is.null(coordx_dyn)) {
      if (ncol(coordx_dyn[[1L]]) != 2L)
        stop("GeoKriglocWeights: anisotropy for dynamic three-dimensional coordinates is not implemented.", call. = FALSE)
      coordx_dyn_aniso <- lapply(coordx_dyn, function(z)
        GeoAniso(z, c(anisopars$angle, anisopars$ratio)))
    }
    loc_aniso <- GeoAniso(loc, c(anisopars$angle, anisopars$ratio))
  }

  ###########################################################################
  ##  Memory + centralized parallel plan
  ###########################################################################
  if (!is.logical(parallel) || length(parallel) != 1L || is.na(parallel))
    stop("parallel must be TRUE or FALSE.", call. = FALSE)
  if (!is.logical(compact) || length(compact) != 1L || is.na(compact))
    stop("compact must be TRUE or FALSE.", call. = FALSE)

  total_parallel_tasks <- if (spacetime) Nloc * Tloc else Nloc
  workers <- .GeoResolveWorkers(parallel, ncores, n_jobs = total_parallel_tasks)
  parallel <- isTRUE(parallel) && workers > 1L

  if (parallel && !spacetime) {
    old_limit <- getOption("future.globals.maxSize")
    old_plan <- future::plan()
    options(future.globals.maxSize = 4 * 1024^3)
    .GeoKriglocSetFuturePlan(workers)
    on.exit(options(future.globals.maxSize = old_limit), add = TRUE)
    on.exit(try(future::plan(old_plan), silent = TRUE), add = TRUE)
  }

  ###########################################################################
  ##  Loop Kriging - SPACE
  ###########################################################################
  if(space){
    neigh = GeoNeighborhood(coordx = coords_aniso, distance = distance, loc = loc_aniso,
                            neighb = neighb, maxdist = maxdist, X = X, M = M,
                            parallel = FALSE, ncores = NULL)

    compact_neigh <- list(indices = neigh$indices)
    cov_template <- pair_cache <- cache_param_vec <- NULL
    if (isTRUE(compact) && !isTRUE(sparse)) {
      cov_template <- .GeoKriglocPrepareSpatialCovariance(
        neigh = compact_neigh, coordx_global = coords_aniso,
        X_global = X, param = param, corrmodel = corrmodel,
        distance = distance, n = n, model = model, radius = radius,
        copula = NULL, M = M
      )
      if (!is.null(cov_template) &&
          .GeoPairCacheEnabled(corrmodel, param, weights = TRUE, spacetime = FALSE)) {
        cache_param_vec <- .GeoKriglocLocalParamVector(cov_template, param)
        pair_cache <- .GeoBuildSpatialPairCache(
          compact_neigh$indices, coords_aniso,
          corrmodel_code = CkCorrModel(corrmodel),
          corrparam = cache_param_vec[cov_template$namescorr],
          distance_code = CheckDistance(distance),
          radius = cov_template$local_cov_radius,
          context = "GeoKriglocWeights"
        )
      }
    }
    task_vars <- list(
      mode = "space",
      neigh = compact_neigh,
      coords_global = coords_aniso,
      coordx_dyn_global = NULL,
      X_global = X,
      M_global = M,
      loc_metric = loc_aniso,
      Xloc = Xloc,
      Mloc = Mloc,
      param = param,
      corrmodel = corrmodel,
      distance = distance,
      n = n,
      sparse = sparse,
      model = model,
      radius = radius,
      copula = copula,
      method = method,
      which = which,
      compact = compact,
      pair_cache = pair_cache,
      cov_template = cov_template,
      cache_param_vec = cache_param_vec,
      corrmodel_code = CkCorrModel(corrmodel),
      distance_code = CheckDistance(distance)
    )
    rm(neigh, compact_neigh)
    invisible(gc(verbose = FALSE))
    weights_list <- .GeoKriglocWeightsRunTasks(
      Nloc, task_vars, parallel = parallel, workers = workers
    )
  }

  ###########################################################################
  ##  Loop Kriging - SPACETIME
  ###########################################################################
  if(spacetime){
    neigh = GeoNeighborhood(coordx = coords_aniso, coordt = coordt,
                            distance = distance, neighb = neighb,
                            coordx_dyn = coordx_dyn_aniso, loc = loc_aniso, time = time,
                            maxdist = maxdist, maxtime = maxtime, X = X, M = M,
                            parallel = FALSE, ncores = NULL)
    total_tasks = Nloc * Tloc
    weights_list = vector("list", total_tasks)

    st_pair_cache <- st_geometry <- corrparam_st <- NULL
    cache_st <- isTRUE(compact) && !isTRUE(sparse) &&
      identical(tolower(method), "cholesky") &&
      .GeoPairCacheEnabled(corrmodel, param, weights = TRUE, spacetime = TRUE)
    if (cache_st) {
      st_geometry <- .GeoSpacetimeObservationGeometry(
        coords_aniso, coordx_dyn_aniso, coordt, context = "GeoKriglocWeights"
      )
      corrparam_st <- .GeoPairCorrelationParam(
        corrmodel, param, context = "GeoKriglocWeights"
      )
      st_pair_cache <- .GeoBuildSpacetimePairCache(
        neigh$indices, st_geometry$coords, st_geometry$times,
        corrmodel_code = CkCorrModel(corrmodel), corrparam = corrparam_st,
        distance_code = CheckDistance(distance), radius = radius,
        context = "GeoKriglocWeights"
      )
    }
    st_cache_vars <- if (!is.null(st_pair_cache)) list(
      pair_cache = st_pair_cache, st_geometry = st_geometry, param = param,
      corrmodel_code = CkCorrModel(corrmodel), corrparam = corrparam_st,
      distance_code = CheckDistance(distance), radius = radius
    ) else NULL

    k = 1
    for(i in 1:Nloc){
      for(j in 1:Tloc){
        param_local = param
        if(!is.null(M) && !is.null(neigh$M)) param_local$mean = neigh$M[[k]]
        coords_neigh = neigh$coordx[[k]]
        # SUGG. 4: check robusto
        if (!is.null(coords_neigh) && !(is.matrix(coords_neigh) && nrow(coords_neigh) == 0) && length(coords_neigh) > 0){
          # SUGG. 3: Xloc riga robusta con indice spacetime
          idx_st <- k
          xloc_row <- if (!is.null(Xloc)) Xloc[idx_st, , drop = FALSE] else NULL

          if (!is.null(st_cache_vars)) {
            weights_list[[k]] <- list(
              weights = .GeoKriglocWeightsCachedSpacetime(
                as.integer(neigh$indices[[k]]), loc_aniso[i, ], time[j],
                st_cache_vars
              ),
              neighbor_indices = as.integer(neigh$indices[[k]])
            )
            k = k + 1L
            next
          }

          coordx_dyn_neigh = if (!is.null(neigh$coordx_dyn)) neigh$coordx_dyn[[k]] else NULL
          weights_result = GeoKrigWeights(
            coordx = if (is.null(coordx_dyn_neigh)) coords_neigh else NULL,
            coordt = neigh$coordt[[k]],
            coordx_dyn = coordx_dyn_neigh,
            loc = matrix(loc_aniso[i, ], nrow = 1),
            time = time[j], 
            X = if(!is.null(neigh$X)) neigh$X[[k]] else NULL,
            Mloc = if(!is.null(Mloc)) Mloc[idx_st] else NULL,
            Xloc = xloc_row,
            sparse = sparse, 
            corrmodel = corrmodel,
            distance = distance, 
            model = model, 
            param = param_local,
            anisopars = NULL,
            radius = radius,
            copula = copula, 
            n = n, 
            method = method,
            which = which
          )
          if (isTRUE(compact)) {
            weights_list[[k]] <- list(
              weights = as.numeric(weights_result$weights),
              neighbor_indices = as.integer(neigh$indices[[k]])
            )
          } else {
            weights_result$neighbor_indices = neigh$indices[[k]]
            weights_list[[k]] = weights_result
          }
        } else {
          weights_list[k] <- list(NULL)
        }
        k = k + 1
      }
    }
  }

  ###########################################################################
  ##  Loop Kriging - BIVARIATE
  ###########################################################################
  if(bivariate){
    neigh = GeoNeighborhood(coordx = coords_aniso, distance = distance,
                            neighb = neighb, coordx_dyn = coordx_dyn_aniso,
                            bivariate = TRUE,
                            loc = loc_aniso, maxdist = maxdist, X = X, M = M,
                            parallel = FALSE, ncores = NULL)

    compact_neigh <- list(
      indices = neigh$indices,
      indices_by_variable = neigh$indices_by_time
    )
    task_vars <- list(
      mode = "bivariate",
      neigh = compact_neigh,
      coords_global = coords_aniso,
      coordx_dyn_global = coordx_dyn_aniso,
      X_global = X,
      M_global = NULL,
      loc_metric = loc_aniso,
      Xloc = Xloc,
      Mloc = Mloc,
      param = param,
      corrmodel = corrmodel,
      distance = distance,
      n = n,
      sparse = sparse,
      model = model,
      radius = radius,
      copula = copula,
      method = method,
      which = which,
      compact = compact
    )
    rm(neigh, compact_neigh)
    invisible(gc(verbose = FALSE))
    weights_list <- .GeoKriglocWeightsRunTasks(
      Nloc, task_vars, parallel = parallel, workers = workers
    )
  }

  ###########################################################################
  ##  Output
  ###########################################################################
  GeoKriglocWeights_out = list(
    bivariate = bivariate,
    coordx = coordx,
    coordy = coordy,
    coordz = coordz,
    coordt = coordt,
    coordx_dyn = coordx_dyn,
    corrmodel = corrmodel,
    data = NULL,
    distance = distance,
    grid = grid,
    loc = loc,
    copula = copula,
    numcoord = if(!is.null(coords)) nrow(coords) else length(coordx),
    numloc = Nloc,
    numtime = if(!is.null(coordt)) length(coordt) else 0,
    numt = Tloc,
    maxdist = maxdist,
    maxtime = maxtime,
    model = model,
    n = n,
    param = param,
    weights = weights_list,
    radius = radius,
    spacetime = spacetime,
    time = time,
    which = which,
    neighb = neighb,
    compact = compact,
    ncores = if (parallel) workers else 1L
  )
  
  structure(c(GeoKriglocWeights_out, call = call), class = c("GeoKriglocWeights"))
}

#########################################################
#########################################################
############ functions for TB############################
#########################################################
#########################################################



##################################################### 
######### bivarate case #############################
#####################################################
tbm2d <- function(coord, coordt, param, corrmodel,L,bivariate){

  N=1; n <- dim(coord)[1]; d <- 2
    if(corrmodel == "Matern"){
       CC=1; N=1
       a <- as.numeric(param['scale']);  nu1 <- as.numeric(param['smooth'])
       a0 = a; nu0 =nu1
       parameters <- list("CC" = CC, "a" = a,"nu1" = nu1,"nu2" = 0 )
       P <- 1;  vtype = 0}

  if (corrmodel=="Bi_matern"||corrmodel=="Bi_Matern"){
    corrmodel <- "Matern"
    a <- matrix(0,2,2)
    a[1,1] <- as.numeric(param['scale_1'])
    a[2,2] <- as.numeric(param['scale_2'])
    a[1,2] <- a[2,1] <- as.numeric(param['scale_12'])
    
    nu1 <- matrix(0,2,2)
    nu1[1,1] <- as.numeric(param['smooth_1'])
    nu1[2,2] <- as.numeric(param['smooth_2'])
    nu1[1,2] <- nu1[2,1] <- as.numeric(param['smooth_12'])
    CC <- matrix(0,2,2)
    CC[1,1] <- CC[2,2] <- 1
    CC[1,2] <- CC[2,1] <- as.numeric(param['pcol'])
    a0 <- min(a)
    nu0 <- min(nu1)
    parameters <- list("C" = CC, "a" = a,"nu1" = nu1,"nu2" =matrix(0,2,2) )
    P<-2; vtype = 0  

  }
   parametersg <- list("a" = a0,"nu1" = nu0)

      A <- matrix(0, P, P*L*N); B <- matrix(0, P, P*L*N)
      G <- matrix(rgamma(P*L*N, nu0, scale = 1),P*L*N,d)
      u <- matrix(rnorm(P*L*N*d), P*L*N, d)/sqrt(G*2)/a0/(2*pi)
      phi <- 2*pi*runif(P*L*N)
      sequen <- c(seq(0,n-0.5, by = ceiling(1e6/P/N)),n)

    m = c()
   for (i in 1:(length(sequen)-1)){ m1 <- sequen[i+1]-sequen[i]; m = c(m1,m)} 
simu11 <- numeric(N * P * n)

result=dotCall64::.C64("for_c",
         SIGNATURE = c("integer","double","double", "double",
                     "double","integer","integer","integer","integer",
                     "double","double","double","double","double",
                     "integer","integer","integer",
                     "double","double","integer","integer","double",
                     "double"), 
                         d,c(a),c(nu1), c(CC), c(parameters$nu2),P,N,L,CkCorrModel(corrmodel),
                         c(u),a0,nu0,c(A),c(B),c(sequen),length(sequen),n,coord,phi,vtype,m,
                         simu1=dotCall64::vector_dc("double",length(simu11)),L,
         INTENT =    c(rep("r",20),
                       "r", "rw","r"),
             PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)$simu1
  simu =  matrix(result,n,P)
  return(simu)
}

#######################################################
######### univariate case #############################
#######################################################

## Lightweight top-level worker used by future.apply.  Keeping this
## function outside tbm2d_uni() prevents it from capturing the complete
## simulation frame when it is serialized for multisession workers.
.tb_call_kernel_chunk <- function(coord_chunk, u, phi, amp, L, CC) {
  coord_chunk <- as.matrix(coord_chunk)
  u <- as.matrix(u)
  Ninner <- nrow(coord_chunk)
  d <- ncol(coord_chunk)

  if (!(d %in% c(2L, 3L)) || ncol(u) != d) {
    stop("TB kernel requires matching two- or three-dimensional coordinates and frequencies.",
         call. = FALSE)
  }

  result <- dotCall64::.C64(
    "TBDnd_amp",
    SIGNATURE = c(
      "double", "double", "double", "double",
      "integer", "integer", "integer", "double"
    ),
    omega = c(u),
    coords = c(coord_chunk),
    phi = phi,
    amp = amp,
    L = as.integer(L),
    n = as.integer(Ninner),
    d = as.integer(d),
    result = dotCall64::vector_dc("double", Ninner),
    INTENT = c("r", "r", "r", "r", "r", "r", "r", "rw"),
    PACKAGE = "GeoModels",
    VERBOSE = 0,
    NAOK = TRUE
  )$result

  sqrt(CC) * result / sqrt(L)
}

safe_besselJ <- function(x, nu) {
  out <- numeric(length(x))
  x <- as.numeric(x)
  nu <- as.numeric(nu)

  tiny_thr <- 1e-6
  big_thr  <- 200

  small <- is.finite(x) & x <= tiny_thr & x >= 0
  if (any(small)) {
    z <- x[small]
    out[small] <- (pmax(z, .Machine$double.eps)/2)^nu / gamma(nu + 1)
  }

  mid <- is.finite(x) & x > tiny_thr & x <= big_thr
  if (any(mid)) {
    out[mid] <- base::besselJ(x[mid], nu)
  }

  big <- is.finite(x) & x > big_thr
  if (any(big)) {
    z <- x[big]
    out[big] <- sqrt(2/(pi*z)) * cos(z - nu*pi/2 - pi/4)
  }

  out[!is.finite(out)] <- 0
  out
}


sphere_area <- function(d) {
  2 * pi^(d / 2) / gamma(d / 2)
}

#########################################################
# beta  MIXTURE METHOD 
#########################################################
C_H <- function(kappa, d) {
  num <- gamma((d + 1) / 2 + kappa) * gamma(1 + kappa) * gamma(d / 2 + 1 + 2 * kappa)
  den <- 2^d * pi^(d / 2) * gamma(0.5 + kappa) * gamma(d / 2 + 1 + kappa) * gamma(d + 1 + 2 * kappa)
  num / den
}

f_T <- function(t, kappa, d, log_scale = FALSE) {
  t <- as.numeric(t)
  out <- numeric(length(t))
  out[t <= 0] <- if (log_scale) -Inf else 0
  if (all(t <= 0)) return(out)

  delta <- (d + 1) / 2 + kappa
  S_d <- sphere_area(d)
  CH  <- C_H(kappa, d)
  const <- S_d * CH * gamma(delta + 0.5)^2 * 4^(2 * delta - 1)

  pos <- t > 0
  x <- t[pos] / 2

  #bval <- besselJ(x, delta - 0.5)
  bval <-safe_besselJ(x, delta - 0.5)
  #big <- x > 200
  #if (any(big)) {
  #  nu <- delta - 0.5
  #  bval[big] <- sqrt(2 / (pi * x[big])) * cos(x[big] - nu * pi / 2 - pi / 4)
  #}

  val <- const * pmax(t[pos], 1e-12)^(d - 2 * delta) * (bval^2)

  if (log_scale) {
    out[pos] <- log(pmax(val, .Machine$double.eps))
  } else {
    out[pos] <- val
  }
  out
}

sample_fT_final <- function(n, kappa, d, T0 = NULL, safety = 1.15, batch = 100000, 
                             max_iter = 1000, optimize_T0 = TRUE) {
  delta <- (d + 1) / 2 + kappa
  S_d   <- sphere_area(d)
  CH    <- C_H(kappa, d)
  B_small <- S_d * CH
  C0 <- S_d * CH * gamma(delta + 0.5)^2 * 4^(2 * delta - 1)
  A_tail <- (4 / pi) * C0
  alpha  <- 2 * kappa + 2
  shape  <- alpha - 1
  
  M1_from_T0 <- function(T0) safety * (B_small * T0^d) / d
  M2_from_T0 <- function(T0) safety * (A_tail / (shape * T0^shape))
  
  if (is.null(T0) && optimize_T0) {
    objective <- function(T0) M1_from_T0(T0) + M2_from_T0(T0)
    T0 <- optimize(objective, interval = c(0.5, 50), maximum = FALSE)$minimum
  } else if (is.null(T0)) {
    t_mode <- tryCatch(
      optimize(function(t) f_T(t, kappa, d), c(1e-3, 40), maximum = TRUE)$maximum,
      error = function(e) 2
    )
    T0 <- max(6, 2 * t_mode)
  }
  
  M1 <- M1_from_T0(T0)
  M2 <- M2_from_T0(T0)
  w1 <- M1 / (M1 + M2)
  w2 <- 1 - w1
  
  g1_rnd <- function(n) T0 * runif(n)^(1 / d)
  g1_pdf <- function(t) ifelse(t >= 0 & t <= T0, d * t^(d - 1) / (T0^d), 0)
  g2_rnd <- function(n) T0 * (1 - runif(n))^(-1 / shape)
  g2_pdf <- function(t) ifelse(t > T0, shape * (T0^shape) / (t^(shape + 1)), 0)
  
  samples <- numeric(0)
  n_total_proposed <- 0
  n_iter <- 0
  
  while (length(samples) < n && n_iter < max_iter) {
    n_iter <- n_iter + 1
    n_batch <- min(batch, (n - length(samples)) * 5)
    
    use_body <- runif(n_batch) < w1
    n_body <- sum(use_body)
    n_tail <- n_batch - n_body
    
    proposals <- numeric(n_batch)
    M_vals <- numeric(n_batch)
    g_vals <- numeric(n_batch)
    
    if (n_body > 0) {
      proposals[use_body] <- g1_rnd(n_body)
      M_vals[use_body] <- M1
      g_vals[use_body] <- g1_pdf(proposals[use_body])
    }
    
    if (n_tail > 0) {
      proposals[!use_body] <- g2_rnd(n_tail)
      M_vals[!use_body] <- M2
      g_vals[!use_body] <- g2_pdf(proposals[!use_body])
    }
    
    f_vals <- f_T(proposals, kappa, d)
    denom <- M_vals * g_vals
    ratio <- ifelse(denom > 0 & is.finite(denom), f_vals / denom, 0)
    ratio[!is.finite(ratio)] <- 0

    if (any(ratio > 1 + 1e-8, na.rm = TRUE)) {
      warning("Envelope violation in sample_fT_final: f/(M*g) exceeded 1; clipping acceptance probabilities.")
    }

    u <- runif(n_batch)
    accept <- u < pmin(1, pmax(0, ratio))
    
    samples <- c(samples, proposals[accept])
    n_total_proposed <- n_total_proposed + n_batch
  }

  if (length(samples) < n) {
    stop("sample_fT_final failed: generated fewer than n samples. Increase max_iter or check the GH parameters.")
  }
  
  samples[seq_len(n)]
}

#####################################################################
simulate_GH_beta <- function(L, d, kappa, mu, l, a) {
  # Caso limite mu = 1: U -> 1 quasi sicuramente
  if (abs(mu - 1) < 1e-8) {
    shapeV1 <- d / 2 + 2 * kappa + 1
    shapeV2 <- mu / 2 - d / 2 - 0.5 - kappa + l
    V <- if (shapeV2 <= 1e-10) rep(1, L) else rbeta(L, shapeV1, shapeV2)
    U <- rep(1, L)   # limite U → 1
  } else if (mu > 1) {
    shapeU1 <- kappa + 1
    shapeU2 <- mu / 2 - 0.5
    shapeV1 <- d / 2 + 2 * kappa + 1
    shapeV2 <- mu / 2 - d / 2 - 0.5 - kappa + l
    U <- rbeta(L, shapeU1, shapeU2)
    V <- if (shapeV2 <= 1e-10) rep(1, L) else rbeta(L, shapeV1, shapeV2)
  } else {
    stop("mu must be >= 1 for the Beta mixture representation.")
  }

  b <- a * sqrt(U * V)
  Tvals <- sample_fT_final(L, kappa, d)
  R <- Tvals / b

  Theta <- matrix(rnorm(L * d), L, d)
  Theta <- Theta / sqrt(rowSums(Theta^2))
  R * Theta
}
###################################################################

#########################################################
# GASPER MIXTURE METHOD 
#########################################################

log_C_H_n <- function(eta, n, d) {
  kappa <- eta + n - d/2
  lgamma(d/2) + lgamma((d + 1)/2 + kappa) +
    lgamma(1 + kappa - n) + lgamma(d/2 + 1 + 2*kappa - n) -
    ((d + 2*n) * log(2) + (d/2) * log(pi) +
       lgamma(d/2 + n) + lgamma(0.5 + kappa - n) +
       lgamma(d/2 + 1 + kappa) + lgamma(d + 1 + 2*kappa))
}

C_4F3_terminating <- function(n, eta, delta, beta, gamma) {
  if (n == 0) return(1.0)
  tk <- 1.0
  s <- 1.0
  for (k in 0:(n - 1)) {
    num <- (-n + k) * (n + 2*eta + k) * (eta + 1 + k) * (delta + k)
    den <- (eta + 0.5 + k) * (beta + k) * (gamma + k) * (k + 1)
    tk  <- tk * (num / den)
    s   <- s + tk
  }
  s
}

w_n_stable <- function(nmax, delta, beta, gamma, d) {
  eta <- 0.5 * (beta + gamma - delta - 1.5)
  logL <- lgamma(delta) + lgamma(beta - d/2) + lgamma(gamma - d/2) -
          (d*log(2) + (d/2)*log(pi) + lgamma(delta - d/2) + lgamma(beta) + lgamma(gamma))

  w_log <- numeric(nmax + 1)
  for (n in 0:nmax) {
    Cn_val <- C_4F3_terminating(n, eta, delta, beta, gamma)
    if (!is.finite(Cn_val) || Cn_val < 0) Cn_val <- 0
    logCH <- log_C_H_n(eta, n, d)
    term1 <- 2 * (lgamma(eta + 1) - lgamma(eta + n + 1))
    term2 <- log(2 * n + 2 * eta) + (lgamma(2 * eta + 1 + n) - lgamma(2 * eta + 1)) -
             log(n + 2 * eta) - lgamma(n + 1)
    w_log[n + 1] <- logL + term1 - 4 * n * log(2) + term2 +
                    log(max(Cn_val, .Machine$double.eps)) - logCH
  }
  m <- max(w_log)
  w <- exp(w_log - m)
  w <- pmax(w, 0)
  w / sum(w)
}
sample_fT_n_adaptive <- function(N, eta, n, d, safety = 1.2) {
  Sd <- sphere_area(d)
  CH <- exp(log_C_H_n(eta, n, d))
  const <- Sd * CH * 4^(2*eta + 2*n) * gamma(1 + eta + n)^2
  nu <- eta + n

  A_body <- Sd * CH
  A_tail <- const * (4 / pi)
  alpha  <- 2*eta + 1 - d

  M1_from_T0 <- function(T0) safety * (A_body / d) * T0^(d + 2*n)
  M2_from_T0 <- function(T0) safety * (A_tail / (alpha * T0^alpha))
  T0 <- optimize(function(T0) M1_from_T0(T0) + M2_from_T0(T0),
                 interval = c(0.5, 60), maximum = FALSE)$minimum

  M1 <- M1_from_T0(T0)
  M2 <- M2_from_T0(T0)
  w1 <- M1 / (M1 + M2)

  g1_rnd <- function(m) T0 * runif(m)^(1 / d)
  g1_pdf <- function(t) ifelse(t >= 0 & t <= T0, d * t^(d - 1) / (T0^d), 0)
  g2_rnd <- function(m) T0 * (1 - runif(m))^(-1 / alpha)
  g2_pdf <- function(t) ifelse(t > T0, alpha * T0^alpha / (t^(alpha + 1)), 0)

  #f_target <- function(t) const * t^(d - 1 - 2*eta) * besselJ(t / 2, nu)^2
  f_target <- function(t) const * t^(d - 1 - 2*eta) * safe_besselJ(t / 2, nu)^2
  samples <- numeric(0)
  while (length(samples) < N) {
    m <- min(10000, (N - length(samples)) * 5)
    use_body <- runif(m) < w1
    t_prop <- numeric(m)
    gval <- numeric(m)
    Mval <- numeric(m)

    if (any(use_body)) {
      nb <- sum(use_body)
      t_prop[use_body] <- g1_rnd(nb)
      gval[use_body]   <- g1_pdf(t_prop[use_body])
      Mval[use_body]   <- M1
    }
    if (any(!use_body)) {
      nt <- sum(!use_body)
      t_prop[!use_body] <- g2_rnd(nt)
      gval[!use_body]   <- g2_pdf(t_prop[!use_body])
      Mval[!use_body]   <- M2
    }

    fval <- f_target(t_prop)
    acc  <- runif(m) < fval / (Mval * gval)
    samples <- c(samples, t_prop[acc])
  }
  samples[1:N]
}

simulate_GH_gasper <- function(L, d, kappa, mu, l, a, nmax = 15) {
  delta <- (d + 1) / 2 + kappa
  beta  <- delta + mu / 2
  gamma <- delta + mu / 2 + l
  eta   <- 0.5 * (beta + gamma - delta - 1.5)
  
  weights <- w_n_stable(nmax, delta, beta, gamma, d)
  n_indices <- sample(0:nmax, size = L, replace = TRUE, prob = weights)
  
  T_samples <- numeric(L)
  
  for (n in 0:nmax) {
    indices_n <- which(n_indices == n)
    n_samples <- length(indices_n)
    
    if (n_samples == 0) next
    
    T_samples[indices_n] <- sample_fT_n_adaptive(
      N = n_samples, 
      eta = eta, 
      n = n, 
      d = d,
      safety = 1.2
    )
  }
  
  R_samples <- T_samples / a
  
  Theta <- matrix(rnorm(L * d), nrow = L, ncol = d)
  Theta <- Theta / sqrt(rowSums(Theta^2))
  
  Omega <- R_samples * Theta
  
  return(Omega)
}

#########################################################
# MAIN SIMULATION FUNCTION
#########################################################

tbm2d_uni <- function(coord, coordt, param, corrmodel, L, bivariate, parallel, ncores){
  
  N <- 1
  n <- dim(coord)[1]
  d <- dim(coord)[2]

  if (is.null(d) || !(d %in% c(2L, 3L))) {
    stop("Current univariate TB/STB implementation supports only 2D or 3D spatial coordinates.",
         call. = FALSE)
  }

  condition <- FALSE

#########################################################################
  if (corrmodel %in% c("GenWend","Genwend","Hypergeometric","GenWend_Matern","Genwend_Matern","Hypergeometric_Matern")) {
    if (corrmodel %in% c("GenWend","Genwend","GenWend_Matern","Genwend_Matern"))   l <- 0.5
    if (corrmodel %in% c("Hypergeometric","Hypergeometric_Matern"))                l <- d/2 + param$smooth
    
    if (corrmodel %in% c("GenWend_Matern","Genwend_Matern","Hypergeometric_Matern"))  mu <- 1/as.numeric(param['power2'])
    else                                                                              mu <- as.numeric(param['power2'])
    
    C2 = mu/2 - d/2 - 0.5 - param$smooth + l >= 0
    C1 = mu >= 1
    condition <- isTRUE(C1 && C2)
    
    CC <- as.double(param['sill'])
    N <- 1
    a  <- as.double(param['scale'])
    nu1 <- as.double(param['smooth'])
    
    if (corrmodel %in% c("GenWend_Matern","Hypergeometric_Matern"))
      other <- 1/as.numeric(param['power2'])
    else
      other <- as.numeric(param['power2'])
    
    a0 <- a
    nu0 <- nu1
    parameters <- list("CC" = CC, "a" = a,"nu1" = nu1, "nu2" = 0, "other" = other)
    P <- 1
    vtype <- 0
    parametersg <- list("a" = a0, "nu1" = nu0, other = other, CC = CC,
                        mu = mu, kappa = param$smooth, l = l)
    model_num <- CkCorrModel(corrmodel)
  }
#########################################################################
  if (corrmodel == "Matern") {
    CC <- as.double(param['sill'])
    N <- 1
    a  <- as.double(param['scale'])
    nu1 <- as.double(param['smooth'])
    other <- as.double(0)
    a0 <- a
    nu0 <- nu1
    parameters <- list("CC" = CC, "a" = a,"nu1" = nu1, "nu2" = 0, "other" = other)
    P <- 1
    vtype <- 0
    parametersg <- list("a" = a0, "nu1" = nu0, other = other, CC = CC)
    model_num <- CkCorrModel(corrmodel)
  }
#########################################################################
  if (corrmodel %in% c("Kummer","Kummer_Matern")) {
    CC <- as.double(param['sill'])
    N <- 1
    a  <- as.double(param['scale'])
    nu1 <- as.double(param['smooth'])
    other <- as.double(param['power2'])
    a0 <- a
    nu0 <- nu1
    parameters <- list("CC" = CC, "a" = a,"nu1" = nu1,"nu2" = 0, "other" = other)
    P <- 1
    vtype <- 0
    parametersg <- list("a" = a0, "nu1" = nu0, other = other, CC = CC)
    model_num <- CkCorrModel(corrmodel)
  }
######################################################################### 
  u <- frequency_sampler(L, parametersg, corrmodel, d, condition)
  #######
  phi <- runif(L, 0, (2*pi))
  amp <- sqrt(-2 * log(runif(L)))
  
  Nlocs <- nrow(coord)
  simu <- NULL

  ## The worker function is defined at file scope and all required values
  ## are passed explicitly.  This avoids both large captured environments
  ## and manual parent-environment manipulation.

  if (!is.null(u)) {

   ## GeoSimapprox decides whether the TB call should use spatial-chunk
## parallelism before entering tbm2d_uni().  Therefore `parallel=TRUE`
## here already means that the inner TB path was selected (automatically
## or via GeoModels.force_tb_inner_parallel).  Do not apply a second,
## inconsistent workload threshold at this lower level.
do_parallel <- isTRUE(parallel)

    if (do_parallel) {

      n.workers <- .GeoResolveWorkers(TRUE, ncores, n_jobs = Nlocs)

      ## One chunk per worker (balanced), instead of fixed-size 1e5 chunks.
      ## This function assumes a `future` plan has ALREADY been set by the
      ## caller (e.g. once, at the top of GeoSimapprox) -- it does not open
      ## or close any plan itself, to avoid repeated worker start/stop
      ## overhead and nested-parallelism issues when this is called many
      ## times (e.g. once per replicate).
      limits <- unique(round(seq(0, Nlocs, length.out = n.workers + 1)))
      chunk_list <- lapply(seq_len(length(limits) - 1), function(k)
        (limits[k] + 1):limits[k + 1])
      coord_chunks <- lapply(
        chunk_list,
        function(indexes) coord[indexes, , drop = FALSE]
      )

      simu <- matrix(NA, Nlocs, 1)
      chunk_results <- future.apply::future_lapply(
        coord_chunks,
        FUN = .tb_call_kernel_chunk,
        u = u,
        phi = phi,
        amp = amp,
        L = L,
        CC = CC,
        future.seed = TRUE,
        future.packages = c("GeoModels", "dotCall64"),
        future.globals = FALSE
      )

      for (k in seq_along(chunk_list)) {
        simu[chunk_list[[k]], 1] <- chunk_results[[k]]
      }

    } else {
      simu <- matrix(.tb_call_kernel_chunk(coord, u, phi, amp, L, CC), Nlocs, 1)
    }
  }

  return(list(simu=simu, freqs=u, parametersg=parametersg))
}
#########################################################
#########################################################
#########################################################
frequency_sampler <- function(L, parametersg, corrmodel, d, condition) {
  
  ini.sample <- NULL
  #########################################################
  if (corrmodel == "Matern") {
    spl <- rgamma(L, parametersg$nu1, scale = 1)
    G <- matrix(spl, nrow = L, ncol = d)
    N <- matrix(rnorm(L * d), nrow = L, ncol = d)
    ini.sample <- (N / sqrt(G*2)) / parametersg$a
    ini.sample <- ini.sample / (2*pi)
  }
  ######################################################### 
  if (corrmodel %in% c("Kummer", "Kummer_Matern")) {
    nu    <- as.numeric(parametersg$nu1)
    alpha <- as.numeric(parametersg$other)

    if (corrmodel == "Kummer") {
      betaK <- as.numeric(parametersg$a)
    } else {
      betaK <- as.numeric(parametersg$a) * sqrt(2 * (alpha + 1))
    }
    spl <- rf(L, df1 = 2*nu, df2 = 2*alpha)
    G <- matrix( (nu / alpha) * spl, nrow = L, ncol = d)
    G <- pmax(G, .Machine$double.eps)
    #G <- matrix( (nu / alpha) * rf(L * d, df1 = 2*nu, df2 = 2*alpha), nrow = L, ncol = d)
    N <- matrix(rnorm(L * d), nrow = L, ncol = d)
    ini.sample <- (N / sqrt(G)) / betaK
    ini.sample <- ini.sample / (2*pi)
  }
#########################################################
  if (corrmodel %in% c("GenWend","GenWend_Matern","Genwend","Genwend_Matern",
                       "Hypergeometric","Hypergeometric_Matern")) {
    
    kappa <- as.numeric(parametersg$kappa)
    mu    <- as.numeric(parametersg$mu)
    l     <- as.numeric(parametersg$l)
    
    if (corrmodel %in% c("GenWend","Genwend","Hypergeometric")) {
      a <- as.numeric(parametersg$a)
    } else if (corrmodel %in% c("GenWend_Matern","Genwend_Matern")) {
      a <- as.numeric(parametersg$a) * 
           (gamma(mu + 2*kappa + 1) / gamma(mu))^(1 / (1 + 2*kappa))
    } else if (corrmodel == "Hypergeometric_Matern") {
      num <- (2*kappa + 1) * gamma((mu + 1)/2 + kappa) * 
             gamma((mu + d + 1)/2 + 2*kappa)
      den <- gamma(mu/2) * gamma((mu + d)/2 + kappa)
      a <- as.numeric(parametersg$a) * (num / den)^(1 / (1 + 2*kappa))
    }
    
    if (condition) {
      #message("==> Using Beta-mixture method")
      ini.sample <- simulate_GH_beta(L, d, kappa, mu, l, a)
      ini.sample <- ini.sample / (2 * pi)
    } else {
      #message("==> Using Gasper mixture method")
      ini.sample <- simulate_GH_gasper(L, d, kappa, mu, l, a, nmax = 15)
      ini.sample <- ini.sample / (2 * pi)
    }
  }
  
  return(ini.sample)
}

GeoVariogramDir <- function(data, coordx, coordy = NULL, coordz = NULL,
                            directions = c(0, 45, 90, 135),
                            tolerance = 22.5,
                            numbins = 13,
                            maxdist = NULL,
                            neighb = NULL,
                            distance = "Eucl",
                            subsample = 1) {
  # ---------------- Checks ----------------
  if (!is.numeric(subsample) || length(subsample) != 1L || !is.finite(subsample) ||
      subsample <= 0 || subsample > 1) {
    stop("subsample must be a single numeric value in (0,1].")
  }

  if (!is.numeric(tolerance) || length(tolerance) != 1L || tolerance <= 0 || tolerance > 90)
    stop("tolerance must be in (0,90].")
  if (!is.numeric(numbins) || length(numbins) != 1L || numbins < 2 || numbins %% 1 != 0)
    stop("numbins must be an integer >= 2.")

  # ---------------- Build coords ----------------
  is_matrix <- is.matrix(coordx) || is.data.frame(coordx)

  if (is_matrix) {
    coords <- as.matrix(coordx)
    if (!(ncol(coords) %in% c(2,3))) stop("coordx must have 2 or 3 columns if matrix/data.frame.")
  } else {
    if (is.null(coordy)) stop("If coordx is a vector, coordy must be provided.")
    coords <- if (is.null(coordz)) cbind(coordx, coordy) else cbind(coordx, coordy, coordz)
  }

  coords <- as.matrix(coords)
  storage.mode(coords) <- "double"
  distance <- match.arg(distance, c("Eucl", "Chor", "Geod"))
  .GeoValidate3D(coords, distance = distance,
                 context = "GeoVariogramDir")

  n0 <- nrow(coords)
  if (length(data) != n0) stop("length(data) must match number of coordinate rows.")

  # ---------------- EARLY spatial subsample ----------------
  if (subsample < 1) {
    m <- max(2L, floor(subsample * n0))
    m <- min(m, n0)
    idx0 <- sort(sample.int(n0, m, replace = FALSE))

    coords <- coords[idx0, , drop = FALSE]
    data   <- data[idx0]
  }

  # update n after subsample
  n <- length(data)

  # ---------------- Defaults ----------------
  if (is.null(maxdist)) maxdist <- Inf

  # Smart neighb selection based on n and maxdist
  if (is.null(neighb)) {
    if (is.finite(maxdist)) {
      if (ncol(coords) == 2) {
        area_total <- diff(range(coords[,1])) * diff(range(coords[,2]))
        if (!is.finite(area_total) || area_total <= 0) stop("Degenerate coordinate extent after subsample.")
        density <- n / area_total
        area_circle <- pi * maxdist^2
        expected_neighb <- density * area_circle
        neighb <- min(n - 1L, max(100L, as.integer(expected_neighb * 2)))
      } else {
        volume_total <- diff(range(coords[,1])) * diff(range(coords[,2])) * diff(range(coords[,3]))
        if (!is.finite(volume_total) || volume_total <= 0) stop("Degenerate coordinate extent after subsample.")
        density <- n / volume_total
        volume_sphere <- (4/3) * pi * maxdist^3
        expected_neighb <- density * volume_sphere
        neighb <- min(n - 1L, max(100L, as.integer(expected_neighb * 2)))
      }
    } else {
      neighb <- n - 1L
    }
  } else {
    if (!is.numeric(neighb) || length(neighb) != 1L || neighb < 1 || neighb %% 1 != 0)
      stop("neighb must be a positive integer.")
    neighb <- min(as.integer(neighb), n - 1L)
  }

  # Adaptive chunking based on neighb size
  if (neighb < 500) {
    chunk_size <- n
  } else if (neighb < 5000) {
    chunk_size <- max(500L, min(n, as.integer(1000e6 / (neighb * 24))))
  } else {
    chunk_size <- max(100L, min(n, as.integer(500e6 / (neighb * 24))))
  }

  # ---------------- Distance bins ----------------
  if (is.finite(maxdist)) {
    maxdist_bin <- maxdist
  } else {
    if (n > 10000) {
      samp <- sample(n, min(1000, n))
      dists_samp <- as.matrix(dist(coords[samp, ]))
      maxdist_bin <- as.numeric(quantile(dists_samp[upper.tri(dists_samp)], 0.95, na.rm = TRUE))
    } else {
      coordinate_ranges <- apply(coords, 2L, range, na.rm = TRUE)
      maxdist_bin <- sqrt(sum((coordinate_ranges[2L, ] -
                               coordinate_ranges[1L, ])^2))
    }
  }

  breaks   <- seq(0, maxdist_bin, length.out = numbins + 1)
  widths   <- diff(breaks)
  centers  <- breaks[-1] - widths/2
  numvario <- numbins
  mm       <- c(breaks[1L], breaks[length(breaks)])

  # ---------------- Directional ranges ----------------
  ang_min <- (directions - tolerance) %% 180
  ang_max <- (directions + tolerance) %% 180
  ndir <- length(directions)

  sums_dir <- replicate(ndir, numeric(numvario), simplify = FALSE)
  cnts_dir <- replicate(ndir, integer(numvario), simplify = FALSE)

  # Helper function: C binning
  .bin_pairs_c <- function(dist_sub, i_sub, j_sub) {
    np <- length(dist_sub)
    if (np == 0L) return(list(moms = numeric(numvario), lbins = integer(numvario)))

    zi <- data[i_sub]
    zj <- data[j_sub]

    binsC  <- double(numbins + 1)
    lbinsC <- integer(numvario)
    momsC  <- double(numvario)

    EV <- dotCall64::.C64(
      "Binned_Variogram2new",
      SIGNATURE = c("double","integer","double","double","double","integer","double","integer","double"),
      INTENT    = c("rw","r","r","r","r","rw","rw","r","r"),
      bins  = binsC,
      np    = as.integer(np),
      data1 = as.double(zi),
      data2 = as.double(zj),
      vdist = as.double(dist_sub),
      lbins = lbinsC,
      moms  = momsC,
      nbins = as.integer(numbins + 1),
      mm    = as.double(mm),
      NAOK = TRUE, PACKAGE = "GeoModels", VERBOSE = 0
    )
    list(moms = EV$moms, lbins = EV$lbins)
  }

  # ---------------- Distance type ----------------
  use_geo <- distance %in% c("Geod", "Chor")
  if (use_geo && !requireNamespace("fields", quietly = TRUE)) {
    stop("Package 'fields' is required by GeoVariogramDir() for geodesic or chordal distance.",
         call. = FALSE)
  }

  proj_sinusoidal <- function(M) {
    lon <- M[,1] * pi / 180
    lat <- M[,2] * pi / 180
    cbind(lon * cos(lat), lat)
  }

  chord_from_arc <- function(theta) 2 * sin(theta / 2)

  coords_knn <- if (use_geo) proj_sinusoidal(coords) else coords

  # ---------------- Chunk processing ----------------
  n_chunks <- ceiling(n / chunk_size)

  for (chunk_id in seq_len(n_chunks)) {
    start_idx <- (chunk_id - 1) * chunk_size + 1
    end_idx <- min(chunk_id * chunk_size, n)
    chunk_indices <- start_idx:end_idx
    n_chunk <- length(chunk_indices)

    query_coords <- coords_knn[chunk_indices, , drop = FALSE]

    if (is.finite(maxdist)) {
      if (use_geo) {
        radius_proj <- maxdist * 1.1
        knn_res <- nabor::knn(coords_knn, query_coords, k = neighb + 1, radius = radius_proj)
      } else {
        knn_res <- nabor::knn(coords_knn, query_coords, k = neighb + 1, radius = maxdist)
      }
    } else {
      knn_res <- nabor::knn(coords_knn, query_coords, k = neighb + 1)
    }

    idx_mat <- knn_res$nn.idx
    dists_knn <- knn_res$nn.dists

    # Compute true distances for geodesic cases
    if (use_geo) {
      n_query <- nrow(query_coords)
      K <- ncol(idx_mat)
      dist_true <- matrix(NA_real_, n_query, K)

      for (r in seq_len(n_query)) {
        idx <- idx_mat[r, ]
        valid <- idx > 0
        if (!any(valid)) next

        global_idx <- chunk_indices[r]
        neighbor_idx <- idx[valid]

        arc <- fields::rdist.earth.vec(
          x1 = matrix(coords[global_idx, 1:2], nrow = 1),
          x2 = coords[neighbor_idx, 1:2, drop = FALSE],
          miles = FALSE, R = 1
        )

        if (distance == "Chor") {
          dist_true[r, which(valid)] <- chord_from_arc(arc)
        } else {
          dist_true[r, which(valid)] <- arc
        }
      }
      dist_mat <- dist_true
    } else {
      dist_mat <- dists_knn
    }

    keep <- idx_mat > 0
    if (is.finite(maxdist)) keep <- keep & (dist_mat <= maxdist)
    if (!any(keep)) next

    row_indices <- matrix(rep(chunk_indices, times = ncol(idx_mat)),
                          nrow = n_chunk, byrow = FALSE)
    i <- as.integer(row_indices[keep])
    j <- as.integer(idx_mat[keep])
    dist <- as.double(dist_mat[keep])

    # Remove self-pairs
    self_pairs <- i == j
    if (any(self_pairs)) {
      i <- i[!self_pairs]
      j <- j[!self_pairs]
      dist <- dist[!self_pairs]
    }
    if (length(i) == 0) next

    dx <- coords[j, 1] - coords[i, 1]
    dy <- coords[j, 2] - coords[i, 2]
    angle <- (atan2(dy, dx) * 180 / pi) %% 180

    mask_dir_mat <- matrix(FALSE, nrow = length(angle), ncol = ndir)
    for (k in seq_along(directions)) {
      if (ang_min[k] < ang_max[k]) {
        mask_dir_mat[,k] <- (angle >= ang_min[k]) & (angle <= ang_max[k])
      } else {
        mask_dir_mat[,k] <- (angle >= ang_min[k]) | (angle <= ang_max[k])
      }
    }

    n_pairs_chunk <- length(i)
    use_c_chunk <- n_pairs_chunk > 5000

    if (use_c_chunk) {
      for (k in seq_along(directions)) {
        mask <- mask_dir_mat[,k]
        if (any(mask)) {
          bp <- .bin_pairs_c(dist[mask], i[mask], j[mask])
          sums_dir[[k]] <- sums_dir[[k]] + bp$moms
          cnts_dir[[k]] <- cnts_dir[[k]] + bp$lbins
        }
      }
    } else {
      diff2 <- (data[j] - data[i])^2 / 2
      bin_idx_all <- cut(dist, breaks, labels = FALSE, include.lowest = TRUE)

      for (k in seq_along(directions)) {
        mask <- mask_dir_mat[,k]
        if (any(mask)) {
          for (b in seq_len(numvario)) {
            bin_mask <- !is.na(bin_idx_all[mask]) & bin_idx_all[mask] == b
            if (any(bin_mask)) {
              vals <- diff2[mask][bin_mask]
              sums_dir[[k]][b] <- sums_dir[[k]][b] + sum(vals, na.rm = TRUE)
              cnts_dir[[k]][b] <- cnts_dir[[k]][b] + length(vals)
            }
          }
        }
      }
    }
  }

  # ---------------- Results ----------------
  result <- vector("list", ndir)
  names(result) <- paste0("dir_", directions, "deg")

  for (k in seq_len(ndir)) {
    cnt <- cnts_dir[[k]]
    gamma <- ifelse(cnt > 0, sums_dir[[k]] / cnt, NA_real_)
    result[[k]] <- list(
      centers = centers,
      gamma = as.numeric(gamma),
      npairs = as.numeric(cnt)
    )
  }

  attr(result, "breaks") <- breaks
  attr(result, "maxdist") <- maxdist_bin
  attr(result, "neighb") <- neighb
  attr(result, "n_chunks") <- n_chunks
  attr(result, "subsample") <- subsample
  attr(result, "direction_definition") <- if (ncol(coords) == 3L) {
    "Azimuth in the xy plane; Euclidean lag distances use x, y, and z."
  } else {
    "Azimuth in the xy plane."
  }
  structure(result, class = "GeoVariogramDir")
}

################################################################################
plot.GeoVariogramDir <- function(x, ...,
                                 main = "Directional Empirical Semivariograms",
                                 pch = 20,
                                 lwd = 1,
                                 col = 1:8,
                                 ylab = "Semivariogram",
                                 xlab = "Distance") {
  directions <- names(x)
  ndir <- length(directions)
  col <- rep(col, length.out = ndir)

  dots <- list(...)
  legend_cex   <- if (!is.null(dots$legend.cex))   dots$legend.cex   else 0.8
  legend_inset <- if (!is.null(dots$legend.inset)) dots$legend.inset else 0.02
  dots$legend.cex <- dots$legend.inset <- NULL

  # limiti default
  all_centers <- unlist(lapply(x, function(vg) vg$centers))
  all_gamma   <- unlist(lapply(x, function(vg) vg$gamma))
  xlim_default <- range(all_centers, na.rm = TRUE)
  ylim_default <- range(all_gamma,   na.rm = TRUE)
  if (!is.finite(ylim_default[1]) || !is.finite(ylim_default[2])) ylim_default <- c(0, 1)

  xlim <- if (!is.null(dots$xlim)) dots$xlim else xlim_default
  ylim <- if (!is.null(dots$ylim)) dots$ylim else ylim_default
  main <- if (!is.null(dots$main)) dots$main else main
  xlab <- if (!is.null(dots$xlab)) dots$xlab else xlab
  ylab <- if (!is.null(dots$ylab)) dots$ylab else ylab
  dots$xlim <- dots$ylim <- dots$main <- dots$xlab <- dots$ylab <- NULL

  # setup margini + xpd per disegnare fuori
  opar <- par(no.readonly = TRUE)
  on.exit(par(opar))
  par(mar = par("mar") + c(0, 0, 0, 4), xpd = NA)  # spazio extra a destra

  do.call(plot, c(list(NA, xlim = xlim, ylim = ylim, xlab = xlab, ylab = ylab, main = main), dots))
  for (i in seq_along(directions)) {
    vg <- x[[i]]
    xvals <- vg$centers
    yvals <- vg$gamma
    if (length(yvals) == 0 || all(is.na(yvals))) next
    lines(xvals, yvals, type = "b", lwd = lwd, pch = pch, col = col[i])
  }

  degs <- as.integer(gsub("dir_(\\d+)deg", "\\1", directions))
  legend_labels <- lapply(degs, function(d) bquote(.(d)*degree))

  usr <- par("usr")
  x_off <- usr[2] + legend_inset * diff(usr[1:2])  # un po' oltre il bordo destro
  y_top <- usr[4]

  legend(x = x_off, y = y_top,
         legend = legend_labels, col = col, lwd = lwd, pch = pch,
         xjust = 0, yjust = 1, bty = "n", cex = legend_cex)
}

####################################################
### File name: GeoLik_original_style_fixed.R
####################################################

### Optim call for log-likelihood maximization
Lik <- function(copula,bivariate,coordx,coordy,coordz,coordt,coordx_dyn,corrmodel,data,fixed,flagcor,flagnuis,grid,lower,
                mdecomp,model,namescorr,namesnuis,namesparam,numcoord,numpairs,numparamcor,numtime,
                optimizer,onlyvar,param,radius,setup,spacetime,sparse,varest,taper,type,upper,ns,X,neighb,MM,aniso,score)
{
  type_code <- suppressWarnings(as.integer(type))
  if (length(type_code) == 1L && !is.na(type_code) && type_code == 3L)
    stop("Likelihood type code 3 is no longer supported.", call. = FALSE)

  #########################################################################
  ## Utilities
  #########################################################################

  combine_param <- function(param, fixed, namesparam) {
    names(param) <- namesparam
    if (!is.null(fixed) && length(fixed) > 0) fixed <- unlist(fixed, use.names = TRUE)
    out <- c(param, fixed)
    out <- as.numeric(out)
    names(out) <- c(namesparam, names(fixed))
    out
  }

  get_mean <- function(X, nuisance, MM = NULL) {
    if (!is.null(MM)) return(as.numeric(MM))
    beta <- .GeoMean_beta(nuisance, ncol(X))
    as.numeric(X %*% beta)
  }

  build_correlation_matrix <- function(corr_vec, ident) {
    corrmat_full <- ident
    lt <- lower.tri(ident)
    corrmat_full[lt] <- corr_vec
    corrmat_full <- t(corrmat_full)
    corrmat_full[lt] <- corr_vec
    corrmat_full
  }

  make_optimize_bounds <- function(param, lower, upper, namesparam) {
    p0 <- as.numeric(param)[1]
    pname <- namesparam[1]

    ll <- if (is.null(lower)) -Inf else as.numeric(lower)[1]
    uu <- if (is.null(upper))  Inf else as.numeric(upper)[1]

    span <- max(10 * (abs(p0) + 1), 10)
    if (startsWith(pname, "mean")) span <- max(100 * (abs(p0) + 1), 100)

    positive_par <- pname %in% c(
      "scale", "scale_s", "scale_t", "sill", "nugget", "power", "power2",
      "shape", "shape1", "shape2", "df", "tail", "tail1", "tail2",
      "min", "max", "rate", "ratio"
    )

    if (!is.finite(ll)) {
      if (positive_par) ll <- .Machine$double.eps else ll <- p0 - span
    }

    if (!is.finite(uu)) {
      if (pname %in% c("scale", "scale_s", "scale_t")) {
        uu <- 4
      } else if (pname == "nugget") {
        uu <- 1 - .Machine$double.eps
      } else if (pname == "angle") {
        uu <- pi
      } else {
        uu <- p0 + span
      }
    }

    if (p0 <= ll) ll <- min(.Machine$double.eps, p0 - span)
    if (p0 >= uu) uu <- p0 + span

    if (!is.finite(ll) || !is.finite(uu) || ll >= uu) {
      stop("invalid lower/upper bounds for optimize", call. = FALSE)
    }

    list(lower = ll, upper = uu)
  }

  make_bobyqa_setup <- function(param, lower, upper, namesparam) {
    p0 <- as.numeric(param)
    ll <- as.numeric(lower)
    uu <- as.numeric(upper)

    if (length(ll) != length(p0) || length(uu) != length(p0)) {
      stop("lower and upper must have the same length as param for bobyqa", call. = FALSE)
    }

    names(ll) <- namesparam
    names(uu) <- namesparam

    span <- pmax(10 * (abs(p0) + 1), 10)
    mean_par <- startsWith(namesparam, "mean")
    span[mean_par] <- pmax(100 * (abs(p0[mean_par]) + 1), 100)

    idx <- is.infinite(ll) & ll < 0
    ll[idx] <- p0[idx] - span[idx]

    idx <- is.infinite(uu) & uu > 0
    uu[idx] <- p0[idx] + span[idx]

    ## standard finite fallbacks for common constrained parameters
    idx <- !is.finite(ll) & namesparam %in% c("scale", "scale_s", "scale_t", "sill", "power", "power2", "shape", "tail", "tail1", "tail2", "ratio")
    ll[idx] <- .Machine$double.eps

    idx <- !is.finite(uu) & namesparam %in% c("scale", "scale_s", "scale_t")
    uu[idx] <- 4
    idx <- !is.finite(uu) & namesparam == "nugget"
    uu[idx] <- 1 - .Machine$double.eps
    idx <- !is.finite(uu) & namesparam == "angle"
    uu[idx] <- pi

    if (any(!is.finite(ll)) || any(!is.finite(uu))) {
      stop("bobyqa requires finite lower and upper bounds", call. = FALSE)
    }

    idx <- p0 <= ll
    if (any(idx)) ll[idx] <- p0[idx] - span[idx]
    idx <- p0 >= uu
    if (any(idx)) uu[idx] <- p0[idx] + span[idx]

    if (any(uu <= ll)) {
      stop("bobyqa requires upper > lower for all parameters", call. = FALSE)
    }

    width <- uu - ll
    finite_width <- width[is.finite(width) & width > 0]
    if (length(finite_width) > 0) {
      rhobeg <- min(0.1 * min(finite_width), 0.2 * max(abs(p0), 1))
    } else {
      rhobeg <- 0.2 * max(abs(p0), 1)
    }
    rhobeg <- max(rhobeg, .Machine$double.eps^0.25)
    rhoend <- rhobeg * 1e-6
    npar <- length(p0)

    list(
      lower = ll,
      upper = uu,
      control = list(
        maxfun = 100000L,
        rhobeg = rhobeg,
        rhoend = rhoend,
        npt = min(2 * npar + 1, (npar + 1) * (npar + 2) / 2),
        iprint = 0
      )
    )
  }

  #########################################################################
  ## C calls for correlations
  #########################################################################

  matr <- function(corrmat,corr,coordx,coordy,coordz,coordt,corrmodel,nuisance,paramcorr,ns,NS,radius)
  {
    cc <- dotCall64::.C64(
      as.character(corrmat),
      SIGNATURE = c(rep("double",5),"integer","double","double","double","integer","integer"),
      cr = dotCall64::vector_dc("double", length(corr)),
      coordx, coordy, coordz, coordt, corrmodel, nuisance, paramcorr, radius, ns, NS,
      INTENT = c("w", rep("r", 10)),
      PACKAGE = 'GeoModels', VERBOSE = 0, NAOK = TRUE
    )$cr
    cc
  }

  matr2 <- function(corrmat,corr,coordx,coordy,coordz,coordt,corrmodel,nuisance,paramcorr,ns,NS,radius,model,mu)
  {
    cc <- dotCall64::.C64(
      as.character(corrmat),
      SIGNATURE = c(rep("double",5),"integer","double","integer","double","double","double","integer","integer","integer"),
      cr = dotCall64::vector_dc("double", length(corr)),
      coordx, coordy, coordz, coordt, corrmodel, c(mu), ns, nuisance, paramcorr, radius, ns, NS, model,
      INTENT = c("rw", rep("r", 13)),
      PACKAGE = 'GeoModels', VERBOSE = 0, NAOK = TRUE
    )$cr
    cc
  }

  #########################################################################
  ## Densities / criteria
  #########################################################################

  LogNormDenStand <- function(const, cova, dimat, mdecomp, stdata, ...) {
    llik <- 1.0e8
    decompvarcov <- MatDecomp(cova, mdecomp)
    if (is.logical(decompvarcov) && !decompvarcov) return(llik)
    logdetvarcov <- MatLogDet(decompvarcov, mdecomp)
    if (is.na(logdetvarcov)) return(llik)
    y <- try(forwardsolve(decompvarcov, stdata, transpose = FALSE), silent = TRUE)
    if (inherits(y, "try-error")) return(llik)
    quad <- sum(y^2)
    if (!is.finite(quad)) return(llik)
    0.5 * (const + logdetvarcov + quad)
  }

  LogNormDenStand_spam <- function(const, cova, dimat, mdecomp, stdata, ...) {
    llik <- 1.0e8
    mcov <- try(spam::as.spam(cova), silent = TRUE)
    if (inherits(mcov, "try-error")) return(llik)
    cholS <- try(spam::chol.spam(mcov), silent = TRUE)
    if (inherits(cholS, "try-error")) return(llik)
    llik <- 0.5 * (const + 2 * c(spam::determinant.spam.chol.NgPeyton(cholS)$modulus) +
                     sum(stdata * spam::solve.spam(cholS, stdata)))
    as.numeric(llik)
  }

  LogNormDenStand22 <- function(const,cova,ident,dimat,mdecomp,nuisance,setup,stdata) {
    llik <- 1.0e8
    decompvarcov <- MatDecomp(cova, mdecomp)
    if (is.logical(decompvarcov) && !decompvarcov) return(llik)
    logdetvarcov <- MatLogDet(decompvarcov, mdecomp)
    if (is.na(logdetvarcov)) return(llik)
    y <- try(forwardsolve(decompvarcov, stdata, transpose = FALSE), silent = TRUE)
    if (inherits(y, "try-error")) return(llik)
    llik <- 0.5 * (const + logdetvarcov + sum(y^2))
    as.numeric(llik)
  }

  LogNormDenStand_biv <- function(const,cova,ident,dimat,mdecomp,nuisance,setup,stdata) {
    llik <- 1.0e8
    ident[lower.tri(ident, diag = TRUE)] <- cova
    ident <- t(ident)
    ident[lower.tri(ident, diag = TRUE)] <- cova
    decompvarcov <- MatDecomp(ident, mdecomp)
    if (is.logical(decompvarcov) && !decompvarcov) return(llik)
    logdetvarcov <- MatLogDet(decompvarcov, mdecomp)
    if (is.na(logdetvarcov)) return(llik)
    y <- try(forwardsolve(decompvarcov, stdata, transpose = FALSE), silent = TRUE)
    if (inherits(y, "try-error")) return(llik)
    llik <- 0.5 * (const + logdetvarcov + sum(y^2))
    as.numeric(llik)
  }

  LogNormDenStand_biv_spam <- function(const,cova,ident,dimat,mdecomp,nuisance,setup,stdata) {
    llik <- 1.0e8
    ident[lower.tri(ident, diag = TRUE)] <- cova
    ident <- t(ident)
    ident[lower.tri(ident, diag = TRUE)] <- cova
    mcov <- try(spam::as.spam(ident), silent = TRUE)
    if (inherits(mcov, "try-error")) return(llik)
    cholS <- try(chol(mcov), silent = TRUE)
    if (inherits(cholS, "try-error")) return(llik)
    llik <- 0.5 * (const + 2 * c(spam::determinant.spam.chol.NgPeyton(cholS)$modulus) +
                     sum(stdata * spam::solve.spam(cholS, stdata)))
    as.numeric(llik)
  }

  CVV <- function(const,cova,dimat,mdecomp,stdata,...) {
    llik <- 1.0e8
    decompvarcov <- MatDecomp(cova, mdecomp)
    if (is.logical(decompvarcov) && !decompvarcov) return(llik)
    invarcov <- MatInv(cova)
    D <- diag(1 / diag(invarcov))
    M <- crossprod(invarcov, D)
    C <- tcrossprod(M, M)
    as.numeric(mean(crossprod(t(crossprod(stdata, C)), stdata)))
  }

  CVV_biv <- function(const,cova,ident,dimat,mdecomp,nuisance,setup,stdata) {
    llik <- 1.0e8
    ident[lower.tri(ident, diag = TRUE)] <- cova
    ident <- t(ident)
    ident[lower.tri(ident, diag = TRUE)] <- cova
    decompvarcov <- MatDecomp(ident, mdecomp)
    if (is.logical(decompvarcov) && !decompvarcov) return(llik)
    invarcov <- MatInv(ident)
    D <- diag(1 / diag(invarcov))
    M <- crossprod(invarcov, D)
    C <- tcrossprod(M, M)
    as.numeric(mean(crossprod(t(crossprod(stdata, C)), stdata)))
  }

  ## Tapering functions are kept for compatibility with type 5/6.
  LogNormDenTap <- function(const,cova,dimat,mdecomp,stdata,setup,...) {
    lliktap <- 1.0e8
    varcovtap <- try(new("spam", entries = cova * setup$taps, colindices = setup$ja,
                         rowpointers = setup$ia, dimension = as.integer(rep(dimat, 2))), silent = TRUE)
    if (inherits(varcovtap, "try-error")) return(lliktap)
    cholvctap <- try(spam::update.spam.chol.NgPeyton(setup$struct, varcovtap), silent = TRUE)
    if (inherits(cholvctap, "try-error")) return(lliktap)
    logdet <- c(spam::determinant(cholvctap)$modulus)
    inv <- spam::solve.spam(cholvctap)
    slot(varcovtap, "entries") <- inv[setup$idx] * setup$taps
    as.numeric(0.5 * (const + 2 * logdet + drop(t(stdata) %*% varcovtap %*% stdata)))
  }

  LogNormDenTap1 <- function(const,cova,dimat,mdecomp,stdata,setup,...) {
    lliktap <- 1.0e8
    varcovtap <- try(new("spam", entries = cova * setup$taps, colindices = setup$ja,
                         rowpointers = setup$ia, dimension = as.integer(rep(dimat, 2))), silent = TRUE)
    if (inherits(varcovtap, "try-error")) return(lliktap)
    cholvctap <- try(spam::update.spam.chol.NgPeyton(setup$struct, varcovtap), silent = TRUE)
    if (inherits(cholvctap, "try-error")) return(lliktap)
    logdet <- c(spam::determinant(cholvctap)$modulus)
    as.numeric(0.5 * (const + 2 * logdet + sum(stdata * spam::solve.spam(cholvctap, stdata))))
  }

  LogNormDenTap_biv <- function(const,cova,ident,dimat,mdecomp,nuisance,setup,stdata) {
    lliktap <- 1.0e8
    varcovtap <- try(new("spam", entries = cova * setup$taps, colindices = setup$ja,
                         rowpointers = setup$ia, dimension = as.integer(rep(dimat, 2))), silent = TRUE)
    if (inherits(varcovtap, "try-error")) return(lliktap)
    cholvctap <- try(spam::chol.spam(varcovtap), silent = TRUE)
    if (inherits(cholvctap, "try-error")) return(lliktap)
    logdet <- c(spam::determinant(cholvctap)$modulus)
    inv <- spam::solve.spam(cholvctap)
    slot(varcovtap, "entries") <- inv[setup$idx] * setup$taps
    as.numeric(0.5 * (const + 2 * logdet + drop(t(stdata) %*% varcovtap %*% stdata)))
  }

  LogNormDenTap1_biv <- function(const,cova,ident,dimat,mdecomp,nuisance,setup,stdata) {
    lliktap <- 1.0e8
    cova[cova == nuisance['sill']] <- nuisance['sill'] + nuisance['nugget']
    varcovtap <- try(new("spam", entries = cova * setup$taps, colindices = setup$ja,
                         rowpointers = setup$ia, dimension = as.integer(rep(dimat, 2))), silent = TRUE)
    if (inherits(varcovtap, "try-error")) return(lliktap)
    cholvctap <- try(spam::update.spam.chol.NgPeyton(setup$struct, varcovtap), silent = TRUE)
    if (inherits(cholvctap, "try-error")) return(lliktap)
    logdet <- c(spam::determinant.spam.chol.NgPeyton(cholvctap)$modulus)
    as.numeric(0.5 * (const + 2 * logdet + sum(stdata * spam::solve.spam(cholvctap, stdata))))
  }

  LogshDenTap1 <- function(const,cova,dimat,mdecomp,nuisance,setup,sill,stdata,...) {
    lliktap <- 1.0e8
    varcovtap <- try(new("spam", entries = setup$taps * cova, colindices = setup$ja,
                         rowpointers = setup$ia, dimension = as.integer(rep(dimat, 2))), silent = TRUE)
    if (inherits(varcovtap, "try-error")) return(lliktap)
    cholvctap <- try(spam::update.spam.chol.NgPeyton(setup$struct, varcovtap), silent = TRUE)
    if (inherits(cholvctap, "try-error")) return(lliktap)
    logdet <- 2 * c(spam::determinant(cholvctap)$modulus)
    skew <- as.numeric(nuisance["skew"])
    delta <- as.numeric(nuisance["tail"])
    Z <- sinh(delta * asinh(stdata) - skew)
    C <- delta * sqrt((1 + Z^2) / (stdata^2 + 1))
    as.numeric(0.5 * (const + const * log(sill) / log(2*pi) + logdet - 2 * sum(log(C)) +
                        sum(Z * spam::solve.spam(cholvctap, Z))))
  }

  #########################################################################
  ## Objective functions
  #########################################################################

  loglik <- function(param, const, coordx, coordy, coordz, coordt, corr, corrmat, corrmodel, data, dimat, fixed, fname,
                     grid, ident, mdecomp, model, namescorr, namesnuis, namesparam, radius, setup, X, ns, NS, MM, aniso, namesaniso) {
    llik <- 1.0e8
    pram <- combine_param(param, fixed, namesparam)
    paramcorr <- as.numeric(pram[namescorr])
    nuisance <- pram[namesnuis]
    Mean <- get_mean(X, nuisance, MM)

    if (aniso) {
      anisopar <- pram[namesaniso]
      coords1 <- GeoAniso(cbind(coordx, coordy, coordz), anisopars = as.numeric(anisopar))
      coordx <- coords1[, 1]
      coordy <- coords1[, 2]
      coordz <- if (ncol(coords1) == 3) coords1[, 3] else NULL
    }

    sill <- as.numeric(nuisance['sill'])
    nugget <- as.numeric(nuisance['nugget'])
    if (!is.finite(sill) || !is.finite(nugget) || sill <= 0 || nugget < 0 || nugget >= 1) return(llik)

    if (fname %in% c("LogNormDenTap", "LogNormDenTap1")) {
      corrv <- matr(corrmat, corr, coordx, coordy, coordz, coordt, corrmodel, nuisance, paramcorr, ns, NS, radius)
      if (length(corrv) == 0 || any(!is.finite(corrv))) return(llik)
      cova_vec <- corrv * sill * (1 - nugget)
      return(do.call(what = fname, args = list(stdata = data - c(Mean), const = const, cova = cova_vec,
                                               dimat = dimat, mdecomp = mdecomp, setup = setup,
                                               nuisance = nuisance, sill = sill)))
    }

    corrv <- matr(corrmat, corr, coordx, coordy, coordz, coordt, corrmodel, nuisance, paramcorr, ns, NS, radius)
    if (length(corrv) == 0 || any(!is.finite(corrv))) return(llik)
    corrmat_full <- build_correlation_matrix(corrv, ident)
    cova <- sill * ((1 - nugget) * corrmat_full + nugget * ident)
    stdata <- data - c(Mean)

    do.call(what = fname, args = list(stdata = stdata, const = const, cova = cova,
                                      dimat = dimat, mdecomp = mdecomp, setup = setup))
  }

  loglik_biv <- function(param,const,coordx,coordy,coordz,coordt,corr,corrmat,corrmodel,data,dimat,fixed,fname,
                         grid,ident,mdecomp,model,namescorr,namesnuis,namesparam,radius,setup,X,ns,NS,MM,aniso,namesaniso) {
    pram <- combine_param(param, fixed, namesparam)
    paramcorr <- as.numeric(pram[namescorr])
    nuisance <- pram[namesnuis]
    sel1 <- substr(names(nuisance), 1, 6) == "mean_1"
    sel2 <- substr(names(nuisance), 1, 6) == "mean_2"
    mm1 <- as.numeric(nuisance[sel1])
    mm2 <- as.numeric(nuisance[sel2])
    X1 <- as.matrix(X[1:ns[1], ])
    X2 <- as.matrix(X[(ns[1] + 1):(ns[2] + ns[1]), ])
    Mean <- as.double(c(X1 %*% mm1, X2 %*% mm2))
    if (!is.null(MM)) Mean <- as.numeric(MM)
    stdata <- data - Mean
    corrv <- matr(corrmat, corr, coordx, coordy, coordz, coordt, corrmodel, nuisance, paramcorr, ns, NS, radius)
    do.call(what = fname, args = list(stdata = stdata, const = const, cova = corrv, ident = ident,
                                      dimat = dimat, mdecomp = mdecomp, nuisance = nuisance, setup = setup))
  }

  loglik_miss_T <- function(param,const,coordx,coordy,coordz,coordt,corr,corrmat,corrmodel,data,dimat,fixed,fname,
                            grid,ident,mdecomp,model,namescorr,namesnuis,namesparam,radius,setup,X,ns,NS,MM,aniso,namesaniso) {
    llik <- 1.0e8
    pram <- combine_param(param, fixed, namesparam)
    paramcorr <- as.numeric(pram[namescorr])
    nuisance <- pram[namesnuis]
    Mean <- get_mean(X, nuisance, MM)

    if (aniso) {
      coords1 <- GeoAniso(cbind(coordx, coordy, coordz), anisopars = as.numeric(pram[namesaniso]))
      coordx <- coords1[,1]; coordy <- coords1[,2]; coordz <- if (ncol(coords1) == 3) coords1[,3] else NULL
    }

    corrv <- matr(corrmat, corr, coordx, coordy, coordz, coordt, corrmodel, nuisance, paramcorr, ns, NS, radius)
    df <- 1 / as.numeric(nuisance['df'])
    sill <- as.numeric(nuisance['sill'])
    nugget <- as.numeric(nuisance['nugget'])
    if (!is.finite(df) || df <= 2 || sill <= 0 || nugget < 0 || nugget >= 1) return(llik)

    corrv <- try(exp(log(df - 2) + 2 * lgamma(0.5 * (df - 1)) -
                       (log(2) + 2 * lgamma(df / 2)) +
                       log(Re(hypergeo::hypergeo(0.5, 0.5, df / 2, corrv^2))) + log(abs(corrv))) * sign(corrv),
                 silent = TRUE)
    if (inherits(corrv, "try-error") || any(!is.finite(corrv))) return(llik)
    corrmat_full <- build_correlation_matrix(corrv, ident)
    cova <- sill * ((1 - nugget) * corrmat_full + nugget * ident)
    LogNormDenStand(stdata = data - c(Mean), const = const, cova = cova, dimat = dimat, mdecomp = mdecomp)
  }

  loglik_miss_skewT <- function(param,const,coordx,coordy,coordz,coordt,corr,corrmat,corrmodel,data,dimat,fixed,fname,
                                grid,ident,mdecomp,model,namescorr,namesnuis,namesparam,radius,setup,X,ns,NS,MM,aniso,namesaniso) {
    llik <- 1.0e8
    pram <- combine_param(param, fixed, namesparam)
    paramcorr <- as.numeric(pram[namescorr])
    nuisance <- pram[namesnuis]
    Mean <- get_mean(X, nuisance, MM)

    if (aniso) {
      coords1 <- GeoAniso(cbind(coordx, coordy, coordz), anisopars = as.numeric(pram[namesaniso]))
      coordx <- coords1[,1]; coordy <- coords1[,2]; coordz <- if (ncol(coords1) == 3) coords1[,3] else NULL
    }

    corrv <- matr(corrmat, corr, coordx, coordy, coordz, coordt, corrmodel, nuisance, paramcorr, ns, NS, radius)
    nu <- 1 / as.numeric(nuisance['df'])
    skew <- as.numeric(nuisance['skew'])
    sill <- as.numeric(nuisance['sill'])
    nugget <- as.numeric(nuisance['nugget'])
    if (!is.finite(nu) || nu <= 2 || abs(skew) >= 1 || sill <= 0 || nugget < 0 || nugget >= 1) return(llik)

    skew2 <- skew^2
    w <- sqrt(1 - skew2)
    KK <- 2 * skew2 / pi
    D1 <- (nu - 1) / 2
    D2 <- nu / 2
    CorSkew <- (2 * skew2 / (pi * w^2 + skew2 * (pi - 2))) *
      (sqrt(1 - corrv^2) + corrv * asin(corrv) - 1) +
      w^2 * corrv / (w^2 + skew2 * (1 - 2 / pi))
    corr3 <- (pi * (nu - 2) * gamma(D1)^2 /
                (2 * (pi * gamma(D2)^2 - skew2 * (nu - 2) * gamma(D1)^2))) *
      (Re(hypergeo::hypergeo(0.5, 0.5, D2, corrv^2)) * ((1 - KK) * CorSkew + KK) - KK)
    if (any(!is.finite(corr3))) return(llik)
    corrmat_full <- build_correlation_matrix(corr3, ident)
    cova <- sill * ((1 - nugget) * corrmat_full + nugget * ident)
    LogNormDenStand(stdata = data - c(Mean), const = const, cova = cova, dimat = dimat, mdecomp = mdecomp)
  }

  loglik_miss_Poisgamma <- function(param,const,coordx,coordy,coordz,coordt,corr,corrmat,corrmodel,data,dimat,fixed,fname,
                                    grid,ident,mdecomp,model,namescorr,namesnuis,namesparam,radius,setup,X,ns,NS,MM,aniso,namesaniso) {
    llik <- 1.0e8
    pram <- combine_param(param, fixed, namesparam)
    paramcorr <- as.numeric(pram[namescorr])
    nuisance0 <- pram[namesnuis]
    Mean <- get_mean(X, nuisance0, MM)
    sel <- names(nuisance0) %in% .GeoMean_names(ncol(X))
    nuisance <- nuisance0[!sel]
    nuisance <- nuisance[order(names(nuisance))]

    if (aniso) {
      coords1 <- GeoAniso(cbind(coordx, coordy, coordz), anisopars = as.numeric(pram[namesaniso]))
      coordx <- coords1[,1]; coordy <- coords1[,2]; coordz <- if (ncol(coords1) == 3) coords1[,3] else NULL
    }

    if (as.numeric(nuisance['nugget']) < 0 || as.numeric(nuisance['nugget']) > 1) return(llik)
    mu <- Mean
    corrv <- matr2(corrmat, corr, coordx, coordy, coordz, coordt, corrmodel, nuisance, paramcorr, ns, NS, radius, 46, mu)
    cova <- ident
    cova[lower.tri(cova)] <- corrv
    cova <- t(cova)
    cova[lower.tri(cova)] <- corrv
    ff <- exp(mu)
    cova[cbind(seq_len(dimat), seq_len(dimat))] <- ff * (1 + ff / as.numeric(nuisance["shape"]))
    LogNormDenStand22(stdata = data - c(ff), const = const, cova = cova, dimat = dimat,
                      ident = ident, mdecomp = mdecomp, nuisance = nuisance, setup = setup)
  }

  loglik_miss_Pois <- function(param,const,coordx,coordy,coordz,coordt,corr,corrmat,corrmodel,data,dimat,fixed,fname,
                               grid,ident,mdecomp,model,namescorr,namesnuis,namesparam,radius,setup,X,ns,NS,MM,aniso,namesaniso) {
    llik <- 1.0e8
    pram <- combine_param(param, fixed, namesparam)
    paramcorr <- as.numeric(pram[namescorr])
    nuisance0 <- pram[namesnuis]
    Mean <- get_mean(X, nuisance0, MM)
    sel <- names(nuisance0) %in% .GeoMean_names(ncol(X))
    nuisance <- nuisance0[!sel]

    if (aniso) {
      coords1 <- GeoAniso(cbind(coordx, coordy, coordz), anisopars = as.numeric(pram[namesaniso]))
      coordx <- coords1[,1]; coordy <- coords1[,2]; coordz <- if (ncol(coords1) == 3) coords1[,3] else NULL
    }

    if (as.numeric(nuisance['nugget']) < 0 || as.numeric(nuisance['nugget']) > 1) return(llik)
    mu <- Mean
    corrv <- matr2(corrmat, corr, coordx, coordy, coordz, coordt, corrmodel, nuisance, paramcorr, ns, NS, radius, 30, mu)
    cova <- ident
    cova[lower.tri(cova)] <- corrv
    cova <- t(cova)
    cova[lower.tri(cova)] <- corrv
    diag(cova) <- exp(mu)
    LogNormDenStand22(stdata = data - c(exp(mu)), const = const, cova = cova, dimat = dimat,
                      ident = ident, mdecomp = mdecomp, nuisance = nuisance, setup = setup)
  }

  LogNormDenStand_Tukey2H <- function(const, cova, dimat, mdecomp, nuisance, sill, setup, stdata) {
    llik <- 1.0e8
    delta1 <- as.numeric(nuisance["tail1"])
    delta2 <- as.numeric(nuisance["tail2"])
    if (delta1 < 0 || delta1 >= 0.5 || delta2 < 0 || delta2 >= 0.5 || sill <= 0) return(llik)

    decompvarcov <- MatDecomp(cova, mdecomp)
    if (is.logical(decompvarcov) && !decompvarcov) return(llik)
    logdetvarcov <- MatLogDet(decompvarcov, mdecomp)
    if (is.na(logdetvarcov)) return(llik)

    pos_idx <- which(stdata >= 0)
    neg_idx <- which(stdata < 0)
    tau_inv <- numeric(dimat)
    log_jacobian <- 0

    if (length(pos_idx) > 0) {
      st_pos <- stdata[pos_idx]
      if (delta1 == 0) {
        tau_inv[pos_idx] <- st_pos
      } else {
        W_pos <- as.numeric(VGAM::lambertW(delta1 * st_pos^2))
        g_pos <- sqrt(W_pos / delta1)
        tau_inv[pos_idx] <- g_pos
        ## For the inverse Tukey-h map,
        ## d tau^{-1}(y) / dy = exp(-W/2) / (1 + W).
        log_jacobian <- log_jacobian +
          sum(-0.5 * W_pos - log1p(W_pos))
      }
    }
    if (length(neg_idx) > 0) {
      st_neg <- stdata[neg_idx]
      if (delta2 == 0) {
        tau_inv[neg_idx] <- st_neg
      } else {
        W_neg <- as.numeric(VGAM::lambertW(delta2 * st_neg^2))
        g_neg <- sqrt(W_neg / delta2)
        tau_inv[neg_idx] <- -g_neg
        log_jacobian <- log_jacobian +
          sum(-0.5 * W_neg - log1p(W_neg))
      }
    }

    y <- try(forwardsolve(decompvarcov, tau_inv, transpose = FALSE), silent = TRUE)
    if (inherits(y, "try-error")) return(llik)
    quad_form <- sum(y^2)
    if (!is.finite(quad_form) || !is.finite(log_jacobian)) return(llik)
    0.5 * (const + dimat * log(sill) + logdetvarcov + quad_form) - log_jacobian
  }

  loglik_tukey2h <- function(param,const,coordx,coordy,coordz,coordt,corr,corrmat,corrmodel,data,dimat,fixed,fname,
                             grid,ident,mdecomp,model,namescorr,namesnuis,namesparam,radius,setup,X,ns,NS,MM,aniso,namesaniso) {
    llik <- 1.0e8
    pram <- combine_param(param, fixed, namesparam)
    paramcorr <- as.numeric(pram[namescorr])
    nuisance <- pram[namesnuis]
    Mean <- get_mean(X, nuisance, MM)

    if (aniso) {
      coords1 <- GeoAniso(cbind(coordx, coordy, coordz), anisopars = as.numeric(pram[namesaniso]))
      coordx <- coords1[,1]; coordy <- coords1[,2]; coordz <- if (ncol(coords1) == 3) coords1[,3] else NULL
    }

    sill <- as.numeric(nuisance['sill'])
    nugget <- as.numeric(nuisance['nugget'])
    tail1 <- as.numeric(nuisance['tail1'])
    tail2 <- as.numeric(nuisance['tail2'])
    if (tail1 < 0 || tail1 >= 0.5 || tail2 < 0 || tail2 >= 0.5 || nugget < 0 || nugget >= 1 || sill <= 0) return(llik)

    nuisance_temp <- nuisance
    nuisance_temp['sill'] <- 1
    nuisance_temp['nugget'] <- 0
    corrv <- matr(corrmat, corr, coordx, coordy, coordz, coordt, corrmodel, nuisance_temp, paramcorr, ns, NS, radius)
    if (any(!is.finite(corrv))) return(llik)
    corrmat_full <- build_correlation_matrix(corrv, ident)
    cova <- (1 - nugget) * corrmat_full + nugget * ident
    stdata <- (data - c(Mean)) / sqrt(sill)
    LogNormDenStand_Tukey2H(const = const, cova = cova, dimat = dimat, mdecomp = mdecomp,
                            nuisance = nuisance, sill = sill, setup = setup, stdata = stdata)
  }

  LogNormDenStand_TukeyH <- function(const, cova, ident, dimat, mdecomp, delta, sill, setup, stdata) {
    llik <- 1.0e8
    if (delta <= 0 || sill <= 0) return(llik)
    decompvarcov <- MatDecomp(cova, mdecomp)
    if (is.logical(decompvarcov) && !decompvarcov) return(llik)
    logdetvarcov <- MatLogDet(decompvarcov, mdecomp)
    if (is.na(logdetvarcov)) return(llik)
    Wval <- as.numeric(VGAM::lambertW(delta * stdata^2))
    IL <- sign(stdata) * sqrt(Wval / delta)
    y <- try(forwardsolve(decompvarcov, IL, transpose = FALSE), silent = TRUE)
    if (inherits(y, "try-error")) return(llik)
    quad_form <- sum(y^2)

    ## Stable form of -2*log|d tau^{-1}(y)/dy|.
    ## Since d tau^{-1}(y)/dy = exp(-W/2)/(1+W), the
    ## contribution is W + 2*log(1+W), including the y=0 limit.
    log_jacobian <- sum(Wval + 2 * log1p(Wval))
    if (!is.finite(quad_form) || !is.finite(log_jacobian)) return(llik)
    0.5 * (const + dimat * log(sill) + logdetvarcov + quad_form + log_jacobian)
  }

  loglik_tukeyh <- function(param,const,coordx,coordy,coordz,coordt,corr,corrmat,corrmodel,data,dimat,fixed,fname,
                            grid,ident,mdecomp,model,namescorr,namesnuis,namesparam,radius,setup,X,ns,NS,MM,aniso,namesaniso) {
    llik <- 1.0e8
    pram <- combine_param(param, fixed, namesparam)
    paramcorr <- as.numeric(pram[namescorr])
    nuisance <- pram[namesnuis]
    Mean <- get_mean(X, nuisance, MM)

    if (aniso) {
      coords1 <- GeoAniso(cbind(coordx, coordy, coordz), anisopars = as.numeric(pram[namesaniso]))
      coordx <- coords1[,1]; coordy <- coords1[,2]; coordz <- if (ncol(coords1) == 3) coords1[,3] else NULL
    }

    sill <- as.numeric(nuisance['sill'])
    tail <- as.numeric(nuisance['tail'])
    nugget <- as.numeric(nuisance['nugget'])
    if (tail <= 0 || tail > 0.5 || nugget < 0 || nugget >= 1 || sill <= 0) return(llik)
    nuisance_temp <- nuisance
    nuisance_temp['sill'] <- 1
    nuisance_temp['nugget'] <- 0
    corrv <- matr(corrmat, corr, coordx, coordy, coordz, coordt, corrmodel, nuisance_temp, paramcorr, ns, NS, radius)
    if (any(!is.finite(corrv))) return(llik)
    corrmat_full <- build_correlation_matrix(corrv, ident)
    cova <- (1 - nugget) * corrmat_full + nugget * ident
    LogNormDenStand_TukeyH(stdata = ((data - c(Mean)) / sqrt(sill)), const = const, cova = cova,
                           dimat = dimat, ident = ident, mdecomp = mdecomp, delta = tail, sill = sill, setup = setup)
  }

  LogNormDenStand_LG <- function(const, cova, ident, dimat, mdecomp, nuisance, sill, setup, stdata, V_data, mu_s) {
    llik <- 1.0e8
    decompvarcov <- MatDecomp(cova, mdecomp)
    if (is.logical(decompvarcov) && !decompvarcov) return(llik)
    logdetvarcov <- MatLogDet(decompvarcov, mdecomp)
    if (is.na(logdetvarcov)) return(llik)
    y <- try(forwardsolve(decompvarcov, stdata, transpose = FALSE, upper.tri = FALSE), silent = TRUE)
    if (inherits(y, "try-error")) return(llik)
    quadterm <- sum(y^2)
    sigma <- sqrt(sill)
    jacobian_term <- -dimat * log(sigma) - sum(log(V_data)) - sum(log(mu_s))
    loglik_val <- -0.5 * (const + logdetvarcov + quadterm) + jacobian_term
    -loglik_val
  }

  loglik_loggauss <- function(param,const,coordx,coordy,coordz,coordt,corr,corrmat,corrmodel,data,dimat,fixed,fname,
                              grid,ident,mdecomp,model,namescorr,namesnuis,namesparam,radius,setup,X,ns,NS,MM,aniso,namesaniso) {
    llik <- 1.0e8
    pram <- combine_param(param, fixed, namesparam)
    paramcorr <- as.numeric(pram[namescorr])
    nuisance <- pram[namesnuis]
    Mean <- get_mean(X, nuisance, MM)

    if (aniso) {
      coords1 <- GeoAniso(cbind(coordx, coordy, coordz), anisopars = as.numeric(pram[namesaniso]))
      coordx <- coords1[,1]; coordy <- coords1[,2]; coordz <- if (ncol(coords1) == 3) coords1[,3] else NULL
    }

    sill <- as.numeric(nuisance["sill"])
    nugget <- as.numeric(nuisance["nugget"])
    if (sill <= 0 || nugget < 0 || nugget >= 1 || any(data <= 0)) return(llik)
    mu_s <- exp(Mean)
    V_data <- data / mu_s
    if (any(V_data <= 0)) return(llik)
    sigma <- sqrt(sill)
    Z_data <- (log(V_data) + sill / 2) / sigma
    nuisance_temp <- nuisance
    nuisance_temp["sill"] <- 1
    corrv <- matr(corrmat, corr, coordx, coordy, coordz, coordt, corrmodel, nuisance_temp, paramcorr, ns, NS, radius)
    if (any(!is.finite(corrv))) return(llik)
    corrmat_full <- build_correlation_matrix(corrv, ident)
    cova <- (1 - nugget) * corrmat_full + nugget * ident
    LogNormDenStand_LG(stdata = Z_data, const = const, cova = cova, dimat = dimat, ident = ident,
                       mdecomp = mdecomp, nuisance = nuisance, sill = sill, setup = setup,
                       V_data = V_data, mu_s = mu_s)
  }

  LogNormDenStand_SH <- function(const,cova,mdecomp,nuisance,sill,setup,stdata, dimat = NULL, ...) {
    llik <- 1.0e8
    decompvarcov <- MatDecomp(cova, mdecomp)
    if (is.logical(decompvarcov) && !decompvarcov) return(llik)
    logdetvarcov <- MatLogDet(decompvarcov, mdecomp)
    if (is.na(logdetvarcov)) return(llik)
    skew <- as.numeric(nuisance["skew"])
    delta <- as.numeric(nuisance["tail"])
    Z <- sinh(delta * asinh(stdata) - skew)
    C <- delta * sqrt((1 + Z^2) / (stdata^2 + 1))
    y <- try(forwardsolve(decompvarcov, Z, transpose = FALSE), silent = TRUE)
    if (inherits(y, "try-error") || any(C <= 0)) return(llik)
    llik <- 0.5 * (const + const * log(sill) / log(2*pi) + logdetvarcov - 2 * sum(log(C)) + sum(y^2))
    as.numeric(llik)
  }

  loglik_sh <- function(param,const,coordx,coordy,coordz,coordt,corr,corrmat,corrmodel,data,dimat,fixed,fname,
                        grid,ident,mdecomp,model,namescorr,namesnuis,namesparam,radius,setup,X,ns,NS,MM,aniso,namesaniso) {
    llik <- 1.0e8
    pram <- combine_param(param, fixed, namesparam)
    paramcorr <- as.numeric(pram[namescorr])
    nuisance <- pram[namesnuis]
    Mean <- get_mean(X, nuisance, MM)

    if (aniso) {
      coords1 <- GeoAniso(cbind(coordx, coordy, coordz), anisopars = as.numeric(pram[namesaniso]))
      coordx <- coords1[,1]; coordy <- coords1[,2]; coordz <- if (ncol(coords1) == 3) coords1[,3] else NULL
    }

    sill <- as.numeric(nuisance['sill'])
    nugget <- as.numeric(nuisance['nugget'])
    tail <- as.numeric(nuisance['tail'])
    if (tail <= 0 || sill <= 0 || nugget < 0 || nugget >= 1) return(llik)
    nuisance_temp <- nuisance
    nuisance_temp['sill'] <- 1
    corrv <- matr(corrmat, corr, coordx, coordy, coordz, coordt, corrmodel, nuisance_temp, paramcorr, ns, NS, radius)
    if (any(!is.finite(corrv))) return(llik)
    corrmat_full <- build_correlation_matrix(corrv, ident)
    cova <- (1 - nugget) * corrmat_full + nugget * ident
    do.call(what = fname, args = list(stdata = ((data - c(Mean)) / sqrt(sill)), const = const, cova = cova,
                                      mdecomp = mdecomp, nuisance = nuisance, sill = sill, setup = setup,
                                      dimat = dimat))
  }

  #########################################################################
  ## Main setup
  #########################################################################

  spacetime_dyn <- FALSE
  NS <- 0
  fname <- NULL
  hessian <- FALSE
  namesaniso <- c("angle", "ratio")

  if (!is.null(coordx_dyn)) spacetime_dyn <- TRUE
  if (grid) {
    a <- expand.grid(coordx, coordy)
    coordx <- a[, 1]
    coordy <- a[, 2]
  }

  if (!spacetime_dyn) dimat <- numcoord * numtime
  if (spacetime_dyn)  dimat <- sum(ns)

  if (is.null(dim(X))) {
    if (!bivariate) X <- as.matrix(rep(1, dimat))
    if ( bivariate) X <- as.matrix(rep(1, ns[1] + ns[2]))
  } else {
    if (!bivariate) num_betas <- ncol(X)
    if ( bivariate) num_betas <- c(ncol(X), ncol(X))
  }

  corrmat <- "CorrelationMat"
  if (model == 36 || model == 47) corrmat <- "CorrelationMat_dis"
  if (spacetime) {
    corrmat <- "CorrelationMat_st_dyn"
    if (model == 36 || model == 47) corrmat <- "CorrelationMat_st_dyn_dis"
  }
  if (bivariate) corrmat <- "CorrelationMat_biv_dyn"

  if (spacetime || bivariate) {
    NS <- cumsum(ns)
    if (spacetime_dyn) {
      data <- t(unlist(data))
      NS <- c(0, NS)[-(length(ns) + 1)]
    } else {
      data <- matrix(t(data), nrow = 1)
      NS <- rep(0, numtime)
    }
  }

  dd <- 0
  if (bivariate) dd <- dimat
  numpairstot <- dimat * (dimat - 1) / 2 + dd
  const <- dimat * log(2*pi)

  if (is.null(neighb)) {
    corr <- double(numpairstot)
    ident <- diag(dimat)
  } else {
    corr <- double(numpairstot)
    ident <- diag(dimat)
  }

  #########################################################################
  ## Select objective and density function
  #########################################################################

  if (model == 1 || model == 20) {
    lname <- 'loglik'
    if (bivariate) lname <- 'loglik_biv'

    if (type == 4) {
      if (bivariate) fname <- 'LogNormDenStand_biv' else fname <- 'LogNormDenStand'
    }
    if (type == 5 || type == 6) {
      corrmat <- "CorrelationMat_tap"
      if (spacetime) corrmat <- "CorrelationMat_st_tap"
      if (bivariate) {
        if (type == 5) fname <- 'LogNormDenTap_biv'
        if (type == 6) fname <- 'LogNormDenTap1_biv'
        corrmat <- "CorrelationMat_biv_tap"
      } else {
        if (type == 5) fname <- 'LogNormDenTap'
        if (type == 6) fname <- 'LogNormDenTap1'
      }
      corr <- double(numpairs)
      tapcorr <- double(numpairs)
      tcor <- matr(corrmat, tapcorr, coordx, coordy, coordz, coordt, setup$tapmodel, c(0,0,1), 1, ns, NS, radius)
      tape <- new("spam", entries = tcor, colindices = setup$ja, rowpointers = setup$ia,
                  dimension = as.integer(rep(dimat, 2)))
      setup$struct <- try(spam::chol.spam(tape, silent = TRUE))
      setup$taps <- tcor
    }
    if (type == 8) {
      if (bivariate) fname <- "CVV_biv" else fname <- "CVV"
    }
    if (sparse) fname <- paste(fname, "_spam", sep = "")
  }

  if (model == 20) {
    lname <- 'loglik_sh'
    if (bivariate) lname <- 'loglik_biv_sh'
    fname <- 'LogNormDenStand_SH'
    if (type == 6) fname <- 'LogshDenTap1'
  }
  if (model == 34) {
    lname <- 'loglik_tukeyh'
    if (bivariate) lname <- 'loglik_biv_tukeyh'
  }
  if (model == 40) {
    lname <- 'loglik_tukey2h'
    if (bivariate) lname <- 'loglik_biv_tukey2h'
  }
  if (model == 47) {
    lname <- 'loglik_miss_Poisgamma'
    if (bivariate) lname <- 'loglik_biv_miss_Poisgamma'
  }
  if (model == 35) {
    lname <- 'loglik_miss_T'
    if (bivariate) lname <- 'loglik_biv_miss_T'
  }
  if (model == 36) {
    lname <- 'loglik_miss_Pois'
    if (bivariate) lname <- 'loglik_biv_miss_Pois'
  }
  if (model == 37) {
    lname <- 'loglik_miss_skewT'
    if (bivariate) lname <- 'loglik_biv_miss_skewT'
  }
  if (model == 22) {
    lname <- 'loglik_loggauss'
    if (bivariate) lname <- 'loglik_biv_loggauss'
  }

  if (type != 5 && type != 6) corrmat <- paste(corrmat, "2", sep = "")
  if (is.null(coordz)) coordz <- double(length(coordx))

  if (!exists(lname, mode = "function")) {
    stop("Objective function not implemented in GeoLik.R: ", lname)
  }
  obj_fun <- get(lname, mode = "function")
  data_opt <- as.numeric(t(data))

  common_args <- list(
    const = const,
    coordx = coordx,
    coordy = coordy,
    coordz = coordz,
    coordt = coordt,
    corr = corr,
    corrmat = corrmat,
    corrmodel = corrmodel,
    data = data_opt,
    dimat = dimat,
    fixed = fixed,
    fname = fname,
    grid = grid,
    ident = ident,
    mdecomp = mdecomp,
    model = model,
    namescorr = namescorr,
    namesnuis = namesnuis,
    namesparam = namesparam,
    radius = radius,
    setup = setup,
    X = X,
    ns = ns,
    NS = NS,
    MM = MM,
    aniso = aniso,
    namesaniso = namesaniso
  )

  #########################################################################
  ## Optimization
  #########################################################################

  if (!onlyvar) {
    ## Safe, conservative controls for full-likelihood optimization.
    ## These match the same spirit used in CompLik2: tighter than the
    ## previous quick defaults, but without changing the likelihood itself.
    opt_reltol <- 1e-12
    opt_pgtol  <- 1e-10
    opt_factr  <- 1e4
    opt_maxit  <- 100000L

    if (length(param) == 1) {
      optimizer <- "optimize"
      opt_bounds <- make_optimize_bounds(param, lower, upper, namesparam)
      Likelihood <- do.call(stats::optimize,
                            c(list(f = obj_fun, lower = opt_bounds$lower, upper = opt_bounds$upper, maximum = FALSE), common_args))
    }

    if (length(param) > 1) {
      if (optimizer == 'L-BFGS-B') {
        Likelihood <- do.call(stats::optim,
                              c(list(par = param, fn = obj_fun, method = 'L-BFGS-B', lower = lower, upper = upper,
                                     hessian = hessian,
                                     control = list(factr = opt_factr, pgtol = opt_pgtol, maxit = opt_maxit, parscale = abs(param) + 1)),
                                common_args))
      }

      if (optimizer %in% c('Nelder-Mead', 'BFGS', 'SANN')) {
        opt_control <- list(reltol = opt_reltol, maxit = opt_maxit, parscale = abs(param) + 1)
        if (optimizer == 'SANN') opt_control <- list(maxit = opt_maxit, parscale = abs(param) + 1)
        Likelihood <- do.call(stats::optim,
                              c(list(par = param, fn = obj_fun, method = optimizer, hessian = hessian,
                                     control = opt_control), common_args))
      }

      if (optimizer == 'nlm') {
        Likelihood <- do.call(stats::nlm,
                              c(list(f = obj_fun, p = param, hessian = hessian, iterlim = opt_maxit, steptol = 1e-6, gradtol = 1e-8),
                                common_args))
      }

      if (optimizer == 'bobyqa') {
        bobyqa_args <- make_bobyqa_setup(param, lower, upper, namesparam)
        bobyqa_args$control$maxfun <- opt_maxit
        Likelihood <- do.call(minqa::bobyqa,
                              c(list(par = param, fn = obj_fun,
                                     lower = bobyqa_args$lower,
                                     upper = bobyqa_args$upper,
                                     control = bobyqa_args$control),
                                common_args))
      }

      if (optimizer == 'nlminb') {
        Likelihood <- do.call(stats::nlminb,
                              c(list(start = param, objective = obj_fun, lower = lower, upper = upper,
                                     control = list(iter.max = opt_maxit, eval.max = opt_maxit,
                                                    rel.tol = opt_reltol, x.tol = 1e-8)),
                                common_args))
      }
    }

    if (!exists("Likelihood")) stop("Optimizer not implemented: ", optimizer)

    #######################################################################
    ## Post-processing
    #######################################################################

    if (optimizer %in% c('Nelder-Mead', 'L-BFGS-B', 'BFGS', 'SANN')) {
      conv_code <- Likelihood$convergence
      conv_msg <- if (!is.null(Likelihood$message)) Likelihood$message else NA_character_
      names(Likelihood$par) <- namesparam
      param <- Likelihood$par
      maxfun <- -Likelihood$value
      Likelihood$value <- maxfun
      Likelihood$convergence_code <- conv_code
      Likelihood$opt_code <- conv_code
      Likelihood$optim_message <- conv_msg
      Likelihood$opt_message <- conv_msg
      Likelihood$counts <- as.numeric(Likelihood$counts[1])

      if (conv_code == 0) {
        Likelihood$convergence <- 'Successful'
      } else if (conv_code == 1) {
        Likelihood$convergence <- paste0(optimizer, ': iteration limit reached')
      } else if (optimizer == 'Nelder-Mead' && conv_code == 10) {
        Likelihood$convergence <- 'Nelder-Mead warning: simplex degenerated; check solution'
      } else {
        Likelihood$convergence <- paste0(optimizer, ' warning: code=', conv_code)
      }
    }

    if (optimizer == 'bobyqa') {
      names(Likelihood$par) <- namesparam
      param <- Likelihood$par
      maxfun <- -Likelihood$fval
      Likelihood$value <- maxfun
      Likelihood$convergence_code <- Likelihood$ierr
      Likelihood$opt_code <- Likelihood$ierr
      Likelihood$optim_message <- Likelihood$msg
      Likelihood$opt_message <- Likelihood$msg
      Likelihood$counts <- if (!is.null(Likelihood$feval)) as.numeric(Likelihood$feval) else NA_real_

      if (Likelihood$ierr == 0) {
        Likelihood$convergence <- 'Successful'
      } else if (Likelihood$ierr == 1) {
        Likelihood$convergence <- 'bobyqa warning: maximum number of function evaluations reached'
      } else if (Likelihood$ierr == 3) {
        Likelihood$convergence <- 'bobyqa warning: trust-region step failed to reduce quadratic model'
      } else if (Likelihood$ierr == 4) {
        Likelihood$convergence <- 'bobyqa warning: bounds too tight relative to rhobeg'
      } else if (Likelihood$ierr == 5) {
        Likelihood$convergence <- 'bobyqa warning: numerical cancellation problem'
      } else {
        Likelihood$convergence <- paste0('bobyqa warning: ierr=', Likelihood$ierr)
      }
    }

    if (optimizer %in% c('nlminb', 'multinlminb')) {
      conv_code <- Likelihood$convergence
      conv_msg <- if (!is.null(Likelihood$message)) Likelihood$message else NA_character_
      names(Likelihood$par) <- namesparam
      param <- Likelihood$par
      maxfun <- -Likelihood$objective
      Likelihood$value <- maxfun
      Likelihood$convergence_code <- conv_code
      Likelihood$opt_code <- conv_code
      Likelihood$optim_message <- conv_msg
      Likelihood$opt_message <- conv_msg
      Likelihood$counts <- as.numeric(Likelihood$iterations)
      if (conv_code == 0) {
        Likelihood$convergence <- 'Successful'
      } else {
        Likelihood$convergence <- paste0(
          'nlminb warning: code=', conv_code,
          if (!is.na(conv_msg)) paste0('; message=', conv_msg) else ''
        )
      }
    }

    if (optimizer == 'nlm') {
      conv_code <- Likelihood$code
      names(Likelihood$estimate) <- namesparam
      param <- Likelihood$estimate
      maxfun <- -as.numeric(Likelihood$minimum)
      Likelihood$value <- maxfun
      Likelihood$par <- param
      Likelihood$param <- param
      Likelihood$convergence_code <- conv_code
      Likelihood$opt_code <- conv_code
      Likelihood$optim_message <- NA_character_
      Likelihood$opt_message <- NA_character_
      Likelihood$counts <- as.numeric(Likelihood$iterations)
      if (conv_code == 1 || conv_code == 2) {
        Likelihood$convergence <- 'Successful'
      } else if (conv_code == 4 || conv_code == 5) {
        Likelihood$convergence <- paste0('nlm warning: code=', conv_code, '; iteration/step limit reached')
      } else {
        Likelihood$convergence <- paste0('nlm warning: code=', conv_code)
      }
    }

    if (optimizer == 'optimize') {
      param <- Likelihood$minimum
      names(param) <- namesparam
      maxfun <- -Likelihood$objective
      Likelihood$value <- maxfun
      Likelihood$par <- param
      Likelihood$param <- param
      Likelihood$convergence <- 'Successful'
      Likelihood$convergence_code <- 0
      Likelihood$optim_message <- NA_character_
      Likelihood$counts <- NA_real_
    }

    serious_failure <- is.null(maxfun) || !is.finite(maxfun) || maxfun <= -1.0e14
    if (serious_failure) {
      Likelihood$convergence <- 'Optimization may have failed: Try with other starting parameters'
      warning('Optimization may have failed: try with other starting values', call. = FALSE)
    }

    if (is.null(Likelihood$par)) Likelihood$par <- param
    names(Likelihood$par) <- namesparam

    numparam <- length(param)
    Likelihood$claic <- NULL
    Likelihood$clbic <- NULL
    if (type == 4) {
      Likelihood$claic <- -2 * (maxfun - numparam)
      Likelihood$clbic <- -2 * maxfun + numparam * log(dimat)
    }
  } else {
    Likelihood <- list(value = NA_real_)
    Likelihood$par <- param
    names(Likelihood$par) <- namesparam
    Likelihood$param <- param
    Likelihood$claic <- NULL
    Likelihood$clbic <- NULL
    Likelihood$convergence <- 'None'
    serious_failure <- FALSE
    maxfun <- Likelihood$value
    numparam <- length(param)
  }

  if (!exists("serious_failure", inherits = FALSE)) serious_failure <- FALSE

  #########################################################################
  ## Hessian, score, asymptotic covariance
  #########################################################################

  if (varest) {
    Likelihood$hessian <- numDeriv::hessian(
      func = obj_fun,
      x = as.numeric(Likelihood$par),
      method = "Richardson",
      const = const,
      coordx = coordx,
      coordy = coordy,
      coordz = coordz,
      coordt = coordt,
      corr = corr,
      corrmat = corrmat,
      corrmodel = corrmodel,
      data = data_opt,
      dimat = dimat,
      fixed = fixed,
      fname = fname,
      grid = grid,
      ident = ident,
      mdecomp = mdecomp,
      model = model,
      namescorr = namescorr,
      namesnuis = namesnuis,
      namesparam = namesparam,
      radius = radius,
      setup = setup,
      X = X,
      ns = ns,
      NS = NS,
      MM = MM,
      aniso = aniso,
      namesaniso = namesaniso
    )
    rownames(Likelihood$hessian) <- namesparam
    colnames(Likelihood$hessian) <- namesparam
  }

  Likelihood$score <- NULL
  if (score) {
    Likelihood$score <- -numDeriv::grad(
      func = obj_fun,
      x = as.numeric(Likelihood$par),
      method = "Richardson",
      const = const,
      coordx = coordx,
      coordy = coordy,
      coordz = coordz,
      coordt = coordt,
      corr = corr,
      corrmat = corrmat,
      corrmodel = corrmodel,
      data = data_opt,
      dimat = dimat,
      fixed = fixed,
      fname = fname,
      grid = grid,
      ident = ident,
      mdecomp = mdecomp,
      model = model,
      namescorr = namescorr,
      namesnuis = namesnuis,
      namesparam = namesparam,
      radius = radius,
      setup = setup,
      X = X,
      ns = ns,
      NS = NS,
      MM = MM,
      aniso = aniso,
      namesaniso = namesaniso
    )
    names(Likelihood$score) <- namesparam
  }

  if (!isTRUE(serious_failure) || identical(Likelihood$convergence, 'None')) {
    if (varest) {
      if ((model == 20 || model == 22 || model == 1 || model == 34 || model == 35 || model == 37) && !(type == 5 || type == 6)) {
        aa <- try(abs(det(Likelihood$hessian)), silent = TRUE)
        if (inherits(aa, "try-error") || !is.finite(aa) || aa <= 1e-12) {
          warning("Asymptotic information matrix is singular", immediate. = TRUE)
          Likelihood$varcov <- NULL
        } else {
          ii <- try(solve(Likelihood$hessian), silent = TRUE)
          if (inherits(ii, "try-error")) {
            warning("Asymptotic information matrix is singular", immediate. = TRUE)
            Likelihood$varcov <- NULL
          } else {
            eig <- eigen(ii, symmetric = TRUE, only.values = TRUE)$values
            if (any(!is.finite(eig)) || min(eig) <= 0) {
              warning("Asymptotic information matrix is not positive-definite")
              Likelihood$varcov <- NULL
            } else {
              Likelihood$varcov <- ii
              dimnames(Likelihood$varcov) <- list(namesparam, namesparam)
              Likelihood$stderr <- sqrt(diag(Likelihood$varcov))
              names(Likelihood$stderr) <- namesparam
            }
          }
        }
      } else if (!is.null(Likelihood$varcov)) {
        eig <- eigen(Likelihood$varcov, symmetric = TRUE, only.values = TRUE)$values
        if (any(!is.finite(eig)) || min(eig) < 0) {
          warning("Asymptotic information matrix is not positive-definite")
          Likelihood$varcov <- 'none'
          Likelihood$stderr <- 'none'
        } else {
          dimnames(Likelihood$varcov) <- list(namesparam, namesparam)
          Likelihood$stderr <- diag(Likelihood$varcov)
          if (any(Likelihood$stderr < 0)) {
            Likelihood$stderr <- 'none'
          } else {
            Likelihood$stderr <- sqrt(Likelihood$stderr)
            names(Likelihood$stderr) <- namesparam
          }
        }
      }
    }
  }

  if (varest && is.null(Likelihood$varcov)) {
    Likelihood$varcov <- 'none'
    Likelihood$stderr <- 'none'
  }

  if (varest) Likelihood$sensmat <- Likelihood$hessian

  return(Likelihood)
}

####################################################
### File name: CompLik.r
####################################################



CompLik <- function(copula,bivariate, coordx, coordy ,coordz,coordt,coordx_dyn,corrmodel, data, distance, flagcorr, flagnuis, fixed,grid,
                           likelihood,lower, model, n, namescorr, namesnuis, namesparam,
                           numparam, numparamcorr, optimizer, onlyvar, param, spacetime, type,
                           upper, varest, weigthed, ns, X,sensitivity,MM,aniso,score)
  {
    ### Define the object function:
    comploglik <- function(param,coords,coordt, corrmodel, data, fixed, fan, n, namescorr, 
                              namesnuis,namesparam,namesaniso,weigthed,X,ns,NS,MM)
      {
        names(param) <- namesparam
        param <- c(param, fixed)
        paramcorr <- param[namescorr]
        nuisance <- param[namesnuis]

        #aniso<-param[namesaniso]

        mean_names <- .GeoMean_names(ncol(X))
        sel <- names(nuisance) %in% mean_names
        if (is.null(MM)) {
          beta <- .GeoMean_beta(nuisance, ncol(X))
          Mean <- as.numeric(X %*% beta)
        } else {
          Mean <- as.numeric(MM)
        }

       other_nuis=as.numeric(nuisance[!sel])   ## or nuis parameters (nugget sill skew df)

        result=dotCall64::.C64(as.character(fan),
        SIGNATURE = c("integer","double","double","double","double",
                         "integer","double","integer","double","double","double","double",
                          "integer","integer","integer","integer"),  
                         corrmodel ,coords[,1],coords[,2] , coordt ,  data , n , paramcorr ,  weigthed , 
                   res=dotCall64::vector_dc("double",1), Mean,0,other_nuis ,
                     ns , NS , 
         INTENT =    c("r","r","r","r","r","r","r","r","w","r", "r","r","r","r","r","r"),
         PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)$res
 
         return(-result)
      }
     comploglik_biv <- function(param,coords,coordt, corrmodel, data, fixed, fan, n, namescorr, namesnuis,namesparam,namesaniso,weigthed,X,ns,NS,MM)
      {

        names(param) <- namesparam
        param <- c(param, fixed)
        paramcorr <- param[namescorr]
        nuisance <- param[namesnuis]
        
        #aniso<-param[namesaniso]

        sel1=substr(names(nuisance),1,6)=="mean_1"
        mm1=as.numeric(nuisance[sel1])
        sel2=substr(names(nuisance),1,6)=="mean_2"
        mm2=as.numeric(nuisance[sel2])
        sel=substr(names(nuisance),1,4)=="mean"
        X1=as.matrix(X[1:ns[1],]);X2=as.matrix(X[(ns[1]+1):(ns[2]+ns[1]),]); 
        other_nuis=as.numeric(nuisance[!sel]) 



        result=dotCall64::.C64(as.character(fan),
        SIGNATURE = c("integer","double","double","double","double","double",
                         "integer","double","integer","double","double","double","double",
                          "integer","integer","integer","integer"),  
                         corrmodel ,coordx,coordy , coordz,coordt ,  data , n , paramcorr ,  weigthed , 
                   res=dotCall64::vector_dc("double",1), c(X1%*%mm1,X2%*%mm2),0,other_nuis ,
                     ns , NS ,
         INTENT =    c("r","r","r","r","r","r","r","r","w","r", "r","r","r","r","r","r","r"),
         PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)$res
        return(-result)
      }
   ############################################################################################################################################
   ############################################################################################################################################
   ############################################################################################################################################
 
    numcoord=length(coordx);numtime=1;spacetime_dyn=FALSE
    if(spacetime) numtime=length(coordt)
    if(bivariate) numtime=2
    if(!is.null(coordx_dyn)) spacetime_dyn=TRUE
    if(!spacetime_dyn)  dimat <- numcoord*numtime#
    if(spacetime_dyn)  dimat <- sum(ns)

    NS=cumsum(ns)
    if(is.null(dim(X)))
    {
       X=as.matrix(rep(1,dimat))
       if((spacetime||bivariate)&& spacetime_dyn)  X=as.matrix(rep(1,NS[numtime]))
    }
    else(if(bivariate) X=rbind(X,X))

  
    fname <- NULL; hessian <- FALSE 
    namesaniso=c("angle","ratio")

    if(all(model==1,likelihood==1,type==2)) fname <- 'Comp_Cond_Gauss'
    if(all(model==1,likelihood==3,type==2)) {fname <- 'Comp_Pair_Gauss'
                                              if(varest ) hessian <- TRUE}
    if(all(model==2,likelihood==3,type==2)){ fname <- 'Comp_Pair_BinomGauss'
                                             if(varest ) hessian <- TRUE}
    if(all(model==11,likelihood==3,type==2)){ fname <- 'Comp_Pair_BinomGauss'
                                              if(varest ) hessian <- TRUE}
    if(all(model==19,likelihood==3,type==2)){ namesnuis=c(namesnuis,"z")
                                              fixed<- c(fixed, list(z=min(n)))
                                              fname <- 'Comp_Pair_Binom2Gauss'
                                              if(varest ) hessian <- TRUE}
    if(all(model==14,likelihood==3,type==2)){ fname <- 'Comp_Pair_BinomnegGauss'
                                              if(varest ) hessian <- TRUE}
    if(all(model==16,likelihood==3,type==2)){ fname <- 'Comp_Pair_BinomnegGauss'
                                              if(varest ) hessian <- TRUE}
    if(all(model==15,likelihood==3,type==2)){ fname <- 'Comp_Pair_PoisbinGauss'
                                              if(varest ) hessian <- TRUE}
    if(all(model==17,likelihood==3,type==2)){ fname <- 'Comp_Pair_PoisbinnegGauss'
                                              if(varest ) hessian <- TRUE}    
    if(all(model==13,likelihood==3,type==2)){ fname <- 'Comp_Pair_WrapGauss'
                                              if(varest ) hessian <- TRUE}
    if(all(model==10,likelihood==3,type==2)){ fname <- 'Comp_Pair_SkewGauss'
                                              if(varest ) hessian <- TRUE}
     if(all(model==21,likelihood==3,type==2)){ fname <- 'Comp_Pair_Gamma'
                                              if(varest ) hessian <- TRUE}
      if(all(model==21,likelihood==1,type==2)){ fname <- 'Comp_Cond_Gamma'
                                              if(varest ) hessian <- TRUE}
    if(all(model==33,likelihood==3,type==2)){ fname <- 'Comp_Pair_Kumaraswamy'
                                              if(varest ) hessian <- TRUE}
      if(all(model==42,likelihood==3,type==2)){ fname <- 'Comp_Pair_Kumaraswamy2'
                                              if(varest ) hessian <- TRUE}
   if(all(model==26,likelihood==3,type==2)){ fname <- 'Comp_Pair_Weibull'
                                              if(varest ) hessian <- TRUE}    
   if(all(model==26,likelihood==1,type==2)){ fname <- 'Comp_Cond_Weibull'
                                              if(varest ) hessian <- TRUE}    
    if(all(model==28,likelihood==3,type==2)){ fname <- 'Comp_Pair_Beta'
                                              if(varest ) hessian <- TRUE}                                   
    if(all(model==24,likelihood==3,type==2)){ fname <- 'Comp_Pair_LogLogistic'
                                              if(varest ) hessian <- TRUE}     
    if(all(model==25,likelihood==3,type==2)){ fname <- 'Comp_Pair_Logistic'
                                              if(varest ) hessian <- TRUE}                                                                              
    if(all(model==23,likelihood==3,type==2)){ fname <- 'Comp_Pair_2Gamma'
                                              if(varest ) hessian <- TRUE}                                        
    if(all(model==22,likelihood==3,type==2)){ fname <- 'Comp_Pair_LogGauss';
                                              if(varest ) hessian <- TRUE}
    if(all(model==18,likelihood==3,type==2)){ fname <- 'Comp_Pair_SkewTGauss'
                                              if(varest ) hessian <- TRUE}  
    if(all(model==27,likelihood==3,type==2)){ fname <- 'Comp_Pair_TWOPIECET'
                                              if(varest ) hessian <- TRUE} 
    if(all(model==39,likelihood==3,type==2)){ fname <- 'Comp_Pair_TWOPIECEBIMODAL'
                                              if(varest ) hessian <- TRUE} 
    if(all(model==29,likelihood==3,type==2)){ fname <- 'Comp_Pair_TWOPIECEGauss'
                                              if(varest ) hessian <- TRUE} 
    if(all(model==31,likelihood==3,type==2)){ fname <- 'Comp_Pair_BinomTWOPIECEGauss'
                                              if(varest ) hessian <- TRUE} 
    if(all(model==32,likelihood==3,type==2)){ fname <- 'Comp_Pair_BinomnegTWOPIECEGauss'
                                              if(varest ) hessian <- TRUE}
    if(all(model==12,likelihood==3,type==2)){ fname <- 'Comp_Pair_T'
                                              if(varest ) hessian <- TRUE}  
    if(all(model==34,likelihood==3,type==2)){ fname <- 'Comp_Pair_Tukeyh' 
                                              if(varest ) hessian <- TRUE} 
    if(all(model==41,likelihood==3,type==2)){ fname <- 'Comp_Pair_Gauss_misp_Tukeygh' 
                                              if(varest ) hessian <- TRUE} 
    if(all(model==40,likelihood==3,type==2)){ fname <- 'Comp_Pair_Tukeyhh' 
                                              if(varest ) hessian <- TRUE} 
    if(all(model==36,likelihood==3,type==2)){ fname <- 'Comp_Pair_Gauss_misp_Pois'
                                              if(varest ) hessian <- TRUE}
    if(all(model==35,likelihood==3,type==2)){ fname <- 'Comp_Pair_Gauss_misp_T'
                                              if(varest ) hessian <- TRUE}
    if(all(model==37,likelihood==3,type==2)){ fname <- 'Comp_Pair_Gauss_misp_SkewT'
                                              if(varest ) hessian <- TRUE}
    if(all(model==20,likelihood==3,type==2)){ fname <- 'Comp_Pair_SinhGauss'
                                              if(varest ) hessian <- TRUE}
    if(all(model==38,likelihood==3,type==2)){ fname <- 'Comp_Pair_TWOPIECETukeyh'
                                              if(varest ) hessian <- TRUE} 
    if(all(model==30,likelihood==3,type==2)){ fname <- 'Comp_Pair_Pois'
                                              if(varest ) hessian <- TRUE}
    if(all(model==46,likelihood==3,type==2)){ fname <- 'Comp_Pair_PoisGamma'
                                              if(varest ) hessian <- TRUE}
    if(all(model==43,likelihood==3,type==2)){ fname <- 'Comp_Pair_PoisZIP'
                                              if(varest ) hessian <- TRUE}
    if(all(model==44,likelihood==3,type==2)){ fname <- 'Comp_Pair_Gauss_misp_PoisZIP'
                                              if(varest ) hessian <- TRUE}
    if(all(model==45,likelihood==3,type==2)){ fname <- 'Comp_Pair_BinomnegGaussZINB'
                                              if(varest ) hessian <- TRUE}
    if(sensitivity) hessian=TRUE
    if(spacetime) fname <- paste(fname,"_st",sep="")
    if(bivariate) fname <- paste(fname,"_biv",sep="")
    fname <- paste(fname,"2",sep="")

    if(!is.null(copula))
    {
        if(copula=="Gaussian") fname <- paste(fname,"GCop",sep="")
        if(copula=="Clayton")     fname <- paste(fname,"BCop",sep="")
    }

    if(aniso) fname <- paste(fname,"_aniso",sep="")

   # path.parent <- getwd()
  
    
   
    
    if((spacetime||bivariate)&&(!spacetime_dyn)){
                                  data=c(t(data))
                                  coordx=rep(coordx,numtime);coordy=rep(coordy,numtime);
                                  if(!is.null) coordz=rep(coordz,numtime);
                }
                if((spacetime||bivariate)&&(spacetime_dyn)) data=unlist(data)          
    
   if(spacetime||bivariate)   NS=c(0,NS)[-(length(ns)+1)]



coords=cbind(coordx,coordy,coordz)

   if(!onlyvar){
  ##############################.  spatial or space time ############################################
   if(!bivariate)           {
    if(length(param)==1) {
         optimizer="optimize"  
     CompLikelihood <- .GeoModels_optimize_1d(
                              f=comploglik, param=param, lower=lower, upper=upper,
                              namesparam=namesparam,
                              coords=coords, coordt=coordt,corrmodel=corrmodel,
                              data=data, fixed=fixed, fan=fname, n=n,
                              namescorr=namescorr, namesnuis=namesnuis,namesaniso=namesaniso,
                              weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)
          }
   if(length(param)>1) {
    if(optimizer=='L-BFGS-B')
      CompLikelihood <- optim(par=param,fn=comploglik, 
                              control=list(factr=1e-10,pgtol=1e-14, maxit=100000), 
                              coords=coords, coordt=coordt,corrmodel=corrmodel, data=data, fixed=fixed,
                              fan=fname, lower=lower, method='L-BFGS-B',n=n,
                              namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso, 
                              upper=upper,weigthed=weigthed,X=X,ns=ns,NS=NS, hessian=FALSE,MM=MM)
   
    if(optimizer=='BFGS') 
        CompLikelihood <- optim(par=param, fn=comploglik,  coords=coords, coordt=coordt,corrmodel=corrmodel, 
                           control=list(factr=1e-10,
                             reltol=1e-14, maxit=100000), data=data, fixed=fixed, fan=fname,
                              hessian=FALSE, method='BFGS',n=n,namescorr=namescorr,
                                  namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso, weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)

      if(optimizer=='Nelder-Mead')
        CompLikelihood <- optim(par=param, fn=comploglik,  coords=coords, coordt=coordt,corrmodel=corrmodel, 
          control=list( reltol=1e-14, maxit=100000), data=data, fixed=fixed, fan=fname,
                              hessian=FALSE, method='Nelder-Mead',n=n,namescorr=namescorr,
                                  namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso, weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)
    
 #    if(optimizer=='lbfgsb3')
 #      CompLikelihood <-lbfgsb3::lbfgsb3(prm=param,fn=comploglik,gr=NULL,lower=lower,upper=upper,
 #                             #control=list(factr=1e-10,pgtol=1e-14, maxit=100000), 
 #                             coords=coords, coordt=coordt,corrmodel=corrmodel, data=data, fixed=fixed,
 #                             fan=fname,n=n,
 #                             namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam, 
 #                             weigthed=weigthed,X=X,ns=ns,NS=NS)

    #if(optimizer=='nmk')
     # CompLikelihood <-dfoptim::nmk(par=param, fn=comploglik, control = list(maxfeval=100000,tol=1e-10),
      #                     coords=coords, coordt=coordt,corrmodel=corrmodel, data=data, fixed=fixed, fan=fname,
       #                    n=n,namescorr=namescorr,namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso, weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)
    #if(optimizer=='nmkb')
    #{
     # CompLikelihood <-dfoptim::nmkb(par=param, fn=comploglik, control = list(maxfeval=100000,tol=1e-10),
      #                   lower=lower,upper=upper,
       #                  coords=coords, coordt=coordt,corrmodel=corrmodel, data=data, fixed=fixed, fan=fname,
        #                 n=n,namescorr=namescorr,namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso, weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)
    #}
    if(optimizer=='nlm')
    CompLikelihood <- nlm(f=comploglik,p=param,steptol = 1e-4, coords=coords, coordt=coordt,corrmodel=corrmodel, data=data, fixed=fixed,
                               fan=fname,hessian=FALSE,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso,  
                               iterlim=100000, weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)
  
    if(optimizer=='nlminb')
     CompLikelihood <-nlminb(objective=comploglik,start=param,coords=coords, coordt=coordt,corrmodel=corrmodel, data=data, fixed=fixed,
                                control = list( iter.max=100000),
                              lower=lower,upper=upper,
                               fan=fname,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam, namesaniso=namesaniso, 
                               weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)

                                   
    }}
     ############################## bivariate  ############################################                           
    if(bivariate)           {
      
    
     if(length(param)==1)
        {
         optimizer="optimize" 
       CompLikelihood <- .GeoModels_optimize_1d(
                              f=comploglik_biv, param=param, lower=lower, upper=upper,
                              namesparam=namesparam,
                              coords=coords, coordt=coordt,corrmodel=corrmodel,
                              data=data, fixed=fixed,fan=fname,n=n,MM=MM,
                              namescorr=namescorr, namesnuis=namesnuis,namesaniso=namesaniso,
                              weigthed=weigthed,X=X,ns=ns,NS=NS)}
      if(length(param)>1) {   
    if(optimizer=='L-BFGS-B'){
      CompLikelihood <- optim(param,comploglik_biv, control=list(pgtol=1e-14, maxit=100000),
                              method='L-BFGS-B',hessian=FALSE,lower=lower, upper=upper,MM=MM,
                              coords=coords, coordt=coordt,corrmodel=corrmodel, 
                              data=data, fixed=fixed,fan=fname,n=n,
                              namescorr=namescorr, namesnuis=namesnuis, namesparam=namesparam,namesaniso=namesaniso, 
                             weigthed=weigthed,X=X,ns=ns,NS=NS )}
  
      if(optimizer=='BFGS')
      CompLikelihood <- optim(param,comploglik_biv, coords=coords, coordt=coordt,corrmodel=corrmodel, control=list(
                              reltol=1e-14, maxit=100000), data=data, fixed=fixed, fan=fname,
                              hessian=FALSE, method='BFGS',n=n,MM=MM,
                              namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam ,namesaniso=namesaniso, weigthed=weigthed,X=X,ns=ns,NS=NS)
   if(optimizer=='Nelder-Mead')
      CompLikelihood <- optim(param,comploglik_biv, coords=coords, coordt=coordt,corrmodel=corrmodel, control=list(
                              reltol=1e-14, maxit=100000), data=data, fixed=fixed, fan=fname,
                              hessian=FALSE, method='Nelder-Mead',n=n,MM=MM,
                              namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam ,namesaniso=namesaniso, weigthed=weigthed,X=X,ns=ns,NS=NS)
       if(optimizer=='nlm') 
        CompLikelihood <- nlm( f=comploglik_biv,p=param, coords=coords, coordt=coordt,corrmodel=corrmodel, data=data, fixed=fixed,
                               fan=fname,hessian=FALSE,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso, MM=MM, 
                               weigthed=weigthed,X=X,ns=ns,NS=NS)
 
    if(optimizer=='nlminb') 
        CompLikelihood <- nlminb( objective=comploglik_biv,start=param, 
                                     control = list( iter.max=100000),
                              lower=lower,upper=upper,
                               coords=coords, coordt=coordt,corrmodel=corrmodel, data=data, fixed=fixed,
                               fan=fname,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso, MM=MM,
                               weigthed=weigthed,X=X,ns=ns,NS=NS)

    
   }}                    
      ########################################################################################   
      ########################################################################################
    # check the optimisation outcome
      if(optimizer=='Nelder-Mead'||optimizer=='multiNelder-Mead'){
        CompLikelihood$value = -CompLikelihood$value
        names(CompLikelihood$par)<- namesparam
        if(CompLikelihood$convergence == 0)
        CompLikelihood$convergence <- 'Successful'
        else
        if(CompLikelihood$convergence == 1)
        CompLikelihood$convergence <- 'Iteration limit reached'
        else
        CompLikelihood$convergence <- "Optimization may have failed"
        if(CompLikelihood$value==-1.0e8) CompLikelihood$convergence <- 'Optimization may have failed: Try with other starting parameters'
    }
        if(optimizer=='nmk'||optimizer=='nmkb'){
        CompLikelihood$value = -CompLikelihood$value
        names(CompLikelihood$par)<- namesparam
        if(CompLikelihood$convergence == 0)
        CompLikelihood$convergence <- 'Successful'
        else CompLikelihood$convergence <- "Optimization may have failed"
        if(CompLikelihood$value==-1.0e8) CompLikelihood$convergence <- 'Optimization may have failed: Try with other starting parameters'
    }

    if(optimizer=='L-BFGS-B'||optimizer=='BFGS'||optimizer=='lbfgsb3c'){
        CompLikelihood$value = -CompLikelihood$value
        names(CompLikelihood$par)<- namesparam
        if(CompLikelihood$convergence == 0)
        CompLikelihood$convergence <- 'Successful'
        else
        if(CompLikelihood$convergence == 1)
        CompLikelihood$convergence <- 'Iteration limit reached'
        else
        CompLikelihood$convergence <- "Optimization may have failed"
        if(CompLikelihood$value==-1.0e8) CompLikelihood$convergence <- 'Optimization may have failed: Try with other starting parameters'
    }

     if(optimizer=='nlm'){
        CompLikelihood$par <- CompLikelihood$estimate
        names(CompLikelihood$par)<- namesparam
        CompLikelihood$value <- -CompLikelihood$minimum
        if(CompLikelihood$code == 1|| CompLikelihood$code == 2)
        CompLikelihood$convergence <- 'Successful'
        else
        if(CompLikelihood$code == 4)
        CompLikelihood$convergence <- 'Iteration limit reached'
        else
        CompLikelihood$convergence <- "Optimization may have failed"
        if(CompLikelihood$value==-1.0e8) CompLikelihood$convergence <- 'Optimization may have failed: Try with other starting parameters'
    }

    if(optimizer=='nlminb'||optimizer=='multinlminb' ){
     
        CompLikelihood$par <- CompLikelihood$par
        names(CompLikelihood$par)<- namesparam
        CompLikelihood$value <- -CompLikelihood$objective
        if(CompLikelihood$convergence == 0) { CompLikelihood$convergence <- 'Successful' }
        else {CompLikelihood$convergence <- "Optimization may have failed" }
        #if(CompLikelihood$objective==-1.0e8) CompLikelihood$convergence <- 'Optimization may have failed: Try with other starting parameters'
    }
    if(optimizer=='optimize'){
    param<-CompLikelihood$minimum
    CompLikelihood$par<-param  
    names(CompLikelihood$par)<- namesparam
    maxfun <- -CompLikelihood$objective
    CompLikelihood$value <- maxfun
    CompLikelihood$convergence <- 'Successful'
    }
  } ##### end if !onlyvar
    else {
          CompLikelihood=as.list(0)
          names(CompLikelihood)="value"
          CompLikelihood$par <- param
          CompLikelihood$claic <- NULL;CompLikelihood$clbic <- NULL;
          CompLikelihood$convergence <- 'Successful'
          if(!bivariate)CompLikelihood$value = - comploglik(param=CompLikelihood$par ,  coords=coords, coordt=coordt,corrmodel=corrmodel, data=data, fixed=fixed, fan=fname,
                             n=n,namescorr=namescorr,namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso, weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)
          else CompLikelihood$value = -comploglik_biv(param=CompLikelihood$par ,  coords=coords, coordt=coordt,corrmodel=corrmodel, data=data, fixed=fixed, fan=fname,
                             n=n,namescorr=namescorr,namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso, weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)
          

          if(hessian) 
          {
               if(!bivariate)  
                CompLikelihood$hessian=numDeriv::hessian(func=comploglik,x=param,method="Richardson",  coords=coords, coordt=coordt,corrmodel=corrmodel, 
                              data=data, fixed=fixed,fan=fname,n=n,
                              namescorr=namescorr, namesnuis=namesnuis, namesparam=namesparam,namesaniso=namesaniso, 
                              weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)
               if(bivariate)  
               CompLikelihood$hessian=numDeriv::hessian(func=comploglik_biv,x=param,method="Richardson",coords=coords, coordt=coordt,corrmodel=corrmodel, 
                             data=data, fixed=fixed,fan=fname,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam, namesaniso=namesaniso, 
                             weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)
               rownames(CompLikelihood$hessian)=namesparam
               colnames(CompLikelihood$hessian)=namesparam
          }
  }

#####################################
#if((sensitivity||varest)&&(is.null(CompLikelihood$hessian)||min(eigen(CompLikelihood$hessian)$values)<0))
if((sensitivity||varest))
  {
if(!bivariate)  

CompLikelihood$hessian=numDeriv::hessian(func=comploglik,x=CompLikelihood$par,method="Richardson",  coords=coords, coordt=coordt,corrmodel=corrmodel, 
                              data=data, fixed=fixed,fan=fname,n=n,
                              namescorr=namescorr, namesnuis=namesnuis, namesparam=namesparam,namesaniso=namesaniso, 
                              weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)
if(bivariate)  
CompLikelihood$hessian=numDeriv::hessian(func=comploglik_biv,x=CompLikelihood$par,method="Richardson",coords=coords, coordt=coordt,corrmodel=corrmodel, 
                             data=data, fixed=fixed,fan=fname,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam, namesaniso=namesaniso, 
                               weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)
rownames(CompLikelihood$hessian)=namesparam
colnames(CompLikelihood$hessian)=namesparam
  }

CompLikelihood$score=NULL
if(score)
{
  if(!bivariate)  
    CompLikelihood$score=numDeriv::grad(func=comploglik,x=CompLikelihood$par,method="Richardson",  coords=coords, coordt=coordt,corrmodel=corrmodel, 
                              data=data, fixed=fixed,fan=fname,n=n,
                              namescorr=namescorr, namesnuis=namesnuis, namesparam=namesparam,namesaniso=namesaniso, 
                              weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM)  
  
  if(bivariate)  
    CompLikelihood$score=numDeriv::grad(func=comploglik_biv,x=CompLikelihood$par,method="Richardson",coords=coords, coordt=coordt,corrmodel=corrmodel, 
                             data=data, fixed=fixed,fan=fname,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam, namesaniso=namesaniso, 
                               weigthed=weigthed,X=X,ns=ns,NS=NS,MM=MM) 
  
  names(CompLikelihood$score)=namesparam
}
####################################
if(hessian) CompLikelihood$sensmat=CompLikelihood$hessian

return(CompLikelihood)
  }

## Internal utilities shared by GeoVarest and GeoVarestbootstrap.
## Keep bootstrap plumbing in one place while the two public functions retain
## their distinct statistical post-processing.

.GeoVarest_preserve_seed <- function(seed) {
  if (is.null(seed)) return(function() invisible(NULL))

  old_kind <- RNGkind()
  had_seed <- exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  old_seed <- if (had_seed) {
    get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  } else {
    NULL
  }

  set.seed(seed)

  function() {
    do.call(RNGkind, as.list(old_kind))
    if (had_seed) {
      assign(".Random.seed", old_seed, envir = .GlobalEnv)
    } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      rm(".Random.seed", envir = .GlobalEnv)
    }
    invisible(NULL)
  }
}

.GeoVarest_prepare_geometry <- function(fit, coordt_use = fit$coordt) {
  dimat <- if (!is.null(fit$coordx_dyn)) sum(fit$ns) else fit$numtime * fit$numcoord

  if (!is.null(fit$X) && !is.null(dim(fit$X)) && ncol(fit$X) == 1L) {
    ncheck <- min(NROW(fit$X), dimat)
    X_use <- if (ncheck > 0L && all(fit$X[seq_len(ncheck), 1L] == 1)) NULL else fit$X
  } else {
    X_use <- fit$X
  }

  coords <- if (!is.null(fit$coordz)) {
    cbind(fit$coordx, fit$coordy, fit$coordz)
  } else {
    cbind(fit$coordx, fit$coordy)
  }

  if (isTRUE(fit$bivariate) && is.null(fit$coordx_dyn)) {
    if (nrow(coords) %% 2L != 0L) {
      stop("bivariate=TRUE but odd number of coordinates", call. = FALSE)
    }
    coords <- coords[seq_len(nrow(coords) / 2L), , drop = FALSE]
  }

  list(dimat = dimat, X = X_use, coords = coords, coordt = coordt_use)
}

.GeoVarest_make_fit_spec <- function(fit, boot_par, model_est, X_use,
                                     optimizer, lower, upper, pair_index,
                                     coordt_use, memdist = TRUE,
                                     weighted = FALSE, include_start = FALSE) {
  list(
    param       = if (isTRUE(include_start)) boot_par$start else NULL,
    fixed       = boot_par$fixed,
    coordt      = coordt_use,
    coordx_dyn  = fit$coordx_dyn,
    copula      = fit$copula,
    anisopars   = boot_par$anisopars,
    est.aniso   = fit$est.aniso,
    aniso_names = if (is.null(boot_par$anisopars)) character(0) else
      intersect(c("angle", "ratio"), names(unlist(boot_par$anisopars, use.names = TRUE))),
    thin_method = fit$thin_method,
    lower       = lower,
    upper       = upper,
    neighb      = fit$neighb,
    p_neighb    = fit$p_neighb,
    corrmodel   = fit$corrmodel,
    model       = model_est,
    n           = fit$n,
    maxdist     = fit$maxdist,
    maxtime     = fit$maxtime,
    memdist     = memdist,
    optimizer   = optimizer,
    grid        = fit$grid,
    likelihood  = fit$likelihood,
    type        = fit$type,
    X           = X_use,
    distance    = fit$distance,
    radius      = fit$radius,
    weighted    = weighted,
    bivariate   = fit$bivariate,
    spacetime   = fit$spacetime,
    pair_index  = pair_index
  )
}

.GeoVarest_start_from_theta <- function(theta_hat, aniso_names = character(0)) {
  out <- as.list(as.numeric(theta_hat))
  names(out) <- names(theta_hat)
  if (length(aniso_names)) out[intersect(names(out), aniso_names)] <- NULL
  out
}

.GeoVarest_fit_args <- function(data, coords, fit_spec, start,
                                score = FALSE) {
  args <- list(
    data        = data,
    start       = start,
    fixed       = fit_spec$fixed,
    coordx      = coords,
    coordt      = fit_spec$coordt,
    coordx_dyn  = fit_spec$coordx_dyn,
    copula      = fit_spec$copula,
    anisopars   = fit_spec$anisopars,
    est.aniso   = fit_spec$est.aniso,
    thin_method = fit_spec$thin_method,
    lower       = fit_spec$lower,
    upper       = fit_spec$upper,
    neighb      = fit_spec$neighb,
    p_neighb    = fit_spec$p_neighb,
    corrmodel   = fit_spec$corrmodel,
    model       = fit_spec$model,
    sparse      = FALSE,
    n           = fit_spec$n,
    maxdist     = fit_spec$maxdist,
    maxtime     = fit_spec$maxtime,
    memdist     = fit_spec$memdist,
    optimizer   = fit_spec$optimizer,
    grid        = fit_spec$grid,
    likelihood  = fit_spec$likelihood,
    type        = fit_spec$type,
    X           = fit_spec$X,
    distance    = fit_spec$distance,
    radius      = fit_spec$radius,
    weighted    = fit_spec$weighted
  )

  if (isTRUE(score)) {
    args$onlyvar <- TRUE
    args$score <- TRUE
    args$sensitivity <- FALSE
    args$varest <- FALSE
  }
  args
}

.GeoVarest_worker_context <- function(fit_spec, coords) {
  list(
    pair_info = .GeoVarest_prepare_pair_info(fit_spec$pair_index, coords, fit_spec),
    fit_context = if (as.character(fit_spec$type)[1L] %in% c("Pairwise", "Independence")) {
      new.env(parent = emptyenv())
    } else {
      NULL
    }
  )
}

.GeoVarest_run_fit <- function(data, coords, fit_spec, start,
                               pair_info, fit_context, score = FALSE) {
  GeoFit_fun <- getFromNamespace("GeoFit", "GeoModels")
  fit_args <- .GeoVarest_fit_args(data, coords, fit_spec, start, score = score)
  out <- NULL
  capture.output({
    out <- .GeoVarest_with_selected_pairs(
      fit_spec$pair_index,
      do.call(GeoFit_fun, fit_args),
      pair_info = pair_info,
      fit_context = fit_context
    )
  }, file = nullfile())
  out
}

.GeoVarest_extract_score <- function(fit_result, target_names) {
  num_params <- length(target_names)
  sc <- unlist(fit_result$score, use.names = TRUE)
  if (length(sc) != num_params || any(!is.finite(sc))) {
    stop("GeoFit returned an invalid score", call. = FALSE)
  }
  if (!is.null(names(sc)) && all(target_names %in% names(sc))) sc <- sc[target_names]

  ll <- if (!is.null(fit_result$logCompLik)) fit_result$logCompLik else fit_result$logLik
  ll <- as.numeric(ll)
  if (length(ll) != 1L || !is.finite(ll)) {
    stop("non-finite log composite likelihood", call. = FALSE)
  }

  ans <- c(-as.numeric(sc), logCompLik = ll)
  names(ans) <- c(target_names, "logCompLik")
  ans
}

.GeoVarest_extract_refit <- function(fit_result, target_names) {
  bad <- function() {
    z <- rep(NA_real_, length(target_names) + 1L)
    names(z) <- c(target_names, "logCompLik")
    z
  }

  if (is.null(fit_result) || !identical(fit_result$convergence, "Successful")) return(bad())
  ll <- if (!is.null(fit_result$logCompLik)) fit_result$logCompLik else fit_result$logLik
  if (length(ll) != 1L || !is.finite(ll)) return(bad())

  th <- unlist(fit_result$param, use.names = TRUE)
  if (!is.null(names(th)) && all(target_names %in% names(th))) th <- th[target_names]
  if (length(th) != length(target_names) || any(!is.finite(th))) return(bad())

  ans <- c(as.numeric(th), logCompLik = as.numeric(ll))
  names(ans) <- c(target_names, "logCompLik")
  ans
}

.GeoVarest_failed_result <- function(target_names, error = NULL) {
  z <- rep(NA_real_, length(target_names) + 1L)
  names(z) <- c(target_names, "logCompLik")
  if (!is.null(error)) attr(z, "error") <- conditionMessage(error)
  z
}

####################################################
### Internal simulation validation helpers
####################################################

.GeoSimulationCanonicalModel <- function(model) {
  if(length(model) != 1L || is.na(model))
    stop("model must be a single non-missing character string.", call. = FALSE)
  model <- gsub("[[:blank:]]", "", as.character(model))
  aliases <- c(
    Gauss = "Gaussian",
    SkewGauss = "SkewGaussian",
    LogGauss = "LogGaussian",
    TwoPieceGauss = "TwoPieceGaussian",
    Binary = "Binomial",
    Bernoulli = "Binomial",
    Geom = "BinomialNeg",
    Geometric = "BinomialNeg",
    tukeyh = "Tukeyh",
    tukeyh2 = "Tukeyh2",
    gamma = "Gamma",
    weibull = "Weibull",
    Loglogistic = "LogLogistic",
    Binomiallogistic = "BinomialLogistic",
    poisson = "Poisson",
    poissongamma = "PoissonGamma"
  )
  if(model %in% names(aliases)) unname(aliases[[model]]) else model
}

.GeoSimulationEffectiveN <- function(model_requested, model, n) {
  requested <- gsub("[[:blank:]]", "", as.character(model_requested)[1L])
  if(requested %in% c("Binary", "Bernoulli")) return(1L)
  if(requested %in% c("Geom", "Geometric")) {
    nn <- .GeoValidateIntegerVector(n, "n", nobs = NULL, minimum = 1L)
    if(length(nn) != 1L || nn[1L] != 1L) {
      stop("For Geometric simulation, n is fixed to 1 (the r = 1 Negative-Binomial case).",
           call. = FALSE)
    }
    return(1L)
  }
  n
}

.GeoValidateNrep <- function(nrep) {
  if(!is.numeric(nrep) || length(nrep) != 1L || !is.finite(nrep) ||
     nrep < 1 || abs(nrep - round(nrep)) > sqrt(.Machine$double.eps)) {
    stop("nrep must be a positive integer.", call. = FALSE)
  }
  as.integer(round(nrep))
}

.GeoValidateIntegerVector <- function(x, name, nobs = NULL, minimum = 1L) {
  x <- as.numeric(x)
  tol <- sqrt(.Machine$double.eps)
  if(!length(x) || any(!is.finite(x)) || any(x < minimum) ||
     any(abs(x - round(x)) > tol)) {
    stop(sprintf("%s must contain integer values greater than or equal to %d.",
                 name, as.integer(minimum)), call. = FALSE)
  }
  if(!is.null(nobs) && !(length(x) %in% c(1L, as.integer(nobs)))) {
    stop(sprintf("%s must have length 1 or the number of simulated observations (%d).",
                 name, as.integer(nobs)), call. = FALSE)
  }
  as.integer(round(x))
}

.GeoValidateLatentIntegerParameters <- function(model, param) {
  tol <- sqrt(.Machine$double.eps)
  pos_integer <- function(x, name, minimum = 1L) {
    x <- as.numeric(x)
    if(length(x) != 1L || !is.finite(x) || x < minimum ||
       abs(x - round(x)) > tol) {
      stop(sprintf("For direct %s simulation, %s must be an integer >= %d because it determines the number of latent Gaussian fields.",
                   model, name, as.integer(minimum)), call. = FALSE)
    }
    as.integer(round(x))
  }

  if(model == "Gamma") {
    param$shape <- pos_integer(param$shape, "param$shape")
  }
  if(model == "Beta") {
    param$shape1 <- pos_integer(param$shape1, "param$shape1")
    param$shape2 <- pos_integer(param$shape2, "param$shape2")
  }
  if(model %in% c("StudentT", "SkewStudentT", "TwoPieceStudentT")) {
    eps_df <- as.numeric(param$df)
    if(length(eps_df) != 1L || !is.finite(eps_df) || eps_df <= 0)
      stop(sprintf("For direct %s simulation, param$df must be the positive reciprocal degrees-of-freedom parameter.", model), call. = FALSE)
    nu <- 1 / eps_df
    if(!is.finite(nu) || nu < 3 || abs(nu - round(nu)) > tol) {
      stop(sprintf("For direct %s simulation, 1/param$df must be an integer >= 3 because the model is constructed from that many latent Gaussian fields.", model), call. = FALSE)
    }
  }
  if(model == "TwoPieceBimodal") {
    param$df <- pos_integer(param$df, "param$df")
  }
  if(model %in% c("PoissonGamma", "PoissonGammaZIP")) {
    shape <- as.numeric(param$shape)
    m <- 2 * shape
    if(length(shape) != 1L || !is.finite(shape) || shape <= 0 ||
       abs(m - round(m)) > tol) {
      stop(sprintf("For direct %s simulation, 2*param$shape must be a positive integer because the Gamma mixing field is generated from squared latent Gaussian fields.", model), call. = FALSE)
    }
  }
  param
}

.GeoValidateSimulationMarginal <- function(model, param) {
  scalar_positive <- function(x, name, strict = TRUE) {
    x <- as.numeric(x)
    if(length(x) != 1L || !is.finite(x) || (strict && x <= 0) || (!strict && x < 0))
      stop(sprintf("%s must be a finite %s scalar.", name,
                   if(strict) "positive" else "non-negative"), call. = FALSE)
    x
  }
  scalar_interval <- function(x, name, lo, hi) {
    x <- as.numeric(x)
    if(length(x) != 1L || !is.finite(x) || x <= lo || x >= hi)
      stop(sprintf("%s must lie strictly between %g and %g.", name, lo, hi), call. = FALSE)
    x
  }

  if(!is.null(param$sill)) scalar_positive(param$sill, "param$sill")
  if(!is.null(param$nugget)) {
    nug <- as.numeric(param$nugget)
    if(length(nug) != 1L || !is.finite(nug) || nug < 0 || nug >= 1)
      stop("param$nugget must be a finite scalar in [0,1).", call. = FALSE)
  }
  if(model %in% c("Weibull", "Kumaraswamy2"))
    scalar_positive(param$shape, "param$shape")
  if(model == "LogLogistic") {
    sh <- as.numeric(param$shape)
    if(length(sh) != 1L || !is.finite(sh) || sh <= 1)
      stop("param$shape must be finite and greater than 1 for LogLogistic simulation under the mean parametrization.", call. = FALSE)
  }
  if(model == "TwoPieceBimodal") {
    sh <- as.numeric(param$shape)
    if(length(sh) != 1L || !is.finite(sh) || sh < 0)
      stop("param$shape must be a finite non-negative scalar for TwoPieceBimodal simulation.", call. = FALSE)
  }
  if(model == "Kumaraswamy") {
    scalar_positive(param$shape1, "param$shape1")
    scalar_positive(param$shape2, "param$shape2")
  }
  if(model %in% c("PoissonZIP", "PoissonGammaZIP", "BinomialNegZINB")) {
    pmu <- as.numeric(param$pmu)
    if(length(pmu) != 1L || !is.finite(pmu))
      stop("param$pmu must be a finite scalar for zero-inflated simulation.", call. = FALSE)
    for(nm in c("nugget1", "nugget2")) {
      ng <- as.numeric(param[[nm]])
      if(length(ng) != 1L || !is.finite(ng) || ng < 0 || ng >= 1)
        stop(sprintf("param$%s must be a finite scalar in [0,1) for zero-inflated simulation.", nm), call. = FALSE)
    }
  }
  if(model %in% c("Beta", "Kumaraswamy", "Kumaraswamy2")) {
    lo <- as.numeric(param$min); hi <- as.numeric(param$max)
    if(length(lo) != 1L || length(hi) != 1L || !is.finite(lo) ||
       !is.finite(hi) || lo >= hi)
      stop(sprintf("Direct %s simulation requires finite param$min < param$max.", model), call. = FALSE)
  }
  if(model == "SinhAsinh") scalar_positive(param$tail, "param$tail")
  if(model == "SkewLaplace") scalar_interval(param$skew, "param$skew", 0, 1)
  if(model %in% c("SkewStudentT", "TwoPieceGaussian", "TwoPieceStudentT",
                 "TwoPieceTukeyh", "TwoPieceBimodal"))
    scalar_interval(param$skew, "param$skew", -1, 1)
  if(model == "Tukeyh") {
    tail <- as.numeric(param$tail)
    if(length(tail) != 1L || !is.finite(tail) || tail < 0 || tail >= 0.5)
      stop("param$tail must lie in [0, 0.5) for Tukeyh simulation.", call. = FALSE)
  }
  if(model == "TwoPieceTukeyh") {
    tail <- as.numeric(param$tail)
    if(length(tail) != 1L || !is.finite(tail) || tail < 0)
      stop("param$tail must be a finite non-negative scalar.", call. = FALSE)
  }
  if(model == "Tukeyh2") {
    for(nm in c("tail1", "tail2")) {
      x <- as.numeric(param[[nm]])
      if(length(x) != 1L || !is.finite(x) || x < 0 || x >= 0.5)
        stop(sprintf("param$%s must lie in [0, 0.5) for Tukeyh2 simulation.", nm), call. = FALSE)
    }
  }
  if(model == "Tukeygh") {
    tail <- as.numeric(param$tail)
    skew <- as.numeric(param$skew)
    if(length(tail) != 1L || !is.finite(tail) || tail < 0)
      stop("param$tail must be a finite non-negative scalar.", call. = FALSE)
    if(length(skew) != 1L || !is.finite(skew))
      stop("param$skew must be a finite scalar.", call. = FALSE)
  }
  param
}

.GeoDirectSimulationModels <- function() {
  c(
    "Gaussian", "Tukeygh", "SkewGaussian", "Binomial", "StudentT",
    "Wrapped", "BinomialNeg", "SkewStudentT", "SinhAsinh", "Gamma",
    "LogGaussian", "LogLogistic", "Logistic", "SkewLaplace", "Weibull",
    "TwoPieceStudentT", "Beta", "TwoPieceGaussian", "Poisson",
    "Kumaraswamy", "Tukeyh", "TwoPieceTukeyh", "TwoPieceBimodal",
    "Tukeyh2", "Kumaraswamy2", "PoissonZIP", "BinomialNegZINB",
    "PoissonGamma", "BinomialLogistic", "PoissonGammaZIP"
  )
}

.GeoValidateDirectSimulationModel <- function(model, simulator = "GeoSim") {
  if(model == "Beta2") {
    stop(sprintf("Direct Beta2 simulation is not implemented in %s(); use GeoSimCopula(..., model = 'Beta2') with an explicit copula.", simulator),
         call. = FALSE)
  }
  if(grepl("_misp_", model, fixed = TRUE) || startsWith(model, "Gaussian_misp_") ||
     startsWith(model, "Binary_misp_")) {
    stop(sprintf("%s is an inferential misspecification model and is not a data-generating model in %s(). Simulate from the corresponding non-misspecified model instead.",
                 model, simulator), call. = FALSE)
  }
  if(!(model %in% .GeoDirectSimulationModels())) {
    stop(sprintf("%s does not implement direct simulation for model = '%s'. Use a supported direct model or GeoSimCopula() when a copula construction is intended.",
                 simulator, model), call. = FALSE)
  }
  invisible(TRUE)
}

.GeoTBCorrelationModels <- function() {
  c("Matern", "GenWend", "Genwend", "Hypergeometric", "HyperGeometric",
    "hypergeometric", "GenWend_Matern", "Genwend_Matern",
    "Hypergeometric_Matern", "HyperGeometric_Matern", "hypergeometric_Matern",
    "Kummer", "Kummer_Matern", "Kummer_matern")
}

.GeoValidateTBCorrelationModel <- function(corrmodel, bivariate = FALSE) {
  if(isTRUE(bivariate)) {
    if(!(corrmodel %in% c("Bi_matern", "Bi_Matern")))
      stop(sprintf("GeoSimapprox(method='TB') does not implement bivariate correlation model '%s'.", corrmodel), call. = FALSE)
  } else if(!(corrmodel %in% .GeoTBCorrelationModels())) {
    stop(sprintf("GeoSimapprox(method='TB') does not implement correlation model '%s'. Use GeoSim() or a supported TB correlation family.", corrmodel), call. = FALSE)
  }
  invisible(TRUE)
}

.GeoValidateCopulaMarginal <- function(model, param) {
  ## Copula margins use quantile transforms; integer restrictions from the
  ## direct latent-field constructions do not apply here.
  if(!is.null(param$sill)) {
    ss <- as.numeric(param$sill)
    if(length(ss) != 1L || !is.finite(ss) || ss <= 0)
      stop("param$sill must be a finite positive scalar.", call. = FALSE)
  }
  if(model == "SinhAsinh") {
    tail <- as.numeric(param$tail); skew <- as.numeric(param$skew)
    if(length(tail) != 1L || !is.finite(tail) || tail <= 0)
      stop("SinhAsinh tail must be a finite positive scalar.", call. = FALSE)
    if(length(skew) != 1L || !is.finite(skew))
      stop("SinhAsinh skew must be a finite scalar.", call. = FALSE)
  }
  if(model == "Tukeyh") {
    tail <- as.numeric(param$tail)
    if(length(tail) != 1L || !is.finite(tail) || tail < 0 || tail >= 0.5)
      stop("Tukeyh tail must lie in [0, 0.5).", call. = FALSE)
  }
  if(model == "TwoPieceTukeyh") {
    tail <- as.numeric(param$tail)
    if(length(tail) != 1L || !is.finite(tail) || tail < 0)
      stop("TwoPieceTukeyh tail must be a finite non-negative scalar.", call. = FALSE)
  }
  if(model == "Tukeyh2") {
    vals <- c(as.numeric(param$tail1), as.numeric(param$tail2))
    if(length(vals) != 2L || any(!is.finite(vals)) || any(vals < 0) || any(vals >= 0.5))
      stop("Tukeyh2 tail1 and tail2 must lie in [0, 0.5).", call. = FALSE)
  }
  if(model == "Tukeygh") {
    vals <- c(as.numeric(param$tail), as.numeric(param$skew))
    if(length(vals) != 2L || any(!is.finite(vals)) || vals[1L] < 0)
      stop("Tukeygh requires finite skew and a finite non-negative tail.", call. = FALSE)
  }
  if(model %in% c("TwoPieceGaussian", "TwoPieceStudentT", "TwoPieceTukeyh")) {
    sk <- as.numeric(param$skew)
    if(length(sk) != 1L || !is.finite(sk) || abs(sk) >= 1)
      stop(sprintf("%s skew must lie strictly between -1 and 1.", model), call. = FALSE)
  }
  if(model == "SkewGaussian") {
    sk <- as.numeric(param$skew)
    if(length(sk) != 1L || !is.finite(sk))
      stop("SkewGaussian skew must be a finite scalar.", call. = FALSE)
  }
  invisible(TRUE)
}

.GeoCopulaSimulationModels <- function() {
  c(
    "Gaussian", "Binomial", "BinomialNeg", "BinomialNegZINB", "Poisson",
    "PoissonZIP", "Gamma", "Weibull", "LogLogistic", "LogGaussian",
    "Logistic", "StudentT", "SkewGaussian", "SkewStudentT", "SinhAsinh",
    "Tukeyh", "Tukeyh2", "SkewLaplace", "Tukeygh", "TwoPieceGaussian",
    "TwoPieceStudentT", "TwoPieceTukeyh", "Kumaraswamy", "Kumaraswamy2",
    "Beta", "Beta2"
  )
}

.GeoValidateCopulaSimulationModel <- function(model) {
  if(!(model %in% .GeoCopulaSimulationModels())) {
    stop(sprintf("GeoSimCopula does not implement marginal model '%s'.", model),
         call. = FALSE)
  }
  invisible(TRUE)
}

.GeoSimulationCanonicalCorrModel <- function(corrmodel) {
  corrmodel <- gsub("[[:blank:]]", "", as.character(corrmodel)[1L])
  aliases <- c(
    matern = "Matern",
    Genwend = "GenWend",
    HyperGeometric = "Hypergeometric",
    hypergeometric = "Hypergeometric",
    Genwend_Matern = "GenWend_Matern",
    HyperGeometric_Matern = "Hypergeometric_Matern",
    hypergeometric_Matern = "Hypergeometric_Matern",
    Kummer_matern = "Kummer_Matern",
    Bi_Matern = "Bi_matern",
    matern_matern = "Matern_Matern",
    Genwend_Genwend = "GenWend_GenWend"
  )
  if(corrmodel %in% names(aliases)) unname(aliases[[corrmodel]]) else corrmodel
}

.GeoValidateExactSimulationMethod <- function(method) {
  method <- gsub("[[:blank:]]", "", as.character(method)[1L])
  if(!(method %in% c("cholesky", "svd")))
    stop("GeoSim method must be either 'cholesky' or 'svd'.", call. = FALSE)
  method
}

.GeoValidateCopulaSimulationMethod <- function(method) {
  method <- gsub("[[:blank:]]", "", as.character(method)[1L])
  key <- tolower(method)
  if (key == "tb") return("TB")
  if (key %in% c("cholesky", "svd")) return(key)
  stop("GeoSimCopula method must be one of 'cholesky', 'svd', or 'TB'.",
       call. = FALSE)
}

.GeoValidateTBDistance <- function(distance, context = "GeoSimapprox") {
  key <- tolower(gsub("[[:blank:]]", "", as.character(distance)[1L]))
  if (!identical(key, "eucl")) {
    stop(sprintf("%s(method='TB') requires distance = 'Eucl' because turning-bands simulation operates on Cartesian coordinates.",
                 context), call. = FALSE)
  }
  invisible(TRUE)
}

.GeoValidateTBControls <- function(L) {
  if(!is.numeric(L) || length(L) != 1L || !is.finite(L) || L < 1 ||
     abs(L - round(L)) > sqrt(.Machine$double.eps))
    stop("L must be a positive integer for turning-bands simulation.", call. = FALSE)
  as.integer(round(L))
}

.GeoTBParallelMode <- function(dime, L, nrep, parallel = FALSE,
                               ncores = NULL, force_inner = FALSE) {
  ## Empirical TB scheduling rule calibrated on the two large-data regimes
  ## exercised by GeoSimapprox/GeoSimcond.  The decision deliberately uses
  ## both the work in one field (N * L) and the total work across replicates
  ## (N * L * nrep): the former determines whether spatial chunking can pay
  ## for worker overhead, while nrep determines when distributing complete
  ## independent fields is preferable.
  if (!isTRUE(parallel)) return("serial")

  if (!is.null(ncores)) {
    nc <- suppressWarnings(as.numeric(ncores)[1L])
    if (is.finite(nc) && nc <= 1) return("serial")
  }

  if (isTRUE(force_inner)) return("inner")

  N <- suppressWarnings(as.numeric(dime)[1L])
  LL <- suppressWarnings(as.numeric(L)[1L])
  R <- suppressWarnings(as.numeric(nrep)[1L])

  if (!is.finite(N) || !is.finite(LL) || !is.finite(R) ||
      N <= 0 || LL <= 0 || R <= 0)
    return("serial")

  get_positive_option <- function(name, default) {
    value <- suppressWarnings(as.numeric(getOption(name, default))[1L])
    if (!is.finite(value) || value <= 0) default else value
  }

  min_one_work <- get_positive_option(
    "GeoModels.tb_parallel_min_one_work", 1e7
  )
  min_total_work <- get_positive_option(
    "GeoModels.tb_parallel_min_total_work", 5e8
  )
  large_N <- get_positive_option(
    "GeoModels.tb_parallel_large_N", 1e5
  )
  heavy_one_work <- get_positive_option(
    "GeoModels.tb_parallel_heavy_one_work", 2e8
  )
  outer_nrep_large <- get_positive_option(
    "GeoModels.tb_parallel_outer_nrep_large", 100
  )
  outer_nrep_small <- get_positive_option(
    "GeoModels.tb_parallel_outer_nrep_small", 200
  )

  work_one <- N * LL
  work_all <- work_one * R

  ## For small individual fields, or too little aggregate work, starting a
  ## multisession plan costs more than it saves.
  if (work_one < min_one_work || work_all < min_total_work)
    return("serial")

  ## Large spatial payloads and very expensive individual fields reach the
  ## outer-replicate crossover sooner.  Otherwise keep using the inner TB
  ## path until there are enough independent replicates to amortize export
  ## and worker-start costs.
  outer_nrep <- if (N >= large_N || work_one >= heavy_one_work)
    outer_nrep_large else outer_nrep_small

  if (R > 1 && R >= outer_nrep) return("outer")

  "inner"
}
.GeoSimulationLatentSpec <- function(model, param, n) {
  k <- 1L
  npoi <- 1
  if(model %in% c("SkewGaussian", "LogGaussian", "TwoPieceGaussian", "TwoPieceTukeyh")) k <- 1L
  if(model %in% c("Weibull", "Wrapped")) k <- 2L
  if(model %in% c("LogLogistic", "Logistic", "SkewLaplace")) k <- 4L
  if(model == "Binomial") k <- max(n)
  if(model == "BinomialLogistic") k <- 2L * max(n)
  if(model %in% c("BinomialNeg", "BinomialNegZINB")) k <- 99999L
  if(model %in% c("Poisson", "PoissonZIP")) {
    k <- 2L
    npoi <- 999999999
  }
  if(model %in% c("PoissonGamma", "PoissonGammaZIP")) {
    k <- 2L + as.integer(round(2 * param$shape))
    npoi <- 999999999
  }
  if(model == "PoissonWeibull") {
    k <- 4L
    npoi <- 999999999
  }
  if(model %in% c("PoissonZIP", "BinomialNegZINB", "PoissonGammaZIP")) {
    param$nugget <- param$nugget1
  }
  if(model == "Gamma") k <- as.integer(param$shape)
  if(model == "Beta") k <- as.integer(param$shape1) + as.integer(param$shape2)
  if(model %in% c("Kumaraswamy", "Kumaraswamy2")) k <- 4L
  if(model == "StudentT") k <- as.integer(round(1 / param$df)) + 1L
  if(model == "TwoPieceBimodal") k <- as.integer(param$df) + 1L
  if(model %in% c("SkewStudentT", "TwoPieceStudentT")) {
    k <- as.integer(round(1 / param$df)) + 2L
  }
  list(k = as.integer(k), npoi = npoi, param = param)
}

.GeoSimulationSquareSums <- function(dd, index, component = 1L) {
  index <- as.integer(index)
  if(!length(index)) return(numeric(dim(dd)[1L]))
  z <- matrix(dd[, component, index, drop = FALSE], nrow = dim(dd)[1L])
  rowSums(z * z)
}

.GeoSimulationBinomialCounts <- function(dd, n) {
  dime <- dim(dd)[1L]
  k <- dim(dd)[3L]
  nn <- if(length(n) == 1L) rep.int(as.integer(n), dime) else as.integer(n)
  out <- numeric(dime)
  for(i in seq_len(k)) out <- out + dd[, 1L, i] * (nn >= i)
  out
}

.GeoSimulationBinomialLogisticCounts <- function(dd, n, mean) {
  dime <- dim(dd)[1L]
  k <- dim(dd)[3L]
  half_k <- k %/% 2L
  nn <- if(length(n) == 1L) rep.int(as.integer(n), dime) else as.integer(n)
  out <- numeric(dime)
  for(j in seq_len(half_k)) {
    i <- 2L * j - 1L
    ee <- 0.5 * (dd[, 1L, i]^2 + dd[, 1L, i + 1L]^2)
    out <- out + as.numeric(mean + log(exp(ee) - 1) > 0) * (nn >= j)
  }
  out
}

.GeoSimulationFormat <- function(sim, grid, space, numxgrid = NULL, numygrid = NULL,
                                 numtime = 1L, numcoord = length(sim), byrow = TRUE) {
  if(!grid) {
    if(space) return(c(sim))
    return(matrix(sim, nrow = numtime, ncol = numcoord, byrow = byrow))
  }
  if(space) return(array(sim, c(numxgrid, numygrid)))
  array(sim, c(numxgrid, numygrid, numtime))
}

.GeoSimulationSplitDynamic <- function(sim, ns) {
  ns <- as.integer(ns)
  ends <- cumsum(ns)
  starts <- c(1L, head(ends, -1L) + 1L)
  values <- c(sim)
  Map(function(first, last) values[first:last], starts, ends)
}

.GeoSimulationTransformUnivariate <- function(model, sim, dd = NULL, simDD = NULL,
                                               mm = 0, vv = 1, sk = 0, tl = 0,
                                               t1l = 0, t2l = 0, bimo = NULL,
                                               param = NULL, k = 1L,
                                               spacetime = FALSE) {
  byrow <- TRUE
  needs_format <- TRUE
  handled <- TRUE

  if(model %in% c("SkewGaussian", "SkewGauss")) {
    sim <- mm + sk * c(abs(dd[, , 1L])) + sqrt(vv) * c(t(simDD))
  } else if(model == "SkewStudentT") {
    denom <- sqrt(.GeoSimulationSquareSums(dd, seq_len(k - 2L)) / (k - 2L))
    bb <- sk * abs(dd[, , k - 1L]) + sqrt(1 - sk^2) * t(simDD)
    sim <- mm + sqrt(vv) * (bb / denom)
  } else if(model == "StudentT") {
    denom <- sqrt(.GeoSimulationSquareSums(dd, seq_len(k - 1L)) / (k - 1L))
    sim <- mm + sqrt(vv) * (c(t(simDD)) / denom)
  } else if(model %in% c("TwoPieceGaussian", "TwoPieceGauss")) {
    base <- t(simDD)
    discrete <- dd[, , 1L]
    pp <- qnorm((1 - sk) / 2)
    sel <- discrete <= pp
    discrete[sel] <- 1 - sk
    discrete[!sel] <- -1 - sk
    sim <- mm + c(sqrt(vv) * (abs(base) * discrete))
  } else if(model == "TwoPieceTukeyh") {
    base <- t(simDD)
    base <- base * exp(tl * base^2 / 2)
    discrete <- dd[, , 1L]
    pp <- qnorm((1 - sk) / 2)
    sel <- discrete <= pp
    discrete[sel] <- 1 - sk
    discrete[!sel] <- -1 - sk
    sim <- mm + c(sqrt(vv) * (abs(base) * discrete))
  } else if(model == "TwoPieceBimodal") {
    sq <- .GeoSimulationSquareSums(dd, seq_len(k - 1L))
    alpha <- 2 * (bimo + 1) / (k - 1L)
    base <- sq / 2^(1 - alpha / 2)
    discrete <- dd[, , k]
    pp <- qnorm((1 - sk) / 2)
    sel <- discrete <= pp
    discrete[sel] <- 1 - sk
    discrete[!sel] <- -1 - sk
    sim <- mm + c(sqrt(vv) * base^(1 / alpha) * discrete)
  } else if(model == "TwoPieceStudentT") {
    denom <- sqrt(.GeoSimulationSquareSums(dd, seq_len(k - 2L)) / (k - 2L))
    base <- c(t(simDD)) / denom
    discrete <- dd[, , k]
    pp <- qnorm((1 - sk) / 2)
    sel <- discrete <= pp
    discrete[sel] <- 1 - sk
    discrete[!sel] <- -1 - sk
    sim <- mm + c(sqrt(vv) * (abs(base) * discrete))
  } else if(model %in% c("LogLogistic", "Logistic", "SkewLaplace")) {
    sim1 <- .GeoSimulationSquareSums(dd, 1:2) / 2
    sim2 <- .GeoSimulationSquareSums(dd, 3:4) / 2
    if(model == "LogLogistic") {
      sim <- exp(mm) * (sim1 / sim2)^(1 / param$shape) /
        (gamma(1 + 1 / param$shape) * gamma(1 - 1 / param$shape))
    } else if(model == "Logistic") {
      sim <- mm + log(sim1 / sim2) * (vv)^(0.5)
    } else {
      sim <- mm + (sim1 / param$skew - sim2 / (1 - param$skew)) * (vv)^(0.5)
    }
  } else if(model %in% c("Gamma", "Weibull")) {
    sq <- .GeoSimulationSquareSums(dd, seq_len(k))
    if(model == "Weibull") {
      sim <- exp(mm) * (sq / 2)^(1 / param$shape) / gamma(1 + 1 / param$shape)
    } else {
      sim <- exp(mm) * sq / k
    }
  } else if(model %in% c("Beta", "Kumaraswamy", "Kumaraswamy2")) {
    if(model == "Beta") {
      sim1 <- .GeoSimulationSquareSums(dd, seq_len(as.integer(param$shape1)))
      sim2 <- .GeoSimulationSquareSums(
        dd,
        seq.int(as.integer(param$shape1) + 1L,
                as.integer(param$shape1) + as.integer(param$shape2))
      )
      sim <- param$min + (param$max - param$min) * sim1 / (sim1 + sim2)
    } else {
      sim1 <- .GeoSimulationSquareSums(dd, 1:2)
      sim2 <- .GeoSimulationSquareSums(dd, 3:4)
      ratio <- sim1 / (sim1 + sim2)
      if(model == "Kumaraswamy") {
        sim <- param$min + (param$max - param$min) *
          (1 - ratio^(1 / param$shape1))^(1 / param$shape2)
      } else {
        med <- plogis(mm)
        shape1 <- as.numeric(param$shape)
        shape2 <- log(0.5) / log1p(-(med^shape1))
        if(any(!is.finite(shape2)) || any(shape2 <= 0)) {
          stop("Kumaraswamy2 produced invalid location-specific shape parameters.", call. = FALSE)
        }
        sim <- param$min + (param$max - param$min) *
          (1 - ratio^(1 / shape2))^(1 / shape1)
      }
    }
  } else if(model == "Wrapped") {
    if(spacetime) mm <- matrix(mm, nrow = nrow(sim), ncol = ncol(sim), byrow = TRUE)
    sim <- (sim + mm) %% (2 * pi)
    needs_format <- FALSE
  } else if(model == "Gaussian") {
    sim <- c(sim)
    byrow <- FALSE
  } else if(model %in% c("LogGaussian", "LogGauss")) {
    base <- c(t(sim))
    sim <- exp(mm) * (exp(sqrt(vv) * base) / exp(vv / 2))
  } else if(model == "Tukeygh") {
    base <- c(t(sim))
    if(!sk && !tl) sim <- mm + sqrt(vv) * base
    if(!sk && tl) sim <- mm + sqrt(vv) * base * exp(tl * base^2 / 2)
    if(!tl && sk) sim <- mm + sqrt(vv) * (exp(sk * base) - 1) / sk
    if(tl && sk) sim <- mm + sqrt(vv) * (exp(sk * base) - 1) * exp(0.5 * tl * base^2) / sk
  } else if(model == "Tukeyh") {
    base <- c(t(sim))
    if(!tl) sim <- mm + sqrt(vv) * base
    if(tl) sim <- mm + sqrt(vv) * base * exp(tl * base^2 / 2)
  } else if(model == "Tukeyh2") {
    base <- c(t(sim))
    sel <- base >= 0
    transformed <- numeric(length(base))
    transformed[sel] <- base[sel] * exp(t1l * base[sel]^2 / 2)
    transformed[!sel] <- base[!sel] * exp(t2l * base[!sel]^2 / 2)
    sim <- mm + sqrt(vv) * transformed
  } else if(model == "SinhAsinh") {
    base <- c(t(sim))
    sim <- mm + sqrt(vv) * sinh((1 / tl) * (asinh(base) + sk))
  } else {
    handled <- FALSE
  }

  list(sim = sim, byrow = byrow, needs_format = needs_format, handled = handled)
}

GeoCV <- function(fit, K = 100, estimation = TRUE,
                  optimizer = NULL, lower = NULL, upper = NULL,
                  n.fold = 0.05, local = FALSE, neighb = NULL, maxdist = NULL,
                  maxtime = NULL, sparse = FALSE, type_krig = "Simple",
                  which = 1, parallel = FALSE, ncores = 6,
                  progress = TRUE, seed = NULL) {
  .dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
  on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)

  ## ---------- sanity checks ----------
  if (!inherits(fit, "GeoFit"))
    stop("fit must be an object of class 'GeoFit'", call. = FALSE)
  if (!is.numeric(n.fold) || length(n.fold) != 1L || !is.finite(n.fold) ||
      n.fold < 0.01 || n.fold > 0.99)
    stop("n.fold must be between 0.01 and 0.99", call. = FALSE)
  if (!is.logical(estimation) || length(estimation) != 1L || is.na(estimation))
    stop("estimation must be TRUE or FALSE", call. = FALSE)
  if (!is.logical(local) || length(local) != 1L || is.na(local))
    stop("local must be TRUE or FALSE", call. = FALSE)
  if (!is.logical(parallel) || length(parallel) != 1L || is.na(parallel))
    stop("parallel must be TRUE or FALSE", call. = FALSE)
  if (!is.logical(progress) || length(progress) != 1L || is.na(progress))
    stop("progress must be TRUE or FALSE", call. = FALSE)
  if (!is.logical(sparse) || length(sparse) != 1L || is.na(sparse))
    stop("sparse must be TRUE or FALSE", call. = FALSE)

  fit_copula_phase1 <- if (is.null(fit$copula)) NULL else as.character(fit$copula)[1L]
  fit_model_code_phase1 <- suppressWarnings(as.integer(CkModel(fit$model)))
  phase1_count_copula <- !is.null(fit_copula_phase1) &&
    fit_copula_phase1 %in% c("Clayton", "SkewGaussian") &&
    fit_model_code_phase1 %in% .GeoGaussianCopulaDiscreteModels
  if (isTRUE(phase1_count_copula) && isTRUE(local))
    stop("GeoCV: Clayton/SkewGaussian copula with Poisson/Binomial/BinomialNeg is currently implemented for global GeoKrig (local=FALSE) only.", call. = FALSE)
  if (identical(fit_copula_phase1, "Clayton") &&
      fit_model_code_phase1 %in% .GeoGaussianCopulaDiscreteModels &&
      isTRUE(estimation))
    stop("GeoCV: fold refitting for discrete Clayton models still requires the Clayton phase-2 GeoFit likelihood. Use estimation=FALSE.", call. = FALSE)

  type_key <- tolower(gsub("[[:blank:]]", "", type_krig))
  if (!(type_key %in% c("simple", "universal")))
    stop("type_krig must be 'Simple' or 'Universal'", call. = FALSE)
  type_krig <- if (identical(type_key, "universal")) "Universal" else "Simple"

  tol_int <- sqrt(.Machine$double.eps)
  if (!is.numeric(K) || length(K) != 1L || !is.finite(K) || K < 2 ||
      abs(K - round(K)) > tol_int)
    stop("K must be an integer >= 2", call. = FALSE)
  K <- as.integer(K)
  if (K > 1000L) stop("K is too large", call. = FALSE)

  ncores <- .GeoValidateNcores(ncores)

  if (!is.null(seed)) {
    if (!is.numeric(seed) || length(seed) != 1L || !is.finite(seed) ||
        abs(seed - round(seed)) > tol_int || abs(seed) > .Machine$integer.max)
      stop("seed must be a finite integer scalar", call. = FALSE)
    seed <- as.integer(seed)
    old_seed <- if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
    } else {
      NULL
    }
    on.exit({
      if (is.null(old_seed)) {
        if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
          rm(".Random.seed", envir = .GlobalEnv)
      } else {
        assign(".Random.seed", old_seed, envir = .GlobalEnv)
      }
    }, add = TRUE)
    set.seed(seed)
  }

  if (parallel &&
      (!requireNamespace("future", quietly = TRUE) ||
       !requireNamespace("future.apply", quietly = TRUE))) {
    warning(
      "Parallel execution requested but future/future.apply is unavailable; using sequential execution.",
      call. = FALSE
    )
    parallel <- FALSE
  }

  has_progressr <- requireNamespace("progressr", quietly = TRUE)
  if (parallel && progress && !has_progressr) {
    warning(
      "Package 'progressr' is required for progress bars in parallel mode; disabling progress.",
      call. = FALSE
    )
    progress <- FALSE
  }

  message("Cross-validation kriging can be time consuming ...")

  ## ---------- model classification ----------
  corr_code <- CkCorrModel(fit$corrmodel)
  spacetime <- CheckST(corr_code)
  bivariate <- CheckBiv(corr_code)
  space <- !spacetime && !bivariate

  if (bivariate && !(which %in% c(1, 2)))
    stop("which must be 1 or 2", call. = FALSE)
  if (local && is.null(maxdist) && is.null(neighb))
    stop("maxdist or neighb is required for local kriging", call. = FALSE)

  fixed_base <- if (is.null(fit$fixed)) {
    list()
  } else if (is.list(fit$fixed)) {
    fit$fixed
  } else {
    as.list(fit$fixed)
  }
  target_mean_name <- if (bivariate) paste0("mean_", which) else "mean"
  target_external_mean <- fixed_base[[target_mean_name]]
  target_external_mean <- !is.null(target_external_mean) &&
    length(target_external_mean) > 1L
  if (identical(type_krig, "Universal") && target_external_mean) {
    warning(
      "The target mean is supplied externally and is treated as known; ",
      "GeoCV uses Simple kriging for these folds.",
      call. = FALSE
    )
    type_krig <- "Simple"
  }
  mean_pattern <- if (bivariate) "^mean_[12][0-9]*$" else "^mean[0-9]*$"
  estimated_mean_names <- if (is.null(names(fit$param))) character(0) else
    grep(mean_pattern, names(fit$param), value = TRUE)
  if (identical(type_krig, "Universal") && !length(estimated_mean_names)) {
    warning(
      "No mean coefficients are estimated in fit; Universal and Simple kriging coincide. ",
      "GeoCV uses Simple kriging.",
      call. = FALSE
    )
    type_krig <- "Simple"
  }

  full_likelihood <- identical(fit$likelihood, "Full")
  if (identical(type_krig, "Universal") && isTRUE(estimation) && !full_likelihood) {
    stop(
      "Universal cross-validation with estimation=TRUE is not available for composite-likelihood refits, because each fold would require its own Godambe covariance matrix. Use type_krig='Simple', set estimation=FALSE with a fit already updated by GeoVarest(), or compute fold-specific Godambe matrices outside GeoCV.",
      call. = FALSE
    )
  }
  if (identical(type_krig, "Universal") && !isTRUE(estimation) && is.null(fit$varcov)) {
    stop(
      "Universal cross-validation with estimation=FALSE requires fit$varcov. For a composite-likelihood fit, update the fitted object with GeoVarest() first.",
      call. = FALSE
    )
  }

  ## Keep the data-generating model (fit$model) distinct from the criterion
  ## used for refitting.  Current GeoFit objects store estimation_model
  ## explicitly; the fallback is only for older misspecified fits.
  model1 <- fit$estimation_model
  if (is.null(model1) || length(model1) != 1L || is.na(model1) || !nzchar(model1)) {
    model1 <- as.character(fit$model)[1L]
    if (isTRUE(fit$missp)) {
      fallback <- c(
        StudentT = "Gaussian_misp_StudentT",
        Poisson = "Gaussian_misp_Poisson",
        PoissonZIP = "Gaussian_misp_PoissonZIP",
        PoissonGamma = "Gaussian_misp_PoissonGamma",
        Binomial = "Gaussian_misp_Binomial",
        SkewStudentT = "Gaussian_misp_SkewStudentT",
        Tukeygh = "Gaussian_misp_Tukeygh"
      )
      if (identical(model1, "BinomialNeg")) {
        stop(
          "GeoCV: this older misspecified BinomialNeg fit does not record the working likelihood. Refit it with the current GeoFit implementation before cross-validation.",
          call. = FALSE
        )
      }
      if (!(model1 %in% names(fallback))) {
        stop(
          "GeoCV: cannot recover the misspecified estimation model from this older GeoFit object. Refit it with the current GeoFit implementation.",
          call. = FALSE
        )
      }
      model1 <- unname(fallback[[model1]])
    }
  } else {
    model1 <- as.character(model1)[1L]
  }

  ## Exact probability scores below use a Gaussian predictive distribution.
  gaussian_scores <- identical(gsub("[[:blank:]]", "", fit$model), "Gaussian")
  if (!gaussian_scores) {
    warning(
      "For non-Gaussian fitted models GeoCV reports RMSE, MAE and MAD only; ",
      "Gaussian probability scores (log-score, Brier, CRPS, PIT, interval score and coverage) are returned as NA.",
      call. = FALSE
    )
  }

  if (is.null(optimizer)) optimizer <- fit$optimizer
  if (is.null(lower)) lower <- fit$lower
  if (is.null(upper)) upper <- fit$upper

  fit_method <- "cholesky"
  if (!is.null(fit$method)) {
    fit_method <- fit$method
  } else if (!is.null(fit$call)) {
    method_expr <- as.list(fit$call)$method
    if (!is.null(method_expr)) {
      method_value <- tryCatch(
        eval(method_expr, envir = parent.frame()),
        error = function(e) NULL
      )
      if (is.character(method_value) && length(method_value) == 1L)
        fit_method <- method_value
    }
  }
  fit_method <- tolower(gsub("[[:blank:]]", "", fit_method))
  if (!(fit_method %in% c("cholesky", "svd"))) fit_method <- "cholesky"

  refit_varest <- identical(type_krig, "Universal") && full_likelihood
  maxtime_use <- if (is.null(maxtime)) fit$maxtime else maxtime

  ## angle/ratio are represented once through anisopars.  GeoFit will append
  ## them to the estimated/fixed parameter blocks according to est.aniso.
  refit_start <- fit$param
  refit_fixed_base <- fixed_base
  if (!is.null(fit$anisopars)) {
    anames <- intersect(c("angle", "ratio"), names(unlist(fit$anisopars, use.names = TRUE)))
    if (length(anames)) {
      if (!is.null(refit_start)) refit_start[intersect(names(refit_start), anames)] <- NULL
      if (!is.null(refit_fixed_base)) refit_fixed_base[intersect(names(refit_fixed_base), anames)] <- NULL
      if (!is.null(lower) && !is.null(names(lower))) lower[intersect(names(lower), anames)] <- NULL
      if (!is.null(upper) && !is.null(names(upper))) upper[intersect(names(upper), anames)] <- NULL
    }
  }

  ## Normalize count-size information once; the remaining fold utilities are
  ## shared in CVUtils.R rather than recreated inside every GeoCV call.
  n_cv <- .GeoCVNormalizeCountSize(fit$n, fit$model)

  ## ---------- spatial case ----------
  results <- NULL
  if (space) {
    coords <- .GeoCVExplicitCoordinates(fit, "GeoCV spatial coordinates")
    N <- nrow(coords)
    data <- .GeoKrigNormalizeData(
      fit$data, nspace = N, coordt = NULL, grid = isTRUE(fit$grid),
      context = "GeoCV spatial data"
    )
    Xfull <- .GeoCVFlattenDesign(fit$X, N, "GeoCV spatial")
    Mfull <- .GeoCVVaryingFixedMean(fixed_base, "mean", N)
    nfull <- .GeoCVFlattenObservationVector(n_cv, N, "fit$n")
    test_samples <- .GeoCVMakeTestSamples(N, K, n.fold)

    cv_iteration <- function(i) {
      test_idx <- test_samples[[i]]
      train_idx <- setdiff(seq_len(N), test_idx)

      data_est <- data[train_idx]
      data_pred <- data[test_idx]
      coords_est <- coords[train_idx, , drop = FALSE]
      coords_pred <- coords[test_idx, , drop = FALSE]
      X_est <- if (!is.null(Xfull)) Xfull[train_idx, , drop = FALSE] else NULL
      X_pred <- if (!is.null(Xfull)) Xfull[test_idx, , drop = FALSE] else NULL
      M_pred <- if (!is.null(Mfull)) Mfull[test_idx] else NULL
      n_est <- .GeoCVSubsetObservationVector(nfull, train_idx)
      n_pred <- .GeoCVSubsetObservationVector(nfull, test_idx)

      fixed_for_est <- refit_fixed_base
      if (!is.null(Mfull)) fixed_for_est$mean <- Mfull[train_idx]

      if (estimation) {
        fit_s <- .GeoCVRefit(
          fit = fit, data = data_est, X = X_est, n = n_est,
          fixed = fixed_for_est, start = refit_start, model = model1,
          optimizer = optimizer, lower = lower, upper = upper,
          sparse = sparse, method = fit_method, varest = refit_varest,
          coordx = coords_est,
          universal = identical(type_krig, "Universal")
        )
      } else {
        fit_s <- fit
        fit_s$data <- data_est
        fit_s$coordx <- coords_est[, 1L]
        fit_s$coordy <- coords_est[, 2L]
        fit_s$coordz <- if (ncol(coords_est) == 3L) coords_est[, 3L] else NULL
        fit_s$coordx_dyn <- NULL
        fit_s$grid <- FALSE
        fit_s$ns <- length(train_idx)
        fit_s$X <- X_est
        fit_s$n <- n_est
        fit_s$fixed <- fixed_for_est
      }

      pr <- .GeoCVPredict(
        fit_s, loc = coords_pred, Xloc = X_pred, Mloc = M_pred, nloc = n_pred,
        local = local, neighb = neighb, maxdist = maxdist, maxtime = maxtime_use,
        type_krig = type_krig, sparse = sparse, method = fit_method
      )

      .GeoCVSummarizeScores(data_pred, pr$pred, pr$mse, gaussian_scores)
    }

    results <- .GeoCVRunIterations(cv_iteration, K, parallel, ncores, progress)
  }

  ## ---------- space-time case ----------
  if (spacetime) {
    coordt <- as.numeric(fit$coordt)
    if (!length(coordt) || any(!is.finite(coordt)))
      stop("fit$coordt is invalid", call. = FALSE)
    T <- length(coordt)

    if (!is.null(fit$coordx_dyn)) {
      coord_blocks <- .GeoNormalizeDynamicCoordinates(
        fit$coordx_dyn, coordt = coordt, context = "GeoCV space-time"
      )
      ns <- vapply(coord_blocks, nrow, integer(1))
      if (is.list(fit$data)) {
        if (length(fit$data) != T)
          stop("GeoCV: fit$data is not aligned with fit$coordt.", call. = FALSE)
        data_blocks <- lapply(seq_len(T), function(k) {
          z <- as.numeric(fit$data[[k]])
          if (length(z) != ns[k] || any(!is.finite(z)))
            stop("GeoCV: dynamic data block ", k, " is invalid.", call. = FALSE)
          z
        })
      } else {
        data_flat0 <- as.numeric(fit$data)
        if (length(data_flat0) != sum(ns) || any(!is.finite(data_flat0)))
          stop("GeoCV: dynamic data are not aligned with coordx_dyn.", call. = FALSE)
        data_blocks <- split(data_flat0, rep(seq_len(T), ns))
      }
    } else {
      coords_fixed <- .GeoCVExplicitCoordinates(fit, "GeoCV fixed space-time coordinates")
      ns <- rep(nrow(coords_fixed), T)
      data_flat0 <- .GeoKrigNormalizeData(
        fit$data, nspace = nrow(coords_fixed), coordt = coordt,
        grid = isTRUE(fit$grid), context = "GeoCV fixed space-time data"
      )
      data_blocks <- split(data_flat0, rep(seq_len(T), ns))
      coord_blocks <- lapply(seq_len(T), function(k) coords_fixed)
    }

    NT <- sum(ns)
    time_group <- rep(seq_len(T), ns)
    coords_flat <- do.call(rbind, coord_blocks)
    data_flat <- as.numeric(unlist(data_blocks, use.names = FALSE))
    Xflat <- .GeoCVFlattenDesign(fit$X, ns, "GeoCV space-time")
    Mflat <- .GeoCVVaryingFixedMean(fixed_base, "mean", NT)
    if (!is.null(Mflat) && !is.null(Xflat))
      stop("GeoCV: covariates and a fixed varying mean are not compatible.", call. = FALSE)
    nflat <- .GeoCVFlattenObservationVector(n_cv, ns, "fit$n")
    test_samples <- .GeoCVMakeTestSamples(NT, K, n.fold, groups = time_group)

    cv_iteration_st <- function(i) {
      test_idx <- test_samples[[i]]
      train_idx <- setdiff(seq_len(NT), test_idx)

      train_by_time <- lapply(seq_len(T), function(k)
        train_idx[time_group[train_idx] == k])
      test_by_time <- lapply(seq_len(T), function(k)
        test_idx[time_group[test_idx] == k])

      coord_train <- lapply(train_by_time, function(idx)
        coords_flat[idx, , drop = FALSE])
      data_train <- lapply(train_by_time, function(idx) data_flat[idx])
      X_train <- if (!is.null(Xflat)) {
        lapply(train_by_time, function(idx) Xflat[idx, , drop = FALSE])
      } else NULL
      M_train <- if (!is.null(Mflat)) Mflat[unlist(train_by_time)] else NULL
      n_train <- .GeoCVSubsetObservationVector(nflat, unlist(train_by_time))

      fixed_for_est <- refit_fixed_base
      if (!is.null(Mflat)) fixed_for_est$mean <- M_train

      if (estimation) {
        fit_s <- .GeoCVRefit(
          fit = fit, data = data_train, X = X_train, n = n_train,
          fixed = fixed_for_est, start = refit_start, model = model1,
          optimizer = optimizer, lower = lower, upper = upper,
          sparse = sparse, method = fit_method, varest = refit_varest,
          coordx_dyn = coord_train, coordt = coordt,
          universal = identical(type_krig, "Universal")
        )
      } else {
        fit_s <- fit
        fit_s$data <- data_train
        fit_s$coordx_dyn <- coord_train
        fit_s$coordt <- coordt
        fit_s$coordx <- unlist(lapply(coord_train, function(z) z[, 1L]))
        fit_s$coordy <- unlist(lapply(coord_train, function(z) z[, 2L]))
        fit_s$coordz <- if (ncol(coords_flat) == 3L) {
          unlist(lapply(coord_train, function(z) z[, 3L]))
        } else NULL
        fit_s$grid <- FALSE
        fit_s$ns <- vapply(coord_train, nrow, integer(1))
        fit_s$X <- X_train
        fit_s$n <- n_train
        fit_s$fixed <- fixed_for_est
      }

      pred_values <- pred_mse <- obs_values <- list()
      out_pos <- 0L
      for (k in seq_len(T)) {
        idx <- test_by_time[[k]]
        if (!length(idx)) next
        out_pos <- out_pos + 1L
        loc_k <- coords_flat[idx, , drop = FALSE]
        Xloc_k <- if (!is.null(Xflat)) Xflat[idx, , drop = FALSE] else NULL
        Mloc_k <- if (!is.null(Mflat)) Mflat[idx] else NULL
        nloc_k <- .GeoCVSubsetObservationVector(nflat, idx)

        pr <- .GeoCVPredict(
          fit_s, loc = loc_k, Xloc = Xloc_k, Mloc = Mloc_k, nloc = nloc_k,
          local = local, neighb = neighb, maxdist = maxdist, maxtime = maxtime_use,
          type_krig = type_krig, sparse = sparse, method = fit_method,
          time = coordt[k]
        )
        pred_values[[out_pos]] <- pr$pred
        pred_mse[[out_pos]] <- if (is.null(pr$mse)) {
          rep(NA_real_, length(pr$pred))
        } else pr$mse
        obs_values[[out_pos]] <- data_flat[idx]
      }

      .GeoCVSummarizeScores(
        unlist(obs_values, use.names = FALSE),
        unlist(pred_values, use.names = FALSE),
        unlist(pred_mse, use.names = FALSE),
        gaussian_scores
      )
    }

    results <- .GeoCVRunIterations(cv_iteration_st, K, parallel, ncores, progress)
  }

  ## ---------- bivariate case ----------
  if (bivariate) {
    ns <- as.integer(fit$ns)
    if (length(ns) != 2L || any(!is.finite(ns)) || any(ns < 1L))
      stop("GeoCV: invalid bivariate block sizes.", call. = FALSE)
    if (ns[which] < 2L)
      stop("GeoCV requires at least two observations in the target bivariate component.", call. = FALSE)

    if (!is.null(fit$coordx_dyn)) {
      if (length(fit$coordx_dyn) != 2L)
        stop("GeoCV: bivariate coordx_dyn must contain two coordinate matrices.", call. = FALSE)
      coords1 <- as.matrix(fit$coordx_dyn[[1L]])
      coords2 <- as.matrix(fit$coordx_dyn[[2L]])
    } else {
      coords_common <- .GeoCVExplicitCoordinates(fit, "GeoCV bivariate coordinates")
      coords1 <- coords2 <- coords_common
    }
    if (nrow(coords1) != ns[1L] || nrow(coords2) != ns[2L])
      stop("GeoCV: bivariate coordinates are not aligned with fit$ns.", call. = FALSE)

    biv_data <- .GeoBivariateDataVector(
      fit$data, ns, context = "GeoCV bivariate"
    )
    data1 <- biv_data[seq_len(ns[1L])]
    data2 <- biv_data[ns[1L] + seq_len(ns[2L])]

    Xflat <- if (!is.null(fit$X)) {
      .GeoBivariateDesign(fit$X, ns, context = "GeoCV bivariate")
    } else NULL
    X1 <- if (!is.null(Xflat)) Xflat[seq_len(ns[1L]), , drop = FALSE] else NULL
    X2 <- if (!is.null(Xflat)) Xflat[ns[1L] + seq_len(ns[2L]), , drop = FALSE] else NULL

    M1 <- .GeoCVVaryingFixedMean(fixed_base, "mean_1", ns[1L])
    M2 <- .GeoCVVaryingFixedMean(fixed_base, "mean_2", ns[2L])
    if ((!is.null(M1) || !is.null(M2)) && !is.null(Xflat))
      stop("GeoCV: bivariate covariates and fixed varying means are not compatible.", call. = FALSE)

    ## GeoKrig's bivariate path uses coefficient-based observed means.  Legacy
    ## fitted objects may nevertheless contain site-specific mean_1 and/or
    ## mean_2 vectors.  Handle each component independently by centering that
    ## component and fixing its coefficient mean to zero; Mloc adds the known
    ## target mean back at prediction sites.
    data1_model <- if (!is.null(M1)) data1 - M1 else data1
    data2_model <- if (!is.null(M2)) data2 - M2 else data2
    fixed_biv <- refit_fixed_base
    start_biv <- refit_start
    if (!is.null(M1)) {
      fixed_biv$mean_1 <- 0
      start_biv <- .GeoCVDropMeanParameters(start_biv, "mean_1")
    }
    if (!is.null(M2)) {
      fixed_biv$mean_2 <- 0
      start_biv <- .GeoCVDropMeanParameters(start_biv, "mean_2")
    }

    nflat <- .GeoCVFlattenObservationVector(n_cv, ns, "fit$n")
    n1 <- if (length(nflat) == 1L) nflat else nflat[seq_len(ns[1L])]
    n2 <- if (length(nflat) == 1L) nflat else nflat[ns[1L] + seq_len(ns[2L])]

    target_n <- ns[which]
    test_samples <- .GeoCVMakeTestSamples(target_n, K, n.fold)

    cv_iteration_biv <- function(i) {
      test_idx <- test_samples[[i]]
      train_idx <- setdiff(seq_len(target_n), test_idx)
      current_fixed <- fixed_biv

      if (which == 1L) {
        data_est <- list(data1_model[train_idx], data2_model)
        data_pred <- data1[test_idx]
        coords_est <- list(coords1[train_idx, , drop = FALSE], coords2)
        coords_pred <- coords1[test_idx, , drop = FALSE]
        X_est <- if (!is.null(Xflat)) {
          rbind(X1[train_idx, , drop = FALSE], X2)
        } else NULL
        X_pred <- if (!is.null(X1)) X1[test_idx, , drop = FALSE] else NULL
        M_pred <- if (!is.null(M1)) M1[test_idx] else NULL
        n_est <- if (length(nflat) == 1L) nflat else c(n1[train_idx], n2)
        n_pred <- if (length(nflat) == 1L) nflat else n1[test_idx]
      } else {
        data_est <- list(data1_model, data2_model[train_idx])
        data_pred <- data2[test_idx]
        coords_est <- list(coords1, coords2[train_idx, , drop = FALSE])
        coords_pred <- coords2[test_idx, , drop = FALSE]
        X_est <- if (!is.null(Xflat)) {
          rbind(X1, X2[train_idx, , drop = FALSE])
        } else NULL
        X_pred <- if (!is.null(X2)) X2[test_idx, , drop = FALSE] else NULL
        M_pred <- if (!is.null(M2)) M2[test_idx] else NULL
        n_est <- if (length(nflat) == 1L) nflat else c(n1, n2[train_idx])
        n_pred <- if (length(nflat) == 1L) nflat else n2[test_idx]
      }

      if (estimation) {
        fit_s <- .GeoCVRefit(
          fit = fit, data = data_est, X = X_est, n = n_est,
          fixed = current_fixed, start = start_biv, model = model1,
          optimizer = optimizer, lower = lower, upper = upper,
          sparse = sparse, method = fit_method, varest = refit_varest,
          coordx_dyn = coords_est,
          universal = identical(type_krig, "Universal")
        )
      } else {
        fit_s <- fit
        fit_s$data <- data_est
        fit_s$coordx_dyn <- coords_est
        fit_s$coordx <- fit_s$coordy <- fit_s$coordz <- NULL
        fit_s$grid <- FALSE
        fit_s$ns <- vapply(coords_est, nrow, integer(1))
        fit_s$X <- X_est
        fit_s$n <- n_est
        fit_s$fixed <- current_fixed
      }

      pr <- .GeoCVPredict(
        fit_s, loc = coords_pred, Xloc = X_pred, Mloc = M_pred, nloc = n_pred,
        local = local, neighb = neighb, maxdist = maxdist, maxtime = maxtime_use,
        type_krig = type_krig, sparse = sparse, method = fit_method,
        which = which
      )

      .GeoCVSummarizeScores(data_pred, pr$pred, pr$mse, gaussian_scores)
    }

    results <- .GeoCVRunIterations(cv_iteration_biv, K, parallel, ncores, progress)
  }

  if (is.null(results))
    stop("Unsupported GeoFit configuration for GeoCV.", call. = FALSE)

  list(
    rmse = results[, "rmse"],
    mae = results[, "mae"],
    mad = results[, "mad"],
    brie = results[, "brie"],
    crps = results[, "crps"],
    lscore = results[, "lscore"],
    pit_mean = results[, "pit_mean"],
    pit = results[, "pit_mean"],
    intscore = results[, "intscore"],
    coverage = results[, "coverage"],
    seed = seed,
    estimation_model = model1
  )
}

########################################
## Parametric bootstrap test for spatial support
########################################
GeoTestsupp_space <- function(data, coordx,
                              start, fixed,
                              model = "Gaussian",
                              h0 = NULL,
                              optimizer = "bobyqa",
                              lower = NULL, upper = NULL,
                              neighb = 5,
                              B = 10000,
                              likelihood = NULL,
                              type = NULL,
                              method = c("cholesky", "TB"),
                              L = 10000,
                              parallel = FALSE,
                              ncores = 6,
                              progress = TRUE,
                              seed = NULL) {

  ## ====== preserve runtime settings and optional RNG seed ======
  cleanup_resources <- .GeoTestPreserveRuntime(gc_on_exit = FALSE)
  on.exit(cleanup_resources(), add = TRUE)
  restore_seed <- .GeoTestSetSeed(seed)
  on.exit(restore_seed(), add = TRUE)

  ## ====== input validation ======
  method <- match.arg(method)

  if (!is.matrix(coordx)) coordx <- as.matrix(coordx)
  if (!is.numeric(data)) data <- as.numeric(data)

  if (!is.numeric(coordx) || any(!is.finite(coordx))) {
    stop("coordx must contain only finite numeric values", call. = FALSE)
  }
  if (any(!is.finite(data))) {
    stop("data must contain only finite numeric values", call. = FALSE)
  }
  if (nrow(coordx) != length(data)) {
    stop(
      "coordx rows (", nrow(coordx), ") must match data length (",
      length(data), ")",
      call. = FALSE
    )
  }
  if (nrow(coordx) < 2L) {
    stop("At least two spatial observations are required", call. = FALSE)
  }

  if (!is.list(start) || length(start) < 1L ||
      is.null(names(start)) || any(names(start) == "")) {
    stop("start must be a non-empty named list", call. = FALSE)
  }

  if (is.null(fixed)) fixed <- list()
  if (!is.list(fixed)) {
    stop("fixed must be NULL or a named list", call. = FALSE)
  }
  if (length(fixed) > 0L &&
      (is.null(names(fixed)) || any(names(fixed) == ""))) {
    stop("fixed must be NULL or a named list", call. = FALSE)
  }

  if (anyDuplicated(names(start)) || anyDuplicated(names(fixed))) {
    stop("start and fixed cannot contain duplicated parameter names", call. = FALSE)
  }

  overlap <- intersect(names(start), names(fixed))
  if (length(overlap) > 0L) {
    stop(
      "Parameters cannot be present in both start and fixed: ",
      paste(overlap, collapse = ", "),
      call. = FALSE
    )
  }

  if (!is.numeric(B) || length(B) != 1L || !is.finite(B) || B < 1) {
    stop("B must be a positive integer", call. = FALSE)
  }
  B <- as.integer(B)

  if (!is.numeric(L) || length(L) != 1L || !is.finite(L) || L < 1) {
    stop("L must be a positive integer", call. = FALSE)
  }
  L <- as.integer(L)

  if (!is.logical(progress) || length(progress) != 1L) {
    stop("progress must be logical (TRUE or FALSE)", call. = FALSE)
  }
  if (!is.logical(parallel) || length(parallel) != 1L) {
    stop("parallel must be logical (TRUE or FALSE)", call. = FALSE)
  }
  if (!is.null(neighb) &&
      (!is.numeric(neighb) || length(neighb) != 1L ||
       !is.finite(neighb) || neighb < 1)) {
    stop("neighb must be a positive integer or NULL", call. = FALSE)
  }
  if (!is.null(neighb)) neighb <- as.integer(round(neighb))

  if (!is.null(h0) &&
      (!is.numeric(h0) || length(h0) != 1L || !is.finite(h0) || h0 <= 0)) {
    stop("h0 must be NULL or a positive finite number", call. = FALSE)
  }
  if (xor(is.null(likelihood), is.null(type))) {
    stop(
      "Either specify both likelihood and type, or leave both NULL for automatic selection.",
      call. = FALSE
    )
  }

  ## ====== constants ======
  max_scale_cap <- 1e6
  min_box_width <- 1e-3
  corrmodel <- "GenWend"
  eps <- 1e-6
  NN <- nrow(coordx)

  ## ====== progress handlers ======
  use_progressr <- .GeoTestConfigureProgress(progress)

  ## ====== automatic likelihood selection ======
  if (is.null(likelihood) && is.null(type)) {
    if (NN > 2000L) {
      likelihood <- "Marginal"
      type <- "Pairwise"
    } else {
      likelihood <- "Full"
      type <- "Standard"
      neighb <- NULL
    }
  }

  if (!(likelihood %in% c("Full", "Marginal", "Conditional"))) {
    stop("Unsupported likelihood", call. = FALSE)
  }
  if (identical(likelihood, "Full") && !identical(type, "Standard")) {
    stop("Full likelihood must be coupled with type = 'Standard'", call. = FALSE)
  }
  if (!identical(likelihood, "Full") && !identical(type, "Pairwise")) {
    stop("Composite likelihood must be coupled with type = 'Pairwise'", call. = FALSE)
  }

  ## ====== robust core handling ======
  core_cfg <- .GeoTestCoreConfig(parallel, ncores, B, strict_positive = FALSE)
  parallel <- core_cfg$parallel
  ncores <- core_cfg$ncores

  ## ====== compute d_min and h0 ======
  nn <- nabor::knn(coordx, coordx, k = 2)
  d_min <- min(nn$nn.dists[, 2])

  if (!is.finite(d_min) || d_min <= 0) {
    stop(
      "The minimum inter-site distance is not positive; check duplicated coordinates.",
      call. = FALSE
    )
  }

  h0_val <- if (is.null(h0)) max(eps, d_min - eps) else as.numeric(h0)
  if (!is.finite(h0_val) || h0_val <= eps) {
    stop("Invalid h0: it must be greater than the numerical tolerance", call. = FALSE)
  }

  ## ====== finite fallback bounds ======
  dat_sd <- try(stats::sd(data), silent = TRUE)
  if (inherits(dat_sd, "try-error") || !is.finite(dat_sd)) dat_sd <- 1
  dat_sd <- max(dat_sd, 1e-6)
  BIG_M <- max(1e3, 100 * dat_sd)
  SILL_MIN <- 1e-8

  mins <- apply(coordx, 2, min)
  maxs <- apply(coordx, 2, max)
  bbox_diag <- sqrt(sum((maxs - mins)^2))
  U_big <- min(max(10 * h0_val, 5 * bbox_diag, 1e-4), max_scale_cap)

  if (U_big <= h0_val) {
    stop(
      "The practical upper bound for scale must be greater than h0. ",
      "Reduce h0 or increase max_scale_cap in the implementation.",
      call. = FALSE
    )
  }

  msg <- sprintf("Testing H0: compact support <= %.5f", h0_val)
  if (is.null(h0)) msg <- paste0(msg, " (independence test)")
  message(msg)

  ## ====== parameter validation ======
  corr_names <- CorrParam(corrmodel)
  nuis_names <- NuisParam(model)
  expected <- sort(unique(c(corr_names, nuis_names)))
  provided <- sort(c(names(start), names(fixed)))

  if (!identical(expected, provided)) {
    missing <- setdiff(expected, provided)
    extra <- setdiff(provided, expected)
    msg_parts <- character(0)
    if (length(missing)) {
      msg_parts <- c(msg_parts, paste0("missing: ", paste(missing, collapse = ", ")))
    }
    if (length(extra)) {
      msg_parts <- c(msg_parts, paste0("extra: ", paste(extra, collapse = ", ")))
    }
    stop(
      "Parameter names in start+fixed must match CorrParam(corrmodel)+NuisParam(model). ",
      paste(msg_parts, collapse = " | "),
      call. = FALSE
    )
  }

  if (!("scale" %in% expected)) {
    stop("The selected correlation model does not contain a scale parameter", call. = FALSE)
  }

  ## The unrestricted model must estimate scale. If the user supplied it
  ## in fixed, use that value as its starting value and remove it from fixed.
  base_start <- start
  base_fixed <- fixed
  if (!("scale" %in% names(base_start))) {
    if (!("scale" %in% names(base_fixed))) {
      stop("scale must be supplied in either start or fixed", call. = FALSE)
    }
    base_start$scale <- base_fixed$scale
    base_fixed$scale <- NULL
  }

  ## ====== bounds utilities ======
  norm_bounds <- function(b, which = c("lower", "upper"), start_names) {
    .GeoTestNormalizeBounds(b, which, start_names, require_list = TRUE)
  }

  widen_boxes <- function(lo, up, starts = NULL, positive = character()) {
    nms <- union(names(lo), names(up))

    for (nm in nms) {
      if (is.null(lo[[nm]]) || !is.finite(lo[[nm]])) {
        lo[[nm]] <- if (nm %in% positive) SILL_MIN else -BIG_M
      }
      if (is.null(up[[nm]]) || !is.finite(up[[nm]])) {
        up[[nm]] <- BIG_M
      }

      w <- up[[nm]] - lo[[nm]]
      if (!is.finite(w) || w <= 0) {
        mid <- if (!is.null(starts) && !is.null(starts[[nm]]) &&
                   is.finite(starts[[nm]])) {
          starts[[nm]]
        } else {
          0
        }
        lo[[nm]] <- mid - min_box_width / 2
        up[[nm]] <- mid + min_box_width / 2
      } else if (w < min_box_width) {
        extra <- (min_box_width - w) / 2
        lo[[nm]] <- lo[[nm]] - extra
        up[[nm]] <- up[[nm]] + extra
      }

      if (nm %in% positive) {
        if (lo[[nm]] < eps) lo[[nm]] <- eps
        if (up[[nm]] <= lo[[nm]]) up[[nm]] <- lo[[nm]] + min_box_width
      }
    }

    list(lower = lo, upper = up)
  }

  clamp_start <- function(starts, lo, up) {
    for (nm in names(starts)) {
      x <- as.numeric(starts[[nm]])
      l <- as.numeric(lo[[nm]])
      u <- as.numeric(up[[nm]])

      if (!is.finite(x)) x <- (l + u) / 2

      gap <- max(1e-10, 1e-8 * max(1, abs(l), abs(u)))
      if (x <= l) x <- l + gap
      if (x >= u) x <- u - gap

      if (!(x > l && x < u)) x <- (l + u) / 2
      starts[[nm]] <- x
    }
    starts
  }

  ## ====== fit helpers ======
  do_fit <- function(data_, starts, fixeds, lows, ups) {
    fixed_arg <- if (length(fixeds) > 0L) fixeds else NULL

    GeoFit(
      data = data_,
      coordx = coordx,
      corrmodel = corrmodel,
      model = model,
      start = starts,
      fixed = fixed_arg,
      optimizer = optimizer,
      lower = lows,
      upper = ups,
      neighb = neighb,
      likelihood = likelihood,
      type = type
    )
  }

  loglik_value <- .GeoTestLogLikValue

  same_pairs <- function(f0, f1) .GeoTestSamePairs(f0, f1, type)

  make_start_from_fit <- function(fit_obj, template, lo, up,
                                  force_scale_above_h0 = FALSE) {
    ans <- template
    est <- unlist(fit_obj$param, use.names = TRUE)
    common <- intersect(names(ans), names(est))
    ans[common] <- as.list(est[common])

    if (force_scale_above_h0 && "scale" %in% names(ans)) {
      scale_try <- max(
        as.numeric(ans$scale),
        h0_val * (1 + 1e-3),
        as.numeric(lo$scale) * (1 + 1e-3)
      )
      ans$scale <- min(scale_try, as.numeric(up$scale) * (1 - 1e-8))
    }

    clamp_start(ans, lo, up)
  }

  fit_unrestricted <- function(data_, start_primary, fixed1, lower1, upper1,
                               fit0_ref = NULL, ll0_ref = NA_real_) {
    candidates <- list()

    first <- suppressWarnings(
      try(do_fit(data_, start_primary, fixed1, lower1, upper1), silent = TRUE)
    )
    if (is.finite(loglik_value(first))) candidates[[length(candidates) + 1L]] <- first

    best_ll <- if (length(candidates)) {
      max(vapply(candidates, loglik_value, numeric(1)))
    } else {
      -Inf
    }

    tol <- 1e-8 * max(1, abs(ll0_ref), abs(best_ll))
    need_fallback <- !is.finite(best_ll) ||
      (is.finite(ll0_ref) && best_ll < ll0_ref - tol)

    fallback_used <- FALSE

    if (need_fallback && !is.null(fit0_ref)) {
      fallback_used <- TRUE
      start_fallback <- make_start_from_fit(
        fit0_ref,
        template = start_primary,
        lo = lower1,
        up = upper1,
        force_scale_above_h0 = TRUE
      )

      retry <- suppressWarnings(
        try(do_fit(data_, start_fallback, fixed1, lower1, upper1), silent = TRUE)
      )
      if (is.finite(loglik_value(retry))) {
        candidates[[length(candidates) + 1L]] <- retry
      }
    }

    if (!length(candidates)) {
      return(list(fit = NULL, fallback = fallback_used, valid = FALSE))
    }

    ll <- vapply(candidates, loglik_value, numeric(1))
    best <- candidates[[which.max(ll)]]
    best_ll <- max(ll)
    tol <- 1e-8 * max(1, abs(ll0_ref), abs(best_ll))

    valid <- !is.finite(ll0_ref) || best_ll >= ll0_ref - tol

    list(fit = best, fallback = fallback_used, valid = valid)
  }

  ## ====== construct H0 and H1 parameterizations ======
  start1 <- base_start
  fixed1 <- base_fixed

  lower1 <- norm_bounds(lower, "lower", start_names = sort(names(start1)))
  upper1 <- norm_bounds(upper, "upper", start_names = sort(names(start1)))

  if (is.null(lower1$scale) || !is.finite(lower1$scale) || lower1$scale <= 0) {
    lower1$scale <- eps
  }
  if (is.null(upper1$scale) || !is.finite(upper1$scale) || upper1$scale <= 0) {
    upper1$scale <- U_big
  }
  upper1$scale <- min(as.numeric(upper1$scale), U_big)

  wb1 <- widen_boxes(
    lower1,
    upper1,
    starts = start1,
    positive = intersect(c("sill", "scale"), names(start1))
  )
  lower1 <- wb1$lower
  upper1 <- wb1$upper

  lower1$scale <- max(as.numeric(lower1$scale), eps)
  upper1$scale <- min(as.numeric(upper1$scale), U_big)

  if (lower1$scale >= h0_val) {
    stop(
      "The lower bound for scale under H1 must be smaller than h0 so that H0 is nested in H1.",
      call. = FALSE
    )
  }
  if (upper1$scale <= h0_val) {
    stop(
      "The upper bound for scale under H1 must be greater than h0.",
      call. = FALSE
    )
  }

  start1 <- clamp_start(start1, lower1, upper1)

  start0 <- start1
  fixed0 <- fixed1

  if (is.null(h0)) {
    start0$scale <- NULL
    fixed0$scale <- h0_val

    lower0 <- lower1[names(start0)]
    upper0 <- upper1[names(start0)]
  } else {
    lower0 <- lower1
    upper0 <- upper1
    upper0$scale <- min(as.numeric(upper0$scale), h0_val)
    lower0$scale <- max(as.numeric(lower0$scale), eps)

    if (lower0$scale >= upper0$scale) {
      stop(
        "The bounds under H0 do not define a non-empty interval for scale.",
        call. = FALSE
      )
    }

    start0 <- clamp_start(start0, lower0, upper0)
  }

  ## ====== observed fits ======
  fit0 <- suppressWarnings(
    try(do_fit(data, start0, fixed0, lower0, upper0), silent = TRUE)
  )
  ll0 <- loglik_value(fit0)
  if (!is.finite(ll0)) {
    stop("The observed fit under H0 failed", call. = FALSE)
  }

  unrestricted_obs <- fit_unrestricted(
    data_ = data,
    start_primary = start1,
    fixed1 = fixed1,
    lower1 = lower1,
    upper1 = upper1,
    fit0_ref = fit0,
    ll0_ref = ll0
  )

  if (!isTRUE(unrestricted_obs$valid) || is.null(unrestricted_obs$fit)) {
    stop(
      "The observed unrestricted fit failed to attain a log-likelihood at least as large as H0.",
      call. = FALSE
    )
  }

  fit1 <- unrestricted_obs$fit
  ll1 <- loglik_value(fit1)

  if (!same_pairs(fit0, fit1)) {
    stop(
      "H0 and H1 must use the same pairs for the composite-likelihood ratio.",
      call. = FALSE
    )
  }

  ll_tol_obs <- 1e-8 * max(1, abs(ll0), abs(ll1))
  Lambda_obs <- if (ll1 < ll0 && ll1 >= ll0 - ll_tol_obs) {
    0
  } else {
    2 * (ll1 - ll0)
  }

  if (!is.finite(Lambda_obs) || Lambda_obs < -ll_tol_obs) {
    stop("Invalid observed likelihood-ratio statistic", call. = FALSE)
  }
  Lambda_obs <- max(0, Lambda_obs)

  ## ====== early exit if unrestricted estimate satisfies H0 ======
  scale_hat <- unlist(c(fit1$param, fit1$fixed), use.names = TRUE)["scale"]
  if (length(scale_hat) == 1L && is.finite(scale_hat) &&
      scale_hat <= h0_val + 1e-10 * max(1, h0_val)) {
    cat("Unrestricted fit satisfies H0 (bootstrap not needed)\n")

    return(invisible(list(
      d_min = round(d_min, 5),
      h0 = round(h0_val, 5),
      lambda_obs = 0,
      pvalue = 1,
      seed = seed,
      method = method,
      B_requested = B,
      bootstrap_successful = 0L,
      bootstrap_success_rate = NA_real_,
      fallback_observed = unrestricted_obs$fallback,
      fallback_bootstrap = 0L,
      B_rep = numeric(0),
      fit_H0 = fit0,
      fit_H1 = fit1
    )))
  }

  ## ====== H0 parameters for simulation ======
  message("Running parametric bootstrap (B = ", B, ") ...")
  parH0 <- c(fit0$param, fit0$fixed)
  param_H0 <- as.list(parH0[expected])

  ## Use observed estimates as bootstrap starting values.
  start0_boot <- as.list(fit0$param)
  start1_boot <- as.list(fit1$param)
  start0_boot <- clamp_start(start0_boot, lower0, upper0)
  start1_boot <- clamp_start(start1_boot, lower1, upper1)

  ## ====== simulations ======
  if (method == "cholesky") {
    data_sim <- GeoModels::GeoSim(
      coordx = coordx,
      corrmodel = corrmodel,
      param = param_H0,
      model = model,
      sparse = TRUE,
      nrep = B,
      progress = FALSE
    )
  } else {
    old_plan_after_sim <- NULL
    if (isTRUE(parallel) && requireNamespace("future", quietly = TRUE)) {
      old_plan_after_sim <- future::plan()
    }

    data_sim <- GeoModels::GeoSimapprox(
      coordx = coordx,
      corrmodel = corrmodel,
      param = param_H0,
      model = model,
      method = "TB",
      L = L,
      nrep = B,
      parallel = parallel,
      ncores = ncores,
      progress = FALSE
    )

    if (!is.null(old_plan_after_sim)) {
      try(future::plan(old_plan_after_sim), silent = TRUE)
    }
  }

  ## ====== bootstrap helper ======
  estimate_fun <- function(Zb) {
    f0 <- suppressWarnings(
      try(do_fit(Zb, start0_boot, fixed0, lower0, upper0), silent = TRUE)
    )
    ll0b <- loglik_value(f0)

    if (!is.finite(ll0b)) {
      return(c(lambda = NA_real_, fallback = 0))
    }

    unrestricted <- fit_unrestricted(
      data_ = Zb,
      start_primary = start1_boot,
      fixed1 = fixed1,
      lower1 = lower1,
      upper1 = upper1,
      fit0_ref = f0,
      ll0_ref = ll0b
    )

    if (!isTRUE(unrestricted$valid) || is.null(unrestricted$fit)) {
      return(c(lambda = NA_real_, fallback = as.numeric(unrestricted$fallback)))
    }

    f1 <- unrestricted$fit
    ll1b <- loglik_value(f1)

    if (!same_pairs(f0, f1)) {
      return(c(lambda = NA_real_, fallback = as.numeric(unrestricted$fallback)))
    }

    tol <- 1e-8 * max(1, abs(ll0b), abs(ll1b))
    if (ll1b < ll0b - tol) {
      return(c(lambda = NA_real_, fallback = as.numeric(unrestricted$fallback)))
    }

    lambda <- if (ll1b < ll0b) 0 else 2 * (ll1b - ll0b)
    c(lambda = max(0, lambda), fallback = as.numeric(unrestricted$fallback))
  }

  ## ====== sequential or parallel bootstrap ======
  if (!parallel) {
    message(sprintf("Performing %d estimations under H0 and H1 sequentially...", B))
    result_list <- .GeoTestSequentialApply(
      data_sim, B, estimate_fun, progress = use_progressr
    )
  } else if (!requireNamespace("future", quietly = TRUE) ||
             !requireNamespace("future.apply", quietly = TRUE)) {
    warning(
      "Parallel bootstrap requested but future/future.apply is unavailable; using sequential evaluation.",
      call. = FALSE
    )
    result_list <- .GeoTestSequentialApply(
      data_sim, B, estimate_fun, progress = FALSE
    )
  } else {
    message(sprintf(
      "Performing %d estimations under H0 and H1 using %d cores...",
      B, ncores
    ))
    data_files <- .GeoTestSaveReplicates(data_sim, B, "supp_space")
    on.exit(unlink(data_files), add = TRUE)
    rm(data_sim)
    gc(verbose = FALSE, full = TRUE)
    pb_boot <- if (use_progressr) {
      progressr::progressor(steps = B)
    } else {
      NULL
    }
    result_list <- .GeoTestParallelFileApply(
      data_files = data_files,
      FUN = estimate_fun,
      ncores = ncores,
      progressor = pb_boot,
      globals = list(coordx = coordx, estimate_fun = estimate_fun),
      progress_every = TRUE,
      error_value = c(lambda = NA_real_, fallback = 0)
    )
  }

  result_mat <- do.call(rbind, result_list)
  Lambda_all <- result_mat[, "lambda"]
  fallback_all <- result_mat[, "fallback"]
  valid <- is.finite(Lambda_all)
  Lambda_boot <- Lambda_all[valid]

  ## ====== results and diagnostics ======
  n_valid <- length(Lambda_boot)
  failed <- B - n_valid
  success_rate <- n_valid / B
  fallback_boot <- sum(fallback_all > 0, na.rm = TRUE)

  if (failed > 0L) {
    warning(sprintf(
      "%d/%d bootstrap replications failed (%.1f%%)",
      failed, B, 100 * failed / B
    ), call. = FALSE)
  }

  min_required <- min(20L, B)
  if (n_valid < min_required) {
    stop(sprintf(
      "Only %d/%d bootstrap replications were successful; at least %d are required.",
      n_valid, B, min_required
    ), call. = FALSE)
  }

  if (success_rate < 0.8) {
    warning(
      sprintf(
        "Only %.1f%% of bootstrap replications were successful; results may be unreliable.",
        100 * success_rate
      ),
      call. = FALSE
    )
  }

  if (B < 99L) {
    warning(
      "B < 99 gives a coarse bootstrap p-value; use a larger B for final analyses.",
      call. = FALSE
    )
  }

  pval <- (1 + sum(Lambda_boot >= Lambda_obs)) / (n_valid + 1)

  res <- list(
    d_min = round(d_min, 5),
    h0 = round(h0_val, 5),
    lambda_obs = round(Lambda_obs, 5),
    pvalue = round(pval, 5),
    seed = seed,
    method = method,
    L = if (method == "TB") L else NULL,
    B_requested = B,
    bootstrap_successful = n_valid,
    bootstrap_failed = failed,
    bootstrap_success_rate = success_rate,
    fallback_observed = unrestricted_obs$fallback,
    fallback_bootstrap = fallback_boot,
    B_rep = Lambda_boot,
    fit_H0 = fit0,
    fit_H1 = fit1
  )

  cat("\n=== FINAL RESULTS ===\n")
  cat("Minimum distance (d_min)           =", res$d_min, "\n")
  cat("H0 threshold (compact support)    <=", res$h0, "\n")
  cat("Observed LRT statistic             =", res$lambda_obs, "\n")
  cat("Valid bootstrap reps               =", n_valid, "/", B, "\n")
  cat("Adaptive H1 fallbacks              =", fallback_boot, "bootstrap reps\n")
  cat("p-value                            =", res$pvalue, "\n\n")

  concl <- if (is.null(h0)) {
    if (res$pvalue < 0.05) {
      "Reject H0 => significant spatial dependence detected"
    } else {
      "Do not reject H0 => data consistent with spatial independence"
    }
  } else {
    if (res$pvalue < 0.05) {
      sprintf(
        "Reject H0 => compact support significantly > %.4f (stronger dependence)",
        h0
      )
    } else {
      sprintf(
        "Do not reject H0 => insufficient evidence that scale > %.4f",
        h0
      )
    }
  }
  cat("Conclusion:", concl, "\n\n")

  invisible(res)
}

####################################################
### Internal coordinate-dimension validation
####################################################

.GeoCoordinateDimension <- function(coordx = NULL, coordy = NULL,
                                    coordz = NULL, coordx_dyn = NULL) {
  if (!is.null(coordx_dyn)) {
    if (!is.list(coordx_dyn) || !length(coordx_dyn)) {
      stop("coordx_dyn must be a non-empty list of coordinate matrices.",
           call. = FALSE)
    }
    dims <- vapply(seq_along(coordx_dyn), function(i) {
      z <- as.matrix(coordx_dyn[[i]])
      if (!(ncol(z) %in% c(2L, 3L))) {
        stop("Each element of coordx_dyn must have two or three columns.",
             call. = FALSE)
      }
      ncol(z)
    }, integer(1L))
    if (length(unique(dims)) != 1L) {
      stop("All elements of coordx_dyn must have the same number of columns.",
           call. = FALSE)
    }
    return(unname(dims[[1L]]))
  }

  if ((is.matrix(coordx) || is.data.frame(coordx)) &&
      is.null(coordy) && is.null(coordz)) {
    return(ncol(as.matrix(coordx)))
  }
  if (!is.null(coordz)) return(3L)
  if (!is.null(coordy)) return(2L)
  if (is.matrix(coordx) || is.data.frame(coordx)) {
    return(ncol(as.matrix(coordx)))
  }
  NA_integer_
}

.GeoNormalizeDynamicCoordinates <- function(
    coordx_dyn, coordt = NULL, context = "This function") {
  if (!is.list(coordx_dyn) || !length(coordx_dyn)) {
    stop(context, ": coordx_dyn must be a non-empty list of coordinate matrices.",
         call. = FALSE)
  }

  blocks <- lapply(seq_along(coordx_dyn), function(i) {
    z <- as.matrix(coordx_dyn[[i]])
    if (!is.numeric(z)) {
      stop(context, ": coordx_dyn[[", i, "]] must be numeric.",
           call. = FALSE)
    }
    if (length(dim(z)) != 2L || !(ncol(z) %in% c(2L, 3L))) {
      stop(context, ": each element of coordx_dyn must have two or three columns.",
           call. = FALSE)
    }
    if (nrow(z) < 1L) {
      stop(context, ": each element of coordx_dyn must contain at least one location.",
           call. = FALSE)
    }
    if (any(!is.finite(z))) {
      stop(context, ": coordx_dyn contains non-finite coordinates.",
           call. = FALSE)
    }
    storage.mode(z) <- "double"
    unname(z)
  })

  dims <- vapply(blocks, ncol, integer(1L))
  if (length(unique(dims)) != 1L) {
    stop(context, ": all elements of coordx_dyn must have the same number of columns.",
         call. = FALSE)
  }

  if (!is.null(coordt)) {
    if (!is.numeric(coordt) || any(!is.finite(coordt))) {
      stop(context, ": coordt must be a finite numeric vector.", call. = FALSE)
    }
    if (length(coordt) != length(blocks)) {
      stop(context, ": coordx_dyn must contain one coordinate matrix for each element of coordt.",
           call. = FALSE)
    }
  }

  blocks
}

.GeoValidateDynamicData <- function(data, coordx_dyn,
                                    context = "This function",
                                    require_list = FALSE) {
  sizes <- vapply(coordx_dyn, nrow, integer(1L))
  total <- sum(sizes)

  if (isTRUE(require_list) && !is.list(data)) {
    stop(context, ": data must be a list with one component per temporal instant when coordx_dyn is used.",
         call. = FALSE)
  }

  if (is.list(data)) {
    if (length(data) != length(coordx_dyn)) {
      stop(context, ": data and coordx_dyn must have the same number of temporal components.",
           call. = FALSE)
    }
    for (i in seq_along(data)) {
      if (!is.numeric(data[[i]]) || any(!is.finite(data[[i]]))) {
        stop(context, ": data[[", i, "]] must be finite and numeric.",
             call. = FALSE)
      }
      if (length(data[[i]]) != sizes[i]) {
        stop(context, ": data[[", i, "]] must contain one value per row of coordx_dyn[[", i, "]].",
             call. = FALSE)
      }
    }
  } else {
    if (!is.numeric(data) || any(!is.finite(data))) {
      stop(context, ": data must be finite and numeric.", call. = FALSE)
    }
    if (length(data) != total) {
      stop(context, ": data must contain one value per dynamic observation (",
           total, " values expected).", call. = FALSE)
    }
  }

  invisible(sizes)
}

.GeoValidateDynamicRows <- function(x, coordx_dyn, name = "X",
                                    context = "This function") {
  if (is.null(x)) return(invisible(NULL))

  sizes <- vapply(coordx_dyn, nrow, integer(1L))
  total <- sum(sizes)

  if (is.list(x)) {
    if (length(x) != length(coordx_dyn)) {
      stop(context, ": ", name,
           " must have one component per temporal instant.", call. = FALSE)
    }
    for (i in seq_along(x)) {
      xi <- as.matrix(x[[i]])
      if (!is.numeric(xi) || any(!is.finite(xi))) {
        stop(context, ": ", name, "[[", i, "]] must be finite and numeric.",
             call. = FALSE)
      }
      if (nrow(xi) != sizes[i]) {
        stop(context, ": ", name, "[[", i,
             "]] must have one row per row of coordx_dyn[[", i, "]].",
             call. = FALSE)
      }
    }
  } else {
    xi <- as.matrix(x)
    if (!is.numeric(xi) || any(!is.finite(xi))) {
      stop(context, ": ", name, " must be finite and numeric.",
           call. = FALSE)
    }
    if (nrow(xi) != total) {
      stop(context, ": ", name,
           " must have one row per dynamic observation (", total,
           " rows expected).", call. = FALSE)
    }
  }

  invisible(sizes)
}

.GeoValidateCoordinateCompatibility <- function(
    coordx = NULL, coordy = NULL, coordz = NULL, coordx_dyn = NULL,
    loc, context = "Kriging") {
  p_obs <- .GeoCoordinateDimension(coordx, coordy, coordz, coordx_dyn)
  loc <- as.matrix(loc)
  p_loc <- ncol(loc)

  if (!(p_loc %in% c(2L, 3L))) {
    stop(context, ": prediction locations must have two or three columns.",
         call. = FALSE)
  }
  if (is.na(p_obs)) return(invisible(c(observed = p_obs, prediction = p_loc)))
  if (!(p_obs %in% c(2L, 3L))) {
    stop(context, ": observed coordinates must have two or three columns.",
         call. = FALSE)
  }
  if (p_obs != p_loc) {
    stop(context, ": observed coordinates are ", p_obs,
         "-dimensional, but prediction locations are ", p_loc,
         "-dimensional. Supply coordinates with the same spatial dimension.",
         call. = FALSE)
  }

  invisible(c(observed = p_obs, prediction = p_loc))
}

.GeoValidate3D <- function(coordx = NULL, coordy = NULL, coordz = NULL,
                           coordx_dyn = NULL, distance = "Eucl",
                           grid = FALSE, approximate = FALSE,
                           method = NULL, context = "This function") {
  p <- .GeoCoordinateDimension(coordx, coordy, coordz, coordx_dyn)
  if (is.na(p)) return(invisible(p))
  if (!(p %in% c(2L, 3L))) {
    stop(context, " requires coordinates with two or three columns.",
         call. = FALSE)
  }

  distance_key <- tolower(gsub("[[:blank:]]", "", as.character(distance)[1L]))
  if (p == 3L && !identical(distance_key, "eucl")) {
    stop(context, ": three-dimensional coordinates require distance = 'Eucl'.",
         call. = FALSE)
  }
  if (p == 3L && isTRUE(grid)) {
    stop(context, ": regular three-dimensional grids are not implemented; ",
         "supply their Cartesian product as an explicit N x 3 matrix with ",
         "grid = FALSE.", call. = FALSE)
  }
  if (p == 3L && isTRUE(approximate)) {
    method_key <- if (is.null(method)) "" else
      toupper(gsub("[[:blank:]]", "", as.character(method)[1L]))
    if (!identical(method_key, "TB")) {
      if (identical(method_key, "CE")) {
        stop(context, ": CE approximate simulation is implemented only for ",
             "two-dimensional regular grids. Use method = 'TB' for a purely ",
             "spatial univariate irregular three-dimensional field, or use ",
             "method = 'cholesky'.", call. = FALSE)
      }
      stop(context, ": approximate three-dimensional simulation is supported ",
           "only by method = 'TB' for purely spatial univariate fields. Use ",
           "method = 'cholesky' otherwise.", call. = FALSE)
    }
  }
  invisible(p)
}

GeoSimcond <- function(estobj = NULL, data, coordx, coordy = NULL, coordz = NULL, coordt = NULL,
                       coordx_dyn = NULL, corrmodel, distance = "Eucl",
                       grid = FALSE, loc, maxdist = NULL, maxtime = NULL, method = "Cholesky",
                       model = "Gaussian", n = 1, nrep = 1, local = FALSE, L = 1000,
                       neighb = NULL, param, anisopars = NULL, radius = 1, sparse = FALSE,
                       time = NULL, copula = NULL, X = NULL, Xloc = NULL, Mloc = NULL,
                       parallel = FALSE, ncores = 6, progress = FALSE, n_iter = 25L,
                       check.duplicates = FALSE, nloc = NULL, mcmc_thin = 1L) {

################################################################################# 
### Helper function for conditional skew gaussian simulation 
################################################################################# 
  # Model-specific conditional engines are implemented in SimcondUtils.R.


###############################################################
############### end helper functions ##########################
###############################################################

###############################################################
############################ MAIN FUNCTION ####################
###############################################################
  call <- match.call()

  ###### Check GeoFit object ######
  if (!is.null(estobj)) {
    if (!inherits(estobj, "GeoFit"))
      stop("need a 'GeoFit' object as input\n")

    data <- estobj$data
    if (!estobj$grid) {
      if (!estobj$bivariate) {
        if (is.null(estobj$coordx_dyn)) coordx <- cbind(estobj$coordx, estobj$coordy, estobj$coordz)
        else coordx <- NULL
      } else {
        if (is.null(estobj$coordx_dyn)) {
          coordx <- estobj$coordx[1:estobj$ns[1]]
          coordy <- estobj$coordy[1:estobj$ns[2]]
        } else {
          coordx <- NULL
        }
      }
    } else {
      coordx <- estobj$coordx; coordy <- estobj$coordy; coordz <- estobj$coordz
    }
    if (length(estobj$coordt) == 1) coordt <- NULL else coordt <- estobj$coordt
    coordx_dyn <- estobj$coordx_dyn
    corrmodel  <- estobj$corrmodel
    model      <- estobj$model
    distance   <- estobj$distance
    grid       <- estobj$grid
    n          <- estobj$n
    param      <- append(estobj$param, estobj$fixed)
    radius     <- estobj$radius
    copula     <- estobj$copula
    anisopars  <- estobj$anisopars
    X          <- estobj$X
  }
  #### end check

  ### Additional checks
  if (is.null(CkModel(model))) stop("The name of the model is not correct\n")
  if (!is.character(corrmodel) || is.null(CkCorrModel(corrmodel)))
    stop("The name of the correlation model is wrong\n")
  if (!(method %in% c("Cholesky", "TB", "CE"))) {
    stop("The method of unconditional simulation is not correct\n")
  }
  corrmodel <- gsub("[[:blank:]]", "", corrmodel)
  model     <- gsub("[[:blank:]]", "", model)
  distance  <- gsub("[[:blank:]]", "", distance)

  # Canonical aliases used throughout GeoModels.
  if (model == "Gauss") model <- "Gaussian"
  if (model == "LogGauss") model <- "LogGaussian"
  if (model == "SkewGauss") model <- "SkewGaussian"
  if (model == "poisson") model <- "Poisson"
  if (model %in% c("Binary", "Bernoulli")) {
    if (!is.null(nloc) && any(as.numeric(nloc) != 1))
      stop("Binary/Bernoulli conditional simulation requires nloc = 1 when supplied.", call. = FALSE)
    model <- "Binomial"
    n <- 1L
    if (is.null(nloc)) nloc <- 1L
  }
  if (model %in% c("Geom", "Geometric")) {
    if (!is.null(nloc) && any(as.numeric(nloc) != 1))
      stop("Geometric conditional simulation requires nloc = 1 when supplied.", call. = FALSE)
    model <- "BinomialNeg"
    n <- 1L
    if (is.null(nloc)) nloc <- 1L
  }

  # The count-constrained sampler mixes rapidly under the validated direct
  # Binomial/Negative-Binomial paths, so its user-facing default is 25 sweeps.
  # Other latent Gibbs engines retain their historical 1000-sweep default when
  # n_iter is omitted; an explicitly supplied n_iter is always respected.
  direct_count_model <- is.null(copula) && model %in% c("Binomial", "BinomialNeg")
  if (missing(n_iter) && !direct_count_model) n_iter <- 1000L

  #### checking cores ######
  ncores <- .GeoResolveWorkers(parallel, ncores)
  parallel <- isTRUE(parallel) && ncores > 1L

  # Determine model type (space / spacetime / bivariate)
  model_ck  <- CkCorrModel(corrmodel)
  bivariate <- CheckBiv(model_ck)
  spacetime <- CheckST(model_ck)
  space     <- !spacetime && !bivariate

  # Phase-1 support is intentionally conservative.  Unsupported domains are
  # rejected before covariance matrices or unconditional simulations are built.
  if (bivariate) {
    stop("GeoSimcond currently supports univariate correlation models only.", call. = FALSE)
  }
  if (spacetime || !is.null(coordt) || !is.null(time)) {
    stop("GeoSimcond space-time conditional simulation is not implemented yet.", call. = FALSE)
  }
  if (!is.null(coordx_dyn)) {
    stop("GeoSimcond dynamic coordinates are not implemented yet.", call. = FALSE)
  }
  if (isTRUE(grid)) {
    stop("GeoSimcond grid = TRUE is not implemented yet; supply explicit coordinates.", call. = FALSE)
  }
  if (!is.logical(local) || length(local) != 1L || is.na(local)) {
    stop("local must be TRUE or FALSE.", call. = FALSE)
  }
  if (isTRUE(sparse)) {
    stop("GeoSimcond sparse = TRUE is not implemented yet.", call. = FALSE)
  }
  if (identical(method, "CE")) {
    stop(paste0(
      "GeoSimcond method = 'CE' is not available for the current explicit-coordinate ",
      "conditional simulation workflow; use method = 'Cholesky' or method = 'TB'."
    ), call. = FALSE)
  }

  .GeoCheckDuplicateLocations(
    coordx = coordx, coordy = coordy, coordz = coordz,
    spacetime = FALSE, bivariate = FALSE, grid = FALSE,
    context = "GeoSimcond", check.duplicates = check.duplicates
  )
  .dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
  on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)

  if (!is.numeric(nrep) || length(nrep) != 1L || !is.finite(nrep) ||
      nrep < 1 || nrep != as.integer(nrep)) {
    stop("nrep must be a positive integer.", call. = FALSE)
  }
  nrep <- as.integer(nrep)
  if (!is.numeric(n_iter) || length(n_iter) != 1L || !is.finite(n_iter) ||
      n_iter < 1 || n_iter != as.integer(n_iter))
    stop("n_iter must be a positive integer.", call. = FALSE)
  n_iter <- as.integer(n_iter)
  if (!is.numeric(mcmc_thin) || length(mcmc_thin) != 1L || !is.finite(mcmc_thin) ||
      mcmc_thin < 1 || mcmc_thin != as.integer(mcmc_thin))
    stop("mcmc_thin must be a positive integer.", call. = FALSE)
  mcmc_thin <- as.integer(mcmc_thin)

  if (is.vector(loc)) loc <- matrix(loc, nrow = 1L)
  loc <- as.matrix(loc)
  if (!is.numeric(loc) || nrow(loc) < 1L || any(!is.finite(loc))) {
    stop("loc must be a non-empty finite numeric matrix.", call. = FALSE)
  }

  # Explicit model/copula support matrix.  This prevents an unsupported model
  # from falling through to an undefined 'res' object.
  supported_plain <- c(
    "Gaussian", "LogGaussian", "Tukeyh", "Tukeyh2", "SinhAsinh",
    "SkewGaussian", "Gamma", "Weibull", "Binomial", "BinomialNeg"
  )
  continuous_copula_margins <- c(
    "Gaussian", "StudentT", "LogGaussian", "Gamma", "Weibull",
    "Beta", "Beta2", "Kumaraswamy", "Kumaraswamy2",
    "Logistic", "SkewLaplace"
  )
  gaussian_copula_margins <- unique(c(
    continuous_copula_margins, "Tukeyh", "Tukeyh2", "SinhAsinh"
  ))
  supported_copula <- list(
    Gaussian = gaussian_copula_margins,
    Clayton = continuous_copula_margins,
    SkewGaussian = continuous_copula_margins
  )
  if (is.null(copula)) {
    if (!(model %in% supported_plain)) {
      stop(sprintf("GeoSimcond does not currently support model = '%s' without a copula.", model),
           call. = FALSE)
    }
  } else {
    copula <- gsub("[[:blank:]]", "", as.character(copula)[1L])
    if (!(copula %in% names(supported_copula))) {
      stop(sprintf("GeoSimcond does not currently support copula = '%s'.", copula),
           call. = FALSE)
    }
    if (!(model %in% supported_copula[[copula]])) {
      stop(sprintf(
        paste0(
          "GeoSimcond implements validated copula conditional simulation only for continuous margins. ",
          "Model '%s' is not supported with copula='%s'. Supported continuous margins are: %s."
        ),
        model, copula, paste(supported_copula[[copula]], collapse = ", ")
      ), call. = FALSE)
    }
  }

  ## Local conditional simulation is implemented through local Gaussian
  ## substitution.  It is therefore available for Gaussian fields, monotone
  ## one-Gaussian transformations, and all continuous margins under a Gaussian
  ## copula.  Gibbs-based latent constructions still require global conditioning.
  if (isTRUE(local)) {
    if (is.null(neighb) && is.null(maxdist)) {
      stop("For local conditional simulation, specify neighb or maxdist.",
           call. = FALSE)
    }
    if (!is.null(neighb)) {
      if (!is.numeric(neighb) || length(neighb) != 1L || !is.finite(neighb) ||
          neighb < 1 || abs(neighb - round(neighb)) > sqrt(.Machine$double.eps)) {
        stop("neighb must be a positive integer for local conditional simulation.",
             call. = FALSE)
      }
      neighb <- as.integer(round(neighb))
    }
    if (!is.null(maxdist) &&
        (!is.numeric(maxdist) || length(maxdist) != 1L || !is.finite(maxdist) || maxdist < 0)) {
      stop("maxdist must be a finite non-negative scalar.", call. = FALSE)
    }

    local_plain <- c("Gaussian", "LogGaussian", "Tukeyh", "Tukeyh2", "SinhAsinh")
    local_supported <- if (is.null(copula)) {
      model %in% local_plain
    } else {
      identical(copula, "Gaussian")
    }
    if (!local_supported) {
      stop(
        paste0(
          "GeoSimcond local=TRUE is currently supported for Gaussian fields, ",
          "LogGaussian/Tukeyh/Tukeyh2/SinhAsinh transformations, and continuous ",
          "margins with copula='Gaussian'. Gibbs-based SkewGaussian, Gamma/Weibull, ",
          "direct Binomial/BinomialNeg models, Clayton and SkewGaussian copulas still require local=FALSE."
        ),
        call. = FALSE
      )
    }
  }

  # Observed coordinates matrix
  if (is.matrix(coordx) || is.data.frame(coordx)) {
    coord_obs <- as.matrix(coordx)
  } else {
    coord_obs <- data.frame(x = coordx)
    if (!is.null(coordy)) coord_obs$y <- coordy
    if (!is.null(coordz)) coord_obs$z <- coordz
    coord_obs <- as.matrix(coord_obs)
  }
  if (!is.numeric(coord_obs) || nrow(coord_obs) < 1L || any(!is.finite(coord_obs))) {
    stop("Observed coordinates must be a non-empty finite numeric matrix.", call. = FALSE)
  }
  if (ncol(loc) != ncol(coord_obs)) {
    stop("Observed and prediction coordinates must have the same dimension.", call. = FALSE)
  }
  data <- as.numeric(data)
  if (length(data) != nrow(coord_obs) || any(!is.finite(data))) {
    stop("data must contain one finite value per observed location.", call. = FALSE)
  }
  if (model == "LogGaussian" && any(data <= 0)) {
    stop("LogGaussian conditional simulation requires strictly positive data.", call. = FALSE)
  }
  if (model %in% c("Gamma", "Weibull") && any(data < 0)) {
    stop(sprintf("%s conditional simulation requires non-negative data.", model),
         call. = FALSE)
  }
  if (!is.null(copula) && model %in% c("Beta", "Beta2")) {
    lower_support <- as.numeric(param$min)
    upper_support <- as.numeric(param$max)
    if (length(lower_support) != 1L || length(upper_support) != 1L ||
        !is.finite(lower_support) || !is.finite(upper_support) ||
        lower_support >= upper_support) {
      stop(sprintf("%s requires finite param$min < param$max.", model), call. = FALSE)
    }
    if (any(data < lower_support | data > upper_support)) {
      stop(sprintf("%s observations must lie inside [param$min, param$max].", model), call. = FALSE)
    }
  }
  if (!is.null(copula) && model %in% c("Poisson", "BinomialNeg", "Binomial")) {
    if (any(data < 0) || any(abs(data - round(data)) > sqrt(.Machine$double.eps))) {
      stop(sprintf("%s observations must be non-negative integers.", model),
           call. = FALSE)
    }
  }
  if (!is.null(copula) && model == "Binomial") {
    if (!is.numeric(n) || length(n) != 1L || !is.finite(n) ||
        n < 1 || n != as.integer(n)) {
      stop("n must be a positive integer for the Binomial model.", call. = FALSE)
    }
    n <- as.integer(n)
    if (any(data > n)) stop("Binomial observations cannot exceed n.", call. = FALSE)
  }
  if (!is.null(copula) && model == "BinomialNeg") {
    if (!is.numeric(n) || length(n) != 1L || !is.finite(n) || n <= 0) {
      stop("n must be strictly positive for the BinomialNeg model.", call. = FALSE)
    }
  }

  # Mean and covariate specification.  The observation and prediction
  # means are selected independently:
  #   observations: param$mean vector, or X %*% beta, or intercept mean;
  #   predictions:  Mloc, or Xloc %*% beta, or intercept mean.
  if (spacetime && is.null(time))
    stop("At least one temporal instant is needed for space-time conditional simulation\n")
  n_obs_mean <- length(data)
  n_loc_mean <- nrow(as.matrix(loc)) * if(spacetime) length(time) else 1L
  external_mean <- .GeoMean_external(param, n = n_obs_mean)
  if (!is.null(external_mean) && !is.null(X))
    stop("A site-specific observation mean and X cannot be used together.\n")
  if (!is.null(Mloc) && !is.null(Xloc))
    stop("Mloc and Xloc are mutually exclusive.\n")
  if (!is.null(external_mean) && is.null(Mloc))
    stop("Mloc is required when the observation mean is site-specific.\n")

  if (!is.null(X)) {
    X <- .GeoMean_design(X, n = n_obs_mean, name = "X")
    p_mean <- ncol(X)
    invisible(.GeoMean_beta(param, p_mean))
    if (is.null(Xloc) && is.null(Mloc))
      stop("Supply either Xloc or Mloc for prediction locations.\n")
  } else {
    if (!is.null(Xloc)) stop("Xloc requires an observation design matrix X.\n")
    p_mean <- 1L
    if (is.null(external_mean)) {
      if (length(setdiff(.GeoMean_supplied_names(param), "mean")))
        stop("Multiple mean coefficients require an explicit design matrix X.\n")
      invisible(.GeoMean_beta(param, 1L))
    }
  }
  if (!is.null(Xloc))
    Xloc <- .GeoMean_design(Xloc, n = n_loc_mean, p = p_mean, name = "Xloc")
  if (!is.null(Mloc)) {
    Mloc <- as.numeric(Mloc)
    if (length(Mloc) == 1L) Mloc <- rep(Mloc, n_loc_mean)
    if (length(Mloc) != n_loc_mean || any(!is.finite(Mloc)))
      stop("Mloc must contain one finite value per prediction point.\n")
  }
  # Do not anisotropize loc here.  Anisotropy is propagated to
  # GeoSim/GeoKrig/GeoKrigWeights/GeoKriglocWeights, which handle it consistently.
  loc_orig <- loc
  # Temporal coordinates to use
  coordt_use <- if (spacetime || bivariate) coordt else NULL

  # Helper for multiplicative models: eta(s) = log(mu(s)) = X(s)' beta.
  # If X/Xloc are supplied, they are used; otherwise a constant mean is assumed.
  # Mloc, when supplied, is interpreted on the eta = log(mu) scale.
  clip_probability <- function(p, context = "probability transform") {
    p <- as.numeric(p)
    if (any(!is.finite(p))) {
      stop(sprintf("%s produced non-finite probabilities.", context), call. = FALSE)
    }
    eps <- .Machine$double.eps
    pmin(pmax(p, eps), 1 - eps)
  }

  get_eta_obs_loc <- function(param, X, Xloc, Mloc, n_obs, n_loc) {
    external <- .GeoMean_external(param, n = n_obs)
    eta_obs <- .GeoMean_vector(param, X = X, n = n_obs, external = external)

    if (!is.null(Mloc)) {
      eta_loc <- as.numeric(Mloc)
      if (length(eta_loc) == 1L) eta_loc <- rep(eta_loc, n_loc)
      if (length(eta_loc) != n_loc)
        stop("Mloc must contain one value per prediction point.\n")
    } else if (!is.null(Xloc)) {
      Xloc_mat <- .GeoMean_design(Xloc, n = n_loc, name = "Xloc")
      beta <- .GeoMean_beta(param, ncol(Xloc_mat))
      eta_loc <- as.numeric(Xloc_mat %*% beta)
    } else {
      if (!is.null(X)) {
        stop(
          "Xloc (or Mloc) must be supplied for prediction when X is used at the observed locations.",
          call. = FALSE
        )
      }
      eta_loc <- rep(.GeoMean_beta(param, 1L)[1L], n_loc)
    }

    list(obs = eta_obs, loc = eta_loc)
  }

############################################################################
######################### not copula models ################################
############################################################################
res <- NULL
if (is.null(copula)) {
  if (model == "Gaussian") {
    res <- .GeoSimcondGaussian(data, corrmodel, nrep, method, L,
                    param,
                    coord_obs, loc, coordt_use, time, X, Xloc, Mloc, distance, radius, anisopars,
                    local, neighb, maxdist, maxtime,
                    space, spacetime, bivariate, parallel, ncores, progress)
  }

  ################################################################################
  ############### monotone transformations of one Gaussian RF ####################
  ################################################################################
  if (model == "LogGaussian") {
    eta <- get_eta_obs_loc(param, X, Xloc, Mloc, n_obs_mean, n_loc_mean)
    vv <- param$sill

    # Multiplicative parametrization:
    # Y(s) = exp(eta(s)) * exp(sqrt(vv) Z(s) - vv/2), hence E[Y(s)] = exp(eta(s)).
    datanorm <- (log(data) - eta$obs + vv/2)/sqrt(vv)   # gaussian scale

    param_gauss <- .GeoMean_intercept_param(param, 0)
    param_gauss$sill <- 1

    res <- .GeoSimcondGaussian(datanorm, corrmodel, nrep, method, L, param_gauss,
                    coord_obs, loc, coordt_use, time, NULL, NULL, NULL, distance, radius, anisopars,
                    local, neighb, maxdist, maxtime,
                    space, spacetime, bivariate, parallel, ncores, progress)

    res <- lapply(res, function(r) exp(eta$loc) * exp(sqrt(vv) * r - vv/2)) # back-transformation
  }

  if (model == "Tukeyh") {
    eta <- get_eta_obs_loc(param, X, Xloc, Mloc, n_obs_mean, n_loc_mean)
    vv <- param$sill
    tail <- .GeoMarginalTukeyTail(param$tail, "tail")
    datanorm <- .GeoMarginalInvLambertH((data - eta$obs)/sqrt(vv), tail)
    param_latent <- .GeoMean_intercept_param(param, 0)
    param_latent$sill <- 1; param_latent$tail <- NULL
    res <- .GeoSimcondGaussian(datanorm, corrmodel, nrep, method, L, param_latent,
                    coord_obs, loc, coordt_use, time, NULL, NULL, NULL, distance, radius, anisopars,
                    local, neighb, maxdist, maxtime,
                    space, spacetime, bivariate, parallel, ncores, progress)
    res <- lapply(res, function(r) eta$loc + sqrt(vv) * .GeoMarginalTukeyH(r, tail))
  }

  if (model == "Tukeyh2") {
    eta <- get_eta_obs_loc(param, X, Xloc, Mloc, n_obs_mean, n_loc_mean)
    vv <- param$sill
    tail1 <- .GeoMarginalTukeyTail(param$tail1, "tail1")
    tail2 <- .GeoMarginalTukeyTail(param$tail2, "tail2")
    datanorm <- .GeoMarginalInvTukeyH2((data - eta$obs)/sqrt(vv), tail1, tail2)
    param_latent <- .GeoMean_intercept_param(param, 0)
    param_latent$sill <- 1; param_latent$tail1 <- NULL; param_latent$tail2 <- NULL
    res <- .GeoSimcondGaussian(datanorm, corrmodel, nrep, method, L, param_latent,
                    coord_obs, loc, coordt_use, time, NULL, NULL, NULL, distance, radius, anisopars,
                    local, neighb, maxdist, maxtime,
                    space, spacetime, bivariate, parallel, ncores, progress)
    res <- lapply(res, function(r) {
      eta$loc + sqrt(vv) * .GeoMarginalTukeyH2(r, tail1, tail2)
    })
  }

  if (model == "SinhAsinh") {
    eta <- get_eta_obs_loc(param, X, Xloc, Mloc, n_obs_mean, n_loc_mean)
    vv <- as.numeric(param$sill)
    skew <- as.numeric(param$skew)
    tail <- as.numeric(param$tail)
    if (length(vv) != 1L || !is.finite(vv) || vv <= 0)
      stop("SinhAsinh sill must be a finite positive scalar.", call. = FALSE)
    if (length(skew) != 1L || !is.finite(skew))
      stop("SinhAsinh skew must be a finite scalar.", call. = FALSE)
    if (length(tail) != 1L || !is.finite(tail) || tail <= 0)
      stop("SinhAsinh tail must be a finite positive scalar.", call. = FALSE)

    ## GeoSim and the SAS construction use
    ## Y = eta + sqrt(vv) * sinh((asinh(Z) + skew) / tail),
    ## whose exact inverse is
    ## Z = sinh(tail * asinh((Y-eta)/sqrt(vv)) - skew).
    datanorm <- sinh(tail * asinh((data - eta$obs)/sqrt(vv)) - skew)
    param_latent <- .GeoMean_intercept_param(param, 0)
    param_latent$sill <- 1; param_latent$skew <- NULL; param_latent$tail <- NULL
    res <- .GeoSimcondGaussian(datanorm, corrmodel, nrep, method, L, param_latent,
                    coord_obs, loc, coordt_use, time, NULL, NULL, NULL, distance, radius, anisopars,
                    local, neighb, maxdist, maxtime,
                    space, spacetime, bivariate, parallel, ncores, progress)
    res <- lapply(res, function(r) {
      eta$loc + sqrt(vv) * sinh((asinh(r) + skew) / tail)
    })
  }
  
  ################################################################################
  ############### non monotone transformations of more than one Gaussian RF ######
  ################################################################################
  if (model == "SkewGaussian") {
    eta <- get_eta_obs_loc(param, X, Xloc, Mloc, n_obs_mean, n_loc_mean)
    param_skew <- .GeoMean_intercept_param(param, 0)
    res <- .GeoSimcondSkewGaussian(coord_obs, loc, data, param_skew, corrmodel, local = local,
                               neighb = neighb, L = L, nrep = nrep, method = method,
                               distance = distance, radius = radius, anisopars = anisopars,
                               parallel = parallel, ncores = ncores, progress = progress,
                               n_iter = n_iter,
                               mean_obs = eta$obs, mean_loc = eta$loc)
  }

  if (model == "Gamma") {
    eta <- get_eta_obs_loc(param, X, Xloc, Mloc, n_obs_mean, n_loc_mean)
    res <- .GeoSimcondGamma(coord_obs, loc, data, param, corrmodel, local = local,
                        neighb = neighb, L = L, nrep = nrep, method = method,
                        distance = distance, radius = radius, anisopars = anisopars,
                        parallel = parallel, ncores = ncores, progress = progress,
                        mean_obs = eta$obs, mean_loc = eta$loc)
  }
  
  if (model == "Weibull") {
    eta <- get_eta_obs_loc(param, X, Xloc, Mloc, n_obs_mean, n_loc_mean)
    param.weibull <- .GeoMean_intercept_param(param, 0)
    param.weibull$shape <- 2
    ck <- gamma((param$shape + 1)/param$shape)
    data.t <- (exp(-eta$obs) * data * ck)^param$shape
    res <- .GeoSimcondGamma(coord_obs, loc, data.t, param.weibull, corrmodel, local = local,
                        neighb = neighb, L = L, nrep = nrep, method = method,
                        distance = distance, radius = radius, anisopars = anisopars,
                        parallel = parallel, ncores = ncores, progress = progress,
                        mean_obs = rep(0, n_obs_mean), mean_loc = rep(0, n_loc_mean))
    res <- lapply(res, function(x) exp(eta$loc) * (x^(1/param$shape))/ck)
  }
  
  if (model == "Poisson")  { stop("Poisson not implemented...") }

  if (model %in% c("Binomial", "BinomialNeg")) {
    eta <- get_eta_obs_loc(param, X, Xloc, Mloc, n_obs_mean, n_loc_mean)
    res <- .GeoSimcondCountRF(
      model = model, coord_obs = coord_obs, loc = loc, data = data,
      n = n, nloc = nloc, param = param, corrmodel = corrmodel,
      mean_obs = eta$obs, mean_loc = eta$loc,
      distance = distance, radius = radius, anisopars = anisopars,
      nrep = nrep, n_iter = n_iter, mcmc_thin = mcmc_thin,
      method = method, local = local, parallel = parallel, progress = progress
    )
  }

} else { 
###########################################################################
################## copula models ###########################################
###########################################################################

  ## Continuous margins are first mapped one-to-one to their copula uniforms.
  ## Gaussian and skew-Gaussian copulas then condition their latent fields;
  ## the Clayton-like copula conditions the Gaussian-square representation
  ## subject to the observed latent ratio.  Discrete margins are intentionally
  ## not routed through the old mid-PIT approximation.

  .GeoCopula_cond_margin <- function(model, param, data) {
    scalar <- function(x, name, lower = -Inf, upper = Inf,
                       lower_open = FALSE, upper_open = FALSE) {
      x <- suppressWarnings(as.numeric(x))
      bad <- length(x) != 1L || !is.finite(x)
      if (!bad) {
        bad <- if (lower_open) x <= lower else x < lower
        bad <- bad || if (upper_open) x >= upper else x > upper
      }
      if (bad) stop(sprintf("Invalid %s for model '%s'.", name, model), call. = FALSE)
      x
    }

    ## Only margins whose marginal scale is parameterized by sill require it.
    ## Positive/bounded margins such as Gamma, Weibull, Beta/Beta2 and
    ## Kumaraswamy/Kumaraswamy2 determine their marginal variance from their
    ## shape parameters, so GeoFit objects for those models legitimately do not
    ## contain param$sill.  Do not reject them before their PIT is constructed.
    sill_models <- c(
      "Gaussian", "StudentT", "LogGaussian", "Logistic", "SkewLaplace",
      "SinhAsinh", "Tukeyh", "Tukeyh2"
    )
    sill <- NULL
    if (model %in% sill_models) {
      sill <- scalar(param$sill, "sill", lower = 0, lower_open = TRUE)
    }
    nugget <- if (is.null(param$nugget)) 0 else
      scalar(param$nugget, "nugget", lower = 0, upper = 1, upper_open = TRUE)

    eta <- NULL
    eta_needed <- model %in% c(
      "Gaussian", "StudentT", "LogGaussian", "Gamma", "Weibull",
      "Beta2", "Kumaraswamy2", "Logistic", "SkewLaplace",
      "SinhAsinh", "Tukeyh", "Tukeyh2"
    )
    if (eta_needed) {
      eta <- get_eta_obs_loc(param, X, Xloc, Mloc, n_obs_mean, n_loc_mean)
    }

    if (model == "Gaussian") {
      u <- stats::pnorm(data, mean = eta$obs, sd = sqrt(sill))
      qloc <- function(p) stats::qnorm(p, mean = eta$loc, sd = sqrt(sill))
    } else if (model == "StudentT") {
      eps_df <- scalar(param$df, "df", lower = 0, lower_open = TRUE)
      df_t <- 1 / eps_df
      u <- stats::pt((data - eta$obs) / sqrt(sill), df = df_t)
      qloc <- function(p) eta$loc + sqrt(sill) * stats::qt(p, df = df_t)
    } else if (model == "LogGaussian") {
      if (any(data <= 0))
        stop("LogGaussian conditional simulation requires strictly positive data.", call. = FALSE)
      u <- stats::plnorm(data, meanlog = eta$obs - sill / 2, sdlog = sqrt(sill))
      qloc <- function(p) stats::qlnorm(p, meanlog = eta$loc - sill / 2, sdlog = sqrt(sill))
    } else if (model == "Gamma") {
      shape <- scalar(param$shape, "shape", lower = 0, lower_open = TRUE)
      if (any(data <= 0))
        stop("Gamma conditional simulation requires strictly positive data.", call. = FALSE)
      mu_obs <- exp(eta$obs); mu_loc <- exp(eta$loc)
      u <- stats::pgamma(data, shape = shape / 2, rate = shape / (2 * mu_obs))
      qloc <- function(p) stats::qgamma(p, shape = shape / 2, rate = shape / (2 * mu_loc))
    } else if (model == "Weibull") {
      shape <- scalar(param$shape, "shape", lower = 0, lower_open = TRUE)
      if (any(data < 0))
        stop("Weibull conditional simulation requires non-negative data.", call. = FALSE)
      mu_obs <- exp(eta$obs); mu_loc <- exp(eta$loc)
      denom <- gamma(1 + 1 / shape)
      u <- stats::pweibull(data, shape = shape, scale = mu_obs / denom)
      qloc <- function(p) stats::qweibull(p, shape = shape, scale = mu_loc / denom)
    } else if (model == "Logistic") {
      u <- stats::plogis(data, location = eta$obs, scale = sqrt(sill))
      qloc <- function(p) stats::qlogis(p, location = eta$loc, scale = sqrt(sill))
    } else if (model == "SinhAsinh") {
      tail <- scalar(param$tail, "tail", lower = 0, lower_open = TRUE)
      skew <- scalar(param$skew, "skew")
      z <- sinh(tail * asinh((data - eta$obs) / sqrt(sill)) - skew)
      u <- stats::pnorm(z)
      qloc <- function(p) {
        zq <- stats::qnorm(p)
        eta$loc + sqrt(sill) * sinh((asinh(zq) + skew) / tail)
      }
    } else if (model == "Tukeyh") {
      tail <- scalar(param$tail, "tail", lower = 0, upper = 0.5, upper_open = TRUE)
      x <- (data - eta$obs) / sqrt(sill)
      u <- stats::pnorm(.GeoMarginalInvLambertH(x, tail))
      qloc <- function(p) {
        eta$loc + sqrt(sill) * .GeoMarginalQTukeyH(p, tail)
      }
    } else if (model == "Tukeyh2") {
      tail1 <- scalar(param$tail1, "tail1", lower = 0, upper = 0.5, upper_open = TRUE)
      tail2 <- scalar(param$tail2, "tail2", lower = 0, upper = 0.5, upper_open = TRUE)
      x <- (data - eta$obs) / sqrt(sill)
      z <- numeric(length(x))
      pos <- x >= 0
      if (any(pos)) z[pos] <- .GeoMarginalInvLambertH(x[pos], tail1)
      if (any(!pos)) z[!pos] <- .GeoMarginalInvLambertH(x[!pos], tail2)
      u <- stats::pnorm(z)
      qloc <- function(p) {
        eta$loc + sqrt(sill) * .GeoMarginalQTukeyH2(p, tail1, tail2)
      }
    } else if (model == "SkewLaplace") {
      skew <- scalar(param$skew, "skew", lower = 0, upper = 1,
                     lower_open = TRUE, upper_open = TRUE)
      u <- .GeoMarginalPSkewLaplace((data - eta$obs) / sqrt(sill), skew)
      qloc <- function(p) eta$loc + sqrt(sill) * .GeoMarginalQSkewLaplace(p, skew)
    } else if (model %in% c("Beta", "Beta2", "Kumaraswamy", "Kumaraswamy2")) {
      pmin0 <- scalar(param$min, "min")
      pmax0 <- scalar(param$max, "max")
      if (pmin0 >= pmax0)
        stop(sprintf("%s requires min < max.", model), call. = FALSE)
      if (any(data < pmin0 | data > pmax0))
        stop(sprintf("%s observations must lie inside [min, max].", model), call. = FALSE)
      y <- (data - pmin0) / (pmax0 - pmin0)

      if (model == "Beta") {
        a <- scalar(param$shape1, "shape1", lower = 0, lower_open = TRUE)
        b <- scalar(param$shape2, "shape2", lower = 0, lower_open = TRUE)
        u <- stats::pbeta(y, shape1 = a, shape2 = b)
        qloc <- function(p) pmin0 + (pmax0 - pmin0) * stats::qbeta(p, shape1 = a, shape2 = b)
      } else if (model == "Beta2") {
        shape <- scalar(param$shape, "shape", lower = 0, lower_open = TRUE)
        mu_obs <- stats::plogis(eta$obs); mu_loc <- stats::plogis(eta$loc)
        u <- stats::pbeta(y, shape1 = mu_obs * shape, shape2 = (1 - mu_obs) * shape)
        qloc <- function(p) pmin0 + (pmax0 - pmin0) *
          stats::qbeta(p, shape1 = mu_loc * shape, shape2 = (1 - mu_loc) * shape)
      } else if (model == "Kumaraswamy") {
        a <- scalar(param$shape1, "shape1", lower = 0, lower_open = TRUE)
        b <- scalar(param$shape2, "shape2", lower = 0, lower_open = TRUE)
        pkuma <- function(x, aa, bb) 1 - (1 - x^aa)^bb
        qkuma <- function(p, aa, bb) (1 - (1 - p)^(1 / bb))^(1 / aa)
        u <- pkuma(y, a, b)
        qloc <- function(p) pmin0 + (pmax0 - pmin0) * qkuma(p, a, b)
      } else {
        a <- scalar(param$shape, "shape", lower = 0, lower_open = TRUE)
        mu_obs <- stats::plogis(eta$obs); mu_loc <- stats::plogis(eta$loc)
        b_obs <- log(0.5) / log1p(-mu_obs^a)
        b_loc <- log(0.5) / log1p(-mu_loc^a)
        if (any(!is.finite(c(b_obs, b_loc))) || any(c(b_obs, b_loc) <= 0))
          stop("Kumaraswamy2 produced invalid location-specific shape parameters.", call. = FALSE)
        pkuma <- function(x, aa, bb) 1 - (1 - x^aa)^bb
        qkuma <- function(p, aa, bb) (1 - (1 - p)^(1 / bb))^(1 / aa)
        u <- pkuma(y, a, b_obs)
        qloc <- function(p) pmin0 + (pmax0 - pmin0) * qkuma(p, a, b_loc)
      }
    } else {
      stop(sprintf("Internal error: unsupported continuous copula margin '%s'.", model),
           call. = FALSE)
    }

    list(
      u_obs = clip_probability(u, "marginal PIT"),
      q_loc = function(p) as.numeric(qloc(clip_probability(p, "predictive PIT"))),
      nugget = nugget
    )
  }

  .GeoCopula_latent_param <- function(param, corrmodel, nugget, skew = NULL) {
    corr_names <- CorrParam(corrmodel)
    out <- as.list(param[corr_names])
    out$mean <- 0
    out$nugget <- nugget
    out$sill <- 1
    if (!is.null(skew)) out$skew <- as.numeric(skew)
    out
  }

  margin <- .GeoCopula_cond_margin(model, param, data)

  if (copula == "Gaussian") {
    datanorm <- stats::qnorm(margin$u_obs)
    param_latent <- .GeoCopula_latent_param(param, corrmodel, margin$nugget)

    res <- .GeoSimcondGaussian(
      datanorm, corrmodel, nrep, method, L, param_latent,
      coord_obs, loc, coordt_use, time, NULL, NULL, NULL,
      distance, radius, anisopars, local, neighb, maxdist, maxtime,
      space, spacetime, bivariate, parallel, ncores, progress
    )
    res <- lapply(res, function(z) margin$q_loc(stats::pnorm(z)))
  }

  if (copula == "Clayton") {
    if (!space) {
      stop("Clayton copula conditional simulation is currently implemented only for purely spatial data.",
           call. = FALSE)
    }
    if (is.null(names(param)) || !("nu" %in% names(param))) {
      stop("For copula='Clayton', param$nu is required.", call. = FALSE)
    }
    nu_clayton <- suppressWarnings(as.numeric(param[["nu", exact = TRUE]]))
    if (length(nu_clayton) != 1L || !is.finite(nu_clayton) || nu_clayton <= 0 ||
        abs(nu_clayton - round(nu_clayton)) > sqrt(.Machine$double.eps)) {
      stop("For copula='Clayton', param$nu must be a positive integer.", call. = FALSE)
    }
    nu_clayton <- as.integer(round(nu_clayton))
    param_latent <- .GeoCopula_latent_param(param, corrmodel, margin$nugget)

    u_res <- .GeoSimcondClaytonCopula(
      coord_obs = coord_obs, loc = loc, u_obs = margin$u_obs,
      nu = nu_clayton, latent_param = param_latent, corrmodel = corrmodel,
      nrep = nrep, method = method, L = L,
      distance = distance, radius = radius, anisopars = anisopars,
      parallel = parallel, ncores = ncores, progress = progress,
      n_iter = n_iter
    )
    res <- lapply(u_res, margin$q_loc)
  }

  if (copula == "SkewGaussian") {
    if (!space) {
      stop("Skew-Gaussian copula conditional simulation is currently implemented only for purely spatial data.",
           call. = FALSE)
    }
    if (is.null(names(param)) || !("nu" %in% names(param))) {
      stop("For copula='SkewGaussian', param$nu (eta) is required.", call. = FALSE)
    }
    eta_cop <- suppressWarnings(as.numeric(param[["nu", exact = TRUE]]))
    if (length(eta_cop) != 1L || !is.finite(eta_cop) || abs(eta_cop) >= 1) {
      stop("For copula='SkewGaussian', param$nu is eta and must lie strictly between -1 and 1.",
           call. = FALSE)
    }
    gamma_eta <- eta_cop / sqrt(1 - eta_cop^2)
    omega_eta <- sqrt(1 + gamma_eta^2)

    datanorm <- sn::qsn(
      margin$u_obs, xi = 0, omega = omega_eta, alpha = gamma_eta
    )
    param_latent <- .GeoCopula_latent_param(
      param, corrmodel, margin$nugget, skew = gamma_eta
    )

    res <- .GeoSimcondSkewGaussian(
      coord_obs, loc, datanorm, param_latent, corrmodel,
      local = local, neighb = neighb, L = L, nrep = nrep,
      method = method, distance = distance, radius = radius,
      anisopars = anisopars, parallel = parallel, ncores = ncores,
      progress = progress, n_iter = n_iter
    )
    res <- lapply(res, function(z) {
      u <- sn::psn(z, xi = 0, omega = omega_eta, alpha = gamma_eta)
      margin$q_loc(u)
    })
  }

###########################################################################
}


if (is.null(res) || !is.list(res) || length(res) < 1L) {
  stop("GeoSimcond did not produce any conditional simulations for the requested model.",
       call. = FALSE)
}
expected_nloc <- nrow(loc)
valid_lengths <- vapply(res, length, integer(1L))
if (any(valid_lengths != expected_nloc)) {
  stop("A conditional simulation has an unexpected number of prediction values.",
       call. = FALSE)
}
if (any(!vapply(res, function(z) all(is.finite(z)), logical(1L)))) {
  stop("At least one conditional simulation contains non-finite values.",
       call. = FALSE)
}

## Summarize without materializing a second nrep x nloc dense matrix.  The
## conditional simulations are already retained in res, so use a streaming
## Welford update for the mean and sample variance.
if (length(res) == 1L) {
  cond_mean <- as.numeric(res[[1L]])
  cond_var <- rep(NA_real_, length(cond_mean))
} else {
  cond_mean <- numeric(expected_nloc)
  cond_M2 <- numeric(expected_nloc)
  k <- 0L
  for (z in res) {
    k <- k + 1L
    z <- as.numeric(z)
    delta <- z - cond_mean
    cond_mean <- cond_mean + delta / k
    cond_M2 <- cond_M2 + delta * (z - cond_mean)
  }
  cond_var <- cond_M2 / (k - 1L)
}
############################################################################
############################################################################
############################ OUTPUT ########################################
GeoSimcond_out <- list(
    bivariate = bivariate,
    coordx = coordx,
    coordy = coordy,
    coordz = coordz,
    coordt = coordt,
    corrmodel = corrmodel,
    data = data,
    distance = distance,
    grid = grid,
    loc = loc,
    copula = copula,
    numcoord = nrow(coord_obs),
    numloc = nrow(loc),
    numtime = length(coordt),
    numt = length(time),
    maxdist = maxdist,
    maxtime = maxtime,
    model = model,
    n = n,
    nloc = nloc,
    param = param,
    method = method,
    local = local,
    neighb = neighb,
    L = L,
    mcmc_thin = mcmc_thin,
    condsim = res,            # list of nrep vectors, each of length numloc
    cond_mean = cond_mean,    # conditional mean across replications
    cond_var = cond_var,      # conditional variance across replications
    radius = radius,
    spacetime = spacetime,
    time = time
  )
  structure(c(GeoSimcond_out, call = call), class = c("GeoSimcond"))
}

# Internal engines used by GeoSimcond().
# Keeping them outside the public function avoids rebuilding large closures on
# every call and keeps the user-facing implementation focused on dispatch.

.GeoSimcondNormalizeSimulations <- function(x, expected, expected_length = NULL) {
  expected <- as.integer(expected)[1L]
  if (is.list(x)) {
    if (length(x) != expected) stop("Unexpected number of simulations.", call. = FALSE)
    out <- lapply(x, as.numeric)
  } else if (is.matrix(x)) {
    if (!is.null(expected_length) &&
        ncol(x) == expected && nrow(x) == expected_length) {
      out <- lapply(seq_len(expected), function(j) as.numeric(x[, j]))
    } else if (!is.null(expected_length) &&
               nrow(x) == expected && ncol(x) == expected_length) {
      out <- lapply(seq_len(expected), function(j) as.numeric(x[j, ]))
    } else if (is.null(expected_length) && ncol(x) == expected) {
      out <- lapply(seq_len(expected), function(j) as.numeric(x[, j]))
    } else if (is.null(expected_length) && nrow(x) == expected) {
      out <- lapply(seq_len(expected), function(j) as.numeric(x[j, ]))
    } else if (expected == 1L &&
               (is.null(expected_length) || length(x) == expected_length)) {
      out <- list(as.numeric(x))
    } else {
      stop("Unexpected matrix format for simulations.", call. = FALSE)
    }
  } else if (expected == 1L) {
    out <- list(as.numeric(x))
  } else {
    stop("Unexpected format of simulations.", call. = FALSE)
  }

  if (!is.null(expected_length) &&
      any(vapply(out, length, integer(1L)) != expected_length)) {
    stop("A simulation has an unexpected length.", call. = FALSE)
  }
  out
}

.GeoSimcondSkewGaussian <- function(
    coord_obs, loc, data, param, corrmodel,
    nrep = 1, method = "Cholesky", local = FALSE,
    neighb = NULL, L = NULL,
    distance = "Eucl", radius = 1, anisopars = NULL,
    parallel = FALSE, ncores = 1, progress = FALSE,
    n_iter = 1000L, batch_size = NULL, max_batch_mb = 256,
    mean_obs = NULL, mean_loc = NULL
) {

  ##########################################################################
  ## Parameters and checks
  ##########################################################################

  sill       <- as.numeric(param$sill)
  skew       <- as.numeric(param$skew)
  sigma_sqrt <- sqrt(sill)

  if (length(sill) != 1L || !is.finite(sill) || sill <= 0) {
    stop("param$sill must be strictly positive.")
  }

  if (length(skew) != 1L || !is.finite(skew)) {
    stop("param$skew must be finite.")
  }

  if (length(nrep) != 1L || !is.finite(nrep) || nrep < 1) {
    stop("nrep must be at least 1.")
  }

  if (length(n_iter) != 1L || !is.finite(n_iter) || n_iter < 1) {
    stop("n_iter must be at least 1.")
  }

  if (local) {
    stop(
      "Local prediction cannot be used with the ",
      "skew-Gaussian Gibbs sampler."
    )
  }

  if (!(method %in% c("Cholesky", "TB", "CE"))) {
    stop("method must be one of 'Cholesky', 'TB', or 'CE'.")
  }

  nrep  <- as.integer(nrep)
  n_iter <- as.integer(n_iter)

  coord_obs <- as.matrix(coord_obs)
  loc       <- as.matrix(loc)
  data      <- as.numeric(data)

  M0 <- nrow(coord_obs)
  N0 <- nrow(loc)

  if(is.null(mean_obs)) mean_obs <- rep(as.numeric(param$mean)[1L], M0)
  if(is.null(mean_loc)) mean_loc <- rep(as.numeric(param$mean)[1L], N0)
  mean_obs <- as.numeric(mean_obs)
  mean_loc <- as.numeric(mean_loc)
  if(length(mean_obs) != M0 || any(!is.finite(mean_obs)))
    stop("mean_obs must contain one finite value per observation.")
  if(length(mean_loc) != N0 || any(!is.finite(mean_loc)))
    stop("mean_loc must contain one finite value per prediction point.")
  ## Existing worker code uses mm for the prediction-side location vector.
  mm <- mean_loc

  if (length(data) != M0) {
    stop(
      "The length of data must match the number ",
      "of observed locations."
    )
  }

  if (N0 < 1L) {
    stop("loc must contain at least one prediction location.")
  }

  ##########################################################################
  ## Standardized latent representation
  ##
  ## Y(s) = mm + skew |G1(s)| + sqrt(sill) G2(s)
  ##
  ## Z(s) = gamma |G1(s)| + G2(s),
  ## gamma = skew / sqrt(sill)
  ##########################################################################

  gamma <- skew / sigma_sqrt
  z_obs <- (data - mean_obs) / sigma_sqrt

  if (any(!is.finite(z_obs))) {
    stop("The standardized observations contain non-finite values.")
  }

  gaussian_case <- isTRUE(gamma == 0)

  keep       <- CorrParam(corrmodel)
  corr_param <- param[keep]

  latent_nugget <- if (is.null(param$nugget)) 0 else as.numeric(param$nugget)
  if (length(latent_nugget) != 1L || !is.finite(latent_nugget) ||
      latent_nugget < 0 || latent_nugget >= 1) {
    stop("param$nugget must be a finite scalar in [0, 1).")
  }

  latent_param <- c(
    corr_param,
    mean = 0,
    nugget = latent_nugget,
    sill = 1
  )

  ##########################################################################
  ## Gaussian covariance matrix and kriging weights
  ##########################################################################

  cat("Computing Kriging weights ...\n")

  GeoW <- GeoKrigWeights(
    coordx = coord_obs,
    corrmodel = corrmodel,
    loc = loc,
    model = "Gaussian",
    distance = distance,
    radius = radius,
    anisopars = anisopars,
    param = latent_param
  )

  Wglob <- GeoW$weights

  if (is.null(Wglob)) {
    stop("GeoKrigWeights did not return the kriging weights.")
  }

  if (is.null(dim(Wglob))) {
    Wglob <- matrix(Wglob, nrow = M0)
  }

  weights_orientation <- GeoW$weights_orientation
  if (is.null(weights_orientation) ||
      !(weights_orientation %in% c("nobs_by_nloc", "nloc_by_nobs"))) {
    stop("GeoKrigWeights returned an unknown weights orientation.")
  }
  if (weights_orientation == "nobs_by_nloc" &&
      (nrow(Wglob) != M0 || ncol(Wglob) != N0)) {
    stop("The nobs_by_nloc kriging weights have incompatible dimensions.")
  }
  if (weights_orientation == "nloc_by_nobs" &&
      (nrow(Wglob) != N0 || ncol(Wglob) != M0)) {
    stop("The nloc_by_nobs kriging weights have incompatible dimensions.")
  }

  # The precision matrix is unnecessary when gamma = 0.
  if (gaussian_case) {

    Sigma.mat.inv <- numeric(0L)

  } else {

    if (is.null(GeoW$covmatrix)) {
      stop(
        "GeoKrigWeights did not return the ",
        "latent covariance matrix."
      )
    }

    Sigma.mat.inv <- MatInv(mtx = GeoW$covmatrix)

    if (!is.matrix(Sigma.mat.inv) ||
        !identical(dim(Sigma.mat.inv), c(M0, M0)) ||
        any(!is.finite(Sigma.mat.inv))) {
      stop(
        "The latent covariance matrix could not be inverted reliably."
      )
    }
  }

  all_coord <- rbind(coord_obs, loc)

  ##########################################################################
  ## Batch size
  ##########################################################################

  fields_per_rep <- if (gaussian_case) 1L else 2L

  if (is.null(batch_size)) {

    if (length(max_batch_mb) != 1L ||
        !is.finite(max_batch_mb) ||
        max_batch_mb <= 0) {
      stop("max_batch_mb must be strictly positive.")
    }

    bytes_per_rep <- fields_per_rep * (M0 + N0) * 8

    batch_size <- floor(
      max_batch_mb * 1024^2 / max(bytes_per_rep, 1)
    )

    batch_size <- max(
      1L,
      min(nrep, as.integer(batch_size))
    )

  } else {

    if (length(batch_size) != 1L ||
        !is.finite(batch_size) ||
        batch_size < 1) {
      stop("batch_size must be at least 1.")
    }

    batch_size <- max(
      1L,
      min(nrep, as.integer(batch_size))
    )
  }

  ##########################################################################
  ## Normalize the output of GeoSim or GeoSimapprox
  ##########################################################################

  ##########################################################################
  ## Generate one batch of unconditional Gaussian simulations
  ##########################################################################

  generate_batch <- function(current_nrep) {

    current_fields <- fields_per_rep * current_nrep

    if (method %in% c("TB", "CE")) {

      sim_obj <- GeoSimapprox(
        coordx = all_coord,
        corrmodel = corrmodel,
        method = method,
        model = "Gaussian",
        param = latent_param,
        L = L,
        nrep = current_fields,
        progress = FALSE,
        parallel = FALSE,
        distance = distance,
        radius = radius,
        anisopars = anisopars
      )

    } else {

      sim_obj <- GeoSim(
        coordx = all_coord,
        corrmodel = corrmodel,
        model = "Gaussian",
        param = latent_param,
        nrep = current_fields,
        progress = FALSE,
        distance = distance,
        radius = radius,
        anisopars = anisopars
      )
    }

    sim_data <- .GeoSimcondNormalizeSimulations(
      sim_obj$data,
      current_fields
    )

    if (gaussian_case) {

      tasks <- lapply(
        seq_len(current_nrep),
        function(j) {
          list(
            g2 = as.numeric(sim_data[[j]])
          )
        }
      )

    } else {

      tasks <- lapply(
        seq_len(current_nrep),
        function(j) {
          list(
            g1 = as.numeric(sim_data[[2L * j - 1L]]),
            g2 = as.numeric(sim_data[[2L * j]])
          )
        }
      )
    }

    rm(sim_obj, sim_data)

    tasks
  }

  ##########################################################################
  ## Self-contained worker
  ##
  ## Only the Gaussian simulations required by the current task are passed
  ## to the worker. The complete batch is not captured as a global object.
  ##########################################################################

  worker_fun <- function(
      task, z_obs, M0, N0, gamma, mm, skew, sigma_sqrt,
      Sigma_mat_inv, Wglob, weights_orientation,
      n_iter, gaussian_case
  ) {

    tryCatch({

      apply_weights <- function(delta) {

        if (weights_orientation == "nobs_by_nloc") {
          as.numeric(crossprod(Wglob, delta))
        } else {
          as.numeric(Wglob %*% delta)
        }
      }

      conditional_field <- function(
          simulation,
          latent_obs
      ) {

        if (length(simulation) != M0 + N0) {
          stop(
            "An unconditional Gaussian simulation ",
            "has an incorrect length."
          )
        }

        sim_obs <- simulation[seq_len(M0)]
        sim_loc <- simulation[M0 + seq_len(N0)]

        sim_loc + apply_weights(
          latent_obs - sim_obs
        )
      }

      ######################################################################
      ## Gaussian special case
      ######################################################################

      if (gaussian_case) {

        g2_loc <- conditional_field(
          task$g2,
          z_obs
        )

        return(
          list(
            ok = TRUE,
            value = mm + sigma_sqrt * g2_loc
          )
        )
      }

      ######################################################################
      ## Initial state satisfying
      ##
      ## z_i = gamma |G1_i| + G2_i
      ######################################################################
      g1_init <- stats::rnorm(M0)
      g2_init <- z_obs - gamma * abs(g1_init)
      ######################################################################
      ## Exact Gibbs kernel implemented in C
      ######################################################################
      gibbs_result <- dotCall64::.C64(
        "skew_gaussian_gibbs_sampler",
        SIGNATURE = c(
          "double", "integer", "double", "double",
          "integer", "double", "double"
        ),
        data_obs = as.double(z_obs),
        n = as.integer(M0),
        Sigma_mat_inv = as.double(Sigma_mat_inv),
        eta = as.double(c(gamma, 0)),
        n_iter = as.integer(n_iter),
        data_x = as.double(g1_init),
        data_y = as.double(g2_init),
        INTENT = c(
          "r", "r", "r", "r",
          "r", "rw", "rw"
        ),
        VERBOSE = 0,
        NAOK = FALSE,
        PACKAGE = "GeoModels"
      )

      g1_obs <- as.numeric(gibbs_result$data_x)
      g2_obs <- as.numeric(gibbs_result$data_y)

      if (length(g1_obs) != M0 ||
          length(g2_obs) != M0 ||
          any(!is.finite(g1_obs)) ||
          any(!is.finite(g2_obs))) {
        stop(
          "The Gibbs sampler returned invalid ",
          "latent values."
        )
      }

      ######################################################################
      ## Check the nonlinear restriction
      ######################################################################
      constraint_error <- max(
        abs(
          z_obs -
            gamma * abs(g1_obs) -
            g2_obs
        )
      )
      constraint_tolerance <- 1e-7 * (
        1 + max(abs(z_obs))
      )

      if (!is.finite(constraint_error) ||
          constraint_error > constraint_tolerance) {
        stop(
          "The Gibbs output does not satisfy the ",
          "skew-Gaussian constraint."
        )
      }
      ######################################################################
      ## Gaussian substitution conditional simulations
      ######################################################################

      g1_loc <- conditional_field(
        task$g1,
        g1_obs
      )

      g2_loc <- conditional_field(
        task$g2,
        g2_obs
      )

      ######################################################################
      ## Back transformation
      ##
      ## Y(s) = mm + skew |G1(s)| + sqrt(sill) G2(s)
      ######################################################################

      list(
        ok = TRUE,
        value =
          mm +
          skew * abs(g1_loc) +
          sigma_sqrt * g2_loc
      )

    }, error = function(e) {

      list(
        ok = FALSE,
        error = conditionMessage(e)
      )
    })
  }

  ##########################################################################
  ## Parallel configuration
  ##########################################################################

  ncores <- .GeoResolveWorkers(parallel, ncores, n_jobs = nrep)
  parallel <- isTRUE(parallel) && nrep > 1L && ncores > 1L
  if (parallel) {
    old_plan <- future::plan()

    on.exit(
      future::plan(old_plan),
      add = TRUE
    )

    old_limit <- getOption(
      "future.globals.maxSize"
    )

    static_size <- sum(
      as.numeric(object.size(Sigma.mat.inv)),
      as.numeric(object.size(Wglob)),
      as.numeric(object.size(z_obs))
    )

    required_limit <- max(
      500 * 1024^2,
      2 * static_size + max_batch_mb * 1024^2
    )

    if (!is.null(old_limit)) {
      required_limit <- max(
        required_limit,
        old_limit
      )
    }

    options(
      future.globals.maxSize = required_limit
    )
    on.exit(
      options(
        future.globals.maxSize = old_limit
      ),
      add = TRUE
    )
    future::plan(
      future::multisession,
      workers = ncores
    )
  }
  ##########################################################################
  ## Batch execution
  ##########################################################################

  run_batches <- function(progressor) {

    res <- vector("list", nrep)
    failure_messages <- character(0L)
    batch_starts <- seq.int(
      1L,
      nrep,
      by = batch_size
    )
    for (batch_start in batch_starts) {

      batch_end <- min(
        nrep,
        batch_start + batch_size - 1L
      )
      current_indices <- batch_start:batch_end
      current_nrep <- length(current_indices)
      tasks <- generate_batch(current_nrep)
      if (parallel && current_nrep > 1L) {

        batch_result <- future.apply::future_lapply(
          tasks,
          worker_fun,
          z_obs = z_obs,
          M0 = M0,
          N0 = N0,
          gamma = gamma,
          mm = mm,
          skew = skew,
          sigma_sqrt = sigma_sqrt,
          Sigma_mat_inv = Sigma.mat.inv,
          Wglob = Wglob,
          weights_orientation = weights_orientation,
          n_iter = n_iter,
          gaussian_case = gaussian_case,
          future.seed = TRUE,

          # One future chunk per worker. This avoids repeatedly
          # serializing Sigma.mat.inv and Wglob for every simulation.
          future.scheduling = 1,

          # worker_fun is self-contained and all data are passed
          # explicitly.
          future.globals = FALSE,

          future.packages = c(
            "GeoModels",
            "dotCall64"
          )
        )

      } else {

        batch_result <- lapply(
          tasks,
          worker_fun,
          z_obs = z_obs,
          M0 = M0,
          N0 = N0,
          gamma = gamma,
          mm = mm,
          skew = skew,
          sigma_sqrt = sigma_sqrt,
          Sigma_mat_inv = Sigma.mat.inv,
          Wglob = Wglob,
          weights_orientation = weights_orientation,
          n_iter = n_iter,
          gaussian_case = gaussian_case
        )
      }

      for (j in seq_along(batch_result)) {

        global_index <- current_indices[j]

        if (isTRUE(batch_result[[j]]$ok)) {

          res[[global_index]] <-
            batch_result[[j]]$value

        } else {

          failure_messages <- c(
            failure_messages,
            sprintf(
              "simulation %d: %s",
              global_index,
              batch_result[[j]]$error
            )
          )
        }

        progressor()
      }

      rm(tasks, batch_result)
      gc(verbose = FALSE)
    }

    ########################################################################
    ## Remove failed simulations
    ########################################################################

    failed <- vapply(
      res,
      is.null,
      logical(1L)
    )
    if (any(failed)) {
      details <- if (length(failure_messages)) {
        paste(
          utils::head(failure_messages, 3L),
          collapse = "; "
        )
      } else {
        "unknown error"
      }
      warning(
        sprintf(
          "%d simulations out of %d failed. %s",
          sum(failed),
          nrep,
          details
        ),
        call. = FALSE
      )
      res <- res[!failed]
    }
    if (!length(res)) {
      stop("All conditional simulations failed.")
    }

    res
  }
  ##########################################################################
  ## Run
  ##########################################################################
  cat(
    "Performing",
    nrep,
    "conditional simulations",
    if (parallel) {
      paste0(" using ", ncores, " cores")
    } else {
      ""
    },
    "...\n"
  )
  if (progress) {
    restore_progress_handlers <- .GeoProgressHandlersPush(
      handler = "txtprogressbar", global = TRUE
    )
    on.exit(restore_progress_handlers(), add = TRUE)
    res <- progressr::with_progress({
      p <- progressr::progressor(
        steps = nrep
      )
      run_batches(p)
    })
  } else {
    res <- run_batches(
      function(...) invisible(NULL)
    )
  }
  res
}
#################################################################################
### Helper function for conditional Clayton-like copula simulation
#################################################################################
.GeoSimcondClaytonCopula <- function(
    coord_obs, loc, u_obs, nu, latent_param, corrmodel,
    nrep = 1, method = "Cholesky", L = 1000,
    distance = "Eucl", radius = 1, anisopars = NULL,
    parallel = FALSE, ncores = 1, progress = FALSE,
    n_iter = 1000L, max_batch_mb = 256
) {
  coord_obs <- as.matrix(coord_obs)
  loc <- as.matrix(loc)
  u_obs <- as.numeric(u_obs)
  M0 <- nrow(coord_obs)
  N0 <- nrow(loc)

  nu_raw <- as.numeric(nu)
  if (length(nu_raw) != 1L || !is.finite(nu_raw) || nu_raw <= 0 ||
      abs(nu_raw - round(nu_raw)) > sqrt(.Machine$double.eps)) {
    stop("For copula='Clayton', param$nu must be a positive integer.", call. = FALSE)
  }
  nu <- as.integer(round(nu_raw))
  d_latent <- nu + 2L

  if (length(u_obs) != M0 || any(!is.finite(u_obs)) ||
      any(u_obs <= 0 | u_obs >= 1)) {
    stop("Clayton latent uniforms must lie strictly between 0 and 1.", call. = FALSE)
  }
  if (!(method %in% c("Cholesky", "TB"))) {
    stop("Clayton copula conditional simulation supports method='Cholesky' or method='TB'.",
         call. = FALSE)
  }

  ## U_nu = R^(nu/2), where R=A/(A+B).
  ratio_obs <- exp((2 / nu) * log(u_obs))
  eps_r <- .Machine$double.eps
  ratio_obs <- pmin(pmax(ratio_obs, eps_r), 1 - eps_r)

  cat("Computing Clayton latent Kriging weights ...\n")
  GeoW <- GeoKrigWeights(
    coordx = coord_obs, corrmodel = corrmodel, loc = loc,
    model = "Gaussian", distance = distance, radius = radius,
    anisopars = anisopars, param = latent_param
  )
  if (is.null(GeoW$covmatrix) || is.null(GeoW$weights)) {
    stop("GeoKrigWeights did not return the covariance matrix and weights required by the Clayton sampler.",
         call. = FALSE)
  }

  Sigma_inv <- MatInv(mtx = GeoW$covmatrix)
  if (!is.matrix(Sigma_inv) || !identical(dim(Sigma_inv), c(M0, M0)) ||
      any(!is.finite(Sigma_inv))) {
    stop("The Clayton latent covariance matrix could not be inverted reliably.", call. = FALSE)
  }

  Wglob <- GeoW$weights
  if (is.null(dim(Wglob))) Wglob <- matrix(Wglob, nrow = M0)
  weights_orientation <- GeoW$weights_orientation
  if (is.null(weights_orientation) ||
      !(weights_orientation %in% c("nobs_by_nloc", "nloc_by_nobs"))) {
    stop("GeoKrigWeights returned an unknown weights orientation.", call. = FALSE)
  }
  if (weights_orientation == "nobs_by_nloc" &&
      (nrow(Wglob) != M0 || ncol(Wglob) != N0)) {
    stop("The nobs_by_nloc Clayton latent weights have incompatible dimensions.",
         call. = FALSE)
  }
  if (weights_orientation == "nloc_by_nobs" &&
      (nrow(Wglob) != N0 || ncol(Wglob) != M0)) {
    stop("The nloc_by_nobs Clayton latent weights have incompatible dimensions.",
         call. = FALSE)
  }

  apply_weights <- function(delta) {
    if (weights_orientation == "nobs_by_nloc") {
      as.numeric(crossprod(Wglob, delta))
    } else {
      as.numeric(Wglob %*% delta)
    }
  }

  initialize_latent <- function() {
    U <- matrix(stats::rnorm(M0 * d_latent), nrow = M0, ncol = d_latent)
    for (ii in seq_len(M0)) {
      q <- sqrt(sum(U[ii, ]^2))
      if (!is.finite(q) || q <= 0) q <- sqrt(d_latent)

      za <- U[ii, seq_len(nu)]
      nz <- sqrt(sum(za^2))
      if (!is.finite(nz) || nz <= 0) {
        za <- rep(0, nu); za[1L] <- 1; nz <- 1
      }
      wb <- U[ii, nu + seq_len(2L)]
      nw <- sqrt(sum(wb^2))
      if (!is.finite(nw) || nw <= 0) {
        wb <- c(1, 0); nw <- 1
      }
      U[ii, seq_len(nu)] <- q * sqrt(ratio_obs[ii]) * za / nz
      U[ii, nu + seq_len(2L)] <- q * sqrt(1 - ratio_obs[ii]) * wb / nw
    }
    U
  }

  run_one <- function(field_list) {
    if (length(field_list) != d_latent)
      stop("Internal Clayton latent-field count mismatch.", call. = FALSE)

    U0 <- initialize_latent()
    gibbs <- dotCall64::.C64(
      "clayton_gibbs_sampler",
      SIGNATURE = c("double", "double", "integer", "integer", "double", "integer"),
      SigmaInv = as.double(Sigma_inv),
      ratio = as.double(ratio_obs),
      n = as.integer(M0),
      nu = as.integer(nu),
      U = as.double(U0),
      nIter = as.integer(n_iter),
      INTENT = c("r", "r", "r", "r", "rw", "r"),
      VERBOSE = 0, NAOK = FALSE, PACKAGE = "GeoModels"
    )
    Uobs <- matrix(gibbs$U, nrow = M0, ncol = d_latent)
    if (any(!is.finite(Uobs)))
      stop("The Clayton latent Gibbs sampler returned non-finite values.", call. = FALSE)

    ## The C sampler enforces A/(A+B)=ratio_obs at every Gibbs update and
    ## performs a final round-off projection onto that manifold.  Recompute the
    ## ratio with scaled row norms rather than raw sums of squares: this is more
    ## stable when one of the two latent radii is very small.
    Zobs <- Uobs[, seq_len(nu), drop = FALSE]
    Wobs <- Uobs[, nu + seq_len(2L), drop = FALSE]
    znorm <- sqrt(rowSums(Zobs^2))
    wnorm <- sqrt(rowSums(Wobs^2))
    qnorm <- sqrt(znorm^2 + wnorm^2)
    ratio_check <- (znorm / qnorm)^2
    ratio_err <- abs(ratio_check - ratio_obs)
    if (any(!is.finite(ratio_check)) || any(!is.finite(ratio_err))) {
      stop("The Clayton latent Gibbs sampler returned an invalid observed ratio constraint.",
           call. = FALSE)
    }
    max_ratio_err <- max(ratio_err)
    ## A discrepancy above this level is no longer attributable to ordinary
    ## floating-point round-off and should still fail loudly.
    if (max_ratio_err > 1e-7) {
      stop(sprintf(
        "The Clayton latent Gibbs sampler violated the observed ratio constraint (max abs error %.3g).",
        max_ratio_err
      ), call. = FALSE)
    }

    Upred <- matrix(0, nrow = N0, ncol = d_latent)
    for (jj in seq_len(d_latent)) {
      simj <- as.numeric(field_list[[jj]])
      if (length(simj) != M0 + N0)
        stop("A Clayton latent Gaussian simulation has an invalid length.", call. = FALSE)
      sim_obs <- simj[seq_len(M0)]
      sim_loc <- simj[M0 + seq_len(N0)]
      Upred[, jj] <- sim_loc + apply_weights(Uobs[, jj] - sim_obs)
    }

    Apred <- rowSums(Upred[, seq_len(nu), drop = FALSE]^2)
    Bpred <- rowSums(Upred[, nu + seq_len(2L), drop = FALSE]^2)
    den <- Apred + Bpred
    if (any(!is.finite(den)) || any(den <= 0))
      stop("Invalid Clayton latent radius at a prediction location.", call. = FALSE)
    beta_pred <- pmin(pmax(Apred / den, 0), 1)
    u_pred <- beta_pred^(nu / 2)
    pmin(pmax(u_pred, .Machine$double.eps), 1 - .Machine$double.eps)
  }

  all_coord <- rbind(coord_obs, loc)
  bytes_per_rep <- d_latent * (M0 + N0) * 8
  batch_size <- floor(max_batch_mb * 1024^2 / max(bytes_per_rep, 1))
  batch_size <- max(1L, min(nrep, as.integer(batch_size)))
  n_batches <- ceiling(nrep / batch_size)
  res <- vector("list", nrep)

  if (isTRUE(parallel) && nrep > 1L) {
    old_plan <- future::plan()
    on.exit(future::plan(old_plan), add = TRUE)
    future::plan(future::multisession, workers = ncores)
  }

  cat("Performing", nrep, "Clayton conditional simulations",
      if (isTRUE(parallel) && nrep > 1L) paste0(" using ", ncores, " cores") else "",
      "...\n")

  out_pos <- 1L
  for (bb in seq_len(n_batches)) {
    nb <- min(batch_size, nrep - out_pos + 1L)
    nfields <- d_latent * nb

    if (method == "TB") {
      gs <- GeoSimapprox(
        coordx = all_coord, corrmodel = corrmodel, method = "TB",
        model = "Gaussian", param = latent_param, L = L,
        nrep = nfields, progress = FALSE, parallel = FALSE,
        distance = distance, radius = radius, anisopars = anisopars
      )
    } else {
      gs <- GeoSim(
        coordx = all_coord, corrmodel = corrmodel,
        model = "Gaussian", param = latent_param,
        nrep = nfields, progress = FALSE,
        distance = distance, radius = radius, anisopars = anisopars
      )
    }
    gdata <- .GeoSimcondNormalizeSimulations(gs$data, nfields, M0 + N0)
    groups <- split(seq_len(nfields), ceiling(seq_len(nfields) / d_latent))

    one_batch <- function(ii) run_one(gdata[groups[[ii]]])
    if (isTRUE(parallel) && nb > 1L) {
      batch_res <- future.apply::future_lapply(
        seq_len(nb), one_batch, future.seed = TRUE,
        future.packages = c("GeoModels", "dotCall64")
      )
    } else {
      batch_res <- lapply(seq_len(nb), one_batch)
    }

    res[out_pos:(out_pos + nb - 1L)] <- batch_res
    out_pos <- out_pos + nb
    if (isTRUE(progress))
      cat(sprintf("  Clayton batch %d/%d completed\n", bb, n_batches))
    rm(gs, gdata, batch_res)
  }

  res
}

################################################################################# 
### Helper function for conditional Gamma simulation 
################################################################################# 
################################################################################# 
.GeoSimcondGamma <- function(coord_obs, loc, data, param, corrmodel,
                         nrep = 1, method = "Cholesky", local = FALSE, neighb = NULL,
                         L = NULL, distance = "Eucl", radius = 1, anisopars = NULL,
                         parallel = FALSE, ncores = 1, progress = FALSE,
                         mean_obs = NULL, mean_loc = NULL) {
  shape_raw <- as.numeric(param$shape)
  if (length(shape_raw) != 1L || !is.finite(shape_raw) || shape_raw <= 0 ||
      abs(shape_raw - round(shape_raw)) > sqrt(.Machine$double.eps)) {
    stop("Direct Gamma conditional simulation requires param$shape to be a positive integer.",
         call. = FALSE)
  }
  shape <- as.integer(round(shape_raw))

  keep <- CorrParam(corrmodel)
  corr_param <- param[keep]
  latent_nugget <- if (is.null(param$nugget)) 0 else as.numeric(param$nugget)
  if (length(latent_nugget) != 1L || !is.finite(latent_nugget) ||
      latent_nugget < 0 || latent_nugget >= 1) {
    stop("param$nugget must be a finite scalar in [0, 1).")
  }
  latent_param <- c(corr_param, list(mean = 0, nugget = latent_nugget, sill = 1))
  
  cat("Computing Kriging weights ...\n")
  if (!local) 
    GeoW <- GeoKrigWeights(coordx = coord_obs, corrmodel = corrmodel,
                           loc = loc, model = "Gaussian",
                           distance = distance, radius = radius,
                           anisopars = anisopars,
                           param = latent_param)
  else stop("Local option cannot be used for Gibbs sampling.\n")
  
  # Static quantities used by all Gibbs iterations.
  Sigma.mat.inv <- MatInv(mtx = GeoW$covmatrix)
  M0 <- nrow(coord_obs)
  N0 <- nrow(loc)
  Wglob <- GeoW$weights
  weights_orientation <- GeoW$weights_orientation
  if (is.null(weights_orientation) ||
      !(weights_orientation %in% c("nobs_by_nloc", "nloc_by_nobs"))) {
    stop("GeoKrigWeights returned an unknown weights orientation.")
  }
  if (!is.matrix(Sigma.mat.inv) ||
      !identical(dim(Sigma.mat.inv), c(M0, M0)) ||
      any(!is.finite(Sigma.mat.inv))) {
    stop("The latent covariance matrix could not be inverted reliably.")
  }
  if (weights_orientation == "nobs_by_nloc" &&
      (nrow(Wglob) != M0 || ncol(Wglob) != N0)) {
    stop("The nobs_by_nloc kriging weights have incompatible dimensions.")
  }
  if (weights_orientation == "nloc_by_nobs" &&
      (nrow(Wglob) != N0 || ncol(Wglob) != M0)) {
    stop("The nloc_by_nobs kriging weights have incompatible dimensions.")
  }
  if(is.null(mean_obs)) mean_obs <- rep(as.numeric(param$mean)[1L], M0)
  if(is.null(mean_loc)) mean_loc <- rep(as.numeric(param$mean)[1L], N0)
  mean_obs <- as.numeric(mean_obs)
  mean_loc <- as.numeric(mean_loc)
  if(length(mean_obs) != M0 || any(!is.finite(mean_obs)))
    stop("mean_obs must contain one finite value per observation.")
  if(length(mean_loc) != N0 || any(!is.finite(mean_loc)))
    stop("mean_loc must contain one finite value per prediction point.")
  data_transformed <- exp(-mean_obs) * data
  if (any(!is.finite(data_transformed)) || any(data_transformed < 0)) {
    stop("Gamma conditional simulation requires finite non-negative observations.")
  }
  # gamma_gibbs_sampler constrains ||U_i||^2 = 2 * y_i.  For the
  # GeoSim Gamma construction, ||U_i||^2 must equal shape * Y_i / exp(eta_i).
  gamma_constraint <- 0.5 * shape * data_transformed
  exp_mean_over_shape <- exp(mean_loc) / shape

  # STRATEGIA MEMORY-EFFICIENT: Genera simulazioni in batch
  if (parallel && nrep > 1) {
    
    # Calcola memoria necessaria per un batch
    coord_sim <- rbind(coord_obs, loc)
    n_total <- nrow(coord_sim)
    
    # Stima: ogni simulazione ~ 8 bytes * n_total
    memory_per_sim_mb <- (8 * n_total) / (1024^2)
    memory_per_batch_mb <- memory_per_sim_mb * shape
    
    # Determina batch size: massimo 200 MB per batch
    max_batch_memory_mb <- 200
    sims_per_batch <- max(1, floor(max_batch_memory_mb / memory_per_batch_mb))
    
    #cat(sprintf("Memory-efficient mode: processing %d simulations per batch\n", sims_per_batch))
    #cat(sprintf("Total batches needed: %d\n", ceiling(nrep / sims_per_batch)))
    
    # Setup progress
    if (progress) {
      restore_progress_handlers <- .GeoProgressHandlersPush(
        handler = "txtprogressbar", global = TRUE
      )
      on.exit(restore_progress_handlers(), add = TRUE)
      pb <- progressr::progressor(
        along = seq_len(nrep), on_exit = FALSE, auto_finish = FALSE
      )
      ## Finish before restoring the temporary global handler, including when
      ## a batch or Gibbs simulation exits early.
      on.exit(pb(type = "finish"), add = TRUE, after = FALSE)
    }
    
    # Parallelization by batch. Preserve the user's future plan.
    old_plan_batch <- future::plan()
    on.exit(try(future::plan(old_plan_batch), silent = TRUE), add = TRUE)
    future::plan(future::multisession, workers = ncores)

    res <- vector("list", nrep)
    n_batches <- ceiling(nrep / sims_per_batch)
    
    for (batch_idx in seq_len(n_batches)) {
      start_idx <- (batch_idx - 1) * sims_per_batch + 1
      end_idx <- min(batch_idx * sims_per_batch, nrep)
      batch_nrep <- end_idx - start_idx + 1
      
      cat(sprintf("\n=== Batch %d/%d: simulations %d to %d ===\n", 
                  batch_idx, n_batches, start_idx, end_idx))
      
      # Genera simulazioni Gaussiane SOLO per questo batch
      if (method == "TB" || method == "CE") {
        Gauss_batch <- GeoSimapprox(
          coordx = coord_sim, 
          corrmodel = corrmodel, 
          method = method, 
          model = "Gaussian", 
          param = latent_param, 
          L = L, 
          nrep = shape * batch_nrep, 
          progress = FALSE, 
          parallel = FALSE,
          distance = distance,
          radius = radius,
          anisopars = anisopars
        )
      } else {
        Gauss_batch <- GeoSim(
          coordx = coord_sim, 
          corrmodel = corrmodel, 
          model = "Gaussian", 
          param = latent_param, 
          nrep = shape * batch_nrep, 
          progress = FALSE,
          distance = distance,
          radius = radius,
          anisopars = anisopars
        )
      }
      
      pair_idx_batch <- split(seq_len(shape * batch_nrep), 
                             ceiling(seq_len(shape * batch_nrep) / shape))
      
      # Normalize also the single-field case returned as a vector.
      Gauss_data_batch <- .GeoSimcondNormalizeSimulations(
        Gauss_batch$data, shape * batch_nrep, M0 + N0
      )
      
      # Variabili essenziali per questo batch
      essential_vars <- list(
        shape = shape,
        data_transformed = data_transformed,
        gamma_constraint = gamma_constraint,
        M0 = M0,
        N0 = N0,
        Sigma.mat.inv = Sigma.mat.inv,
        Wglob = Wglob,
        weights_orientation = weights_orientation,
        exp_mean_over_shape = exp_mean_over_shape,
        pair_idx = pair_idx_batch
      )
      
      # Worker function ottimizzata
      run_simulation_worker <- function(local_i, vars, Gauss_data) {
        shape <- vars$shape
        data_transformed <- vars$data_transformed
        gamma_constraint <- vars$gamma_constraint
        M0 <- vars$M0
        N0 <- vars$N0
        Sigma.mat.inv <- vars$Sigma.mat.inv
        Wglob <- vars$Wglob
        weights_orientation <- vars$weights_orientation
        exp_mean_over_shape <- vars$exp_mean_over_shape
        pair_idx <- vars$pair_idx
        
        # Gibbs sampling
        random_sign <- matrix(ifelse(runif(M0 * shape) <= 0.5, 1, -1), 
                             nrow = M0, ncol = shape)
        ini_vals <- random_sign * matrix(rep(sqrt(data_transformed), shape),
                                         nrow = M0, ncol = shape)
        
        gibbs.result <- tryCatch({
          dotCall64::.C64("gamma_gibbs_sampler",
                          SIGNATURE = c("double", "double", "int", "int", 
                                       "double", "int", "int"),
                          SigmaInv = as.double(Sigma.mat.inv),
                          y        = as.double(gamma_constraint),
                          n        = as.integer(M0),
                          v        = as.integer(shape),
                          U        = as.double(ini_vals),
                          nIte     = as.integer(1000),
                          nRep     = as.integer(150),
                          INTENT   = c("r", "r", "r", "r", "rw", "r", "r"), 
                          VERBOSE = 0, 
                          NAOK = FALSE, 
                          PACKAGE = "GeoModels")
        }, error = function(e) {
          return(list(U = NULL))
        })
        
        if (is.null(gibbs.result$U)) return(NULL)
        
        gibbs_U <- matrix(gibbs.result$U, nrow = M0, ncol = shape)
        
        # Processamento
        idx_xy <- pair_idx[[local_i]]
        res_i <- matrix(0, nrow = N0, ncol = shape)
        
        for (kk in seq_len(shape)) {
          simX <- Gauss_data[[idx_xy[kk]]]
          sim_train_x <- simX[seq_len(M0)]
          sim_valid_x <- simX[(M0 + 1):(N0 + M0)]
          delta <- gibbs_U[, kk] - sim_train_x
          sk_pred_x <- if (weights_orientation == "nobs_by_nloc") {
            as.numeric(crossprod(Wglob, delta))
          } else {
            as.numeric(Wglob %*% delta)
          }
          res_i[, kk] <- sim_valid_x + sk_pred_x
        }
        
        return(exp_mean_over_shape * rowSums(res_i^2))
      }
      
      # Parallelizzazione del batch corrente
      old_limit <- options(future.globals.maxSize = 2000 * 1024^2)
      on.exit(options(old_limit), add = TRUE)
      
      batch_results <- future.apply::future_lapply(
        seq_len(batch_nrep),
        function(local_i) {
          result <- run_simulation_worker(local_i, essential_vars, Gauss_data_batch)
          return(result)
        },
        future.seed = TRUE,
        future.packages = c("dotCall64", "GeoModels"),
        future.globals = structure(TRUE, add = c("essential_vars", "Gauss_data_batch"))
      )
      
      # Salva risultati del batch
      for (local_i in seq_len(batch_nrep)) {
        res[[start_idx + local_i - 1]] <- batch_results[[local_i]]
        if (progress) {
          pb(sprintf("Simulation %d/%d completed", start_idx + local_i - 1, nrep))
        }
      }
      
      # Pulizia memoria dopo ogni batch
      rm(Gauss_batch, Gauss_data_batch, batch_results, essential_vars)
      gc()
    }
    
  } else {
    # Modalità sequenziale (codice originale)
    if (method == "TB" || method == "CE") {
      Gauss.all <- GeoSimapprox(
        coordx = rbind(coord_obs, loc), 
        corrmodel = corrmodel, 
        method = method, 
        model = "Gaussian", 
        param = latent_param, 
        L = L, 
        nrep = shape * nrep, 
        progress = FALSE, 
        parallel = FALSE,
        distance = distance,
        radius = radius,
        anisopars = anisopars
      )
    } else {
      Gauss.all <- GeoSim(
        coordx = rbind(coord_obs, loc), 
        corrmodel = corrmodel, 
        model = "Gaussian", 
        param = latent_param, 
        nrep = shape * nrep, 
        progress = FALSE,
        distance = distance,
        radius = radius,
        anisopars = anisopars
      )
    }
    
    Gauss_data <- .GeoSimcondNormalizeSimulations(
      Gauss.all$data, shape * nrep, M0 + N0
    )
    pair_idx <- split(seq_len(shape * nrep), ceiling(seq_len(shape * nrep) / shape))
    
    cat("Performing", nrep, "conditional simulations ...\n")
    if (progress) {
      restore_progress_handlers <- .GeoProgressHandlersPush(
        handler = "txtprogressbar", global = TRUE
      )
      on.exit(restore_progress_handlers(), add = TRUE)
      pb <- progressr::progressor(
        along = seq_len(nrep), on_exit = FALSE, auto_finish = FALSE
      )
      ## Finish before restoring the temporary global handler, including when
      ## a batch or Gibbs simulation exits early.
      on.exit(pb(type = "finish"), add = TRUE, after = FALSE)
    }
    
    res <- vector("list", nrep)
    
    for (i in seq_len(nrep)) {
      random_sign <- matrix(ifelse(runif(M0 * shape) <= 0.5, 1, -1), 
                           nrow = M0, ncol = shape)
      ini_vals <- random_sign * matrix(rep(sqrt(data_transformed), shape),
                                         nrow = M0, ncol = shape)
      
      gibbs.result <- dotCall64::.C64("gamma_gibbs_sampler",
                                      SIGNATURE = c("double", "double", "int", "int", 
                                                   "double", "int", "int"),
                                      SigmaInv = as.double(Sigma.mat.inv),
                                      y        = as.double(gamma_constraint),
                                      n        = as.integer(M0),
                                      v        = as.integer(shape),
                                      U        = as.double(ini_vals),
                                      nIte     = as.integer(1000),
                                      nRep     = as.integer(150),
                                      INTENT   = c("r", "r", "r", "r", "rw", "r", "r"), 
                                      VERBOSE = 0, 
                                      NAOK = FALSE, 
                                      PACKAGE = "GeoModels")
      
      if (is.null(gibbs.result$U)) {
        warning(paste("Gibbs sampler failed at simulation", i))
        next
      }
      
      gibbs_U <- matrix(gibbs.result$U, nrow = M0, ncol = shape)
      
      idx_xy <- pair_idx[[i]]
      res_i <- matrix(0, nrow = N0, ncol = shape)
      
      for (kk in seq_len(shape)) {
        simX <- Gauss_data[[idx_xy[kk]]]
        sim_train_x <- simX[seq_len(M0)]
        sim_valid_x <- simX[(M0 + 1):(N0 + M0)]
        delta <- gibbs_U[, kk] - sim_train_x
        sk_pred_x <- if (weights_orientation == "nobs_by_nloc") {
          as.numeric(crossprod(Wglob, delta))
        } else {
          as.numeric(Wglob %*% delta)
        }
        res_i[, kk] <- sim_valid_x + sk_pred_x
      }
      
      res[[i]] <- exp_mean_over_shape * rowSums(res_i^2)
      
      if (progress) {
        pb(sprintf("Simulation %d/%d completed", i, nrep))
      }
    }
  }
  
  # Rimuove risultati nulli
  res <- res[!sapply(res, is.null)]
  failed_sims <- nrep - length(res)
  if (failed_sims > 0) {
    warning(paste(failed_sims, "simulazioni su", nrep, "sono fallite"))
  }
  
  return(res)
}
################################################################################# 
### internal function: conditional simulation of Gaussian RF 
################################################################################# 
.GeoSimcondPrepareLocalWeightMap <- function(weights_list, n_obs, n_loc) {
  if (!is.list(weights_list) || length(weights_list) != n_loc) {
    stop("GeoKriglocWeights returned an incompatible local weights list.")
  }

  lens <- integer(n_loc)
  idx_parts <- vector("list", n_loc)
  weight_parts <- vector("list", n_loc)

  for (j in seq_len(n_loc)) {
    wj <- weights_list[[j]]
    if (is.null(wj)) next

    idx <- as.integer(wj$neighbor_indices)
    ww <- as.numeric(wj$weights)
    if (length(idx) != length(ww)) {
      stop("A local kriging system returned inconsistent indices and weights.")
    }
    if (!length(idx)) next
    if (anyNA(idx) || any(idx < 1L | idx > n_obs)) {
      stop("A local kriging system returned invalid observation indices.")
    }
    if (any(!is.finite(ww))) {
      stop("A local kriging system returned non-finite weights.")
    }

    lens[j] <- length(idx)
    idx_parts[[j]] <- idx
    weight_parts[[j]] <- ww
  }

  active <- which(lens > 0L)
  if (!length(active)) {
    return(list(
      index = integer(0L), weight = numeric(0L), target = integer(0L),
      active = integer(0L), nnz = 0L, sparse_weights = NULL
    ))
  }

  out <- list(
    index = as.integer(unlist(idx_parts[active], use.names = FALSE)),
    weight = as.numeric(unlist(weight_parts[active], use.names = FALSE)),
    target = rep(as.integer(active), times = lens[active]),
    active = as.integer(active),
    nnz = as.integer(sum(lens)),
    sparse_weights = NULL
  )

  ## spam is already a GeoModels dependency.  Build the n_loc x n_obs local
  ## weight operator once so that multiple conditional Gaussian replicates can
  ## share compiled sparse matrix multiplication.  Keep the compact triplet map
  ## as a fallback for compatibility with older spam installations.
  out$sparse_weights <- tryCatch({
    sp <- spam::spam(
      list(i = out$target, j = out$index, values = out$weight),
      nrow = n_loc, ncol = n_obs
    )
    ## Validate the sparse multiplication method before discarding the triplet
    ## representation.  If this probe fails, retain the compact base-R fallback.
    probe <- as.numeric(sp %*% numeric(n_obs))
    if (length(probe) != n_loc || any(!is.finite(probe))) NULL else sp
  }, error = function(e) NULL)

  if (!is.null(out$sparse_weights)) {
    ## Avoid storing both sparse and triplet representations for massive local
    ## graphs.  nnz is retained for the empty/non-empty decision.
    out$index <- integer(0L)
    out$weight <- numeric(0L)
    out$target <- integer(0L)
    out$active <- integer(0L)
  }

  out
}

.GeoSimcondLocalWeightProduct <- function(x, weight_map, n_loc) {
  out <- numeric(n_loc)
  if (is.null(weight_map$nnz) || weight_map$nnz == 0L) return(out)

  if (!is.null(weight_map$sparse_weights)) {
    ans <- as.numeric(weight_map$sparse_weights %*% as.numeric(x))
    if (length(ans) != n_loc || any(!is.finite(ans))) {
      stop("Sparse local conditioning returned invalid values.")
    }
    return(ans)
  }

  weighted <- weight_map$weight * as.numeric(x)[weight_map$index]
  dim(weighted) <- c(length(weighted), 1L)
  sums <- rowsum(weighted, group = weight_map$target, reorder = FALSE)
  out[weight_map$active] <- as.numeric(sums[, 1L])
  out
}

.GeoSimcondLocalConditionalBatch <- function(sim_data_list, krig_pred_vec, weight_map,
                                    n_obs, n_loc, progressor = NULL,
                                    target_temp_mb = 128) {
  nrep <- length(sim_data_list)
  out <- vector("list", nrep)
  if (!nrep) return(out)
  if (is.null(progressor)) progressor <- function(...) NULL

  ## For one realization or an empty local graph the compact vector path is
  ## cheaper and preserves the maxdist-only no-neighbor semantics exactly.
  if (is.null(weight_map$nnz) || weight_map$nnz == 0L ||
      is.null(weight_map$sparse_weights) || nrep == 1L) {
    for (i in seq_len(nrep)) {
      out[[i]] <- .GeoSimcondOne(
        i, sim_data_list, krig_pred_vec, TRUE,
        weight_map, NA_character_, n_obs, n_loc
      )
      progressor()
    }
    return(out)
  }

  ## Bound only the temporary dense matrices created by the sparse batch.  The
  ## unconditional simulations are already retained by the simulator.  This
  ## avoids a large nrep * nnz temporary object when many realizations are asked
  ## for on a massive data set.
  target_bytes <- max(1, as.numeric(target_temp_mb)) * 1024^2
  bytes_per_rep <- 8 * (n_obs + 4 * n_loc)
  block_size <- max(
    1L,
    min(nrep, as.integer(floor(target_bytes / max(bytes_per_rep, 1))))
  )
  blocks <- split(seq_len(nrep), ceiling(seq_len(nrep) / block_size))

  sparse_ok <- TRUE
  for (ids in blocks) {
    nb <- length(ids)
    obs_mat <- matrix(
      unlist(lapply(ids, function(i) sim_data_list[[i]][seq_len(n_obs)]),
             use.names = FALSE),
      nrow = n_obs, ncol = nb
    )
    loc_mat <- matrix(
      unlist(lapply(ids, function(i) sim_data_list[[i]][n_obs + seq_len(n_loc)]),
             use.names = FALSE),
      nrow = n_loc, ncol = nb
    )

    obs_pred <- if (sparse_ok) {
      tryCatch(
        as.matrix(weight_map$sparse_weights %*% obs_mat),
        error = function(e) NULL
      )
    } else NULL

    if (is.null(obs_pred) || !identical(dim(obs_pred), c(n_loc, nb)) ||
        any(!is.finite(obs_pred))) {
      ## Defensive fallback: do not make local conditional simulation depend on
      ## sparse-matrix method details of a particular spam version.
      sparse_ok <- FALSE
      for (i in ids) {
        out[[i]] <- .GeoSimcondOne(
          i, sim_data_list, krig_pred_vec, TRUE,
          weight_map, NA_character_, n_obs, n_loc
        )
        progressor()
      }
      next
    }

    cond_mat <- loc_mat - obs_pred
    cond_mat <- cond_mat + krig_pred_vec
    out[ids] <- lapply(seq_len(nb), function(k) as.numeric(cond_mat[, k]))
    for (unused in ids) progressor()
  }

  out
}

.GeoSimcondGaussianMeans <- function(param, X, Xloc, Mloc, n_obs, n_loc) {
  external <- .GeoMean_external(param, n = n_obs)
  mu_obs <- .GeoMean_vector(
    param, X = X, n = n_obs, external = external, xname = "X"
  )

  if (!is.null(Mloc)) {
    mu_loc <- as.numeric(Mloc)
    if (length(mu_loc) == 1L) mu_loc <- rep(mu_loc, n_loc)
    if (length(mu_loc) != n_loc || any(!is.finite(mu_loc))) {
      stop("Mloc must contain one finite value per prediction point.")
    }
  } else if (!is.null(Xloc)) {
    Xloc_mat <- .GeoMean_design(Xloc, n = n_loc, name = "Xloc")
    beta <- .GeoMean_beta(param, ncol(Xloc_mat))
    mu_loc <- as.numeric(Xloc_mat %*% beta)
  } else {
    if (!is.null(X)) {
      stop("Xloc or Mloc is required when X is supplied.")
    }
    mu_loc <- .GeoMean_vector(param, X = NULL, n = n_loc, xname = "Xloc")
  }

  if (length(mu_obs) != n_obs || length(mu_loc) != n_loc ||
      any(!is.finite(c(mu_obs, mu_loc)))) {
    stop("Could not construct finite Gaussian means for local conditioning.")
  }

  list(obs = as.numeric(mu_obs), loc = as.numeric(mu_loc))
}

.GeoSimcondOne <- function(i, sim_data_list, krig_pred_vec, local_flag,
                             weights_obj, weights_orientation, n_obs, n_loc) {
  sim_i <- sim_data_list[[i]]
  sim_nc_obs_data <- sim_i[seq_len(n_obs)]
  sim_nc_loc_data <- sim_i[n_obs + seq_len(n_loc)]

  if (local_flag) {
    sim_nc_obs_pred <- .GeoSimcondLocalWeightProduct(sim_nc_obs_data, weights_obj, n_loc)
  } else {
    W <- weights_obj
    xvec <- as.vector(sim_nc_obs_data)

    if (identical(weights_orientation, "nobs_by_nloc")) {
      sim_nc_obs_pred <- as.vector(crossprod(W, xvec))
    } else if (identical(weights_orientation, "nloc_by_nobs")) {
      sim_nc_obs_pred <- as.vector(W %*% xvec)
    } else {
      stop("Unknown kriging weights orientation.")
    }
  }

  krig_pred_vec + sim_nc_loc_data - sim_nc_obs_pred
}

#################################################################################
### main function: conditional Gaussian simulations
#################################################################################
.GeoSimcondGaussian <- function(data, corrmodel, nrep, method, L,
                     param,
                     coord_obs, loc, coordt_use, time, X, Xloc, Mloc, distance, radius, anisopars,
                     local, neighb, maxdist, maxtime,
                     space, spacetime, bivariate, parallel, ncores, progress) {

  #############################################
  # Input validation
  #############################################
  if (is.null(loc) || nrow(loc) == 0) {
    stop("loc must be a non-empty matrix")
  }
  if (nrep < 1) {
    stop("nrep must be at least 1")
  }
  if (local && is.null(neighb) && is.null(maxdist)) {
    stop("For local conditional simulation, specify neighb or maxdist.",
         call. = FALSE)
  }

  #############################################
  # Build combined coordinates
  #############################################
  coord_sim <- rbind(coord_obs, loc)
  time_sim  <- c(coordt_use, time)
  n_obs <- length(data)
  n_loc <- if(!is.null(Mloc)) length(Mloc) else if(!is.null(Xloc)) nrow(Xloc) else nrow(loc)

  ## The substitution residual must be a zero-mean Gaussian field:
  ## Z_c = E(Z | data) + Z*loc - W Z*obs.
  param_sim <- .GeoMean_intercept_param(param, 0)
  X_sim <- NULL

  #############################################
  # Unconditional simulations
  #############################################
  if (method %in% c("TB", "CE") && isTRUE(parallel)) {
    cat("Performing", nrep, "unconditional simulations using", method,
        "with", ncores, "cores ...\n")
  } else {
    cat("Performing", nrep, "unconditional simulations using", method, "...\n")
  }

  if (method == "Cholesky") {
    sim_args <- list(
      coordx = coord_sim, coordt = time_sim, corrmodel = corrmodel,
      progress = FALSE, X = X_sim, nrep = nrep, distance = distance,
      radius = radius, anisopars = anisopars
    )
    sim_args <- c(sim_args, list(param = param_sim))
    ## GeoSim() prints its own Cholesky banner when nrep > 1.  Suppress that
    ## internal banner here so GeoSimcond reports this phase exactly once.
    invisible(utils::capture.output(
      sim_nc <- do.call(GeoSim, sim_args),
      type = "output"
    ))
  } else if (method == "TB" || method == "CE") {
    sim_args_approx <- list(
      coordx = coord_sim, coordt = time_sim, corrmodel = corrmodel,
      progress = FALSE, method = method, L = L, parallel = parallel,
      ncores = ncores,
      X = X_sim, nrep = nrep, distance = distance, radius = radius,
      anisopars = anisopars
    )
    sim_args_approx <- c(sim_args_approx, list(param = param_sim))
    sim_nc <- do.call(GeoSimapprox, sim_args_approx)
  } else {
    stop("Unsupported unconditional simulation method.")
  }

  sim_data_list <- .GeoSimcondNormalizeSimulations(
    sim_nc$data, nrep, n_obs + n_loc
  )
  rm(sim_nc)

  #############################################
  # Prediction and kriging weights
  #############################################
  weights_orientation <- NA_character_

  if (local) {
    local_desc <- character(0L)
    if (!is.null(neighb)) {
      local_desc <- c(local_desc, paste0(neighb, " nearest neighbours"))
    }
    if (!is.null(maxdist)) {
      local_desc <- c(local_desc, paste0("maxdist = ", format(maxdist)))
    }
    if (!is.null(maxtime)) {
      local_desc <- c(local_desc, paste0("maxtime = ", format(maxtime)))
    }
    local_desc <- if (length(local_desc)) paste(local_desc, collapse = "; ") else "local"
    weight_parallel_note <- if (isTRUE(parallel)) paste0("; ", ncores, " cores") else ""
    cat("Computing compact local kriging weights (", local_desc,
        weight_parallel_note, ") ...\n", sep = "")

    weights_args <- list(
      coordx = coord_obs, coordt = coordt_use,
      corrmodel = corrmodel, loc = loc,
      X = X, Xloc = Xloc, Mloc = Mloc, time = time,
      distance = distance, radius = radius, anisopars = anisopars,
      param = param, neighb = neighb, maxdist = maxdist,
      maxtime = maxtime, parallel = parallel, ncores = ncores,
      compact = TRUE
    )
    krig_weights_obj <- do.call(GeoKriglocWeights, weights_args)
    local_weights <- .GeoSimcondPrepareLocalWeightMap(
      krig_weights_obj$weights, n_obs = n_obs, n_loc = n_loc
    )
    rm(krig_weights_obj)

    means <- .GeoSimcondGaussianMeans(
      param = param, X = X, Xloc = Xloc, Mloc = Mloc,
      n_obs = n_obs, n_loc = n_loc
    )
    krig_pred_vec <- means$loc + .GeoSimcondLocalWeightProduct(
      data - means$obs, local_weights, n_loc
    )
    rm(means)

    weights_obj <- local_weights
    rm(local_weights)
  } else {
    cat("Computing global kriging predictor and weights ...\n")
    krig_sim_args <- list(
      coordx = coord_obs, coordt = coordt_use,
      data = data, corrmodel = corrmodel,
      loc = loc, X = X, Xloc = Xloc, Mloc = Mloc, time = time,
      distance = distance, radius = radius, anisopars = anisopars,
      param = param
    )
    krig_sim <- do.call(GeoKrig, krig_sim_args)
    krig_pred_vec <- as.numeric(krig_sim$pred)
    rm(krig_sim)

    weights_args <- list(
      coordx = coord_obs, coordt = coordt_use,
      corrmodel = corrmodel, loc = loc,
      X = X, Xloc = Xloc, Mloc = Mloc, time = time,
      distance = distance, radius = radius, anisopars = anisopars,
      param = param
    )
    krig_weights_obj <- do.call(GeoKrigWeights, weights_args)
    W <- krig_weights_obj$weights
    weights_orientation <- krig_weights_obj$weights_orientation
    if (is.null(weights_orientation) ||
        !(weights_orientation %in% c("nobs_by_nloc", "nloc_by_nobs"))) {
      stop("GeoKrigWeights returned an unknown weights orientation.")
    }
    if (weights_orientation == "nobs_by_nloc" &&
        (nrow(W) != n_obs || ncol(W) != n_loc)) {
      stop("The nobs_by_nloc kriging weights have incompatible dimensions.")
    }
    if (weights_orientation == "nloc_by_nobs" &&
        (nrow(W) != n_loc || ncol(W) != n_obs)) {
      stop("The nloc_by_nobs kriging weights have incompatible dimensions.")
    }
    rm(krig_weights_obj)
    weights_obj <- W
    rm(W)
  }

  if (length(krig_pred_vec) != n_loc || any(!is.finite(krig_pred_vec))) {
    stop("The kriging predictor returned invalid conditional means.")
  }

  #############################################
  # Conditional simulations with progress bar
  #############################################
  if (local) {
    cat("Applying local conditional correction to", nrep,
        "simulations ...\n")
  } else if (isTRUE(parallel) && nrep > 1L) {
    cat("Performing", nrep, "global conditional simulations in parallel using",
        ncores, "cores ...\n")
  } else {
    cat("Performing", nrep, "global conditional simulations ...\n")
  }

  if (progress) {
    restore_progress_handlers <- .GeoProgressHandlersPush(
      handler = "txtprogressbar", global = TRUE
    )
    on.exit(restore_progress_handlers(), add = TRUE)
    p <- progressr::progressor(
      along = seq_len(nrep), on_exit = FALSE, auto_finish = FALSE
    )
    ## Finish before restoring the temporary global handler on every exit path.
    on.exit(p(type = "finish"), add = TRUE, after = FALSE)
  } else {
    p <- function(...) NULL
  }

  if (space) {
    ## For local conditioning the expensive stages (TB/CE simulation and local
    ## kriging-weight construction) can already use the requested workers.
    ## Keep the final substitution in the main R process: with compact local
    ## weights it is a cheap vectorized operation, while a PSOCK cluster would
    ## duplicate sim_data_list and the local weight map on every worker.
    if (local) {
      ## Fast path shared by Gaussian fields, one-Gaussian monotone transforms
      ## and Gaussian-copula margins.  The local graph is identical for every
      ## realization, so apply the same sparse operator in bounded batches.
      sim_cond <- .GeoSimcondLocalConditionalBatch(
        sim_data_list = sim_data_list,
        krig_pred_vec = krig_pred_vec,
        weight_map = weights_obj,
        n_obs = n_obs,
        n_loc = n_loc,
        progressor = p
      )
    } else if (parallel && !is.null(ncores) && nrep > 1) {
      memory_needed <- object.size(sim_data_list) +
                       object.size(weights_obj) +
                       object.size(krig_pred_vec)
      memory_gb <- as.numeric(memory_needed) / 1024^3

      if (memory_gb > 10) {
        chunk_size <- max(1, min(20, floor(nrep / 4)))
        chunks <- split(seq_len(nrep), ceiling(seq_len(nrep) / chunk_size))
        sim_cond <- vector("list", nrep)

        for (chunk_idx in seq_along(chunks)) {
          chunk_indices <- chunks[[chunk_idx]]
          cl <- parallel::makeCluster(min(ncores, length(chunk_indices)))
          chunk_sim_data_only <- sim_data_list[chunk_indices]
          krig_pred_only <- krig_pred_vec
          weights_only <- weights_obj
          orientation_only <- weights_orientation

          compute_one <- .GeoSimcondOne
          chunk_results <- tryCatch({
            parallel::clusterExport(cl, "compute_one", envir = environment())
            parallel::parLapply(cl, seq_along(chunk_indices), function(local_i) {
              compute_one(
                local_i, chunk_sim_data_only, krig_pred_only, FALSE,
                weights_only, orientation_only, n_obs, n_loc
              )
            })
          }, finally = {
            parallel::stopCluster(cl)
          })

          for (local_i in seq_along(chunk_indices)) {
            sim_cond[[chunk_indices[local_i]]] <- chunk_results[[local_i]]
            p()
          }
          rm(chunk_sim_data_only, krig_pred_only, weights_only,
             orientation_only, chunk_results)
          if (chunk_idx %% 5 == 0) gc()
        }
      } else {
        cl <- parallel::makeCluster(ncores)
        compute_one <- .GeoSimcondOne
        sim_cond <- tryCatch({
          parallel::clusterExport(cl, "compute_one", envir = environment())
          parallel::parLapply(cl, seq_len(nrep), function(i) {
            compute_one(
              i, sim_data_list, krig_pred_vec, FALSE,
              weights_obj, weights_orientation, n_obs, n_loc
            )
          })
        }, finally = {
          parallel::stopCluster(cl)
        })

        for (i in seq_len(nrep)) p()
      }
    } else {
      sim_cond <- vector("list", nrep)
      for (i in seq_len(nrep)) {
        sim_cond[[i]] <- tryCatch({
          .GeoSimcondOne(
            i, sim_data_list, krig_pred_vec, FALSE,
            weights_obj, weights_orientation, n_obs, n_loc
          )
        }, error = function(e) {
          warning(paste("Simulation", i, "failed:", e$message))
          NULL
        })
        p()
      }

      sim_cond <- sim_cond[!vapply(sim_cond, is.null, logical(1L))]
      failed_sims <- nrep - length(sim_cond)
      if (failed_sims > 0) {
        warning(paste(failed_sims, "simulations out of", nrep, "failed"))
      }
    }
  } else {
    stop("Gauss_cd currently not implemented for spacetime models")
  }

  sim_cond
}

#################################################################################
### Direct Binomial / Negative-Binomial count-constrained conditional sampler
#################################################################################
.GeoSimcondCountRF <- function(
    model, coord_obs, loc, data, n, nloc, param, corrmodel,
    mean_obs, mean_loc, distance = "Eucl", radius = 1,
    anisopars = NULL, nrep = 1L, n_iter = 25L, mcmc_thin = 1L,
    method = "Cholesky", local = FALSE, parallel = FALSE, progress = FALSE
) {
  if (!(model %in% c("Binomial", "BinomialNeg"))) {
    stop("Internal error: unsupported count model in .GeoSimcondCountRF().",
         call. = FALSE)
  }
  if (!identical(method, "Cholesky")) {
    stop(
      paste0(
        "Direct ", model, " conditional simulation currently uses the exact dense ",
        "latent Gaussian covariance and requires method = 'Cholesky'."
      ),
      call. = FALSE
    )
  }
  if (isTRUE(local)) {
    stop(
      paste0(
        "Direct ", model, " conditional simulation currently requires local = FALSE; ",
        "the count-constrained latent Gibbs sampler conditions on the full observed field."
      ),
      call. = FALSE
    )
  }
  if (!is.numeric(mcmc_thin) || length(mcmc_thin) != 1L ||
      !is.finite(mcmc_thin) || mcmc_thin < 1 ||
      mcmc_thin != as.integer(mcmc_thin)) {
    stop("mcmc_thin must be a positive integer.", call. = FALSE)
  }
  mcmc_thin <- as.integer(mcmc_thin)

  coord_obs <- as.matrix(coord_obs)
  loc <- as.matrix(loc)
  data <- as.numeric(data)
  mean_obs <- as.numeric(mean_obs)
  mean_loc <- as.numeric(mean_loc)
  m <- nrow(coord_obs)
  p <- nrow(loc)

  if (length(mean_obs) != m || length(mean_loc) != p ||
      any(!is.finite(c(mean_obs, mean_loc)))) {
    stop("Invalid latent probit means for count conditional simulation.", call. = FALSE)
  }
  tol <- sqrt(.Machine$double.eps)
  if (any(data < 0) || any(abs(data - round(data)) > tol)) {
    stop(sprintf("%s observations must be non-negative integers.", model),
         call. = FALSE)
  }
  y <- as.integer(round(data))

  sizes <- .GeoKrigCountSizes(
    model = model, n = n, nloc = nloc,
    nobs = m, npred = p, context = "GeoSimcond"
  )

  if (model == "Binomial") {
    if (any(y > sizes$obs)) {
      stop("Binomial observations cannot exceed the corresponding number of trials n.",
           call. = FALSE)
    }
    free_len <- as.integer(sizes$obs)
    success_count <- y
    forced_success <- integer(m)
    Lobs <- max(free_len)
    model_code <- 0L
  } else {
    r <- as.integer(sizes$obs[1L])
    free_len <- as.integer(y + r - 1L)
    success_count <- rep.int(r - 1L, m)
    forced_success <- as.integer(y + r)
    Lobs <- max(forced_success)
    model_code <- 1L
  }

  latent_cells <- as.double(Lobs) * as.double(m)
  if (!is.finite(latent_cells) || latent_cells > .Machine$integer.max) {
    stop("The latent count-constraint array is too large for the current implementation.",
         call. = FALSE)
  }
  if (latent_cells * 16 / 1024^2 > 1024) {
    warning(
      "The count-constrained latent Gibbs state exceeds approximately 1 GB; conditional simulation may be memory intensive.",
      call. = FALSE
    )
  }

  ## The direct count constructions threshold unit-variance Gaussian fields.
  ## Build exactly that latent covariance, retaining the fitted nugget and
  ## correlation parameters but removing marginal count parameters.
  corr_names <- CorrParam(corrmodel)
  latent_param <- as.list(param[corr_names])
  latent_param$mean <- 0
  latent_param$nugget <- if (is.null(param$nugget)) 0 else as.numeric(param$nugget)[1L]
  latent_param$sill <- 1

  all_coord <- rbind(coord_obs, loc)
  latent_cov <- GeoCovmatrix(
    coordx = all_coord, corrmodel = corrmodel, distance = distance,
    grid = FALSE, model = "Gaussian", n = 1,
    param = latent_param, anisopars = anisopars, radius = radius,
    sparse = FALSE, copula = NULL, X = NULL,
    check.duplicates = FALSE
  )$covmatrix
  Rall <- as.matrix(latent_cov)
  rm(latent_cov)

  expected_dim <- m + p
  if (!identical(dim(Rall), c(expected_dim, expected_dim)) ||
      any(!is.finite(Rall))) {
    stop("Could not construct a finite latent Gaussian covariance matrix.",
         call. = FALSE)
  }
  Rall <- 0.5 * (Rall + t(Rall))
  io <- seq_len(m)
  il <- m + seq_len(p)
  Roo <- Rall[io, io, drop = FALSE]
  Rol <- Rall[io, il, drop = FALSE]
  Rll <- Rall[il, il, drop = FALSE]
  rm(Rall)

  chol_obs <- tryCatch(chol(Roo), error = function(e) NULL)
  if (is.null(chol_obs)) {
    stop("The latent observation covariance matrix is not positive definite.",
         call. = FALSE)
  }
  Q <- chol2inv(chol_obs)
  Q <- 0.5 * (Q + t(Q))
  W <- Q %*% Rol

  cond_cov <- Rll - crossprod(Rol, W)
  cond_cov <- 0.5 * (cond_cov + t(cond_cov))
  chol_cond <- tryCatch(chol(cond_cov), error = function(e) NULL)
  if (is.null(chol_cond)) {
    stop(
      paste0(
        "The latent conditional covariance at loc is singular or not positive definite. ",
        "For direct count conditional simulation, prediction locations must not duplicate ",
        "observed locations or create an exactly singular prediction covariance."
      ),
      call. = FALSE
    )
  }
  chol_loc <- tryCatch(chol(Rll), error = function(e) NULL)
  if (is.null(chol_loc)) {
    stop("The latent covariance among prediction locations is not positive definite.",
         call. = FALSE)
  }

  if (isTRUE(parallel)) {
    warning(
      "parallel=TRUE is currently ignored by the native single-chain count-constrained Gibbs sampler.",
      call. = FALSE
    )
  }
  if (isTRUE(progress)) {
    message(
      sprintf(
        "Running count-constrained latent Gibbs sampler: %d burn-in sweeps, %d retained simulations, thinning %d.",
        as.integer(n_iter), as.integer(nrep), mcmc_thin
      )
    )
  }

  ans <- .Call(
    "GeoCountCondSim",
    Q,
    as.numeric(mean_obs),
    as.integer(free_len),
    as.integer(success_count),
    as.integer(forced_success),
    W,
    as.numeric(mean_loc),
    chol_cond,
    chol_loc,
    as.integer(sizes$pred),
    as.integer(model_code),
    as.integer(n_iter),
    as.integer(mcmc_thin),
    as.integer(nrep),
    as.integer(Lobs),
    PACKAGE = "GeoModels"
  )

  if (!is.matrix(ans) || !identical(dim(ans), c(p, as.integer(nrep))) ||
      any(!is.finite(ans))) {
    stop("The native count conditional simulator returned an invalid result.",
         call. = FALSE)
  }

  lapply(seq_len(nrep), function(j) as.numeric(ans[, j]))
}

## Robust one-dimensional optimization used by GeoFit composite-likelihood
## paths.  The historical code called stats::optimize() over the complete
## automatic parameter interval.  For positive parameters that interval can
## be enormous (for example [.Machine$double.eps, 1e12]); even after a log
## transformation, optimize() assumes a unimodal objective and can converge
## to a remote numerical basin while ignoring the supplied starting value.
##
## The helper below first brackets a local minimum around the supplied start
## on the natural transformed scale (log scale for positive parameters), then
## refines only inside that bracket with stats::optimize().  This mirrors the
## start-dependent behaviour of the multidimensional optimizers used by
## GeoFit and avoids pathological jumps across many orders of magnitude.
.GeoModels_optimize_1d <- function(f, param, lower, upper, namesparam, ...) {
    if (length(param) != 1L)
        stop(".GeoModels_optimize_1d requires exactly one parameter", call. = FALSE)

    p0 <- as.numeric(param[1L])
    lo <- as.numeric(lower[1L])
    hi <- as.numeric(upper[1L])
    pname <- if (length(namesparam)) as.character(namesparam[1L]) else "parameter"

    if (!is.finite(p0))
        stop("Invalid starting value for one-parameter optimization: ", pname,
             call. = FALSE)

    ## optimize() needs finite bounds.  Normally GeoFit has already supplied
    ## them, but keep this helper safe for legacy/internal calls too.
    if (!is.finite(lo) || !is.finite(hi)) {
        if (p0 > 0 && (is.na(lo) || !is.finite(lo) || lo >= 0)) {
            if (!is.finite(lo)) lo <- max(.Machine$double.eps, p0 / 1e4)
            if (!is.finite(hi)) hi <- max(p0 * 1e4, p0 + 1)
        } else {
            span <- max(10, 20 * (1 + abs(p0)))
            if (!is.finite(lo)) lo <- p0 - span
            if (!is.finite(hi)) hi <- p0 + span
        }
    }

    if (!is.finite(lo) || !is.finite(hi) || lo >= hi)
        stop("Invalid interval for one-parameter optimization: ", pname,
             call. = FALSE)

    use_log <- is.finite(p0) && p0 > 0 && lo >= 0 && hi > 0

    if (use_log) {
        lo_pos <- max(lo, .Machine$double.eps)
        aa <- log(lo_pos)
        bb <- log(hi)
        z0 <- log(min(max(p0, lo_pos), hi))
        back <- exp
        ff <- function(z, ...) f(exp(z), namesparam = namesparam, ...)
        base_step <- log(2)
        parameter_scale <- "log"
    } else {
        aa <- lo
        bb <- hi
        z0 <- min(max(p0, lo), hi)
        back <- identity
        ff <- function(z, ...) f(z, namesparam = namesparam, ...)
        base_step <- max(0.25, 0.25 * (1 + abs(z0)))
        parameter_scale <- "linear"
    }

    if (!is.finite(aa) || !is.finite(bb) || aa >= bb)
        stop("Invalid transformed interval for one-parameter optimization: ",
             pname, call. = FALSE)

    ## Safe scalar objective evaluation for the bracketing stage.  Invalid
    ## trial points are treated as +Inf but genuine errors at the starting
    ## point are still exposed below if no finite bracket can be constructed.
    eval_one <- function(z) {
        val <- tryCatch(ff(z, ...), error = function(e) Inf)
        val <- suppressWarnings(as.numeric(val)[1L])
        if (!is.finite(val)) Inf else val
    }

    f0 <- eval_one(z0)
    if (!is.finite(f0))
        stop("Non-finite objective at the starting value for one-parameter optimization: ",
             pname, call. = FALSE)

    ## Start-centred adaptive bracketing.  The first probes are a factor of 2
    ## apart for positive parameters.  The radius is expanded only when the
    ## objective is still descending, so a nearby likelihood basin is not
    ## abandoned merely because the formal automatic bounds are extremely
    ## broad.
    step <- min(base_step, (bb - aa) / 4)
    if (!is.finite(step) || step <= 0) step <- (bb - aa) / 4

    z_left <- max(aa, z0 - step)
    z_right <- min(bb, z0 + step)
    f_left <- eval_one(z_left)
    f_right <- eval_one(z_right)

    bracket <- NULL

    if (f0 <= f_left && f0 <= f_right) {
        bracket <- c(z_left, z_right)
    } else {
        direction <- if (f_left < f_right) -1 else 1
        z_prev <- z0
        f_prev <- f0
        z_curr <- if (direction < 0) z_left else z_right
        f_curr <- if (direction < 0) f_left else f_right
        radius <- step

        for (ii in seq_len(40L)) {
            if (!is.finite(f_curr) || f_curr >= f_prev) {
                bracket <- sort(c(z_curr, z_prev))
                break
            }

            radius <- radius * 1.6
            z_next <- z0 + direction * radius
            z_next <- min(max(z_next, aa), bb)

            if (identical(z_next, z_curr) || abs(z_next - z_curr) <= .Machine$double.eps) {
                ## The objective kept improving up to an admissible boundary.
                edge2 <- if (direction < 0) min(z_prev, z_curr) else max(z_prev, z_curr)
                edge1 <- if (direction < 0) aa else bb
                bracket <- sort(c(edge1, edge2))
                break
            }

            f_next <- eval_one(z_next)

            if (!is.finite(f_next) || f_next >= f_curr) {
                bracket <- sort(c(z_prev, z_next))
                break
            }

            z_prev <- z_curr
            f_prev <- f_curr
            z_curr <- z_next
            f_curr <- f_next
        }
    }

    if (is.null(bracket) || length(bracket) != 2L ||
        !all(is.finite(bracket)) || bracket[1L] >= bracket[2L]) {
        ## Very defensive fallback: use a modest start-centred interval rather
        ## than the complete gigantic automatic interval.
        radius <- min(max(base_step * 4, (bb - aa) / 100), (bb - aa) / 2)
        bracket <- c(max(aa, z0 - radius), min(bb, z0 + radius))
    }

    ans <- stats::optimize(
        f = ff,
        interval = bracket,
        maximum = FALSE,
        ...
    )

    ans$minimum <- back(ans$minimum)
    ans$parameter_scale <- parameter_scale
    ans$search_interval <- back(bracket)
    ans$start <- p0
    ans
}

## Internal utilities for GeoCV.
## These functions centralize cross-validation plumbing; spatial, space-time
## and bivariate fold construction remain in GeoCV itself.

.GeoCVMetricNames <- c(
  "rmse", "mae", "mad", "lscore", "brie",
  "crps", "pit_mean", "intscore", "coverage"
)

.GeoCVSummarizeScores <- function(obs, pred, mse, gaussian_scores) {
  obs <- as.numeric(c(obs))
  pred <- as.numeric(c(pred))
  mse <- as.numeric(c(mse))

  if (length(obs) != length(pred))
    stop("length of observations and predictions does not match", call. = FALSE)
  if (!gaussian_scores && length(mse) == 0L)
    mse <- rep(NA_real_, length(pred))
  if (length(mse) != length(pred))
    stop("length of predictive MSE and predictions does not match", call. = FALSE)

  ok_err <- is.finite(obs) & is.finite(pred)
  if (!any(ok_err))
    stop("no finite observation/prediction pairs are available", call. = FALSE)

  err <- obs[ok_err] - pred[ok_err]
  ans <- c(
    rmse = sqrt(mean(err^2)),
    mae = mean(abs(err)),
    mad = stats::median(abs(err)),
    lscore = NA_real_, brie = NA_real_, crps = NA_real_,
    pit_mean = NA_real_, intscore = NA_real_, coverage = NA_real_
  )

  if (gaussian_scores) {
    ok_prob <- is.finite(obs) & is.finite(pred) & is.finite(mse) & mse > 0
    if (any(ok_prob)) {
      pp <- GeoScores(
        obs[ok_prob], pred = pred[ok_prob], mse = mse[ok_prob],
        score = c("brie", "crps", "lscore", "pit", "intscore", "coverage")
      )
      required <- c("LogScore", "Brier", "CRPS", "PIT", "IS95", "Cvg95")
      missing_scores <- setdiff(required, names(pp))
      if (length(missing_scores)) {
        stop(
          "GeoScores did not return the required components: ",
          paste(missing_scores, collapse = ", "), call. = FALSE
        )
      }
      pit_value <- mean(as.numeric(pp$PIT), na.rm = TRUE)
      if (is.nan(pit_value)) pit_value <- NA_real_
      ans[c("lscore", "brie", "crps", "pit_mean", "intscore", "coverage")] <- c(
        as.numeric(pp$LogScore), as.numeric(pp$Brier), as.numeric(pp$CRPS),
        pit_value, as.numeric(pp$IS95), as.numeric(pp$Cvg95)
      )
    }
  }

  ans[.GeoCVMetricNames]
}

.GeoCVBindScores <- function(x) {
  if (!is.list(x) || !length(x))
    stop("No cross-validation results were returned", call. = FALSE)
  sizes <- vapply(x, length, integer(1))
  bad <- which(sizes != length(.GeoCVMetricNames))
  if (length(bad)) {
    stop(
      "Invalid score vector returned by cross-validation iteration(s): ",
      paste(bad, collapse = ", "), call. = FALSE
    )
  }
  ans <- do.call(rbind, x)
  if (is.null(dim(ans)) || ncol(ans) != length(.GeoCVMetricNames))
    stop("Cross-validation score results have invalid dimensions", call. = FALSE)
  colnames(ans) <- .GeoCVMetricNames
  ans
}

.GeoCVRunIterations <- function(iteration_fun, K, parallel, ncores, progress) {
  n.cores <- .GeoResolveWorkers(parallel, ncores, n_jobs = K)
  if (n.cores <= 1L) {
    out <- vector("list", K)
    if (isTRUE(progress)) {
      cat("Performing", K, "cross-validations...\n")
      pb <- txtProgressBar(min = 0, max = K, style = 3)
      on.exit(close(pb), add = TRUE)
    }
    for (i in seq_len(K)) {
      out[[i]] <- iteration_fun(i)
      if (isTRUE(progress)) setTxtProgressBar(pb, i)
    }
    return(.GeoCVBindScores(out))
  }

  old_plan <- future::plan()
  on.exit(try(future::plan(old_plan), silent = TRUE), add = TRUE)
  future::plan(future::multisession, workers = n.cores)

  if (isTRUE(progress)) {
    out <- progressr::with_progress({
      p <- progressr::progressor(along = seq_len(K))
      future.apply::future_lapply(
        seq_len(K),
        function(i) {
          z <- iteration_fun(i)
          p()
          z
        },
        future.seed = TRUE, future.stdout = FALSE, future.conditions = "none"
      )
    })
  } else {
    out <- future.apply::future_lapply(
      seq_len(K), iteration_fun,
      future.seed = TRUE, future.stdout = FALSE, future.conditions = "none"
    )
  }
  .GeoCVBindScores(out)
}

.GeoCVMakeTestSamples <- function(n, K, n.fold, groups = NULL) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n) || n < 2L)
    stop("At least two observations are required for cross-validation.", call. = FALSE)

  ntest <- max(1L, min(n - 1L, as.integer(round(n * n.fold))))
  if (is.null(groups))
    return(lapply(seq_len(K), function(i) sort(sample.int(n, ntest))))

  if (length(groups) != n)
    stop("Internal error: fold groups are not aligned with observations.", call. = FALSE)
  groups <- as.integer(factor(groups, levels = unique(groups)))
  group_sizes <- tabulate(groups)
  max_test <- sum(pmax(group_sizes - 1L, 0L))
  if (max_test < 1L) {
    stop(
      "Space-time cross-validation cannot remove observations while retaining at least one training observation at every time.",
      call. = FALSE
    )
  }
  ntest <- min(ntest, max_test)

  lapply(seq_len(K), function(i) {
    remaining <- group_sizes
    selected <- integer(0)
    for (idx in sample.int(n)) {
      g <- groups[idx]
      if (remaining[g] > 1L) {
        selected <- c(selected, idx)
        remaining[g] <- remaining[g] - 1L
        if (length(selected) == ntest) break
      }
    }
    if (length(selected) != ntest)
      stop("Unable to construct a valid space-time cross-validation fold.", call. = FALSE)
    sort(selected)
  })
}

.GeoCVFlattenDesign <- function(X, ns, context) {
  if (is.null(X)) return(NULL)
  total <- sum(ns)
  if (is.list(X) && !is.data.frame(X) && !is.matrix(X)) {
    if (length(X) != length(ns))
      stop(context, ": X must contain one design matrix per block.", call. = FALSE)
    blocks <- lapply(seq_along(ns), function(i) {
      z <- as.matrix(X[[i]])
      if (!is.numeric(z) || nrow(z) != ns[i] || any(!is.finite(z)))
        stop(context, ": X block ", i, " is not aligned with the data.", call. = FALSE)
      z
    })
    p <- unique(vapply(blocks, ncol, integer(1)))
    if (length(p) != 1L)
      stop(context, ": all X blocks must have the same number of columns.", call. = FALSE)
    return(do.call(rbind, blocks))
  }
  X <- as.matrix(X)
  if (!is.numeric(X) || nrow(X) != total || any(!is.finite(X)))
    stop(context, ": X must have one finite numeric row per observation.", call. = FALSE)
  X
}

.GeoCVFlattenObservationVector <- function(x, ns, name) {
  if (is.null(x)) stop(name, " is missing", call. = FALSE)
  if (length(x) == 1L && !is.list(x)) {
    if (!is.numeric(x) || !is.finite(x))
      stop(name, " must be finite and numeric", call. = FALSE)
    return(as.numeric(x))
  }
  z <- if (is.list(x)) unlist(x, use.names = FALSE) else as.numeric(x)
  if (!is.numeric(z) || length(z) != sum(ns) || any(!is.finite(z)))
    stop(name, " must be scalar or contain one finite value per observation", call. = FALSE)
  as.numeric(z)
}

.GeoCVSubsetObservationVector <- function(x, idx) {
  if (is.null(x) || length(x) == 1L) x else x[idx]
}

.GeoCVNormalizeCountSize <- function(x, model, name = "fit$n") {
  if (is.null(x)) return(x)
  model <- .GeoPredictionCanonicalModel(as.character(model)[1L])
  z <- if (is.list(x)) unlist(x, use.names = FALSE) else as.numeric(x)
  if (!length(z) || any(!is.finite(z)))
    stop(name, " must contain finite numeric values", call. = FALSE)

  if (model %in% c("BinomialNeg", "BinomialNegZINB")) {
    if (any(z < 1) || any(abs(z - round(z)) > sqrt(.Machine$double.eps)))
      stop(name, " must contain positive integers for direct Negative-Binomial fields.", call. = FALSE)
    if (length(z) > 1L && any(z != z[1L]))
      stop("GeoCV: for direct Negative-Binomial fields n is the common number r of successes and cannot vary by observation.", call. = FALSE)
    return(as.integer(round(z[1L])))
  }

  if (model %in% c("Binary", "Bernoulli", "Geom", "Geometric")) {
    if (any(abs(z - 1) > sqrt(.Machine$double.eps)))
      stop("GeoCV: Binary/Bernoulli and Geometric fields require the common size parameter n=1.", call. = FALSE)
    return(1L)
  }
  x
}

.GeoCVVaryingFixedMean <- function(fixed, name, expected) {
  value <- fixed[[name]]
  if (is.null(value) || length(value) <= 1L) return(NULL)
  value <- as.numeric(unlist(value, use.names = FALSE))
  if (length(value) != expected || any(!is.finite(value)))
    stop("fit$fixed$", name, " is not aligned with the observations.", call. = FALSE)
  value
}

.GeoCVExplicitCoordinates <- function(fit, context) {
  .GeoKrigCoordinateMatrix(
    coordx = fit$coordx, coordy = fit$coordy, coordz = fit$coordz,
    grid = isTRUE(fit$grid), context = context
  )
}

.GeoCVDropMeanParameters <- function(z, prefix) {
  if (is.null(z) || is.null(names(z))) return(z)
  z[!grepl(paste0("^", prefix, "[0-9]*$"), names(z))]
}

.GeoCVRefit <- function(fit, data, X, n, fixed, start, model,
                        optimizer, lower, upper, sparse, method, varest,
                        coordx = NULL, coordx_dyn = NULL, coordt = NULL,
                        universal = FALSE) {
  args <- list(
    data = data, coordx = coordx, coordx_dyn = coordx_dyn,
    corrmodel = fit$corrmodel, X = X,
    likelihood = fit$likelihood, type = fit$type, grid = FALSE,
    copula = fit$copula, anisopars = fit$anisopars,
    est.aniso = fit$est.aniso, model = model,
    radius = fit$radius, n = n,
    maxdist = fit$maxdist, neighb = fit$neighb,
    p_neighb = fit$p_neighb, distance = fit$distance,
    optimizer = optimizer, lower = lower, upper = upper,
    start = start, fixed = fixed,
    weighted = isTRUE(fit$weighted),
    thin_method = if (!is.null(fit$thin_method)) fit$thin_method else "bernoulli",
    memdist = TRUE, method = method, sparse = sparse, varest = varest
  )
  if (!is.null(coordt)) {
    args$coordt <- coordt
    args$maxtime <- fit$maxtime
  }
  ans <- suppressWarnings(.GeoWithDuplicateCheckSuppressed(do.call(GeoFit, args)))
  if (isTRUE(universal) && is.null(ans$varcov)) {
    stop("The refitted model did not return varcov required for Universal kriging.", call. = FALSE)
  }
  ans
}

.GeoCVPredict <- function(fit, loc, Xloc, Mloc, nloc,
                          local, neighb, maxdist, maxtime,
                          type_krig, sparse, method,
                          time = NULL, which = NULL) {
  args <- list(
    fit, loc = loc, mse = TRUE,
    Xloc = Xloc, Mloc = Mloc, nloc = nloc,
    type_krig = type_krig, sparse = sparse, method = method
  )
  if (!is.null(time)) args$time <- time
  if (!is.null(which)) args$which <- which

  if (!local) return(.GeoWithDuplicateCheckSuppressed(do.call(GeoKrig, args)))

  args$neighb <- neighb
  args$maxdist <- maxdist
  if (!is.null(time)) args$maxtime <- maxtime
  args$progress <- FALSE
  .GeoWithDuplicateCheckSuppressed(do.call(GeoKrigloc, args))
}

####################################################
### File name: CompIndLik2.r
####################################################

### Optim call for Indipendence Composite log-likelihood maximization

CompIndLik2 <- function(bivariate, coordx, coordy ,coordz,coordt,coordx_dyn, data, flagcorr, flagnuis, fixed,grid,
                           lower, model, n, namescorr, namesnuis, namesparam,
                           numparam,  optimizer, onlyvar, param, spacetime, type,
                           upper,namesupper, varest, ns, X,sensitivity,copula,MM,score)
  {

############# utility functions ############
lambertW0_custom <- function(x, tol = 1e-12, maxiter = 100) {
  if (length(x) > 1) {
    return(sapply(x, function(xi) lambertW0_custom(xi, tol, maxiter)))
  }
  if (is.na(x) || is.infinite(x)) return(x)
  if (x < -1/exp(1)) return(NaN)
  if (x == 0) return(0)
  if (x == -1/exp(1)) return(-1)
  if (x > 3) {
    # Per valori grandi: W(x) ≈ log(x) - log(log(x))
    w <- log(x) - log(log(x))
  } else if (x > -0.32358) {
    if (x < 1.5) {
      # Serie di potenze modificata
      w <- x * (1 - x + (3/2) * x^2 - (8/3) * x^3 + (125/24) * x^4)
    } else {
      # Approssimazione logaritmica migliorata
      lx <- log(x)
      w <- lx - log(lx) + lx/(1 + lx)
    }
  } else {
    sigma <- -1 - exp(1) * x
    if (sigma > 0) {
      p <- sqrt(2 * sigma)
      w <- -1 + p * (1 - p/3 + 11*p^2/72 - 43*p^3/540 + 769*p^4/17280)
    } else {
      w <- log(-x) - log(-log(-x))
    }
  }
  for (iter in 1:maxiter) {
    ew <- exp(w)
    wew <- w * ew
    f <- wew - x
    if (abs(f) <= tol * (1 + abs(x))) break
    fprime <- ew * (1 + w)
    if (abs(fprime) < 1e-20) break  # Evita divisione per zero
    delta <- f / fprime
    w_new <- w - delta
    if (abs(delta) <= tol * (1 + abs(w))) break
    w <- w_new
  }
  return(w)
}

one_log_SkewLaplace <- function(z, mu, sill, skew) {
  
  sigma <- sqrt(sill)
  z_std <- (z - mu) / sigma  # vettoriale
  base <- log(skew * (1 - skew)) - log(sigma)
  slope <- ifelse(z_std >= 0, -(1 - skew), skew)
  res <- base + slope * z_std
  res
}


one_log_tukeyh <- function(data, mm, sill, tail) {
  q <- (data - mm) / sqrt(sill)
  aa <- tail * q * q
  lambert_vals <- lambertW0_custom(aa)
  x <- sign(q) * sqrt(lambert_vals / tail)
  extra <- 1 / (1 + lambert_vals)
  return(log(dnorm(x) * x * extra / (q * sqrt(sill))))
}

indloglik <- function(fan, data, mm, nuis) {

## gaussian and misspecified gaussian
    if(fan == "Ind_Pair_Gauss") {
        sill = nuis[1]
        res = sum(dnorm(data, mean = mm, sd = sqrt(sill), log = TRUE))
    }

    if(fan == "Ind_Pair_Gauss_misp_T") {
        df = 1/nuis[1]; sill = nuis[2]
        vv = sill * df/(df - 2)
        res = sum(dnorm(data, mean = mm, sd = sqrt(vv), log = TRUE))
    }

    if(fan == "Ind_Pair_Gauss_misp_SkewT") {
        df = 1/nuis[1]; sill = nuis[2]; skew = nuis[3]
        kk = gamma(0.5*(df - 1))/gamma(0.5*(df))
        me = sqrt(sill)*sqrt(df/pi)*kk*skew
        vv = sill*(df/(df - 2) - me^2)
        res = sum(dnorm(data, mean = (mm + me), sd = sqrt(vv), log = TRUE))
    }

    if(fan == "Ind_Pair_Gauss_misp_Tukeygh") {
        sill = nuis[1]; eta = nuis[2]; tail = nuis[3]
        eta2 = eta*eta; u = 1 - tail
        me = sqrt(sill)*(exp(eta2/(2*u)) - 1)/(eta*sqrt(u))
        vv = sill*((exp(2*eta2/(1 - 2*tail)) - 2*exp(eta2/(2*(1 - 2*tail))) + 1)/(eta2*sqrt(1 - 2*tail))) - me*me
        res = sum(dnorm(data, mean = mm + me, sd = sqrt(vv), log = TRUE))
    }

## non gaussian over R
    if(fan == "Ind_Pair_T") {
        sill = nuis[2]; df = 1/nuis[1]
        res = sum(dt((data - mm)/sqrt(sill), df, log = TRUE) - 0.5*log(sill))
    }
    
    if(fan == "Ind_Pair_Logistic") {
        sill = nuis[1]
        res = sum(dlogis(data, mm, sqrt(sill), log = TRUE))
    }
    
    if(fan == "Ind_Pair_SkewGauss") {
        sk = nuis[2]; sill = nuis[1]
        omega = sk*sk + sill
        q = data - mm
        res = sum(log(2) - 0.5*log(omega) + dnorm(q/(sqrt(omega)), log = TRUE) + 
                 pnorm((sk*q)/(sqrt(sill)*sqrt(omega)), log.p = TRUE))
    }
    
    if(fan == "Ind_Pair_SkewLaplace") {
        skew = nuis[2]; sill = nuis[1]
        res = sum(one_log_SkewLaplace(data, mm, sill, skew))
    }

    if(fan == "Ind_Pair_SinhGauss") {
        skew = nuis[2]; sill = nuis[1]; tail = nuis[3]
        q = (data - mm)/(sqrt(sill))
        b1 = tail*asinh(q) - skew; Z1 = sinh(b1)
        res = sum(-0.5*log(q^2 + 1) - 0.5*log(2*pi*sill) + log(cosh(b1)) + log(tail) - Z1*Z1/2)
    }

    if(fan == "Ind_Pair_Tukeyh") {
        sill = nuis[1]; tail = nuis[2]
        res = sum(one_log_tukeyh(data, mm, sill, tail))
    }
    
    if(fan == "Ind_Pair_Tukeyhh") {
        sill = nuis[1]; tail1 = nuis[3]; tail2 = nuis[2]
        res = sum(log(exp(one_log_tukeyh(data, mm, sill, tail2))*I(data >= mm) +
                     exp(one_log_tukeyh(data, mm, sill, tail1))*I(data < mm)))
    }
    
    if(fan == "Ind_Pair_TWOPIECEGauss") {
        sill = nuis[1]; eta = nuis[2]
        y = (data - mm)/sqrt(sill)
        res = sum(log(dnorm(y/(1 - eta))*I(y >= 0) + dnorm(y/(1 + eta))*I(y < 0)) - 0.5*log(sill))
    }
    
    if(fan == "Ind_Pair_TWOPIECETukeyh") {
        sill = nuis[1]; eta = nuis[2]; tail = nuis[3]
        y = (data - mm)/sqrt(sill)
        res = sum(log(exp(one_log_tukeyh(y/(1 - eta), 0, 1, tail))*I(y >= 0) + 
                     exp(one_log_tukeyh(y/(1 + eta), 0, 1, tail))*I(y < 0)) - 0.5*log(sill))
    }
    
    if(fan == "Ind_Pair_TWOPIECET") {
        sill = nuis[2]; eta = nuis[3]; tail = 1/nuis[1]
        y = (data - mm)/sqrt(sill)
        res = sum(log(dt(y/(1 - eta), df = tail)*I(y >= 0) + dt(y/(1 + eta), df = tail)*I(y < 0)) - 0.5*log(sill))
    }

## non gaussian over R^+
    if(fan == "Ind_Pair_Gamma") {
        shape = nuis[2]
        res = sum(dgamma(data, shape = shape/2, scale = 1/(shape/(2*exp(mm))), log = TRUE))
    }
    
    if(fan == "Ind_Pair_Weibull") {
        shape = nuis[2]
        res = sum(dweibull(data, shape = shape, scale = exp(mm)/(gamma(1 + 1/shape)), log = TRUE))
    }
    
    if(fan == "Ind_Pair_LogGauss") {
        sill = nuis[1]
        c1 = mm - sill/2
        res = sum(dlnorm(data, meanlog = c1, sdlog = sqrt(sill), log = TRUE))
    }
    
    if(fan == "Ind_Pair_LogLogistic") {
        dllogis1 <- function(x, shape, rate = 1, scale = 1/rate, log = FALSE) {
            z <- x / scale
            num <- (shape / scale) * z^(shape - 1)
            denom <- (1 + z^shape)^2
            dens <- num / denom
            if (log) dens <- log(dens)
            return(dens)
        }
        shape = nuis[2]
        ci = gamma(1 + 1/shape)*gamma(1 - 1/shape)
        res = sum(dllogis1(data, shape, scale = exp(mm)/ci, log = TRUE))
    }

## non Gaussian bounded support
    if(fan == "Ind_Pair_Beta2") {
        mmax = nuis[4]; mmin = nuis[3]
        shape = nuis[2]
        me = 1/(1 + exp(-mm))
        res = sum(dbeta((data - mmin)/(mmax - mmin), me*shape, (1 - me)*shape, log = TRUE) - log(mmax - mmin))
    }
    
    if(fan == "Ind_Pair_Kumaraswamy2") {
        mmax = nuis[4]; mmin = nuis[3]
        shape = nuis[2]
        q = (data - mmin)/(mmax - mmin); k = 1 - q^shape
        m1 = 1/(1 + exp(-mm))
        shapei = log(0.5)/log1p(-m1^shape)
        res = sum(log(shapei) + log(shape) + (shape - 1)*log(q) + (shapei - 1)*log(k) - log(mmax - mmin))
    }

#### discrete
    if(fan == "Ind_Pair_Pois") {
        mm = exp(mm)
        res = sum(dpois(data, mm, log = TRUE))
    }
    
    if(fan == "Ind_Pair_Gauss_misp_Pois") {
        mm = exp(mm)
        res = sum(dnorm(data, mean = mm, sd = sqrt(mm), log = TRUE))
    }
    
      if(fan == "Ind_Pair_BinomGauss") {
        mm_safe <- pmax(pmin(mm, 20), -20)
        prob <- pnorm(mm_safe)
        prob <- pmax(pmin(prob, 1 - .Machine$double.eps), .Machine$double.eps)
        
        if(any(!is.finite(prob))) return(1e8)
        res <- sum(dbinom(data, n, prob, log = TRUE))
        if(!is.finite(res)) return(1e8)
    }
    
    
  if(fan == "Ind_Pair_BinomGauss_misp") {
        mm_safe <- pmax(pmin(mm, 20), -20)
        prob <- pnorm(mm_safe)
        prob <- pmax(pmin(prob, 1 - .Machine$double.eps), .Machine$double.eps)
        
        if(any(!is.finite(prob))) return(1e8)
        
        mm_adj <- n * prob
        vv <- pmax(mm_adj * (1 - prob), 1e-10)
        res <- sum(dnorm(data, mean = mm_adj, sd = sqrt(vv), log = TRUE))
        if(!is.finite(res)) return(1e8)
    }

    
   if(fan == "Ind_Pair_BinomnegGauss") {
        mm_safe <- pmax(pmin(mm, 20), -20)
        prob <- pnorm(mm_safe)
        prob <- pmax(pmin(prob, 1 - .Machine$double.eps), .Machine$double.eps)
        
        if(any(!is.finite(prob))) return(1e8)
        res <- sum(dnbinom(data, n, prob, log = TRUE))
        if(!is.finite(res)) return(1e8)
    }
    
    if(fan == "Ind_Pair_PoisGamma") {
        mm = exp(mm)
        res = sum(dnbinom(data, nuis[2], mu = mm, log = TRUE))
    }

    if(fan == "Ind_Pair_PoisGammaZIP") {
        mm = exp(mm)
        pp = pnorm(nuis[3])
        pp = pmax(pmin(pp, 1 - 1e-18), 1e-18)
        res1 = sum(log(pp + (1 - pp)*dnbinom(data[data == 0], nuis[2], mu = mm[data == 0], 0)))
        res2 = sum(log(1 - pp) + dnbinom(data[data != 0], nuis[2], mu = mm[data != 0], 1))
        res = res1 + res2
        
        if(!is.finite(res)) {
            return(1e8)
        }
    }

    if(fan == "Ind_Pair_Gauss_misp_PoisGamma") {
        mm = exp(mm)
        res = sum(dnorm(data, mean = mm, sd = sqrt(mm*(1 + mm/nuis[2])), log = TRUE))
    }
    
    if(fan == "Ind_Pair_PoisZIP") {
        mm = exp(mm)
        pp = pnorm(nuis[3])
        pp = pmax(pmin(pp, 1 - 1e-18), 1e-18)
        res1 = sum(log(pp + (1 - pp)*dpois(data[data == 0], mm[data == 0], 0)))
        res2 = sum(log(1 - pp) + dpois(data[data != 0], mm[data != 0], 1))
        res = res1 + res2
        
        if(!is.finite(res)) {
            return(1e8)
        }
    }
    
    if(fan == "Ind_Pair_Gauss_misp_PoisZIP") {
        mm = exp(mm)
        pp = pnorm(nuis[3])
        pp = pmax(pmin(pp, 1 - 1e-18), 1e-18)
        res1 = sum(log(pp + (1 - pp)*dnorm(data[data == 0], mean = mm[data == 0], sd = sqrt(mm[data == 0]), log = FALSE)))
        res2 = sum(log(1 - pp) + dnorm(data[data != 0], mean = mm[data != 0], sd = sqrt(mm[data != 0]), log = TRUE))
        res = res1 + res2
        
        if(!is.finite(res)) {
            return(1e8)
        }
    }
    
  if(fan == "Ind_Pair_BinomnegGaussZINB") {
        mm_safe <- pmax(pmin(mm, 20), -20)
        prob <- pnorm(mm_safe)
        prob <- pmax(pmin(prob, 1 - .Machine$double.eps), .Machine$double.eps)
        pp <- pnorm(nuis[3])
        pp <- pmax(pmin(pp, 1 - .Machine$double.eps), .Machine$double.eps)
        
        if(any(!is.finite(prob)) || any(!is.finite(pp))) return(1e8)
        
        res1 <- if(any(data == 0)) sum(log(pp + (1 - pp) * dnbinom(data[data == 0], n, prob[data == 0], log = FALSE))) else 0
        res2 <- if(any(data != 0)) sum(log(1 - pp) + dnbinom(data[data != 0], n, prob[data != 0], log = TRUE)) else 0
        res <- res1 + res2
        if(!is.finite(res)) return(1e8)
    }

    return(-res)
}

compindloglik2 <- function(param, data, fixed, fan, n, namesnuis, namesparam, X, MM) {
    names(param) <- namesparam
    param <- c(param, fixed)
    nuisance <- param[namesnuis]
    mean_names <- .GeoMean_names(ncol(X))
    sel <- names(nuisance) %in% mean_names
    other_nuis <- as.numeric(nuisance[!sel])
    
    Mean <- if (is.null(MM)) {
      as.numeric(X %*% .GeoMean_beta(nuisance, ncol(X)))
    } else {
      as.numeric(MM)
    }
    result <- indloglik(fan, data, Mean, other_nuis)
    if(!is.finite(result)) result <- 1e8
    return(result)
}

 compindloglik_biv2 <- function(param, data1,data2,fixed, fan, n,   ## to do...
                           namesnuis,namesparam,X,MM)
      {

        names(param) <- namesparam
        param <- c(param, fixed)
        nuisance <- param[namesnuis]
        sel1=substr(names(nuisance),1,6)=="mean_1"
        mm1=as.numeric(nuisance[sel1])
        sel2=substr(names(nuisance),1,6)=="mean_2"
        mm2=as.numeric(nuisance[sel2])
        sel=substr(names(nuisance),1,4)=="mean"
        X1=as.matrix(X[1:ns[1],]);X2=as.matrix(X[(ns[1]+1):(ns[2]+ns[1]),]); 
        other_nuis=as.numeric(nuisance[!sel]) 
        Mean=c(X1%*%mm1,X2%*%mm2)
        res=double(1)
     
        result=4
        return(-result)
      }

   ##################################################################################################
   ############### starting function ###############################################################
   ##################################################################################################

    numcoord=length(coordx);numtime=1;spacetime_dyn=FALSE
    if(spacetime) numtime=length(coordt)
    if(bivariate) numtime=2
    if(!is.null(coordx_dyn)) spacetime_dyn=TRUE

    dimat=numcoord*numtime#
    if((spacetime||bivariate)) dimat <- sum(ns)
    NS=cumsum(ns)
    if(is.null(dim(X))){X=as.matrix(rep(1,dimat))}
    #else(if(bivariate) X=rbind(X,X))


    fname <- NULL; hessian <- FALSE
     
###################### pairwise ###############################################

    # --- lookup vector ---------------------------------------------------------
#  (keep only the models that are actually used)
lookup <- c(
  Gauss                    =  1,
  BinomGauss               = 11,         
  BinomnegGauss            = 16,  
  WrapGauss                = 13,
  SkewGauss                = 10,
  SkewLaplace              = 59,
  Gamma                    = 21,
  Kumaraswamy2             = 42,
  Beta2                    = 50,
  Weibull                  = 26,
  LogLogistic              = 24,
  Logistic                 = 25,
  LogGauss                 = 22,
  TWOPIECET                = 27,
  TWOPIECEGauss            = 29,
  T                        = 12,
  Tukeyh                   = 34,
  Tukeyhh                  = 40,
  Gauss_misp_Tukeygh       = 41,
  Gauss_misp_Pois          = 36,
  Gauss_misp_T             = 35,
  Gauss_misp_SkewT         = 37,
  SinhGauss                = 20,
  TWOPIECETukeyh           = 38,
  Pois                     = 30,
  PoisGamma                = 46,
  PoisGammaZIP             = 57,
  PoisZIP                  = 43,
  Gauss_misp_PoisZIP       = 44,
  BinomnegGaussZINB        = 45,
  Gauss_misp_PoisGamma     = 47,
  BinomGauss_misp          = 51
)

# reverse the map: model numbers → names
model2name <- setNames(rep(names(lookup), lengths(lookup)),
                       unlist(lookup, use.names = FALSE))
fname <- paste0("Ind_Pair_", model2name[ as.character(model) ])
if (is.na(fname)) {
  stop(
    "Independence likelihood is not implemented for model code ",
    model,
    call. = FALSE
  )
}

if (bivariate) {
  stop(
    "Independence likelihood initialization is not implemented for bivariate models",
    call. = FALSE
  )
}

########################################################################                                            
    if(sensitivity) hessian=TRUE
    
     if((spacetime||bivariate)&&(!spacetime_dyn))    data=c(t(data))
     if((spacetime||bivariate)&&(spacetime_dyn))     data=unlist(data)          
     if(spacetime||bivariate)   NS=c(0,NS)[-(length(ns)+1)]

####     
tot=c(param,fixed) ## all the parameters
## deleting corr para
tot=tot[is.na(pmatch(names(tot),namescorr))]
## deleting nugget
tot=tot[names(tot)!='nugget'];namesnuis=namesnuis[namesnuis!="nugget"]
param=tot[pmatch(namesparam,names(tot))];param=param[!is.na(names(param))];param=param[!is.na(param)];

if(model  %in%  c(2,14,16,21,42,50,26,24,25,30,46,43,11))  ## model where sill must be fixed = to 1
 {param=param[names(param)!='sill'];a=1; names(a)="sill";
  if(is.null(unlist(fixed['sill']))) fixed=c(fixed,a)}

if(!is.null(copula))
    {if(copula=="Clayton")
           {param=param[names(param)!='nu'];a=2; names(a)="nu";
           if(is.null(unlist(fixed['nu']))) fixed=c(fixed,a)}}

namesparam <- names(param)

## Match bounds by parameter name.  Bounds supplied by WlsStart may also
## contain correlation parameters, which have already been removed here.
subset_ind_bound <- function(bound, target_names) {
  if (is.null(bound) || length(bound) == 0L)
    return(rep(NA_real_, length(target_names)))

  bb <- unlist(bound, use.names = TRUE)

  if (!is.null(names(bb)) && all(target_names %in% names(bb)))
    return(as.numeric(bb[target_names]))

  if (length(bb) == length(target_names))
    return(as.numeric(bb))

  rep(NA_real_, length(target_names))
}

lower <- subset_ind_bound(lower, namesparam)
upper <- subset_ind_bound(upper, namesparam)
names(lower) <- namesparam
names(upper) <- namesparam

## Backward-compatible initialization used by the previous implementation.
## This changes only the starting point; the likelihood itself is not altered.
sel_mean_zero <- startsWith(names(param), "mean") &
                 is.finite(param) &
                 param == 0
param[sel_mean_zero] <- 0.1

param <- as.numeric(param)

## CompIndLik2 has native implementations for these optimizers.  Other
## GeoFit optimizers are mapped to Nelder-Mead for the preliminary fit.
supported_ind_optimizers <- c(
  "L-BFGS-B", "BFGS", "SANN", "Nelder-Mead",
  "nlm", "nlminb", "bobyqa"
)

if (!(optimizer %in% supported_ind_optimizers)) {
  warning(
    "Optimizer '", optimizer,
    "' is not implemented for independence likelihood; using Nelder-Mead",
    call. = FALSE
  )
  optimizer <- "Nelder-Mead"
}

## Bounded optimizers require finite marginal bounds.  If the spatial
## bounds cannot be matched to the marginal parameters, use Nelder-Mead.
if (optimizer %in% c("L-BFGS-B", "nlminb", "bobyqa") &&
    (any(!is.finite(lower)) || any(!is.finite(upper)))) {
  warning(
    "Finite marginal bounds are unavailable; using Nelder-Mead for the independence fit",
    call. = FALSE
  )
  optimizer <- "Nelder-Mead"
}

   if(!onlyvar){
  ##############################.  spatial or space time ############################################
   if(!bivariate)           {
    if (length(param) == 1L) {

      optimizer <- "optimize"
      lower <- as.numeric(lower[1])
      upper <- as.numeric(upper[1])

      if (!is.finite(lower) || !is.finite(upper) || lower >= upper) {
        pname <- namesparam[1]
        p0 <- param[1]

        if (startsWith(pname, "mean")) {
          lower <- p0 - 20 * (1 + abs(p0))
          upper <- p0 + 20 * (1 + abs(p0))
        } else {
          stop(
            "Finite lower and upper bounds are required for the one-parameter independence fit of '",
            pname, "'",
            call. = FALSE
          )
        }
      }

     CompLikelihood <- optimize(f= compindloglik2,    
                              data=data, fixed=fixed, fan=fname,  lower=lower, n=n,
                               namesnuis=namesnuis,namesparam=namesparam, maximum = FALSE,
                              upper= upper,  X=X,MM=MM)}
   if(length(param)>1) {
    
    if(optimizer=='L-BFGS-B'){
      CompLikelihood <- optim(par=param,fn= compindloglik2, 
                              control=list(factr=1e-10,pgtol=1e-14, maxit=100000), 
                                data=data, fixed=fixed,
                              fan=fname, lower=lower, method='L-BFGS-B',n=n,
                               namesnuis=namesnuis,namesparam=namesparam, 
                              upper=upper,  X=X,MM=MM,   hessian=FALSE)
  }

    if(optimizer=='BFGS') 
        CompLikelihood <- optim(par=param, fn= compindloglik2,     
                           control=list(factr=1e-10,
                             reltol=1e-14, maxit=100000),data=data, fixed=fixed, fan=fname,
                              hessian=FALSE, method='BFGS',n=n,
                              namesnuis=namesnuis,namesparam=namesparam,  X=X,MM=MM)
   if(optimizer=='SANN'){ 
      CompLikelihood <- optim(par=param, fn= compindloglik2,     
                           control=list(factr=1e-10,
                             reltol=1e-14, maxit=100000),data=data, fixed=fixed, fan=fname,
                              hessian=FALSE, method='SANN',n=n,
                              namesnuis=namesnuis,namesparam=namesparam,  X=X,MM=MM)
  }
      if(optimizer=='Nelder-Mead')
        CompLikelihood <- optim(par=param, fn= compindloglik2,     
          control=list( reltol=1e-14, maxit=100000), data=data, fixed=fixed, fan=fname,
                              hessian=FALSE, method='Nelder-Mead',n=n,
                                  namesnuis=namesnuis,namesparam=namesparam,  X=X,MM=MM)
    if(optimizer=='nlm')
    CompLikelihood <- nlm(f= compindloglik2,p=param,steptol = 1e-4,    data=data, fixed=fixed,
                               fan=fname,hessian=FALSE,n=n, namesnuis=namesnuis,namesparam=namesparam, 
                               iterlim=100000,   X=X,MM=MM)
  
    if(optimizer=='nlminb')
     CompLikelihood <-nlminb(objective= compindloglik2,start=param,   data=data, fixed=fixed,
                                control = list( iter.max=100000),
                              lower=lower,upper=upper,
                               fan=fname,n=n, namesnuis=namesnuis,namesparam=namesparam, 
                                 X=X,MM=MM)    
    if (optimizer == 'bobyqa') 
             CompLikelihood <- minqa::bobyqa(par = param,fn = compindloglik2,lower = lower,upper = upper,
            control = list(maxfun = 100000, rhobeg = 0.5, rhoend = 1e-8),data = data,fixed = fixed,fan = fname,n = n,
             namesnuis = namesnuis,namesparam = namesparam,X = X,MM = MM)               
    }}
######################################################################################
############################## bivariate  ############################################ 
######################################################################################                          
    if(bivariate)           { 
     if(length(param)==1)
        {
         optimizer="optimize" 

       CompLikelihood <- optimize(f= compindloglik_biv2,     
                              data=data, fixed=fixed,fan=fname,  lower=lower,n=n,
                               namesnuis=namesnuis, namesparam=namesparam,maximum = FALSE,
                              upper=upper,  X=X,MM=MM )}
      if(length(param)>1) {   
    if(optimizer=='L-BFGS-B'){
      CompLikelihood <- optim(param, compindloglik_biv2, control=list(pgtol=1e-14, maxit=100000),
                              method='L-BFGS-B',hessian=FALSE,lower=lower, upper=upper,
                                  
                              data=data, fixed=fixed,fan=fname,n=n,
                               namesnuis=namesnuis, namesparam=namesparam,
                               X=X ,MM=MM )}

    
      if(optimizer=='BFGS')
      CompLikelihood <- optim(param, compindloglik_biv2,     control=list(
                              reltol=1e-14, maxit=100000), data=data, fixed=fixed, fan=fname,
                              hessian=FALSE, method='BFGS',n=n,
                               namesnuis=namesnuis,namesparam=namesparam ,  X=X,MM=MM )
   if(optimizer=='Nelder-Mead')
      CompLikelihood <- optim(param, compindloglik_biv2,     control=list(
                              reltol=1e-14, maxit=100000), data=data, fixed=fixed, fan=fname,
                              hessian=FALSE, method='Nelder-Mead',n=n,
                               namesnuis=namesnuis,namesparam=namesparam ,  X=X,MM=MM )
     if(optimizer=='nlm') 
        CompLikelihood <- nlm( f= compindloglik_biv2,p=param,     data=data, fixed=fixed,
                               fan=fname,hessian=FALSE,n=n, namesnuis=namesnuis,namesparam=namesparam, 
                                 X=X,MM=MM )
    if(optimizer=='nlminb') 
        CompLikelihood <- nlminb( objective= compindloglik_biv2,start=param, 
                                     control = list( iter.max=100000),
                              lower=lower,upper=upper,
                                   data=data, fixed=fixed,
                               fan=fname,n=n, namesnuis=namesnuis,namesparam=namesparam, 
                                 X=X ,MM=MM)
   }
 }  

                   
      ########################################################################################   
      ########################################################################################
    # check the optimisation outcome
      if (optimizer %in% c('Nelder-Mead', 'SANN')){
        CompLikelihood$value = -CompLikelihood$value
        names(CompLikelihood$par)<- namesparam
        if(CompLikelihood$convergence == 0)
        CompLikelihood$convergence <- 'Successful'
        else
        if(CompLikelihood$convergence == 1)
        CompLikelihood$convergence <- 'Iteration limit reached'
        else
        CompLikelihood$convergence <- "Optimization may have failed"
        if(CompLikelihood$value==-1.0e8) CompLikelihood$convergence <- 'Optimization may have failed: Try with other starting parameters'
    }
        if(optimizer=='nmk'||optimizer=='nmkb'){
        CompLikelihood$value = -CompLikelihood$value
        names(CompLikelihood$par)<- namesparam
        if(CompLikelihood$convergence == 0)
        CompLikelihood$convergence <- 'Successful'
        else CompLikelihood$convergence <- "Optimization may have failed"
        if(CompLikelihood$value==-1.0e8) CompLikelihood$convergence <- 'Optimization may have failed: Try with other starting parameters'
    }
      
    if(optimizer=='L-BFGS-B'||optimizer=='BFGS'){
        CompLikelihood$value = -CompLikelihood$value
        names(CompLikelihood$par)<- namesparam
        if(CompLikelihood$convergence == 0)
        CompLikelihood$convergence <- 'Successful'
        else
        if(CompLikelihood$convergence == 1)
        CompLikelihood$convergence <- 'Iteration limit reached'
        else
        CompLikelihood$convergence <- "Optimization may have failed"
        if(CompLikelihood$value==-1.0e8) CompLikelihood$convergence <- 'Optimization may have failed: Try with other starting parameters'
    }

     if(optimizer=='nlm'){
        CompLikelihood$par <- CompLikelihood$estimate
        names(CompLikelihood$par)<- namesparam
        CompLikelihood$value <- -CompLikelihood$minimum
        if(CompLikelihood$code == 1|| CompLikelihood$code == 2)
        CompLikelihood$convergence <- 'Successful'
        else
        if(CompLikelihood$code == 4)
        CompLikelihood$convergence <- 'Iteration limit reached'
        else
        CompLikelihood$convergence <- "Optimization may have failed"
        if(CompLikelihood$value==-1.0e8) CompLikelihood$convergence <- 'Optimization may have failed: Try with other starting parameters'
    }

    if(optimizer=='nlminb'||optimizer=='multinlminb'){
        CompLikelihood$par <- CompLikelihood$par
        names(CompLikelihood$par)<- namesparam
        CompLikelihood$value <- -CompLikelihood$objective
        if(CompLikelihood$convergence == 0) { CompLikelihood$convergence <- 'Successful' }
        else {CompLikelihood$convergence <- "Optimization may have failed" }
        if(CompLikelihood$objective==-1.0e8) CompLikelihood$convergence <- 'Optimization may have failed: Try with other starting parameters'
    }
    if (optimizer == 'bobyqa') {
      names(CompLikelihood$par) <- namesparam
      CompLikelihood$value <- -CompLikelihood$fval
      ierr <- if (!is.null(CompLikelihood$ierr)) CompLikelihood$ierr else
              if (!is.null(CompLikelihood$ierror)) CompLikelihood$ierror else NA_integer_
      if (isTRUE(ierr == 0L)) {
        CompLikelihood$convergence <- 'Successful'
      } else {
        msg <- if (!is.null(CompLikelihood$msg)) CompLikelihood$msg else
               'Optimization may have failed'
        CompLikelihood$convergence <- msg
      }
      if (is.null(CompLikelihood$counts))
        CompLikelihood$counts <- CompLikelihood$feval
    }

    if(optimizer=='optimize'){
    param<-CompLikelihood$minimum
    CompLikelihood$par<-param  
    names(CompLikelihood$par)<- namesparam
    maxfun <- -CompLikelihood$objective
    CompLikelihood$value <- maxfun
    CompLikelihood$convergence <- 'Successful'
    }
  } ##### end if !onlyvar
    else {
          CompLikelihood=as.list(0)
          names(CompLikelihood)="value"
          CompLikelihood$par <- param
          ## param is converted to an unnamed numeric vector above so that the
          ## optimizers receive a plain numeric vector.  In the onlyvar path no
          ## optimizer runs to restore the parameter names; preserve the same
          ## named output contract as the fitted Independence paths.
          names(CompLikelihood$par) <- namesparam
          CompLikelihood$claic <- NULL;CompLikelihood$clbic <- NULL;
          CompLikelihood$convergence <- 'Successful'
          if(!bivariate) CompLikelihood$value = -  compindloglik2(param=CompLikelihood$par ,     
                              data=data, fixed=fixed, fan=fname,
                             n=n,namesnuis=namesnuis,namesparam=namesparam,  X=X,MM=MM)
          else CompLikelihood$value = - compindloglik_biv2(param=CompLikelihood$par ,     
                data=data, fixed=fixed, fan=fname,
                             n=n,namesnuis=namesnuis,namesparam=namesparam,  X=X,MM=MM)

          if(hessian) 
          {
               if(!bivariate)  
                CompLikelihood$hessian=numDeriv::hessian(func= compindloglik2,x=param,method="Richardson",     
                              data=data,fixed=fixed,fan=fname,n=n,
                               namesnuis=namesnuis, namesparam=namesparam,
                                X=X,MM=MM )
               if(bivariate)  
               CompLikelihood$hessian=numDeriv::hessian(func= compindloglik_biv2,x=param,method="Richardson",   
                             data=data, fixed=fixed,fan=fname,n=n, namesnuis=namesnuis,namesparam=namesparam, 
                               X=X ,MM=MM)
               rownames(CompLikelihood$hessian)=namesparam
               colnames(CompLikelihood$hessian)=namesparam
          }
  }

#####################################
if((sensitivity||varest))
  {
if(!bivariate)  
CompLikelihood$hessian=numDeriv::hessian(func= compindloglik2,x=CompLikelihood$par,method="Richardson",      
                              data=data, fixed=fixed,fan=fname,n=n,
                               namesnuis=namesnuis, namesparam=namesparam,
                                X=X ,MM=MM)
if(bivariate)  
CompLikelihood$hessian=numDeriv::hessian(func= compindloglik_biv2,x=CompLikelihood$par,method="Richardson",    
                          data=data, fixed=fixed,fan=fname,n=n, namesnuis=namesnuis,namesparam=namesparam, 
                                 X=X ,MM=MM)
rownames(CompLikelihood$hessian)=namesparam
colnames(CompLikelihood$hessian)=namesparam
  }

CompLikelihood$score=NULL
if(score)
  {
if(!bivariate)  
CompLikelihood$score=numDeriv::grad(func= compindloglik2,x=CompLikelihood$par,method="Richardson",      
                              data=data, fixed=fixed,fan=fname,n=n,
                               namesnuis=namesnuis, namesparam=namesparam,
                                X=X ,MM=MM)
if(bivariate)  
CompLikelihood$score=numDeriv::grad(func= compindloglik_biv2,x=CompLikelihood$par,method="Richardson",    
                          data=data, fixed=fixed,fan=fname,n=n, namesnuis=namesnuis,namesparam=namesparam, 
                                 X=X ,MM=MM)
names(CompLikelihood$score)=namesparam
  }



if (sensitivity || varest)
  CompLikelihood$sensmat <- CompLikelihood$hessian
else
  CompLikelihood$sensmat <- NULL

if (is.null(CompLikelihood$claic))  CompLikelihood$claic <- NULL
if (is.null(CompLikelihood$clbic))  CompLikelihood$clbic <- NULL
if (is.null(CompLikelihood$stderr)) CompLikelihood$stderr <- NULL
if (is.null(CompLikelihood$varcov)) CompLikelihood$varcov <- NULL
if (is.null(CompLikelihood$varimat)) CompLikelihood$varimat <- NULL
if (is.null(CompLikelihood$counts)) CompLikelihood$counts <- NULL
if (is.null(CompLikelihood$message)) CompLikelihood$message <- NULL

    return(CompLikelihood)
  }

####################################################
### File name: CompLik2.R
####################################################

### Optim call for Composite log-likelihood maximization


CompLik2 <- function(copula,bivariate, coordx, coordy ,coordz,coordt,coordx_dyn,corrmodel, data, distance, flagcorr, flagnuis, fixed,grid,
                           likelihood,lower, model, n, namescorr, namesnuis, namesparam,
                           numparam, numparamcorr, optimizer, onlyvar, param, spacetime, type,
                           upper, varest,  weigthed, ns, X,sensitivity,
                           colidx,rowidx,neighb,MM,aniso,score)
  {

comploglik2 <- function(param, colidx, rowidx, corrmodel, coords, data1, data2, fixed, fan, n, namescorr,
                              namesnuis, namesparam, namesaniso, weigthed, X, MM, aniso, type_cop, cond_pair) {

    names(param) <- namesparam
    param <- c(param, fixed)

    ## Guard copula and marginal parameter domains at every objective
    ## evaluation. This is essential for unconstrained optimizers such as
    ## Nelder-Mead, especially for the skew-Gaussian eta parameter.
    if (type_cop > 0L) {
        bad <- FALSE
        get1 <- function(nm) {
            if (!(nm %in% names(param))) return(NA_real_)
            suppressWarnings(as.numeric(param[[nm]]))
        }

        nug <- get1("nugget")
        if (!is.finite(nug) || nug < 0 || nug >= 1) bad <- TRUE

        if ("sill" %in% names(param)) {
            sv <- get1("sill")
            if (!is.finite(sv) || sv <= 0) bad <- TRUE
        }
        if (model == 12L) {
            dfv <- get1("df")
            if (!is.finite(dfv) || dfv <= 0) bad <- TRUE
        }
        if (model %in% c(21L, 26L, 50L, 42L)) {
            sh <- get1("shape")
            if (!is.finite(sh) || sh <= 0) bad <- TRUE
        }
        if (model %in% c(28L, 33L)) {
            sh1 <- get1("shape1"); sh2 <- get1("shape2")
            if (!is.finite(sh1) || !is.finite(sh2) || sh1 <= 0 || sh2 <= 0) bad <- TRUE
        }
        if (model %in% c(28L, 33L, 50L, 42L)) {
            lo <- get1("min"); hi <- get1("max")
            if (!is.finite(lo) || !is.finite(hi) || lo >= hi) bad <- TRUE
        }
        if (model == 59L) {
            sk <- get1("skew")
            if (!is.finite(sk) || sk <= 0 || sk >= 1) bad <- TRUE
        }

        if (type_cop == 2L) {
            nu_cop <- get1("nu")
            if (!is.finite(nu_cop) || nu_cop <= 0 ||
                abs(nu_cop - round(nu_cop)) > sqrt(.Machine$double.eps)) bad <- TRUE
        }
        if (type_cop == 3L) {
            eta_cop <- get1("nu")
            if (!is.finite(eta_cop) || abs(eta_cop) >= 1) bad <- TRUE
        }

        if (bad) return(1e20)
    }

    paramcorr <- param[namescorr]
    nuisance <- param[namesnuis]
    sel <- sel_mean

    if (is.null(MM)) {
        beta <- .GeoMean_beta(nuisance, ncol(X))
        Mean <- as.numeric(X %*% beta)
    } else {
        Mean <- as.numeric(MM)
    }
    other_nuis <- nuisance[!sel]

    if (aniso) {
        coords1 <- GeoAniso(coords, anisopars = param[namesaniso])
        c1 <- as.vector(t(coords1[colidx, , drop = FALSE]))
        c2 <- as.vector(t(coords1[rowidx, , drop = FALSE]))

        result <- dotCall64::.C64(
            as.character(fan),
            SIGNATURE = c("integer", "double", "double", "double", "double",
                         "integer", "integer", "double", "integer", "double",
                         "double", "double", "double", "integer", "integer"),
            corrmodel, c1, c2, data1, data2,
            n_col, n_row, paramcorr, weigthed,
            res = dotCall64::vector_dc("double", 1),
            Mean[colidx], Mean[rowidx], other_nuis, type_cop, cond_pair,
            INTENT = c(rep("r", 9), "w", rep("r", 5)),
            PACKAGE = 'GeoModels',
            VERBOSE = 0,
            NAOK = TRUE
        )$res

    } else {
        result <- dotCall64::.C64(
            as.character(fan),
            SIGNATURE = c("integer", "double", "double", "integer", "integer",
                         "double", "integer", "double", "double", "double",
                         "double", "integer", "integer"),
            corrmodel, data1, data2,
            n_col, n_row, paramcorr, weigthed,
            res = dotCall64::vector_dc("double", 1),
            Mean[colidx], Mean[rowidx], other_nuis, type_cop, cond_pair,
            INTENT = c(rep("r", 7), "w", rep("r", 5)),
            PACKAGE = 'GeoModels',
            VERBOSE = 0,
            NAOK = TRUE
        )$res
    }

    ## Optimizers work on the average negative composite log-likelihood.
    ## obj_scale is defined after the pair selection and is constant in theta.
    return(-result / obj_scale)
}
##################################################################
##################################################################
comploglik_biv2 <- function(param,colidx,rowidx, corrmodel, coords,data1,data2,fixed, fan, n,
                          namescorr, namesnuis,namesparam,namesaniso,weigthed,X,MM,aniso,type_cop,cond_pair)
      {

        names(param) <- namesparam
        param <- c(param, fixed)
        paramcorr <- param[namescorr]
        nuisance <- param[namesnuis]
        anisopar <- param[namesaniso]

        sel1=substr(names(nuisance),1,6)=="mean_1"
        mm1=as.numeric(nuisance[sel1])
        sel2=substr(names(nuisance),1,6)=="mean_2"
        mm2=as.numeric(nuisance[sel2])
        sel=substr(names(nuisance),1,4)=="mean"
        X1=as.matrix(X[1:ns[1],]);X2=as.matrix(X[(ns[1]+1):(ns[2]+ns[1]),]);
        other_nuis=as.numeric(nuisance[!sel])
        Mean=c(X1%*%mm1,X2%*%mm2)
        res=double(1)

        result=dotCall64::.C64(as.character(fan),
          SIGNATURE = c("integer","double","double", "integer","double","integer","double","double","double","double"),
                        corrmodel,data1, data2, n,paramcorr,weigthed, res=res,Mean[colidx],Mean[rowidx],other_nuis,
          INTENT =    c(rep("r",6),"w", rep("r", 3)),
             PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)$res
        ## Optimizers work on the average negative composite log-likelihood.
        ## obj_scale is defined after the pair selection and is constant in theta.
        return(-result / obj_scale)
      }

   ##################################################################################################
   ############### starting function ################################################################
   ##################################################################################################

    numcoord=length(coordx);numtime=1;spacetime_dyn=FALSE
    if(spacetime) numtime=length(coordt)
    if(bivariate) numtime=2
    if(!is.null(coordx_dyn)) spacetime_dyn=TRUE

    dimat=numcoord*numtime
    if((spacetime||bivariate)) dimat <- sum(ns)
    NS=cumsum(ns)
    if(is.null(dim(X))){X=as.matrix(rep(1,dimat))}

    fname <- NULL; hessian <- FALSE
    namesaniso=c("angle","ratio")

    if(length(n)==1) n=rep(n,dimat)
####################### conditional ##############################################
    if(all(model==1, likelihood==1,type==2)) fname <- 'Comp_Cond_Gauss'
    if(all(model==21,likelihood==1,type==2)) fname <- 'Comp_Cond_Gamma'
    if(all(model==26,likelihood==1,type==2)) fname <- 'Comp_Cond_Weibull'
    if(all(model==22,likelihood==1,type==2)) fname <- 'Comp_Cond_LogGauss'
    if(all(model==28,likelihood==1,type==2)) fname <- 'Comp_Cond_Beta'
    if(all(model==50,likelihood==1,type==2)) fname <- 'Comp_Cond_Beta2'
    if(all(model==13,likelihood==1,type==2)) fname <- 'Comp_Cond_WrapGauss'
    if(all(model==33,likelihood==1,type==2)) fname <- 'Comp_Cond_Kumaraswamy'
    if(all(model==42,likelihood==1,type==2)) fname <- 'Comp_Cond_Kumaraswamy2'
    if(all(model==20,likelihood==1,type==2)) fname <- 'Comp_Cond_SinhGauss'
    if(all(model==38,likelihood==1,type==2)) fname <- 'Comp_Cond_TWOPIECETukeyh'
    if(all(model==27,likelihood==1,type==2)) fname <- 'Comp_Cond_TWOPIECET'
    if(all(model==29,likelihood==1,type==2)) fname <- 'Comp_Cond_TWOPIECEGauss'
    if(all(model==10,likelihood==1,type==2)) fname <- 'Comp_Cond_SkewGauss'
    if(all(model==12,likelihood==1,type==2)) fname <- 'Comp_Cond_T'
    if(all(model==34,likelihood==1,type==2)) fname <- 'Comp_Cond_Tukeyh'
    if(all(model==40,likelihood==1,type==2)) fname <- 'Comp_Cond_Tukeyhh'
    if(all(model==35,likelihood==1,type==2)) fname <- 'Comp_Cond_Gauss_misp_T'
    if(all(model==39,likelihood==1,type==2)) fname <- 'Comp_Cond_TWOPIECEBIMODAL'
    if(all(model==31,likelihood==1,type==2)) fname <- 'Comp_Cond_BinomTWOPIECEGauss'
    if(all(model==32,likelihood==1,type==2)) fname <- 'Comp_Cond_BinomnegTWOPIECEGauss'
    if(all(model==30,likelihood==1,type==2)) fname <- 'Comp_Cond_Pois'
    if(all(model==46,likelihood==1,type==2)) fname <- 'Comp_Cond_PoisGamma'
    if(all(model==36,likelihood==1,type==2)) fname <- 'Comp_Cond_Gauss_misp_Pois'
    if(all(model==2,likelihood==1,type==2))  fname <- 'Comp_Cond_BinomGauss'
    if(all(model==11,likelihood==1,type==2)&length(n)==1) fname <- 'Comp_Cond_BinomGauss'
    if(all(model==51,likelihood==1,type==2)&length(n)>=1) fname <- 'Comp_Cond_BinomNNGauss_misp'
    if(all(model==11,likelihood==1,type==2)&length(n)>1) fname <- 'Comp_Cond_BinomNNGauss'
    if(all(model==49,likelihood==1,type==2)&length(n)==1) fname <- 'Comp_Cond_BinomLogi'
    if(all(model==49,likelihood==1,type==2)&length(n)>1) fname <- 'Comp_Cond_BinomNNLogi'
    if(all(model==14,likelihood==1,type==2)) fname <- 'Comp_Cond_BinomnegGauss'
    if(all(model==16,likelihood==1,type==2)) fname <- 'Comp_Cond_BinomnegGauss'
    if(all(model==54,likelihood==1,type==2)) fname <- 'Comp_Cond_BinomnegBinary'
    if(all(model==43,likelihood==1,type==2)) fname <- 'Comp_Cond_PoisZIP'
    if(all(model==57,likelihood==1,type==2)) fname <- 'Comp_Cond_PoisGammaZIP'
    if(all(model==44,likelihood==1,type==2)) fname <- 'Comp_Cond_Gauss_misp_PoisZIP'
    if(all(model==45,likelihood==1,type==2)) fname <- 'Comp_Cond_BinomnegGaussZINB'
    if(all(model==56,likelihood==1,type==2)) fname <- 'Comp_Cond_BinomnegGaussZINB1'
    if(all(model==24,likelihood==1,type==2)) fname <- 'Comp_Cond_LogLogistic'
    if(all(model==25,likelihood==1,type==2)) fname <- 'Comp_Cond_Logistic'
    if(all(model==59,likelihood==1,type==2)) fname <- 'Comp_Cond_SkewLaplace'
    if(all(model==41,likelihood==1,type==2)) fname <- 'Comp_Cond_Gauss_misp_Tukeygh'
    if(all(model==37,likelihood==1,type==2)) fname <- 'Comp_Cond_Gauss_misp_SkewT'
    if(all(model==47,likelihood==1,type==2)) fname <- 'Comp_Cond_Gauss_misp_PoisGamma'
###################### pairwise ###############################################
    if(all(model==1,likelihood==3,type==2)) fname <- 'Comp_Pair_Gauss'
    if(all(model==2,likelihood==3,type==2)) fname <- 'Comp_Pair_BinomGauss'
    if(all(model==11,likelihood==3,type==2)&length(n)==1) fname <- 'Comp_Pair_BinomGauss'
    if(all(model==51,likelihood==3,type==2)&length(n)>=1) fname <- 'Comp_Pair_BinomNNGauss_misp'
    if(all(model==11,likelihood==3,type==2)&length(n)>1) fname <- 'Comp_Pair_BinomNNGauss'
    if(all(model==49,likelihood==3,type==2)&length(n)==1) fname <- 'Comp_Pair_BinomLogi'
    if(all(model==49,likelihood==3,type==2)&length(n)>1) fname <- 'Comp_Pair_BinomNNLogi'
    if(all(model==19,likelihood==3,type==2)) { namesnuis=c(namesnuis,"z")
                                              fixed<- c(fixed, list(z=min(n)))
                                              fname <- 'Comp_Pair_Binom2Gauss'}
    if(all(model==14,likelihood==3,type==2)) fname <- 'Comp_Pair_BinomnegGauss'
    if(all(model==16,likelihood==3,type==2)) fname <- 'Comp_Pair_BinomnegGauss'
    if(all(model==54,likelihood==3,type==2)) fname <- 'Comp_Pair_BinomnegBinary'
    if(all(model==15,likelihood==3,type==2)) fname <- 'Comp_Pair_PoisbinGauss'
    if(all(model==17,likelihood==3,type==2)) fname <- 'Comp_Pair_PoisbinnegGauss'
    if(all(model==13,likelihood==3,type==2)) fname <- 'Comp_Pair_WrapGauss'
    if(all(model==10,likelihood==3,type==2)) fname <- 'Comp_Pair_SkewGauss'
    if(all(model==21,likelihood==3,type==2)) fname <- 'Comp_Pair_Gamma'
    if(all(model==33,likelihood==3,type==2)) fname <- 'Comp_Pair_Kumaraswamy'
    if(all(model==42,likelihood==3,type==2)) fname <- 'Comp_Pair_Kumaraswamy2'
    if(all(model==28,likelihood==3,type==2)) fname <- 'Comp_Pair_Beta'
    if(all(model==50,likelihood==3,type==2)) fname <- 'Comp_Pair_Beta2'
    if(all(model==26,likelihood==3,type==2)) fname <- 'Comp_Pair_Weibull'
    if(all(model==24,likelihood==3,type==2)) fname <- 'Comp_Pair_LogLogistic'
    if(all(model==25,likelihood==3,type==2)) fname <- 'Comp_Pair_Logistic'
    if(all(model==59,likelihood==3,type==2)) fname <- 'Comp_Pair_SkewLaplace'
    if(all(model==23,likelihood==3,type==2)) fname <- 'Comp_Pair_2Gamma'
    if(all(model==22,likelihood==3,type==2)) fname <- 'Comp_Pair_LogGauss'
    if(all(model==18,likelihood==3,type==2)) fname <- 'Comp_Pair_SkewTGauss'
    if(all(model==27,likelihood==3,type==2)) fname <- 'Comp_Pair_TWOPIECET'
    if(all(model==39,likelihood==3,type==2)) fname <- 'Comp_Pair_TWOPIECEBIMODAL'
    if(all(model==29,likelihood==3,type==2)) fname <- 'Comp_Pair_TWOPIECEGauss'
    if(all(model==31,likelihood==3,type==2)) fname <- 'Comp_Pair_BinomTWOPIECEGauss'
    if(all(model==32,likelihood==3,type==2)) fname <- 'Comp_Pair_BinomnegTWOPIECEGauss'
    if(all(model==12,likelihood==3,type==2)) fname <- 'Comp_Pair_T'
    if(all(model==34,likelihood==3,type==2)) fname <- 'Comp_Pair_Tukeyh'
    if(all(model==40,likelihood==3,type==2)) fname <- 'Comp_Pair_Tukeyhh'
    if(all(model==41,likelihood==3,type==2)) fname <- 'Comp_Pair_Gauss_misp_Tukeygh'
    if(all(model==36,likelihood==3,type==2)) fname <- 'Comp_Pair_Gauss_misp_Pois'
    if(all(model==35,likelihood==3,type==2)) fname <- 'Comp_Pair_Gauss_misp_T'
    if(all(model==37,likelihood==3,type==2)) fname <- 'Comp_Pair_Gauss_misp_SkewT'
    if(all(model==20,likelihood==3,type==2)) fname <- 'Comp_Pair_SinhGauss'
    if(all(model==38,likelihood==3,type==2)) fname <- 'Comp_Pair_TWOPIECETukeyh'
    if(all(model==30,likelihood==3,type==2)) fname <- 'Comp_Pair_Pois'
    if(all(model==46,likelihood==3,type==2)) fname <- 'Comp_Pair_PoisGamma'
    if(all(model==43,likelihood==3,type==2)) fname <- 'Comp_Pair_PoisZIP'
    if(all(model==57,likelihood==3,type==2)) fname <- 'Comp_Pair_PoisGammaZIP'
    if(all(model==44,likelihood==3,type==2)) fname <- 'Comp_Pair_Gauss_misp_PoisZIP'
    if(all(model==45,likelihood==3,type==2)) fname <- 'Comp_Pair_BinomnegGaussZINB'
    if(all(model==56,likelihood==3,type==2)) fname <- 'Comp_Pair_BinomnegGaussZINB1'
    if(all(model==47,likelihood==3,type==2)) fname <- 'Comp_Pair_Gauss_misp_PoisGamma'
 #############################################################################
    if(sensitivity) hessian=TRUE
    if(spacetime) fname <- paste(fname,"_st",sep="")
    if(bivariate) fname <- paste(fname,"_biv",sep="")

    type_cop=0
    cond_pair=0

  if(!is.null(copula))
    {
        fname <- paste(fname,"Cop",sep="")
        if(copula=="Gaussian")      {type_cop=1; }
        if(copula=="Clayton")       {type_cop=2; }
        if(copula=="SkewGaussian")  {type_cop=3; }
        if(all(likelihood==1,type==2)) {cond_pair=1;  fname=gsub("Cond", "Pair", fname)}
        if(all(likelihood==3,type==2)) {cond_pair=0}
    }

    fname <- paste(fname,"2mem",sep="")
    if(aniso) fname <- paste(fname,"_aniso",sep="")

     if((spacetime||bivariate)&&(!spacetime_dyn))    data=c(t(data))
     if((spacetime||bivariate)&&(spacetime_dyn))     data=unlist(data)
     if(spacetime||bivariate)   NS=c(0,NS)[-(length(ns)+1)]

###### selecting data with indexes from composite likelihood

   if(is.null(neighb)) {colidx=colidx+1; rowidx=rowidx+1}
   data1=data[colidx]; data2=data[rowidx]

   ## Count observations do not change during optimization.  Validate and
   ## convert them once, then reuse the integer cache in every Poisson-Gamma
   ## pairwise likelihood evaluation (base, ZIP, and Gaussian-misspecified).
   if(!bivariate && model %in% c(46L,47L,57L)) {
     pg_cache <- .C("SetPGPairCounts",
       as.double(data1),as.double(data2),as.integer(length(data1)),
       ok=integer(1),PACKAGE="GeoModels",NAOK=TRUE)$ok
     if(pg_cache[1L] != 1L)
       stop("Poisson-Gamma data must be non-negative integer counts or NA")
   }

   ## Constant scaling of the objective.  The optimizer and numerical score
   ## use the average pairwise objective, while CompLikelihood$value is
   ## converted back to the total composite log-likelihood below.
   obj_scale <- max(1, length(data1))

   if((model==11||model==49||model==51)&&length(n)>1) {n1=n[colidx];n2=n[rowidx];n=c(n1,n2)}

   ## Quantities below are invariant across objective evaluations.
   ## Cache them once in the CompLik2 frame; comploglik2 accesses them
   ## through lexical scoping without changing its call signature.
   if(!bivariate) {
     sel_mean <- namesnuis %in% .GeoMean_names(ncol(X))
     n_col <- n[colidx]
     n_row <- n[rowidx]
   }

##################
lname="comploglik2"

coords=cbind(coordx,coordy,coordz)

## Original optimizer controls.
## Keep these unchanged to preserve the same numerical path as the previous code.
## The changes below only reclassify optimizer warnings after convergence.
opt_reltol <- 1e-12
opt_pgtol  <- 1e-10
opt_factr  <- 1e4
opt_maxit  <- 100000L

if(!onlyvar){
  ############################## spatial or space time ############################################
   if(!bivariate)           {
    if(length(param)==1) {
         optimizer="optimize"
     CompLikelihood <- .GeoModels_optimize_1d(
                              f=eval(as.name(lname)), param=param, lower=lower, upper=upper,
                              namesparam=namesparam,
                              colidx=colidx,rowidx=rowidx,corrmodel=corrmodel, coords=coords,
                              data1=data1,data2=data2, fixed=fixed, fan=fname, n=n,
                              namescorr=namescorr, namesnuis=namesnuis,namesaniso=namesaniso,
                              weigthed=weigthed,X=X, MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)}
   if(length(param)>1) {

    if(optimizer=='L-BFGS-B')
      CompLikelihood <- optim(par=param,fn=eval(as.name(lname)),
                              control=list(factr=opt_factr, pgtol=opt_pgtol, maxit=opt_maxit),
                              colidx=colidx,rowidx=rowidx,corrmodel=corrmodel, coords=coords,data1=data1,data2=data2, fixed=fixed,
                              fan=fname, lower=lower, method='L-BFGS-B',n=n,
                              namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam, namesaniso=namesaniso,
                              upper=upper,weigthed=weigthed,X=X,  hessian=FALSE,MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)

    if(optimizer=='BFGS')
        CompLikelihood <- optim(par=param, fn=eval(as.name(lname)),  colidx=colidx,rowidx=rowidx,corrmodel=corrmodel,  coords=coords,
                           control=list(reltol=opt_reltol, maxit=opt_maxit),data1=data1,data2=data2, fixed=fixed, fan=fname,
                              hessian=FALSE, method='BFGS',n=n,namescorr=namescorr,
                                  namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso,weigthed=weigthed,X=X,MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)
    if(optimizer=='SANN')
        CompLikelihood <- optim(par=param, fn=eval(as.name(lname)),  colidx=colidx,rowidx=rowidx,corrmodel=corrmodel,  coords=coords,
                           control=list(reltol=opt_reltol, maxit=opt_maxit),data1=data1,data2=data2, fixed=fixed, fan=fname,
                              hessian=FALSE, method='SANN',n=n,namescorr=namescorr,
                                  namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso,weigthed=weigthed,X=X, MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)

      if(optimizer=='Nelder-Mead')
        CompLikelihood <- optim(par=param, fn=eval(as.name(lname)),  colidx=colidx,rowidx=rowidx,corrmodel=corrmodel,  coords=coords,
          control=list(reltol=opt_reltol, maxit=opt_maxit), data1=data1,data2=data2, fixed=fixed, fan=fname,
                              hessian=FALSE, method='Nelder-Mead',n=n,namescorr=namescorr,
                                  namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso,weigthed=weigthed,X=X, MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)


    if(optimizer=='nlm')
    CompLikelihood <- nlm(f=eval(as.name(lname)),p=param,steptol = 1e-4, colidx=colidx,rowidx=rowidx,corrmodel=corrmodel,  coords=coords,data1=data1,data2=data2, fixed=fixed,
                               fan=fname,hessian=FALSE,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso,
                               iterlim=opt_maxit, weigthed=weigthed,X=X,MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)


     if(optimizer=='bobyqa')
  {
     CompLikelihood <-minqa::bobyqa(par=param, fn = get(lname),lower=lower,upper=upper,
                        control = list(maxfun=opt_maxit),
                        colidx=colidx,rowidx=rowidx,corrmodel=corrmodel, coords=coords, data1=data1,data2=data2, fixed=fixed,fan=fname,n=n,namescorr=namescorr,
                       namesnuis=namesnuis,namesparam=namesparam, namesaniso=namesaniso, weigthed=weigthed,X=X,MM=MM,
                        aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)
    }

    if(optimizer=='nlminb'){

     CompLikelihood <-nlminb(objective=eval(as.name(lname)),start=param,colidx=colidx,rowidx=rowidx,corrmodel=corrmodel, coords=coords, data1=data1,data2=data2, fixed=fixed,
                                control = list(iter.max=opt_maxit, eval.max=opt_maxit, rel.tol=opt_reltol, x.tol=1e-8),
                              lower=lower,upper=upper,hessian=FALSE,
                               fan=fname,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam, namesaniso=namesaniso,
                               weigthed=weigthed,X=X, MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)
    }
  }
}
######################################################################################
############################## bivariate  ############################################
######################################################################################
    if(bivariate)           {
     if(length(param)==1)
        {
         optimizer="optimize"
       CompLikelihood <- .GeoModels_optimize_1d(
                              f=comploglik_biv2, param=param, lower=lower, upper=upper,
                              namesparam=namesparam,
                              colidx=colidx,rowidx=rowidx,corrmodel=corrmodel, coords=coords,
                              data1=data1,data2=data2, fixed=fixed,fan=fname,n=n,
                              namescorr=namescorr, namesnuis=namesnuis,namesaniso=namesaniso,
                              weigthed=weigthed,X=X,MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)}
      if(length(param)>1) {
    if(optimizer=='L-BFGS-B'){
      CompLikelihood <- optim(param,comploglik_biv2, control=list(factr=opt_factr, pgtol=opt_pgtol, maxit=opt_maxit),
                              method='L-BFGS-B',hessian=FALSE,lower=lower, upper=upper,
                               colidx=colidx,rowidx=rowidx,corrmodel=corrmodel,  coords=coords,
                              data1=data1,data2=data2, fixed=fixed,fan=fname,n=n,
                              namescorr=namescorr, namesnuis=namesnuis, namesparam=namesparam,namesaniso=namesaniso,
                             weigthed=weigthed,X=X,MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)}

      if(optimizer=='BFGS')
      CompLikelihood <- optim(param,comploglik_biv2,  colidx=colidx,rowidx=rowidx,corrmodel=corrmodel,  coords=coords,control=list(reltol=opt_reltol, maxit=opt_maxit), data1=data1,data2=data2, fixed=fixed, fan=fname,
                              hessian=FALSE, method='BFGS',n=n,
                              namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam ,namesaniso=namesaniso,weigthed=weigthed,X=X,MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)
   if(optimizer=='Nelder-Mead')
      CompLikelihood <- optim(param,comploglik_biv2,  colidx=colidx,rowidx=rowidx,corrmodel=corrmodel,  coords=coords,control=list(reltol=opt_reltol, maxit=opt_maxit), data1=data1,data2=data2, fixed=fixed, fan=fname,
                              hessian=FALSE, method='Nelder-Mead',n=n,
                              namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam ,namesaniso=namesaniso,weigthed=weigthed,X=X,MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)
    if(optimizer=='nlm')
        CompLikelihood <- nlm( f=comploglik_biv2,p=param,  colidx=colidx,rowidx=rowidx,corrmodel=corrmodel, coords=coords, data1=data1,data2=data2, fixed=fixed,
                               fan=fname,hessian=FALSE,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam, namesaniso=namesaniso,
                               weigthed=weigthed,X=X,MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)

   if(optimizer=='bobyqa')   {
        CompLikelihood <- minqa::bobyqa(fn=comploglik_biv2,par=param,
                                     control = list(maxfun=opt_maxit),
                              lower=lower,upper=upper,
                                colidx=colidx,rowidx=rowidx,corrmodel=corrmodel,  coords=coords,data1=data1,data2=data2, fixed=fixed,
                               fan=fname,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam, namesaniso=namesaniso,
                               weigthed=weigthed,X=X,MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)
                      }
    if(optimizer=='nlminb')
        CompLikelihood <- nlminb( objective=comploglik_biv2,start=param,
                                     control = list(iter.max=opt_maxit, eval.max=opt_maxit, rel.tol=opt_reltol, x.tol=1e-8),
                              lower=lower,upper=upper, hessian=FALSE,
                                colidx=colidx,rowidx=rowidx,corrmodel=corrmodel,  coords=coords,data1=data1,data2=data2, fixed=fixed,
                               fan=fname,n=n,namescorr=namescorr, namesnuis=namesnuis,namesparam=namesparam, namesaniso=namesaniso,
                               weigthed=weigthed,X=X,MM=MM,aniso=aniso,type_cop=type_cop,cond_pair=cond_pair)


   }
 }
########################################################################################
########################################################################################
    # Check the optimisation outcome.
    # Important: all optimizers minimize -loglik.  Therefore checks for penalized
    # values must be done on the minimized objective BEFORE changing sign.

    CompLikelihood$optimizer <- optimizer
    CompLikelihood$opt_code <- NA_integer_
    CompLikelihood$opt_message <- NULL
    CompLikelihood$opt_warning <- NULL
    CompLikelihood$warning <- NULL

    penalized_minimum <- function(obj) {
      is.null(obj) || length(obj) == 0L || any(!is.finite(obj)) || any(obj >= 1.0e8)
    }

    bad_parameter <- function(par) {
      is.null(par) || length(par) == 0L || any(!is.finite(par)) || any(abs(par) >= 1.0e14)
    }

    usable_fit <- function(min_obj, par) {
      !penalized_minimum(min_obj) && !bad_parameter(par)
    }

    make_opt_warning <- function(opt_name, code, msg = NULL) {
      out <- paste0(opt_name, ' warning: code=', code)
      if(!is.null(msg) && length(msg) > 0L && !is.na(msg) && nzchar(as.character(msg))) {
        out <- paste0(out, '; ', as.character(msg))
      }
      out
    }

    set_success_or_failure <- function(obj, min_obj, par, opt_name, code, msg = NULL,
                                       success_codes = 0L, iter_codes = integer(0L)) {
      obj$opt_code <- code
      obj$opt_message <- msg

      if(usable_fit(min_obj, par)) {
        obj$convergence <- 'Successful'
        if(!(length(code) == 1L && code %in% success_codes)) {
          if(length(code) == 1L && code %in% iter_codes) {
            obj$opt_warning <- paste0(opt_name, ': iteration limit reached')
          } else {
            obj$opt_warning <- make_opt_warning(opt_name, code, msg)
          }
        }
      } else {
        obj$convergence <- 'Optimization may have failed: Try with other starting parameters'
      }
      obj
    }

    if(optimizer=='Nelder-Mead'||optimizer=='multiNelder-Mead'||optimizer=='SANN'){
        opt_code <- CompLikelihood$convergence
        opt_msg  <- if(!is.null(CompLikelihood$message)) CompLikelihood$message else NULL
        min_obj  <- CompLikelihood$value

        CompLikelihood$value = -min_obj * obj_scale
        names(CompLikelihood$par)<- namesparam

        CompLikelihood <- set_success_or_failure(CompLikelihood, min_obj, CompLikelihood$par,
                                                 optimizer, opt_code, opt_msg,
                                                 success_codes = 0L, iter_codes = 1L)

        CompLikelihood$counts=as.numeric(CompLikelihood$counts[1])
    }

    if(optimizer=='nmk'||optimizer=='nmkb'){
        opt_code <- CompLikelihood$convergence
        min_obj  <- CompLikelihood$value

        CompLikelihood$value = -min_obj * obj_scale
        names(CompLikelihood$par)<- namesparam

        CompLikelihood <- set_success_or_failure(CompLikelihood, min_obj, CompLikelihood$par,
                                                 optimizer, opt_code, NULL,
                                                 success_codes = 0L)

        CompLikelihood$counts=as.numeric(CompLikelihood$feval)
    }

    if(optimizer=='L-BFGS-B'||optimizer=='BFGS'||optimizer=='lbfgsb3c'){
        opt_code <- CompLikelihood$convergence
        opt_msg  <- if(!is.null(CompLikelihood$message)) CompLikelihood$message else NULL
        min_obj  <- CompLikelihood$value

        CompLikelihood$value = -min_obj * obj_scale
        names(CompLikelihood$par)<- namesparam

        CompLikelihood <- set_success_or_failure(CompLikelihood, min_obj, CompLikelihood$par,
                                                 optimizer, opt_code, opt_msg,
                                                 success_codes = 0L, iter_codes = 1L)

        CompLikelihood$counts=as.numeric(CompLikelihood$counts[1])
    }

    if(optimizer=='nlm'){
        opt_code <- CompLikelihood$code
        min_obj  <- CompLikelihood$minimum

        CompLikelihood$par <- CompLikelihood$estimate
        names(CompLikelihood$par)<- namesparam
        CompLikelihood$value <- -min_obj * obj_scale

        CompLikelihood <- set_success_or_failure(CompLikelihood, min_obj, CompLikelihood$par,
                                                 optimizer, opt_code, NULL,
                                                 success_codes = c(1L, 2L),
                                                 iter_codes = 4L)

        CompLikelihood$counts=as.numeric(CompLikelihood$iterations)
    }

    if(optimizer=='bobyqa'){
        opt_code <- CompLikelihood$ierr
        opt_msg  <- if(!is.null(CompLikelihood$msg)) CompLikelihood$msg else NULL
        min_obj  <- CompLikelihood$fval

        names(CompLikelihood$par)<- namesparam
        CompLikelihood$value <- -min_obj * obj_scale

        CompLikelihood <- set_success_or_failure(CompLikelihood, min_obj, CompLikelihood$par,
                                                 optimizer, opt_code, opt_msg,
                                                 success_codes = 0L)

        CompLikelihood$counts=as.numeric(CompLikelihood$feval)
    }

    if(optimizer=='nlminb'||optimizer=='multinlminb'){
        opt_code <- CompLikelihood$convergence
        opt_msg  <- CompLikelihood$message
        min_obj  <- CompLikelihood$objective

        names(CompLikelihood$par)<- namesparam
        CompLikelihood$value <- -min_obj * obj_scale

        CompLikelihood <- set_success_or_failure(CompLikelihood, min_obj, CompLikelihood$par,
                                                 optimizer, opt_code, opt_msg,
                                                 success_codes = 0L)

        CompLikelihood$counts=as.numeric(CompLikelihood$iterations)
    }

    if(optimizer=='optimize'){
        param <- CompLikelihood$minimum
        min_obj <- CompLikelihood$objective

        CompLikelihood$par <- param
        names(CompLikelihood$par)<- namesparam
        CompLikelihood$value <- -min_obj * obj_scale
        CompLikelihood$counts <- NULL

        CompLikelihood <- set_success_or_failure(CompLikelihood, min_obj, CompLikelihood$par,
                                                 optimizer, 0L, NULL,
                                                 success_codes = 0L)
    }
  } ##### end if !onlyvar
    else {
          CompLikelihood=as.list(0)
          names(CompLikelihood)="value"
          CompLikelihood$par <- param
          CompLikelihood$claic <- NULL;CompLikelihood$clbic <- NULL;
          CompLikelihood$convergence <- 'Successful'
          CompLikelihood$opt_code <- 0L
          CompLikelihood$opt_message <- NULL
          CompLikelihood$opt_warning <- NULL
          CompLikelihood$warning <- NULL
          if(!bivariate) CompLikelihood$value = - obj_scale * comploglik2(param=CompLikelihood$par ,  colidx=colidx,rowidx=rowidx,corrmodel=corrmodel,  coords=coords,
                              data1=data1,data2=data2, fixed=fixed, fan=fname,type_cop=type_cop,
                             n=n,namescorr=namescorr,namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso,weigthed=weigthed,X=X,MM=MM,aniso=aniso,cond_pair=cond_pair)
          else CompLikelihood$value = - obj_scale * comploglik_biv2(param=CompLikelihood$par ,  colidx=colidx,rowidx=rowidx,corrmodel=corrmodel,  coords=coords,
                data1=data1,data2=data2, fixed=fixed, fan=fname,type_cop=type_cop,
                             n=n,namescorr=namescorr,namesnuis=namesnuis,namesparam=namesparam,namesaniso=namesaniso,weigthed=weigthed,X=X,MM=MM,aniso=aniso,cond_pair=cond_pair)
          ## The sensitivity matrix, when requested, is computed once below
          ## using the numerically scaled parameterization.
  }

#####################################
if (sensitivity || varest)
{
  ## The objective used by the optimizer is the average negative
  ## composite log-likelihood, -CL / obj_scale.  The sensitivity matrix
  ## must be returned on the scale of the total negative composite
  ## log-likelihood.
  ##
  ## Numerical differentiation on the original parameter scale can be
  ## unstable when parameters have very different magnitudes.  We first
  ## obtain a preliminary Hessian, use its diagonal to define a local
  ## rescaling, recompute the Hessian in the scaled parameterization, and
  ## finally transform it back to the original parameter scale.  In exact
  ## arithmetic this gives the same Hessian; the change is only numerical.
  ##
  ## The scaled Hessian is the primary operational estimate.  When its
  ## unit-free diagnostics indicate weak curvature or poor conditioning, a
  ## second Richardson evaluation with a different step is computed in the
  ## same scaled parameterization.  This second matrix is diagnostic only:
  ## it never replaces or regularizes the primary sensitivity matrix.

  theta_hat <- as.numeric(CompLikelihood$par)
  names(theta_hat) <- namesparam
  ntheta <- length(theta_hat)

  if (!bivariate) {
    objective_theta <- function(theta) {
      comploglik2(
        param = theta,
        colidx = colidx,
        rowidx = rowidx,
        corrmodel = corrmodel,
        coords = coords,
        data1 = data1,
        data2 = data2,
        fixed = fixed,
        fan = fname,
        n = n,
        namescorr = namescorr,
        namesnuis = namesnuis,
        namesparam = namesparam,
        namesaniso = namesaniso,
        weigthed = weigthed,
        X = X,
        MM = MM,
        aniso = aniso,
        type_cop = type_cop,
        cond_pair = cond_pair
      )
    }
  } else {
    objective_theta <- function(theta) {
      comploglik_biv2(
        param = theta,
        colidx = colidx,
        rowidx = rowidx,
        corrmodel = corrmodel,
        coords = coords,
        data1 = data1,
        data2 = data2,
        fixed = fixed,
        fan = fname,
        n = n,
        namescorr = namescorr,
        namesnuis = namesnuis,
        namesparam = namesparam,
        namesaniso = namesaniso,
        weigthed = weigthed,
        X = X,
        MM = MM,
        aniso = aniso,
        type_cop = type_cop,
        cond_pair = cond_pair
      )
    }
  }

  symmetrize_hessian <- function(HH) {
    (HH + t(HH)) / 2
  }

  relative_matrix_difference <- function(A, B) {
    denom <- max(1, max(abs(A)), max(abs(B)))
    max(abs(A - B)) / denom
  }

  ## Preliminary Hessian of the average negative composite
  ## log-likelihood.  It is used only to select numerically appropriate
  ## perturbation scales.  Multiplying by obj_scale here would make every
  ## perturbation unnecessarily smaller by sqrt(obj_scale).
  H0_average <- numDeriv::hessian(
    func = objective_theta,
    x = theta_hat,
    method = "Richardson"
  )
  H0_average <- symmetrize_hessian(H0_average)

  if (any(!is.finite(H0_average))) {
    stop(
      "The preliminary numerical sensitivity matrix contains non-finite values.",
      call. = FALSE
    )
  }

  ## Local diagonal rescaling.  Use the estimated curvature whenever it is
  ## finite and positive.  A parameter-based fallback is used only for
  ## invalid or zero preliminary diagonal entries; no global curvature floor
  ## is imposed, because legitimate curvatures can differ by many orders of
  ## magnitude.
  hdiag <- abs(diag(H0_average))
  valid_diag <- is.finite(hdiag) & hdiag > 0

  fallback_scale <- pmax(abs(theta_hat), 1)
  scale_vec <- numeric(ntheta)
  scale_vec[valid_diag] <- 1 / sqrt(hdiag[valid_diag])
  scale_vec[!valid_diag] <- fallback_scale[!valid_diag]

  if (any(!is.finite(scale_vec)) || any(scale_vec <= 0)) {
    stop(
      "Unable to construct a finite positive parameter scaling for the sensitivity matrix.",
      call. = FALSE
    )
  }

  A <- diag(scale_vec, nrow = ntheta, ncol = ntheta)

  objective_scaled <- function(eta) {
    theta <- theta_hat + as.vector(A %*% eta)
    objective_theta(theta)
  }

  compute_scaled_hessian_average <- function(method_args = NULL) {
    hessian_args <- list(
      func = objective_scaled,
      x = rep(0, ntheta),
      method = "Richardson"
    )
    if (!is.null(method_args)) hessian_args$method.args <- method_args

    HH <- do.call(numDeriv::hessian, hessian_args)
    HH <- symmetrize_hessian(HH)

    if (any(!is.finite(HH))) {
      stop(
        "The numerical sensitivity matrix contains non-finite values.",
        call. = FALSE
      )
    }
    HH
  }

  ## Primary Hessian with respect to eta, where
  ## theta = theta_hat + A eta.  Restore the total composite-likelihood
  ## scale only after numerical differentiation.
  H_scaled_average <- compute_scaled_hessian_average()
  H_scaled <- obj_scale * H_scaled_average
  H_scaled <- symmetrize_hessian(H_scaled)

  ## Since H_eta = A H_theta A, transform back as
  ## H_theta = A^{-1} H_eta A^{-1}.
  Ainv <- diag(1 / scale_vec, nrow = ntheta, ncol = ntheta)
  CompLikelihood$hessian <- Ainv %*% H_scaled %*% Ainv
  CompLikelihood$hessian <- symmetrize_hessian(CompLikelihood$hessian)

  rownames(CompLikelihood$hessian) <- namesparam
  colnames(CompLikelihood$hessian) <- namesparam

  names(scale_vec) <- namesparam
  rownames(H_scaled_average) <- namesparam
  colnames(H_scaled_average) <- namesparam
  rownames(H_scaled) <- namesparam
  colnames(H_scaled) <- namesparam

  eig_primary <- eigen(
    H_scaled,
    symmetric = TRUE,
    only.values = TRUE
  )$values
  cond_primary <- tryCatch(
    kappa(H_scaled),
    error = function(e) Inf
  )

  max_abs_eig <- max(abs(eig_primary))
  eigen_ratio_primary <- if (is.finite(max_abs_eig) && max_abs_eig > 0) {
    min(eig_primary) / max_abs_eig
  } else {
    NA_real_
  }

  ## Adaptive diagnostic check.  Thresholds determine only whether a second
  ## numerical evaluation is useful; they never alter the returned Hessian.
  hessian_condition_threshold <- 1000
  hessian_eigenratio_threshold <- 1e-3
  hessian_relative_tolerance <- 1e-2
  hessian_check_d <- 0.05

  needs_stability_check <-
    !is.finite(cond_primary) ||
    cond_primary > hessian_condition_threshold ||
    !is.finite(eigen_ratio_primary) ||
    eigen_ratio_primary <= 0 ||
    abs(eigen_ratio_primary) < hessian_eigenratio_threshold

  hessian_check_performed <- FALSE
  hessian_stable <- TRUE
  hessian_check_error <- NULL
  H_check_scaled_average <- NULL
  H_check_scaled <- NULL
  eig_check <- rep(NA_real_, ntheta)
  cond_check <- NA_real_
  relative_difference_check <- NA_real_
  smallest_eigenvalue_sign_change <- NA

  if (isTRUE(needs_stability_check)) {
    hessian_check_performed <- TRUE

    H_check_result <- tryCatch(
      compute_scaled_hessian_average(
        method_args = list(d = hessian_check_d)
      ),
      error = function(e) e
    )

    if (inherits(H_check_result, "error")) {
      hessian_stable <- FALSE
      hessian_check_error <- conditionMessage(H_check_result)
      warning(
        paste0(
          "The adaptive sensitivity-matrix check failed: ",
          hessian_check_error,
          ". The primary scaled Richardson Hessian has been retained."
        ),
        call. = FALSE
      )
    } else {
      H_check_scaled_average <- H_check_result
      H_check_scaled <- obj_scale * H_check_scaled_average
      H_check_scaled <- symmetrize_hessian(H_check_scaled)

      eig_check <- eigen(
        H_check_scaled,
        symmetric = TRUE,
        only.values = TRUE
      )$values
      cond_check <- tryCatch(
        kappa(H_check_scaled),
        error = function(e) Inf
      )

      relative_difference_check <- relative_matrix_difference(
        H_scaled_average,
        H_check_scaled_average
      )
      smallest_eigenvalue_sign_change <-
        sign(min(eig_primary)) != sign(min(eig_check))

      hessian_stable <-
        is.finite(relative_difference_check) &&
        relative_difference_check <= hessian_relative_tolerance &&
        !isTRUE(smallest_eigenvalue_sign_change)

      rownames(H_check_scaled_average) <- namesparam
      colnames(H_check_scaled_average) <- namesparam
      rownames(H_check_scaled) <- namesparam
      colnames(H_check_scaled) <- namesparam

      if (!isTRUE(hessian_stable)) {
        warning(
          paste0(
            "The numerical sensitivity matrix is unstable across ",
            "Richardson step sizes. The primary scaled Hessian has been ",
            "retained; standard errors and composite-likelihood information ",
            "criteria should be interpreted with caution."
          ),
          call. = FALSE
        )
      } else if (min(eig_primary) <= 0 || min(eig_check) <= 0) {
        warning(
          paste0(
            "The sensitivity matrix is reproducibly non-positive-definite ",
            "under the adaptive Richardson check. The primary scaled ",
            "Hessian has been retained."
          ),
          call. = FALSE
        )
      }
    }
  }

  ## Backward-compatible attributes plus adaptive stability diagnostics.
  attr(CompLikelihood$hessian, "parameter_scaling") <- scale_vec
  attr(CompLikelihood$hessian, "scaled_hessian") <- H_scaled
  attr(CompLikelihood$hessian, "scaled_hessian_average") <- H_scaled_average
  attr(CompLikelihood$hessian, "scaled_condition") <- cond_primary
  attr(CompLikelihood$hessian, "scaled_eigenvalues") <- eig_primary
  attr(CompLikelihood$hessian, "hessian_method") <-
    "scaled_parameterization_Richardson_with_adaptive_check"
  attr(CompLikelihood$hessian, "objective_scale") <- obj_scale
  attr(CompLikelihood$hessian, "hessian_check_performed") <-
    hessian_check_performed
  attr(CompLikelihood$hessian, "hessian_stable") <- hessian_stable
  attr(CompLikelihood$hessian, "hessian_positive_definite") <-
    all(is.finite(eig_primary)) && min(eig_primary) > 0
  attr(CompLikelihood$hessian, "hessian_stability") <- list(
    condition_threshold = hessian_condition_threshold,
    eigenratio_threshold = hessian_eigenratio_threshold,
    relative_tolerance = hessian_relative_tolerance,
    check_d = hessian_check_d,
    primary_condition = cond_primary,
    primary_eigenvalues = eig_primary,
    primary_min_eigen_ratio = eigen_ratio_primary,
    check_performed = hessian_check_performed,
    check_error = hessian_check_error,
    check_condition = cond_check,
    check_eigenvalues = eig_check,
    relative_scaled_difference = relative_difference_check,
    smallest_eigenvalue_sign_change = smallest_eigenvalue_sign_change,
    stable = hessian_stable,
    check_scaled_hessian_average = H_check_scaled_average,
    check_scaled_hessian = H_check_scaled
  )
}
CompLikelihood$score=NULL
if(score)
{
  if(!bivariate)
    CompLikelihood$score=numDeriv::grad(func=comploglik2,
                                        x=as.numeric(CompLikelihood$par),
                                        method="Richardson",
                                        colidx=colidx, rowidx=rowidx,
                                        corrmodel=corrmodel, coords=coords,
                                        data1=data1, data2=data2, fixed=fixed,
                                        fan=fname, n=n, type_cop=type_cop,
                                        namescorr=namescorr, namesnuis=namesnuis,
                                        namesparam=namesparam, namesaniso=namesaniso,
                                        weigthed=weigthed, X=X, MM=MM, aniso=aniso,
                                        cond_pair=cond_pair)

  if(bivariate)
    CompLikelihood$score=numDeriv::grad(func=comploglik_biv2,
                                        x=as.numeric(CompLikelihood$par),
                                        method="Richardson",
                                        colidx=colidx, rowidx=rowidx,
                                        corrmodel=corrmodel, coords=coords,
                                        data1=data1, data2=data2, fixed=fixed,
                                        fan=fname, n=n, type_cop=type_cop,
                                        namescorr=namescorr, namesnuis=namesnuis,
                                        namesparam=namesparam, namesaniso=namesaniso,
                                        weigthed=weigthed, X=X, MM=MM, aniso=aniso,
                                        cond_pair=cond_pair)

  ## numDeriv differentiates the normalized objective -CL / obj_scale.
  ## Restore the gradient to the scale of the total negative composite
  ## log-likelihood. GeoVarest will only need to reverse its sign.
  CompLikelihood$score <- obj_scale * CompLikelihood$score
  names(CompLikelihood$score) <- namesparam
}
####################################
       serious_failure <- is.null(CompLikelihood$value) ||
                          length(CompLikelihood$value) == 0L ||
                          !is.finite(CompLikelihood$value) ||
                          CompLikelihood$value <= -1.0e14 ||
                          grepl('^Optimization may have failed', CompLikelihood$convergence)

       if(serious_failure) {
           CompLikelihood$warning <- 'Optimization may have failed: try with other starting values'
           warning(CompLikelihood$warning, call. = FALSE)
       } else {
           ## Soft optimizer warnings are kept in CompLikelihood$opt_warning,
           ## but no R warning is emitted. This avoids false alarms when the
           ## optimum is finite, non-penalized, and usable.
           if(varest){
              stop("subsambling is not working")
           }
       }
      if(hessian) CompLikelihood$sensmat=CompLikelihood$hessian
    return(CompLikelihood)
  }

####################################################
### File name: GeoSpoutlier.r
####################################################

GeoOutlier <- function(data, coordx, coordy = NULL, coordz = NULL,
                       coordt = NULL, coordx_dyn = NULL,
                       distance = "Eucl", grid = FALSE, neighb = 10,
                       alpha = 0.001, method = "Z-Median", radius = 1,
                       bivariate = FALSE, X = NULL) {
  if (!is.numeric(alpha) || length(alpha) != 1L || !is.finite(alpha) ||
      alpha <= 0 || alpha >= 1) {
    stop("alpha must be between 0 and 1", call. = FALSE)
  }
  if (!is.logical(bivariate) || length(bivariate) != 1L || is.na(bivariate))
    stop("bivariate must be logical", call. = FALSE)
  if (!is.numeric(neighb) || length(neighb) != 1L || !is.finite(neighb) ||
      neighb < 2 || neighb != as.integer(neighb)) {
    stop("neighb must be an integer >= 2", call. = FALSE)
  }
  neighb <- as.integer(neighb)
  if (!identical(method, "Z-Median")) stop("method is not valid", call. = FALSE)
  if ((is.numeric(coordt) && length(coordt) > 1L) || !is.null(coordx_dyn)) {
    stop("GeoOutlier is currently implemented only for purely spatial data.",
         call. = FALSE)
  }
  if (isTRUE(bivariate)) {
    stop("GeoOutlier is currently implemented only for univariate data.",
         call. = FALSE)
  }

  if ((is.matrix(coordx) || is.data.frame(coordx)) &&
      is.null(coordy) && is.null(coordz)) {
    coords <- as.matrix(coordx)
  } else {
    if (is.null(coordy)) {
      stop("coordy is required when coordx is not a coordinate matrix.",
           call. = FALSE)
    }
    coords <- if (is.null(coordz)) {
      if (isTRUE(grid)) as.matrix(expand.grid(coordx, coordy))
      else cbind(coordx, coordy)
    } else {
      if (isTRUE(grid)) as.matrix(expand.grid(coordx, coordy, coordz))
      else cbind(coordx, coordy, coordz)
    }
  }
  coords <- as.matrix(coords)
  storage.mode(coords) <- "double"
  .GeoValidate3D(coords, distance = distance, grid = grid,
                 context = "GeoOutlier")

  data <- as.numeric(c(t(data)))
  if (length(data) != nrow(coords)) {
    stop("The number of observations must match the number of coordinate rows.",
         call. = FALSE)
  }
  if (neighb > nrow(coords)) {
    stop("neighb cannot exceed the number of observations.", call. = FALSE)
  }
  if (!is.null(X) && nrow(as.matrix(X)) != length(data)) {
    stop("X must have one row per observation.", call. = FALSE)
  }

  distance <- gsub("[[:blank:]]", "", distance)
  coords_original <- coords
  if (distance %in% c("Geod", "Chor")) {
    if (!requireNamespace("mapproj", quietly = TRUE)) {
      stop("Package 'mapproj' is required by GeoOutlier() for geodesic or chordal distance.",
           call. = FALSE)
    }
    prj <- mapproj::mapproject(coords[, 1L], coords[, 2L],
                              projection = "sinusoidal")
    coords <- radius * cbind(prj$x, prj$y)
  } else if (!identical(distance, "Eucl")) {
    stop("distance must be one of 'Eucl', 'Geod', or 'Chor'.", call. = FALSE)
  }

  aa <- nabor::knn(coords, coords, k = neighb)
  local_values <- lapply(seq_len(nrow(coords)), function(i) {
    data[aa$nn.idx[i, -1L]]
  })
  res <- matrix(unlist(local_values, use.names = FALSE),
                ncol = neighb - 1L, byrow = TRUE)
  local_median <- apply(res, 1L, stats::median)
  deviation <- data - local_median
  sd_deviation <- stats::sd(deviation)
  if (!is.finite(sd_deviation) || sd_deviation <= 0) {
    selected <- rep(FALSE, length(deviation))
  } else {
    zscore <- (deviation - mean(deviation)) / sd_deviation
    selected <- zscore > stats::qnorm(1 - alpha / 2) |
      zscore < stats::qnorm(alpha / 2)
  }

  output_coords <- if (distance %in% c("Geod", "Chor")) coords_original else coords
  if (is.null(X)) {
    cbind(output_coords, data = data)[selected, , drop = FALSE]
  } else {
    cbind(output_coords, data = data, as.matrix(X))[selected, , drop = FALSE]
  }
}

####################################################
### File name: GeoCorrFct.r
####################################################

GeoCorrFct<- function(x,t=NULL,corrmodel, model="Gaussian",distance="Eucl",  
                                  param, radius=6371,n=1,covariance=FALSE,variogram=FALSE)

{
  
  #############################################################################################
CorrelationFct <- function(bivariate,corrmodel, lags, lagt, numlags, numlagt, mu,model, nuisance,param,N)
    {


       if(!bivariate) { 
         nn=numlags*numlagt
         #print(nn)
         p=dotCall64::.C64('VectCorrelation',SIGNATURE = c("double","integer","double","integer","integer","double",
                  "integer","double","double","double","integer"),  
    corr=dotCall64::vector_dc("double",nn),corrmodel,lags,numlags, numlagt,mu,model,nuisance,param,lagt,N,
         INTENT =c("rw",rep("r",10)),PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)

             cc=p$corr
                    }
        else    {
                nn=numlags*4
                            p=dotCall64::.C64('VectCorrelation_biv',SIGNATURE = c("double","double","integer","double","integer","integer","double",
                  "integer","double","double","double","integer"),  
    corr=dotCall64::vector_dc("double",nn),vario=double(nn),corrmodel,lags,numlags, numlagt,mu,model,nuisance,param,lagt,N,
         INTENT =c("rw",rep("r",11)),PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)

                 cc=c(p$corr,p$vario)   

                    }
        return(cc)
    }
  #############################################################################################
  #################### end internal function ##################################################
  #############################################################################################
    call <- match.call()

   # opar=par(no.readonly = TRUE)
#on.exit(par(opar))

   # if(covariance&&variogram||!covariance&&!variogram) covariance=TRUE;variogram=FALSE
 
    model <- .GeoCovCanonicalModel(model)
    if(is.null(CkCorrModel (corrmodel))) stop("The name of the correlation model  is not correct\n")
    if(is.null(CkModel(model)))   stop("The name of the  model  is not correct\n")
    if(!is.numeric(x)) stop("Distances must be numeric\n")
    if(sum(x<0)>=1) stop("Distances must be positive\n")
    if(!is.list(param)) stop("param  must be a list\n")
    .GeoValidateHoleWendlandParameters(corrmodel, param, d=2L, context="GeoCorrFct")

    spacetime<-CheckST(CkCorrModel(corrmodel))
    bivariate<-CheckBiv(CkCorrModel(corrmodel))

    supported_models <- c(
      "Gaussian","SkewGaussian","StudentT","SkewStudentT","Tukeyh","Tukeyh2",
      "Tukeygh","Kumaraswamy","Kumaraswamy2","SinhAsinh","TwoPieceStudentT",
      "TwoPieceTukeyh","TwoPieceGaussian","Gamma","Weibull","LogLogistic",
      "LogGaussian","Binary","Bernoulli","Binomial","BinomialNeg","Geom","Geometric","Poisson",
      "PoissonGamma","SkewLaplace"
    )
    if(!(model %in% supported_models))
      stop("GeoCorrFct: observed-scale correlation/covariance is not implemented for model = '",
           model, "'.", call.=FALSE)
    if(bivariate && model != "Gaussian")
      stop("GeoCorrFct: bivariate observed-scale curves are currently implemented only for model = 'Gaussian'.",
           call.=FALSE)

    if(model %in% c("Binary","Bernoulli")) {
      if(!is.numeric(n) || length(n) != 1L || !is.finite(n) || as.integer(round(n)) != 1L)
        stop("GeoCorrFct: Binary/Bernoulli is the one-trial field; n must equal 1.", call.=FALSE)
      n <- 1L
    }
    if(model == "Binomial") {
      if(!is.numeric(n) || length(n) != 1L || !is.finite(n) || n < 1 ||
         abs(n-round(n)) > sqrt(.Machine$double.eps))
        stop("GeoCorrFct: a lag-only Binomial curve requires one common positive integer n; use GeoCovmatrix for site-specific n.", call.=FALSE)
      n <- as.integer(round(n))
    }
    if(model %in% c("BinomialNeg","Geom","Geometric")) {
      if(model %in% c("Geom","Geometric")) n <- 1L
      if(!is.numeric(n) || length(n) != 1L || !is.finite(n) || n < 1 ||
         abs(n-round(n)) > sqrt(.Machine$double.eps))
        stop("GeoCorrFct: Negative-Binomial n is the common number r of successes and must be one positive integer scalar.", call.=FALSE)
      n <- as.integer(round(n))
    }

    if(!bivariate) {
      if(is.null(param$sill)) param$sill <- 1
      if(is.null(param$nugget)) param$nugget <- 0
    }

if(is.null(t)) t=0
   
mu=0;nuisance=0
mm=0
num_beta=c(1,1)

nx=length(x)
if(spacetime) nt=length(t)
else nt=1

num_betas=c(1,1)
if(!bivariate) {
  supplied_mean <- .GeoMean_supplied_names(param)
  if(length(setdiff(supplied_mean, "mean")))
    stop("GeoCorrFct() has no design matrix X; only the intercept parameter 'mean' is allowed.\n")
  if(is.null(param$mean)) param$mean <- 0
  if(length(param$mean) != 1L || !is.numeric(param$mean) || !is.finite(param$mean))
    stop("GeoCorrFct() requires a finite scalar intercept 'mean'.\n")
}

mu=if(!bivariate) as.numeric(param$mean) else 0
  ## selecting nuisance mean and correlation parameters
      if(!bivariate){
        corr_names <- CorrelationPar(CkCorrModel(corrmodel))
        parcorr <- unlist(param[corr_names], use.names=FALSE)
        storage.mode(parcorr) <- "double"

        nuis_names <- NuisParam(model,FALSE,num_betas)
        nuisance <- unlist(param[nuis_names], use.names=TRUE)
        storage.mode(nuisance) <- "double"
        sel <- names(nuisance) %in% .GeoMean_names(1L)
        mm <- as.numeric(nuisance[sel])
        nuisance <- nuisance[!sel]
        }
      if(bivariate){
        if(!covariance) param$sill_1=param$sill_2=1
        corr_names <- CorrelationPar(CkCorrModel(corrmodel))
        parcorr <- unlist(param[corr_names], use.names=FALSE)
        storage.mode(parcorr) <- "double"

        nuis_names <- NuisParam(model,bivariate,num_betas)
        nuisance <- unlist(param[nuis_names], use.names=TRUE)
        storage.mode(nuisance) <- "double"
    }

correlation <- CorrelationFct(bivariate,CkCorrModel(corrmodel), x, t, nx, nt,mu,
                                     CkModel(model), nuisance,parcorr,n)

if(is.null(t)) t=0
if(length(t)>1)  A=as.matrix(expand.grid(x,t)) 


####### Gaussian
   if(model=="Gaussian"){
        if(bivariate){  

                        cova11 <- correlation[1:nx]
                        cova12 <- correlation[(nx+1):(2*nx)]
                        cova22 <- correlation[(3*nx+1):(4*nx)]
                        variogram11  <- correlation[(4*nx+1):(5*nx)]
                        variogram12  <- correlation[(5*nx+1):(6*nx)]
                        variogram22  <- correlation[(7*nx+1):(8*nx)]
                      
                               }
        else { 
             vs=as.numeric(nuisance["sill"])

             if(length(t)>1) cova <- correlation*(1-as.numeric(nuisance["nugget"])) + as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
             else   cova <- correlation*(1-as.numeric(nuisance["nugget"]))+as.numeric(nuisance["nugget"])*I(x==0)
        }
}



if(!bivariate){
##########################
###### non Gaussian cases
##########################

   if(model=="SkewGaussian") {    
            if(bivariate) {}
              else {

               if(length(t)>1) correlation1=(1-as.numeric(nuisance['nugget']) )*correlation  + as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
               else correlation1=(1-as.numeric(nuisance['nugget']) )*correlation  +as.numeric(nuisance["nugget"])*I(x==0)

              vv=as.numeric(nuisance['sill']);sk=as.numeric(nuisance['skew']);sk2=sk^2;corr2=correlation^2;  
              cc=(2*sk2)*(sqrt(1-corr2) + correlation*asin(correlation)-1)/(pi*vv+sk2*(pi-2)) + (correlation1*vv)/(vv+sk2*(1-2/pi))
              vs=(vv+sk2*(1-2/pi))
              cova=cc; 
               }
                   }
  ##########################################
   if(model=="StudentT")        { if(bivariate) {}
                        else {
                              
                 if(length(t)>1) correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
                 else            correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(x==0)

                              nu=1/as.numeric(nuisance['df']);sill=as.numeric(nuisance['sill'])
                              vs=sill*(nu)/(nu-2)
                              cc=((nu-2)*gamma((nu-1)/2)^2*Re(hypergeo::hypergeo(0.5,0.5 ,nu/2 ,correlation^2))*correlation1)/(2*gamma(nu/2)^2)
                              
                              cova=cc;
                               }
                  }

      ##########################################
   if(model=="SkewStudentT")        { if(bivariate) {}
                        else {
                              
                 if(length(t)>1) correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
                 else            correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(x==0)

                              nu=1/as.numeric(nuisance['df']);sill=as.numeric(nuisance['sill']);sk=as.numeric(nuisance['skew'])
                              skew2=sk*sk; f=(nu-1)/2; w=sqrt(1-skew2);y=correlation1;
                              CorSkew=(2*skew2/(pi*w*w+skew2*(pi-2)))*(sqrt(1-y*y)+y*asin(y)-1)+w*w*correlation/(w*w+skew2*(1-2/pi)) ;
                              mm=sqrt(nu)*gamma(f)*sk/(sqrt(pi)*gamma(nu/2));
                cc=(pi*(nu-2)*gamma(f)^2/(2*(pi*gamma(nu/2)^2-skew2*(nu-2)*gamma(f)^2)))*(Re(hypergeo::hypergeo(0.5,0.5,nu/2,y*y))*((1-2*skew2/pi)*CorSkew+2*skew2/pi)-2*skew2/pi);
                               vs=sill*(nu/(nu-2)-mm*mm);
                              cova=cc;
                               }
                  }
##########################################

 if(model=="Tukeyh")        { if(bivariate) {}
                        else {
                             

    if(length(t)>1) correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
             else   correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(x==0)

                              h=as.numeric(nuisance['tail'])
                              sill=as.numeric(nuisance['sill'])
                              vs=  sill*(1-2*h)^(-1.5)    
                              cc=(correlation*(1-2*h)^(1.5))/((1-h)^2-(h*correlation)^2)^(1.5)
                             
                              cova=cc;
                             }
                  }
   if(model=="Tukeyh2")        { if(bivariate) {}
                        else {
                if(length(t)>1) correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
                else       correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(x==0)
                              hr=as.numeric(nuisance['tail1']); hl=as.numeric(nuisance['tail2'])
                              mom_hh <- .GeoTukeyh2StandardMoments(hr, hl)
                              cova=.GeoTukeyh2Correlation(correlation, hr, hl)
                              vs=as.numeric(nuisance['sill'])*mom_hh$variance
                             }
                  }   
##########################################
  if(model=="Tukeygh")       { if(bivariate) {}
                        else {
                             
                          if(length(t)>1)
               correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)  
                          else     correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(x==0)

                              rho=correlation
                              tail=as.numeric(nuisance['tail'])
                              sill=as.numeric(nuisance['sill'])
                              eta=as.numeric(nuisance['skew'])
                              rho2=rho*rho; eta2=eta*eta; tail2=tail*tail;

                        if(tail>1e-05&&abs(eta)>1e-05){
                               rho2=rho*rho; eta2=eta*eta; tail2=tail*tail;
                               u=1-tail; a=1+rho;
                               mu=(exp(eta2/(2*u))-1)/(eta*sqrt(u));
                               var0=(exp(2*eta2/(1-2*tail))-2*exp(eta2/(2*(1-2*tail)))+1)/(eta2*sqrt(1-2*tail))-mu*mu
                               vs=sill*var0
                               A1=exp(a*eta2/(1-tail*a));
                               A2=2*exp(0.5*eta2*  (1-tail*(1-rho2))  / (u*u- tail2*rho2)  );
                               A3=eta2*sqrt(u*u- rho2*tail*tail);
                               cc=((A1-A2+1)/A3-mu*mu)/var0
                               
                               cova=cc;
                            } 
                        if(tail<=1e-05&&abs(eta)>1e-05){
                              var0=( -exp(eta^2)+exp(eta^2*2))*eta^(-2)
                              vs=sill*var0
                              cc=(( -exp(eta^2)+exp(eta^2*(1+rho)))*eta^(-2))/var0
                              
                              cova=cc;
                            } 
                        if(tail>1e-05&&abs(eta)<=1e-05){
                              ## When skew is approximately zero, Tukey-gh reduces to Tukey-h.
                              ## This avoids the undefined object `h` previously used here and
                              ## keeps the formula consistent with the Tukeyh block above.
                              vs=sill*(1-2*tail)^(-1.5)
                              cc=(rho*(1-2*tail)^(1.5))/((1-tail)^2-(tail*rho)^2)^(1.5)
                              
                              cova=cc;
                            } 
                        if(tail<=1e-05&&abs(eta)<=1e-05){
                            cova=rho;
                            vs=sill
                            }  
                }}


if(model=="Kumaraswamy"||model=="Kumaraswamy2")  { if(bivariate) {}
else {


if(length(t)>1)
               correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0) 
else correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(x==0)

if(model=="Kumaraswamy"){
ga=as.numeric(nuisance['shape2'])
eta=as.numeric(nuisance['shape1'])
}
if(model=="Kumaraswamy2"){
ga=as.numeric(nuisance['shape'])
pmed=plogis(mu)
eta=log(0.5)/log1p(-(pmed^ga))
}
mm=eta*beta(1+1/ga,eta)
ssup=as.numeric(nuisance['max'])-as.numeric(nuisance['min'])
sill=ssup^2*(eta*beta(1+2/ga,eta)-mm^2)
###
NN=length(correlation);
res=double(NN)
bb=.C("corr_kuma_vec",as.double(correlation),as.double(eta),as.double(ga), 
     res=as.double(res),
    as.integer(NN),PACKAGE='GeoModels', NAOK=TRUE)
rho=bb$res
cova=rho;
vs=sill
 }
}

##########################


if(model == "SinhAsinh") {
  if(bivariate) {} else {
    nugget <- as.numeric(nuisance['nugget'])
    d <- as.numeric(nuisance['tail'])
    e <- as.numeric(nuisance['skew'])
    sill <- as.numeric(nuisance['sill'])
    if(d <= 0 || !is.finite(d) || !is.finite(e))
      stop("GeoCorrFct: SinhAsinh requires finite skew and a positive finite tail.", call.=FALSE)
    if(length(t)>1)
      rho_lat <- correlation*(1-nugget) + nugget*I(A[,1]==0&A[,2]==0)
    else
      rho_lat <- correlation*(1-nugget) + nugget*I(x==0)
    mom_sas <- .GeoSASStandardMoments(e, d)
    cova <- corrsas(rho_lat,e,d)
    vs <- sill*mom_sas$variance
  }
}

##########################################
 if(model=="TwoPieceStudentT")        { if(bivariate) {}
                        else {

                            if(length(t)>1)
                                   correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
                              else correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(x==0)

                              nu=1/as.numeric(nuisance['df']); sk=as.numeric(nuisance['skew']);sill=as.numeric(nuisance['sill'])
                              sk2=sk^2
                              vs=sill* ((nu/(nu-2))*(1+3*sk2) - 4*sk2*(nu/pi)*(gamma(0.5*(nu-1))/gamma(0.5*nu))^2)
                              corr2=correlation1^2;sk2=sk^2

                                  a1=Re(hypergeo::hypergeo(0.5,0.5,nu/2,corr2))
                                  a2=correlation1 *asin(correlation1) + (1-corr2)^(0.5)
                                  ll=qnorm((1-sk)/2)
                                  p11=pbivnorm::pbivnorm(ll,ll, rho = correlation, recycle = TRUE)
                                  a3=3*sk2 + 2*sk + 4*p11 - 1
                                  KK=( nu*(nu-2)*gamma((nu-1)/2)^2) / (nu*pi*gamma(nu/2)^2*(3*sk2+1)-4*sk2*nu*(nu-2)*gamma((nu-1)/2)^2 )
                                  cc= KK*(a1*a2*a3-4*sk2);
                              ##
                              cova=cc;
                               }
                  } 
  ##########################################
  if(model=="TwoPieceTukeyh")        { if(bivariate) {}
                        else {
                              

                              if(length(t)>1) correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
                              else correlation1=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(x==0)


                              tail=as.numeric(nuisance['tail']); sk=nuisance['skew'];sill=nuisance['sill']
                              corr2=correlation1^2;sk2=sk^2
                              gg2=(1-(1-corr2)*tail)^2
                              xx=corr2/gg2
                              A=(asin(sqrt(xx))*sqrt(xx)+sqrt(1-xx))/(1-xx)^(1.5)
                              ll=qnorm((1-sk)/2)
                              p11=pbivnorm::pbivnorm(ll,ll, rho = correlation, recycle = TRUE)
                              a3=3*sk2 + 2*sk + 4*p11 - 1
                              mm=8*sk2/(pi*(1-tail)^2); 
                              ff=(1+3*sk2)/(1-2*tail)^(1.5)
                              M=(2*(1-corr2)^(3/2))/(pi*gg2)
                              cc=  (M*A*a3-mm)/( ff- mm)
                              vs=sill*(ff- mm) 
                              cova=cc;
                               }
                  }  
##########################################
 if(model=="TwoPieceGaussian")        { 
                        if(bivariate) {}
                        else {        

                         if(length(t)>1) correlation1=correlation*(1-as.numeric(nuisance['nugget']))+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
                        else correlation1=correlation*(1-as.numeric(nuisance['nugget']))+as.numeric(nuisance["nugget"])*I(x==0)


                        corr2=sqrt(1-correlation1^(2))
                        sk=as.numeric(nuisance['skew']); sk2=sk^2
                        ll=qnorm((1-sk)/2)
                        p11=pbivnorm::pbivnorm(ll,ll, rho = correlation, recycle = TRUE)
                        KK=3*sk2+2*sk+ 4*p11 - 1
                        cc=(2*((corr2 + correlation1*asin(correlation1))*KK)- 8*sk2)/(3*pi*sk2  -  8*sk2   +pi   )
                        vs= as.numeric(nuisance['sill'])*(1+3*sk2-8*sk2/pi)
                        cova=cc;
                        }
                  } 
    ##########################################
 if(model=="SkewLaplace") { if(bivariate) {}
                        else {
                          if(length(t)>1) rho_lat=correlation*(1-as.numeric(nuisance['nugget']))+as.numeric(nuisance['nugget'])*I(A[,1]==0&A[,2]==0)
                          else rho_lat=correlation*(1-as.numeric(nuisance['nugget']))+as.numeric(nuisance['nugget'])*I(x==0)
                          sk=as.numeric(nuisance['skew'])
                          cova=rho_lat^2
                          vs=as.numeric(nuisance['sill'])*(1/sk^2+1/(1-sk)^2)
                        }
                  }
    ##########################################
 if(model=="Gamma")        { if(bivariate) {}
                        else {
                              
                          if(length(t)>1) correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
                          else   correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(x==0)



                              vs=2*exp(mm)^2/as.numeric(nuisance['shape'])
                              cc=correlation^2
                              cova=cc;
                            }
                  }
##########################################
if(model=="Weibull")        { if(bivariate) {} 
                        else {
                         

                        if(length(t)>1)   correlation=correlation*(1-as.numeric(nuisance['nugget'] )  )+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
                        else correlation=correlation*(1-as.numeric(nuisance['nugget'] )  )+as.numeric(nuisance["nugget"])*I(x==0)


                        sh=as.numeric(nuisance["shape"])
                        vs=exp(mm)^2*(gamma(1+2/sh)/gamma(1+1/sh)^2-1)
                        cc=.GeoWeibullCorr(correlation, sh)
                        cova=cc;
                     }
                    }
##########################################
if(model=="LogLogistic")    { if(bivariate) {}  
                      else { 


                     if(length(t)>1)  correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
                     else correlation=correlation*(1-as.numeric(nuisance['nugget'] ))+as.numeric(nuisance["nugget"])*I(x==0)


                     sh=as.numeric(nuisance["shape"])
                     vs=exp(mm)^2*(2*sh*sin(pi/sh)^2/(pi*sin(2*pi/sh))-1)
                     cc=((pi*sin(2*pi/sh))/(2*sh*(sin(pi/sh))^2-pi*sin(2*pi/sh)))*
                                    (Re(hypergeo::hypergeo(-1/sh, -1/sh, 1,correlation^2))*
                                     Re(hypergeo::hypergeo( 1/sh,  1/sh, 1,correlation^2)) -1)
                      cova=cc;
                    }
                    }
if(model=="LogGaussian")    { if(bivariate) {}  
                      else {    
                    
                    if(length(t)>1)  correlation=correlation*(1-as.numeric(nuisance["nugget"]))  +as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
                    else  correlation=correlation*(1-as.numeric(nuisance["nugget"]))   +as.numeric(nuisance["nugget"])*I(x==0)


                    vvar=as.numeric(nuisance["sill"])
                    yy=vvar*correlation
                    cc<-(exp(yy)-1)/(exp(vvar)-1)
                    vs<-exp(2*mm)*(exp(vvar)-1)
                    cova=cc; 
                     }
                  }
#############################################
if(model %in% c("Binary","Bernoulli","Binomial","BinomialNeg","Geom","Geometric"))
   {
                    if(!bivariate) {
                           pp=pnorm(mu)
                           if(model %in% c("Binary","Bernoulli")) vs=pp*(1-pp)
                           if(model=="Binomial") vs=as.numeric(n)*pp*(1-pp)
                           if(model=="BinomialNeg") vs=as.numeric(n)*(1-pp)/pp^2
                           if(model %in% c("Geom","Geometric")) vs=(1-pp)/pp^2
                           ## VectCorrelation returns the observed-scale
                           ## correlation.  Do not rely on numerical evaluation
                           ## of the bivariate Gaussian cdf at latent rho = 1:
                           ## that is a singular boundary point.  Enforce the
                           ## exact same-site endpoint directly, so that
                           ## correlation(0)=1, covariance(0)=Var(Y), and
                           ## variogram(0)=0 by construction.
                           cova=correlation
                           if(spacetime) {
                             if(length(t)>1L) zero_lag <- (A[,1] == 0 & A[,2] == 0)
                             else zero_lag <- (x == 0 & t[1L] == 0)
                           } else zero_lag <- (x == 0)
                           cova[zero_lag] <- 1
                           }
   }
if(model=="Poisson") {
                    if(bivariate) {}
                    if(!bivariate) {   
                           if(length(t)>1) correlation=(1-as.numeric(nuisance['nugget']))*correlation  +as.numeric(nuisance["nugget"])*I(A[,1]==0&A[,2]==0)
                           else correlation=(1-as.numeric(nuisance['nugget']))*correlation  +as.numeric(nuisance["nugget"])*I(x==0)
                           corr2=correlation^2    
                           vs=exp(mu);
                           z=2*vs/(1-corr2)
                           cc=corr2*(1-(besselI(z,0,expon.scaled = TRUE)+besselI(z,1,expon.scaled = TRUE)))
                           cova=cc
                         }
                      }
if(model=="PoissonGamma") {
        if(bivariate) {}
        if(!bivariate) {
          ## PEM / Poisson-Erlang mixture correlation.  Corollary 5.1 is an
          ## off-diagonal expression and contains terms involving
          ## 1-rho^2.  At the same space-time point rho=1 is a singular
          ## numerical boundary for the hypergeometric representation, while
          ## the theoretical endpoint is exactly Corr{M(s),M(s)}=1.  Evaluate
          ## the closed form only away from zero lag and impose the endpoint
          ## directly.
          if(length(t)>1L) {
            zero_lag <- (A[,1] == 0 & A[,2] == 0)
          } else {
            zero_lag <- (x == 0 & t[1L] == 0)
          }

          nug <- as.numeric(nuisance['nugget'])
          latent_corr <- (1-nug)*correlation + nug*zero_lag
          a <- as.numeric(nuisance['shape'])
          mean_pg <- exp(mu)

          ## Use the compiled stationary PEM kernel shared with covariance
          ## matrices and kriging.  Integer kappa values use terminating 2F1
          ## polynomials; half-integers/general values use the compiled
          ## evaluator rather than R-level integration/continued fractions.
          cc <- .Call("GeoCorrPGsVecCall",
                      as.double(latent_corr), as.double(mean_pg), as.double(a),
                      PACKAGE="GeoModels")
          cc[zero_lag] <- 1

          ## Var{M(s)} = mu * (1 + mu/kappa), with kappa = shape.
          vv <- mean_pg*(1+mean_pg/a)
          vs <- vv
          cova <- cc
        }
    }


  } ## end not bivariate                 
#################################################################

if(!bivariate){
   if(!covariance) vs=1
   res=cova*vs; 
   if(variogram) res=vs*(1-cova)
 }
if(bivariate)
{
 if(!variogram) res=rbind(cova11,cova12,cova22)
 else  res=rbind(variogram11,variogram12,variogram22)
}

GeoCorrFct <- list(corr=res,
                   distances=x,
                    times=t,
                    model=model,
                    distance=distance,  
                    param=param,
                    radius=radius,
                    n=n,
                    covariance=covariance,
                    variogram=variogram,
                    spacetime=spacetime,
                    bivariate=bivariate)

structure(c(GeoCorrFct, call = call), class = c("GeoCorrFct"))




}




plot.GeoCorrFct<- function(x,type="p",...)
  {

if(!inherits(x,"GeoCorrFct"))       stop("Enter an object obtained from the function GeoCorrFct\n")

opar=par(no.readonly = TRUE)
on.exit(par(opar))

#old.par <- par(no.readonly = TRUE)

space=!x$bivariate&&!x$spacetime
################################# spatial ###################################################################
if(space) {
    if(type=="p") { plot(x$distances,x$corr,xlab="Distances",type="p",axes=FALSE,...);axis(1);axis(2)}
    if(type=="l") { 
                   if(as.numeric(x$param$nugget)>0){maxcor=max(x$corr)
                                                    plot.new(); plot.window(xlim=c(0,max(x$distances)),ylim= c(0,maxcor))
                                                    if(!x$variogram) points(0,maxcor,pch=20) else points(0,0,pch=20)
                                                    sel=x$distances>0
                                                    lines(x$distances[sel],x$corr[sel],...)
                                                    axis(1);axis(2)}
                   else  {   plot(x$distances,x$corr,type="l",axes=FALSE,...)        ;axis(1);axis(2)}                    
                  } 
   }

if(x$bivariate){
##################################  bivariate ##################################################################
  if(type=="p") {
par(mfrow=c(1,3))
 plot(x$distances,x$corr[1,],xlab="Distances",main="1st component",axes=FALSE,type="p",...);axis(1);axis(2)
 plot(x$distances,x$corr[2,],xlab="Distances",main="Cross component",axes=FALSE,type="p",...);axis(1);axis(2)
 plot(x$distances,x$corr[3,],xlab="Distances",main="2nd component",axes=FALSE,type="p",...);axis(1);axis(2)
}
  if(type=="l") {
  par(mfrow=c(1,3))  
###########
          if(as.numeric(x$param$nugget_1)>0){
                                                    maxcor=max(x$corr[1,])
                                                    plot.new(); plot.window(xlim=c(0,max(x$distances)),ylim= c(0,maxcor));title(main="1st component")
                                                    if(!x$variogram) points(0,maxcor,pch=20) else points(0,0,pch=20)
                                                    sel=x$distances>0
                                                    lines(x$distances[sel],x$corr[1,][sel],...)
                                                    axis(1);axis(2)
          }
          else   { plot(x$distances,x$corr[1,],xlab="Distances",main="1st component",axes=FALSE,type="l",...);axis(1);axis(2) }
##########
          if(as.numeric(x$param$nugget_1)>0||as.numeric(x$param$nugget_2)>0) {

                                                    maxcor=max(x$corr[2,])
                                                    mincor=min(x$corr[2,])
                                                    plot.new(); plot.window(xlim=c(0,max(x$distances)),ylim= c(mincor,maxcor));title(main="Cross component")
                                                    if(!x$variogram) {
                                                    bb=as.numeric(x$param$pcol*sqrt( (x$param$nugget_1+x$param$sill_1)*(x$param$nugget_2+x$param$sill_2)))
                                                    points(0,bb,pch=20) }
                                                    else points(0,0,pch=20)
                                                    sel=x$distances>0
                                                    lines(x$distances[sel],(x$corr[2,])[sel],...)
                                                    axis(1);axis(2)
          }
          else  plot(x$distances,x$corr[2,],xlab="Distances",main="Cross component",axes=FALSE,type="l",...);axis(1);axis(2)
##########
          if(as.numeric(x$param$nugget_2)>0){

                                                    maxcor=max(x$corr[3,])
                                                    plot.new(); plot.window(xlim=c(0,max(x$distances)),ylim= c(0,maxcor));title(main="2nd component")
                                                    if(!x$variogram) points(0,maxcor,pch=20) else points(0,0,pch=20)
                                                    sel=x$distances>0
                                                    lines(x$distances[sel],x$corr[3,][sel],...)
                                                    axis(1);axis(2)
          }
          else   { plot(x$distances,x$corr[3,],xlab="Distances",main="2nd component",axes=FALSE,type="l",...);axis(1);axis(2) }
  }
}
####################################################################################################
if(x$spacetime){
par(mfrow=c(1,3))
cc=matrix(x$corr,nrow=length(x$distances),ncol=length(x$times))

persp(cc,x= x$distances,y=x$times, theta = 20, phi = 30, 
     ticktype = "detailed",zlab="",xlab="Distance",ylab="Time")
# A=as.matrix(expand.grid(x$distances,x$times)) 
#scatterplot3d::scatterplot3d(A[,1],A[,2], c(cc),
#                              type="h",highlight.3d=TRUE,cex.axis=.7,cex.lab=.7,
#                              main="",
#                              ,xlab="Distance",ylab="Time")


 if(type=="p") {  plot( x$distances,cc[,1],type="p",xlab="Spatial Distances",main="Spatial Marginal",axes=FALSE,...)    ;axis(1);axis(2)
               
                  plot( x$times,cc[1,],    type="p",xlab="Times",main="Temporal marginal",axes=FALSE,...)    ;axis(1);axis(2)
              }

 if(type=="l") { 

    if(as.numeric(x$param$nugget)>0){

        maxcor=max(cc[1,])
        plot.new(); plot.window(xlim=c(0,max(x$distances)),ylim= c(0,maxcor));title(main="Spatial Marginal")
        if(!x$variogram) points(0,maxcor,pch=20) else points(0,0,pch=20)
        sel=x$distances>0
        lines(x$distances[sel],(cc[,1])[sel],...)
         axis(1);axis(2)

        maxcor=max(cc[,1])
        plot.new(); plot.window(xlim=c(0,max(x$times)),ylim= c(0,maxcor));title(main="Temporal marginal")
        if(!x$variogram) points(0,maxcor,pch=20) else points(0,0,pch=20)
        sel=x$times>0
        lines(x$times[sel],(cc[1,])[sel],...)
        axis(1);axis(2)
     }
    else { plot( x$distances,cc[,1],type="l",xlab="Spatial Distances",main="Spatial Marginal",axes=FALSE,...)    ;axis(1);axis(2)
            plot( x$times,cc[1,],    type="l",xlab="Times",main="Temporal marginal",axes=FALSE,...)    ;axis(1);axis(2)
         }   
   }
}
################################################################################################
invisible()
}  

# Internal marginal-distribution utilities shared by diagnostics and simulation.

.GeoMarginalInvLambertH <- function(x, tail) {
  tail <- as.numeric(tail)[1L]
  if (!is.finite(tail)) stop("The tail parameter must be finite.", call. = FALSE)
  x <- as.numeric(x)
  if (abs(tail) < .Machine$double.eps) return(x)
  value <- sqrt(VGAM::lambertW(tail * x * x) / tail)
  sign(x) * value
}

.GeoMarginalTukeyTail <- function(tail, name = "tail") {
  tail <- as.numeric(tail)
  if (length(tail) != 1L || !is.finite(tail) || tail < 0 || tail >= 0.5) {
    stop(name, " must lie in [0, 0.5).", call. = FALSE)
  }
  tail
}

.GeoMarginalTukeyH <- function(z, tail) {
  z <- as.numeric(z)
  z * exp(0.5 * as.numeric(tail)[1L] * z^2)
}

.GeoMarginalTukeyH2 <- function(z, tail1, tail2) {
  z <- as.numeric(z)
  out <- numeric(length(z))
  pos <- z >= 0
  if (any(pos)) out[pos] <- .GeoMarginalTukeyH(z[pos], tail1)
  if (any(!pos)) out[!pos] <- .GeoMarginalTukeyH(z[!pos], tail2)
  out
}

.GeoMarginalQTukeyH <- function(p, tail) {
  .GeoMarginalTukeyH(stats::qnorm(p), tail)
}

.GeoMarginalQTukeyH2 <- function(p, tail1, tail2) {
  .GeoMarginalTukeyH2(stats::qnorm(p), tail1, tail2)
}

.GeoMarginalDTukeyH <- function(z, tail) {
  tail <- as.numeric(tail)[1L]
  z <- as.numeric(z)
  if (!is.finite(tail) || abs(tail) < .Machine$double.eps) {
    return(stats::dnorm(z))
  }
  b <- .GeoMarginalInvLambertH(z, tail)
  stats::dnorm(b) / (exp(0.5 * tail * b^2) * (1 + tail * b^2))
}

.GeoMarginalQSkewLaplace <- function(p, skew) {
  p <- as.numeric(p)
  skew <- as.numeric(skew)[1L]
  out <- rep(NA_real_, length(p))
  right <- p >= skew
  if (any(right)) {
    out[right] <- -(log1p(-p[right]) - log1p(-skew)) / skew
  }
  if (any(!right)) {
    out[!right] <- (log(p[!right]) - log(skew)) / (1 - skew)
  }
  out
}

.GeoMarginalPSkewLaplace <- function(x, skew) {
  x <- as.numeric(x)
  skew <- as.numeric(skew)[1L]
  out <- numeric(length(x))
  left <- x < 0
  if (any(left)) out[left] <- skew * exp((1 - skew) * x[left])
  if (any(!left)) out[!left] <- 1 - (1 - skew) * exp(-skew * x[!left])
  out
}

.GeoMarginalQTwoPieceGaussian <- function(p, skew) {
  p <- as.numeric(p); skew <- as.numeric(skew)[1L]
  out <- numeric(length(p))
  split <- 0.5 * (1 + skew)
  left <- p > 0 & p < split
  right <- p >= split & p <= 1
  if (any(left)) out[left] <- (1 + skew) * stats::qnorm(p[left] / (1 + skew))
  if (any(right)) out[right] <- (1 - skew) * stats::qnorm((p[right] - skew) / (1 - skew))
  out
}

.GeoMarginalQTwoPieceStudentT <- function(p, skew, df) {
  p <- as.numeric(p); skew <- as.numeric(skew)[1L]; df <- as.numeric(df)[1L]
  out <- numeric(length(p))
  split <- 0.5 * (1 + skew)
  left <- p > 0 & p < split
  right <- p >= split & p <= 1
  if (any(left)) out[left] <- (1 + skew) * stats::qt(p[left] / (1 + skew), df = df)
  if (any(right)) out[right] <- (1 - skew) * stats::qt((p[right] - skew) / (1 - skew), df = df)
  out
}

.GeoMarginalQTwoPieceTukeyH <- function(p, skew, tail) {
  p <- as.numeric(p); skew <- as.numeric(skew)[1L]
  out <- numeric(length(p))
  split <- 0.5 * (1 + skew)
  left <- p > 0 & p < split
  right <- p >= split & p <= 1
  if (any(left)) {
    out[left] <- (1 + skew) * .GeoMarginalQTukeyH(p[left] / (1 + skew), tail)
  }
  if (any(right)) {
    out[right] <- (1 - skew) * .GeoMarginalQTukeyH((p[right] - skew) / (1 - skew), tail)
  }
  out
}

.GeoMarginalInvTukeyH2 <- function(x, tail1, tail2) {
  x <- as.numeric(x)
  out <- numeric(length(x))
  pos <- x >= 0
  if (any(pos)) out[pos] <- .GeoMarginalInvLambertH(x[pos], tail1)
  if (any(!pos)) out[!pos] <- .GeoMarginalInvLambertH(x[!pos], tail2)
  out
}

.GeoMarginalQLogLogistic <- function(p, shape, rate = 1, scale = 1 / rate,
                                     lower.tail = TRUE, log.p = FALSE) {
  if (missing(shape)) stop("argument 'shape' is missing, with no default")
  if (shape <= 0) stop("'shape' must be positive")
  if (scale <= 0) stop("'scale' must be positive")
  if (log.p) p <- exp(p)
  if (any(p < 0 | p > 1)) stop("probabilities must be between 0 and 1")
  if (!lower.tail) p <- 1 - p
  out <- numeric(length(p))
  out[p == 0] <- 0
  out[p == 1] <- Inf
  valid <- p > 0 & p < 1
  if (any(valid)) {
    pv <- p[valid]
    out[valid] <- scale * (pv / (1 - pv))^(1 / shape)
  }
  out
}

.GeoMarginalDLogLogistic <- function(x, shape, rate = 1, scale = 1 / rate) {
  if (missing(shape)) stop("argument 'shape' is missing, with no default")
  if (shape <= 0) stop("'shape' must be positive")
  if (scale <= 0) stop("'scale' must be positive")
  out <- numeric(length(x))
  pos <- x > 0 & is.finite(x)
  z <- x[pos] / scale
  out[pos] <- (shape / scale) * z^(shape - 1) / (1 + z^shape)^2
  out
}

# Canonical aliases shared by marginal diagnostics.
.GeoMarginalModelAliases <- c(
  Gauss = "Gaussian",
  SkewGauss = "SkewGaussian",
  LogGauss = "LogGaussian",
  Loglogistic = "LogLogistic",
  TwoPieceGauss = "TwoPieceGaussian",
  tukeyh = "Tukeyh",
  tukeyh2 = "Tukeyh2",
  gamma = "Gamma",
  weibull = "Weibull"
)

.GeoMarginalCanonicalModel <- function(model) {
  model <- as.character(model)[1L]
  if (model %in% names(.GeoMarginalModelAliases)) {
    model <- unname(.GeoMarginalModelAliases[[model]])
  }
  model
}

.GeoMarginalSplitCDF <- function(x, fun_left, fun_right) {
  x <- as.numeric(x)
  out <- numeric(length(x))
  right <- x >= 0
  if (any(right)) out[right] <- fun_right(x[right])
  if (any(!right)) out[!right] <- fun_left(x[!right])
  out
}

.GeoMarginalPTukeyH <- function(x, tail) {
  stats::pnorm(.GeoMarginalInvLambertH(x, tail))
}

.GeoMarginalPTukeyH2 <- function(x, tail1, tail2) {
  stats::pnorm(.GeoMarginalInvTukeyH2(x, tail1, tail2))
}

.GeoMarginalDSkewLaplace <- function(x, skew) {
  x <- as.numeric(x)
  skew <- as.numeric(skew)[1L]
  skew * (1 - skew) * ifelse(
    x < 0,
    exp((1 - skew) * x),
    exp(-skew * x)
  )
}

.GeoMarginalPTwoPieceGaussian <- function(x, skew) {
  skew <- as.numeric(skew)[1L]
  .GeoMarginalSplitCDF(
    x,
    fun_left = function(z) (1 + skew) * stats::pnorm(z / (1 + skew)),
    fun_right = function(z) skew + (1 - skew) * stats::pnorm(z / (1 - skew))
  )
}

.GeoMarginalPTwoPieceStudentT <- function(x, skew, df) {
  skew <- as.numeric(skew)[1L]
  df <- as.numeric(df)[1L]
  .GeoMarginalSplitCDF(
    x,
    fun_left = function(z) (1 + skew) * stats::pt(z / (1 + skew), df = df),
    fun_right = function(z) skew + (1 - skew) * stats::pt(z / (1 - skew), df = df)
  )
}

.GeoMarginalPTwoPieceTukeyH <- function(x, skew, tail) {
  skew <- as.numeric(skew)[1L]
  .GeoMarginalSplitCDF(
    x,
    fun_left = function(z) (1 + skew) * .GeoMarginalPTukeyH(z / (1 + skew), tail),
    fun_right = function(z) skew + (1 - skew) * .GeoMarginalPTukeyH(z / (1 - skew), tail)
  )
}

.GeoMarginalPLogLogistic <- function(x, shape, rate = 1, scale = 1 / rate) {
  if (missing(shape)) stop("argument 'shape' is missing, with no default")
  if (length(shape) != 1L || !is.finite(shape) || shape <= 0) {
    stop("'shape' must be a positive finite scalar")
  }
  scale <- as.numeric(scale)
  if (any(scale <= 0, na.rm = TRUE)) stop("'scale' must be positive")
  z <- (x / scale)^shape
  z / (1 + z)
}

####################################################
### File name: WeightedLeastSquare.r
####################################################


### Procedures are in alphabetical order.

print.GeoWLS <- function(x, digits = max(3, getOption("digits") - 3), ...)
  {
    if(x$weighted)
      method <- 'Weighted Least Squares'
    else
      method <- 'Least Squares'

    cat('\n##############################################################')
    cat('\nResults:', method,'Fitting of Gaussian Random Fields.\n')
    cat('\nCovariance model:', x$corrmodel, '\n')
    cat('Number of spatial coordinates:', x$numcoord, '\n')
    cat('Number of dependent temporal realisations:', x$numtime, '\n')
    cat('Number of estimated parameters:', length(x$param), '\n')
    cat('The value of the', method, 'at the minimum:',-x$wls,'\n')
    cat('Number of spatial bins', length(x$bins),'\n')
    cat('Number of temporal bins', length(x$bint),'\n')
    #cat('Min and max spatial distances:', x$srange,'\n')
    #if(length(x$coordt)>1) cat('Min and max temporal interval:', x$trange,'\n')

    cat('\nEstimated parameters:\n')
    print.default(x$param, digits = digits, print.gap = 2,
                  quote = FALSE)

    cat('\n##############################################################\n')
    invisible(x)
  }



WlsStart <- function(coordx, coordy, coordz,coordt, coordx_dyn, corrmodel, data, distance, fcall, fixed, grid,
                    likelihood, maxdist, neighb,maxtime, model, n, param, 
                    parscale, paramrange, radius, start, taper, tapsep, type, varest, 
                    weighted, copula,X,memdist,nosym,p_neighb = 1, thin_method = "bernoulli")
  {
    if (!isTRUE(memdist)) {
      warning(
        "'memdist = FALSE' is deprecated and ignored; WlsStart now always precomputes the selected pair structure for GeoFit.",
        call. = FALSE
      )
    }
    memdist <- TRUE

    fit_context <- getOption("GeoModels.fit_context_cache", NULL)
    if (is.environment(fit_context) &&
        exists("wls_initparam", envir = fit_context, inherits = FALSE)) {
      initparam <- get("wls_initparam", envir = fit_context, inherits = FALSE)

      ## Reproduce the data layout prepared by StartParam on the first call.
      ## Only values change between bootstrap replications; dimensions and
      ## ordering remain fixed.  This preserves the exact input seen by the
      ## likelihood while avoiding the repeated static setup.
      data_template <- initparam$data
      if (!is.list(data_template)) {
        data <- as.vector(data)
        if (!is.null(dim(data_template)) &&
            length(data) == length(data_template)) {
          dim(data) <- dim(data_template)
        }
      }
      initparam$data <- data
      return(initparam)
    }

    # Determines the range of the parameters for a given correlation
    SetRangeParam <- function(namesparam, numparam)
    {
        low <- 0#.Machine$double.eps
        big<- Inf
        lower <- NULL
        upper <- NULL
        # Check for the param set:
        for(i in 1:numparam){ 
            if (grepl("^mean[0-9]*$", namesparam[i]) ||
                grepl("^mean_[12][0-9]*$", namesparam[i])) {
                lower <- c(lower, -big)
                upper <- c(upper, big)
            }
            if(namesparam[i]=='skew'){
                lower <- c(lower, -big)
                upper <- c(upper, big)}
            if(namesparam[i]=='tail'){
                lower <- c(lower, low)
                upper <- c(upper, if (model %in% c('Tukeyh','tukeyh')) 0.5 - sqrt(.Machine$double.eps) else big)}
            if(namesparam[i] %in% c('tail1','tail2')){
                lower <- c(lower, low)
                upper <- c(upper, 0.5 - sqrt(.Machine$double.eps))}
           if(namesparam[i]=='df'){
                lower <- c(lower, 0)
                upper <- c(upper, big)}
            if(namesparam[i]=='nugget'){
                lower <- c(lower, low)
                upper <- c(upper, big)}
             if(namesparam[i]=='nugget1'){
                lower <- c(lower, low)
                upper <- c(upper, big)}
          if(namesparam[i]=='nugget2'){
                lower <- c(lower, low)
                upper <- c(upper, big)}
            if(namesparam[i]=='power'){
                lower <- c(lower, low)
                upper <- c(upper, 2)}
            if(namesparam[i]=='power_s'){
                lower <- c(lower, low)
                upper <- c(upper, 2)}
            if(namesparam[i]=='power_t'){
                lower <- c(lower, low)
                upper <- c(upper, 2)}
            if(namesparam[i]=='power1'){
                lower <- c(lower, low)
                upper <- c(upper, 2)}
            if(namesparam[i]=='scale'||namesparam[i]=='scale_11'||namesparam[i]=='scale_12'||namesparam[i]=='scale_22'){
                lower <- c(lower, low)
                upper <- c(upper, big)}
                  if(namesparam[i]=='smooth_11'||namesparam[i]=='smooth_12'||namesparam[i]=='smooth_22'){
                lower <- c(lower, low)
                upper <- c(upper, big)}  
            if(namesparam[i]=='power_2'||namesparam[i]=='power2_11'||namesparam[i]=='power2_12'||namesparam[i]=='power2_22'){
                lower <- c(lower, low)
                upper <- c(upper, big)}       
            if(namesparam[i]=='scale_s'){
                lower <- c(lower, low)
                upper <- c(upper, big)}
            if(namesparam[i]=='scale_t'){
                lower <- c(lower, low)
                upper <- c(upper, big)}
            if(namesparam[i]=='sep'){
                lower <- c(lower, low)
                upper <- c(upper, 1)}
            if(namesparam[i]=='pcol'){
                lower <- c(lower, -1)
                upper <- c(upper,  1)}
            if(namesparam[i]=='sill'||namesparam[i]=='sill_1'||namesparam[i]=='sill_2'||namesparam[i]=='nugget_1'||namesparam[i]=='nugget2_2'){
                lower <- c(lower, low)
                upper <- c(upper, big)}
            if(namesparam[i]=='smooth'){
                lower <- c(lower, low)
                upper <- c(upper, big)}
              }
              
        return(list(lower=lower, upper=upper))
    }
    
    ### Initialization parameters:
    
    initparam <- StartParam(coordx, coordy, coordz,coordt, coordx_dyn, corrmodel, data, distance, fcall, fixed,
                           grid, likelihood, maxdist,neighb, maxtime, model, n, 
                           param, parscale, paramrange, radius,  start, taper, tapsep,
                           "GeoWLS", type,  weighted,copula, X, memdist, nosym,
                           p_neighb, thin_method)
  

    
    if(!is.null(initparam$error))     stop(initparam$error)
    if(length(coordt)>0&&is.list(X)) X=X[[1]]

    bivariate<-CheckBiv(CkCorrModel(corrmodel))
    

    if(!bivariate) {if(is.null(X))  {X=1;num_betas=1} 
                    else num_betas=ncol(X)  }

    if( bivariate) {if(is.null(X))  {X=1;num_betas=c(1,1)} 
                    else 
                       { if(is.list(X)) num_betas=c(ncol(X[[1]]),ncol(X[[2]])) 
                         else num_betas=c(ncol(X),ncol(X))
                       } 
                   } 

    
    ### Set the initial type of likelihood objects:
    initparam$type <- CkType(type)
   # if(substr(model,1,6)=='Binary'||substr(model,1,8)=='Binomial'||substr(model,1,11)=='BinomialNeg'||substr(model,1,4)=='Geom'||substr(model,1,4)=='Pois') return(initparam)
    start_user <- if(is.null(start)) numeric(0) else unlist(start)
    fixed_user <- if(is.null(fixed)) numeric(0) else unlist(fixed)

    .reconcile_mean_parameters <- function(initparam, expected,
                                           start_values, fixed_values) {
      start_names <- names(start_values)
      fixed_names <- names(fixed_values)
      if(is.null(start_names)) start_names <- character(0)
      if(is.null(fixed_names)) fixed_names <- character(0)

      overlap <- intersect(expected, intersect(start_names, fixed_names))
      if(length(overlap)) {
        stop("Mean coefficient(s) cannot be supplied in both start and fixed: ",
             paste(overlap, collapse=", "), call.=FALSE)
      }

      ## StartParam supplies data-driven defaults for the mean coefficients.
      ## They are placeholders only: explicit fixed values have precedence,
      ## followed by explicit starting values.  Missing coefficients remain
      ## estimated and use the StartParam defaults as their initial values.
      defaults <- c(initparam$param, initparam$fixed)
      par_now <- initparam$param
      fix_now <- initparam$fixed
      if(is.null(par_now)) par_now <- numeric(0)
      if(is.null(fix_now)) fix_now <- numeric(0)

      for(nm in expected) {
        par_now <- par_now[names(par_now) != nm]
        fix_now <- fix_now[names(fix_now) != nm]

        if(nm %in% fixed_names) {
          value <- as.numeric(fixed_values[[nm]])
          if(length(value) != 1L || !is.finite(value))
            stop("Fixed mean coefficient '", nm, "' must be a finite scalar.",
                 call.=FALSE)
          fix_now[nm] <- value
          initparam$flagnuis[nm] <- 0
        } else {
          if(nm %in% start_names) {
            value <- as.numeric(start_values[[nm]])
          } else if(nm %in% names(defaults)) {
            value <- as.numeric(defaults[[nm]])
          } else {
            value <- 0
          }
          if(length(value) != 1L || !is.finite(value))
            stop("Starting value for mean coefficient '", nm,
                 "' must be a finite scalar.", call.=FALSE)
          par_now[nm] <- value
          initparam$flagnuis[nm] <- 1
        }
      }

      if(length(par_now)) par_now <- par_now[sort(names(par_now))]
      initparam$param <- par_now
      initparam$fixed <- if(length(fix_now)) fix_now else NULL
      initparam$namesparam <- names(initparam$param)
      initparam$numparam <- length(initparam$param)
      initparam$namesfixed <- names(initparam$fixed)
      initparam$numfixed <- length(initparam$fixed)
      initparam
    }

    ### Parameters omitted from start use the conservative defaults created by
    ### StartParam().  Repair defaults that would otherwise lie on an invalid
    ### or numerically dangerous boundary.  Explicit user starts are never
    ### changed here.
    expected_mean <- if(!bivariate) {
      .GeoMean_names(num_betas)
    } else {
      c(.GeoMean_names(num_betas[1], "mean_1"),
        .GeoMean_names(num_betas[2], "mean_2"))
    }
    start_names_user <- names(start_user)
    if(is.null(start_names_user)) start_names_user <- character(0)

    .repair_automatic_starts <- function(initparam, start_names, mean_names,
                                         model_name, data_values) {
      par_now <- initparam$param
      if(is.null(par_now) || !length(par_now)) return(initparam)

      auto_names <- setdiff(names(par_now), c(start_names, mean_names))
      zz <- suppressWarnings(as.numeric(unlist(data_values, use.names = FALSE)))
      zz <- zz[is.finite(zz)]
      vv <- if(length(zz) > 1L) stats::var(zz) else NA_real_
      if(!is.finite(vv) || vv <= 0) vv <- 1
      rr <- if(length(zz)) range(zz) else c(0, 1)
      span <- diff(rr)
      if(!is.finite(span) || span <= 0) span <- max(1, abs(rr[1L]))
      pad <- max(sqrt(.Machine$double.eps), 0.01 * span)

      for(nm in auto_names) {
        cur <- suppressWarnings(as.numeric(par_now[[nm]]))
        if(length(cur) != 1L || !is.finite(cur)) cur <- NA_real_

        if(grepl("^nugget($|_)", nm)) cur <- 0.05
        else if(grepl("^sill($|_)", nm)) cur <- vv
        else if(grepl("^scale($|_|[st]$)", nm) || grepl("^scale_[0-9]", nm)) {
          if(!is.finite(cur) || cur <= 0) cur <- 1
        }
        else if(grepl("^smooth($|_)", nm)) cur <- 0.5
        else if(grepl("^power($|_)", nm) || grepl("^power[12]($|_)", nm)) cur <- 1
        else if(nm == "sep") cur <- 0.5
        else if(nm == "pcol") cur <- 0
        else if(grepl("^shape($|[12]$|_)", nm)) cur <- 1
        else if(grepl("^df($|_)", nm)) cur <- 0.2
        else if(grepl("^tail($|[12]$|_)", nm)) cur <- 0.05
        else if(grepl("^skew($|_)", nm)) {
          cur <- if(model_name == "SkewLaplace") 0.5 else 0
        }
        else if(nm == "pmu") cur <- 0
        else if(nm == "min") cur <- rr[1L] - pad
        else if(nm == "max") cur <- rr[2L] + pad
        else if(!is.finite(cur)) cur <- 1

        par_now[[nm]] <- cur
      }

      ## Ensure an automatically estimated bounded-support interval has
      ## strictly positive width even for constant observations.
      if(all(c("min", "max") %in% names(par_now)) &&
         !any(c("min", "max") %in% start_names)) {
        if(!is.finite(par_now[["min"]]) || !is.finite(par_now[["max"]]) ||
           par_now[["min"]] >= par_now[["max"]]) {
          par_now[["min"]] <- rr[1L] - pad
          par_now[["max"]] <- rr[2L] + pad
        }
      }

      initparam$param <- par_now
      initparam$namesparam <- names(par_now)
      initparam$numparam <- length(par_now)
      initparam
    }

    initparam <- .repair_automatic_starts(
      initparam, start_names_user, expected_mean, model, data
    )

    ###### continuous and discrete models
    if((model %in% c('Gaussian','Gauss','Chisq','LogLogistic','Logistic','SkewLaplace','Gamma','Gamma2','Beta','Beta2','LogGaussian','LogGauss','Binomial_TwoPieceGaussian','Binomial_TwoPieceGauss',
          'Tukeygh','Tukeyh','Tukeyh2','Kumaraswamy','Kumaraswamy2','Weibull','SkewGaussian','SkewGauss','SinhAsinh','StudentT','SkewStudentT',
          "Gaussian_misp_StudentT","Gaussian_misp_Poisson","Gaussian_misp_Tukeygh",
          "Binomial","Geometric","PoisBin","BinomialNeg","Binary_misp_BinomialNeg","Binomial2","PoisBinNeg","Poisson","BinomialNegZINB","BinomialLogistic","Gaussian_misp_Binomial",
          "Gaussian_misp_BinomialNeg","PoissonZIP1","BinomialNegZINB1",
          "Gaussian_misp_SkewStudentT","PoissonGamma","PoissonWeibull","Gaussian_misp_PoissonGamma","PoissonGammaZIP1",
          "TwoPieceStudentT",'Wrapped',"TwoPieceGaussian","TwoPieceGauss","TwoPieceTukeyh","TwoPieceBimodal")) &
          (type %in% c('Standard','Pairwise','Tapering','Tapering1','Independence'))) {

      if(!initparam$bivariate) {
        mean_names <- .GeoMean_names(num_betas)
        initparam <- .reconcile_mean_parameters(
          initparam, mean_names, start_user, fixed_user
        )
      } else {
        mean_names <- c(.GeoMean_names(num_betas[1], "mean_1"),
                        .GeoMean_names(num_betas[2], "mean_2"))
        initparam <- .reconcile_mean_parameters(
          initparam, mean_names, start_user, fixed_user
        )
      }

    }

    paramrange <- SetRangeParam(names(initparam$param), length(initparam$param))
    initparam$lower <- paramrange$lower
    initparam$upper <- paramrange$upper

    if (is.environment(fit_context)) {
      assign("wls_initparam", initparam, envir = fit_context)
    }
    return(initparam)
  }

####################################################################################################################
####################################################################################################################
####################################################################################################################
####################################################################################################################
GeoWLS <- function(data, coordx, coordy=NULL,coordz=NULL, coordt=NULL,  coordx_dyn=NULL, corrmodel, distance="Eucl",
                         fixed=NULL,grid=FALSE, maxdist=NULL, neighb=NULL,maxtime=NULL,
                         optimizer='Nelder-Mead', numbins=NULL, radius=1,  start=NULL,
                         weighted=FALSE,optimization=TRUE,numbins_t=NULL)
  {
    call <- match.call()
    grid_requested <- isTRUE(grid)

    ## GeoWLS is a Gaussian-only variogram least-squares procedure.  Keep the
    ## model label internal because shared setup code and GeoCovariogram use it,
    ## but do not expose a public `model` argument.
    model <- 'Gaussian'

    if (grid_requested) {
      if (!is.null(coordx_dyn)) stop("grid=TRUE cannot be combined with coordx_dyn")
      is_biv_wls <- isTRUE(CheckBiv(CkCorrModel(corrmodel)))
      gg <- .GeoVariogram_normalize_grid(
        data=data, coordx=coordx, coordy=coordy, coordz=coordz,
        coordt=coordt, bivariate=is_biv_wls
      )
      data <- gg$data
      coordx <- gg$coordx
      coordy <- gg$coordy
      coordz <- gg$coordz
      grid <- FALSE
    }

    if (!is.null(maxdist)) {
      if (!is.numeric(maxdist) || any(!is.finite(maxdist)) || any(maxdist <= 0))
        stop("maxdist must contain positive finite numeric values")
    }
    .GeoVariogram_validate_data(data, context="GeoWLS")

    ### Check the parameters given in input:
    checkinput <- CkInput(coordx, coordy, coordz,coordt, coordx_dyn, corrmodel, data, distance,"Fitting", fixed, grid, 'None',
                              maxdist, maxtime,  model,NULL,  optimizer, NULL, radius, start, NULL,
                             NULL, 'GeoWLS', FALSE, weighted, NULL,NULL)
    

    if(!is.null(checkinput$error))
      stop(checkinput$error)
    # Check the number of spatial breakpoints (legacy GeoModels meaning:
    # numbins breakpoints define numbins - 1 empirical spatial bins).
    if(is.null(numbins)) numbins <- 13L
    if(!is.numeric(numbins) || length(numbins)!=1L || !is.finite(numbins) ||
       numbins < 2 || numbins %% 1 != 0)
      stop('numbins must be an integer >= 2\n')
    numbins <- as.integer(numbins)
    if(!is.null(numbins_t)) {
      if(!is.numeric(numbins_t) || length(numbins_t)!=1L || !is.finite(numbins_t) ||
         numbins_t < 1 || numbins_t %% 1 != 0)
        stop('numbins_t must be a positive integer\n')
      numbins_t <- as.integer(numbins_t)
    }
    ### Define the object function for the weighted least squares method:
    WLsquare <- function(param,bins, bint, corrmodel, fixed, fun, lenbins, moments,
                         namescorr, namesnuis, numbins, numbint)
      {
        param <- c(param, fixed)#set the parameters set:
        paramcorr <- param[namescorr]#set the correlation parameters:
        nuisance <- param[namesnuis]#set the nuisance parameters:

     if(fun=='WLeastSquare_G')
        result <- .C('WLeastSquare_G', as.double(bins), as.double(bint), as.integer(corrmodel),
                     as.double(lenbins), as.double(moments), as.integer(numbins),
                     as.integer(numbint), as.double(nuisance), as.double(paramcorr),
                     res=double(1), PACKAGE='GeoModels', NAOK=TRUE)$res
     if(fun=='LeastSquare_G')
        result <- .C('LeastSquare_G', as.double(bins), as.double(bint), as.integer(corrmodel),
                     as.double(lenbins), as.double(moments), as.integer(numbins),
                     as.integer(numbint), as.double(nuisance), as.double(paramcorr),
                     res=double(1), PACKAGE='GeoModels', NAOK=TRUE)$res
       return(result)
      }
    ### Initializes global variables:
    GeoWLS <- NULL
    fname <- NULL
    variogramt <- NULL
    variogramst <- NULL
    ### Initializes the parameter values:
    parscale <- NULL
    initparam <- StartParam(coordx, coordy, coordz,coordt, coordx_dyn,corrmodel, data, distance, "Fitting", fixed, grid,
    'None', maxdist, neighb,maxtime,  model, NULL,  NULL,parscale, TRUE, radius, start,NULL,  NULL,
                           'GeoWLS', 'GeoWLS', FALSE, NULL, NULL, FALSE, FALSE,
                           1, "bernoulli")


    if(!is.null(initparam$error))
      stop(initparam$error)
     coordx=initparam$coordx
     coordy=initparam$coordy
     coordz=initparam$coordz
     coordt=initparam$coordt
     on.exit({
       try(.C('DeleteGlobalVar', PACKAGE='GeoModels', NAOK=TRUE), silent=TRUE)
     }, add=TRUE)
     ## Native variogram routines always index the third coordinate.  Use an
     ## explicit zero vector in 2D instead of passing NULL or a scalar zero.
     if(is.null(coordz)) coordz <- rep(0, length(coordx))
    ###### ----------- START Estimation of the empirical variogram ---------- #####
    numvario <- numbins-1
    bins <- double(numbins) # vector of spatial bins
    moments <- double(numvario) # vector of spatial moments
    lenbins <- integer(numvario) # vector of spatial bin sizes
    ### Checks the type of variogram:
    fname <- 'Binned_Variogram'


    if(initparam$spacetime){### Computes the spatial-temporal variogram:
      spacetime_dyn <- !is.null(coordx_dyn)
      ns <- as.integer(initparam$ns)
      NS <- as.integer(c(0L, cumsum(ns))[-(length(ns) + 1L)])

      ti <- .GeoVariogram_time_bins(
        coordt, maxtime,
        numbins_t = if (is.null(numbins_t)) numbins - 1L else numbins_t
      )
      time_breaks <- ti$breaks
      temporal_centers <- ti$centers
      ntbin <- length(time_breaks) - 1L

      bint_native <- as.double(time_breaks)
      momentt <- double(ntbin)
      lenbint_native <- integer(ntbin)
      momentst <- double(numvario * ntbin)
      lenbinst_native <- integer(numvario * ntbin)

      ## The native routines expect observations in time-block order,
      ## consistently with the coordinates expanded by StartParam().
      if(!spacetime_dyn) data_native <- c(t(data))
      else data_native <- unlist(data, use.names = FALSE)

      if(spacetime_dyn) {
        EV=.C("Binned_Variogram_st2_dyn", bins=bins, bint=bint_native,
             as.double(coordx),as.double(coordy),as.double(coordz),
             as.double(coordt),as.double(data_native),
             lenbins=lenbins,lenbinst=lenbinst_native,lenbint=lenbint_native,
             moments=moments, momentst=momentst, momentt=momentt,
             as.integer(numbins), as.integer(length(time_breaks)),
             as.integer(ns),as.integer(NS), PACKAGE='GeoModels', NAOK=TRUE)
      } else {
        EV=.C("Binned_Variogram_st2", bins=bins, bint=bint_native,
             as.double(coordx),as.double(coordy),as.double(coordz),
             as.double(coordt),as.double(data_native),
             lenbins=lenbins,lenbinst=lenbinst_native,lenbint=lenbint_native,
             moments=moments, momentst=momentst, momentt=momentt,
             as.integer(numbins), as.integer(length(time_breaks)),
             as.integer(ns),as.integer(NS), PACKAGE='GeoModels', NAOK=TRUE)
      }

      bins_native <- EV$bins
      spatial_count <- EV$lenbins
      spatial_sum <- EV$moments
      temporal_count <- EV$lenbint
      temporal_sum <- EV$momentt
      st_count <- matrix(EV$lenbinst, nrow=numvario, ncol=ntbin, byrow=TRUE)
      st_sum <- matrix(EV$momentst, nrow=numvario, ncol=ntbin, byrow=TRUE)

      centers <- bins_native[-length(bins_native)] + diff(bins_native)/2
      variograms <- rep(NA_real_, numvario)
      ii <- spatial_count > 0L
      variograms[ii] <- spatial_sum[ii] / spatial_count[ii]

      variogramt <- rep(NA_real_, ntbin)
      ii <- temporal_count > 0L
      variogramt[ii] <- temporal_sum[ii] / temporal_count[ii]

      st_mean <- matrix(NA_real_, nrow=numvario, ncol=ntbin)
      ii <- st_count > 0L
      st_mean[ii] <- st_sum[ii] / st_count[ii]
      variogramst <- as.numeric(as.vector(t(st_mean)))

      ## Build the rectangular WLS system without dropping empty cells.  The
      ## C objective now skips zero-count cells, preserving the lag geometry.
      moment_full <- rbind(temporal_sum, st_sum)
      moment_full <- cbind(c(0, spatial_sum), moment_full)
      count_full <- rbind(temporal_count, st_count)
      count_full <- cbind(c(1L, spatial_count), count_full)

      bins <- c(-bins_native[1L], bins_native)
      bint <- c(0, temporal_centers)
      moments <- moment_full
      lenbins <- count_full
      numbins <- length(bins)
      numbint <- length(bint)
    }
    else{### Computes the spatial variogram:
      numbint <- 1L
      bint <- 0
      EV=.C("Binned_Variogram2", bins=bins, as.double(coordx),as.double(coordy),
         as.double(coordz),as.double(coordt),as.double(initparam$data), lenbins=lenbins,
         moments=moments, as.integer(numbins),PACKAGE='GeoModels', NAOK=TRUE)
      bins <- EV$bins
      lenbins <- EV$lenbins
      moments <- EV$moments

      centers <- bins[-length(bins)] + diff(bins)/2
      variograms <- rep(NA_real_, length(lenbins))
      ii <- lenbins > 0L
      variograms[ii] <- moments[ii] / lenbins[ii]

      if(!is.null(initparam$param['scale']) && any(is.finite(variograms))) {
        initparam$param['scale'] <- centers[which.max(replace(variograms, !is.finite(variograms), -Inf))]
      }
    }
    ###### ----------- END Estimation of the empirical variogram ---------- #####



    ###### ---------- START model fitting by weighted least squares method ----------######
    #if(model=='Gaussian'||model=='Gauss') # Gaussian random field:
      if(weighted) fname <- 'WLeastSquare_G'
      else fname <- 'LeastSquare_G'


if(optimization){
    ### Computes estimates by the weighted least squares method:
    if(optimizer=='L-BFGS-B')
      fitted <- optim(initparam$param, WLsquare, bins=bins, bint=bint, corrmodel=initparam$corrmodel,
                      fixed=initparam$fixed, fun=fname, lenbins=lenbins, method=optimizer,
                      moments=moments, namescorr=initparam$namescorr, namesnuis=initparam$namesnuis,
                      numbins=numbins, numbint=numbint, control=list(fnscale=-1, factr=1, pgtol=1e-14,
                      maxit = 1e8), lower=initparam$lower, upper=initparam$upper, hessian=FALSE)
    else
      fitted <- optim(initparam$param, WLsquare, bins=bins, bint=bint, corrmodel=initparam$corrmodel,
                      fixed=initparam$fixed, fun=fname, lenbins=lenbins, method=optimizer,
                      moments=moments, namescorr=initparam$namescorr, namesnuis=initparam$namesnuis,
                      numbins=numbins, numbint=numbint, control=list(fnscale=-1, reltol=1e-14, maxit=1e8),
                      hessian=FALSE)
     }  
     else{ res=WLsquare(unlist(start),bins=bins, bint=bint, corrmodel=initparam$corrmodel,
                      fixed=initparam$fixed, fun=fname, lenbins=lenbins, 
                      moments=moments, namescorr=initparam$namescorr, namesnuis=initparam$namesnuis,
                      numbins=numbins, numbint=numbint)
              fitted=list(convergence=NULL,counts=NULL,message=NULL,par=unlist(start),value=res)

            }


    ###### ---------- END model fitting by weighted least squares method ----------######
    ### Set the output:
    ## Keep the structural metadata required by downstream methods such as
    ## GeoCovariogram().  These fields describe the fitted Gaussian field;
    ## they do not re-introduce a user-selectable model argument.
    ##
    ## StartParam() represents the default intercept-only design by the scalar
    ## value 1.  Preserve the same meaning but store an explicit n x 1 matrix
    ## in GeoWLS objects.  Mean utilities intentionally reject ambiguous
    ## vectors, so this canonical representation is important for downstream
    ## methods (notably GeoCovariogram).
    X_out <- initparam$X
    if(!initparam$bivariate) {
      nobs_out <- length(unlist(initparam$data, use.names=FALSE))
      if(is.null(X_out) ||
         (is.atomic(X_out) && is.null(dim(X_out)) && length(X_out)==1L &&
          length(initparam$num_betas)==1L && initparam$num_betas==1L)) {
        X_out <- matrix(1, nrow=nobs_out, ncol=1L)
      } else {
        X_out <- .GeoMean_design(X_out, n=nobs_out, p=initparam$num_betas,
                                 name="GeoWLS$X")
      }
    }

    GeoWLS <- list(      bins=bins,
                         bint=bint,
                         centers=centers,
                         bivariate=initparam$bivariate,
                         spacetime=initparam$spacetime,
                         coordx = initparam$coordx,
                         coordy = initparam$coordy,
                         coordz = initparam$coordz,
                         coordt = initparam$coordt,
                         coordx_dyn=coordx_dyn,
                         convergence = fitted$convergence,
                         corrmodel = corrmodel,
                         copula=NULL,
                         data = initparam$data,
                         distance=initparam$distance,
                         fixed = initparam$fixed,
                         grid = grid_requested,
                         iterations = fitted$counts,
                         maxdist =maxdist,
                         maxtime = maxtime,
                         numbins_t = numbins_t,
                         message = fitted$message,
                         model=model,
                         ## `n` is irrelevant for continuous Gaussian fields,
                         ## but shared covariance utilities pass it to the native
                         ## correlation engine.  Use the same harmless scalar
                         ## default as GeoCorrFct() instead of returning NULL.
                         n=1L,
                         numbetas=initparam$num_betas,
                         numcoord=initparam$numcoord,
                         numtime=initparam$numtime,
                         param = fitted$par,
                         radius=initparam$radius,
                         X=X_out,
                         variograms = variograms,
                         variogramt = variogramt,
                         variogramst = variogramst,
                         weighted = weighted,
                         wls = fitted$value)
    structure(c(GeoWLS, call = call), class = c("GeoWLS"))
  }

####################################################
### File name: GeoQQ.r (cleaned; comments in English)
####################################################
GeoQQ <- function(fit, type="Q", add=FALSE, ylim=c(0,1), xlim=NULL, breaks=10, ...)
{
  ###### utilities ##########
  # Log-logistic quantile with shape (>0) and scale (>0)

  # Log-logistic density (fix: density, not quantile)
  ##################################

  if(!inherits(fit,"GeoFit"))  stop("A GeoFit object is needed as input\n")
  if(identical(fit$residual_type, "Pearson")) {
    stop("GeoQQ() is not defined for Pearson-residual objects. Use GeoQQ() or GeoPit() on the original discrete GeoFit object, and use GeoResiduals() output for residual maps or variograms.",
         call. = FALSE)
  }
  if(type!="Q" && type!="D") stop("Type can be Q or D \n")

  model <- fit$model
  discrete_models <- c(
    "Binary", "Bernoulli", "Binomial", "BinomialLogistic",
    "Geom", "Geometric", "BinomialNeg", "Poisson", "PoissonGamma",
    "PoissonZIP", "PoissonZIP1", "PoissonGammaZIP", "PoissonGammaZIP1",
    "BinomialNegZINB", "BinomialNegZINB1"
  )

  pp_tmp <- c(unlist(fit$param), unlist(fit$fixed))
  external_mean <- !is.null(.GeoMean_external(fit$fixed))
  non_intercept_X <- FALSE
  if (!is.null(fit$X)) {
    Xqq <- .GeoMean_design(
      fit$X, n = length(unlist(fit$data, use.names = FALSE)), name = "X"
    )
    non_intercept_X <- ncol(Xqq) != 1L ||
      any(abs(Xqq[, 1L] - 1) > sqrt(.Machine$double.eps))
  }
  if (external_mean || non_intercept_X) {
    if (model %in% discrete_models) {
      stop(
        "GeoQQ() does not define a single theoretical marginal distribution for discrete fits with a nonconstant mean. Use GeoPit(fit) for site-specific marginal diagnostics.",
        call. = FALSE
      )
    }
    stop("GeoQQ() requires an intercept-only marginal fit. ",
         "Please use GeoResiduals(fit) before calling GeoQQ().")
  }

  .GeoQQ_constant_n <- function(n, model_name) {
    nn <- as.numeric(n)
    if (!length(nn) || any(!is.finite(nn)) || any(nn <= 0) || any(nn != round(nn))) {
      stop("For model '", model_name, "', n must contain positive integers.", call. = FALSE)
    }
    if (length(nn) > 1L && any(nn != nn[1L])) {
      stop(
        "GeoQQ() requires a constant n for model '", model_name,
        "'. Use GeoPit(fit) when n is site-specific.",
        call. = FALSE
      )
    }
    nn[1L]
  }

  if (model %in% c("Binomial", "BinomialLogistic", "BinomialNeg")) {
    fit$n <- .GeoQQ_constant_n(fit$n, model)
  }

  copula <- if(!is.null(fit$copula)) fit$copula else NULL

  fit$param <- unlist(fit$param); fit$fixed <- unlist(fit$fixed)
  pp <- c(fit$param, fit$fixed)

  # Safe extraction: single-bracket returns NA when name is absent (no error)
  MM <- as.numeric(pp["mean"])
  VV <- as.numeric(pp["sill"])
  if (is.na(MM)) MM <- 0


  # Which models actually require 'sill' (marginal variance)?
  .requires_VV <- function(m) {
    m %in% c(
      "Gaussian","Gaussian_misp_Binomial","Gaussian_misp_Poisson","Gaussian_misp_BinomialNeg",
      "SkewGaussian","SkewLaplace",
      "StudentT","Gaussian_misp_StudentT",
      "SkewStudentT","Gaussian_misp_SkewStudentT",
      "LogGaussian","Logistic",
      "SinhAsinh","Tukeyh","Tukeyh2",
      "Tukeygh","Gaussian_misp_Tukeygh",
      "TwoPieceGaussian","TwoPieceStudentT","TwoPieceTukeyh"
    )
  }

  # Save and restore par settings
  opar <- par(no.readonly = TRUE); on.exit(par(opar))

  # Extract main from ... and remove it to avoid conflicts
  dots <- list(...)
  main_title <- if("main" %in% names(dots)) dots$main else NULL
  if("main" %in% names(dots)) dots <- dots[names(dots) != "main"]
  
  # Helper to get main title (use default if main_title is NULL)
  .get_main <- function(default) {
    if(is.null(main_title)) default else main_title
  }

  # xlim helpers
  has_xlim <- !is.null(xlim)
  .XL <- function(def) { if (has_xlim) xlim else def }
  .GRID <- function(def_range, n = 512) {
    xr <- if (has_xlim) xlim else def_range
    seq(xr[1], xr[2], length.out = n)
  }

  # Library checks for skew/t families when needed
  need_sn   <- model %in% c("SkewGaussian","SkewStudentT","Gaussian_misp_SkewStudentT")
  need_VGAM <- model %in% c("Tukeyh","Tukeyh2","TwoPieceTukeyh","Tukeygh","Gaussian_misp_Tukeygh")
  if (need_sn && !requireNamespace("sn", quietly = TRUE))
    stop("Package 'sn' is required for skew-* models.")
  if (need_VGAM && !requireNamespace("VGAM", quietly = TRUE))
    stop("Package 'VGAM' is required for Tukey-h/gh models.")

  ########################################
  #### QQ plot
  ########################################
  if(type=="Q") {
    xlab <- "Theoretical Quantiles"
    ylab <- "Sample Quantiles"

    if(!fit$bivariate){
      dd <- if (is.list(fit$coordx_dyn)) unlist(fit$data) else c(t(fit$data))
      dd <- dd[is.finite(dd)]
      if (!length(dd)) stop("empty data")

      N <- length(dd)
      probabilities  <- (1:N)/(N+1)
      probabilities1 <- c(0.25,0.75)
      q_e  <- quantile(dd, probabilities)
      q_e1 <- quantile(dd, probabilities1)

      if (.requires_VV(model) && is.na(VV)) {
        stop("'sill' parameter is required for model '", model,
             "' but is missing in 'fit'. Provide 'sill' in fit$param/fixed.")
      }

      # ---- Models ----
      if(model %in% c("Beta2")){
        mm <- 1/(1+exp(-MM)); sh <- as.numeric(pp["shape"])
        pmin <- as.numeric(pp["min"]); pmax <- as.numeric(pp["max"])
        q_t  <- pmin + (pmax-pmin)*qbeta(probabilities,  shape1=mm*sh, shape2=(1-mm)*sh)
        q_t1 <- pmin + (pmax-pmin)*qbeta(probabilities1, shape1=mm*sh, shape2=(1-mm)*sh)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Beta qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("Kumaraswamy2")){
        mm <- 1/(1+exp(-MM)); sh <- as.numeric(pp["shape"])
        pmin <- as.numeric(pp["min"]); pmax <- as.numeric(pp["max"])
        shape2 <- log(0.5)/log1p(-(mm^sh))
        q_t  <- pmin + (pmax-pmin)*(1 - (1 - probabilities)^(1/shape2))^(1/sh)
        q_t1 <- pmin + (pmax-pmin)*(1 - (1 - probabilities1)^(1/shape2))^(1/sh)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Kumaraswamy qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("Gaussian","Gaussian_misp_Binomial","Gaussian_misp_Poisson","Gaussian_misp_BinomialNeg")) {
        q_t  <- qnorm(probabilities,  mean=MM, sd=sqrt(VV))
        q_t1 <- qnorm(probabilities1, mean=MM, sd=sqrt(VV))
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Gaussian qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("Binomial")) {
        q_t  <- qbinom(probabilities,  size=fit$n, prob=pnorm(pp["mean"]))
        q_t1 <- qbinom(probabilities1, size=fit$n, prob=pnorm(pp["mean"]))
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Binomial qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("BinomialLogistic")) {
        q_t  <- qbinom(probabilities,  size=fit$n, prob=plogis(MM))
        q_t1 <- qbinom(probabilities1, size=fit$n, prob=plogis(MM))
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Binomial-Logistic qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("BinomialNeg")) {
        q_t  <- qnbinom(probabilities,  size=fit$n, prob=pnorm(pp["mean"]))
        q_t1 <- qnbinom(probabilities1, size=fit$n, prob=pnorm(pp["mean"]))
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Binomial Neg qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("Poisson")) {
        q_t  <- qpois(probabilities,  lambda=exp(pp["mean"]))
        q_t1 <- qpois(probabilities1, lambda=exp(pp["mean"]))
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Poisson qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("PoissonGamma")) {
        ff <- exp(pp["mean"])
        sh <- as.numeric(pp["shape"])
        q_t  <- qnbinom(probabilities,  size=sh, prob=sh/(ff+sh))
        q_t1 <- qnbinom(probabilities1, size=sh, prob=sh/(ff+sh))
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Poisson qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("SkewGaussian")) {
        omega <- as.numeric(sqrt((pp["skew"]^2 + pp["sill"])/pp["sill"]))
        alpha <- as.numeric(pp["skew"]/sqrt(pp["sill"]))
        q_t  <- MM + sqrt(VV)*sn::qsn(probabilities,  xi=0, omega=omega, alpha=alpha)
        q_t1 <- MM + sqrt(VV)*sn::qsn(probabilities1, xi=0, omega=omega, alpha=alpha)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Skew Gaussian qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model%in%c("StudentT","Gaussian_misp_StudentT")){
        df <- as.numeric(1/pp["df"])
        q_t  <- MM + sqrt(VV)*qt(probabilities,  df=df)
        q_t1 <- MM + sqrt(VV)*qt(probabilities1, df=df)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("t qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model%in%c("SkewStudentT","Gaussian_misp_SkewStudentT")){
        alpha <- as.numeric(pp["skew"])
        nu    <- as.numeric(1/pp["df"])
        q_t  <- MM + sqrt(VV)*sn::qst(probabilities,  xi=0, omega=1, alpha=alpha, nu=nu)
        q_t1 <- MM + sqrt(VV)*sn::qst(probabilities1, xi=0, omega=1, alpha=alpha, nu=nu)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("skewt qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("Weibull")){
        shape <- pp["shape"]
        q_t  <- exp(MM)*qweibull(probabilities,  shape=shape, scale=1/(gamma(1+1/shape)))
        q_t1 <- exp(MM)*qweibull(probabilities1, shape=shape, scale=1/(gamma(1+1/shape)))
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Weibull qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("Gamma")){
        shape <- pp["shape"]
        q_t  <- exp(MM)*qgamma(probabilities,  shape=shape/2, rate=shape/2)
        q_t1 <- exp(MM)*qgamma(probabilities1, shape=shape/2, rate=shape/2)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Gamma qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("LogGaussian")){
        q_t  <- qlnorm(probabilities,  meanlog = MM - VV/2, sdlog = sqrt(VV))
        q_t1 <- qlnorm(probabilities1, meanlog = MM - VV/2, sdlog = sqrt(VV))
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("LogGaussian qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("LogLogistic")){
        shape <- pp["shape"]
        cc <- gamma(1+1/shape)*gamma(1-1/shape)
        q_t  <- .GeoMarginalQLogLogistic(probabilities,  shape=shape, scale=exp(MM)/cc)
        q_t1 <- .GeoMarginalQLogLogistic(probabilities1, shape=shape, scale=exp(MM)/cc)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("LogLogistic qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("Logistic")){
        q_t  <- qlogis(probabilities,  location=MM, scale=sqrt(VV))
        q_t1 <- qlogis(probabilities1, location=MM, scale=sqrt(VV))
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Logistic qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("SinhAsinh")){
        tail <- as.numeric(pp["tail"]); skew <- as.numeric(pp["skew"])
        q_t  <- MM + sqrt(VV)*sinh( (asinh(qnorm(probabilities))/tail) + skew/tail )
        q_t1 <- MM + sqrt(VV)*sinh( (asinh(qnorm(probabilities1))/tail) + skew/tail )
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Sas qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("Tukeyh")){
        tail <- as.numeric(pp["tail"])
        uu <- qnorm(probabilities); uu1 <- qnorm(probabilities1)
        q_t  <- MM + sqrt(VV)*uu * exp(0.5*tail*uu^2)
        q_t1 <- MM + sqrt(VV)*uu1* exp(0.5*tail*uu1^2)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Tukey-h qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("Tukeyh2")){
        tail1 <- as.numeric(pp["tail1"]); tail2 <- as.numeric(pp["tail2"])
        q_t  <- MM + sqrt(VV)*.GeoMarginalQTukeyH2(probabilities, tail1, tail2)
        q_t1 <- MM + sqrt(VV)*.GeoMarginalQTukeyH2(probabilities1, tail1, tail2)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Tukey-hh qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("SkewLaplace")){
        skew <- as.numeric(pp["skew"])
        q_t  <- MM + sqrt(VV)*.GeoMarginalQSkewLaplace(probabilities, skew)
        q_t1 <- MM + sqrt(VV)*.GeoMarginalQSkewLaplace(probabilities1, skew)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Skew Laplace qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("Tukeygh","Gaussian_misp_Tukeygh")){
        tail <- as.numeric(pp["tail"]); skew <- as.numeric(pp["skew"])
        uu <- qnorm(probabilities); uu1 <- qnorm(probabilities1)
        if (abs(skew) < 1e-8) {
          q_t  <- MM + sqrt(VV) * (uu  * exp(0.5*tail*uu^2))
          q_t1 <- MM + sqrt(VV) * (uu1 * exp(0.5*tail*uu1^2))
        } else {
          q_t  <- MM + sqrt(VV) * ((exp(skew*uu)  - 1) * exp(0.5*tail*uu^2)  / skew)
          q_t1 <- MM + sqrt(VV) * ((exp(skew*uu1) - 1) * exp(0.5*tail*uu1^2) / skew)
        }
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Tukey-gh qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("TwoPieceGaussian")){
        skew <- as.numeric(pp["skew"])
        q_t  <- MM + sqrt(VV)*.GeoMarginalQTwoPieceGaussian(probabilities, skew)
        q_t1 <- MM + sqrt(VV)*.GeoMarginalQTwoPieceGaussian(probabilities1, skew)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Two-Piece Gaussian qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("TwoPieceStudentT")){
        skew <- as.numeric(pp["skew"]); df <- 1/as.numeric(pp["df"])
        q_t  <- MM + sqrt(VV)*.GeoMarginalQTwoPieceStudentT(probabilities, skew, df)
        q_t1 <- MM + sqrt(VV)*.GeoMarginalQTwoPieceStudentT(probabilities1, skew, df)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Two-Piece Student qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      if(model %in% c("TwoPieceTukeyh")){
        skew <- as.numeric(pp["skew"]); tail <- as.numeric(pp["tail"])
        q_t  <- MM + sqrt(VV)*.GeoMarginalQTwoPieceTukeyH(probabilities, skew, tail)
        q_t1 <- MM + sqrt(VV)*.GeoMarginalQTwoPieceTukeyH(probabilities1, skew, tail)
        do.call(plot, c(list(x=q_t, y=q_e, main=.get_main("Two-Piece Tukey-h qq-plot"), 
                             xlab=xlab, ylab=ylab, xlim=.XL(range(q_t, finite=TRUE))), dots))
      }

      x <- q_t1; y <- q_e1
      slope <- diff(y)/diff(x); int <- y[1L] - slope * x[1L]
      abline(int, slope, pch=20)
    }

    if(fit$bivariate){
      par(mfrow=c(1,2))
      if(is.list(fit$coordx_dyn)){ dd1 <- fit$data[[1]]; dd2 <- fit$data[[2]]}
      else  {dd1 <- fit$data[1,]; dd2 <- fit$data[2,];}
      dd1 <- dd1[is.finite(dd1)]; dd2 <- dd2[is.finite(dd2)]

      if(model %in% c("Gaussian")) {
        if (has_xlim) {
          qqnorm(dd1, main=.get_main("First Gaussian qq-plot"), xlim=xlim); abline(0,1)
          qqnorm(dd2, main=.get_main("Second Gaussian qq-plot"), xlim=xlim); abline(0,1)
        } else {
          qqnorm(dd1, main=.get_main("First Gaussian qq-plot")); abline(0,1)
          qqnorm(dd2, main=.get_main("Second Gaussian qq-plot")); abline(0,1)
        }
      }

      if(model %in% c("SkewGaussian")){
        omega1 <- sqrt((pp["skew_1"]^2 + pp["sill_1"])/pp["sill_1"])
        alpha1 <- pp["skew_1"]/sqrt(pp["sill_1"])
        probabilities <- (1:length(dd1))/(length(dd1)+1)
        q_t1 <- sn::qsn(probabilities, xi=0, omega=as.numeric(omega1), alpha=as.numeric(alpha1))
        q_e1 <- quantile(dd1, probabilities)
        do.call(plot, c(list(x=q_t1, y=q_e1, main=.get_main("First Skew Gaussian qq-plot"),
                             xlim=.XL(range(q_t1, finite=TRUE))), dots))
        aa <- lm(q_e1~1+q_t1); abline(as.numeric(aa$coefficients[1]),as.numeric(aa$coefficients[2]))

        omega2 <- sqrt((pp["skew_2"]^2 + pp["sill_2"])/pp["sill_2"])
        alpha2 <- pp["skew_2"]/sqrt(pp["sill_2"])
        probabilities <- (1:length(dd2))/(length(dd2)+1)
        q_t2 <- sn::qsn(probabilities, xi=0, omega=as.numeric(omega2), alpha=as.numeric(alpha2))
        q_e2 <- quantile(dd2, probabilities)
        do.call(plot, c(list(x=q_t2, y=q_e2, main=.get_main("Second Skew Gaussian qq-plot"),
                             xlim=.XL(range(q_t2, finite=TRUE))), dots))
        aa <- lm(q_e2~1+q_t2); abline(as.numeric(aa$coefficients[1]),as.numeric(aa$coefficients[2]))
      }
    }
  }

  ##########################################################
  #### Density / histogram
  ##########################################################
  if(type=="D") {
    if(!fit$bivariate){

      dd <- if (is.list(fit$coordx_dyn)) unlist(fit$data) else c(t(fit$data))
      dd <- dd[is.finite(dd)]
      if (!length(dd)) stop("empty data")

      if (.requires_VV(model) && is.na(VV)) {
        stop("'sill' parameter is required for model '", model,
             "' but is missing in 'fit'. Provide 'sill' in fit$param/fixed.")
      }

      .hist_once <- function(main_title_default, xr_default) {
        if(!add) {
          do.call(hist, c(list(x=dd, freq=FALSE, xlim=.XL(xr_default), ylab="Density", xlab="",
                               main=.get_main(main_title_default), ylim=ylim, breaks=breaks), dots))
        }
      }

      # ------------ Continuous models ------------
      if(model %in% c("Beta2")){
        mm <- 1/(1+exp(-MM)); sh <- as.numeric(pp["shape"])
        pmin <- as.numeric(pp["min"]); pmax <- as.numeric(pp["max"])
        xr <- .XL(c(pmin, pmax))
        ll <- .GRID(xr, n = 800)
        ds <- ifelse(ll>=pmin & ll<=pmax,
                     dbeta((ll-pmin)/(pmax-pmin), shape1=mm*sh, shape2=(1-mm)*sh)/(pmax-pmin), 0)
        .hist_once("Beta Histogram", xr_default = c(pmin,pmax))
        lines(ll, ds, ...)
      }

      if(model %in% c("Kumaraswamy2")){
        sh <- as.numeric(pp["shape"])
        pmin <- as.numeric(pp["min"]); pmax <- as.numeric(pp["max"])
        xr <- .XL(c(pmin,pmax))
        ll <- .GRID(xr, n = 800)
        dkuma <- function(x,MM,sh,pmin,pmax){
          out <- numeric(length(x))
          idx <- which(x>=pmin & x<=pmax)
          if (length(idx)) {
            q <- (x[idx]-pmin)/(pmax-pmin)
            m1 <- 1/(1+exp(-MM))
            shapei <- log(0.5)/log1p(-m1^sh)
            out[idx] <- shapei*sh*q^(sh-1)*(1-q^sh)^(shapei-1)/(pmax-pmin)
          }
          out
        }
        ds <- dkuma(ll, MM, sh, pmin, pmax)
        .hist_once("Kumaraswamy Histogram", xr_default = c(pmin,pmax))
        lines(ll, ds, ...)
      }

      if(model%in%c("Gaussian")){
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        .hist_once("Gaussian Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, dnorm(ll, mean=MM, sd=sqrt(VV)), ...)
      }

      if(model%in%c("StudentT","Gaussian_misp_StudentT")){
        df <- as.numeric(1/pp["df"])
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        .hist_once("Student T Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, dt((ll - MM)/sqrt(VV), df = df) / sqrt(VV), ...)
      }

      if(model%in%c("SkewStudentT","Gaussian_misp_SkewStudentT")){
        alpha <- as.numeric(pp["skew"]); nu <- as.numeric(1/pp["df"])
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        d_st <- sn::dst((ll-MM)/sqrt(VV), xi=0, omega=1, alpha=alpha, nu=nu)/sqrt(VV)
        .hist_once("Skew-T Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, d_st, ...)
      }

      if(model %in% c("SkewGaussian")){
        skew <- as.numeric(pp["skew"]); sill <- as.numeric(pp["sill"])
        omega <- sqrt((skew^2 + sill)/sill); alpha <- skew/sqrt(sill)
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        d_sn <- sn::dsn((ll-MM)/sqrt(VV), xi=0, omega=omega, alpha=alpha)/sqrt(VV)
        .hist_once("Skew Gaussian Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, d_sn, ...)
      }

      if(model %in% c("Weibull")){
        shape <- pp["shape"]
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        d_w <- ifelse(ll>0, dweibull(ll, shape=shape, scale=exp(MM)/(gamma(1+1/shape))), 0)
        .hist_once("Weibull Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, d_w, ...)
      }

     if(model %in% c("Gamma")){
       shape <- as.numeric(pp["shape"])
       xr <- .XL(range(dd, finite = TRUE))
       xr[1] <- max(0, xr[1])
       ll <- seq(from = max(xr[1], .Machine$double.eps), to = xr[2], length.out = 512)
       k    <- shape/2; rate <- shape/(2*exp(MM))  # coerente con qq
       d_g <- dgamma(ll, shape = k, rate = rate)
       .hist_once("Gamma Histogram", xr_default = xr)
       lines(ll, d_g, ...)
}


      if(model %in% c("LogGaussian")){
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        d_l <- ifelse(ll>0, dlnorm(ll, meanlog = MM - VV/2, sdlog = sqrt(VV)), 0)
        .hist_once("LogGaussian Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, d_l, ...)
      }

      if(model %in% c("Logistic")){
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        d_l <- dlogis(ll, location = MM, scale = sqrt(VV))
        .hist_once("Logistic Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, d_l, ...)
      }

      if(model %in% c("LogLogistic")){
        shape <- pp["shape"]
        cc <- gamma(1+1/shape)*gamma(1-1/shape)
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        d_l <- .GeoMarginalDLogLogistic(ll, shape = shape, scale = exp(MM)/cc)
        .hist_once("LogLogistic Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, d_l, ...)
      }

      if(model %in% c("SinhAsinh")){
        tail <- as.numeric(pp["tail"]); skew <- as.numeric(pp["skew"])
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        dsas <- function(z, skew, tail){
          s <- sinh(tail*asinh(z) - skew)
          a <- (2*pi*(1+z^2))^(-0.5)
          c0 <- sqrt(1+s^2)
          d0 <- exp(-0.5*s^2)
          tail*c0*d0*a
        }
        ds <- dsas((ll-MM)/sqrt(VV), skew, tail)/sqrt(VV)
        .hist_once("SAS Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, ds, ...)
      }

      if(model %in% c("SkewLaplace")){
        skew <- as.numeric(pp["skew"])
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        dslaplace <- function(x, sk) {
          if (sk <= 0 || sk >= 1) stop("sk debe estar en (0,1)")
          f <- numeric(length(x))
          sel1 <- x < 0
          f[sel1] <- sk * (1 - sk) * exp((1 - sk) * x[sel1])
          sel2 <- !sel1
          f[sel2] <- sk * (1 - sk) * exp(-sk * x[sel2])
          return(f)
        }
        ds <- dslaplace((ll-MM)/sqrt(VV), skew)/sqrt(VV)
        .hist_once("Skew Laplace Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, ds, ...)
      }

      if(model %in% c("Tukeyh")){
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        ds <- .GeoMarginalDTukeyH((ll-MM)/sqrt(VV), as.numeric(pp["tail"]))/sqrt(VV)
        .hist_once("Tukey-h Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, ds, ...)
      }

      if(model %in% c("Tukeyh2")){
        tail1 <- as.numeric(pp["tail1"]); tail2 <- as.numeric(pp["tail2"])
        xr <- .XL(range(dd, finite=TRUE))
        ll <- .GRID(xr)
        zs <- (ll-MM)/sqrt(VV)
        ds <- numeric(length(zs))
        idx1 <- which(zs>=0); idx2 <- which(zs<0)
        ds[idx1] <- .GeoMarginalDTukeyH(zs[idx1], tail1)
        ds[idx2] <- .GeoMarginalDTukeyH(zs[idx2], tail2)
        ds <- ds / sqrt(VV)
        .hist_once("Tukey-hh Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, ds, ...)
      }

      if(model %in% c("TwoPieceGaussian")){
        skew <- as.numeric(pp["skew"])
        xr <- .XL(range(dd, finite=TRUE)); ll <- .GRID(xr)
        z <- (ll-MM)/sqrt(VV)
        ds <- numeric(length(z))
        idx1 <- which(z>=0); idx2 <- which(z<0)
        ds[idx1] <- dnorm(z[idx1]/(1-skew),0,1) / sqrt(VV)
        ds[idx2] <- dnorm(z[idx2]/(1+skew),0,1) / sqrt(VV)
        .hist_once("Two-Piece Gaussian Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, ds, ...)
      }

      if(model %in% c("TwoPieceStudentT")){
        skew <- as.numeric(pp["skew"]); df <- 1/as.numeric(pp["df"])
        xr <- .XL(range(dd, finite=TRUE)); ll <- .GRID(xr)
        z <- (ll-MM)/sqrt(VV)
        ds <- numeric(length(z))
        idx1 <- which(z>=0); idx2 <- which(z<0)
        ds[idx1] <- dt(z[idx1]/(1-skew), df=df) / sqrt(VV)
        ds[idx2] <- dt(z[idx2]/(1+skew), df=df) / sqrt(VV)
        .hist_once("Two-Piece Student Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, ds, ...)
      }

      if(model %in% c("TwoPieceTukeyh")){
        skew <- as.numeric(pp["skew"]); tail <- as.numeric(pp["tail"])
        xr <- .XL(range(dd, finite=TRUE)); ll <- .GRID(xr)
        z <- (ll-MM)/sqrt(VV)
        ds <- numeric(length(z))
        idx1 <- which(z>=0); idx2 <- which(z<0)
        ds[idx1] <- .GeoMarginalDTukeyH(z[idx1]/(1-skew), tail) / sqrt(VV)
        ds[idx2] <- .GeoMarginalDTukeyH(z[idx2]/(1+skew), tail) / sqrt(VV)
        .hist_once("Two-Piece Tukey-h Histogram", xr_default = range(dd, finite=TRUE))
        lines(ll, ds, ...)
      }

      # ------------ Discrete models ------------
      .disc_support <- function() {
        if (has_xlim) {
          ys <- seq(ceiling(xlim[1]), floor(xlim[2]), by=1)
          if (length(ys) == 0) ys <- sort(unique(as.numeric(dd)))
        } else {
          ys <- sort(unique(as.numeric(dd)))
        }
        ys
      }

      if(model %in% c("BinomialNeg")) {
        count <- as.numeric(table(dd)); ll_emp <- count/sum(count)
        y_emp <- sort(unique(as.numeric(dd)))
        ys <- .disc_support()
        ds <- dnbinom(ys, size=fit$n, prob=pnorm(MM))
        if(!add) {
          do.call(plot, c(list(x=y_emp, y=ll_emp, type="h", col="blue", 
                               main=.get_main("Binomial Negative Histogram"),
                               ylab="Density", xlab="", xlim=.XL(range(c(y_emp, ys))), 
                               ylim=ylim, lwd=2), dots))
        }
        points(ys, ds, type="p", lwd=3)
        lines(ys, ds)
      }

      if(model %in% c("Binomial")) {
        count <- as.numeric(table(dd)); ll_emp <- count/sum(count)
        y_emp <- sort(unique(as.numeric(dd)))
        ys <- .disc_support()
        ds <- dbinom(ys, size=fit$n, prob=pnorm(MM))
        if(!add) {
          do.call(plot, c(list(x=y_emp, y=ll_emp, type="h", col="blue", 
                               main=.get_main("Binomial Histogram"),
                               ylab="Density", xlab="", xlim=.XL(range(c(y_emp, ys))), 
                               ylim=ylim, lwd=2), dots))
        }
        points(ys, ds, type="p", lwd=3)
        lines(ys, ds)
      }

      if(model %in% c("BinomialLogistic")) {
        count <- as.numeric(table(dd)); ll_emp <- count/sum(count)
        y_emp <- sort(unique(as.numeric(dd)))
        ys <- .disc_support()
        ds <- dbinom(ys, size=fit$n, prob=plogis(MM))
        if(!add) {
          do.call(plot, c(list(x=y_emp, y=ll_emp, type="h", col="blue", 
                               main=.get_main("Binomial-Logistic Histogram"),
                               ylab="Density", xlab="", xlim=.XL(range(c(y_emp, ys))), 
                               ylim=ylim, lwd=2), dots))
        }
        points(ys, ds, type="p", lwd=3)
        lines(ys, ds)
      }

      if(model %in% c("Poisson")) {
        count <- as.numeric(table(dd)); ll_emp <- count/sum(count)
        y_emp <- sort(unique(as.numeric(dd)))
        ys <- .disc_support()
        ds <- dpois(ys, lambda=exp(MM))
        if(!add) {
          do.call(plot, c(list(x=y_emp, y=ll_emp, type="h", col="blue", 
                               main=.get_main("Poisson Histogram"),
                               ylab="Density", xlab="", xlim=.XL(range(c(y_emp, ys))), 
                               ylim=ylim, lwd=2), dots))
        }
        points(ys, ds, type="p", lwd=3)
        lines(ys, ds)
      }

      if(model %in% c("PoissonGamma")) {
        count <- as.numeric(table(dd)); ll_emp <- count/sum(count)
        y_emp <- sort(unique(as.numeric(dd)))
        ys <- .disc_support()
        ff <- exp(MM); sh <- as.numeric(pp["shape"])
        ds <- dnbinom(ys, size=sh, prob=sh/(ff+sh))
        if(!add) {
          do.call(plot, c(list(x=y_emp, y=ll_emp, type="h", col="blue", 
                               main=.get_main("Poisson Gamma (NegBin) Histogram"),
                               ylab="Density", xlab="", xlim=.XL(range(c(y_emp, ys))), 
                               ylim=ylim, lwd=2), dots))
        }
        points(ys, ds, type="p", lwd=3)
        lines(ys, ds)
      }
    }
  }

  invisible()
}

GeoCovmatrix <- function(estobj=NULL,coordx=NULL, coordy=NULL, coordz=NULL, coordt=NULL,coordx_dyn=NULL,corrmodel, distance="Eucl", grid=FALSE,
                     model="Gaussian", n=1, param, anisopars=NULL,radius=1,
                       sparse=FALSE,copula=NULL,X=NULL,spobj=NULL,check.duplicates=FALSE)

{
  call <- match.call()

  ########################################################################################################
  ##########  Internal function: computing covariance matrix #############################################
  ########################################################################################################




    Cmatrix <- function(bivariate, coordx, coordy,coordz, coordt,corrmodel, dime, n, ns, NS, nuisance, numpairs,
                           numpairstot, model, paramcorr, setup, radius, spacetime, spacetime_dyn,type,copula,ML,other_nuis)
    {
###################################################################################
############### computing correlation #############################################
###################################################################################

if(model %in% c(1,9,34,12,20,18,39,27,38,29,21,26,24,10,22,40,28,33,42,59,13))
{

  if(type=="Standard") {
    if(spacetime) 
       cr=dotCall64::.C64('CorrelationMat_st_dyn2',SIGNATURE = c(rep("double",5),"integer","double","double","double","integer","integer"),
            corr=dotCall64::vector_dc("double",numpairstot),coordx,coordy,coordz,coordt,
            corrmodel,nuisance,paramcorr,radius,ns,NS,
            INTENT = c("w",rep("r",10)),
            PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
     if(bivariate) 
       cr=dotCall64::.C64('CorrelationMat_biv_dyn2',SIGNATURE = c(rep("double",5),"integer","double","double","double","integer","integer"),
            corr=dotCall64::vector_dc("double",numpairstot),coordx,coordy,coordz,coordt,
            corrmodel,nuisance,paramcorr,radius,ns,NS,
            INTENT = c("w",rep("r",10)),
            PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
   if(!bivariate&&!spacetime)  
        cr=dotCall64::.C64('CorrelationMat2',SIGNATURE = c(rep("double",5),"integer","double","double","double","integer","integer"),
            corr=dotCall64::vector_dc("double",numpairstot),coordx,coordy,coordz,coordt,
            corrmodel,nuisance,paramcorr,radius,ns,NS,
            INTENT = c("w",rep("r",10)),
            PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
    
}
###############################################################
if(type=="Tapering")  {
        fname <- 'CorrelationMat_tap'

        if(spacetime) fname <- 'CorrelationMat_st_tap'
       if(bivariate) fname <- 'CorrelationMat_biv_tap'
#############
cr=dotCall64::.C64(fname,SIGNATURE = c("double","double","double","double","double","integer","double","double","double","integer","integer"),
     corr=dotCall64::vector_dc("double",numpairs), coordx,coordy,coordz,coordt,corrmodel,nuisance, paramcorr,radius,ns,NS,
 INTENT = c("w","r","r","r","r","r","r","r", "r", "r", "r"),
            PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
#############
     ## deleting correlation equual  to 1 because there are problems  with hipergeometric function
        sel=(abs(cr$corr-1)<.Machine$double.eps);cr$corr[sel]=0
      }
    }  
    

###################################################################################
###################################################################################

if(model==1)   ## gaussian case
{
  if(!bivariate)
     {
      corr=cr$corr*(1-as.numeric(nuisance['nugget']))
      vv=nuisance['sill']
     }
if(bivariate){}
}
###############################################################
if(model==9)  {  ## TukeyGH

if(!bivariate)   {
          corr=cr$corr*(1-as.numeric(nuisance['nugget']))
          h=as.numeric(nuisance['tail'])
          g=as.numeric(nuisance['skew'])
          ss=as.numeric(nuisance['sill'])

  if(abs(g)<=1e-05 && h<=1e-05) { vv=ss }
  if(abs(g)>1e-05&&h<=1e-05) {  #
              aa=( -exp(g^2)+exp(g^2*2))*g^(-2)
              corr <- (( -exp(g^2)+exp(g^2*(1+corr)))*g^(-2))/aa
              vv=  aa*ss
              }
   if(abs(g)<=1e-05&& h>1e-05){ ##
              aa=(1-2*h)^(1.5) # variance
              corr <- aa*corr/( (1-h)^2-h^2*corr^2 )^(1.5)
              vv=aa*ss
               }
  if(h>1e-05&&abs(g)>1e-05){ # ok
                  rho=corr; rho2=rho*rho;
                  tail=h; eta=g
                  eta2=eta*eta; tail2=tail*tail;
                  u=1-tail;a=1+rho;
                  A1=exp(a*eta2/(1-tail*a));
                  A2=2*exp(0.5*eta2*  (1-tail*(1-rho2))  / (u*u- tail2*rho2)  );
                  A3=eta2*sqrt(u*u- rho2*tail2)
                  mu=(exp(eta2/(2*u))-1)/(eta*sqrt(u));
                  cova=ss*((A1-A2+1)/A3-mu*mu)
                  vari=ss*((exp(2*eta2/(1-2*tail))-2*exp(eta2/(2*(1-2*tail)))+1)/(eta2*
                           sqrt(1-2*tail))-mu*mu);
                  corr <-cova/vari
                  vv=vari
            }
}
if(bivariate){}
}
###############################################################
if(model==40)  {  ## TukeyH2

if(!bivariate)    {
    corr_lat=cr$corr*(1-as.numeric(nuisance['nugget']))
    hr=as.numeric(nuisance['tail1'])
    hl=as.numeric(nuisance['tail2'])
    mom_hh <- .GeoTukeyh2StandardMoments(hr, hl)
    corr <- .GeoTukeyh2Correlation(corr_lat, hr, hl)
    vv=as.numeric(nuisance['sill'])*mom_hh$variance
       }
if(bivariate){}
}

###############################################################
if(model==34)  {  ## TukeyH

if(!bivariate)    {
       corr=cr$corr*(1-as.numeric(nuisance['nugget']))
    h=nuisance['tail']
   if(!h)     vv=as.numeric(nuisance['sill'])
   if(h){     aa=(1-2*h)^(-1.5) # variance
              #corr <- aa*corr/( (1-h)^2-h^2*corr^2 )^(1.5)
              #corr <- (-corr/((1+h*(corr-1))*(-1+h+h*corr)*(1+h*(-2+h-h*corr^2))^0.5))/aa
              corr =(corr*(1-2*h)^(1.5))/((1-h)^2-(h*corr)^2)^(1.5)
              vv=aa*as.numeric(nuisance['sill'])
              }
       }
if(bivariate){}
}
######################################################################
if(model==12)   ##  student case
    {
if(!bivariate){
   rho_base=cr$corr
   rho_num=rho_base*(1-as.numeric(nuisance['nugget']))
   nu=as.numeric(1/nuisance['df'])
   # In the constructive Student-t RF the common scale-mixture component
   # depends on the nugget-free latent correlation, while the Gaussian
   # numerator carries the nugget attenuation.
   if(nu<170) corr=((nu-2)*gamma((nu-1)/2)^2*
                    Re(hypergeo::hypergeo(0.5,0.5,nu/2,rho_base^2))*rho_num)/
                   (2*gamma(nu/2)^2)
   vv=as.numeric(nuisance['sill'])*(nu)/(nu-2)
}
if(bivariate){}
}

###############################################################
  if(model==18)   ##  skew student case
    {
     if(!bivariate)
{
    corr=cr$corr*(1-as.numeric(nuisance['nugget']))
    nu=as.numeric(1/nuisance['df']); sk=as.numeric(nuisance['skew'])
    skew2=sk*sk;l=nu/2; f=(nu-1)/2; w=sqrt(1-skew2);y=corr;
    CorSkew=(2*skew2/(pi*w*w+skew2*(pi-2)))*(sqrt(1-y*y)+y*asin(y)-1)+w*w*cr$corr/(w*w+skew2*(1-2/pi)) ;
    mm=sqrt(nu)*gamma(f)*sk/(sqrt(pi)*gamma(l));
    corr=(pi*(nu-2)*gamma(f)^2/(2*(pi*gamma(l)^2-skew2*(nu-2)*gamma(f)^2)))*(Re(hypergeo::hypergeo(0.5,0.5,l,y*y))
                                *((1-2*skew2/pi)*CorSkew+2*skew2/pi)-2*skew2/pi);
    vv=as.numeric(nuisance['sill'])*(nu/(nu-2)-mm*mm);
}
if(bivariate){}
}
###############################################################

if (model == 20) {  ## SAS
  nugget <- as.numeric(nuisance['nugget'])
  d <- as.numeric(nuisance['tail'])
  e <- as.numeric(nuisance['skew'])
  sill <- as.numeric(nuisance['sill'])
  mom_sas <- .GeoSASStandardMoments(e, d)
  corr_base <- cr$corr * (1 - nugget)
  corr <- corrsas(corr_base, e, d)
  vv <- sill * mom_sas$variance
}


###############################################################
 if(model==39)   ##  two piece bimodal case
    {
if(!bivariate)
{
 corr=cr$corr*(1-as.numeric(nuisance['nugget']))
 nu=as.numeric(nuisance['df']); sk=as.numeric(nuisance['skew'])
 delta=as.numeric(nuisance['shape'])
 alpha=2*(delta+1)/nu
 nn=2^(1-alpha/2)
 ll=qnorm((1-sk)/2)
 p11=pbivnorm::pbivnorm(ll,ll, rho = cr$corr, recycle = TRUE)
 corr2=corr^2;sk2=sk^2
 a1=Re(hypergeo::hypergeo(-1/alpha ,-1/alpha,nu/2,corr2))
 a3=3*sk2 + 2*sk + 4*p11 - 1

 MM=(2^(2/alpha)*(gamma(nu/2 + 1/alpha))^2)
 vari=2^(2/alpha)*(gamma(nu/2 + 2/alpha))*gamma(nu/2)* (1+3*sk2) - sk2*2^(2/alpha+2)*gamma(nu/2+1/alpha)^2
 corr= MM*(a1*a3-4*sk2)/vari
 vv=as.numeric(nuisance['sill'])*vari/(nn^(2/alpha)*gamma(nu/2)^2)
}

if(bivariate){}
}
###############################################################
if(model==27)   ##  two piece student case case
    {
  if(!bivariate)
{
          corr=cr$corr*(1-as.numeric(nuisance['nugget']))
          nu=as.numeric(1/nuisance['df']); sk=as.numeric(nuisance['skew'])
          corr2=corr^2;sk2=sk^2
          a1=Re(hypergeo::hypergeo(0.5,0.5,nu/2,corr2))
          a2=corr*asin(corr) + (1-corr2)^(0.5)
          ll=qnorm((1-sk)/2)
          p11=pbivnorm::pbivnorm(ll,ll, rho = cr$corr, recycle = TRUE)
          a3=3*sk2 + 2*sk + 4*p11 - 1
          KK=( nu*(nu-2)*gamma((nu-1)/2)^2) / (nu*pi*gamma(nu/2)^2*(3*sk2+1)-4*sk2*nu*(nu-2)*gamma((nu-1)/2)^2 )
          corr= KK*(a1*a2*a3-4*sk2);
          ttemp=gamma(0.5*(nu-1))/gamma(0.5*nu)
          vv=as.numeric(nuisance['sill'])*((nu/(nu-2))*(1+3*sk2) - 4*sk2*(nu/pi)*ttemp^2)
      }

if(bivariate){}
}
###############################################################
if(model==38)   ##  two piece tukey h  case
    {
      if(!bivariate)
{
          corr=cr$corr*(1-as.numeric(nuisance['nugget']))
          tail=as.numeric(nuisance['tail']); sk=as.numeric(nuisance['skew'])
          corr2=corr^2;sk2=sk^2;
          gg2=(1-(1-corr2)*tail)^2
          xx=corr2/gg2
          A=(asin(sqrt(xx))*sqrt(xx)+sqrt(1-xx))/(1-xx)^(1.5)
          ll=qnorm((1-sk)/2)
          p11=pbivnorm::pbivnorm(ll,ll, rho = cr$corr, recycle = TRUE)
          a3=3*sk2 + 2*sk + 4*p11 - 1
          mm=8*sk2/(pi*(1-tail)^2);
          ff=(1+3*sk2)/(1-2*tail)^(1.5)
          M=(2*(1-corr2)^(3/2))/(pi*gg2)
          corr=  (M*A*a3-mm)/( ff- mm)
          vv= as.numeric(nuisance['sill'])*((1-2*tail)^(-1.5)* (1+3*(sk2)) - 4*(sk2)*2/(pi*(1-tail)^2))
  }
if(bivariate){}
}
###############################################################
if(model==29)   ##  two piece gaussian case
    {
      if(!bivariate)
{
          corr1=cr$corr*(1-as.numeric(nuisance['nugget']))
          sk=as.numeric(nuisance['skew']);
          corr2=sqrt(1-corr1^2); sk2=sk^2
          ll=qnorm((1-sk)/2)
          p11=pbivnorm::pbivnorm(ll,ll, rho = cr$corr, recycle = TRUE)
          KK=3*sk2+2*sk+ 4*p11 - 1
          corr=(2*((corr2 + corr1*asin(corr1))*KK)- 8*sk2)/(3*pi*sk2  -  8*sk2   +pi   )
          vv=as.numeric(nuisance['sill'])*(1+3*sk2-8*sk2/pi)
  }
if(bivariate){}
}
###############################################################
if(model==10)  {  ##  skew Gaussian case

          if(!bivariate){
              corr=cr$corr*(1-as.numeric(nuisance['nugget']))
              sk=as.numeric(nuisance['skew'])
              corr2=cr$corr^2; ; sk2=sk^2; vv=as.numeric(nuisance['sill'])
              corr=(2*sk2)*(sqrt(1-corr2) + cr$corr*asin(cr$corr)-1)/(pi*vv+sk2*(pi-2)) + (corr*vv)/(vv+sk2*(1-2/pi))
              vv=vv+as.numeric(nuisance['skew'])^2*(1-2/pi)
     }
      if(bivariate){}
}
if(model==59)  {  ##  skewLaplace case

          if(!bivariate){
              corr=cr$corr*(1-as.numeric(nuisance['nugget']))
              sk=as.numeric(nuisance['skew'])
              corr=corr^2
              vv=as.numeric(nuisance['sill'])*(1/sk^2 + 1/(1-sk)^2)
     }
      if(bivariate){}
}
#################################################################################
################ covariance matrix for models defined on the real line ##########
#################################################################################
if(model %in% c(1,9,34,12,20,18,39,27,38,29,10,40,59)){
if(!bivariate)
{
 if(type=="Standard"){
     # Builds the covariance matrix:
        varcov <-  diag(dime)
        varcov[lower.tri(varcov)] <- corr
        varcov <- t(varcov)
        varcov[lower.tri(varcov)] <- corr
        varcov=varcov*vv
      }
    if(type=="Tapering")  {
          vcov <- vv*corr;
          varcov <- new("spam",entries=vcov,colindices=setup$ja,
                             rowpointers=setup$ia,dimension=as.integer(rep(dime,2)))
          diag(varcov)=vv

        }
}
#####
    if(bivariate)      {
       if(type=="Standard"){
          corr <- cr$corr
          varcov<-diag(dime)
          varcov[lower.tri(varcov,diag=TRUE)] <- corr
          varcov <- t(varcov)
          varcov[lower.tri(varcov,diag=TRUE)] <- corr
        }
        if(type=="Tapering")  {

          varcov <-new("spam",entries=cr$corr,colindices=setup$ja,
                         rowpointers=setup$ia,dimension=as.integer(rep(dime,2)))
        }
      }
}
#################################################################################
############# covariance models defined on a bounded support of the real line ###
#################################################################################
if(model==28)   ##  beta case
    {
      if(!bivariate) {
         corr=cr$corr*(1-as.numeric(nuisance['nugget']))
         corr2=corr^2
         shape1=as.numeric(nuisance['shape1']);
         shape2=as.numeric(nuisance['shape2']);
         ssup=as.numeric(nuisance['max'])-as.numeric(nuisance['min'])
         cc=0.5*(shape1+shape2)
         vv=ssup^2*shape1*shape2/((cc+1)*(shape1+shape2)^2)        ## variance remember min and max!
         idx=which(abs(corr2)>1e-10);corr22=corr2[idx]
######################
  #nu=shape1;alpha=shape2
  nu2=shape1/2;alpha2=shape2/2
  res=0;ss=0;k=0;A=rep(0,length(corr22))
  while(k<=100){
    p1=2*(lgamma(cc+k)-lgamma(cc)+lgamma(nu2+1+k)-lgamma(nu2+1))
    p2=lgamma(k+1)+(lgamma(nu2+k)-lgamma(nu2))+2*(lgamma(cc+1+k)-lgamma(cc+1))
    b1=p1-p2
    b2=log(hypergeo::genhypergeo(U=c(cc+k,cc+k,alpha2), L=c(cc+k+1,cc+k+1), polynomial=TRUE,maxiter=1000, z=corr22))
    b3=k*log(corr22)
    sum=exp(b1+b2+b3)
    res=res+sum
    A=res
    if (all(sum<1e-6)){
      break
    }
    k=k+1
  }
######################
         corr[idx]=A
         corr=shape1*(cc + 1 ) * ((1-corr2)^(cc) *corr -1)/shape2 ## correlation
         corr[-idx]=0
        }
     if(bivariate){}
}

if(model==33||model==42)   ##  Kumaraswamy case
    {
      if(!bivariate) {
#######################

      corr=cr$corr*(1-as.numeric(nuisance['nugget']))
      if(model==33){
                ga=as.numeric(nuisance['shape2'])
                eta=as.numeric(nuisance['shape1'])
                   }
      if(model==42){
                ga=as.numeric(nuisance['shape'])
                mu_kuma2=as.numeric(ML)
                if(length(mu_kuma2) != 1L && (max(mu_kuma2)-min(mu_kuma2)) > sqrt(.Machine$double.eps))
                  stop("GeoCovmatrix: plain Kumaraswamy2 covariance is currently implemented only for a constant mean; use the copula covariance engine for location-dependent margins.", call.=FALSE)
                pmed=plogis(mu_kuma2[1L])
                eta=log(0.5)/log1p(-(pmed^ga))
                   }

      mm=eta*beta(1+1/ga,eta)
      NN=length(corr);
      res=double(NN)
      bb=.C("corr_kuma_vec",as.double(corr),as.double(eta),as.double(ga),
      res=as.double(res),as.integer(NN),PACKAGE='GeoModels', NAOK=TRUE)
      corr=bb$res
      ssup=as.numeric(nuisance['max'])-as.numeric(nuisance['min'])
      vv=ssup^2*(eta*beta(1+2/ga,eta)-mm^2)
      }
      if(bivariate){}
}

#################################################################################
# covariance matrix for models defined on a bounded support of the  real line  ##
#################################################################################
if(model %in% c(28,33,42)){
if(!bivariate)
{
 if(type=="Standard"){
     # Builds the covariance matrix:
        varcov <-  diag(dime)
        varcov[lower.tri(varcov)] <- corr
        varcov <- t(varcov)
        varcov[lower.tri(varcov)] <- corr
        varcov=varcov*vv
      }
    if(type=="Tapering")  {
          vcov <- vv*corr;
          varcov <- new("spam",entries=vcov,colindices=setup$ja,
                             rowpointers=setup$ia,dimension=as.integer(rep(dime,2)))
          diag(varcov)=vv
        }
}
#####
    if(bivariate)      {
       if(type=="Standard"){
          corr <- cr$corr
          varcov<-diag(dime)
          varcov[lower.tri(varcov,diag=TRUE)] <- corr
          varcov <- t(varcov)
          varcov[lower.tri(varcov,diag=TRUE)] <- corr
        }
        if(type=="Tapering")  {
          varcov <-new("spam",entries=cr$corr,colindices=setup$ja,
                         rowpointers=setup$ia,dimension=as.integer(rep(dime,2)))
        }
      }
}

#################################################################################
################ wrapped Gaussian RF                     #######################
#################################################################################


if(model==13)   ##  wrapped
    {
      if(!bivariate) {
         corr=cr$corr*(1-as.numeric(nuisance['nugget']))
         vv=sinh(as.numeric(nuisance['sill']))
         corr=sinh(as.numeric(nuisance['sill'])*corr)/vv
        }
     if(bivariate){}
    }

#################################################################################
################ models defined on the positive real line #######################
#################################################################################
if(model %in% c(21,26,24,22)) 
    {

         mu = ML   ## mean  function

if(model==21)   ##  gamma case
    {
      if(!bivariate) {
         corr=cr$corr*(1-as.numeric(nuisance['nugget']))
         corr=corr^2   ### gamma correlation
         vv=exp(mu)^2 * 2/as.numeric(nuisance['shape'])
        }
     if(bivariate){}
    }
###############################################################
if(model==26)   ##  weibull case
    {
        if(!bivariate) {
         corr=cr$corr*(1-as.numeric(nuisance['nugget']))
         sh=as.numeric(nuisance['shape'])
         vv=exp(mu)^2 * (gamma(1+2/sh)/gamma(1+1/sh)^2-1)
         ## Use Eq. (6) directly.  The previous Euler-transformed form
         ## involved 2F1(1+1/sh,1+1/sh;1;rho^2), which is numerically
         ## singular for the hypergeo backend near rho^2=1.
         corr=.GeoWeibullCorr(corr, sh)
        }
  if(bivariate) {}
    }
###############################################################
    if(model==24)   ## log-logistic case
    {
       if(!bivariate) {
       corr=cr$corr*(1-as.numeric(nuisance['nugget']))
       sh=as.numeric(nuisance['shape'])
       vv=(exp(mu))^2* (2*sh*sin(pi/sh)^2/(pi*sin(2*pi/sh))-1)
       corr= ((pi*sin(2*pi/sh))/(2*sh*(sin(pi/sh))^2-pi*sin(2*pi/sh)))*
             (Re(hypergeo::hypergeo(-1/sh,-1/sh ,1 ,corr^2))* Re(hypergeo::hypergeo(1/sh,1/sh ,1 ,corr^2)) -1)
        }
   if(bivariate){}
    }
###############################################################
if(model==22)  {  ## Log Gaussian
      if(!bivariate) {
      corr=cr$corr*(1-as.numeric(nuisance['nugget']))
      vvar=as.numeric(nuisance['sill'])
      corr=(exp(vvar*corr)-1)/(exp(vvar)-1)
            vv=(exp(mu))^2*(exp(vvar)-1) 
             }
    if(bivariate){}
  }
#################################################################################
################ covariance matrix for models defined on the positive real line #
#################################################################################

       if(type=="Standard"){
     # Builds the covariance matrix:
        varcov <-  diag(dime)
        varcov[lower.tri(varcov)] <- corr
        varcov <- t(varcov)
        varcov[lower.tri(varcov)] <- corr
      }
    if(type=="Tapering")  {
          varcov <- new("spam",entries=corr,colindices=setup$ja,
                             rowpointers=setup$ia,dimension=as.integer(rep(dime,2)))
          diag(varcov)=1
        }
    V=vv%*%t(vv)
   
    varcov=varcov*sqrt(V)
}

###############################################################
################################ end continous models #########
###############################################################

###############################################################
################################ start discrete #models ########
###############################################################
if(model %in% c(2,11,30,16,14,43,45,46,57,58)){ #  binomial (negative)Gaussian type , Poisson (inflated) poissongamma poisson gamma zip

if(model==2||model==11)
{if(length(n)==1) n=rep(n,dime)}
if(model==14) n=rep.int(1L,dime)
if(model %in% c(16,45)) n=rep.int(as.integer(n[1L]),dime)
if(!bivariate){
mu=ML

if(type=="Standard")  {

    fname <-'CorrelationMat_dis2'
    if(spacetime) fname <- 'CorrelationMat_st_dyn_dis2'

  cr=dotCall64::.C64(fname,SIGNATURE =
      c("double","double","double","double","double","integer","double","integer","double","double","double","integer","integer","integer"),
        corr=dotCall64::vector_dc("double",numpairstot), coordx,coordy,coordz, coordt,corrmodel,c(mu), n,other_nuis,paramcorr,radius,ns,NS,model,
  INTENT = c("w","r","r","r","r","r","r","r", "r", "r","r", "r", "r", "r"),
             PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)

     corr=cr$corr # ojo que corr en este caso es una covarianza y ya va con el nugget
     varcov <-  diag(dime)
     varcov[lower.tri(varcov)] <- corr
     varcov <- t(varcov)
     varcov[lower.tri(varcov)] <- corr

}
############################
############################
 if(type=="Tapering")  {
        tap <-new("spam",entries=setup$taps,colindices=setup$ja,
                         rowpointers=setup$ia,dimension=as.integer(rep(dime,2)))
        idx=spam::triplet(tap)$indices
        
        fname <- "CorrelationMat_dis_tap"
        if(spacetime) fname <- "CorrelationMat_st_dis_tap"
     
if(spacetime){


cr<-dotCall64::.C64("CorrelationMat_st_dis_tap",corr=dotCall64::vector_dc("double",numpairs),coordx,coordy,coordz,coordt,
    corrmodel,other_nuis,paramcorr,radius,ns,NS,n[idx[,1]],n[idx[,2]],mu[idx[,1]],mu[idx[,2]],model,
    SIGNATURE=c("double","double","double","double","double","integer","double",
        "double","double","integer","integer","integer","integer","double","double","integer"),
    INTENT=c("w",rep("r",15)),
          PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
}
else
{
cr<-dotCall64::.C64("CorrelationMat_dis_tap",corr=dotCall64::vector_dc("double",numpairs),coordx,coordy,coordz,coordt,corrmodel,
    other_nuis,paramcorr,radius,ns,NS,n[idx[,1]],n[idx[,2]],mu[idx[,1]],mu[idx[,2]],model,
   SIGNATURE=c("double","double","double","double","double","integer","double",
        "double","double","integer","integer","integer","integer","double","double","integer"),
    INTENT=c("w",rep("r",15)),
           PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
}
    varcov <-new("spam",entries=cr$corr,colindices=setup$ja,
            rowpointers=setup$ia,dimension=as.integer(rep(dime,2)))

        }
            ## updating the diagonal with variance
  if(model %in% c(2,11)) { pg=pnorm(mu);  vv=pg*(1-pg)*n; diag(varcov)=vv }
  if(model %in% c(14))   { pg=pnorm(mu); vv=  (1-pg)/pg^2; diag(varcov)=vv }
  if(model %in% c(16))   { pg=pnorm(mu); vv=n*(1-pg)/pg^2; diag(varcov)=vv }
  if(model %in% c(30))   { vv=exp(mu); diag(varcov)=vv } ## poisson


  if(model %in% c(46))  
   { mm=exp(mu); vv=mm*(1+mm/as.numeric(param$shape)); diag(varcov)=vv } ## poissongamma

    if(model %in% c(43))   { mm=exp(mu); pg=pnorm(param$pmu)
                           vv=(1-pg)*mm*(1+pg*mm)
                           diag(varcov)=vv }

  if(model %in% c(45))   {
                            p=pnorm(param$pmu)
                            MM=pnorm(mu)
                            vv=n*(1-MM)*(1-p)*(1+n*p*(1-MM)) /MM^2
                            diag(varcov)=vv 
                        }

  if(model %in% c(57,58))   { mm=exp(mu); kk=as.numeric(param$shape);
                            pg=pnorm(param$pmu)
                           vv=(1-pg)*mm*(1+pg*mm+(mm/kk))
                           diag(varcov)=vv }
}

}
###############################################################
################################ end discrete #models #########
###############################################################
return(varcov)
}

#############################################################################################
#################### end internal function ##################################################
#############################################################################################


##############################################################################
###### extracting GeoFit object informations if necessary              #######
##############################################################################
if(!is.null(estobj)){
   if(!inherits(estobj,"GeoFit"))
               stop("need  a 'GeoFit' object as input\n")
if(!estobj$grid){  #not regular grid 

 if(!estobj$bivariate){  if(is.null(estobj$coordx_dyn)) coordx=cbind(estobj$coordx,estobj$coordy,estobj$coordz)
                         else coordx_dyn=estobj$coordx_dyn
                      } ## spatial (temporal) non regular case
 else  {    if(is.null(estobj$coordx_dyn))  {
                                              n1 <- as.integer(estobj$ns[1L])
                                              idx1 <- seq_len(n1)
                                              coordx=estobj$coordx[idx1]    # bivariate, common static locations
                                              coordy=estobj$coordy[idx1]
                                              coordz=if(is.null(estobj$coordz)) NULL else estobj$coordz[idx1]
                                            }  
            else {coordx_dyn=estobj$coordx_dyn}                                      # bivariate  dynamic  
       }
 }
else  { coordx=estobj$coordx; 
        coordy=estobj$coordy
        coordz=estobj$coordz
      }
   coordt=estobj$coordt
   coordx_dyn=estobj$coordx_dyn
   corrmodel=estobj$corrmodel
   model=estobj$model
   distance=estobj$distance
   grid=estobj$grid
   n=estobj$n
   param=append(estobj$param,estobj$fixed)
   radius=estobj$radius
   copula=estobj$copula
   anisopars=estobj$anisopars
    if(!is.null(anisopars))  {param$angle=NULL; param$ratio=NULL}
   X=estobj$X

}
################# end extracting information###################################################
model <- .GeoCovCanonicalModel(model)
if( !is.character(corrmodel)|| is.null(CkCorrModel(corrmodel)))       stop("the name of the correlation model is wrong")
 if(is.null(CkModel(model))) stop("The name of the  model  is not correct\n")


 compact_models <- c(
  "GenWend_Matern","GenWend_Matern2","Wend0", "wend0","Wend1", "wend1","Wend2", "wend2","Genwend", "GenWend",
  "Hypergeometric2","Hypergeometric","Hypergeometric_Matern","Hypergeometric_Matern2","GenWend_Hole", "GenWend_hole",
  "GenWend_Matern_Hole", "GenWend_Matern_hole","Wen0_space", "wen0_space","Wen0_time",  "wen0_time",
  "Wen1_space", "wen1_space","Wen1_time",  "wen1_time","Wen2_space", "wen2_space","Wen2_time",  "wen2_time",
  "Gneiting_wen_S", "gneiting_wen_S", "Gneiting_Wen_S","Gneiting_wen_T", "gneiting_wen_T","Genwend_Genwend_nosep",
  "Wend0_Wend0","Wend0_Wend1","Wend0_Wend2","Wend1_Wend0","Wend1_Wend1","Wend1_Wend2","Wend2_Wend0","Wend2_Wend1","Wend2_Wend2",
  "GenWend_GenWend","Bi_wend0_sep", "Bi_Wend0_contr","Bi_wend0","Bi_wend1_sep","Bi_Wend1_contr","Bi_wend1",
  "Bi_wend2_sep","Bi_wend2","Bi_Wend2_contr","Bi_GenWend_sep", "Bi_genWend_sep","Bi_GenWend_contr", "Bi_genWend_contr", "Bi_GenWend", "Bi_genWend"
)

if (sparse && !(corrmodel %in% compact_models)) {
  stop("Sparse option should be used only for compactly supported covariance models\n")
}

bivariate<-CheckBiv(CkCorrModel(corrmodel))
spacetime<-CheckST(CkCorrModel(corrmodel))
space=!spacetime&&!bivariate

## Normalize the optional copula field before input checking.
if(!is.null(copula)) {
  copula <- as.character(copula)[1L]
  if(is.na(copula) || !nzchar(trimws(copula))) copula <- NULL
}
if(!is.null(copula) && !(copula %in% c("Gaussian", "Clayton", "SkewGaussian")))
  stop("GeoCovmatrix supports copula='Gaussian', 'Clayton', or 'SkewGaussian'.",
       call. = FALSE)
if(!is.null(copula) && copula %in% c("Clayton", "SkewGaussian") &&
   !(as.integer(CkModel(model)) %in% .GeoCopulaPredictionModels))
  stop("GeoCovmatrix supports copula='", copula,
       "' only for the validated copula prediction margins.", call. = FALSE)

if(is.null(copula)) {
  plain_supported <- c(
    "Gaussian","Tukeygh","Tukeyh","Tukeyh2","StudentT","SinhAsinh",
    "SkewStudentT","TwoPieceBimodal","TwoPieceStudentT","TwoPieceTukeyh",
    "TwoPieceGaussian","SkewGaussian","SkewLaplace","Beta","Kumaraswamy",
    "Kumaraswamy2","Wrapped","Gamma","Weibull","LogLogistic","LogGaussian",
    "Binary","Bernoulli","Binomial","Geometric","Geom","BinomialNeg",
    "Poisson","PoissonZIP","BinomialNegZINB","PoissonGamma",
    "PoissonGammaZIP","PoissonGammaZIP1"
  )
  if(!(model %in% plain_supported))
    stop("GeoCovmatrix: plain observed-scale covariance is not implemented for model = '",
         model, "'.", call.=FALSE)
  if(bivariate && model != "Gaussian")
    stop("GeoCovmatrix: plain bivariate observed-scale covariance is currently implemented only for model = 'Gaussian'.",
         call.=FALSE)
  if(model %in% c("PoissonGammaZIP","PoissonGammaZIP1") && (spacetime || sparse))
    stop("GeoCovmatrix: PoissonGammaZIP/PoissonGammaZIP1 covariance is currently implemented only for dense purely spatial designs.",
         call.=FALSE)
}

Mobs_cov <- NULL

tapsep=NULL ;   maxdist=NULL; maxtime=NULL; type="Standard"    #setted for older compatibility
##############################################################################
###### extracting sp object informations if necessary              ###########
##############################################################################

if(!is.null(spobj)) {
   if(space||bivariate){
        a=sp2Geo(spobj,NULL); coordx=a$coords 
       if(!a$pj) {if(distance!="Chor") distance="Geod"}
    }
   if(spacetime){
        a=sp2Geo(spobj,NULL); coordx=a$coords ; coordt=a$coordt 
        if(!a$pj) {if(distance!="Chor") distance="Geod"}
     }
}

## Preserve the user's grid declaration in the returned object.  Internally a
## regular grid is expanded to explicit locations and then processed as an
## irregular coordinate matrix.
grid_input <- isTRUE(grid)

.GeoValidate3D(
  coordx = coordx, coordy = coordy, coordz = coordz,
  coordx_dyn = coordx_dyn, distance = distance, grid = grid,
  context = "GeoCovmatrix"
)

## Hole-effect generalized Wendland validity depends on the Euclidean
## dimension through the turning-bands construction.  Infer d from the
## coordinate representation before flattening it.
hole_d <- 2L
if(!is.null(coordz)) hole_d <- 3L
if(is.matrix(coordx) && ncol(coordx) >= 3L) hole_d <- 3L
if(is.list(coordx_dyn) && length(coordx_dyn) &&
   any(vapply(coordx_dyn, function(z) is.matrix(z) && ncol(z) >= 3L, logical(1L))))
  hole_d <- 3L
.GeoValidateHoleWendlandParameters(corrmodel, param, d=hole_d, context="GeoCovmatrix")

if(!is.null(coordx_dyn)) {
  coordx_dyn <- .GeoNormalizeDynamicCoordinates(
    coordx_dyn,
    coordt = if(spacetime) coordt else NULL,
    context = "GeoCovmatrix"
  )
  if(spacetime) {
    .GeoValidateDynamicRows(X, coordx_dyn, name = "X",
                            context = "GeoCovmatrix")
  }
}
if(bivariate) {
  biv_info <- .GeoBivariateCoordinateInfo(
    coordx = coordx, coordy = coordy, coordz = coordz,
    coordx_dyn = coordx_dyn, grid = grid, context = "GeoCovmatrix"
  )
  if(biv_info$dynamic) coordx_dyn <- biv_info$blocks
  if(!is.null(X)) {
    X <- .GeoBivariateDesign(X, biv_info$ns, context = "GeoCovmatrix")
  }
}

.GeoCheckDuplicateLocations(
  coordx = coordx, coordy = coordy, coordz = coordz, coordt = coordt,
  coordx_dyn = coordx_dyn, spacetime = spacetime, bivariate = bivariate,
  grid = grid, context = "GeoCovmatrix", check.duplicates = check.duplicates
)

#######################################################
     ## setting zero mean and nugget if no mean or nugget is fixed
    if(!bivariate){
    if(is.null(param$mean)) param$mean<-0
    ## CkInput(..., fcall = "Simulation") validates the union of the
    ## Gaussian nuisance parameters and the declared model nuisance
    ## parameters.  Hence the auxiliary generic nugget must exist here even
    ## for double-nugget ZIP models.  It is removed *after* StartParam(),
    ## before building the positional nuisance vector passed to
    ## CorrelationMat_dis2 (see cleanup below).
    if(is.null(param$nugget)) param$nugget<-0
    #if(length(param$mean)>1) MM=fixed$mean
    if(is.null(param$sill)) param$sill<-1  
}
    else{
    if(is.null(param$mean_1)) param$mean_1<-0
    if(is.null(param$mean_2)) param$mean_2<-0
    if(is.null(param$nugget_1)) param$nugget_1<-0
    if(is.null(param$nugget_2)) param$nugget_2<-0
    sill_names <- c("sill_1", "sill_2")
    missing_sills <- sill_names[vapply(sill_names, function(nm)
      is.null(param[[nm]]), logical(1L))]
    if(length(missing_sills))
      stop("GeoCovmatrix: bivariate models require parameter(s) ",
           paste(missing_sills, collapse=", "), ".", call.=FALSE)

}

    if(is.null(coordx_dyn)){
    coordx=unname(coordx);coordy=unname(coordy);coordz=unname(coordz)}
    else{coordx_dyn=lapply(coordx_dyn, unname)}


    #if the covariance is compact supported  and option sparse is used
    #then set the code as a tapering and an object spam is returned
if(sparse) {
    covmod=CkCorrModel(corrmodel)
    if(covmod %in% c(10,11,13,15,19,6,7,21,22,23,26,29,30,
                     63,64,65,66,67,68,
                     69,70,71,72,73,74,75,76,77,78,
                     111,112,129,113,114,131,132,130,134,
                     115,116,120))
    {
      type="Tapering"
      if(bivariate){
                  #taper="unit_matrix_biv"
                  if(covmod %in% c(111,113,115,130)) maxdist=c(param$scale,param$scale,param$scale)
                  if(covmod %in% c(112,114,116,134)) maxdist=c(param$scale_1,param$scale_12,param$scale_2)
                  if(covmod %in% c(120,129,131,132)) maxdist=c(param$scale_1,0.5*(param$scale_1+param$scale_2),param$scale_2)
      }
      if(spacetime)
      {

      maxdist=param$scale_s;maxtime=param$scale_t

       if(covmod==63||covmod==65||covmod==67) {  tapsep=c(param$power2_s,param$power_t,param$scale_s,param$scale_t,param$sep) }
       if(covmod==64||covmod==66||covmod==68) {  tapsep=c(param$power_s,param$power2_t,param$scale_s,param$scale_t,param$sep) }
    }
      if(space){  ### spatial Gen Wend (reparametrized)
        maxdist=param$scale
        if(covmod==6)   maxdist=as.numeric(param$scale*exp((lgamma(2*param$smooth+1/param$power2+1)-lgamma(1/param$power2))/ (1+2*param$smooth) ))
        if(covmod==7)   maxdist=as.numeric(param$scale*exp((lgamma(2*param$smooth+param$power2+1)-lgamma(param$power2))/ (1+2*param$smooth) ))
        if(covmod==23)
              {
              log_num <- log(2) * (2 * param$smooth + 1) + lgamma((1/param$power2 + 1) / 2 + param$smooth) + lgamma((1/param$power2 + 2 + 1) / 2 + 2 * param$smooth)
              log_den <- lgamma(1/param$power2 / 2) +lgamma((1/param$power2 + 2) / 2 + param$smooth)
              maxdist <- param$scale*exp(log_num - log_den)^(1/(1+2*param$smooth))
              }
        
        if(covmod==30)    {
              log_num <- log(2) * (2 * param$smooth + 1) + lgamma((param$power2 + 1) / 2 + param$smooth) + lgamma((param$power2 + 2 + 1) / 2 + 2 * param$smooth)
              log_den <- lgamma(param$power2 / 2) +lgamma((param$power2 + 2) / 2 + param$smooth)
              maxdist <- param$scale*exp(log_num - log_den)^(1/(1+2*param$smooth))
              }
        ## Reparameterized hole-effect GW is GW_{nu*a,xi-1/2,nu,d,k};
        ## hence its exact compact support is nu*a, not a.
        if(covmod==29) maxdist=as.numeric(param$power2*param$scale)
      }
  }
  taper=corrmodel
}
##### sill parameter fixed for some models

if(!bivariate){

if(model %in% c("Weibull","Poisson","Binomial","Gamma","LogLogistic",
        "BinomialNeg","Binary","Bernoulli","Geom","Geometric","Gaussian_misp_Poisson",
        'PoissonZIP','Gaussian_misp_PoissonZIP','BinomialNegZINB', 'PoissonGammaZIP','PoissonGammaZIP1','PoissonGamma',
        'PoissonZIP1','Gaussian_misp_PoissonZIP1','BinomialNegZINB1',
        'Beta2','Kumaraswamy2','Beta','Kumaraswamy')){
if(is.null(param$sill)) param$sill=1
else                    param$sill=1
}
}


coords=NULL
coordx_dyn_orig=coordx_dyn

if(is.null(coordx_dyn)){
 #######################################
    if(grid) { coords=as.matrix(expand.grid(coordx,coordy)); grid=FALSE }
    else     {  if(!is.null(coordy)) { coords=as.matrix(cbind(coordx,coordy,coordz))}
                else                 { coords=as.matrix(coordx);}
                if(ncol(coords)==3) coordz=coords[,3]
                else                coordz=NULL
             }
    coords_orig=coords
    if(!is.null(anisopars)) {  coords=GeoAniso(coords,c(anisopars$angle,anisopars$ratio))}
}
else {
    coords_orig=do.call(rbind, coordx_dyn_orig)
    coordx_dyn_metric=coordx_dyn_orig
    if(!is.null(anisopars)) {
      if(ncol(coords_orig)!=2L)
        stop("GeoCovmatrix: anisotropy for dynamic three-dimensional coordinates is not implemented.", call.=FALSE)
      coordx_dyn_metric=lapply(coordx_dyn_orig, function(z)
        GeoAniso(z,c(anisopars$angle,anisopars$ratio)))
    }
    coordx_dyn=coordx_dyn_metric
    coords=do.call(rbind, coordx_dyn)
    if(ncol(coords)==3) coordz=coords[,3]
    else                coordz=NULL
}
coordz_orig=if(ncol(coords_orig)==3L) coords_orig[,3] else NULL
 #######################################
#######################################

    if(!bivariate) {
      nobs_mean_cov <- if(!is.null(coordx_dyn)) {
        sum(vapply(coordx_dyn, function(z) nrow(as.matrix(z)), integer(1L)))
      } else {
        nrow(coords) * if(is.null(coordt)) 1L else length(coordt)
      }
      if(!is.numeric(n) || !length(n) || any(!is.finite(n)))
        stop("GeoCovmatrix: n must be finite numeric.", call.=FALSE)
      if(any(n < 1) || any(abs(n-round(n)) > sqrt(.Machine$double.eps)))
        stop("GeoCovmatrix: n must contain positive integer values.", call.=FALSE)
      if(is.null(copula) && model %in% c("Binary","Bernoulli")) {
        if(length(n) != 1L || as.integer(round(n[1L])) != 1L)
          stop("GeoCovmatrix: Binary/Bernoulli is the one-trial field; n must equal 1.", call.=FALSE)
        n <- 1L
      }
      if(is.null(copula) && model %in% c("Geom","Geometric")) {
        if(length(n) != 1L || as.integer(round(n[1L])) != 1L)
          stop("GeoCovmatrix: Geometric is the r = 1 Negative-Binomial field; n must equal 1.", call.=FALSE)
        n <- 1L
      }
      if(is.null(copula) && model %in% c("BinomialNeg","BinomialNegZINB")) {
        if(length(n) != 1L)
          stop("GeoCovmatrix: for direct Negative-Binomial fields n is the common number r of successes and must be a positive integer scalar.", call.=FALSE)
        n <- as.integer(round(n[1L]))
      } else if(model == "Binomial" &&
                length(n) != 1L && length(n) != nobs_mean_cov) {
        stop("GeoCovmatrix: Binomial n must be scalar or contain exactly one value per observation (",
             nobs_mean_cov, ").", call.=FALSE)
      }
      Mobs_cov <- .GeoMean_external(param, n = nobs_mean_cov)
      if(!is.null(Mobs_cov) && !is.null(X))
        stop("A site-specific mean vector and X cannot be used together.\n")
      if(!is.null(X)) {
        X <- .GeoMean_design(X, n = nobs_mean_cov, name = "X")
        invisible(.GeoMean_beta(param, ncol(X)))
      } else {
        supplied_mean <- .GeoMean_supplied_names(param)
        if(is.null(Mobs_cov) && length(setdiff(supplied_mean, "mean")))
          stop("Multiple mean coefficients require an explicit design matrix X.\n")
        if(is.null(Mobs_cov)) invisible(.GeoMean_beta(param, 1L))
      }
      if(!is.null(Mobs_cov)) param$mean <- 0
    }


    checkinput <- CkInput(coords[,1], coords[,2],coordz, coordt, coordx_dyn, corrmodel, NULL, distance, "Simulation",
                             NULL, grid, NULL, maxdist, maxtime,  model=model, n,  NULL,
                              param, radius, NULL, taper, tapsep,  "Standard", NULL, NULL,copula,X)




    
    if(!is.null(checkinput$error)) stop(checkinput$error)
    spacetime_dyn=FALSE
    if(!is.null(coordx_dyn))  spacetime_dyn=TRUE
    # Initialising the parameters:
    
    initparam <- StartParam(coords[,1], coords[,2],coordz, coordt,coordx_dyn, corrmodel, NULL, distance, "Simulation",
                           NULL, grid, NULL, maxdist, NULL,maxtime, model, n,
                           param, NULL, NULL, radius, NULL, taper, tapsep,  type, 
                           type, FALSE,copula,X,FALSE)

    if(is.null(coordz)) cc=cbind(initparam$coordx,initparam$coordy,0)
    else cc=cbind(initparam$coordx,initparam$coordy,initparam$coordz)

    if(!spacetime_dyn) dime=initparam$numcoord*initparam$numtime
    else               dime=sum(initparam$ns)

    if(!initparam$bivariate) numpairstot=dime*(dime-1)*0.5
    if(initparam$bivariate)  numpairstot=dime*(dime-1)*0.5+dime
    if(!is.null(initparam$error)) stop(initparam$error)
    setup<-initparam$setup
    if(initparam$type=="Tapering")
    {
       corr <- double(initparam$numpairs)
       if(ncol(cc)==2)  ccz=corr
       if(ncol(cc)==3)  ccz=cc[,3]

       #tapmod <- setup$tapmodel
       ### unit taperssss ####
       if(sparse){
           if(spacetime) tapmod=230
           if(bivariate) tapmod=147
           if(space) tapmod=36
           }
       else(tapmod=CkCorrModel(taper))
   
      if(initparam$spacetime) fname= "CorrelationMat_st_tap"
      else {
         if(!initparam$spacetime) fname= "CorrelationMat_tap"
         if(initparam$bivariate) fname= "CorrelationMat_biv_tap"
     }

if(is.null(tapsep)) tapsep=1
   tp=dotCall64::.C64(fname,SIGNATURE =
         c("double","double","double","double","double","integer","double","double","double","integer","integer"),
        tapcorr=dotCall64::vector_dc("double",initparam$numpairs),
        cc[,1],cc[,2],ccz,initparam$coordt,tapmod, 1,tapsep,1,1,1,
  INTENT = c("w","r","r","r","r","r","r","r", "r", "r","r"),
             PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)

        setup$taps<-tp$tapcorr

    }
 ### end tapering   
    if(is.null(X))  initparam$X=as.matrix(rep(1,dime))

    if(bivariate) {if(is.null(X))  initparam$X=as.matrix(rep(1,initparam$ns[1]+initparam$ns[2])) }
    if(!space){
          initparam$NS=cumsum(initparam$ns);
            if(spacetime_dyn){  initparam$NS=c(0,initparam$NS)[-(length(initparam$ns)+1)]}
            else{               initparam$NS=rep(0,initparam$numtime)}
    }
    if(is.null(initparam$NS)) initparam$NS=0

    if(initparam$model %in% c(43,45,57)) initparam$namesnuis=initparam$namesnuis[!initparam$namesnuis %in% "nugget"]



other_nuis=NULL
ML=0
##### sill parameter fixed for some models
if(!bivariate){

if(model %in% c("Weibull","Poisson","Binomial","Gamma","LogLogistic",
        "BinomialNeg","Binary","Bernoulli","Geom","Geometric","Gaussian_misp_Poisson",
        'PoissonZIP','Gaussian_misp_PoissonZIP','BinomialNegZINB', 'PoissonGammaZIP','PoissonGammaZIP1','PoissonGamma',
        'PoissonZIP1','Gaussian_misp_PoissonZIP1','BinomialNegZINB1',"LogGaussian",
        'Beta2','Kumaraswamy2','Beta','Kumaraswamy')) 
  {
  ### setting mean because mean affects the covariance in this case
  parc <- initparam$param[initparam$namescorr]
  inip <- initparam$param[!(names(initparam$param) %in% names(parc))]
  mean_names <- .GeoMean_names(ncol(initparam$X))
  other_nuis2 <- inip[!(names(inip) %in% mean_names)]
  ## Build the ZIP nuisance vectors in the exact positional order expected by
  ## CorrelationMat_dis2.  StartParam() deliberately carries an auxiliary
  ## Gaussian `nugget` for Simulation-input validation; for double-nugget ZIP
  ## models that auxiliary entry must never reach the C kernel.
  zip_order <- switch(as.character(initparam$model),
    "43" = c("nugget1", "nugget2", "pmu"),
    "44" = c("nugget1", "nugget2", "pmu"),
    "45" = c("nugget1", "nugget2", "pmu"),
    "57" = c("nugget1", "nugget2", "pmu", "shape"),
    "58" = c("nugget", "pmu", "shape"),
    NULL)
  if(!is.null(zip_order)) {
    missing_zip <- setdiff(zip_order, names(other_nuis2))
    if(length(missing_zip))
      stop("GeoCovmatrix: internal ZIP nuisance construction is missing ",
           paste(missing_zip, collapse=", "), ".", call.=FALSE)
    other_nuis <- as.numeric(other_nuis2[zip_order])
  } else {
    other_nuis <- as.numeric(other_nuis2[order(names(other_nuis2))])
  }
  if(!is.null(Mobs_cov)) {
    ML <- Mobs_cov
  } else {
    beta <- .GeoMean_beta(initparam$param, ncol(initparam$X))
    ML <- as.numeric(initparam$X %*% beta)
  }
                             } 
}


######  calling main correlation functions
if(is.null(copula)) {
    covmatrix<- Cmatrix(initparam$bivariate,cc[,1],cc[,2],cc[,3],initparam$coordt,initparam$corrmodel,dime,n,initparam$ns,
                        initparam$NS, initparam$param[initparam$namesnuis],
                        initparam$numpairs,numpairstot,initparam$model,
                        initparam$param[initparam$namescorr],setup,initparam$radius,initparam$spacetime,spacetime_dyn,initparam$type,copula,ML,other_nuis)
                    }
else {
    covmatrix<- Cmatrix_copula(initparam$bivariate,cc[,1],cc[,2],cc[,3],initparam$coordt,initparam$corrmodel,dime,n,initparam$ns,
                        initparam$NS, initparam$param[initparam$namesnuis],
                        initparam$numpairs,numpairstot,initparam$model,
                        initparam$param[initparam$namescorr],setup,initparam$radius,initparam$spacetime,spacetime_dyn,initparam$type,copula,ML,other_nuis)
                    }


    if(type=="Tapering") sparse=TRUE

    # Delete the global variables:
    if(!space)
                      .C('DeleteGlobalVar', PACKAGE='GeoModels', NAOK=TRUE)
    else                     
                      .C('DeleteGlobalVar2', PACKAGE='GeoModels', NAOK=TRUE)
 ##### setting names nuis parameters
    if(initparam$bivariate)   initparam$numtime=2
    if(!initparam$bivariate){namesnuis = NuisParam(model, bivariate,ncol(initparam$X),copula)}
    else                    {namesnuis =initparam$namesnuis}

if(is.null(X)) initparam$X=NULL
#######################################
    # Return the objects list:
    CovMat <- list(bivariate =  initparam$bivariate,
                   coordx = coords_orig[,1],
                   coordy = coords_orig[,2],
                   coordz = coordz_orig,
                   coordt = initparam$coordt,
                   coordx_dyn = coordx_dyn_orig,
                   covmatrix=covmatrix,
                   copula=copula,
                   corrmodel = corrmodel,
                   distance = distance,
                   grid=   grid_input,
                   nozero=initparam$setup$nozero,
                   n=n,
                   ns=initparam$ns,
                   NS=initparam$NS,
                   model=initparam$model,
                   namescorr = initparam$namescorr,
                   namesnuis=namesnuis,
                   namessim = initparam$namessim,
                   numblock = initparam$numblock,
                   numcoord = initparam$numcoord,
                   numcoordx = initparam$numcoordx,
                   numcoordy = initparam$numcoordy,
                   numtime = initparam$numtime,
                   param = param,
                   ## Kept separate because param must remain a scalar parameter
                   ## vector for covariance and kriging internals.
                   mean_external = if(!initparam$bivariate) Mobs_cov else NULL,
                   radius = radius,
                   setup=setup,
                   spacetime = initparam$spacetime,
                   sparse=sparse,
    
                   X=initparam$X)
    CovMat$call <- call
    structure(CovMat, class = c("GeoCovmatrix"))
}

print.GeoCovmatrix <- function(x, digits = max(3, getOption("digits") - 3), ...)
  {
  is_bivariate <- isTRUE(x$bivariate)
  if(is_bivariate) biv <- "Bivariate" else biv <- "Univariate"

  if(is_bivariate) type <- "Spatial Bivariate"
  else if(isTRUE(x$spacetime)) type <- "Spatio-temporal"
  else type <- "Spatial"

  dd <- dim(x$covmatrix)
  dime <- if(length(dd)) dd[1L] else length(x$covmatrix)

  model <- paste0("model code ", x$model)
  if(x$model==1) model <- 'Gaussian'
  if(x$model==21) model <- 'Gamma'
  if(x$model==39) model <- 'TwoPieceBimodal'
  if(x$model==24) model <- 'LogLogistic'
  if(x$model==30) model <- 'Poisson'
  if(x$model==46) model <- 'PoissonGamma'
  if(x$model==43) model <- 'PoissonZIP'
  if(x$model==50) model <- 'Beta2'
  if(x$model==12) model <- 'StudentT'
  if(x$model==9) model <- 'Tukeygh '
  if(x$model==18) model <- 'SkewStudentT'
  if(x$model==25) model <- 'Logistic'
  if(x$model==34) model <- 'Tukeyh'
  if(x$model==40) model <- 'Tukeyh2'
  if(x$model==23) model <- 'Gamma2'
  if(x$model==22) model <- 'LogGaussian'
  if(x$model==10) model <- 'SkewGaussian'
  if(x$model==59) model <- 'SkewLaplace'
  if(x$model==27) model <- 'TwoPieceStudentT'
  if(x$model==38) model <- 'TwoPieceTukeyh'
  if(x$model==29) model <- 'TwoPieceGaussian'
  if(x$model==20) model <- 'SinhAsinh'
  if(x$model==13) model <- 'Wrapped'
  if(x$model==26) model <- 'Weibull'
  if(x$model==11) model <- 'Binomial'
  if(x$model==49) model <- 'BinomialLogistic'
  if(x$model==33) model <- 'Kumaraswamy'
  if(x$model==42) model <- 'Kumaraswamy2'
  if(x$model==28) model <- 'Beta'
  if(x$model==31) model <- 'Binomial_TwoPieceGauss'
  if(x$model==32) model <- 'BinomialNeg_TwoPieceGauss'
  if(x$model==19) model <- 'Binomial2'
  if(x$model==16) model <- 'BinomialNeg'
  if(x$model==45) model <- 'BinomialNegZINB'
  if(x$model==14) model <- 'Geometric'
  if(x$model==15) model <- 'PoisBin'
  if(x$model==17) model <- 'PoisBinNeg'

    cat('\n##################################################################')
    cat('\n Covariance matrix type:', type,'\n')
    cat('\n Dimension:',dime,'X',dime,'\n')
    cat('\n Model:', model, '\n')
    if(!is.null(x$copula)) cat('\n Copula:', x$copula, '\n')
    cat('\n Correlation model:', x$corrmodel, '\n')
    if(!is.null(x$coordx_dyn)) {
      cat('\n Spatial block sizes:', paste(x$ns, collapse = ', '), '\n')
    } else {
      cat('\n Number of spatial coordinates:', x$numcoord, '\n')
    }
    if(isTRUE(x$spacetime))
      cat('\n Number of dependent temporal realisations:', length(x$coordt), '\n')
    cat('\n Type of the random field:', biv, '\n')
    cat('\n Number of parameters:', length(x$param), '\n')
    cat('\nParameters:\n')
    print.default(unlist(x$param), digits = digits, print.gap = 2,
                  quote = FALSE)
    cat('\n##################################################################\n')
    invisible(x)
  }

####################################################
### File name: GeoKrigloc.r
####################################################


###########################################################################
##  Future backend helpers                                                ##
###########################################################################

.GeoKriglocSetFuturePlan <- function(workers) {
  backend <- tolower(as.character(
    getOption("GeoModels.GeoKrigloc_future_backend", "auto")
  )[1L])
  if (!(backend %in% c("auto", "multicore", "multisession"))) {
    stop(
      "GeoKrigloc: option 'GeoModels.GeoKrigloc_future_backend' must be one of 'auto', 'multicore', or 'multisession'.",
      call. = FALSE
    )
  }

  can_multicore <- tryCatch(
    isTRUE(future::supportsMulticore()),
    error = function(e) FALSE
  )
  if ((identical(backend, "multicore") ||
       (identical(backend, "auto") && can_multicore)) && can_multicore) {
    future::plan(future::multicore, workers = workers)
    return("multicore")
  }
  if (identical(backend, "multicore") && !can_multicore) {
    warning(
      "GeoKrigloc: multicore futures are not supported in this R session; using multisession.",
      call. = FALSE
    )
  }

  libs <- .libPaths()
  pkg_dir <- tryCatch(find.package("GeoModels", quiet = TRUE),
                      error = function(e) character(0))
  if (length(pkg_dir) == 1L && nzchar(pkg_dir)) libs <- c(dirname(pkg_dir), libs)
  libs <- unique(vapply(
    libs[nzchar(libs)], normalizePath, character(1L),
    winslash = "/", mustWork = FALSE
  ))
  future::plan(future::multisession, workers = workers, rscript_libs = libs)
  "multisession"
}

###########################################################################
##  Top-level worker — defined OUTSIDE GeoKrigloc 
###########################################################################

.perform_kriging_worker <- function(k, neigh, param, corrmodel, distance,
                                    n, nloc, model, type_krig, sparse, anisopars,
                                    radius, mse, copula, varcov, method,
                                    loc, time, Xloc, Mloc, M,
                                    task_indices, is_space, spacetime,
                                    which) {
  .dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
  on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)


  if (!is.null(neigh$numpoints) && length(neigh$numpoints) >= k &&
      neigh$numpoints[k] == 0L) {
    return(list(indices = as.integer(k), pred = NA_real_,
                mse = if (mse) NA_real_ else NULL))
  }

  local_param <- param
  if (!is.null(M)) local_param$mean <- neigh$M[[k]]

  Xloc_k <- if (!is.null(Xloc)) Xloc[k, , drop = FALSE] else NULL

  Mloc_k <- if (!is.null(Mloc)) Mloc[k] else NULL
  nloc_k <- if (is.null(nloc) || length(nloc) == 1L) nloc else nloc[k]
  n_k <- n
  if (length(n) > 1L && !is.null(neigh$indices) && length(neigh$indices) >= k) {
    idx_n <- neigh$indices[[k]]
    if (length(n) >= max(idx_n, 0L)) n_k <- n[idx_n]
  }

  if (is_space) {
    pr <- GeoKrig(
      loc       = loc[k, ],
      data      = neigh$data[[k]],
      coordx    = neigh$coordx[[k]],
      corrmodel = corrmodel,
      distance  = distance,
      n         = n_k,
      nloc      = nloc_k,
      method    = method,
      X         = neigh$X[[k]],
      Xloc      = Xloc_k,
      Mloc      = Mloc_k,
      type_krig = type_krig,
      sparse    = sparse,
      model     = model,
      param     = local_param,
      anisopars = anisopars,
      radius    = radius,
      mse       = mse,
      copula    = copula,
      varcov    = varcov
    )

  } else {
    i <- task_indices[k, 1L]
    j <- task_indices[k, 2L]

    if (spacetime) {
      coordx_dyn_k <- if (!is.null(neigh$coordx_dyn)) neigh$coordx_dyn[[k]] else NULL
      pr <- GeoKrig(
        data      = neigh$data[[k]],
        coordx    = if (is.null(coordx_dyn_k)) neigh$coordx[[k]] else NULL,
        coordt    = neigh$coordt[[k]],
        coordx_dyn = coordx_dyn_k,
        loc       = loc[i, ],
        time      = time[j],
        X         = neigh$X[[k]],
        Mloc      = Mloc_k,
        Xloc      = Xloc_k,
        type_krig = type_krig,
        sparse    = sparse,
        corrmodel = corrmodel,
        distance  = distance,
        model     = model,
        param     = local_param,
        anisopars = anisopars,
        radius    = radius,
        mse       = mse,
        varcov    = varcov,
        copula    = copula,
        n         = n_k,
        nloc      = nloc_k,
        method    = method
      )
    } else {
      # bivariate: preserve the two variable-specific coordinate blocks.
      coordx_dyn_k <- if (!is.null(neigh$coordx_dyn)) neigh$coordx_dyn[[k]] else NULL
      if (!is.null(neigh$block_sizes) && any(neigh$block_sizes[[k]] == 0L)) {
        return(list(pred = NA_real_, mse = if (mse) NA_real_ else NULL))
      }
      pr <- GeoKrig(
        data      = neigh$data[[k]],
        coordx    = if (is.null(coordx_dyn_k)) neigh$coordx[[k]] else NULL,
        coordx_dyn = coordx_dyn_k,
        loc       = loc[i, ],
        X         = neigh$X[[k]],
        Mloc      = Mloc_k,
        Xloc      = Xloc_k,
        type_krig = type_krig,
        sparse    = sparse,
        corrmodel = corrmodel,
        distance  = distance,
        model     = model,
        param     = local_param,
        varcov    = varcov,
        anisopars = anisopars,
        radius    = radius,
        mse       = mse,
        copula    = copula,
        n         = n_k,
        nloc      = nloc_k,
        method    = method,
        which     = which
      )
    }
  }

  list(indices = as.integer(k), pred = pr$pred,
       mse = if (mse) pr$mse else NULL)
}


###########################################################################
##  Main function                                                        ##
###########################################################################

GeoKrigloc = function(estobj=NULL, data, coordx=NULL, coordy=NULL, coordz=NULL, coordt=NULL, coordx_dyn=NULL,
                      corrmodel, distance="Eucl", grid=FALSE, loc, neighb=NULL,
                      maxdist=NULL, maxtime=NULL, method="cholesky", model="Gaussian",
                      n=1, nloc=NULL, mse=FALSE, param, anisopars=NULL,
                      radius=1, sparse=FALSE, time=NULL, type="Standard",
                      type_krig="Simple", weigthed=TRUE, which=1,
                      copula=NULL, X=NULL, Xloc=NULL, Mloc=NULL, varcov=NULL,
                      spobj=NULL, spdata=NULL, parallel=FALSE, ncores=6, progress=TRUE,
                      check.duplicates=FALSE)
{


  ###########################################################################
  ##  0. Parallel helpers                                                  ##
  ###########################################################################

  setup_parallel <- function(workers) {
    old_opts <- options(
      warn = -1,
      verbose = FALSE,
      future.globals.maxSize = 4 * 1024^3
    )
    old_plan <- future::plan()
    .GeoKriglocSetFuturePlan(workers)
    list(old_opts = old_opts, old_plan = old_plan)
  }

  cleanup_parallel <- function(cfg) {
    options(cfg$old_opts)
    try(future::plan(cfg$old_plan), silent = TRUE)
    invisible(gc(verbose = FALSE))
  }

  choose_workers_safe <- function(neigh, max_cores, n_jobs,
                                      payload = neigh) {
    max_cores <- min(as.integer(max_cores), max(1L, as.integer(n_jobs)))
    if (max_cores <= 1L) return(1L)

    nn <- if (!is.null(neigh$numpoints)) {
      suppressWarnings(max(as.integer(neigh$numpoints), na.rm = TRUE))
    } else if (!is.null(neigh$indices)) {
      suppressWarnings(max(vapply(neigh$indices, length, integer(1L)),
                           na.rm = TRUE))
    } else if (!is.null(neigh$coordx)) {
      suppressWarnings(max(vapply(neigh$coordx, function(z) {
        if (is.null(z)) 0L else nrow(as.matrix(z))
      }, integer(1L)), na.rm = TRUE))
    } else 0L
    if (!is.finite(nn) || nn < 1L) nn <- 1L

    p <- if (!is.null(neigh$X)) {
      suppressWarnings(max(vapply(neigh$X, function(z) {
        if (is.null(z)) 0L else ncol(as.matrix(z))
      }, integer(1L)), na.rm = TRUE))
    } else 0L
    if (!is.finite(p) || p < 0L) p <- 0L

    ## Conservative dense-local estimate: covariance, factor/solve workspaces,
    ## cross-covariance/design temporaries and R object overhead.
    nn <- as.double(nn)
    p <- as.double(p)
    bytes_per_job <- 8 * (nn^2 + nn * max(1, p) + 8 * nn) * 4
    if (!is.finite(bytes_per_job) || bytes_per_job <= 0) return(1L)

    ## Multisession workers each receive their own neighborhood object;
    ## multicore workers can initially share it through copy-on-write.
    backend <- tolower(as.character(
      getOption("GeoModels.GeoKrigloc_future_backend", "auto")
    )[1L])
    can_multicore <- tryCatch(
      isTRUE(future::supportsMulticore()), error = function(e) FALSE
    )
    multisession <- identical(backend, "multisession") ||
      (identical(backend, "auto") && !can_multicore)
    if (multisession) {
      bytes_per_job <- bytes_per_job + as.numeric(object.size(payload))
    }

    free_ram <- tryCatch({
      if (requireNamespace("memuse", quietly = TRUE)) {
        as.numeric(memuse::Sys.meminfo()["freeram"], units = "b")
      } else 2 * 1024^3
    }, error = function(e) 2 * 1024^3)

    safe <- max(1L, as.integer(floor(0.70 * free_ram / bytes_per_job)))
    min(max_cores, safe)
  }


  ###########################################################################
  ##  1. Input processing                                                  ##
  ###########################################################################

  call = match.call()

  if (!is.null(estobj)) {
    if (!inherits(estobj, "GeoFit")) stop("need a 'GeoFit' object as input\n")
    data = estobj$data
    if (!estobj$grid) {
      if (!estobj$bivariate) {
        if (is.null(estobj$coordx_dyn))
          coordx = cbind(estobj$coordx, estobj$coordy, estobj$coordz)
        else
          coordx_dyn = estobj$coordx_dyn
      } else {
        if (is.null(estobj$coordx_dyn)) {
          coordx = estobj$coordx[1:estobj$ns[1]]
          coordy = estobj$coordy[1:estobj$ns[2]]
        } else {
          coordx_dyn = estobj$coordx_dyn
        }
      }
    } else {
      coordx = estobj$coordx
      coordy = estobj$coordy
      coordz = estobj$coordz
    }
    if (isTRUE(estobj$spacetime)) coordt = estobj$coordt else coordt = NULL
    coordx_dyn = estobj$coordx_dyn
    corrmodel  = estobj$corrmodel
    model      = estobj$model
    distance   = estobj$distance
    grid       = estobj$grid
    n          = estobj$n
    param      = append(estobj$param, estobj$fixed)
    radius     = estobj$radius
    copula     = estobj$copula
    anisopars  = estobj$anisopars
    X          = estobj$X
    varcov     = estobj$varcov
    if (!is.null(estobj$sparse)) sparse = estobj$sparse
  }

  model_requested <- gsub("[[:blank:]]", "", as.character(model)[1L])
  model <- .GeoPredictionCanonicalModel(model_requested)
  if (is.null(CkModel(model))) stop("The name of the model is not correct")
  if (!is.character(corrmodel) || is.null(CkCorrModel(corrmodel)))
    stop("The name of the correlation model is wrong")
  corrmodel <- gsub("[[:blank:]]", "", corrmodel)
  if (!is.null(copula)) {
    copula <- as.character(copula)[1L]
    if (is.na(copula) || !nzchar(gsub("[[:blank:]]", "", copula))) {
      copula <- NULL
    } else {
      copula <- gsub("[[:blank:]]", "", copula)
    }
  }
  if (!is.null(copula) && !(copula %in% c("Gaussian", "Clayton", "SkewGaussian")))
    stop("GeoKrigloc currently implements copula-based linear prediction for 'Gaussian', 'Clayton', and 'SkewGaussian'.", call. = FALSE)
  distance <- gsub("[[:blank:]]", "", distance)

  bivariate = CheckBiv(CkCorrModel(corrmodel))
  spacetime = CheckST(CkCorrModel(corrmodel))
  space     = !spacetime && !bivariate
  .GeoKrigValidatePredictionModel(model, copula = copula, bivariate = bivariate, context = "GeoKrigloc")

  method <- .GeoKrigMatchMethod(method, sparse = sparse, context = "GeoKrigloc")
  type_key <- tolower(gsub("[[:blank:]]", "", type))
  if (!identical(type_key, "standard")) {
    warning("GeoKrigloc currently computes the Standard local kriging system; ",
            "the supplied type is ignored.", call. = FALSE)
  }
  type <- "Standard"
  krig_key <- tolower(gsub("[[:blank:]]", "", type_krig))
  if (!(krig_key %in% c("simple", "universal")))
    stop("type_krig must be 'Simple' or 'Universal'")
  type_krig <- if (krig_key == "simple") "Simple" else "Universal"
  if (!is.logical(weigthed) || length(weigthed) != 1L || is.na(weigthed))
    stop("weigthed must be TRUE or FALSE")
  if (!isTRUE(weigthed))
    warning("weigthed is retained for backward compatibility and has no effect for type = 'Standard'.", call. = FALSE)

  ## GeoFit stores a single spatial realization as a 1 x n matrix.
  ## GeoNeighborhood expects spatial data as a vector (or an n-row object).
  ## Normalize only this representation; space-time and bivariate formats
  ## must retain their matrix/list structure.
  if (space && is.matrix(data) && nrow(data) == 1L) {
    data <- as.numeric(data[1L, , drop = TRUE])
  }

  if (!is.null(spobj)) {
    a = sp2Geo(spobj, spdata); coordx = a$coords
    if (!a$pj && distance != "Chor") distance = "Geod"
    if (spacetime) coordt = a$coordt
    if (!is.null(a$Y)) data <- a$Y
   if (!is.null(a$X)) X <- a$X
  }

  if (spacetime && !is.null(coordx_dyn)) {
    coordx_dyn <- .GeoNormalizeDynamicCoordinates(
      coordx_dyn, coordt = coordt, context = "GeoKrigloc"
    )
    .GeoValidateDynamicData(data, coordx_dyn, context = "GeoKrigloc")
    .GeoValidateDynamicRows(X, coordx_dyn, name = "X", context = "GeoKrigloc")
  }

  biv_info <- NULL
  if (bivariate) {
    biv_info <- .GeoBivariateCoordinateInfo(
      coordx = coordx, coordy = coordy, coordz = coordz,
      coordx_dyn = coordx_dyn, grid = grid,
      context = "GeoKrigloc"
    )
    data <- .GeoBivariateDataVector(data, biv_info$ns,
                                    context = "GeoKrigloc")
    if (!is.null(X)) {
      X <- .GeoBivariateDesign(X, biv_info$ns,
                               context = "GeoKrigloc")
    }
    if (biv_info$dynamic) {
      coordx_dyn <- biv_info$blocks
      coordx <- coordy <- coordz <- NULL
    }
  }

  if (is.null(coordx_dyn)) {
    coords <- .GeoKrigCoordinateMatrix(
      coordx = coordx, coordy = coordy, coordz = coordz, grid = grid,
      context = "GeoKrigloc"
    )
    if (!bivariate) {
      data <- .GeoKrigNormalizeData(
        data, nspace = nrow(coords),
        coordt = if (spacetime) coordt else NULL,
        grid = grid, context = "GeoKrigloc"
      )
    }
  } else {
    coordx = coordy = coordz = coords = NULL
  }

  .GeoCheckDuplicateLocations(
    coordx = if (is.null(coordx_dyn)) coords else NULL,
    coordt = coordt, coordx_dyn = coordx_dyn,
    spacetime = spacetime, bivariate = bivariate, grid = FALSE,
    context = "GeoKrigloc", check.duplicates = check.duplicates
  )
  .dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
  on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)

  if (is.vector(loc) && !is.list(loc)) loc <- matrix(loc, nrow = 1L)
  loc <- as.matrix(loc)
  loc_orig <- loc
  if (!is.numeric(loc) || !nrow(loc) || any(!is.finite(loc)))
    stop("loc must be a finite numeric N x 2 or N x 3 matrix")
  if (spacetime) {
    if (is.null(time) || !is.numeric(time) || !length(time) || any(!is.finite(time)))
      stop("time must be a non-empty finite numeric vector for space-time kriging")
    time <- as.numeric(time)
  }

  .GeoValidate3D(
    coordx = if (is.null(coordx_dyn)) coords else NULL,
    coordx_dyn = coordx_dyn, distance = distance, grid = grid,
    context = "GeoKrigloc"
  )
  .GeoValidate3D(
    coordx = loc, distance = distance, grid = FALSE,
    context = "GeoKrigloc prediction locations"
  )
  .GeoValidateCoordinateCompatibility(
    coordx = if (is.null(coordx_dyn)) coords else NULL,
    coordx_dyn = coordx_dyn, loc = loc, context = "GeoKrigloc"
  )

  ## Select neighbors and compute covariances in exactly the same metric.
  coords_metric <- coords
  coordx_dyn_metric <- coordx_dyn
  loc_metric <- loc
  worker_anisopars <- anisopars
  if (!is.null(anisopars)) {
    if (!identical(distance, "Eucl"))
      stop("GeoKrigloc: anisopars requires distance = 'Eucl'.")
    if (!is.null(coords_metric))
      coords_metric <- GeoAniso(coords_metric, c(anisopars$angle, anisopars$ratio))
    if (!is.null(coordx_dyn_metric))
      coordx_dyn_metric <- lapply(coordx_dyn_metric, function(z)
        GeoAniso(as.matrix(z), c(anisopars$angle, anisopars$ratio)))
    loc_metric <- GeoAniso(loc_metric, c(anisopars$angle, anisopars$ratio))
    ## Coordinates sent to GeoKrig are already transformed.
    worker_anisopars <- NULL
    param$angle <- NULL
    param$ratio <- NULL
  }

  Nloc = nrow(loc)
  if (is.null(Nloc)) Nloc = 1
  if (spacetime && is.null(time))
    stop("At least one prediction time is required for space-time kriging.")
  Tloc = length(time)
  if (bivariate) Tloc = 1

  ## Normalize the same three mean specifications used by GeoKrig():
  ## intercept only, X beta, or an external observation mean plus Mloc.
  if (!bivariate) {
    nobs_mean <- length(unlist(data, use.names = FALSE))
    M <- .GeoMean_external(param, n = nobs_mean)
    if (!is.null(M) && !is.null(X))
      stop("A fixed observation mean and X cannot be used together.")
    if (!is.null(Mloc) && !is.null(Xloc))
      stop("Use either Mloc or Xloc, not both.")
    if (!is.null(M) && is.null(Mloc))
      stop("Mloc is required when the observation mean is supplied as a vector.")
    if (!is.null(X)) {
      X <- .GeoMean_design(X, n = nobs_mean, name = "X")
      p_mean <- ncol(X)
      invisible(.GeoMean_beta(param, p_mean))
      if (is.null(Xloc) && is.null(Mloc))
        stop("Xloc or Mloc is required when X is supplied.")
    } else {
      p_mean <- 1L
      if (!is.null(Xloc)) stop("X must be supplied when Xloc is used.")
      if (is.null(M)) {
        if (length(setdiff(.GeoMean_supplied_names(param), "mean")))
          stop("Parameters mean1, mean2, ... require an explicit design matrix X; without X the mean is intercept-only.")
        invisible(.GeoMean_beta(param, 1L))
      }
    }

    total_mean_tasks <- if (space) Nloc else Nloc * Tloc
    if (!is.null(Mloc)) {
      if (!is.numeric(Mloc) || any(!is.finite(Mloc)))
        stop("Mloc must be finite and numeric.")
      Mloc <- as.numeric(Mloc)
      if (length(Mloc) == 1L) Mloc <- rep(Mloc, total_mean_tasks)
      if (length(Mloc) != total_mean_tasks)
        stop("Mloc must have one value per local prediction task.")
      Xloc <- NULL
    } else if (!is.null(Xloc)) {
      Xloc <- .GeoMean_design(Xloc, n = total_mean_tasks, p = p_mean, name = "Xloc")
    }
  } else {
    M <- NULL
    p_mean <- if (is.null(X)) 1L else ncol(X)
    invisible(.GeoBivariateValidateMeanParameters(
      param, p_mean, context = "GeoKrigloc"
    ))
    if (!is.null(Mloc) && !is.null(Xloc))
      stop("Use either Mloc or Xloc, not both.")
    if (!is.null(X) && is.null(Xloc) && is.null(Mloc))
      stop("Xloc or Mloc is required when X is supplied.")
    if (is.null(X) && !is.null(Xloc))
      stop("X must be supplied when Xloc is used.")
    if (!is.null(Mloc)) {
      if (!is.numeric(Mloc) || any(!is.finite(Mloc)))
        stop("Mloc must be finite and numeric.")
      Mloc <- as.numeric(Mloc)
      if (length(Mloc) == 1L) Mloc <- rep(Mloc, Nloc)
      if (length(Mloc) != Nloc)
        stop("Mloc must have one value per prediction location.")
      Xloc <- NULL
    } else if (!is.null(Xloc)) {
      Xloc <- .GeoMean_design(Xloc, n = Nloc, p = p_mean, name = "Xloc")
    }
  }

  if (identical(type_krig, "Universal") && (!is.null(M) || !is.null(Mloc))) {
    warning("External observation/prediction means are treated as known; ",
            "type_krig is therefore set to 'Simple'.", call. = FALSE)
    type_krig <- "Simple"
  }
  if (identical(type_krig, "Universal") && isTRUE(mse) && is.null(varcov)) {
    stop("Universal kriging with mse = TRUE requires varcov (the covariance matrix of the estimated parameters).")
  }


  ###########################################################################
  ##  2. Pre-compute task indices for spacetime / bivariate               ##
  ###########################################################################

  task_indices <- if (spacetime || bivariate) {
    idx <- matrix(0L, nrow = Nloc * Tloc, ncol = 2L)
    k   <- 1L
    for (i in seq_len(Nloc)) {
      for (j in seq_len(Tloc)) {
        idx[k, 1L] <- i
        idx[k, 2L] <- j
        k <- k + 1L
      }
    }
    idx
  } else {
    NULL
  }


  ###########################################################################
  ##  3. Neighbourhood setup                                               ##
  ###########################################################################

  if (space) {
    neigh       = GeoNeighborhood(data, coordx = coords_metric, distance = distance,
                                  loc = loc_metric, neighb = neighb, maxdist = maxdist,
                                  X = X, M = M, parallel = FALSE, ncores = ncores)
    total_tasks <- Nloc

  } else if (spacetime) {
    neigh       = GeoNeighborhood(data, coordx = coords_metric, coordt = coordt,
                                  distance = distance, neighb = neighb,
                                  coordx_dyn = coordx_dyn_metric, loc = loc_metric, time = time,
                                  maxdist = maxdist, maxtime = maxtime,
                                  X = X, M = M, parallel = FALSE, ncores = ncores)
    total_tasks <- Nloc * Tloc

  } else {   # bivariate
    neigh       = GeoNeighborhood(data, coordx = coords_metric, distance = distance,
                                  neighb = neighb, coordx_dyn = coordx_dyn_metric,
                                  bivariate = TRUE,
                                  loc = loc_metric, maxdist = maxdist,
                                  X = X, M = M, parallel = FALSE, ncores = ncores)
    total_tasks <- Nloc * Tloc
  }


  ###########################################################################
  ##  4. Progress bar setup                                                ##
  ###########################################################################

  if (!is.logical(progress)) stop("progress must be logical (TRUE/FALSE)")

  if (progress) {
    restore_progress_handlers <- .GeoProgressHandlersPush(
      handler = "txtprogressbar", global = TRUE
    )
    on.exit(restore_progress_handlers(), add = TRUE)
    pb <- progressr::progressor(
      along = seq_len(total_tasks), on_exit = FALSE, auto_finish = FALSE
    )
    ## Finish before restoring the temporary global handler, including on
    ## errors or user interrupts.
    on.exit(pb(type = "finish"), add = TRUE, after = FALSE)
  } else {
    pb <- function(...) {}
  }

  # Pre-allocate result vectors
  res1 <- numeric(total_tasks)
  res2 <- if (mse) numeric(total_tasks) else NULL


  ###########################################################################
  ##  5. Kriging jobs                                                      ##
  ###########################################################################

  ## GeoKrigloc deliberately uses GeoKrig as its single local prediction
  ## engine.  GeoNeighborhood defines each local data set; GeoKrig then
  ## performs exactly the same validated kriging calculation used elsewhere
  ## in the package.  Keeping one numerical path avoids duplicating covariance,
  ## copula, mean, MSE, and solver logic inside GeoKrigloc.
  jobs <- seq_len(total_tasks)
  worker_args <- list(
    neigh = neigh, param = param, corrmodel = corrmodel,
    distance = distance, n = n, nloc = nloc, model = model,
    type_krig = type_krig, sparse = sparse,
    anisopars = worker_anisopars, radius = radius, mse = mse,
    copula = copula, varcov = varcov, method = method,
    loc = loc_metric, time = time, Xloc = Xloc, Mloc = Mloc, M = M,
    task_indices = task_indices, is_space = space,
    spacetime = spacetime, which = which
  )


  ###########################################################################
  ##  6. Execute jobs                                                       ##
  ###########################################################################

  parallel_exec <- isTRUE(parallel) && total_tasks > 1L
  workers <- 1L
  if (parallel_exec) {
    requested <- .GeoResolveWorkers(TRUE, ncores, n_jobs = total_tasks)

    ## A numeric ncores value is an explicit worker request.  Respect it up to
    ## the number of detected cores and available prediction tasks.  The RAM
    ## safety heuristic is reserved for ncores = NULL, which remains available
    ## as an opt-in automatic/safe mode.
    if (is.null(ncores)) {
      workers <- choose_workers_safe(
        neigh, requested, total_tasks, payload = worker_args
      )
      if (workers < requested) {
        warning("Reducing cores from ", requested, " to ", workers,
                " to avoid memory pressure.")
      }
    } else {
      workers <- requested
    }
    parallel_exec <- workers > 1L
  }

  if (parallel_exec) {
    config <- setup_parallel(workers)
    on.exit(cleanup_parallel(config), add = TRUE)
    if (progress) cat("Performing local kriging using", workers, "cores...\n")

    res <- future.apply::future_lapply(
      jobs,
      function(k) {
        worker <- get(
          ".perform_kriging_worker",
          envir = asNamespace("GeoModels"),
          inherits = FALSE
        )
        ans <- do.call(worker, c(list(k = as.integer(k)), worker_args))
        if (progress) pb(sprintf("k=%d", k))
        ans
      },
      future.globals = c("worker_args", "progress", "pb"),
      future.packages = "GeoModels",
      future.seed = TRUE,
      future.stdout = FALSE,
      future.conditions = if (progress) "condition" else "none"
    )
  } else {
    res <- vector("list", total_tasks)
    for (k in jobs) {
      ans <- do.call(
        .perform_kriging_worker,
        c(list(k = as.integer(k)), worker_args)
      )
      res[[k]] <- ans
      if (progress) pb(sprintf("k=%d", k))
    }
  }

  for (ans in res) {
    res1[ans$indices] <- as.numeric(ans$pred)
    if (mse) res2[ans$indices] <- as.numeric(ans$mse)
  }


  ###########################################################################
  ##  7. Format output                                                     ##
  ###########################################################################

  varpred <- NULL

  if (spacetime || bivariate) {
    pred <- matrix(res1, nrow = Tloc, ncol = Nloc)
    if (mse) varpred <- matrix(res2, nrow = Tloc, ncol = Nloc)
  } else {
    pred <- res1
    if (mse) varpred <- res2
  }

  if (Tloc == 1) {
    pred    <- c(pred)
    varpred <- c(varpred)
  }

  GeoKrigloc_out <- list(
    bivariate = bivariate,
    coordx    = coordx,
    coordy    = coordy,
    coordz    = coordz,
    coordt    = coordt,
    coordx_dyn = coordx_dyn,
    corrmodel = corrmodel,
    data      = data,
    distance  = distance,
    grid      = grid,
    loc       = loc_orig,
    copula    = copula,
    numcoord  = if (exists("coords", inherits = FALSE)) nrow(coords) else NA,
    numloc    = Nloc,
    numtime   = length(coordt),
    numt      = Tloc,
    maxdist   = maxdist,
    maxtime   = maxtime,
    model     = model,
    fit_model = model_requested,
    n         = n,
    param     = param,
    pred      = pred,
    radius    = radius,
    spacetime = spacetime,
    time      = time,
    type       = type,
    type_krig = type_krig,
    method     = method,
    weigthed   = weigthed,
    which     = which,
    mse       = varpred,
    varcov    = varcov
  )

  structure(c(GeoKrigloc_out, call = call), class = c("GeoKrigloc"))
}

# Estimates J = Var{score(theta_hat)} using GeoFit(score=TRUE),
# then G^{-1} = H^{-1} J H^{-1}.

# Library paths to use for multisession workers.  In addition to the
# current .libPaths(), explicitly prepend the library that contains the
# GeoModels namespace currently in use.  This makes bootstrap parallelism
# robust when GeoModels is installed in a personal/non-standard library and
# also avoids relying on version-specific future inheritance of .libPaths().
.GeoVarest_future_libpaths <- function() {
  libs <- .libPaths()

  pkg_dir <- tryCatch(
    find.package("GeoModels", quiet = TRUE),
    error = function(e) character(0)
  )

  if (length(pkg_dir) == 1L && nzchar(pkg_dir)) {
    libs <- c(dirname(pkg_dir), libs)
  }

  libs <- libs[nzchar(libs)]
  libs <- vapply(
    libs,
    function(z) normalizePath(z, winslash = "/", mustWork = FALSE),
    character(1L),
    USE.NAMES = FALSE
  )
  unique(libs)
}

# Configure the future backend used by GeoVarest bootstrap work.  On Unix-like
# systems a forked multicore backend avoids serializing the fitted model and
# reloading GeoModels in every worker, which is materially faster for large TB
# jobs.  future::supportsMulticore() already disables this route in environments
# where forking is unsafe (for example some IDE configurations).  Windows and
# unsupported environments transparently fall back to multisession with the
# explicit library-path propagation required by GeoModels.
.GeoVarest_set_future_plan <- function(workers) {
  backend <- tolower(as.character(
    getOption("GeoModels.GeoVarest_future_backend", "auto")
  )[1L])

  if (!(backend %in% c("auto", "multicore", "multisession"))) {
    stop(
      "GeoVarest: option 'GeoModels.GeoVarest_future_backend' must be one of 'auto', 'multicore', or 'multisession'.",
      call. = FALSE
    )
  }

  can_multicore <- tryCatch(
    isTRUE(future::supportsMulticore()),
    error = function(e) FALSE
  )

  use_multicore <- identical(backend, "multicore") ||
    (identical(backend, "auto") && can_multicore)

  if (use_multicore && can_multicore) {
    future::plan(future::multicore, workers = workers)
    return("multicore")
  }

  if (identical(backend, "multicore") && !can_multicore) {
    warning(
      "GeoVarest: multicore futures are not supported in this R session; using multisession.",
      call. = FALSE
    )
  }

  future::plan(
    future::multisession,
    workers = workers,
    rscript_libs = .GeoVarest_future_libpaths()
  )
  "multisession"
}


# Return the exact pair graph stored by the original pairwise fit.  Reusing
# it avoids rebuilding nearest-neighbour indices for every bootstrap sample.
# For old deterministic fits that do not contain valid indices, the caller can
# safely fall back to rebuilding the deterministic graph.  Missing indices are
# an error only for stochastic thinning, because rebuilding would change the
# estimator conditionally on the selected graph.
.GeoVarest_selected_pairs <- function(fit) {
  thin_method <- if (is.null(fit$thin_method)) "bernoulli" else fit$thin_method
  thin_method <- tolower(as.character(thin_method))[1L]
  p_neighb <- if (is.null(fit$p_neighb)) 1 else fit$p_neighb
  p_neighb <- as.numeric(p_neighb)[1L]
  stochastic <- identical(thin_method, "match") ||
    (thin_method %in% c("bernoulli", "fixedbudget", "fixed_budget", "targetbalanced") &&
       is.finite(p_neighb) && p_neighb < 1)

  if (!identical(as.character(fit$type)[1L], "Pairwise")) return(NULL)

  if (!is.null(fit$memdist) && !isTRUE(fit$memdist) && stochastic) {
    stop(
      "This GeoFit object was created with the deprecated memdist = FALSE path and does not contain a reusable stochastic-thinning graph. Refit the model with the current GeoFit implementation before calling GeoVarest or GeoVarestbootstrap.",
      call. = FALSE
    )
  }

  colidx <- fit$colidx
  rowidx <- fit$rowidx
  valid <- is.numeric(colidx) && is.numeric(rowidx) &&
    length(colidx) == length(rowidx) && length(colidx) >= 1L &&
    all(is.finite(colidx)) && all(is.finite(rowidx))

  if (!valid) {
    if (stochastic) {
      stop(
        "The original stochastic-thinning fit does not contain valid pair indices.",
        call. = FALSE
      )
    }
    return(NULL)
  }

  if (!is.integer(colidx)) colidx <- as.integer(colidx)
  if (!is.integer(rowidx)) rowidx <- as.integer(rowidx)

  list(colidx = colidx, rowidx = rowidx)
}

.GeoVarest_with_selected_pairs <- function(pair_index, expr, pair_info = NULL,
                                                fit_context = NULL) {
  run <- function() .GeoWithDuplicateCheckSuppressed(force(expr))
  if (is.null(pair_index) && is.null(fit_context)) return(run())

  old_options <- options(
    GeoModels.selected_pair_index = pair_index,
    GeoModels.selected_pair_info = pair_info,
    GeoModels.fit_context_cache = fit_context
  )
  on.exit(options(old_options), add = TRUE)
  run()
}

# Precompute transient pair distances once per worker for the common purely
# spatial univariate pairwise case.  The cache lives only for the worker call;
# it is never stored in the returned GeoFit object.  More complicated layouts
# retain exact pair-index reuse and let StartParam rebuild their distances.
.GeoVarest_prepare_pair_info <- function(pair_index, coords, fit_spec) {
  if (is.null(pair_index) ||
      !identical(as.character(fit_spec$type)[1L], "Pairwise") ||
      isTRUE(fit_spec$bivariate) || isTRUE(fit_spec$spacetime) ||
      (!is.null(fit_spec$coordt) && length(fit_spec$coordt) > 1L) ||
      !is.null(fit_spec$coordx_dyn)) {
    return(NULL)
  }

  coords <- as.matrix(coords)
  if (nrow(coords) < 1L ||
      max(c(pair_index$colidx, pair_index$rowidx)) > nrow(coords)) {
    return(NULL)
  }

  distance_code <- switch(
    if (is.null(fit_spec$distance)) "Eucl" else as.character(fit_spec$distance)[1L],
    eucl = 0L, Eucl = 0L,
    chor = 1L, Chor = 1L,
    geod = 2L, Geod = 2L,
    proj = 3L, Proj = 3L,
    NULL
  )
  if (is.null(distance_code) || identical(distance_code, 3L)) return(NULL)

  pair_fun <- getFromNamespace(".GeoSelectedPairInfo", "GeoModels")
  tryCatch(
    pair_fun(
      pair_index = pair_index,
      coords = coords,
      distance = distance_code,
      radius = if (is.null(fit_spec$radius)) 1 else fit_spec$radius,
      spacetime = FALSE,
      bivariate = FALSE,
      ns = NULL,
      coordt = NULL
    ),
    error = function(e) NULL
  )
}

.GeoVarest_release_fit_context <- function(fit_spec) {
  if (identical(as.character(fit_spec$type)[1L], "Pairwise")) {
    try(
      .C("DeleteGlobalVar2", PACKAGE = "GeoModels", NAOK = TRUE),
      silent = TRUE
    )
  }
  invisible(NULL)
}


.GeoVarest_estimation_model <- function(fit, context = "GeoVarest") {
  stored <- fit$estimation_model
  if (!is.null(stored) && length(stored) == 1L && !is.na(stored) && nzchar(stored)) {
    return(as.character(stored))
  }

  model <- as.character(fit$model)[1L]
  if (!isTRUE(fit$missp)) return(model)

  ## Backward compatibility for GeoFit objects created before
  ## estimation_model was stored explicitly.  BinomialNeg is ambiguous
  ## because both Gaussian- and binary-working misspecifications exist; such
  ## old objects must be refitted so that the inferential criterion is known.
  fallback <- c(
    StudentT = "Gaussian_misp_StudentT",
    Poisson = "Gaussian_misp_Poisson",
    PoissonZIP = "Gaussian_misp_PoissonZIP",
    PoissonGamma = "Gaussian_misp_PoissonGamma",
    Binomial = "Gaussian_misp_Binomial",
    SkewStudentT = "Gaussian_misp_SkewStudentT",
    Tukeygh = "Gaussian_misp_Tukeygh"
  )

  if (identical(model, "BinomialNeg")) {
    stop(
      context, ": this older misspecified BinomialNeg fit does not record the working likelihood. Refit it with the current GeoFit before bootstrap variance estimation.",
      call. = FALSE
    )
  }
  if (model %in% names(fallback)) return(unname(fallback[[model]]))

  stop(
    context, ": cannot recover the misspecified estimation model from this older GeoFit object. Refit it with the current GeoFit implementation.",
    call. = FALSE
  )
}

.GeoVarest_bootstrap_parameters <- function(fit) {
  start <- fit$param
  fixed <- fit$fixed
  anisopars <- fit$anisopars

  if (!is.null(anisopars)) {
    anames <- intersect(c("angle", "ratio"), names(unlist(anisopars, use.names = TRUE)))
    if (length(anames)) {
      if (!is.null(start)) start[intersect(names(start), anames)] <- NULL
      if (!is.null(fixed)) fixed[intersect(names(fixed), anames)] <- NULL
    }
  }

  list(
    start = start,
    fixed = fixed,
    sim_param = append(start, fixed),
    anisopars = anisopars
  )
}

.GeoVarest_validate_controls <- function(K, parallel, ncores, min_success_rate, context) {
  tol <- sqrt(.Machine$double.eps)
  if (!is.numeric(K) || length(K) != 1L || !is.finite(K) || K < 2 ||
      abs(K - round(K)) > tol) {
    stop(context, ": K must be an integer >= 2", call. = FALSE)
  }
  if (!is.logical(parallel) || length(parallel) != 1L || is.na(parallel)) {
    stop(context, ": parallel must be TRUE or FALSE", call. = FALSE)
  }
  ncores <- .GeoValidateNcores(ncores, paste0(context, ": ncores"))
  if (!is.numeric(min_success_rate) || length(min_success_rate) != 1L ||
      !is.finite(min_success_rate) || min_success_rate <= 0 || min_success_rate > 1) {
    stop(context, ": min_success_rate must be in (0,1]", call. = FALSE)
  }

  list(
    K = as.integer(round(K)),
    ncores = ncores,
    min_success_rate = as.numeric(min_success_rate)
  )
}

.GeoVarest_check_success <- function(n_successful, K, min_success_rate, context) {
  min_success <- max(2L, as.integer(ceiling(min_success_rate * K)))
  if (n_successful < min_success) {
    stop(
      sprintf(
        "%s: only %d of %d bootstrap replications succeeded (%.1f%%); at least %d (%.1f%%) are required.",
        context, n_successful, K, 100 * n_successful / K,
        min_success, 100 * min_success_rate
      ),
      call. = FALSE
    )
  }
  if (n_successful < K) {
    warning(
      sprintf("%s: %d of %d bootstrap replications failed; variance estimates use the %d successful replications.",
              context, K - n_successful, K, n_successful),
      call. = FALSE
    )
  }
  invisible(min_success)
}

.GeoVarest_preflight_simulation <- function(fit, method, model_sim, sim_param,
                                             context = "GeoVarest") {
  if (!is.null(fit$copula)) {
    if (!(method %in% c("cholesky", "TB"))) {
      stop(context, ": copula bootstrap simulation supports method='cholesky' or method='TB'.",
           call. = FALSE)
    }
    ## Copula simulations use their own latent construction. Validate the
    ## marginal at the copula layer and, for TB, validate the latent backend.
    model_check <- .GeoSimulationCanonicalModel(model_sim)
    tryCatch({
      .GeoValidateCopulaSimulationModel(model_check)
      .GeoValidateCopulaMarginal(model_check, sim_param)
      if (identical(method, "TB")) {
        if (isTRUE(fit$bivariate))
          stop("turning-bands copula simulation is univariate only.", call. = FALSE)
        if (isTRUE(fit$spacetime))
          stop("turning-bands copula simulation is not implemented for spatio-temporal models.", call. = FALSE)
        .GeoValidateTBDistance(
          if (is.null(fit$distance)) "Eucl" else fit$distance,
          context = context
        )
        .GeoValidateTBCorrelationModel(.GeoSimulationCanonicalCorrModel(fit$corrmodel), bivariate = FALSE)
      }
    }, error = function(e) {
      stop(context, ": fitted copula model cannot be simulated: ",
           conditionMessage(e), call. = FALSE)
    })
    return(invisible(TRUE))
  }

  simulator <- if (identical(method, "cholesky")) "GeoSim" else "GeoSimapprox"
  model_check <- .GeoSimulationCanonicalModel(model_sim)
  tryCatch({
    .GeoValidateDirectSimulationModel(model_check, simulator = simulator)
    .GeoValidateLatentIntegerParameters(model_check, sim_param)
    .GeoValidateSimulationMarginal(model_check, sim_param)
  }, error = function(e) {
    stop(context, ": fitted model cannot be simulated by ", simulator, ": ", conditionMessage(e), call. = FALSE)
  })
  invisible(TRUE)
}

.GeoVarest_reject_full <- function(fit, context) {
  if (identical(as.character(fit$likelihood)[1L], "Full") &&
      identical(as.character(fit$type)[1L], "Standard")) {
    stop(
      context,
      ": Full/Standard likelihood fits do not use bootstrap Godambe variance estimation. Refit with GeoFit(..., likelihood='Full', type='Standard', varest=TRUE) to obtain Hessian-based standard errors.",
      call. = FALSE
    )
  }
  invisible(TRUE)
}

## Evaluate an expression under a deterministic scalar seed without changing
## the RNG state seen by the caller.  GeoVarest uses this for per-replication
## bootstrap streams so results do not depend on worker scheduling.
.GeoVarest_with_rng_kind <- function(rng_kind, expr) {
  if (is.null(rng_kind) || !length(rng_kind)) return(force(expr))

  rng_kind <- as.character(rng_kind)
  if (length(rng_kind) < 2L || length(rng_kind) > 3L || any(!nzchar(rng_kind))) {
    stop("GeoVarest: invalid internal RNG kind specification.", call. = FALSE)
  }

  old_kind <- RNGkind()
  old_seed <- if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  } else NULL

  restore <- function() {
    do.call(RNGkind, as.list(old_kind))
    if (is.null(old_seed)) {
      if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
        rm(".Random.seed", envir = .GlobalEnv)
      }
    } else {
      assign(".Random.seed", old_seed, envir = .GlobalEnv)
    }
  }
  on.exit(restore(), add = TRUE)

  do.call(RNGkind, as.list(rng_kind))
  force(expr)
}

.GeoVarest_seeded_eval <- function(seed_value, expr, rng_kind = NULL) {
  if (is.null(seed_value) || length(seed_value) != 1L || !is.finite(seed_value)) {
    return(.GeoVarest_with_rng_kind(rng_kind, force(expr)))
  }

  seed_value <- as.integer(seed_value)

  .GeoVarest_with_rng_kind(rng_kind, {
    set.seed(seed_value)
    force(expr)
  })
}

## Simulate all TB replications assigned to one worker in one GeoSimapprox
## call.  GeoSimapprox receives the original per-replication seeds through an
## internal option and reseeds immediately around each do_one_sim() call.
## Therefore batching removes repeated setup/validation overhead without
## changing the simulated fields.  The hidden option below is useful for
## regression tests and emergency fallback; it is not part of the public API.
.GeoVarest_tb_simulate_batch <- function(ids, sim_spec, sim_seeds, L, rng_kind = NULL) {
  ids <- as.integer(ids)
  if (!length(ids)) return(list())

  GeoSimapprox_fun <- getFromNamespace("GeoSimapprox", "GeoModels")
  use_batch <- !identical(
    getOption("GeoModels.GeoVarest_tb_batch", FALSE),
    FALSE
  )

  make_args <- function(nrep) {
    args <- sim_spec
    args$nrep <- as.integer(nrep)
    args$progress <- FALSE
    args$method <- "TB"
    args$L <- L
    ## GeoVarest already parallelizes across persistent worker chunks.
    ## Avoid nested futures inside each batch.
    args$parallel <- FALSE
    args$ncores <- NULL
    args
  }

  normalize_data <- function(sim_obj, nrep) {
    if (nrep == 1L) return(list(sim_obj$data))
    if (!is.list(sim_obj$data) || length(sim_obj$data) != nrep) {
      stop("GeoSimapprox returned an unexpected batched data layout.", call. = FALSE)
    }
    sim_obj$data
  }

  if (use_batch) {
    old_opt <- options(
      GeoModels.GeoSimapprox_rep_seeds = as.integer(sim_seeds[ids])
    )
    on.exit(options(old_opt), add = TRUE)

    batch_try <- tryCatch({
      sim_obj <- .GeoVarest_with_rng_kind(
        rng_kind,
        do.call(
          GeoSimapprox_fun,
          make_args(length(ids))
        )
      )
      normalize_data(sim_obj, length(ids))
    }, error = function(e) e)

    ## Restore immediately before a possible legacy fallback; the on.exit
    ## above remains as protection against interrupts.
    options(old_opt)

    if (!inherits(batch_try, "error")) return(batch_try)
    ## Preserve the old per-replication failure behavior if an unexpected
    ## model/layout cannot use the batch path.
  }

  lapply(ids, function(k) {
    .GeoVarest_seeded_eval(
      sim_seeds[[k]],
      do.call(GeoSimapprox_fun, make_args(1L)),
      rng_kind = rng_kind
    )$data
  })
}

# Simulate exactly one TB replication under its deterministic replication seed.
# This is the default large-job path used by GeoVarest: a worker simulates one
# field, scores it immediately, releases it, and only then advances to the next
# replication.  Peak memory is therefore O(n) per worker rather than O(K*n),
# which is important when L is large (e.g. L >= 10000).
.GeoVarest_tb_simulate_one <- function(k, sim_spec, sim_seeds, L, rng_kind = NULL) {
  k <- as.integer(k)[1L]
  GeoSimapprox_fun <- getFromNamespace("GeoSimapprox", "GeoModels")

  args <- sim_spec
  args$nrep <- 1L
  args$progress <- FALSE
  args$method <- "TB"
  args$L <- L
  ## Outer GeoVarest parallelism already occupies the available workers.
  ## Nested TB futures would oversubscribe the machine.
  args$parallel <- FALSE
  args$ncores <- NULL

  .GeoVarest_seeded_eval(
    sim_seeds[[k]],
    do.call(GeoSimapprox_fun, args),
    rng_kind = rng_kind
  )$data
}

# Internal worker for TB-based GeoVarest calculations.
# Workers are persistent, but bootstrap fields are streamed one at a time:
# simulate -> score -> release.  This keeps memory bounded for large L/K while
# still reusing the selected pair graph, pair distances and GeoFit static setup
# within each worker.
.GeoVarest_score_worker <- function(ids, mode = c("tb", "file"),
                                      sim_spec, fit_spec, theta_hat,
                                      data_files = NULL, sim_seeds = NULL,
                                      L = NULL, rng_kind = NULL,
                                      progressor = NULL) {
  .dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
  on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)

  mode <- match.arg(mode)
  ids <- as.integer(ids)
  target_names <- names(theta_hat)
  ctx <- .GeoVarest_worker_context(fit_spec, sim_spec$coordx)
  on.exit(.GeoVarest_release_fit_context(fit_spec), add = TRUE)

  use_stream <- identical(mode, "tb") && !identical(
    getOption("GeoModels.GeoVarest_tb_stream", TRUE), FALSE
  )
  sim_data <- NULL
  if (identical(mode, "tb") && !use_stream) {
    sim_data <- .GeoVarest_tb_simulate_batch(
      ids, sim_spec, sim_seeds, L, rng_kind = rng_kind
    )
  }

  start <- .GeoVarest_start_from_theta(theta_hat, fit_spec$aniso_names)
  out <- vector("list", length(ids))
  names(out) <- as.character(ids)

  for (ii in seq_along(ids)) {
    k <- ids[[ii]]
    current_file <- if (identical(mode, "file")) data_files[[k]] else NULL

    out[[ii]] <- tryCatch({
      current_data <- if (identical(mode, "file")) {
        readRDS(current_file)
      } else if (use_stream) {
        .GeoVarest_tb_simulate_one(k, sim_spec, sim_seeds, L, rng_kind = rng_kind)
      } else {
        sim_data[[ii]]
      }

      ff <- .GeoVarest_run_fit(
        current_data, sim_spec$coordx, fit_spec, start,
        ctx$pair_info, ctx$fit_context, score = TRUE
      )
      .GeoVarest_extract_score(ff, target_names)
    }, error = function(e) {
      .GeoVarest_failed_result(target_names, e)
    }, finally = {
      if (!is.null(current_file)) try(unlink(current_file), silent = TRUE)
    })

    gc_every <- if (identical(mode, "tb")) 25L else 5L
    if (ii %% gc_every == 0L) gc(verbose = FALSE)
    if (!is.null(progressor)) {
      progressor(sprintf(if (identical(mode, "tb")) "replication %d" else "score %d", k))
    }
  }

  if (!is.null(sim_data)) rm(sim_data)
  gc(verbose = FALSE, full = identical(mode, "file"))
  out
}



GeoVarest <- function(fit, K = 100, sparse = FALSE,
                      method = c("cholesky", "TB", "CE"),
                      alpha = 0.95, L = 10000,
                      parallel = FALSE, ncores = 6, progress = TRUE,
                      seed = NULL, min_success_rate = 0.8) {
  .dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
  on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)


  `%||%` <- function(a, b) if (!is.null(a)) a else b

  method <- match.arg(method)

  .GeoVarest_reject_full(fit, "GeoVarest")
  ctl <- .GeoVarest_validate_controls(
    K, parallel, ncores, min_success_rate, "GeoVarest"
  )
  K <- ctl$K
  ncores <- ctl$ncores
  min_success_rate <- ctl$min_success_rate

  .GeoValidate3D(
    coordx = fit$coordx, coordy = fit$coordy, coordz = fit$coordz,
    coordx_dyn = fit$coordx_dyn,
    distance = if (is.null(fit$distance)) "Eucl" else fit$distance,
    grid = isTRUE(fit$grid),
    approximate = !identical(method, "cholesky"), method = method,
    context = "GeoVarest"
  )

  fit_dimension <- .GeoCoordinateDimension(
    fit$coordx, fit$coordy, fit$coordz, fit$coordx_dyn
  )
  if (identical(fit_dimension, 3L) && identical(method, "TB")) {
    if (isTRUE(fit$bivariate) || isTRUE(fit$spacetime)) {
      stop("GeoVarest: three-dimensional TB simulation is supported only for purely spatial univariate fits.",
           call. = FALSE)
    }
    if (!is.null(fit$anisopars)) {
      stop("GeoVarest: three-dimensional TB simulation does not support anisopars; transform the coordinates explicitly.",
           call. = FALSE)
    }
  }

  ## ---------------------------------------------------------------
  ## 0. Checks
  ## ---------------------------------------------------------------
  if (!is.logical(progress) || length(progress) != 1L) {
    stop("progress must be logical", call. = FALSE)
  }

  if (!(is.numeric(alpha) && length(alpha) == 1L && alpha > 0 && alpha < 1)) {
    stop("alpha must be a single numeric in (0,1)", call. = FALSE)
  }


  if (is.null(fit$sensmat)) {
    stop("Sensitivity matrix is missing: use sensitivity = TRUE in GeoFit", call. = FALSE)
  }

  restore_rng <- .GeoVarest_preserve_seed(seed)
  on.exit(restore_rng(), add = TRUE)

  ## ---------------------------------------------------------------
  ## Progress handlers
  ## ---------------------------------------------------------------
  use_progressr <- isTRUE(progress) && requireNamespace("progressr", quietly = TRUE)

  if (isTRUE(progress) && !use_progressr) {
    warning(
      "progress=TRUE but progressr is not available; using text messages only.",
      call. = FALSE
    )
  }

  if (use_progressr) {
    restore_progress_handlers <- .GeoProgressHandlersPush(
      handler = "txtprogressbar", global = TRUE
    )
    on.exit(restore_progress_handlers(), add = TRUE)
  }

  gm_fun <- function(name) {
    if (requireNamespace("GeoModels", quietly = TRUE) &&
        exists(name, envir = asNamespace("GeoModels"), inherits = FALSE)) {
      return(getFromNamespace(name, "GeoModels"))
    }
    if (exists(name, mode = "function", inherits = TRUE)) {
      return(get(name, mode = "function", inherits = TRUE))
    }
    stop("Function ", name, " is not available. Load GeoModels first.", call. = FALSE)
  }


  set_proj_env_for_workers <- function() {
    get_proj_path <- function() {
      path <- Sys.getenv("PROJ_DATA", unset = "")

      if (!nzchar(path)) {
        path <- Sys.getenv("PROJ_LIB", unset = "")
      }

     # if (!nzchar(path) && requireNamespace("sf", quietly = TRUE)) {
     #   path <- tryCatch(sf::sf_proj_info("path"), error = function(e) "")
     # }

      if (length(path) != 1L || !nzchar(path)) {
        return("")
      }

      normalizePath(path, winslash = "/", mustWork = FALSE)
    }

    proj_path <- get_proj_path()

    if (!nzchar(proj_path)) {
      return(character(0))
    }

    Sys.setenv(PROJ_DATA = proj_path)
    Sys.setenv(PROJ_LIB  = proj_path)

    c(PROJ_DATA = proj_path, PROJ_LIB = proj_path)
  }

  GeoFit_fun <- gm_fun("GeoFit")
  GeoSim_fun <- gm_fun("GeoSim")

  ## Use an internal representation for purely spatial models without
  ## modifying the GeoFit object returned to the user. GeoFit may store
  ## absence of time as the scalar 0, whereas GeoSim/GeoFit internal calls
  ## should receive NULL.
  coordt_use <- fit$coordt
  if (is.null(coordt_use) || length(coordt_use) <= 1L) {
    coordt_use <- NULL
  }

  fit$thin_method <- fit$thin_method %||% "bernoulli"
  fit$p_neighb    <- fit$p_neighb %||% 1
  selected_pair_index <- .GeoVarest_selected_pairs(fit)

  optimizer <- fit$optimizer %||% "Nelder-Mead"
  lower     <- fit$lower
  upper     <- fit$upper

  theta_hat <- as.numeric(unlist(fit$param, use.names = TRUE))
  names(theta_hat) <- names(unlist(fit$param, use.names = TRUE))

  num_params <- length(theta_hat)
  if (num_params < 1L) {
    stop("fit$param is empty", call. = FALSE)
  }

  H <- as.matrix(fit$sensmat)
  storage.mode(H) <- "double"

  if (!all(dim(H) == c(num_params, num_params))) {
    stop("fit$sensmat has incompatible dimension", call. = FALSE)
  }

  H <- (H + t(H)) / 2
  rownames(H) <- colnames(H) <- names(theta_hat)

  safe_inverse <- function(A, label) {
    A <- as.matrix(A)
    storage.mode(A) <- "double"
    A <- (A + t(A)) / 2

    out <- try(solve(A), silent = TRUE)

    if (!inherits(out, "try-error") && all(is.finite(out))) {
      return((out + t(out)) / 2)
    }

    ee <- eigen(A, symmetric = TRUE)
    tol <- max(dim(A)) * max(abs(ee$values)) * .Machine$double.eps
    keep <- abs(ee$values) > tol

    if (!any(keep)) {
      stop("Cannot invert ", label, call. = FALSE)
    }

    out <- ee$vectors[, keep, drop = FALSE] %*%
      diag(1 / ee$values[keep], nrow = sum(keep)) %*%
      t(ee$vectors[, keep, drop = FALSE])

    warning("Used spectral generalized inverse for ", label, call. = FALSE)
    (out + t(out)) / 2
  }

  geometry <- .GeoVarest_prepare_geometry(fit, coordt_use)
  dimat <- geometry$dimat
  X_use <- geometry$X
  coords <- geometry$coords

  selected_pair_info <- .GeoVarest_prepare_pair_info(
    selected_pair_index,
    coords,
    list(
      type = fit$type,
      bivariate = fit$bivariate,
      spacetime = fit$spacetime,
      coordx_dyn = fit$coordx_dyn,
      distance = fit$distance,
      radius = fit$radius
    )
  )
  fit_context_main <- if (as.character(fit$type)[1L] %in% c("Pairwise", "Independence")) new.env(parent = emptyenv()) else NULL
  on.exit(.GeoVarest_release_fit_context(list(type = fit$type)), add = TRUE)

  ## Data-generating model and inferential criterion are distinct for
  ## misspecified fits.  GeoFit now records both explicitly.
  model_sim <- as.character(fit$model)[1L]
  model_est <- .GeoVarest_estimation_model(fit, "GeoVarest")
  boot_par <- .GeoVarest_bootstrap_parameters(fit)

  .GeoVarest_preflight_simulation(
    fit, method, model_sim, boot_par$sim_param, context = "GeoVarest"
  )

  estimated_size_mb <- (K * dimat * 8) / (1024^2)

  ## For TB simulations, never send GeoSimapprox's local do_one_sim(),
  ## RFfct1(), and simu_approx() closures to multisession workers.  Those
  ## closures retain the complete simulation frame and can become hundreds
  ## of MiB each.  Instead, stream complete simulation/score replications
  ## through a small namespace-level worker.  Streaming is also activated
  ## sequentially when retaining all K simulated fields would exceed 500 MiB.
  tb_stream_mode <- identical(method, "TB") && is.null(fit$copula) &&
    ((isTRUE(parallel) && K > 1L) || estimated_size_mb > 500)

  if (estimated_size_mb > 500) {
    warning(
      sprintf("Estimated simulated dataset size: %.1f MB", estimated_size_mb),
      call. = FALSE
    )
  }

  ## ---------------------------------------------------------------
  ## 1--3. Streaming TB simulation and score evaluation
  ## ---------------------------------------------------------------
  if (tb_stream_mode) {
    sim_spec <- list(
      coordx     = coords,
      coordt     = coordt_use,
      coordx_dyn = fit$coordx_dyn,
      anisopars  = boot_par$anisopars,
      corrmodel  = fit$corrmodel,
      model      = model_sim,
      param      = boot_par$sim_param,
      grid       = fit$grid,
      X          = fit$X,
      n          = fit$n,
      distance   = fit$distance,
      radius     = fit$radius
    )

    fit_spec <- .GeoVarest_make_fit_spec(
      fit, boot_par, model_est, X_use, optimizer, lower, upper,
      selected_pair_index, coordt_use, memdist = TRUE,
      weighted = fit$weighted %||% FALSE
    )

    ## One deterministic simulation seed per replication.  The result is
    ## reproducible for a fixed GeoVarest(seed=...) independently of worker
    ## scheduling and of the number of score workers.
    sim_seeds <- sample.int(.Machine$integer.max, K, replace = FALSE)
    ## future.apply uses L'Ecuyer-CMRG streams in multisession workers when
    ## future.seed=TRUE.  The bootstrap's explicit scalar replicate seeds must
    ## nevertheless be interpreted under the caller's RNG kind; otherwise the
    ## same scalar seed produces different TB fields in sequential and parallel
    ## execution.  Carry the caller RNG kind explicitly to every worker.
    sim_rng_kind <- RNGkind()

    workers <- .GeoResolveWorkers(parallel, ncores, n_jobs = K)

    use_outer_stream <- isTRUE(parallel) && workers > 1L &&
      requireNamespace("future", quietly = TRUE) &&
      requireNamespace("future.apply", quietly = TRUE)

    if (isTRUE(parallel) && !use_outer_stream) {
      warning(
        "Parallel TB evaluation requested but future/future.apply is unavailable; using sequential streaming.",
        call. = FALSE
      )
    }

    worker_fun <- if (requireNamespace("GeoModels", quietly = TRUE) &&
                      exists(".GeoVarest_score_worker",
                             envir = asNamespace("GeoModels"), inherits = FALSE)) {
      getFromNamespace(".GeoVarest_score_worker", "GeoModels")
    } else {
      .GeoVarest_score_worker
    }

    ids <- seq_len(K)

    if (use_outer_stream) {
      old_plan <- future::plan()
      on.exit(try(future::plan(old_plan), silent = TRUE), add = TRUE)

      old_proj_data <- Sys.getenv("PROJ_DATA", unset = NA)
      old_proj_lib  <- Sys.getenv("PROJ_LIB", unset = NA)
      on.exit({
        if (is.na(old_proj_data)) Sys.unsetenv("PROJ_DATA") else
          Sys.setenv(PROJ_DATA = old_proj_data)
        if (is.na(old_proj_lib)) Sys.unsetenv("PROJ_LIB") else
          Sys.setenv(PROJ_LIB = old_proj_lib)
      }, add = TRUE)

      set_proj_env_for_workers()
      future_backend <- .GeoVarest_set_future_plan(workers)

      if (progress) {
        cat("Performing", K,
            "TB simulations and scores on", workers, "cores...\n")
      }

      chunks <- split(ids, rep(seq_len(workers), length.out = K))
      explicit_payload <- sum(vapply(
        list(sim_spec, fit_spec, theta_hat, sim_seeds, sim_rng_kind),
        function(x) as.numeric(utils::object.size(x)),
        numeric(1)
      ))
      old_max_size <- getOption("future.globals.maxSize")
      on.exit(options(future.globals.maxSize = old_max_size), add = TRUE)
      options(
        future.globals.maxSize = max(500 * 1024^2, 1.5 * explicit_payload)
      )

      run_chunks <- function(progressor = NULL) {
        future.apply::future_lapply(
          chunks,
          FUN = worker_fun,
          mode = "tb",
          sim_spec = sim_spec,
          fit_spec = fit_spec,
          theta_hat = theta_hat,
          sim_seeds = sim_seeds,
          L = L,
          rng_kind = sim_rng_kind,
          progressor = progressor,
          future.seed = TRUE,
          future.scheduling = 1,
          future.globals = FALSE,
          future.packages = "GeoModels"
        )
      }

      result_chunks <- if (use_progressr) {
        progressr::with_progress({
          pb <- progressr::progressor(steps = K)
          pb_worker <- if (as.numeric(utils::object.size(pb)) <= 1024^2) pb else NULL
          run_chunks(pb_worker)
        })
      } else {
        run_chunks(NULL)
      }

      score_list <- vector("list", K)
      for (chunk_result in result_chunks) {
        ids_chunk <- as.integer(names(chunk_result))
        score_list[ids_chunk] <- unname(chunk_result)
      }
    } else {
      if (progress) cat("Performing", K, "TB simulations and scores sequentially...\n")
      if (use_progressr) {
        score_list <- progressr::with_progress({
          pb <- progressr::progressor(steps = K)
          unname(worker_fun(
            ids = ids,
            mode = "tb",
            sim_spec = sim_spec,
            fit_spec = fit_spec,
            theta_hat = theta_hat,
            sim_seeds = sim_seeds,
            L = L,
            rng_kind = sim_rng_kind,
            progressor = pb
          ))
        })
      } else {
        score_list <- unname(worker_fun(
          ids = ids,
          mode = "tb",
          sim_spec = sim_spec,
          fit_spec = fit_spec,
          theta_hat = theta_hat,
          sim_seeds = sim_seeds,
          L = L,
          rng_kind = sim_rng_kind,
          progressor = NULL
        ))
      }
    }

  } else {

  ## ---------------------------------------------------------------
  ## 1. Simulate K datasets
  ## ---------------------------------------------------------------
  sim_args <- list(
    coordx     = coords,
    coordt     = coordt_use,
    coordx_dyn = fit$coordx_dyn,
    anisopars  = boot_par$anisopars,
    corrmodel  = fit$corrmodel,
    model      = model_sim,
    param      = boot_par$sim_param,
    grid       = fit$grid,
    X          = fit$X,
    n          = fit$n,
    distance   = fit$distance,
    radius     = fit$radius,
    nrep       = K,
    progress   = progress
  )

  if (is.null(fit$copula)) {
    if (method == "cholesky") {
      data_sim_full <- do.call(
        GeoSim_fun,
        c(sim_args, list(sparse = sparse, method = method))
      )
    } else {
      GeoSimapprox_fun <- gm_fun("GeoSimapprox")

      ## GeoSimapprox may parallelize over nrep = K. This simulation phase
      ## is completed before score evaluation starts, so the score
      ## parallelism below is not nested. We defensively restore the current
      ## future plan before entering the score phase.
      old_plan_before_sim <- NULL
      if (isTRUE(parallel) && requireNamespace("future", quietly = TRUE)) {
        old_plan_before_sim <- future::plan()
      }

      data_sim_full <- do.call(
        GeoSimapprox_fun,
        c(
          sim_args,
          list(
            method   = method,
            L        = L,
            parallel = parallel,
            ncores   = ncores
          )
        )
      )

      if (!is.null(old_plan_before_sim)) {
        try(future::plan(old_plan_before_sim), silent = TRUE)
      }
    }
  } else {
    if (!(method %in% c("cholesky", "TB"))) {
      stop("Copula simulation supports method = 'cholesky' or method = 'TB'.",
           call. = FALSE)
    }

    GeoSimCopula_fun <- gm_fun("GeoSimCopula")
    copula_args <- list(copula = fit$copula, sparse = sparse, method = method)
    if (identical(method, "TB")) {
      copula_args <- c(copula_args, list(L = L, parallel = parallel, ncores = ncores))
    }

    data_sim_full <- do.call(
      GeoSimCopula_fun,
      c(sim_args, copula_args)
    )
  }

  data_sim <- data_sim_full$data
  rm(data_sim_full)
  gc(verbose = FALSE, full = TRUE)

  ## ---------------------------------------------------------------
  ## 2. Evaluate composite likelihood and score through GeoFit
  ## ---------------------------------------------------------------
  make_start <- function(par) {
    par <- as.numeric(par)
    names(par) <- names(theta_hat)
    .GeoVarest_start_from_theta(par, if (is.null(boot_par$anisopars)) character(0) else c("angle", "ratio"))
  }

  geofit_onlyvar <- function(current_data, par, want_score) {
    call_it <- function() {
      GeoFit_fun(
        data        = current_data,
        start       = make_start(par),
        fixed       = boot_par$fixed,
        coordx      = coords,
        coordt      = coordt_use,
        coordx_dyn  = fit$coordx_dyn,
        copula      = fit$copula,
        anisopars   = boot_par$anisopars,
        est.aniso   = fit$est.aniso,
        thin_method = fit$thin_method,
        lower       = lower,
        upper       = upper,
        neighb      = fit$neighb,
        p_neighb    = fit$p_neighb,
        corrmodel   = fit$corrmodel,
        model       = model_est,
        sparse      = FALSE,
        n           = fit$n,
        maxdist     = fit$maxdist,
        maxtime     = fit$maxtime,
        memdist     = TRUE,
        optimizer   = optimizer,
        grid        = fit$grid,
        likelihood  = fit$likelihood,
        type        = fit$type,
        X           = X_use,
        distance    = fit$distance,
        radius      = fit$radius,
        onlyvar     = TRUE,
        score       = want_score,
        sensitivity = FALSE,
        varest      = FALSE,
        weighted    = fit$weighted %||% FALSE
      )
    }

    .GeoVarest_with_selected_pairs(
      selected_pair_index,
      call_it(),
      pair_info = selected_pair_info,
      fit_context = fit_context_main
    )
  }

  geofit_score <- function(current_data, par) {
    capture.output({
      ff <- geofit_onlyvar(current_data, par, TRUE)
    }, file = nullfile())
    ans <- .GeoVarest_extract_score(ff, names(theta_hat))
    list(score = ans[seq_along(theta_hat)], logCompLik = unname(ans[[length(ans)]]))
  }

  make_score <- function(k, current_data) {
    tryCatch({
      out <- geofit_score(current_data, theta_hat)

      c(as.numeric(out$score), logCompLik = as.numeric(out$logCompLik))

    }, error = function(e) {
      z <- rep(NA_real_, num_params + 1L)
      attr(z, "error") <- conditionMessage(e)
      z
    })
  }

  ## ---------------------------------------------------------------
  ## 3. Score loop. Parallelization is only here.
  ## ---------------------------------------------------------------
  if (progress) {
    cat("Computing", K, "scores ...\n")
  }

  use_parallel <- isTRUE(parallel) && K > 1L &&
    requireNamespace("future", quietly = TRUE) &&
    requireNamespace("future.apply", quietly = TRUE)

  if (isTRUE(parallel) && !use_parallel) {
    warning(
      "Parallel evaluation requested but future/future.apply is unavailable; using sequential evaluation.",
      call. = FALSE
    )
  }

  if (use_parallel) {
    ncores <- .GeoResolveWorkers(TRUE, ncores, n_jobs = K)
    use_parallel <- ncores > 1L
  }

  if (use_parallel) {
    old_plan <- future::plan()
    on.exit(try(future::plan(old_plan), silent = TRUE), add = TRUE)

    ## Workers receive an explicit compact payload below.  Avoid measuring
    ## nested closures here: their captured environments can be very large and
    ## are not exported by the current implementation.
    old_max_size <- getOption("future.globals.maxSize")
    on.exit(options(future.globals.maxSize = old_max_size), add = TRUE)
    safety_margin <- 1.5
    default_cap <- 500 * 1024^2

    old_proj_data <- Sys.getenv("PROJ_DATA", unset = NA)
    old_proj_lib  <- Sys.getenv("PROJ_LIB",  unset = NA)

    on.exit({
      if (is.na(old_proj_data)) {
        Sys.unsetenv("PROJ_DATA")
      } else {
        Sys.setenv(PROJ_DATA = old_proj_data)
      }
      if (is.na(old_proj_lib)) {
        Sys.unsetenv("PROJ_LIB")
      } else {
        Sys.setenv(PROJ_LIB = old_proj_lib)
      }
    }, add = TRUE)

    rscript_envs <- set_proj_env_for_workers()

    future::plan(
      future::multisession,
      workers = ncores,
      rscript_libs = .GeoVarest_future_libpaths()
    )

    temp_dir <- tempdir()
    tag <- format(Sys.time(), "%Y%m%d_%H%M%S")

    data_files <- vapply(seq_len(K), function(k) {
      fp <- file.path(temp_dir, sprintf("scoreJ_%s_%04d.rds", tag, k))
      saveRDS(data_sim[[k]], fp, compress = FALSE)
      fp
    }, character(1))

    on.exit(unlink(data_files), add = TRUE)

    rm(data_sim)
    gc(verbose = FALSE, full = TRUE)

    ## Use the package-level worker with an explicit payload.  This avoids
    ## exporting the nested GeoFit closures (and their captured fit object) to
    ## every multisession worker.  It is especially important for large data
    ## and also keeps deterministic/Independence fits on their original path.
    score_worker <- if (requireNamespace("GeoModels", quietly = TRUE) &&
                        exists(".GeoVarest_score_worker",
                               envir = asNamespace("GeoModels"), inherits = FALSE)) {
      getFromNamespace(".GeoVarest_score_worker", "GeoModels")
    } else {
      .GeoVarest_score_worker
    }

    score_sim_spec <- list(
      coordx = coords,
      coordt = coordt_use,
      coordx_dyn = fit$coordx_dyn
    )

    score_fit_spec <- .GeoVarest_make_fit_spec(
      fit, boot_par, model_est, X_use, optimizer, lower, upper,
      selected_pair_index, coordt_use, memdist = TRUE,
      weighted = fit$weighted %||% FALSE
    )

    chunks <- split(seq_len(K), rep(seq_len(ncores), length.out = K))

    explicit_payload <- sum(vapply(
      list(data_files, score_sim_spec, score_fit_spec, theta_hat),
      function(x) as.numeric(utils::object.size(x)),
      numeric(1)
    ))
    options(
      future.globals.maxSize = max(default_cap, explicit_payload * safety_margin)
    )

    run_score_chunks <- function(progressor = NULL) {
      future.apply::future_lapply(
        chunks,
        FUN = score_worker,
        mode = "file",
        data_files = data_files,
        sim_spec = score_sim_spec,
        fit_spec = score_fit_spec,
        theta_hat = theta_hat,
        progressor = progressor,
        future.seed = TRUE,
        future.globals = FALSE,
        future.packages = "GeoModels"
      )
    }

    result_chunks <- if (use_progressr) {
      progressr::with_progress({
        pb <- progressr::progressor(steps = K)
        pb_worker <- if (as.numeric(utils::object.size(pb)) <= 1024^2) pb else NULL
        run_score_chunks(pb_worker)
      })
    } else {
      run_score_chunks(NULL)
    }

    score_list <- vector("list", K)
    for (chunk_result in result_chunks) {
      ids_chunk <- as.integer(names(chunk_result))
      score_list[ids_chunk] <- unname(chunk_result)
    }

  } else {
    if (use_progressr) {
      score_list <- progressr::with_progress({
        pb <- progressr::progressor(along = seq_len(K))
        out <- vector("list", K)

        for (k in seq_len(K)) {
          out[[k]] <- make_score(k, data_sim[[k]])
          if (k %% 5L == 0L) gc(verbose = FALSE)
          pb(sprintf("score %d", k))
        }

        out
      })
    } else {
      score_list <- vector("list", K)

      for (k in seq_len(K)) {
        if (progress && (k == 1L || k %% 10L == 0L || k == K)) {
          cat("score", k, "of", K, "\n")
        }

        score_list[[k]] <- make_score(k, data_sim[[k]])
        if (k %% 5L == 0L) gc(verbose = FALSE)
      }
    }

    rm(data_sim)
    gc(verbose = FALSE, full = TRUE)
  }

  } # end non-streaming branch

  errors <- vapply(
    score_list,
    function(x) attr(x, "error") %||% "",
    character(1)
  )

  score_raw <- do.call(rbind, score_list)
  colnames(score_raw) <- c(names(theta_hat), "logCompLik")

  valid <- apply(
    score_raw[, seq_len(num_params), drop = FALSE],
    1L,
    function(z) all(is.finite(z))
  ) & is.finite(score_raw[, num_params + 1L])

  score_out <- score_raw[valid, , drop = FALSE]
  n_successful <- nrow(score_out)

  if (n_successful < max(2L, as.integer(ceiling(min_success_rate * K)))) {
    shown <- utils::head(errors[nzchar(errors)], 8L)
    msg <- sprintf(
      "GeoVarest: only %d of %d score-bootstrap replications succeeded (%.1f%%); at least %.1f%% are required.",
      n_successful, K, 100 * n_successful / K, 100 * min_success_rate
    )
    if (length(shown) > 0L) {
      msg <- paste0(msg, "\nFirst errors:\n- ", paste(shown, collapse = "\n- "))
    }
    stop(msg, call. = FALSE)
  }
  .GeoVarest_check_success(n_successful, K, min_success_rate, "GeoVarest")

  if (progress) {
    cat("Successful score evaluations:", n_successful, "out of", K, "\n")
  }

  ## ---------------------------------------------------------------
  ## 4. Sandwich/Godambe and criteria
  ## ---------------------------------------------------------------
  score_mat <- score_out[, seq_len(num_params), drop = FALSE]

  J <- stats::var(score_mat)
  J <- (J + t(J)) / 2
  rownames(J) <- colnames(J) <- names(theta_hat)

  ## ---------------------------------------------------------------
  ## Numerically stable sandwich through diagonal rescaling
  ##
  ## Let D = diag{1 / sqrt(diag(H))}. Then
  ##   Hs = D H D,  Js = D J D,
  ## and, in exact arithmetic,
  ##   H^{-1} J H^{-1}
  ##     = D Hs^{-1} Js Hs^{-1} D.
  ##
  ## This is algebraically identical to the unscaled sandwich, but it
  ## avoids unstable inversion when parameters have very different scales.
  ## ---------------------------------------------------------------
  hdiag <- diag(H)

  if (any(!is.finite(hdiag)) || any(hdiag <= 0)) {
    stop(
      "The sensitivity matrix has non-positive or non-finite diagonal elements.",
      call. = FALSE
    )
  }

  scale_vec <- 1 / sqrt(hdiag)
  names(scale_vec) <- names(theta_hat)

  D <- diag(scale_vec, nrow = num_params, ncol = num_params)
  rownames(D) <- colnames(D) <- names(theta_hat)

  H_scaled <- D %*% H %*% D
  H_scaled <- (H_scaled + t(H_scaled)) / 2
  rownames(H_scaled) <- colnames(H_scaled) <- names(theta_hat)

  J_scaled <- D %*% J %*% D
  J_scaled <- (J_scaled + t(J_scaled)) / 2
  rownames(J_scaled) <- colnames(J_scaled) <- names(theta_hat)

  H_scaled_inv <- safe_inverse(H_scaled, "scaled H/sensmat")
  rownames(H_scaled_inv) <- colnames(H_scaled_inv) <- names(theta_hat)

  Ginv_scaled <- H_scaled_inv %*% J_scaled %*% H_scaled_inv
  Ginv_scaled <- (Ginv_scaled + t(Ginv_scaled)) / 2
  rownames(Ginv_scaled) <- colnames(Ginv_scaled) <- names(theta_hat)

  ## Transform the covariance matrix back to the original parameter scale.
  Ginv <- D %*% Ginv_scaled %*% D
  Ginv <- (Ginv + t(Ginv)) / 2
  rownames(Ginv) <- colnames(Ginv) <- names(theta_hat)

  ## H^{-1} on the original parameter scale, retained for compatibility.
  Hinv <- D %*% H_scaled_inv %*% D
  Hinv <- (Hinv + t(Hinv)) / 2
  rownames(Hinv) <- colnames(Hinv) <- names(theta_hat)

  ## Invert the covariance on the scaled parameter system.  Directly
  ## inverting Ginv on the original scale can recreate the same purely
  ## unit-driven ill-conditioning that motivated the scaling of H.
  G_scaled <- safe_inverse(Ginv_scaled, "scaled Ginv/varcov")
  rownames(G_scaled) <- colnames(G_scaled) <- names(theta_hat)

  Dinv <- diag(1 / scale_vec, nrow = num_params, ncol = num_params)
  rownames(Dinv) <- colnames(Dinv) <- names(theta_hat)

  G <- Dinv %*% G_scaled %*% Dinv
  G <- (G + t(G)) / 2
  rownames(G) <- colnames(G) <- names(theta_hat)

  stderr <- sqrt(pmax(0, diag(Ginv)))
  names(stderr) <- names(theta_hat)

  ## Equivalent to tr(H^{-1} J), evaluated on the stable scaled system.
  penalty <- sum(diag(H_scaled_inv %*% J_scaled))
  penalty_alt <- sum(diag(H_scaled %*% Ginv_scaled))

  fit_loglik <- if (!is.null(fit$logCompLik)) fit$logCompLik else fit$logLik
  fit_loglik <- as.numeric(fit_loglik)

  if (length(fit_loglik) != 1L || !is.finite(fit_loglik)) {
    stop("fit does not contain a finite logCompLik/logLik value", call. = FALSE)
  }

  lik <- as.character(fit$likelihood)
  typ <- as.character(fit$type)

  if ((lik == "Marginal" && typ %in% c("Independence", "Pairwise")) ||
      (lik == "Conditional" && typ == "Pairwise")) {

    claic <- -2 * fit_loglik + 2 * penalty
    clbic <- -2 * fit_loglik + log(dimat) * penalty
    fit$varimat <- H %*% Ginv %*% H

  } else {
    claic <- clbic <- NA_real_
  }

  z_alpha <- stats::qnorm(1 - (1 - alpha) / 2)

  fit$stderr <- stderr
  fit$varcov <- Ginv
  fit$godambe <- G
  fit$Jmat <- J
  fit$Hinv <- Hinv

  ## Numerical diagnostics for the rescaled sensitivity matrix.
  fit$H_scaled <- H_scaled
  fit$J_scaled <- J_scaled
  fit$Ginv_scaled <- Ginv_scaled
  fit$godambe_scaled <- G_scaled
  fit$parameter_scaling <- scale_vec
  fit$H_scaled_condition <- kappa(H_scaled)
  fit$H_scaled_eigenvalues <- eigen(
    H_scaled,
    symmetric = TRUE,
    only.values = TRUE
  )$values

  fit$claic <- claic
  fit$clic <- claic
  fit$clbic <- clbic
  fit$clic_penalty <- penalty
  fit$clic_penalty_alt <- penalty_alt

  fit$conf.int <- rbind(
    theta_hat - z_alpha * stderr,
    theta_hat + z_alpha * stderr
  )
  colnames(fit$conf.int) <- names(theta_hat)
  rownames(fit$conf.int) <- c("Lower", "Upper")

  fit$pvalues_godambe <- 2 * stats::pnorm(-abs(theta_hat / stderr))
  names(fit$pvalues_godambe) <- names(theta_hat)

  ## Keep the standard component name for compatibility with GeoModels.
  fit$pvalues <- fit$pvalues_godambe
  fit$pvalues_type <- "Wald p-values based on score-bootstrap Godambe variance"

  fit$scores <- score_mat
  fit$score_logCompLik <- score_out[, "logCompLik"]

  fit$score_failures <- data.frame(
    iteration = which(!valid),
    error = errors[!valid],
    stringsAsFactors = FALSE
  )

  fit$bootstrap_type <- "parametric_score_J_geofit_score"
  fit$bootstrap_K <- K
  fit$bootstrap_successful <- n_successful
  fit$bootstrap_success_rate <- n_successful / K
  fit$bootstrap_min_success_rate <- min_success_rate
  fit$bootstrap_pair_graph <- if (is.null(selected_pair_index)) "deterministic" else "reused_original"
  fit$bootstrap_thin_method <- fit$thin_method
  fit$bootstrap_p_neighb <- fit$p_neighb
  fit$bootstrap_thinning_seed_rule <- if (is.null(selected_pair_index)) {
    "not applicable"
  } else {
    "not used; the original selected-pair graph is reused"
  }

  fit
}

############################################################
### Circulant-embedding helpers for spatial and separable
### spatio-temporal Gaussian random fields.
############################################################

.CEValidateScalar <- function(x, name, lower = -Inf, upper = Inf,
                              lower_open = FALSE, upper_open = FALSE) {
  if (length(x) != 1L || !is.numeric(x) || !is.finite(x)) {
    stop(name, " must be a finite numeric scalar.", call. = FALSE)
  }
  bad_lower <- if (lower_open) x <= lower else x < lower
  bad_upper <- if (upper_open) x >= upper else x > upper
  if (bad_lower || bad_upper) {
    left <- if (lower_open) "(" else "["
    right <- if (upper_open) ")" else "]"
    stop(name, " must belong to ", left, lower, ", ", upper, right, ".",
         call. = FALSE)
  }
  as.numeric(x)
}

.CERegularAxis <- function(x, n, name) {
  if (!is.numeric(x) || any(!is.finite(x))) {
    stop(name, " must be a finite numeric vector.", call. = FALSE)
  }
  if (length(n) != 1L || !is.numeric(n) || !is.finite(n) ||
      n != as.integer(n) || n < 2L) {
    stop("The dimension associated with ", name,
         " must be an integer greater than or equal to two.", call. = FALSE)
  }
  n <- as.integer(n)
  if (length(x) != n) {
    stop(name, " must contain exactly ", n, " coordinates.", call. = FALSE)
  }
  dx <- diff(x)
  if (any(dx <= 0)) {
    stop(name, " must be strictly increasing.", call. = FALSE)
  }
  tol <- sqrt(.Machine$double.eps) * max(1, abs(dx[1L]), max(abs(x)))
  if (max(abs(dx - dx[1L])) > tol) {
    stop(name, " must be equally spaced for circulant embedding.",
         call. = FALSE)
  }
  unname(as.numeric(x))
}

.CEValidateMaxExt <- function(max.ext) {
  if (length(max.ext) != 1L || !is.numeric(max.ext) || !is.finite(max.ext) ||
      max.ext != as.integer(max.ext) || max.ext < 1L) {
    stop("max.ext must be a positive integer.", call. = FALSE)
  }
  as.integer(max.ext)
}

.CEVarianceParameters <- function(param) {
  sill <- if (is.null(param$sill)) 1 else param$sill
  nugget <- if (is.null(param$nugget)) 0 else param$nugget
  sill <- .CEValidateScalar(sill, "param$sill", lower = 0,
                            lower_open = TRUE)
  nugget <- .CEValidateScalar(nugget, "param$nugget", lower = 0,
                              upper = 1)
  list(sill = sill, nugget = nugget)
}

.CEStandardCorrelationParam <- function(param) {
  out <- param
  out$mean <- 0
  out$sill <- 1
  out$nugget <- 0
  out
}

.CEEigenInfo <- function(lambda) {
  if (!is.numeric(lambda) || any(!is.finite(lambda))) {
    stop("The circulant embedding produced non-finite eigenvalues.",
         call. = FALSE)
  }
  scale <- max(1, max(abs(lambda)))
  tol <- sqrt(.Machine$double.eps) * scale
  list(valid = min(lambda) >= -tol, tol = tol,
       values = pmax(lambda, 0), minimum = min(lambda))
}

.CESeparableModel <- function(corrmodel, param) {
  make_param <- function(prefix = "") {
    suffix <- if (nzchar(prefix)) paste0("_", prefix) else ""
    out <- list(nugget = 0, sill = 1)
    for (nm in c("scale", "smooth", "power2")) {
      value <- param[[paste0(nm, suffix)]]
      if (!is.null(value)) out[[nm]] <- value
    }
    out
  }

  if (identical(corrmodel, "Matern_Matern")) {
    return(list(model = "Matern", spatial = make_param("s"),
                temporal = make_param("t")))
  }
  if (identical(corrmodel, "GenWend_GenWend")) {
    return(list(model = "GenWend", spatial = make_param("s"),
                temporal = make_param("t")))
  }
  if (identical(corrmodel, "GenWend_Matern_GenWend_Matern")) {
    return(list(model = "GenWend_Matern", spatial = make_param("s"),
                temporal = make_param("t")))
  }
  if (identical(corrmodel, "Kummer_Kummer")) {
    return(list(model = "Kummer", spatial = make_param("s"),
                temporal = make_param("t")))
  }
  if (identical(corrmodel, "Kummer_Matern_Kummer_Matern")) {
    return(list(model = "Kummer_Matern", spatial = make_param("s"),
                temporal = make_param("t")))
  }

  stop("CE spatio-temporal simulation is implemented only for supported separable correlation models.",
       call. = FALSE)
}

.CETemporalEmbedding <- function(time.seq, corrmodel, param, max.ext) {
  nt <- length(time.seq)
  if (nt < 2L) {
    stop("At least two times are required for a temporal embedding.",
         call. = FALSE)
  }
  dt <- diff(time.seq)
  tol_time <- sqrt(.Machine$double.eps) *
    max(1, abs(dt[1L]), max(abs(time.seq)))
  if (any(dt <= 0)) {
    stop("coordt must be strictly increasing for CE simulation.",
         call. = FALSE)
  }
  if (max(abs(dt - dt[1L])) > tol_time) {
    stop("coordt must be equally spaced for CE spatio-temporal simulation.",
         call. = FALSE)
  }

  base_size <- 2L * (nt - 1L)
  last <- NULL
  for (attempt in seq_len(max.ext)) {
    emb_size <- as.integer(base_size * 2^(attempt - 1L))
    half <- emb_size %/% 2L
    lags <- (0:half) * dt[1L]
    rho <- as.numeric(GeoCorrFct(
      x = lags, corrmodel = corrmodel, model = "Gaussian",
      param = param
    )$corr)
    second_half <- if (half > 1L) rev(rho[2:half]) else numeric(0)
    first_row <- c(rho, second_half)
    lambda <- Re(fft(first_row, inverse = TRUE))
    info <- .CEEigenInfo(lambda)
    last <- list(size = emb_size, lambda = info$values,
                 valid = info$valid, minimum = info$minimum,
                 attempt = attempt)
    if (info$valid) return(last)
  }

  warning("The temporal circulant embedding has negative eigenvalues at the maximum extension; negative eigenvalues were truncated to zero.",
          call. = FALSE)
  last$truncated <- TRUE
  last
}

######################## space-time case ########################
## Separable simulation using Cholesky in space and a circulant
## embedding/FFT in time. Spatial locations must be fixed.
CE_Space_Time <- function(coords, time.seq, param, corrmodel, distance,
                          max.ext = 1L) {
  coords <- as.matrix(coords)
  if (!is.numeric(coords) || nrow(coords) < 1L ||
      !(ncol(coords) %in% c(2L, 3L)) || any(!is.finite(coords))) {
    stop("coords must be a finite numeric matrix with two or three columns.",
         call. = FALSE)
  }
  if (!is.numeric(time.seq) || !length(time.seq) ||
      any(!is.finite(time.seq))) {
    stop("coordt must be a non-empty finite numeric vector.",
         call. = FALSE)
  }
  max.ext <- .CEValidateMaxExt(max.ext)
  variance <- .CEVarianceParameters(param)
  sep <- .CESeparableModel(corrmodel, param)
  sep$spatial$mean <- 0
  sep$temporal$mean <- 0

  spatial_cov <- GeoCovmatrix(
    coordx = coords, corrmodel = sep$model, model = "Gaussian",
    param = sep$spatial, distance = distance
  )$covmatrix
  spatial_cov <- as.matrix(spatial_cov)
  diag(spatial_cov) <- 1
  chol_space <- tryCatch(
    t(chol(spatial_cov)),
    error = function(e) {
      stop("The spatial correlation matrix is not positive definite in CE_Space_Time: ",
           conditionMessage(e), call. = FALSE)
    }
  )

  ns <- nrow(coords)
  nt <- length(time.seq)
  if (nt == 1L) {
    base <- matrix(as.numeric(chol_space %*% rnorm(ns)), nrow = 1L)
  } else {
    time.seq <- as.numeric(time.seq - time.seq[1L])
    embedding <- .CETemporalEmbedding(
      time.seq, sep$model, sep$temporal, max.ext
    )
    emb_size <- embedding$size
    spectrum <- sqrt(embedding$lambda)
    freq <- matrix(complex(real = 0, imaginary = 0),
                   nrow = emb_size, ncol = ns)
    for (j in seq_len(emb_size)) {
      z <- complex(real = rnorm(ns), imaginary = rnorm(ns))
      freq[j, ] <- spectrum[j] * as.vector(chol_space %*% z)
    }
    temporal <- matrix(complex(real = 0, imaginary = 0),
                       nrow = emb_size, ncol = ns)
    for (j in seq_len(ns)) {
      temporal[, j] <- fft(freq[, j], inverse = TRUE) / sqrt(emb_size)
    }
    base <- Re(temporal[seq_len(nt), , drop = FALSE])
  }

  if (variance$nugget > 0) {
    noise <- matrix(rnorm(nt * ns), nrow = nt, ncol = ns)
    field <- sqrt(variance$sill) *
      (sqrt(1 - variance$nugget) * base +
         sqrt(variance$nugget) * noise)
  } else {
    field <- sqrt(variance$sill) * base
  }

  ## Temporal blocks: all spatial sites at time 1, then time 2, etc.
  c(t(field))
}

############################################################
SimCE <- function(M, N, x, y, z = NULL, corrmodel, param,
                  mean.val = 0, max.ext = 1L) {
  if (!is.null(z)) {
    stop("Spatial CE simulation is implemented only for two-dimensional grids.",
         call. = FALSE)
  }
  x <- .CERegularAxis(x, M, "x")
  y <- .CERegularAxis(y, N, "y")
  max.ext <- .CEValidateMaxExt(max.ext)
  variance <- .CEVarianceParameters(param)

  out <- GRF.CE.Sim(
    nsim = 1L, x = x, y = y, M = length(x), N = length(y),
    corrmodel = corrmodel, param = .CEStandardCorrelationParam(param),
    max.ext = max.ext
  )

  mean.val <- as.numeric(mean.val)
  ngrid <- length(x) * length(y)
  if (any(!is.finite(mean.val)) || !(length(mean.val) %in% c(1L, ngrid))) {
    stop("mean.val must be a finite scalar or contain one value per grid point.",
         call. = FALSE)
  }
  if (length(mean.val) == 1L) mean.val <- rep(mean.val, ngrid)

  latent <- as.matrix(out$X)
  if (variance$nugget > 0) {
    nugget_noise <- matrix(rnorm(length(latent)), nrow = nrow(latent),
                           ncol = ncol(latent))
    field <- sqrt(variance$sill) *
      (sqrt(1 - variance$nugget) * latent +
         sqrt(variance$nugget) * nugget_noise)
  } else {
    field <- sqrt(variance$sill) * latent
  }
  out$X <- sweep(field, 1L, mean.val, "+")
  out
}

##################
GRF.CE.Sim <- function(nsim, x, y, M, N, corrmodel, param, max.ext) {
  if (length(nsim) != 1L || !is.numeric(nsim) || !is.finite(nsim) ||
      nsim != as.integer(nsim) || nsim < 1L) {
    stop("nsim must be a positive integer.", call. = FALSE)
  }
  nsim <- as.integer(nsim)
  max.ext <- .CEValidateMaxExt(max.ext)
  decomposition <- NULL
  selected_attempt <- max.ext

  for (attempt in seq_len(max.ext)) {
    ext <- 2^attempt
    candidate <- Cov_Mat_CE(
      x = x, y = y, M = M, N = N, corrmodel = corrmodel,
      param = param, ext = ext, approx = FALSE
    )
    decomposition <- candidate
    selected_attempt <- attempt
    if (isTRUE(candidate$valid)) break
  }

  approx.flag <- !isTRUE(decomposition$valid)
  if (approx.flag) {
    warning("The spatial circulant embedding has negative eigenvalues at the maximum extension; negative eigenvalues were truncated to zero.",
            call. = FALSE)
    decomposition$prefix <- sqrt(pmax(decomposition$eigenvalues, 0))
  }

  simulation <- Cov_Mat_CE.sim(decomposition, nsim = nsim)
  list(
    X = simulation,
    grid.points = as.matrix(decomposition$grid),
    k0 = selected_attempt,
    approx.flag = approx.flag,
    min.eigenvalue = decomposition$min.eigenvalue
  )
}

##########################################
Cov_Mat_CE <- function(x, y, M, N, corrmodel, param, ext = 2,
                       approx = FALSE) {
  mygrid <- grid.prep(x, y, M = M, N = N, ext = ext)
  rx <- mygrid$M.ext * mygrid$cell.width
  ry <- mygrid$N.ext * mygrid$cell.height

  dx_abs <- abs(mygrid$x.ext[1L] - mygrid$x.ext)
  dx <- pmin(dx_abs, rx - dx_abs)
  dy_abs <- abs(mygrid$y.ext[1L] - mygrid$y.ext)
  dy <- pmin(dy_abs, ry - dy_abs)
  torus <- expand.grid(dx = dx, dy = dy)
  distances <- matrix(
    sqrt(torus[, 1L]^2 + torus[, 2L]^2),
    nrow = mygrid$M.ext, ncol = mygrid$N.ext
  )

  sigma_row <- my.GeoCorrFct(
    x = c(distances), corrmodel = corrmodel,
    param = .CEStandardCorrelationParam(param)
  )
  sigma_row <- matrix(sigma_row, nrow = mygrid$M.ext,
                      ncol = mygrid$N.ext)
  lambda <- Re(fft(sigma_row, inverse = TRUE))
  info <- .CEEigenInfo(lambda)
  if (isTRUE(approx)) info$values <- pmax(lambda, 0)

  list(
    dp = length(lambda), d = dim(lambda), sdp = sqrt(length(lambda)),
    M = M, N = N, prefix = sqrt(info$values),
    eigenvalues = lambda, valid = info$valid || isTRUE(approx),
    min.eigenvalue = info$minimum,
    grid = expand.grid(x = mygrid$x, y = mygrid$y)
  )
}

##################
grid.prep <- function(x, y, M, N, ext = 2) {
  x <- .CERegularAxis(x, M, "x")
  y <- .CERegularAxis(y, N, "y")
  if (length(ext) != 1L || !is.numeric(ext) || !is.finite(ext) ||
      ext != as.integer(ext) || ext < 2L) {
    stop("The embedding extension must be an integer greater than or equal to two.",
         call. = FALSE)
  }
  ext <- as.integer(ext)
  M.ext <- ext * M
  N.ext <- ext * N
  cell.width <- diff(x)[1L]
  cell.height <- diff(y)[1L]
  x.ext <- x[1L] + (0:(M.ext - 1L)) * cell.width
  y.ext <- y[1L] + (0:(N.ext - 1L)) * cell.height

  list(
    M = M, N = N, x = x, y = y,
    cell.width = cell.width, cell.height = cell.height,
    M.ext = M.ext, N.ext = N.ext,
    x.ext = x.ext, y.ext = y.ext
  )
}

##################
my.GeoCorrFct <- function(x, corrmodel, param) {
  if (identical(corrmodel, "none")) {
    result <- rep(0, length(x))
    result[x == 0] <- 1
    return(result)
  }
  as.numeric(GeoCorrFct(
    x = x, corrmodel = corrmodel, model = "Gaussian",
    param = .CEStandardCorrelationParam(param)
  )$corr)
}

##################
Cov_Mat_CE.sim <- function(ex1, nsim = 2L) {
  dp <- ex1$dp
  d <- ex1$d
  sdp <- ex1$sdp
  M <- ex1$M
  N <- ex1$N
  prefix <- ex1$prefix
  if (length(prefix) != dp || any(!is.finite(prefix))) {
    stop("Invalid spectral decomposition in Cov_Mat_CE.sim.",
         call. = FALSE)
  }

  X <- matrix(0, nrow = M * N, ncol = nsim)
  for (i in seq_len(nsim)) {
    standard <- rnorm(dp)
    spectral <- prefix * (fft(matrix(standard, d[1L], d[2L])) / sdp)
    draw <- Re(fft(spectral, inverse = TRUE) / sdp)
    X[, i] <- as.vector(draw[seq_len(M), seq_len(N), drop = FALSE])
  }
  X
}

####################################################
### File name: GeoNeighborhood.R
####################################################

GeoNeighborhood <- function(data = NULL, coordx = NULL, coordy = NULL,
                            coordz = NULL, coordt = NULL,
                            coordx_dyn = NULL, bivariate = FALSE,
                            distance = "Eucl", grid = FALSE, loc,
                            neighb = NULL, maxdist = NULL,
                            maxtime = NULL, radius = 1, time = NULL,
                            X = NULL, M = NULL, spobj = NULL,
                            spdata = NULL, parallel = FALSE,
                            ncores = 6, check.duplicates = FALSE) {

  ##########################################################################
  ## Small validation helpers
  ##########################################################################

  .is_scalar_logical <- function(x) {
    is.logical(x) && length(x) == 1L && !is.na(x)
  }

  .check_nonnegative_scalar <- function(x, name, allow_null = TRUE) {
    if (is.null(x) && allow_null) return(NULL)
    if (!is.numeric(x) || length(x) != 1L || !is.finite(x) || x < 0) {
      stop(name, " must be a finite non-negative scalar.", call. = FALSE)
    }
    as.numeric(x)
  }

  .check_numeric_matrix <- function(x, name, ncol_allowed = c(2L, 3L)) {
    if (is.data.frame(x)) x <- as.matrix(x)
    if (!is.matrix(x) || !is.numeric(x)) {
      stop(name, " must be a numeric matrix.", call. = FALSE)
    }
    if (nrow(x) < 1L || !(ncol(x) %in% ncol_allowed)) {
      stop(name, " must have at least one row and two or three columns.",
           call. = FALSE)
    }
    if (any(!is.finite(x))) {
      stop(name, " contains non-finite values.", call. = FALSE)
    }
    unname(x)
  }

  .check_spherical_coords <- function(x, name) {
    if (ncol(x) != 2L) {
      stop(name, " must have exactly two columns (longitude, latitude) ",
           "for Geod or Chor distance.", call. = FALSE)
    }
    if (any(x[, 2L] < -90 | x[, 2L] > 90)) {
      stop(name, " contains latitudes outside [-90, 90].", call. = FALSE)
    }
    invisible(TRUE)
  }

  .normalise_neighb <- function(x) {
    if (is.null(x)) return(NULL)
    if (!is.numeric(x) || length(x) != 1L || !is.finite(x) || x < 1) {
      stop("neighb must be a finite positive integer.", call. = FALSE)
    }
    xr <- round(x)
    if (abs(x - xr) > sqrt(.Machine$double.eps)) {
      stop("neighb must be an integer.", call. = FALSE)
    }
    as.integer(xr)
  }

  .build_coords <- function(coordx, coordy, coordz, grid) {
    if (is.null(coordx)) return(NULL)

    if ((is.matrix(coordx) || is.data.frame(coordx)) &&
        is.null(coordy) && is.null(coordz)) {
      return(.check_numeric_matrix(as.matrix(coordx), "coordx"))
    }

    if (is.null(coordy)) {
      stop("coordy is required when coordx is not a coordinate matrix.",
           call. = FALSE)
    }

    if (!is.numeric(coordx) || !is.numeric(coordy) ||
        any(!is.finite(coordx)) || any(!is.finite(coordy))) {
      stop("coordx and coordy must be finite numeric vectors.", call. = FALSE)
    }

    if (!is.null(coordz) &&
        (!is.numeric(coordz) || any(!is.finite(coordz)))) {
      stop("coordz must be a finite numeric vector.", call. = FALSE)
    }

    if (isTRUE(grid)) {
      out <- if (is.null(coordz)) {
        as.matrix(expand.grid(coordx, coordy))
      } else {
        as.matrix(expand.grid(coordx, coordy, coordz))
      }
    } else {
      if (length(coordx) != length(coordy)) {
        stop("coordx and coordy must have the same length.", call. = FALSE)
      }
      if (!is.null(coordz) && length(coordz) != length(coordx)) {
        stop("coordz must have the same length as coordx and coordy.",
             call. = FALSE)
      }
      out <- if (is.null(coordz)) {
        cbind(coordx, coordy)
      } else {
        cbind(coordx, coordy, coordz)
      }
    }

    .check_numeric_matrix(out, "coordinates")
  }

  .map <- function(index, fun) {
    if (!isTRUE(parallel) || length(index) < 2L) {
      return(lapply(index, fun))
    }

    workers <- .GeoResolveWorkers(TRUE, ncores, n_jobs = length(index))
    if (workers <= 1L) return(lapply(index, fun))

    if (!requireNamespace("future", quietly = TRUE) ||
        !requireNamespace("future.apply", quietly = TRUE)) {
      stop("parallel=TRUE requires the future and future.apply packages.",
           call. = FALSE)
    }

    old_plan <- future::plan()
    on.exit(future::plan(old_plan), add = TRUE)
    future::plan(future::multisession, workers = workers)

    future.apply::future_lapply(
      index,
      fun,
      future.seed = TRUE,
      future.globals = TRUE,
      future.packages = "nabor"
    )
  }

  ##########################################################################
  ## Exact spatial distances used to verify/order the nabor candidates
  ##########################################################################

  .central_angle <- function(ref, query) {
    deg_to_rad <- pi / 180
    phi_q <- query[2L] * deg_to_rad
    phi_r <- ref[, 2L] * deg_to_rad
    delta_lon <- (ref[, 1L] - query[1L]) * deg_to_rad

    sin_q <- sin(phi_q)
    cos_q <- cos(phi_q)
    sin_r <- sin(phi_r)
    cos_r <- cos(phi_r)

    x <- cos_r * sin(delta_lon)
    y <- cos_q * sin_r - sin_q * cos_r * cos(delta_lon)
    z <- sin_q * sin_r + cos_q * cos_r * cos(delta_lon)

    atan2(sqrt(x * x + y * y), z)
  }

  .exact_distance <- function(ref, query, distance, radius) {
    if (!nrow(ref)) return(numeric(0L))

    if (identical(distance, "Eucl")) {
      delta <- sweep(ref, 2L, query, FUN = "-")
      return(sqrt(rowSums(delta * delta)))
    }

    angle <- .central_angle(ref, query)
    if (identical(distance, "Geod")) {
      radius * angle
    } else {
      2 * radius * sin(angle / 2)
    }
  }

  .to_xyz <- function(x, radius) {
    lat <- x[, 2L] * pi / 180
    lon <- x[, 1L] * pi / 180
    radius * cbind(
      cos(lat) * cos(lon),
      cos(lat) * sin(lon),
      sin(lat)
    )
  }

  .search_radius <- function(maxdist, distance, radius) {
    if (is.null(maxdist)) return(NULL)
    if (maxdist == 0) return(NULL)

    if (identical(distance, "Geod")) {
      if (maxdist >= pi * radius) return(NULL)
      return(2 * radius * sin(maxdist / (2 * radius)))
    }

    if (identical(distance, "Chor")) {
      if (maxdist >= 2 * radius) return(NULL)
      return(maxdist)
    }

    maxdist
  }

  .find_neighbors <- function(ref, query, distance, maxdist, neighb, radius) {
    ref <- .check_numeric_matrix(ref, "reference coordinates")
    query <- .check_numeric_matrix(query, "prediction locations")

    if (ncol(ref) != ncol(query)) {
      stop("Reference coordinates and prediction locations must have ",
           "the same number of columns.", call. = FALSE)
    }

    if (distance %in% c("Geod", "Chor")) {
      .check_spherical_coords(ref, "reference coordinates")
      .check_spherical_coords(query, "prediction locations")
      ref_search <- .to_xyz(ref, radius)
      query_search <- .to_xyz(query, radius)
    } else {
      ref_search <- ref
      query_search <- query
    }

    nref <- nrow(ref)
    k <- if (is.null(neighb)) nref else min(neighb, nref)
    if (k < 1L) stop("No reference coordinates are available.", call. = FALSE)

    search_radius <- .search_radius(maxdist, distance, radius)
    cutoff_tol <- if (is.null(maxdist)) {
      0
    } else {
      64 * .Machine$double.eps * max(1, abs(maxdist))
    }

    finish_one <- function(idx, dist_hint, qrow) {
      ok <- !is.na(idx) & idx > 0L & idx <= nref & is.finite(dist_hint)
      idx <- as.integer(idx[ok])
      if (!length(idx)) {
        return(list(indices = integer(0L), distances = numeric(0L)))
      }

      idx <- unique(idx)
      exact <- .exact_distance(ref[idx, , drop = FALSE], qrow,
                               distance, radius)

      if (!is.null(maxdist)) {
        keep <- exact <= maxdist + cutoff_tol
        idx <- idx[keep]
        exact <- exact[keep]
      }

      if (!length(idx)) {
        return(list(indices = integer(0L), distances = numeric(0L)))
      }

      ord <- order(exact, idx)
      idx <- idx[ord]
      exact <- exact[ord]

      if (!is.null(neighb) && length(idx) > neighb) {
        idx <- idx[seq_len(neighb)]
        exact <- exact[seq_len(neighb)]
      }

      list(indices = idx, distances = exact)
    }

    ## Radius-only neighborhoods may contain all reference points. Processing
    ## one query at a time avoids allocating a nquery x nref result matrix.
    radius_only <- is.null(neighb) && !is.null(maxdist)

    if (radius_only) {
      return(.map(seq_len(nrow(query_search)), function(i) {
        args <- list(
          data = ref_search,
          query = matrix(query_search[i, ], nrow = 1L),
          k = k
        )
        if (!is.null(search_radius)) args$radius <- search_radius
        ans <- do.call(nabor::knn, args)
        finish_one(ans$nn.idx[1L, ], ans$nn.dists[1L, ],
                   query[i, ])
      }))
    }

    args <- list(data = ref_search, query = query_search, k = k)
    if (!is.null(search_radius)) args$radius <- search_radius
    ans <- do.call(nabor::knn, args)

    lapply(seq_len(nrow(query)), function(i) {
      finish_one(ans$nn.idx[i, ], ans$nn.dists[i, ], query[i, ])
    })
  }

  .select_times <- function(target_time, coordt, maxtime) {
    delta <- abs(coordt - target_time)
    idx <- if (is.null(maxtime)) {
      seq_along(coordt)
    } else {
      which(delta <= maxtime +
              64 * .Machine$double.eps * max(1, abs(maxtime)))
    }
    if (!length(idx)) return(integer(0L))
    idx[order(coordt[idx], idx)]
  }

  ##########################################################################
  ## Handle sp/spacetime objects before determining the data structure
  ##########################################################################

  if (!is.null(spobj)) {
    spatial_info <- sp2Geo(spobj, spdata)
    coordx <- spatial_info$coords
    if (!is.null(spatial_info$coordt)) coordt <- spatial_info$coordt
    if (!isTRUE(spatial_info$pj) && !identical(distance, "Chor")) {
      distance <- "Geod"
    }
    if (!is.null(spatial_info$Y)) data <- spatial_info$Y
    if (!is.null(spatial_info$X)) X <- spatial_info$X
  }

  ##########################################################################
  ## Main input validation
  ##########################################################################

  if (missing(loc)) stop("Parameter 'loc' is required.", call. = FALSE)
  if (!.is_scalar_logical(bivariate)) {
    stop("bivariate must be TRUE or FALSE.", call. = FALSE)
  }
  if (!.is_scalar_logical(grid)) {
    stop("grid must be TRUE or FALSE.", call. = FALSE)
  }
  if (!.is_scalar_logical(parallel)) {
    stop("parallel must be TRUE or FALSE.", call. = FALSE)
  }

  distance <- match.arg(distance, c("Eucl", "Chor", "Geod"))
  neighb <- .normalise_neighb(neighb)
  maxdist <- .check_nonnegative_scalar(maxdist, "maxdist")
  maxtime <- .check_nonnegative_scalar(maxtime, "maxtime")

  if (!is.numeric(radius) || length(radius) != 1L ||
      !is.finite(radius) || radius <= 0) {
    stop("radius must be a finite positive scalar.", call. = FALSE)
  }
  radius <- as.numeric(radius)

  if (is.vector(loc) && !is.list(loc)) loc <- matrix(loc, nrow = 1L)
  loc <- .check_numeric_matrix(as.matrix(loc), "loc")

  coords <- .build_coords(coordx, coordy, coordz, grid)

  .GeoValidate3D(
    coordx = coords, coordx_dyn = coordx_dyn,
    distance = distance, grid = grid,
    context = "GeoNeighborhood"
  )
  .GeoValidate3D(
    coordx = loc, distance = distance, grid = FALSE,
    context = "GeoNeighborhood prediction locations"
  )
  spacetime <- !is.null(coordt)
  is_dynamic <- !is.null(coordx_dyn)

  ## A two-element dynamic coordinate list without temporal coordinates is the
  ## standard representation of a bivariate field on different locations.
  if (!isTRUE(bivariate) && !spacetime && is_dynamic &&
      is.list(coordx_dyn) && length(coordx_dyn) == 2L) {
    bivariate <- TRUE
  }

  if (spacetime && isTRUE(bivariate)) {
    stop("Simultaneous bivariate and spatiotemporal neighborhood selection ",
         "is not implemented.", call. = FALSE)
  }

  if (!spacetime && !isTRUE(bivariate) &&
      is.null(neighb) && is.null(maxdist)) {
    stop("For spatial data, specify neighb or maxdist.", call. = FALSE)
  }

  if (spacetime) {
    if (!is.numeric(coordt) || !length(coordt) || any(!is.finite(coordt))) {
      stop("coordt must be a non-empty finite numeric vector.", call. = FALSE)
    }
    if (is.null(time) || !is.numeric(time) || !length(time) ||
        any(!is.finite(time))) {
      stop("time must be a non-empty finite numeric vector for ",
           "spatiotemporal neighborhoods.", call. = FALSE)
    }
    coordt <- as.numeric(coordt)
    time <- as.numeric(time)
  }

  if (distance %in% c("Geod", "Chor")) {
    .check_spherical_coords(loc, "loc")
  }

  if (!is_dynamic && is.null(coords)) {
    stop("Observed coordinates are missing.", call. = FALSE)
  }

  if (!is.null(coords) && distance %in% c("Geod", "Chor")) {
    .check_spherical_coords(coords, "observed coordinates")
  }

  if (spacetime && is_dynamic) {
    if (!is.list(coordx_dyn) || length(coordx_dyn) != length(coordt)) {
      stop("For dynamic spatiotemporal data, coordx_dyn must be a list ",
           "with one coordinate matrix for each element of coordt.",
           call. = FALSE)
    }
    coordx_dyn <- lapply(seq_along(coordx_dyn), function(i) {
      z <- .check_numeric_matrix(coordx_dyn[[i]],
                                 paste0("coordx_dyn[[", i, "]]"))
      if (distance %in% c("Geod", "Chor")) {
        .check_spherical_coords(z, paste0("coordx_dyn[[", i, "]]"))
      }
      z
    })
  }

  if (isTRUE(bivariate) && is_dynamic) {
    if (!is.list(coordx_dyn) || length(coordx_dyn) != 2L) {
      stop("For a bivariate field on different locations, coordx_dyn ",
           "must contain two coordinate matrices.", call. = FALSE)
    }
    coordx_dyn <- lapply(seq_len(2L), function(i) {
      z <- .check_numeric_matrix(coordx_dyn[[i]],
                                 paste0("coordx_dyn[[", i, "]]"))
      if (distance %in% c("Geod", "Chor")) {
        .check_spherical_coords(z, paste0("coordx_dyn[[", i, "]]"))
      }
      z
    })
  }

  .GeoCheckDuplicateLocations(
    coordx = coords, coordt = coordt, coordx_dyn = coordx_dyn,
    spacetime = spacetime, bivariate = bivariate, grid = FALSE,
    context = "GeoNeighborhood", check.duplicates = check.duplicates
  )

  ##########################################################################
  ## Component subsetting helpers
  ##########################################################################

  .subset_rows <- function(x, idx, n_total, name) {
    if (is.null(x)) return(NULL)
    if (is.vector(x) && !is.list(x)) {
      if (length(x) != n_total) {
        stop(name, " has an incompatible length.", call. = FALSE)
      }
      return(x[idx])
    }
    x <- as.matrix(x)
    if (nrow(x) != n_total) {
      stop(name, " has an incompatible number of rows.", call. = FALSE)
    }
    x[idx, , drop = FALSE]
  }

  .subset_static_data <- function(x, flat_idx, time_idx, spatial_idx,
                                  nt, ns) {
    if (is.null(x)) return(NULL)
    if (is.matrix(x)) {
      if (identical(dim(x), c(nt, ns))) {
        return(as.numeric(t(x[time_idx, spatial_idx, drop = FALSE])))
      }
      if (identical(dim(x), c(ns, nt))) {
        return(as.numeric(x[spatial_idx, time_idx, drop = FALSE]))
      }
    }
    if (is.vector(x) && !is.list(x) && length(x) == nt * ns) {
      return(x[flat_idx])
    }
    stop("data has an incompatible format for static spatiotemporal data.",
         call. = FALSE)
  }

  .dynamic_piece <- function(x, time_idx, local_idx, offsets, name,
                             matrix_result = FALSE) {
    if (is.null(x)) return(NULL)

    if (is.list(x)) {
      if (length(x) < time_idx) {
        stop(name, " has too few temporal components.", call. = FALSE)
      }
      piece <- x[[time_idx]]
      if (matrix_result) {
        piece <- as.matrix(piece)
        return(piece[local_idx, , drop = FALSE])
      }
      return(piece[local_idx])
    }

    global_idx <- offsets[time_idx] + local_idx
    if (matrix_result) {
      x <- as.matrix(x)
      if (nrow(x) < max(global_idx, 0L)) {
        stop(name, " has an incompatible number of rows.", call. = FALSE)
      }
      return(x[global_idx, , drop = FALSE])
    }

    if (length(x) < max(global_idx, 0L)) {
      stop(name, " has an incompatible length.", call. = FALSE)
    }
    x[global_idx]
  }

  ##########################################################################
  ## Pure spatial processor
  ##########################################################################

  .process_spatial <- function() {
    n_obs <- nrow(coords)
    neighbors <- .find_neighbors(coords, loc, distance, maxdist,
                                 neighb, radius)

    pieces <- .map(seq_len(nrow(loc)), function(i) {
      idx <- neighbors[[i]]$indices
      list(
        coordx = coords[idx, , drop = FALSE],
        indices = idx,
        distances = neighbors[[i]]$distances,
        data = .subset_rows(data, idx, n_obs, "data"),
        X = .subset_rows(X, idx, n_obs, "X"),
        M = .subset_rows(M, idx, n_obs, "M"),
        numpoints = length(idx),
        numcoord = length(idx),
        numtime = 1L,
        block_sizes = length(idx)
      )
    })

    list(
      coordx = lapply(pieces, `[[`, "coordx"),
      coordx_dyn = NULL,
      coordt = NULL,
      coordt_obs = NULL,
      indices = lapply(pieces, `[[`, "indices"),
      indices_by_time = NULL,
      distances = lapply(pieces, `[[`, "distances"),
      data = if (is.null(data)) NULL else lapply(pieces, `[[`, "data"),
      X = if (is.null(X)) NULL else lapply(pieces, `[[`, "X"),
      M = if (is.null(M)) NULL else lapply(pieces, `[[`, "M"),
      numpoints = vapply(pieces, `[[`, integer(1L), "numpoints"),
      numcoord = vapply(pieces, `[[`, integer(1L), "numcoord"),
      numtime = vapply(pieces, `[[`, integer(1L), "numtime"),
      block_sizes = lapply(pieces, `[[`, "block_sizes")
    )
  }

  ##########################################################################
  ## Static spatiotemporal processor
  ##########################################################################

  .process_spacetime_static <- function() {
    ns <- nrow(coords)
    nt <- length(coordt)
    n_total <- ns * nt

    spatial_neighbors <- .find_neighbors(coords, loc, distance, maxdist,
                                         neighb, radius)
    temporal_neighbors <- lapply(time, .select_times,
                                 coordt = coordt, maxtime = maxtime)

    ## Local prediction tasks use location-major order throughout GeoKrigloc:
    ## all requested times for location 1, then all times for location 2, etc.
    ## Do not use expand.grid(location, target_time) here because its first
    ## argument varies fastest and would therefore create time-major blocks.
    task_grid <- data.frame(
      location = rep(seq_len(nrow(loc)), each = length(time)),
      target_time = rep(seq_along(time), times = nrow(loc))
    )

    pieces <- .map(seq_len(nrow(task_grid)), function(k) {
      i <- task_grid$location[k]
      j <- task_grid$target_time[k]
      sidx <- spatial_neighbors[[i]]$indices
      tidx <- temporal_neighbors[[j]]

      flat_idx <- if (length(sidx) && length(tidx)) {
        as.integer(unlist(lapply(tidx, function(tt) sidx + ns * (tt - 1L)),
                          use.names = FALSE))
      } else {
        integer(0L)
      }

      list(
        coordx = coords[sidx, , drop = FALSE],
        coordt = coordt[tidx],
        coordt_obs = rep(coordt[tidx], each = length(sidx)),
        indices = flat_idx,
        indices_by_time = rep(list(sidx), length(tidx)),
        distances = spatial_neighbors[[i]]$distances,
        data = .subset_static_data(data, flat_idx, tidx, sidx, nt, ns),
        X = .subset_rows(X, flat_idx, n_total, "X"),
        M = .subset_rows(M, flat_idx, n_total, "M"),
        numpoints = length(flat_idx),
        numcoord = length(sidx),
        numtime = length(tidx),
        block_sizes = rep(length(sidx), length(tidx))
      )
    })

    list(
      coordx = lapply(pieces, `[[`, "coordx"),
      coordx_dyn = NULL,
      coordt = lapply(pieces, `[[`, "coordt"),
      coordt_obs = lapply(pieces, `[[`, "coordt_obs"),
      indices = lapply(pieces, `[[`, "indices"),
      indices_by_time = lapply(pieces, `[[`, "indices_by_time"),
      distances = lapply(pieces, `[[`, "distances"),
      data = if (is.null(data)) NULL else lapply(pieces, `[[`, "data"),
      X = if (is.null(X)) NULL else lapply(pieces, `[[`, "X"),
      M = if (is.null(M)) NULL else lapply(pieces, `[[`, "M"),
      numpoints = vapply(pieces, `[[`, integer(1L), "numpoints"),
      numcoord = vapply(pieces, `[[`, integer(1L), "numcoord"),
      numtime = vapply(pieces, `[[`, integer(1L), "numtime"),
      block_sizes = lapply(pieces, `[[`, "block_sizes")
    )
  }

  ##########################################################################
  ## Dynamic spatiotemporal processor
  ##########################################################################

  .process_spacetime_dynamic <- function() {
    nt <- length(coordt)
    sizes <- vapply(coordx_dyn, nrow, integer(1L))
    offsets <- c(0L, cumsum(sizes))[seq_len(nt)]
    temporal_neighbors <- lapply(time, .select_times,
                                 coordt = coordt, maxtime = maxtime)

    ## Local prediction tasks use location-major order throughout GeoKrigloc:
    ## all requested times for location 1, then all times for location 2, etc.
    ## Do not use expand.grid(location, target_time) here because its first
    ## argument varies fastest and would therefore create time-major blocks.
    task_grid <- data.frame(
      location = rep(seq_len(nrow(loc)), each = length(time)),
      target_time = rep(seq_along(time), times = nrow(loc))
    )

    pieces <- .map(seq_len(nrow(task_grid)), function(k) {
      i <- task_grid$location[k]
      j <- task_grid$target_time[k]
      tidx <- temporal_neighbors[[j]]

      if (!length(tidx)) {
        return(list(
          coordx = matrix(numeric(0L), nrow = 0L, ncol = ncol(loc)),
          coordx_dyn = list(), coordt = numeric(0L),
          coordt_obs = numeric(0L), indices = integer(0L),
          indices_by_time = list(), distances = list(), data = NULL,
          X = NULL, M = NULL, numpoints = 0L, numcoord = 0L,
          numtime = 0L, block_sizes = integer(0L)
        ))
      }

      blocks <- lapply(tidx, function(tt) {
        nb <- .find_neighbors(
          coordx_dyn[[tt]],
          matrix(loc[i, ], nrow = 1L),
          distance, maxdist, neighb, radius
        )[[1L]]

        idx <- nb$indices
        list(
          coords = coordx_dyn[[tt]][idx, , drop = FALSE],
          local_indices = idx,
          global_indices = offsets[tt] + idx,
          distances = nb$distances,
          data = .dynamic_piece(data, tt, idx, offsets, "data"),
          X = .dynamic_piece(X, tt, idx, offsets, "X", TRUE),
          M = .dynamic_piece(M, tt, idx, offsets, "M")
        )
      })

      block_sizes <- vapply(blocks, function(z) nrow(z$coords), integer(1L))
      keep_blocks <- block_sizes > 0L
      if (!any(keep_blocks)) {
        return(list(
          coordx = matrix(numeric(0L), nrow = 0L, ncol = ncol(loc)),
          coordx_dyn = list(), coordt = numeric(0L),
          coordt_obs = numeric(0L), indices = integer(0L),
          indices_by_time = list(), distances = list(), data = NULL,
          X = NULL, M = NULL, numpoints = 0L, numcoord = 0L,
          numtime = 0L, block_sizes = integer(0L)
        ))
      }
      blocks <- blocks[keep_blocks]
      tidx <- tidx[keep_blocks]
      block_sizes <- block_sizes[keep_blocks]
      coord_blocks <- lapply(blocks, `[[`, "coords")
      flat_coords <- do.call(rbind, coord_blocks)

      list(
        coordx = flat_coords,
        coordx_dyn = coord_blocks,
        coordt = coordt[tidx],
        coordt_obs = rep(coordt[tidx], block_sizes),
        indices = as.integer(unlist(lapply(blocks, `[[`, "global_indices"),
                                    use.names = FALSE)),
        indices_by_time = lapply(blocks, `[[`, "local_indices"),
        distances = lapply(blocks, `[[`, "distances"),
        data = if (is.null(data)) NULL else
          unlist(lapply(blocks, `[[`, "data"), use.names = FALSE),
        X = if (is.null(X)) NULL else do.call(rbind, lapply(blocks, `[[`, "X")),
        M = if (is.null(M)) NULL else
          unlist(lapply(blocks, `[[`, "M"), use.names = FALSE),
        numpoints = sum(block_sizes),
        numcoord = sum(block_sizes),
        numtime = length(tidx),
        block_sizes = block_sizes
      )
    })

    list(
      coordx = lapply(pieces, `[[`, "coordx"),
      coordx_dyn = lapply(pieces, `[[`, "coordx_dyn"),
      coordt = lapply(pieces, `[[`, "coordt"),
      coordt_obs = lapply(pieces, `[[`, "coordt_obs"),
      indices = lapply(pieces, `[[`, "indices"),
      indices_by_time = lapply(pieces, `[[`, "indices_by_time"),
      distances = lapply(pieces, `[[`, "distances"),
      data = if (is.null(data)) NULL else lapply(pieces, `[[`, "data"),
      X = if (is.null(X)) NULL else lapply(pieces, `[[`, "X"),
      M = if (is.null(M)) NULL else lapply(pieces, `[[`, "M"),
      numpoints = vapply(pieces, `[[`, integer(1L), "numpoints"),
      numcoord = vapply(pieces, `[[`, integer(1L), "numcoord"),
      numtime = vapply(pieces, `[[`, integer(1L), "numtime"),
      block_sizes = lapply(pieces, `[[`, "block_sizes")
    )
  }

  ##########################################################################
  ## Bivariate processor
  ##########################################################################

  .process_bivariate <- function() {
    if (!is_dynamic) {
      n1 <- nrow(coords)
      n2 <- n1
      neighbors <- .find_neighbors(coords, loc, distance, maxdist,
                                   neighb, radius)

      pieces <- .map(seq_len(nrow(loc)), function(i) {
        idx <- neighbors[[i]]$indices
        global_idx <- c(idx, n1 + idx)

        data_i <- NULL
        if (!is.null(data)) {
          if (is.matrix(data) && nrow(data) == 2L && ncol(data) == n1) {
            data_i <- c(data[1L, idx], data[2L, idx])
          } else if (is.list(data) && length(data) == 2L) {
            data_i <- c(data[[1L]][idx], data[[2L]][idx])
          } else if (is.vector(data) && length(data) == n1 + n2) {
            data_i <- data[global_idx]
          } else {
            stop("data has an incompatible format for bivariate data.",
                 call. = FALSE)
          }
        }

        M_i <- NULL
        if (!is.null(M)) {
          if (is.list(M) && length(M) == 2L) {
            M_i <- c(M[[1L]][idx], M[[2L]][idx])
          } else if (is.vector(M) && length(M) == n1 + n2) {
            M_i <- M[global_idx]
          } else if (is.vector(M) && length(M) == 2L) {
            M_i <- M
          } else {
            stop("M has an incompatible format for bivariate data.",
                 call. = FALSE)
          }
        }

        X_i <- NULL
        if (!is.null(X)) {
          if (is.list(X) && length(X) == 2L) {
            X_i <- rbind(as.matrix(X[[1L]])[idx, , drop = FALSE],
                         as.matrix(X[[2L]])[idx, , drop = FALSE])
          } else {
            Xm <- as.matrix(X)
            if (nrow(Xm) == n1 + n2) {
              X_i <- Xm[global_idx, , drop = FALSE]
            } else if (nrow(Xm) == n1) {
              X_i <- rbind(Xm[idx, , drop = FALSE],
                           Xm[idx, , drop = FALSE])
            } else {
              stop("X has an incompatible number of rows for bivariate data.",
                   call. = FALSE)
            }
          }
        }

        list(
          coordx = coords[idx, , drop = FALSE],
          coordx_dyn = list(coords[idx, , drop = FALSE],
                            coords[idx, , drop = FALSE]),
          indices = global_idx,
          indices_by_variable = list(idx, idx),
          distances = list(neighbors[[i]]$distances,
                           neighbors[[i]]$distances),
          data = data_i, X = X_i, M = M_i,
          numpoints = length(global_idx),
          numcoord = length(idx), numtime = 1L,
          block_sizes = c(length(idx), length(idx))
        )
      })
    } else {
      coords1 <- coordx_dyn[[1L]]
      coords2 <- coordx_dyn[[2L]]
      n1 <- nrow(coords1)
      n2 <- nrow(coords2)
      neighbors1 <- .find_neighbors(coords1, loc, distance, maxdist,
                                    neighb, radius)
      neighbors2 <- .find_neighbors(coords2, loc, distance, maxdist,
                                    neighb, radius)

      pieces <- .map(seq_len(nrow(loc)), function(i) {
        idx1 <- neighbors1[[i]]$indices
        idx2 <- neighbors2[[i]]$indices
        global_idx <- c(idx1, n1 + idx2)
        block1 <- coords1[idx1, , drop = FALSE]
        block2 <- coords2[idx2, , drop = FALSE]

        data_i <- NULL
        if (!is.null(data)) {
          if (is.list(data) && length(data) == 2L) {
            data_i <- c(data[[1L]][idx1], data[[2L]][idx2])
          } else if (is.vector(data) && length(data) == n1 + n2) {
            data_i <- data[global_idx]
          } else {
            stop("data has an incompatible format for bivariate data ",
                 "on different locations.", call. = FALSE)
          }
        }

        M_i <- NULL
        if (!is.null(M)) {
          if (is.list(M) && length(M) == 2L) {
            M_i <- c(M[[1L]][idx1], M[[2L]][idx2])
          } else if (is.vector(M) && length(M) == n1 + n2) {
            M_i <- M[global_idx]
          } else if (is.vector(M) && length(M) == 2L) {
            M_i <- M
          } else {
            stop("M has an incompatible format for bivariate data ",
                 "on different locations.", call. = FALSE)
          }
        }

        X_i <- NULL
        if (!is.null(X)) {
          if (is.list(X) && length(X) == 2L) {
            X_i <- rbind(as.matrix(X[[1L]])[idx1, , drop = FALSE],
                         as.matrix(X[[2L]])[idx2, , drop = FALSE])
          } else {
            Xm <- as.matrix(X)
            if (nrow(Xm) != n1 + n2) {
              stop("X has an incompatible number of rows for bivariate data ",
                   "on different locations.", call. = FALSE)
            }
            X_i <- Xm[global_idx, , drop = FALSE]
          }
        }

        list(
          coordx = rbind(block1, block2),
          coordx_dyn = list(block1, block2),
          indices = global_idx,
          indices_by_variable = list(idx1, idx2),
          distances = list(neighbors1[[i]]$distances,
                           neighbors2[[i]]$distances),
          data = data_i, X = X_i, M = M_i,
          numpoints = length(global_idx),
          numcoord = length(global_idx), numtime = 1L,
          block_sizes = c(length(idx1), length(idx2))
        )
      })
    }

    list(
      coordx = lapply(pieces, `[[`, "coordx"),
      coordx_dyn = lapply(pieces, `[[`, "coordx_dyn"),
      coordt = NULL,
      coordt_obs = NULL,
      indices = lapply(pieces, `[[`, "indices"),
      indices_by_time = lapply(pieces, `[[`, "indices_by_variable"),
      distances = lapply(pieces, `[[`, "distances"),
      data = if (is.null(data)) NULL else lapply(pieces, `[[`, "data"),
      X = if (is.null(X)) NULL else lapply(pieces, `[[`, "X"),
      M = if (is.null(M)) NULL else lapply(pieces, `[[`, "M"),
      numpoints = vapply(pieces, `[[`, integer(1L), "numpoints"),
      numcoord = vapply(pieces, `[[`, integer(1L), "numcoord"),
      numtime = vapply(pieces, `[[`, integer(1L), "numtime"),
      block_sizes = lapply(pieces, `[[`, "block_sizes")
    )
  }

  ##########################################################################
  ## Dispatch and output
  ##########################################################################

  results <- if (isTRUE(bivariate)) {
    .process_bivariate()
  } else if (spacetime && is_dynamic) {
    .process_spacetime_dynamic()
  } else if (spacetime) {
    .process_spacetime_static()
  } else {
    .process_spatial()
  }

  list(
    data = results$data,
    coordx = results$coordx,
    coordx_dyn = results$coordx_dyn,
    coordt = results$coordt,
    coordt_obs = results$coordt_obs,
    indices = results$indices,
    indices_by_time = results$indices_by_time,
    distances = results$distances,
    distance = distance,
    numpoints = results$numpoints,
    numcoord = results$numcoord,
    numtime = results$numtime,
    block_sizes = results$block_sizes,
    radius = radius,
    spacetime = spacetime,
    bivariate = bivariate,
    dynamic = is_dynamic,
    X = results$X,
    M = results$M
  )
}

####################################################
### File name: GeoSim.r
####################################################


# Simulate spatial and spatio-temporal random felds:
GeoSim <- function(coordx=NULL, coordy=NULL,coordz=NULL, coordt=NULL, coordx_dyn=NULL,corrmodel, distance="Eucl", grid=FALSE,
     method="cholesky",model='Gaussian', n=1, param, anisopars=NULL, radius=1,
      sparse=FALSE,X=NULL,spobj=NULL,nrep=1,progress=TRUE,check.duplicates=FALSE)
{
    call <- match.call()
####################################################################
############ internal function #####################################
####################################################################
##############################################################################
     RFfct1<- function(ccov,dime,nuisance,simd,X,ns)
    {
        numcoord=ccov$numcoord; numtime=ccov$numtime;grid=ccov$grid;
        spacetime=ccov$spacetime;bivariate=ccov$bivariate

        if(!bivariate) {
          external_mean <- .GeoMean_external(nuisance, n = dime)
          if(!is.null(external_mean)) {
            sim <- external_mean + as.numeric(simd)
          } else {
            X <- .GeoMean_design(X, n = dime, name = "X")
            beta <- .GeoMean_beta(nuisance, ncol(X))
            sim <- as.numeric(X %*% beta) + as.numeric(simd)
          }
        }
        if(bivariate) {
          X <- .GeoBivariateDesign(X, ns, context = "GeoSim")
          mu_biv <- .GeoBivariateObservedMean(
            nuisance, X, ns, context = "GeoSim"
          )
          sim <- mu_biv + as.numeric(simd)
        }
            if(!spacetime&&!bivariate) sim <- c(sim)
            else sim <- matrix(sim, nrow=numtime, ncol=numcoord,byrow=TRUE)
        return(sim)
    }
####################################################################
############# END internal functions ###############################
####################################################################


    if(!is.character(corrmodel) || length(corrmodel) != 1L) stop("the name of the correlation model is wrong")
    corrmodel <- .GeoSimulationCanonicalCorrModel(corrmodel)
    if(is.null(CkCorrModel(corrmodel))) stop("the name of the correlation model is wrong")
    model_requested <- gsub("[[:blank:]]", "",model)
    model <- .GeoSimulationCanonicalModel(model_requested)
    n <- .GeoSimulationEffectiveN(model_requested, model, n)
    nrep <- .GeoValidateNrep(nrep)
    .GeoValidateDirectSimulationModel(model, simulator = "GeoSim")

    distance=gsub("[[:blank:]]", "",distance)
    method <- .GeoValidateExactSimulationMethod(method)

if(is.null(coordz))
{
    if(grid) { xgrid=coordx;ygrid=coordy;
               numxgrid=length(xgrid);numygrid=length(ygrid) }
}
else {if(grid) stop("grid can not be coupled with coordz \n ")}


spacetime_dyn=FALSE


param <- lapply(param, unname) ## carlo
param_input <- param

# Enforce the finite-moment Tukey-hh parameter domain at the public
# simulation entry point.  Keep this explicit even though the common
# marginal validator performs the same check: GeoSim/GeoSimapprox must
# never proceed with h_l or h_r >= 1/2.
if (identical(model, "Tukeyh2")) {
    for (nm in c("tail1", "tail2")) {
        value <- suppressWarnings(as.numeric(param[[nm]]))
        if (length(value) != 1L || !is.finite(value) || value < 0 || value >= 0.5)
            stop(sprintf("param$%s must lie in [0, 0.5) for Tukeyh2 simulation.", nm),
                 call. = FALSE)
    }
}

param <- .GeoValidateLatentIntegerParameters(model, param)
param <- .GeoValidateSimulationMarginal(model, param)
##############################################################################
###### extracting sp object informations if necessary              ###########
##############################################################################
bivariate<-CheckBiv(CkCorrModel(corrmodel))
spacetime<-CheckST(CkCorrModel(corrmodel))
space=!spacetime&&!bivariate
if(bivariate && !(model %in% c("Gaussian", "Gauss"))) {
  stop("GeoSim currently supports bivariate simulation only for model = 'Gaussian'.",
       call. = FALSE)
}
if(!is.null(spobj)) {
   if(space||bivariate){
        a=sp2Geo(spobj,NULL); coordx=a$coords 
       if(!a$pj) {if(distance!="Chor") distance="Geod"}
    }
   if(spacetime){
        a=sp2Geo(spobj,NULL); coordx=a$coords ; coordt=a$coordt 
        if(!a$pj) {if(distance!="Chor") distance="Geod"}
     }
}
###############################################################
###############################################################


    .GeoValidate3D(
      coordx = coordx, coordy = coordy, coordz = coordz,
      coordx_dyn = coordx_dyn, distance = distance, grid = grid,
      context = "GeoSim"
    )

    if(spacetime && !is.null(coordx_dyn)) {
      coordx_dyn <- .GeoNormalizeDynamicCoordinates(
        coordx_dyn, coordt = coordt, context = "GeoSim"
      )
      .GeoValidateDynamicRows(X, coordx_dyn, name = "X", context = "GeoSim")
    }
    if(bivariate) {
      biv_info <- .GeoBivariateCoordinateInfo(
        coordx = coordx, coordy = coordy, coordz = coordz,
        coordx_dyn = coordx_dyn, grid = grid, context = "GeoSim"
      )
      if(biv_info$dynamic) coordx_dyn <- biv_info$blocks
      if(!is.null(X)) {
        X <- .GeoBivariateDesign(X, biv_info$ns, context = "GeoSim")
      }
    }

    .GeoCheckDuplicateLocations(
      coordx = coordx, coordy = coordy, coordz = coordz, coordt = coordt,
      coordx_dyn = coordx_dyn, spacetime = spacetime, bivariate = bivariate,
      grid = grid, context = "GeoSim", check.duplicates = check.duplicates
    )
    .dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
    on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)

    if(!is.null(coordx_dyn))  spacetime_dyn=TRUE
    unname(coordt);
    if(is.null(coordx_dyn)){
    unname(coordx);unname(coordy)}

if(!bivariate) { if(is.null(param$sill))  param$sill=1 }


################################################################################
################ starting spatial and spatiotemporal case#######################
################################################################################
 if(!bivariate)
    {
 ############### check and construct the mean specification
 ntime_mean <- if(is.null(coordt)) 1L else length(coordt)
 if(!is.null(coordx_dyn)) {
   nobs_mean <- sum(vapply(coordx_dyn, function(z) nrow(as.matrix(z)), integer(1L)))
 } else if(isTRUE(grid)) {
   nz_mean <- if(is.null(coordz)) 1L else length(coordz)
   nobs_mean <- length(coordx) * length(coordy) * nz_mean * ntime_mean
 } else {
   nspace_mean <- if(is.matrix(coordx) || is.data.frame(coordx)) nrow(as.matrix(coordx)) else length(coordx)
   nobs_mean <- nspace_mean * ntime_mean
 }
 if(model %in% c("Binomial", "BinomialLogistic")) {
   n <- .GeoValidateIntegerVector(n, "n", nobs = nobs_mean, minimum = 1L)
 }
 if(model %in% c("Geom", "Geometric")) {
   n <- .GeoValidateIntegerVector(n, "n", nobs = NULL, minimum = 1L)
   if(length(n) != 1L || as.integer(n[1L]) != 1L)
     stop("For direct Geometric simulation, n is fixed to 1 (the r = 1 Negative-Binomial case).", call.=FALSE)
   n <- 1L
 }
 if(model %in% c("BinomialNeg", "BinomialNegZINB")) {
   n <- .GeoValidateIntegerVector(n, "n", nobs = NULL, minimum = 1L)
   if(length(n) != 1L)
     stop("For direct Negative-Binomial simulation, n is the common number r of successes and must be a positive integer scalar.", call.=FALSE)
 }
 external_mean <- .GeoMean_external(param, n = nobs_mean)
 if(!is.null(external_mean) && !is.null(X))
   stop("A site-specific mean vector and X cannot be used together.\n")
 if(!is.null(X)) {
   X <- .GeoMean_design(X, n = nobs_mean, name = "X")
   num_betas <- ncol(X)
   invisible(.GeoMean_beta(param, num_betas))
 } else {
   supplied_mean <- .GeoMean_supplied_names(param)
   if(is.null(external_mean) && length(setdiff(supplied_mean, "mean")))
     stop("Multiple mean coefficients require an explicit design matrix X.\n")
   num_betas <- 1L
   if(is.null(external_mean)) invisible(.GeoMean_beta(param, 1L))
 }
 param_check <- param
 if(!is.null(external_mean)) param_check$mean <- 0
 if( !all(names(param_check) %in% c(CorrParam(corrmodel), NuisParam2(model,bivariate,num_betas=num_betas))) )
   stop("Nuisance and correlation parameters must be included in param\n")
#################################
    if(model %in% c("SkewGaussian","Beta",'Kumaraswamy','Kumaraswamy2','LogGaussian',
        "Binomial","BinomialNeg",#"BinomialNegZINB",
                    "StudentT","SkewStudentT","Poisson","TwoPieceTukeyh","PoissonZIP","PoissonGamma","PoissonGammaZIP","PoissonWeibull",
                     "TwoPieceBimodal", "TwoPieceStudentT","TwoPieceGaussian","TwoPieceGauss","Tukeyh","Tukeyh2","Tukeygh","SinhAsinh",
                    "Gamma","Weibull","LogLogistic","Logistic","SkewLaplace","BinomialLogistic"))
       {
           if(spacetime_dyn && is.list(X)) X <- do.call(rbind, lapply(X, as.matrix))
           mm <- .GeoMean_vector(param, X = X, n = nobs_mean, external = external_mean)
           param <- .GeoMean_zero(param, num_betas)

        if((model %in% c("SkewGaussian","TwoPieceGaussian","Logistic","SkewLaplace",
          "TwoPieceGauss","Gamma","Weibull","LogLogistic","Poisson","PoissonZIP","Tukeyh","Tukeyh2","PoissonGamma","PoissonGammaZIP","PoissonWeibull",
          'LogGaussian',"TwoPieceTukeyh","TwoPieceBimodal", "Tukeygh","SinhAsinh",
          "Binomial","BinomialNeg",
                    "StudentT","SkewStudentT","TwoPieceStudentT","Gaussian")))  
        {
          vv<-param$sill;
          param$sill=1
        }


        if(model%in% c("SkewGaussian","SkewStudentT","TwoPieceTukeyh","TwoPieceBimodal",
               "TwoPieceStudentT","TwoPieceGaussian","TwoPieceGauss"))
               { sk<-param$skew
               if(model%in% c("TwoPieceTukeyh")) tl<-param$tail
               if(model%in% c("TwoPieceBimodal")) bimo<-param$shape
               }
       }
#################################
  if(model %in% c("Tukeygh","SinhAsinh"))  {
          param$mean=0
          sk<-param$skew; tl<-param$tail}
    if(model %in% c("Tukeyh"))  {
          param$mean=0
          tl<-param$tail}
         if(model %in% c("Tukeyh2"))  {
          param$mean=0
          t1l<-param$tail1
          t2l<-param$tail2
          }
#################################
   if(model %in% c("Wrapped"))  {
            mm <- 2 * atan(.GeoMean_vector(param, X = X, n = nobs_mean,
                                           external = external_mean)) + pi
            param <- .GeoMean_zero(param, num_betas)
        }

if(model%in% c("SkewGaussian","StudentT","SkewStudentT","TwoPieceTukeyh",
               "TwoPieceStudentT","TwoPieceGaussian"))
     { nugget=param$nugget;param$nugget=0}  ### ojo!!


zz=param[grep("mean", names(param))]   
ll=list(sill=param$sill,nugget=param$nugget)
mc=append(zz,ll)

pc=param[CorrParam(corrmodel)]



ccov = GeoCovmatrix(coordx=coordx, coordy=coordy,coordz=coordz, coordt=coordt, coordx_dyn=coordx_dyn, corrmodel=corrmodel,
                   distance=distance,grid=grid,model="Gaussian", n=n,
                param=append(mc,pc), anisopars=anisopars, radius=radius, sparse=sparse,copula=NULL,
                X=X)

######################
# matrix decomposition and square root
if(!sparse) {
   decompvarcov <- MatDecomp(ccov$covmatrix,method)
   if(is.logical(decompvarcov)){print(" Covariance matrix is not positive definite");stop()}
   sqrtvarcov <- MatSqrt(decompvarcov,method)
}
else {
 cholS <- spam::chol(ccov$covmatrix)
 # cholS is the upper triangular part of the permutated matrix Sigma
 iord <- spam::ordering(cholS, inv=TRUE)
 R <- spam::as.spam(cholS)
}

######################
##############################################
## putting 2 nugget
 if(model%in% c("PoissonZIP","BinomialNegZINB","PoissonGammaZIP"))
{
  II=diag(nrow(ccov$covmatrix));
  II[lower.tri(II)] <- (1-param$nugget2)/(1-param$nugget1)
  II <- t(II);
  II[lower.tri(II)] <- (1-param$nugget2)/(1-param$nugget1)
  ccov_with_nug=(ccov$covmatrix)*(II)

if(!sparse) {
  decompvarcov1 <- MatDecomp(ccov_with_nug,method)
  if(is.logical(decompvarcov1)){print(" Covariance matrix is not positive definite");stop()}
  sqrtvarcov1 <- MatSqrt(decompvarcov1,method)
  }
  else {
   cholS1 <- spam::chol(ccov_with_nug)
   iord1 <- spam::ordering(cholS1, inv=TRUE)
   R1 <- spam::as.spam(cholS1)
  }
}
##############################################
## putting 1 nugget
if(model%in% c("SkewGaussian","StudentT","SkewStudentT","TwoPieceTukeyh",
               "TwoPieceStudentT","TwoPieceGaussian"))
{
  II=diag(nrow(ccov$covmatrix));
  II[lower.tri(II)] <- (1-nugget);
  II <- t(II);
  II[lower.tri(II)] <- (1-nugget)
  ccov_with_nug=ccov$covmatrix*II

if(!sparse) {
  decompvarcov1 <- MatDecomp(ccov_with_nug,method)
if(is.logical(decompvarcov1)){print(" Covariance matrix is not positive definite");stop()}
  sqrtvarcov1 <- MatSqrt(decompvarcov1,method)
  }
  else {   
    cholS1 <- spam::chol(ccov_with_nug)
    iord1 <- spam::ordering(cholS1, inv=TRUE)
    R1 <- spam::as.spam(cholS1)
   }
}
##########################################################################################################################################
 numtime=1
 numcoord=ccov$numcoord;
 if(spacetime) numtime=ccov$numtime;
 dime<-numcoord*numtime

########################

   if(spacetime_dyn) {
        coords=NULL
       coords=do.call(rbind,args=c(coordx_dyn))
      ns <- vapply(coordx_dyn, nrow, integer(1L))
      if(ncol(coords)==2) {coordx <- coords[,1]; coordy <- coords[,2];coordz=NULL}
      if(ncol(coords)==3) {coordx <- coords[,1]; coordy <- coords[,2];coordz=coords[,3]}
       dime=sum(ns)
      ccov$numtime=1
   }
   else { dime=nrow(ccov$covmatrix) }
########################
    
SIM=list()
###################################################
### starting number of replicates
###################################################
if(nrep>1){

cat("Performing", nrep, "simulations using Cholesky method...\n")}
if(progress){
restore_progress_handlers <- .GeoProgressHandlersPush(
  handler = "txtprogressbar", global = TRUE
)
on.exit(restore_progress_handlers(), add = TRUE)
pb <- progressr::progressor(
  along = seq_len(nrep), on_exit = FALSE, auto_finish = FALSE
)
## Finish the progressor before restoring the progressr handlers.  This is
## essential because progressr may temporarily capture stdout while a global
## handler is active; leaving the progressor unfinished can leave that sink
## open after GeoSim() returns.
on.exit(pb(type = "finish"), add = TRUE, after = FALSE)

}

for( L in seq_len(nrep)){
    spec <- .GeoSimulationLatentSpec(model, param, n)
    k <- spec$k
    npoi <- spec$npoi
    param <- spec$param
    

  ################################################################################
  ################################################################################
   if(model %in% c("BinomialNeg","BinomialNegZINB")) {
      dd <- NULL
   } else {
      dd <- array(0, dim = c(dime, 1, k))
   }

 #########################################

#########################################################
KK=1;ssp=double(dime)

if(model%in% c("SkewGaussian","StudentT","SkewStudentT","TwoPieceTukeyh",
               "TwoPieceStudentT","TwoPieceGaussian"))
{
  ss=matrix(rnorm(dime) , nrow=dime, ncol = 1)
  if(sparse) simD=as.numeric((array(c(ss),c(1,dime)) %*% R1) [,iord1] )
    else     #simD=crossprod(sqrtvarcov1,ss) 
             simD=sqrtvarcov1%*%ss

      ccov1=ccov
      ccov1$covmatrix=ccov_with_nug
      if(space) simDD <- c(simD)
      else simDD <- matrix(simD, nrow=ccov1$numtime, ncol=ccov1$numcoord,byrow=TRUE)
}

  # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  # INIT efficient counters for BinomialNeg / BinomialNegZINB (univariate)
  if(model %in% c("BinomialNeg","BinomialNegZINB")){
    rvec <- rep.int(as.integer(n[1L]), dime)
    counts   <- integer(dime)     # successi cumulati
    finished <- logical(dime)     # chi ha raggiunto n successi
    nb_vals  <- integer(dime)     # output (tempo al successo r) = trial - r
    trial    <- 0L                # contatore di tentativi (iterazioni Bernoulli)
  }
  # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

  if(model %in% c("Poisson", "PoissonZIP", "PoissonGamma", "PoissonGammaZIP", "PoissonWeibull")) {
    poi_counts <- integer(dime)
  }

  while(KK<=npoi) {
  for(i in 1:k) {

    ss=matrix(rnorm(dime) , nrow=dime, ncol = 1)

    #### simulating N(0,1) with matrix decomposition using sparse or dense matrices without nugget!
      if(sparse)   simd=as.numeric((array(c(ss),c(1,dime)) %*% R) [,iord] )
      else         #simd=crossprod(sqrtvarcov,ss) 
                   simd=sqrtvarcov%*%ss



    #######################################################################
    nuisance<-param[ccov$namesnuis]
    sim<-RFfct1(ccov,dime,nuisance,simd,ccov$X,ns)

    ####################################
    ####### starting cases #############
    ####################################
     #if(model %in% c("Binomial", "BinomialNeg","BinomialNegZINB")) {
      #   simdim <- dim(sim)
       #  sim <- as.numeric(sim>0)
        # dim(sim) <- simdim
         # }
 if(model %in% c("Binomial", "BinomialNeg","BinomialNegZINB")) {
    simdim <- dim(sim)
    sim_vec <- if(is.null(simdim)) c(sim) else c(t(sim))
    eta <- if(length(mm) == 1) rep(mm, length(sim_vec)) else c(mm)
    sim_vec <- as.numeric(sim_vec + eta > 0)
    if(is.null(simdim)) sim <- sim_vec
    else sim <- matrix(sim_vec, nrow=simdim[1], ncol=simdim[2], byrow=TRUE)
    }    
    ####################################
    if(model %in% c("Weibull","SkewGaussian","SkewGauss","Binomial","BinomialLogistic","Poisson","PoissonGamma","PoissonGammaZIP","PoissonWeibull","PoissonZIP","Beta","Kumaraswamy","Kumaraswamy2",
              "LogGaussian","TwoPieceTukeyh",
                "Gamma","LogLogistic","Logistic","SkewLaplace","StudentT",
                "SkewStudentT","TwoPieceStudentT","TwoPieceGaussian","TwoPieceGauss","TwoPieceBimodal")) {
        dd[,,i]=t(sim)

     }
     ####################################
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    if(model %in% c("BinomialNeg","BinomialNegZINB")){
                 # vec Bernoulli corrente (ordine identico a c(t(sim)) usato prima)
                 trial <- trial + 1L
                 vec <- c(t(sim))
                 idx <- !finished
                 if(any(idx)){
                   counts[idx] <- counts[idx] + vec[idx]
                   new_finished <- (counts >= rvec) & !finished
                   if(any(new_finished)){
                     nb_vals[new_finished] <- trial - rvec[new_finished]
                     finished <- finished | new_finished
                   }
                 }
                 if(all(finished)) {break;}### stopping rule (tutti chiusi)
               }
    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

    }
 ####################################
  if(model %in% c("Poisson","PoissonZIP"))   {
   pois1=0.5*(dd[,,1]^2+dd[,,2]^2)
   ssp=ssp+c(pois1)
   active <- ssp <= c(exp(mm))
   poi_counts <- poi_counts + as.integer(active)
   if(!any(active)) break  ## stopping rule
 }
  ####################################
 if(model %in% c("PoissonGamma","PoissonGammaZIP"))   {
   if(KK==1) sim3 <- .GeoSimulationSquareSums(dd, 3:k)
   #################################
   pois1=0.5*(dd[,,1]^2+dd[,,2]^2)
   ssp=ssp+c(pois1)
   threshold <- c(exp(mm)*sim3/(k-2))
   active <- ssp <= threshold
   poi_counts <- poi_counts + as.integer(active)
   if(!any(active)) break  ## stopping rule
 }
if(model %in% c("PoissonWeibull"))   {

  if(KK==1) sim3 <- .GeoSimulationSquareSums(dd, 3:k)
   #################################
   pois1=0.5*(dd[,,1]^2+dd[,,2]^2)
   ssp=ssp+c(pois1)
   threshold <- c(exp(mm)*(sim3/2)^(1/param$shape)/(gamma(1+1/param$shape)))
   active <- ssp <= threshold
   poi_counts <- poi_counts + as.integer(active)
   if(!any(active)) break  ## stopping rule
 }

 KK=KK+1
}

 ####### end for #########################
 ###############################################################################################
 #### simulation for discrete random field based on indipendent copies  of GRF ######
 ###############################################################################################
 if(model %in% c("Binomial","BinomialLogistic","Poisson","PoissonGamma","PoissonGammaZIP","PoissonWeibull","PoissonZIP","BinomialNeg","BinomialNegZINB"))   {

   if(model %in% c("Poisson","PoissonGamma","PoissonWeibull"))   {sim=poi_counts;byrow=TRUE}
    
    if(model %in% c("PoissonZIP","PoissonGammaZIP"))   {
      ss=matrix(rnorm(dime) , nrow=dime, ncol = 1)

      if(sparse)   a=as.numeric((array(c(ss),c(1,dime)) %*% R1) [,iord1] )
      else         #a=crossprod(sqrtvarcov1,ss) 
                       a=sqrtvarcov1%*%ss

     ###
      a[a<as.numeric(param$pmu)]=0;a[a!=0]=1
      sim=a*poi_counts;
      byrow=TRUE
      }
########################################
   if(model == "Binomial") {
                  sim <- .GeoSimulationBinomialCounts(dd, n)
                  byrow <- TRUE }
########################################
if(model == "BinomialLogistic") {
                  sim <- .GeoSimulationBinomialLogisticCounts(dd, n, mm)
                  byrow <- TRUE }
#######################################
   if(model %in% c("BinomialNeg"))   {
          if(any(!finished)) {
              stop("Some locations did not reach n successes. Increase the maximum number of trials.")
          }
          sim <- nb_vals
          byrow=TRUE
          }
  if(model %in% c("BinomialNegZINB"))   {
          if(any(!finished)) {
              stop("Some locations did not reach n successes. Increase the maximum number of trials.")
          }
          sim <- nb_vals
      ss=matrix(rnorm(dime) , nrow=dime, ncol = 1)

      if(sparse)   a=as.numeric((array(c(ss),c(1,dime)) %*% R1) [,iord1] )
      else         #a=crossprod(sqrtvarcov1,ss) 
                   a=sqrtvarcov1%*%ss
     ###
          a[a<as.numeric(param$pmu)]=0;a[a!=0]=1
          sim=a*sim
          byrow=TRUE
          }
   # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
#############################################
############### formatting data #############
#############################################

    sim <- .GeoSimulationFormat(
      sim, grid = grid, space = space, numxgrid = numxgrid, numygrid = numygrid,
      numtime = numtime, numcoord = numcoord, byrow = byrow
    )
}
#########################################################################################################
#### Common transformation of latent Gaussian fields for univariate models.
#########################################################################################################
    transformed <- .GeoSimulationTransformUnivariate(
      model = model, sim = sim, dd = dd, simDD = get0("simDD", ifnotfound = NULL, inherits = FALSE),
      mm = get0("mm", ifnotfound = 0, inherits = FALSE),
      vv = get0("vv", ifnotfound = 1, inherits = FALSE),
      sk = get0("sk", ifnotfound = 0, inherits = FALSE),
      tl = get0("tl", ifnotfound = 0, inherits = FALSE),
      t1l = get0("t1l", ifnotfound = 0, inherits = FALSE),
      t2l = get0("t2l", ifnotfound = 0, inherits = FALSE),
      bimo = get0("bimo", ifnotfound = NULL, inherits = FALSE),
      param = param, k = k, spacetime = spacetime
    )
    if(transformed$handled) {
      sim <- transformed$sim
      if(transformed$needs_format) {
        sim <- .GeoSimulationFormat(
          sim, grid = grid, space = space, numxgrid = numxgrid, numygrid = numygrid,
          numtime = numtime, numcoord = numcoord, byrow = transformed$byrow
        )
      }
    }
##################################################################
###########. formatting data for space time dynamic case. #########
    if(spacetime_dyn) sim <- .GeoSimulationSplitDynamic(sim, ns)
###################################################################

SIM[[L]]=sim
if(progress) pb()
} 
######################################################################
##########  end of replicates ########################################
######################################################################
}
###################################################################
################ end spatial and spatiotemporal case###############
###################################################################


################################################################
################ starting bivariate  case#######################
################################################################
if(bivariate){
  sel1 <- substr(names(param), 1, 6) == "mean_1"
  sel2 <- substr(names(param), 1, 6) == "mean_2"
  num_betas <- c(sum(sel1), sum(sel2))

  zz <- param[grep("mean", names(param))]
  pc <- param[CorrParam(corrmodel)]
  ccov <- GeoCovmatrix(
    coordx = coordx, coordy = coordy, coordz = coordz, coordt = coordt,
    coordx_dyn = coordx_dyn, corrmodel = corrmodel, distance = distance,
    grid = grid, model = "Gaussian", n = n, param = append(pc, zz),
    anisopars = anisopars, radius = radius, sparse = sparse, copula = NULL, X = X
  )

  if(!sparse) {
    decompvarcov <- MatDecomp(ccov$covmatrix, method)
    if(is.logical(decompvarcov)) {
      print(" Covariance matrix is not positive definite")
      stop()
    }
    sqrtvarcov <- MatSqrt(decompvarcov, method)
  } else {
    cholS <- spam::chol(ccov$covmatrix)
    iord <- spam::ordering(cholS, inv = TRUE)
    R <- spam::as.spam(cholS)
  }

  ns <- as.integer(ccov$ns)
  if(length(ns) != 2L || any(ns < 1L))
    stop("GeoSim: invalid bivariate block sizes returned by GeoCovmatrix.")
  dime <- nrow(ccov$covmatrix)
  if(dime != sum(ns))
    stop("GeoSim: covariance dimension does not match the bivariate block sizes.")

  SIM <- vector("list", nrep)
  for(L in seq_len(nrep)) {
    if(spacetime_dyn) {
      ccov$numtime <- 1L
      ccov$numcoord <- dime
    }
    numcoord <- ccov$numcoord
    numtime <- ccov$numtime
    ss <- matrix(rnorm(dime), nrow = dime, ncol = 1L)
    if(sparse) simd <- as.numeric((array(c(ss), c(1L, dime)) %*% R)[, iord])
    else simd <- sqrtvarcov %*% ss

    nuisance <- ccov$param[ccov$namesnuis]
    sim <- RFfct1(ccov, dime, nuisance, simd, ccov$X, ns)
    sim <- .GeoSimulationFormat(
      c(sim), grid = grid, space = FALSE, numxgrid = numxgrid, numygrid = numygrid,
      numtime = numtime, numcoord = numcoord, byrow = FALSE
    )
    if(spacetime_dyn) sim <- .GeoSimulationSplitDynamic(sim, ns)
    SIM[[L]] <- sim
  }
}

######################################################################
##########  end of bivariate  ########################################
######################################################################

if(nrep==1) SIM=SIM[[1]] 

##################################################################
#######################################
    if(ccov$bivariate)   ccov$numtime=1

    # Delete the global variables:
    # Return the objects list:
    GeoSim <- list(bivariate = bivariate,
    coordx = ccov$coordx,
    coordy = ccov$coordy,
    coordz = ccov$coordz,
    coordt = ccov$coordt,
    coordx_dyn =coordx_dyn,
    corrmodel = corrmodel,
    data = SIM,
    distance = distance,
    grid = grid,
    model = model_requested,
    method=method,
    n=n,
    numcoord = ccov$numcoord,
    numtime = ccov$numtime,
    param = param_input,
    radius = radius,
    #randseed=.Random.seed,
    spacetime = spacetime,
    sparse=ccov$sparse,
    nrep=nrep,
    X=X)
#}
##############################################
    structure(c(GeoSim, call = call), class = c("GeoSim"))
}

## Fast exact duplicate-location validation.
## The native checker uses an open-addressing hash table and stops at the
## first duplicate, so the expected cost is O(n) and no distance matrix is
## created. Internal resampling/refitting paths can suppress a repeated check
## after the original design has already been validated.

.GeoDuplicateCheckOption <- "GeoModels.skip_duplicate_coordinate_check"

.GeoWithDuplicateCheckSuppressed <- function(expr) {
  old <- getOption(.GeoDuplicateCheckOption, FALSE)
  options(structure(list(TRUE), names = .GeoDuplicateCheckOption))
  on.exit(options(structure(list(old), names = .GeoDuplicateCheckOption)), add = TRUE)
  force(expr)
}

.GeoDuplicateCheckEnabled <- function() {
  !isTRUE(getOption(.GeoDuplicateCheckOption, FALSE))
}

.GeoSuppressDuplicateChecksForNestedCalls <- function() {
  old <- getOption(.GeoDuplicateCheckOption, FALSE)
  options(structure(list(TRUE), names = .GeoDuplicateCheckOption))
  old
}

.GeoRestoreDuplicateCheckOption <- function(old) {
  options(structure(list(old), names = .GeoDuplicateCheckOption))
  invisible(NULL)
}

.GeoCoordMatrixForDuplicateCheck <- function(coordx, coordy = NULL, coordz = NULL,
                                              context = "GeoModels") {
  if (!is.null(coordy)) {
    z <- if (is.null(coordz)) cbind(coordx, coordy) else cbind(coordx, coordy, coordz)
  } else {
    z <- as.matrix(coordx)
  }
  z <- as.matrix(z)
  if (!is.numeric(z) || !nrow(z) || !(ncol(z) %in% 1:3) || any(!is.finite(z))) {
    stop(context, ": coordinates must be a finite numeric matrix.", call. = FALSE)
  }
  storage.mode(z) <- "double"
  unname(z)
}

.GeoCheckDuplicateMatrix <- function(z, context, label = "observation locations") {
  hit <- .Call("GeoDuplicateCoordinates", z, PACKAGE = "GeoModels")
  if (hit[2L] > 0) {
    stop(
      context, ": duplicated ", label, " are not allowed. First duplicate: observations ",
      format(hit[1L], scientific = FALSE, trim = TRUE), " and ",
      format(hit[2L], scientific = FALSE, trim = TRUE), ".",
      call. = FALSE
    )
  }
  invisible(TRUE)
}

.GeoCheckDuplicateLocations <- function(coordx = NULL, coordy = NULL, coordz = NULL,
                                         coordt = NULL, coordx_dyn = NULL,
                                         spacetime = FALSE, bivariate = FALSE,
                                         grid = FALSE, context = "GeoModels",
                                         check.duplicates = FALSE) {
  if (!is.logical(check.duplicates) || length(check.duplicates) != 1L || is.na(check.duplicates)) {
    stop(context, ": check.duplicates must be TRUE or FALSE.", call. = FALSE)
  }
  if (!isTRUE(check.duplicates) || !.GeoDuplicateCheckEnabled()) return(invisible(TRUE))

  check_time <- function(tt) {
    if (is.null(tt)) return(invisible(TRUE))
    if (!is.numeric(tt) || any(!is.finite(tt)))
      stop(context, ": coordt must be finite and numeric.", call. = FALSE)
    jj <- anyDuplicated(as.numeric(tt))
    if (jj) {
      ii <- match(tt[jj], tt[seq_len(jj - 1L)])
      stop(context, ": duplicated time coordinates are not allowed. First duplicate: times ",
           ii, " and ", jj, ".", call. = FALSE)
    }
    invisible(TRUE)
  }

  if (isTRUE(grid) && is.null(coordx_dyn)) {
    axes <- Filter(Negate(is.null), list(coordx, coordy, coordz))
    labs <- c("x", "y", "z")[seq_along(axes)]
    for (j in seq_along(axes)) {
      a <- as.numeric(axes[[j]])
      if (any(!is.finite(a))) stop(context, ": grid coordinates must be finite.", call. = FALSE)
      k <- anyDuplicated(a)
      if (k) stop(context, ": duplicated values in the ", labs[j],
                  " grid axis would create duplicated locations.", call. = FALSE)
    }
    if (isTRUE(spacetime)) check_time(coordt)
    return(invisible(TRUE))
  }

  if (isTRUE(bivariate)) {
    if (!is.null(coordx_dyn)) {
      if (!is.list(coordx_dyn) || length(coordx_dyn) != 2L)
        stop(context, ": bivariate coordx_dyn must contain two coordinate blocks.", call. = FALSE)
      for (j in 1:2) {
        z <- .GeoCoordMatrixForDuplicateCheck(coordx_dyn[[j]], context = context)
        .GeoCheckDuplicateMatrix(z, context,
                                 paste0("observation locations within variable ", j))
      }
    } else {
      z <- .GeoCoordMatrixForDuplicateCheck(coordx, coordy, coordz, context)
      .GeoCheckDuplicateMatrix(z, context, "observation locations")
    }
    return(invisible(TRUE))
  }

  if (isTRUE(spacetime) && !is.null(coordx_dyn)) {
    check_time(coordt)
    if (!is.list(coordx_dyn))
      stop(context, ": coordx_dyn must be a list of coordinate matrices.", call. = FALSE)
    for (j in seq_along(coordx_dyn)) {
      z <- .GeoCoordMatrixForDuplicateCheck(coordx_dyn[[j]], context = context)
      .GeoCheckDuplicateMatrix(z, context,
                               paste0("space-time observation locations within time slice ", j))
    }
    return(invisible(TRUE))
  }

  z <- .GeoCoordMatrixForDuplicateCheck(coordx, coordy, coordz, context)
  .GeoCheckDuplicateMatrix(z, context,
                           if (isTRUE(spacetime)) "spatial locations" else "observation locations")
  if (isTRUE(spacetime)) check_time(coordt)
  invisible(TRUE)
}

####################################################
### File name: GeoCorrFct_Cop.R
####################################################

GeoCorrFct_Cop <- function(x, t = NULL, corrmodel, model = "Gaussian", copula = "Gaussian",
                           distance = "Eucl", param, radius = 6371, n = 1,
                           covariance = FALSE, variogram = FALSE)
{
  ############################################################################
  ## Latent Gaussian correlation from the native correlation engine
  ############################################################################
  CorrelationFct <- function(corrmodel, lags, lagt, numlags, numlagt,
                             mu, paramcorr, N) {
    p <- dotCall64::.C64(
      "VectCorrelation",
      SIGNATURE = c("double", "integer", "double", "integer", "integer",
                    "double", "integer", "double", "double", "double", "integer"),
      corr = dotCall64::vector_dc("double", numlags * numlagt),
      corrmodel, lags, numlags, numlagt, mu, CkModel("Gaussian"),
      c(nugget = 0, sill = 1), paramcorr, lagt, N,
      INTENT = c("rw", rep("r", 10)),
      PACKAGE = "GeoModels", VERBOSE = 0, NAOK = TRUE
    )
    p$corr
  }

  ############################################################################
  ## Checks
  ############################################################################
  call <- match.call()

  corr_id <- CkCorrModel(corrmodel)
  model_id <- CkModel(model)
  if (is.null(corr_id))
    stop("The name of the correlation model is not correct\n")
  if (is.null(model_id))
    stop("The name of the model is not correct\n")
  if (!is.numeric(x) || any(!is.finite(x)))
    stop("Distances must be finite numeric values\n")
  if (any(x < 0))
    stop("Distances must be positive\n")
  if (!is.list(param))
    stop("param must be a list\n")

  copula <- as.character(copula)[1L]
  if (is.na(copula) || !(copula %in% c("Gaussian", "Clayton", "SkewGaussian"))) {
    stop("GeoCorrFct_Cop supports copula='Gaussian', 'Clayton', or 'SkewGaussian'.",
         call. = FALSE)
  }
  if (!(as.integer(model_id) %in% .GeoCopulaPredictionModels)) {
    stop("GeoCorrFct_Cop supports the validated copula prediction margins, including Poisson, Binomial, and BinomialNeg.",
         call. = FALSE)
  }

  spacetime <- CheckST(corr_id)
  bivariate <- CheckBiv(corr_id)
  if (bivariate)
    stop("GeoCorrFct_Cop is implemented only for univariate copula margins.", call. = FALSE)

  ## GeoCorrFct_Cop is a lag-only function and has no design matrix.  Therefore
  ## location-dependent margins can only use an intercept-only marginal predictor.
  supplied_mean <- .GeoMean_supplied_names(param)
  if (length(setdiff(supplied_mean, "mean"))) {
    stop("GeoCorrFct_Cop() has no design matrix X; only the intercept parameter 'mean' is allowed.",
         call. = FALSE)
  }

  param_work <- param
  if (is.null(param_work$mean)) param_work$mean <- 0
  if (length(param_work$mean) != 1L || !is.numeric(param_work$mean) ||
      !is.finite(param_work$mean)) {
    stop("GeoCorrFct_Cop() requires a finite scalar intercept 'mean'.", call. = FALSE)
  }
  ## Keep the same convenience convention as GeoCorrFct(): when a scale-based
  ## continuous margin needs sill and it is omitted, use unit sill.
  if (is.null(param_work$sill)) param_work$sill <- 1
  if (is.null(param_work$nugget)) param_work$nugget <- 0

  nugget <- suppressWarnings(as.numeric(param_work$nugget)[1L])
  if (!is.finite(nugget) || nugget < 0 || nugget >= 1)
    stop("GeoCorrFct_Cop requires nugget in [0,1).", call. = FALSE)

  if (identical(copula, "Clayton")) {
    nu <- suppressWarnings(as.numeric(param_work$nu)[1L])
    if (!is.finite(nu) || nu <= 0 ||
        abs(nu - round(nu)) > sqrt(.Machine$double.eps)) {
      stop("The GeoModels Clayton-like copula requires a finite positive integer 'nu'.",
           call. = FALSE)
    }
  }
  if (identical(copula, "SkewGaussian")) {
    eta_cop <- suppressWarnings(as.numeric(param_work$nu)[1L])
    if (!is.finite(eta_cop) || abs(eta_cop) >= 1)
      stop("The skew-Gaussian copula requires nu (eta) in (-1,1).", call. = FALSE)
  }

  ############################################################################
  ## Setup and latent correlation
  ############################################################################
  x <- as.numeric(x)
  nx <- length(x)
  if (is.null(t)) t <- 0
  if (!is.numeric(t) || any(!is.finite(t)) || any(t < 0))
    stop("Temporal distances must be finite non-negative numeric values.", call. = FALSE)
  t <- as.numeric(t)

  if (spacetime) {
    nt <- length(t)
    lag_grid <- as.matrix(expand.grid(x, t))
  } else {
    t <- 0
    nt <- 1L
    lag_grid <- cbind(x, 0)
  }

  corr_names <- CorrelationPar(corr_id)
  if (is.null(corr_names))
    stop("Correlation parameters are not defined for this correlation model.", call. = FALSE)
  missing_corr <- setdiff(corr_names, names(param_work))
  if (length(missing_corr)) {
    stop("Missing correlation parameter(s): ", paste(missing_corr, collapse = ", "), ".",
         call. = FALSE)
  }
  parcorr <- suppressWarnings(as.numeric(unlist(param_work[corr_names], use.names = FALSE)))
  if (length(parcorr) != length(corr_names) || any(!is.finite(parcorr)))
    stop("Correlation parameters must be finite numeric scalars.", call. = FALSE)

  eta <- as.numeric(param_work$mean)

  ## Pass only marginal/copula nuisance parameters to the covariance engine.
  ## Correlation parameters do not affect the copula transformation itself and
  ## including them would unnecessarily fragment the numerical caches.
  cop_names <- NuisParam2(model, FALSE, 1L, copula = copula)
  params_cop <- unlist(param_work[cop_names], use.names = TRUE)
  if (as.integer(model_id) %in% c(11L, 16L)) {
    nn <- suppressWarnings(as.numeric(n))
    if (length(nn) != 1L || !is.finite(nn) || nn < 1 ||
        abs(nn - round(nn)) > sqrt(.Machine$double.eps))
      stop("GeoCorrFct_Cop requires a positive integer n for Binomial/BinomialNeg margins.", call. = FALSE)
    params_cop["n"] <- nn
  }

  latent_rho <- CorrelationFct(corr_id, x, t, nx, nt, eta, parcorr, n)
  latent_rho <- as.numeric(latent_rho)
  if (length(latent_rho) != nrow(lag_grid) || any(!is.finite(latent_rho)))
    stop("Non-finite latent correlations returned by the correlation model.", call. = FALSE)
  if (any(abs(latent_rho) > 1 + 1e-8))
    stop("The latent correlation model returned values outside [-1,1].", call. = FALSE)
  latent_rho <- pmax(-1, pmin(1, latent_rho))

  ## The nugget acts on the latent correlation for distinct observations, exactly
  ## as in Cmatrix_copula().  At the exact zero lag GeoCorrFct_Cop represents the
  ## same random variable, hence its covariance is the marginal variance.
  rho_cop <- (1 - nugget) * latent_rho
  zero_lag <- lag_grid[, 1L] == 0 & lag_grid[, 2L] == 0
  rho_cop[zero_lag] <- 1

  ############################################################################
  ## Observed-scale covariance using the common three-copula dispatcher
  ############################################################################
  nout <- length(rho_cop)
  cova <- .GeoCopulaCrossCovFast(
    copula = copula,
    rho = rho_cop,
    model_type = model_id,
    params = params_cop,
    eta1 = rep(eta, nout),
    eta2 = rep(eta, nout)
  )

  moments <- .GeoCopulaMarginalMoments(model_id, params_cop, eta)
  marginal_variance <- as.numeric(moments$variance)[1L]
  if (!is.finite(marginal_variance) || marginal_variance <= 0)
    stop("Invalid marginal variance in GeoCorrFct_Cop().", call. = FALSE)

  ## Enforce the exact same-variable endpoint.  The fast numerical engines are
  ## calibrated to this limit, but setting it explicitly avoids tiny quadrature
  ## or truncation errors in plotted correlation/variogram curves.
  cova[zero_lag] <- marginal_variance
  corr_obs <- cova / marginal_variance
  corr_obs[zero_lag] <- 1

  if (isTRUE(variogram)) {
    if (isTRUE(covariance)) {
      res <- marginal_variance - cova
    } else {
      res <- 1 - corr_obs
    }
    res[zero_lag] <- 0
  } else if (isTRUE(covariance)) {
    res <- cova
  } else {
    res <- corr_obs
  }

  out <- list(
    corr = res,
    distances = x,
    times = t,
    model = model,
    copula = copula,
    distance = distance,
    param = param_work,
    radius = radius,
    n = n,
    covariance = covariance,
    variogram = variogram,
    spacetime = spacetime,
    bivariate = FALSE
  )

  structure(c(out, call = call), class = c("GeoCorrFct"))
}

####################################################
### File name: GeoScatterplot.R
####################################################

.GeoScatterplot_native_models <- c(
    "Gaussian", "SkewGaussian", "Tukeyh", "StudentT",
    "SinhAsinh", "LogGaussian", "Gamma", "Weibull"
)

## Continuous marginal models supported by the pairwise copula likelihood
## and by the shared bivariate-density engine.
.GeoScatterplot_copula_models <- c(
    "Gaussian", "StudentT", "LogGaussian", "Gamma", "Weibull",
    "Beta", "Beta2", "Kumaraswamy", "Kumaraswamy2",
    "Logistic", "SkewLaplace"
)

## Models for which the marginal CDF is continuous and can therefore be used
## on the Uniform/Gaussian diagnostic scales.
.GeoScatterplot_continuous_models <- c(
    "Gaussian", "SkewGaussian", "StudentT", "SkewStudentT",
    "Gaussian_misp_StudentT", "Gaussian_misp_SkewStudentT",
    "LogGaussian", "LogLogistic", "Gamma", "Weibull",
    "Beta", "Beta2", "Kumaraswamy", "Kumaraswamy2",
    "Logistic", "SkewLaplace", "SinhAsinh", "Tukeyh", "Tukeyh2",
    "TwoPieceGaussian", "TwoPieceStudentT",
    "TwoPieceTukeyh", "TwoPieceBimodal"
)

.GeoScatterplot_normalize_model <- function(model) {
    .GeoMarginalCanonicalModel(model)
}

.GeoScatterplot_three_columns <- function(z, validate = FALSE) {
    z <- as.matrix(z)
    storage.mode(z) <- "double"
    if (ncol(z) == 2L) z <- cbind(z, 0)
    if (isTRUE(validate) && ncol(z) != 3L) {
        stop("coordinates must have two or three columns")
    }
    z
}

.GeoScatterplot_representative_lag <- function(x, method, empty = 0) {
    x <- x[is.finite(x)]
    if (!length(x)) return(empty)
    if (identical(method, "median")) stats::median(x) else mean(x)
}

.GeoScatterplot_match_scale <- function(scale) {
    if (length(scale) > 1L) scale <- scale[1L]
    if (length(scale) != 1L || is.na(scale)) {
        stop("scale must be one of 'Original', 'Uniform', or 'Gaussian'")
    }

    key <- tolower(gsub("[[:blank:]_-]", "", as.character(scale)))
    aliases <- c(
        original = "Original",
        response = "Original",
        raw = "Original",
        uniform = "Uniform",
        pit = "Uniform",
        gaussian = "Gaussian",
        normal = "Gaussian"
    )

    if (!(key %in% names(aliases))) {
        stop("scale must be one of 'Original', 'Uniform', or 'Gaussian'")
    }
    unname(aliases[[key]])
}

.GeoScatterplot_parameters <- function(object) {
    fixed <- object$fixed

    ## An externally supplied mean can be a vector. It is not a model
    ## parameter and must not be flattened into the parameter vector.
    if (!is.null(.GeoMean_external(fixed))) fixed$mean <- NULL

    out <- c(
        unlist(object$param, use.names = TRUE),
        unlist(fixed, use.names = TRUE)
    )

    if (length(out)) {
        out <- out[!duplicated(names(out))]
    }
    out
}

.GeoScatterplot_fitted_location <- function(object, n) {
    external_mean <- .GeoMean_external(object$fixed, n = n)
    if (!is.null(external_mean)) return(external_mean)

    par <- .GeoScatterplot_parameters(object)
    if (is.null(object$X)) {
        X <- matrix(1, nrow = n, ncol = 1L)
    } else {
        X <- .GeoMean_design(object$X, n = n, name = "X")
    }
    beta <- .GeoMean_beta(par, ncol(X))
    as.numeric(X %*% beta)
}

.GeoScatterplot_coordinates <- function(object) {
    if (is.matrix(object$coordx) || is.data.frame(object$coordx)) {
        coords <- as.matrix(object$coordx)
    } else {
        if (is.null(object$coordy)) {
            stop("The GeoFit object does not contain valid spatial coordinates")
        }
        if (is.null(object$coordz)) {
            coords <- cbind(object$coordx, object$coordy)
        } else {
            coords <- cbind(object$coordx, object$coordy, object$coordz)
        }
    }

    storage.mode(coords) <- "double"
    if (ncol(coords) == 2L) coords <- cbind(coords, 0)
    if (ncol(coords) != 3L) {
        stop("The GeoFit object must contain two- or three-dimensional coordinates")
    }
    coords
}

.GeoScatterplot_spacetime_layout <- function(object) {
    y <- .GeoScatterplot_data_vector(object)

    if (!isTRUE(object$spacetime)) {
        coords <- .GeoScatterplot_coordinates(object)
        if (nrow(coords) != length(y)) {
            stop("The number of coordinates differs from the number of observations")
        }
        return(list(
            data = y, coords_obs = coords, coords_base = coords,
            coords_dyn = NULL, time_obs = rep(0, length(y)),
            time_values = 0, ns = nrow(coords), dynamic = FALSE
        ))
    }

    coordt <- as.numeric(object$coordt)
    if (length(coordt) < 2L || any(!is.finite(coordt))) {
        stop("The spatio-temporal GeoFit object does not contain valid temporal coordinates")
    }

    if (is.null(object$coordx_dyn)) {
        coords_base <- .GeoScatterplot_coordinates(object)
        nsite <- nrow(coords_base)
        nt <- length(coordt)
        expected <- nsite * nt
        if (length(y) != expected) {
            stop(
                "For fixed spatial locations, the number of observations must equal ",
                "the number of sites times the number of temporal instants"
            )
        }

        coords_obs <- coords_base[
            rep(seq_len(nsite), times = nt), , drop = FALSE
        ]
        time_obs <- rep(coordt, each = nsite)

        return(list(
            data = y, coords_obs = coords_obs, coords_base = coords_base,
            coords_dyn = NULL, time_obs = time_obs,
            time_values = coordt, ns = rep.int(nsite, nt), dynamic = FALSE
        ))
    }

    if (!is.list(object$coordx_dyn) ||
        length(object$coordx_dyn) != length(coordt)) {
        stop("coordx_dyn must be a list with one coordinate matrix per temporal instant")
    }

    coords_dyn <- lapply(object$coordx_dyn, function(z) {
        z <- as.matrix(z)
        storage.mode(z) <- "double"
        if (ncol(z) == 2L) z <- cbind(z, 0)
        if (ncol(z) != 3L) {
            stop("Each element of coordx_dyn must have two or three columns")
        }
        z
    })
    ns <- vapply(coords_dyn, nrow, integer(1))
    coords_obs <- do.call(rbind, coords_dyn)
    time_obs <- rep(coordt, times = ns)

    if (length(y) != nrow(coords_obs)) {
        stop("The dynamic coordinates and the observations have different lengths")
    }

    list(
        data = y, coords_obs = coords_obs, coords_base = NULL,
        coords_dyn = coords_dyn, time_obs = time_obs,
        time_values = coordt, ns = ns, dynamic = TRUE
    )
}

.GeoScatterplot_unique_tol <- function(x, tol = NULL) {
    x <- sort(as.numeric(x[is.finite(x)]))
    if (!length(x)) return(numeric(0))
    if (is.null(tol)) {
        tol <- sqrt(.Machine$double.eps) * max(1, abs(x))
    }
    out <- x[1L]
    if (length(x) == 1L) return(out)
    for (v in x[-1L]) {
        local_tol <- max(tol, sqrt(.Machine$double.eps) * max(1, abs(v), abs(out[length(out)])))
        if (abs(v - out[length(out)]) > local_tol) out <- c(out, v)
    }
    out
}

.GeoScatterplot_distinct_time_lags <- function(coordt, maxtime = NULL) {
    tt <- sort(as.numeric(coordt))
    if (!length(tt) || any(!is.finite(tt))) {
        stop("Temporal coordinates must be finite")
    }
    if (length(tt) == 1L) return(0)

    tol <- sqrt(.Machine$double.eps) * max(1, abs(tt))
    dd <- diff(tt)
    regular <- length(dd) <= 1L ||
        max(abs(dd - stats::median(dd))) <= max(tol, sqrt(.Machine$double.eps) * max(1, abs(dd)))

    if (regular) {
        ## Every lag of a regular grid is represented by a difference from
        ## the first time point; this avoids constructing a T by T matrix.
        possible <- c(0, tt[-1L] - tt[1L])
    } else {
        ## Irregular grids can genuinely have O(T^2) distinct lags.  Build
        ## them incrementally so memory is proportional to the distinct lag
        ## set rather than to a dense outer-difference matrix.
        possible <- 0
        nt <- length(tt)
        for (i in seq_len(nt - 1L)) {
            d <- tt[(i + 1L):nt] - tt[i]
            if (!is.null(maxtime) && length(maxtime) == 1L && is.finite(maxtime)) {
                d <- d[d <= maxtime + tol]
            }
            if (length(d)) {
                possible <- .GeoScatterplot_unique_tol(c(possible, d), tol = tol)
            }
        }
    }

    possible <- .GeoScatterplot_unique_tol(possible, tol = tol)
    if (!is.null(maxtime) && length(maxtime) == 1L && is.finite(maxtime) && maxtime >= 0) {
        possible <- possible[possible <= maxtime + tol]
    }
    possible
}

.GeoScatterplot_time_lags <- function(object, time.lag = NULL) {
    if (!isTRUE(object$spacetime)) return(0)

    coordt <- as.numeric(object$coordt)
    maxtime <- if (is.null(time.lag)) object$maxtime else NULL
    possible <- .GeoScatterplot_distinct_time_lags(coordt, maxtime = maxtime)
    tol <- sqrt(.Machine$double.eps) * max(1, abs(coordt), abs(possible))

    if (is.null(time.lag)) return(possible)

    if (!is.numeric(time.lag) || !length(time.lag) ||
        any(!is.finite(time.lag)) || any(time.lag < 0)) {
        stop("time.lag must contain non-negative finite temporal lags")
    }

    matched <- vapply(time.lag, function(u) {
        k <- which.min(abs(possible - u))
        local_tol <- max(tol, sqrt(.Machine$double.eps) * max(1, abs(u)))
        if (!length(k) || abs(possible[k] - u) > local_tol) {
            stop(
                "Temporal lag ", signif(u, 6),
                " is not available from the temporal coordinates in the GeoFit object"
            )
        }
        possible[k]
    }, numeric(1))

    .GeoScatterplot_unique_tol(matched, tol = tol)
}

.GeoScatterplot_neighbor_coordinates <- function(coords, distance = "Eucl") {
    distance_key <- tolower(gsub("[[:blank:]]", "", as.character(distance)[1L]))

    trim_one <- function(z) {
        z <- as.matrix(z)
        storage.mode(z) <- "double"
        if (distance_key != "eucl" && ncol(z) == 3L) {
            z <- z[, 1:2, drop = FALSE]
        }
        z
    }

    if (is.list(coords)) lapply(coords, trim_one) else trim_one(coords)
}

.GeoScatterplot_spatial_pairs_maxdist <- function(data, coords, maxdist,
                                                  bins, distance = "Eucl",
                                                  radius = 1) {
    coords <- as.matrix(coords)
    storage.mode(coords) <- "double"
    data <- as.numeric(data)
    n <- nrow(coords)

    if (length(data) != n) {
        stop("The data and coordinates must have the same length")
    }
    if (length(maxdist) != 1L || !is.finite(maxdist) || maxdist <= 0) {
        stop("maxdist must be a positive finite scalar")
    }
    if (!is.numeric(bins) || length(bins) < 2L ||
        any(!is.finite(bins)) || any(diff(bins) <= 0)) {
        stop("bins must be a strictly increasing finite numeric vector")
    }
    if (n < 2L) {
        return(list(
            bin = numeric(0), v1 = numeric(0), v2 = numeric(0),
            i = integer(0), j = integer(0), lags = numeric(0)
        ))
    }

    distance_key <- tolower(gsub("[[:blank:]]", "", as.character(distance)[1L]))
    if (!(distance_key %in% c("eucl", "geod", "chor"))) {
        stop("Unsupported distance in GeoScatterplot")
    }
    if (!is.numeric(radius) || length(radius) != 1L ||
        !is.finite(radius) || radius <= 0) {
        stop("radius must be a positive finite scalar")
    }

    max_cells <- getOption("GeoModels.scatterplot.maxdist_cells", 2000000L)
    if (!is.numeric(max_cells) || length(max_cells) != 1L ||
        !is.finite(max_cells) || max_cells < n) {
        max_cells <- max(n, 2000000L)
    }
    block_size <- max(1L, min(n, as.integer(floor(max_cells / n))))
    step <- bins[2L] - bins[1L]
    nbins <- length(bins) - 1L

    if (distance_key == "eucl") {
        coords_norm2 <- rowSums(coords * coords)
    } else {
        lon <- coords[, 1L] * pi / 180
        lat <- coords[, 2L] * pi / 180
        sin_lat <- sin(lat)
        cos_lat <- cos(lat)
    }

    out_bin <- out_v1 <- out_v2 <- out_lag <- list()
    out_i <- out_j <- list()
    out_k <- 0L

    starts <- seq.int(1L, n, by = block_size)
    for (start in starts) {
        rows <- start:min(n, start + block_size - 1L)

        if (distance_key == "eucl") {
            q <- coords[rows, , drop = FALSE]
            d2 <- outer(rowSums(q * q), coords_norm2, "+") -
                2 * tcrossprod(q, coords)
            d2[d2 < 0] <- 0
            dmat <- sqrt(d2)
        } else {
            cs <- outer(sin_lat[rows], sin_lat, "*") +
                outer(cos_lat[rows], cos_lat, "*") *
                cos(outer(lon[rows], lon, "-"))
            cs[cs > 1] <- 1
            cs[cs < -1] <- -1
            theta <- acos(cs)
            dmat <- if (distance_key == "geod") {
                radius * theta
            } else {
                2 * radius * sin(theta / 2)
            }
        }

        hit <- which(is.finite(dmat) & dmat <= maxdist, arr.ind = TRUE)
        if (!nrow(hit)) next
        i <- rows[hit[, 1L]]
        j <- hit[, 2L]
        keep <- i < j & is.finite(data[i]) & is.finite(data[j])
        if (!any(keep)) next

        i <- i[keep]
        j <- j[keep]
        lagv <- dmat[hit[keep, , drop = FALSE]]

        bin_idx <- floor((lagv - bins[1L]) / step) + 1L
        at_upper <- lagv == bins[length(bins)]
        bin_idx[at_upper] <- nbins
        safe_idx <- pmax(1L, pmin(nbins, bin_idx))
        valid_bin <- bin_idx >= 1L & bin_idx <= nbins &
            lagv >= bins[safe_idx] & lagv <= bins[safe_idx + 1L]
        if (!any(valid_bin)) next

        i <- i[valid_bin]
        j <- j[valid_bin]
        lagv <- lagv[valid_bin]
        bin_idx <- bin_idx[valid_bin]

        out_k <- out_k + 1L
        out_bin[[out_k]] <- bins[bin_idx]
        out_v1[[out_k]] <- data[i]
        out_v2[[out_k]] <- data[j]
        out_i[[out_k]] <- i
        out_j[[out_k]] <- j
        out_lag[[out_k]] <- lagv
    }

    if (!out_k) {
        return(list(
            bin = numeric(0), v1 = numeric(0), v2 = numeric(0),
            i = integer(0), j = integer(0), lags = numeric(0)
        ))
    }

    list(
        bin = unlist(out_bin, use.names = FALSE),
        v1 = unlist(out_v1, use.names = FALSE),
        v2 = unlist(out_v2, use.names = FALSE),
        i = as.integer(unlist(out_i, use.names = FALSE)),
        j = as.integer(unlist(out_j, use.names = FALSE)),
        lags = unlist(out_lag, use.names = FALSE)
    )
}

.GeoScatterplot_distance_label <- function(bins, k) {
    left <- "["
    right <- if (k == length(bins) - 1L) "]" else ")"
    paste0(left, signif(bins[k], 3), ", ", signif(bins[k + 1L], 3), right)
}

.GeoScatterplot_bin_membership <- function(lags, bins, k) {
    if (k < length(bins) - 1L) {
        lags >= bins[k] & lags < bins[k + 1L]
    } else {
        lags >= bins[k] & lags <= bins[k + 1L]
    }
}

.GeoScatterplot_nuisance <- function(object, model, par) {
    nb <- object$numbetas
    if (is.null(nb)) nb <- 1L

    ## The marginal nuisance vector is kept separate from the copula
    ## parameter. This avoids relying on model-specific copula indexing.
    nuisance_names <- NuisParam2(
        model = model,
        bivariate = FALSE,
        num_betas = nb,
        copula = NULL
    )
    nuisance_names <- nuisance_names[substr(nuisance_names, 1L, 4L) != "mean"]

    out <- numeric(length(nuisance_names))
    names(out) <- nuisance_names

    for (k in seq_along(nuisance_names)) {
        nm <- nuisance_names[k]
        if (nm %in% names(par)) {
            out[k] <- as.numeric(par[[nm]])
        } else if (nm == "sill" && model %in% c("Gamma", "Weibull", "Beta", "Beta2", "Kumaraswamy", "Kumaraswamy2")) {
            ## GeoFit removes structurally fixed sill values from its output.
            out[k] <- 1
        } else {
            stop("Parameter '", nm, "' is missing from the GeoFit object")
        }
    }

    out
}

.GeoScatterplot_copula_parameter <- function(object, model, par) {
    if (is.null(object$copula) || object$copula == "Gaussian") return(0)

    nb <- object$numbetas
    if (is.null(nb)) nb <- 1L

    base_names <- NuisParam2(
        model = model, bivariate = FALSE,
        num_betas = nb, copula = NULL
    )
    full_names <- NuisParam2(
        model = model, bivariate = FALSE,
        num_betas = nb, copula = object$copula
    )

    extra <- setdiff(full_names, base_names)
    extra <- extra[extra %in% names(par)]

    if (!length(extra)) {
        candidates <- if (object$copula == "Clayton") {
            c("nu", "copula_nu")
        } else {
            c("skew_cop", "copula_skew", "skewcop", "skew")
        }
        extra <- candidates[candidates %in% names(par)]
    }

    if (length(extra) != 1L) {
        stop(
            "Unable to identify the parameter of the ", object$copula,
            " copula in the GeoFit object"
        )
    }

    as.numeric(par[[extra]])
}

.GeoScatterplot_density_setup <- function(object, model) {
    par <- .GeoScatterplot_parameters(object)
    nuisance <- .GeoScatterplot_nuisance(object, model, par)

    ## A Marginal + Independence fit estimates only the univariate marginal
    ## model. Correlation and copula parameters, even when supplied in fixed,
    ## do not belong to the fitted independence likelihood and must therefore
    ## not enter the theoretical bivariate contour.
    independence <- identical(object$likelihood, "Marginal") &&
        identical(object$type, "Independence")

    if (independence) {
        return(list(
            model = as.integer(CkModel(model)),
            model_name = model,
            independence = TRUE,
            corrmodel_name = NULL,
            cormod = NA_integer_,
            type_cop = 0L,
            copula_name = "Independence",
            copula_par = 0,
            corr_names = character(0),
            corrpar = numeric(0),
            corrpar_named = numeric(0),
            nuisance = as.double(nuisance),
            nuisance_named = nuisance
        ))
    }

    corr_names <- CorrParam(object$corrmodel)

    if (!all(corr_names %in% names(par))) {
        missing <- corr_names[!(corr_names %in% names(par))]
        stop("Missing correlation parameter(s): ", paste(missing, collapse = ", "))
    }

    type_cop <- if (is.null(object$copula)) {
        0L
    } else {
        switch(
            object$copula,
            Gaussian = 1L,
            Clayton = 2L,
            SkewGaussian = 3L,
            stop("Unsupported copula in the GeoFit object")
        )
    }

    list(
        model = as.integer(CkModel(model)),
        model_name = model,
        independence = FALSE,
        corrmodel_name = object$corrmodel,
        cormod = as.integer(CkCorrModel(object$corrmodel)),
        type_cop = as.integer(type_cop),
        copula_name = object$copula,
        copula_par = .GeoScatterplot_copula_parameter(object, model, par),
        corr_names = corr_names,
        corrpar = as.double(par[corr_names]),
        corrpar_named = par[corr_names],
        nuisance = as.double(nuisance),
        nuisance_named = nuisance
    )
}

.GeoScatterplot_residual_n <- function(n, nobs, model) {
    if (is.null(n))
        stop("Residual-scale scatterplots for model '", model, "' require n.")
    z <- if (is.list(n)) unlist(n, use.names = FALSE) else as.numeric(n)
    if (length(z) == 1L) z <- rep(z, nobs)
    if (length(z) != nobs || any(!is.finite(z)) || any(z <= 0) ||
        any(abs(z - round(z)) > sqrt(.Machine$double.eps))) {
        stop("Residual-scale scatterplots for model '", model,
             "' require positive integer n with length 1 or the number of observations.")
    }
    as.numeric(z)
}

.GeoScatterplot_residual_scale <- function(data, mean, model, object = NULL) {
    if (model %in% c("LogGaussian", "Gamma", "Weibull")) {
        sc <- exp(mean)
        if (any(!is.finite(sc)) || any(sc <= 0)) {
            stop("Invalid fitted multiplicative mean")
        }
        return(data / sc)
    }

    if (model %in% .GeoScatterplot_native_models) {
        return(data - mean)
    }

    ## Count-model residual plots use the same deterministic Pearson scale as
    ## GeoResiduals().  This provides a meaningful common scale for pair plots
    ## without pretending that a continuous bivariate density is available.
    pearson <- function(obs, fitted_mean, fitted_var) {
        if (any(!is.finite(fitted_mean)) || any(!is.finite(fitted_var)) ||
            any(fitted_var < 0))
            stop("Invalid fitted count moments for residual-scale scatterplot")
        (obs - fitted_mean) / sqrt(pmax(fitted_var, .Machine$double.eps))
    }

    if (model == "Poisson") {
        fitted_mean <- exp(mean)
        return(pearson(data, fitted_mean, fitted_mean))
    }

    if (model %in% c("Binary", "Bernoulli")) {
        prob <- pnorm(mean)
        return(pearson(data, prob, prob * (1 - prob)))
    }

    if (model == "Binomial") {
        if (is.null(object)) stop("A GeoFit object is required for Binomial residuals")
        nn <- .GeoScatterplot_residual_n(object$n, length(data), model)
        prob <- pnorm(mean)
        fitted_mean <- nn * prob
        return(pearson(data, fitted_mean, nn * prob * (1 - prob)))
    }

    if (model %in% c("Geom", "Geometric", "BinomialNeg")) {
        if (is.null(object)) stop("A GeoFit object is required for Negative-Binomial residuals")
        nn <- if (model %in% c("Geom", "Geometric")) {
            rep(1, length(data))
        } else {
            .GeoScatterplot_residual_n(object$n, length(data), model)
        }
        prob <- pnorm(mean)
        fitted_mean <- nn * (1 - prob) / prob
        fitted_var <- nn * (1 - prob) / prob^2
        return(pearson(data, fitted_mean, fitted_var))
    }

    if (model == "PoissonGamma") {
        if (is.null(object)) stop("A GeoFit object is required for PoissonGamma residuals")
        par <- .GeoScatterplot_parameters(object)
        shape <- suppressWarnings(as.numeric(par["shape"]))
        if (length(shape) != 1L || !is.finite(shape) || shape <= 0)
            stop("PoissonGamma residual-scale scatterplots require a positive fitted shape parameter")
        fitted_mean <- exp(mean)
        fitted_var <- fitted_mean * (1 + fitted_mean / shape)
        return(pearson(data, fitted_mean, fitted_var))
    }

    stop("Residual-scale scatterplots are not implemented for model '", model, "'")
}

.GeoScatterplot_data_vector <- function(object) {
    if (is.list(object$data)) {
        out <- unlist(object$data, use.names = FALSE)
    } else {
        out <- c(t(object$data))
    }
    as.numeric(out)
}

.GeoScatterplot_pit_data <- function(object, scale) {
    if (!(scale %in% c("Uniform", "Gaussian"))) {
        stop("The PIT scale must be either 'Uniform' or 'Gaussian'")
    }

    pit_object <- GeoPit(object, type = scale)
    .GeoScatterplot_data_vector(pit_object)
}

.GeoScatterplot_nuisance_value <- function(nuisance, name,
                                                       default = NA_real_) {
    if (name %in% names(nuisance)) {
        value <- suppressWarnings(as.numeric(nuisance[[name]]))
        if (length(value) == 1L && is.finite(value)) return(value)
    }
    default
}

.GeoScatterplot_marginal_quantile <- function(u, mean, model, nuisance) {
    eps <- sqrt(.Machine$double.eps)
    u <- pmin(1 - eps, pmax(eps, u))
    sill <- .GeoScatterplot_nuisance_value(nuisance, "sill", 1)
    if (!is.finite(sill) || sill <= 0) {
        stop("The fitted marginal variance must be positive")
    }
    sd <- sqrt(sill)
    out <- NULL

    if (model == "Gaussian") {
        out <- mean + sd * stats::qnorm(u)
    }

    if (model == "SkewGaussian") {
        skew <- .GeoScatterplot_nuisance_value(nuisance, "skew")
        if (!is.finite(skew)) stop("The SkewGaussian skew parameter is missing")
        omega <- sqrt((skew^2 + sill) / sill)
        alpha <- skew / sd
        out <- mean + sd * sn::qsn(u, xi = 0, omega = omega, alpha = alpha)
    }

    if (model == "Tukeyh") {
        tail <- .GeoScatterplot_nuisance_value(nuisance, "tail")
        if (!is.finite(tail) || tail < 0) {
            stop("The Tukey-h tail parameter must be non-negative")
        }
        out <- mean + sd * .GeoMarginalQTukeyH(u, tail)
    }

    if (model == "StudentT") {
        inv_df <- .GeoScatterplot_nuisance_value(nuisance, "df")
        if (!is.finite(inv_df) || inv_df <= 0) {
            stop("The StudentT inverse degrees-of-freedom parameter must be positive")
        }
        out <- mean + sd * stats::qt(u, df = 1 / inv_df)
    }

    if (model == "SinhAsinh") {
        skew <- .GeoScatterplot_nuisance_value(nuisance, "skew")
        tail <- .GeoScatterplot_nuisance_value(nuisance, "tail")
        if (!is.finite(skew) || !is.finite(tail) || tail <= 0) {
            stop("The SinhAsinh skew and tail parameters are invalid")
        }
        z <- stats::qnorm(u)
        out <- mean + sd * sinh((asinh(z) + skew) / tail)
    }

    if (model == "LogGaussian") {
        out <- stats::qlnorm(u, meanlog = mean - sill / 2, sdlog = sd)
    }

    if (model == "Gamma") {
        shape <- .GeoScatterplot_nuisance_value(nuisance, "shape")
        if (!is.finite(shape) || shape <= 0) {
            stop("The Gamma shape parameter must be positive")
        }
        out <- stats::qgamma(
            u, shape = shape / 2,
            rate = shape / (2 * exp(mean))
        )
    }

    if (model == "Weibull") {
        shape <- .GeoScatterplot_nuisance_value(nuisance, "shape")
        if (!is.finite(shape) || shape <= 0) {
            stop("The Weibull shape parameter must be positive")
        }
        out <- stats::qweibull(
            u, shape = shape,
            scale = exp(mean) / gamma(1 + 1 / shape)
        )
    }

    if (model == "Beta") {
        shape1 <- .GeoScatterplot_nuisance_value(nuisance, "shape1")
        shape2 <- .GeoScatterplot_nuisance_value(nuisance, "shape2")
        lower <- .GeoScatterplot_nuisance_value(nuisance, "min")
        upper <- .GeoScatterplot_nuisance_value(nuisance, "max")
        if (!is.finite(shape1) || !is.finite(shape2) ||
            shape1 <= 0 || shape2 <= 0 || !is.finite(lower) ||
            !is.finite(upper) || upper <= lower) {
            stop("The fitted Beta parameters are invalid")
        }
        out <- lower + (upper - lower) * stats::qbeta(u, shape1, shape2)
    }

    if (model == "Beta2") {
        shape <- .GeoScatterplot_nuisance_value(nuisance, "shape")
        lower <- .GeoScatterplot_nuisance_value(nuisance, "min")
        upper <- .GeoScatterplot_nuisance_value(nuisance, "max")
        mu <- stats::plogis(mean)
        if (!is.finite(shape) || shape <= 0 || !is.finite(lower) ||
            !is.finite(upper) || upper <= lower) {
            stop("The fitted Beta2 parameters are invalid")
        }
        out <- lower + (upper - lower) * stats::qbeta(
            u, shape1 = mu * shape, shape2 = (1 - mu) * shape
        )
    }

    if (model == "Kumaraswamy") {
        shape1 <- .GeoScatterplot_nuisance_value(nuisance, "shape1")
        shape2 <- .GeoScatterplot_nuisance_value(nuisance, "shape2")
        lower <- .GeoScatterplot_nuisance_value(nuisance, "min")
        upper <- .GeoScatterplot_nuisance_value(nuisance, "max")
        if (!is.finite(shape1) || !is.finite(shape2) ||
            shape1 <= 0 || shape2 <= 0 || !is.finite(lower) ||
            !is.finite(upper) || upper <= lower) {
            stop("The fitted Kumaraswamy parameters are invalid")
        }
        qq <- (1 - (1 - u)^(1 / shape2))^(1 / shape1)
        out <- lower + (upper - lower) * qq
    }

    if (model == "Kumaraswamy2") {
        shape <- .GeoScatterplot_nuisance_value(nuisance, "shape")
        lower <- .GeoScatterplot_nuisance_value(nuisance, "min")
        upper <- .GeoScatterplot_nuisance_value(nuisance, "max")
        mu <- stats::plogis(mean)
        if (!is.finite(shape) || shape <= 0 || !is.finite(lower) ||
            !is.finite(upper) || upper <= lower || mu <= 0 || mu >= 1) {
            stop("The fitted Kumaraswamy2 parameters are invalid")
        }
        shape2 <- log(0.5) / log1p(-mu^shape)
        qq <- (1 - (1 - u)^(1 / shape2))^(1 / shape)
        out <- lower + (upper - lower) * qq
    }

    if (model == "Logistic") {
        out <- mean + sd * stats::qlogis(u)
    }

    if (model == "SkewLaplace") {
        skew <- .GeoScatterplot_nuisance_value(nuisance, "skew")
        if (!is.finite(skew) || skew <= 0 || skew >= 1) {
            stop("The SkewLaplace skew parameter must lie in (0,1)")
        }
        out <- mean + sd * .GeoMarginalQSkewLaplace(u, skew)
    }

    if (is.null(out)) {
        stop("Marginal quantiles are not implemented for model '", model, "'")
    }
    out
}

.GeoScatterplot_marginal_density <- function(x, mean, model, nuisance) {
    sill <- .GeoScatterplot_nuisance_value(nuisance, "sill", 1)
    if (!is.finite(sill) || sill <= 0) {
        stop("The fitted marginal variance must be positive")
    }
    sd <- sqrt(sill)
    out <- NULL

    if (model == "Gaussian") {
        out <- stats::dnorm(x, mean = mean, sd = sd)
    }

    if (model == "SkewGaussian") {
        skew <- .GeoScatterplot_nuisance_value(nuisance, "skew")
        if (!is.finite(skew)) stop("The SkewGaussian skew parameter is missing")
        omega <- sqrt((skew^2 + sill) / sill)
        alpha <- skew / sd
        out <- sn::dsn((x - mean) / sd, xi = 0, omega = omega, alpha = alpha) / sd
    }

    if (model == "Tukeyh") {
        tail <- .GeoScatterplot_nuisance_value(nuisance, "tail")
        if (!is.finite(tail) || tail < 0) {
            stop("The Tukey-h tail parameter must be non-negative and finite")
        }
        out <- .GeoMarginalDTukeyH((x - mean) / sd, tail) / sd
    }

    if (model == "StudentT") {
        inv_df <- .GeoScatterplot_nuisance_value(nuisance, "df")
        if (!is.finite(inv_df) || inv_df <= 0) {
            stop("The StudentT inverse degrees-of-freedom parameter must be positive")
        }
        out <- stats::dt((x - mean) / sd, df = 1 / inv_df) / sd
    }

    if (model == "SinhAsinh") {
        skew <- .GeoScatterplot_nuisance_value(nuisance, "skew")
        tail <- .GeoScatterplot_nuisance_value(nuisance, "tail")
        if (!is.finite(skew) || !is.finite(tail) || tail <= 0) {
            stop("The SinhAsinh skew and tail parameters are invalid")
        }
        xx <- (x - mean) / sd
        bb <- tail * asinh(xx) - skew
        z <- sinh(bb)
        jac <- tail * cosh(bb) / sqrt(1 + xx^2)
        out <- stats::dnorm(z) * jac / sd
    }

    if (model == "LogGaussian") {
        out <- stats::dlnorm(x, meanlog = mean - sill / 2, sdlog = sd)
    }

    if (model == "Gamma") {
        shape <- .GeoScatterplot_nuisance_value(nuisance, "shape")
        if (!is.finite(shape) || shape <= 0) {
            stop("The Gamma shape parameter must be positive")
        }
        out <- stats::dgamma(
            x, shape = shape / 2,
            rate = shape / (2 * exp(mean))
        )
    }

    if (model == "Weibull") {
        shape <- .GeoScatterplot_nuisance_value(nuisance, "shape")
        if (!is.finite(shape) || shape <= 0) {
            stop("The Weibull shape parameter must be positive")
        }
        out <- stats::dweibull(
            x, shape = shape,
            scale = exp(mean) / gamma(1 + 1 / shape)
        )
    }

    if (model == "Beta") {
        shape1 <- .GeoScatterplot_nuisance_value(nuisance, "shape1")
        shape2 <- .GeoScatterplot_nuisance_value(nuisance, "shape2")
        lower <- .GeoScatterplot_nuisance_value(nuisance, "min")
        upper <- .GeoScatterplot_nuisance_value(nuisance, "max")
        width <- upper - lower
        if (!is.finite(shape1) || !is.finite(shape2) ||
            shape1 <= 0 || shape2 <= 0 || !is.finite(width) || width <= 0) {
            stop("The fitted Beta parameters are invalid")
        }
        out <- stats::dbeta((x - lower) / width, shape1, shape2) / width
    }

    if (model == "Beta2") {
        shape <- .GeoScatterplot_nuisance_value(nuisance, "shape")
        lower <- .GeoScatterplot_nuisance_value(nuisance, "min")
        upper <- .GeoScatterplot_nuisance_value(nuisance, "max")
        width <- upper - lower
        mu <- stats::plogis(mean)
        if (!is.finite(shape) || shape <= 0 || !is.finite(width) || width <= 0) {
            stop("The fitted Beta2 parameters are invalid")
        }
        out <- stats::dbeta(
            (x - lower) / width,
            shape1 = mu * shape,
            shape2 = (1 - mu) * shape
        ) / width
    }

    if (model == "Kumaraswamy") {
        shape1 <- .GeoScatterplot_nuisance_value(nuisance, "shape1")
        shape2 <- .GeoScatterplot_nuisance_value(nuisance, "shape2")
        lower <- .GeoScatterplot_nuisance_value(nuisance, "min")
        upper <- .GeoScatterplot_nuisance_value(nuisance, "max")
        width <- upper - lower
        if (!is.finite(shape1) || !is.finite(shape2) ||
            shape1 <= 0 || shape2 <= 0 || !is.finite(width) || width <= 0) {
            stop("The fitted Kumaraswamy parameters are invalid")
        }
        z <- (x - lower) / width
        out <- ifelse(
            z > 0 & z < 1,
            shape1 * shape2 * z^(shape1 - 1) *
                (1 - z^shape1)^(shape2 - 1) / width,
            0
        )
    }

    if (model == "Kumaraswamy2") {
        shape <- .GeoScatterplot_nuisance_value(nuisance, "shape")
        lower <- .GeoScatterplot_nuisance_value(nuisance, "min")
        upper <- .GeoScatterplot_nuisance_value(nuisance, "max")
        width <- upper - lower
        mu <- stats::plogis(mean)
        if (!is.finite(shape) || shape <= 0 || !is.finite(width) ||
            width <= 0 || mu <= 0 || mu >= 1) {
            stop("The fitted Kumaraswamy2 parameters are invalid")
        }
        shape2 <- log(0.5) / log1p(-mu^shape)
        z <- (x - lower) / width
        out <- ifelse(
            z > 0 & z < 1,
            shape * shape2 * z^(shape - 1) *
                (1 - z^shape)^(shape2 - 1) / width,
            0
        )
    }

    if (model == "Logistic") {
        out <- stats::dlogis(x, location = mean, scale = sd)
    }

    if (model == "SkewLaplace") {
        skew <- .GeoScatterplot_nuisance_value(nuisance, "skew")
        if (!is.finite(skew) || skew <= 0 || skew >= 1) {
            stop("The SkewLaplace skew parameter must lie in (0,1)")
        }
        out <- .GeoMarginalDSkewLaplace((x - mean) / sd, skew) / sd
    }

    if (is.null(out)) {
        stop("Marginal densities are not implemented for model '", model, "'")
    }
    out
}

.GeoScatterplot_contour_levels <- function(z, nlevels,
                                                  scale = "Original") {
    zz <- z[is.finite(z) & z > 0]
    if (length(zz) < 2L || diff(range(zz)) <= 0) return(numeric(0))

    if (scale == "Gaussian") {
        ## Levels analogous to those used in the copula comparison figures:
        ## 0.02, 0.06, 0.10, ... . If many are available, retain an evenly
        ## spaced subset controlled by nlevels.
        upper <- max(zz)
        lev <- seq(0.02, upper, by = 0.04)
        lev <- lev[lev > min(zz) & lev < upper]
        if (length(lev) > nlevels) {
            idx <- unique(round(seq(1, length(lev), length.out = nlevels)))
            lev <- lev[idx]
        }
        if (length(lev)) return(lev)
    }

    ## Copula densities can increase sharply near the corners of the unit
    ## square. Quantile-based levels prevent a few extreme grid values from
    ## determining all automatically selected contours.
    if (scale == "Uniform") {
        probs <- seq(0.20, 0.90, length.out = nlevels)
        lev <- unique(as.numeric(stats::quantile(
            zz, probs = probs, names = FALSE, type = 8
        )))
        lev <- lev[lev > min(zz) & lev < max(zz)]
        if (length(lev)) return(lev)
    }

    lev <- pretty(range(zz), n = nlevels)
    lev <- lev[lev > min(zz) & lev < max(zz)]

    if (!length(lev)) {
        lev <- seq(min(zz), max(zz), length.out = nlevels + 2L)
        lev <- lev[-c(1L, length(lev))]
    }
    lev
}

.GeoScatterplot_response_density_grid <- function(setup, xgrid, ygrid,
                                                  lag, time_lag = 0,
                                                  mean1, mean2) {
    gx <- rep(xgrid, times = length(ygrid))
    gy <- rep(ygrid, each = length(xgrid))

    ans <- .C(
        "GeoBivDensity",
        model = as.integer(setup$model),
        cormod = as.integer(setup$cormod),
        type_cop = as.integer(setup$type_cop),
        copula_par = as.double(setup$copula_par),
        x = as.double(gx),
        y = as.double(gy),
        n = as.integer(length(gx)),
        lag = as.double(lag),
        lagt = as.double(time_lag),
        par = as.double(setup$corrpar),
        mean1 = as.double(mean1),
        mean2 = as.double(mean2),
        nuis = as.double(setup$nuisance),
        NN1 = as.integer(1L),
        NN2 = as.integer(1L),
        res = double(length(gx)),
        PACKAGE = "GeoModels",
        NAOK = TRUE
    )$res

    z <- matrix(ans, nrow = length(xgrid), ncol = length(ygrid))
    rho_info <- .GeoScatterplot_copula_rho(setup, lag, time_lag)
    attr(z, "rho") <- rho_info$rho
    attr(z, "rho_latent") <- rho_info$rho_latent
    attr(z, "nugget") <- rho_info$nugget
    z
}

.GeoScatterplot_copula_rho <- function(setup, lag, time_lag = 0) {
    if (length(lag) != 1L || !is.finite(lag) || lag < 0) {
        stop("The representative spatial lag must be a non-negative finite number")
    }
    if (length(time_lag) != 1L || !is.finite(time_lag) || time_lag < 0) {
        stop("The representative temporal lag must be a non-negative finite number")
    }

    corr_param <- as.list(setNames(
        as.numeric(setup$corrpar_named), names(setup$corrpar_named)
    ))
    param <- c(list(mean = 0, nugget = 0, sill = 1), corr_param)

    rho_latent <- as.numeric(GeoCorrFct(
        x = lag,
        t = time_lag,
        corrmodel = setup$corrmodel_name,
        model = "Gaussian",
        param = param,
        covariance = FALSE,
        variogram = FALSE
    )$corr)[1L]

    nugget <- .GeoScatterplot_nuisance_value(
        setup$nuisance_named, "nugget", 0
    )
    rho <- (1 - nugget) * rho_latent

    if (!is.finite(rho_latent) || !is.finite(rho)) {
        stop("Unable to evaluate the fitted correlation at the representative lag")
    }

    eps <- 1e-10
    rho <- max(-1 + eps, min(1 - eps, rho))
    rho_latent <- max(-1 + eps, min(1 - eps, rho_latent))

    list(rho = rho, rho_latent = rho_latent, nugget = nugget)
}

.GeoScatterplot_direct_copula_grid <- function(setup, xgrid, ygrid,
                                                     lag, time_lag = 0, scale) {
    if (setup$type_cop <= 0L) {
        stop("A fitted copula is required for direct copula-density evaluation")
    }

    if (scale == "Uniform") {
        eps <- sqrt(.Machine$double.eps)
        ugrid <- pmin(1 - eps, pmax(eps, xgrid))
        vgrid <- pmin(1 - eps, pmax(eps, ygrid))
    } else if (scale == "Gaussian") {
        ugrid <- stats::pnorm(xgrid)
        vgrid <- stats::pnorm(ygrid)
    } else {
        stop("Direct copula-density evaluation is available only on the Uniform or Gaussian scale")
    }

    gu <- rep(ugrid, times = length(vgrid))
    gv <- rep(vgrid, each = length(ugrid))
    rho_info <- .GeoScatterplot_copula_rho(setup, lag, time_lag)

    ans <- .C(
        "GeoCopulaDensity",
        type_cop = as.integer(setup$type_cop),
        copula_par = as.double(setup$copula_par),
        u = as.double(gu),
        v = as.double(gv),
        n = as.integer(length(gu)),
        rho = as.double(rho_info$rho),
        res = double(length(gu)),
        PACKAGE = "GeoModels",
        NAOK = TRUE
    )$res

    copula_density <- matrix(
        ans, nrow = length(ugrid), ncol = length(vgrid)
    )
    copula_density[
        !is.finite(copula_density) | copula_density < 0
    ] <- 0

    attr(copula_density, "rho") <- rho_info$rho
    attr(copula_density, "rho_latent") <- rho_info$rho_latent
    attr(copula_density, "nugget") <- rho_info$nugget
    attr(copula_density, "copula_parameter") <- setup$copula_par
    attr(copula_density, "copula") <- setup$copula_name

    if (scale == "Uniform") return(copula_density)

    gaussian_jacobian <- outer(
        stats::dnorm(xgrid), stats::dnorm(ygrid)
    )
    gaussian_density <- copula_density * gaussian_jacobian
    gaussian_density[
        !is.finite(gaussian_density) | gaussian_density < 0
    ] <- 0
    attr(gaussian_density, "rho") <- rho_info$rho
    attr(gaussian_density, "rho_latent") <- rho_info$rho_latent
    attr(gaussian_density, "nugget") <- rho_info$nugget
    attr(gaussian_density, "copula_parameter") <- setup$copula_par
    attr(gaussian_density, "copula") <- setup$copula_name
    gaussian_density
}

.GeoScatterplot_independence_density_grid <- function(setup, xgrid, ygrid,
                                                              mean1, mean2, scale) {
    if (scale == "Original") {
        fx <- .GeoScatterplot_marginal_density(
            x = xgrid, mean = mean1, model = setup$model_name,
            nuisance = setup$nuisance_named
        )
        fy <- .GeoScatterplot_marginal_density(
            x = ygrid, mean = mean2, model = setup$model_name,
            nuisance = setup$nuisance_named
        )
        z <- outer(fx, fy)
    } else if (scale == "Uniform") {
        ## Under independence, the copula density on (0,1)^2 is identically 1.
        z <- matrix(1, nrow = length(xgrid), ncol = length(ygrid))
    } else if (scale == "Gaussian") {
        ## After the marginal PIT and Gaussian-score transformation, an
        ## independence fit has joint density phi(z1) * phi(z2).
        z <- outer(stats::dnorm(xgrid), stats::dnorm(ygrid))
    } else {
        stop("Unsupported scale for an independence contour")
    }

    z[!is.finite(z) | z < 0] <- 0
    attr(z, "rho") <- 0
    attr(z, "rho_latent") <- 0
    attr(z, "nugget") <- NA_real_
    attr(z, "copula") <- "Independence"
    attr(z, "copula_parameter") <- NA_real_
    z
}

.GeoScatterplot_density_grid <- function(setup, xgrid, ygrid, lag,
                                         time_lag = 0, mean1, mean2, scale) {
    if (isTRUE(setup$independence)) {
        return(.GeoScatterplot_independence_density_grid(
            setup = setup, xgrid = xgrid, ygrid = ygrid,
            mean1 = mean1, mean2 = mean2, scale = scale
        ))
    }

    if (scale == "Original") {
        return(.GeoScatterplot_response_density_grid(
            setup = setup, xgrid = xgrid, ygrid = ygrid,
            lag = lag, time_lag = time_lag, mean1 = mean1, mean2 = mean2
        ))
    }

    ## For fitted copula models, evaluate the copula density directly.
    ## This is both simpler and more stable than transforming the complete
    ## response density back through the fitted marginal distribution.
    if (setup$type_cop > 0L) {
        return(.GeoScatterplot_direct_copula_grid(
            setup = setup, xgrid = xgrid, ygrid = ygrid,
            lag = lag, time_lag = time_lag, scale = scale
        ))
    }

    ## Native non-Gaussian random fields do not have a separate copula
    ## evaluator. For these models, recover the copula density from the
    ## fitted response density and fitted marginal densities.
    if (scale == "Uniform") {
        eps <- sqrt(.Machine$double.eps)
        ugrid <- pmin(1 - eps, pmax(eps, xgrid))
        vgrid <- pmin(1 - eps, pmax(eps, ygrid))
    } else {
        ugrid <- stats::pnorm(xgrid)
        vgrid <- stats::pnorm(ygrid)
    }

    x_response <- .GeoScatterplot_marginal_quantile(
        u = ugrid, mean = mean1, model = setup$model_name,
        nuisance = setup$nuisance_named
    )
    y_response <- .GeoScatterplot_marginal_quantile(
        u = vgrid, mean = mean2, model = setup$model_name,
        nuisance = setup$nuisance_named
    )

    response_density <- .GeoScatterplot_response_density_grid(
        setup = setup,
        xgrid = x_response,
        ygrid = y_response,
        lag = lag,
        time_lag = time_lag,
        mean1 = mean1,
        mean2 = mean2
    )

    fx <- .GeoScatterplot_marginal_density(
        x = x_response, mean = mean1, model = setup$model_name,
        nuisance = setup$nuisance_named
    )
    fy <- .GeoScatterplot_marginal_density(
        x = y_response, mean = mean2, model = setup$model_name,
        nuisance = setup$nuisance_named
    )

    denominator <- outer(fx, fy)
    copula_density <- response_density / denominator
    copula_density[!is.finite(copula_density) | copula_density < 0] <- 0
    attr(copula_density, "rho") <- attr(response_density, "rho")
    attr(copula_density, "rho_latent") <- attr(response_density, "rho_latent")
    attr(copula_density, "nugget") <- attr(response_density, "nugget")

    if (scale == "Uniform") return(copula_density)

    gaussian_jacobian <- outer(stats::dnorm(xgrid), stats::dnorm(ygrid))
    gaussian_density <- copula_density * gaussian_jacobian
    gaussian_density[!is.finite(gaussian_density) | gaussian_density < 0] <- 0
    attr(gaussian_density, "rho") <- attr(copula_density, "rho")
    attr(gaussian_density, "rho_latent") <- attr(copula_density, "rho_latent")
    attr(gaussian_density, "nugget") <- attr(copula_density, "nugget")
    gaussian_density
}

.GeoScatterplot_panel <- function(x, y, main, lag, time_lag = 0,
                                  setup, contour, mean1, mean2, scale,
                                  ngrid, nlevels, levels,
                                  contour.col, contour.lwd, contour.labels,
                                  point.col, dots) {
    keep <- is.finite(x) & is.finite(y)
    x <- x[keep]
    y <- y[keep]

    if (!length(x)) {
        plot.new()
        title(main = main)
        mtext("No pairs", side = 3, line = -2)
        return(list(
            x = x, y = y, lag = lag, time_lag = time_lag,
            density = NULL, xgrid = NULL, ygrid = NULL, levels = NULL
        ))
    }

    plot_args <- modifyList(
        list(
            x = x, y = y, col = point.col,
            xlab = "", ylab = "", main = main
        ),
        dots
    )
    do.call(graphics::plot, plot_args)
    graphics::abline(0, 1)

    if (!isTRUE(contour)) {
        return(list(
            x = x, y = y, lag = lag, time_lag = time_lag,
            density = NULL, xgrid = NULL, ygrid = NULL, levels = NULL
        ))
    }

    usr <- graphics::par("usr")
    if (scale == "Uniform") {
        eps <- 1 / (100 * ngrid)
        xlim <- c(max(0, usr[1L]), min(1, usr[2L]))
        ylim <- c(max(0, usr[3L]), min(1, usr[4L]))
        xgrid <- seq(max(eps, xlim[1L]), min(1 - eps, xlim[2L]),
                     length.out = ngrid)
        ygrid <- seq(max(eps, ylim[1L]), min(1 - eps, ylim[2L]),
                     length.out = ngrid)
    } else {
        xgrid <- seq(usr[1L], usr[2L], length.out = ngrid)
        ygrid <- seq(usr[3L], usr[4L], length.out = ngrid)
    }

    z <- .GeoScatterplot_density_grid(
        setup = setup,
        xgrid = xgrid,
        ygrid = ygrid,
        lag = lag,
        time_lag = time_lag,
        mean1 = mean1,
        mean2 = mean2,
        scale = scale
    )

    lev <- levels
    if (is.null(lev)) {
        lev <- .GeoScatterplot_contour_levels(z, nlevels, scale = scale)
    }

    if (length(lev)) {
        graphics::contour(
            xgrid, ygrid, z,
            levels = lev,
            add = TRUE,
            drawlabels = contour.labels,
            col = contour.col,
            lwd = contour.lwd
        )
    }

    rho <- attr(z, "rho")
    rho_latent <- attr(z, "rho_latent")
    nugget <- attr(z, "nugget")
    copula_parameter <- attr(z, "copula_parameter")
    copula_name <- attr(z, "copula")

    reflection <- NULL
    if (scale %in% c("Uniform", "Gaussian") &&
        nrow(z) == ncol(z) && length(xgrid) == length(ygrid)) {
        z_reflected <- z[nrow(z):1L, ncol(z):1L, drop = FALSE]
        delta <- abs(z - z_reflected)
        zmax <- max(z, na.rm = TRUE)
        zsum <- sum(z, na.rm = TRUE)
        reflection <- c(
            max_absolute = max(delta, na.rm = TRUE),
            max_relative = if (is.finite(zmax) && zmax > 0) {
                max(delta, na.rm = TRUE) / zmax
            } else NA_real_,
            L1_relative = if (is.finite(zsum) && zsum > 0) {
                sum(delta, na.rm = TRUE) / zsum
            } else NA_real_
        )
    }

    list(
        x = x, y = y, lag = lag, time_lag = time_lag,
        density = z, xgrid = xgrid, ygrid = ygrid, levels = lev,
        rho = rho, rho_latent = rho_latent, nugget = nugget,
        copula = copula_name, copula_parameter = copula_parameter,
        reflection = reflection
    )
}

.GeoScatterplot_GeoFit <- function(object, maxdist, neighb, time.lag,
                                   numbins, radius, contour, residuals,
                                   scale, gaussian.range, ngrid, nlevels,
                                   levels, contour.col, contour.lwd,
                                   contour.labels, point.col, lag.method,
                                   dots) {
    if (!inherits(object, "GeoFit")) stop("object must inherit from 'GeoFit'")
    if (isTRUE(object$bivariate)) stop("Bivariate GeoFit objects are not supported yet")

    spacetime <- isTRUE(object$spacetime)
    if (!spacetime && !is.null(time.lag)) {
        stop("time.lag can be used only with a spatio-temporal GeoFit object")
    }
    model <- .GeoScatterplot_normalize_model(object$model)
    scale <- .GeoScatterplot_match_scale(scale)
    independence <- identical(object$likelihood, "Marginal") &&
        identical(object$type, "Independence")

    if (scale != "Original" && identical(object$residual_type, "Pearson")) {
        stop(
            "Uniform and Gaussian scatterplot scales are not defined for Pearson-residual objects. Use the original discrete GeoFit object only for discrete PIT diagnostics, or use scale='Original' for Pearson residual scatterplots.",
            call. = FALSE
        )
    }
    if (scale != "Original" && !(model %in% .GeoScatterplot_continuous_models)) {
        stop(
            "Uniform and Gaussian scatterplot scales require a fitted continuous marginal model.",
            call. = FALSE
        )
    }

    graphics::par(pty = "s")

    if (length(gaussian.range) != 1L || !is.finite(gaussian.range) ||
        gaussian.range <= 0) {
        stop("gaussian.range must be a positive finite number")
    }

    if (scale != "Original" && isTRUE(residuals)) {
        stop("residuals can be used only with scale = 'Original'")
    }

    if (isTRUE(contour)) {
        if (isTRUE(object$missp)) {
            stop("Contours are not available for misspecified likelihood models")
        }

        if (isTRUE(independence)) {
            if (!(model %in% unique(c(
                .GeoScatterplot_native_models,
                .GeoScatterplot_copula_models
            )))) {
                stop(
                    "Independence contours are currently available only for ",
                    "continuous marginal models supported by GeoScatterplot"
                )
            }
        } else if (is.null(object$copula)) {
            if (!(model %in% .GeoScatterplot_native_models)) {
                stop(
                    "Native fitted contours are available only for: ",
                    paste(.GeoScatterplot_native_models, collapse = ", ")
                )
            }
        } else {
            if (!(object$copula %in% c("Gaussian", "SkewGaussian", "Clayton"))) {
                stop("Unsupported copula in the GeoFit object")
            }
            if (!(model %in% .GeoScatterplot_copula_models)) {
                stop(
                    "Copula contours are currently available for the continuous ",
                    "marginal models: ",
                    paste(.GeoScatterplot_copula_models, collapse = ", ")
                )
            }
        }
    }

    if (length(numbins) != 1L || is.na(numbins) || numbins < 1L ||
        numbins != as.integer(numbins)) {
        stop("numbins must be a positive integer")
    }
    numbins <- as.integer(numbins)

    if (isTRUE(contour)) {
        if (length(ngrid) != 1L || is.na(ngrid) || ngrid < 20L ||
            ngrid != as.integer(ngrid)) {
            stop("ngrid must be an integer not smaller than 20")
        }
        ngrid <- as.integer(ngrid)

        if (length(nlevels) != 1L || is.na(nlevels) || nlevels < 1L ||
            nlevels != as.integer(nlevels)) {
            stop("nlevels must be a positive integer")
        }
        nlevels <- as.integer(nlevels)
    }

    lag.method <- match.arg(lag.method, choices = c("median", "mean"))

    layout <- .GeoScatterplot_spacetime_layout(object)

    .GeoValidate3D(
        coordx = object$coordx, coordy = object$coordy,
        coordz = object$coordz, coordx_dyn = object$coordx_dyn,
        distance = object$distance, grid = isTRUE(object$grid),
        context = "GeoScatterplot"
    )
    yraw <- layout$data
    mu <- .GeoScatterplot_fitted_location(object, length(yraw))

    if (scale == "Original") {
        yplot <- if (isTRUE(residuals)) {
            .GeoScatterplot_residual_scale(yraw, mu, model, object = object)
        } else {
            yraw
        }
    } else {
        yplot <- .GeoScatterplot_pit_data(object, scale = scale)
        if (length(yplot) != length(yraw)) {
            stop("GeoPit() returned a transformed data vector with an invalid length")
        }
    }

    if (isTRUE(contour) && scale == "Original" && !isTRUE(residuals)) {
        mu_range <- range(mu, finite = TRUE)
        if (length(mu_range) != 2L || diff(mu_range) > sqrt(.Machine$double.eps)) {
            stop(
                "A single original-scale contour requires a constant fitted location; ",
                "use residuals = TRUE or select scale = 'Uniform' or 'Gaussian'"
            )
        }
    }

    contour_mean <- if (scale == "Original" && !isTRUE(residuals)) mu[1L] else 0

    if (is.null(neighb) && is.null(maxdist)) {
        if (!is.null(object$neighb) && length(object$neighb)) {
            neighb <- object$neighb
        } else if (!is.null(object$maxdist) && length(object$maxdist) == 1L &&
                   is.finite(object$maxdist)) {
            maxdist <- object$maxdist
        } else {
            stop("Specify either neighb or maxdist")
        }
    }

    setup <- if (isTRUE(contour)) {
        .GeoScatterplot_density_setup(object, model)
    } else {
        NULL
    }

    panel_dots <- dots
    if (scale == "Uniform") {
        if (is.null(panel_dots$xlim)) panel_dots$xlim <- c(0, 1)
        if (is.null(panel_dots$ylim)) panel_dots$ylim <- c(0, 1)
        if (is.null(panel_dots$xaxs)) panel_dots$xaxs <- "i"
        if (is.null(panel_dots$yaxs)) panel_dots$yaxs <- "i"
        if (is.null(panel_dots$asp)) panel_dots$asp <- 1
    }
    if (scale == "Gaussian") {
        if (is.null(panel_dots$xlim)) {
            panel_dots$xlim <- c(-gaussian.range, gaussian.range)
        }
        if (is.null(panel_dots$ylim)) {
            panel_dots$ylim <- c(-gaussian.range, gaussian.range)
        }
        if (is.null(panel_dots$xaxs)) panel_dots$xaxs <- "i"
        if (is.null(panel_dots$yaxs)) panel_dots$yaxs <- "i"
        if (is.null(panel_dots$asp)) panel_dots$asp <- 1
    }

    if (!is.null(object$anisopars)) {
        if (tolower(object$distance) != "eucl") {
            stop("Anisotropy with a non-Euclidean distance is not supported")
        }
        if (layout$dynamic) {
            layout$coords_dyn <- lapply(
                layout$coords_dyn,
                function(z) .GeoScatterplot_three_columns(GeoAniso(z, anisopars = object$anisopars))
            )
            layout$coords_obs <- do.call(rbind, layout$coords_dyn)
        } else {
            layout$coords_base <- .GeoScatterplot_three_columns(
                GeoAniso(layout$coords_base, anisopars = object$anisopars)
            )
            if (spacetime) {
                nsite <- nrow(layout$coords_base)
                layout$coords_obs <- layout$coords_base[
                    rep(seq_len(nsite), times = length(layout$time_values)),
                    , drop = FALSE
                ]
            } else {
                layout$coords_obs <- layout$coords_base
            }
        }
    }

    time_lags <- .GeoScatterplot_time_lags(object, time.lag)
    temporal_tol <- sqrt(.Machine$double.eps) *
        max(1, abs(layout$time_values), abs(time_lags))

    panel_output <- list()
    panel_index <- 1L

    if (!spacetime) {
        coords_work <- layout$coords_obs

        if (!is.null(neighb)) {
            if (any(!is.finite(neighb)) || any(neighb < 1L) ||
                any(neighb != as.integer(neighb))) {
                stop("neighb must contain positive integers")
            }
            neighb <- as.integer(neighb)

            nc <- ceiling(sqrt(length(neighb)))
            nr <- ceiling(length(neighb) / nc)
            graphics::par(mfrow = c(nr, nc))

            neigh_coords <- .GeoScatterplot_neighbor_coordinates(
                coords_work, object$distance
            )
            for (k in seq_along(neighb)) {
                sel <- GeoNeighIndex(
                    coordx = neigh_coords, distance = object$distance,
                    neighb = neighb[k], radius = radius
                )
                i <- as.integer(sel$colidx)
                j <- as.integer(sel$rowidx)
                lagv <- as.numeric(sel$lags)
                lag <- .GeoScatterplot_representative_lag(lagv, lag.method)

                main <- paste0("Neighb = ", neighb[k], "; h = ", signif(lag, 3))
                panel_output[[panel_index]] <- .GeoScatterplot_panel(
                    x = yplot[i], y = yplot[j], main = main,
                    lag = lag, time_lag = 0, setup = setup,
                    contour = contour, mean1 = contour_mean,
                    mean2 = contour_mean, scale = scale, ngrid = ngrid,
                    nlevels = nlevels, levels = levels,
                    contour.col = contour.col, contour.lwd = contour.lwd,
                    contour.labels = contour.labels, point.col = point.col,
                    dots = panel_dots
                )
                panel_output[[panel_index]]$indices <- cbind(colidx = i, rowidx = j)
                panel_output[[panel_index]]$pair_lags <- lagv
                panel_output[[panel_index]]$pair_time_lags <- rep(0, length(lagv))
                panel_index <- panel_index + 1L
            }
        } else {
            if (length(maxdist) != 1L || !is.finite(maxdist) || maxdist <= 0) {
                stop("maxdist must be a positive finite scalar")
            }

            bins <- seq(0, maxdist, length.out = numbins + 1L)
            pairs <- .GeoScatterplot_spatial_pairs_maxdist(
                data = yplot, coords = coords_work, maxdist = maxdist,
                bins = bins, distance = object$distance, radius = radius
            )

            nc <- ceiling(sqrt(numbins))
            nr <- ceiling(numbins / nc)
            graphics::par(mfrow = c(nr, nc))

            for (k in seq_len(numbins)) {
                use <- pairs$bin == bins[k]
                lag <- (bins[k] + bins[k + 1L]) / 2
                main <- paste0(
                    .GeoScatterplot_distance_label(bins, k),
                    "; h = ", signif(lag, 3)
                )

                panel_output[[panel_index]] <- .GeoScatterplot_panel(
                    x = pairs$v1[use], y = pairs$v2[use], main = main,
                    lag = lag, time_lag = 0, setup = setup,
                    contour = contour, mean1 = contour_mean,
                    mean2 = contour_mean, scale = scale, ngrid = ngrid,
                    nlevels = nlevels, levels = levels,
                    contour.col = contour.col, contour.lwd = contour.lwd,
                    contour.labels = contour.labels, point.col = point.col,
                    dots = panel_dots
                )
                panel_output[[panel_index]]$indices <- cbind(
                    colidx = pairs$i[use], rowidx = pairs$j[use]
                )
                panel_output[[panel_index]]$pair_lags <- pairs$lags[use]
                panel_output[[panel_index]]$pair_time_lags <- rep(0, sum(use))
                panel_index <- panel_index + 1L
            }
        }
    } else {
        if (!is.null(neighb)) {
            if (any(!is.finite(neighb)) || any(neighb < 1L) ||
                any(neighb != as.integer(neighb))) {
                stop(
                    "For spatio-temporal objects, ",
                    "'neighb' must contain positive integers"
                )
            }
            neighb <- as.integer(neighb)

            npanels <- length(neighb) * length(time_lags)
            nc <- ceiling(sqrt(npanels))
            nr <- ceiling(npanels / nc)
            graphics::par(mfrow = c(nr, nc))

            max_time <- if (length(time_lags)) max(time_lags) else 0
            selection_cache <- vector("list", length(neighb))

            for (kk in seq_along(neighb)) {
                K <- neighb[kk]
                spatial_input <- if (layout$dynamic) NULL else
                    .GeoScatterplot_neighbor_coordinates(
                        layout$coords_base, object$distance
                    )
                dynamic_input <- if (layout$dynamic)
                    .GeoScatterplot_neighbor_coordinates(
                        layout$coords_dyn, object$distance
                    ) else NULL
                sel <- GeoNeighIndex(
                    coordx = spatial_input,
                    coordt = layout$time_values,
                    coordx_dyn = dynamic_input,
                    distance = object$distance,
                    neighb = K,
                    maxdist = NULL,
                    maxtime = max_time,
                    radius = radius,
                    bivariate = FALSE
                )
                valid <- is.finite(sel$colidx) & is.finite(sel$rowidx) &
                    sel$colidx >= 1L & sel$rowidx >= 1L &
                    sel$colidx <= length(yplot) & sel$rowidx <= length(yplot) &
                    is.finite(sel$lags) & is.finite(sel$lagt)
                selection_cache[[kk]] <- list(
                    colidx = as.integer(sel$colidx[valid]),
                    rowidx = as.integer(sel$rowidx[valid]),
                    lags = as.numeric(sel$lags[valid]),
                    lagt = as.numeric(sel$lagt[valid])
                )
            }

            for (u in time_lags) {
                for (kk in seq_along(neighb)) {
                    sel <- selection_cache[[kk]]
                    i_all <- sel$colidx
                    j_all <- sel$rowidx
                    lag_all <- sel$lags
                    time_all <- sel$lagt
                    use <- abs(time_all - u) <= temporal_tol

                    i <- i_all[use]
                    j <- j_all[use]
                    lagv <- lag_all[use]
                    timev <- time_all[use]
                    lag <- .GeoScatterplot_representative_lag(lagv, lag.method, empty = 0)
                    u_rep <- .GeoScatterplot_representative_lag(timev, lag.method, empty = u)

                    main <- paste0(
                        "Neighb = ", neighb[kk], "; h = ", signif(lag, 3),
                        "; u = ", signif(u_rep, 3)
                    )

                    panel_output[[panel_index]] <- .GeoScatterplot_panel(
                        x = yplot[i], y = yplot[j], main = main,
                        lag = lag, time_lag = u_rep, setup = setup,
                        contour = contour, mean1 = contour_mean,
                        mean2 = contour_mean, scale = scale, ngrid = ngrid,
                        nlevels = nlevels, levels = levels,
                        contour.col = contour.col, contour.lwd = contour.lwd,
                        contour.labels = contour.labels, point.col = point.col,
                        dots = panel_dots
                    )
                    panel_output[[panel_index]]$indices <- cbind(colidx = i, rowidx = j)
                    panel_output[[panel_index]]$pair_lags <- lagv
                    panel_output[[panel_index]]$pair_time_lags <- timev
                    panel_output[[panel_index]]$neighb <- neighb[kk]
                    panel_index <- panel_index + 1L
                }
            }
        } else {
            if (length(maxdist) != 1L || !is.finite(maxdist) || maxdist <= 0) {
                stop("maxdist must be a positive finite scalar")
            }

            max_time <- if (length(time_lags)) max(time_lags) else 0
            K_all <- max(1L, max(layout$ns))
            sel <- GeoNeighIndex(
                coordx = if (layout$dynamic) NULL else
                    .GeoScatterplot_neighbor_coordinates(
                        layout$coords_base, object$distance
                    ),
                coordt = layout$time_values,
                coordx_dyn = if (layout$dynamic)
                    .GeoScatterplot_neighbor_coordinates(
                        layout$coords_dyn, object$distance
                    ) else NULL,
                distance = object$distance,
                neighb = K_all,
                maxdist = maxdist,
                maxtime = max_time,
                radius = radius,
                bivariate = FALSE
            )

            i_all <- as.integer(sel$colidx)
            j_all <- as.integer(sel$rowidx)
            lag_all <- as.numeric(sel$lags)
            time_all <- as.numeric(sel$lagt)
            valid <- is.finite(i_all) & is.finite(j_all) &
                i_all >= 1L & j_all >= 1L &
                i_all <= length(yplot) & j_all <= length(yplot) &
                is.finite(lag_all) & is.finite(time_all) & lag_all <= maxdist
            i_all <- i_all[valid]
            j_all <- j_all[valid]
            lag_all <- lag_all[valid]
            time_all <- time_all[valid]
            bins <- seq(0, maxdist, length.out = numbins + 1L)

            npanels <- numbins * length(time_lags)
            nc <- ceiling(sqrt(npanels))
            nr <- ceiling(npanels / nc)
            graphics::par(mfrow = c(nr, nc))

            for (u in time_lags) {
                time_use <- abs(time_all - u) <= temporal_tol
                for (k in seq_len(numbins)) {
                    spatial_use <- .GeoScatterplot_bin_membership(
                        lag_all, bins, k
                    )
                    use <- time_use & spatial_use
                    lag <- (bins[k] + bins[k + 1L]) / 2
                    main <- paste0(
                        .GeoScatterplot_distance_label(bins, k),
                        "; h = ", signif(lag, 3),
                        "; u = ", signif(u, 3)
                    )

                    panel_output[[panel_index]] <- .GeoScatterplot_panel(
                        x = yplot[i_all[use]], y = yplot[j_all[use]],
                        main = main, lag = lag, time_lag = u, setup = setup,
                        contour = contour, mean1 = contour_mean,
                        mean2 = contour_mean, scale = scale, ngrid = ngrid,
                        nlevels = nlevels, levels = levels,
                        contour.col = contour.col, contour.lwd = contour.lwd,
                        contour.labels = contour.labels, point.col = point.col,
                        dots = panel_dots
                    )
                    panel_output[[panel_index]]$indices <- cbind(
                        colidx = i_all[use], rowidx = j_all[use]
                    )
                    panel_output[[panel_index]]$pair_lags <- lag_all[use]
                    panel_output[[panel_index]]$pair_time_lags <- time_all[use]
                    panel_index <- panel_index + 1L
                }
            }
        }
    }

    invisible(list(
        call = match.call(), model = model,
        copula = if (isTRUE(independence)) "Independence" else object$copula,
        independence = isTRUE(independence),
        contour = isTRUE(contour), scale = scale,
        residuals = isTRUE(residuals), spacetime = spacetime,
        time_lags = time_lags, panels = panel_output
    ))
}

.GeoScatterplot_raw_time_indices <- function(coordt, times = NULL) {
    coordt <- as.numeric(coordt)
    if (length(coordt) < 2L || any(!is.finite(coordt))) {
        stop("coordt must contain at least two finite temporal coordinates")
    }
    if (anyDuplicated(coordt)) {
        stop("coordt must not contain duplicated temporal coordinates")
    }

    if (is.null(times)) return(seq_along(coordt))
    if (!is.numeric(times) || !length(times) || any(!is.finite(times))) {
        stop("times must contain finite temporal coordinates or valid time indices")
    }

    tol <- sqrt(.Machine$double.eps) * max(1, abs(coordt), abs(times))
    idx <- vapply(times, function(tt) {
        jj <- which.min(abs(coordt - tt))
        if (length(jj) && abs(coordt[jj] - tt) <= tol) return(as.integer(jj))

        if (tt == as.integer(tt) && tt >= 1L && tt <= length(coordt)) {
            return(as.integer(tt))
        }

        stop(
            "The temporal instant ", signif(tt, 8),
            " is neither present in coordt nor a valid time index"
        )
    }, integer(1L))

    idx <- sort(unique(idx))
    if (length(idx) < 2L) {
        stop("times must select at least two temporal instants")
    }
    idx
}

.GeoScatterplot_raw_time_lags <- function(coordt, time.lag = NULL) {
    coordt <- as.numeric(coordt)
    possible <- .GeoScatterplot_distinct_time_lags(coordt)
    tol <- sqrt(.Machine$double.eps) * max(1, abs(coordt), abs(possible))

    if (is.null(time.lag)) return(possible)
    if (!is.numeric(time.lag) || !length(time.lag) ||
        any(!is.finite(time.lag)) || any(time.lag < 0)) {
        stop("time.lag must contain non-negative finite temporal lags")
    }

    matched <- vapply(time.lag, function(u) {
        jj <- which.min(abs(possible - u))
        local_tol <- max(tol, sqrt(.Machine$double.eps) * max(1, abs(u)))
        if (!length(jj) || abs(possible[jj] - u) > local_tol) {
            stop(
                "Temporal lag ", signif(u, 8),
                " is not available from the selected temporal coordinates"
            )
        }
        possible[jj]
    }, numeric(1L))

    .GeoScatterplot_unique_tol(matched, tol = tol)
}

.GeoScatterplot_raw_spacetime_layout <- function(data, coords = NULL,
                                                  coordt, coordx_dyn = NULL,
                                                  times = NULL) {
    coordt <- as.numeric(coordt)
    time_idx <- .GeoScatterplot_raw_time_indices(coordt, times)
    coordt_use <- coordt[time_idx]
    nt <- length(coordt)

    if (is.null(coordx_dyn)) {
        coords <- .GeoScatterplot_three_columns(coords, validate = TRUE)
        nsite <- nrow(coords)

        if (is.matrix(data) || is.data.frame(data)) {
            data_mat <- as.matrix(data)
            if (!identical(dim(data_mat), c(nt, nsite))) {
                stop(
                    "For fixed spatio-temporal locations, data must be a ",
                    "length(coordt) by nrow(coordx) matrix"
                )
            }
        } else {
            data_vec <- as.numeric(data)
            if (length(data_vec) != nt * nsite) {
                stop(
                    "For fixed spatio-temporal locations, length(data) must equal ",
                    "length(coordt) * nrow(coordx)"
                )
            }
            data_mat <- matrix(data_vec, nrow = nt, ncol = nsite, byrow = TRUE)
        }

        data_mat <- data_mat[time_idx, , drop = FALSE]
        y <- as.numeric(t(data_mat))
        ns <- rep.int(nsite, length(time_idx))
        coords_obs <- coords[rep(seq_len(nsite), times = length(time_idx)),
                             , drop = FALSE]
        time_obs <- rep(coordt_use, each = nsite)

        return(list(
            data = y,
            coords_base = coords,
            coords_dyn = NULL,
            coords_obs = coords_obs,
            time_obs = time_obs,
            time_values = coordt_use,
            ns = ns,
            dynamic = FALSE
        ))
    }

    if (!is.list(coordx_dyn) || length(coordx_dyn) != nt) {
        stop("coordx_dyn must be a list with one coordinate matrix per temporal instant")
    }
    coords_dyn_all <- lapply(coordx_dyn, .GeoScatterplot_three_columns, validate = TRUE)
    ns_all <- vapply(coords_dyn_all, nrow, integer(1L))
    if (any(ns_all <= 0L)) {
        stop("Each element of coordx_dyn must contain at least one location")
    }

    if (is.list(data)) {
        if (length(data) != nt) {
            stop("For dynamic locations, data must have one component per temporal instant")
        }
        data_all <- lapply(seq_len(nt), function(k) {
            z <- as.numeric(data[[k]])
            if (length(z) != ns_all[k]) {
                stop("data[[", k, "]] and coordx_dyn[[", k, "]] have different lengths")
            }
            z
        })
    } else if ((is.matrix(data) || is.data.frame(data)) &&
               length(unique(ns_all)) == 1L &&
               identical(dim(as.matrix(data)), c(nt, ns_all[1L]))) {
        data_mat <- as.matrix(data)
        data_all <- lapply(seq_len(nt), function(k) as.numeric(data_mat[k, ]))
    } else {
        data_vec <- as.numeric(data)
        if (length(data_vec) != sum(ns_all)) {
            stop(
                "For dynamic locations, length(data) must equal the total ",
                "number of rows in coordx_dyn"
            )
        }
        ends <- cumsum(ns_all)
        starts <- c(1L, ends[-length(ends)] + 1L)
        data_all <- Map(function(a, b) data_vec[a:b], starts, ends)
    }

    coords_dyn <- coords_dyn_all[time_idx]
    data_use <- data_all[time_idx]
    ns <- ns_all[time_idx]
    y <- as.numeric(unlist(data_use, use.names = FALSE))
    coords_obs <- do.call(rbind, coords_dyn)
    time_obs <- rep(coordt_use, times = ns)

    list(
        data = y,
        coords_base = NULL,
        coords_dyn = coords_dyn,
        coords_obs = coords_obs,
        time_obs = time_obs,
        time_values = coordt_use,
        ns = ns,
        dynamic = TRUE
    )
}

.GeoScatterplot_raw_spacetime <- function(data, coords = NULL, coordt,
                                          coordx_dyn = NULL, times = NULL,
                                          time.lag = NULL, distance = "Eucl",
                                          maxdist = NULL, neighb = NULL,
                                          numbins = 4L, radius = 1,
                                          point.col, lag.method, dots,
                                          call) {
    layout <- .GeoScatterplot_raw_spacetime_layout(
        data = data, coords = coords, coordt = coordt,
        coordx_dyn = coordx_dyn, times = times
    )
    time_lags <- .GeoScatterplot_raw_time_lags(
        layout$time_values, time.lag = time.lag
    )
    temporal_tol <- sqrt(.Machine$double.eps) *
        max(1, abs(layout$time_values), abs(time_lags))

    lag.method <- match.arg(lag.method, choices = c("median", "mean"))
    if (is.null(neighb) && is.null(maxdist)) {
        stop("Specify either neighb or maxdist")
    }
    if (!is.null(neighb) && !is.null(maxdist)) {
        warning("Both neighb and maxdist were supplied; neighb is used")
    }

    max_time <- if (length(time_lags)) max(time_lags) else 0
    panel_index <- 1L
    panels <- list()

    if (!is.null(neighb)) {
        if (any(!is.finite(neighb)) || any(neighb < 1L) ||
            any(neighb != as.integer(neighb))) {
            stop("neighb must contain positive integers")
        }
        neighb <- as.integer(neighb)
        npanels <- length(neighb) * length(time_lags)
        nc <- ceiling(sqrt(npanels))
        nr <- ceiling(npanels / nc)
        graphics::par(mfrow = c(nr, nc))

        selection_cache <- lapply(neighb, function(K) {
            GeoNeighIndex(
                coordx = if (layout$dynamic) NULL else
                    .GeoScatterplot_neighbor_coordinates(layout$coords_base, distance),
                coordt = layout$time_values,
                coordx_dyn = if (layout$dynamic)
                    .GeoScatterplot_neighbor_coordinates(layout$coords_dyn, distance) else NULL,
                distance = distance,
                neighb = K,
                maxdist = NULL,
                maxtime = max_time,
                radius = radius,
                bivariate = FALSE
            )
        })

        for (u in time_lags) {
            for (kk in seq_along(neighb)) {
                sel <- selection_cache[[kk]]
                i_all <- as.integer(sel$colidx)
                j_all <- as.integer(sel$rowidx)
                lag_all <- as.numeric(sel$lags)
                time_all <- as.numeric(sel$lagt)

                valid <- is.finite(i_all) & is.finite(j_all) &
                    i_all >= 1L & j_all >= 1L &
                    i_all <= length(layout$data) & j_all <= length(layout$data) &
                    is.finite(lag_all) & is.finite(time_all)
                use <- valid & abs(time_all - u) <= temporal_tol

                i <- i_all[use]
                j <- j_all[use]
                lagv <- lag_all[use]
                timev <- time_all[use]
                lag <- .GeoScatterplot_representative_lag(lagv, lag.method, empty = 0)
                u_rep <- .GeoScatterplot_representative_lag(timev, lag.method, empty = u)
                main <- paste0(
                    "Neighb = ", neighb[kk], "; h = ", signif(lag, 3),
                    "; u = ", signif(u_rep, 3)
                )

                panels[[panel_index]] <- .GeoScatterplot_panel(
                    x = layout$data[i], y = layout$data[j], main = main,
                    lag = lag, time_lag = u_rep, setup = NULL,
                    contour = FALSE, mean1 = 0, mean2 = 0,
                    scale = "Original", ngrid = 20L, nlevels = 1L,
                    levels = NULL, contour.col = "red", contour.lwd = 1,
                    contour.labels = FALSE, point.col = point.col,
                    dots = dots
                )
                panels[[panel_index]]$indices <- cbind(colidx = i, rowidx = j)
                panels[[panel_index]]$pair_lags <- lagv
                panels[[panel_index]]$pair_time_lags <- timev
                panels[[panel_index]]$neighb <- neighb[kk]
                panel_index <- panel_index + 1L
            }
        }
    } else {
        if (length(maxdist) != 1L || !is.finite(maxdist) || maxdist <= 0) {
            stop("maxdist must be a positive finite scalar")
        }

        K_all <- max(1L, max(layout$ns))
        sel <- GeoNeighIndex(
            coordx = if (layout$dynamic) NULL else
                .GeoScatterplot_neighbor_coordinates(layout$coords_base, distance),
            coordt = layout$time_values,
            coordx_dyn = if (layout$dynamic)
                .GeoScatterplot_neighbor_coordinates(layout$coords_dyn, distance) else NULL,
            distance = distance,
            neighb = K_all,
            maxdist = maxdist,
            maxtime = max_time,
            radius = radius,
            bivariate = FALSE
        )

        i_all <- as.integer(sel$colidx)
        j_all <- as.integer(sel$rowidx)
        lag_all <- as.numeric(sel$lags)
        time_all <- as.numeric(sel$lagt)
        valid <- is.finite(i_all) & is.finite(j_all) &
            i_all >= 1L & j_all >= 1L &
            i_all <= length(layout$data) & j_all <= length(layout$data) &
            is.finite(lag_all) & is.finite(time_all) & lag_all <= maxdist
        i_all <- i_all[valid]
        j_all <- j_all[valid]
        lag_all <- lag_all[valid]
        time_all <- time_all[valid]

        bins <- seq(0, maxdist, length.out = numbins + 1L)
        npanels <- numbins * length(time_lags)
        nc <- ceiling(sqrt(npanels))
        nr <- ceiling(npanels / nc)
        graphics::par(mfrow = c(nr, nc))

        for (u in time_lags) {
            time_use <- abs(time_all - u) <= temporal_tol
            for (k in seq_len(numbins)) {
                spatial_use <- .GeoScatterplot_bin_membership(
                    lag_all, bins, k
                )
                use <- time_use & spatial_use
                lag <- (bins[k] + bins[k + 1L]) / 2
                main <- paste0(
                    .GeoScatterplot_distance_label(bins, k),
                    "; h = ", signif(lag, 3),
                    "; u = ", signif(u, 3)
                )

                panels[[panel_index]] <- .GeoScatterplot_panel(
                    x = layout$data[i_all[use]], y = layout$data[j_all[use]],
                    main = main, lag = lag, time_lag = u, setup = NULL,
                    contour = FALSE, mean1 = 0, mean2 = 0,
                    scale = "Original", ngrid = 20L, nlevels = 1L,
                    levels = NULL, contour.col = "red", contour.lwd = 1,
                    contour.labels = FALSE, point.col = point.col,
                    dots = dots
                )
                panels[[panel_index]]$indices <- cbind(
                    colidx = i_all[use], rowidx = j_all[use]
                )
                panels[[panel_index]]$pair_lags <- lag_all[use]
                panels[[panel_index]]$pair_time_lags <- time_all[use]
                panel_index <- panel_index + 1L
            }
        }
    }

    invisible(list(
        call = call, spacetime = TRUE, time_lags = time_lags,
        time_values = layout$time_values, panels = panels
    ))
}

GeoScatterplot <- function(data, coordx = NULL, coordy = NULL, coordz = NULL,
                           coordt = NULL, coordx_dyn = NULL,
                           distance = "Eucl", grid = FALSE,
                           maxdist = NULL, neighb = NULL,
                           times = NULL, time.lag = NULL, numbins = 4, radius = 1,
                           bivariate = FALSE,
                           contour = inherits(data, "GeoFit"), residuals = FALSE,
                           scale = c("Original", "Gaussian", "Uniform"),
                           gaussian.range = 3, ngrid = 80, nlevels = 6, levels = NULL,
                           contour.col = "red", contour.lwd = 1.5,
                           contour.labels = FALSE,
                           point.col = grDevices::adjustcolor("#481567FF", 0.45),
                           lag.method = c("median", "mean"), ...) {

    call <- match.call()
    opar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(opar))

    if (inherits(data, "GeoFit")) {
        fit_radius <- if (missing(radius)) data$radius else radius
        if (is.null(fit_radius) || !is.finite(fit_radius)) fit_radius <- 1

        return(.GeoScatterplot_GeoFit(
            object = data,
            maxdist = maxdist,
            neighb = neighb,
            time.lag = time.lag,
            numbins = numbins,
            radius = fit_radius,
            contour = contour,
            residuals = residuals,
            scale = scale,
            gaussian.range = gaussian.range,
            ngrid = ngrid,
            nlevels = nlevels,
            levels = levels,
            contour.col = contour.col,
            contour.lwd = contour.lwd,
            contour.labels = contour.labels,
            point.col = point.col,
            lag.method = lag.method,
            dots = list(...)
        ))
    }

    raw_coord_dim <- .GeoCoordinateDimension(
        coordx = coordx, coordy = coordy, coordz = coordz,
        coordx_dyn = coordx_dyn
    )

    scale <- .GeoScatterplot_match_scale(scale)
    if (scale != "Original") {
        stop("Uniform and Gaussian scales require a GeoFit object as the first argument")
    }

    if (isTRUE(contour)) {
        stop("Fitted contours require a GeoFit object as the first argument")
    }

    CheckDistance(distance)
    spacetime <- !isTRUE(bivariate) && !is.null(coordt) &&
        is.numeric(coordt) && length(coordt) > 1L

    if (length(numbins) != 1 || is.na(numbins) || numbins < 1 ||
        numbins != as.integer(numbins)) {
        stop("numbins must be a positive integer")
    }
    numbins <- as.integer(numbins)

    if (!spacetime && !is.null(time.lag)) {
        stop("time.lag can be used only for spatio-temporal data")
    }
    if (!spacetime && !is.null(times)) {
        stop("times can be used only for spatio-temporal data")
    }

    if (!is.null(coordx_dyn) && spacetime) {
        .GeoValidate3D(
            coordx_dyn = coordx_dyn, distance = distance, grid = grid,
            context = "GeoScatterplot"
        )
        return(.GeoScatterplot_raw_spacetime(
            data = data, coords = NULL, coordt = coordt,
            coordx_dyn = coordx_dyn, times = times,
            time.lag = time.lag, distance = distance,
            maxdist = maxdist, neighb = neighb, numbins = numbins,
            radius = radius, point.col = point.col,
            lag.method = lag.method, dots = list(...), call = call
        ))
    }

    if (grid) {
        if (is.null(coordy)) stop("coordy must be provided when grid=TRUE")
        if (is.null(coordz)) {
            a <- as.matrix(expand.grid(coordx, coordy))
            coordx <- a[, 1]
            coordy <- a[, 2]
            coordz <- rep(0, length(coordx))
        } else {
            a <- as.matrix(expand.grid(coordx, coordy, coordz))
            coordx <- a[, 1]
            coordy <- a[, 2]
            coordz <- a[, 3]
        }
    }

    if (is.null(coordx_dyn)) {
        if (is.matrix(coordx) || is.data.frame(coordx)) {
            coordx <- as.matrix(coordx)
            if (ncol(coordx) == 1) {
                if (is.null(coordy)) stop("coordy must be provided when coordx has one column")
                if (is.null(coordz)) coordz <- rep(0, nrow(coordx))
                coords <- cbind(coordx[, 1], coordy, coordz)
            } else if (ncol(coordx) == 2) {
                coords <- cbind(coordx[, 1], coordx[, 2], 0)
            } else {
                coords <- coordx[, 1:3, drop = FALSE]
            }
        } else {
            if (is.null(coordy)) {
                stop("coordy must be provided when coordx is not a coordinate matrix")
            }
            if (is.null(coordz)) coordz <- rep(0, length(coordx))
            coords <- cbind(coordx, coordy, coordz)
        }

        coords <- as.matrix(coords)
        storage.mode(coords) <- "double"

        if (spacetime) {
            validate_coords <- if (identical(raw_coord_dim, 2L))
                coords[, 1:2, drop = FALSE] else coords
            .GeoValidate3D(
                coordx = validate_coords, distance = distance, grid = grid,
                context = "GeoScatterplot"
            )
            return(.GeoScatterplot_raw_spacetime(
                data = data, coords = coords, coordt = coordt,
                coordx_dyn = NULL, times = times,
                time.lag = time.lag, distance = distance,
                maxdist = maxdist, neighb = neighb, numbins = numbins,
                radius = radius, point.col = point.col,
                lag.method = lag.method, dots = list(...), call = call
            ))
        }

        if (bivariate) coords <- rbind(coords, coords)

        data <- as.numeric(t(data))
        numcoord <- nrow(coords)

        if (bivariate) {
            ns <- rep(numcoord / 2, 2)
        } else if (!is.null(coordt)) {
            ns <- rep(numcoord, length(coordt))
        } else {
            ns <- numcoord
        }
    } else {
        coords <- do.call(rbind, coordx_dyn)
        coords <- as.matrix(coords)
        storage.mode(coords) <- "double"
        if (ncol(coords) == 2) coords <- cbind(coords, 0)

        data <- as.numeric(unlist(data))
        ns <- vapply(coordx_dyn, nrow, integer(1L))
    }

    validate_coords <- if (identical(raw_coord_dim, 2L))
        coords[, 1:2, drop = FALSE] else coords
    .GeoValidate3D(
        coordx = validate_coords, distance = distance, grid = grid,
        context = "GeoScatterplot"
    )

    if (ncol(coords) == 2) coords <- cbind(coords, 0)
    if (ncol(coords) != 3) stop("coordinates must have two or three columns")

    if (length(data) != nrow(coords)) {
        stop("length(data) must match the number of coordinate rows after processing")
    }

    .panel_layout <- function(n) {
        nc <- ceiling(sqrt(n))
        nr <- ceiling(n / nc)
        c(nr, nc)
    }

    .call_pairs <- function(data, coords, bins, maxdist) {
        .GeoScatterplot_spatial_pairs_maxdist(
            data = data, coords = coords, maxdist = maxdist,
            bins = bins, distance = distance, radius = radius
        )
    }

    ##################################################
    ##### Scatterplot based on distances
    ##################################################

    if (is.null(neighb)) {
        if (is.null(maxdist)) stop("maxdist must be provided when neighb is NULL")

        if (!bivariate && !spacetime) {
            if (length(maxdist) != 1) stop("maxdist must be a scalar for the univariate spatial case")

            bins <- seq(0, maxdist, length.out = numbins + 1)
            vvv <- .call_pairs(data, coords, bins, maxdist)

            lay <- .panel_layout(numbins)
            par(mfrow = lay)

            for (i in seq_len(numbins)) {
                sel <- vvv$bin == bins[i]
                v111 <- vvv$v1[sel]
                v222 <- vvv$v2[sel]

                main <- .GeoScatterplot_distance_label(bins, i)

                if (length(v111) == 0 || length(v222) == 0) {
                    plot.new()
                    title(main = main)
                    mtext("No pairs", side = 3, line = -2)
                    message("There are no points between ", signif(bins[i], 3),
                            " and ", signif(bins[i + 1], 3))
                } else {
                    plot_args <- modifyList(
                        list(x = v111, y = v222, col = point.col,
                             xlab = "", ylab = "", main = main),
                        list(...)
                    )
                    do.call(graphics::plot, plot_args)
                    graphics::abline(0, 1)
                }
            }
        }

        if (bivariate) {
            if (length(maxdist) == 1) maxdist <- rep(maxdist, 2)
            if (length(maxdist) != 2) stop("maxdist must have length one or two for bivariate=TRUE")

            data1 <- data[seq_len(ns[1])]
            data2 <- data[(ns[1] + 1):(ns[1] + ns[2])]
            coords1 <- coords[seq_len(ns[1]), , drop = FALSE]
            coords2 <- coords[(ns[1] + 1):(ns[1] + ns[2]), , drop = FALSE]

            bins1 <- seq(0, maxdist[1], length.out = numbins + 1)
            bins2 <- seq(0, maxdist[2], length.out = numbins + 1)

            vvv1 <- .call_pairs(data1, coords1, bins1, maxdist[1])
            vvv2 <- .call_pairs(data2, coords2, bins2, maxdist[2])

            lay <- .panel_layout(2 * numbins)
            par(mfrow = lay)

            for (i in seq_len(numbins)) {
                sel <- vvv1$bin == bins1[i]
                v111 <- vvv1$v1[sel]
                v222 <- vvv1$v2[sel]
                cc <- if (length(v111) > 1 && length(v222) > 1) stats::cor(v111, v222) else NA_real_
                main <- paste("cor =", signif(cc, 3),
                              .GeoScatterplot_distance_label(bins1, i))

                if (length(v111) == 0 || length(v222) == 0) {
                    plot.new(); title(main = main); mtext("No pairs", side = 3, line = -2)
                } else {
                    plot_args <- modifyList(
                        list(x = v111, y = v222, col = point.col,
                             xlab = "", ylab = "", main = main),
                        list(...)
                    )
                    do.call(graphics::plot, plot_args)
                    graphics::abline(0, 1)
                }
            }

            for (i in seq_len(numbins)) {
                sel <- vvv2$bin == bins2[i]
                v111 <- vvv2$v1[sel]
                v222 <- vvv2$v2[sel]
                cc <- if (length(v111) > 1 && length(v222) > 1) stats::cor(v111, v222) else NA_real_
                main <- paste("cor =", signif(cc, 3),
                              .GeoScatterplot_distance_label(bins2, i))

                if (length(v111) == 0 || length(v222) == 0) {
                    plot.new(); title(main = main); mtext("No pairs", side = 3, line = -2)
                } else {
                    plot_args <- modifyList(
                        list(x = v111, y = v222, col = point.col,
                             xlab = "", ylab = "", main = main),
                        list(...)
                    )
                    do.call(graphics::plot, plot_args)
                    graphics::abline(0, 1)
                }
            }
        }
    }

    ##################################################
    ##### Scatterplot based on neighborhood
    ##################################################

    if (!is.null(neighb)) {
        ln <- length(neighb)
        lay <- .panel_layout(ln)

        if (!bivariate && !spacetime) {
            par(mfrow = lay)
            neigh_coords <- .GeoScatterplot_neighbor_coordinates(coords, distance)
            for (i in seq_len(ln)) {
                sel <- GeoNeighIndex(
                    coordx = neigh_coords, distance = distance,
                    neighb = neighb[i], radius = radius
                )
                data1 <- data[sel$colidx]
                data2 <- data[sel$rowidx]
                plot_args <- modifyList(
                    list(x = data1, y = data2, col = point.col,
                         xlab = "", ylab = "",
                         main = paste("Neighb =", neighb[i])),
                    list(...)
                )
                do.call(graphics::plot, plot_args)
                graphics::abline(0, 1)
            }
        }

        if (bivariate) {
            stop("Neighborhood scatterplot for bivariate=TRUE is not implemented yet")
        }

    }

    invisible(call)
}

## Internal parallel-worker utilities shared across GeoModels.

.GeoValidateNcores <- function(ncores, context = "ncores") {
  if (is.null(ncores)) return(NULL)
  tol <- sqrt(.Machine$double.eps)
  if (!is.numeric(ncores) || length(ncores) != 1L || !is.finite(ncores) ||
      ncores < 1 || abs(ncores - round(ncores)) > tol) {
    stop(context, " must be NULL or a positive integer", call. = FALSE)
  }
  as.integer(round(ncores))
}

.GeoResolveWorkers <- function(parallel = TRUE, ncores = NULL, n_jobs = Inf,
                               auto_max = 6L, context = "ncores") {
  if (!is.logical(parallel) || length(parallel) != 1L || is.na(parallel))
    stop("parallel must be TRUE or FALSE", call. = FALSE)

  ncores <- .GeoValidateNcores(ncores, context)
  coremax <- suppressWarnings(parallel::detectCores())
  if (is.na(coremax) || coremax < 1L) coremax <- 1L
  coremax <- as.integer(coremax)

  jobs <- suppressWarnings(as.numeric(n_jobs)[1L])
  if (is.na(jobs) || jobs <= 0) jobs <- 1
  if (!is.finite(jobs)) jobs <- Inf

  if (!isTRUE(parallel) || coremax <= 1L || jobs <= 1) return(1L)

  if (is.null(ncores)) {
    auto_max <- suppressWarnings(as.numeric(auto_max)[1L])
    if (!is.finite(auto_max) || auto_max < 1) auto_max <- 6L
    workers <- min(as.integer(auto_max), max(1L, coremax - 1L))
  } else {
    ## An explicit request may use all detected cores; the automatic path
    ## leaves one core free and is capped at six workers package-wide.
    workers <- min(ncores, coremax)
  }

  if (is.finite(jobs)) workers <- min(workers, as.integer(floor(jobs)))
  as.integer(max(1L, workers))
}

####################################################
### File name: GeoCovDisplay.R
####################################################

GeoCovDisplay <- function(covmatrix, limits = FALSE, pch = 2) {

  ## Internal display controls.
  tol <- 1e-150
  max_points <- 250000L
  max_raster_cells <- 1000000L
  main <- NULL

  ##########################################################################
  ## Input checks
  ##########################################################################

  if (!inherits(covmatrix, "GeoCovmatrix")) {
    stop("A GeoCovmatrix object is needed as input.")
  }

  if (is.null(covmatrix$covmatrix)) {
    stop("The GeoCovmatrix object does not contain a covariance matrix.")
  }

  if (length(limits) != 1L || is.na(limits)) {
    stop("limits must be TRUE or FALSE.")
  }

  limits <- isTRUE(limits)

  if (length(tol) != 1L || !is.finite(tol) || tol < 0) {
    stop("tol must be a finite non-negative scalar.")
  }

  if (length(max_points) != 1L ||
      !is.finite(max_points) || max_points < 1) {
    stop("max_points must be a positive finite scalar.")
  }

  if (length(max_raster_cells) != 1L ||
      !is.finite(max_raster_cells) || max_raster_cells < 4) {
    stop("max_raster_cells must be a finite scalar not smaller than 4.")
  }

  max_points <- as.integer(max_points)
  max_raster_cells <- as.integer(max_raster_cells)

  covmat <- covmatrix$covmatrix
  is_spam <- inherits(covmat, "spam")
  is_dense_matrix <- is.matrix(covmat)

  if (!is_spam && !is_dense_matrix) {
    stop(
      "covmatrix$covmatrix must be either a base R matrix or a spam matrix."
    )
  }

  ##########################################################################
  ## Utilities
  ##########################################################################

  make_raster_dimensions <- function(nr, nc, max_cells) {
    ratio <- nc / nr

    nx <- floor(sqrt(max_cells * ratio))
    nx <- max(1L, min(nc, as.integer(nx)))

    ny <- floor(max_cells / nx)
    ny <- max(1L, min(nr, as.integer(ny)))

    ## If clipping ny permits a wider raster without exceeding the budget,
    ## increase nx accordingly.
    nx <- max(1L, min(nc, as.integer(floor(max_cells / ny))))

    c(nx = nx, ny = ny)
  }

  add_to_raster <- function(z, rows, cols, nr, nc) {
    if (!length(rows)) {
      return(z)
    }

    nx <- nrow(z)
    ny <- ncol(z)

    bx <- pmin(
      nx,
      as.integer(floor((cols - 1) * nx / nc) + 1L)
    )

    ## Matrix row 1 is displayed at the top of the graph.
    by_top <- pmin(
      ny,
      as.integer(floor((rows - 1) * ny / nr) + 1L)
    )
    by <- ny - by_top + 1L

    z[cbind(bx, by)] <- 1L
    z
  }

  draw_empty_frame <- function() {
    graphics::plot.new()
    graphics::plot.window(
      xlim = c(0, 1),
      ylim = c(0, 1),
      xaxs = "i",
      yaxs = "i"
    )

    if (!is.null(main)) {
      graphics::title(main = main)
    }
  }

  draw_points <- function(rows, cols, nr, nc) {
    draw_empty_frame()

    if (length(rows)) {
      graphics::points(
        x = (cols - 0.5) / nc,
        y = 1 - (rows - 0.5) / nr,
        pch = pch
      )
    }
  }

  draw_raster <- function(z) {
    nx <- nrow(z)
    ny <- ncol(z)

    ## image() needs at least two coordinates in each direction. Degenerate
    ## one-row or one-column cases are therefore displayed as points.
    if (nx < 2L || ny < 2L) {
      idx <- which(z != 0, arr.ind = TRUE)
      draw_empty_frame()

      if (nrow(idx)) {
        graphics::points(
          x = (idx[, 1L] - 0.5) / nx,
          y = (idx[, 2L] - 0.5) / ny,
          pch = pch
        )
      }
      return(invisible(NULL))
    }

    graphics::image(
      x = (seq_len(nx) - 0.5) / nx,
      y = (seq_len(ny) - 0.5) / ny,
      z = z,
      zlim = c(0, 1),
      col = grDevices::gray.colors(2L, start = 1, end = 0),
      xaxt = "n",
      yaxt = "n",
      xlab = "",
      ylab = "",
      xaxs = "i",
      yaxs = "i",
      main = main,
      useRaster = TRUE
    )
  }

  draw_limits <- function(nr, nc) {
    if (!limits) {
      return(invisible(NULL))
    }

    horizontal_cuts <- numeric(0L)
    vertical_cuts <- numeric(0L)

    if (isTRUE(covmatrix$bivariate)) {
      ns <- covmatrix$ns

      if (!is.null(ns) &&
          length(ns) >= 2L &&
          all(is.finite(ns)) &&
          all(ns > 0) &&
          sum(ns) == nr &&
          nr == nc) {
        cuts <- cumsum(ns) / sum(ns)
        cuts <- cuts[-length(cuts)]
        horizontal_cuts <- c(horizontal_cuts, cuts)
        vertical_cuts <- c(vertical_cuts, cuts)
      } else {
        ## Backward-compatible fallback when block sizes are unavailable.
        horizontal_cuts <- c(horizontal_cuts, 0.5)
        vertical_cuts <- c(vertical_cuts, 0.5)
      }
    }

    if (isTRUE(covmatrix$spacetime)) {
      nt <- suppressWarnings(as.integer(covmatrix$numtime))
      ns <- covmatrix$ns

      if (!is.null(ns) &&
          length(ns) > 1L &&
          all(is.finite(ns)) &&
          all(ns > 0) &&
          sum(ns) == nr &&
          nr == nc) {
        cuts <- cumsum(ns) / sum(ns)
        cuts <- cuts[-length(cuts)]
      } else if (length(nt) == 1L && is.finite(nt) && nt > 1L) {
        cuts <- seq_len(nt - 1L) / nt
      } else {
        cuts <- numeric(0L)
      }

      horizontal_cuts <- c(horizontal_cuts, cuts)
      vertical_cuts <- c(vertical_cuts, cuts)
    }

    horizontal_cuts <- unique(horizontal_cuts)
    vertical_cuts <- unique(vertical_cuts)

    if (length(horizontal_cuts)) {
      graphics::abline(h = horizontal_cuts)
    }
    if (length(vertical_cuts)) {
      graphics::abline(v = vertical_cuts)
    }

    invisible(NULL)
  }

  ##########################################################################
  ## Extract and display a spam sparsity pattern without densification
  ##########################################################################

  display_spam <- function(x) {
    entries <- methods::slot(x, "entries")
    colindices <- methods::slot(x, "colindices")
    rowpointers <- methods::slot(x, "rowpointers")
    dims <- methods::slot(x, "dimension")

    nr <- as.integer(dims[1L])
    nc <- as.integer(dims[2L])

    if (length(rowpointers) != nr + 1L) {
      stop("Invalid spam matrix: inconsistent row pointers.")
    }

    if (length(entries) != length(colindices)) {
      stop("Invalid spam matrix: entries and column indices differ in length.")
    }

    if (any(!is.finite(entries))) {
      stop("The covariance matrix contains non-finite stored entries.")
    }

    keep <- abs(entries) > tol
    nnz <- sum(keep)
    total_cells <- as.double(nr) * as.double(nc)

    if (as.double(nnz) == total_cells) {
      message("No image plot: the covariance matrix is dense.")
      return(invisible(list(
        nrow = nr,
        ncol = nc,
        nonzero = nnz,
        mode = "dense"
      )))
    }

    old_par <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(old_par), add = TRUE)

    ## For a moderate number of stored non-zero entries, plotting their
    ## coordinates directly is exact and requires O(nnz) rather than O(nr*nc)
    ## memory.
    if (nnz <= max_points) {
      row_counts <- diff(rowpointers)
      rows <- rep.int(seq_len(nr), row_counts)

      rows <- rows[keep]
      cols <- colindices[keep]

      draw_points(rows, cols, nr, nc)
      draw_limits(nr, nc)

      return(invisible(list(
        nrow = nr,
        ncol = nc,
        nonzero = nnz,
        mode = "points"
      )))
    }

    ## For very large sparse patterns, aggregate non-zero entries into a
    ## bounded raster. The full nr x nc logical matrix is never allocated.
    raster_dims <- make_raster_dimensions(
      nr,
      nc,
      max_raster_cells
    )

    z <- matrix(
      0L,
      nrow = raster_dims[["nx"]],
      ncol = raster_dims[["ny"]]
    )

    ## Process row blocks to avoid allocating a row-index vector of length nnz.
    target_entries_per_chunk <- 1000000L
    stored_nnz <- length(entries)
    mean_per_row <- max(1, ceiling(stored_nnz / max(nr, 1L)))
    rows_per_chunk <- max(
      1L,
      as.integer(floor(target_entries_per_chunk / mean_per_row))
    )

    row_starts <- seq.int(1L, nr, by = rows_per_chunk)

    for (r1 in row_starts) {
      r2 <- min(nr, r1 + rows_per_chunk - 1L)
      p1 <- rowpointers[r1]
      p2 <- rowpointers[r2 + 1L] - 1L

      if (p1 > p2) {
        next
      }

      local_counts <- diff(rowpointers[r1:(r2 + 1L)])
      local_rows <- rep.int(r1:r2, local_counts)
      local_keep <- keep[p1:p2]

      if (!any(local_keep)) {
        next
      }

      local_rows <- local_rows[local_keep]
      local_cols <- colindices[p1:p2][local_keep]

      z <- add_to_raster(
        z,
        local_rows,
        local_cols,
        nr,
        nc
      )
    }

    draw_raster(z)
    draw_limits(nr, nc)

    invisible(list(
      nrow = nr,
      ncol = nc,
      nonzero = nnz,
      mode = "raster",
      raster_dimension = dim(z)
    ))
  }

  ##########################################################################
  ## Display an ordinary dense matrix
  ##########################################################################

  display_dense_matrix <- function(x) {
    nr <- nrow(x)
    nc <- ncol(x)

    if (any(!is.finite(x))) {
      stop("The covariance matrix contains non-finite values.")
    }

    total_cells <- as.double(nr) * as.double(nc)

    ## Direct image for moderate matrices.
    if (total_cells <= max_raster_cells) {
      nonzero <- abs(x) > tol
      nnz <- sum(nonzero)

      if (as.double(nnz) == total_cells) {
        message("No image plot: the covariance matrix is dense.")
        return(invisible(list(
          nrow = nr,
          ncol = nc,
          nonzero = nnz,
          mode = "dense"
        )))
      }

      old_par <- graphics::par(no.readonly = TRUE)
      on.exit(graphics::par(old_par), add = TRUE)

      ## Transpose after reversing rows to preserve the conventional matrix
      ## display with row 1 at the top, as in the original implementation.
      z <- t(nonzero[nr:1L, , drop = FALSE])
      storage.mode(z) <- "integer"

      draw_raster(z)
      draw_limits(nr, nc)

      return(invisible(list(
        nrow = nr,
        ncol = nc,
        nonzero = nnz,
        mode = "direct"
      )))
    }

    ## A large dense input is already resident in memory. To avoid allocating
    ## an additional nr x nc logical matrix, scan it in row blocks and build a
    ## bounded display raster only.
    raster_dims <- make_raster_dimensions(
      nr,
      nc,
      max_raster_cells
    )

    z <- matrix(
      0L,
      nrow = raster_dims[["nx"]],
      ncol = raster_dims[["ny"]]
    )

    nnz <- 0
    rows_per_chunk <- max(1L, min(nr, 1000L))
    row_starts <- seq.int(1L, nr, by = rows_per_chunk)

    for (r1 in row_starts) {
      r2 <- min(nr, r1 + rows_per_chunk - 1L)
      block <- x[r1:r2, , drop = FALSE]
      idx <- which(abs(block) > tol, arr.ind = TRUE)

      if (!nrow(idx)) {
        next
      }

      rows <- idx[, 1L] + r1 - 1L
      cols <- idx[, 2L]
      nnz <- nnz + nrow(idx)

      z <- add_to_raster(z, rows, cols, nr, nc)
    }

    if (as.double(nnz) == total_cells) {
      message("No image plot: the covariance matrix is dense.")
      return(invisible(list(
        nrow = nr,
        ncol = nc,
        nonzero = nnz,
        mode = "dense"
      )))
    }

    old_par <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(old_par), add = TRUE)

    draw_raster(z)
    draw_limits(nr, nc)

    invisible(list(
      nrow = nr,
      ncol = nc,
      nonzero = nnz,
      mode = "raster",
      raster_dimension = dim(z)
    ))
  }

  ##########################################################################
  ## Dispatch
  ##########################################################################

  if (is_spam) {
    return(display_spam(covmat))
  }

  display_dense_matrix(covmat)
}

GeoTestIsotropy <- function(data, coordx,
                            start, fixed,
                            optimizer = "bobyqa",
                            model = "Gaussian",
                            corrmodel = "Matern",
                            lower = NULL, upper = NULL,
                            B = 1000,
                            likelihood = NULL,
                            type = NULL,
                            copula = NULL,
                            neighb = 5,
                            method = c("cholesky", "TB"),
                            L = 10000,
                            n = 1,
                            angle_starts = c(pi / 4, 3 * pi / 4),
                            batch_size = NULL,
                            max_attempts = NULL,
                            parallel = FALSE,
                            ncores = 6,
                            progress = TRUE,
                            seed = NULL) {

  `%||%` <- function(x, y) if (!is.null(x)) x else y

  ## ====== preserve runtime settings and optional RNG seed ======
  ## Do not use .GeoTestPreserveRuntime() here.  That helper inspects and
  ## restores global progressr handlers, which is unsafe when this function
  ## is called inside tryCatch()/withCallingHandlers() and can trigger
  ## "should not be called with handlers on the stack" even when
  ## progress = FALSE.  GeoTestIsotropy manages its own progress state below.
  old_future_plan <- if (requireNamespace("future", quietly = TRUE)) future::plan() else NULL
  on.exit({
    if (!is.null(old_future_plan)) try(future::plan(old_future_plan), silent = TRUE)
    gc(verbose = FALSE, full = TRUE)
  }, add = TRUE)
  restore_seed <- .GeoTestSetSeed(seed)
  on.exit(restore_seed(), add = TRUE)

  ## ====== INITIAL INPUT VALIDATION ======
  method <- match.arg(method)

  if (!is.matrix(coordx)) coordx <- as.matrix(coordx)
  if (!is.numeric(data)) data <- as.numeric(data)

  if (ncol(coordx) != 2L) {
    stop("GeoTestIsotropy currently requires two-dimensional coordinates", call. = FALSE)
  }
  if (any(!is.finite(coordx))) {
    stop("coordx contains non-finite values", call. = FALSE)
  }
  if (any(!is.finite(data))) {
    stop("data contains non-finite values", call. = FALSE)
  }
  if (nrow(coordx) != length(data)) {
    stop(
      "coordx rows (", nrow(coordx), ") must match data length (",
      length(data), ")",
      call. = FALSE
    )
  }

  if (!is.list(start) || is.null(names(start)) || any(names(start) == "")) {
    stop("start must be a named list", call. = FALSE)
  }
  if (is.null(fixed)) fixed <- list()
  if (!is.list(fixed) || (length(fixed) > 0L &&
                          (is.null(names(fixed)) || any(names(fixed) == "")))) {
    stop("fixed must be a named list", call. = FALSE)
  }

  duplicated_parameters <- intersect(names(start), names(fixed))
  if (length(duplicated_parameters) > 0L) {
    stop(
      "Parameters cannot be present in both start and fixed: ",
      paste(duplicated_parameters, collapse = ", "),
      call. = FALSE
    )
  }

  if (!is.numeric(B) || length(B) != 1L || !is.finite(B) || B < 1 ||
      abs(B - round(B)) > sqrt(.Machine$double.eps)) {
    stop("B must be a positive integer", call. = FALSE)
  }
  B <- as.integer(round(B))

  if (!is.numeric(L) || length(L) != 1L || !is.finite(L) || L < 1 ||
      abs(L - round(L)) > sqrt(.Machine$double.eps)) {
    stop("L must be a positive integer", call. = FALSE)
  }
  L <- as.integer(round(L))

  if (!is.logical(parallel) || length(parallel) != 1L) {
    stop("parallel must be logical", call. = FALSE)
  }
  if (!is.logical(progress) || length(progress) != 1L) {
    stop("progress must be logical", call. = FALSE)
  }

  if (!is.numeric(angle_starts) || length(angle_starts) < 1L ||
      any(!is.finite(angle_starts))) {
    stop("angle_starts must contain at least one finite numeric angle", call. = FALSE)
  }
  angle_starts <- as.numeric(angle_starts) %% pi
  angle_starts <- angle_starts[!duplicated(round(angle_starts, 12L))]

  NN <- nrow(coordx)

  model_canonical <- .GeoSimulationCanonicalModel(model)
  n_used <- .GeoSimulationEffectiveN(model, model_canonical, n)
  if (model_canonical %in% c("Binomial", "BinomialLogistic", "BinomialNeg", "BinomialNegZINB")) {
    n_used <- .GeoValidateIntegerVector(n_used, "n", nobs = NN, minimum = 1L)
  }

  ## ====== AUTOMATIC LIKELIHOOD SELECTION ======
  if (is.null(likelihood) && is.null(type)) {
    if (NN > 2000) {
      likelihood <- "Marginal"
      type <- "Pairwise"
    } else {
      likelihood <- "Full"
      type <- "Standard"
      neighb <- NULL
    }
  } else if (xor(is.null(likelihood), is.null(type))) {
    stop(
      "Either specify both 'likelihood' and 'type', or leave both NULL for automatic selection",
      call. = FALSE
    )
  }

  if (identical(type, "Pairwise")) {
    if (is.null(neighb) || !is.numeric(neighb) || length(neighb) != 1L ||
        !is.finite(neighb) || neighb < 1) {
      stop("neighb must be a positive integer for pairwise likelihood", call. = FALSE)
    }
    neighb <- as.integer(round(neighb))
  } else {
    neighb <- NULL
  }

  if (!is.null(copula) && identical(method, "TB")) {
    .GeoValidateTBDistance("Eucl", context = "GeoTestIsotropy")
    .GeoValidateTBCorrelationModel(
      .GeoSimulationCanonicalCorrModel(corrmodel),
      bivariate = FALSE
    )
  }

  ## ====== ROBUST CORE SELECTION ======
  core_cfg <- .GeoTestCoreConfig(parallel, ncores, B, strict_positive = FALSE)
  parallel <- core_cfg$parallel
  ncores <- core_cfg$ncores

  ## Use the same progress-bar machinery adopted by the other parallel
  ## GeoModels functions (e.g. GeoNeighbSelect and GeoKrigloc).  In
  ## particular, install one temporary txtprogressbar handler and keep a
  ## single progressor owned by the main R process.  Do not use
  ## .GeoTestConfigureProgress() here: changing global progressr handlers
  ## from inside nested calling handlers can trigger
  ## "should not be called with handlers on the stack".
  use_progressr <- isTRUE(progress) && requireNamespace("progressr", quietly = TRUE)
  if (isTRUE(progress) && !use_progressr) {
    warning(
      "progress=TRUE but progressr is not available; progress bar disabled.",
      call. = FALSE
    )
  }
  if (use_progressr) {
    restore_progress_handlers <- .GeoProgressHandlersPush(
      handler = "txtprogressbar", global = TRUE
    )
    on.exit(restore_progress_handlers(), add = TRUE)
  }

  if (is.null(batch_size)) {
    batch_size <- if (parallel) max(2L, 2L * ncores) else min(10L, B)
  }
  if (!is.numeric(batch_size) || length(batch_size) != 1L ||
      !is.finite(batch_size) || batch_size < 1 ||
      abs(batch_size - round(batch_size)) > sqrt(.Machine$double.eps)) {
    stop("batch_size must be NULL or a positive integer", call. = FALSE)
  }
  batch_size <- min(B, as.integer(round(batch_size)))

  if (is.null(max_attempts)) {
    max_attempts <- max(B + 20L, as.integer(ceiling(1.5 * B)))
  }
  if (!is.numeric(max_attempts) || length(max_attempts) != 1L ||
      !is.finite(max_attempts) || max_attempts < B ||
      abs(max_attempts - round(max_attempts)) > sqrt(.Machine$double.eps)) {
    stop("max_attempts must be NULL or an integer greater than or equal to B", call. = FALSE)
  }
  max_attempts <- as.integer(round(max_attempts))

  corrmodel_used <- corrmodel
  model_used <- model

  ## ====== PARAMETER VALIDATION ======
  corr_names <- CorrParam(corrmodel_used)
  nuis_names <- NuisParam(model_used)
  expected <- sort(unique(c(corr_names, nuis_names)))
  provided <- sort(c(names(start), names(fixed)))

  if (!identical(expected, provided)) {
    missing_parameters <- setdiff(expected, provided)
    extra_parameters <- setdiff(provided, expected)
    msg <- character(0)

    if (length(missing_parameters) > 0L) {
      msg <- c(msg, paste0("missing: ", paste(missing_parameters, collapse = ", ")))
    }
    if (length(extra_parameters) > 0L) {
      msg <- c(msg, paste0("extra: ", paste(extra_parameters, collapse = ", ")))
    }

    stop(
      "Parameter names in start+fixed must match CorrParam(corrmodel)+NuisParam(model). ",
      paste(msg, collapse = " | "),
      call. = FALSE
    )
  }

  ## Bounds apply only to parameters in start.
  start_names <- sort(names(start))

  norm_bounds <- function(b, which = c("lower", "upper")) {
    .GeoTestNormalizeBounds(b, which, start_names, require_list = FALSE)
  }

  lower <- norm_bounds(lower, "lower")
  upper <- norm_bounds(upper, "upper")

  lower_vec <- unlist(lower, use.names = TRUE)
  upper_vec <- unlist(upper, use.names = TRUE)
  start_vec <- unlist(start, use.names = TRUE)[names(lower_vec)]

  if (any(lower_vec >= upper_vec)) {
    stop("Each lower bound must be smaller than its upper bound", call. = FALSE)
  }
  if (any(start_vec < lower_vec | start_vec > upper_vec)) {
    stop("Starting values must lie within the specified bounds", call. = FALSE)
  }

  ## ====== ANISOTROPY AND FIT HELPERS ======
  aniso_H0 <- list(angle = 0, ratio = 1)
  aniso_ratio_start <- 1.2

  loglik_value <- .GeoTestLogLikValue

  strip_anisotropy <- function(par) {
    par <- as.list(par)
    par[setdiff(names(par), c("angle", "ratio"))]
  }

  fit_H0 <- function(Z, start_values = start) {
    GeoFit(
      data = Z,
      coordx = coordx,
      corrmodel = corrmodel_used,
      likelihood = likelihood,
      optimizer = optimizer,
      model = model_used,
      type = type,
      neighb = neighb,
      start = start_values,
      fixed = fixed,
      copula = copula,
      lower = lower,
      upper = upper,
      n = n_used,
      anisopars = aniso_H0
    )
  }

  fit_H1_once <- function(Z, start_values = start,
                          anisopars_values = list(angle = 0, ratio = 1.2)) {
    GeoFit(
      data = Z,
      coordx = coordx,
      corrmodel = corrmodel_used,
      likelihood = likelihood,
      optimizer = optimizer,
      model = model_used,
      type = type,
      neighb = neighb,
      start = start_values,
      fixed = fixed,
      copula = copula,
      lower = lower,
      upper = upper,
      n = n_used,
      anisopars = anisopars_values,
      est.aniso = c(TRUE, TRUE)
    )
  }

  same_pairs <- function(fit0_obj, fit1_obj) {
    .GeoTestSamePairs(fit0_obj, fit1_obj, type, require_indices = TRUE)
  }

  fit_H1_multistart <- function(Z, fit0_obj, start_values = start,
                                angles = angle_starts,
                                ratio_start = aniso_ratio_start) {
    ll0 <- loglik_value(fit0_obj)
    if (!is.finite(ll0)) {
      return(list(
        fit = NULL,
        attempts = 0L,
        valid_attempts = 0L,
        error = "H0 fit has a non-finite likelihood"
      ))
    }

    ratio_start <- as.numeric(ratio_start)[1L]
    if (!is.finite(ratio_start) || ratio_start <= 1) ratio_start <- 1.2

    angles <- as.numeric(angles) %% pi
    angles <- angles[is.finite(angles)]
    angles <- angles[!duplicated(round(angles, 12L))]
    if (length(angles) < 1L) angles <- 0

    fits <- lapply(angles, function(angle_start) {
      suppressWarnings(try(
        fit_H1_once(
          Z,
          start_values = start_values,
          anisopars_values = list(angle = angle_start, ratio = ratio_start)
        ),
        silent = TRUE
      ))
    })

    ll_values <- vapply(fits, loglik_value, numeric(1))
    valid <- is.finite(ll_values)

    if (!any(valid)) {
      return(list(
        fit = NULL,
        attempts = length(fits),
        valid_attempts = 0L,
        error = "All anisotropic multi-start fits failed"
      ))
    }

    valid_indices <- which(valid)
    best_index <- valid_indices[which.max(ll_values[valid])]
    best_fit <- fits[[best_index]]

    if (!same_pairs(fit0_obj, best_fit)) {
      return(list(
        fit = NULL,
        attempts = length(fits),
        valid_attempts = sum(valid),
        error = "H0 and H1 do not use the same pair set"
      ))
    }

    list(
      fit = best_fit,
      attempts = length(fits),
      valid_attempts = sum(valid),
      starting_angle = angles[best_index],
      error = NULL
    )
  }

  ## ====== OBSERVED FITS ======
  fit0 <- fit_H0(data)
  if (!is.finite(loglik_value(fit0))) {
    stop("Observed isotropic fit returned a non-finite log-likelihood", call. = FALSE)
  }

  fit1_result <- fit_H1_multistart(
    data,
    fit0_obj = fit0,
    start_values = start,
    angles = angle_starts,
    ratio_start = aniso_ratio_start
  )

  if (is.null(fit1_result$fit)) {
    stop("Observed anisotropic fit failed: ", fit1_result$error, call. = FALSE)
  }

  fit1 <- fit1_result$fit
  LL0 <- loglik_value(fit0)
  LL1 <- loglik_value(fit1)

  if (!is.finite(LL0) || !is.finite(LL1)) {
    stop("Observed H0/H1 fit returned non-finite log-likelihood values", call. = FALSE)
  }
  if (!same_pairs(fit0, fit1)) {
    stop(
      "Observed H0 and H1 fits must use the same pairs for the composite-likelihood ratio",
      call. = FALSE
    )
  }

  tol_obs <- 1e-8 * max(1, abs(LL0))
  observed_boundary_selected <- LL1 < LL0 - tol_obs
  LRT_obs <- if (observed_boundary_selected) 0 else max(0, 2 * (LL1 - LL0))

  ratio_fit <- as.numeric(fit1$anisopars$ratio %||% NA_real_)
  angle_fit <- as.numeric(fit1$anisopars$angle %||% NA_real_)
  ratio_hat <- if (observed_boundary_selected) 1 else ratio_fit
  angle_hat <- if (observed_boundary_selected) NA_real_ else angle_fit

  if (!is.finite(ratio_hat) || ratio_hat < 1) {
    stop("Observed anisotropy-ratio estimate is invalid", call. = FALSE)
  }

  ## Use observed estimates as bootstrap starting values. Angle and ratio
  ## remain in anisopars and are therefore removed from start_H1_boot.
  start_H0_boot <- strip_anisotropy(fit0$param)
  start_H1_boot <- strip_anisotropy(fit1$param)

  ## Use the same H1 multi-start angles for the observed fit and every
  ## bootstrap refit. Keeping this set fixed makes the numerical search
  ## comparable across the observed and bootstrap samples and controls cost.
  bootstrap_angle_starts <- angle_starts
  bootstrap_ratio_start <- if (is.finite(ratio_fit) && ratio_fit > 1) ratio_fit else 1.2

  ## ====== NULL MODEL FOR PARAMETRIC BOOTSTRAP ======
  par0 <- c(fit0$param, fit0$fixed)
  param_H0 <- as.list(par0[expected])

  if (is.null(copula)) {
    latent_check <- suppressWarnings(try(
      .GeoValidateLatentIntegerParameters(model_canonical, param_H0),
      silent = TRUE
    ))
    if (inherits(latent_check, "try-error")) {
      cond <- attr(latent_check, "condition")
      msg <- if (!is.null(cond)) conditionMessage(cond) else as.character(latent_check)
      extra <- if (identical(model_canonical, "Gamma")) {
        paste0(
          " For the direct Gamma latent-squares model used in the paper, ",
          "fix 'shape' to a positive integer (put it in 'fixed') so the fitted ",
          "null model can be simulated exactly by the latent-Gaussian construction."
        )
      } else {
        ""
      }
      stop(
        "The fitted null model cannot be simulated by the direct latent-Gaussian bootstrap: ",
        msg, extra,
        call. = FALSE
      )
    }
  }

  simulate_null <- function(nrep, workers = ncores) {
    if (is.null(copula)) {
      if (method == "cholesky") {
        ## GeoSim() currently prints an informational line for Cholesky
        ## simulation even when progress=FALSE. Capture that output here so
        ## GeoTestIsotropy() can present a single bootstrap progress bar.
        sim_out <- NULL
        invisible(utils::capture.output(
          sim_out <- GeoSim(
            coordx = coordx,
            corrmodel = corrmodel_used,
            param = param_H0,
            model = model_used,
            n = n_used,
            nrep = nrep,
            anisopars = aniso_H0,
            progress = FALSE
          ),
          type = "output"
        ))
        sim_out
      } else {
        plan_before_sim <- future::plan()
        on.exit(try(future::plan(plan_before_sim), silent = TRUE), add = TRUE)
        GeoSimapprox(
          coordx = coordx,
          corrmodel = corrmodel_used,
          model = model_used,
          method = "TB",
          L = L,
          n = n_used,
          param = param_H0,
          nrep = nrep,
          anisopars = aniso_H0,
          progress = FALSE,
          parallel = parallel && workers > 1L,
          ncores = workers
        )
      }
    } else {
      GeoSimCopula(
        coordx = coordx,
        corrmodel = corrmodel_used,
        method = method,
        param = param_H0,
        model = model_used,
        n = n_used,
        copula = copula,
        nrep = nrep,
        anisopars = aniso_H0,
        L = L,
        parallel = parallel && workers > 1L,
        ncores = workers,
        progress = FALSE
      )
    }
  }

  bootstrap_one <- function(Zb) {
    f0b <- suppressWarnings(try(
      fit_H0(Zb, start_values = start_H0_boot),
      silent = TRUE
    ))

    if (inherits(f0b, "try-error") || !is.finite(loglik_value(f0b))) {
      return(list(
        valid = FALSE,
        lrt = NA_real_,
        ratio = NA_real_,
        angle = NA_real_,
        boundary_selected = FALSE,
        h1_attempts = 0L,
        h1_valid_attempts = 0L,
        error = "H0 fit failed"
      ))
    }

    f1_result <- fit_H1_multistart(
      Zb,
      fit0_obj = f0b,
      start_values = start_H1_boot,
      angles = bootstrap_angle_starts,
      ratio_start = bootstrap_ratio_start
    )

    if (is.null(f1_result$fit)) {
      return(list(
        valid = FALSE,
        lrt = NA_real_,
        ratio = NA_real_,
        angle = NA_real_,
        boundary_selected = FALSE,
        h1_attempts = f1_result$attempts,
        h1_valid_attempts = f1_result$valid_attempts,
        error = f1_result$error
      ))
    }

    ll0b <- loglik_value(f0b)
    ll1b <- loglik_value(f1_result$fit)

    if (!is.finite(ll0b) || !is.finite(ll1b)) {
      return(list(
        valid = FALSE,
        lrt = NA_real_,
        ratio = NA_real_,
        angle = NA_real_,
        boundary_selected = FALSE,
        h1_attempts = f1_result$attempts,
        h1_valid_attempts = f1_result$valid_attempts,
        error = "Non-finite bootstrap likelihood"
      ))
    }

    tol_b <- 1e-8 * max(1, abs(ll0b))
    boundary_selected <- ll1b < ll0b - tol_b
    ratio_b <- if (boundary_selected) {
      1
    } else {
      as.numeric(f1_result$fit$anisopars$ratio %||% NA_real_)
    }
    angle_b <- if (boundary_selected) {
      NA_real_
    } else {
      as.numeric(f1_result$fit$anisopars$angle %||% NA_real_)
    }

    if (!is.finite(ratio_b) || ratio_b < 1) {
      return(list(
        valid = FALSE,
        lrt = NA_real_,
        ratio = NA_real_,
        angle = NA_real_,
        boundary_selected = boundary_selected,
        h1_attempts = f1_result$attempts,
        h1_valid_attempts = f1_result$valid_attempts,
        error = "Invalid bootstrap anisotropy-ratio estimate"
      ))
    }

    list(
      valid = TRUE,
      lrt = if (boundary_selected) 0 else max(0, 2 * (ll1b - ll0b)),
      ratio = ratio_b,
      angle = angle_b,
      boundary_selected = boundary_selected,
      h1_attempts = f1_result$attempts,
      h1_valid_attempts = f1_result$valid_attempts,
      error = NULL
    )
  }

  process_batch <- function(sim, batch_n, batch_id) {
    ## Keep one package-standard progress bar for the whole bootstrap.  In the
    ## sequential branch the master updates it immediately after each valid
    ## refit.  In the parallel branch the master polls individual futures and
    ## updates it as each valid refit actually completes, avoiding buffered or
    ## batchwise jumps in the display.
    if (!parallel) {
      out <- vector("list", batch_n)
      for (b in seq_len(batch_n)) {
        out[[b]] <- bootstrap_one(.GeoTestGetRep(sim, b))
        if (!is.null(pb) && isTRUE(out[[b]]$valid) &&
            is.finite(out[[b]]$lrt) && is.finite(out[[b]]$ratio)) {
          pb(sprintf("b=%d", b))
        }
      }
      out
    } else {
      data_files <- .GeoTestSaveReplicates(
        sim, batch_n, sprintf("isotropy_batch%04d", batch_id)
      )
      on.exit(unlink(data_files), add = TRUE)
      rm(sim)
      gc(verbose = FALSE, full = TRUE)
      .GeoTestParallelFileApply(
        data_files = data_files,
        FUN = bootstrap_one,
        ncores = min(ncores, batch_n),
        progressor = pb,
        future_packages = "GeoModels"
      )
    }
  }

  ## ====== BOOTSTRAP: collect exactly B valid replications ======
  ## One progressor is created for the complete bootstrap, using the same
  ## txtprogressbar handler as the other parallel GeoModels functions.  The
  ## bar counts valid replications and is advanced by the main R process when
  ## each refit has genuinely completed.
  if (use_progressr) {
    pb <- progressr::progressor(
      steps = B, on_exit = FALSE, auto_finish = FALSE
    )
    on.exit(pb(type = "finish"), add = TRUE, after = FALSE)
  } else {
    pb <- NULL
  }

  run_bootstrap <- function() {
    valid_results <- vector("list", 0L)
    failure_records <- vector("list", 0L)
    total_attempts <- 0L
    batch_id <- 0L

    while (length(valid_results) < B && total_attempts < max_attempts) {
      outstanding <- B - length(valid_results)
      remaining_attempts <- max_attempts - total_attempts
      batch_n <- min(batch_size, outstanding, remaining_attempts)
      if (batch_n < 1L) break

      batch_id <- batch_id + 1L
      workers_batch <- min(ncores, batch_n)
      sim <- simulate_null(batch_n, workers = workers_batch)
      batch_results <- process_batch(sim, batch_n, batch_id)
      if (exists("sim", inherits = FALSE)) rm(sim)
      gc(verbose = FALSE, full = TRUE)

      for (j in seq_along(batch_results)) {
        total_attempts <- total_attempts + 1L
        ans <- batch_results[[j]]

        if (isTRUE(ans$valid) && is.finite(ans$lrt) && is.finite(ans$ratio)) {
          valid_results[[length(valid_results) + 1L]] <- ans
        } else {
          failure_records[[length(failure_records) + 1L]] <- data.frame(
            attempt = total_attempts,
            error = ans$error %||% "Unknown bootstrap failure",
            h1_attempts = as.integer(ans$h1_attempts %||% 0L),
            h1_valid_attempts = as.integer(ans$h1_valid_attempts %||% 0L),
            stringsAsFactors = FALSE
          )
        }
      }

    }

    list(
      valid_results = valid_results,
      failure_records = failure_records,
      total_attempts = total_attempts,
      batch_id = batch_id
    )
  }

  bootstrap_state <- run_bootstrap()

  valid_results <- bootstrap_state$valid_results
  failure_records <- bootstrap_state$failure_records
  total_attempts <- bootstrap_state$total_attempts
  batch_id <- bootstrap_state$batch_id

  n_valid <- length(valid_results)
  if (n_valid < B) {
    stop(
      "GeoTestIsotropy obtained only ", n_valid, " valid bootstrap replications after ",
      total_attempts, " attempts (requested B = ", B, "). Increase 'max_attempts', ",
      "improve starting values/bounds, or inspect optimization failures.",
      call. = FALSE
    )
  }

  LRT_boot <- vapply(valid_results, function(x) x$lrt, numeric(1))
  ratio_boot <- vapply(valid_results, function(x) x$ratio, numeric(1))
  angle_boot <- vapply(valid_results, function(x) x$angle, numeric(1))
  boundary_boot <- vapply(
    valid_results, function(x) isTRUE(x$boundary_selected), logical(1)
  )
  h1_attempts <- vapply(
    valid_results, function(x) as.integer(x$h1_attempts %||% 0L), integer(1)
  )
  h1_valid_attempts <- vapply(
    valid_results, function(x) as.integer(x$h1_valid_attempts %||% 0L), integer(1)
  )

  pval_lrt <- (1 + sum(LRT_boot >= LRT_obs)) / (B + 1)
  pval_ratio <- (1 + sum(ratio_boot >= ratio_hat)) / (B + 1)

  if (!is.null(pb)) pb(type = "finish")

  if (length(failure_records) > 0L) {
    failure_table <- do.call(rbind, failure_records)
    rownames(failure_table) <- NULL
  } else {
    failure_table <- data.frame(
      attempt = integer(0),
      error = character(0),
      h1_attempts = integer(0),
      h1_valid_attempts = integer(0),
      stringsAsFactors = FALSE
    )
  }

  ## ====== RESULTS ======
  cat("\n=== FINAL RESULTS ===\n")
  cat("Observed WCLRT/LRT statistic =", round(LRT_obs, 5), "\n")
  cat("Observed anisotropy ratio    =", round(ratio_hat, 5), "\n")
  cat("Valid bootstrap reps         =", B, "\n")
  cat("Total bootstrap attempts     =", total_attempts, "\n")
  cat("Failed attempts replaced     =", nrow(failure_table), "\n")
  cat("H1 multi-start angles        =", length(angle_starts), "\n")
  if (is.finite(angle_hat)) {
    cat("Estimated anisotropy angle   =", round(angle_hat, 5), "\n")
  } else {
    cat("Estimated anisotropy angle   = NA (boundary solution)\n")
  }
  cat("MWCLT p-value (ratio)        =", round(pval_ratio, 5), "\n")
  cat("WCLRT p-value                =", round(pval_lrt, 5), "\n\n")

  conclusion_ratio <- if (pval_ratio < 0.05) {
    "Reject H0 => significant anisotropy detected"
  } else {
    "Do not reject H0 => data consistent with isotropy"
  }
  conclusion_lrt <- if (pval_lrt < 0.05) {
    "Reject H0 => significant anisotropy detected"
  } else {
    "Do not reject H0 => data consistent with isotropy"
  }

  cat("MWCLT conclusion:", conclusion_ratio, "\n")
  cat("WCLRT conclusion:", conclusion_lrt, "\n\n")

  invisible(list(
    ## Backward-compatible primary result: WCLRT/LRT.
    statistic = LRT_obs,
    pvalue = pval_lrt,
    ## Explicit results for both tests described in the paper.
    statistic_lrt = LRT_obs,
    pvalue_lrt = pval_lrt,
    statistic_ratio = ratio_hat,
    pvalue_ratio = pval_ratio,
    pvalues = c(MWCLT = pval_ratio, WCLRT = pval_lrt),
    ratio_hat = ratio_hat,
    angle_hat = angle_hat,
    corrmodel = corrmodel_used,
    model = model_used,
    likelihood = likelihood,
    type = type,
    method = method,
    parallel = parallel,
    ncores = if (parallel) ncores else 1L,
    L = if (method == "TB") L else NULL,
    n = n_used,
    seed = seed,
    fit_H0 = fit0,
    fit_H1 = fit1,
    observed_boundary_selected = observed_boundary_selected,
    observed_multistart_angles = angle_starts,
    observed_H1_attempts = fit1_result$attempts,
    observed_H1_valid_attempts = fit1_result$valid_attempts,
    ## Backward-compatible alias for the LRT bootstrap statistics.
    B_rep = LRT_boot,
    B_rep_lrt = LRT_boot,
    B_rep_ratio = ratio_boot,
    B_rep_angle = angle_boot,
    bootstrap_B = B,
    bootstrap_successful = B,
    bootstrap_success_rate = 1,
    bootstrap_total_attempts = total_attempts,
    bootstrap_failed_attempts = nrow(failure_table),
    bootstrap_boundary_solutions = sum(boundary_boot),
    bootstrap_H1_attempts = h1_attempts,
    bootstrap_H1_valid_attempts = h1_valid_attempts,
    bootstrap_failures = failure_table,
    batch_size = batch_size,
    max_attempts = max_attempts,
    ## Retained for compatibility with objects produced by earlier versions.
    observed_fallback_used = FALSE,
    bootstrap_fallbacks = 0L,
    bootstrap_attempts = h1_attempts
  ))
}

####################################################
### File name: Utility.r
####################################################
# Check if the correlation is bivariate
CheckBiv <- function(numbermodel)
{
    CheckBiv <- NULL
    if(!is.null(numbermodel)){
    if(numbermodel >= 101 & numbermodel <= 140) CheckBiv <- TRUE
    else CheckBiv <- FALSE
    }
    return(CheckBiv)
}
# Check if the correlation is spatial or spatial-temporal
CheckST <- function(numbermodel)
{
    CheckST <- NULL
     if(!is.null(numbermodel)){
    if(numbermodel >40 & numbermodel <= 100) CheckST <- TRUE
    else  CheckST <- FALSE
     }
    return(CheckST)
}


# Check the type of distances
CheckDistance <- function(distance) {
  out <- switch(distance,
                eucl = 0, Eucl = 0, 
                chor = 1, Chor = 1, 
                geod = 2, Geod = 2,
                stop("Invalid type of  di distance: use 'Eucl', 'Chor' o 'Geod'."))
  return(out)
}
### Procedures are in alphabetical order.
CkCorrModel <- function(corrmodel)
  {
    CkCorrModel <- NULL
    # Correlation function are in alphabetical order
    CkCorrModel <- switch(corrmodel,
            #spatial models
                             cauchy=1,Cauchy=1,
                             Matern1=2,
                             Matern2=3,
                             exponential=4,Exponential=4,Exp=4,exp=4,Matern0=4,
                             dagum = 5, Dagum = 5,
                             GenWend_Matern=6,Genwend_Matern=6,
                             GenWend_Matern2=7,Genwend_Matern2=7,
                             Gencauchy=8,GenCauchy=8,
                             Shkarofski=10,shkarofski=10,
                             Wend0=11,wend0=11,
                             stable=12,Stable=12,
                             Wend1=13,wend1=13,
                             matern=14,Matern=14,
                             Wend2=15,wend2=15,
                             wave=16,Wave=16,
                             Multiquadric=17,multiquadric=17,
                             Sinpower=18,sinpower=18,
                             Genwend=19,GenWend=19,
                             F_sphere=20,F_Sphere=20,
                             Hypergeometric2=21,HyperGeometric2=21, hypergeometric2=21,
                             Hypergeometric=22,HyperGeometric=22, hypergeometric=22,
                             Hypergeometric_Matern=23,HyperGeometric_Matern=23, hypergeometric_Matern=23,
                             Hypergeometric_Matern2=30,Hypergeometric_matern2=30, hypergeometric_Matern2=30,
                             Kummer=24,Kummer=24,
                             Kummer_Matern=25,Kummer_matern=25,
                             GenWend_Hole=26,GenWend_hole=26,
                             Matern_Hole=27,Matern_hole=27,
                             Schoenberg=28,schoenberg=28,
                             GenWend_Matern_Hole=29,GenWend_Matern_hole=29,
                             Gaussian=31,
                             Spherical=32,
             # spatial-temporal non-separable models
                             gneiting=42,Gneiting=42,  #ok
                             iacocesare=44,Iacocesare=44, #ok
                             porcu=46,Porcu=46,
                             stein=48,Stein=48,          #ok
                             porcu1=50,Porcu1=50,
                             gneiting_GC=52,Gneiting_GC=52, #ok
                             gneiting_GC2=54,Gneiting_GC2=54,
                             sinpower_st=56,Sinpower_st=56,    #ok
                             multiquadric_st=58,Multiquadric_st=58,   #ok
                             gneiting_mat_T=61,Gneiting_mat_T=61, #ok
                             gneiting_mat_S=62,Gneiting_mat_S=62,Gneiting_Mat_S=62, #ok
                             Wen0_space=63,wen0_space=63,  #ok
                             Wen0_time=64,wen0_time=64,    #ok
                             Wen1_space=65,wen1_space=65,  #ok
                             Wen1_time=66,wen1_time=66,    #ok
                             Wen2_space=67,wen2_space=67,  #ok
                             Wen2_time=68,wen2_time=68,    #ok
                             Matern_Matern_sep=86,
                             Gneiting_wen_S=87,gneiting_wen_S=87,Gneiting_Wen_S=87,
                             Gneiting_wen_T=88,gneiting_wen_T=88, #ok
                             Matern_Matern_nosep=89,
                             Genwend_Genwend_nosep=85,
              # spatial-temporal separable models
                             Wend0_Wend0=69,wend0_wend0=69, #ok
                             Wend0_Wend1=70,wend0_wend1=70, #ok
                             Wend0_Wend2=71,wend0_wend2=71, #ok
                             Wend1_Wend0=72,wend1_wend0=72, #ok
                             Wend1_Wend1=73,wend1_wend1=73, #ok
                             Wend1_Wend2=74,wend1_wend2=74, #ok
                             Wend2_Wend0=75,wend2_wend0=75, #ok
                             Wend2_Wend1=76,wend2_wend1=76, #ok
                             Wend2_Wend2=77,wend2_wend2=77, #ok 
                             GenWend_GenWend=78,Genwend_Genwend=78, #ok 
                             exp_cauchy=82,Exp_Cauchy=82,
                             exp_exp=84,Exp_Exp=84,
                             Matern_Matern=86, matern_matern=86,  #ok
                             Matern_Matern_temp=95,# new model
                             stable_stable=94,Stable_Stable=94,
                             prove=96,
              # Bivariate models
                             Bi_wend0_sep=111,Bi_Wend0_sep=111,
                             Bi_Wend0_contr=129,Bi_wend0_contr=129,
                             Bi_wend0=112,Bi_Wend0=112,
                             Bi_F_sphere=117,Bi_F_sphere=117,
                              Bi_F_sphere_sep=119,Bi_F_sphere_sep=119,
                              Bi_F_sphere_contr=121,Bi_F_sphere_contr=121,
                             Bi_wend1_sep=113,Bi_Wend1_sep=113,
                             Bi_Wend1_contr=131,Bi_wend1_contr=131,
                             Bi_wend1=114,Bi_Wend1=114,  
                             Bi_wend2_sep=115,Bi_Wend2_sep=115,
                             Bi_wend2=116,Bi_Wend2=116,
                             Bi_Wend2_contr=120,Bi_wend2_contr=120,
                             Bi_matern_contr=118,Bi_Matern_contr=118, 
                             Bi_matern_sep=122,  Bi_Matern_sep=122, 
                             Bi_matern=128,Bi_Matern=128,
                             Bi_LMC_contr=124,
                             Bi_LMC=126,
                             Bi_GenWend_sep=130,Bi_genWend_sep=130,
                             Bi_GenWend_contr=134,Bi_genWend_contr=134,
                             Bi_GenWend=132,Bi_genWend=132,
                             Bi_Matern_Cauchy=136,  
                             Bi_GenMatern_Cauchy=137,   Bi_genMatern_Cauchy=137,
                             Bi_genCauchy=138,Bi_GenCauchy=138,
                             Bi_Stable=139,Bi_stable=139,   
            ########Tapers
                   ##spatial
                             Bohman=28,
                            # Wendland0=30,wendland0=30,
                            # Wendland1=32,wendland1=32,
                            # Wendland2=34, wendland2=34,
                             unit_matrix=36,
                            
                   ##bivariate
                             Bi_Wendland1=140,Bi_wendland1=140,
                             Bi_Wendland2=142,Bi_wendland2=142,
                             Bi_Wendland3=144,Bi_wendland3=144,
                             Bi_wendland_asy=146,Bi_wendland_asy=146,
                             unit_matrix_biv=147,
                   ##spacetime separable
                             Wendland0_Wendland0=200,
                             Wendland0_Wendland1=202,
                             Wendland0_Wendland2=204,
                             Wendland1_Wendland0=206,
                             Wendland1_Wendland1=208,
                             Wendland1_Wendland2=210,
                             Wendland2_Wendland0=212,
                             Wendland2_Wendland1=214,
                             Wendland2_Wendland2=216,
                    ##spacetime no separable
                             Wendland0_time=218,
                             Wendland1_time=220,
                             Wendland2_time=222,
                             Wendland0_space=224,
                             Wendland1_space=226,
                             Wendland2_space=228,
                             unit_matrix_st=230)
    return(CkCorrModel)
  }



# checking models valid only on the sphere
CheckSph<- function(numbermodel)
  {
    Check <- FALSE
    if(numbermodel %in% c(17,18,56,58,20,117))  Check=TRUE    
    return(Check)
  }


CkInput <- function(coordx, coordy, coordz,coordt, coordx_dyn, corrmodel, data, distance, fcall, fixed, grid,
                      likelihood, maxdist, maxtime,  model, n,  optimizer, param,
                       radius, start, taper, tapsep, type, varest, weighted,
                       copula,X)
  {
    error <- NULL
    replicates=1


    # START Include internal functions:
    CheckParamRange <- function(param)
    { 
        ## An empty parameter set is valid: there is simply nothing to check.
        ## This is required when GeoFit/GeoFit2 use automatic initialization
        ## (start = NULL, normalized internally to an empty list).
        if (is.null(param) || length(param) == 0L) return(TRUE)

      #  if(!is.na(param['df'])) if(param['df'] > 1/2 || param['df'] < 0 ) return(FALSE)
        #if(!is.na(param['tail'])) if(param['tail'] >0.5) return(FALSE)
        if(!is.na(param['shape'])) if(param['shape'] <0) return(FALSE)
        if(!is.na(param['nugget'])) if(param['nugget'] < 0||param['nugget'] >= 1) return(FALSE)
        if(!is.na(param['nugget_1'])) if(param['nugget_1'] < 0) return(FALSE)
        if(!is.na(param['nugget_2'])) if(param['nugget_2'] < 0) return(FALSE)
        if(!is.na(param['power'])) if(param['power'] <=0 || param['power'] > 2) return(FALSE)
        if(!is.na(param['power_s'])) if(param['power_s'] <=0 || param['power_s'] > 2) return(FALSE)
        if(!is.na(param['power_t'])) if(param['power_t'] <=0 || param['power_t'] > 2) return(FALSE)
       # if(!is.na(param['power1'])) if(param['power1'] <=0 || param['power1'] > 2) return(FALSE)
        if(!is.na(param['power2'])) if(param['power2'] <= 0) return(FALSE)
        if(!is.na(param['power2_1'])) if(param['power2_1'] <= 0) return(FALSE)
        if(!is.na(param['power2_12'])) if(param['power2_12'] <= 0) return(FALSE)
        if(!is.na(param['power2_2'])) if(param['power2_2'] <= 0) return(FALSE)
        #if(!is.na(param['sep'])) if(param['sep'] < 0 || param['sep'] > 1) return(FALSE)
        if(!is.na(param['scale'])) if(param['scale'] <= 0) return(FALSE)
        if(!is.na(param['scale_s'])) if(param['scale_s'] <= 0) return(FALSE)
        if(!is.na(param['scale_t'])) if(param['scale_t'] <= 0) return(FALSE)
        if(!is.na(param['scale_1']))   if(param['scale_1'] < 0) return(FALSE)
        if(!is.na(param['scale_2']))   if(param['scale_2'] < 0) return(FALSE)
        if(!is.na(param['scale_12']))   if(param['scale_12'] <= 0) return(FALSE)
        if(!is.na(param['sill'])) if(param['sill'] <= 0) return(FALSE)
        if(!is.na(param['sill_1'])) if(param['sill_1'] <= 0) return(FALSE)
        if(!is.na(param['sill_2'])) if(param['sill_2'] <= 0) return(FALSE)
        #if(!is.na(param['smooth'])) if(param['smooth'] <= 0) return(FALSE)
        if(!is.na(param['smooth_s'])) if(param['smooth_s'] < 0) return(FALSE)
        if(!is.na(param['smooth_t'])) if(param['smooth_t'] < 0) return(FALSE)
        #if(!is.na(param['smooth_1'])) if(param['smooth_1'] < 0) return(FALSE)
        if(!is.na(param['smooth_2'])) if(param['smooth_2'] < 0) return(FALSE)
        #if(!is.na(param['smooth_12'])) if(param['smooth_12'] < 0) return(FALSE)
        if(!is.na(param['pcol'])) if(param['pcol'] > 1 || param['pcol'] < -1  ) return(FALSE)
        return(TRUE)  
    }
   
    bivariate=CheckBiv(CkCorrModel(corrmodel))

    dimension_error <- tryCatch({
      .GeoValidate3D(
        coordx = coordx, coordy = coordy, coordz = coordz,
        coordx_dyn = coordx_dyn, distance = distance, grid = grid,
        context = if (is.null(fcall)) "GeoModels" else as.character(fcall)
      )
      NULL
    }, error = function(e) conditionMessage(e))
    if (!is.null(dimension_error)) {
      return(list(error = paste0(dimension_error, "\n")))
    }
    if(!bivariate)  if(length(coordt)>0&&is.list(X)) X=X[[1]]
    if(!bivariate) {if(is.null(X))  {X=1;num_betas=1} 
                    else num_betas=ncol(X)  }
    if( bivariate) {if(is.null(X))  {X=1;num_betas=c(1,1)} 
                    else { 
                               if(is.list(X))  num_betas=c(ncol(X[[1]]),ncol(X[[2]]))
                               else  num_betas=c(ncol(X),ncol(X))
                          } }
    #bivariate<-CheckBiv(CheckCorrModel(corrmodel))
    #spacetime<-CheckST(CheckCorrModel(corrmodel))
    #if(is.null(bivariate)&&is.null(bivariate))
    # END Include internal functions
    if(!is.null(copula)) {
       {if(copula!="Gaussian"&&copula!="Clayton"&&copula!="SkewGaussian") 
        error <- 'Copula model is wrong\n'
        return(list(error=error))}
       }

    ### check not kriging inputs   
    if(fcall!="Kriging"){

    ### START Checks inserted input
    # START common check fitting and simulation
    if(missing(coordx_dyn)){
        if(missing(coordx) || !is.numeric(coordx)){
        error <- 'insert a suitable set of numeric coordinates\n'
        return(list(error=error))}
        if(!is.null(coordy) & !is.numeric(coordy)){
        error <- 'insert a suitable set of numeric coordinates\n'
        return(list(error=error))}
          if(!is.null(coordz) & !is.numeric(coordz)){
        error <- 'insert a suitable set of numeric coordinates\n'
        return(list(error=error))}
    }
    if(missing(corrmodel) || !is.character(corrmodel)){
        error <- 'insert the correlation model\n'
        return(list(error=error))}
    if(!is.null(grid) & !is.logical(grid)){
        error <- 'the parameter grid need to be a logic value\n'
        return(list(error=error))}
    if(!is.null(model) & !is.character(model)){
        error <- 'insert the name of the random field\n'
        return(list(error=error))}
    if(is.null(CkModel(model))){
        error <- 'the model name of the random field is not correct\n'
        return(list(error=error))}
    if(is.null(replicates) || (abs(replicates-round(replicates))>0) || replicates<1){
        error <- 'the parameter replicates need to be a positive integer\n'
        return(list(error=error))}
    if(is.null(CheckDistance(distance))){
        error <- 'the name of the distance is not correct\n'
        return(list(error=error))}
    if(is.null(CkCorrModel(corrmodel))){
        error <- 'the name of the correlation model is not correct\n'
        return(list(error=error))}
    if(radius<0){
        error <- 'the radius of the sphere must be positive\n'
        return(list(error=error))}
    ## Count-field size semantics.
    ##
    ## Direct Binomial RFs allow site-specific n(s), as in the repeated-trial
    ## construction.  Direct Negative-Binomial RFs instead use one common
    ## positive integer r (stored in the public argument `n`) for the whole
    ## field; Geometric is the special case r = 1.  Gaussian-copula margins
    ## are a different construction and retain their existing marginal-size
    ## handling.
    model_code_n <- CkModel(model)
    binomial_n_models <- c(11L, 15L, 19L, 31L, 49L, 51L)
    negbin_common_r_models <- c(16L, 45L, 52L, 54L, 56L)
    if(model_code_n %in% c(2L, binomial_n_models, 14L, negbin_common_r_models)) {
      bad_n <- is.null(n) || !is.numeric(n) || !length(n) ||
               any(!is.finite(n)) || any(n < 1) ||
               any(abs(n - round(n)) > sqrt(.Machine$double.eps))

      if(!bad_n && is.null(copula) && model_code_n == 2L) {
        bad_n <- length(n) != 1L || as.integer(round(n[1L])) != 1L
        if(bad_n) {
          error <- 'for the direct Binary/Bernoulli random field n is fixed to 1\n'
          return(list(error=error))
        }
      }

      if(!bad_n && is.null(copula) && model_code_n == 14L) {
        bad_n <- length(n) != 1L || as.integer(round(n[1L])) != 1L
        if(bad_n) {
          error <- 'for the direct Geometric random field n is fixed to 1 (the r = 1 Negative-Binomial case)\n'
          return(list(error=error))
        }
      }

      if(!bad_n && is.null(copula) && model_code_n %in% negbin_common_r_models) {
        bad_n <- length(n) != 1L
        if(bad_n) {
          error <- 'for direct Negative-Binomial random fields n is the common number r of successes and must be one positive integer scalar\n'
          return(list(error=error))
        }
      }

      if(!bad_n && identical(fcall, "Fitting") &&
         model_code_n %in% binomial_n_models && length(n) > 1L) {
        nobs_n <- length(unlist(data, use.names = FALSE))
        bad_n <- length(n) != nobs_n
      }
      if(bad_n) {
        error <- 'the parameter n must contain positive integer values; for Binomial it may be scalar or one value per observation\n'
        return(list(error=error))
      }
    }

    # START check fitting
    if(fcall=="Fitting"){
         if(!is.null(coordx_dyn))
            {
                if(!is.list(coordx_dyn)){
                error<-"Dynamical coordinates must be a list object"
              return(list(error=error))}
              if(!is.list(data)){
                error<-"Data must be a list"
              return(list(error=error))}
          }
        else {if(missing(data) || !is.numeric(data)){
              error <- 'insert a numeric vector or matrix of data\n'
              return(list(error=error))}}
        if(!is.null(fixed) & !is.list(fixed)){
            error <- 'insert fixed values as a list of parameters\n'
            return(list(error=error))}
        if(!is.null(fixed)){ 
            namfixed <- names(fixed)
     
        if(!all(namfixed %in% c(NuisParam2(model,CheckBiv(CkCorrModel(corrmodel)),num_betas,copula),CorrelationPar(CkCorrModel(corrmodel))))){
                error <- 'some names of the fixed parameters is/are not correct\n'
                return(list(error=error))}
        if(!CheckParamRange(unlist(fixed))){
            error <- 'some fixed values are out of the range\n'
            return(list(error=error))}}
           else {namfixed=NULL}   
        if(!is.null(likelihood) & !is.character(likelihood)){
            error <- 'insert the type of likelihood objects\n'
            return(list(error=error))}
        for(i in 1:length(maxdist)){
        if(!is.null(maxdist[i])){
            error <- "insert a positive numeric value for the maximum spatial distance\n"
            if(!is.numeric(maxdist)) return(list(error=error))
            else if(maxdist[i]<0) return(list(error=error))}}
        if(!is.null(maxtime)){
            error <- "insert a positive numeric value for the maximum time interval\n"
            if(!is.numeric(maxtime)) return(list(error=error))
            else if(maxtime<0) return(list(error=error))}
        if(!is.null(optimizer) & !is.character(optimizer)){
            error <- 'insert the type of maximising algorithm\n'
            return(list(error=error))}
        if(!is.null(varest) & !is.logical(varest)){
            error <- 'the parameter std.err need to be a logical value\n'
            return(list(error=error))}


        if(!is.null(type) & !is.character(type)){
            error <- 'insert the configuration of the likelihood objects\n'
            return(list(error=error))}
        if(is.null(CkType(type))){
            error <- 'the type name of the likelihood objects is not correct\n'
            return(list(error=error))}
        if(is.null(CkLikelihood(likelihood))){
           error <- 'the setting name of the likelihood objects is not correct\n'
           return(list(error=error))}
        if(likelihood == "Full"){
            if(!any(type == c("Standard", "Tapering","Tapering1","CV","Tapering2"))){
                error <- 'insert a type name of the likelihood objects compatible with the full likelihood\n'
                return(list(error=error))}}
        if(likelihood == "Marginal"){
            if(!any(type == c("Pairwise","Independence"))){
                error <- 'Marginal likelihood is compatible only with Pairwise or Independence type\n'
                return(list(error=error))}}
        if(likelihood == "Conditional" && type != "Pairwise") {
            error <- 'Conditional likelihood is compatible only with Pairwise type\n'
            return(list(error=error))
        }
        if(likelihood == "Marginal_2" && type != "Pairwise") {
            error <- 'Marginal_2 likelihood is compatible only with Pairwise type\n'
            return(list(error=error))
        }
        if(varest & (likelihood == "Conditional" ||
            likelihood == "Marginal"|| likelihood == "Marginal_2")){
            error <- 'insert the type of estimation method for the variances\n'
            return(list(error=error))}

  ### check on starting parameters         
  if(!is.null(start)){
            if(!is.list(start)){
                error <- 'insert starting values as a list of parameters\n'
                return(list(error=error))}
 namstart <- names(start)

  ## Mean coefficients may be split between start and fixed, and omitted
  ## coefficients receive automatic starting values in WlsStart().  Their
  ## names and their correspondence with the columns of X are validated by
  ## the dedicated univariate/bivariate mean helpers before this generic
  ## input checker is called.  Do not require every coefficient here.
  if(!all(namstart %in% c(NuisParam2(model,CheckBiv(CkCorrModel(corrmodel)),num_betas,copula), CorrelationPar(CkCorrModel(corrmodel)))))
    { error <- 'some names of the starting parameters is/are not correct\n'
      return(list(error=error))}

  if(!CheckParamRange(unlist(start))){
                error <- 'some starting values are out of the range\n'
                return(list(error=error))
  }

   if(!is.null(fixed))
    if(any(namstart %in% namfixed)){
        error <- 'some fixed parameter name/s is/are matching with starting parameter name/s\n'
        return(list(error=error))}
  }
    ### end check on starting parameters  

  #### START - checks the format of the coordinates and dataset##
  if(!is.null(coordx_dyn))
     {AS=1}
  else
    {
    dimdata <- dim(data) # set the data dimension

    if(is.null(coordt)) # START 1) spatial random field
      {   
        if(CheckST(CkCorrModel(corrmodel)))
          {
            error <- 'temporal coordinates are missing\n'
            return(list(error=error))}
            if(grid) # START regular grid
              {
                if(is.null(coordy)) {
                  error <- 'coordy is required when grid=TRUE\n'
                  return(list(error=error))
                }
                spatial_dims <- c(length(coordx), length(coordy))
                if(!is.null(coordz)) spatial_dims <- c(spatial_dims, length(coordz))
                expected_dims <- if(CheckBiv(CkCorrModel(corrmodel)))
                                   c(spatial_dims, 2L) else spatial_dims
                if(is.null(dimdata) || !identical(as.integer(dimdata), as.integer(expected_dims))) {
                  error <- paste0('grid data dimensions must be ',
                                  paste(expected_dims, collapse=' x '), '\n')
                  return(list(error=error))
                }
              } # END regular grid
            else # START irregular grid
              {
                numsite <- length(data)
                if(CheckBiv(CkCorrModel(corrmodel))) numsite<- length(data)/2
                if(is.null(numsite))
                  {error <- c('insert a vector of spatial observations\n')
                    return(list(error=error))}
                if(is.null(coordy))
                  { dimcoord <- dim(coordx)
                    if(is.null(dimcoord))
                      {error <- c('insert a suitable set of coordinates\n')
                      return(list(error=error))}
                    else
                      { if(dimcoord[1]!=numsite || (dimcoord[2]!=2 && dimcoord[2]!=3 ))
                          {
                            error <- c('the number of coordinates does not match with the number of spatial observations\n')
                            return(list(error=error))
                          }}
                   }
                else
                  { if(length(coordx)!=length(coordy))
                      {error <- c('the number of the two coordinates does not match\n')
                        return(list(error=error))}
                    if(length(coordx)!=numsite)
                      {error <- c('the number of coordinates does not match with the number of spatial observations\n')
                        return(list(error=error))}
                  }
              }
      } # END 1) spatial random field
    else # 2) case: spatial-temporal random field
      {
        if(!is.numeric(coordt))
          { error <- 'insert a numerical vector of temporal coordinates\n'
            return(list(error=error))}
        if(length(coordt)<=1)
          { error <- 'insert a numerical vector of temporal coordinates\n'
            return(list(error=error))}
    if(grid) # START regular grid
              {
                if(is.null(coordy)) {
                  error <- 'coordy is required when grid=TRUE\n'
                  return(list(error=error))
                }
                spatial_dims <- c(length(coordx), length(coordy))
                if(!is.null(coordz)) spatial_dims <- c(spatial_dims, length(coordz))
                expected_dims <- c(spatial_dims, length(coordt))
                if(is.null(dimdata) || !identical(as.integer(dimdata), as.integer(expected_dims))) {
                  error <- paste0('space-time grid data dimensions must be ',
                                  paste(expected_dims, collapse=' x '), '\n')
                  return(list(error=error))
                }
              } # END regular grid
            else # START irregular grid
              {
                if(is.list(coordx_dyn)) {
                 if(length(coordx_dyn)!= length(coordt))
                      { error <- c('the time coordinate does not match with spatial dynamic coordinates number\n')
                    return(list(error=error))}
                }
                if(is.null(dimdata))
                  { error <- c('insert a matrix of t x d spatial-temporal observations\n')
                    return(list(error=error))}
                if(length(dimdata)!=2)
                  { error <- c('the dimension of the data matrix is not correct\n')
                    return(list(error=error))}
                if(is.null(coordy))
                  {  
                    if(dimdata[2]!=nrow(coordx) || dimdata[2]!=nrow(coordx))
                      { error <- c('the number of coordinates does not match with the number of spatial observations\n')
                        return(list(error=error))  }
                  }
                else
                  {
                    if(length(coordx)!=length(coordy))
                      {error <- c('the number of the  spatial coordinates does not match\n')
                        return(list(error=error))}
                    if(length(coordx)!=dimdata[2])
                      { error <- c('the number of coordinates does not match with the number of spatial observations\n')
                        return(list(error=error))}
                  }
                if(dimdata[1]!=length(coordt))
                  { error <- c('the time coordinate does not match with the number of the matrix rows\n')
                    return(list(error=error))}
              } # END irregular grid
      }
      # END - check the format of the inserted coordinates and dataset
      } 
    }
    # END check fitting
    # START check simulation
    if(fcall=="Simulation"){
        if(type=="Tapering")
          {
  
         if(!is.null(coordx_dyn))
            {if(!is.list(coordx_dyn)){
                error<-"Dynamical coordinates mst be a list object"
              return(list(error=error))}}
          if(is.list(coordx_dyn)) {
                 if(length(coordx_dyn)!= length(coordt))
                      { error <- c('the time coordinate does not match with spatial dynamic coordinates number\n')
                    return(list(error=error))}}
            if(!is.null(maxdist)){
            error <- "insert a positive numeric value for the maximum spatial distance\n"
            if(!is.numeric(maxdist)) return(list(error=error))
            else if(maxdist<0) return(list(error=error))}
            if(!is.null(maxtime)){
            error <- "insert a positive numeric value for the maximum temporal distance\n"
            if(!is.numeric(maxtime)) return(list(error=error))
            else if(maxtime<0) return(list(error=error))}
            if(!is.null(tapsep)){
            error <- "separability parameter of spacetime taper must be between 0  and 1\n"
            if(!is.numeric(tapsep)) return(list(error=error))
            else if(tapsep<0||tapsep>1) return(list(error=error))}
          }  ## end tapering

        if(is.null(param) || !is.list(param)){
            error <- 'insert the parameters as a list\n'
            return(list(error=error))}
       biv<-CheckBiv(CkCorrModel(corrmodel))
 
       a1=unique(c(NuisParam2("Gaussian",biv,num_betas,NULL),
                    NuisParam2(model,biv,num_betas,copula)))
       a2=CorrelationPar(CkCorrModel(corrmodel))
  
        if(length(param)!= length(c(a1,a2)))
             {
            error <- "some parameters are missing or does not match with the declared model\n"
            return(list(error=error))}

        if(!all( names(param) %in% c(unique(c(NuisParam2("Gaussian",biv,num_betas,copula),NuisParam2(model,biv,num_betas,copula))),
                                      CorrelationPar(CkCorrModel(corrmodel))))){
            error <- 'some names of the parameters are not correct\n'
            return(list(error=error))}
           
        if(is.list(coordx_dyn)) {
                if(biv) coordt=c(1,2)
                 if(length(coordx_dyn)!= length(coordt)){
                      { error <- c('the number of temporal instants does not match with  the dynamic  spatial coordinates number\n')
                    return(list(error=error))}}    
        }
        if(!CheckParamRange(unlist(param))){
            error <- 'some parameters are out of the range\n'
          return(list(error=error))}
     }# END check simulation
  
    }
 else{   

    if(missing(coordx)) {
    error <- "spatial locations must be a matrix of dimension 2\n"
    return(list(error=error))}
    else     {
    if(is.vector(coordx)&&!length(coordx)==2)    {
           error <- "spatial locations must be a vector of dimension 2\n"
           return(list(error=error))}
    if(is.matrix(coordx)&&!ncol(coordx)==2)       {
           error <- "spatial locations must be  a matrix of dimension 2\n"
           return(list(error=error))}
    }
   if((!is.null(coordt)&&!is.numeric(coordt))  ){
           error <- "time  must be a vector\n"
           return(list(error=error))}
   if(!type %in% c("simple","ordinary","Simple","Ordinary")){
           error <-"kriging type can be  Simple or Ordinary\n"
   return(list(error=error))}
   if(missing(data) || !is.numeric(data)){
           error <- "insert a numeric vector of data\n"
           return(list(error=error))}
  if(is.list(coordx_dyn)) {
                 if(length(coordx_dyn)!= length(coordt))
                      { error <- c('the time coordinate does not match with spatial dynamic coordinates number\n')
                    return(list(error=error))}}
        }

      # END check kriging
  }

CkLikelihood <- function(likelihood)
  {
    CkLikelihood <- switch(likelihood,
                              None=0,
                              Conditional=1,
                              Full=2,
                              Marginal=3,
                              Marginal_2=1)
    return(CkLikelihood)
  }

CkModel <- function(model)
  {
    CkModel <- switch(model,
                         None=0,
                         Gaussian=1,Gauss=1,
                         Binary=2,Bernoulli=2,
                         Tukeygh=9,
                         SkewGaussian=10,SkewGauss=10,
                         Binomial=11,
                         StudentT=12,
                         Wrapped=13,
                         Geom=14,Geometric=14,
                         PoisBin=15,
                         BinomialNeg=16,
                         PoisBinNeg=17,
                         SkewStudentT=18,
                         Binomial2=19,
                         SinhAsinh=20,
                         Gamma=21,
                         LogGaussian=22,LogGauss=22,
                         Gamma2=23,
                         LogLogistic=24,
                         Logistic=25,
                          SkewLaplace=59,
                         Weibull=26,
                         TwoPieceStudentT=27,
                         Beta=28,
                         TwoPieceGaussian=29,TwoPieceGauss=29,
                         Poisson=30,poisson=30,
                         Binomial_TwoPieceGaussian=31,
                         Binomial_TwoPieceGauss=31,
                         BinomialNeg_TwoPieceGaussian=32,
                         BinomialNeg_TwoPieceGauss=32,
                         Kumaraswamy=33,
                         Tukeyh=34,tukeyh=34,
                         Gaussian_misp_StudentT=35,
                         Gaussian_misp_Poisson=36,
                         Gaussian_misp_SkewStudentT=37,
                         TwoPieceTukeyh=38,
                         TwoPieceBimodal=39,
                         Tukeyh2=40,tukeyh2=40,
                         Gaussian_misp_Tukeygh=41,
                         Kumaraswamy2=42,
                         PoissonZIP=43,
                         Gaussian_misp_PoissonZIP=44,
                         BinomialNegZINB=45,
                         PoissonGamma=46,poissongamma=46,
                         Gaussian_misp_PoissonGamma=47,
                         BinomialLogistic=49,Binomiallogistic=49,
                         Beta2=50,
                         Gaussian_misp_Binomial=51,
                         Gaussian_misp_BinomialNeg=52,
                         PoissonZIP1=53,
                         Binary_misp_BinomialNeg=54,
                         BinomialNegZINB1=56,
                         PoissonGammaZIP=57,
                         PoissonGammaZIP1=58
                         )
    return(CkModel)
  }
  
CkType <- function(type)
  {
    CkType <- switch(type,
                        Pairwise=2,
                        Standard=4,
                        Tapering=5,
                        Tapering2=5,
                        Tapering1=6,
                        GeoWLS=7,
                        CV=8,
                        Independence=9)
    return(CkType)
  }


.GeoValidateHoleWendlandParameters <- function(corrmodel, param, d=2L,
                                                context="GeoModels") {
  code <- CkCorrModel(corrmodel)
  if(is.null(code) || !(as.integer(code) %in% c(26L,29L)))
    return(invisible(TRUE))
  if(!is.list(param))
    stop(context, ": param must be a list.", call.=FALSE)

  req <- c("power1","power2","scale","smooth")
  miss <- req[!vapply(req, function(nm) {
    z <- param[[nm]]
    is.numeric(z) && length(z)==1L && is.finite(z)
  }, logical(1L))]
  if(length(miss))
    stop(context, ": ", corrmodel, " requires finite scalar parameter(s): ",
         paste(miss, collapse=", "), ".", call.=FALSE)

  k <- as.numeric(param$power1)
  nu <- as.numeric(param$power2)
  a <- as.numeric(param$scale)
  sm <- as.numeric(param$smooth)
  if(k < 0 || abs(k-round(k)) > sqrt(.Machine$double.eps))
    stop(context, ": for ", corrmodel,
         ", power1 is the hole-effect order k and must be a non-negative integer.",
         call.=FALSE)
  k <- as.integer(round(k))
  d <- as.integer(d)[1L]
  if(is.na(d) || d < 1L) d <- 2L
  deff <- d + 2L*k

  nu_min <- function(xi, dd) {
    if(dd == 1L && xi < 0)
      (sqrt(8*xi + 9) - 1)/2
    else
      xi + (dd + 1)/2
  }

  if(as.integer(code) == 26L) {
    xi <- sm
    if(a <= 0 || xi <= -0.5)
      stop(context, ": GenWend_Hole requires scale > 0 and smooth (xi) > -1/2.",
           call.=FALSE)
  } else {
    ## GW_{nu*a, xi-1/2, nu, d, k}: public smooth is the Matern-like xi.
    if(a <= 0 || sm <= 0)
      stop(context, ": GenWend_Matern_Hole requires scale > 0 and smooth > 0.",
           call.=FALSE)
    xi <- sm - 0.5
  }

  lower <- nu_min(xi, deff)
  if(nu < lower - 100*.Machine$double.eps*max(1,abs(lower)))
    stop(context, ": invalid ", corrmodel,
         " parameters: power2 (nu) must be >= ", format(lower, digits=15),
         " for smooth=", format(sm, digits=15), ", power1 (k)=", k,
         " and dimension d=", d, ".", call.=FALSE)
  invisible(TRUE)
}

#####  names of the correlation models ###############
#CorrParam <- function(corrmodel)
#{
#  val <- CorrelationPar(CkCorrModel(corrmodel))
#  if (is.null(val)) warning("Unrecognized correlation model ")
#  return(val)
#}

CorrParam <- function(corrmodel)
{
  val_code <- CkCorrModel(corrmodel)
  if (is.null(val_code) || length(val_code) == 0)
    stop("Unrecognized correlation model")

  val <- CorrelationPar(val_code)
  if (is.null(val))
    stop("Parameters not defined for this correlation code")

  return(val)
}


#####  names of the correlation models ###############
CorrelationPar <- function(corrmodel) {

  param_map <- list(
    "1" = c("power2", "scale"),
    "2" = c("scale"),
    "3" = c("scale"),
    "4" = c("scale"),
    "5" = c("power1", "power2", "scale"),
    "6" = c("power2", "scale", "smooth"),
    "7" = c("power2", "scale", "smooth"),
    "8" = c("power1", "power2", "scale"),
    "10" = c("scale_1", "scale_2", "smooth"),
    "11" = c("power2", "scale"),
    "12" = c("power", "scale"),
    "13" = c("power2", "scale"),
    "14" = c("scale", "smooth"),
    "15" = c("power2", "scale"),
    "16" = c("scale"),
    "17" = c("power", "scale"),
    "18" = c("power"),
    "19" = c("power2", "scale", "smooth"),
    "20" = c("scale", "smooth"),
    "21" = c("power1", "power2", "scale", "smooth"),
    "22" = c("power2", "scale", "smooth"),
    "23" = c("power2", "scale", "smooth"),
    "24" = c("power2", "scale", "smooth"),
    "25" = c("power2", "scale", "smooth"),
    "26" = c("power1", "power2", "scale", "smooth"),
    "27" = c("power1", "scale", "smooth"),
    "28" = c("scale"),
    "29" = c("power1", "power2", "scale", "smooth"),
    "30" = c("power2", "scale", "smooth"),
    "31" = c("scale"),
    "32" = c("scale"),
    "42" = c("power_s", "power_t", "scale_s", "scale_t", "sep"),
    "44" = c("power2", "power_s", "power_t", "scale_s", "scale_t"),
    "46" = c("power_s", "power_t", "scale_s", "scale_t", "sep"),
    "48" = c("power_t", "scale_s", "scale_t", "smooth_s"),
    "50" = c("power_s", "power_t", "scale_s", "scale_t", "sep"),
    "52" = c("power_s", "power_t", "scale_s", "scale_t", "sep"),
    "54" = c("power_s", "power_t", "scale_s", "scale_t", "sep"),
    "56" = c("scale_s", "scale_t", "smooth_t"),
    "58" = c("power_s", "power_t", "scale_s", "scale_t"),
    "60" = c("power_s", "power_t", "scale_s", "scale_t", "sep"),
    "61" = c("power_s", "power2_s", "scale_s", "scale_t", "sep", "smooth_t"),
    "62" = c("power_t", "power2_t", "scale_s", "scale_t", "sep", "smooth_s"),
    "63" = c("power_t", "power2_s", "power2_t", "scale_s", "scale_t", "sep"),
    "64" = c("power_s", "power2_s", "power2_t", "scale_s", "scale_t", "sep"),
    "65" = c("power_t", "power2_s", "power2_t", "scale_s", "scale_t", "sep"),
    "66" = c("power_s", "power2_s", "power2_t", "scale_s", "scale_t", "sep"),
    "67" = c("power_t", "power2_s", "power2_t", "scale_s", "scale_t", "sep"),
    "68" = c("power_s", "power2_s", "power2_t", "scale_s", "scale_t", "sep"),
    "69" = c("power2_s", "power2_t", "scale_s", "scale_t"),
    "70" = c("power2_s", "power2_t", "scale_s", "scale_t"),
    "71" = c("power2_s", "power2_t", "scale_s", "scale_t"),
    "72" = c("power2_s", "power2_t", "scale_s", "scale_t"),
    "73" = c("power2_s", "power2_t", "scale_s", "scale_t"),
    "74" = c("power2_s", "power2_t", "scale_s", "scale_t"),
    "75" = c("power2_s", "power2_t", "scale_s", "scale_t"),
    "76" = c("power2_s", "power2_t", "scale_s", "scale_t"),
    "77" = c("power2_s", "power2_t", "scale_s", "scale_t"),
    "78" = c("scale_s", "scale_t", "smooth_s", "smooth_t", "power2_s", "power2_t"),
    "82" = c("power2", "scale_s", "scale_t"),
    "84" = c("scale_s", "scale_t"),
    "85" = c("scale_s", "scale_t", "power2_s", "power2_t", "smooth_s", "smooth_t", "sep"),
    "86" = c("scale_s", "scale_t", "smooth_s", "smooth_t"),
    "95" = c("scale_s", "scale_t", "smooth_s", "smooth_t","sep"),
    "87" = c("power_t", "power2_s", "power2_t", "scale_s", "scale_t", "sep", "smooth_s"),
    "88" = c("power_s", "power2_s", "power2_t", "scale_s", "scale_t", "sep", "smooth_t"),
    "89" = c("scale_s", "scale_t", "smooth_s", "smooth_t", "sep"),
    "94" = c("power_s", "power_t", "scale_s", "scale_t"),
    "96" = c("power2_s", "power2_t", "scale_s", "scale_t"),
    "111" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2", "scale"),
    "112" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2_1", "power2_12", "power2_2", "scale_1", "scale_12", "scale_2"),
    "113" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2", "scale"),
    "114" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2_1", "power2_12", "power2_2", "scale_1", "scale_12", "scale_2"),
    "115" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2", "scale"),
    "116" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2_1", "power2_12", "power2_2", "scale_1", "scale_12", "scale_2"),
    "117" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "scale_1", "scale_12", "scale_2", "smooth_1", "smooth_12", "smooth_2"),
    "118" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "scale_1", "scale_2", "smooth_1", "smooth_2"),
    "119" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "scale", "smooth"),
    "120" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2_1", "power2_2", "scale_1", "scale_2"),
    "121" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "scale_1", "scale_2", "smooth_1", "smooth_2"),
    "122" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "scale", "smooth"),
    "124" = c("a_1", "a_12", "a_2", "nugget_1", "nugget_2", "scale_1", "scale_2"),
    "126" = c("a_1", "a_12", "a_2", "a_21", "nugget_1", "nugget_2", "scale_1", "scale_2"),
    "128" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "scale_1", "scale_12", "scale_2", "smooth_1", "smooth_12", "smooth_2"),
    "129" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2_1", "power2_2", "scale_1", "scale_2"),
    "130" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2", "scale", "smooth"),
    "131" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2_1", "power2_2", "scale_1", "scale_2"),
    "132" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2_1", "power2_12", "power2_2", "scale_1", "scale_12", "scale_2", "smooth_1", "smooth_12", "smooth_2"),
    "134" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "power2_1", "power2_2", "scale_1", "scale_2", "smooth_1", "smooth_2"),
    "136" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "scale_1", "scale_12", "scale_2", "smooth_1", "smooth_12", "smooth_2", "power2_2"),
    "137" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "scale_1", "scale_12", "scale_2", "smooth_1", "smooth_12", "smooth_2", "power2_1", "power2_12", "power2_2"),
    "138" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "scale_1", "scale_12", "scale_2", "power1_1", "power1_12", "power1_2", "power2_1", "power2_12", "power2_2"),
    "139" = c("sill_1", "sill_2", "nugget_1", "nugget_2", "pcol", "scale_1", "scale_12", "scale_2", "power_1", "power_12", "power_2")
  )
  
  corrmodel_char <- as.character(corrmodel)
  if (corrmodel_char %in% names(param_map)) {
    return(param_map[[corrmodel_char]])
  } else {
    return(NULL)
  }
}

  #############################################################  
  #############################################################
NuisParam2 <- function(model, bivariate = FALSE, num_betas = c(1, 1), copula = NULL) {
  if (is.null(CkModel(model))) {
    stop("The name of the model is not correct.\n")
  }

  # Normalizza num_betas per univariati
  if (!bivariate && all(num_betas == c(1, 1))) {
    num_betas <- 1
  }

  #############################################################
  # Univariate case
  #############################################################
  if (!bivariate) {
    mm <- if (num_betas == 1) "mean" else {
      c("mean", paste0("mean", 1:(num_betas - 1)))
    }

    base_params <- c(mm, "nugget", "sill")

    # Aggiungi nu se copula è Clayton o SkewGaussian
    add_nu <- !is.null(copula) && copula %in% c("Clayton", "SkewGaussian")

    # Lista modelli con parametri specifici
    model_list <- list(
      simple = c("Gaussian", "Gauss", "Binomial", "Gaussian_misp_Binomial","Gaussian_misp_BinomialNeg",
                 "BinomialLogistic", "Binomial2", "BinomialNeg", "Binary", "Bernoulli",
                 "Poisson", "Gaussian_misp_Poisson", "Geom", "Geometric", "Wrapped",
                 "PoisBin", "PoisBinNeg", "LogGaussian", "LogGauss", "Logistic"),
      
      double_nugget = c("PoissonZIP", "Gaussian_misp_PoissonZIP", "BinomialNegZINB"),
      single_nugget_zip = c("PoissonZIP1", "Gaussian_misp_PoissonZIP1", "BinomialNegZINB1"),
      
      shape1 = c("PoissonGammaZIP1", "Gaussian_misp_PoissonGammaZIP1"),
      shape2 = c("PoissonGammaZIP", "Gaussian_misp_PoissonGammaZIP"),
      
      shape_only = c("Weibull", "weibull", "Gamma", "gamma", "LogLogistic",
                     "Loglogistic", "PoissonGamma", "PoissonWeibull",
                     "Gaussian_misp_PoissonGamma"),
      
      beta_like = c("Beta2", "Kumaraswamy2"),
      beta_full = c("Beta", "Kumaraswamy"),
      
      skew = c("SkewGaussian", "SkewGauss", "TwoPieceGaussian", "TwoPieceGauss",  "SkewLaplace",
               "Binomial_TwoPieceGaussian", "Binomial_TwoPieceGauss",
               "BinomialNeg_TwoPieceGaussian", "BinomialNeg_TwoPieceGauss"),
      
      skew_t = c("SkewStudentT", "TwoPieceStudentT", "Gaussian_misp_SkewStudentT"),
      bimodal = "TwoPieceBimodal",
      
      student_t = c("StudentT", "Gaussian_misp_StudentT"),
      
      tukey_h = c("Tukeyh", "tukeyh"),
      tukey_h2 = c("Tukeyh2", "tukeyh2"),
      tukey_gh = c("Tukeygh", "SinhAsinh", "TwoPieceTukeyh", "Gaussian_misp_Tukeygh")
    )

    # Assegnazione parametri
    if (model %in% model_list$simple) {
      param <- base_params
    } else if (model %in% model_list$double_nugget) {
      param <- c(mm, "nugget1", "nugget2", "pmu", "sill")
    } else if (model %in% model_list$single_nugget_zip) {
      param <- c(mm, "nugget", "pmu", "sill")
    } else if (model %in% model_list$shape1) {
      param <- c(mm, "nugget", "pmu", "sill", "shape")
    } else if (model %in% model_list$shape2) {
      param <- c(mm, "nugget1", "nugget2", "pmu", "sill", "shape")
    } else if (model %in% model_list$shape_only) {
      param <- c(mm, "nugget", "sill", "shape")
    } else if (model %in% model_list$beta_like) {
      param <- c(mm, "nugget", "sill", "shape", "min", "max")
    } else if (model %in% model_list$beta_full) {
      param <- c(mm, "nugget", "sill", "shape1", "shape2", "min", "max")
    } else if (model %in% model_list$skew) {
      param <- c(mm, "nugget", "sill", "skew")
    } else if (model %in% model_list$skew_t) {
      param <- c(mm, "df", "nugget", "sill", "skew")
    } else if (model %in% model_list$bimodal) {
      param <- c(mm, "df", "nugget", "sill", "shape", "skew")
    } else if (model %in% model_list$student_t) {
      param <- c(mm, "df", "nugget", "sill")
    } else if (model %in% model_list$tukey_h) {
      param <- c(mm, "nugget", "sill", "tail")
    } else if (model %in% model_list$tukey_h2) {
      param <- c(mm, "nugget", "sill", "tail1", "tail2")
    } else if (model %in% model_list$tukey_gh) {
      param <- c(mm, "nugget", "sill", "skew", "tail")
    } else {
      stop("Model not recognized in univariate case.")
    }

    if (add_nu) param <- c(param, "nu")
    return(param)
  }

  #############################################################
  # Bivariate case
  #############################################################
  if (bivariate) {
    mm1 <- if (num_betas[1] == 1) "mean_1" else c("mean_1", paste0("mean_1", 1:(num_betas[1] - 1)))
    mm2 <- if (num_betas[2] == 1) "mean_2" else c("mean_2", paste0("mean_2", 1:(num_betas[2] - 1)))
    mm <- c(mm1, mm2)

    model_list <- list(
      biv_simple = c("Gaussian", "Gauss", "Binomial", "BinomialLogistic", "Binomial2",
                     "BinomialNeg", "Geom", "Geometric", "Poisson", "Wrapped", "PoisBin",
                     "PoisBinNeg", "LogGaussian", "LogGauss", "Logistic"),
      
      biv_skew = c("SkewGaussian", "SkewGauss", "TwoPieceGaussian", "TwoPieceGauss"),
      biv_shape = c("Weibull", "Gamma", "LogGauss", "LogGaussian", "LogLogistic"),
      biv_t = c("StudentT", "Gaussian_misp_StudentT"),
      biv_tukeyh = c("Tukeyh", "tukeyh"),
      biv_tukeygh = c("Tukeygh", "SinhAsinh", "TwoPieceTukeyh")
    )

    if (model %in% model_list$biv_simple) {
      param <- mm
    } else if (model %in% model_list$biv_skew) {
      param <- c(mm, "skew_1", "skew_2")
    } else if (model %in% model_list$biv_shape) {
      param <- c(mm, "shape_1", "shape_2")
    } else if (model %in% model_list$biv_t) {
      param <- c(mm, "df_1", "df_2")
    } else if (model %in% model_list$biv_tukeyh) {
      param <- c(mm, "tail_1", "tail_2")
    } else if (model %in% model_list$biv_tukeygh) {
      param <- c(mm, "skew_1", "skew_2", "tail_1", "tail_2")
    } else {
      stop("Model not recognized in bivariate case.")
    }

    return(param)
  }

  stop("Unexpected error: check inputs.")
}

# Wrapper NuisParam
NuisParam <- function(model, bivariate = FALSE, num_betas = c(1, 1), copula = NULL) {
  a <- NuisParam2(model, bivariate, num_betas, copula)
  models_no_sill <- c(
    "Weibull", "Poisson", "Binomial", "Gamma", "LogLogistic",
    "BinomialNeg", "Binary", "Bernoulli", "Geom", "Geometric", "Gaussian_misp_Poisson",
    "PoissonGammaZIP", "PoissonGamma", "PoissonGammaZIP1", "Gaussian_misp_BinomialNeg","Gaussian_misp_Binomial",
    "PoissonZIP", "Gaussian_misp_PoissonZIP", "BinomialNegZINB",
    "PoissonZIP1", "Gaussian_misp_PoissonZIP1", "BinomialNegZINB1",
    "Beta2", "Kumaraswamy2", "Beta", "Kumaraswamy"
  )
  if (model %in% models_no_sill) {
    a <- a[a != "sill"]
  }
  return(a)
}
# Wrapper MargParam
MargParam <- function(model, bivariate = FALSE, num_betas = c(1, 1), copula = NULL) {
  a <- NuisParam(model = model, bivariate = bivariate, num_betas = num_betas, copula = copula)
  a <- a[a != "nugget"]
  a
}
####################################################################################
#########################################################################################
#########################################################################################
#########################################################################################

.GeoSelectedPairInfo <- function(pair_index, coords, distance, radius,
                                 spacetime = FALSE, bivariate = FALSE,
                                 ns = NULL, coordt = NULL) {
  if (!is.list(pair_index) ||
      is.null(pair_index$colidx) || is.null(pair_index$rowidx)) {
    stop("Internal selected-pair specification is invalid.", call. = FALSE)
  }

  colidx <- pair_index$colidx
  rowidx <- pair_index$rowidx

  if (!is.numeric(colidx) || !is.numeric(rowidx) ||
      length(colidx) != length(rowidx) || length(colidx) < 1L ||
      any(!is.finite(colidx)) || any(!is.finite(rowidx)) ||
      any(colidx != floor(colidx)) || any(rowidx != floor(rowidx))) {
    stop("Internal selected-pair indices must be finite integer vectors of equal positive length.",
         call. = FALSE)
  }

  colidx <- as.integer(colidx)
  rowidx <- as.integer(rowidx)
  coords <- as.matrix(coords)

  if (nrow(coords) < 1L ||
      any(colidx < 1L | colidx > nrow(coords)) ||
      any(rowidx < 1L | rowidx > nrow(coords))) {
    stop("Internal selected-pair indices are outside the coordinate range.",
         call. = FALSE)
  }

  ## Bound temporary allocations for massive pair graphs.  The output lag
  ## vector is unavoidable, but coordinate differences are created only for
  ## one block at a time rather than as M x d matrices.
  block_size <- getOption("GeoModels.selected_pair_block_size", 1000000L)
  block_size <- suppressWarnings(as.integer(block_size)[1L])
  if (!is.finite(block_size) || block_size < 1L) block_size <- 1000000L

  npairs <- length(colidx)
  lags <- numeric(npairs)
  starts <- seq.int(1L, npairs, by = block_size)

  for (from in starts) {
    to <- min(npairs, from + block_size - 1L)
    take <- from:to
    ii <- colidx[take]
    jj <- rowidx[take]

    if (distance == 0L) {
      squared <- numeric(length(take))
      for (dimension in seq_len(ncol(coords))) {
        delta <- coords[ii, dimension] - coords[jj, dimension]
        squared <- squared + delta * delta
      }
      lags[take] <- sqrt(squared)
    } else {
      if (ncol(coords) < 2L) {
        stop("Geodesic and chordal selected-pair distances require longitude and latitude.",
             call. = FALSE)
      }

      lon1 <- coords[ii, 1L] * pi / 180
      lat1 <- coords[ii, 2L] * pi / 180
      lon2 <- coords[jj, 1L] * pi / 180
      lat2 <- coords[jj, 2L] * pi / 180

      cosine <- sin(lat1) * sin(lat2) +
        cos(lat1) * cos(lat2) * cos(lon1 - lon2)
      cosine <- pmax(-1, pmin(1, cosine))
      angle <- acos(cosine)

      lags[take] <- if (distance == 2L) {
        radius * angle
      } else {
        2 * radius * sin(0.5 * angle)
      }
    }
  }

  if (any(!is.finite(lags))) {
    stop("Non-finite distances were obtained for the reused selected pairs.",
         call. = FALSE)
  }

  lagt <- 0
  first <- second <- NULL

  if (isTRUE(spacetime)) {
    if (is.null(ns) || is.null(coordt) || length(ns) != length(coordt)) {
      stop("Internal space-time dimensions are inconsistent for selected-pair reuse.",
           call. = FALSE)
    }

    cumulative <- cumsum(as.integer(ns))
    lagt <- numeric(npairs)
    for (from in starts) {
      to <- min(npairs, from + block_size - 1L)
      take <- from:to
      tc <- as.integer(findInterval(colidx[take] - 1L, cumulative) + 1L)
      tr <- as.integer(findInterval(rowidx[take] - 1L, cumulative) + 1L)
      if (any(tc < 1L | tc > length(coordt)) ||
          any(tr < 1L | tr > length(coordt))) {
        stop("Internal selected-pair time indices are invalid.", call. = FALSE)
      }
      lagt[take] <- abs(as.numeric(coordt[tc]) - as.numeric(coordt[tr]))
    }
  }

  if (isTRUE(bivariate)) {
    if (is.null(ns) || length(ns) != 2L) {
      stop("Internal bivariate dimensions are inconsistent for selected-pair reuse.",
           call. = FALSE)
    }
    n1 <- as.integer(ns[1L])
    first <- as.integer(colidx > n1)
    second <- as.integer(rowidx > n1)
  }

  list(
    colidx = colidx,
    rowidx = rowidx,
    lags = lags,
    lagt = lagt,
    first = first,
    second = second
  )
}

StartParam <- function(coordx, coordy,coordz ,coordt,coordx_dyn, corrmodel, data, distance, fcall,
                      fixed, grid,likelihood,  maxdist, neighb,maxtime, model, n, 
                      param, parscale,paramrange, radius, start, taper, tapsep, 
                      type,typereal,  weighted,copula, X,memdist,nosym,p_neighb,thin_method)
{

    ## GeoVarest may temporarily provide the exact graph used by the original
    ## fit.  Pair distances can also be supplied as a transient worker-local
    ## cache, avoiding repeated O(M) distance calculations across bootstrap
    ## replications without enlarging the returned GeoFit object.
    selected_pair_index <- getOption("GeoModels.selected_pair_index", NULL)
    selected_pair_info <- getOption("GeoModels.selected_pair_info", NULL)

####################################  
### START internal functions:
#################################### 
    replicates=1
    # Check if the correlation is bivariate
    CheckBiv <- function(corrmodel)
    {
        CheckBiv <- NULL
        if(corrmodel >= 101 & corrmodel <= 140) CheckBiv <- TRUE
        else CheckBiv <- FALSE
        return(CheckBiv)
    }
    # Check if the correlation is spatial or spatial-temporal
    CheckST <- function(corrmodel)
    {
        CheckST <- NULL
        if(corrmodel >40 & corrmodel <= 100) CheckST <- TRUE
        else  CheckST <- FALSE
        return(CheckST)
    }
    # Check the type of distances
    CheckDistance<- function(distance)
    {
        CheckDistance <- NULL
        CheckDistance <- switch(distance,
                                eucl=0,
                                Eucl=0,
                                chor=1,
                                Chor=1,
                                geod=2,
                                Geod=2,
                                proj=3,
                                Proj=3)
        return(CheckDistance)
    }
##############################################################################################    
newtap<- function(coords,numcoord, coordt,numtime, distance,maxdist,maxtime,spacetime,bivariate,radius)
### using nearest.dist...
    {
      if(distance==0) method1="euclidean"
      if(distance==2||distance==1) method1="greatcircle"
if(method1=="greatcircle"){
      gb=spam::nearest.dist( x=coords,method = method1,
             delta = maxdist*360/(radius*2*pi), upper = NULL,miles=FALSE, R=radius)
      if(distance==2) gb@entries=radius*gb@entries             ##GC
      if(distance==1) gb@entries=2*radius*sin(0.5*gb@entries)  ##CH
      }

if(method1=="euclidean")
      gb=spam::nearest.dist(x=coords,method = method1,
                         delta = maxdist, upper =NULL,miles=FALSE, R=1)
      numpairs=length(gb@entries)
 ##loading only good distances..
   dotCall64::.C64("SetGlobalVar2",
        SIGNATURE = c("integer","integer",
             "double","integer","double",
            "double","double","integer",
            "integer","integer","integer","integer"),  
       numcoord,  numtime,  
       gb@entries,numpairs,srange[2],
       1,1,1, # to change for spacetime sparse
       spacetime,bivariate,1,1, INTENT =    c(rep("r",12)),
         PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
   
    colidx=gb@colindices
    rowidx=gb@rowpointers
    nozero=numpairs/(numcoord)^2
  return(list(colidx=colidx,rowidx=rowidx,numpairs=numpairs,nozero=nozero))
    }
####################################
### END Includes internal functions
####################################

    # Set the correlation and  if the correlation is space-time(spacetime=T and bivariate=F) or bivariate (F o T)  or univariate (case spacetime=F and bivariate=F)p
    corrmodel<-CkCorrModel(corrmodel)
    bivariate <- CheckBiv(corrmodel); if(bivariate) coordt=c(1,2)
    spacetime <- CheckST(corrmodel)
    isdyn=!is.null(coordx_dyn)
    space=!(spacetime||bivariate)
    ####################
    if(!bivariate)
    {
        if(is.null(X))  {X=1;num_betas=1}
        else 
          {if(is.list(X))  num_betas=ncol(X[[1]])
           else  num_betas=ncol(X) 
          }
    }
    if(bivariate)
    {
        if(is.null(X))  {X=1;num_betas=c(1,1)}
        else
                        { if(is.list(X))  num_betas=c(ncol(X[[1]]),ncol(X[[2]]))
                          else  num_betas=c(ncol(X),ncol(X)) }
    }
    ####################
    namesnuis <- NuisParam2(model,bivariate,num_betas,copula)
    ltimes=length(coordt)


    ### case gris handling coordz
    if(grid) {
              if(is.null(coordz)) { cc=as.matrix(expand.grid(coordx,coordy));coordx=cc[,1];coordy=cc[,2]; coordz=NULL}
              else {  
                    # if(is.null(coordz)) cc=as.matrix(expand.grid(coordx,coordy,0));
                     cc=as.matrix(expand.grid(coordx,coordy,coordz));
                     coordx=cc[,1];coordy=cc[,2]; coordz=cc[,3];
            

                  }
             }

    ### Set returning variables and initialize the model parameters:
    # Initialises the starting and fixed parameters' names
    error <- NULL
    ns<-NULL
    namesfixed <- namesstart <- namessim <- NULL
    numfixed <- numstart <- 0
    # Set the model, likelihood, correlation and the nuisance parameters:
    model <- CkModel(model)
    flagnuis <- NULL
    namescorr <- CorrelationPar(corrmodel)
    numparamcorr <- length(namescorr)
    paramcorr <- rep(1, numparamcorr)
    names(paramcorr) <- namescorr
    flagcorr <- NULL

    ### START settings the data structure:

    # set the coordinates sizes:
    if(is.null(coordx_dyn))  
    {
    
      if(is.null(coordy)){
                         if(ncol(coordx)==2) {
                                 coordy=coordx[,2];
                                 coordx=coordx[,1];
                                 coordz=NULL}
                        else { coordz=coordx[,3];coordy=coordx[,2];coordx=coordx[,1];}
                          }

      numcoord <- numcoordx <- numcoordy <-numcoordz <-length(coordx)

      if(bivariate && !is.null(nrow(coordx)) && !is.null(nrow(coordy))&& !is.null(nrow(coordz))  )   {  #heterotopic case
        numcoordx=nrow(coordx); 
        numcoordy=nrow(coordy);
        numcoordz=nrow(coordz);
        numcoord=numcoordy+numcoordx+numcoordz}
      
      ns<-rep(numcoord,ltimes)
    }
    else     ##  dynamic coords
    {
       env <- new.env()
       coords=do.call(rbind,args=c(coordx_dyn),envir = env) 

       if(is.list(X))  X=do.call(rbind,args=c(X),envir = env)
       if(ncol(coords)==2) ns=lengths(coordx_dyn)/2 
       if(ncol(coords)==3) ns=lengths(coordx_dyn)/3
       
       if(ncol(coords)==2) {coordx <- coords[,1]; coordy <- coords[,2];coordz=NULL}
       if(ncol(coords)==3) { coordx <- coords[,1]; coordy <- coords[,2];coordz <- coords[,3]}
       numcoord <- numcoordx <- numcoordy <-numcoordz <- length(coordx)
    }


   if(!space && is.null(coordx_dyn)) {coordx=rep(coordx,ltimes);coordy=rep(coordy,ltimes);coordz=rep(coordz,ltimes);}
    
    NS=cumsum(ns)
    if(!space)   NS=c(0,NS)[-(length(ns)+1)]

    # initialize tapering variables:
    tapering=ia=idx=ja=colidx=rowidx=integer(1)
    nozero<-NULL
    tapmodel=0
    cutoff <- FALSE
    distance<-CheckDistance(distance)
    ### END settings the data structure
    # START code for the simulation procedure


##################################################################################################################
    .set_fixed_default <- function(fixed, name, value) {
      if(is.null(fixed)) fixed <- list()
      if(!is.list(fixed)) fixed <- as.list(fixed)
      if(is.null(names(fixed))) names(fixed) <- rep("", length(fixed))
      if(!(name %in% names(fixed))) fixed[[name]] <- value
      fixed
    }

    if(fcall=="Fitting"){
        ### Parameters' settings:
        nuisance=nuisance1=nuisance2=NULL
        likelihood <- CkLikelihood(likelihood)
        type <- CkType(type)
     if((!bivariate&&num_betas==1)||(bivariate&all(num_betas==c(1,1))))
     {
          if(model %in% c(1,10,12,18,9,20,13,21,22,23,24,25,26,27,28,29,31,32,33,34,35,36,37,38,39,40,41,42,46,47,48,50,59))   ### continous model 
          {
           if(!bivariate) {
                           mu <- mean(unlist(data))
                           if(any(type==c(1, 7,8)))     # Checks the type of likelihood
                             fixed <- .set_fixed_default(fixed, "mean", mu)
                           nuisance <- c(mu, 0, var(c(unlist(data))))

                           if(likelihood==2 && (CkType(typereal)==5 || CkType(typereal)==7) ) tapering <- 1
                           if(model %in% c(10,29,31,32,59))         nuisance <- c(nuisance,0)
                           if(model %in% c(18,20,27,37,38,40,41))      nuisance <- c(0,nuisance,0)
                            if(model %in% c(39))      nuisance <- c(0,0,nuisance,0)
                           if(model %in% c(21,24,12,26,34,35,46,47,48))   nuisance <- c(0,nuisance)
                           #if(model %in% c(23,28,33))  nuisance <- c(0,0,0,nuisance)
                           if(model %in% c(23,28,33))  nuisance <- c(0,0,0,nuisance,0,0)
                           if(model %in% c(42,50))  nuisance <- c(0,nuisance,0,0)
                       }
             if(bivariate) {
                           if(is.null(coordx_dyn)) { mu1 <- mean(data[1,]); mu2 <- mean(data[2,])}
                           else                   { mu1 <- mean(data[[1]]); mu2 <- mean(data[[2]])}
                           if(any(type==c(1, 7, 8))) {   # Checks the type of likelihood
                             fixed <- .set_fixed_default(fixed, "mean_1", mu1)
                             fixed <- .set_fixed_default(fixed, "mean_2", mu2)
                           }
                           nuisance <- c(mu1,mu2)
                           if(model %in% c(10,29,31,32,59))  {nuisance <- c(nuisance,0.1,0.2)}
                           if(model %in% c(26,46,47,48,42,50))  {nuisance <- c(nuisance,0.1,0.2)}
                           if(model %in% c(21))  {nuisance <- c(nuisance,0.1)}
                          
                           if(likelihood==2 && (CkType(typereal)==5 || CkType(typereal)==7)) tapering <- 1
                 }
        }

if(model %in% c(11,13,14,15,16,19,17,30,45,49,51,52,53,54,56,58)){                                                        #discrete
            mu=0
            if(any(type==c(1, 7,8)))     # Checks the type of likelihood
              fixed <- .set_fixed_default(fixed, "mean", mu)                        

            nuisance <- c(mu, 0, 1)

            if(model==45) nuisance <- c(mu, 0, 0,0,1)
            if(model==53||model==56) nuisance <- c(mu, 0,0,1)
            if(model==58) nuisance <- c(mu, 0,0,0,1)
        }
 if(model %in% c(43,44)) nuisance <- c(0, 0, 0,0, 1)
 if(model %in% c(57)) nuisance <-    c(0, 0, 0,0, 0,1)

}


 if((!bivariate&&num_betas>1)||(bivariate&&num_betas[1]>1&&num_betas[2]>1) )
 {

   if(model %in% c(1,10,12,18,9,20,13,21,22,23,24,25,26,27,28,29,31,32,33,34,35,36,37,38,39,40,41,42,46,47,48,50,59))  ### continous models
{
    if(!bivariate)
    {
         if(any(type==c(1, 7,8))) { # Checks the type of likelihood
            mu <- mean(unlist(data))
            fixed <- .set_fixed_default(fixed, "mean", mu)
            if(num_betas > 1)
              for(i in seq_len(num_betas-1))
                fixed <- .set_fixed_default(fixed, paste0("mean", i), 1)
         }
            for(i in 1:num_betas) nuisance=c(nuisance,1);
            nuisance=c(nuisance,0,var(c(unlist(data))))
             if(model %in% c(10,29,31,32,59))        nuisance=c(nuisance,1)  
             if(model %in% c(21,24,12,26,34,35,46,47,48))  nuisance=c(nuisance,1) 
             if(model %in% c(18,20,27,37,38,40,41))     nuisance=c(1,nuisance,1) 
             if(model %in% c(39))     nuisance=c(1,1,nuisance,1) 
          #  if(model %in% c(23,28,33))         nuisance=c(nuisance,1,1,1)  
            if(model %in% c(23,28,33))         nuisance=c(nuisance,1,1,1,1,1)  
            if(model %in% c(42,50))  nuisance=c(nuisance,1,1,1) 
    }
    if(bivariate)
    {
            if(any(type==c(1, 7,8))) { # Checks the type of likelihood
              if(!is.list(data)) {
                data_biv <- as.matrix(data)
                mu1 <- mean(data_biv[1,])
                mu2 <- mean(data_biv[2,])
              } else {
                mu1 <- mean(data[[1]])
                mu2 <- mean(data[[2]])
              }
              fixed <- .set_fixed_default(fixed, "mean_1", mu1)
              fixed <- .set_fixed_default(fixed, "mean_2", mu2)
              if(num_betas[1] > 1)
                for(i in seq_len(num_betas[1]-1))
                  fixed <- .set_fixed_default(fixed, paste0("mean_1", i), 1)
              if(num_betas[2] > 1)
                for(i in seq_len(num_betas[2]-1))
                  fixed <- .set_fixed_default(fixed, paste0("mean_2", i), 1)
            }
            for(i in 1:num_betas[1]) nuisance1=c(nuisance1,1);
            for(i in 1:num_betas[2]) nuisance2=c(nuisance2,1);
            nuisance=c(nuisance1,nuisance2 )
             if(model %in% c(10,29,31,32,59))        nuisance=c(nuisance,1,1)  
             if(model %in% c(21,24,12,26,34,35,46,47,48,42,50))  nuisance=c(nuisance,1,1) 
             if(model %in% c(18,20,27,37,38,40,41))     nuisance=c(1,nuisance,1) 
              if(model %in% c(39))     nuisance=c(1,1,nuisance,1) 
            #if(model %in% c(23,28,33))         nuisance=c(nuisance,1,1,1)
            if(model %in% c(23,28,33))         nuisance=c(nuisance,1,1,1,1,1)  

    }
}  ### end continous models
     if(model %in% c(2,11,13,14,15,16,19,17,30,49,51,52,54,43,45,53,56,57,58)) {  # discrete models

        if(any(type==c(1, 7,8))) { # Checks the type of likelihood
          mu <- mean(unlist(data))
          fixed <- .set_fixed_default(fixed, "mean", mu)
          if(num_betas > 1)
            for(i in seq_len(num_betas-1))
              fixed <- .set_fixed_default(fixed, paste0("mean", i), 1)
        }

         if(model %in% c(2,11,13,14,15,16,19,17,30,49,51,52,54))   nuisance <- c(0,rep(0,num_betas-1) ,0, 1) 

       
  
     if(model %in% c(43,45)) nuisance <- c(0,rep(1,num_betas-1) ,0, 0,0,1)
     if(model %in% c(57)) nuisance <-    c(0,rep(1,num_betas-1) ,0, 0,0,0,1)

     if(model %in% c(53,56)) nuisance <- c(0,rep(1,num_betas-1) , 0,0,1)
     if(model %in% c(58)) nuisance <- c(0,rep(1,num_betas-1) , 0,0,0,1)

        }
     #
}

 if(!is.null(copula)) {
     if(copula=="Clayton") nuisance=c(nuisance,2)
     if(copula=="SkewGaussian") nuisance=c(nuisance,0)
 }
# Update the parameter vector     
        names(nuisance) <- namesnuis
        namesparam <- sort(c(namescorr, namesnuis))
        param <- c(nuisance, paramcorr)
        param <- param[namesparam]
        numparam <- length(param)
        flag <- rep(1, numparam)
        namesflag <- namesparam
        names(flag) <- namesflag
        all_parameter_names <- namesparam
        # Update the parameters with fixed values:
     
        if(!is.null(fixed)){
            fixed <- unlist(fixed)
            namesfixed <- names(fixed)
            numfixed <- length(namesfixed)

            bad_fixed <- setdiff(namesfixed, all_parameter_names)
            if(length(bad_fixed))
                stop("Invalid parameter name(s) in 'fixed': ",
                     paste(bad_fixed, collapse=", "),
                     ". Valid parameter names are: ",
                     paste(all_parameter_names, collapse=", "), ".", call.=FALSE)
            idx_fixed <- match(namesfixed, namesflag)
            flag[idx_fixed] <- 0
            param <- param[-match(namesfixed, namesparam)]
                          numparamcorr <- numparamcorr-sum(namesfixed %in% namescorr)
            namesparam <- names(param)
            numparam <- length(param)   

        }
        else { }

        flagcorr <- flag[namescorr]
        flagnuis <- flag[namesnuis]
        # Update the parameters with starting values:
        if(!is.null(start)){
            start <- unlist(start)
            namesstart <- names(start)

                 if(any(type==c(1, 7,8))){    # Checks the type of likelihood

                if(!bivariate) {   # univariate case
                       if(any(model==c(1,10,12,18,20,9,13,21,22,23,24,25,26,27,28,29,31,32,33,34,35,36,37,38,39,40,41,42,46,47,48,50,
                        11,14,15,16,19,17,30,45,49,51,52,54,53,56,58,59)))
                       if(any(namesstart == 'mean'))  start <- start[!namesstart == 'mean']
                       if(num_betas>1)
                       for(i in 1:(num_betas-1)) 
                       {  
                         if(any(namesstart == paste("mean",i,sep="")))  {
                              namesstart <- names(start) 
                              if(any(model==c(1,10,12,18,20,9,13,21,22,23,24,25,26,27,28,29,31,32,33,34,35,36,37,38,39,40,41,42,46,47,48,50,
                        11,14,15,16,19,17,30,45,49,51,52,54,53,56,58,59)))
                                                 start <- start[!namesstart == paste("mean",i,sep="")]
                            }
                       }
                }
                if(bivariate) {          
                                  if(any(namesstart == 'mean_1'))  start <- start[!namesstart == 'mean_1']        
                                  if(any(namesstart == 'mean_2'))  {namesstart <- names(start) ; start <- start[!namesstart == 'mean_2']}
                      
                       if(num_betas[1]>1)
                       for(i in 1:(num_betas[1]-1)) {  if(any(namesstart == paste("mean_1",i,sep="")))  {namesstart <- names(start) ; 
                       if(any(model==c(1,10,12,18,20,9,13,21,22,23,24,25,26,27,28,29,31,32,33,34,35,36,37,38,39,40,41,42,46,47,48,50,59)))
                                                 start <- start[!namesstart == paste("mean_1",i,sep="")]}}            
                       if(num_betas[2]>1)
                       for(i in 1:(num_betas[2]-1)) {  if(any(namesstart == paste("mean_2",i,sep="")))  {namesstart <- names(start) ; 
                       if(any(model==c(1,10,12,18,20,9,13,21,22,23,24,25,26,27,28,29,31,32,33,34,35,36,37,38,39,40,41,42,46,47,48,50,59)))
                                                 start <- start[!namesstart == paste("mean_2",i,sep="")]}}  
                                  }
                }

            namesstart <- names(start)
            numstart <- length(start)
            bad_start <- setdiff(namesstart, all_parameter_names)
            if(length(bad_start))
                stop("Invalid parameter name(s) in 'start': ",
                     paste(bad_start, collapse=", "),
                     ". Valid parameter names are: ",
                     paste(all_parameter_names, collapse=", "), ".", call.=FALSE)
            overlap_start_fixed <- intersect(namesstart, namesfixed)
            if(length(overlap_start_fixed))
                stop("Parameter(s) supplied in both 'start' and 'fixed': ",
                     paste(overlap_start_fixed, collapse=", "), ".", call.=FALSE)
            idx_start <- match(namesstart, namesparam)
            if(anyNA(idx_start))
                stop("Internal parameter-name reconciliation failed for: ",
                     paste(namesstart[is.na(idx_start)], collapse=", "), ".", call.=FALSE)
            param[idx_start] <- start
            }
        ### set the scale of the parameters:
        # Insert here!
        # set the range of the parameters if its the case
        paramrange=TRUE
        #if(paramrange) paramrange <- SetRangeParam(namesparam, numparam)
        #else 
        paramrange <- list(lower=NULL, upper=NULL)
        ### Set the data format:
        if(!space){ # setting spam indexes
            if(spacetime) numtime <- ltimes
            if(bivariate) numtime <- 2
                }
        else{              #    setting spam indexes
            numtime <- 1
            coordt <- 0
            data <- matrix(data, ncol=numcoord, nrow=replicates)
            }
      
       #if((typereal=="Tapering"&&type=="Tapering")||(typereal=="Tapering1"&&type=="Tapering1")||(typereal=="Tapering2"&&type=="Tapering2")){
     
        if(typereal=="Tapering"||typereal=="Tapering1"||typereal=="Tapering2"){
        tapering<-1

        if(!space){
        idx<-integer((numcoord*numtime)^2)
        ja<-integer((numcoord*numtime)^2)
        ia<-integer(numcoord*numtime+1)}
          else {idx=ja=ia=0}
        tapmodel<-CkCorrModel(taper)
                }
              ###### ojo aca!! if conditional then I used nn2 using "all" the indeces
        if(likelihood==1&&is.numeric(maxdist)&&is.null(neighb))   
        neighb= min(numcoord*numtime,500)

                                             
    }
# END code for the fitting procedure
    
##################################################################################################################
# START code for the simulation procedure
    if(fcall=="Simulation"){

        neighb=NULL;likelihood=2
        namesnuis <- sort(unique(c(namesnuis,NuisParam2("Gaussian",bivariate,num_betas,copula))))
        param <- unlist(param)
        numparam <- length(param)
        namesparam <- names(param)

        if(!bivariate) if(any(model!=c(43,45,53,56,57,58)))  namessim <- c("mean","sill","nugget","scale",namescorr[!namescorr=="scale"])
       if(bivariate)  namessim <- c("mean_1","mean_2","scale",namescorr[!namescorr=="scale"])  

        if(spacetime) numtime <- ltimes
        else {numtime <- 1; coordt <- 0}
        if(bivariate) numtime <- 2

        if((typereal=="Tapering"&&type=="Tapering")||(typereal=="Tapering1"&&type=="Tapering1")||(typereal=="Tapering2"&&type=="Tapering2")){
                 tapering<-1
                 nt=numcoord*numtime
               if(!space){  
                idx<-integer(nt^2)
                ja<-integer(nt^2)
                ia<-integer(nt+1)
                 }
                else {idx=ja=ia=0}
                tapmodel<-CkCorrModel(taper)
        }
        K=neighb
}  # END code for the simulation procedure
#####################################################################################
numpairs <- integer(1); srange <- double(1); trange <- double(1)

if(typereal=="Independence"){ maxdist=NULL;maxtime=NULL;K=neighb}
  
#################
distC=FALSE
if(!tapering) { if(is.null(neighb)&is.numeric(maxdist)) distC=TRUE  }### just for maxdist parameter
################
  
    if(is.null(maxdist)) srange<-c(srange,double(1)) else {srange<-c(srange,as.double(maxdist))}                # cutoff<-TRUE
    if(is.null(maxtime)) trange<-c(trange,double(1)) else {trange<-c(trange,as.double(maxtime))}                # cutoff<-TRUE
    isinit <- as.integer(1)
    if(is.null(tapsep))  tapsep=c(0.5,0.5)
    else  {if(length(tapsep)==1) tapsep=c(tapsep,0)}
    mem=FALSE
    if(tapering||memdist)  { mem=TRUE }   #### NB

    if(mem&&!tapering)  
      {        
         
                nn=numcoord*numtime
                if(spacetime&&isdyn)  nn=sum(ns)
                if(is.null(neighb)){
                    if(typereal=="Independence") colidx=rowidx=0
                    else         colidx=rowidx=integer(nn*(nn-1)/2)}
           
    }
    if(bivariate) {
    if(!srange[1]&&!srange[2])  srange=c(srange,0,0)
    if(is.na(srange[3])) srange[3]=srange[2];
    if(is.na(srange[4])) srange[4]=srange[2];}
    ###
    if(CheckSph(corrmodel))   radius=1
    ###
    aa=double(5);for(i in 1:length(tapsep)) aa[i]=tapsep[i];tapsep=aa

if(fcall=="Fitting"&likelihood==2&!is.null(neighb)) mem=FALSE # Vecchia gp case
if(fcall=="Fitting"&likelihood==2||fcall=="Simulation") mem=FALSE 
if(tapering) mem=TRUE

###################### using new "spam" with neighdist(just for space)#########################################################

if(tapering&space){
    cc=cbind(coordx, coordy,coordz);numcoord=nrow(cc);numtime=length(coordt)
    atap=newtap(cc,numcoord, coordt,numtime, distance,maxdist,maxtime,spacetime,bivariate,radius)
    ja=colidx=atap$colidx
    ia=rowidx=atap$rowidx
    numpairs=atap$numpairs
    nozero=atap$nozero
    K=NULL
}
else{          # all the rest
##############################################################
## loading distances in memory using brute force C routine ###
#############################################################
### aca paso solo para  simular o maximum likelihood o variogram 
### o si hay CL with  maxdist!!!
  
if(distC||fcall=="Simulation"||(fcall=="Fitting"&likelihood==2)||(fcall=="Fitting"&typereal=="GeoWLS")) {
if(fcall=="Fitting"&mem==TRUE&(!space)&!tapering)   {vv=length(NS); numcoord=NS[vv]+ns[vv]} # number of space time point in the case of coordxdyn

if(is.null(coordz)) coordz=double(numcoordx*numtime) ## is it necessary?

srange[which(srange==Inf)]=1e+50;trange[which(trange==Inf)]=1e+50
gb=.C('SetGlobalVar',as.integer(bivariate), as.double(coordx), as.double(coordy),as.double(coordz), as.double(coordt),as.integer(grid),ia=as.integer(ia),idx=as.integer(idx),  #7
           isinit=as.integer(isinit),ja=as.integer(ja), as.integer(mem), as.integer(numcoord),as.integer( numcoordx),  as.integer(numcoordy), as.integer(numcoordz), #6
           numpairs=as.integer(numpairs), as.double(radius),as.double(srange), as.double(tapsep),  as.integer(spacetime), #5
            as.integer(numtime),as.double(trange), as.integer(tapering), as.integer(tapmodel),as.integer(distance),as.integer(weighted), #6
           colidx= as.integer(colidx),rowidx= as.integer(rowidx), # 2
            as.integer(ns), as.integer(NS), as.integer(isdyn),
      PACKAGE='GeoModels', NAOK=TRUE)

  #if(!sum(coordz)) coordz=NULL

rm(colidx);rm(rowidx)
if(type=="Tapering") {rm(idx);rm(ja);rm(ia)}
##
## number  of selected pairs
numpairs <- gb$numpairs
## indexes for composite 
    colidx=gb$colidx
    rowidx=gb$rowidx
    colidx <- colidx[1:numpairs]
    rowidx  <- rowidx[1:numpairs]
## indexes for tapering 
    idx <- gb$idx
    ja <- gb$ja 
    ia <- gb$ia;
    isinit <- gb$isinit
    nozero <- numpairs/(numcoord*numtime)^2
    idx <- idx[1:numpairs]
    ja  <- ja[1:numpairs]
    K=neighb
  
}
#######################################################################
else   
###############################################################
################### loading distances in memory using nabor package 
#### it works when CL  using neighb  or maxdist AND neighb 
#############################################################
{ 
if(typereal!="Independence") {
  ########################## 
if(distance==0) distance1="Eucl";
if(distance==2) distance1="Geod";
if(distance==1) distance1="Chor";

if(all(neighb==0.5)) neighb=NULL ## ojo!!
if(sum(!is.finite(maxdist))) maxdist=NULL

if(space)   #  spatial case
{
##########################################
  
  K=neighb
  x=cbind(coordx, coordy,coordz)

  if (is.null(selected_pair_index)) {
    sol=GeoNeighIndex(coordx=x,distance=distance1,
      maxdist=maxdist,neighb=K,radius=radius,p_neighb=p_neighb,thin_method=thin_method)
  } else {
    reused <- if (!is.null(selected_pair_info)) {
      selected_pair_info
    } else {
      .GeoSelectedPairInfo(
        selected_pair_index, x, distance, radius,
        spacetime = FALSE, bivariate = FALSE, ns = ns, coordt = coordt
      )
    }
    ## GeoNeighIndex returns target indices in rowidx and neighbour indices in
    ## colidx. StartParam swaps them before returning, so preserve that schema.
    sol <- list(
      rowidx = reused$colidx,
      colidx = reused$rowidx,
      lags = reused$lags,
      lagt = reused$lagt
    )
  }

###
  nn = length(sol$lags)
  if (nn == 0L) stop("No pairwise contributions were retained. Increase 'p_neighb', 'neighb', or 'maxdist'.", call. = FALSE)
  sol$lagt=0
  gb=list(); gb$colidx=sol$colidx;
             gb$rowidx=sol$rowidx ;
             gb$numpairs=nn

             

  ## loading space distances in memory 
  mmm=1;ttt=1
if(weighted)  mmm=max(sol$lags)
  
    ss <- dotCall64::.C64("SetGlobalVar2",
           SIGNATURE = c("integer", "integer",
            "double",
             "integer", "double",
                         "double", 
                         "integer", "double", "integer", "integer",
                         "integer", "integer"),
           as.integer(numcoord),
           as.integer(numtime),
           as.double(sol$lags),
           as.integer(nn),
           as.double(mmm),
           as.double(sol$lagt),
           as.integer(nn),
           as.double(ttt),
           as.integer(spacetime),
           as.integer(bivariate),
           as.integer(1),
           as.integer(1),
           INTENT = rep("r", 12), 
           PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)

} 
##############################################   
if(spacetime)   #  space time  case
{      
  K=neighb
  x=cbind(coordx, coordy,coordz)
 
  if (is.null(selected_pair_index)) {
    sol=GeoNeighIndex(coordx=x[1:numcoord,],
      coordx_dyn=coordx_dyn,
      coordt=coordt,distance=distance1,maxdist=maxdist,neighb=K,maxtime=maxtime,radius=radius,
      p_neighb=p_neighb,thin_method=thin_method)
  } else {
    reused <- if (!is.null(selected_pair_info)) {
      selected_pair_info
    } else {
      .GeoSelectedPairInfo(
        selected_pair_index, x, distance, radius,
        spacetime = TRUE, bivariate = FALSE, ns = ns, coordt = coordt
      )
    }
    sol <- list(
      rowidx = reused$colidx,
      colidx = reused$rowidx,
      lags = reused$lags,
      lagt = reused$lagt
    )
  }
  if (!length(sol$colidx)) stop("No pairwise contributions were retained. Increase 'p_neighb', 'neighb', 'maxdist', or 'maxtime'.", call. = FALSE)
  gb=list(); gb$colidx=sol$colidx;
             gb$rowidx=sol$rowidx ;
             nn=length(gb$colidx)
             gb$numpairs=nn
  ## loading space time distances in memory   
  mmm=1;ttt=1
if(weighted) { mmm=max(sol$lags) ;ttt=max(sol$lagt)}

  ss=dotCall64::.C64("SetGlobalVar2",
        SIGNATURE = c("integer","integer",
             "double",
             "integer","double",
              "double",
              "integer","double",
            "integer","integer","integer","integer"),  
       numcoord,  numtime, 
        sol$lags,
        nn,mmm, 
      sol$lagt,
      nn,ttt,
       spacetime,bivariate,1,1, INTENT =    c(rep("r",12)),
         PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
} 
##############################################  
if(bivariate)   # bivariate case 
{ 
  K=neighb
  x=cbind(coordx, coordy,coordz)

  if (is.null(selected_pair_index)) {
    sol=GeoNeighIndex(coordx=x, coordx_dyn=coordx_dyn, distance=distance1,maxdist=maxdist,neighb=K,maxtime=maxtime,radius=radius,bivariate=TRUE,
      p_neighb=p_neighb,thin_method=thin_method)
  } else {
    reused <- if (!is.null(selected_pair_info)) {
      selected_pair_info
    } else {
      .GeoSelectedPairInfo(
        selected_pair_index, x, distance, radius,
        spacetime = FALSE, bivariate = TRUE, ns = ns, coordt = NULL
      )
    }
    sol <- list(
      rowidx = reused$colidx,
      colidx = reused$rowidx,
      lags = reused$lags,
      first = reused$first,
      second = reused$second
    )
  }
  ###    deleting symmetric indexes with associate distances

  if (!length(sol$colidx)) stop("No pairwise contributions were retained. Increase 'p_neighb', 'neighb', or 'maxdist'.", call. = FALSE)

  gb=list(); gb$colidx=sol$colidx;
             gb$rowidx=sol$rowidx ;
             nn=length(gb$colidx)
             gb$numpairs=nn
## loading space time distances in memory   
  mmm=1
if(weighted) { mmm=max(sol$lags)}
  
  
  ss=dotCall64::.C64("SetGlobalVar2",
        SIGNATURE = c("integer","integer",
             "double",
             "integer","double",
              "double","integer","double",
            "integer","integer","integer","integer"),  
       numcoord,  2,  
       sol$lags,
       nn,mmm, 
       1,nn,1,spacetime,bivariate,sol$first,sol$second,
        INTENT =    c(rep("r",12)),
         PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)

} #### end bivariate case

    numpairs <- gb$numpairs
    colidx=gb$rowidx 
    rowidx=gb$colidx
    idx <- 0;ja <- 0;ia <- 0
    isinit <- 1
    nozero <- numpairs/(numcoord*numtime)^2
    idx <- 0;ja  <- 0
}  
############################################## 
# end neighboord case
##############################################
if(is.null(coordt)) coordt=1
}}

########################################################################################
########################################################################################
    ### Returned list of objects:
    return(list(bivariate=bivariate,coordx=coordx,coordy=coordy,coordz=coordz,coordt=coordt,corrmodel=corrmodel,
                colidx = as.integer(colidx), rowidx = as.integer(rowidx),
                data=data,distance=distance,
                error=error,flagcorr=flagcorr,flagnuis=flagnuis,fixed=fixed,likelihood=likelihood,
                lower=paramrange$lower,model=model,n=n,namescorr=namescorr,namesfixed=namesfixed,
                namesnuis=namesnuis,namesparam=namesparam,namessim=namessim,namesstart=namesstart,ns=ns,NS=NS,
                num_betas=num_betas,
                numcoord=numcoord,numcoordx=numcoordx,numcoordy=numcoordy,neighb=K,
                numfixed=numfixed,numpairs=numpairs,numparam=numparam,numparamcorr=numparamcorr,
                numstart=numstart,numtime=numtime,param=param,
                setup=list(                ## setup is a list
                ia=ia,idx=idx,ja=ja,nozero=nozero,tapmodel=tapmodel,tapsep=tapsep),  radius=radius,                            ## with tapered matrix informations
                spacetime=spacetime,srange=srange,start=start,upper=paramrange$upper,type=type,
                trange=trange,weighted=weighted,X=X))
}
## Temporarily configure progressr without leaking handler state to the user session.
.GeoProgressHandlersPush <- function(handler = NULL, global = NULL) {
  old_handlers <- progressr::handlers()
  old_global <- progressr::handlers(global = NA)

  if (!is.null(handler)) progressr::handlers(handler)
  if (!is.null(global)) progressr::handlers(global = isTRUE(global))

  function() {
    try(progressr::handlers(old_handlers), silent = TRUE)
    ## Do not wrap global handler changes in try()/tryCatch():
    ## base::globalCallingHandlers() cannot be modified with handlers on stack.
    progressr::handlers(global = old_global)
    invisible(NULL)
  }
}

GeoScores <- function(data_to_pred,
                      probject = NULL,
                      pred = NULL,
                      mse = NULL,
                      score = c("pe", "crps", "intscore", "coverage"),
                      lower95 = NULL,
                      upper95 = NULL,
                      threshold = 0.5,
                      na.rm = TRUE) {

  score <- match.arg(score,
                     choices = c("brie", "crps", "lscore", "pit", "pe", "intscore", "coverage"),
                     several.ok = TRUE)

  if (!is.logical(na.rm) || length(na.rm) != 1L || is.na(na.rm)) {
    stop("na.rm must be TRUE or FALSE", call. = FALSE)
  }
  if (!is.numeric(threshold) || length(threshold) != 1L || !is.finite(threshold)) {
    stop("threshold must be a finite numeric scalar", call. = FALSE)
  }

  y <- as.numeric(c(data_to_pred))
  from_krig <- !is.null(probject) &&
    (inherits(probject, "GeoKrig") || inherits(probject, "GeoKrigloc"))
  from_simcond <- !is.null(probject) && inherits(probject, "GeoSimcond")

  if (!is.null(probject) && !from_krig && !from_simcond) {
    stop("probject must be an object of class GeoKrig, GeoKrigloc, or GeoSimcond", call. = FALSE)
  }

  if (xor(is.null(lower95), is.null(upper95))) {
    stop("lower95 and upper95 must be supplied together", call. = FALSE)
  }
  has_interval <- !is.null(lower95) && !is.null(upper95)

  check_mask <- function(mask, what) {
    if (!na.rm && !all(mask)) {
      stop("invalid or non-finite values found for ", what, call. = FALSE)
    }
    if (!any(mask)) {
      stop("no valid observations available for ", what, call. = FALSE)
    }
    mask
  }

  ## -----------------------------------------------------------------------
  ## Conditional-simulation objects: use the empirical predictive
  ## distribution represented by condsim. No Gaussian approximation is used.
  ## -----------------------------------------------------------------------
  if (from_simcond) {
    cs <- probject$condsim
    if (is.null(cs)) {
      stop("GeoSimcond object does not contain conditional simulations", call. = FALSE)
    }

    if (is.list(cs)) {
      if (length(cs) < 1L) {
        stop("GeoSimcond object does not contain conditional simulations", call. = FALSE)
      }
      lens <- vapply(cs, function(z) length(c(z)), integer(1L))
      if (any(lens != lens[1L])) {
        stop("conditional simulations have inconsistent lengths", call. = FALSE)
      }
      sim <- do.call(rbind, lapply(cs, function(z) as.numeric(c(z))))
    } else {
      sim <- as.matrix(cs)
      storage.mode(sim) <- "double"
    }

    if (length(dim(sim)) != 2L || nrow(sim) < 1L || ncol(sim) < 1L) {
      stop("GeoSimcond conditional simulations have an invalid shape", call. = FALSE)
    }

    ## GeoSimcond stores replications by rows and prediction locations by
    ## columns. Accept the transposed orientation defensively for external
    ## objects that preserve the same class contract.
    if (ncol(sim) != length(y) && nrow(sim) == length(y)) sim <- t(sim)
    if (ncol(sim) != length(y)) {
      stop("length of data_to_pred and GeoSimcond prediction locations does not match", call. = FALSE)
    }

    n <- ncol(sim)
    B <- nrow(sim)
    finite_draw <- is.finite(sim)
    if (!na.rm && !all(finite_draw)) {
      stop("invalid or non-finite conditional simulations found", call. = FALSE)
    }
    draw_count <- colSums(finite_draw)
    draw_ok <- draw_count > 0L

    if (!is.null(probject$cond_mean) && length(c(probject$cond_mean)) == n) {
      point_pred <- as.numeric(c(probject$cond_mean))
    } else {
      point_pred <- vapply(seq_len(n), function(j) {
        z <- sim[, j]
        z <- z[is.finite(z)]
        if (length(z)) mean(z) else NA_real_
      }, numeric(1L))
    }

    if (has_interval) {
      lower95 <- as.numeric(c(lower95))
      upper95 <- as.numeric(c(upper95))
      if (length(lower95) != n || length(upper95) != n) {
        stop("length of lower95/upper95 and prediction locations does not match", call. = FALSE)
      }
    }

    out <- list()

    if ("pe" %in% score) {
      idx <- check_mask(is.finite(y) & is.finite(point_pred), "point prediction scores")
      err <- y[idx] - point_pred[idx]
      out$MAE <- mean(abs(err))
      out$RMSPE <- sqrt(mean(err^2))
    }

    ## Empirical CRPS in O(B log B) per prediction location:
    ## E|X-y| - 1/(2)E|X-X'| for the empirical predictive distribution.
    if ("crps" %in% score) {
      crps_i <- rep(NA_real_, n)
      valid_y <- is.finite(y) & draw_ok
      if (!na.rm && !all(valid_y)) {
        stop("invalid values found for conditional-simulation CRPS", call. = FALSE)
      }
      for (j in which(valid_y)) {
        z <- sim[, j]
        z <- sort(z[is.finite(z)])
        m <- length(z)
        coeff <- 2 * seq_len(m) - m - 1
        crps_i[j] <- mean(abs(z - y[j])) - sum(coeff * z) / (m * m)
      }
      idx <- check_mask(is.finite(crps_i), "conditional-simulation CRPS")
      out$CRPS <- mean(crps_i[idx])
    }

    if ("pit" %in% score) {
      pit <- rep(NA_real_, n)
      valid_y <- is.finite(y) & draw_ok
      if (!na.rm && !all(valid_y)) {
        stop("invalid values found for conditional-simulation PIT", call. = FALSE)
      }
      for (j in which(valid_y)) {
        z <- sim[, j]
        z <- z[is.finite(z)]
        pit[j] <- mean(z <= y[j])
      }
      if (!any(is.finite(pit))) {
        stop("no valid observations available for conditional-simulation PIT", call. = FALSE)
      }
      out$PIT <- pit
    }

    if ("brie" %in% score) {
      brier_i <- rep(NA_real_, n)
      valid_y <- is.finite(y) & draw_ok
      if (!na.rm && !all(valid_y)) {
        stop("invalid values found for conditional-simulation Brier score", call. = FALSE)
      }
      for (j in which(valid_y)) {
        z <- sim[, j]
        z <- z[is.finite(z)]
        prob <- mean(z > threshold)
        obs <- as.numeric(y[j] > threshold)
        brier_i[j] <- (prob - obs)^2
      }
      idx <- check_mask(is.finite(brier_i), "conditional-simulation Brier score")
      out$Brier <- mean(brier_i[idx])
    }

    if ("intscore" %in% score || "coverage" %in% score) {
      if (has_interval) {
        l <- lower95
        u <- upper95
        interval_ok <- is.finite(y) & is.finite(l) & is.finite(u) & l <= u
      } else {
        l <- rep(NA_real_, n)
        u <- rep(NA_real_, n)
        for (j in which(draw_ok)) {
          z <- sim[, j]
          z <- z[is.finite(z)]
          qq <- as.numeric(stats::quantile(z, probs = c(0.025, 0.975),
                                           names = FALSE, type = 7))
          l[j] <- qq[1L]
          u[j] <- qq[2L]
        }
        interval_ok <- is.finite(y) & is.finite(l) & is.finite(u) & l <= u
      }

      idx <- check_mask(interval_ok, "conditional-simulation prediction-interval scores")
      yi <- y[idx]
      li <- l[idx]
      ui <- u[idx]
      alpha <- 0.05
      if ("intscore" %in% score) {
        out$IS95 <- mean((ui - li) +
                           (2 / alpha) * (li - yi) * (yi < li) +
                           (2 / alpha) * (yi - ui) * (yi > ui))
      }
      if ("coverage" %in% score) {
        out$Cvg95 <- mean(yi >= li & yi <= ui)
      }
    }

    if ("lscore" %in% score) {
      out$LogScore <- NA_real_
      warning(
        "LogScore is not computed automatically from a GeoSimcond object because conditional draws do not by themselves provide the predictive density at the observed value. Other requested simulation-based scores were computed.",
        call. = FALSE
      )
    }

    return(out)
  }

  ## -----------------------------------------------------------------------
  ## GeoKrig/GeoKrigloc or direct pred/mse input. Distributional scores are
  ## always computed from the Gaussian N(pred, mse) convention when requested.
  ## For a non-Gaussian GeoKrig object this is a Gaussian predictive
  ## approximation based on the OLP and its MSE; interpretation is left to the
  ## user. For a Gaussian model it is the natural conditional distribution.
  ## -----------------------------------------------------------------------
  if (from_krig) {
    pred <- as.numeric(c(probject$pred))
    mse <- if (is.null(probject$mse)) NULL else as.numeric(c(probject$mse))
  } else {
    if (is.null(pred)) {
      stop("GeoKrig/GeoKrigloc/GeoSimcond object or pred is mandatory", call. = FALSE)
    }
    pred <- as.numeric(c(pred))
    if (!is.null(mse)) mse <- as.numeric(c(mse))
  }

  if (length(y) != length(pred)) {
    stop("length of data_to_pred and pred does not match", call. = FALSE)
  }
  n <- length(pred)

  if (has_interval) {
    lower95 <- as.numeric(c(lower95))
    upper95 <- as.numeric(c(upper95))
    if (length(lower95) != n || length(upper95) != n) {
      stop("length of lower95/upper95 and pred does not match", call. = FALSE)
    }
  }

  point_requested <- "pe" %in% score
  distribution_scores <- intersect(score, c("brie", "crps", "lscore", "pit"))
  interval_scores <- intersect(score, c("intscore", "coverage"))
  needs_distribution <- length(distribution_scores) > 0L
  needs_interval <- length(interval_scores) > 0L
  needs_gaussian_distribution <- needs_distribution || (needs_interval && !has_interval)

  se <- NULL
  if (needs_gaussian_distribution) {
    if (!is.null(mse)) {
      if (length(mse) != n) stop("length of mse and pred does not match", call. = FALSE)
      se <- suppressWarnings(sqrt(mse))
    } else if (has_interval) {
      se <- (upper95 - lower95) / (2 * qnorm(0.975))
      mse <- se^2
    } else {
      stop("mse or both lower95 and upper95 are mandatory for Gaussian probabilistic scores", call. = FALSE)
    }
  }

  point_ok <- is.finite(y) & is.finite(pred)
  distribution_ok <- point_ok
  if (needs_gaussian_distribution) {
    distribution_ok <- distribution_ok & is.finite(se) & se > 0 &
      is.finite(mse) & mse > 0
  }
  interval_ok <- point_ok
  if (needs_interval && has_interval) {
    interval_ok <- interval_ok & is.finite(lower95) & is.finite(upper95) &
      lower95 < upper95
  } else if (needs_interval) {
    interval_ok <- distribution_ok
  }

  out <- list()

  if (point_requested) {
    idx <- check_mask(point_ok, "point prediction scores")
    err <- y[idx] - pred[idx]
    out$MAE <- mean(abs(err))
    out$RMSPE <- sqrt(mean(err^2))
  }

  if (needs_distribution) {
    idx <- check_mask(distribution_ok, "Gaussian predictive-distribution scores")
    yd <- y[idx]
    pd <- pred[idx]
    sed <- se[idx]
    msed <- mse[idx]
    errd <- yd - pd
    std <- errd / sed

    if ("brie" %in% score) {
      probs <- 1 - pnorm(threshold, mean = pd, sd = sed)
      obs <- as.numeric(yd > threshold)
      out$Brier <- mean((probs - obs)^2)
    }

    if ("lscore" %in% score) {
      out$LogScore <- mean(0.5 * (log(2 * pi * msed) + std^2))
    }

    if ("pit" %in% score) {
      out$PIT <- pnorm(yd, mean = pd, sd = sed)
    }

    if ("crps" %in% score) {
      out$CRPS <- mean(
        sed * (std * (2 * pnorm(std) - 1) +
                 2 * dnorm(std) - 1 / sqrt(pi))
      )
    }
  }

  if (needs_interval) {
    idx <- check_mask(interval_ok, "prediction-interval scores")
    yi <- y[idx]

    if (has_interval) {
      l <- lower95[idx]
      u <- upper95[idx]
    } else {
      sei <- se[idx]
      pi <- pred[idx]
      alpha <- 0.05
      l <- pi + sei * qnorm(alpha / 2)
      u <- pi + sei * qnorm(1 - alpha / 2)
    }

    alpha <- 0.05
    if ("intscore" %in% score) {
      out$IS95 <- mean((u - l) +
                         (2 / alpha) * (l - yi) * (yi < l) +
                         (2 / alpha) * (yi - u) * (yi > u))
    }
    if ("coverage" %in% score) {
      out$Cvg95 <- mean(yi >= l & yi <= u)
    }
  }

  out
}

######################################################################################################
######################################################################################################
######################################################################################################
######################################################################################################


### Internal cache environment for covariance utilities
if (!exists(".GeoModels_env", inherits = FALSE)) {
  .GeoModels_env <- new.env(parent = emptyenv())
}

.GeoModels_get_cache <- function(name) {
  if (is.null(.GeoModels_env[[name]]) ||
      !is.environment(.GeoModels_env[[name]])) {
    .GeoModels_env[[name]] <- new.env(parent = emptyenv())
  }
  .GeoModels_env[[name]]
}


### Observed-scale correlation for the constructive Weibull random field.
###
### For shape = kappa, GeoModels uses the correlation from Eq. (6) of the
### Weibull-process paper,
###
###   C(kappa) * { 2F1(-1/kappa,-1/kappa;1;rho^2) - 1 }.
###
### The exact endpoint rho^2 = 1 is one.  Evaluating hypergeo::hypergeo()
### at that boundary is unnecessary and, for some shapes (notably kappa=2),
### can make the numerical backend attempt a divergent integral.  Handle the
### endpoints analytically and evaluate the hypergeometric function only for
### 0 < rho^2 < 1.
.GeoWeibullCorr <- function(rho, shape) {
  shape <- suppressWarnings(as.numeric(shape))
  if (length(shape) != 1L || !is.finite(shape) || shape <= 0)
    stop("Weibull shape must be a finite positive scalar.", call. = FALSE)

  rho <- as.numeric(rho)
  if (!length(rho)) return(numeric())
  if (any(!is.finite(rho)))
    stop("Weibull latent correlations must be finite.", call. = FALSE)

  ## Correlations supplied by the covariance core should lie in [-1,1].
  ## Clamp only round-off excursions before squaring.
  rho <- pmax(-1, pmin(1, rho))
  z <- rho * rho

  out <- numeric(length(z))
  at_zero <- (z == 0)
  at_one <- (z == 1)
  mid <- !(at_zero | at_one)

  out[at_zero] <- 0
  out[at_one] <- 1

  if (any(mid)) {
    ## C(kappa) = Gamma(1+1/kappa)^2 /
    ##            {Gamma(1+2/kappa)-Gamma(1+1/kappa)^2}.
    ## Compute it through lgamma/expm1 to avoid avoidable overflow.
    delta <- lgamma(1 + 2 / shape) - 2 * lgamma(1 + 1 / shape)
    denom_ratio <- expm1(delta)
    if (!is.finite(denom_ratio) || denom_ratio <= 0)
      stop("Invalid Weibull correlation normalizing constant.", call. = FALSE)

    factor <- 1 / denom_ratio
    hz <- Re(hypergeo::hypergeo(-1 / shape, -1 / shape, 1, z[mid]))
    out[mid] <- factor * (hz - 1)
  }

  ## Protect covariance matrices against tiny numerical overshoots while
  ## retaining the exact theoretical endpoints.
  out[abs(out) < 32 * .Machine$double.eps] <- 0
  out[abs(out - 1) < 64 * .Machine$double.eps] <- 1
  pmax(0, pmin(1, out))
}

### decomposition of a square  matrix
MatDecomp <- function(mtx, method) {
  # Controlli preliminari veloci
  if (!is.matrix(mtx) || anyNA(mtx) || any(!is.finite(mtx))) return(FALSE)
  
  if (method == "cholesky") {
    if (!isSymmetric(mtx, tol = 1e-8)) return(FALSE)
    mat.decomp <- tryCatch(
      FastGP::rcppeigen_get_chol(mtx),
      error = function(e) FALSE
    )
    if (isFALSE(mat.decomp)) return(FALSE)
    d <- diag(mat.decomp)
    if (any(d <= 0, na.rm = TRUE)) return(FALSE) 
  } else if (method == "svd") {
    mat.decomp <- tryCatch(
      svd(mtx),
      error = function(e) FALSE
    )
    if (isFALSE(mat.decomp)) return(FALSE)
    d <- mat.decomp$d
    if (any(d <= 0, na.rm = TRUE)) return(FALSE)
    if (!is.finite(sum(log(d)))) return(FALSE)
  }
  mat.decomp
}

MatSqrt <- function(mat.decomp, method) {
  if (method == "cholesky") return(mat.decomp)
  if (method == "svd") {
    d_sqrt <- sqrt(mat.decomp$d)
    return(t(t(mat.decomp$v) * d_sqrt))
  }
  stop("Invalid method")
}
MatInv <- function(mtx) {
  # Aggiunto controllo preliminare
  if (!is.matrix(mtx) || anyNA(mtx)) return(FALSE)
  tryCatch(
    FastGP::rcppeigen_invert_matrix(mtx),
    error = function(e) FALSE
  )
}
MatLogDet <- function(mat.decomp, method) {
  if (method == "cholesky") {
    d <- diag(mat.decomp)
    if (any(d <= 0, na.rm = TRUE)) return(NA)
    2 * sum(log(d))
    
  } else if (method == "svd") {
    d <- mat.decomp$d
    if (any(d <= 0, na.rm = TRUE)) return(NA)
    sum(log(d))
    
  } else {
    NA
  }
}

# utility function for geokrig
getInvC <- function(covmatrix, CC, mse = TRUE) {
  if (!covmatrix$sparse) {
    U <- tryCatch({
      FastGP::rcppeigen_get_chol(covmatrix$covmatrix)
    }, error = function(e) {
      stop("Covariance matrix is not positive definite")
    })
    vec <- forwardsolve(U, CC)
    Invc <- forwardsolve(U, vec, transpose = TRUE)
    mse_val <- if (mse) crossprod(vec) else NULL
  } else {
    cc <- if (spam::is.spam(covmatrix$covmatrix)) covmatrix$covmatrix else spam::as.spam(covmatrix$covmatrix)
    U <- tryCatch({
      spam::chol.spam(cc)
    }, error = function(e) {
      stop("Covariance matrix is not positive definite")
    })
    vec <- spam::forwardsolve(U, CC)
    Invc <- spam::backsolve(U, vec)
    #mse_val <- if (mse) as.matrix(spam::crossprod.spam(vec)) else NULL
    mse_val <- if (mse) as.matrix(crossprod(vec)) else NULL
    
  }
  list(a = Invc, b = mse_val)
}
  
######################################################################################################
######################################################################################################
######################################################################################################
######################################################################################################
.GeoTukeyh2StandardMoments <- function(tail1, tail2) {
  hr <- suppressWarnings(as.numeric(tail1))[1L]
  hl <- suppressWarnings(as.numeric(tail2))[1L]
  if (!is.finite(hr) || !is.finite(hl) || hr < 0 || hl < 0 ||
      hr >= 0.5 || hl >= 0.5) {
    stop("Tukeyh2 tail1 and tail2 must lie in [0, 0.5).", call. = FALSE)
  }
  mm <- (hr - hl) / (sqrt(2 * pi) * (1 - hl) * (1 - hr))
  vv <- 0.5 * (1 - 2 * hl)^(-1.5) +
        0.5 * (1 - 2 * hr)^(-1.5) - mm^2
  list(mean = mm, variance = vv)
}

.GeoTukeyh2Correlation <- function(rho, tail1, tail2) {
  mom <- .GeoTukeyh2StandardMoments(tail1, tail2)
  hr <- as.numeric(tail1)[1L]
  hl <- as.numeric(tail2)[1L]
  rho <- as.numeric(rho)
  if (!length(rho)) return(numeric(0))
  out <- rep(NA_real_, length(rho))
  good <- is.finite(rho) & !is.na(rho)
  if (!any(good)) return(out)

  rr0 <- pmax(-1, pmin(1, rho[good]))
  ans <- numeric(length(rr0))
  zero <- abs(rr0) < 1e-15
  one <- abs(rr0 - 1) < 1e-15
  work <- !(zero | one)
  ans[zero] <- 0
  ans[one] <- 1

  if (any(work)) {
    eps <- 1e-12
    rr <- pmax(-1 + 1e-12, pmin(1 - 1e-12, rr0[work]))
    x1 <- 1 - (1 - rr^2) * hr
    y1 <- 1 - (1 - rr^2) * hl
    x2 <- pmax((1 - hr)^2 - (rr * hr)^2, eps)
    y2 <- pmax((1 - hl)^2 - (rr * hl)^2, eps)
    gg <- pmax(1 - hl - hr + (1 - rr^2) * hl * hr, eps)

    ax1 <- pmin(1, pmax(-1, rr / x1))
    ay1 <- pmin(1, pmax(-1, rr / y1))
    x1y1 <- pmax(x1 * y1, eps)
    axy <- pmin(1, pmax(-1, rr / sqrt(x1y1)))

    h1 <- sqrt(pmax(1 - rr^2 / x1^2, 0)) + (rr / x1) * asin(ax1)
    h2 <- sqrt(pmax(1 - rr^2 / y1^2, 0)) + (rr / y1) * asin(ay1)
    h3 <- sqrt(pmax(1 - rr^2 / x1y1, 0)) +
          (rr / sqrt(x1y1)) * asin(axy)

    p1 <- x1 * h1 / (2 * pi * x2^(3/2)) + rr / (4 * x2^(3/2))
    p2 <- y1 * h2 / (2 * pi * y2^(3/2)) + rr / (4 * y2^(3/2))
    p3 <- -sqrt(x1y1) * h3 / (2 * pi * gg^(3/2)) + rr / (4 * gg^(3/2))
    ans[work] <- (p1 + p2 + 2 * p3 - mom$mean^2) / mom$variance
  }
  ans <- pmax(-1, pmin(1, ans))
  out[good] <- ans
  out
}

.GeoSASStandardMoments <- function(skew, tail) {
  e <- suppressWarnings(as.numeric(skew))[1L]
  d <- suppressWarnings(as.numeric(tail))[1L]
  if (!is.finite(e) || !is.finite(d) || d <= 0)
    stop("SinhAsinh requires finite skew and a positive finite tail.", call. = FALSE)
  mm <- sinh(e / d) * exp(0.25) *
    (besselK(0.25, (d + 1) / (2 * d)) +
       besselK(0.25, abs(1 - d) / (2 * d))) / sqrt(8 * pi)
  vv <- cosh(2 * e / d) * exp(0.25) *
    (besselK(0.25, (d + 2) / (2 * d)) +
       besselK(0.25, abs(2 - d) / (2 * d))) / sqrt(32 * pi) - 0.5 - mm^2
  if (!is.finite(vv) || vv <= 0)
    stop("SinhAsinh marginal variance is not finite and positive.", call. = FALSE)
  list(mean = mm, variance = vv)
}

.GeoPlainTransformedMarginalMoments <- function(model_type, params, eta = 0) {
  model_type <- as.integer(model_type)
  eta <- as.numeric(eta)
  if (!length(eta) || any(!is.finite(eta)))
    stop("Marginal locations must be finite.", call. = FALSE)
  sill <- suppressWarnings(as.numeric(params["sill"]))[1L]
  if (!is.finite(sill) || sill <= 0)
    stop("sill must be finite and positive.", call. = FALSE)

  if (model_type == 34L) {
    h <- suppressWarnings(as.numeric(params["tail"]))[1L]
    if (!is.finite(h) || h < 0 || h >= 0.5)
      stop("Tukeyh tail must lie in [0, 0.5).", call. = FALSE)
    return(list(mean = eta,
                variance = rep(sill * (1 - 2 * h)^(-1.5), length(eta))))
  }
  if (model_type == 40L) {
    mom <- .GeoTukeyh2StandardMoments(params["tail1"], params["tail2"])
    return(list(mean = eta + sqrt(sill) * mom$mean,
                variance = rep(sill * mom$variance, length(eta))))
  }
  if (model_type == 20L) {
    mom <- .GeoSASStandardMoments(params["skew"], params["tail"])
    return(list(mean = eta + sqrt(sill) * mom$mean,
                variance = rep(sill * mom$variance, length(eta))))
  }
  stop("Unsupported plain transformed model id ", model_type, ".", call. = FALSE)
}

.GeoPlainTransformedCrossCov <- function(rho, model_type, params) {
  model_type <- as.integer(model_type)
  rho <- as.numeric(rho)
  sill <- suppressWarnings(as.numeric(params["sill"]))[1L]
  if (!is.finite(sill) || sill <= 0)
    stop("sill must be finite and positive.", call. = FALSE)

  if (model_type == 34L) {
    h <- suppressWarnings(as.numeric(params["tail"]))[1L]
    if (!is.finite(h) || h < 0 || h >= 0.5)
      stop("Tukeyh tail must lie in [0, 0.5).", call. = FALSE)
    if (h == 0) return(sill * rho)
    v0 <- (1 - 2 * h)^(-1.5)
    cc <- rho * (1 - 2 * h)^(1.5) /
      pmax((1 - h)^2 - (h * rho)^2, .Machine$double.eps)^(1.5)
    return(sill * v0 * cc)
  }
  if (model_type == 40L) {
    mom <- .GeoTukeyh2StandardMoments(params["tail1"], params["tail2"])
    return(sill * mom$variance *
             .GeoTukeyh2Correlation(rho, params["tail1"], params["tail2"]))
  }
  if (model_type == 20L) {
    ## `params` may be either a named numeric vector or a list.  Using
    ## params["skew"] / params["tail"] preserves list type in the latter
    ## case, while corrsas() deliberately requires numeric scalar inputs.
    ## Normalize once here so plain and Gaussian-copula SAS kriging share
    ## exactly the same covariance implementation.
    skew <- suppressWarnings(as.numeric(params[["skew"]]))[1L]
    tail <- suppressWarnings(as.numeric(params[["tail"]]))[1L]
    mom <- .GeoSASStandardMoments(skew, tail)
    return(sill * mom$variance * corrsas(rho, skew, tail))
  }
  stop("Unsupported plain transformed model id ", model_type, ".", call. = FALSE)
}

corrsas <- function(corr, skew, tail, max_coeff = NULL) {
    d <- tail
    e <- skew
    # ========================================================================
    if (!is.numeric(skew) || !is.numeric(tail)) {
        stop("skew and tail must be numeric")
    }
    if (length(skew) != 1 || length(tail) != 1) {
        stop("skew and tail must be scalars")
    }
    if (tail <= 0) {
        stop("tail must be positive")
    }
    if (!is.numeric(corr)) {
        stop("corr must be numeric")
    }
    if (length(corr) == 0) {
        return(numeric(0))
    }
    ## Exact Gaussian identity and independence do not require the
    ## Hermite/integration machinery below.
    if (abs(skew) < .Machine$double.eps && abs(tail - 1) < .Machine$double.eps)
        return(as.numeric(corr))
    if (all(corr == 0)) return(rep(0, length(corr)))
    
    # Determinazione automatica max_coeff
    if (is.null(max_coeff)) {
        stability_indicator <- abs(e/d) + abs(d - 1) + max(abs(corr))
        is_stable <- stability_indicator < 2.0
        
        if (is_stable) {
            base_terms <- 6
            adjustment <- min(ceiling(stability_indicator), 6)
            max_coeff <- base_terms + adjustment
        } else {
            base_terms <- 8
            adjustment <- min(ceiling(stability_indicator - 2), 6)
            max_coeff <- base_terms + adjustment
        }
        max_coeff <- min(max_coeff, 12)
        max_coeff <- max(max_coeff, 6)
    }
    
    # Pre-calcolo di costanti
    sqrt_8pi <- sqrt(8 * pi)
    sqrt_32pi <- sqrt(32 * pi)
    sqrt_2pi <- sqrt(2 * pi)
    exp_0.25 <- exp(0.25)
    
    # ========================================================================
    # ========================================================================
    compute_bessel_terms <- function(d) {
        tryCatch({
            k1 <- besselK(0.25, (d + 1)/(2*d))
            k2 <- besselK(0.25, abs(1 - d)/(2*d))
            k3 <- besselK(0.25, (d + 2)/(2*d))
            k4 <- besselK(0.25, abs(2 - d)/(2*d))
            if (any(!is.finite(c(k1, k2, k3, k4)))) {return(NULL)}
            list(k1 = k1, k2 = k2, k3 = k3, k4 = k4)
        }, error = function(e) {
            warning("Bessel function computation failed: ", e$message)
            NULL
        })
    }

    bessel_vals <- compute_bessel_terms(d)
    if (is.null(bessel_vals)) {
        return(rep(NA_real_, length(corr)))
    }
    
    #
    besselK_1 <- bessel_vals$k1
    besselK_2 <- bessel_vals$k2
    besselK_3 <- bessel_vals$k3
    besselK_4 <- bessel_vals$k4
    
    # 
    if (abs(e/d) < 0.01) {
        # Per valori molto piccoli, usa serie di Taylor
        x <- e/d
        sinh_e_d <- x * (1 + x^2/6)
        cosh_2e_d <- 1 + 2*x^2 * (1 + x^2/3)
    } else {
        sinh_e_d <- sinh(e/d)
        cosh_2e_d <- cosh(2*e/d)
    }
    
    mm <- sinh_e_d * exp_0.25 * (besselK_1 + besselK_2) / sqrt_8pi
    vv <- cosh_2e_d * exp_0.25 * (besselK_3 + besselK_4) / sqrt_32pi - 0.5 - mm^2
    
    if (!is.finite(vv) || abs(vv) < .Machine$double.eps) {
        return(rep(NA_real_, length(corr)))
    }
    
    # Pre-calcolo di j_vec e gamma terms
    j_vec <- 1:max_coeff
    gamma_j_plus_1 <- gamma(j_vec + 1)
    # ========================================================================
    integrand <- function(z, alpha, kappa, j, r) {
        z_sq <- z^2
        if (j - 2*r == 0) {
            z_pow <- 1
        } else {
            z_pow <- z^(j - 2*r)
        }
        aa <- z + sqrt(z_sq + 1)
        exp_term <- exp(-z_sq/2 + alpha/kappa)
        exp_term * z_pow * (aa^(1/kappa) - exp(-2*alpha/kappa) * aa^(-1/kappa))
    }
    # ========================================================================
    # ========================================================================
    II_cache <- .GeoModels_get_cache("II_global_cache")
    MAX_CACHE_SIZE <- 5000
    # 
    cache_size <- length(names(II_cache))
    if (cache_size >= MAX_CACHE_SIZE) {
        # Rimuovi metà degli elementi (strategia semplice)
        keys_to_remove <- sample(names(II_cache), floor(MAX_CACHE_SIZE / 2))
        rm(list = keys_to_remove, envir = II_cache)
    }
    
    # =======================================================================
    # ========================================================================
    II <- function(alpha, kappa, j, r) {
        key <- paste(round(alpha, 8), round(kappa, 8), j, r, sep = "_")
        if (exists(key, envir = II_cache)) {
            return(get(key, envir = II_cache))
        }
        val <- tryCatch({
            # Limiti finiti: l'integranda decade come exp(-z²/2)
            # Oltre ±6 il contributo è trascurabile (< 1e-9)
            integrate(integrand, 
                     lower = -6, 
                     upper = 6, 
                     alpha = alpha, 
                     kappa = kappa, 
                     j = j, 
                     r = r,
                     rel.tol = 1e-7,
                     abs.tol = 1e-9,
                     subdivisions = 150,
                     stop.on.error = FALSE)$value
        }, error = function(e) {
            warning("Integration failed for key ", key, ": ", e$message, call. = FALSE)
            0
        })
        # 
        if (is.finite(val)) {
            assign(key, val, envir = II_cache)
        } else {
            val <- 0
        }
        
        val
    }
    kahan_sum <- function(x) {
        if (length(x) == 0) return(0)
        s <- 0
        c <- 0
        for (xi in x) {
            y <- xi - c
            t <- s + y
            c <- (t - s) - y
            s <- t
        }
        s
    }
    # ========================================================================
    # ========================================================================
    coeffs <- numeric(max_coeff)
    for (j in j_vec) {
        max_r <- floor(j/2)
        if (max_r < 0) {
            coeffs[j] <- 0
            next
        }
        
        rr <- 0:max_r
        II_vals <- sapply(rr, function(r_val) II(e, d, j, r_val))
        gamma_r_plus_1 <- gamma(rr + 1)
        gamma_j_minus_2r_plus_1 <- gamma(j - 2*rr + 1)
        valid_gamma <- is.finite(gamma_r_plus_1) & is.finite(gamma_j_minus_2r_plus_1) & 
                      gamma_r_plus_1 > 0 & gamma_j_minus_2r_plus_1 > 0
        if (any(valid_gamma)) {
            terms <- II_vals[valid_gamma] * (-1)^rr[valid_gamma] / 
                    (2^(rr[valid_gamma] + 1) * gamma_r_plus_1[valid_gamma] * 
                     gamma_j_minus_2r_plus_1[valid_gamma])
            
            if (is.finite(gamma_j_plus_1[j]) && gamma_j_plus_1[j] > 0) {
                coeffs[j] <- gamma_j_plus_1[j] * kahan_sum(terms) / sqrt_2pi
            }
        }
    }
    valid_idx <- is.finite(coeffs) & !is.na(coeffs) & coeffs != 0
    if (!any(valid_idx)) {
        return(rep(NA_real_, length(corr)))
    }
    coeffs <- coeffs[valid_idx]
    j_vec_valid <- j_vec[valid_idx]
    gamma_terms_valid <- gamma_j_plus_1[valid_idx]
    coeffs_sq <- coeffs^2
    
    # ========================================================================
    # ========================================================================
    corrsas_inner <- function(rho) {
        if (!is.finite(rho) || is.na(rho)) return(NA_real_)
        if (rho == 0) return(0)
        if (rho == 1) return(1)
        
        # Gestione overflow con log-scale per correlazioni alte
        max_j <- max(j_vec_valid)
        if (abs(rho) > 0.99 && max_j > 50) {
            log_rho <- log(abs(rho))
            if (max_j * abs(log_rho) > 700) return(NA_real_)
            
            # Calcolo in log-scale
            log_terms <- 2*log(abs(coeffs)) + j_vec_valid * log_rho - log(gamma_terms_valid)
            max_log <- max(log_terms[is.finite(log_terms)])
            if (!is.finite(max_log)) return(NA_real_)
            
            terms_scaled <- exp(log_terms - max_log)
            numerator <- exp(max_log) * kahan_sum(terms_scaled)
        } else {
            rho_powers <- rho^j_vec_valid
            if (any(is.infinite(rho_powers))) return(NA_real_) 
            numerator <- kahan_sum(coeffs_sq * rho_powers / gamma_terms_valid)
        }
        result <- numerator / vv
        if (is.finite(result)) result else NA_real_
    }
    
    # ========================================================================
    # ========================================================================
    if (length(corr) == 1) {
        return(corrsas_inner(corr))
    } else {
        return(vapply(corr, corrsas_inner, numeric(1)))
    }
}



.GeoCovCanonicalModel <- function(model) {
  if(!is.character(model) || length(model) != 1L || is.na(model)) return(model)
  model <- gsub("[[:blank:]]", "", model)
  aliases <- c(
    Gauss = "Gaussian", SkewGauss = "SkewGaussian", LogGauss = "LogGaussian",
    TwoPieceGauss = "TwoPieceGaussian", TwopieceGauss = "TwoPieceGaussian",
    TwopieceT = "TwoPieceStudentT", TwopieceTukeyh = "TwoPieceTukeyh",
    Loglogistic = "LogLogistic",
    poisson = "Poisson", poissongamma = "PoissonGamma",
    Gaussian_misp_StudentT = "StudentT",
    Gaussian_misp_Poisson = "Poisson",
    Gaussian_misp_SkewStudentT = "SkewStudentT",
    Gaussian_misp_Tukeygh = "Tukeygh",
    Gaussian_misp_PoissonZIP = "PoissonZIP",
    Gaussian_misp_PoissonZIP1 = "PoissonZIP1",
    Gaussian_misp_PoissonGamma = "PoissonGamma",
    Gaussian_misp_Binomial = "Binomial",
    Gaussian_misp_BinomialNeg = "BinomialNeg",
    Binary_misp_BinomialNeg = "BinomialNeg"
  )
  if(model %in% names(aliases)) unname(aliases[[model]]) else model
}

######################################################################################################
############## functions for gaussiam copula #########################################################
######################################################################################################
# marginal variances
variance_disp <- function(model_type, params) {
  model_type <- as.integer(model_type)
  eta <- suppressWarnings(as.numeric(params["mean"]))
  if (!length(eta) || !is.finite(eta[1L])) eta <- 0
  eta <- eta[1L]

  switch(as.character(model_type),
    "1"  = as.numeric(params["sill"]),  # Gaussian
    "12" = {
      df_t <- 1 / as.numeric(params["df"])
      as.numeric(params["sill"]) * df_t / (df_t - 2)
    },  # StudentT
    "22" = exp(2 * eta) * (exp(as.numeric(params["sill"])) - 1),  # LogGaussian
    "30" = exp(eta),  # Poisson
    "21" = 2 * exp(2 * eta) / as.numeric(params["shape"]),  # Gamma
    "26" = exp(2 * eta) * (gamma(1 + 2 / as.numeric(params["shape"])) /
      (gamma(1 + 1 / as.numeric(params["shape"]))^2) - 1),  # Weibull
    "11" = {
      n <- as.numeric(params["n"])
      pp <- pnorm(eta)
      n * pp * (1 - pp)
    },  # Binomial
    "16" = {
      n <- as.numeric(params["n"])
      pp <- pnorm(eta)
      n * (1 - pp) / pp^2
    },  # BinomialNeg
    "28" = {
      a <- as.numeric(params["shape1"]); b <- as.numeric(params["shape2"])
      rr <- as.numeric(params["max"] - params["min"])
      rr^2 * a * b / ((a + b)^2 * (a + b + 1))
    },  # Beta
    "50" = {
      m <- plogis(eta); phi <- as.numeric(params["shape"])
      rr <- as.numeric(params["max"] - params["min"])
      rr^2 * m * (1 - m) / (phi + 1)
    },  # Beta2
    "33" = {
      a <- as.numeric(params["shape1"]); b <- as.numeric(params["shape2"])
      rr <- as.numeric(params["max"] - params["min"])
      m1 <- b * beta(1 + 1 / a, b)
      m2 <- b * beta(1 + 2 / a, b)
      rr^2 * (m2 - m1^2)
    },  # Kumaraswamy
    "42" = {
      a <- as.numeric(params["shape"]); med <- plogis(eta)
      b <- log(0.5) / log1p(-med^a)
      rr <- as.numeric(params["max"] - params["min"])
      m1 <- b * beta(1 + 1 / a, b)
      m2 <- b * beta(1 + 2 / a, b)
      rr^2 * (m2 - m1^2)
    },  # Kumaraswamy2
    "25" = pi^2 * as.numeric(params["sill"]) / 3,  # Logistic
    "59" = {
      sk <- as.numeric(params["skew"]); ss <- as.numeric(params["sill"])
      ss * (1 - 2 * sk + 2 * sk^2) / (sk^2 * (1 - sk)^2)
    },  # SkewLaplace
    "18" = {
      df <- 1 / as.numeric(params["df"])
      skew <- as.numeric(params["skew"])
      mu <- as.numeric(params["mean"])
      mu * ((df / (df - 2)) * (1 + skew^2) -
              df * skew^2 * gamma(0.5 * (df - 1)) / (pi * gamma(0.5 * df)))
    },  # legacy SkewStudentT expression retained for compatibility
    "10" = as.numeric(params["sill"]) + as.numeric(params["skew"])^2 * (1 - 2 / pi),
    "34" = {
      h <- suppressWarnings(as.numeric(params["tail"]))
      if (!length(h) || is.na(h)) h <- suppressWarnings(as.numeric(params["h"]))
      if (!is.finite(h) || h < 0 || h >= 0.5)
        stop("Tukeyh tail must lie in [0, 0.5).", call. = FALSE)
      as.numeric(params["sill"]) * (1 - 2 * h)^(-1.5)
    },
    "40" = {
      mom <- .GeoTukeyh2StandardMoments(params["tail1"], params["tail2"])
      as.numeric(params["sill"]) * mom$variance
    },
    "20" = {
      mom <- .GeoSASStandardMoments(params["skew"], params["tail"])
      as.numeric(params["sill"]) * mom$variance
    },
    stop(paste("Model not supported:", model_type))
  )
}

# Quantile of the observed margin evaluated at a standard-uniform probability.
# The location parameter is taken from params["mean"].  This function is used
# by the Hermite representation of Gaussian-copula covariances.
quantile_disp <- function(p, model_type, params) {
  model_type <- as.integer(model_type)
  eta <- suppressWarnings(as.numeric(params["mean"]))
  if (!length(eta) || !is.finite(eta[1L])) eta <- 0
  eta <- eta[1L]
  p <- pmin(pmax(as.numeric(p), .Machine$double.eps), 1 - .Machine$double.eps)

  switch(as.character(model_type),
    "1" = qnorm(p, mean = eta, sd = sqrt(as.numeric(params["sill"]))),
    "12" = eta + sqrt(as.numeric(params["sill"])) *
      qt(p, df = 1 / as.numeric(params["df"])),
    "22" = qlnorm(p, meanlog = eta - as.numeric(params["sill"]) / 2,
                   sdlog = sqrt(as.numeric(params["sill"]))),
    "30" = qpois(p, lambda = exp(eta)),
    "21" = exp(eta) * qgamma(p, shape = as.numeric(params["shape"]) / 2,
                              rate = as.numeric(params["shape"]) / 2),
    "26" = exp(eta) * qweibull(
      p, shape = as.numeric(params["shape"]),
      scale = 1 / gamma(1 + 1 / as.numeric(params["shape"]))
    ),
    "11" = {
      n <- as.numeric(params["n"])
      qbinom(p, size = n, prob = pnorm(eta))
    },
    "16" = {
      n <- as.numeric(params["n"])
      qnbinom(p, size = n, prob = pnorm(eta))
    },
    "28" = {
      lo <- as.numeric(params["min"]); hi <- as.numeric(params["max"])
      lo + (hi - lo) * qbeta(p, shape1 = as.numeric(params["shape1"]),
                             shape2 = as.numeric(params["shape2"]))
    },
    "50" = {
      lo <- as.numeric(params["min"]); hi <- as.numeric(params["max"])
      m <- plogis(eta); phi <- as.numeric(params["shape"])
      lo + (hi - lo) * qbeta(p, shape1 = m * phi, shape2 = (1 - m) * phi)
    },
    "33" = {
      lo <- as.numeric(params["min"]); hi <- as.numeric(params["max"])
      a <- as.numeric(params["shape1"]); b <- as.numeric(params["shape2"])
      lo + (hi - lo) * (1 - (1 - p)^(1 / b))^(1 / a)
    },
    "42" = {
      lo <- as.numeric(params["min"]); hi <- as.numeric(params["max"])
      a <- as.numeric(params["shape"]); med <- plogis(eta)
      b <- log(0.5) / log1p(-med^a)
      lo + (hi - lo) * (1 - (1 - p)^(1 / b))^(1 / a)
    },
    "25" = qlogis(p, location = eta, scale = sqrt(as.numeric(params["sill"]))),
    "59" = {
      sk <- as.numeric(params["skew"]); ss <- sqrt(as.numeric(params["sill"]))
      out <- numeric(length(p)); right <- p >= sk
      out[right] <- eta - (ss / sk) *
        (log1p(-p[right]) - log1p(-sk))
      out[!right] <- eta + (ss / (1 - sk)) *
        (log(p[!right]) - log(sk))
      out
    },
    "18" = eta + sqrt(as.numeric(params["sill"])) *
      sn::qst(p, xi = 0, omega = 1, alpha = as.numeric(params["skew"]),
              nu = 1 / as.numeric(params["df"])),
    "10" = eta + sqrt(as.numeric(params["sill"])) *
      sn::qsn(p, xi = 0,
              omega = sqrt((as.numeric(params["skew"])^2 + as.numeric(params["sill"])) /
                             as.numeric(params["sill"])),
              alpha = as.numeric(params["skew"]) / sqrt(as.numeric(params["sill"]))),
    "34" = {
      z <- qnorm(p)
      h <- suppressWarnings(as.numeric(params["tail"]))
      if (!length(h) || is.na(h)) h <- suppressWarnings(as.numeric(params["h"]))
      if (!is.finite(h) || h < 0 || h >= 0.5)
        stop("Tukeyh tail must lie in [0, 0.5).", call. = FALSE)
      eta + sqrt(as.numeric(params["sill"])) * z * exp(0.5 * h * z^2)
    },
    "40" = {
      z <- qnorm(p)
      hr <- as.numeric(params["tail1"]); hl <- as.numeric(params["tail2"])
      .GeoTukeyh2StandardMoments(hr, hl)
      h <- ifelse(z >= 0, hr, hl)
      eta + sqrt(as.numeric(params["sill"])) * z * exp(0.5 * h * z^2)
    },
    "20" = {
      z <- qnorm(p)
      d <- as.numeric(params["tail"]); e <- as.numeric(params["skew"])
      .GeoSASStandardMoments(e, d)
      eta + sqrt(as.numeric(params["sill"])) * sinh((asinh(z) + e) / d)
    },
    stop(paste("Model not supported:", model_type))
  )
}

#####################################################################
# coefficients a_k in copula gaussian covariance 
####################################################################
hermite_single <- function(k, t) {
    if (k == 0) return(rep(1, length(t)))
    if (k == 1) return(t)
    H_prev <- rep(1, length(t))
    H_curr <- t
    if (k > 1) {
      for (i in 2:k) {
        H_next <- t * H_curr - (i - 1) * H_prev
        H_prev <- H_curr
        H_curr <- H_next
      }
    }
    return(H_curr)
  }
compute_ak_vectorized <- function(M, model_type, params) {
  ak <- numeric(M)
  failed_integrations <- c()
  for (k in 1:M) {
    integrand <- function(t) {
      p <- pnorm(t)
      q <- quantile_disp(p, model_type, params)
      q[!is.finite(q)] <- 0
      result <- q * hermite_single(k, t) * dnorm(t)
      # Controllo aggiuntivo per valori non finiti
      result[!is.finite(result)] <- 0
      return(result)
    }
    # Strategia di integrazione progressiva
    result <- NA
    # Tentativo 1: Integrazione standard con più suddivisioni
    result <- tryCatch({
      integrate(integrand, lower = -12, upper = 12, 
                rel.tol = 1e-6, abs.tol = 1e-10, 
                subdivisions = 2000L)$value
    }, error = function(e) NA, warning = function(w) NA)

    # Tentativo 2: Limiti più ristretti se il primo fallisce
    if (is.na(result)) {
      result <- tryCatch({
        integrate(integrand, lower = -8, upper = 8, 
                  rel.tol = 1e-5, abs.tol = 1e-8, 
                  subdivisions = 1000L)$value
      }, error = function(e) NA, warning = function(w) NA)
    }
    if (is.na(result)) {
      if (k > 20) {
        # Per k elevati, i coefficienti tendono rapidamente a zero
        result <- 0
        #message(paste("Coefficient a_", k, " approximated to 0 (high k)", sep = ""))
      } else {
        # Integrazione molto conservativa
        result <- tryCatch({
          integrate(integrand, lower = -6, upper = 6, 
                    rel.tol = 1e-4, abs.tol = 1e-6,
                    subdivisions = 500L, stop.on.error = FALSE)$value
        }, error = function(e) {
          failed_integrations <<- c(failed_integrations, k)
          0
        })
      }
    }
    
    ak[k] <- ifelse(is.finite(result), result, 0)
  }
  
  # Report sui fallimenti
  if (length(failed_integrations) > 0) {
    message(paste("Integration failed for k =", paste(failed_integrations, collapse = ", "), 
                  "- coefficients set to 0"))
  }
  
  return(ak)
}
# main fnction


# Cached standard-normal Gauss-Hermite rule and Hermite design matrix.
# This avoids rebuilding the same M x nquad basis in repeated kriging calls.
.GeoGaussianHermiteRule <- function(M, nquad) {
  M <- as.integer(M); nquad <- as.integer(nquad)
  key <- paste0("M", M, "|n", nquad)
  cache <- .GeoModels_get_cache("gaussian_hermite_rule")
  if (exists(key, envir = cache, inherits = FALSE))
    return(get(key, envir = cache, inherits = FALSE))

  rule <- .GeoNormalQuadrature(nquad)
  H <- .GeoHermiteBasis(rule$nodes, M)
  out <- list(
    nodes = rule$nodes,
    weights = rule$weights,
    prob = pnorm(rule$nodes),
    H = H,
    WH = H * rule$weights
  )
  assign(key, out, envir = cache)
  out
}

# Fast Gauss-Hermite coefficients for continuous Gaussian-copula margins.
# The rule returned by .GeoNormalQuadrature() integrates expectations under
# a standard normal law directly, so all coefficients are obtained in one
# matrix product instead of M adaptive one-dimensional integrations.
.GeoGaussianCopulaAKGH <- function(M, model_type, params, nquad = NULL) {
  M <- as.integer(M)
  model_type <- as.integer(model_type)
  if (length(M) != 1L || is.na(M) || M < 1L)
    stop("M must be a positive integer.", call. = FALSE)

  if (is.null(nquad)) {
    ## Skew-Laplace has a kink in its quantile function at p = skew and needs
    ## a denser rule.  Smooth continuous margins are already very accurate
    ## with a substantially smaller rule.
    nquad <- if (model_type == 59L) 256L else 80L
  }
  nquad <- as.integer(nquad)

  rule <- .GeoGaussianHermiteRule(M, nquad)
  qq <- quantile_disp(rule$prob, model_type, params)
  if (any(!is.finite(qq)))
    stop("Non-finite marginal quantiles while computing Gaussian-copula covariance.",
         call. = FALSE)

  as.numeric(crossprod(rule$WH, qq))
}

# Vectorized polynomial evaluator for
#   sum_{k=1}^M coefficient[k] * rho^k.
# Horner evaluation uses M vector operations, rather than one R callback for
# every covariance entry.
.GeoGaussianCopulaPolynomial <- function(rho, coefficient) {
  rho <- as.numeric(rho)
  if (!length(rho)) return(numeric(0))

  good <- is.finite(rho) & !is.na(rho)
  out <- rep(NA_real_, length(rho))
  if (!any(good)) return(out)

  rr <- rho[good]
  coefficient <- as.numeric(coefficient)
  M <- length(coefficient)
  if (!M) {
    out[good] <- 0
    return(out)
  }

  acc <- rep(coefficient[M], length(rr))
  if (M > 1L) {
    for (k in seq.int(M - 1L, 1L))
      acc <- coefficient[k] + rr * acc
  }
  val <- rr * acc
  val[abs(rr) < 1e-12] <- 0
  val[!is.finite(val)] <- 0
  out[good] <- val
  out
}

# Gaussian-copula covariance.  Continuous margins use one cached
# Gauss-Hermite coefficient calculation and vectorized Horner evaluation.
# Legacy/discrete margins retain the adaptive-integration coefficient path.
gaussian_copula_cov <- function(rho, model_type, nuisance, M = 30,
                                cache_ak = TRUE, auto_reduce_M = TRUE,
                                verbose = FALSE, nquad = NULL) {
  model_type <- as.integer(model_type)
  M <- as.integer(M)
  continuous_models <- c(1L, 12L, 20L, 21L, 22L, 25L, 26L,
                         28L, 33L, 34L, 40L, 42L, 50L, 59L)
  use_gh <- model_type %in% continuous_models

  ## For a Gaussian margin the Gaussian-copula covariance is exactly
  ## sill * rho, so no quadrature or series evaluation is needed.
  if (model_type == 1L) {
    sill <- suppressWarnings(as.numeric(unlist(nuisance["sill"], use.names = FALSE)))
    if (length(sill) == 1L && is.finite(sill))
      return(.GeoGaussianCopulaPolynomial(rho, sill))
  }

  if (use_gh && is.null(nquad))
    nquad <- if (model_type == 59L) 256L else 80L

  cache_key <- paste(
    if (use_gh) "gh" else "adaptive",
    model_type,
    paste(names(nuisance), nuisance, sep = "=", collapse = ";"),
    M,
    if (use_gh) as.integer(nquad) else "",
    sep = "|"
  )
  ak_cache <- .GeoModels_get_cache("ak_cache")

  if (cache_ak && exists(cache_key, envir = ak_cache, inherits = FALSE)) {
    ak_coeffs <- get(cache_key, envir = ak_cache, inherits = FALSE)
    if (verbose) message("Coefficients a_k loaded from cache")
  } else {
    if (verbose) {
      method <- if (use_gh) paste0("Gauss-Hermite (n = ", nquad, ")") else "adaptive integration"
      message("Computing coefficients a_k with ", method, " for M = ", M)
    }

    ak_coeffs <- if (use_gh) {
      .GeoGaussianCopulaAKGH(M, model_type, nuisance, nquad = nquad)
    } else {
      compute_ak_vectorized(M, model_type, nuisance)
    }

    if (auto_reduce_M) {
      significant_coeffs <- which(abs(ak_coeffs) > 1e-10)
      if (length(significant_coeffs) > 0L) {
        effective_M <- max(significant_coeffs)
        if (effective_M < M * 0.7) {
          M_new <- min(effective_M + 5L, M)
          ak_coeffs <- ak_coeffs[seq_len(M_new)]
        }
      }
    }

    if (cache_ak)
      assign(cache_key, ak_coeffs, envir = ak_cache)
  }

  if (length(ak_coeffs) >= 10L) {
    last_coeffs <- abs(tail(ak_coeffs, 5L))
    if (all(last_coeffs < 1e-8) && verbose)
      message("Series converges well - last coefficients are small")
  }

  k <- seq_along(ak_coeffs)
  coefficient <- (ak_coeffs^2) / factorial(k)
  covariances <- .GeoGaussianCopulaPolynomial(rho, coefficient)

  if (verbose && any(is.finite(covariances))) {
    rr <- range(covariances[is.finite(covariances)])
    message("Computed with ", length(ak_coeffs), " coefficients")
    message("Covariance range: ", round(rr[1L], 6), " - ", round(rr[2L], 6))
  }

  covariances
}


################################################################################
## Gaussian-copula covariance helpers used by GeoCovmatrix/GeoKrig.           ##
################################################################################
.GeoGaussianCopulaContinuousModels <- c(1L, 12L, 20L, 21L, 22L, 25L, 26L,
                                         28L, 33L, 34L, 40L, 42L, 50L, 59L)
## Discrete count margins supported by the observed-scale copula covariance
## engines.  They are kept separate because, unlike most continuous margins,
## all three have location-dependent quantile functions.
.GeoGaussianCopulaDiscreteModels <- c(11L, 16L, 30L) # Binomial, BinomialNeg, Poisson
.GeoCopulaPredictionModels <- c(.GeoGaussianCopulaContinuousModels,
                                .GeoGaussianCopulaDiscreteModels)

.GeoCopulaParamScalar <- function(params, name, required = TRUE, default = NULL) {
  nms <- names(params)
  if (is.null(nms) || !(name %in% nms)) {
    if (required) stop("Missing copula marginal parameter '", name, "'.", call. = FALSE)
    return(default)
  }
  x <- suppressWarnings(as.numeric(params[[name]]))
  if (length(x) != 1L || !is.finite(x))
    stop("Copula marginal parameter '", name, "' must be a finite scalar.", call. = FALSE)
  x
}

.GeoCopulaEtaVector <- function(eta, n, default = 0) {
  if (is.null(eta) || !length(eta)) eta <- default
  eta <- as.numeric(eta)
  if (length(eta) == 1L) eta <- rep(eta, n)
  if (length(eta) != n || any(!is.finite(eta)))
    stop("Internal copula location vector has incompatible length or non-finite values.",
         call. = FALSE)
  eta
}

.GeoCopulaMarginalMoments <- function(model_type, params, eta = 0) {
  model_type <- as.integer(model_type)
  eta <- as.numeric(eta)
  if (!length(eta) || any(!is.finite(eta)))
    stop("Copula marginal locations must be finite.", call. = FALSE)

  sill <- function() .GeoCopulaParamScalar(params, "sill")
  lohi <- function() {
    lo <- .GeoCopulaParamScalar(params, "min")
    hi <- .GeoCopulaParamScalar(params, "max")
    if (lo >= hi) stop("Copula bounded margins require min < max.", call. = FALSE)
    c(lo, hi)
  }

  if (model_type == 1L) {                         # Gaussian
    ss <- sill()
    return(list(mean = eta, variance = rep(ss, length(eta))))
  }
  if (model_type %in% c(20L, 34L, 40L)) {         # SAS / Tukeyh / Tukeyh2
    return(.GeoPlainTransformedMarginalMoments(model_type, params, eta))
  }
  if (model_type == 12L) {                        # StudentT
    ss <- sill(); eps <- .GeoCopulaParamScalar(params, "df"); df_t <- 1 / eps
    if (df_t <= 2) stop("StudentT linear prediction requires more than 2 degrees of freedom.", call. = FALSE)
    return(list(mean = eta, variance = rep(ss * df_t / (df_t - 2), length(eta))))
  }
  if (model_type == 22L) {                        # LogGaussian
    ss <- sill(); mm <- exp(eta)
    return(list(mean = mm, variance = mm^2 * (exp(ss) - 1)))
  }
  if (model_type == 30L) {                        # Poisson
    mm <- exp(eta)
    return(list(mean = mm, variance = mm))
  }
  if (model_type == 11L) {                        # Binomial
    nn <- .GeoCopulaParamScalar(params, "n")
    if (nn < 1 || abs(nn - round(nn)) > sqrt(.Machine$double.eps))
      stop("Binomial copula prediction requires a positive integer n.", call. = FALSE)
    pp <- pnorm(eta)
    return(list(mean = nn * pp, variance = nn * pp * (1 - pp)))
  }
  if (model_type == 16L) {                        # BinomialNeg
    nn <- .GeoCopulaParamScalar(params, "n")
    if (nn < 1 || abs(nn - round(nn)) > sqrt(.Machine$double.eps))
      stop("BinomialNeg copula prediction requires a positive integer n.", call. = FALSE)
    pp <- pnorm(eta)
    mm <- nn * (1 - pp) / pp
    return(list(mean = mm, variance = nn * (1 - pp) / pp^2))
  }
  if (model_type == 21L) {                        # Gamma
    sh <- .GeoCopulaParamScalar(params, "shape")
    if (sh <= 0) stop("Gamma shape must be positive.", call. = FALSE)
    mm <- exp(eta)
    return(list(mean = mm, variance = 2 * mm^2 / sh))
  }
  if (model_type == 26L) {                        # Weibull
    sh <- .GeoCopulaParamScalar(params, "shape")
    if (sh <= 0) stop("Weibull shape must be positive.", call. = FALSE)
    mm <- exp(eta)
    vv0 <- gamma(1 + 2 / sh) / gamma(1 + 1 / sh)^2 - 1
    return(list(mean = mm, variance = mm^2 * vv0))
  }
  if (model_type == 25L) {                        # Logistic
    ss <- sill()
    return(list(mean = eta, variance = rep(pi^2 * ss / 3, length(eta))))
  }
  if (model_type == 59L) {                        # SkewLaplace
    ss <- sill(); sk <- .GeoCopulaParamScalar(params, "skew")
    if (sk <= 0 || sk >= 1) stop("SkewLaplace skew must lie in (0,1).", call. = FALSE)
    shift <- sqrt(ss) * (1 / sk - 1 / (1 - sk))
    vv <- ss * (1 - 2 * sk + 2 * sk^2) / (sk^2 * (1 - sk)^2)
    return(list(mean = eta + shift, variance = rep(vv, length(eta))))
  }
  if (model_type == 28L) {                        # Beta
    lr <- lohi(); a <- .GeoCopulaParamScalar(params, "shape1")
    b <- .GeoCopulaParamScalar(params, "shape2")
    if (a <= 0 || b <= 0) stop("Beta shape parameters must be positive.", call. = FALSE)
    rr <- lr[2L] - lr[1L]
    mm <- lr[1L] + rr * a / (a + b)
    vv <- rr^2 * a * b / ((a + b)^2 * (a + b + 1))
    return(list(mean = rep(mm, length(eta)), variance = rep(vv, length(eta))))
  }
  if (model_type == 50L) {                        # Beta2
    lr <- lohi(); phi <- .GeoCopulaParamScalar(params, "shape")
    if (phi <= 0) stop("Beta2 shape must be positive.", call. = FALSE)
    m <- plogis(eta); rr <- lr[2L] - lr[1L]
    return(list(mean = lr[1L] + rr * m,
                variance = rr^2 * m * (1 - m) / (phi + 1)))
  }
  if (model_type == 33L) {                        # Kumaraswamy
    lr <- lohi(); a <- .GeoCopulaParamScalar(params, "shape1")
    b <- .GeoCopulaParamScalar(params, "shape2")
    if (a <= 0 || b <= 0) stop("Kumaraswamy shape parameters must be positive.", call. = FALSE)
    m1 <- b * beta(1 + 1 / a, b); m2 <- b * beta(1 + 2 / a, b)
    rr <- lr[2L] - lr[1L]
    return(list(mean = rep(lr[1L] + rr * m1, length(eta)),
                variance = rep(rr^2 * (m2 - m1^2), length(eta))))
  }
  if (model_type == 42L) {                        # Kumaraswamy2
    lr <- lohi(); a <- .GeoCopulaParamScalar(params, "shape")
    if (a <= 0) stop("Kumaraswamy2 shape must be positive.", call. = FALSE)
    med <- plogis(eta)
    b <- log(0.5) / log1p(-med^a)
    if (any(!is.finite(b)) || any(b <= 0))
      stop("Kumaraswamy2 produced invalid location-specific shape parameters.", call. = FALSE)
    m1 <- b * beta(1 + 1 / a, b); m2 <- b * beta(1 + 2 / a, b)
    rr <- lr[2L] - lr[1L]
    return(list(mean = lr[1L] + rr * m1,
                variance = rr^2 * (m2 - m1^2)))
  }

  stop("Copula marginal moments are not implemented for model id ", model_type, ".",
       call. = FALSE)
}

.GeoNormalQuadrature <- function(n = 60L) {
  n <- as.integer(n)
  key <- paste0("n", n)
  cache <- .GeoModels_get_cache("normal_quadrature")
  if (exists(key, envir = cache, inherits = FALSE)) return(get(key, envir = cache))
  if (n < 10L) stop("Internal Gaussian quadrature requires at least 10 nodes.", call. = FALSE)
  J <- matrix(0, n, n)
  off <- sqrt(seq_len(n - 1L))
  J[cbind(seq_len(n - 1L), 2:n)] <- off
  J[cbind(2:n, seq_len(n - 1L))] <- off
  ee <- eigen(J, symmetric = TRUE)
  oo <- order(ee$values)
  out <- list(nodes = ee$values[oo], weights = ee$vectors[1L, oo]^2)
  assign(key, out, envir = cache)
  out
}

.GeoHermiteBasis <- function(x, M) {
  x <- as.numeric(x); M <- as.integer(M)
  H <- matrix(0, nrow = length(x), ncol = M)
  if (M < 1L) return(H)
  H[, 1L] <- x
  if (M >= 2L) {
    H[, 2L] <- x^2 - 1
    if (M >= 3L) {
      hm2 <- rep(1, length(x)); hm1 <- x
      for (k in 2:M) {
        hk <- x * hm1 - (k - 1) * hm2
        H[, k] <- hk
        hm2 <- hm1; hm1 <- hk
      }
    }
  }
  H
}

.GeoCopulaHermiteGHIndexed <- function(model_type, params, eta,
                                             M = 30L, nquad = 60L) {
  eta <- as.numeric(eta)
  uq <- unique(eta)
  rule <- .GeoGaussianHermiteRule(M, nquad)
  p <- rule$prob
  WH <- rule$WH
  out_u <- matrix(NA_real_, nrow = length(uq), ncol = M)

  cache <- .GeoModels_get_cache("copula_hermite_gh")
  base_names <- setdiff(names(params), .GeoMean_names(max(1L, sum(grepl("^mean", names(params))))))
  base_key <- paste(model_type,
                    paste(base_names, unlist(params[base_names]), sep = "=", collapse = ";"),
                    M, nquad, sep = "|")

  for (i in seq_along(uq)) {
    key <- paste0(base_key, "|eta=", format(uq[i], digits = 17, scientific = TRUE))
    if (exists(key, envir = cache, inherits = FALSE)) {
      out_u[i, ] <- get(key, envir = cache)
    } else {
      pp <- params
      pp["mean"] <- uq[i]
      q <- quantile_disp(p, model_type, pp)
      if (any(!is.finite(q)))
        stop("Non-finite marginal quantiles while computing Gaussian-copula covariance.", call. = FALSE)
      aa <- as.numeric(crossprod(WH, q))
      out_u[i, ] <- aa
      assign(key, aa, envir = cache)
    }
  }
  list(coef = out_u, index = match(eta, uq))
}

.GeoCopulaHermiteGH <- function(model_type, params, eta, M = 30L, nquad = 60L) {
  z <- .GeoCopulaHermiteGHIndexed(model_type, params, eta, M = M, nquad = nquad)
  z$coef[z$index, , drop = FALSE]
}

# Native location-dependent Gaussian-copula series evaluator.  The historical
# R implementation expanded one M-column Hermite matrix for every covariance
# pair.  For dense covariance and kriging blocks that can be hundreds of MB
# even though the same endpoint appears in many pairs.  Here only the unique
# endpoint coefficient rows are retained and native code follows integer row
# indices while accumulating the identical Hermite series.
.GeoGaussianCovSeriesIndexedC <- function(rho, z1, z2) {
  rho <- as.numeric(rho)
  n <- length(rho)
  if (!n) return(numeric(0))
  A1 <- as.matrix(z1$coef); A2 <- as.matrix(z2$coef)
  idx1 <- as.integer(z1$index); idx2 <- as.integer(z2$index)
  degree <- ncol(A1)
  if (degree < 1L || ncol(A2) != degree || length(idx1) != n ||
      length(idx2) != n || anyNA(idx1) || anyNA(idx2) ||
      any(idx1 < 1L | idx1 > nrow(A1)) || any(idx2 < 1L | idx2 > nrow(A2)))
    stop("Internal indexed Gaussian-copula series dimensions are inconsistent.",
         call. = FALSE)

  dotCall64::.C64(
    "GeoGaussianCovSeriesIndexed",
    SIGNATURE = c("double", "integer", "double", "integer", "integer",
                  "double", "integer", "integer", "integer", "double",
                  "double"),
    rho, as.integer(n), as.double(A1), as.integer(nrow(A1)), idx1,
    as.double(A2), as.integer(nrow(A2)), idx2, as.integer(degree),
    factorial(seq_len(degree)),
    out = dotCall64::vector_dc("double", n),
    INTENT = c(rep("r", 10), "w"), NAOK = TRUE, PACKAGE = "GeoModels"
  )$out
}

.GeoGaussianCopulaCrossCov <- function(rho, model_type, params,
                                       eta1 = 0, eta2 = eta1,
                                       M = 30L, nquad = 60L) {
  model_type <- as.integer(model_type)
  rho <- as.numeric(rho)
  n <- length(rho)
  if (!n) return(numeric(0))
  eta1 <- .GeoCopulaEtaVector(eta1, n)
  eta2 <- .GeoCopulaEtaVector(eta2, n)
  if (any(!is.finite(rho)) || any(abs(rho) > 1 + 1e-10))
    stop("Invalid latent correlations in Gaussian-copula covariance.", call. = FALSE)
  rho <- pmax(-1, pmin(1, rho))

  ## For additive location families the covariance does not depend on eta.
  additive <- c(1L, 12L, 20L, 25L, 34L, 40L, 59L)
  ## Positive multiplicative families have q_eta = exp(eta) q_0.
  multiplicative <- c(21L, 22L, 26L)
  ## Fixed-shape bounded families do not depend on the regression location.
  fixed_bounded <- c(28L, 33L)

  pp <- params
  pp["mean"] <- 0

  ## Tukey-h, Tukey-hh and sinh-arcsinh are monotone transforms of a
  ## single Gaussian RF.  Under a Gaussian copula their covariance is
  ## therefore exactly the same as in the corresponding plain model; do
  ## not approximate it with a truncated Hermite expansion.
  if (model_type %in% c(20L, 34L, 40L)) {
    return(.GeoPlainTransformedCrossCov(rho, model_type, pp))
  }
  if (model_type %in% c(additive, fixed_bounded)) {
    return(gaussian_copula_cov(rho, model_type, pp, M = M))
  }
  if (model_type %in% multiplicative) {
    base <- gaussian_copula_cov(rho, model_type, pp, M = M)
    return(exp(eta1 + eta2) * base)
  }
  if (model_type %in% c(50L, 42L, .GeoGaussianCopulaDiscreteModels)) {
    ## Count quantiles are step functions.  A denser normal quadrature reduces
    ## aliasing while preserving the fast indexed/cached covariance path.
    nquad_loc <- if (model_type %in% .GeoGaussianCopulaDiscreteModels)
      max(as.integer(nquad), 120L) else as.integer(nquad)
    if (isTRUE(getOption("GeoModels.gaussian.indexed_series", TRUE))) {
      z1 <- .GeoCopulaHermiteGHIndexed(
        model_type, params, eta1, M = M, nquad = nquad_loc
      )
      z2 <- .GeoCopulaHermiteGHIndexed(
        model_type, params, eta2, M = M, nquad = nquad_loc
      )
      return(.GeoGaussianCovSeriesIndexedC(rho, z1, z2))
    }

    ## Historical reference path retained behind an internal option for exact
    ## regression comparisons.  It intentionally materializes pair-expanded
    ## Hermite matrices and should not be used for production-sized problems.
    A1 <- .GeoCopulaHermiteGH(model_type, params, eta1, M = M, nquad = nquad_loc)
    A2 <- .GeoCopulaHermiteGH(model_type, params, eta2, M = M, nquad = nquad_loc)
    fac <- factorial(seq_len(M))
    out <- numeric(n)
    rp <- rho
    for (k in seq_len(M)) {
      term <- (A1[, k] * A2[, k] / fac[k]) * rp
      term[!is.finite(term)] <- 0
      out <- out + term
      rp <- rp * rho
    }
    return(out)
  }

  ## Backward-compatible fallback for legacy Gaussian-copula margins.
  gaussian_copula_cov(rho, model_type, params, M = M)
}

.GeoCopulaMarginalMeanDerivative <- function(model_type, params, eta) {
  model_type <- as.integer(model_type); eta <- as.numeric(eta)
  if (model_type %in% c(1L, 12L, 20L, 25L, 34L, 40L, 59L)) return(rep(1, length(eta)))
  if (model_type %in% c(21L, 22L, 26L, 30L)) return(exp(eta))
  if (model_type == 11L) {
    nn <- .GeoCopulaParamScalar(params, "n")
    return(nn * dnorm(eta))
  }
  if (model_type == 16L) {
    nn <- .GeoCopulaParamScalar(params, "n")
    pp <- pnorm(eta)
    return(-nn * dnorm(eta) / pp^2)
  }
  if (model_type == 50L) {
    lo <- .GeoCopulaParamScalar(params, "min"); hi <- .GeoCopulaParamScalar(params, "max")
    m <- plogis(eta)
    return((hi - lo) * m * (1 - m))
  }
  if (model_type %in% c(28L, 33L)) return(rep(0, length(eta)))
  if (model_type == 42L) {
    ## Stable centered finite-difference derivative of the Kumaraswamy2 mean.
    h <- 1e-5 * pmax(1, abs(eta))
    mp <- .GeoCopulaMarginalMoments(model_type, params, eta + h)$mean
    mm <- .GeoCopulaMarginalMoments(model_type, params, eta - h)$mean
    return((mp - mm) / (2 * h))
  }
  rep(1, length(eta))
}




################################################################################
## Fast Clayton-like copula covariance helpers used by GeoCovmatrix/GeoKrig.  ##
##                                                                            ##
## The bivariate Clayton-like density is expensive because it contains an      ##
## Appell F4 evaluation.  Kriging must therefore never call adaptive 2-D       ##
## integration separately for every covariance entry.  The implementation     ##
## below builds and caches a deterministic Gauss-Legendre representation of   ##
## the copula joint law on a modest grid of rho^2 values.  Marginal quantile   ##
## transforms are then applied with BLAS matrix products and the resulting     ##
## covariance table is interpolated.                                           ##
################################################################################

.GeoUnitQuadrature <- function(n = 36L) {
  n <- as.integer(n)
  if (length(n) != 1L || !is.finite(n) || n < 12L)
    stop("Internal unit-interval quadrature requires at least 12 nodes.", call. = FALSE)
  key <- paste0("n", n)
  cache <- .GeoModels_get_cache("unit_quadrature")
  if (exists(key, envir = cache, inherits = FALSE)) return(get(key, envir = cache))

  k <- seq_len(n - 1L)
  off <- k / sqrt(4 * k^2 - 1)
  J <- matrix(0, n, n)
  J[cbind(k, k + 1L)] <- off
  J[cbind(k + 1L, k)] <- off
  ee <- eigen(J, symmetric = TRUE)
  oo <- order(ee$values)
  x <- ee$values[oo]
  w <- 2 * ee$vectors[1L, oo]^2
  out <- list(nodes = (x + 1) / 2, weights = w / 2)
  assign(key, out, envir = cache)
  out
}

.GeoBalanceCopulaQuadrature <- function(P, w, tol = 1e-11, maxit = 80L) {
  P <- (P + t(P)) / 2
  P[!is.finite(P) | P < 0] <- 0
  tiny <- .Machine$double.xmin
  if (!any(P > 0)) stop("Copula quadrature produced zero mass.", call. = FALSE)

  for (it in seq_len(maxit)) {
    rs <- rowSums(P)
    s <- sqrt(w / pmax(rs, tiny))
    P <- sweep(P, 1L, s, `*`)
    P <- sweep(P, 2L, s, `*`)
    if (max(abs(rowSums(P) - w)) < tol) break
  }
  ## One final symmetric correction substantially reduces accumulated roundoff.
  P <- (P + t(P)) / 2
  rs <- rowSums(P)
  s <- sqrt(w / pmax(rs, tiny))
  P <- sweep(P, 1L, s, `*`)
  P <- sweep(P, 2L, s, `*`)
  P / sum(P)
}

.GeoClaytonKernelTable <- function(nu, nquad = 36L, nrho = 55L) {
  nu <- suppressWarnings(as.numeric(nu))
  if (length(nu) != 1L || !is.finite(nu) || nu <= 0 ||
      abs(nu - round(nu)) > sqrt(.Machine$double.eps))
    stop("Clayton-like linear prediction requires a positive integer nu.", call. = FALSE)
  nu <- as.integer(round(nu))
  nquad <- as.integer(nquad); nrho <- as.integer(nrho)
  if (nrho < 15L) stop("Internal Clayton covariance grid is too small.", call. = FALSE)

  ## The interior grid is concentrated near both independence and the
  ## comonotone limit.  rho^2=1 is handled analytically, avoiding evaluation
  ## of the singular limiting density.
  tt <- seq(0, 1, length.out = nrho - 1L)
  xgrid <- c(0.995 * 0.5 * (1 - cos(pi * tt)), 1)
  use_symmetry <- isTRUE(getOption("GeoModels.clayton.symmetric_kernel", TRUE))
  kernel_mode <- if (use_symmetry) "tri" else "full"
  key <- paste0("nu=", nu, "|q=", nquad, "|r=", nrho,
                "|v=3|mode=", kernel_mode)
  cache <- .GeoModels_get_cache("clayton_kernel_table")
  if (exists(key, envir = cache, inherits = FALSE)) return(get(key, envir = cache))

  rule <- .GeoUnitQuadrature(nquad)
  ww <- outer(rule$weights, rule$weights)

  ## The Clayton-like copula density is exchangeable in (u, v).  Evaluate
  ## only one triangular half of the quadrature grid, then mirror it before
  ## balancing.  For the default 36-point rule this reduces each expensive
  ## Appell-F4 density slice from 1296 to 666 native evaluations.
  if (use_symmetry) {
    ij <- which(upper.tri(matrix(FALSE, nquad, nquad), diag = TRUE),
                arr.ind = TRUE)
    u <- rule$nodes[ij[, 1L]]
    v <- rule$nodes[ij[, 2L]]
    density_evaluations <- nrow(ij)
  } else {
    u <- rep(rule$nodes, times = nquad)
    v <- rep(rule$nodes, each = nquad)
    ij <- NULL
    density_evaluations <- nquad * nquad
  }

  joint <- array(0, dim = c(nquad, nquad, nrho))
  joint[, , 1L] <- ww
  joint[, , nrho] <- diag(rule$weights, nrow = nquad, ncol = nquad)

  if (nrho > 2L) {
    for (j in 2:(nrho - 1L)) {
      rho <- sqrt(xgrid[j])
      ans <- .C(
        "GeoCopulaDensity",
        as.integer(2L), as.double(nu),
        as.double(u), as.double(v), as.integer(length(u)),
        as.double(rho), res = double(length(u)),
        PACKAGE = "GeoModels", NAOK = TRUE
      )$res

      if (use_symmetry) {
        dens <- matrix(0, nrow = nquad, ncol = nquad)
        dens[ij] <- ans
        dens[cbind(ij[, 2L], ij[, 1L])] <- ans
      } else {
        dens <- matrix(ans, nrow = nquad, ncol = nquad)
      }

      P <- dens * ww
      joint[, , j] <- .GeoBalanceCopulaQuadrature(P, rule$weights)
    }
  }

  out <- list(x = xgrid, joint = joint, nodes = rule$nodes,
              weights = rule$weights, nu = nu,
              symmetric_kernel = use_symmetry,
              density_evaluations_per_slice = density_evaluations)
  assign(key, out, envir = cache)
  out
}

.GeoClaytonQuantileRows <- function(model_type, params, eta, nodes, weights) {
  eta <- as.numeric(eta)
  Q <- matrix(NA_real_, nrow = length(eta), ncol = length(nodes))
  V <- .GeoCopulaMarginalMoments(model_type, params, eta)$variance
  if (length(V) == 1L) V <- rep(V, length(eta))

  for (i in seq_along(eta)) {
    pp <- params
    pp["mean"] <- eta[i]
    q <- quantile_disp(nodes, model_type, pp)
    if (any(!is.finite(q)))
      stop("Non-finite marginal quantiles in Clayton-like covariance.", call. = FALSE)
    qc <- q - sum(weights * q)
    vq <- sum(weights * qc^2)
    if (!is.finite(vq) || vq <= 0 || !is.finite(V[i]) || V[i] <= 0)
      stop("Invalid marginal variance in Clayton-like covariance.", call. = FALSE)
    ## Calibrate the deterministic quadrature to the exact marginal variance.
    Q[i, ] <- qc * sqrt(V[i] / vq)
  }
  Q
}

.GeoInterp1Grid <- function(grid, values, x) {
  x <- pmax(grid[1L], pmin(grid[length(grid)], as.numeric(x)))
  ii <- findInterval(x, grid, all.inside = TRUE)
  ii <- pmin(ii, length(grid) - 1L)
  x0 <- grid[ii]; x1 <- grid[ii + 1L]
  a <- ifelse(x1 > x0, (x - x0) / (x1 - x0), 0)
  (1 - a) * values[ii] + a * values[ii + 1L]
}

.GeoInterp3Grid <- function(g1, g2, gx, tab, z1, z2, x) {
  z1 <- pmax(g1[1L], pmin(g1[length(g1)], as.numeric(z1)))
  z2 <- pmax(g2[1L], pmin(g2[length(g2)], as.numeric(z2)))
  x  <- pmax(gx[1L], pmin(gx[length(gx)], as.numeric(x)))

  i1 <- pmin(findInterval(z1, g1, all.inside = TRUE), length(g1) - 1L)
  i2 <- pmin(findInterval(z2, g2, all.inside = TRUE), length(g2) - 1L)
  ir <- pmin(findInterval(x,  gx, all.inside = TRUE), length(gx) - 1L)
  a1 <- (z1 - g1[i1]) / pmax(g1[i1 + 1L] - g1[i1], .Machine$double.eps)
  a2 <- (z2 - g2[i2]) / pmax(g2[i2 + 1L] - g2[i2], .Machine$double.eps)
  ar <- (x  - gx[ir]) / pmax(gx[ir + 1L] - gx[ir], .Machine$double.eps)

  at <- function(d1, d2, dr) tab[cbind(i1 + d1, i2 + d2, ir + dr)]
  c000 <- at(0L,0L,0L); c100 <- at(1L,0L,0L)
  c010 <- at(0L,1L,0L); c110 <- at(1L,1L,0L)
  c001 <- at(0L,0L,1L); c101 <- at(1L,0L,1L)
  c011 <- at(0L,1L,1L); c111 <- at(1L,1L,1L)
  c00 <- (1-a1)*c000 + a1*c100
  c10 <- (1-a1)*c010 + a1*c110
  c01 <- (1-a1)*c001 + a1*c101
  c11 <- (1-a1)*c011 + a1*c111
  c0 <- (1-a2)*c00 + a2*c10
  c1 <- (1-a2)*c01 + a2*c11
  (1-ar)*c0 + ar*c1
}



# Native Clayton interpolation helpers.  The historical R interpolators above
# are retained as regression references.  These wrappers move interval search,
# interpolation, and (for multiplicative margins) exp(eta1 + eta2) scaling into
# one native pass, avoiding large temporary vectors on warm-cache covariance
# calls.
.GeoClaytonInterp1GridC <- function(grid, values, x,
                                     eta1 = NULL, eta2 = NULL,
                                     multiplicative = FALSE) {
  grid <- as.numeric(grid)
  values <- as.numeric(values)
  x <- as.numeric(x)
  ng <- length(grid); n <- length(x)
  if (!n) return(numeric(0))
  if (ng < 2L || length(values) != ng)
    stop("Internal Clayton interpolation table has incompatible dimensions.",
         call. = FALSE)

  multiplicative <- isTRUE(multiplicative)
  if (multiplicative) {
    eta1 <- .GeoCopulaEtaVector(eta1, n)
    eta2 <- .GeoCopulaEtaVector(eta2, n)
  } else {
    ## C64 needs valid addresses, although the native routine does not
    ## dereference eta1/eta2 when multiplicative == 0.
    eta1 <- 0
    eta2 <- 0
  }

  dotCall64::.C64(
    "GeoClaytonInterp1",
    SIGNATURE = c("double", "integer", "double", "double", "double",
                  "double", "integer", "integer", "double"),
    grid, as.integer(ng), values, x, eta1, eta2, as.integer(n),
    as.integer(multiplicative),
    out = dotCall64::vector_dc("double", n),
    INTENT = c("r", "r", "r", "r", "r", "r", "r", "r", "w"),
    NAOK = TRUE, PACKAGE = "GeoModels"
  )$out
}

.GeoClaytonInterp3GridC <- function(eta_grid, x_grid, tab,
                                     eta1, eta2, x) {
  eta_grid <- as.numeric(eta_grid)
  x_grid <- as.numeric(x_grid)
  x <- as.numeric(x)
  n <- length(x); ne <- length(eta_grid); nx <- length(x_grid)
  if (!n) return(numeric(0))
  if (ne < 2L || nx < 2L || length(tab) != ne * ne * nx)
    stop("Internal Clayton location interpolation table has incompatible dimensions.",
         call. = FALSE)
  eta1 <- .GeoCopulaEtaVector(eta1, n)
  eta2 <- .GeoCopulaEtaVector(eta2, n)

  dotCall64::.C64(
    "GeoClaytonInterp3",
    SIGNATURE = c("double", "integer", "double", "integer", "double",
                  "double", "double", "double", "integer", "double"),
    eta_grid, as.integer(ne), x_grid, as.integer(nx), tab,
    eta1, eta2, x, as.integer(n),
    out = dotCall64::vector_dc("double", n),
    INTENT = c("r", "r", "r", "r", "r", "r", "r", "r", "r", "w"),
    NAOK = TRUE, PACKAGE = "GeoModels"
  )$out
}

.GeoClaytonInterp1Grid <- function(grid, values, x,
                                    eta1 = NULL, eta2 = NULL,
                                    multiplicative = FALSE) {
  if (isTRUE(getOption("GeoModels.clayton.native_lookup", TRUE))) {
    return(.GeoClaytonInterp1GridC(
      grid, values, x, eta1 = eta1, eta2 = eta2,
      multiplicative = multiplicative
    ))
  }
  out <- .GeoInterp1Grid(grid, values, x)
  if (isTRUE(multiplicative))
    out <- exp(.GeoCopulaEtaVector(eta1, length(out)) +
               .GeoCopulaEtaVector(eta2, length(out))) * out
  out
}

.GeoClaytonInterp3Grid <- function(eta_grid, x_grid, tab,
                                    eta1, eta2, x) {
  if (isTRUE(getOption("GeoModels.clayton.native_lookup", TRUE))) {
    return(.GeoClaytonInterp3GridC(
      eta_grid, x_grid, tab, eta1 = eta1, eta2 = eta2, x = x
    ))
  }
  .GeoInterp3Grid(eta_grid, eta_grid, x_grid, tab, eta1, eta2, x)
}

.GeoClaytonCovLookupTable <- function(model_type, params, nu,
                                           nquad = 36L, nrho = 55L,
                                           neta = 51L, eta_range = NULL) {
  model_type <- as.integer(model_type)
  kernel <- .GeoClaytonKernelTable(nu, nquad = nquad, nrho = nrho)
  additive <- c(1L, 12L, 20L, 25L, 34L, 40L, 59L)
  multiplicative <- c(21L, 22L, 26L)
  fixed_bounded <- c(28L, 33L)
  location_dependent <- c(50L, 42L, .GeoGaussianCopulaDiscreteModels)

  nms <- names(params)
  keep <- if (is.null(nms)) seq_along(params) else !grepl("^mean[0-9]*$", nms)
  ppkey <- params[keep]
  if (!is.null(names(ppkey))) ppkey <- ppkey[order(names(ppkey))]
  base_key <- paste(
    "m", model_type,
    paste(names(ppkey), format(as.numeric(unlist(ppkey, use.names = FALSE)), digits = 17, scientific = TRUE),
          sep = "=", collapse = ";"),
    "nu", as.integer(round(nu)), "q", nquad, "r", nrho, "e", neta,
    "kernel", if (isTRUE(kernel$symmetric_kernel)) "tri" else "full",
    "v5", sep = "|"
  )
  cache <- .GeoModels_get_cache("clayton_cov_lookup")

  if (model_type %in% c(additive, multiplicative, fixed_bounded)) {
    key <- paste0(base_key, "|fixed")
    if (exists(key, envir = cache, inherits = FALSE)) return(get(key, envir = cache))
    Q <- .GeoClaytonQuantileRows(model_type, params, 0,
                                 kernel$nodes, kernel$weights)
    cv <- numeric(length(kernel$x))
    for (j in seq_along(kernel$x))
      cv[j] <- as.numeric(Q %*% kernel$joint[, , j] %*% t(Q))
    out <- list(kind = if (model_type %in% multiplicative) "multiplicative" else "fixed",
                x = kernel$x, cv = cv)
    assign(key, out, envir = cache)
    return(out)
  }

  if (model_type %in% location_dependent) {
    ## Beta2/Kumaraswamy2 and the count margins have location-dependent
    ## quantile transforms.  Do not
    ## build a huge fixed grid reaching probabilities arbitrarily close to 0/1:
    ## low-order uniform quadrature cannot resolve such nearly-degenerate beta
    ## margins and can return zero numerical variance.  Instead build a compact
    ## eta grid around the locations actually requested.  The expensive Clayton
    ## copula kernel (Appell F4 evaluations) is still shared session-wide; only
    ## the inexpensive marginal transform table depends on this eta band.
    er <- suppressWarnings(as.numeric(eta_range))
    er <- er[is.finite(er)]
    if (!length(er)) er <- 0
    elo <- floor(min(er)) - 1
    ehi <- ceiling(max(er)) + 1
    if (!is.finite(elo) || !is.finite(ehi)) stop("Invalid Clayton marginal-location range.", call. = FALSE)
    if (ehi <= elo) ehi <- elo + 2
    ## Keep a minimum two-unit band so nearby GeoKrigloc neighborhoods reuse
    ## the same cached table rather than generating a new table for each site.
    if ((ehi - elo) < 2) {
      mid <- 0.5 * (elo + ehi)
      elo <- floor(mid - 1)
      ehi <- ceiling(mid + 1)
    }
    key <- paste0(base_key, "|eta=", format(elo, scientific = FALSE), ":",
                  format(ehi, scientific = FALSE))
    if (exists(key, envir = cache, inherits = FALSE)) return(get(key, envir = cache))

    ## Keep the interpolation spacing reasonably fine on wide regression
    ## ranges without changing the expensive copula kernel resolution.
    neta_eff <- max(as.integer(neta), as.integer(4L * ceiling(ehi - elo) + 1L))
    eg <- seq(elo, ehi, length.out = neta_eff)
    Q <- .GeoClaytonQuantileRows(model_type, params, eg,
                                 kernel$nodes, kernel$weights)
    ng <- length(eg); nr <- length(kernel$x)
    tab <- array(0, dim = c(ng, ng, nr))
    for (j in seq_len(nr)) {
      tmp <- Q %*% kernel$joint[, , j] %*% t(Q)
      tab[, , j] <- (tmp + t(tmp)) / 2
    }
    out <- list(kind = "location", x = kernel$x, eta = eg, tab = tab,
                eta_range = c(elo, ehi))
    assign(key, out, envir = cache)
    return(out)
  }

  stop("Unsupported margin in Clayton-like covariance.", call. = FALSE)
}

.GeoClaytonCopulaCrossCovFast <- function(rho, model_type, params,
                                           eta1 = 0, eta2 = eta1,
                                           nquad = 36L, nrho = 55L,
                                           neta = 51L, eta_range = NULL) {
  model_type <- as.integer(model_type)
  if (!(model_type %in% .GeoCopulaPredictionModels))
    stop("Clayton-like linear prediction is not implemented for model id ",
         model_type, ".", call. = FALSE)
  if (model_type %in% .GeoGaussianCopulaDiscreteModels)
    nquad <- max(as.integer(nquad), 48L)
  rho <- as.numeric(rho); n <- length(rho)
  if (!n) return(numeric(0))
  eta1 <- .GeoCopulaEtaVector(eta1, n)
  eta2 <- .GeoCopulaEtaVector(eta2, n)
  if (any(!is.finite(rho)) || any(abs(rho) > 1 + 1e-10))
    stop("Invalid latent correlations in Clayton-like covariance.", call. = FALSE)
  x <- pmin(1, pmax(0, rho^2))

  nu <- .GeoCopulaParamScalar(params, "nu")
  if (nu <= 0 || abs(nu - round(nu)) > sqrt(.Machine$double.eps))
    stop("Clayton-like linear prediction requires a positive integer nu.", call. = FALSE)
  lookup_eta_range <- if (is.null(eta_range)) {
    range(c(eta1, eta2), finite = TRUE)
  } else {
    eta_range
  }
  lookup <- .GeoClaytonCovLookupTable(
    model_type, params, nu,
    nquad = nquad, nrho = nrho, neta = neta,
    eta_range = lookup_eta_range
  )

  if (identical(lookup$kind, "fixed")) {
    out <- .GeoClaytonInterp1Grid(lookup$x, lookup$cv, x)
  } else if (identical(lookup$kind, "multiplicative")) {
    out <- .GeoClaytonInterp1Grid(
      lookup$x, lookup$cv, x, eta1 = eta1, eta2 = eta2,
      multiplicative = TRUE
    )
  } else {
    out <- .GeoClaytonInterp3Grid(
      lookup$eta, lookup$x, lookup$tab, eta1, eta2, x
    )
  }
  out[x < 1e-15] <- 0
  out
}


################################################################################
## Fast skew-Gaussian copula covariance helpers used by GeoCovmatrix/GeoKrig. ##
##                                                                            ##
## Write the observed-scale transform as                                      ##
##   g_s(G1,G2) = F_s^{-1[F_Zeta{gamma_eta |G1| + G2}].                      ##
## A bivariate Hermite expansion of g_s gives                                 ##
##   Cov{Y_i,Y_j} = sum_{a+b>=1} f_i(a,b) f_j(a,b) rho^(a+b).                ##
## Truncating this expansion is much cheaper than evaluating the bivariate     ##
## skew-Gaussian density for every covariance entry. It also preserves a       ##
## positive-semidefinite covariance construction: each total-degree term is   ##
## a diagonally scaled Hadamard power of the latent correlation matrix.        ##
##                                                                            ##
## The expensive univariate skew-normal transform is computed only once for   ##
## each copula eta. Location-dependent Beta2/Kumaraswamy2 coefficients are    ##
## tabulated on a compact eta grid; interpolation is done on the coefficient  ##
## Gram matrices, which is equivalent to linearly interpolating the Hermite   ##
## feature vectors and therefore retains the PSD construction.                 ##
################################################################################

.GeoSkewGaussianUniformGH <- function(eta, nquad = 60L) {
  eta <- suppressWarnings(as.numeric(eta))
  if (length(eta) != 1L || !is.finite(eta) || abs(eta) >= 1)
    stop("Skew-Gaussian linear prediction requires eta in (-1,1).", call. = FALSE)
  nquad <- as.integer(nquad)
  if (nquad < 20L)
    stop("Internal skew-Gaussian quadrature requires at least 20 nodes.", call. = FALSE)

  key <- paste0("eta=", format(eta, digits = 17, scientific = TRUE),
                "|q=", nquad, "|v=2")
  cache <- .GeoModels_get_cache("skewgaussian_uniform_gh")
  if (exists(key, envir = cache, inherits = FALSE))
    return(get(key, envir = cache, inherits = FALSE))

  ## g(G1,G2) depends on G1 only through |G1|.  Integrating that coordinate
  ## directly under its half-normal distribution is substantially more
  ## accurate for |eta| close to one than applying ordinary Gauss-Hermite
  ## quadrature across the kink at G1=0.
  half_rule <- .GeoUnitQuadrature(nquad)
  a_nodes <- qnorm((1 + half_rule$nodes) / 2)
  normal_rule <- .GeoNormalQuadrature(nquad)

  gamma_eta <- eta / sqrt(1 - eta^2)
  omega_eta <- sqrt(1 + gamma_eta^2)
  z <- outer(gamma_eta * a_nodes, normal_rule$nodes, `+`)
  u <- sn::psn(as.numeric(z), xi = 0, omega = omega_eta, alpha = gamma_eta)
  u <- matrix(pmin(1 - 1e-12, pmax(1e-12, u)), nrow = nquad, ncol = nquad)

  out <- list(
    half_nodes = a_nodes, half_weights = half_rule$weights,
    normal_nodes = normal_rule$nodes, normal_weights = normal_rule$weights,
    u = u, eta = eta, nquad = nquad
  )
  assign(key, out, envir = cache)
  out
}

.GeoHermiteBasis0 <- function(x, M) {
  M <- as.integer(M)
  cbind(`0` = rep(1, length(x)), .GeoHermiteBasis(x, M))
}

.GeoSkewGaussianFeatureGrams <- function(model_type, params, eta_copula,
                                          eta_grid = 0,
                                          degree = 36L, nquad = 60L) {
  model_type <- as.integer(model_type)
  degree <- as.integer(degree); nquad <- as.integer(nquad)
  if (degree < 4L || nquad <= degree)
    stop("Internal skew-Gaussian Hermite settings are inconsistent.", call. = FALSE)
  eta_grid <- as.numeric(eta_grid)
  if (!length(eta_grid) || any(!is.finite(eta_grid)))
    stop("Invalid skew-Gaussian marginal-location grid.", call. = FALSE)

  gh <- .GeoSkewGaussianUniformGH(eta_copula, nquad = nquad)
  B1 <- .GeoHermiteBasis0(gh$half_nodes, degree)
  ## Averaging over the unobserved sign of G1 eliminates all odd Hermite
  ## orders exactly because the response transform depends on |G1| only.
  if (degree >= 1L) B1[, seq.int(2L, degree + 1L, by = 2L)] <- 0
  B2 <- .GeoHermiteBasis0(gh$normal_nodes, degree)
  WB1 <- B1 * gh$half_weights
  WB2 <- B2 * gh$normal_weights
  sfac <- sqrt(factorial(0:degree))
  normmat <- outer(sfac, sfac)
  degmat <- outer(0:degree, 0:degree, `+`)
  use <- degmat > 0 & degmat <= degree

  nms <- names(params)
  keep <- if (is.null(nms)) seq_along(params) else !grepl("^mean[0-9]*$", nms)
  ppkey <- params[keep]
  if (!is.null(names(ppkey))) ppkey <- ppkey[order(names(ppkey))]
  key <- paste(
    "m", model_type,
    paste(names(ppkey),
          format(as.numeric(unlist(ppkey, use.names = FALSE)), digits = 17, scientific = TRUE),
          sep = "=", collapse = ";"),
    "eta", format(eta_copula, digits = 17, scientific = TRUE),
    "loc", paste(format(eta_grid, digits = 15, scientific = TRUE), collapse = ","),
    "d", degree, "q", nquad, "v1", sep = "|"
  )
  cache <- .GeoModels_get_cache("skewgaussian_feature_grams")
  if (exists(key, envir = cache, inherits = FALSE))
    return(get(key, envir = cache, inherits = FALSE))

  ng <- length(eta_grid)
  feature_by_degree <- vector("list", degree)
  for (k in seq_len(degree)) {
    idx <- which(degmat == k & use, arr.ind = TRUE)
    feature_by_degree[[k]] <- matrix(0, nrow = ng, ncol = nrow(idx))
    attr(feature_by_degree[[k]], "index") <- idx
  }

  for (ii in seq_len(ng)) {
    pp <- params
    pp["mean"] <- eta_grid[ii]
    qv <- quantile_disp(as.numeric(gh$u), model_type, pp)
    if (any(!is.finite(qv)))
      stop("Non-finite marginal quantiles in skew-Gaussian covariance.", call. = FALSE)
    Q <- matrix(qv, nrow = nquad, ncol = nquad)
    mom <- .GeoCopulaMarginalMoments(model_type, params, eta_grid[ii])
    Q <- Q - as.numeric(mom$mean)[1L]

    ## E[g H_a(G1) H_b(G2)] / sqrt(a! b!) are orthonormal Hermite
    ## feature coefficients.  Only total degree <= degree is retained.
    A <- crossprod(WB1, Q %*% WB2) / normmat
    A[!use] <- 0

    ## Calibrate the truncated feature norm to the exact marginal variance.
    ## This makes the rho=1 limit exact for equal margins while preserving the
    ## PSD feature representation (a common positive scale is applied at each
    ## marginal location).
    vtr <- sum(A^2)
    vex <- as.numeric(mom$variance)[1L]
    if (!is.finite(vtr) || vtr <= 0 || !is.finite(vex) || vex <= 0)
      stop("Invalid marginal variance in skew-Gaussian covariance.", call. = FALSE)
    A <- A * sqrt(vex / vtr)

    for (k in seq_len(degree)) {
      idx <- attr(feature_by_degree[[k]], "index")
      feature_by_degree[[k]][ii, ] <- A[idx]
    }
  }

  grams <- lapply(feature_by_degree, function(F) tcrossprod(F))
  out <- list(eta = eta_grid, grams = grams, degree = degree,
              eta_copula = eta_copula, nquad = nquad)
  assign(key, out, envir = cache)
  out
}

.GeoInterp2Weights <- function(g, z1, z2) {
  g <- as.numeric(g)
  z1 <- pmax(g[1L], pmin(g[length(g)], as.numeric(z1)))
  z2 <- pmax(g[1L], pmin(g[length(g)], as.numeric(z2)))
  if (length(g) < 2L)
    stop("Internal interpolation grid must contain at least two points.", call. = FALSE)
  i1 <- pmin(findInterval(z1, g, all.inside = TRUE), length(g) - 1L)
  i2 <- pmin(findInterval(z2, g, all.inside = TRUE), length(g) - 1L)
  a1 <- (z1 - g[i1]) / pmax(g[i1 + 1L] - g[i1], .Machine$double.eps)
  a2 <- (z2 - g[i2]) / pmax(g[i2 + 1L] - g[i2], .Machine$double.eps)
  list(i1 = as.integer(i1), i2 = as.integer(i2), a1 = a1, a2 = a2)
}

.GeoInterp2Grid <- function(g, tab, z1, z2) {
  z1 <- pmax(g[1L], pmin(g[length(g)], as.numeric(z1)))
  z2 <- pmax(g[1L], pmin(g[length(g)], as.numeric(z2)))
  if (length(g) == 1L) return(rep(tab[1L, 1L], max(length(z1), length(z2))))
  w <- .GeoInterp2Weights(g, z1, z2)
  i1 <- w$i1; i2 <- w$i2; a1 <- w$a1; a2 <- w$a2
  c00 <- tab[cbind(i1, i2)]
  c10 <- tab[cbind(i1 + 1L, i2)]
  c01 <- tab[cbind(i1, i2 + 1L)]
  c11 <- tab[cbind(i1 + 1L, i2 + 1L)]
  (1 - a1) * (1 - a2) * c00 +
    a1 * (1 - a2) * c10 +
    (1 - a1) * a2 * c01 +
    a1 * a2 * c11
}

.GeoSkewGaussianCovLookupTable <- function(model_type, params, eta_copula,
                                            degree = 36L, nquad = 60L,
                                            neta = 51L, eta_range = NULL) {
  model_type <- as.integer(model_type)
  additive <- c(1L, 12L, 20L, 25L, 34L, 40L, 59L)
  multiplicative <- c(21L, 22L, 26L)
  fixed_bounded <- c(28L, 33L)
  location_dependent <- c(50L, 42L, .GeoGaussianCopulaDiscreteModels)

  nms <- names(params)
  keep <- if (is.null(nms)) seq_along(params) else !grepl("^mean[0-9]*$", nms)
  ppkey <- params[keep]
  if (!is.null(names(ppkey))) ppkey <- ppkey[order(names(ppkey))]
  base_key <- paste(
    "m", model_type,
    paste(names(ppkey),
          format(as.numeric(unlist(ppkey, use.names = FALSE)), digits = 17, scientific = TRUE),
          sep = "=", collapse = ";"),
    "eta", format(eta_copula, digits = 17, scientific = TRUE),
    "d", degree, "q", nquad, "e", neta, "v2", sep = "|"
  )
  cache <- .GeoModels_get_cache("skewgaussian_cov_lookup")

  if (model_type %in% c(additive, multiplicative, fixed_bounded)) {
    key <- paste0(base_key, "|fixed")
    if (exists(key, envir = cache, inherits = FALSE))
      return(get(key, envir = cache, inherits = FALSE))
    ff <- .GeoSkewGaussianFeatureGrams(
      model_type, params, eta_copula, eta_grid = 0,
      degree = degree, nquad = nquad
    )
    coef <- vapply(ff$grams, function(G) G[1L, 1L], numeric(1L))
    out <- list(
      kind = if (model_type %in% multiplicative) "multiplicative" else "fixed",
      coef = coef, degree = seq_along(coef)
    )
    assign(key, out, envir = cache)
    return(out)
  }

  if (model_type %in% location_dependent) {
    er <- suppressWarnings(as.numeric(eta_range))
    er <- er[is.finite(er)]
    if (!length(er)) er <- 0
    ## Quantize the requested band to two-unit boundaries.  This makes the
    ## same table reusable by many GeoKrigloc neighborhoods while still
    ## avoiding the numerically degenerate extreme-location grid that a fixed
    ## global range would create for Beta2/Kumaraswamy2.
    elo <- 2 * floor((min(er) - 0.5) / 2)
    ehi <- 2 * ceiling((max(er) + 0.5) / 2)
    if (!is.finite(elo) || !is.finite(ehi))
      stop("Invalid skew-Gaussian marginal-location range.", call. = FALSE)
    if (ehi <= elo) ehi <- elo + 4
    if ((ehi - elo) < 4) {
      mid <- 0.5 * (elo + ehi)
      elo <- 2 * floor((mid - 1) / 2)
      ehi <- elo + 4
    }
    key <- paste0(base_key, "|eta=", format(elo, scientific = FALSE), ":",
                  format(ehi, scientific = FALSE))
    if (exists(key, envir = cache, inherits = FALSE))
      return(get(key, envir = cache, inherits = FALSE))

    neta_eff <- max(as.integer(neta), as.integer(4L * ceiling(ehi - elo) + 1L))
    eg <- seq(elo, ehi, length.out = neta_eff)
    ff <- .GeoSkewGaussianFeatureGrams(
      model_type, params, eta_copula, eta_grid = eg,
      degree = degree, nquad = nquad
    )
    out <- list(kind = "location", eta = eg, grams = ff$grams,
                gram_flat = unlist(ff$grams, use.names = FALSE),
                degree = seq_along(ff$grams), eta_range = c(elo, ehi))
    assign(key, out, envir = cache)
    return(out)
  }

  stop("Unsupported margin in skew-Gaussian covariance.", call. = FALSE)
}


# Reference evaluator retained for numerical regression tests.  It mirrors the
# historical skew-Gaussian covariance loop exactly at R level.
.GeoCovSeriesForwardR <- function(rho, coefficient, scale = NULL) {
  rho <- as.numeric(rho)
  n <- length(rho)
  if (!n) return(numeric(0))
  coefficient <- as.numeric(coefficient)
  out <- numeric(n)
  rp <- rho
  scaled <- !is.null(scale)
  if (scaled) {
    scale <- as.numeric(scale)
    if (length(scale) == 1L) scale <- rep(scale, n)
    if (length(scale) != n)
      stop("Internal covariance-series scale has incompatible length.", call. = FALSE)
  }
  for (k in seq_along(coefficient)) {
    ck <- rep(coefficient[k], n)
    if (scaled) ck <- ck * scale
    out <- out + ck * rp
    rp <- rp * rho
  }
  out[abs(rho) < 1e-15] <- 0
  out
}

# Native evaluator for fixed-coefficient covariance series.  One native pass
# replaces degree full-vector R passes and avoids the temporary ck/rp vectors
# created by the historical implementation.  The C loop uses the same
# low-to-high power accumulation order as .GeoCovSeriesForwardR().
.GeoCovSeriesForwardC <- function(rho, coefficient, scale = NULL) {
  rho <- as.numeric(rho)
  n <- length(rho)
  if (!n) return(numeric(0))
  coefficient <- as.numeric(coefficient)
  degree <- length(coefficient)
  if (!degree) return(numeric(n))

  scaled <- !is.null(scale)
  if (scaled) {
    scale <- as.numeric(scale)
    if (length(scale) == 1L) scale <- rep(scale, n)
    if (length(scale) != n)
      stop("Internal covariance-series scale has incompatible length.", call. = FALSE)
  } else {
    ## C64 still needs a valid address even though the native routine does not
    ## dereference scale when use_scale == 0.
    scale <- 1
  }

  dotCall64::.C64(
    "GeoCovSeriesForward",
    SIGNATURE = c("double", "integer", "double", "integer",
                  "double", "integer", "double"),
    rho, as.integer(n), coefficient, as.integer(degree),
    scale, as.integer(scaled),
    out = dotCall64::vector_dc("double", n),
    INTENT = c("r", "r", "r", "r", "r", "r", "w"),
    NAOK = TRUE, PACKAGE = "GeoModels"
  )$out
}

.GeoCovSeriesForward <- function(rho, coefficient, scale = NULL) {
  if (isTRUE(getOption("GeoModels.covseries.native", TRUE)))
    .GeoCovSeriesForwardC(rho, coefficient, scale = scale)
  else
    .GeoCovSeriesForwardR(rho, coefficient, scale = scale)
}

# Reference evaluator for location-dependent covariance coefficients.  This is
# the historical R path retained for regression tests and for disabling the
# native kernels through GeoModels.covseries.native.
.GeoCovSeriesInterp2ForwardR <- function(rho, eta_grid, grams, eta1, eta2) {
  rho <- as.numeric(rho)
  n <- length(rho)
  if (!n) return(numeric(0))
  eta1 <- .GeoCopulaEtaVector(eta1, n)
  eta2 <- .GeoCopulaEtaVector(eta2, n)
  out <- numeric(n)
  rp <- rho
  for (k in seq_along(grams)) {
    ck <- .GeoInterp2Grid(eta_grid, grams[[k]], eta1, eta2)
    out <- out + ck * rp
    rp <- rp * rho
  }
  out[abs(rho) < 1e-15] <- 0
  out
}

# Native evaluator for Beta2/Kumaraswamy2 skew-Gaussian covariance tables.
# Interpolation indices and weights are computed once in R, rather than once
# per Hermite degree, and the degree loop plus bilinear interpolation are then
# evaluated in one native pass over all covariance pairs.
.GeoCovSeriesInterp2ForwardC <- function(rho, eta_grid, grams, eta1, eta2,
                                             gram_flat = NULL) {
  rho <- as.numeric(rho)
  n <- length(rho)
  if (!n) return(numeric(0))
  eta_grid <- as.numeric(eta_grid)
  ng <- length(eta_grid)
  if (ng < 2L)
    stop("Internal covariance interpolation grid is too short.", call. = FALSE)
  degree <- length(grams)
  if (!degree) return(numeric(n))
  eta1 <- .GeoCopulaEtaVector(eta1, n)
  eta2 <- .GeoCopulaEtaVector(eta2, n)
  w <- .GeoInterp2Weights(eta_grid, eta1, eta2)

  if (is.null(gram_flat)) {
    gram_flat <- unlist(lapply(grams, function(z) {
      z <- as.matrix(z)
      if (!identical(dim(z), c(ng, ng)))
        stop("Internal covariance interpolation table has inconsistent dimensions.",
             call. = FALSE)
      as.numeric(z)
    }), use.names = FALSE)
  } else {
    gram_flat <- as.numeric(gram_flat)
    if (length(gram_flat) != degree * ng * ng)
      stop("Internal flattened covariance table has incompatible length.",
           call. = FALSE)
  }

  dotCall64::.C64(
    "GeoCovSeriesInterp2Forward",
    SIGNATURE = c("double", "integer", "double", "integer", "integer",
                  "integer", "integer", "double", "double", "double"),
    rho, as.integer(n), gram_flat, as.integer(ng), as.integer(degree),
    w$i1, w$i2, w$a1, w$a2,
    out = dotCall64::vector_dc("double", n),
    INTENT = c("r", "r", "r", "r", "r", "r", "r", "r", "r", "w"),
    NAOK = TRUE, PACKAGE = "GeoModels"
  )$out
}

.GeoCovSeriesInterp2Forward <- function(rho, eta_grid, grams, eta1, eta2,
                                            gram_flat = NULL) {
  if (isTRUE(getOption("GeoModels.covseries.native", TRUE)))
    .GeoCovSeriesInterp2ForwardC(
      rho, eta_grid, grams, eta1, eta2, gram_flat = gram_flat
    )
  else
    .GeoCovSeriesInterp2ForwardR(rho, eta_grid, grams, eta1, eta2)
}

.GeoSkewGaussianCopulaCrossCovFast <- function(rho, model_type, params,
                                                eta1 = 0, eta2 = eta1,
                                                degree = 36L, nquad = 60L,
                                                neta = 51L, eta_range = NULL) {
  model_type <- as.integer(model_type)
  if (!(model_type %in% .GeoCopulaPredictionModels))
    stop("Skew-Gaussian linear prediction is not implemented for model id ",
         model_type, ".", call. = FALSE)
  if (model_type %in% .GeoGaussianCopulaDiscreteModels) {
    degree <- max(as.integer(degree), 48L)
    nquad <- max(as.integer(nquad), 80L)
  }
  rho <- as.numeric(rho); n <- length(rho)
  if (!n) return(numeric(0))
  eta1 <- .GeoCopulaEtaVector(eta1, n)
  eta2 <- .GeoCopulaEtaVector(eta2, n)
  if (any(!is.finite(rho)) || any(abs(rho) > 1 + 1e-10))
    stop("Invalid latent correlations in skew-Gaussian covariance.", call. = FALSE)
  rho <- pmax(-1, pmin(1, rho))

  eta_copula <- .GeoCopulaParamScalar(params, "nu")
  if (abs(eta_copula) >= 1)
    stop("Skew-Gaussian linear prediction requires nu (eta) in (-1,1).", call. = FALSE)

  ## eta=0 is exactly the Gaussian copula.
  if (abs(eta_copula) < 1e-10) {
    return(.GeoGaussianCopulaCrossCov(
      rho, model_type, params, eta1 = eta1, eta2 = eta2
    ))
  }

  lookup_eta_range <- if (is.null(eta_range)) {
    range(c(eta1, eta2), finite = TRUE)
  } else {
    eta_range
  }
  lookup <- .GeoSkewGaussianCovLookupTable(
    model_type, params, eta_copula,
    degree = degree, nquad = nquad, neta = neta,
    eta_range = lookup_eta_range
  )

  ## Fixed and multiplicative margins have one coefficient vector shared by
  ## all pairs.  Location-dependent Beta2/Kumaraswamy2 margins use the same
  ## cached Hermite Gram tables, with bilinear interpolation and series
  ## accumulation fused in a native pass.
  if (!identical(lookup$kind, "location")) {
    scale <- if (identical(lookup$kind, "multiplicative"))
      exp(eta1 + eta2) else NULL
    return(.GeoCovSeriesForward(rho, lookup$coef, scale = scale))
  }

  .GeoCovSeriesInterp2Forward(
    rho, lookup$eta, lookup$grams, eta1 = eta1, eta2 = eta2,
    gram_flat = lookup$gram_flat
  )
}


## Common observed-scale covariance dispatcher for the three copula families
## supported by GeoModels, including the audited count margins used by linear
## prediction.  Keeping the dispatch in one place
## guarantees that GeoCovmatrix(), GeoCovariogram(), and GeoKrig() use the
## same numerical covariance engine.
.GeoCopulaCrossCovFast <- function(copula, rho, model_type, params,
                                    eta1 = 0, eta2 = eta1, eta_range = NULL) {
  if (is.null(copula) || length(copula) != 1L || is.na(copula))
    stop("A single copula name is required for copula covariance.", call. = FALSE)
  copula <- as.character(copula)
  if (!(copula %in% c("Gaussian", "Clayton", "SkewGaussian")))
    stop("Copula covariance is implemented for 'Gaussian', 'Clayton', and 'SkewGaussian'.",
         call. = FALSE)

  model_type <- as.integer(model_type)
  if (!(model_type %in% .GeoCopulaPredictionModels))
    stop("Observed-scale covariance for copula='", copula,
         "' is not implemented for model id ", model_type, ".", call. = FALSE)

  if (identical(copula, "Gaussian")) {
    return(.GeoGaussianCopulaCrossCov(
      rho, model_type, params, eta1 = eta1, eta2 = eta2
    ))
  }
  if (identical(copula, "Clayton")) {
    return(.GeoClaytonCopulaCrossCovFast(
      rho, model_type, params, eta1 = eta1, eta2 = eta2,
      eta_range = eta_range
    ))
  }
  .GeoSkewGaussianCopulaCrossCovFast(
    rho, model_type, params, eta1 = eta1, eta2 = eta2,
    eta_range = eta_range
  )
}


################################################################################
## Memory-bounded dense assembly for observed-scale copula covariance matrices.##
##                                                                            ##
## The native correlation routines return the strict lower triangle in         ##
## column-major order.  The former implementation materialized, at the same    ##
## time, a full pair-index matrix, two pairwise eta vectors, the complete       ##
## transformed covariance vector, and a lower.tri() logical matrix.  For large ##
## n those temporary objects can be comparable in size to the covariance       ##
## matrix itself.  Here the same ordered correlation vector is consumed in     ##
## bounded contiguous column blocks and written directly to the final matrix.  ##
################################################################################
.GeoDenseCopulaCovarianceBlocked <- function(raw_corr, dime, copula, model_type,
                                              params, eta_obs, variances,
                                              nugget_scale,
                                              block_pairs = 1000000L) {
  dime <- as.integer(dime)
  if (length(dime) != 1L || is.na(dime) || dime < 1L)
    stop("Invalid covariance dimension.", call. = FALSE)

  expected <- as.double(dime) * as.double(dime - 1L) / 2
  if (length(raw_corr) != expected)
    stop("Internal copula covariance ordering mismatch.", call. = FALSE)

  variances <- as.numeric(variances)
  if (length(variances) == 1L) variances <- rep(variances, dime)
  if (length(variances) != dime || any(!is.finite(variances)))
    stop("Invalid marginal variances in copula covariance assembly.", call. = FALSE)

  eta_obs <- as.numeric(eta_obs)
  if (length(eta_obs) == 1L) eta_obs <- rep(eta_obs, dime)
  if (length(eta_obs) != dime || any(!is.finite(eta_obs)))
    stop("Invalid marginal locations in copula covariance assembly.", call. = FALSE)

  nugget_scale <- as.numeric(nugget_scale)[1L]
  if (!is.finite(nugget_scale))
    stop("Invalid nugget scaling in copula covariance assembly.", call. = FALSE)

  block_pairs <- suppressWarnings(as.double(block_pairs)[1L])
  if (!is.finite(block_pairs) || block_pairs < 1)
    block_pairs <- 1000000
  block_pairs <- floor(block_pairs)

  varcov <- matrix(0, dime, dime)
  diag(varcov) <- variances
  if (dime <= 1L) return(varcov)

  ## A single global location range makes blockwise Clayton/SkewGaussian
  ## interpolation numerically identical to a one-shot covariance transform.
  eta_range <- range(eta_obs, finite = TRUE)
  location_sensitive <- as.integer(model_type) %in% c(21L, 22L, 26L, 42L, 50L,
                                                       .GeoGaussianCopulaDiscreteModels)

  pos <- 1
  col_start <- 1L
  while (col_start < dime) {
    first_len <- dime - col_start
    ncols <- max(1, floor(block_pairs / first_len))
    ncols <- min(ncols, dime - col_start)
    cols <- seq.int(col_start, length.out = ncols)
    lens <- dime - cols
    np <- sum(lens)

    idx <- seq.int(pos, length.out = np)
    rho <- as.numeric(raw_corr[idx]) * nugget_scale

    if (location_sensitive) {
      ## Generate only block-sized eta vectors; no global n(n-1)/2 pair index
      ## matrix (or list of per-column eta vectors) is ever materialized.
      eta1 <- numeric(np)
      eta2 <- numeric(np)
      eta_pos <- 1L
      for (ii in seq_along(cols)) {
        j <- cols[ii]
        nr <- lens[ii]
        take <- seq.int(eta_pos, length.out = nr)
        eta1[take] <- eta_obs[seq.int(j + 1L, dime)]
        eta2[take] <- eta_obs[j]
        eta_pos <- eta_pos + nr
      }
    } else {
      eta1 <- 0
      eta2 <- 0
    }

    cv <- .GeoCopulaCrossCovFast(
      copula, rho, model_type, params,
      eta1 = eta1, eta2 = eta2, eta_range = eta_range
    )
    if (length(cv) != np || any(!is.finite(cv)))
      stop("Invalid block in copula covariance assembly.", call. = FALSE)

    ## Write each strict-lower column and its transpose directly.  This avoids
    ## lower.tri() logical matrices and avoids a second full covariance vector.
    local_pos <- 1L
    for (ii in seq_along(cols)) {
      j <- cols[ii]
      nr <- lens[ii]
      rows <- seq.int(j + 1L, dime)
      take <- seq.int(local_pos, length.out = nr)
      vals <- cv[take]
      varcov[rows, j] <- vals
      varcov[j, rows] <- vals
      local_pos <- local_pos + nr
    }

    pos <- pos + np
    col_start <- col_start + ncols
  }

  if ((pos - 1) != length(raw_corr))
    stop("Internal copula covariance block accounting mismatch.", call. = FALSE)
  varcov
}


###############################################
Cmatrix_copula <- function(bivariate, coordx, coordy,coordz, coordt,corrmodel, dime, n, ns, NS, nuisance, numpairs,
                           numpairstot, model, paramcorr, setup, radius, spacetime, spacetime_dyn,type,copula,ML,other_nuis)
    {
###################################################################################
############### computing correlation #############################################
###################################################################################

  if(type=="Standard") {
    if(spacetime) 
       cr=dotCall64::.C64('CorrelationMat_st_dyn2',SIGNATURE = c(rep("double",5),"integer","double","double","double","integer","integer"),
            corr=dotCall64::vector_dc("double",numpairstot),coordx,coordy,coordz,coordt,
            corrmodel,nuisance,paramcorr,radius,ns,NS,
            INTENT = c("w",rep("r",10)),
            PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
     if(bivariate) 
       cr=dotCall64::.C64('CorrelationMat_biv_dyn2',SIGNATURE = c(rep("double",5),"integer","double","double","double","integer","integer"),
            corr=dotCall64::vector_dc("double",numpairstot),coordx,coordy,coordz,coordt,
            corrmodel,nuisance,paramcorr,radius,ns,NS,
            INTENT = c("w",rep("r",10)),
            PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
   if(!bivariate&&!spacetime)  
        cr=dotCall64::.C64('CorrelationMat2',SIGNATURE = c(rep("double",5),"integer","double","double","double","integer","integer"),
            corr=dotCall64::vector_dc("double",numpairstot),coordx,coordy,coordz,coordt,
            corrmodel,nuisance,paramcorr,radius,ns,NS,
            INTENT = c("w",rep("r",10)),
            PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
    
}
###############################################################
if(type=="Tapering")  {
        fname <- 'CorrelationMat_tap'
        if(spacetime) fname <- 'CorrelationMat_st_tap'
       if(bivariate) fname <- 'CorrelationMat_biv_tap'
#############
cr=dotCall64::.C64(fname,SIGNATURE = c("double","double","double","double","double","integer","double","double","double","integer","integer"),
     corr=dotCall64::vector_dc("double",numpairs), coordx,coordy,coordz,coordt,corrmodel,nuisance, paramcorr,radius,ns,NS,
 INTENT = c("w","r","r","r","r","r","r","r", "r", "r", "r"),
            PACKAGE='GeoModels', VERBOSE = 0, NAOK = TRUE)
#############
     ## deleting correlation equual  to 1 because there are problems  with hipergeometric function
        sel=(abs(cr$corr-1)<.Machine$double.eps);cr$corr[sel]=0
      }
  
raw_corr <- cr$corr
cr <- NULL
nugget_scale <- 1 - as.numeric(nuisance['nugget'])
if (!length(nugget_scale) || !is.finite(nugget_scale[1L])) nugget_scale <- 1
nugget_scale <- nugget_scale[1L]
if(model == 11L) {
  ## GeoKrig expands a scalar Binomial trial count to one value per
  ## observation before covariance construction.  The Gaussian-copula
  ## Hermite covariance below currently uses one common marginal size, so
  ## collapse an equal vector back to its scalar value.  Do not silently
  ## use the first value when trial counts are genuinely location-specific.
  n_bin <- as.numeric(n)
  if (!length(n_bin) || any(!is.finite(n_bin)))
    stop("Cmatrix_copula: Binomial n must contain finite values.", call. = FALSE)
  if (length(n_bin) > 1L && any(n_bin != n_bin[1L]))
    stop("Cmatrix_copula: copula-based Binomial covariance currently requires a common n across locations; location-specific n is not yet implemented.", call. = FALSE)
  nuisance["n"] <- n_bin[1L]
}
if(model == 16L) {
  ## Negative-Binomial n is a common field parameter by construction.
  nuisance["n"] <- as.numeric(n)[1L]
}

if (!(copula %in% c("Gaussian", "Clayton", "SkewGaussian"))) {
  stop("Cmatrix_copula() currently supports Gaussian, Clayton-like, and skew-Gaussian copulas.")
}

common_copula_model <- as.integer(model) %in% .GeoGaussianCopulaContinuousModels ||
  (copula %in% c("Clayton", "SkewGaussian") &&
   as.integer(model) %in% .GeoGaussianCopulaDiscreteModels)
if (!bivariate && common_copula_model) {
  ## ML contains the location predictor X beta for models whose covariance
  ## changes with the marginal location (positive-scale margins, Beta2 and
  ## Kumaraswamy2).  For additive/fixed-shape margins it is harmless to use
  ## the intercept because their covariance does not depend on location.
  eta0 <- suppressWarnings(as.numeric(nuisance["mean"]))
  if (!length(eta0) || !is.finite(eta0[1L])) eta0 <- 0
  eta_obs <- if (is.numeric(ML) && length(ML) == dime && all(is.finite(ML))) {
    as.numeric(ML)
  } else {
    rep(eta0[1L], dime)
  }

  moments <- .GeoCopulaMarginalMoments(model, nuisance, eta_obs)
  vv <- as.numeric(moments$variance)

  if (type == "Standard") {
    varcov <- .GeoDenseCopulaCovarianceBlocked(
      raw_corr = raw_corr, dime = dime, copula = copula, model_type = model,
      params = nuisance, eta_obs = eta_obs, variances = vv,
      nugget_scale = nugget_scale
    )
  } else if (type == "Tapering") {
    corr <- as.numeric(raw_corr) * nugget_scale
    nr <- diff(setup$ia)
    row_id <- rep(seq_len(dime), nr)
    col_id <- as.integer(setup$ja)
    if (length(row_id) != length(corr) || length(col_id) != length(corr) ||
        any(col_id < 1L | col_id > dime))
      stop("Internal sparse copula covariance ordering mismatch.", call. = FALSE)
    cova <- .GeoCopulaCrossCovFast(
      copula, corr, model, nuisance,
      eta1 = eta_obs[row_id], eta2 = eta_obs[col_id],
      eta_range = range(eta_obs, finite = TRUE)
    )
    varcov <- new("spam", entries = cova, colindices = setup$ja,
                  rowpointers = setup$ia, dimension = as.integer(rep(dime, 2)))
    diag(varcov) <- vv
  }
} else {
  ## Legacy fallback for older Gaussian-copula models not handled by the
  ## common observed-scale covariance engine above.
  if (copula %in% c("Clayton", "SkewGaussian"))
    stop(sprintf("%s covariance is not implemented for model id %s.", copula, model), call. = FALSE)
  corr <- as.numeric(raw_corr) * nugget_scale
  cova <- gaussian_copula_cov(corr, model, nuisance)
  vv <- variance_disp(model, nuisance)
  if(!bivariate) {
    if(type=="Standard") {
      varcov <- diag(dime)
      varcov[lower.tri(varcov)] <- cova
      varcov <- t(varcov)
      varcov[lower.tri(varcov)] <- cova
      diag(varcov) <- vv
    }
    if(type=="Tapering") {
      varcov <- new("spam", entries=cova, colindices=setup$ja,
                    rowpointers=setup$ia, dimension=as.integer(rep(dime,2)))
      diag(varcov) <- vv
    }
  }
}

return(varcov)
}
######################################################################################################

GeoKrigWeights <- function(
  coordx, coordy = NULL, coordz = NULL, coordt = NULL, coordx_dyn = NULL,
  corrmodel, distance = "Eucl", grid = FALSE, loc,
  method = "cholesky", model = "Gaussian", n = 1, nloc = NULL,
  param, anisopars = NULL, radius = 1, sparse = FALSE, time = NULL,
  which = 1, copula = NULL, X = NULL, Xloc = NULL, Mloc = NULL
) {
  ############################################################################
  ## GeoKrigWeights: global kriging weights
  ##
  ## Conservative updated version.
  ## Main reliable target: Gaussian latent/global weights.
  ## The default orientation is kept as nobs x nloc for compatibility with
  ## existing GeoSimcond code using crossprod(W, x).
  ############################################################################

  call <- match.call()

  ## Internal defaults kept for compatibility with previous internal code.
  ## They are intentionally not formal arguments, so the public API matches
  ## the existing Rd documentation.
  return_covmatrix <- TRUE
  return_CC <- TRUE
  weights_orientation <- "nobs_by_nloc"

  method <- gsub("[[:blank:]]", "", method)
  model <- .GeoPredictionCanonicalModel(gsub("[[:blank:]]", "", model))
  corrmodel <- gsub("[[:blank:]]", "", corrmodel)
  distance <- gsub("[[:blank:]]", "", distance)

  if (is.null(CkModel(model))) {
    stop("The name of the model is not correct\n")
  }
  corrmodel_code <- CkCorrModel(corrmodel)
  if (!is.character(corrmodel) || is.null(corrmodel_code)) {
    stop("the name of the correlation model is wrong")
  }

  bivariate <- CheckBiv(corrmodel_code)
  spacetime <- CheckST(corrmodel_code)

  ## This version is intentionally conservative.  GeoSimcond uses this function
  ## for Gaussian latent weights, so do not silently return approximate/wrong
  ## weights for models whose cross-covariance requires model-specific code.
  if (!identical(model, "Gaussian")) {
    stop("GeoKrigWeights currently returns validated weights only for model = 'Gaussian'. ",
         "For non-Gaussian prediction use GeoKrig() or GeoKrigloc().", call. = FALSE)
  }
  if (!is.null(copula)) {
    stop("GeoKrigWeights currently requires copula = NULL; conditional-simulation ",
         "code must request weights on the latent Gaussian scale.", call. = FALSE)
  }
  if (bivariate && !(which == 1 || which == 2)) {
    stop("which parameter must be 1 or 2")
  }

  if (spacetime && !is.null(coordx_dyn)) {
    coordx_dyn <- .GeoNormalizeDynamicCoordinates(
      coordx_dyn, coordt = coordt, context = "GeoKrigWeights"
    )
    .GeoValidateDynamicRows(X, coordx_dyn, name = "X",
                            context = "GeoKrigWeights")
  }

  biv_info <- NULL
  if (bivariate) {
    biv_info <- .GeoBivariateCoordinateInfo(
      coordx = coordx, coordy = coordy, coordz = coordz,
      coordx_dyn = coordx_dyn, grid = grid,
      context = "GeoKrigWeights"
    )
    if (biv_info$dynamic) {
      coordx_dyn <- biv_info$blocks
      coordx <- coordy <- coordz <- NULL
    }
    if (!is.null(X)) {
      X <- .GeoBivariateDesign(X, biv_info$ns,
                               context = "GeoKrigWeights")
    }
  }

  ## Location matrix
  if (is.vector(loc)) {
    loc <- matrix(loc, nrow = 1)
  } else if (!is.matrix(loc)) {
    loc <- as.matrix(loc)
  }
  if (!(ncol(loc) %in% c(2L, 3L))) {
    stop("loc parameter must be a matrix N x 2 or N x 3")
  }


  .GeoValidate3D(
    coordx = coordx, coordy = coordy, coordz = coordz,
    coordx_dyn = coordx_dyn, distance = distance, grid = grid,
    context = "GeoKrigWeights"
  )
  .GeoValidate3D(
    coordx = loc, distance = distance, grid = FALSE,
    context = "GeoKrigWeights prediction locations"
  )
  .GeoValidateCoordinateCompatibility(
    coordx = coordx, coordy = coordy, coordz = coordz,
    coordx_dyn = coordx_dyn, loc = loc, context = "GeoKrigWeights"
  )

  if (spacetime && is.null(time)) {
    stop("At least one temporal instant is needed for space-time kriging\n")
  }
  if (is.null(time)) time <- 0

  ## Mean inputs are normalized after dimat/dimat2 are known.
  if (!is.null(X) && !is.matrix(X) && !is.list(X)) X <- as.matrix(X)
  Mobs <- if (!bivariate) .GeoMean_external(param) else NULL
  if (!is.null(Mobs) && !is.null(X))
    stop("A fixed observation mean and X cannot be used together.")
  if (!is.null(Mloc) && !is.null(Xloc))
    stop("Use either Mloc or Xloc, not both.")
  if (!is.null(Mobs) && is.null(Mloc))
    stop("Mloc is required when the observation mean is supplied as a vector.")
  if (!is.null(X) && is.null(Xloc) && is.null(Mloc))
    stop("Xloc or Mloc is required when X is supplied.")
  if (is.null(X) && !is.null(Xloc))
    stop("X must be supplied when Xloc is used.")
  if (is.null(X) && is.null(Mobs) &&
      length(setdiff(.GeoMean_supplied_names(param), "mean")))
    stop("Parameters mean1, mean2, ... require an explicit design matrix X; without X the mean is intercept-only.")

  ## Prediction locations in the metric used by Corr_c.
  ## GeoCovmatrix handles the observed covariance.  For the cross-covariance we
  ## apply the same anisotropic map to ccc and loc below.  If the caller already
  ## passes anisotropized coordinates, anisopars must be NULL.
  loc_metric <- loc
  if (!is.null(anisopars)) {
    loc_metric <- GeoAniso(loc_metric, c(anisopars$angle, anisopars$ratio))
  }

  numloc <- nrow(loc_metric)
  tloc <- length(time)
  if (!tloc) tloc <- 1L
  if (bivariate) tloc <- 1L

  if (ncol(loc_metric) == 2L) {
    locx <- loc_metric[, 1L]
    locy <- loc_metric[, 2L]
    locz <- rep(0, numloc)
  } else {
    locx <- loc_metric[, 1L]
    locy <- loc_metric[, 2L]
    locz <- loc_metric[, 3L]
  }

  ############################################################################
  ## Build covariance matrix at observed sites
  ############################################################################
  param_cov <- param
  if (!is.null(Mobs)) param_cov$mean <- 0
  covmatrix <- GeoCovmatrix(
    coordx = coordx, coordy = coordy, coordz = coordz, coordt = coordt,
    coordx_dyn = coordx_dyn, corrmodel = corrmodel, distance = distance,
    grid = grid, model = model, n = n, param = param_cov,
    anisopars = anisopars, radius = radius, sparse = sparse,
    copula = copula, X = X
  )
  covmatrix$param <- unlist(covmatrix$param)

  spacetime_dyn <- !is.null(covmatrix$coordx_dyn)
  if (!spacetime_dyn) {
    dimat <- covmatrix$numcoord * covmatrix$numtime
  } else {
    dimat <- sum(covmatrix$ns)
  }
  dimat2 <- numloc * tloc

  if (!bivariate) {
    if (is.null(X)) X <- covmatrix$X
    X <- .GeoMean_design(X, n = dimat, name = "X")
    p_mean <- ncol(X)
    invisible(.GeoMean_beta(param_cov, p_mean))
    if (!is.null(Mobs) && length(Mobs) != dimat)
      stop("The fixed observation mean must have one value per observation.")

    if (!is.null(Mloc)) {
      if (!is.numeric(Mloc) || any(!is.finite(Mloc)))
        stop("Mloc must be finite and numeric.")
      Mloc <- as.numeric(Mloc)
      if (length(Mloc) == 1L) Mloc <- rep(Mloc, dimat2)
      if (length(Mloc) != dimat2)
        stop("Mloc must have one value per prediction point/time combination.")
      Xloc <- NULL
    } else {
      Xloc <- .GeoMean_design(Xloc, n = dimat2, p = p_mean, name = "Xloc")
    }
  } else {
    X <- .GeoBivariateDesign(
      if (is.null(X)) covmatrix$X else X,
      covmatrix$ns, context = "GeoKrigWeights"
    )
    p_mean <- ncol(X)
    invisible(.GeoBivariateValidateMeanParameters(
      covmatrix$param, p_mean, context = "GeoKrigWeights"
    ))
    if (!is.null(Mloc)) {
      if (!is.numeric(Mloc) || any(!is.finite(Mloc)))
        stop("Mloc must be finite and numeric.")
      Mloc <- as.numeric(Mloc)
      if (length(Mloc) == 1L) Mloc <- rep(Mloc, dimat2)
      if (length(Mloc) != dimat2)
        stop("Mloc must have one value per prediction location.")
      Xloc <- NULL
    } else {
      Xloc <- .GeoMean_design(Xloc, n = dimat2, p = p_mean, name = "Xloc")
    }
  }

  ############################################################################
  ## Observed coordinates stacked as required by Corr_c
  ############################################################################
  NS <- 0L
  if (spacetime || bivariate) {
    NS <- cumsum(covmatrix$ns)
    NS <- c(0L, NS)[-(length(covmatrix$ns) + 1L)]
  }

  if (!grid) {
    if (!spacetime && !bivariate) {
      if (is.null(covmatrix$coordz)) {
        ccc <- cbind(covmatrix$coordx, covmatrix$coordy, 0)
      } else {
        ccc <- cbind(covmatrix$coordx, covmatrix$coordy, covmatrix$coordz)
      }
    }
  }

  if (grid) {
    if (is.null(coordz)) {
      ccc <- as.matrix(expand.grid(covmatrix$coordx, covmatrix$coordy))
      ccc <- cbind(ccc, 0)
    } else {
      ccc <- as.matrix(expand.grid(covmatrix$coordx, covmatrix$coordy, covmatrix$coordz))
    }
    grid <- FALSE
  } else {
    if ((spacetime || bivariate) && !spacetime_dyn) {
      if (!is.null(covmatrix$coordz)) {
        ccc <- cbind(rep(covmatrix$coordx, covmatrix$numtime),
                     rep(covmatrix$coordy, covmatrix$numtime),
                     rep(covmatrix$coordz, covmatrix$numtime))
      } else {
        ccc <- cbind(rep(covmatrix$coordx, covmatrix$numtime),
                     rep(covmatrix$coordy, covmatrix$numtime), 0)
      }
    }
    if ((spacetime || bivariate) && spacetime_dyn) {
      ccc <- do.call(rbind, args = c(covmatrix$coordx_dyn))
      if (ncol(ccc) == 2L) ccc <- cbind(ccc, 0)
    }
  }

  if (!exists("ccc")) stop("Could not build observed coordinate matrix")
  ccc <- as.matrix(ccc)
  if (ncol(ccc) == 2L) ccc <- cbind(ccc, 0)
  if (nrow(ccc) != dimat) stop("Observed coordinate matrix has wrong number of rows")

  if (!is.null(anisopars)) {
    ccc <- GeoAniso(ccc, c(anisopars$angle, anisopars$ratio))
  }

  ############################################################################
  ## Cross-correlation and cross-covariance CC
  ############################################################################
  corrmodel_num <- CkCorrModel(covmatrix$corrmodel)
  distance_num <- CheckDistance(covmatrix$distance)
  corrparam <- covmatrix$param[covmatrix$namescorr]

  if (!(covmatrix$model %in% c(1))) {
    stop("Internal error: GeoKrigWeights is restricted to Gaussian model code.", call. = FALSE)
  }

  cc_call <- dotCall64::.C64(
    "Corr_c",
    SIGNATURE = c("double", "double", "double", "double", "double", "integer",
                  "integer", "double", "double", "double", "integer", "integer",
                  "integer", "integer", "integer", "integer", "double", "integer",
                  "integer", "double", "integer", "integer", "double"),
    corri = dotCall64::vector_dc("double", dimat * dimat2),
    ccc[, 1L], ccc[, 2L], ccc[, 3L], covmatrix$coordt,
    corrmodel_num, as.integer(0),
    locx, locy, locz,
    covmatrix$numcoord, numloc, tloc,
    covmatrix$ns, NS, covmatrix$numtime,
    corrparam,
    covmatrix$spacetime, covmatrix$bivariate,
    time, distance_num, as.integer(which - 1L), covmatrix$radius,
    INTENT = c("w", rep("r", 22)),
    NAOK = TRUE, PACKAGE = "GeoModels"
  )

  rho <- cc_call$corri

  if (bivariate) {
    ## Corr_c evaluates the bivariate covariance model directly.  The returned
    ## vector already includes marginal sills and cross-covariance scaling.
    corri <- rho
  } else {
    nug <- as.numeric(covmatrix$param["nugget"])
    sill <- as.numeric(covmatrix$param["sill"])
    if (is.na(nug)) nug <- 0
    if (is.na(sill)) stop("sill parameter is missing")
    corri <- (1 - nug) * rho
    vvar <- sill
  }

  CC <- if (bivariate) {
    matrix(corri, nrow = dimat, ncol = dimat2)
  } else {
    matrix(corri * as.numeric(vvar), nrow = dimat, ncol = dimat2)
  }
  if (nrow(CC) != dimat || ncol(CC) != dimat2) {
    stop("Wrong dimension of cross-covariance matrix CC")
  }

  ############################################################################
  ## Final weights
  ############################################################################
  weights_result <- getInvC(covmatrix, CC, mse = FALSE)
  krig_weights <- weights_result$a

  ## Default is the historical orientation used by GeoSimcond:
  ## kriging vector = crossprod(weights, data), i.e. weights is nobs x nloc.
  if (weights_orientation == "nloc_by_nobs") {
    krig_weights <- t(krig_weights)
  }

  result <- list(
    weights = krig_weights,
    weights_orientation = weights_orientation,
    model = model,
    corrmodel = corrmodel,
    coordx = covmatrix$coordx,
    coordy = covmatrix$coordy,
    coordz = covmatrix$coordz,
    coordx_dyn = covmatrix$coordx_dyn,
    bivariate = bivariate,
    spacetime = covmatrix$spacetime,
    param = covmatrix$param,
    xloc = Xloc,
    Mloc = Mloc,
    mean_obs = Mobs,
    tloc = tloc,
    loc = loc,
    loc_metric = loc_metric,
    call = call
  )

  if (return_covmatrix) result$covmatrix <- covmatrix$covmatrix
  if (return_CC) result$CC <- CC

  return(result)
}

####################################################
### File name: GeoFit.r
####################################################

.GeoFit_validate_likelihood_type <- function(likelihood, type) {
  if (!is.character(likelihood) || length(likelihood) != 1L || is.na(likelihood))
    stop("likelihood must be a single character string.", call. = FALSE)
  if (!is.character(type) || length(type) != 1L || is.na(type))
    stop("type must be a single character string.", call. = FALSE)

  if (identical(likelihood, "Difference") || identical(type, "Difference")) {
    stop("The Difference composite likelihood is no longer supported.", call. = FALSE)
  }

  allowed <- switch(likelihood,
    Full = c("Standard", "Tapering", "Tapering1", "Tapering2", "CV"),
    Marginal = c("Pairwise", "Independence"),
    Conditional = "Pairwise",
    Marginal_2 = "Pairwise",
    NULL
  )
  if (is.null(allowed)) {
    stop("The setting name of the likelihood objects is not correct.", call. = FALSE)
  }
  if (!(type %in% allowed)) {
    stop(sprintf(
      "likelihood='%s' is not compatible with type='%s'. Allowed type value(s): %s.",
      likelihood, type, paste(allowed, collapse = ", ")
    ), call. = FALSE)
  }
  invisible(TRUE)
}


.GeoFit_parameter_suggestions <- function(bad, valid) {
  valid <- unique(valid[nzchar(valid)])
  if (!length(valid)) return(character(0))
  out <- character(0)
  for (nm in bad) {
    dd <- as.numeric(utils::adist(nm, valid, ignore.case = TRUE))
    if (!length(dd) || all(!is.finite(dd))) next
    jj <- which.min(dd)
    lim <- max(1, min(3, ceiling(nchar(nm) / 2)))
    if (is.finite(dd[jj]) && dd[jj] <= lim)
      out <- c(out, sprintf("'%s' -> '%s'", nm, valid[jj]))
  }
  out
}

.GeoFit_unit_sill_models <- function() {
  c(
    "Weibull", "Poisson", "Binomial", "Gamma", "LogLogistic",
    "BinomialNeg", "Binary", "Bernoulli", "Geom", "Geometric",
    "Gaussian_misp_Poisson", "Binary_misp_BinomialNeg",
    "PoissonZIP", "Gaussian_misp_PoissonZIP", "BinomialNegZINB", "BinomNeg",
    "PoissonZIP1", "Gaussian_misp_PoissonZIP1", "BinomialNegZINB1",
    "Gaussian_misp_PoissonGamma", "PoissonGamma", "PoissonGammaZIP",
    "PoissonGammaZIP1", "Beta2", "Kumaraswamy2", "Beta", "Kumaraswamy"
  )
}

.GeoFit_normalize_unit_sill <- function(start, fixed, model, bivariate) {
  if (bivariate || !(model %in% .GeoFit_unit_sill_models())) return(fixed)

  if (!is.null(start$sill))
    stop("'sill' is fixed at 1 for this model and must not be supplied in 'start'.",
         call. = FALSE)

  if (is.null(fixed)) fixed <- list()
  if (!is.null(fixed$sill)) {
    fsill <- fixed$sill
    ok <- is.numeric(fsill) && length(fsill) == 1L && !is.na(fsill) &&
      is.finite(fsill) && isTRUE(all.equal(as.numeric(fsill), 1,
                                           tolerance = sqrt(.Machine$double.eps)))
    if (!ok)
      stop("'sill' is fixed at 1 for this model; if supplied in 'fixed', it must equal 1.",
           call. = FALSE)
  }
  fixed$sill <- 1
  fixed
}

.GeoFit_validate_parameter_names <- function(start, fixed, model, corrmodel,
                                             copula, bivariate, X) {
  num_betas <- if (!bivariate) {
    if (is.null(X)) 1L
    else if (is.list(X)) ncol(X[[1L]])
    else ncol(as.matrix(X))
  } else {
    if (is.null(X)) c(1L, 1L)
    else if (is.list(X)) c(ncol(X[[1L]]), ncol(X[[2L]]))
    else rep(ncol(as.matrix(X)), 2L)
  }

  nuisance_start <- unique(NuisParam(model, bivariate, num_betas, copula = copula))
  nuisance_fixed <- unique(NuisParam2(model, bivariate, num_betas, copula = copula))
  corr <- unique(CorrParam(corrmodel))

  check_one <- function(x, arg, nuisance) {
    valid <- unique(c(nuisance, corr))
    if (is.null(x) || !length(x)) return(invisible(TRUE))
    if (!is.list(x)) return(invisible(TRUE))  # CkInput reports the type error.

    nm <- names(x)
    if (is.null(nm) || anyNA(nm) || any(!nzchar(nm))) {
      stop(sprintf("GeoFit: every entry of '%s' must have a parameter name.", arg),
           call. = FALSE)
    }
    dup <- unique(nm[duplicated(nm)])
    if (length(dup)) {
      stop(sprintf("GeoFit: duplicated parameter name(s) in '%s': %s.",
                   arg, paste(dup, collapse = ", ")), call. = FALSE)
    }

    bad <- setdiff(nm, valid)
    if (!length(bad)) return(invisible(TRUE))

    cp <- if (is.null(copula)) "" else sprintf(" with copula='%s'", copula)
    msg <- sprintf(
      "GeoFit: invalid parameter name(s) in '%s': %s.\nValid nuisance/marginal parameter names for model='%s'%s: %s.\nValid correlation parameter names for corrmodel='%s': %s.",
      arg, paste(bad, collapse = ", "), model, cp,
      if (length(nuisance)) paste(nuisance, collapse = ", ") else "none",
      corrmodel, if (length(corr)) paste(corr, collapse = ", ") else "none"
    )
    sug <- .GeoFit_parameter_suggestions(bad, valid)
    if (length(sug))
      msg <- paste0(msg, "\nPossible correction(s): ", paste(sug, collapse = "; "), ".")
    stop(msg, call. = FALSE)
  }

  check_one(start, "start", nuisance_start)
  check_one(fixed, "fixed", nuisance_fixed)

  if (!is.null(start) && !is.null(fixed) && is.list(start) && is.list(fixed)) {
    overlap <- intersect(names(start), names(fixed))
    if (length(overlap)) {
      stop(sprintf(
        "GeoFit: parameter(s) cannot be supplied in both 'start' and 'fixed': %s.",
        paste(overlap, collapse = ", ")
      ), call. = FALSE)
    }
  }
  invisible(TRUE)
}

.GeoFit_check_st_pairwise_support <- function(model, likelihood) {
  code <- CkModel(model)
  ## These sets mirror the space-time native routines registered in
  ## GeoModels_init.c.  Spatial-only kernels are deliberately not considered
  ## here: this guard exists only to prevent an unavailable _st2mem dispatch.
  marginal_codes <- c(
    1L, 2L, 10L, 11L, 12L, 13L, 14L, 16L, 20L, 21L, 22L, 24L, 25L,
    26L, 27L, 28L, 29L, 30L, 33L, 34L, 35L, 36L, 38L, 39L, 40L, 42L,
    43L, 44L, 45L, 46L, 47L, 49L, 59L
  )
  conditional_codes <- marginal_codes
  allowed <- if (identical(likelihood, "Marginal")) marginal_codes else conditional_codes

  if (!(code %in% allowed)) {
    stop(sprintf(
      "Model '%s' is not implemented for spatio-temporal %s pairwise composite likelihood.",
      model, tolower(likelihood)
    ), call. = FALSE)
  }
  invisible(TRUE)
}

GeoFit <- function(data, coordx, coordy = NULL, coordz = NULL, coordt = NULL,
                   coordx_dyn = NULL, copula = NULL, corrmodel = NULL,
                   distance = "Eucl", fixed = NULL, anisopars = NULL,
                   est.aniso = c(FALSE, FALSE), grid = FALSE,
                   likelihood = "Marginal", lower = NULL, maxdist = Inf,
                   neighb = NULL, p_neighb = 1, maxtime = Inf, memdist = TRUE,
                   method = "cholesky", model = "Gaussian", n = 1,
                   onlyvar = FALSE, optimizer = "Nelder-Mead", radius = 1,
                   score = FALSE, sensitivity = FALSE, sparse = FALSE,
                   start = NULL, thin_method = "bernoulli", type = "Pairwise",
                   upper = NULL, varest = FALSE, weighted = FALSE, X = NULL,
                    spobj = NULL, spdata = NULL, check.duplicates = FALSE)
{
  ###########  first preliminary check  ###############
  call <- match.call()

  if (!isTRUE(memdist)) {
    warning(
      "'memdist = FALSE' is deprecated and ignored; GeoFit now always precomputes the selected pair structure.",
      call. = FALSE
    )
  }
  memdist <- TRUE

  suppressWarnings({

    ## start is genuinely optional.  WlsStart/StartParam provide conservative
    ## internal initial values for parameters omitted by the user.
    if (is.null(start)) start <- list()
    grid_requested <- isTRUE(grid)

    ## normalize strings early (before checks)
    if (!is.null(corrmodel)) corrmodel <- gsub("[[:blank:]]", "", corrmodel)
    if (!is.null(model))     model     <- gsub("[[:blank:]]", "", model)
    if (!is.null(distance))  distance  <- gsub("[[:blank:]]", "", distance)
    if (!is.null(optimizer)) optimizer <- gsub("[[:blank:]]", "", optimizer)
    if (!is.null(likelihood))likelihood<- gsub("[[:blank:]]", "", likelihood)
    if (!is.null(type))      type      <- gsub("[[:blank:]]", "", type)

    .GeoFit_validate_likelihood_type(likelihood, type)

    ## default corrmodel for Independence+Marginal
    if (is.null(corrmodel) && likelihood == "Marginal" && type == "Independence") {
      if (is.null(coordt)) {
        corrmodel <- "Exponential"
        tlist <- list(nugget = 0, scale = 1)
      } else {
        corrmodel <- "Exp_Exp"
        tlist <- list(nugget = 0, scale_s = 1, scale_t = 1)
      }
      fixed <- append(fixed, tlist)
    }
    nosym = FALSE
    ## checks on corrmodel/model/copula
    if (!is.character(corrmodel) || is.null(CkCorrModel(corrmodel)))
      stop("the name of the correlation model is wrong\n")
    model_code <- CkModel(model)
    if (is.null(model_code))
      stop("The name of the model is not correct\n")

    ## GeoLik has explicit full-likelihood objectives only for this set of
    ## model codes.  Older calls for other models (notably Gamma/Weibull)
    ## reached GeoLik with no objective name selected and failed with internal
    ## errors such as `object 'lname' not found`.  Stop here with a stable,
    ## user-facing capability message instead.
    if (likelihood == "Full") {
      full_model_codes <- c(1L, 20L, 22L, 34L, 35L, 36L, 37L, 40L, 47L)
      if (!(model_code %in% full_model_codes)) {
        stop(
          sprintf(
            "GeoFit: Full likelihood is not implemented for model='%s'. Use pairwise likelihood where available.",
            model
          ),
          call. = FALSE
        )
      }
    }

    ## Preserve the inferential model separately from the data-generating
    ## model.  For misspecified likelihoods these are intentionally different
    ## and bootstrap routines must be able to recover the original criterion.
    estimation_model <- model

    if (!is.null(copula)) {
      if (!copula %in% c("Clayton", "Gaussian", "SkewGaussian"))
        stop("the type of copula is wrong\n")
    }

    ## Independence requires Marginal likelihood
    if (type == "Independence" && likelihood != "Marginal")
      stop("Independence likelihood must be coupled with Marginal likelihood\n")

    ## Pairwise requirements (early)
    if (type == "Pairwise") {
      if (is.null(coordt)) {
        if (is.null(neighb) && is.null(maxdist))
          stop("neighb or maxdist must be fixed\n")
      } else {
        ## space-time: require (neighb OR maxdist) AND maxtime
        if ((is.null(neighb) && is.null(maxdist)) || is.null(maxtime))
          stop("neighb or maxdist and maxtime must be fixed\n")
      }
    }

    taper <- NULL
    tapsep <- NULL

    if (likelihood == "Marginal" && type == "Independence") {
      anisopars <- NULL
      est.aniso <- c(FALSE, FALSE)
    }

    if (!is.null(X) && is.null(coordx_dyn)) X <- as.matrix(X)

    if (!is.null(neighb)) {
      if (!is.numeric(neighb) || length(neighb) != 1L || !is.finite(neighb) ||
          neighb < 1 || abs(neighb - round(neighb)) > sqrt(.Machine$double.eps)) {
        stop("neighb must be a finite positive integer", call. = FALSE)
      }
      neighb <- as.integer(round(neighb))
    }

    if (!is.null(anisopars) && !is.list(anisopars))
      stop("anisopars must be a list with two elements\n")

    if (!is.character(optimizer)) stop("invalid optimizer\n")
    if (!is.character(distance))  stop("invalid distance\n")

  
    bivariate <- CheckBiv(CkCorrModel(corrmodel))
    spacetime <- CheckST(CkCorrModel(corrmodel))
    space <- !spacetime && !bivariate

    if (bivariate && type == "Pairwise") {
      if (CkModel(model) != 1L || likelihood != "Marginal") {
        stop(
          "Bivariate pairwise fitting is currently implemented only for model='Gaussian' with likelihood='Marginal'.",
          call. = FALSE
        )
      }
    }

    if (spacetime && type == "Pairwise" &&
        likelihood %in% c("Marginal", "Conditional", "Marginal_2")) {
      .GeoFit_check_st_pairwise_support(
        model, if (likelihood == "Marginal_2") "Conditional" else likelihood
      )
    }

    ## Pairwise copula likelihood is currently a univariate spatial implementation.
    ## Keep the accepted marginal set synchronized with biv_cop() and the
    ## native Comp_Pair_*Cop2mem wrappers.
    if (!is.null(copula) && type == "Pairwise") {
      if (!space)
        stop("Pairwise copula fitting is currently implemented only for univariate purely spatial data.",
             call. = FALSE)
      if (!(likelihood %in% c("Marginal", "Conditional")))
        stop("Copula fitting with type='Pairwise' requires likelihood='Marginal' or 'Conditional'.",
             call. = FALSE)

      model_cop <- switch(model,
        Gauss = "Gaussian",
        LogGauss = "LogGaussian",
        weibull = "Weibull",
        gamma = "Gamma",
        model
      )
      continuous_copula_margins <- c(
        "Gaussian", "StudentT", "LogGaussian", "Gamma", "Weibull",
        "Beta", "Beta2", "Kumaraswamy", "Kumaraswamy2",
        "Logistic", "SkewLaplace"
      )
      discrete_copula_margins <- c("Poisson", "Binomial", "BinomialNeg")
      supported_copula_margins <- if (copula %in% c("Gaussian", "SkewGaussian"))
        c(continuous_copula_margins, discrete_copula_margins)
      else
        continuous_copula_margins

      if (!(model_cop %in% supported_copula_margins)) {
        stop(sprintf(
          "Marginal model '%s' is not implemented for pairwise %s-copula fitting. Supported margins are: %s.",
          model, copula, paste(supported_copula_margins, collapse = ", ")
        ), call. = FALSE)
      }

      ## The Clayton option in GeoModels is the constructive Clayton-like RF
      ## of Bevilacqua et al.; its nu parameter is a positive integer. Because
      ## nu is discrete, it must be fixed rather than optimized continuously.
      if (copula == "Clayton") {
        ## Use exact name matching: '$nu' can partially match 'nugget' in R.
        has_nu_fixed <- !is.null(fixed) && "nu" %in% names(fixed)
        has_nu_start <- !is.null(start) && "nu" %in% names(start)
        if (!has_nu_fixed || has_nu_start)
          stop("For copula='Clayton', nu must be supplied only in fixed and must be a positive integer.",
               call. = FALSE)
        nu_clay <- as.numeric(fixed[["nu", exact = TRUE]])
        if (length(nu_clay) != 1L || !is.finite(nu_clay) || nu_clay <= 0 ||
            abs(nu_clay - round(nu_clay)) > sqrt(.Machine$double.eps))
          stop("For copula='Clayton', fixed$nu must be a finite positive integer.",
               call. = FALSE)
      }

      ## For the skew-Gaussian copula, nu is the bounded eta parameter in the
      ## paper. It may be estimated (start$nu) or fixed (fixed$nu), but it must
      ## be supplied in exactly one of them. Exact name matching is essential:
      ## in R, fixed$nu can otherwise partially match fixed$nugget.
      ## The objective also guards every optimizer proposal so Nelder-Mead
      ## cannot leave (-1,1).
      if (copula == "SkewGaussian") {
        has_nu_start <- !is.null(start) && "nu" %in% names(start)
        has_nu_fixed <- !is.null(fixed) && "nu" %in% names(fixed)
        if (identical(has_nu_start, has_nu_fixed))
          stop("For copula='SkewGaussian', nu (eta) must be supplied exactly once: either in start (estimated) or in fixed.",
               call. = FALSE)
        nu0 <- if (has_nu_start) start[["nu", exact = TRUE]] else fixed[["nu", exact = TRUE]]
        nu0 <- suppressWarnings(as.numeric(nu0))
        if (length(nu0) != 1L || !is.finite(nu0) || abs(nu0) >= 1)
          stop("For copula='SkewGaussian', nu (eta) must lie strictly between -1 and 1.",
               call. = FALSE)
      }
    }

    ###### checking if neighb or maxdist or maxtime has been specified when using cl
    if (space) {
      if (type == "Pairwise" && (likelihood == "Marginal" || likelihood == "Conditional")) {
        if (is.null(neighb) && isTRUE(maxdist == Inf))
          stop("neighb or maxdist must be specificed when using marginal or conditional pairwise composite likelihood\n")
      }
    }

    if (bivariate) {
      if (type == "Pairwise" && (likelihood == "Marginal" || likelihood == "Conditional")) {
        ## keep your original intent, but avoid sum(is.null(...)) which is odd
        if (is.null(neighb) && isTRUE(maxdist == Inf))
          stop("neighb or maxdist must be specificed when using marginal or conditional pairwise composite likelihood\n")
      }
    }

    if (spacetime) {
      if (type == "Pairwise" && (likelihood == "Marginal" || likelihood == "Conditional")) {
        if ((is.null(neighb) && isTRUE(maxdist == Inf)) && isTRUE(maxtime == Inf))
          stop("neighb or maxdist and maxtime must be specificed when using marginal or conditional pairwise composite likelihood\n")
        if ((is.null(neighb) && isTRUE(maxdist == Inf)) && isTRUE(maxtime < Inf))
          stop("neighb or maxdist must be specificed when using marginal or conditional pairwise composite likelihood\n")
        if ((!is.null(neighb) || isTRUE(maxdist < Inf)) && isTRUE(maxtime == Inf))
          stop("maxtime must be specificed when using marginal or conditional pairwise composite likelihood\n")
      }
    }
    ########

    ##############################################################################
    ###### extracting sp object informations if necessary              ###########
    ##############################################################################
    if (!is.null(spobj)) {
      if (space || bivariate) {
        a <- sp2Geo(spobj, spdata)
        coordx <- a$coords
        if (!a$pj) { if (distance != "Chor") distance <- "Geod" }
      }
      if (spacetime) {
        a <- sp2Geo(spobj, spdata)
        coordx <- a$coords
        coordt <- a$coordt
        if (!a$pj) { if (distance != "Chor") distance <- "Geod" }
      }
      if (!is.null(a$Y)) data <- a$Y
      if (!is.null(a$X)) X <- a$X
    }

    if (!is.null(anisopars) && tolower(distance) != "eucl") {
      stop("anisopars requires distance='Eucl'", call. = FALSE)
    }

    ## Canonicalize regular-grid inputs before generic input checking and
    ## likelihood setup.  This avoids dimension-dependent grid branches and
    ## makes spatial, space-time and bivariate grids use the same explicit
    ## coordinate/data layouts as irregular designs.
    if (grid_requested) {
      if (!is.null(coordx_dyn))
        stop("grid=TRUE cannot be combined with coordx_dyn", call. = FALSE)
      gg <- .GeoVariogram_normalize_grid(
        data = data, coordx = coordx, coordy = coordy, coordz = coordz,
        coordt = coordt, bivariate = bivariate
      )
      data <- gg$data
      coordx <- gg$coordx
      coordy <- gg$coordy
      coordz <- gg$coordz
      grid <- FALSE
    }

    if (spacetime && !is.null(coordx_dyn)) {
      coordx_dyn <- .GeoNormalizeDynamicCoordinates(
        coordx_dyn, coordt = coordt, context = "GeoFit"
      )
      .GeoValidateDynamicData(
        data, coordx_dyn, context = "GeoFit", require_list = TRUE
      )
      .GeoValidateDynamicRows(X, coordx_dyn, name = "X", context = "GeoFit")
    }
    if (bivariate) {
      biv_info <- .GeoBivariateCoordinateInfo(
        coordx = coordx, coordy = coordy, coordz = coordz,
        coordx_dyn = coordx_dyn, grid = grid, context = "GeoFit"
      )
      if (biv_info$dynamic) coordx_dyn <- biv_info$blocks
      biv_data <- .GeoBivariateDataVector(
        data, biv_info$ns, context = "GeoFit", allow_na = TRUE
      )
      ## Downstream fitting code historically expects a 2 x n matrix for
      ## common locations and a two-element list for heterotopic locations.
      data <- if (biv_info$dynamic) {
        list(
          biv_data[seq_len(biv_info$ns[1L])],
          biv_data[biv_info$ns[1L] + seq_len(biv_info$ns[2L])]
        )
      } else {
        matrix(biv_data, nrow = 2L, byrow = TRUE)
      }
      if (!is.null(X)) {
        X <- .GeoBivariateDesign(X, biv_info$ns, context = "GeoFit")
      }
      .GeoBivariateValidateFitMeans(
        start, fixed, X, biv_info$ns, context = "GeoFit"
      )
    }

    .GeoCheckDuplicateLocations(
      coordx = coordx, coordy = coordy, coordz = coordz, coordt = coordt,
      coordx_dyn = coordx_dyn, spacetime = spacetime, bivariate = bivariate,
      grid = grid, context = "GeoFit", check.duplicates = check.duplicates
    )
    .dup_old <- .GeoSuppressDuplicateChecksForNestedCalls()
    on.exit(.GeoRestoreDuplicateCheckOption(.dup_old), add = TRUE)

    ##############################################################################
    ###### Normalize and validate the univariate mean specification
    ##
    ## A vector fixed$mean is an externally supplied mean at the observations.
    ## It must be removed before CkInput/WlsStart, otherwise unlist() turns it
    ## into apparent coefficients mean1, mean2, ... .  A scalar fixed$mean is
    ## instead the intercept coefficient beta[1].
    MM <- NULL
    if (!bivariate) {
      nobs_mean <- length(unlist(data, use.names = FALSE))
      mean_check <- .GeoMean_validate_fit(start, fixed, X, nobs_mean)
      X <- mean_check$X
      MM <- mean_check$external
      if (!is.null(MM)) fixed$mean <- 0
    }

    ## Validate user-supplied parameter names before internal defaults (such as
    ## nugget=0 or the fixed unit sill used by constructive count/positive
    ## models) are added.  Exact matching is intentional: partial names such as
    ## `shap` must not silently match `shape`.
    .GeoFit_validate_parameter_names(
      start = start, fixed = fixed, model = model, corrmodel = corrmodel,
      copula = copula, bivariate = bivariate, X = X
    )

    ##############################################################################
    ###### setting nugget if missing
    if (!bivariate) {
      if (!sum(substr(names(unlist(append(start, fixed))), 1, 6) == "nugget"))
        fixed$nugget <- 0
    }

    ###############################################################
    ###############################################################
    fixed <- .GeoFit_normalize_unit_sill(start, fixed, model, bivariate)

    ##### all parameters are estimated
    allest <- FALSE
    if (!bivariate) {
      if ((length(c(CorrParam(corrmodel), NuisParam2(model, bivariate, 2, copula = copula))) == length(start)) &&
          is.null(fixed)) {
        fixed <- list(nugget = 0)
        tempstart <- start
        start$nugget <- NULL
        allest <- TRUE
      }
    }

    #############################################################################
    checkinput <- CkInput(coordx, coordy, coordz, coordt, coordx_dyn, corrmodel, data,
                          distance, "Fitting", fixed, grid, likelihood, maxdist, maxtime,
                          model, n, optimizer, NULL, radius, start, taper, tapsep, type,
                          varest, weighted, copula, X)

    if (!is.null(checkinput$error))
      stop(checkinput$error)

    ### Initialization global variables:
    GeoFit <- NULL
    sensmat <- varcov <- varimat <- parscale <- NULL

    ### Initialization parameters:
    coordt <- unname(coordt)
    if (is.null(coordx_dyn)) {
      coordx <- unname(coordx)
      coordy <- unname(coordy)
      coordz <- unname(coordz)
    }

    initparam <- WlsStart(coordx, coordy, coordz, coordt, coordx_dyn, corrmodel, data,
                          distance, "Fitting", fixed, grid, likelihood, maxdist, neighb,
                          maxtime, model, n, NULL, parscale, optimizer == "L-BFGS-B",
                          radius, start, taper, tapsep, type, varest, weighted, copula,
                          X, memdist, nosym, p_neighb, thin_method)

    if (identical(type, "Pairwise") &&
        (!is.numeric(initparam$numpairs) || length(initparam$numpairs) != 1L ||
         !is.finite(initparam$numpairs) || initparam$numpairs < 1L)) {
      stop("No pairwise contributions were retained. Increase 'p_neighb', 'neighb', 'maxdist', or 'maxtime'.",
           call. = FALSE)
    }

    ### fixing initparam if all parameters are estimated
    if (allest) {
      if (!bivariate) {
        bb <- initparam$num_betas
        ccpp <- CorrParam(corrmodel)
        aabb <- NuisParam2(model, bivariate, bb, copula = copula)
        initparam$namescorr <- ccpp
        initparam$namesnuis <- aabb
        initparam$param <- unlist(tempstart)
        initparam$namesparam <- names(initparam$param)
        initparam$fixed <- NULL
        oo <- double(length(ccpp)); names(oo) <- ccpp; oo[names(oo) != "kl"] <- 1
        initparam$flagcorr <- oo
        oo <- double(length(aabb)); names(oo) <- aabb; oo[names(oo) != "kl"] <- 1
        initparam$flagnuis <- oo
      }
    }

    if (type == "Independence") {
      marginal_names <- NuisParam(
        model, bivariate = initparam$bivariate,
        num_betas = initparam$num_betas - 1
      )
      if (!any(marginal_names %in% names(initparam$param)))
        stop("No marginal parameters to estimate")
    }

    ## moving sill from starting to fixed parameters if necessary (in some model sill must be 1)
    if (sum(initparam$namesparam == "sill") == 1) {
      if (initparam$model %in% c(2,14,16,21,28,33,42,50,26,24,30,46,43,11,54)) {
        initparam$param <- initparam$param[initparam$namesparam != "sill"]
        initparam$namesparam <- names(initparam$param)
        a <- 1; names(a) <- "sill"
        initparam$fixed <- c(initparam$fixed, a)
      }
    }

    if (!is.null(initparam$error)) stop(initparam$error)

    ## After WlsStart() has initialized any missing regression coefficients,
    ## enforce the canonical mapping beta = (mean, mean1, ...) to the columns
    ## of X.  This check is deliberately post-initialization so callers may
    ## omit starting values and let WlsStart() provide them.
    if (!bivariate && !is.null(X) && is.null(MM)) {
      p_mean <- ncol(X)
      if (length(initparam$num_betas) != 1L ||
          is.na(initparam$num_betas) || initparam$num_betas != p_mean) {
        stop(
          "The number of mean coefficients initialized by WlsStart() does not match ncol(X).",
          call. = FALSE
        )
      }
      expected_mean <- .GeoMean_names(p_mean)
      available_mean <- unique(c(names(initparam$param), names(initparam$fixed)))
      missing_mean <- setdiff(expected_mean, available_mean)
      unexpected_mean <- setdiff(
        .GeoMean_supplied_names(c(initparam$param, initparam$fixed)),
        expected_mean
      )
      if (length(missing_mean) || length(unexpected_mean)) {
        msg <- paste0(
          "The columns of X must correspond, in order, to coefficients ",
          paste(expected_mean, collapse = ", "), "."
        )
        if (length(missing_mean))
          msg <- paste0(msg, " Missing: ", paste(missing_mean, collapse = ", "), ".")
        if (length(unexpected_mean))
          msg <- paste0(msg, " Unexpected: ", paste(unexpected_mean, collapse = ", "), ".")
        stop(msg, call. = FALSE)
      }
    }

    if (bivariate && !is.null(X)) {
      p_mean <- ncol(X)
      if (length(initparam$num_betas) != 2L ||
          any(is.na(initparam$num_betas)) ||
          any(initparam$num_betas != c(p_mean, p_mean))) {
        stop(
          "The number of bivariate mean coefficients initialized by WlsStart() does not match ncol(X).",
          call. = FALSE
        )
      }
      expected_mean <- c(
        .GeoMean_names(p_mean, "mean_1"),
        .GeoMean_names(p_mean, "mean_2")
      )
      available_mean <- unique(c(names(initparam$param), names(initparam$fixed)))
      supplied_mean <- unique(c(
        .GeoMean_supplied_names(c(initparam$param, initparam$fixed), "mean_1"),
        .GeoMean_supplied_names(c(initparam$param, initparam$fixed), "mean_2")
      ))
      missing_mean <- setdiff(expected_mean, available_mean)
      unexpected_mean <- setdiff(supplied_mean, expected_mean)
      if (length(missing_mean) || length(unexpected_mean)) {
        msg <- paste0(
          "The columns of X must correspond, in order, to bivariate coefficients ",
          paste(expected_mean, collapse = ", "), "."
        )
        if (length(missing_mean))
          msg <- paste0(msg, " Missing: ", paste(missing_mean, collapse = ", "), ".")
        if (length(unexpected_mean))
          msg <- paste0(msg, " Unexpected: ", paste(unexpected_mean, collapse = ", "), ".")
        stop(msg, call. = FALSE)
      }
    }

    ## Defensive fallback: some StartParam/WlsStart paths may not carry this
    ## flag into initparam.  Downstream methods such as GeoCovariogram()
    ## expect every GeoFit object to contain a non-NULL logical bivariate field.
    if (is.null(initparam$bivariate)) initparam$bivariate <- bivariate

    if (!(optimizer %in% c("L-BFGS-B","nlminb","nlm","nmkb","nmk","multiNelder-Mead",
                           "multinlminb","BFGS","Nelder-Mead","optimize","SANN",
                           "bobyqa","sbplx")))
      stop("optimizer is not correct\n")

    if ((optimizer %in% c("L-BFGS-B","nlminb","nmkb","multinlminb","bobyqa","sbplx")) &&
        is.null(lower) && is.null(upper))
      stop("lower and upper bound are missing\n")

    ######################## handling lower and upper bound parameters####################################
    if (optimizer %in% c("L-BFGS-B","nlminb","nmkb","multinlminb","multiNelder-Mead","bobyqa","sbplx") ||
        length(initparam$param) == 1) {

      if (!is.null(lower) || !is.null(upper)) {
        if (!is.list(lower) || !is.list(upper)) stop("lower and upper bound must be a list\n")
        if (sum(unlist(lower) > unlist(upper)) > 0)
          stop("some values of the lower bound is greater of the upper bound\n")

        if (sum(names(lower) == "sill") == 1) {
          if (initparam$model %in% c(2,14,16,21,42,50,26,24,30,46,43,11,54)) {
            lower <- lower[names(lower) != "sill"]
            upper <- upper[names(upper) != "sill"]
          }
        }

        ## setting alphabetic order
        lower <- lower[order(names(lower))]
        upper <- upper[order(names(upper))]

        npar <- length(initparam$param)
        ll <- as.numeric(unlist(lower, use.names = TRUE))
        uu <- as.numeric(unlist(upper, use.names = TRUE))
        names(ll) <- names(lower)
        names(uu) <- names(upper)

        if (length(ll) != npar || length(uu) != npar)
          stop("lower and upper bound must be of the same length of starting values\n")

        if (sum(sort(names(initparam$param)) == sort(names(upper))) < npar ||
            sum(sort(names(initparam$param)) == sort(names(lower))) < npar)
          stop("the names of parameters in the lower and/or upper bounds do not match with starting parameters names.\n")

        ll[ll == 0] <- .Machine$double.eps
        uu[uu == Inf] <- 1e+12
        initparam$upper <- uu
        initparam$lower <- ll
      }
    }

    ## When only one parameter is estimated, the composite-likelihood code
    ## switches internally to stats::optimize().  optimize() requires finite
    ## scalar bounds; automatic bounds from WlsStart() can be 0 and Inf
    ## for parameters such as scale.  Sanitize only the internal bounds here.
    if (!onlyvar && length(initparam$param) == 1) {
      pname <- names(initparam$param)

      if (is.null(initparam$lower) || is.null(initparam$upper))
        stop("lower and upper bounds are required when only one parameter is estimated\n")

      lb <- initparam$lower
      ub <- initparam$upper

      if (!is.null(names(lb)) && pname %in% names(lb)) lb <- lb[pname] else lb <- lb[1]
      if (!is.null(names(ub)) && pname %in% names(ub)) ub <- ub[pname] else ub <- ub[1]

      lb <- as.numeric(lb)
      ub <- as.numeric(ub)

      if (length(lb) != 1 || length(ub) != 1 || is.na(lb) || is.na(ub))
        stop("invalid lower/upper bounds for optimize()\n")

      p0 <- as.numeric(initparam$param[1L])
      if (!is.finite(p0))
        stop("invalid starting value for one-parameter optimization\n")

      if (lb == 0) lb <- .Machine$double.eps

      ## Positive parameters keep the broad automatic upper bound; the
      ## one-dimensional composite-likelihood optimizer now works on the log
      ## scale, so [eps, 1e12] no longer destroys resolution near the start.
      ## For genuinely unbounded parameters (for example a single mean), use
      ## a finite start-centred interval because stats::optimize() requires it.
      if (!is.finite(lb) && !is.finite(ub)) {
        span <- 20 * (1 + abs(p0))
        lb <- p0 - span
        ub <- p0 + span
      } else {
        if (!is.finite(lb)) lb <- p0 - 20 * (1 + abs(p0))
        if (!is.finite(ub)) ub <- 1e+12
      }

      if (!is.finite(lb) || !is.finite(ub) || lb >= ub)
        stop("invalid lower/upper bounds for optimize()\n")

      names(lb) <- pname
      names(ub) <- pname
      initparam$lower <- lb
      initparam$upper <- ub
    }
    ############################################################

    ## MM was extracted before CkInput/WlsStart.  Keep the internal scalar
    ## placeholder fixed at zero; all likelihood paths receive MM explicitly.
    if (!is.null(MM)) initparam$mean <- 0

    ###################################################################################
    ####################################################################
    #############  handling anisotropy parameters ######################
    ####################################################################
    update.aniso <- function(param, namesparam, fixed, namesfixed, lower, upper,
                             anisopars, estimate_aniso)
    {
      un_anisopars <- unlist(anisopars)
      namesaniso <- names(un_anisopars)

      anisostart <- unlist(anisopars)[estimate_aniso]
      anisofixed <- unlist(anisopars)[!estimate_aniso]

      if (length(anisostart) == 0) anisostart <- NULL
      if (length(anisofixed) == 0) anisofixed <- NULL

      ll <- c(0, 1)
      uu <- c(pi, 1e+25)
      lwr <- c(lower, ll[estimate_aniso])
      upr <- c(upper, uu[estimate_aniso])

      param <- c(param, anisostart)
      fixed <- c(fixed, anisofixed)
      namesparam <- names(param)
      namesfixed <- names(fixed)

      if (sum(!is.na(fixed[namesaniso]))) { # updating fixed values
        if (!estimate_aniso[2] && estimate_aniso[1]) fixed["ratio"] <- un_anisopars["ratio"]
        if (!estimate_aniso[1] && estimate_aniso[2]) fixed["angle"] <- un_anisopars["angle"]
        if (!estimate_aniso[1] && !estimate_aniso[2]) {
          fixed["angle"] <- un_anisopars["angle"]
          fixed["ratio"] <- un_anisopars["ratio"]
        }
      }

      list(param = param, namesparam = namesparam,
           fixed = fixed, namesfixed = namesfixed,
           lower = lwr, upper = upr)
    }

    aniso <- FALSE
    if (!is.null(anisopars)) {
      aniso <- TRUE
      namesaniso <- c("angle","ratio")
      qq <- update.aniso(initparam$param, initparam$namesparam,
                         initparam$fixed, initparam$namesfixed,
                         initparam$lower, initparam$upper,
                         anisopars, est.aniso)
      initparam$param <- qq$param
      initparam$fixed <- qq$fixed
      initparam$namesparam <- qq$namesparam
      initparam$namesfixed <- qq$namesfixed
      initparam$lower <- qq$lower
      initparam$upper <- qq$upper
    }

    ###################################################################################
    ###################################################################################

    ## Full likelihood:
    if (likelihood == "Full")
      fitted <- Lik(copula, initparam$bivariate, initparam$coordx, initparam$coordy,
                    initparam$coordz, initparam$coordt, coordx_dyn, initparam$corrmodel,
                    unname(initparam$data), initparam$fixed, initparam$flagcorr,
                    initparam$flagnuis, grid, initparam$lower, method, initparam$model,
                    initparam$namescorr, initparam$namesnuis, initparam$namesparam,
                    initparam$numcoord, initparam$numpairs, initparam$numparamcorr,
                    initparam$numtime, optimizer, onlyvar, initparam$param, initparam$radius,
                    initparam$setup, initparam$spacetime, sparse, varest, taper, initparam$type,
                    initparam$upper, initparam$ns, unname(initparam$X), initparam$neighb,
                    MM, aniso, score)

    ## Composite likelihood.  The selected pair structure is always
    ## precomputed and reused during optimization, Hessian and score
    ## evaluation.  The legacy CompLik() path is intentionally retired.
    if ((likelihood %in% c("Marginal","Conditional","Marginal_2")) && type == "Pairwise") {
      fitted <- CompLik2(copula, initparam$bivariate, initparam$coordx, initparam$coordy,
                         initparam$coordz, initparam$coordt, coordx_dyn, initparam$corrmodel,
                         unname(initparam$data), initparam$distance, initparam$flagcorr,
                         initparam$flagnuis, initparam$fixed, grid, initparam$likelihood,
                         initparam$lower, initparam$model, initparam$n, initparam$namescorr,
                         initparam$namesnuis, initparam$namesparam, initparam$numparam,
                         initparam$numparamcorr, optimizer, onlyvar, initparam$param,
                         initparam$spacetime, initparam$type, initparam$upper, varest,
                         initparam$weighted, initparam$ns, unname(initparam$X), sensitivity,
                         initparam$colidx, initparam$rowidx, initparam$neighb, MM, aniso, score)
    }

    if (likelihood == "Marginal" && type == "Independence")
      fitted <- CompIndLik2(initparam$bivariate, initparam$coordx, initparam$coordy, initparam$coordz,
                            initparam$coordt, coordx_dyn, unname(initparam$data),
                            initparam$flagcorr, initparam$flagnuis, initparam$fixed, grid,
                            initparam$lower, initparam$model, initparam$n, initparam$namescorr,
                            initparam$namesnuis, initparam$namesparam, initparam$numparam,
                            optimizer, onlyvar, initparam$param, initparam$spacetime, initparam$type,
                            initparam$upper, names(upper), varest, initparam$ns, unname(initparam$X),
                            sensitivity, copula, MM, score)

    ## misspecified models
    missp <- FALSE
    if (model == "Gaussian_misp_Tukeygh")      { model <- "Tukeygh";      missp <- TRUE }
    if (model == "Gaussian_misp_Poisson")      { model <- "Poisson";      missp <- TRUE }
    if (model == "Gaussian_misp_Binomial")     { model <- "Binomial";     missp <- TRUE }
    if (model == "Gaussian_misp_BinomialNeg")  { model <- "BinomialNeg";  missp <- TRUE }
    if (model == "Gaussian_misp_PoissonGamma") { model <- "PoissonGamma"; missp <- TRUE }
    if (model == "Gaussian_misp_PoissonZIP")   { model <- "PoissonZIP";   missp <- TRUE }
    if (model == "Gaussian_misp_StudentT")     { model <- "StudentT";     missp <- TRUE }
    if (model == "Gaussian_misp_SkewStudentT") { model <- "SkewStudentT"; missp <- TRUE }
    if (model == "Binary_misp_BinomialNeg")    { model <- "BinomialNeg";  missp <- TRUE }

    numtime <- 1
    if (initparam$spacetime) numtime <- length(coordt)
    if (initparam$bivariate) numtime <- 2
    dimat <- initparam$numcoord

    keep_fit_context <- is.environment(
      getOption("GeoModels.fit_context_cache", NULL)
    )
    if (!(likelihood == "Marginal" && type == "Independence") &&
        !keep_fit_context) {
      .C("DeleteGlobalVar2", PACKAGE = "GeoModels", NAOK = TRUE)
    }

    ff <- as.list(initparam$fixed)
    if (!is.null(MM)) ff$mean <- MM

    if (length(initparam$param) == 1) optimizer <- "optimize"

    if (aniso) anisopars <- as.list(c(fitted$par, ff)[namesaniso])

    if (is.null(unlist(ff))) ff <- NULL

    ## special case: maxdist and neighb = NULL
   # if (likelihood != "Full") {
   #   if (is.null(neighb) && is.numeric(maxdist) && likelihood == "Marginal") {
   #     fitted$value <- 2 * fitted$value
   #     initparam$numpairs <- 2 * initparam$numpairs
   #   }
   # }

    if (!is.null(coordt) && is.null(coordx_dyn)) {
      if (is.null(coordz)) {
        initparam$coordx <- initparam$coordx[1:(length(initparam$coordx) / length(initparam$coordt))]
        initparam$coordy <- initparam$coordy[1:(length(initparam$coordy) / length(initparam$coordt))]
      } else {
        initparam$coordx <- initparam$coordx[1:(length(initparam$coordx) / length(initparam$coordt))]
        initparam$coordy <- initparam$coordy[1:(length(initparam$coordy) / length(initparam$coordt))]
        initparam$coordz <- initparam$coordz[1:(length(initparam$coordz) / length(initparam$coordt))]
      }
    }

    if (model %in% c("Weibull","Poisson","Binomial","Gamma","PoissonGamma","Gaussian_misp_PoissonGamma",
                     "LogLogistic","BinomialNeg","BinomNeg","Binary","Bernoulli","Geom","Geometric","Binary_misp_BinomialNeg",
                     "Gaussian_misp_Poisson","PoissonZIP","PoissonGammaZIP","PoissonGammaZIP1",
                     "Gaussian_misp_PoissonZIP","BinomialNegZINB","PoissonZIP1","Gaussian_misp_PoissonZIP1",
                     "BinomialNegZINB1","Beta2","Kumaraswamy2","Beta","Kumaraswamy")) {
      if (!is.null(ff$sill)) ff$sill <- NULL
    }

    conf.int <- NULL
    pvalues <- NULL

    if (!is.null(initparam$coordz) && all(initparam$coordz == 0)) initparam$coordz <- NULL

    if (likelihood == "Full" && type == "Standard") {
      if (varest) {
        alpha <- 0.95
        conf.int <- pvalues <- NULL
        if (is.numeric(fitted$stderr)) {
          aa <- qnorm(1 - (1 - alpha) / 2) * fitted$stderr
          pp <- as.numeric(fitted$par)
          low <- pp - aa
          upp <- pp + aa
          conf.int <- rbind(low, upp)
          pvalues <- 2 * pnorm(-abs(pp / fitted$stderr))
        }
      }
    }

    if (bivariate && is.null(coordx_dyn)) {
      initparam$coordx <- initparam$coordx[1:dimat]
      initparam$coordy <- initparam$coordy[1:dimat]
    }

    ### Set the output object:
    GeoFit <- list(
      anisopars = anisopars,
      bivariate = if (is.null(initparam$bivariate)) bivariate else initparam$bivariate,
      claic = fitted$claic,
      clbic = fitted$clbic,
      colidx = as.integer(initparam$colidx),
      coordx = initparam$coordx,
      coordy = initparam$coordy,
      coordz = initparam$coordz,
      coordt = initparam$coordt,
      coordx_dyn = coordx_dyn,
      duplicate_locations_checked = isTRUE(check.duplicates),
      conf.int = conf.int,
      convergence = fitted$convergence,
      copula = copula,
      corrmodel = corrmodel,
      data = initparam$data,
      distance = distance,
      est.aniso = est.aniso,
      fixed = ff,
      grid = grid_requested,
      iterations = fitted$counts,
      likelihood = likelihood,
      logCompLik = fitted$value,
      lower = lower,
      message = fitted$message,
      model = model,
      estimation_model = estimation_model,
      n = initparam$n,
      ns = initparam$ns,
      numbetas = initparam$num_betas,
      numcoord = initparam$numcoord,
      numtime = initparam$numtime,
      optimizer = optimizer,
      param = as.list(fitted$par),
      p_neighb = p_neighb,
      thin_method = thin_method,
      memdist = TRUE,
      nozero = initparam$setup$nozero,
      score = fitted$score,
      maxdist = maxdist,
      maxtime = maxtime,
      neighb = initparam$neighb,
      numpairs = initparam$numpairs,
      missp = missp,
      pvalues = pvalues,
      radius = radius,
      rowidx = as.integer(initparam$rowidx),
      spacetime = initparam$spacetime,
      stderr = fitted$stderr,
      sensmat = fitted$sensmat,
      upper = upper,
      varcov = fitted$varcov,
      varimat = fitted$varimat,
      type = type,
      weighted = initparam$weighted,
      X = X
    )

    structure(c(GeoFit, call = call), class = c("GeoFit"))
  })
}

################################################################################

print.GeoFit <- function(x, digits = max(3, getOption("digits") - 3), ...)
{
  if (x$likelihood == "Full") {
    method <- "Likelihood"
    if (x$type == "Tapering") { claic <- "CLAIC"; clbic <- "CLBIC" }
    else { claic <- "AIC"; clbic <- "BIC" }
  } else {
    method <- "Composite-Likelihood"
    claic <- "CLAIC"
    clbic <- "CLBIC"
  }

  missp <- ""
  if (isTRUE(x$missp)) missp <- "misspecified"

  if (x$model %in% c("Gaussian","Gauss")) { process <- "Gaussian"; model <- "Gaussian" }
  if (x$model == "Gamma") { process <- "Gamma"; model <- "Gamma" }
  if (x$model == "TwoPieceBimodal") { process <- "TwoPieceBimodal"; model <- "TwoPieceBimodal" }
  if (x$model == "LogLogistic") { process <- "LogLogistic"; model <- "LogLogistic" }
  if (x$model == "Gaussian_misp_Poisson") { process <- "Poisson"; model <- "Misspecified Gaussian Poisson" }
  if (x$model == "Gaussian_misp_Binomial") { process <- "Binomial"; model <- "Misspecified Gaussian Binomial" }
  if (x$model == "Gaussian_misp_PoissonZIP") { process <- "PoissonZIP"; model <- "Misspecified Gaussian Poisson Inflated" }
  if (x$model == "Poisson") { process <- "Poisson"; model <- "Poisson" }
  if (x$model == "PoissonGamma") { process <- "PoissonGamma"; model <- "PoissonGamma" }
  if (x$model == "Gaussian_misp_PoissonGamma") { process <- "PoissonGamma"; model <- "Misspecified Gaussian PoissonGamma" }
  if (x$model == "PoissonZIP") { process <- "PoissonZIP"; model <- "PoissonZIP" }
  if (x$model == "PoissonGammaZIP") { process <- "PoissonGammaZIP"; model <- "PoissonGammaZIP" }
  if (x$model == "Beta2") { process <- "Beta2"; model <- "Beta2" }
  if (x$model == "Gaussian_misp_StudentT") { process <- "StudentT"; model <- "Misspecified Gaussian StudentT" }
  if (x$model == "StudentT") { process <- "StudentT"; model <- "StudentT" }
  if (x$model == "Gaussian_misp_Tukeygh") { process <- "Tukeygh"; model <- "Misspecified Gaussian Tukeygh" }
  if (x$model == "Tukeygh") { process <- "Tukeygh"; model <- "Tukeygh" }
  if (x$model == "Gaussian_misp_SkewStudentT") { process <- "SkewStudentT"; model <- "Misspecified Gaussian SkewStudentT" }
  if (x$model == "SkewStudentT") { process <- "SkewStudentT"; model <- "SkewStudentT" }
  if (x$model == "Logistic") { process <- "Logistic"; model <- "Logistic" }
  if (x$model == "Tukeyh") { process <- "Tukeyh"; model <- "Tukeyh" }
  if (x$model == "Tukeyh2") { process <- "Tukeyh2"; model <- "Tukeyh2" }
  if (x$model == "Gamma2") { process <- "Gamma2"; model <- "Gamma2" }
  if (x$model %in% c("LogGauss","LogGaussian")) { process <- "Log Gaussian"; model <- "LogGaussian" }
  if (x$model %in% c("SkewGauss","SkewGaussian")) { process <- "Skew Gaussian"; model <- "SkewGaussian" }
  if (x$model == "SkewLaplace") { process <- "Skew Laplace"; model <- "SkewLaplace" }
  if (x$model == "TwoPieceStudentT") { process <- "TwoPiece StudentT"; model <- "TwoPieceStudentT" }
  if (x$model == "TwoPieceTukeyh") { process <- "TwoPiece Tukeyh"; model <- "TwoPieceTukeyh" }
  if (x$model %in% c("TwoPieceGaussian","TwoPieceGauss")) { process <- "TwoPiece Gaussian"; model <- "TwoPieceGaussian" }
  if (x$model == "SinhAsinh") { process <- "SinhAsinh"; model <- "SinhAsinh" }
  if (x$model == "Wrapped") { process <- "Wrapped"; model <- "Wrapped" }
  if (x$model == "Weibull") { process <- "Weibull"; model <- "Weibull" }
  if (x$model == "Binomial") { process <- "Binomial"; model <- "Binomial" }
  if (x$model == "BinomialLogistic") { process <- "BinomialLogistic"; model <- "BinomialLogistic" }
  if (x$model == "Kumaraswamy") { process <- "Kumaraswamy"; model <- "Kumaraswamy" }
  if (x$model == "Kumaraswamy2") { process <- "Kumaraswamy2"; model <- "Kumaraswamy2" }
  if (x$model == "Beta") { process <- "Beta"; model <- "Beta" }
  if (x$model %in% c("Binomial_TwoPieceGaussian","Binomial_TwoPieceGauss")) { process <- "Binomial TwoPiece Gaussian"; model <- "Binomial_TwoPieceGauss" }
  if (x$model %in% c("BinomialNeg_TwoPieceGaussian","BinomialNeg_TwoPieceGauss")) { process <- "Negative Binomial TwoPiece Gaussian"; model <- "BinomialNeg_TwoPieceGauss" }
  if (x$model == "Binomial2") { process <- "Binomial"; model <- "Binomial2" }
  if (x$model == "BinomialNeg") { process <- "BinomialNeg"; model <- "BinomialNeg" }
  if (x$model == "Binary_misp_BinomialNeg") { process <- "BinomialNeg"; model <- "Misspecified Binary BinomialNeg" }
  if (x$model == "BinomialNegLogistic") { process <- "BinomialNegLogistic"; model <- "BinomialNegLogistic" }
  if (x$model == "BinomialNegZINB") { process <- "BinomialNegZINB"; model <- "BinomialNegZINB" }
  if (x$model %in% c("Geom","Geometric")) { process <- "Geometric"; model <- "Geometric" }
  if (x$model == "PoisBin") { process <- "Poisson Binomial"; model <- "PoisBin" }
  if (x$model == "PoisBinNeg") { process <- "Poisson NegBinomial"; model <- "PoisBinNeg" }

  if (x$bivariate) { biv <- "bivariate"; x$numtime <- 1 }
  else { biv <- "univariate" }

  cat("\n##################################################################")
  cat("\nMaximum", missp, method, "Fitting of", process, "Random Fields\n")
  if (!is.null(x$copula)) cat("\nCopula:", x$copula, "\n")
  cat("\nSetting:", x$likelihood, method, "\n")
  cat("\nModel:", model, "\n")
  cat("\nDistance:", x$distance, "\n")
  cat("\nType of the likelihood objects:", x$type, x$method, "\n")

  ## Pair-thinning information, when available.  For backward
  ## compatibility with older GeoFit objects, retrieve thin_method from
  ## the object if present, otherwise from the stored call; if it is not
  ## available but thinning was used, report the default.
  p_retained <- NULL
  if (!is.null(x$p_neighb)) {
    p_retained <- suppressWarnings(as.numeric(x$p_neighb)[1])
  } else if (!is.null(x$call) && !is.null(as.list(x$call)$p_neighb)) {
    p_retained <- suppressWarnings(as.numeric(as.character(as.list(x$call)$p_neighb))[1])
  }

  thin_method <- NULL
  if (!is.null(x$thin_method)) {
    thin_method <- as.character(x$thin_method)[1]
  } else if (!is.null(x$call) && !is.null(as.list(x$call)$thin_method)) {
    thin_method <- paste(deparse(as.list(x$call)$thin_method), collapse = "")
    thin_method <- gsub('^"|"$', "", thin_method)
    thin_method <- gsub("^'|'$", "", thin_method)
  }

  if (!is.null(p_retained) && is.finite(p_retained) && p_retained < 1) {
    if (is.null(thin_method) || !nzchar(thin_method)) thin_method <- "bernoulli"
    cat("Pair thinning method:", thin_method, "\n")
    cat("(Expected) Retained pairs (%):",
        format(100 * p_retained, digits = digits, nsmall = 2), "\n")
  }

  cat("\nCovariance model:", x$corrmodel, "\n")
  cat("\nOptimizer:", x$optimizer, "\n")
  cat("\nNumber of spatial coordinates:", x$numcoord, "\n")
  if (x$spacetime) cat("Number of dependent temporal realisations:", x$numtime, "\n")
  cat("Type of the random field:", biv, "\n")
  cat("Number of estimated parameters:", length(x$param), "\n")
  cat("\nType of convergence:", x$convergence, "")
  cat("\nMaximum log-", method, " value: ",
      format(x$logCompLik, digits = digits, nsmall = 2), "\n", sep = "")

  if (!is.null(x$claic))
    cat(claic, ":", format(x$claic, digits = digits), "\n")

  if (!is.null(x$clbic))
    cat(clbic, ":", format(x$clbic, digits = digits), "\n")

  cat("\nEstimated parameters:\n")
  print.default(unlist(x$param), digits = digits, print.gap = 2, quote = FALSE)

  if (!is.null(x$stderr)) {
    cat("\nStandard errors:\n")
    print.default(x$stderr, digits = digits, print.gap = 2, quote = FALSE)
  }

  cat("\n##################################################################\n")
  invisible(x)
}


summary.GeoFit <- function(object, digits = max(3, getOption("digits") - 3), ...) {
  print.GeoFit(object, digits = digits, ...)
}

GeoAniso <- function(coords,
                     anisopars = c(0, 1),
                     inverse = FALSE) {

  tol = sqrt(.Machine$double.eps)
  if (!is.matrix(coords) || !is.numeric(coords)) {
    stop("coords must be a numeric matrix.")
  }

  ncc <- ncol(coords)

  if (!(ncc %in% c(2L, 3L))) {
    stop("coords must have two or three columns.")
  }

  if (any(!is.finite(coords))) {
    stop("coords contains non-finite values.")
  }

  if (!is.numeric(anisopars) ||
      length(anisopars) != 2L ||
      any(!is.finite(anisopars))) {
    stop(
      "anisopars must contain two finite numeric values: ",
      "angle and anisotropy ratio."
    )
  }

  if (!is.logical(inverse) ||
      length(inverse) != 1L ||
      is.na(inverse)) {
    stop("inverse must be TRUE or FALSE.")
  }

  if (!is.numeric(tol) ||
      length(tol) != 1L ||
      !is.finite(tol) ||
      tol < 0) {
    stop("tol must be a finite non-negative scalar.")
  }

  angle <- anisopars[1L]
  stretch <- anisopars[2L]

  if (angle < 0 || angle > pi) {
    stop("The anisotropy angle must be between 0 and pi radians.")
  }

  if (stretch < 1 - tol) {
    stop("The anisotropy ratio must be greater than or equal to 1.")
  }

  # Accept harmless floating-point deviations slightly below one.
  stretch <- max(stretch, 1)

  rotation <- matrix(
    c(
      cos(angle), -sin(angle),
      sin(angle),  cos(angle)
    ),
    nrow = 2L,
    ncol = 2L
  )

  transform_2d <- function(x) {
    if (inverse) {
      # Inverse of rotation %*% diag(c(1, 1 / stretch)).
      x %*% diag(c(1, stretch)) %*% t(rotation)
    } else {
      x %*% rotation %*% diag(c(1, 1 / stretch))
    }
  }

  if (ncc == 2L) {
    coordstransf <- transform_2d(coords)
  } else {
    if (!all(abs(coords[, 3L]) <= tol)) {
      stop(
        "Anisotropy for genuine three-dimensional coordinates ",
        "is not implemented."
      )
    }

    coordstransf <- coords
    coordstransf[, 1:2] <- transform_2d(
      coords[, 1:2, drop = FALSE]
    )
  }

  dimnames(coordstransf) <- dimnames(coords)
  coordstransf
}

#! /bin/sh
# Run Inno Setup Compiler to create installers for geoid datasets

INNO="c:/Program Files/Inno Setup 5/ISCC.exe"
test -f "$INNO" || INNO="c:/Program Files (x86)/Inno Setup 5/ISCC.exe"

GEOIDDIR=/usr/local/share/GeographicLib
test -d "$GEOIDDIR"/installers || mkdir -p "$GEOIDDIR"/installers
GEOIDDIR=`cygpath -w $GEOIDDIR`
(
cat <<EOF
egm84-30    EGM84 30'
egm84-15    EGM84 15'
egm96-15    EGM96 15'
egm96-5     EGM96 5'
egm2008-5   EGM2008 5'
egm2008-2_5 EGM2008 2.5'
egm2008-1   EGM2008 1'
EOF
) | while read prefix name; do
    "$INNO" geoid-installers.iss \
        /dGEOIDDIR="$GEOIDDIR" \
        /dPREFIX="$prefix" \
        /dNAME="$name" > $prefix.log 2>&1
done
exit 0

#! /bin/sh
for m in egm84 egm96 egm2008 wgs84; do
  tar cfCjv ../../data-distrib/gravity-distrib/$m.tar.bz2 .. gravity/$m.egm{,.cof}
  (
      cd ..
      rm -f ../data-distrib/gravity-distrib/$m.zip
      zip ../data-distrib/gravity-distrib/$m.zip gravity/$m.egm{,.cof}
  )
done

#!/bin/sh
#
# A simple script to add/reset autotools support
# This requires that autoconf and autoconf-archive are installed

run ()
{
    echo "Running: $*"
    eval $*

    if test $? != 0 ; then
        echo "Error: while running '$*'"
        exit 1
    fi
}

mv m4/pkg.m4 ./
rm -rf autom4te.cache m4
mkdir m4
mv pkg.m4 m4/
run aclocal
run autoheader
LIBTOOLIZE=libtoolize
type $LIBTOOLIZE > /dev/null 2>&1 || LIBTOOLIZE=g$LIBTOOLIZE
run $LIBTOOLIZE --force --copy
run automake --add-missing --copy --force-missing --foreign
run autoconf
run autoreconf

rm -rf autom4te.cache configure~

#! /bin/sh
# Convert a pod file into a usage function for the GeographicLib utilities.

SOURCE=$1
NAME=`basename $SOURCE .pod`
VERSION=$2

(
cat<<EOF
int usage(int retval, bool brief) {
  if (brief)
    ( retval ? std::cerr : std::cout ) << "Usage:\n\\
EOF

$POD2MAN $SOURCE | nroff -c -man 2>/dev/null | $COL -b -x |
sed -e 1,/SYNOPSIS/d -e '/^$/,$d' -e 's/  / /g' -e 's/$/\\n\\/' -e 's/"/\\"/g'

cat <<EOF
\n\\
For full documentation type:\n\\
    $NAME --help\n\\
or visit:\n\\
    https://geographiclib.sourceforge.io/C++/$VERSION/$NAME.1.html\n";
  else
    ( retval ? std::cerr : std::cout ) << "Man page:\n\\
EOF

$POD2MAN $SOURCE | nroff -c -man 2>/dev/null | $COL -b -x | head --lines -4 |
tail --lines +5 | sed -e 's/\\/\\\\/g' -e 's/$/\\n\\/' -e 's/"/\\"/g'

cat <<EOF
";
  return retval;
}
EOF
) | # Break long strings assuming that files don't contain the ~ character
tr '\n' '~' | sed -e's/\\~/"~"/g' -e's/""//g' | tr '~' '\n'

#! /bin/sh
cd $HOME
for d in geographiclib \
             git/geographiclib-{c,fortran,java,js,octave,python,doc}; do
    (
        cd $d
        git ls-tree -r HEAD --name-only |
            sed -e "s%^%$HOME/$d/%"
    )
done

#! /bin/sh
#for m in emm2010 wmm2010 igrf11 wmm2015 wmm2015v2 wmm2020 igrf13 wmm2025 wmmhr2025 igrf14; do
for m in igrf14; do
  tar cfCjv ../../data-distrib/magnetic-distrib/$m.tar.bz2 .. magnetic/$m.wmm{,.cof}
  (
      cd ..
      rm -f ../data-distrib/magnetic-distrib/$m.zip
      zip ../data-distrib/magnetic-distrib/$m.zip magnetic/$m.wmm{,.cof}
  )
done

#! /bin/sh -ex

# This is the script to build the utilities for the cgi-bin scripts.  Note:

# 0: To access the machine for running this command, use
#   ssh -t karney,geographiclib@shell.sourceforge.net create
# Once logged in the following paths are available
# git: /home/git/p/geographiclib/code.git
WEB=/home/project-web/geographiclib
FRS=/home/frs/project/geographiclib

# 1: The utilities used long double (GEOGRAPHICLIB_PRECISION=3)

# 2: GeoidEval needs access to geoid data which is installed in
# $WEB/geoids.  It find this with
#   export GEOGRAPHICLIB_DATA=..

# 3: Static libraries are used so the utilities are self contained
# executables.

VERSION=2.0
FULLVERSION=$VERSION-alpha
rm -rf /tmp/GeographicLib-$VERSION /tmp/geog-$VERSION
tar xfpzC $FRS/distrib-C++/GeographicLib-$FULLVERSION.tar.gz /tmp
cd /tmp/GeographicLib-$VERSION
# N.B. $HOME/cmake/bin is in PATH for cmake
cmake \
    -D CMAKE_INSTALL_PREFIX=/tmp/geog-$VERSION \
    -D BUILD_SHARED_LIBS=OFF \
    -D GEOGRAPHICLIB_PRECISION=3 \
    -D EXAMPLEDIR= -B BUILD -S .
cd BUILD
make
make install
mkdir -p $WEB/bin-$VERSION
cd /tmp/geog-$VERSION/bin
install CartConvert ConicProj GeodesicProj GeoConvert GeodSolve GeoidEval Gravity MagneticField Planimeter RhumbSolve TransverseMercatorProj $WEB/bin-$VERSION/

#! /bin/sh
#
# tar.gz and zip distrib files copied to $DEVELSOURCE
# html documentation rsync'ed to $WEBDIST/htdocs/C++/$VERSION/
#
# Windows version ready to build in
# $WINDOWSBUILD/GeographicLib-$VERSION/BUILD-vc10{,-x64}
# after ./build installer is copied to
# $DEVELSOURCE/GeographicLib-$VERSION-win{32,64}.exe
#
# Built version ready to install in /usr/local in
# relc/GeographicLib-$VERSION/BUILD-system
#
# gita - check out from git, create package with cmake
# gitb - check out from git, check various builds in the non-release tree
# gitr - new release branch
# relb - release package, build with autoconf
# relc - release package, build with cmake
# relx - cmake release package inventory
# rely - autoconf release package inventory
# instb - installed files, autoconf
# instc - installed files, cmake

set -e
umask 0022

# The following files contain version information:
#   CMakeLists.txt (PROJECT_VERSION_* LIBVERSION_*)
#   NEWS
#   configure.ac (AC_INIT, GEOGRAPHICLIB_VERSION_* LT_*)
#   develop/test-distribution.sh
#   doc/GeographicLib.dox.in (3 places)

# maxima
#   maxima/geodesic.mac

START=`date +%s`
DATE=`date +%F`
VERSION=2.5
SUFFIX=
DISTVERSION=$VERSION$SUFFIX
BRANCH=main
TEMP=/home/scratch/geographiclib-dist
if test `hostname` = petrel; then
    DEVELSOURCE=$HOME/geographiclib
    WINDEVELSOURCE=//datalake-pr-smb/vt-open/ckarney/geographiclib
    WINDOWSBUILD=/var/tmp
else
    DEVELSOURCE=/u/geographiclib
    WINDEVELSOURCE=//datalake-pr-smb/vt-open/ckarney/geographiclib
    WINDOWSBUILD=/u/temp
fi
WINDOWSBUILDWIN=//datalake-pr-smb/vt-open/ckarney/temp
GITSOURCE=file://$DEVELSOURCE
WEBDIST=/home/ckarney/web/geographiclib-web
mkdir -p $WEBDIST/htdocs/C++
NUMCPUS=4
HAVEINTEL=

test -d $TEMP || mkdir $TEMP
rm -rf $TEMP/*
mkdir $TEMP/gita # Package creation via cmake
mkdir $TEMP/gitb # Non release testing
mkdir $TEMP/gitr # For release branch
(cd $TEMP/gitr; git clone -b $BRANCH $GITSOURCE)
(cd $TEMP/gita; git clone -b $BRANCH file://$TEMP/gitr/geographiclib)
(cd $TEMP/gitb; git clone -b $BRANCH file://$TEMP/gitr/geographiclib)

echo ==============================================================
echo Make a source package in $TEMP/gita/geographiclib/BUILD

cd $TEMP/gita/geographiclib
cmake -S . -B BUILD
(cd BUILD && make dist)
# cp $TEMP/gita/geographiclib/BUILD/distrib/GeographicLib-$DISTVERSION.{zip,tar.gz} $DEVELSOURCE/data-distrib/distrib-C++/

echo ==============================================================
echo Unpack source package in $TEMP/rel bcx

mkdir $TEMP/rel{b,c,x}
tar xfpzC BUILD/distrib/GeographicLib-$DISTVERSION.tar.gz $TEMP/relb # Version for autoconf
tar xfpzC BUILD/distrib/GeographicLib-$DISTVERSION.tar.gz $TEMP/relc # Version for cmake
tar xfpzC BUILD/distrib/GeographicLib-$DISTVERSION.tar.gz $TEMP/relx # for listing

echo ==============================================================
echo Unpack devel cmake distribution in $TEMP/relx and list in $TEMP/files.x
(
    cd $TEMP/relx
    find . -type f | sort -u > ../files.x
)

echo ==============================================================
echo Make a release for Windows testing in $WINDOWSBUILD/GeographicLib-$VERSION
rm -rf $WINDOWSBUILD/GeographicLib-$VERSION

unzip -qq -d $WINDOWSBUILD BUILD/distrib/GeographicLib-$DISTVERSION.zip

# for ver in 10 11 12 14 15 16; do
for ver in 14 15 16 17; do
    for arch in win32 x64; do
        pkg=vc$ver-$arch
        gen="Visual Studio $ver"
        installer=
        # N.B. update CPACK_NSIS_INSTALL_ROOT in CMakeLists.txt and
        # update documentation examples if VS version for binary
        # installer changes.
        test "$ver" = 14 && installer=y
        (
            echo "#! /bin/sh -exv"
            echo echo ========== cmake $pkg ==========
            echo b=c:/scratch/geog-$pkg
            echo rm -rf \$b //datalake-pr-smb/vt-open/ckarney/pkg-$pkg/GeographicLib-$VERSION/\*
            echo 'unset GEOGRAPHICLIB_DATA'
            echo cmake -G \"$gen\" -A $arch -D BUILD_BOTH_LIBS=ON -D CMAKE_INSTALL_PREFIX=//datalake-pr-smb/vt-open/ckarney/pkg-$pkg/GeographicLib-$VERSION -D PACKAGE_DEBUG_LIBS=ON -D CONVERT_WARNINGS_TO_ERRORS=ON -D EXAMPLEDIR= -S . -B \$b
            echo cmake --build \$b --config Debug   --target ALL_BUILD
            echo cmake --build \$b --config Debug   --target testprograms
            echo cmake --build \$b --config Debug   --target RUN_TESTS
            echo cmake --build \$b --config Debug   --target INSTALL
            echo cmake --build \$b --config Release --target ALL_BUILD
            echo cmake --build \$b --config Release --target exampleprograms
            echo cmake --build \$b --config Release --target experimental
            echo cmake --build \$b --config Release --target testprograms
            echo cmake --build \$b --config Release --target RUN_TESTS
            echo cmake --build \$b --config Release --target INSTALL
            echo cmake --build \$b --config Release --target PACKAGE
            test "$installer" &&
                echo cp \$b/GeographicLib-$DISTVERSION-*.exe $WINDEVELSOURCE/ ||
                    true
        ) > $WINDOWSBUILD/GeographicLib-$VERSION/build-$pkg
        chmod +x $WINDOWSBUILD/GeographicLib-$VERSION/build-$pkg
    done
done
cat > $WINDOWSBUILD/GeographicLib-$VERSION/test-all <<'EOF'
#! /bin/sh
(
    for d in build-*; do
        ./$d
    done
) >& build.log
EOF
chmod +x $WINDOWSBUILD/GeographicLib-$VERSION/test-all

echo ==============================================================
echo Set up release branch in $TEMP/gitr/geographiclib

cd $TEMP/gitr/geographiclib
git checkout release
git config user.email karney@alum.mit.edu
git ls-files | sort > ../files.old
(
    cd $TEMP/gitb/geographiclib
    git ls-files | sort
) > ../files.current
cut -f3- -d/ $TEMP/files.x | sort > ../files.new
(
    cd ..
    comm -23 files.old files.new | grep -v gitattributes > files.delete || true
    comm -23 files.new files.current > files.relonly
    comm -12 files.new files.current > files.main
    comm -13 files.delete files.new > files.add
    comm -13 files.old files.new > files.add
    comm -12 files.old files.new > files.common
)

(
    S=$TEMP/relx/GeographicLib-$VERSION
    C=$TEMP/gitb/geographiclib
    test -s ../files.delete && xargs git rm -f < ../files.delete
    cat ../files.main ../files.relonly |
        sed -e 's%[^/][^/]*$%%' -e 's%/$%%' | sort -u | grep -v '^$' |
        while read d; do
            test -d $d || mkdir -p $d
        done
    while read f; do
        if cmp $S/$f $f >& /dev/null; then :
        else
            cp -p $S/$f $f
            git add $f
        fi
    done < ../files.relonly
    while read f; do
        test -d `dirname $f` || mkdir `dirname $f`
        git checkout $BRANCH $f
        if cmp $S/$f $f >& /dev/null; then :
        else
            cp -p $S/$f $f
            git add $f
        fi
    done < ../files.main
    git checkout $BRANCH .gitattributes
    for i in 1 2 3 4 5; do
        find -type d -empty | xargs -r rmdir
    done
)
rm -rf GeographicLib-$VERSION
rm -f java/.gitignore
for ((i=0; i<7; ++i)); do
    find * -type d -empty | xargs -r rmdir
done

echo ==============================================================
echo CMake build in $TEMP/relc/GeographicLib-$VERSION/BUILD install to $TEMP/instc
cd $TEMP/relc/GeographicLib-$VERSION
cmake -D BUILD_BOTH_LIBS=ON -D BUILD_DOCUMENTATION=ON -D USE_BOOST_FOR_EXAMPLES=ON -D CONVERT_WARNINGS_TO_ERRORS=ON -D CMAKE_INSTALL_PREFIX=$TEMP/instc -S . -B BUILD
(
    cd BUILD
    make package_source
    make -j$NUMCPUS all testprograms
    make test
    make exampleprograms
    make -j$NUMCPUS experimental
    make install
    # rsync -a --delete doc/html/ $WEBDIST/htdocs/C++/$VERSION/
)

echo ==============================================================
echo Test building installed examples in $TEMP/instc
cd $TEMP/relc/GeographicLib-$VERSION
env GeographicLib_DIR=$TEMP/instc cmake -S $TEMP/instc/share/doc/GeographicLib-dev -B BUILD-examples-shared
(
    cd BUILD-examples-shared
    make -j$NUMCPUS
    ./example-Geodesic
)
env GeographicLib_DIR=$TEMP/instc cmake -S $TEMP/instc/share/doc/GeographicLib-dev -B BUILD-examples-static -D GeographicLib_USE_STATIC_LIBS=ON
(
    cd BUILD-examples-static
    make -j$NUMCPUS
    ./example-Geodesic
)

echo ==============================================================
echo List installed files fron cmake build in $TEMP/instc to $TEMP/files.c
(
    cd $TEMP/instc
    find . -type f | sort -u > ../files.c
)

echo ==============================================================
echo Make distribution from release tree with cmake

cd $TEMP/relc/GeographicLib-$VERSION
cmake -S . -B BUILD-dist
(cd BUILD-dist && make package_source)

echo ==============================================================
echo Unpack release cmake distribution in $TEMP/relz and list in $TEMP/files.z
mkdir $TEMP/relz
tar xfpzC BUILD-dist/GeographicLib-$VERSION.tar.gz $TEMP/relz

(
    cd $TEMP/relz
    find . -type f | sort -u > ../files.z
)

echo ==============================================================
echo CMake build in $TEMP/relc/GeographicLib-$VERSION/BUILD-system install to /usr/local

cmake -D BUILD_BOTH_LIBS=ON -D CONVERT_WARNINGS_TO_ERRORS=ON -S . -B BUILD-system
(cd BUILD-system && make -j$NUMCPUS all testprograms && make test)

if test "$HAVEINTEL"; then
    echo ==============================================================
    echo CMake build for intel in $TEMP/relc/GeographicLib-$VERSION/BUILD-intel
    env FC=ifort CC=icc CXX=icpc cmake -D BUILD_BOTH_LIBS=ON -D CONVERT_WARNINGS_TO_ERRORS=ON -S . -B BUILD-intel
    (
        cd BUILD-intel
        make -j$NUMCPUS all testprograms
        make test
        make exampleprograms
        make -j$NUMCPUS experimental
    )
fi

echo ==============================================================
echo Check build with configure in $TEMP/relb/GeographicLib-$VERSION/BUILD-config to $TEMP/instb

cd $TEMP/relb/GeographicLib-$VERSION
mkdir BUILD-config
cd BUILD-config
../configure --prefix=$TEMP/instb
make -j$NUMCPUS
make install
cd ..

if test "$HAVEINTEL"; then
echo ==============================================================
echo Check build with configure + intell in $TEMP/relb/GeographicLib-$VERSION/BUILD-config-intel

    mkdir BUILD-config-intel
    cd BUILD-config-intel
    env FC=ifort CC=icc CXX=icpc ../configure
    make -j$NUMCPUS
    cd ..
fi

echo ==============================================================
echo Make source dist with autoconf $TEMP/relb/GeographicLib-$VERSION/BUILD-dist

mkdir BUILD-dist
cd BUILD-dist
../configure
make dist-gzip

echo ==============================================================
echo Unpack release autoconf distribution in $TEMP/rely and list in $TEMP/files.y

mkdir $TEMP/rely
tar xfpzC geographiclib-$VERSION.tar.gz $TEMP/rely
mv $TEMP/rely/{geographiclib,GeographicLib}-$VERSION
cd $TEMP/rely
find . -type f | sort -u > ../files.y

echo ==============================================================
echo List installed files fron autoconf build in $TEMP/instb to $TEMP/files.b

mv $TEMP/instb/share/doc/{geographiclib,GeographicLib}
cd $TEMP/instb
find . -type f | sort -u > ../files.b

echo ==============================================================
echo CMake build of devel tree in $TEMP/gitb/geographiclib/BUILD

cd $TEMP/gitb/geographiclib
cmake -D BUILD_BOTH_LIBS=ON -D BUILD_DOCUMENTATION=ON -D USE_BOOST_FOR_EXAMPLES=ON -D CONVERT_WARNINGS_TO_ERRORS=ON -S . -B BUILD
(cd BUILD && make -j$NUMCPUS && make -j$NUMCPUS develprograms)

cp $DEVELSOURCE/include/mpreal.h include/
for p in 1 3 4 5; do
    echo ==============================================================
    echo CMake build of devel tree at precision $p in $TEMP/gitb/geographiclib/BUILD-$p
    mkdir BUILD-$p
    cmake -D USE_BOOST_FOR_EXAMPLES=ON -D GEOGRAPHICLIB_PRECISION=$p -S . -B BUILD-$p
    (
        cd BUILD-$p
        make -j$NUMCPUS all testprograms
        if test $p -ne 1; then
            make test
        fi
        make -j$NUMCPUS develprograms
        make exampleprograms
        make -j$NUMCPUS experimental
    )
done

echo ==============================================================
echo Compile and run little test program
cd $TEMP
cat > testprogram.cpp <<EOF
#include <iostream>
#include <iomanip>

#include <GeographicLib/Constants.hpp>
#include <GeographicLib/DMS.hpp>
#include <GeographicLib/LambertConformalConic.hpp>

int main() {
  using namespace GeographicLib;
  double
    // These are the constants for Pennsylvania South, EPSG:3364
    // https://www.spatialreference.org/ref/epsg/3364/
    a = Constants::WGS84_a(),   // major radius
    f = 1/298.257222101,        // inverse flattening (GRS80)
    lat1 = DMS::Decode(40,58),  // standard parallel 1
    lat2 = DMS::Decode(39,56),  // standard parallel 2
    k1 = 1,                     // scale on std parallels
    lat0 =  DMS::Decode(39,20), // latitude of origin
    lon0 = -DMS::Decode(77,45), // longitude of origin
    fe = 600000,                // false easting
    fn = 0;                     // false northing
  LambertConformalConic PASouth(a, f, lat1, lat2, k1);
  double x0, y0;
  PASouth.Forward(lon0, lat0, lon0, x0, y0); // Transform origin point
  x0 -= fe; y0 -= fn;           // Combine result with false origin

  double lat = 39.95, lon = -75.17;    // Philadelphia
  double x, y;
  PASouth.Forward(lon0, lat, lon, x, y);
  x -= x0; y -= y0;             // Philadelphia in PA South coordinates

  std::cout << std::fixed << std::setprecision(3)
            << x << " " << y << "\n";
  return 0;
}
EOF
for i in  b c; do
    cp testprogram.cpp testprogram$i.cpp
    g++ -c -g -O3 -I$TEMP/inst$i/include testprogram$i.cpp
    g++ -g -o testprogram$i testprogram$i.o -Wl,-rpath=$TEMP/inst$i/lib \
        -L$TEMP/inst$i/lib -lGeographicLib
    ./testprogram$i
done

echo ==============================================================
echo Verify library versions of cmake and autoconf builds are the same and other checks

libversion=`find $TEMP/instc/lib -type f \
-name 'libGeographicLib.so.*.*' -printf "%f" |
sed 's/libGeographicLib\.so\.//'`
test -f $TEMP/instb/lib/libGeographicLib.so.$libversion ||
echo autoconf/cmake library so mismatch

CONFIG_FILE=$TEMP/gitr/geographiclib/configure
CONFIG_MAJOR=`grep ^GEOGRAPHICLIB_VERSION_MAJOR= $CONFIG_FILE | cut -f2 -d=`
CONFIG_MINOR=`grep ^GEOGRAPHICLIB_VERSION_MINOR= $CONFIG_FILE | cut -f2 -d=`
CONFIG_PATCH=`grep ^GEOGRAPHICLIB_VERSION_PATCH= $CONFIG_FILE | cut -f2 -d=`
CONFIG_VERSIONA=`grep ^PACKAGE_VERSION= $CONFIG_FILE | cut -f2 -d= |
cut -f2 -d\'`
CONFIG_VERSION=$CONFIG_MAJOR.$CONFIG_MINOR
test "$CONFIG_PATCH" = 0 || CONFIG_VERSION=$CONFIG_VERSION.$CONFIG_PATCH
test "$CONFIG_VERSION"  = "$VERSION" || echo autoconf version number mismatch
test "$CONFIG_VERSIONA" = "$VERSION" || echo autoconf version string mismatch

cd $TEMP/relx/GeographicLib-$VERSION
(
    echo Files with trailing spaces:
    find . -type f | grep -E -v 'config\.guess|Makefile\.in|\.m4|\.png|\.gif' |
        while read f; do
            tr -d '\r' < $f | grep ' $' > /dev/null && echo $f || true
        done
    echo
    echo Files with tabs:
    find . -type f |
        grep -E -v '[Mm]akefile|\.html|\.m4|\.png|\.gif' |
        grep -E -v \
             '\.sh|depcomp|install-sh|/config\.|configure$|compile|missing' |
        xargs grep -l  '	' || true
    echo
    echo Files with multiple newlines:
    find . -type f |
        grep -E -v \
           '/Makefile\.in|\.1\.html|\.png|\.gif|/ltmain|/config|\.m4' |
        while read f; do
            tr 'X\n' 'xX' < $f | grep XXX > /dev/null && echo $f || true
        done
    echo
    echo Files with no newline at end:
    find . -type f |
        grep -E -v '\.png|\.gif' |
        while read f; do
            n=`tail -1 $f | wc -l`; test $n -eq 0 && echo $f || true
        done
    echo
    echo Files with extra newlines at end:
    find . -type f |
        grep -E -v '/configure|/ltmain.sh|\.png|\.gif|\.1\.html' |
        while read f; do
            n=`tail -1 $f | wc -w`; test $n -eq 0 && echo $f || true
        done
    echo
) > $TEMP/badfiles.txt
cat $TEMP/badfiles.txt
cat > $TEMP/tasks.txt <<EOF
# install built version
sudo make -C $TEMP/relc/GeographicLib-$VERSION/BUILD-system install

# commit and tag release branch
cd $TEMP/gitr/geographiclib
# Check .gitignore files!
git add -A
git commit -m "Version $VERSION ($DATE)"
git tag -m "Version $VERSION ($DATE)" r$VERSION
git push
git push --tags

# tag main branch
cd $DEVELSOURCE
git tag -m "Version $VERSION ($DATE)" v$VERSION
git push --all
git push --tags

# Also to do
# post release notices
# set default download files
# update home brew
#   dir = /usr/local/Homebrew/Library/Taps/homebrew/homebrew-core
#   branch = geographiclib/$VERSION
#   file = Formula/geographiclib.rb
#   brew install --build-from-source geographiclib
#   commit message = geographiclib $VERSION
# update vcpkg git@github.com:microsoft/vcpkg
#   dir = ports/geographiclib
#   ./vcpkg install 'geographiclib[tools]'
#   binaries in installed/x64-linux/tools/geographiclib
#   libs in installed/x64-linux/{include,lib,debug/lib}
#   ./vcpkg x-add-version geographiclib
#   commit message = [geographiclib] Update to version $VERSION
# update conda-forge
#   url = git@github.com:conda-forge/geographiclib-cpp-feedstock
#   conda build recipe
# update binaries for cgi applications
EOF
echo cat $TEMP/tasks.txt
cat $TEMP/tasks.txt
END=`date +%s`
echo Elapsed time $((END-START)) secs

#! /bin/sh
# Run Inno Setup Compiler to create installers for magnetic datasets

INNO="c:/Program Files/Inno Setup 6/ISCC.exe"
test -f "$INNO" || INNO="c:/Program Files (x86)/Inno Setup 6/ISCC.exe"

MAGNETICDIR=..
(
cat <<EOF
igrf14  IGRF14
EOF
) | while read prefix name; do
    "$INNO" magnetic-installers.iss \
        //dMAGNETICDIR="$MAGNETICDIR" \
        //dPREFIX="$prefix" \
        //dNAME="$name" > $prefix.log 2>&1
done
exit 0

cat <<EOF
wmm2010 WMM2010
emm2010 EMM2010
igrf11  IGRF11
wmm2015 WMM2015
igrf12  IGRF12
emm2015 EMM2015
emm2017 EMM2017
wmm2015v2 WMM2015v2
wmm2020 WMM2020
igrf13  IGRF13
wmm2025 WMM2025
wmmhr2025 WMMHR2025
igrf14  IGRF14
EOF

#! /bin/sh
# Run Inno Setup Compiler to create installers for geoid datasets

INNO="c:/Program Files/Inno Setup 5/ISCC.exe"
test -f "$INNO" || INNO="c:/Program Files (x86)/Inno Setup 5/ISCC.exe"

GRAVITYDIR=..
GRAVITYDIR=`cygpath -w $GRAVITYDIR`
(
cat <<EOF
egm84 EGM84
egm96 EGM96
egm2008 EGM2008
wgs84 WGS84
grs80 GRS80
EOF
) | while read prefix name; do
    "$INNO" gravity-installers.iss \
        /dGRAVITYDIR="$GRAVITYDIR" \
        /dPREFIX="$prefix" \
        /dNAME="$name" > $prefix.log 2>&1
done
exit 0

\name{GeoSim}
\alias{GeoSim}
\alias{print.GeoSim}
\encoding{UTF-8}
\title{Simulation of Gaussian and non-Gaussian random fields.}
\description{
  Simulates a realization of a Gaussian or non-Gaussian spatial, spatio-temporal,
  or spherical random field, and Gaussian bivariate random fields, for a specified
  covariance or correlation model.
  The covariance parameters are supplied through \code{param}; available correlation
  models are documented in \code{\link{GeoCovmatrix}}.
}
\usage{
GeoSim(coordx=NULL, coordy=NULL,coordz=NULL, coordt=NULL, coordx_dyn=NULL, corrmodel, 
 distance="Eucl", grid=FALSE, method="cholesky", 
 model='Gaussian', n=1, param,anisopars=NULL,radius=1, 
 sparse=FALSE,X=NULL,spobj=NULL,nrep=1,progress=TRUE,check.duplicates=FALSE)
}
\arguments{
 \item{coordx}{May be \code{NULL} when \code{coordx_dyn} or \code{spobj} is supplied. Otherwise, a numeric (\eqn{d \times 2}{d x 2})-matrix or 
 (\eqn{d \times 3}{d x 3})-matrix. Coordinates on a sphere for a fixed radius \code{radius} 
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordt}{A numeric vector giving the temporal coordinates at which the field is simulated. Optional argument; the default is \code{NULL}, in which case a spatial random field is expected. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{For dynamic simulation sites, a list with one two- or three-column coordinate matrix per element of \code{coordt}.  The output at time \eqn{t} follows the row order of \code{coordx_dyn[[t]]}.  See \code{\link{GeoModels-spacetime-ordering}}.}
 \item{corrmodel}{String; the name of a correlation model, for the
 see \code{\link{GeoCovmatrix}} for the list of implemented correlation models.}
 \item{distance}{String; the name of the spatial distance. The default
 is \code{Eucl}, the Euclidean distance. See \code{\link{GeoFit}} for details.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data
 are interpreted as spatial or spatio-temporal realisations on a set
 of non-equispaced spatial sites (irregular grid).}
 \item{method}{String; the type of matrix decomposition used in the simulation. 
 Default is cholesky. The other possible choices is \code{svd}.} 
 \item{model}{String; the type of RF and therefore the densities associated to the likelihood
 objects. \code{Gaussian} is the default, see the Section
 Details.}
 \item{n}{Positive integer size parameter. For Binomial it may be scalar or site-specific; for direct Negative Binomial it is the single common number \eqn{r} of successes, with Geometric corresponding to \eqn{r=1}.} 
 \item{param}{A list of parameter values required in the
 simulation procedure of random fields, see \bold{Examples}.}
 \item{anisopars}{ A list of two elements "angle" and "ratio" i.e. the anisotropy angle and the anisotropy ratio, respectively.}
 \item{radius}{Numeric; a value indicating the radius of the sphere when using the great 
 circle distance. Default value is 1.} 
 \item{sparse}{Logical; if \code{TRUE} then cholesky decomposition is performed
 using sparse matrices algorithms (spam packake). It should be used
 with compactly supported covariance models.\code{FALSE} is the default.}
 \item{X}{Numeric design matrix for the mean.  For fixed locations, rows are ordered by time blocks, with all sites at the first time followed by all sites at the second time.  For dynamic locations, supply either a stacked matrix in temporal-block order or a list with \code{X[[t]]} aligned with \code{coordx_dyn[[t]]}.}
 \item{spobj}{An object of class \code{sp} or \code{spacetime}. For space-time objects, the current \code{sp2Geo()} conversion uses sequential temporal indices and does not preserve irregular original time spacing; to retain irregular temporal distances, use the explicit-coordinate interface with numeric \code{coordt}.}
 \item{nrep}{Numeric; Numbers of indipendent replicates.}
 \item{progress}{Logic; If TRUE then a progress bar is shown.}
 
\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
}


\details{

Bivariate simulation is currently implemented only for \code{model = "Gaussian"};
other bivariate marginal models are rejected explicitly.  \code{Binary} and
\code{Bernoulli} are aliases of a binomial field with \code{n = 1}; \code{Geom}
and \code{Geometric} are aliases of a negative-binomial field with
\code{n = 1}. Direct simulation of \code{model = "Beta2"} is not implemented
by \code{GeoSim}; use \code{\link{GeoSimCopula}} with an explicit copula instead.
The function stops explicitly rather than returning a latent Gaussian draw.
Models whose names contain \code{Gaussian_misp_} are inferential working models and
are rejected as data-generating mechanisms. Other model names that are valid elsewhere
in GeoModels but do not have a direct simulator are also rejected explicitly.

Several direct non-Gaussian constructions use an integer number of independent latent
Gaussian fields. Consequently, \code{Gamma} requires integer \code{shape};
\code{Beta} requires integer \code{shape1} and \code{shape2}; and for
\code{StudentT}, \code{SkewStudentT}, and \code{TwoPieceStudentT},
\code{1/df} must be an integer at least 3. For \code{PoissonGamma} and
\code{PoissonGammaZIP}, \code{2*shape} must be a positive integer. These
parameters are never rounded silently. These restrictions apply to the direct latent-field
constructions in \code{GeoSim}; they do not apply to quantile-transformed copula margins
in \code{GeoSimCopula}.

\code{nrep} must be a positive integer. For binomial and negative-binomial fields,
\code{n} must contain positive integers and may be scalar or have one value per
simulated observation. The returned \code{param} component preserves the parameter
list supplied by the user rather than the standardized latent-Gaussian working parameters.

For univariate simulation, the location parameter is
\eqn{\mu=X\beta}, with coefficients named \code{mean}, \code{mean1}, and so
on in the order of the columns of \code{X}.  If \code{X=NULL}, the simulation
is intercept-only and uses scalar \code{mean}.  Alternatively,
\code{param$mean} may be a vector with one value per simulated observation;
this external mean is mutually exclusive with \code{X}.

For the Tukey transformed-Gaussian models, \code{Tukeyh} requires
\code{0 <= tail < 0.5}, while \code{Tukeyh2} requires both
\code{0 <= tail1 < 0.5} and \code{0 <= tail2 < 0.5}.  In \code{Tukeyh2},
\code{tail1} is the right-tail parameter and \code{tail2} is the left-tail
parameter. Zero is an allowed boundary and recovers the Gaussian transformation
on the corresponding side. \code{SinhAsinh} requires a strictly positive
\code{tail}.
}

\section{Spatio-temporal ordering}{
With fixed locations, a simulated space-time realization is returned as a
\eqn{T \times N}{T x N} matrix: rows correspond to \code{coordt} and columns
to rows of \code{coordx}.  The corresponding internal and \code{X} row order
is \code{c(t(data))}, i.e. time then site.

With dynamic locations, the simulated realization is a list of length
\eqn{T}.  Element \code{data[[t]]} has one value per row of
\code{coordx_dyn[[t]]}, in the same row order.  Replicates, when requested,
contain objects with this same fixed or dynamic layout.  See
\code{\link{GeoModels-spacetime-ordering}}.
}

\value{
 Returns an object of class \code{GeoSim}.
 An object of class \code{GeoSim} is a list containing
 at most the following components:
 
 \item{bivariate}{Logical:\code{TRUE} if the Gaussian RF is bivariate, otherwise \code{FALSE};}
 \item{coordx}{A \eqn{d}{d}-dimensional vector of spatial coordinates;}
 \item{coordy}{A \eqn{d}{d}-dimensional vector of spatial coordinates;}
 \item{coordz}{A \eqn{d}{d}-dimensional vector of spatial coordinates;}
 \item{coordt}{A \eqn{t}{t}-dimensional vector of temporal coordinates;}
 \item{coordx_dyn}{A list of dynamical (in time) spatial coordinates;}
 \item{corrmodel}{The correlation model; see \code{\link{GeoCovmatrix}}.}
 \item{data}{The simulated data.  For fixed-location space-time simulation this is a matrix with times in rows and sites in columns; for dynamic locations it is a list with one vector per time, aligned with \code{coordx_dyn}.  See \code{\link{GeoModels-spacetime-ordering}}.}
 \item{distance}{The type of spatial distance;}
 \item{method}{The method of simulation}
 \item{model}{The type of RF, see \code{\link{GeoFit}}.}
 \item{n}{The Binomial number of trials; for direct Negative Binomial, the common number \eqn{r} of successes.} 
 \item{numcoord}{The number of spatial coordinates;}
 \item{numtime}{The number the temporal realisations of the RF;}
 \item{param}{The parameter list supplied to the simulation call;}
 \item{radius}{The radius of the sphere if coordinates are passed in lon/lat format;}
 \item{spacetime}{\code{TRUE} if spatio-temporal and \code{FALSE} if
 spatial RF;}
 \item{nrep}{The number of indipendent replicates;}
}



\seealso{
\code{\link{GeoCovmatrix}} for covariance matrix construction,
  \code{\link{GeoFit}} for parameter estimation,
  \code{\link{GeoSimcond}} and \code{\link{GeoSimapprox}} for conditional and approximate simulation.
}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}


\examples{
library(GeoModels)


################################################################
###
### Example 1. Simulation of a spatial Gaussian RF 
### with Matern and Generalized Wendland correlations
###############################################################

# Define the spatial-coordinates of the points:
x <- runif(500);y <- runif(500)
coords=cbind(x,y)
set.seed(261)
# Simulation of a spatial Gaussian RF with Matern correlation function
data1 <- GeoSim(coordx=coords, corrmodel="Matern", param=list(smooth=0.5,
 mean=0,sill=1,scale=0.4/3,nugget=0))$data

set.seed(261)
data2 <- GeoSim(coordx=coords, corrmodel="GenWend", param=list(smooth=0,
 power2=4,mean=0,sill=1,scale=0.4,nugget=0))$data
opar=par(no.readonly = TRUE)
par(mfrow=c(1,2))
if (requireNamespace("fields", quietly = TRUE)) {
 fields::quilt.plot(coords, data1, main = "Matern", xlab = "", ylab = "")
}
if (requireNamespace("fields", quietly = TRUE)) {
 fields::quilt.plot(coords, data2, main = "Wendland", xlab = "", ylab = "")
}
par(opar)
 

################################################################
###
### Example 2. Simulation of a spatial geometric RF 
### with underlying Wend0 correlation
###
################################################################

# Define the spatial-coordinates of the points:
x <- runif(800);y <- runif(800)
coords <- cbind(x,y)
set.seed(251)
# Simulation of a spatial Binomial RF:
sim <- GeoSim(coordx=coords, corrmodel="Wend0",
 model="BinomialNeg",n=1,sparse=TRUE,
 param=list(nugget=0,mean=0,scale=.2,power2=4))

if (requireNamespace("fields", quietly = TRUE)) {
 fields::quilt.plot(
  coords, sim$data, nlevel = max(sim$data),
  col = terrain.colors(max(sim$data + 1))
 )
}

################################################################
###
### Example 3. Simulation of a spatial Weibull RF
### with underlying Matern correlation on a regular grid
###
###############################################################
# Define the spatial-coordinates of the points:
x <- seq(0,1,0.032)
y <- seq(0,1,0.032)
set.seed(261)
# Simulation of a spatial Gaussian RF with Matern correlation function
data1 <- GeoSim(x,y,grid=TRUE, corrmodel="Matern",model="Weibull", 
 param=list(shape=1.2,mean=0,scale=0.3/3,nugget=0,smooth=0.5))$data
if (requireNamespace("fields", quietly = TRUE)) {
 fields::image.plot(x, y, data1, main = "Weibull RF", xlab = "", ylab = "")
}

################################################################
###
### Example 4. Simulation of a spatial t RF
### with with underlying Generalized Wendland correlation 
###
###############################################################
# Define the spatial-coordinates of the points:
x <- seq(0,1,0.03)
y <- seq(0,1,0.03)
set.seed(268)
# Simulation of a spatial Gaussian RF with Matern correlation function
data1 <- GeoSim(x,y,grid=TRUE, corrmodel="GenWend",model="StudentT", sparse=TRUE,
 param=list(df=1/4,mean=0,sill=1,scale=0.3,nugget=0,smooth=1,power2=5))$data
if (requireNamespace("fields", quietly = TRUE)) {
 fields::image.plot(
  x, y, data1, col = terrain.colors(100), main = "Student-t RF",
  xlab = "", ylab = ""
 )
}


################################################################
###
### Example 5. Simulation of a sinhasinh RF
### with underlying Wend0 correlation.
###
###############################################################

# Define the spatial-coordinates of the points:
x <- runif(500, 0, 2)
y <- runif(500, 0, 2)
coords <- cbind(x,y)
set.seed(261)
corrmodel="Wend0"
# Simulation of a spatial Gaussian RF:
param=list(power2=4,skew=0,tail=1,
 mean=0,sill=1,scale=0.2,nugget=0) ## gaussian case
data0 <- GeoSim(coordx=coords, corrmodel=corrmodel,
 model="SinhAsinh", param=param,sparse=TRUE)$data
plot(density(data0),xlim=c(-7,7))

param=list(power2=4,skew=0,tail=0.7,
 mean=0,sill=1,scale=0.2,nugget=0) ## heavy tails
data1 <- GeoSim(coordx=coords, corrmodel=corrmodel,
 model="SinhAsinh", param=param,sparse=TRUE)$data
lines(density(data1),lty=2)

param=list(power2=4,skew=0.5,tail=1,
 mean=0,sill=1,scale=0.2,nugget=0) ## asymmetry
data2 <- GeoSim(coordx=coords, corrmodel=corrmodel,
 model="SinhAsinh", param=param,sparse=TRUE)$data
lines(density(data2),lty=3)

################################################################
###
### Example 6. Simulation of a bivariate Gaussian RF
### with bivariate Matern correlation model
###
###############################################################

# Define the spatial-coordinates of the points:
x <- runif(500, 0, 2)
y <- runif(500, 0, 2)
coords <- cbind(x,y)

# Simulation of a bivariate spatial Gaussian RF:
# with a separable Bivariate Matern
param=list(mean_1=4,mean_2=2,smooth_1=0.5,smooth_2=0.5,smooth_12=0.5,
 scale_1=0.12,scale_2=0.1,scale_12=0.15,
 sill_1=1,sill_2=1,nugget_1=0,nugget_2=0,pcol=0.5)
data <- GeoSim(coordx=coords,corrmodel="Bi_matern",
 param=param)$data
opar=par(no.readonly = TRUE)
par(mfrow=c(1,2))
if (requireNamespace("fields", quietly = TRUE)) {
 fields::quilt.plot(
  coords, data[1, ], col = terrain.colors(100), main = "1",
  xlab = "", ylab = ""
 )
}
if (requireNamespace("fields", quietly = TRUE)) {
 fields::quilt.plot(
  coords, data[2, ], col = terrain.colors(100), main = "2",
  xlab = "", ylab = ""
 )
}
par(opar)


################################################################
###
### Example 7. Simulation of a spatio temporal Gaussian random field.
### observed on fixed location sites with double Matern correlation 
###
###############################################################



coordt=1:5

# Define the spatial-coordinates of the points:
x <- runif(50, 0, 2)
y <- runif(50, 0, 2)
coords <- cbind(x,y)

param<-list(nugget=0,mean=0,scale_s=0.2/3,scale_t=2/3,sill=1,smooth_s=0.5,smooth_t=0.5)
data <- GeoSim(coordx=coords, coordt=coordt, corrmodel="Matern_Matern",
 param=param)$data
dim(data)

################################################################
###
### Example 8. Simulation of a spatio temporal Gaussian random field.
### observed on dynamical location sites with double Matern correlation 
###
###############################################################

# Define the dynamical spatial-coordinates of the points:

coordt=1:5
coordx_dyn=list()
maxN=30
set.seed(8)
for(k in 1:length(coordt))
{
NN=sample(1:maxN,size=1)
x <- runif(NN, 0, 1)
y <- runif(NN, 0, 1)
coordx_dyn[[k]]=cbind(x,y)
}
coordx_dyn

param<-list(nugget=0,mean=0,scale_s=0.2/3,scale_t=2/3,sill=1,smooth_s=0.5,smooth_t=0.5)
data <- GeoSim(coordx_dyn=coordx_dyn, coordt=coordt, corrmodel="Matern_Matern",
 param=param)$data
## spatial realization at first temporal instants
data[[1]]
## spatial realization at third temporal instants
data[[3]]




################################################################
###
### Example 9. Simulation of a Gaussian RF 
### with a Wend0 correlation in the north emisphere of the planet earth
### using geodesic distance
###############################################################
distance="Geod";radius=6371

NN=3000 ## total point on the sphere on lon/lat format
set.seed(80)
coords=cbind(runif(NN,-180,180),runif(NN,0,90))
## Set the wendland parameters
corrmodel <- "Wend0"
param<-list(mean=0,sill=1,nugget=0,scale=1000,power2=3)
# Simulation of a spatial Gaussian RF on the sphere
#set.seed(2)
data <- GeoSim(coordx=coords,corrmodel=corrmodel,sparse=TRUE,
 distance=distance,radius=radius,param=param)$data
#require(globe)
#globe::globeearth(eye=place("newyorkcity"))
#globe::globepoints(loc=coords,pch=20,col = cm.colors(length(data),alpha=0.4)[rank(data)])




}
\keyword{Simulation}

\name{plot.GeoVariogram}
\alias{plot.GeoVariogram}
\encoding{UTF-8}
\title{Plot empirical spatial, spatio-temporal and bivariate semivariograms}

\description{
 Plot an empirical spatial, spatio-temporal, or bivariate semivariogram
 returned by \code{\link{GeoVariogram}}.
}

\usage{
\method{plot}{GeoVariogram}(x, \dots)
}

\arguments{
 \item{x}{an object of the class \code{"GeoVariogram"} }
 \item{\dots}{other arguments to be passed to the function
 \code{\link{plot}} }
}

\details{
 Spatial objects are shown as a standard lag plot. Spatio-temporal objects
 show the empirical surface and its spatial and temporal margins. Bivariate
 objects show the two marginal semivariograms and the cross-semivariogram;
 missing empirical bins are handled as \code{NA}.
}

\value{
 Produces a plot.
 No values are returned.
}

\seealso{
 \code{\link{GeoVariogram}} for variogram computation and examples.
}

\keyword{Variogram}

\name{winds.coords}
\alias{winds.coords}
\title{Weather Stations of the Irish Daily Wind Speeds}
\usage{data(winds.coords)}
\description{
 A data frame containing information about the weather stations
 where the data are recorded in Ireland.
}
\format{A data frame containing site - the name of the city (character),
 abbr - the abbrevation (character), elev - the elevation (numeric),
 lat - latitude (numeric) and lon - longitude.}
\source{
 Haslett, J. and Raftery, A. E. (1989),
 Space-time modelling with long-memory dependence: assessing
 Ireland's wind-power resource (with discussion),
 \emph{Applied Statistics}, 38, 1--50.
}
\keyword{datasets}

\name{GeoVarest}
\alias{GeoVarest}
\encoding{UTF-8}
\title{Score-based variance estimation for \code{GeoFit} objects}

\description{
 The function updates a fitted \code{GeoFit} object by estimating the
 variability matrix of the composite likelihood score through parametric simulation.
 The fitted model is used to generate \code{K} independent datasets. For each simulated dataset, the composite likelihood score is evaluated at the original estimate \eqn{\hat\theta}, without refitting the model. The empirical
 variance of these simulated scores provides an estimate of the variability matrix \eqn{J}. Together with the sensitivity matrix \eqn{H}, computed by
 \code{GeoFit} when \code{sensitivity = TRUE}, this yields the Godambe
 sandwich covariance matrix
 \deqn{
 G^{-1} = H^{-1} J H^{-1}.
 }
 The updated object contains standard errors, Wald confidence intervals,
 p-values, the estimated matrices \eqn{J}, \eqn{H^{-1}} and \eqn{G^{-1}},
 and composite likelihood information criteria based on the penalty
 \eqn{\mathrm{tr}(H^{-1}J)}.
}
\usage{
GeoVarest(fit, K = 100, sparse = FALSE,
 method = c("cholesky", "TB", "CE"),
 alpha = 0.95, L = 10000,
 parallel = FALSE, ncores = 6, progress = TRUE, seed = NULL,
 min_success_rate = 0.8)
}
\arguments{
 \item{fit}{A fitted object obtained from \code{\link{GeoFit}}. The object
 must contain the sensitivity matrix, hence \code{GeoFit} should be called
 with \code{sensitivity = TRUE}. Full/Standard likelihood fits are not accepted;
 use \code{GeoFit(..., likelihood = "Full", type = "Standard", varest = TRUE)}
 for Hessian-based standard errors.}

 \item{K}{The number of simulations used in the parametric score bootstrap.}

 \item{sparse}{Logical; if \code{TRUE}, then Cholesky decomposition is
 performed using sparse matrix algorithms.}

 \item{method}{String; the method of simulation. The default is
 \code{"cholesky"}. For large data sets the options \code{"TB"} and
 \code{"CE"} call approximate simulation methods; see
 \code{\link{GeoSimapprox}}.}

 \item{alpha}{Numeric; the level of the confidence interval.}

 \item{L}{Numeric; the number of lines in the turning bands method.}

 \item{parallel}{Logical; default \code{FALSE}. If \code{TRUE}, the score evaluation step is
 parallelized.}

 \item{ncores}{Positive integer or \code{NULL}; default \code{6}. With \code{parallel=TRUE}, an explicit integer requests that many workers, capped by detected cores and the number of bootstrap jobs. Set \code{ncores=NULL} for automatic selection, capped at six workers and normally leaving one detected core free.}

 \item{progress}{Logical; if \code{TRUE}, progress information is shown.}

 \item{seed}{Optional integer seed for reproducibility of the simulated
 samples.}
 \item{min_success_rate}{Minimum fraction of successful score-bootstrap
 replications required to return variance estimates. The default is 0.8.}
}
\details{
For spatio-temporal fits, the original numeric \code{coordt} values stored in \code{fit} are reused. Irregularly spaced times are supported with \code{method = "cholesky"}; approximate simulation methods retain the restrictions documented in \code{\link{GeoSimapprox}}.


 Let \eqn{cl(\theta)} denote the composite log-likelihood and let
 \eqn{U(\theta) = \nabla cl(\theta)} be the corresponding composite score.
 The function simulates \code{K} data sets from the fitted model and evaluates
 the composite score at the fitted parameter value \eqn{\hat\theta}. The
 variability matrix is estimated as
 \deqn{\hat J = Var\{U_1(\hat\theta), \ldots, U_K(\hat\theta)\}.}

 If \eqn{H} is the sensitivity matrix stored in \code{fit$sensmat}, the
 inverse Godambe matrix is estimated by
 \deqn{\widehat{G^{-1}} = H^{-1} \hat J H^{-1}.}
 Standard errors are obtained from the square root of the diagonal of
 \eqn{\widehat{G^{-1}}}.

 For composite likelihoods, the penalty used in the information criterion is
 \deqn{tr(H^{-1}\hat J),}
 and the composite likelihood information criterion is computed as
 \deqn{-2 cl(\hat\theta) + 2 tr(H^{-1}\hat J).}

 Differently from \code{\link{GeoVarestbootstrap}}, this function does not
 refit the model for each simulated data set. It estimates the variability
 matrix of the score and then computes the sandwich/Godambe covariance
 matrix.

 For stochastic nearest-neighbor fits, the realized retained-pair graph stored
 in the original fit is reused unchanged for every simulated data set. Thus the
 score bootstrap estimates variability conditional on the selected pair graph,
 in agreement with the current implementation described in the stochastic-NN
 methodology.


Parallel multisession workers are started with the package-library paths of the calling R session, including the library containing the loaded GeoModels installation.

For \code{method = "TB"} without a copula, parallel workers are persistent and replications are streamed one at a time (simulate, evaluate, release). This keeps peak memory bounded when both \code{L} and \code{K} are large while still reusing the selected pair graph and static fitting context within each worker. For fitted copula models, \code{method="TB"} is simulated through the turning-bands backend of \code{GeoSimCopula}; the copula path is currently not streamed replicate-by-replicate. On platforms where \pkg{future} reports forked multicore execution as safe, GeoModels uses it automatically to reduce worker startup and serialization overhead; otherwise it falls back to multisession with the current GeoModels library path propagated explicitly.
}
\value{
 Returns an updated object of class \code{GeoFit}. The following components
 are added or updated:
 \item{stderr}{Estimated standard errors obtained from the inverse Godambe
 matrix.}
 \item{varcov}{Estimated inverse Godambe matrix \eqn{\widehat{G^{-1}}}.}
 \item{godambe}{Estimated Godambe matrix.}
 \item{Jmat}{Estimated variability matrix of the composite score.}
 \item{Hinv}{Inverse, or generalized inverse, of the sensitivity matrix.}
 \item{claic}{Composite likelihood AIC-type criterion.}
 \item{clic}{Same value as \code{claic}.}
 \item{clbic}{Composite likelihood BIC-type criterion.}
 \item{clic_penalty}{Penalty term \eqn{tr(H^{-1}\hat J)}.}
 \item{conf.int}{Wald-type confidence intervals based on the estimated
 standard errors.}
 \item{pvalues}{Wald-type p-values.}
 \item{scores}{Matrix of successful bootstrap score evaluations.}
 \item{score_logCompLik}{Composite log-likelihood values corresponding to the
 successful score evaluations.}
 \item{score_failures}{Data frame with failed score evaluations, if any.}
}

\section{Three-dimensional coordinates}{
For a purely spatial univariate fit with explicit irregular three-dimensional
Euclidean coordinates, \code{method = "TB"} is supported when no
\code{anisopars} were used. Method \code{"CE"} remains restricted to
regular two-dimensional grids. Use \code{method = "cholesky"} for
bivariate, spatio-temporal, anisotropic, or other unsupported
three-dimensional cases.
}
\seealso{
\code{\link{GeoFit}} for the fitted objects used as input
}
\author{
Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},
\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl},
\url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},
\url{https://www.researchgate.net/profile/Christian-Caamano}
}
\examples{


library(GeoModels)

################################################################
###
### Example 1. Test on the parameter
### of a regression model using conditional composite likelihood
###
###############################################################
set.seed(342)
model="Gaussian" 
# Define the spatial-coordinates of the points:
NN=3500
x = runif(NN, 0, 1)
y = runif(NN, 0, 1)
coords = cbind(x,y)
# Parameters
mean=1; mean1=-1.25; # regression parameters
 sill=1 # variance

# matrix covariates
X=cbind(rep(1,nrow(coords)),runif(nrow(coords)))

# model correlation 
corrmodel="Matern"
smooth=0.5;scale=0.1; nugget=0;

# simulation
param=list(smooth=smooth,mean=mean,mean1=mean1,
 sill=sill,scale=scale,nugget=nugget)
data = GeoSim(coordx=coords, corrmodel=corrmodel,
 model=model, param=param,X=X)$data

I=Inf

fixed=list(nugget=nugget,smooth=smooth)
start=list(mean=mean,mean1=mean1,scale=scale,sill=sill)

lower=list(mean=-I,mean1=-I,scale=0,sill=0)
upper=list(mean=I,mean1=I,scale=I,sill=I)
# Maximum pairwise composite-likelihood fitting of the RF:
fit = GeoFit(data=data,coordx=coords,corrmodel=corrmodel, model=model,
 likelihood="Conditional",type="Pairwise",sensitivity=TRUE,
 lower=lower,upper=upper,neighb=3,
 optimizer="nlminb",X=X,
 start=start,fixed=fixed)

unlist(fit$param)


#fit_update=GeoVarest(fit,K=100,parallel=TRUE)
#fit_update$stderr
#fit_update$conf.int
#fit_update$pvalues

}
\keyword{Composite}

\name{GeoTestsupp_space}
\alias{GeoTestsupp_space}
\title{Parametric Bootstrap Test for Spatial Support}
\description{
Performs a parametric bootstrap likelihood-ratio test for the spatial
\emph{scale}, interpreted as the compact-support radius of the
\code{"GenWend"} correlation model. The null model either constrains the
scale by \eqn{\mathrm{scale} \le h_0}{scale <= h0} or, when \code{h0 = NULL},
represents spatial independence by fixing the support below the minimum
observed inter-site distance. The alternative estimates the scale without the
null upper constraint.
}
\usage{
GeoTestsupp_space(data, coordx,
 start, fixed,
 model = "Gaussian",
 h0 = NULL,
 optimizer = "bobyqa",
 lower = NULL, upper = NULL,
 neighb = 5,
 B = 10000,
 likelihood = NULL,
 type = NULL,
 method = c("cholesky", "TB"),
 L = 10000,
 parallel = FALSE,
 ncores = 6,
 progress = TRUE,
 seed = NULL)
}
\arguments{
  \item{data}{Numeric vector containing the spatial observations. Missing or
  non-finite values are not allowed.}

  \item{coordx}{Numeric matrix whose rows contain the spatial coordinates.
  Its number of rows must equal \code{length(data)}. Duplicated locations are
  not allowed because the minimum inter-site distance must be positive.}

  \item{start}{Non-empty named list containing parameters to be estimated.
  Together with \code{fixed}, it must contain exactly the parameters returned
  by \code{CorrParam("GenWend")} and \code{NuisParam(model)}.}

  \item{fixed}{Named list containing parameters kept fixed, or \code{NULL}.
  A parameter cannot appear in both \code{start} and \code{fixed}. If
  \code{scale} is supplied in \code{fixed}, it is internally moved to the
  estimated-parameter list for construction of the unrestricted model.}

  \item{model}{Marginal model passed to \code{GeoFit}. Default is
  \code{"Gaussian"}.}

  \item{h0}{Positive finite threshold defining
  \eqn{H_0:\mathrm{scale}\le h_0}{H0: scale <= h0}. If \code{NULL}, the
  support is fixed at \code{max(1e-6, d_min - 1e-6)}, where
  \eqn{d_{min}}{d_min} is the minimum inter-site distance. This removes
  dependence among the sampled locations; for Gaussian random fields it is
  equivalent to independence of the sampled observations. It should not be
  confused with the process-level null hypothesis tested by
  \code{GeoTestIndependence}.}

  \item{optimizer}{Optimization method passed to \code{GeoFit}. Default is
  \code{"bobyqa"}.}

  \item{lower, upper}{Named lists containing bounds for parameters in
  \code{start}. Missing bounds are replaced internally by finite fallback
  bounds. The null upper bound for \code{scale} is always reimposed after any
  numerical widening of the parameter box.}

  \item{neighb}{Number of nearest neighbors used for pairwise composite
  likelihood. It is ignored for full likelihood. When full likelihood is
  selected automatically, it is internally set to \code{NULL}.}

  \item{B}{Positive integer giving the requested number of parametric
  bootstrap replications. Default is \code{10000}. Values below 99 are
  allowed but generate a warning because they give a coarse bootstrap
  \emph{p}-value.}

  \item{likelihood}{Likelihood passed to \code{GeoFit}. Supported values are
  \code{"Full"}, \code{"Marginal"}, and \code{"Conditional"}. Either both
  \code{likelihood} and \code{type} must be supplied, or both must be
  \code{NULL}.}

  \item{type}{Likelihood type passed to \code{GeoFit}. Full likelihood must
  use \code{"Standard"}; marginal or conditional composite likelihood must
  use \code{"Pairwise"}.}

  \item{method}{Simulation method used under the fitted null model.
  \code{"cholesky"} calls \code{GeoSim}; \code{"TB"} calls
  \code{GeoSimapprox}.}

  \item{L}{Number of spectral components used by \code{GeoSimapprox} when
  \code{method = "TB"}. Default is \code{10000}. Ignored for Cholesky
  simulation.}

  \item{parallel}{Logical; default \code{FALSE}. Set \code{TRUE} to evaluate bootstrap fits in parallel when the required packages are available.}

  \item{ncores}{Positive integer or \code{NULL}; default \code{6}. With \code{parallel=TRUE}, an explicit integer requests that many workers, capped by detected cores and the number of bootstrap replications. Set \code{ncores=NULL} for automatic selection, capped at six workers and normally leaving one detected core free.}

  \item{progress}{Logical value indicating whether progress information should
  be displayed through \pkg{progressr}.}

  \item{seed}{Optional finite numeric seed. When supplied, the test is
  reproducible and the user's previous random-number-generator state is
  restored when the function exits.}
}
\details{
The correlation model is fixed internally to \code{"GenWend"}. Therefore,
the parameter called \code{scale} is interpreted as the compact-support
radius.

\strong{Likelihood selection.}
When both \code{likelihood} and \code{type} are \code{NULL}, the function uses
full likelihood with \code{type = "Standard"} for at most 2000 observations,
and marginal pairwise composite likelihood for larger datasets. User-supplied
combinations are checked for consistency.

\strong{Null and alternative models.}
For a numeric \code{h0}, the null model estimates \code{scale} subject to
\eqn{\mathrm{scale}\le h_0}{scale <= h0}, while the unrestricted model uses a
data-dependent upper bound strictly larger than \code{h0}. For
\code{h0 = NULL}, the null model fixes the support radius just below the
minimum observed inter-site distance. Since the generalized Wendland
correlation is compactly supported, all distinct observations are then
uncorrelated under the null model.

The lower bound under the alternative must be smaller than \code{h0}, and its
upper bound must be larger than \code{h0}, so that the null parameter space is
nested in the alternative. The function stops if these conditions are not
satisfied.

\strong{Likelihood-ratio statistic.}
Let \eqn{\ell_0}{l0} and \eqn{\ell_1}{l1} denote the fitted log-likelihood or
log-composite-likelihood values under the null and unrestricted models. The
observed statistic is
\deqn{\Lambda_{\mathrm{obs}} = 2(\ell_1-\ell_0).}{Lambda_obs = 2*(l1-l0).}
Tiny negative differences attributable to numerical tolerance are set to zero.
A materially smaller unrestricted likelihood is treated as an optimization
failure rather than automatically converted to a zero statistic.

For pairwise composite likelihood, the function verifies that the null and
unrestricted fits use identical \code{rowidx} and \code{colidx} pair sets. A
likelihood-ratio comparison based on different pair sets is not accepted.

\strong{Adaptive unrestricted-fit fallback.}
The unrestricted model is first fitted once using the primary starting values.
Only if this fit fails or returns a likelihood smaller than the null likelihood
beyond numerical tolerance is a second fit attempted. The fallback starts from
the fitted null parameters and places \code{scale} slightly above \code{h0}.
The better valid unrestricted fit is retained. The same adaptive rule is used
for each bootstrap dataset.

\strong{Parametric bootstrap.}
Bootstrap datasets are simulated from the fitted null model. Each dataset is
refitted under both hypotheses, using the observed null and unrestricted
estimates as starting values. Replications with failed fits, incompatible pair
sets, non-finite likelihoods, or an unresolved ordering
\eqn{\ell_1 < \ell_0}{l1 < l0} are discarded. If \eqn{B_{\mathrm{valid}}}{B_valid}
replications are successful, the bootstrap \emph{p}-value is
\deqn{
\widehat p =
\frac{1 + \sum_{b=1}^{B_{\mathrm{valid}}}
I(\Lambda_b \ge \Lambda_{\mathrm{obs}})}
{B_{\mathrm{valid}} + 1}.
}{
p_hat = (1 + sum(Lambda_b >= Lambda_obs)) / (B_valid + 1).
}
At least \code{min(20, B)} valid replications are required. A warning is issued
when fewer than 80 percent of the requested replications are successful.

\strong{Early exit.}
If the unrestricted estimate itself satisfies the null constraint, the
likelihood-ratio statistic is zero and the function returns \code{pvalue = 1}
without running the bootstrap.

\strong{Parallel computation.}
For \code{method = "TB"}, simulation may use parallel computation inside
\code{GeoSimapprox}. Bootstrap refits are performed in a separate phase. In
the parallel bootstrap, each simulated dataset is written to a separate RDS
file so that each worker reads only its own replicate. The user's
\pkg{future} plan, \pkg{progressr} handlers, RNG state, and temporary files are
restored or removed on exit.
}
\value{
An invisible list. When the bootstrap is run, it contains:

  \item{d_min}{Minimum inter-site distance.}
  \item{h0}{Threshold actually used under the null hypothesis.}
  \item{lambda_obs}{Observed likelihood-ratio statistic.}
  \item{pvalue}{Parametric-bootstrap \emph{p}-value.}
  \item{seed}{Seed supplied by the user, or \code{NULL}.}
  \item{method}{Simulation method used.}
  \item{L}{Number of turning-bands components when \code{method = "TB"},
  otherwise \code{NULL}.}
  \item{B_requested}{Requested number of bootstrap replications.}
  \item{bootstrap_successful}{Number of valid bootstrap replications.}
  \item{bootstrap_failed}{Number of failed or discarded bootstrap
  replications.}
  \item{bootstrap_success_rate}{Fraction of requested replications that were
  valid.}
  \item{fallback_observed}{Logical value indicating whether the adaptive
  unrestricted-fit fallback was used for the observed data.}
  \item{fallback_bootstrap}{Number of bootstrap replications in which the
  adaptive unrestricted-fit fallback was attempted.}
  \item{B_rep}{Vector of valid bootstrap likelihood-ratio statistics.}
  \item{fit_H0}{Fitted \code{GeoFit} object under the null hypothesis.}
  \item{fit_H1}{Fitted \code{GeoFit} object under the unrestricted
  alternative.}

For an early exit, \code{lambda_obs = 0}, \code{pvalue = 1},
\code{bootstrap_successful = 0}, \code{bootstrap_success_rate = NA}, and
\code{B_rep} is empty. In the current implementation, \code{L} and
\code{bootstrap_failed} are omitted from this early-return object.
}
\note{
The test concerns the compact-support radius of the generalized Wendland
correlation model and is not a generic range-parameter test for arbitrary
correlation functions. For final inference, a substantially larger value of
\code{B} than that used in quick examples is recommended.
}
\seealso{
\code{\link{GeoFit}}, \code{\link{GeoSim}},
\code{\link{GeoSimapprox}}, \code{\link{GeoTestIndependence}},
\code{\link{GeoTestIsotropy}}, \code{\link{GeoTests}}
}
\examples{
\dontrun{
library(GeoModels)

set.seed(123)
n <- 300
coords <- cbind(runif(n), runif(n))

param <- list(
  mean = 0,
  nugget = 0,
  sill = 1,
  scale = 0.20,
  smooth = 0.5,
  power2 = 4
)

z <- GeoSim(
  coordx = coords,
  corrmodel = "GenWend",
  model = "Gaussian",
  param = param
)$data

start <- list(
  mean = 0,
  sill = 1,
  scale = 0.15
)
fixed <- list(
  nugget = 0,
  smooth = 0.5,
  power2 = 4
)

lower <- list(mean = -Inf, sill = 1e-6, scale = 1e-6)
upper <- list(mean = Inf, sill = Inf, scale = 1)

## Test H0: scale <= 0.10 against scale > 0.10.
ans <- GeoTestsupp_space(
  data = z,
  coordx = coords,
  start = start,
  fixed = fixed,
  h0 = 0.10,
  lower = lower,
  upper = upper,
  B = 199,
  likelihood = "Full",
  type = "Standard",
  parallel = FALSE,
  seed = 321
)

ans$pvalue
ans$lambda_obs
ans$bootstrap_success_rate

## Independence test: support fixed below the minimum inter-site distance.
ans_ind <- GeoTestsupp_space(
  data = z,
  coordx = coords,
  start = start,
  fixed = fixed,
  h0 = NULL,
  lower = lower,
  upper = upper,
  B = 199,
  likelihood = "Full",
  type = "Standard",
  parallel = FALSE,
  seed = 654
)
}
}
\author{
Moreno Bevilacqua
\email{moreno.bevilacqua89@gmail.com}
\url{https://sites.google.com/view/moreno-bevilacqua/home}
\cr
Victor Morales Onate
\email{victor.morales@uv.cl}
\url{https://sites.google.com/site/moralesonatevictor/}
\cr
Christian Caamano-Carrillo
\email{chcaaman@ubiobio.cl}
\url{https://www.researchgate.net/profile/Christian-Caamano}
}
\keyword{Bootstrap}
\keyword{Composite}
\keyword{HypothesisTest}
\keyword{Spatial}

\name{GeoFit}
\alias{GeoFit}
\alias{print.GeoFit}
\encoding{UTF-8}
\title{Maximum-Likelihood-Based Fitting of Gaussian and non-Gaussian random fields.}
\description{
 Maximum weighted (stochastic) composite-likelihood fitting for Gaussian and some non-Gaussian univariate spatial, spatio-temporal and bivariate spatial random fields. The function allows fixing any of the parameters and setting upper/lower bounds in the optimization.
 Different optimization methods can be used.
}
\usage{
GeoFit(data, coordx, coordy=NULL, coordz=NULL, coordt=NULL,
 coordx_dyn=NULL, copula=NULL, corrmodel=NULL, distance="Eucl",
 fixed=NULL, anisopars=NULL, est.aniso=c(FALSE,FALSE),
 grid=FALSE, likelihood="Marginal", lower=NULL, maxdist=Inf,
 neighb=NULL, p_neighb=1, maxtime=Inf, memdist=TRUE,
 method="cholesky", model="Gaussian", n=1, onlyvar=FALSE,
 optimizer="Nelder-Mead", radius=1, score=FALSE,
 sensitivity=FALSE, sparse=FALSE, start=NULL,
 thin_method="bernoulli", type="Pairwise", upper=NULL, 
 varest=FALSE, weighted=FALSE,
 X=NULL, spobj=NULL, spdata=NULL, check.duplicates=FALSE)
}

\arguments{
 \item{data}{A \eqn{d}{d}-dimensional vector (a single spatial realisation) or a (\eqn{d \times d}{d x d})-matrix (a single spatial realisation on regular grid)
 or a (\eqn{t \times d}{t x d})-matrix (a single spatio-temporal realisation) or a (\eqn{d \times d \times t \times n}{d x d x t x n})-array
 (a single spatio-temporal realisation on regular grid).
 See Details for the accepted data layouts.}

 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix.
 Coordinates on a sphere for a fixed radius \code{radius} are passed in lon/lat format expressed in decimal degrees.}

 \item{coordy}{A numeric vector giving 1-dimension of spatial coordinates; optional argument, default is \code{NULL}.}

 \item{coordz}{A numeric vector giving 1-dimension of spatial coordinates; optional argument, default is \code{NULL}.}

 \item{coordt}{A numeric vector assigning one dimension of the observation-time coordinates. Optional argument, default is \code{NULL}; if \code{NULL}, a spatial random field is expected. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}

 \item{coordx_dyn}{For dynamic locations, a list with one numeric coordinate matrix per temporal instant.  If \eqn{T} is \code{length(coordt)}, the list must have length \eqn{T}; \code{coordx_dyn[[t]]} must have \eqn{N_t} rows and two or three columns.  Its row order must match \code{data[[t]]} and, when supplied as a list, \code{X[[t]]}.  See \code{\link{GeoModels-spacetime-ordering}}.}

 \item{copula}{String; the copula used by pairwise copula likelihoods. Supported values are \code{"Gaussian"}, \code{"Clayton"} (the constructive Clayton-like spatial copula), and \code{"SkewGaussian"}.}

 \item{corrmodel}{String; the name of a correlation model; see \code{\link{GeoCovmatrix}} for the list of implemented correlation models.}

 \item{distance}{String; the name of the spatial distance. Default is \code{"Eucl"} (Euclidean distance). See Details for the accepted options.}

 \item{fixed}{An optional named list giving the values of the parameters that will be considered as known values. Parameter names must match exactly the nuisance/marginal or correlation parameters supported by the selected model and correlation model; invalid names are reported explicitly.
 The listed parameters for a given model/correlation function will not be estimated.}

 \item{anisopars}{A list of two elements: \code{"angle"} and \code{"ratio"}, i.e. the anisotropy angle and the anisotropy ratio, respectively. Geometric anisotropy is available only with \code{distance="Eucl"}.}

 \item{est.aniso}{A bivariate logical vector providing which anisotropy parameters must be estimated.}

 \item{grid}{Logical; if \code{FALSE} (default) the data are interpreted as spatial or spatio-temporal realisations on a set of non-equispaced spatial sites. If \code{TRUE}, \code{coordx}, \code{coordy} (and optionally \code{coordz}) are interpreted as grid axes and the data dimensions must match those axes; grid inputs are internally converted to the canonical explicit-coordinate layout before fitting.}

 \item{likelihood}{String; the configuration of the composite likelihood. \code{"Marginal"} is the default; see Details for the accepted options.}

 \item{lower}{An optional named list giving lower bounds for parameters when the optimizer is \code{L-BFGS-B}, \code{nlminb}, \code{bobyqa} or \code{optimize}.
 Names must match parameters that are being estimated.}

 \item{maxdist}{Numeric; an optional positive value indicating the maximum spatial distance considered in the composite computation.
 See Details for more information.}

 \item{neighb}{Numeric; an optional positive integer indicating the order of neighborhood in the composite likelihood computation.
 See Details for more information.}

 \item{p_neighb}{Numeric scalar in \eqn{(0,1]}. If \code{1} (default), no thinning is applied. For \code{thin_method="bernoulli"}, it is the candidate-pair inclusion probability in the constant-weight design and controls the retained size in expectation. For \code{thin_method="FixedBudget"}, it defines the exact global retained-pair budget \eqn{K=\mathrm{round}(p_{neighb}d)}, where \eqn{d} is the candidate-pair count.}

 \item{maxtime}{Numeric; an optional non-negative maximum temporal-distance threshold, expressed in the same units as \code{coordt}, used to select pairwise composite-likelihood contributions.}

 \item{memdist}{Deprecated logical argument retained for backward compatibility. The selected pair structure is always precomputed and reused during composite-likelihood optimization. Supplying \code{FALSE} produces a warning and is treated as \code{TRUE}.}

 \item{method}{String; the type of matrix decomposition/linear algebra backend used in likelihood computations.
 Default is \code{"cholesky"}. Another possible choice is \code{"svd"} (when available).}

 \item{model}{String; the type of random field (and associated density) used in the likelihood objects. Default is \code{"Gaussian"}; see Details for the accepted options.}

 \item{n}{Positive integer size parameter. For direct Binomial models it may be scalar or contain one value per observation. For direct Negative-Binomial models it is the common number \eqn{r} of successes and must be scalar; Geometric corresponds to \eqn{r=1}. Non-integer or non-positive values are rejected.}

 \item{onlyvar}{Logical; if \code{TRUE} (and \code{varest=TRUE}) only the variance-covariance matrix is computed without optimizing.
 Default is \code{FALSE}.}

 \item{optimizer}{String; the optimization algorithm (see \code{\link{optim}} for details). Default is \code{"Nelder-Mead"}.
 Other possible choices are \code{"nlm"}, \code{"BFGS"}, \code{"SANN"}, \code{"L-BFGS-B"}, \code{"nlminb"}, \code{"bobyqa"}.
 For \code{"L-BFGS-B"}, \code{"nlminb"} and \code{"bobyqa"} bounds can be passed via \code{lower} and \code{upper}.
 In the one-dimensional case, \code{optimize} is used.}

 \item{radius}{Numeric; the radius of the sphere in the case of lon-lat coordinates. Default is \code{1}.}

 \item{score}{Logical; if \code{TRUE} the score function is computed. Default is \code{FALSE}.}

 \item{sensitivity}{Logical; if \code{TRUE} the sensitivity matrix is computed. Default is \code{FALSE}.}

 \item{sparse}{Logical; if \code{TRUE} then maximum likelihood / composite likelihood may exploit sparse-matrix algorithms (e.g., \pkg{spam}).
 Typically used with compactly supported covariance models. Default is \code{FALSE}.}

 \item{start}{An optional named list with initial values for parameters to be estimated. Default is \code{NULL}. Parameter names must match exactly the nuisance/marginal or correlation parameters supported by the selected model and correlation model; invalid names are reported explicitly. Parameters omitted from \code{start} receive conservative internal initial values; explicitly supplied values are preserved (see Details).}

 \item{thin_method}{String; thinning scheme in stochastic weighted pairwise
 likelihood (used when \code{p_neighb < 1}). Default is \code{"bernoulli"}
 (independent Bernoulli thinning). An alternative thinning scheme is
 \code{"FixedBudget"}. The legacy name \code{"TargetBalanced"} is accepted as an alias for \code{"FixedBudget"}.}

 \item{type}{String; the type of likelihood objects. If \code{"Pairwise"} (default) the composite likelihood is formed by pairwise components (see Details).}

 \item{upper}{An optional named list giving upper bounds for parameters when the optimizer is \code{L-BFGS-B}, \code{nlminb}, \code{bobyqa} or \code{optimize}.
 Names must match parameters that are being estimated.}

 \item{varest}{Logical; if \code{TRUE} the estimates variances and standard errors are returned. For composite likelihood estimation it is deprecated.
 Use \code{sensitivity=TRUE} and update the object using \code{GeoVarest}. Default is \code{FALSE}.}

 \item{weighted}{Logical; if \code{TRUE} the likelihood objects are weighted; see Details for the accepted options. Default is \code{FALSE}.}

 \item{X}{Numeric design matrix for the linear mean \eqn{X\beta}.  Its columns correspond in order to \code{mean}, \code{mean1}, \ldots.  For fixed-location space-time data, rows follow \code{c(t(data))}: all sites at the first time, then all sites at the second time, and so on.  For dynamic locations, use either a stacked matrix in time-wise order or a list with \code{X[[t]]} aligned with \code{coordx_dyn[[t]]}.  See \code{\link{GeoModels-spacetime-ordering}}.}

 \item{spobj}{An object of class \code{sp} or \code{spacetime}. For space-time objects, the current \code{sp2Geo()} conversion uses sequential temporal indices and does not preserve irregular original time spacing; to retain irregular temporal distances, use the explicit-coordinate interface with numeric \code{coordt}.}

 \item{spdata}{Character; the name of the data component in the \code{sp} or \code{spacetime} object.}
\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
}
\details{

For a univariate model the mean is specified as
\deqn{\mu = X\beta.}
The entries of \eqn{\beta} are represented by the parameters \code{mean},
\code{mean1}, and so on.  The columns of \code{X} correspond to these
coefficients in exactly this order.  When \code{X=NULL}, an intercept-only model is used, which is
equivalent to a one-column matrix of ones and the single coefficient
\code{mean}.  Missing starting values for regression coefficients and other
estimated parameters are initialized internally. Explicit entries in \code{start}
always define starting values for parameters that remain estimated, including
\code{likelihood="Full", type="Standard"}; they are never converted to fixed
mean coefficients. A coefficient is fixed only when it is supplied in \code{fixed}.
Supplied coefficient names must be compatible with \code{ncol(X)}. For difficult
non-Gaussian models, user-supplied starting values can still improve numerical
optimization.

A site-specific known mean can instead be supplied as a vector in
\code{fixed$mean}, with one value per observation.  This external mean is
mutually exclusive with \code{X} and with estimated mean coefficients in
\code{start}.  A scalar \code{fixed$mean} is not an external vector: it is the
fixed intercept coefficient.

 \code{GeoFit} provides weighted and stochastic weighted composite-likelihood
 estimation based on pairs for Gaussian and non-Gaussian random fields,
 including nearest-neighbor and stochastic nearest-neighbor pairwise likelihoods;
 see Caamaño-Carrillo et al. (2024) and Bevilacqua et al. (2026).
 It also provides independence composite-likelihood estimation. The accepted
likelihood/type combinations are checked explicitly: \code{likelihood="Full"}
uses the full-likelihood types (including \code{"Standard"}).  Full-likelihood
objectives are currently implemented for Gaussian, SinhAsinh, LogGaussian,
Tukeyh, Tukeyh2, and the internally supported misspecified-Gaussian full
likelihoods. Other margins, including Gamma and Weibull, are rejected before
optimization with an explicit capability message;
\code{likelihood="Marginal"} is used with \code{"Pairwise"} or
\code{"Independence"}; and \code{likelihood="Conditional"} is used with
\code{"Pairwise"}. The historical Difference composite likelihood is no longer
supported. For space-time pairwise fitting, GeoFit checks that a registered
space-time native kernel exists for the requested marginal model and stops
before optimization when it does not. Bivariate pairwise fitting is currently
implemented only for the Gaussian model with marginal pairwise likelihood.

For pairwise copula fitting, the current implementation is univariate and purely
spatial.  Continuous margins supported with the Gaussian, Clayton-like, and
skew-Gaussian copulas are \code{"Gaussian"}, \code{"StudentT"},
\code{"LogGaussian"}, \code{"Gamma"}, \code{"Weibull"},
\code{"Beta"}, \code{"Beta2"}, \code{"Kumaraswamy"},
\code{"Kumaraswamy2"}, \code{"Logistic"}, and \code{"SkewLaplace"}.
The Gaussian and skew-Gaussian copulas additionally support the discrete margins
\code{"Poisson"}, \code{"Binomial"}, and \code{"BinomialNeg"}. For the
skew-Gaussian copula these probabilities are evaluated as copula-rectangle
probabilities; when \code{nu=0} the calculation reduces exactly to the Gaussian
copula likelihood.
For \code{copula="Clayton"}, \code{nu} is the positive-integer parameter of
the constructive Clayton-like random field and must be supplied in
\code{fixed}; it is not a continuously estimable copula parameter.  For
\code{copula="SkewGaussian"}, \code{nu} is the bounded asymmetry parameter
\eqn{\eta\in(-1,1)} and invalid optimizer proposals are rejected by the
objective function.

 The optimization method is specified using \code{optimizer}. The default method is \code{Nelder-Mead}; other available methods are
 \code{nlm}, \code{BFGS}, \code{SANN}, \code{L-BFGS-B}, \code{bobyqa}, and \code{nlminb}. In the last three cases,
 bounds can be specified using \code{lower} and \code{upper}.

 Depending on the dimension of \code{data} and on the name of the correlation model,
 the observations are assumed to be a realization of a spatial, spatio-temporal or bivariate random field.
 Specifically, with \code{data}, \code{coordx}, \code{coordy}, \code{coordt}:
 \itemize{
 \item If \code{data} is a numeric \eqn{d}{d}-dimensional vector and \code{coordx}, \code{coordy} are two numeric \eqn{d}{d}-dimensional vectors
 (or \code{coordx} is a (\eqn{d \times 2}{d x 2})-matrix and \code{coordy=NULL}), then the data are interpreted as a single spatial realisation observed on \eqn{d}{d} spatial sites;
 \item If \code{data} is a numeric (\eqn{t \times d}{t x d})-matrix and \code{coordt} is a numeric \eqn{t}{t}-dimensional vector, then the data are interpreted as a single spatio-temporal realisation observed on \code{d} sites and \code{t} times;
 \item If \code{data} is a numeric (\eqn{2 \times d}{2 x d})-matrix, then the data are interpreted as a single bivariate spatial realisation observed on \code{d} spatial sites;
 \item If \code{data} is a list, \code{coordx_dyn} is a list and \code{coordt} is a numeric \eqn{t}{t}-dimensional vector, then the data are interpreted as a spatio-temporal realisation observed on dynamical spatial sites (different locations for each time) and for \code{t} times.
 }

 It is also possible to specify a matrix of covariates using \code{X}. Specifically:
 \itemize{
 \item In the spatial case, \code{X} must be a (\eqn{d \times k}{d x k}) matrix associated to \code{data} (a \eqn{d}{d}-vector);
 \item In the spatio-temporal case, \code{X} must be a (\eqn{N \times k}{N x k}) matrix associated to \code{data} (a \eqn{t \times d}{t x d}-matrix), where \eqn{N=t\times d};
 \item In the bivariate case, \code{X} must be a (\eqn{N \times k}{N x k}) matrix associated to \code{data} (a \eqn{2 \times d}{2 x d}-matrix), where \eqn{N=2\times d}.
 }

 The \code{distance} parameter allows different kinds of spatial distances:
 \enumerate{
 \item \code{Eucl}, Euclidean distance (default);
 \item \code{Chor}, chordal distance;
 \item \code{Geod}, geodesic distance.
 }

 The \code{likelihood} parameter represents the composite-likelihood configuration:
 \enumerate{
 \item \code{Conditional}, composite likelihood formed by conditionals;
 \item \code{Marginal}, composite likelihood formed by marginals (default);
 \item \code{Full}, standard likelihood.
 }

 It must be coupled with \code{type}:
 \enumerate{
 \item \code{Pairwise}, composite likelihood based on pairs;
 \item \code{Independence}, composite likelihood based on independence;
 \item \code{Standard}, standard likelihood.
 }

 Observation coordinates are checked for exact duplicates before fitting. For space-time data, the full space-time point must be unique; the same spatial site observed at different times is allowed. The check is hash-based and is performed once at the user-facing fit, not during bootstrap/refit iterations that reuse an already validated design. If stochastic thinning retains no pairwise contributions, fitting stops with an explicit error rather than optimizing an empty criterion.

For \code{model="PoissonGamma"} and \code{model="PoissonGammaZIP"}, pairwise fitting accepts every finite \code{shape>0}. This is the continuous Kibble--Gamma pairwise extension used by the marginal moments, correlation function, and bivariate probabilities. The constructive random-field simulation based on a finite sum of squared Gaussian fields is more restrictive and requires \eqn{2\,shape} to be a positive integer; \code{GeoSim} enforces that simulation constraint explicitly.

Stochastic thinning of nearest-neighbor pairs can be enabled via \code{p_neighb<1}. The argument \code{thin_method}
 controls the thinning scheme (default \code{"bernoulli"}).
}
\section{Spatio-temporal ordering}{
For fixed spatial locations, \code{data} is a \eqn{T \times N}{T x N}
matrix: row \eqn{t} corresponds to \code{coordt[t]} and column \eqn{i} to
row \eqn{i} of \code{coordx}.  Observation-level quantities are ordered as
\code{c(t(data))}, hence \code{X} and a vector \code{fixed$mean} use the
order time then site.

For dynamic locations, \code{coordx_dyn}, \code{data}, and optionally a
list-valued \code{X} have one aligned element per time.  Element \eqn{t}
contains the coordinates, responses, and covariate rows observed at
\code{coordt[t]}; internal concatenation is by increasing list index.
See \code{\link{GeoModels-spacetime-ordering}} for the complete convention.
}

\value{
 Returns an object of class \code{GeoFit}.
 An object of class \code{GeoFit} is a list containing at most the following components:
 \item{bivariate}{Logical: \code{TRUE} if the random field is bivariate, otherwise \code{FALSE}.}
 \item{clic}{The composite information criterion after a \code{GeoVarest} call; if the full likelihood is considered then it coincides with AIC.}
 \item{coordx}{A \eqn{d}{d}-dimensional vector of spatial coordinates.}
 \item{coordy}{A \eqn{d}{d}-dimensional vector of spatial coordinates.}
 \item{coordt}{A \eqn{t}{t}-dimensional vector of temporal coordinates.}
 \item{coordx_dyn}{A list of dynamical (in time) spatial coordinates.}
 \item{conf.int}{Confidence intervals for standard maximum likelihood estimation.}
 \item{convergence}{A string that denotes if convergence is reached.}
 \item{copula}{The type of copula.}
 \item{corrmodel}{The correlation model.}
 \item{data}{The vector/matrix/array (or list) of data.}
 \item{distance}{The type of spatial distance.}
 \item{fixed}{A list of fixed parameters.}
 \item{iterations}{The number of iterations used by the numerical routine.}
 \item{likelihood}{The configuration of the composite likelihood.}
 \item{logCompLik}{The value of the log composite-likelihood at the maximum.}
 \item{maxdist}{The maximum spatial distance used in the weighted composite likelihood (or \code{NULL}).}
 \item{maxtime}{The maximum temporal-distance threshold used in the composite likelihood, expressed in the same units as \code{coordt}.}
 \item{message}{Extra message passed from the numerical routines.}
 \item{model}{The density associated to the likelihood objects.}
 \item{estimation_model}{The inferential model/working likelihood supplied to
 \code{GeoFit}. For misspecified fits this differs from \code{model}, which
 records the data-generating response model used by simulation, diagnostics,
 and prediction routines.}
 \item{missp}{\code{TRUE} if a misspecified Gaussian model is used in the composite likelihood.}
 \item{n}{The Binomial number of trials, or the common Negative-Binomial number \eqn{r} of successes.}
 \item{neighb}{The order of spatial neighborhood in the composite likelihood computation.}
 \item{ns}{The number of (different) location sites in the bivariate case.}
 \item{numcoord}{The number of spatial coordinates.}
 \item{numtime}{The number of temporal realisations.}
 \item{param}{A list of parameter estimates.}
 \item{radius}{The radius of the sphere in the case of great-circle distance.}
 \item{stderr}{Standard errors for standard maximum likelihood estimation.}
 \item{sensmat}{The sensitivity matrix.}
 \item{varcov}{The variance-covariance matrix of the estimates.}
 \item{type}{The type of likelihood objects.}
 \item{X}{The matrix of covariates.}
}

\references{

 \bold{General Composite-likelihood}:

 Varin, C., Reid, N. and Firth, D. (2011). An overview of composite
 likelihood methods. \emph{Statistica Sinica}, \bold{21}, 5--42.

 Varin, C. and Vidoni, P. (2005). A note on composite likelihood inference
 and model selection. \emph{Biometrika}, \bold{92}, 519--528.

 \bold{Non-Gaussian random fields}:

 Alegría, A., Caro, S., Bevilacqua, M., Porcu, E. and Clarke, J. (2017).
 Estimating covariance functions of multivariate skew-Gaussian random fields
 on the sphere. \emph{Spatial Statistics}, \bold{22}, 388--402.
 \doi{10.1016/j.spasta.2017.05.003}

 Alegría, A., Bevilacqua, M. and Porcu, E. (2016). Likelihood-based inference
 for multivariate space-time wrapped-Gaussian fields.
 \emph{Journal of Statistical Computation and Simulation},
 \bold{86}(13), 2583--2597.

 Bevilacqua, M., Caamaño-Carrillo, C. and Gaetan, C. (2020). On modelling
 positive continuous data with spatio-temporal dependence.
 \emph{Environmetrics}, \bold{31}(7), e2628. \doi{10.1002/env.2628}

 Bevilacqua, M., Caamaño-Carrillo, C., Arellano-Valle, R. B. and
 Morales-Oñate, V. (2021). Non-Gaussian geostatistical modeling using
 (skew) t processes. \emph{Scandinavian Journal of Statistics},
 \bold{48}(1), 212--245. \doi{10.1111/sjos.12447}

 Blasi, F., Caamaño-Carrillo, C., Bevilacqua, M. and Furrer, R. (2022).
 A selective view of climatological data and likelihood estimation.
 \emph{Spatial Statistics}, \bold{50}, 100596.
 \doi{10.1016/j.spasta.2022.100596}

 Bevilacqua, M., Caamaño-Carrillo, C., Arellano-Valle, R. B. and
 Gómez, C. (2022). A class of random fields with two-piece marginal
 distributions for modeling point-referenced data with spatial outliers.
 \emph{TEST}, \bold{31}(3), 644--674. \doi{10.1007/s11749-021-00797-5}

 Morales-Navarrete, D., Bevilacqua, M., Caamaño-Carrillo, C. and
 Castro, L. M. (2024). Modelling point referenced spatial count data:
 A Poisson process approach. \emph{Journal of the American Statistical
 Association}, \bold{119}(545), 664--677.
 \doi{10.1080/01621459.2022.2140053}

 Bevilacqua, M., Alvarado, E. and Caamaño-Carrillo, C. (2024).
 A flexible Clayton-like spatial copula with application to bounded support
 data. \emph{Journal of Multivariate Analysis}, \bold{201}, 105277.
 \doi{10.1016/j.jmva.2023.105277}

 \bold{Weighted composite-likelihood for (non-)Gaussian random fields}:

 Bevilacqua, M., Gaetan, C., Mateu, J. and Porcu, E. (2012).
 Estimating space and space-time covariance functions for large data sets:
 a weighted composite likelihood approach.
 \emph{Journal of the American Statistical Association, Theory and Methods},
 \bold{107}, 268--280. \doi{10.1080/01621459.2011.646928}

 Bevilacqua, M. and Gaetan, C. (2015). Comparing composite likelihood methods
 based on pairs for spatial Gaussian random fields.
 \emph{Statistics and Computing}, \bold{25}(5), 877--892.
 \doi{10.1007/s11222-014-9471-3}

 Caamaño-Carrillo, C., Bevilacqua, M., López, C. and Morales-Oñate, V. (2024).
 Nearest neighbours weighted composite likelihood based on pairs for
 (non-)Gaussian massive spatial data with an application to Tukey-hh random
 fields estimation. \emph{Computational Statistics and Data Analysis},
 \bold{191}, 107887. \doi{10.1016/j.csda.2023.107887}

 Bevilacqua, M., Cuevas-Pacheco, F. and Caamaño-Carrillo, C. (2026).
Fast stochastic nearest neighbor pairwise composite likelihood for massive
spatial datasets. \emph{arXiv preprint}, arXiv:2607.06142.
\doi{10.48550/arXiv.2607.06142}

}


\seealso{
\code{\link{GeoCovmatrix}} for covariance matrix construction,
  \code{\link{GeoSim}} for simulation,
  \code{\link{GeoKrig}} for prediction,
  \code{\link{GeoVariogram}} and \code{\link{GeoWLS}} for variogram-based tools,
  \code{\link{GeoVarest}} for bootstrap variance estimation.
}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}
 
\examples{
library(GeoModels)

###############################################################
############ Examples of spatial Gaussian random fields ################
###############################################################


# Define the spatial-coordinates of the points:
set.seed(3)
N=300 # number of location sites
x <- runif(N, 0, 1)
y <- runif(N, 0, 1)
coords <- cbind(x,y)

# Define spatial matrix covariates and regression parameters
X=cbind(rep(1,N),runif(N))
mean <- 0.2
mean1 <- -0.5

# Set the covariance model's parameters:
corrmodel <- "Matern"
sill <- 1
nugget <- 0
scale <- 0.2/3
smooth=0.5


param<-list(mean=mean,mean1=mean1,sill=sill,nugget=nugget,scale=scale,smooth=smooth)

# Simulation of the spatial Gaussian random field:
data <- GeoSim(coordx=coords,corrmodel=corrmodel, param=param,X=X)$data



################################################################
###
### Example 0. Maximum independence composite likelihood fitting of
### a Gaussian random field (no dependence parameters)
### 
###############################################################
# setting starting parameters to be estimated
start<-list(mean=mean,mean1=mean1,sill=sill)

fit1 <- GeoFit(data=data,coordx=coords,likelihood="Marginal",
 type="Independence", start=start,X=X)
print(fit1)


################################################################
###
### Example 1. Maximum conditional pairwise likelihood fitting of
### a Gaussian random field using Nelder-Mead
### 
###############################################################
# setting fixed and starting parameters to be estimated
fixed<-list(nugget=nugget,smooth=smooth)
start<-list(mean=mean,mean1=mean1,scale=scale,sill=sill)

fit1 <- GeoFit(data=data,coordx=coords,corrmodel=corrmodel, 
 neighb=3,likelihood="Conditional",optimizer="Nelder-Mead",
 type="Pairwise", start=start,fixed=fixed,X=X)
print(fit1)

################################################################
###
### Example 2. Maximum stochastic marginal pairwise likelihood 
### fitting of a Gaussian random field using Nelder-Mead
### 
###############################################################

#N=100000 # number of location sites
#x <- runif(N, 0, 1)
#y <- runif(N, 0, 1)
#coords <- cbind(x,y)
#X=cbind(rep(1,N),runif(N))
#data <- GeoSimapprox(coordx=coords,method="TB",L=20000,parallel=TRUE,
#	  corrmodel=corrmodel, param=param,X=X)$data
#fixed<-list(nugget=nugget)
#start<-list(mean=mean,mean1=mean1,scale=scale,sill=sill,smooth=smooth)

#fit2 <- GeoFit(data=data,coordx=coords,corrmodel=corrmodel,
# neighb=3,likelihood="Marginal",optimizer="Nelder-Mead",
# p_neighb=0.2, thin_method="bernoulli",
# type="Pairwise", start=start,fixed=fixed,X=X)
#print(fit2)

################################################################
###
### Example 3. Standard Maximum likelihood fitting of
### a Gaussian random field using nlminb
###
###############################################################
# Define the spatial-coordinates of the points:
set.seed(3)
N=250 # number of location sites
x <- runif(N, 0, 1)
y <- runif(N, 0, 1)
coords <- cbind(x,y)

param<-list(mean=mean,sill=sill,nugget=nugget,scale=scale,smooth=smooth)

data <- GeoSim(coordx=coords,corrmodel=corrmodel, param=param)$data

# setting fixed and parameters to be estimated
fixed<-list(nugget=nugget,smooth=smooth)
start<-list(mean=mean,scale=scale,sill=sill)

I=Inf
lower<-list(mean=-I,scale=0,sill=0)
upper<-list(mean=I,scale=I,sill=I)
fit2 <- GeoFit(data=data,coordx=coords,corrmodel=corrmodel,
 optimizer="nlminb",upper=upper,lower=lower,
 likelihood="Full",type="Standard", 
 start=start,fixed=fixed)
print(fit2)


###############################################################
############ Examples of spatial non-Gaussian random fields #############
###############################################################


################################################################
###
### Example 4. Maximum pairwise likelihood fitting of a Weibull random field 
### with Generalized Wendland correlation with Nelder-Mead
### 
###############################################################
set.seed(524)
# Define the spatial-coordinates of the points:
N=300
x <- runif(N, 0, 1)
y <- runif(N, 0, 1)
coords <- cbind(x,y)
X=cbind(rep(1,N),runif(N))
mean=1; mean1=2 # regression parameters
nugget=0
shape=2
scale=0.2
smooth=0

model="Weibull"
corrmodel="GenWend"
param=list(mean=mean,mean1=mean1,scale=scale,
 shape=shape,nugget=nugget,power2=4,smooth=smooth)
# Simulation of a non-stationary Weibull random field:
data <- GeoSim(coordx=coords, corrmodel=corrmodel,model=model,X=X,
 param=param)$data


fixed<-list(nugget=nugget,power2=4,smooth=smooth)
start<-list(mean=mean,mean1=mean1,scale=scale,shape=shape)

# Maximum independence likelihood:
fit <- GeoFit(data=data, coordx=coords, X=X,
 likelihood="Marginal", type="Independence", corrmodel=corrmodel,
 model=model, start=start, fixed=fixed)
print(unlist(fit$param))

## estimating dependence parameter fixing vector mean parameter
Xb <- as.numeric(X \%*\% unlist(fit$param)[1:2])
fixed<-list(nugget=nugget,power2=4,smooth=smooth,mean=Xb)
start<-list(scale=scale,shape=shape)

# Maximum conditional composite-likelihood fitting of the random fields:
fit1 <- GeoFit(data=data,coordx=coords,corrmodel=corrmodel, model=model,
 neighb=3,likelihood="Conditional",type="Pairwise",
 optimizer="Nelder-Mead",
 start=start,fixed=fixed)
print(unlist(fit1$param))



### joint estimation of the dependence parameter and mean parameters
fixed<-list(nugget=nugget,power2=4,smooth=smooth)
start<-list(mean=mean,mean1=mean1,scale=scale,shape=shape)
fit2 <- GeoFit(data=data,coordx=coords,corrmodel=corrmodel, model=model,
 neighb=3,likelihood="Conditional",type="Pairwise",X=X,
 optimizer="Nelder-Mead",
 start=start,fixed=fixed)
print(unlist(fit2$param))



################################################################
###
### Example 5. Maximum pairwise likelihood fitting of
### a Skew-Gaussian spatial random fields with Wendland correlation
###
###############################################################
set.seed(261)
model="SkewGaussian"
# Define the spatial-coordinates of the points:
x <- runif(500, 0, 1)
y <- runif(500, 0, 1)
coords <- cbind(x,y)

corrmodel="Wend0"
mean=0;nugget=0
sill=1
skew=-4.5
power2=4
c_supp=0.2

# model parameters
param=list(power2=power2,skew=skew,
 mean=mean,sill=sill,scale=c_supp,nugget=nugget)
data <- GeoSim(coordx=coords, corrmodel=corrmodel,model=model, param=param)$data

plot(density(data))
fixed=list(power2=power2,nugget=nugget)
start=list(scale=c_supp,skew=skew,mean=mean,sill=sill)
lower=list(scale=0,skew=-I,mean=-I,sill=0)
upper=list(scale=I,skew=I,mean=I,sill=I)
# Maximum marginal pairwise likelihood:
fit1 <- GeoFit(data=data,coordx=coords,corrmodel=corrmodel, model=model,
 neighb=3,likelihood="Marginal",type="Pairwise",
 optimizer="bobyqa",lower=lower,upper=upper,
 start=start,fixed=fixed)
print(unlist(fit1$param))


################################################################
###
### Example 6. Maximum pairwise likelihood fitting of 
### a Bernoulli random field with exponential correlation 
###
###############################################################

set.seed(422)
N=250
x <- runif(N, 0, 1)
y <- runif(N, 0, 1)
coords <- cbind(x,y)
mean=0.1; mean1=0.8; mean2=-0.5 # regression parameters
X=cbind(rep(1,N),runif(N),runif(N)) # matrix covariates
corrmodel <- "Wend0"
param=list(mean=mean,mean1=mean1,mean2=mean2,nugget=0,scale=0.2,power2=4)
# Simulation of the spatial Binomial-Gaussian random field:
data <- GeoSim(coordx=coords, corrmodel=corrmodel, model="Binomial", n=1,X=X,
 param=param)$data


## estimating the marginal parameters using independence cl
fixed <- list(power2=4,scale=0.2,nugget=0)
start <- list(mean=mean,mean1=mean1,mean2=mean2)

# Maximum independence likelihood:
fit <- GeoFit(data=data, coordx=coords, n=1, X=X,
 likelihood="Marginal", type="Independence", corrmodel=corrmodel,
 model="Binomial", start=start, fixed=fixed)
 
print(fit)


## estimating dependence parameter fixing vector mean parameter
Xb <- as.numeric(X \%*\% unlist(fit$param))
fixed <- list(nugget=0,power2=4,mean=Xb)
start <- list(scale=0.2)
lower <- list(scale=0)
upper <- list(scale=2)
# Maximum Marginal pairwise likelihood:
fit1 <- GeoFit(data=data, coordx=coords, corrmodel=corrmodel, n=1,
 likelihood="Marginal", type="Pairwise", neighb=3,
 model="Binomial", start=start, fixed=fixed,
 lower=list(scale=0.001), upper=list(scale=1))
 
print(fit1)


## estimating jointly marginal and dependence parameters
fixed <- list(nugget=0,power2=4)
start <- list(mean=mean,mean1=mean1,mean2=mean2,scale=0.2)

# Maximum conditional pairwise likelihood:
fit2 <- GeoFit(data=data, coordx=coords, corrmodel=corrmodel, n=1, X=X,
 likelihood="Marginal", type="Pairwise", neighb=3,
 model="Binomial", start=start, fixed=fixed)
 
print(fit2)


###############################################################
######### Examples of Gaussian spatio-temporal random fields ###########
###############################################################
set.seed(52)
# Define the temporal sequence:
time <- seq(1, 9, 1)

# Define the spatial-coordinates of the points:
x <- runif(20, 0, 1)
y <- runif(20, 0, 1)
coords=cbind(x,y)

# Set the covariance model's parameters:
scale_s=0.2/3;scale_t=1
smooth_s=0.5;smooth_t=0.5
sill=1
nugget=0
mean=0

param<-list(mean=0,scale_s=scale_s,scale_t=scale_t,
 smooth_t=smooth_t, smooth_s=smooth_s ,sill=sill,nugget=nugget)

# Simulation of the spatio-temporal Gaussian random field:
data <- GeoSim(coordx=coords,coordt=time,corrmodel="Matern_Matern",
 param=param)$data

################################################################
###
### Example 7. Maximum pairwise likelihood fitting of a
### space time Gaussian random fields with double-exponential correlation
###
###############################################################
# Fixed parameters
fixed<-list(nugget=nugget,smooth_s=smooth_s,smooth_t=smooth_t)
# Starting value for the estimated parameters
start<-list(mean=mean,scale_s=scale_s,scale_t=scale_t,sill=sill)

# Maximum composite-likelihood fitting of the random fields:
fit <- GeoFit(data=data,coordx=coords,coordt=time,
 corrmodel="Matern_Matern",maxtime=1,neighb=3,
 likelihood="Marginal",type="Pairwise",
 start=start,fixed=fixed)
print(fit)



###############################################################
######### Examples of a bivariate Gaussian random field ###########
###############################################################

################################################################
### Example 8. Maximum pairwise likelihood fitting of a
### bivariate Gaussian random fields with separable Bivariate matern 
### (cross) correlation model 
###############################################################

# Define the spatial-coordinates of the points:
set.seed(89)
x <- runif(300, 0, 1)
y <- runif(300, 0, 1)
coords=cbind(x,y)
# parameters
param=list(mean_1=0,mean_2=0,scale=0.1,smooth=0.5,sill_1=1,sill_2=1,
 nugget_1=0,nugget_2=0,pcol=0.2)

# Simulation of a spatial bivariate Gaussian random field:
data <- GeoSim(coordx=coords, corrmodel="Bi_Matern_sep", 
 param=param)$data

# selecting fixed and estimated parameters
fixed=list(mean_1=0,mean_2=0,nugget_1=0,nugget_2=0,smooth=0.5)
start=list(sill_1=var(data[1,]),sill_2=var(data[2,]),
 scale=0.1,pcol=cor(data[1,],data[2,]))


# Maximum marginal pairwise likelihood
fitcl<- GeoFit(data=data, coordx=coords, corrmodel="Bi_Matern_sep",
 likelihood="Marginal",type="Pairwise",
 start=start,fixed=fixed,
 neighb=3)
print(fitcl)

}
\keyword{Composite}

\name{GeoCorrFct_Cop}
\alias{GeoCorrFct_Cop}
\encoding{UTF-8}
\title{Correlation, Covariance, and Variogram Functions for Spatial Copula Models}
\description{
Computes the observed-scale correlation, covariance, or semivariogram of a
univariate spatial or spatio-temporal random field constructed with a Gaussian,
Clayton-like, or skew-Gaussian copula. The function uses the same numerical
second-order covariance engines as \code{GeoCovmatrix()}, \code{GeoCovariogram()},
and \code{GeoKrig()}.
}
\usage{
GeoCorrFct_Cop(x, t = NULL, corrmodel,
model = "Gaussian", copula = "Gaussian",
distance = "Eucl", param, radius = 6371,
n = 1, covariance = FALSE, variogram = FALSE)
}

\arguments{
 \item{x}{Numeric vector of non-negative spatial distances.}
 \item{t}{Optional numeric vector of non-negative temporal distances for a
 spatio-temporal correlation model.}
 \item{corrmodel}{String giving the latent Gaussian correlation model. See
 \code{\link{GeoCovmatrix}} for implemented correlation models.}
 \item{model}{String giving the marginal distribution. Copula covariance is
 implemented for \code{Gaussian}, \code{StudentT}, \code{LogGaussian},
 \code{Gamma}, \code{Weibull}, \code{Beta}, \code{Beta2},
 \code{Kumaraswamy}, \code{Kumaraswamy2}, \code{Logistic},
 \code{SkewLaplace}, \code{Tukeyh}, \code{Tukeyh2}, and
 \code{SinhAsinh}.}
 \item{copula}{String giving the copula. Supported values are
 \code{Gaussian}, \code{Clayton}, and \code{SkewGaussian}. For the
 Clayton-like copula, \code{param$nu} must be a positive integer. For the
 skew-Gaussian copula, \code{param$nu} is the asymmetry parameter and must
 belong to \eqn{(-1,1)}.}
 \item{distance}{String giving the spatial distance. The default is
 \code{Eucl}. The supplied \code{x} values are already distances; this
 argument is retained for consistency with the other GeoModels functions.}
 \item{param}{List of correlation, marginal, nugget, and copula parameters.
 Since this is a lag-only function and has no design matrix, location-dependent
 margins use an intercept-only marginal predictor \code{mean}.}
 \item{radius}{Numeric radius of the sphere when great-circle distances are
 used. The default is 6371 km.}
 \item{n}{Numeric number of trials, retained for interface compatibility.}
 \item{covariance}{Logical. If \code{TRUE} and \code{variogram=FALSE}, return
 observed-scale covariance. If \code{FALSE}, return observed-scale
 correlation.}
 \item{variogram}{Logical. If \code{TRUE}, return a semivariogram. With
 \code{covariance=TRUE}, this is the observed-scale semivariogram
 \eqn{V_Y-C_Y(h)}. With \code{covariance=FALSE}, it is the normalized
 semivariogram \eqn{1-\rho_Y(h)}.}
}

\details{
The spatial dependence supplied by \code{corrmodel} is interpreted as the
correlation \eqn{\rho(h)} of the latent Gaussian random field underlying the
copula. The marginal covariance is then obtained on the observed scale using
the selected copula and marginal quantile transformation.

For Gaussian copulas, the implementation uses the Gaussian-copula covariance
engine. The \code{Tukeyh}, \code{Tukeyh2}, and \code{SinhAsinh} margins are
monotone transforms of one Gaussian field, so their Gaussian-copula covariance
is evaluated with the corresponding exact transformed-Gaussian covariance rather
than a truncated Hermite expansion. Clayton-like covariances use the cached
deterministic quadrature/interpolation engine, avoiding adaptive two-dimensional
integration at every lag. Skew-Gaussian covariances use the cached bivariate
Hermite representation. Consequently, repeated calls with the same copula and
marginal parameters can reuse cached numerical objects.

The nugget is applied to the latent correlation for non-zero lags. At the exact
zero spatial and temporal lag, the function represents the same random variable
and therefore returns correlation one, the marginal variance for covariance,
and zero for the semivariogram.
}

\value{
An object of class \code{GeoCorrFct}. Its \code{corr} component contains the
requested correlation, covariance, or semivariogram values. The object also
contains the spatial distances, temporal distances, marginal model, copula,
parameters, and flags describing the requested scale.
}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{
library(GeoModels)

################################################################
### Correlation of a mean-reparametrized Beta random field
### with a Matern latent correlation model.
################################################################

x <- seq(0, 0.4, 0.02)

param <- list(smooth = 0.5, scale = 0.2 / 3, nugget = 0,
              mean = 0, min = 0, max = 1, shape = 2)

corr_gauss <- GeoCorrFct_Cop(
    x = x, corrmodel = "Matern", param = param,
    copula = "Gaussian", model = "Beta2"
)
plot(corr_gauss, ylab = "Correlation", lwd = 2)

## Clayton-like and skew-Gaussian copulas are selected through param$nu.
## Their fast covariance engines are also used by GeoCovmatrix and GeoKrig.
}

\name{CompIndLik2}
\alias{CompIndLik2}
\encoding{UTF-8}
\title{Optimizes the Composite indipendence log-likelihood}
\description{
 Subroutine called by GeoFit.
 The procedure estimates the model parameters by maximisation of the indipendence 
 composite log-likelihood.
}
\usage{
CompIndLik2(bivariate, coordx, coordy ,coordz,coordt,
coordx_dyn, data, flagcorr, flagnuis, fixed,grid,
 lower, model, n, namescorr, namesnuis, 
 namesparam,
 numparam, optimizer, onlyvar, 
 param, spacetime, type,
 upper, namesupper, varest, ns, X,
 sensitivity,copula,MM,score)
}
\arguments{
 \item{bivariate}{Logical; if \code{TRUE} then the data come froma a bivariate random field. 
 Otherwise from a univariate random field.}
 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix 
 Coordinates on a sphere for a fixed radius \code{radius} 
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordt}{A numeric vector assigning 1-dimension of
 temporal coordinates. Optional argument, the default is \code{NULL}
 then a spatial random field is expected.}
 \item{coordx_dyn}{A list of \eqn{m} numeric (\eqn{d_t \times 2}{d x 2})-matrices
 containing dynamical (in time) spatial coordinates. Optional argument, the default is \code{NULL} }
 \item{data}{A numeric vector or a (\eqn{n \times d}{n x d})-matrix or (\eqn{d \times d
 \times n}{d x d x n})-matrix of observations.}
 \item{flagcorr}{A numeric vector of binary values denoting which
 paramerters of the correlation function will be estimated.}
 \item{flagnuis}{A numeric vector of binary values denoting which
 nuisance paramerters will be estimated.}
 \item{fixed}{A numeric vector of parameters that
 will be considered as known values.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data
 are interpreted as a vector or a (\eqn{n \times d}{n x d})-matrix,
 instead if \code{TRUE} then (\eqn{d \times d
 \times n}{d x d x n})-matrix is considered.}

 \item{lower}{An optional named list giving the values for the lower bound of the space parameter
 when the optimizer is \code{L-BFGS-B} or \code{nlminb} or \code{optimize}. The 
 names of the list must be the same of the names in the \code{start} list.}
 \item{model}{Numeric; the id value of the density associated to the likelihood objects.}
 \item{n}{Numeric; number of trials in a binomial random fields.} 
 \item{namescorr}{String; the names of the correlation parameters.}
 \item{namesnuis}{String; the names of the nuisance parameters.}
 \item{namesparam}{String; the names of the parameters to be maximised.}
 \item{numparam}{Numeric; the number of parameters to be maximised.}
 \item{optimizer}{String; the optimization algorithm
 (see \code{\link{optim}} for details). \code{Nelder-Mead} is the
 default. Other possible choices are \code{nlm}, \code{BFGS} \code{L-BFGS-B} and \code{nlminb}. 
 In these last two cases upper and lower bounds can be passed by the user. In the case
 of one-dimensional optimization, the function \code{optimize} is used.} 
 \item{onlyvar}{Logical; if \code{TRUE} (and varest is \code{TRUE}) only
 the variance covariance matrix is computed
 without optimizing. \code{FALSE} is the default.} 
 \item{param}{A numeric vector of parameters values.}
 \item{spacetime}{Logical; if \code{TRUE} the random field is spatio-temporal otherwise is a spatial field.}
 \item{type}{String; the type of the likelihood objects. If \code{Pairwise} (the
 default) then the marginal composite likelihood is formed by pairwise marginal likelihoods.}
 \item{upper}{An optional named list giving the values for the upper bound 
 of the space parameter when the optimizer is or \code{L-BFGS-B} or \code{nlminb} or \code{optimize}. 
 The names of the list must be the same of the names in the \code{start} list.}
 \item{namesupper}{String; the names of the upper limit of the parameters.}
 \item{varest}{Logical; if \code{TRUE} the estimate variances and standard errors are returned.
 \code{FALSE} is the default.}
 \item{ns}{Numeric; Number of (dynamical) temporal instants. } 
 \item{X}{Numeric; Matrix of space-time covariates in the linear mean specification.}
 \item{sensitivity}{Logical; if \code{TRUE} then the sensitivy matrix is computed}
 \item{copula}{String; the type of copula. It can be "Clayton" or "Gaussian"}
 \item{MM}{Numeric;a non constant fixed mean }
 \item{score}{Logical; should score function be computed?}
}

\value{Return a list from an \code{optim} call.}

\seealso{\code{\link{GeoFit}}}


\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\keyword{Composite}

\name{GeoModels-3D}
\alias{GeoModels-3D}
\title{Three-dimensional spatial coordinates in GeoModels}
\description{
Conventions and current limitations for purely spatial models whose locations
have three Cartesian coordinates.
}
\details{
For an irregular three-dimensional spatial data set, supply either an
\eqn{N \times 3} numeric matrix in \code{coordx}, or three equal-length
vectors through \code{coordx}, \code{coordy}, and \code{coordz}. Use
\code{grid = FALSE} and \code{distance = "Eucl"}.

Euclidean lags use all three coordinates. The principal fitting, covariance,
variogram, Cholesky simulation, neighborhood, and global/local kriging paths
support this representation.

Regular three-dimensional grids specified by three coordinate axes and
\code{grid = TRUE} are not implemented. Supply their Cartesian product as an
explicit \eqn{N \times 3} matrix and use \code{grid = FALSE}.

For purely spatial univariate fields, spectral turning-bands simulation accepts
explicit irregular three-dimensional Euclidean coordinates through
\code{GeoSimapprox(..., method = "TB", grid = FALSE)}. The same TB simulator
can be used by \code{GeoVarest()} and \code{GeoVarestbootstrap()} for such
fits. Three-dimensional TB does not support bivariate or spatio-temporal
models, and does not accept \code{anisopars}; transform coordinates explicitly
when needed. CE remains restricted to regular two-dimensional grids. Use
\code{method = "cholesky"} for unsupported three-dimensional cases.

Distances \code{"Geod"} and \code{"Chor"} require exactly two columns
(longitude and latitude) and cannot be combined with a third coordinate.

For \code{GeoVariogramDir()}, three-dimensional lag lengths use all three
coordinates, while direction classes are horizontal azimuths computed from
the \eqn{x}- and \eqn{y}-coordinate differences. This is not a full
three-dimensional directional analysis with azimuth and elevation.
}
\seealso{
\code{\link{GeoFit}}, \code{\link{GeoCovmatrix}},
\code{\link{GeoVariogram}}, \code{\link{GeoVariogramDir}},
\code{\link{GeoSim}}, \code{\link{GeoSimapprox}},
\code{\link{GeoKrig}}, \code{\link{GeoKrigloc}}
}

\name{CkCorrModel}
\alias{CkCorrModel}
\encoding{UTF-8}
\title{Checking Correlation Model}
\description{
 The procedure controls if the correlation model
 inserted is correct.
}
\usage{
CkCorrModel(corrmodel)
}
\arguments{
 \item{corrmodel}{String; the name of a correlation model, for the
 description see \code{\link{GeoCovmatrix}}.}
}

\details{The procedure controls if the correlation model is correct}


\value{Return a number associated to a given correlation model if the model
is considered in the package. Otherwise return NULL.} 



\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\keyword{Composite}

\name{GeoVariogramDir}
\alias{GeoVariogramDir}
\encoding{UTF-8}
\title{Empirical directional semivariogram}

\description{
 Computes empirical semivariograms in multiple directions (e.g., 0, 45, 90, 135 degrees) to assess
 spatial anisotropy.
}

\usage{
GeoVariogramDir(data, coordx, coordy = NULL, coordz = NULL,
 directions = c(0, 45, 90, 135), tolerance = 22.5, numbins = 13,
 maxdist = NULL, neighb = NULL, distance = "Eucl",
 subsample = 1)
}

\arguments{
 \item{data}{A numeric vector containing the observed values at each location.}
 \item{coordx}{
 Spatial coordinates. Either a numeric vector giving the first coordinate, or a matrix with 2 (or 3) columns.
 If a matrix is provided, \code{coordy} and \code{coordz} are ignored.
 }
 \item{coordy}{
 A numeric vector of the second coordinate. Optional; defaults to \code{NULL}.
 }
 \item{coordz}{
 A numeric vector of the third coordinate (if needed). Optional; defaults to \code{NULL}.
 }
 \item{directions}{
 Numeric vector giving the principal directions (in degrees) for which the semivariogram is computed
 (default: \code{c(0, 45, 90, 135)}).
 }
 \item{tolerance}{
 Angular tolerance (in degrees) around each direction (default: 22.5).
 }
 \item{numbins}{
 Number of distance bins for the empirical semivariogram (default: 13).
 }
 \item{maxdist}{
 Maximum spatial distance to consider between pairs. If \code{NULL}, an internal default is used.
 }
 \item{neighb}{
 Number of nearest neighbors to use for each location. If \code{NULL}, an internal default is used.
 }
 \item{distance}{
 Type of distance metric to use (default: \code{"Eucl"}). See \code{\link{GeoFit}} for options.
 }
 \item{subsample}{
 Numeric in \eqn{(0,1]}{(0,1]}. Proportion of spatial locations used to compute the directional semivariograms
 (useful for large datasets). Default is 1 (use all locations).
 }
}

\details{
 The function computes empirical semivariograms for several directions by:
 \itemize{
 \item Selecting pairs of points within \code{maxdist} and among the \code{neighb} nearest neighbors using
 \code{\link{GeoNeighIndex}}.
 \item Computing squared differences for each selected pair.
 \item Assigning each pair to a directional class if the vector connecting the pair falls within the
 specified angular tolerance around a given direction.
 \item Binning pairs by distance and computing the average squared difference (semivariogram) within each bin.
 }

 The direction is defined in the \emph{xy-plane} even in 3D. For 2D data, set \code{coordz = NULL}.

 The \code{subsample} argument can be used to reduce computational cost by randomly selecting a subset of
 spatial locations prior to pair construction. This is particularly useful for large datasets.
}

\value{
 A list of class \code{"GeoVariogramDir"} with one element for each direction. Each element is a list with components:
 \item{centers}{Centers of the distance bins.}
 \item{gamma}{Empirical semivariogram values for each bin.}
 \item{npairs}{Number of point pairs in each bin.}
}


\section{Three-dimensional coordinates}{
With three-dimensional Euclidean coordinates, lag distances use all three
coordinates, but direction classes are azimuths in the \eqn{xy} plane.
Distances \code{"Geod"} and \code{"Chor"} require exactly two coordinate
columns. The automatic maximum lag includes the extent of the z coordinate.
}
\seealso{
 \code{\link{GeoVariogram}}, \code{\link{GeoNeighIndex}}
}

\examples{
require(GeoModels)
set.seed(960)
NN <- 2500
coords <- cbind(runif(NN), runif(NN))
scale <- 0.5/3
param <- list(mean = 0, sill = 1, nugget = 0, scale = scale, smooth = 0.5)
corrmodel <- "Matern"

set.seed(951)
data <- GeoSim(coordx = coords, corrmodel = corrmodel,
 model = "Gaussian", param = param)$data

vario_dir <- GeoVariogramDir(data = data, coordx = coords, maxdist = 0.4,
 subsample = 0.5)

plot(vario_dir, ylim = c(0,1))
}

\keyword{Variogram}

\name{GeoKrigWeights}
\alias{GeoKrigWeights}
\encoding{UTF-8}
\title{
Compute validated Gaussian kriging weights for spatial and spatio-temporal random fields
}
\description{
Given a set of spatial locations (and possibly temporal instants), the function
returns Gaussian kriging weights and the covariance quantities used to build the
kriging system.  Model-specific non-Gaussian prediction is provided by
\code{\link{GeoKrig}} and \code{\link{GeoKrigloc}} rather than by this low-level
weights helper.
}
\usage{
GeoKrigWeights(
 coordx, coordy = NULL, coordz = NULL, coordt = NULL,
 coordx_dyn = NULL, corrmodel, distance = "Eucl",
 grid = FALSE, loc, method = "cholesky", model = "Gaussian",
 n = 1, nloc = NULL, param, anisopars = NULL, radius = 1,
 sparse = FALSE, time = NULL, which = 1, copula = NULL,
 X = NULL, Xloc = NULL, Mloc=NULL)
}
\arguments{
 \item{coordx}{Numeric (\eqn{d\times 2}{d x 2}) or (\eqn{d\times 3}{d x 3}) matrix
 of spatial coordinates. Coordinates on a sphere are accepted
 (lon/lat in decimal degrees) when \code{distance = "Sphere"}.}
 \item{coordy}{Optional numeric vector giving an additional spatial coordinate
 dimension. Ignored if \code{coordx} is already a matrix.}
 \item{coordz}{Optional numeric vector giving a third spatial coordinate
 dimension.}
 \item{coordt}{Optional numeric vector containing the temporal coordinates of the observations. If missing, a purely spatial random field is assumed. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{List of one two- or three-column observation-coordinate matrix per temporal instant.  Blocks are concatenated by time; rows within each block retain their matrix order.  See \code{\link{GeoModels-spacetime-ordering}}.}
 \item{corrmodel}{Character string naming a valid correlation model.
 See \code{\link{GeoCovmatrix}} for admissible choices.}
 \item{distance}{Character string specifying the spatial distance.
 Default is \code{"Eucl"} (Euclidean). See \code{\link{GeoFit}}.}
 \item{grid}{Logical. If \code{TRUE}, coordinates are interpreted as
 defining a regular grid (see \code{\link{GeoKrig}}).}
 \item{loc}{Numeric (\eqn{n\times 2}{n x 2}) matrix of locations for which
 the kriging weights are required.}
 \item{method}{Character string indicating the matrix factorisation
 used to solve the kriging system: \code{"cholesky"} (default) or \code{"svd"}.}
 \item{model}{Character string.  The validated public implementation currently
 requires \code{"Gaussian"} (the alias \code{"Gauss"} is accepted).}
 \item{n}{Integer. Number of trials for Binomial random fields
 (default \code{1}).}
 \item{nloc}{Integer. Number of trials for the prediction locations
 in Binomial random fields (default \code{1}).}
 \item{param}{Named list of covariance and mean parameters.
 See \code{\link{CorrParam}} and \code{\link{GeoCovmatrix}}.}
 \item{anisopars}{List with components \code{angle} and \code{ratio}
 defining geometric anisotropy (optional).}
 \item{radius}{Positive numeric value: sphere radius when
 coordinates are lon/lat (default \code{1}).}
 \item{sparse}{Logical. If \code{TRUE}, sparse‐matrix algorithms
 (package \pkg{spam}) are employed. Only effective with compactly
 supported covariance functions.}
 \item{time}{Numeric vector giving the temporal instants for which weights are required. Values need not be equally spaced and are interpreted on the same numeric time scale as \code{coordt}. Ignored if \code{coordt} is missing.}
 \item{which}{Integer (\code{1} or \code{2}) selecting the variable whose
 weights are returned in the bivariate case.}
 \item{copula}{Must be \code{NULL}.  Copula conditional-simulation code requests
 these weights on the latent Gaussian scale.}
 \item{X}{Numeric design matrix at observation locations.  Fixed-location observations are ordered time then site; dynamic observations are ordered by the temporal blocks of \code{coordx_dyn}.}
 \item{Xloc}{Numeric design matrix at prediction tasks, ordered location then time: all requested times for the first row of \code{loc}, then all requested times for the second row, and so on.}
 \item{Mloc}{Numeric vector of known prediction means in the same location-major order as \code{Xloc}.  Use either \code{Mloc} or \code{Xloc}, not both.}
}
\details{

Mean inputs follow \code{\link{GeoKrig}}.  Coefficients \code{mean},
\code{mean1}, and so on correspond in order to the columns of \code{X}; an
intercept-only model uses only \code{mean}.  A vector \code{param$mean} is a
known observation mean and is mutually exclusive with \code{X}.  At
prediction locations, use either \code{Xloc} or \code{Mloc}, not both.

The function builds the kriging system
\deqn{\Sigma \mathbf{w} = \boldsymbol{\sigma}_0}{Sigma w = sigma_0}
where \eqn{\Sigma}{Sigma} is the covariance matrix between observed
locations and \eqn{\boldsymbol{\sigma}_0}{sigma_0} the vector of
covariances between observed and prediction locations. 
The solution \eqn{\mathbf{w}}{w} is returned together with \eqn{\Sigma}{Sigma}
and, optionally, \eqn{\Sigma^{-1}}{Sigma^-1}.
The returned weights solve the Gaussian covariance system only.  Mean inputs are
validated for ordering compatibility, but no non-Gaussian covariance transform is
performed here.  No actual prediction is carried out; for full Gaussian or
non-Gaussian prediction use \code{\link{GeoKrig}}.
}
\section{Spatio-temporal ordering}{
Observation-level rows follow time-major order.  For fixed sites this is the
order of \code{c(t(data))}; for dynamic sites it is the row-binding of
\code{coordx_dyn} by list element.  Prediction tasks follow location-major
order, so rows of \code{Xloc} and elements of \code{Mloc} enumerate all
requested times for the first prediction location before moving to the next
location.  See \code{\link{GeoModels-spacetime-ordering}}.
}

\value{
A list containing:
 \item{weights}{Numeric matrix of Gaussian kriging weights.  The validated
 orientation is observations by prediction tasks, as recorded in
 \code{weights_orientation}.}
 \item{weights_orientation}{Character string describing the weight-matrix
 orientation.}
 \item{covmatrix}{Observed Gaussian covariance matrix.}
 \item{CC}{Observation--prediction Gaussian cross-covariance matrix.}
 \item{model}{The canonical model name (currently \code{"Gaussian"}).}
 \item{corrmodel}{Input correlation model.}
 \item{param}{Parameters used to construct the Gaussian covariance matrix.}
 \item{bivariate}{Logical indicating a bivariate Gaussian correlation model.}
 \item{spacetime}{Logical indicating a space-time correlation model.}
 \item{loc}{Prediction locations.}
 \item{tloc}{Number of requested prediction times.}
 \item{Xloc,Mloc,mean_obs}{Normalized mean inputs retained for alignment and
 conditional-simulation workflows.}
}
\references{
 Gaetan, C. and Guyon, X. (2010) \emph{Spatial Statistics and Modeling}.
 Springer-Verlag, New York.
}
\seealso{
 \code{\link{GeoKrig}} for full prediction,
 \code{\link{GeoKrigloc}} for local prediction,
 \code{\link{GeoCovmatrix}} for covariance model details,
}
\author{
 Moreno Bevilacqua, \email{moreno.bevilacqua@uai.cl}, 
 Víctor Morales-Oñate, \email{victor.morales@uv.cl}, 
 Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl}
}


\examples{
library(GeoModels)
################################################################

################################################################
###
### Example 1. Spatial kriging weights for
### Gaussian random fields with Gen wendland correlation.
###
################################################################

model="Gaussian"
set.seed(79)
x = runif(300, 0, 1)
y = runif(300, 0, 1)
coords=cbind(x,y)

corrmodel = "GenWend"
mean=0; sill=5; nugget=0
scale=0.2;smooth=0;power2=4

param=list(mean=mean,sill=sill,nugget=nugget,scale=scale,smooth=smooth,power2=power2)

# Simulation of the spatial Gaussian random field:
data = GeoSim(coordx=coords, corrmodel=corrmodel,model=model,
 param=param)$data

xx=seq(0,1,0.25)
loc_to_pred=as.matrix(expand.grid(xx,xx))

W=GeoKrigWeights(,coordx=coords,loc=loc_to_pred,corrmodel=corrmodel,
 model=model,param=param)

dim(W$weights) ### kriging weights
}

\keyword{Kriging}

\name{GeoNeighbSelect}
\alias{GeoNeighbSelect}
\encoding{UTF-8}
\title{Neighborhood Selection for Pairwise Composite Likelihood Estimation}
\description{
The procedure fits pairwise composite likelihood models over user-specified spatial
or space-time neighborhoods and selects the candidate minimizing the sum of squared
differences between the fitted and empirical semivariograms. A
\code{GeoVariogram} object must be supplied.
}
\usage{
GeoNeighbSelect(data, coordx, coordy=NULL,coordz=NULL, coordt=NULL, coordx_dyn=NULL,
 copula=NULL,corrmodel=NULL, distance="Eucl",fixed=NULL,anisopars=NULL,
 est.aniso=c(FALSE,FALSE), grid=FALSE, likelihood='Marginal',lower=NULL,
 neighb=c(1,2,3,4,5),p_neighb=1,maxtime=Inf, memdist=TRUE,model='Gaussian',
 n=1, ncores=6,optimizer='Nelder-Mead', parallel=FALSE,
 bivariate=FALSE,radius=1, start=NULL,type='Pairwise', upper=NULL,
 weighted=FALSE,X=NULL,spobj=NULL,spdata=NULL,vario=NULL,progress=TRUE,
 check.duplicates=FALSE)
}
\arguments{
 \item{data}{A \eqn{d}{d}-dimensional vector (a single spatial realisation) or a (\eqn{d \times d}{d x d})-matrix (a single spatial realisation on
 regular grid)
 or a
 (\eqn{t \times d}{t x d})-matrix (a single spatio-temporal realisation) or an (\eqn{d \times d \times t \times n }{d x d x t})-array
 (a single spatio-temporal realisation on regular grid).
 See Details for the accepted data layouts.}
 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix
 Coordinates on a sphere for a fixed radius \code{radius}
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}

 \item{coordt}{A numeric vector assigning one dimension of temporal coordinates. Optional argument; the default is \code{NULL}, in which case a spatial random field is expected. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{A list of \eqn{m} numeric (\eqn{d_t \times 2}{d x 2})-matrices
 containing dynamical (in time) spatial coordinates. Optional argument, the default is \code{NULL}}
 \item{copula}{String; the type of copula. It can be "Clayton" or "Gaussian"}
 \item{corrmodel}{String; the name of a correlation model, for the
 see \code{\link{GeoCovmatrix}} for the list of implemented correlation models.}
 \item{distance}{String; the name of the spatial distance. The default is \code{Eucl}, the Euclidean distance.
 See Details for the accepted options.}
 \item{fixed}{An optional named list giving the values of the parameters that
 will be considered as known values. The listed parameters for a
 given correlation function will be not estimated.}
 \item{anisopars}{ A list of two elements: "angle" and "ratio" i.e. the anisotropy angle and the anisotropy ratio, respectively.}
 \item{est.aniso}{ A bivariate logical vector providing which anisotropic parameters must be estimated.}


 \item{grid}{Logical; if \code{FALSE} (the default) the data
 are interpreted as spatial or spatio-temporal realisations on a set
 of non-equispaced spatial sites (irregular grid).}

 \item{likelihood}{String; the configuration of the composite
 likelihood. \code{Marginal} is the default (see \code{\link{GeoFit}} for details).}

 \item{lower}{An optional named list giving the values for the lower bound of the space parameter
 when the optimizer is \code{L-BFGS-B} or \code{nlminb} or \code{bobyqa} or \code{optimize}. The
 names of the list must be the same of the names in the \code{start} list.}


 \item{neighb}{Numeric; a vector of positive integers indicating the
 order of neighborhood in the weight function of composite likelihood
 (see \code{\link{GeoFit}} for details).}

 \item{p_neighb}{Numeric; a value in \eqn{(0,1]} specifying the expected fraction
of nearest-neighbor pairs retained through stochastic thinning. If equal to 1
(the default), no thinning is applied. Values below 1 use independent Bernoulli
thinning, so the selected neighborhood depends on the thinning realization. Use a
fixed random seed, or \code{p_neighb=1}, when a stable comparison is required.}

 \item{maxtime}{Numeric vector of non-negative maximum temporal-distance thresholds, expressed in the same units as \code{coordt}, considered when selecting the neighborhood configuration.}


 \item{memdist}{Deprecated logical argument retained for backward compatibility. The selected pair structure is always precomputed and reused. Supplying \code{FALSE} produces a warning and is treated as \code{TRUE}.}

 \item{model}{String; the type of random fields and therefore the densities associated to the likelihood
 objects. \code{Gaussian} is the default, see \code{\link{GeoFit}} for details.}

 \item{n}{Numeric; number of trials in a binomial random fields; number of successes in a negative binomial random fields }
 \item{ncores}{Positive integer or \code{NULL}; default \code{6}. With \code{parallel=TRUE}, an explicit integer requests that many workers, capped by detected cores and the number of candidate fits. Set \code{ncores=NULL} for automatic selection, capped at six workers and normally leaving one detected core free.}

 \item{optimizer}{String; the optimization algorithm
 (see \code{\link{optim}} for details). \code{Nelder-Mead} is the
 default. Other possible choices are \code{nlm}, \code{BFGS}, \code{SANN}, \code{L-BFGS-B} and \code{nlminb} and
 \code{bobyqa}.
 In these last three cases upper and lower bounds can be passed by the user. In the case
 of one-dimensional optimization, the function \code{optimize} is used.}

 \item{parallel}{Logical; default \code{FALSE}. Set \code{TRUE} to evaluate candidate fits in parallel using a \code{future} plan.}
 \item{bivariate}{Logical; if \code{TRUE} the bivariate case is considered.}

 \item{radius}{Numeric; the radius of the sphere in the case of lon-lat coordinates.
 Default value is 1.}

 \item{start}{An optional named list with the initial values of the
 parameters that are used by the numerical routines in maximization
 procedure. \code{NULL} is the default (see \code{\link{GeoFit}} for details).}

 \item{type}{String; the type of the likelihood objects. If \code{Pairwise} (the
 default) then the marginal composite likelihood is formed by
 pairwise marginal likelihoods (see \code{\link{GeoFit}} for details).}

 \item{upper}{An optional named list giving the values for the upper bound
 of the space parameter when the optimizer is or \code{L-BFGS-B} or \code{bobyqa} or \code{nlminb} or \code{optimize}.
 The names of the list must be the same of the names in the \code{start} list.}

 \item{weighted}{Logical; if \code{TRUE} the likelihood objects are
 weighted (see \code{\link{GeoFit}} for details). If \code{FALSE} (the
 default) the composite likelihood is not weighted.}

 \item{X}{Numeric; Matrix of spatio(temporal)covariates in the linear mean specification.}

 \item{spobj}{An object of class sp or spacetime}
 \item{spdata}{Character:The name of data in the sp or spacetime object}
 \item{vario}{An object of class \code{GeoVariogram} obtained using
\code{\link{GeoVariogram}}. Its spatial, bivariate, or space-time structure must
match the selected correlation model.}
 \item{progress}{Logic; If TRUE then a progress bar is shown.}
\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
}
\details{
For a spatial univariate model, the criterion compares the fitted and empirical
spatial semivariograms at the empirical bin centers. For a bivariate model, the two
marginal semivariograms and the cross-semivariogram are compared in the order
\eqn{(11,12,22)}. For a space-time model, the spatial and temporal margins and
the complete rectangular space-time semivariogram surface are compared jointly.
Only finite empirical cells contribute to the criterion; a candidate is rejected
when its corresponding fitted values are non-finite or dimensionally incompatible.

The absolute minimum defines the best candidate. The function also reports a more
parsimonious neighborhood using the existing plateau rule: the smallest neighborhood
whose criterion is within 15 percent of the minimum. If every candidate fit fails or
produces a non-finite criterion, the function stops with an error instead of selecting
an arbitrary candidate.
}
\value{
A list with components:
\itemize{
  \item \code{best_neighb}: neighborhood associated with the smallest criterion;
  \item \code{best_maxtime}: selected temporal neighborhood for a space-time model,
  otherwise \code{NULL};
  \item \code{best_estimates}: named parameter estimates for the best candidate;
  \item \code{res}: variogram sum-of-squares criterion for every candidate;
  \item \code{estimates}: matrix of parameter estimates, with columns aligned to
  the names in \code{start};
  \item \code{sugg_neighb}: parsimonious neighborhood selected by the 15 percent
  plateau rule;
  \item \code{sugg_time}: corresponding temporal neighborhood for a space-time
  model, otherwise \code{NULL}.
}
}
\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{

library(GeoModels)


######### spatial case
set.seed(32)
N=500 # number of location sites
x <- runif(N, 0, 1)
y <- runif(N, 0, 1)
coords <- cbind(x,y)
mean <- 0.2
# Set the covariance model's parameters:
corrmodel <- "Matern"
sill <- 1;nugget <- 0
scale <- 0.2/3;smooth=0.5

model="Gaussian"
param<-list(mean=mean,sill=sill,nugget=nugget,scale=scale,smooth=smooth)
# Simulation
data <- GeoSim(coordx=coords,corrmodel=corrmodel, param=param,model=model)$data
I=Inf
fixed<-list(nugget=nugget)
start<-list(mean=mean,scale=scale,smooth=smooth,sill=sill)
lower<-list(mean=-I,scale=0,sill=0,smooth=0)
upper<-list(mean=I,scale=I,sill=I,smooth=I)

vario = GeoVariogram(coordx=coords,data=data,maxdist=0.3,numbins=15)

neighb=c(1,2,3,4) ## trying different neighbs
selK <- GeoNeighbSelect(vario=vario,data=data,coordx=coords,corrmodel=corrmodel,
 model=model,neighb=neighb,
 likelihood="Conditional",type="Pairwise",parallel=FALSE,
 optimizer="nlminb",lower=lower,upper=upper,
 start=start,fixed=fixed)
print(selK$best_neighb) ## selected neighbor




}
\keyword{Composite}

\name{GeoFit2}
\alias{GeoFit2}
\encoding{UTF-8}
\title{Fitting Gaussian and Non-Gaussian Random Fields with Automatic Marginal Initialization}
\description{
 A univariate starting-value wrapper around \code{\link{GeoFit}}. The function optionally
 performs a preliminary fit under spatial independence to estimate available
 marginal parameters, uses those estimates to update the starting values, and
 then calls \code{GeoFit} for the requested final fit. If the preliminary
 independence fit is unavailable or fails, the original starting values are
 retained. Bivariate correlation models are not supported.
}
\usage{
GeoFit2(data, coordx = NULL, coordy = NULL, coordz = NULL, coordt = NULL,
 coordx_dyn = NULL, copula = NULL, corrmodel = NULL,
 distance = "Eucl", fixed = NULL, anisopars = NULL,
 est.aniso = c(FALSE, FALSE), grid = FALSE,
 likelihood = "Marginal", lower = NULL, maxdist = Inf,
 neighb = NULL, p_neighb = 1, maxtime = Inf, memdist = TRUE,
 method = "cholesky", model = "Gaussian", n = 1,
 onlyvar = FALSE, optimizer = "Nelder-Mead", radius = 1,
 score = FALSE, sensitivity = FALSE, sparse = FALSE,
 start = NULL, thin_method = "bernoulli", type = "Pairwise",
 upper = NULL, varest = FALSE, weighted = FALSE, X = NULL,
 spobj = NULL, spdata = NULL, independence_start = TRUE,
 independence_optimizer = "Nelder-Mead",
 warn_independence_failure = TRUE, check.duplicates = FALSE)
}
\arguments{
 \item{data}{A \eqn{d}{d}-dimensional vector (a single spatial realisation) or a (\eqn{d \times d}{d x d})-matrix (a single spatial realisation on regular grid)
 or a (\eqn{t \times d}{t x d})-matrix (a single spatio-temporal realisation) or an (\eqn{d \times d \times t \times n}{d x d x t x n})-array
 (a single spatio-temporal realisation on regular grid).
 See Details for the accepted data layouts.}

 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix.
 Coordinates on a sphere for a fixed radius \code{radius} are passed in lon/lat format expressed in decimal degrees.
 It may be omitted when coordinates are supplied through \code{coordx_dyn} or \code{spobj}.}

 \item{coordy}{A numeric vector giving 1-dimension of spatial coordinates; optional argument, default is \code{NULL}.}

 \item{coordz}{A numeric vector giving 1-dimension of spatial coordinates; optional argument, default is \code{NULL}.}

 \item{coordt}{A numeric vector assigning one dimension of the observation-time coordinates. Optional argument, default is \code{NULL}; if \code{NULL}, a spatial random field is expected. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}

 \item{coordx_dyn}{A list of \eqn{m} numeric (\eqn{d_t \times 2}{d_t x 2})-matrices containing dynamical (in time) spatial coordinates.
 Optional argument, default is \code{NULL}.}

 \item{copula}{String; the type of copula. It can be \code{"Clayton"} or \code{"Gaussian"}.}

 \item{corrmodel}{String; the name of a correlation model; see \code{\link{GeoCovmatrix}} for the list of implemented correlation models.}

 \item{distance}{String; the name of the spatial distance. Default is \code{"Eucl"} (Euclidean distance). See Details for the accepted options.}

 \item{fixed}{An optional named list giving the values of the parameters that will be considered as known values.
 The listed parameters for a given correlation function will not be estimated.}

 \item{anisopars}{A list of two elements: \code{"angle"} and \code{"ratio"}, i.e. the anisotropy angle and the anisotropy ratio, respectively. Geometric anisotropy requires \code{distance="Eucl"}.}

 \item{est.aniso}{A bivariate logical vector providing which anisotropic parameters must be estimated.}

 \item{grid}{Logical; if \code{FALSE} (default) the data are interpreted as spatial or spatio-temporal realisations on a set of non-equispaced spatial sites. If \code{TRUE}, the grid axes and data dimensions are handled by \code{GeoFit} and converted internally to the canonical explicit-coordinate layout before fitting.}

 \item{likelihood}{String; the configuration of the composite likelihood. \code{"Marginal"} is the default; see Details for the accepted options.}

 \item{lower}{An optional named list giving lower bounds for parameters when the optimizer is \code{L-BFGS-B}, \code{nlminb}, \code{bobyqa} or \code{optimize}.
 Names must match parameters that are being estimated.}

 \item{maxdist}{Numeric; an optional positive value indicating the maximum spatial distance considered in the composite likelihood computation.
 See Details for more information.}

 \item{neighb}{Numeric; an optional positive integer indicating the order of neighborhood in the composite likelihood computation.
 See Details for more information.}

 \item{p_neighb}{Numeric scalar in \eqn{(0,1]}. If \code{1} (default), no thinning is applied. For \code{thin_method="bernoulli"}, it is the candidate-pair inclusion probability in the constant-weight design and controls the retained size in expectation. For \code{thin_method="FixedBudget"}, it defines the exact global retained-pair budget \eqn{K=\mathrm{round}(p_{neighb}d)}, where \eqn{d} is the candidate-pair count.}

 \item{maxtime}{Numeric; an optional non-negative maximum temporal-distance threshold, expressed in the same units as \code{coordt}, used to select pairwise composite-likelihood contributions.}

 \item{memdist}{Deprecated logical argument retained for backward compatibility. The selected pair structure is always precomputed and reused during composite-likelihood optimization. Supplying \code{FALSE} produces a warning and is treated as \code{TRUE}.}

 \item{method}{String; the type of matrix decomposition used in the likelihood computation. Default is \code{"cholesky"}.
 Another possible choice is \code{"svd"} (when available).}

 \item{model}{String; the type of RF and therefore the densities associated to the likelihood objects.
 \code{"Gaussian"} is the default; see Details for the accepted options.}

 \item{n}{Positive integer scalar, or one positive integer per observation, for models using a Binomial/Negative-Binomial trial or success count.}

 \item{onlyvar}{Logical; if \code{TRUE} (and \code{varest=TRUE}) only the variance-covariance matrix is computed without optimizing.
 Default is \code{FALSE}.}

 \item{optimizer}{String; the optimization algorithm (see \code{\link{optim}} for details). \code{"Nelder-Mead"} is the default.
 Other possible choices are \code{"nlm"}, \code{"BFGS"}, \code{"SANN"}, \code{"L-BFGS-B"}, \code{"nlminb"}, \code{"bobyqa"}.
 In these last three cases upper and lower bounds can be passed by the user. In the one-dimensional case, \code{optimize} is used.}

 \item{radius}{Numeric; the radius of the sphere in the case of lon-lat coordinates. Default value is \code{1}.}

 \item{score}{Logical; should score function be computed? Default is \code{FALSE}.}

 \item{sensitivity}{Logical; if \code{TRUE} then the sensitivity matrix is computed.}

 \item{sparse}{Logical; if \code{TRUE} then maximum likelihood is computed using sparse matrix algorithms (e.g., \pkg{spam}).
 It should be used with compactly supported covariance models. Default is \code{FALSE}.}

 \item{start}{An optional named list with initial values for parameters used by the numerical routines in the maximization procedure.
 Default is \code{NULL}; omitted parameters receive internal starting values (see Details).}

 \item{thin_method}{String; thinning scheme used when \code{p_neighb < 1}. Default is \code{"bernoulli"} (independent Bernoulli thinning). An alternative thinning scheme is \code{"FixedBudget"}. The legacy name \code{"TargetBalanced"} is accepted as an alias for \code{"FixedBudget"}.}

 \item{type}{String; the type of the likelihood objects. If \code{"Pairwise"} (default) then the marginal composite likelihood is formed by pairwise marginal likelihoods (see Details).}

 \item{upper}{An optional named list giving upper bounds for parameters when the optimizer is \code{L-BFGS-B}, \code{nlminb}, \code{bobyqa} or \code{optimize}.
 Names must match parameters that are being estimated.}

 \item{varest}{Logical; if \code{TRUE} the estimates' variances and standard errors are returned. For composite likelihood estimation it is deprecated.
 Use \code{sensitivity=TRUE} and update the object using \code{GeoVarest}. Default is \code{FALSE}.}

 \item{weighted}{Logical; if \code{TRUE} the likelihood objects are weighted; see Details for the accepted options. Default is \code{FALSE}.}

 \item{X}{Numeric; matrix of spatio(temporal) covariates in the linear mean specification.}

 \item{spobj}{An object of class \code{sp} or \code{spacetime}.}

 \item{spdata}{Character; the name of data in the \code{sp} or \code{spacetime} object.}


 \item{independence_start}{Logical; if \code{TRUE} (default), \code{GeoFit2}
 attempts a preliminary marginal fit under spatial independence before the final
 fit. The preliminary fit is skipped when the requested fit already has
 \code{type="Independence"}. Bivariate correlation models are not supported by
 \code{GeoFit2}; use \code{GeoFit} directly for those models.}

 \item{independence_optimizer}{String; optimizer used only for the preliminary
 independence fit. Default is \code{"Nelder-Mead"}. The optimizer used for the
 final fit is still specified by \code{optimizer}.}

 \item{warn_independence_failure}{Logical; if \code{TRUE} (default), a warning
 is issued when the preliminary independence fit is unavailable or fails. In
 that case the supplied \code{start} values are used unchanged.}
\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
}
\details{
 \code{GeoFit2} does not implement a second fitting engine. It is a wrapper
 around the canonical \code{\link{GeoFit}} function.

 When \code{independence_start=TRUE}, the function first constructs a
 preliminary call to \code{GeoFit} with \code{likelihood="Marginal"} and
 \code{type="Independence"}. The preliminary fit estimates the marginal
 parameters available for the selected model. When \code{start} is supplied,
matching marginal entries and omitted mean coefficients are updated as before.
When \code{start=NULL}, all eligible marginal estimates from the independence
fit are used, while dependence parameters use the conservative automatic
initial values prepared by \code{GeoFit}. Parameters supplied in \code{fixed}
always take precedence.

 The final model is then fitted by calling \code{GeoFit} with the original
 requested likelihood, likelihood-object type, optimizer, bounds, pair-selection
 settings, and the updated starting values. Consequently, the objective function,
 parameter constraints, and returned standard \code{GeoFit} components are the
 same as in a direct call to \code{GeoFit}; only the starting values may differ.

 For Gaussian models, \code{sill} is the total marginal variance and
 \code{nugget} attenuates off-diagonal correlation. Therefore an independence
 estimate may initialize \code{sill}, while a user-supplied \code{nugget} is
 preserved unchanged and is never added to or subtracted from \code{sill}.

 Bivariate correlation models are not supported by \code{GeoFit2}; use
 \code{GeoFit} directly. The preliminary initialization is skipped when
 \code{type="Independence"}. If an independence likelihood is not
 implemented for the selected model, or if the preliminary optimization fails,
 the final fit still proceeds using the supplied starting values or, when
\code{start=NULL}, the internal starting values prepared by \code{GeoFit}. This
behavior can be controlled with \code{warn_independence_failure}.

 Stochastic thinning of nearest-neighbor pairs in the final fit can be enabled
 through \code{p_neighb < 1}; \code{thin_method} specifies the thinning
 scheme.
}
\value{
 Returns an object of class \code{GeoFit}.
 An object of class \code{GeoFit} is a list containing at most the following components:
 \item{bivariate}{Logical: \code{TRUE} if the Gaussian RF is bivariate, otherwise \code{FALSE}.}
 \item{clic}{The composite information criterion after a \code{GeoVarest} call; if the full likelihood is considered then it coincides with AIC.}
 \item{coordx}{A \eqn{d}{d}-dimensional vector of spatial coordinates.}
 \item{coordy}{A \eqn{d}{d}-dimensional vector of spatial coordinates.}
 \item{coordt}{A \eqn{t}{t}-dimensional vector of temporal coordinates.}
 \item{coordx_dyn}{A list of dynamical (in time) spatial coordinates.}
 \item{conf.int}{Confidence intervals for standard maximum likelihood estimation.}
 \item{convergence}{A string that denotes if convergence is reached.}
 \item{copula}{The type of copula.}
 \item{corrmodel}{The correlation model.}
 \item{data}{The vector or matrix or array (or list) of data.}
 \item{distance}{The type of spatial distance.}
 \item{fixed}{A list of fixed parameters.}
 \item{iterations}{The number of iterations used by the numerical routine.}
 \item{likelihood}{The configuration of the composite likelihood.}
 \item{logCompLik}{The value of the log composite-likelihood at the maximum.}
 \item{maxdist}{The maximum spatial distance used in the weighted composite likelihood (or \code{NULL}).}
 \item{maxtime}{The maximum temporal-distance threshold used in the composite likelihood, expressed in the same units as \code{coordt}.}
 \item{message}{Extra message passed from the numerical routines.}
 \item{model}{The density associated to the likelihood objects.}
 \item{missp}{\code{TRUE} if a misspecified Gaussian model is used in the composite likelihood.}
 \item{n}{The number of trials in a binomial RF; the number of successes in a negative binomial random field.}
 \item{neighb}{The order of spatial neighborhood in the composite likelihood computation.}
 \item{ns}{The number of (different) location sites in the bivariate case.}
 \item{numcoord}{The number of spatial coordinates.}
 \item{numtime}{The number of temporal realisations of the random field.}
 \item{param}{A list of parameter estimates.}
 \item{radius}{The radius of the sphere in the case of great-circle distance.}
 \item{stderr}{Standard errors for standard maximum likelihood estimation.}
 \item{sensmat}{The sensitivity matrix.}
 \item{varcov}{The variance-covariance matrix of the estimates.}
 \item{type}{The type of the likelihood objects.}
 \item{X}{The matrix of covariates.}
 \item{start_original}{The named list of starting values supplied by the user.}
 \item{start_used}{The named list of starting values passed to the final
 \code{GeoFit} call after any successful marginal initialization.}
 \item{independence_fit}{The preliminary independence \code{GeoFit} object, or
 \code{NULL} when the preliminary fit was skipped or failed.}
 \item{independence_start_message}{A character string describing why the
 preliminary initialization was unavailable or failed, or \code{NULL} when no
 such message was generated.}

}

\references{
 The methodological references for maximum weighted composite-likelihood
 fitting of Gaussian and non-Gaussian random fields are reported in
 \code{\link{GeoFit}}.
}

\seealso{
\code{\link{GeoCovmatrix}} for covariance matrix construction,
  \code{\link{GeoSim}} for simulation,
  \code{\link{GeoKrig}} for prediction,
  \code{\link{GeoVariogram}} and \code{\link{GeoWLS}} for variogram-based tools,
  \code{\link{GeoVarest}} for bootstrap variance estimation.
}

\author{
Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com}, \url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl}, \url{https://www.researchgate.net/profile/Christian-Caamano}
}
\examples{


library(GeoModels)


###############################################################
############ Examples of spatial Gaussian random fields ################
###############################################################

################################################################
###
### Example 1 : Maximum pairwise conditional likelihood fitting 
### of a Gaussian RF with Matern correlation
###
###############################################################
model="Gaussian"
# Define the spatial-coordinates of the points:
set.seed(3)
N=400 # number of location sites
x <- runif(N, 0, 1)
set.seed(6)
y <- runif(N, 0, 1)
coords <- cbind(x,y)

# Define spatial matrix covariates
X=cbind(rep(1,N),runif(N))

# Set the covariance model's parameters:
corrmodel <- "Matern"
mean <- 0.2
mean1 <- -0.5
sill <- 1
nugget <- 0
scale <- 0.2/3
smooth=0.5
param<-list(mean=mean,mean1=mean1,sill=sill,nugget=nugget,scale=scale,smooth=smooth)

# Simulation of the spatial Gaussian RF:
data <- GeoSim(coordx=coords,model=model,corrmodel=corrmodel, param=param,X=X)$data

fixed<-list(nugget=nugget,smooth=smooth)
start<-list(mean=mean,mean1=mean1,scale=scale,sill=sill)

################################################################
###
### Maximum pairwise likelihood fitting of
### Gaussian random fields with exponential correlation.
### 
###############################################################
fit1 <- GeoFit2(data=data,coordx=coords,corrmodel=corrmodel, 
 neighb=3,likelihood="Conditional",
 type="Pairwise", start=start,fixed=fixed,X=X)
print(fit1)




###############################################################
############ Examples of spatial non-Gaussian random fields #############
###############################################################


################################################################
###
### Example 2. Maximum pairwise likelihood fitting of 
### a LogGaussian RF with Generalized Wendland correlation
### 
###############################################################
set.seed(524)
# Define the spatial-coordinates of the points:
N=500
x <- runif(N, 0, 1)
y <- runif(N, 0, 1)
coords <- cbind(x,y)
X=cbind(rep(1,N),runif(N))
mean=1; mean1=2 # regression parameters
nugget=0
sill=0.5
scale=0.2
smooth=0

model="LogGaussian"
corrmodel="GenWend"
param=list(mean=mean,mean1=mean1,sill=sill,scale=scale,
 nugget=nugget,power2=4,smooth=smooth)
# Simulation of a non stationary LogGaussian RF:
data <- GeoSim(coordx=coords, corrmodel=corrmodel,model=model,X=X,
 param=param)$data

fixed<-list(nugget=nugget,power2=4,smooth=smooth)
start<-list(mean=mean,mean1=mean1,scale=scale,sill=sill)
I=Inf
lower<-list(mean=-I,mean1=-I,scale=0,sill=0)
upper<-list(mean= I,mean1= I,scale=I,sill=I)

# Maximum pairwise composite-likelihood fitting of the RF:
fit <- GeoFit2(data=data,coordx=coords,corrmodel=corrmodel, model=model,
 neighb=3,likelihood="Conditional",type="Pairwise",X=X,
 optimizer="nlminb",lower=lower,upper=upper,
 start=start,fixed=fixed)
print(unlist(fit$param))


################################################################
###
### Example 3. Maximum pairwise likelihood fitting of
### SinhAsinh random fields with Wendland0 correlation
###
###############################################################
set.seed(261)
model="SinhAsinh"
# Define the spatial-coordinates of the points:
x <- runif(500, 0, 1)
y <- runif(500, 0, 1)
coords <- cbind(x,y)

corrmodel="Wend0"
mean=0;nugget=0
sill=1
skew=-0.5
tail=1.5
power2=4
c_supp=0.2

# model parameters
param=list(power2=power2,skew=skew,tail=tail,
 mean=mean,sill=sill,scale=c_supp,nugget=nugget)
data <- GeoSim(coordx=coords, corrmodel=corrmodel,model=model, param=param)$data

plot(density(data))
fixed=list(power2=power2,nugget=nugget)
start=list(scale=c_supp,skew=skew,tail=tail,mean=mean,sill=sill)
# Maximum pairwise likelihood:
fit1 <- GeoFit2(data=data,coordx=coords,corrmodel=corrmodel, model=model,
 neighb=3,likelihood="Marginal",type="Pairwise",
 start=start,fixed=fixed)
print(unlist(fit1$param))



}
\keyword{Composite}

\name{Lik}
\alias{Lik}
\encoding{UTF-8}
\title{Optimizes the Log Likelihood}
\description{
 Subroutine called by GeoFit.
 The procedure estimates the model parameters by maximization of the
 log-likelihood.
}
\usage{
Lik(copula,bivariate,coordx,coordy,coordz,coordt,
coordx_dyn,corrmodel,data,fixed,flagcor,flagnuis,
 grid,lower,mdecomp,model,namescorr,
 namesnuis,namesparam,numcoord,
 numpairs,numparamcor,numtime,optimizer,
 onlyvar,param,radius,setup,
 spacetime,sparse,varest,taper,type,
 upper,ns,X,neighb,MM,aniso,score)
}
\arguments{
 \item{copula}{String; the type of copula. It can be "Beta" or "Gaussian"}
 \item{bivariate}{Logical; if \code{TRUE} then the data come froma a bivariate random field.
 Otherwise from a univariate random field.}
 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix 
 Coordinates on a sphere for a fixed radius \code{radius} 
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}

 \item{coordt}{A numeric vector assigning one dimension of temporal coordinates. Optional argument; the default is \code{NULL}, in which case a spatial random field is expected. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{A list of \eqn{m} numeric (\eqn{d_t \times 2}{d x 2})-matrices
 containing dynamical (in time) spatial coordinates. Optional argument, the default is \code{NULL}
 } 
 \item{corrmodel}{Numeric; the id of the correlation model.}
 \item{data}{A numeric vector or a (\eqn{n \times d}{n x d})-matrix or (\eqn{d \times d
 \times n}{d x d x n})-matrix of observations.}
 \item{flagcor}{A numeric vector of flags denoting which correlation
 parameters have to be estimated.}
 \item{flagnuis}{A numeric verctor of flags denoting which nuisance
 parameters have to estimated.}
 \item{fixed}{A numeric vector of parameters that
 will be considered as known values.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data
 are interpreted as a vector or a (\eqn{n \times d}{n x d})-matrix,
 instead if \code{TRUE} then (\eqn{d \times d
 \times n}{d x d x n})-matrix is considered.}
 \item{lower}{An optional named list giving the values for the lower bound of the space parameter
 when the optimizer is \code{L-BFGS-B} or \code{nlminb} or \code{optimize}. The 
 names of the list must be the same of the names in the \code{start} list.}
 \item{model}{Numeric; the id value of the density associated to the likelihood
 objects.}
 \item{namescorr}{String; the names of the correlation parameters.}
 \item{namesnuis}{String; the names of the nuisance parameters.}
 \item{namesparam}{String; the names of the parameters to be maximised.}
 \item{numcoord}{Numeric; the number of coordinates.}
 \item{numpairs}{Numeric; the number of pairs.}
 \item{numparamcor}{Numeric; the number of the correlation parameters.}
 \item{numtime}{Numeric; the number of temporal observations.}
 \item{mdecomp}{String; the type of matrix decomposition used in the simulation. Default is cholesky.
 The other possible choices is \code{svd} (Singular values decomposition).} 
 \item{optimizer}{String; the optimization algorithm
 (see \code{\link{optim}} for details). \code{Nelder-Mead} is the
 default. Other possible choices are \code{nlm}, \code{BFGS} \code{L-BFGS-B} and \code{nlminb}. 
 In these last two cases upper and lower bounds can be passed by the user. In the case
 of one-dimensional optimization, the function \code{optimize} is used.} 
 \item{onlyvar}{Logical; if \code{TRUE} (and varest is \code{TRUE}) only the variance covariance matrix is computed
 without optimizing. \code{FALSE} is the default.} 
 \item{param}{A numeric vector of parameters.}
 \item{sparse}{Logical; if \code{TRUE} then maximum likelihood is computed
 using sparse matrices algorithms.\code{FALSE} is the default.}
 \item{radius}{Numeric; the radius of the sphere when considering data on a sphere.}
 \item{ns}{Numeric: vector of number of location sites for each temporal instants} 
 \item{setup}{A List of useful components for the estimation based on the maximum tapered likelihood.}
 \item{spacetime}{Logical; if the random field is spatial
 (\code{FALSE}) or spatio-temporal (\code{TRUE}).}
 \item{varest}{Logical; if \code{TRUE} the estimate' variances and
 standard errors are returned.
 \code{FALSE} is the default.}
 \item{taper}{String; the name of the taper correlation function.}
 \item{type}{String; the type of the likelihood objects. If \code{Pairwise} (the
 default) then the marginal composite likelihood is formed by
 pairwise marginal likelihoods.}
\item{upper}{An optional named list giving the values for the upper bound 
 of the space parameter when the optimizer is or \code{L-BFGS-B} or \code{nlminb} or \code{optimize}. 
 The names of the list must be the same of the names in the \code{start} list.}
 \item{X}{Numeric; Matrix of spatio(temporal)covariates in the linear mean specification.}
 \item{neighb}{Numeric;parameter for vecchia approximation using GPvecchia package}
 \item{MM}{Numeric;a non constant fixed mean }
 \item{aniso}{Logical; should anisotropy be considered?}
 \item{score}{Logical; should score function be computed?}
}

\value{Return a list from an \code{optim} call.}

\seealso{\code{\link{GeoFit}}}



\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\keyword{Composite}

\name{corrsas}
\alias{corrsas}
\encoding{UTF-8}
\title{Correlation Function for Sinh-Arcsinh random fields}
\description{
 Computes the correlations f for a random field transformed via the sinh-arcsinh (SAS) distribution. This transformation introduces flexible skewness and tail behavior to an underlying Gaussian field. The resulting correlation is derived via an infinite Hermite expansion, as described in Equation (16) of Blasi et al. (2022).
}
\usage{
corrsas(corr, skew, tail, max_coeff = NULL)
}
\arguments{
 \item{corr}{A numeric vector of correlation values of the underlying standard Gaussian random field.}
 \item{skew}{A numeric value representing the skewness parameter \eqn{\alpha} of the sinh-arcsinh transformation. Positive values induce right-skewness, negative values left-skewness.}
 \item{tail}{A positive numeric value representing the tailweight parameter \eqn{\kappa}. Values less than 1 yield heavier tails than Gaussian, while values greater than 1 produce lighter tails.}
 \item{max_coeff}{Optional integer. The maximum number of Hermite coefficients used in the infinite series expansion. If \code{NULL}, a default truncation value is used internally.}
}
\details{
 The correlation of the sinh-arcsinh transformed field is computed as:
 \deqn{
 \rho_{SAS}(h) = \sum_{j=1}^{\infty} \frac{\xi_j^2(\alpha, \kappa)}{j!} \rho(h)^j
 }
 where \eqn{\rho(h)} is the correlation function of the underlying Gaussian field and \eqn{\xi_j(\alpha, \kappa)} are Hermite coefficients depending on the skewness and tail parameters. This series is truncated at \code{max_coeff} terms for computational feasibility.

 See Equation (16) in Blasi et al. (2022) for the full derivation.
}
\value{
 A numeric vector of adjusted correlation values corresponding to the SAS-transformed process.
}
\references{
 Blasi, F., Caamaño-Carrillo, C., Bevilacqua, M., Furrer, R. (2022). A selective view of climatological data and likelihood estimation. \emph{Spatial Statistics}, 50, 100596. \doi{10.1016/j.spasta.2022.100596}
}

\examples{
# Example usage:
rho <- seq(0, 1, length.out = 50)
rho_sas <- corrsas(rho, skew = 0.5, tail = 0.8, max_coeff = 20)
plot(rho, rho_sas, type = "l", main = "SAS Correlation", 
 xlab = "Original Correlation", ylab = "Transformed Correlation")
}

\name{GeoTestIndependence}
\alias{GeoTestIndependence}
\title{Parametric bootstrap test for absence of spatial dependence}
\description{
Performs a parametric bootstrap likelihood-ratio test of spatial independence
against a spatially dependent random-field model. Under the null hypothesis,
the observations at distinct sites are fitted using the marginal independence
likelihood. Under the alternative hypothesis, a full spatial likelihood is
fitted with the nugget proportion estimated below its independence boundary.
The bootstrap is required because the null hypothesis corresponds to a boundary
point and the spatial correlation parameters are not identified under the null.
}
\usage{
GeoTestIndependence(data, coordx,
 start, fixed = NULL,
 corrmodel = "Matern",
 model = "Gaussian",
 optimizer = "bobyqa",
 optimizer_ind = "Nelder-Mead",
 lower = NULL, upper = NULL,
 X = NULL, n = 1,
 distance = "Eucl", radius = 1,
 anisopars = NULL,
 est.aniso = c(FALSE, FALSE),
 sparse = FALSE,
 B = 1000,
 parallel = FALSE,
 ncores = 6,
 progress = TRUE,
 seed = NULL)
}
\arguments{
 \item{data}{Numeric vector of observations.}

 \item{coordx}{Matrix or object coercible to a matrix containing the spatial
 coordinates. The number of rows must equal \code{length(data)}. Duplicate
 locations are not allowed.}

 \item{start}{Non-empty named list containing the parameters to be estimated
 under the spatial alternative. It must include, either directly or through
 \code{fixed}, all parameters required by \code{model} and \code{corrmodel}.
 The parameter \code{nugget} is always estimated under the alternative; if it
 is supplied in \code{fixed}, it is automatically moved to \code{start}.}

 \item{fixed}{Optional named list containing parameters kept fixed under the
 spatial alternative. Parameters cannot occur in both \code{start} and
 \code{fixed}.}

 \item{corrmodel}{Correlation model used under the spatial alternative. The
 default is \code{"Matern"}.}

 \item{model}{Marginal model for the observations. The default is
 \code{"Gaussian"}.}

 \item{optimizer}{Optimization method passed to \code{GeoFit} for the full
 spatial fit under \eqn{H_1}. The default is \code{"bobyqa"}.}

 \item{optimizer_ind}{Optimization method used for the marginal independence
 fit under \eqn{H_0}. The default is \code{"Nelder-Mead"}. If the requested
 method fails, the function also tries \code{"Nelder-Mead"} as a fallback and
 retains the successful fit with the largest finite log-likelihood.}

 \item{lower, upper}{Optional named lists of lower and upper bounds for the
 parameters estimated under the spatial alternative. Names must refer only to
 parameters in \code{start}, after the possible promotion of \code{nugget}
 from \code{fixed}. The upper bound for \code{nugget} must be one, or omitted,
 so that the independence model belongs to the closure of the alternative.}

 \item{X}{Optional design matrix for the mean or marginal regression model.
 Its number of rows must equal \code{length(data)}.}

 \item{n}{Model-specific parameter passed to \code{GeoFit} and \code{GeoSim}.
 The default is \code{1}.}

 \item{distance}{Distance metric passed to \code{GeoFit}, \code{GeoDistances},
 and \code{GeoSim}. The default is \code{"Eucl"}.}

 \item{radius}{Positive radius used for spherical or geodesic distances. The
 default is \code{1}.}

 \item{anisopars}{Optional list of anisotropy parameters passed to the spatial
 fit under the alternative.}

 \item{est.aniso}{Logical vector of length two indicating whether the
 anisotropy angle and ratio are estimated under the alternative. The default is
 \code{c(FALSE, FALSE)}.}

 \item{sparse}{Logical value passed to the full-likelihood spatial fit. The
 default is \code{FALSE}.}

 \item{B}{Number of parametric bootstrap replications. The default is
 \code{1000}. Values below \code{99} produce a coarse bootstrap \emph{p}-value
 and generate a warning.}

 \item{parallel}{Logical; default \code{FALSE}. Set \code{TRUE} to evaluate bootstrap fits in parallel when the required packages are available.}

 \item{ncores}{Positive integer or \code{NULL}; default \code{6}. With \code{parallel=TRUE}, an explicit integer requests that many workers, capped by detected cores and the number of bootstrap replications. Set \code{ncores=NULL} for automatic selection, capped at six workers and normally leaving one detected core free.}

 \item{progress}{Logical value indicating whether progress bars should be
 displayed through \pkg{progressr}.}

 \item{seed}{Optional finite numeric seed. When supplied, the previous global
 random-number-generator state is restored when the function exits.}
}
\details{
The test compares
\deqn{H_0: \text{absence of spatial dependence}}
with
\deqn{H_1: \text{a spatially dependent random-field model}.}

Under \eqn{H_0}, the function fits
\code{GeoFit(..., likelihood = "Marginal", type = "Independence")}. Only
marginal and regression parameters are estimated. Spatial correlation
parameters and the nugget proportion are absent from the null fit because they
are not identified under independence.
The null fit uses \code{optimizer_ind}; if necessary, a
\code{"Nelder-Mead"} fallback is attempted. The optimizer associated with the
selected null fit is returned in \code{optimizer_independence}.

Under \eqn{H_1}, the function fits
\code{GeoFit(..., likelihood = "Full", type = "Standard")} using
\code{corrmodel}. The nugget parameter is estimated on
\eqn{[0,1)}. Independence is represented by the boundary
\eqn{\texttt{nugget}=1}, where the spatial correlation contribution vanishes.
The upper optimization bound is internally set to a value immediately below
one because \code{GeoFit} requires \code{nugget < 1}.

The observed likelihood-ratio statistic is
\deqn{\Lambda_{obs}=2\{\ell_1-\ell_0\},}
where \eqn{\ell_0} is the maximized independence log-likelihood and
\eqn{\ell_1} is the maximized full spatial log-likelihood. If the unrestricted
spatial fit reaches the independence boundary, the statistic is set to zero
and the function returns \code{pvalue = 1} without running the bootstrap.

The usual chi-squared likelihood-ratio approximation is not used because the
null value is on the boundary and the spatial correlation parameters are not
identified under \eqn{H_0}. Instead, the null distribution is approximated by
parametric bootstrap. The bootstrap \emph{p}-value is
\deqn{
\widehat p =
\frac{1+\sum_{b=1}^{B_{valid}}
I(\Lambda_b^*\geq\Lambda_{obs})}
{B_{valid}+1}.
}

Bootstrap datasets are generated exactly independently at the observed sites.
Because \code{GeoSim} does not accept \code{nugget = 1}, the implementation
uses an auxiliary \code{GenWend} correlation model with compact support equal
to one half of the minimum inter-site distance. Consequently, all off-diagonal
correlations at the observed locations are exactly zero. This auxiliary model
is used only as a simulation device; the fitted null model remains the marginal
independence model.

The observed parameter estimates are used as starting values in the bootstrap
refits. If the first spatial fit fails to attain at least the independence
log-likelihood, adaptive fallback fits are attempted using alternative initial
values for \code{nugget}. When the best spatial fit is numerically equivalent
to the null fit, the bootstrap statistic is treated as a boundary value equal
to zero.

When \code{parallel = TRUE}, bootstrap fits are evaluated with
\pkg{future.apply}. Each simulated dataset is written temporarily to a separate
RDS file so that workers do not receive the complete simulation object. The
user's previous \pkg{future} plan, \pkg{progressr} handlers, and random-number-
generator state are restored on exit.

The current implementation is restricted to univariate models with a single
nugget proportion named \code{nugget}. It compares the independence likelihood
only with the full spatial likelihood. It does not directly compare the
independence likelihood with a pairwise composite likelihood, because the two
criteria are not on the same scale.
}
\value{
An invisible list with components:

\item{lambda_obs}{Observed likelihood-ratio statistic.}

\item{pvalue}{Parametric bootstrap \emph{p}-value.}

\item{nugget_hat}{Estimated nugget proportion under the spatial alternative.}

\item{boundary_observed}{Logical value indicating whether the unrestricted
spatial fit was classified as an independence-boundary solution.}

\item{optimizer_independence}{Optimizer used by the selected independence fit
under \eqn{H_0}. This can differ from \code{optimizer_ind} when the
\code{"Nelder-Mead"} fallback is selected.}

\item{optimizer_spatial}{Optimizer used for the full spatial fit under
\eqn{H_1}.}

\item{seed}{Seed supplied by the user, or \code{NULL}.}

\item{B_requested}{Requested number of bootstrap replications.}

\item{bootstrap_successful}{Number of valid bootstrap replications.}

\item{bootstrap_failed}{Number of failed bootstrap replications.}

\item{bootstrap_success_rate}{Fraction of successful bootstrap replications.
This is \code{NA} when the observed statistic is zero and the bootstrap is not
run.}

\item{fallback_observed}{Logical value indicating whether adaptive starting
values were required for the observed spatial fit.}

\item{fallback_bootstrap}{Number of bootstrap replications in which adaptive
starting values were used for the spatial fit.}

\item{boundary_bootstrap}{Number of valid bootstrap replications classified as
independence-boundary solutions.}

\item{simulation_support}{Compact-support value used by the auxiliary
\code{GenWend} model to simulate independent observations. This component is
absent when the observed statistic is zero and the bootstrap is skipped.}

\item{B_rep}{Vector of valid bootstrap likelihood-ratio statistics.}

\item{fit_H0}{\code{GeoFit} object fitted under spatial independence.}

\item{fit_H1}{\code{GeoFit} object fitted under the spatial alternative.}
}
\note{
For Gaussian random fields, the null hypothesis is equivalent to mutual
independence of the observations at distinct spatial locations. For
non-Gaussian models, the interpretation as independence requires that zero
latent spatial dependence imply independence in the corresponding observed
random-field construction.

A small bootstrap \emph{p}-value indicates evidence of spatial dependence. For
final analyses, a value such as \code{B = 999} or larger is recommended.
}
\seealso{
\code{\link{GeoFit}},
\code{\link{GeoSim}},
\code{\link{GeoDistances}},
\code{\link{GeoTestsupp_space}},
\code{\link{GeoTestIsotropy}}
}
\examples{
\dontrun{
library(GeoModels)

set.seed(3)
N <- 300
coords <- cbind(runif(N), runif(N))

X <- cbind(1, runif(N))

param <- list(
  mean = 0.2,
  mean1 = -0.5,
  sill = 1,
  nugget = 0,
  scale = 0.2 / 3,
  smooth = 0.5
)

data <- GeoSim(
  coordx = coords,
  corrmodel = "Matern",
  model = "Gaussian",
  param = param,
  X = X
)$data

start <- list(
  mean = 0,
  mean1 = 0,
  sill = 1,
  nugget = 0.1,
  scale = 0.1
)

fixed <- list(smooth = 0.5)

lower <- list(
  mean = -Inf,
  mean1 = -Inf,
  sill = 1e-6,
  nugget = 0,
  scale = 1e-6
)

upper <- list(
  mean = Inf,
  mean1 = Inf,
  sill = Inf,
  nugget = 1,
  scale = 1
)

ans <- GeoTestIndependence(
  data = data,
  coordx = coords,
  X = X,
  start = start,
  fixed = fixed,
  corrmodel = "Matern",
  model = "Gaussian",
  optimizer = "bobyqa",
  optimizer_ind = "Nelder-Mead",
  lower = lower,
  upper = upper,
  B = 199,
  parallel = TRUE,
  seed = 123
)

ans$pvalue
ans$lambda_obs
ans$nugget_hat
ans$bootstrap_success_rate
}
}
\author{
Moreno Bevilacqua
\email{moreno.bevilacqua89@gmail.com}
\url{https://sites.google.com/view/moreno-bevilacqua/home}
}
\keyword{Bootstrap}
\keyword{Likelihood}
\keyword{HypothesisTest}
\keyword{Spatial}

\name{GeoResiduals}
\alias{GeoResiduals}
\encoding{UTF-8}
\title{Compute Residuals from a Fitted GeoModels Model}
\description{
 The procedure return a GeoFit object associated to the estimated residuals. For a random field Y defined on the real line (Gaussian, Skew Gaussian, Tukeyh etcc) they are computed as (Y-m)/sqrt(v) where m and v are the estimated mean and variance respectively.
 For a random field Y defined on the positive real line (Gamma, Weibull, Log-Gaussian) they are computed as Y/m where m is the estimated mean.
 In the first case residuals have zero mean and unut variance with a specific distribution defined on the real line.
 In the second case residuals have unit mean with a specific distribution defined on the positive real line.
 For the discrete Bernoulli/Binary, Binomial, Geometric/negative-binomial,
 Poisson, and Poisson--Gamma models, the function returns Pearson residuals
 computed from the fitted marginal mean and variance. The paper specifies these marginal moments but does
 not prescribe a residual definition; Pearson residuals are an explicit package
 convention.
For continuous models, the returned object can be coupled with
\code{\link{GeoQQ}} and \code{\link{GeoCovariogram}}. For discrete Pearson
residuals, use the returned object for residual maps and variograms, and apply
\code{GeoQQ()} or \code{GeoPit()} to the original fitted object for marginal
assessment.
}
\usage{GeoResiduals(fit)}
\arguments{
 \item{fit}{A fitted object obtained from the
 \code{\link{GeoFit}}.}
}

\details{

The fitted location is evaluated from the full mean specification: a
site-specific fixed mean vector when present, otherwise \eqn{X\beta} with
coefficients \code{mean}, \code{mean1}, and so on matched in order to the
columns of \code{X}.  The returned residual object is converted to an
intercept-only, standardized representation for subsequent diagnostics.  All
other fields of the original \code{GeoFit} object, including neighbourhood,
thinning, anisotropy, convergence, and pair-selection metadata, are preserved.

For \code{Beta2} and \code{Kumaraswamy2}, the residual transformation uses
the fitted support \code{min}/\code{max} and the corresponding fitted marginal
CDF, so the standardized residual scale is the unit interval.

For the discrete models described in the package paper, Pearson residuals are
used: \eqn{(Y-E(Y))/\sqrt{Var(Y)}}.  Thus \eqn{p=\Phi(\mu)} for
Bernoulli, Binomial and negative-binomial fields, while
\eqn{\lambda=\exp(\mu)} for Poisson fields.  For Poisson--Gamma fields,
with fitted mean \eqn{\lambda} and shape \eqn{\kappa}, the fitted variance is
\eqn{\lambda(1+\lambda/\kappa)}.  These deterministic residuals
are intended for residual maps and variograms; they are not randomized quantile
residuals, and \code{GeoQQ()} intentionally rejects the Pearson-residual object.  Bivariate residuals are currently supported only
for Gaussian models.
}

\value{
 Returns an (updated) object of class \code{GeoFit}.  The component
 \code{residual_type} records whether the data are Pearson or model-standardized residuals.
}


\seealso{\code{\link{GeoFit}}.}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}



\examples{
library(GeoModels)



###########################
###Example 1: Residuals using a Gaussian RF
###########################
set.seed(211)
model="Gaussian";
N=700 # number of location sites
# Set the coordinates of the points:
x = runif(N, 0, 1)
y = runif(N, 0, 1)
coords=cbind(x,y)

# regression parameters
mean = 5
mean1=0.8

X=cbind(rep(1,N),runif(N))
# correlation parameters:
corrmodel = "Wend0"
sill = 1
nugget = 0
scale = 0.3
power2=4

param=list(mean=mean,mean1=mean1, sill=sill, nugget=nugget, 
 scale=scale,power2=power2)
# Simulation of the Gaussian RF:
data = GeoSim(coordx=coords, corrmodel=corrmodel, X=X,model=model,param=param)$data

start=list(mean=mean,mean1=mean1, scale=scale,sill=sill)
fixed=list(nugget=nugget,power2=power2)
# Maximum composite-likelihood fitting 
fit = GeoFit(data,coordx=coords, corrmodel=corrmodel,model=model,X=X,
 likelihood="Conditional",type='Pairwise',start=start,
 fixed=fixed,neighb=3)

res=GeoResiduals(fit)
mean(res$data) # should be approx 0
var(res$data) # should be approx 1
# checking goodness of fit marginal model
GeoQQ(res);GeoQQ(res,type="D",col="red",ylim=c(0,0.5),breaks=20);
# Empirical estimation of the variogram for the residuals:
vario = GeoVariogram(res$data,coordx=coords,maxdist=0.5)
# Comparison between empirical amd estimated semivariogram for the residuals
GeoCovariogram(res, show.vario=TRUE, vario=vario,pch=20)





###########################
###Example 2: Residuals using a Weibull RF
###########################
model="Weibull";shape=4
N=700 # number of location sites
# Set the coordinates of the points:
x = runif(N, 0, 1)
y = runif(N, 0, 1)
coords=cbind(x,y)


# regression parameters
mean = 5
mean1=0.8

X=cbind(rep(1,N),runif(N))
# correlation parameters:
corrmodel = "Wend0"
sill = 1
nugget = 0
scale = 0.3
power2=4

param=list(mean=mean,mean1=mean1, sill=sill, nugget=nugget, 
	 scale=scale,shape=shape,power2=power2)
# Simulation of the Gaussian RF:
data = GeoSim(coordx=coords, corrmodel=corrmodel, X=X,model=model,param=param)$data

I=Inf
start=list(mean=mean,mean1=mean1, scale=scale,shape=shape)
lower=list(mean=-I,mean1=-I, scale=0,shape=0)
upper=list(mean= I,mean1= I, scale=I,shape=I)
fixed=list(nugget=nugget,sill=sill,power2=power2)
# Maximum composite-likelihood fitting 
fit = GeoFit(data,coordx=coords, corrmodel=corrmodel,model=model,X=X,
 likelihood="Conditional",type='Pairwise',start=start,
 optimizer="nlminb", lower=lower,upper=upper,
 fixed=fixed,neighb=3)


res=GeoResiduals(fit)
mean(res$data) # should be approx 1
# checking goodness of fit marginal model
GeoQQ(res);GeoQQ(res,type="D",lwd=2,ylim=c(0,1.7),breaks=20);
# Empirical estimation of the variogram for the residuals:
vario = GeoVariogram(res$data,coordx=coords,maxdist=0.5)
# Comparison between empirical amd estimated semivariogram for the residuals
GeoCovariogram(res, show.vario=TRUE, vario=vario,pch=20)

}

\keyword{Composite}

\name{GeoNeighborhood}
\alias{GeoNeighborhood}
\encoding{UTF-8}
\title{Spatio (temporal) neighborhood selection for local kriging.}
\description{
Given a set of spatio (temporal) locations and data, the procedure selects a spatio (temporal) neighborhood associated to some given spatio (temporal) locations.
The neighborhood is computed using a fixed spatio (temporal) threshold
or considering a fixed number of spatio (temporal) neighbors.
}

\usage{
GeoNeighborhood(data=NULL, coordx, coordy=NULL,coordz=NULL,
coordt=NULL, coordx_dyn=NULL, bivariate=FALSE,
 distance="Eucl", grid=FALSE, 
 loc, neighb=NULL,maxdist=NULL,
 maxtime=NULL, radius=1, time=NULL, 
 X=NULL,M=NULL,spobj=NULL,spdata=NULL,
 parallel=FALSE,ncores=6,check.duplicates=FALSE)
}
\arguments{
 \item{data}{An optional \eqn{d}{d}-dimensional vector (a single spatial realisation) or a (\eqn{d \times d}{d x d})-matrix (a single spatial realisation on regular grid)
 or a
 (\eqn{t \times d}{t x d})-matrix (a single spatio-temporal realisation) or an (\eqn{d \times d \times t \times n }{d x d x t})-array
 (a single spatio-temporal realisation on regular grid).}
 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix 
 Coordinates on a sphere for a fixed radius \code{radius} 
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordt}{A numeric vector giving one dimension of the observation-time coordinates. Optional argument; the default is \code{NULL}, in which case a spatial random field is expected. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{A list of \eqn{m} numeric (\eqn{d_t \times 2}{d x 2})-matrices
 containing dynamical (in time) spatial coordinates. Optional argument, the default is \code{NULL}
 }
 \item{bivariate}{If TRUE then data is considered as spatial bivariate data.}
 \item{distance}{String; the name of the spatial distance. The default
 is \code{Eucl}, the Euclidean distance. See \code{\link{GeoFit}} for details.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data
 are interpreted as spatial or spatio-temporal realisations on a set
 of non-equispaced spatial sites (irregular grid).}
 \item{loc}{A (\eqn{1 \times 2}{1 x 2})-matrix giving the spatial coordinate
 of the location for which a neighborhood is computed.}
 \item{neighb}{Numeric; an optional positive integer indicating the 
 order of spatial neighborhood.}
 \item{maxdist}{Numeric; a positive value indicating the maximum
 spatial distance considered in the spatial neighborhood
 selection.}
 \item{maxtime}{Numeric; an optional non-negative maximum temporal-distance threshold, expressed in the same units as \code{coordt}.}
 \item{radius}{Numeric; a value indicating the radius of the sphere when using the great 
 circle distance. Default value is 1.} 
 \item{time}{Numeric; the temporal coordinate at which the neighborhood is requested, interpreted on the same numeric time scale as \code{coordt}.}
 \item{X}{Numeric; an optional Matrix of spatio (temporal) covariates.}
 \item{M}{Numeric; an estimated spatio (temporal) mean vector.}
 \item{spobj}{An object of class \code{sp} or \code{spacetime}. For space-time objects, the current \code{sp2Geo()} conversion uses sequential temporal indices and does not preserve irregular original time spacing; to retain irregular temporal distances, use the explicit-coordinate interface with numeric \code{coordt}.}
 \item{spdata}{Character:The name of data in the sp or spacetime object}
 \item{parallel}{Logical; default \code{FALSE}. Set \code{TRUE} to enable parallel execution where supported.}
 \item{ncores}{Positive integer or \code{NULL}; default \code{6}. With \code{parallel=TRUE}, an explicit integer requests that many workers, capped by detected cores and the number of available jobs. Set \code{ncores=NULL} for automatic selection, capped at six workers and normally leaving one detected core free.}
\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
}


\details{
Observed locations are checked for exact duplicates before neighborhood construction. In space-time data, the same spatial site may occur at different times, but duplicate space-time observations are not allowed.
}
\value{
 Returns a list containing the following information:
 \item{coordx}{A list of the matrix coordinates of the computed spatial neighborhood ;}
 \item{coordt}{A vector of the computed temporal neighborhood;}
 \item{data}{A list of the vector of data associated with the spatio (temporal) neighborhood;}
 \item{distance}{The type of spatial distance;}
 \item{numcoord}{The vector of numbers of location sites involved the spatial neighborhood;}
 \item{numtime}{The vector of numbers of temporal insttants involved the temporal neighborhood;}
 \item{radius}{The radius of the sphere if coordinates are passed in lon/lat format;}
 \item{spacetime}{\code{TRUE} if spatio-temporal and \code{FALSE} if spatial RF;}
 \item{X}{The matrix of spatio (temporal) covariates associated with the computed spatio (temporal) neighborhood;}
}



\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}


\examples{
library(GeoModels)
##########################################
#### Example: spatial neighborhood ######
##########################################
set.seed(75)
coords=cbind(runif(500),runif(500))

param=list(nugget=0,mean=0,scale=0.2,sill=1,
 power2=4,smooth=1)

data_all = GeoSim(coordx=coords, corrmodel="GenWend", 
 param=param)$data

plot(coords)
##two locations 
loc_to_pred=matrix(c(0.3,0.5,0.7,0.2),2,2)

points(loc_to_pred,pch=20)
neigh=GeoNeighborhood(data_all, coordx=coords, 
 loc=loc_to_pred,neighb=8)

# two Neighborhoods 
neigh$coordx
points(neigh$coordx[[1]],pch=20,col="red")
points(neigh$coordx[[2]],pch=20,col="blue")
# associated data
neigh$data


###################################################
#### Example: spatio temporal spatial neighborhood# 
###################################################

set.seed(78)
coords=matrix(runif(80),40,2)
coordt=seq(0,6,0.25)

param=list(nugget=0,mean=0,scale_s=0.2/3,scale_t=0.25/3,sill=2)

data_all = GeoSim(coordx=coords, coordt=coordt,corrmodel="Exp_Exp", 
 param=param)$data
## two location to predict
loc_to_pred=matrix(runif(4),2,2)
## three temporal instants to predict
time=c(1,2)

plot(coords,xlim=c(0,1),ylim=c(0,1))
points(loc_to_pred,pch=20)

neigh=GeoNeighborhood(data_all, coordx=coords, coordt=coordt,
 loc=loc_to_pred,time=time,neighb=3,maxtime=0.5)

# first spatio-temporal neighborhoods 
# with associated data
neigh$coordx[[1]]
neigh$coordt[[1]]
neigh$data[[1]]

plot(coords)
points(loc_to_pred,pch=20)
points(neigh$coordx[[1]],col="red",pch=20)

###################################################
#### Example: bivariate spatial neighborhood ##### 
###################################################

set.seed(79)
coords=matrix(runif(100),50,2)

param=list(mean_1=0,mean_2=0,scale=0.12,smooth=0.5,
 sill_1=1,sill_2=1,nugget_1=0,nugget_2=0,pcol=0.5)

data_all = GeoSim(coordx=coords,corrmodel="Bi_matern_sep",
 param=param)$data
## two location to predict
loc_to_pred=matrix(runif(4),2,2)

neigh=GeoNeighborhood(data_all, coordx=coords,bivariate=TRUE,
 loc=loc_to_pred,neighb=5)

plot(coords)
points(loc_to_pred,pch=20)
points(neigh$coordx[[1]],col="red",pch=20)
points(neigh$coordx[[2]],col="red",pch=20)

}

\name{CkType}
\alias{CkType}
\encoding{UTF-8}
\title{Checking Likelihood Objects}
\description{
Subroutine called by \code{InitParam}. \\
The procedure controls the type of likelihood objects inserted by the users.
}
\usage{
CkType(type)
}
\arguments{
 \item{type}{String; the type of likelihood object. Supported values are
 \code{Pairwise}, \code{Standard}, \code{Tapering}, \code{Tapering1},
 \code{Tapering2}, \code{GeoWLS}, \code{CV}, and \code{Independence}.}
}
\value{
The function returns a numeric positive integer, or \code{NULL} if the type of likelihood is invalid.
}
\details{
The procedure checks the likelihood object.
}
\seealso{\code{\link{GeoFit}}}
\author{
Moreno Bevilacqua \email{moreno.bevilacqua89@gmail.com} \url{https://sites.google.com/view/moreno-bevilacqua/home} \\
Víctor Morales Oñate \email{victor.morales@uv.cl} \url{https://sites.google.com/site/moralesonatevictor/} \\
Christian Caamaño-Carrillo \email{chcaaman@ubiobio.cl} \url{https://www.researchgate.net/profile/Christian-Caamano}
}
\keyword{Composite}

\name{GeoModels-spacetime-ordering}
\alias{GeoModels-spacetime-ordering}
\alias{spacetime-ordering}
\alias{spacetime-data-order}
\encoding{UTF-8}
\title{Ordering Conventions for Spatio-temporal Data, Covariates and Predictions}
\description{
Defines the ordering used by \pkg{GeoModels} for fixed-location and
dynamic-location spatio-temporal data, observation-level means and
covariates, and kriging prediction tasks.
}
\details{
The conventions below apply consistently to data, coordinates, design
matrices, known mean vectors, and prediction outputs.
}
\section{Irregular temporal coordinates}{
For the explicit-coordinate space-time interfaces, \code{coordt} is a numeric
coordinate vector and does not need to be equally spaced. Covariance,
likelihood, pair-selection, variogram, neighborhood and kriging calculations
use temporal lags derived from the supplied values, typically
\eqn{|t_i-t_j|}. Consequently, \code{maxtime} is a temporal-distance threshold
expressed in the same units as \code{coordt}, not an index lag or an order of
neighboring time points.

Important exceptions are explicit. Space-time
\code{GeoSimapprox(..., method = "CE")} requires finite, strictly increasing,
equally spaced temporal coordinates, and space-time turning-bands simulation
is not implemented. \code{GeoSimcond} and \code{GeoOutlier} currently reject
space-time input.

The \code{spobj} route has a separate legacy limitation: the current
\code{sp2Geo()} conversion of \code{STFDF}/\code{STIDF} objects replaces their
time positions by sequential indices \code{1, 2, ..., T}. Therefore an
irregular original time spacing is not preserved through \code{spobj}. When
irregular temporal distances matter, pass the data and coordinates through the
explicit interface and supply the numeric \code{coordt} values directly.
}

\section{Fixed spatial locations}{
Let \eqn{T} be \code{length(coordt)} and let \eqn{N} be the number of rows of
\code{coordx}. The data must be a \eqn{T \times N}{T x N} matrix. Row
\eqn{t} corresponds to \code{coordt[t]}, and column \eqn{i} corresponds to
row \eqn{i} of \code{coordx}.

The internal observation order is time-major and is equivalent to
\code{c(t(data))}:
\preformatted{
time 1: site 1, site 2, ..., site N
time 2: site 1, site 2, ..., site N
...
time T: site 1, site 2, ..., site N
}
Any observation-level vector, including a known mean vector, must use this
order. A design matrix \code{X} must have \eqn{NT} rows in the same order.
For example, if \code{X_time} has one row per time,
\code{X_time[rep(seq_len(T), each = N), , drop = FALSE]} creates the
corresponding observation-level design.
}

\section{Dynamic spatial locations}{
For dynamic locations, \code{coordx_dyn} must be a list of length \eqn{T}.
The element \code{coordx_dyn[[t]]} is an \eqn{N_t \times 2}{N_t x 2} or
\eqn{N_t \times 3}{N_t x 3} matrix containing the locations observed at
\code{coordt[t]}.

The data must be a list of length \eqn{T}; \code{data[[t]]} must contain
\eqn{N_t} values, and its \eqn{i}-th value corresponds to row \eqn{i} of
\code{coordx_dyn[[t]]}. The internal order is the time-wise concatenation
\code{unlist(data, use.names = FALSE)}, matched by
\code{do.call(rbind, coordx_dyn)}:
\preformatted{
all observations at time 1,
then all observations at time 2,
...,
then all observations at time T.
}
A design matrix \code{X} can be supplied either as a
\eqn{(\sum_t N_t) \times p}{sum(N_t) x p} matrix in this concatenated order,
or as a list of length \eqn{T} with \code{X[[t]]} having \eqn{N_t} rows and
the same row order as \code{coordx_dyn[[t]]}. A known observation-level mean
uses the same time-wise concatenation.
}

\section{Mean parameters}{
For a univariate linear mean,
\deqn{\mu = X\beta.}
The regression vector \eqn{\beta} is represented by the parameters
\code{mean}, \code{mean1}, \code{mean2}, and so on. The columns of \code{X}
correspond to these coefficients in exactly this order. An intercept-only
model is equivalent to a one-column matrix of ones and the single coefficient
\code{mean}. A supplied observation-level mean vector is a known mean and is
not interpreted as regression coefficients.
}

\section{Kriging prediction order}{
Let \eqn{N_{loc}} be the number of rows of \code{loc} and let
\eqn{T_{loc}} be \code{length(time)}. Kriging prediction tasks are ordered
location-major:
\preformatted{
location 1: time 1, time 2, ..., time Tloc
location 2: time 1, time 2, ..., time Tloc
...
location Nloc: time 1, time 2, ..., time Tloc
}
Therefore the rows of \code{Xloc} and the elements of \code{Mloc} must follow
this order. If \code{X_time_pred} has one row per prediction time, use
\code{X_time_pred[rep(seq_len(Tloc), times = Nloc), , drop = FALSE]}.
Equivalently, if \code{mu_time} contains one prediction mean per time, use
\code{rep(mu_time, times = Nloc)} for \code{Mloc}.

Dynamic coordinates describe the observed sites only and do not change the
prediction-task order. Spatio-temporal kriging output is a
\eqn{T_{loc} \times N_{loc}}{Tloc x Nloc} matrix: rows correspond to
\code{time} and columns correspond to rows of \code{loc}.
}
\seealso{
\code{\link{GeoFit}}, \code{\link{GeoSim}}, \code{\link{GeoVariogram}},
\code{\link{GeoKrig}}, and \code{\link{GeoKrigloc}}.
}
\examples{
T <- 3
N <- 2
data <- matrix(seq_len(T * N), nrow = T, ncol = N)
c(t(data))

X_time <- cbind(1, seq_len(T))
X <- X_time[rep(seq_len(T), each = N), , drop = FALSE]

Nloc <- 2
Xloc <- X_time[rep(seq_len(T), times = Nloc), , drop = FALSE]
Mloc <- rep(seq_len(T), times = Nloc)

## Irregular observation times are allowed in the explicit interface.
coordt_irregular <- c(0, 0.25, 1.1, 2.8)
## Temporal lags are based on differences between these values; for example,
abs(outer(coordt_irregular, coordt_irregular, "-"))
}
\keyword{spatial}

\name{GeoVarestbootstrap} 
\alias{GeoVarestbootstrap}
\encoding{UTF-8}
\title{Update a \code{GeoFit} object using parametric bootstrap for std error estimation.}
\description{
 The procedure updates a \code{GeoFit} object using a
 classical parametric bootstrap. The fitted model is simulated and refitted
 for each bootstrap replication; the empirical covariance matrix of the
 successful parameter estimates is used as the bootstrap covariance estimate.
 A faster score-bootstrap alternative is \code{GeoVarest}. Full/Standard
 likelihood fits are not accepted; obtain their Hessian-based standard errors
 directly with \code{GeoFit(..., varest = TRUE)}.}
\usage{
GeoVarestbootstrap(
  fit, K = 100, sparse = FALSE,
  optimizer = NULL, lower = NULL, upper = NULL,
  method = "cholesky", alpha = 0.95, L = 10000,
  parallel = FALSE, ncores = 6, progress = TRUE,
  seed = NULL, min_success_rate = 0.8
)
}
\arguments{
 \item{fit}{A fitted object obtained from the
 \code{\link{GeoFit}}.}
 \item{K}{The number of simulations in the parametric bootstrap.}
 \item{sparse}{Logical; if \code{TRUE} then cholesky decomposition is performed
 using sparse matrices algorithms (spam packake).}
 \item{optimizer}{The type of optimization algorithm (see \code{\link{GeoFit}} for details). If NULL then the optimization algorithm of 
 the object fit is chosen.}
 \item{lower}{An optional named list giving the values for the lower bound of the space parameter
 when the optimizer is \code{L-BFGS-B} or \code{nlminb} or \code{optimize}.}
 \item{upper}{An optional named list giving the values for the upper bound of the space parameter
 when the optimizer is \code{L-BFGS-B} or \code{nlminb} or \code{optimize}.}
 \item{method}{String; The method of simulation. Default is \code{cholesky}. For large data set three options are \code{TB} or \code{CE} (see the \code{GeoSimapprox}) function. } 
 \item{alpha}{ Numeric; The level of the confidence interval. }
 \item{L}{Numeric; the number of lines in the turning band method.} 
 \item{parallel}{Logical; default \code{FALSE}. If \code{TRUE} then 
 the estimation step is parallelized 
 }
 \item{ncores}{Positive integer or \code{NULL}; default \code{6}. With \code{parallel=TRUE}, an explicit integer requests that many workers, capped by detected cores and the number of bootstrap jobs. Set \code{ncores=NULL} for automatic selection, capped at six workers and normally leaving one detected core free.}
 \item{progress}{Logic; If TRUE then a progress bar is shown.}
  \item{seed}{Optional integer seed for reproducibility of the simulated
 samples.}
  \item{min_success_rate}{Minimum fraction of successful bootstrap refits
 required to return variance estimates. The default is 0.8.}
 }
\details{
For spatio-temporal fits, the original numeric \code{coordt} values stored in \code{fit} are reused. Irregularly spaced times are supported with \code{method = "cholesky"}; approximate simulation methods retain the restrictions documented in \code{\link{GeoSimapprox}}.

 The function updates a \code{GeoFit} object by refitting each simulated
data set. A sensitivity matrix is not required to estimate bootstrap standard
errors and confidence intervals. If \code{fit$sensmat} is available, CLIC and
CLBIC penalties are also computed. For stochastic nearest-neighbor fits, the
realized retained-pair graph from the original fit is reused unchanged across
all bootstrap refits.

Parallel multisession workers are started with the package-library paths of the calling R session, including the library containing the loaded GeoModels installation.

For \code{method = "TB"} without a copula, parallel workers are persistent and replications are streamed one at a time (simulate, evaluate, release). This keeps peak memory bounded when both \code{L} and \code{K} are large while still reusing the selected pair graph and static fitting context within each worker. For fitted copula models, \code{method="TB"} is simulated through the turning-bands backend of \code{GeoSimCopula}; the copula path is currently not streamed replicate-by-replicate. On platforms where \pkg{future} reports forked multicore execution as safe, GeoModels uses it automatically to reduce worker startup and serialization overhead; otherwise it falls back to multisession with the current GeoModels library path propagated explicitly.
}
\value{
 Returns an updated object of class \code{GeoFit}. The main updated components
 include \code{stderr}, \code{varcov}, \code{godambe} when invertible,
\code{conf.int}, \code{pvalues}, and \code{estimates}. When a compatible
\code{sensmat} is available, \code{claic}/\code{clic}, \code{clbic},
and \code{clic_penalty} are also updated.
}



\section{Three-dimensional coordinates}{
For a purely spatial univariate fit with explicit irregular three-dimensional
Euclidean coordinates, \code{method = "TB"} is supported when no
\code{anisopars} were used. Method \code{"CE"} remains restricted to
regular two-dimensional grids. Use \code{method = "cholesky"} for
bivariate, spatio-temporal, anisotropic, or other unsupported
three-dimensional cases.
}
\seealso{
\code{\link{GeoFit}} for the fitted objects used as input
}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{


library(GeoModels)

################################################################
###
### Example 1. Test on the parameter
### of a regression model using conditional composite likelihood
###
###############################################################
set.seed(342)
model="Gaussian" 
# Define the spatial-coordinates of the points:
NN=3500
x = runif(NN, 0, 1)
y = runif(NN, 0, 1)
coords = cbind(x,y)
# Parameters
mean=1; mean1=-1.25; # regression parameters
 sill=1 # variance

# matrix covariates
X=cbind(rep(1,nrow(coords)),runif(nrow(coords)))

# model correlation 
corrmodel="Matern"
smooth=0.5;scale=0.1; nugget=0;

# simulation
param=list(smooth=smooth,mean=mean,mean1=mean1,
 sill=sill,scale=scale,nugget=nugget)
data = GeoSim(coordx=coords, corrmodel=corrmodel,
 model=model, param=param,X=X)$data

I=Inf

fixed=list(nugget=nugget,smooth=smooth)
start=list(mean=mean,mean1=mean1,scale=scale,sill=sill)

lower=list(mean=-I,mean1=-I,scale=0,sill=0)
upper=list(mean=I,mean1=I,scale=I,sill=I)
# Maximum pairwise composite-likelihood fitting of the RF:
fit = GeoFit(data=data,coordx=coords,corrmodel=corrmodel, model=model,
 likelihood="Conditional",type="Pairwise",sensitivity=TRUE,
 lower=lower,upper=upper,neighb=3,
 optimizer="nlminb",X=X,
 start=start,fixed=fixed)

unlist(fit$param)


#fit_update=GeoVarestbootstrap(fit,K=100,parallel=TRUE)
#fit_update$stderr
#fit_update$conf.int
#fit_update$pvalues

}


\keyword{Composite}

\name{GeoCovDisplay}
\alias{GeoCovDisplay}
\encoding{UTF-8}
\title{Image plot displaying the pattern of the sparsness of a covariance matrix.}
\description{
 Image plot displaying the pattern of the sparsness of a covariance matrix.
}
\usage{
GeoCovDisplay(covmatrix,limits=FALSE,pch=2)
}
\arguments{
 \item{covmatrix}{An object returned by \code{\link{GeoCovmatrix}}.}
 \item{limits}{Logical; If TRUE and the covariance matrix is spatiotemporal or spatial bivariate
 then vertical and horizontal lines are added to the image plot.} 
 \item{pch}{Type of symbols to use in the image plot.}
}

\details{ 
 For a given covariance matrix object (\code{\link{GeoCovmatrix}})
 the function diplays the pattern of the sparsness of a covariance matrix
 where the white color represents 0 entries and black color represents non zero entries}
\value{Produces a plot. No values are returned.}


\seealso{\code{\link{GeoCovmatrix}}}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{

library(GeoModels)


 # Define the spatial-coordinates of the points:
x <- runif(100, 0, 2)
y <- runif(100, 0, 2)
coords=cbind(x,y)
matrix1 <- GeoCovmatrix(coordx=coords, corrmodel="GenWend", param=list(smooth=0,
 power2=4,sill=1,scale=0.2,nugget=0))
 
GeoCovDisplay(matrix1)

}
\keyword{Sparsness pattern}

\name{GeoDistances}
\alias{GeoDistances}
\title{Compute Distance Matrices (Euclidean, Chordal, or Geodesic)}
\description{
Compute a full distance matrix between coordinates in 2D or 3D, using Euclidean, 
Chordal, or Geodesic distance. The computation is done in C for efficiency, 
with parameters passed from R via \code{dotCall64}.
}
\usage{
GeoDistances(coordx = NULL, coordy = NULL, coordz = NULL,
 distance = c("Eucl", "Chor", "Geod"), radius = 1)
}
\arguments{
 \item{coordx}{A numeric matrix of coordinates. Must have 2 or 3 columns 
 (for 2D or 3D) (if coordy and coordz are not provided). Otherwise
 coordinates can be given via 
 \code{coordx}, \code{coordy}, and optionally \code{coordz}.}
 \item{coordy}{Optional numeric vector of coordinates.}
 \item{coordz}{Optional numeric vector of coordinates (for 3D Euclidean).}
 \item{distance}{Type of distance to compute: 
 \code{"Eucl"} for Euclidean, 
 \code{"Chor"} for chordal, 
 \code{"Geod"} for geodesic.}
 \item{radius}{Radius of the sphere for geodesic or chordal distances. 
 Defaults to 1. Use Earth radius (e.g., 6371 Km) for geographic distances.}
}
\details{
\itemize{
 \item Euclidean distance: straight-line distance in 2D or 3D space.
 \item Chordal distance: Euclidean distance between points projected on the 
 unit sphere, scaled by the given radius.
 \item Geodesic distance: shortest path along the surface of a sphere of 
 radius \code{radius}.
}
}
\value{
A symmetric numeric matrix of size \code{n x n}, where \code{n} is the number 
of points. Each entry contains the distance between the corresponding pair 
of coordinates.
}
\examples{
# Example with Euclidean distance
coords <- cbind(c(0, 10), c(0, 10))
GeoDistances(coords, distance = "Eucl")

# Example with geodesic distance (approx distance between Rome and New York)
rome <- c(12.4964, 41.9028) #
ny <- c(-74.0060, 40.7128)
coords <- rbind(rome, ny)
GeoDistances(coords, distance = "Geod", radius = 6371)
}

\keyword{Distances}

\name{GeoCorrFct}
\alias{GeoCorrFct}
\encoding{UTF-8}
\title{Spatial and Spatio-temporal correlation or covariance of (non) Gaussian random fields}
\description{
 The function computes the correlations of 
 a spatial (or spatio-temporal or bivariate spatial) Gaussian or non-Gaussian randomm field
 for a given correlation model and a set of spatial (temporal) distances.
}
\usage{
GeoCorrFct(x,t=NULL,corrmodel, model="Gaussian",
distance="Eucl", param, radius=6371,n=1,
covariance=FALSE,variogram=FALSE)
}

\arguments{
 \item{x}{A set of spatial distances.}
 \item{t}{A set of (optional) temporal distances.}
 \item{corrmodel}{String; the name of a correlation model, for the
 description see \code{\link{GeoCovmatrix}}.}
 \item{model}{String; the type of random field. See \code{\link{GeoFit}}.}
 \item{distance}{String; the name of the spatial distance. The default
 is \code{Eucl}, the Euclidean distance. See \code{\link{GeoFit}}.}
 \item{param}{A list of parameter values required for the covariance
 model. For univariate curves, omitted \code{mean}, \code{sill}, and
 \code{nugget} use the defaults 0, 1, and 0, respectively.}
 \item{radius}{Numeric; a value indicating the radius of the sphere 
 when using covariance models valid using the great circle distance. 
 Default value is the radius of the earth in Km (i.e. 6371)} 
 \item{n}{Positive integer size parameter. For Binomial it is the common number of trials used by this lag-only curve. For Negative Binomial it is the single common number \eqn{r} of successes; Geometric corresponds to \eqn{r=1}. Site-specific Binomial sizes are handled by \code{GeoCovmatrix()} rather than by a single lag-only curve. Default is \eqn{1}.}
 \item{covariance}{Logic; if TRUE then the covariance is returned. Default is FALSE}
 \item{variogram}{Logical; if FALSE the correlation/covariance is returned. If TRUE, \code{covariance=FALSE} returns the normalized semivariogram \eqn{1-\rho(h)}, while \code{covariance=TRUE} returns the semivariogram on the response scale.}
 }


\value{Returns correlations or covariances values associated to a given parametric spatial and temporal correlation models.}


\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian, Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{
library(GeoModels)

################################################################
###
### Example 1. Covariance of a Gaussian random field with underlying 
### Matern correlation model with nugget
###
###############################################################
# Define the spatial distances
x = seq(0,1,0.002)
# Correlation Parameters for Matern model 
CorrParam("Matern")
NuisParam("Gaussian")
# Matern Parameters 
param=list(sill=2,smooth=0.5,scale=0.2/3,nugget=0.2,mean=0)
cc= GeoCorrFct(x=x, corrmodel="Matern", covariance=TRUE,
 param=param,model="Gaussian")
plot(cc,ylab="Corr",lwd=2,main="Matern correlation",type="l")

################################################################
###
### Example 2. Covariance of a Gaussian random field with underlying 
### Generalized Wendland-Matern correlation model
###
###############################################################
CorrParam("GenWend_Matern")
NuisParam("Gaussian")
# GenWend Matern Parameters 
param=list(sill=2,smooth=1,scale=0.1,nugget=0,power2=1/4,mean=0)
cc= GeoCorrFct(x=x, corrmodel="GenWend_Matern", param=param,model="Gaussian",covariance=FALSE)
plot(cc,ylab="Cov",lwd=2,,main="GenWend covariance",type="l")

################################################################
###
### Example 3. Semivariogram of a Tukeyh random field with underlying 
### Generalized Wendland correlation model
###
###############################################################
CorrParam("GenWend")
NuisParam("Tukeyh")
x = seq(0,1,0.005)
param=list(sill=1,smooth=1,scale=0.5,nugget=0,power2=5,tail=0.1,mean=0)
cc= GeoCorrFct(x=x, corrmodel="GenWend", param=param,model="Tukeyh",variogram=TRUE)
plot(cc,ylab="Corr",lwd=2,main="Tukey semivariogram",type="l")

################################################################
###
### Example 4. Semi-Variogram of a LoggGaussian random field with underlying 
### Kummer correlation model
###
###############################################################
CorrParam("Kummer")
NuisParam("LogGaussian")
# GenWend Matern Parameters 
param=list(smooth=1,sill=0.5,scale=0.1,nugget=0,power2=1,mean=0)
cc= GeoCorrFct(x=x, corrmodel="Kummer", param=param,model="LogGaussian",
 ,covariance=TRUE,variogram=TRUE)
plot(cc,ylab="Semivario",lwd=2,
 main="LogGaussian semivariogram",type="l")


################################################################
###
### Example 5. Covariance of Poisson random field with underlying 
### Matern correlation model
###
###############################################################
CorrParam("Matern")
NuisParam("Poisson")
x = seq(0,1,0.005)
param=list(scale=0.6/3,nugget=0,smooth=0.5,mean=2)
cc= GeoCorrFct(x=x, corrmodel="Matern", param=param,model="Poisson",covariance=TRUE)
plot(cc,ylab="Cov",lwd=2,
 main="Poisson covariance",type="l")

################################################################
###
### Example 6. Space time semivariogram of a Gaussian random field 
### with separable Matern correlation model
###
###############################################################

## spatial and temporal distances 
h<-seq(0,3,by=0.04)
times<-seq(0,3,by=0.04)

# Correlation Parameters for the space time separable Matern model 
CorrParam("Matern")
NuisParam("Gaussian")
# Matern Parameters 
param=list(sill=1,scale_s=0.6/3,scale_t=0.5,nugget=0,mean=0,smooth_s=1.5,smooth_t=0.5)
cc= GeoCorrFct(x=h,t=times,corrmodel="Matern_Matern", param=param,
 model="Gaussian",variogram=TRUE)
plot(cc,lwd=2,type="l")


################################################################
###
### Example 7. Correlation of a bivariate Gaussian random field 
### with underlying separable bivariate Matern correlation model
###
###############################################################
# Define the spatial distances
x = seq(0,1,0.005)
 #Correlation Parameters for the bivariate sep Matern model 
CorrParam("Bi_Matern")
 #Matern Parameters 
param=list(sill_1=1,sill_2=1,smooth_1=0.5,smooth_2=1,smooth_12=0.75,
 scale_1=0.2/3, scale_2=0.2/3, scale_12=0.2/3,
 mean_1=0,mean_2=0,nugget_1=0,nugget_2=0,pcol=-0.2)
cc= GeoCorrFct(x=x, corrmodel="Bi_Matern", param=param,model="Gaussian")
plot(cc,ylab="corr",lwd=2,type="l")
}

=head1 NAME

Gravity -- compute the earth's gravity field

=head1 SYNOPSIS

B<Gravity> [ B<-n> I<name> ] [ B<-d> I<dir> ]
[ B<-N> I<Nmax> ] [ B<-M> I<Mmax> ]
[ B<-G> | B<-D> | B<-A> | B<-H> ] [ B<-c> I<lat> I<h> ]
[ B<-w> ] [ B<-p> I<prec> ]
[ B<-v> ]
[ B<--comment-delimiter> I<commentdelim> ]
[ B<--version> | B<-h> | B<--help> ]
[ B<--input-file> I<infile> | B<--input-string> I<instring> ]
[ B<--line-separator> I<linesep> ]
[ B<--output-file> I<outfile> ]

=head1 DESCRIPTION

B<Gravity> reads in positions on standard input and prints out the
gravitational field on standard output.

The input line is of the form I<lat> I<lon> I<h>.  I<lat> and I<lon> are
the latitude and longitude expressed as decimal degrees or degrees,
minutes, and seconds; for details on the allowed formats for latitude
and longitude, see the C<GEOGRAPHIC COORDINATES> section of
GeoConvert(1).  I<h> is the height above the ellipsoid in meters; this
quantity is optional and defaults to 0.  Alternatively, the gravity
field can be computed at various points on a circle of latitude
(constant I<lat> and I<h>) via the B<-c> option; in this case only the
longitude should be given on the input lines.  The quantities printed
out are governed by the B<-G> (default), B<-D>, B<-A>, or B<-H> options.

All the supported gravity models, except for grs80, use WGS84 as the
reference ellipsoid I<a> = 6378137 m, I<f> = 1/298.257223563, I<omega> =
7292115e-11 rad/s, and I<GM> = 3986004.418e8 m^3/s^2.

=head1 OPTIONS

=over

=item B<-n> I<name>

use gravity field model I<name> instead of the default C<egm96>.  See
L</MODELS>.

=item B<-d> I<dir>

read gravity models from I<dir> instead of the default.  See
L</MODELS>.

=item B<-N> I<Nmax>

limit the degree of the model to I<Nmax>.

=item B<-M> I<Mmax>

limit the order of the model to I<Mmax>.

=item B<-G>

compute the acceleration due to gravity (including the centrifugal
acceleration due the the earth's rotation) B<g>.  The output consists of
I<gx> I<gy> I<gz> (all in m/s^2), where the I<x>, I<y>, and I<z>
components are in easterly, northerly, and up directions, respectively.
Usually I<gz> is negative.

=item B<-D>

compute the gravity disturbance B<delta> = B<g> - B<gamma>, where
B<gamma> is the "normal" gravity due to the reference ellipsoid .  The
output consists of I<deltax> I<deltay> I<deltaz> (all in mGal, 1 mGal =
10^-5 m/s^2), where the I<x>, I<y>, and I<z> components are in easterly,
northerly, and up directions, respectively.  Note that I<deltax> =
I<gx>, because I<gammax> = 0.

=item B<-A>

computes the gravitational anomaly.  The output consists of 3 items
I<Dg01> I<xi> I<eta>, where I<Dg01> is in mGal (1 mGal = 10^-5 m/s^2)
and I<xi> and I<eta> are in arcseconds.  The gravitational anomaly
compares the gravitational field B<g> at I<P> with the normal gravity
B<gamma> at I<Q> where the I<P> is vertically above I<Q> and the
gravitational potential at I<P> equals the normal potential at I<Q>.
I<Dg01> gives the difference in the magnitudes of these two vectors and
I<xi> and I<eta> give the difference in their directions (as northerly
and easterly components).  The calculation uses a spherical
approximation to match the results of the NGA's synthesis programs.

=item B<-H>

compute the height of the geoid above the reference ellipsoid (in
meters).  In this case, I<h> should be zero.  The results accurately
match the results of the NGA's synthesis programs.  GeoidEval(1) can
compute geoid heights much more quickly by interpolating on a grid of
precomputed results; however the results from GeoidEval(1) are only
accurate to a few millimeters.

=item B<-c> I<lat> I<h>

evaluate the field on a circle of latitude given by I<lat> and I<h>
instead of reading these quantities from the input lines.  In this case,
B<Gravity> can calculate the field considerably more quickly.  If geoid
heights are being computed (the B<-H> option), then I<h> must be zero.

=item B<-w>

toggle the longitude first flag (it starts off); if the flag is on, then
on input and output, longitude precedes latitude (except that, on input,
this can be overridden by a hemisphere designator, I<N>, I<S>, I<E>,
I<W>).

=item B<-p> I<prec>

set the output precision to I<prec>.  By default I<prec> is 5 for
acceleration due to gravity, 3 for the gravity disturbance and anomaly,
and 4 for the geoid height.

=item B<-v>

print information about the gravity model on standard error before
processing the input.

=item B<--comment-delimiter> I<commentdelim>

set the comment delimiter to I<commentdelim> (e.g., "#" or "//").  If
set, the input lines will be scanned for this delimiter and, if found,
the delimiter and the rest of the line will be removed prior to
processing and subsequently appended to the output line (separated by a
space).

=item B<--version>

print version and exit.

=item B<-h>

print usage, the default gravity path and name, and exit.

=item B<--help>

print full documentation and exit.

=item B<--input-file> I<infile>

read input from the file I<infile> instead of from standard input; a file
name of "-" stands for standard input.

=item B<--input-string> I<instring>

read input from the string I<instring> instead of from standard input.
All occurrences of the line separator character (default is a semicolon)
in I<instring> are converted to newlines before the reading begins.

=item B<--line-separator> I<linesep>

set the line separator character to I<linesep>.  By default this is a
semicolon.

=item B<--output-file> I<outfile>

write output to the file I<outfile> instead of to standard output; a
file name of "-" stands for standard output.

=back

=head1 MODELS

B<Gravity> computes the gravity field using one of the following models

    egm84, earth gravity model 1984.  See
      https://earth-info.nga.mil/index.php?dir=wgs84&action=wgs84#tab_egm84
    egm96, earth gravity model 1996.  See
      https://earth-info.nga.mil/index.php?dir=wgs84&action=wgs84#tab_egm96
    egm2008, earth gravity model 2008.  See
      https://earth-info.nga.mil/index.php?dir=wgs84&action=wgs84#tab_egm2008
    wgs84, world geodetic system 1984.  This returns the normal
      gravity for the WGS84 ellipsoid.
    grs80, geodetic reference system 1980.  This returns the normal
      gravity for the GRS80 ellipsoid.

These models approximate the gravitation field above the surface of the
earth.  By default, the C<egm96> gravity model is used.  This may
changed by setting the environment variable
C<GEOGRAPHICLIB_GRAVITY_NAME> or with the B<-n> option.

The gravity models will be loaded from a directory specified at compile
time.  This may changed by setting the environment variables
C<GEOGRAPHICLIB_GRAVITY_PATH> or C<GEOGRAPHICLIB_DATA>, or with the
B<-d> option.  The B<-h> option prints the default gravity path and
name.  Use the B<-v> option to ascertain the full path name of the data
file.

Instructions for downloading and installing gravity models are available
at L<https://geographiclib.sourceforge.io/C++/doc/gravity.html#gravityinst>.

=head1 ENVIRONMENT

=over

=item B<GEOGRAPHICLIB_GRAVITY_NAME>

Override the compile-time default gravity name of C<egm96>.  The B<-h>
option reports the value of B<GEOGRAPHICLIB_GRAVITY_NAME>, if defined,
otherwise it reports the compile-time value.  If the B<-n> I<name>
option is used, then I<name> takes precedence.

=item B<GEOGRAPHICLIB_GRAVITY_PATH>

Override the compile-time default gravity path.  This is typically
C</usr/local/share/GeographicLib/gravity> on Unix-like systems and
C<C:/ProgramData/GeographicLib/gravity> on Windows systems.  The B<-h>
option reports the value of B<GEOGRAPHICLIB_GRAVITY_PATH>, if defined,
otherwise it reports the compile-time value.  If the B<-d> I<dir> option
is used, then I<dir> takes precedence.

=item B<GEOGRAPHICLIB_DATA>

Another way of overriding the compile-time default gravity path.  If it
is set (and if B<GEOGRAPHICLIB_GRAVITY_PATH> is not set), then
$B<GEOGRAPHICLIB_DATA>/gravity is used.

=back

=head1 ERRORS

An illegal line of input will print an error message to standard output
beginning with C<ERROR:> and causes B<Gravity> to return an exit
code of 1.  However, an error does not cause B<Gravity> to
terminate; following lines will be converted.

=head1 EXAMPLES

The gravity field from EGM2008 at the top of Mount Everest

    echo 27:59:17N 86:55:32E 8820 | Gravity -n egm2008
    => -0.00001 0.00103 -9.76782

=head1 SEE ALSO

GeoConvert(1), GeoidEval(1), geographiclib-get-gravity(8).

=head1 AUTHOR

B<Gravity> was written by Charles Karney.

=head1 HISTORY

B<Gravity> was added to GeographicLib, L<https://geographiclib.sourceforge.io>,
in version 1.16.

\name{getInvC}
\alias{getInvC}
\encoding{UTF-8}
\title{Utility function for kriging computation}
\description{Utility function that computes
 (\eqn{\Sigma^{-1} \mathbf{CC}}) where \eqn{\Sigma} is a covariance matrix and 
 \eqn{\mathbf{CC}} is the matrix of of covariances between the observed location and the locations to predict.
 Optionally, the quadratic form
 \eqn{\mathbf{cc}^\top \Sigma^{-1} \mathbf{cc}} can be computed. Both dense and sparse matrix representations are supported.
}
\usage{
getInvC(covmatrix, CC, mse = TRUE)
}
\arguments{
 \item{covmatrix}{A covariance matrix:
 \describe{
 \item{\code{covmatrix}}{A covariance matrix (dense or sparse).}
 \item{\code{sparse}}{Logical scalar indicating whether the matrix
 is stored in sparse format.}
 }
 }
 \item{CC}{Numeric matrix to be multiplied by the inverse of the
 covariance matrix (length must match the dimension of
 \code{covmatrix}).}
 \item{mse}{Logical. If \code{TRUE} (default) the quadratic form
 \eqn{\mathbf{cc}^\top \Sigma^{-1} \mathbf{cc}} is also returned.}
}
\value{
 A list with components
 \describe{
 \item{\code{a}}{Numeric vector: the product
 \eqn{\Sigma^{-1} \mathbf{cc}}.}
 \item{\code{b}}{Numeric scalar: the quadratic form
 \eqn{\mathbf{cc}^\top \Sigma^{-1} \mathbf{cc}} (only if
 \code{mse = TRUE}).}
 }
}
\details{
 For dense matrices the function uses the Cholesky decomposition provided
 by \code{FastGP::rcppeigen_get_chol}. 
 For sparse matrices (class \code{spam}) the factorisation is performed
 with \code{spam::chol.spam}. 
 If the covariance matrix is not positive definite an error is thrown.
}
\author{
 Moreno Bevilacqua, \email{moreno.bevilacqua@uai.cl}
}

\keyword{Kriging}

=head1 NAME

Planimeter -- compute the area of geodesic polygons

=head1 SYNOPSIS

B<Planimeter> [ B<-r> ] [ B<-s> ] [ B<-l> ] [ B<-e> I<a> I<f> ]
[ B<-w> ] [ B<-p> I<prec> ] [ B<-G> | B<-Q> | B<-R> ] [ B<-E> ]
[ B<--geoconvert-input> ]
[ B<--comment-delimiter> I<commentdelim> ]
[ B<--version> | B<-h> | B<--help> ]
[ B<--input-file> I<infile> | B<--input-string> I<instring> ]
[ B<--line-separator> I<linesep> ]
[ B<--output-file> I<outfile> ]

=head1 DESCRIPTION

Measure the area of a geodesic polygon.  Reads polygon vertices from
standard input, one per line.  Vertices are be given as latitude and
longitude.  By default latitude precedes longitude, however this
convention is reversed with the B<-w> flag and a hemisphere designator
(I<N>, I<S>, I<E>, I<W>) can be used to disambiguate the coordinates.
The end of input, a blank line, or a line which can't be interpreted
as a vertex signals the end of one polygon and the start of the next.
For each polygon print a summary line with the number of points, the
perimeter (in meters), and the area (in meters^2).

The edges of the polygon are given by the I<shortest> geodesic (or
rhumb line) between consecutive vertices.  In certain cases, there may
be two or many such shortest path, and in that case, the polygon is
not uniquely specified by its vertices.  For geodesics, this only
happens with very long edges (for the WGS84 ellipsoid, any edge
shorter than 19970 km is uniquely specified by its end points).  In
such cases, insert an additional vertex near the middle of the long
edge to define the boundary of the polygon.

By default, polygons traversed in a counter-clockwise direction return a
positive area and those traversed in a clockwise direction return a
negative area.  This sign convention is reversed if the B<-r> option is
given.

Of course, encircling an area in the clockwise direction is equivalent
to encircling the rest of the ellipsoid in the counter-clockwise
direction.  The default interpretation used by B<Planimeter> is the one
that results in a smaller magnitude of area; i.e., the magnitude of the
area is less than or equal to one half the total area of the ellipsoid.
If the B<-s> option is given, then the interpretation used is the one
that results in a positive area; i.e., the area is positive and less
than the total area of the ellipsoid.

Arbitrarily complex polygons are allowed.  In the case of
self-intersecting polygons the area is accumulated "algebraically",
e.g., the areas of the 2 loops in a figure-8 polygon will partially
cancel.  Polygons may include one or both poles.  There is no need to
close the polygon.

=head1 OPTIONS

=over

=item B<-r>

toggle whether counter-clockwise traversal of the polygon returns a
positive (the default) or negative result.

=item B<-s>

toggle whether to return a signed result (the default) or not.

=item B<-l>

toggle whether the vertices represent a polygon (the default) or a
polyline.  For a polyline, the number of points and the length of the
path joining them is returned; the path is not closed and the area is
not reported.

=item B<-e> I<a> I<f>

specify the ellipsoid via the equatorial radius, I<a> and
the flattening, I<f>.  Setting I<f> = 0 results in a sphere.  Specify
I<f> E<lt> 0 for a prolate ellipsoid.  A simple fraction, e.g., 1/297,
is allowed for I<f>.  By default, the WGS84 ellipsoid is used, I<a> =
6378137 m, I<f> = 1/298.257223563.  If entering vertices as UTM/UPS or
MGRS coordinates, use the default ellipsoid, since the conversion of
these coordinates to latitude and longitude always uses the WGS84
parameters.

=item B<-w>

toggle the longitude first flag (it starts off); if the flag is on, then
when reading geographic coordinates, longitude precedes latitude (this
can be overridden by a hemisphere designator, I<N>, I<S>, I<E>, I<W>).

=item B<-p> I<prec>

set the output precision to I<prec> (default 6); the perimeter is given
(in meters) with I<prec> digits after the decimal point; the area is
given (in meters^2) with (I<prec> - 5) digits after the decimal point.

=item B<-G>

the edges joining the vertices are geodesics.  This is the default
option and is recommended for terrestrial applications.  This option,
B<-G>, and the following two options, B<-Q> and B<-R>, are mutually
exclusive.

=item B<-Q>

map the points to the authalic sphere and compute the area of the
resulting spherical polygon.  The area will be reasonable accurate
I<provided> that the edges are sufficiently short.  The perimeter
calculation is not accurate.

=item B<-R>

the edges joining the vertices are rhumb lines instead of geodesics.

=item B<-E>

use the exact equations for the geodesic B<-G>, authalic B<-Q>, and
rhumb B<-R> calculations instead of series expansions.  For the
geodesic and rhumb methods, the area is computed by applying discrete
sine transforms to the integrand in the expression for the area.
These are more accurate, albeit slower, than the (default) series
expansions for |I<f>| E<gt> 0.02 and will give high accuracy for -99
E<lt> I<f> E<lt> 0.99.  It is not necessary to specify this option for
terrestrial applications.

=item B<--geoconvert-input>

The input lines are interpreted in the same way as GeoConvert(1)
allowing the coordinates for the vertices to be given as UTM/UPS or
MGRS coordinates, as well as latitude and longitude.  B<CAUTION>:
GeoConvert assumes the coordinates refer to the WGS84 ellipsoid
(disregarding the B<-e> flag) and MGRS coordinates signify the center
of the corresponding MGRS square.

=item B<--comment-delimiter> I<commentdelim>

set the comment delimiter to I<commentdelim> (e.g., "#" or "//").  If
set, the input lines will be scanned for this delimiter and, if found,
the delimiter and the rest of the line will be removed prior to
processing.  For a given polygon, the last such string found will be
appended to the output line (separated by a space).

=item B<--version>

print version and exit.

=item B<-h>

print usage and exit.

=item B<--help>

print full documentation and exit.

=item B<--input-file> I<infile>

read input from the file I<infile> instead of from standard input; a file
name of "-" stands for standard input.

=item B<--input-string> I<instring>

read input from the string I<instring> instead of from standard input.
All occurrences of the line separator character (default is a semicolon)
in I<instring> are converted to newlines before the reading begins.

=item B<--line-separator> I<linesep>

set the line separator character to I<linesep>.  By default this is a
semicolon.

=item B<--output-file> I<outfile>

write output to the file I<outfile> instead of to standard output; a
file name of "-" stands for standard output.

=back

=head1 EXAMPLES

Example (the area of the 100km MGRS square 18SWK)

   Planimeter --geoconvert-input <<EOF
   18n 500000 4400000
   18n 600000 4400000
   18n 600000 4500000
   18n 500000 4500000
   EOF
   => 4 400139.532959 10007388597.2

The following code takes the output from gdalinfo and reports the area
covered by the data (assuming the edges of the image are geodesics).

   #! /bin/sh
   grep -E '^((Upper|Lower) (Left|Right)|Center) ' |
   sed -e 's/d /d/g' -e "s/' /'/g" | tr -s '(),\r\t' ' ' | awk '{
       if ($1 $2 == "UpperLeft")
           ul = $6 " " $5;
       else if ($1 $2 == "LowerLeft")
           ll = $6 " " $5;
       else if ($1 $2 == "UpperRight")
           ur = $6 " " $5;
       else if ($1 $2 == "LowerRight")
           lr = $6 " " $5;
       else if ($1 == "Center") {
           printf "%s\n%s\n%s\n%s\n\n", ul, ll, lr, ur;
           ul = ll = ur = lr = "";
       }
   }
   ' | Planimeter | cut -f3 -d' '

=head1 ACCURACY

Using the B<-G> option (the default), the accuracy was estimated by
computing the error in the area for 10^7 approximately regular
polygons on the WGS84 ellipsoid.  The centers and the orientations of
the polygons were uniformly distributed, the number of vertices was
log-uniformly distributed in [3, 300], and the center to vertex
distance log-uniformly distributed in [0.1 m, 9000 km].

The maximum error in the perimeter was 200 nm, and the maximum error
in the area was

   0.0013 m^2 for perimeter < 10 km
   0.0070 m^2 for perimeter < 100 km
   0.070 m^2 for perimeter < 1000 km
   0.11 m^2 for all perimeters

=head1 SEE ALSO

GeoConvert(1), GeodSolve(1).

An online version of this utility is availbable at
L<https://geographiclib.sourceforge.io/cgi-bin/Planimeter>.

The algorithm for the area of geodesic polygon is
given in Section 6 of C. F. F. Karney, I<Algorithms for geodesics>,
J. Geodesy 87, 43-55 (2013);
DOI L<https://doi.org/10.1007/s00190-012-0578-z>;
addenda: L<https://geographiclib.sourceforge.io/geod-addenda.html>.

The algorithm for the area of a rhumb polygon is given in Section 3 of
C. F. F. Karney, I<The area of rhumb polygons>,
Stud. Geophys. Geod. 68(3--4), 99--120 (2024);
DOI L<https://doi.org/10.1007/s11200-024-0709-z>.

=head1 AUTHOR

B<Planimeter> was written by Charles Karney.

=head1 HISTORY

B<Planimeter> was added to GeographicLib,
L<https://geographiclib.sourceforge.io>, in version 1.4.

\name{CorrelationPar}
\alias{CorrelationPar}
\encoding{UTF-8}
\title{Lists the Parameters of a Correlation Model}
\description{
 Subroutine called by InitParam and other procedures.
 The procedure returns a list with the parameters of a given
 correlation model.
}
\usage{
CorrelationPar(corrmodel)
}
\arguments{
 \item{corrmodel}{Integer; an integer associated to a given correlation model.}
}

\value{Return a vector string of correlation parameters.}


\details{The function returns a list with the Parameters of a Correlation Model}

\seealso{\code{\link{GeoFit}}}



\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}


\keyword{Composite}

\name{GeoNeighIndex}
\alias{GeoNeighIndex}
\encoding{UTF-8}
\title{Spatial, spatio-temporal, or bivariate nearest-neighbour indices}
\description{
Builds directed nearest-neighbour candidate pairs for spatial, spatio-temporal,
or bivariate data. Candidate pairs can optionally be thinned using independent
Bernoulli sampling or an exact fixed-budget, stratum-wise design.
}
\usage{
GeoNeighIndex(coordx=NULL, coordy=NULL, coordz=NULL, coordt=NULL,
 coordx_dyn=NULL, distance="Eucl", neighb=4,
 maxdist=NULL, maxtime=1, radius=1,
 bivariate=FALSE, p_neighb=1,
 thin_method="bernoulli", check.duplicates=FALSE)
}
\arguments{
 \item{coordx}{A numeric \eqn{N \times 2}{N x 2} or \eqn{N \times 3}{N x 3}
 coordinate matrix. Longitude/latitude coordinates can be supplied for spherical
 distances. It may be omitted when \code{coordx_dyn} is supplied.}

 \item{coordy}{Optional numeric vector giving a second spatial coordinate when
 coordinates are supplied separately.}

 \item{coordz}{Optional numeric vector giving a third spatial coordinate when
 coordinates are supplied separately.}

 \item{coordt}{A numeric vector giving the temporal coordinates. Optional
 argument, default is \code{NULL}; if \code{NULL}, a purely spatial random
 field is expected. Temporal coordinates may be irregularly spaced; temporal
 lags are computed from the supplied coordinate values.}

 \item{coordx_dyn}{A list of numeric coordinate matrices. For spatio-temporal
 data, the list must contain one coordinate matrix per element of \code{coordt}.
 For bivariate data with different spatial supports, it must contain the two
 variable-specific coordinate matrices. Optional argument, default is
 \code{NULL}.}

 \item{distance}{Character string specifying the spatial distance. Default is
 \code{"Eucl"}. See \code{\link{GeoFit}}.}

 \item{neighb}{A positive integer giving the nearest-neighbour candidate size.
 In the bivariate case, a length-three positive integer vector can be supplied
 for within-variable 1, cross-variable, and within-variable 2 pairs.}

 \item{maxdist}{Optional non-negative maximum spatial distance. In the bivariate
 case, a length-three vector can be supplied for the three pair types.}

 \item{maxtime}{A non-negative numeric value giving the maximum temporal
 distance in the same units as \code{coordt}. Spatio-temporal candidate pairs
 satisfy the corresponding temporal-distance restriction; see Details.}

 \item{radius}{Numeric radius used for spherical distances. Default is
 \code{1}.}

 \item{bivariate}{Logical; if \code{FALSE} (default), construct univariate
 spatial or spatio-temporal pairs. If \code{TRUE}, construct bivariate spatial
 pairs.}

 \item{p_neighb}{Numeric scalar in \eqn{(0,1]}. If \code{1}, the candidate set
 is returned without thinning. For \code{thin_method="bernoulli"}, it is the
 marginal inclusion probability in the constant-probability design and hence
 controls the retained size in expectation. For
 \code{thin_method="FixedBudget"}, it defines the exact global retained-pair
 budget \eqn{K=\mathrm{round}(p_{neighb} d)}, where \eqn{d} is the candidate
 pair count.}

 \item{thin_method}{Character string selecting the thinning design. Accepted
 values are \code{"bernoulli"} and \code{"FixedBudget"}, case-insensitively.
 The legacy name \code{"TargetBalanced"} is accepted as a backward-compatible
 alias for \code{"FixedBudget"}.}
\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
}
\details{
The unthinned output is a directed candidate graph: \code{rowidx} identifies
candidate targets and \code{colidx} their neighbours. For a purely spatial
configuration with \eqn{N} locations and candidate size \eqn{m}, the regular
case contains approximately \eqn{Nm} directed candidate pairs.

For chordal and geodesic distances, nearest-neighbour ordering is built in
three-dimensional unit-sphere coordinates. Because chordal distance is monotone
in great-circle distance, this gives the correct global nearest-neighbour order,
including near the dateline and poles. Reported lags are then evaluated in the
requested spherical metric and scaled by \code{radius}.

For spatio-temporal data, temporal coordinates are treated as numeric
coordinates, not as equally spaced indices. Candidate temporal separations are
computed as
\deqn{|t_i-t_j|,}
and \code{maxtime} is interpreted as a temporal-distance threshold in the same
units as \code{coordt}. Thus irregularly spaced temporal coordinates are
supported directly.

With \code{thin_method="bernoulli"} and \code{p_neighb < 1}, candidate pairs
are retained independently. In the current constant-weight implementation the
inclusion probability is \code{p_neighb}, so the retained count is random with
expectation approximately \eqn{p_{neighb}d}.

With \code{thin_method="FixedBudget"} and \code{p_neighb < 1}, exactly
\eqn{K=\mathrm{round}(p_{neighb}d)} candidate pairs are retained, subject only
to truncation to the available candidate count. The budget is allocated across
natural candidate strata using proportional quotas with randomized residual
allocation, followed by sampling without replacement within each stratum. For
purely spatial data, strata are the target-specific nearest-neighbour lists.
For spatio-temporal data, target and temporal-lag descriptors define the
strata; for bivariate data, the variable-pair descriptors are also included.
The resulting global retained size is exact.

For \code{thin_method="FixedBudget"}, the exact target is \eqn{round(p d)}. Hence very small \code{p_neighb} values may legitimately return zero retained pairs. This is a valid graph-selection result; \code{GeoFit} rejects such a graph because a pairwise likelihood cannot be estimated without pair contributions. Exact duplicate observation locations are rejected before nearest-neighbor construction.

When \code{p_neighb = 1}, no thinning is applied, independently of
\code{thin_method}.
}
\value{
A list containing the pair indices and distances required by downstream
GeoModels procedures. Depending on the setting, components include:
\item{rowidx}{Directed target indices.}
\item{colidx}{Directed neighbour indices.}
\item{lags}{Spatial distances for the retained pairs.}
\item{lagt}{Temporal distances for spatio-temporal pairs.}
\item{first}{First variable indicator for bivariate pairs.}
\item{second}{Second variable indicator for bivariate pairs.}
\item{maxdist}{Spatial-distance threshold, when stored for the selected case.}
\item{neighb}{Nearest-neighbour candidate size, when stored for the selected case.}

The function returns only the retained pair structure; it does not append
summary fields such as candidate or retained counts. These can be obtained from
the lengths of the returned index vectors.
}
\author{
Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com}, \url{https://sites.google.com/view/moreno-bevilacqua/home},
Victor Morales Onate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamano-Carrillo, \email{chcaaman@ubiobio.cl}, \url{https://www.researchgate.net/profile/Christian-Caamano}
}
\examples{
require(GeoModels)
NN <- 400
coords <- cbind(runif(NN), runif(NN))

## Full directed 5-NN candidate graph.
sel <- GeoNeighIndex(coordx=coords, neighb=5)
length(sel$rowidx)

## Bernoulli thinning: retained size is random with expectation 20 percent
## of the candidate graph.
set.seed(1)
sel_ber <- GeoNeighIndex(coordx=coords, neighb=5,
 p_neighb=0.2, thin_method="bernoulli")
length(sel_ber$rowidx)

## Fixed-budget thinning: the global retained size is exact.
set.seed(1)
sel_fix <- GeoNeighIndex(coordx=coords, neighb=5,
 p_neighb=0.2, thin_method="FixedBudget")
length(sel_fix$rowidx)

## Irregular temporal coordinates: maxtime is a distance threshold.
times <- c(0, 0.25, 1.1, 2.8)
sel_st <- GeoNeighIndex(coordx=coords[1:30, ], coordt=times,
 neighb=3, maxtime=1)
head(sel_st$lagt)
}
\keyword{nearest neighbour}
\keyword{spatio-temporal}

\name{GeoPit}
\alias{GeoPit}
\encoding{UTF-8}
\title{Probability Integral Transform for Fitted GeoModels Objects}
\description{
Transforms observations or prediction targets using the fitted marginal
cumulative distribution function. The output can be represented on the
uniform probability-integral-transform scale or on the standard Gaussian-score
scale.
}
\usage{
GeoPit(object, type = c("Uniform", "Gaussian"), data_to_pred = NULL)
}
\arguments{
  \item{object}{An object of class \code{GeoFit}, \code{GeoKrig}, or
  \code{GeoKrigloc}. The spatial bivariate case is not currently supported.}
  \item{type}{Character string. \code{"Uniform"} returns fitted PIT values
  \eqn{F_i(Y_i)}. \code{"Gaussian"} returns
  \eqn{\Phi^{-1}\{F_i(Y_i)\}}. Partial matching is performed by
  \code{match.arg}.}
  \item{data_to_pred}{Optional observed values at the prediction locations when
  \code{object} is a \code{GeoKrig} or \code{GeoKrigloc} object. This is
  required unless the object already contains a \code{data_to_pred} component.}
}
\details{

For a \code{GeoFit} object, marginal transformations use the complete fitted
mean: a site-specific fixed vector when present, otherwise \eqn{X\beta} with
coefficients \code{mean}, \code{mean1}, and so on matched to the columns of
\code{X}.  The intercept-only case uses scalar \code{mean}.

For a \code{GeoFit} object, site-specific fitted means are reconstructed from
the design matrix and fitted regression coefficients. An externally supplied
fixed mean vector takes precedence. The fitted marginal CDF is then evaluated
at each observation. Copula parameters do not enter the PIT because the
transformation is marginal. The Student-t and skew Student-t transformations
use the fitted degrees of freedom \eqn{1/df} without integer rounding. The
continuous PIT implementation includes the bounded Kumaraswamy families,
Logistic, and SkewLaplace marginals used by the copula diagnostics.

For \code{type = "Gaussian"}, PIT values are clipped only for numerical
protection before applying the standard-normal quantile function. For discrete
marginal models, the current transformation is the ordinary non-randomized PIT
and therefore is not exactly uniform. For \code{PoissonGamma}, the fitted
marginal CDF is the negative-binomial CDF with mean \code{exp(mean)} (or the
site-specific fitted mean) and size parameter \code{shape}. Pearson-residual objects returned by
\code{GeoResiduals()} are rejected; PIT diagnostics for discrete models must use
the original fitted object.

For \code{GeoKrig} and \code{GeoKrigloc} objects, predictive PIT is currently
available only for Gaussian models and requires predictive MSE values.  Linear
kriging for the non-Gaussian models does not by itself define the full conditional
predictive distribution needed for a valid PIT.
}
\value{
The input object with its \code{data} component replaced by the transformed
values.
}
\examples{
\dontrun{
fit_u <- GeoPit(fit, type = "Uniform")
fit_z <- GeoPit(fit, type = "Gaussian")
}
}
\keyword{spatial}
\keyword{distribution}

=head1 NAME

GeoidEval -- look up geoid heights

=head1 SYNOPSIS

B<GeoidEval> [ B<-n> I<name> ] [ B<-d> I<dir> ] [ B<-l> ]
[ B<-a> | B<-c> I<south> I<west> I<north> I<east> ] [ B<-w> ]
[ B<-z> I<zone> ] [ B<--msltohae> ] [ B<--haetomsl> ]
[ B<-v> ]
[ B<--comment-delimiter> I<commentdelim> ]
[ B<--version> | B<-h> | B<--help> ]
[ B<--input-file> I<infile> | B<--input-string> I<instring> ]
[ B<--line-separator> I<linesep> ]
[ B<--output-file> I<outfile> ]

=head1 DESCRIPTION

B<GeoidEval> reads in positions on standard input and prints out the
corresponding heights of the geoid model above the WGS84 ellipsoid on
standard output.

Positions are given as latitude and longitude, UTM/UPS, or MGRS, in any
of the formats accepted by GeoConvert(1).  (MGRS coordinates signify the
I<center> of the corresponding MGRS square.)  If the B<-z> option is
specified then the specified zone is prepended to each line of input
(which must be in UTM/UPS coordinates).  This allows a file with UTM
eastings and northings in a single zone to be used as standard input.

More accurate results for the geoid height are provided by Gravity(1).
This utility can also compute the direction of gravity accurately.

The height of the geoid above the ellipsoid, I<N>, is sometimes called
the geoid undulation.  It can be used to convert a height above the
ellipsoid, I<h>, to the corresponding height above the geoid (the
orthometric height, roughly the height above mean sea level), I<H>,
using the relations

=over

I<h> = I<N> + I<H>, E<nbsp>E<nbsp>I<H> = -I<N> + I<h>.

=back

=head1 OPTIONS

=over

=item B<-n> I<name>

use I<name> for the geoid model instead of the default C<egm96-5>.  See
L</GEOIDS>.

=item B<-d> I<dir>

read geoid data from I<dir> instead of the default.  See
L</GEOIDS>.

=item B<-l>

use bilinear interpolation instead of cubic.  See
L</INTERPOLATION>.

=item B<-a>

cache the entire data set in memory.  See L</CACHE>.

=item B<-c> I<south> I<west> I<north> I<east>

cache the data bounded by I<south> I<west> I<north> I<east> in memory.
The first two arguments specify the SW corner of the cache and the last
two arguments specify the NE corner.  The B<-w> flag specifies that
longitude precedes latitude for these corners, provided that it appears
before B<-c>.  See L</CACHE>.

=item B<-w>

toggle the longitude first flag (it starts off); if the flag is on, then
when reading geographic coordinates, longitude precedes latitude (this
can be overridden by a hemisphere designator, I<N>, I<S>, I<E>, I<W>).

=item B<-z> I<zone>

prefix each line of input by I<zone>, e.g., C<38n>.  This should be used
when the input consists of UTM/UPS eastings and northings.

=item B<--msltohae>

standard input should include a final token on each line which is
treated as a height (in meters) above the geoid and the output echoes
the input line with the height converted to height above ellipsoid
(HAE).  If B<-z> I<zone> is specified then the I<third> token is treated
as the height; this makes it possible to convert LIDAR data where each
line consists of: easting northing height intensity.

=item B<--haetomsl>

this is similar to B<--msltohae> except that the height token is treated
as a height (in meters) above the ellipsoid and the output echoes the
input line with the height converted to height above the geoid (MSL).

=item B<-v>

print information about the geoid model on standard error before
processing the input.

=item B<--comment-delimiter> I<commentdelim>

set the comment delimiter to I<commentdelim> (e.g., "#" or "//").  If
set, the input lines will be scanned for this delimiter and, if found,
the delimiter and the rest of the line will be removed prior to
processing and subsequently appended to the output line (separated by a
space).

=item B<--version>

print version and exit.

=item B<-h>

print usage, the default path and name for geoid models, and exit.

=item B<--help>

print full documentation and exit.

=item B<--input-file> I<infile>

read input from the file I<infile> instead of from standard input; a file
name of "-" stands for standard input.

=item B<--input-string> I<instring>

read input from the string I<instring> instead of from standard input.
All occurrences of the line separator character (default is a semicolon)
in I<instring> are converted to newlines before the reading begins.

=item B<--line-separator> I<linesep>

set the line separator character to I<linesep>.  By default this is a
semicolon.

=item B<--output-file> I<outfile>

write output to the file I<outfile> instead of to standard output; a
file name of "-" stands for standard output.

=back

=head1 GEOIDS

B<GeoidEval> computes geoid heights by interpolating on the data in a
regularly spaced table (see L</INTERPOLATION>).  The following geoid
grids are available (however, some may not be installed):

                                  bilinear error    cubic error
   name         geoid    grid     max      rms      max      rms
   egm84-30     EGM84    30'      1.546 m  70 mm    0.274 m  14 mm
   egm84-15     EGM84    15'      0.413 m  18 mm    0.021 m  1.2 mm
   egm96-15     EGM96    15'      1.152 m  40 mm    0.169 m  7.0 mm
   egm96-5      EGM96     5'      0.140 m  4.6 mm   .0032 m  0.7 mm
   egm2008-5    EGM2008   5'      0.478 m  12 mm    0.294 m  4.5 mm
   egm2008-2_5  EGM2008   2.5'    0.135 m  3.2 mm   0.031 m  0.8 mm
   egm2008-1    EGM2008   1'      0.025 m  0.8 mm   .0022 m  0.7 mm

By default, the C<egm96-5> geoid model is used.  This may changed by
setting the environment variable C<GEOGRAPHICLIB_GEOID_NAME> or with
the B<-n> option.  The errors listed here are estimates of the
quantization and interpolation errors in the reported heights compared
to the specified geoid.

The geoid model data will be loaded from a directory specified at
compile time.  This may changed by setting the environment variables
C<GEOGRAPHICLIB_GEOID_PATH> or C<GEOGRAPHICLIB_DATA>, or with the
B<-d> option.  The B<-h> option prints the default geoid path and
name.  Use the B<-v> option to ascertain the full path name of the
data file.

Instructions for downloading and installing geoid data are available at
L<https://geographiclib.sourceforge.io/C++/doc/geoid.html#geoidinst>.

B<NOTE>: all the geoids above apply to the WGS84 ellipsoid (I<a> =
6378137 m, I<f> = 1/298.257223563) only.

=head1 INTERPOLATION

Cubic interpolation is used to compute the geoid height unless B<-l> is
specified in which case bilinear interpolation is used.  The cubic
interpolation is based on a least-squares fit of a cubic polynomial to a
12-point stencil

   . 1 1 .
   1 2 2 1
   1 2 2 1
   . 1 1 .

The cubic is constrained to be independent of longitude when evaluating
the height at one of the poles.  Cubic interpolation is considerably
more accurate than bilinear; however it results in small discontinuities
in the returned height on cell boundaries.

=head1 CACHE

By default, the data file is randomly read to compute the geoid heights
at the input positions.  Usually this is sufficient for interactive use.
If many heights are to be computed, use B<-c> I<south> I<west> I<north>
I<east> to notify B<GeoidEval> to read a rectangle of data into memory;
heights within the this rectangle can then be computed without any disk
access.  If B<-a> is specified all the geoid data is read; in the case
of C<egm2008-1>, this requires about 0.5 GB of RAM.  The evaluation of
heights outside the cached area causes the necessary data to be read
from disk.  Use the B<-v> option to verify the size of the cache.

Regardless of whether any cache is requested (with the B<-a> or B<-c>
options), the data for the last grid cell in cached.  This allows
the geoid height along a continuous path to be returned with little
disk overhead.

=head1 ENVIRONMENT

=over

=item B<GEOGRAPHICLIB_GEOID_NAME>

Override the compile-time default geoid name of C<egm96-5>.  The B<-h>
option reports the value of B<GEOGRAPHICLIB_GEOID_NAME>, if defined,
otherwise it reports the compile-time value.  If the B<-n> I<name>
option is used, then I<name> takes precedence.

=item B<GEOGRAPHICLIB_GEOID_PATH>

Override the compile-time default geoid path.  This is typically
C</usr/local/share/GeographicLib/geoids> on Unix-like systems and
C<C:/ProgramData/GeographicLib/geoids> on Windows systems.  The B<-h>
option reports the value of B<GEOGRAPHICLIB_GEOID_PATH>, if defined,
otherwise it reports the compile-time value.  If the B<-d> I<dir> option
is used, then I<dir> takes precedence.

=item B<GEOGRAPHICLIB_DATA>

Another way of overriding the compile-time default geoid path.  If it
is set (and if B<GEOGRAPHICLIB_GEOID_PATH> is not set), then
$B<GEOGRAPHICLIB_DATA>/geoids is used.

=back

=head1 ERRORS

An illegal line of input will print an error message to standard output
beginning with C<ERROR:> and causes B<GeoidEval> to return an exit code
of 1.  However, an error does not cause B<GeoidEval> to terminate;
following lines will be converted.

=head1 ABBREVIATIONS

The geoid is usually approximated by an "earth gravity model". The
models published by the NGA are:

=over

=item B<EGM84>

An earth gravity model published by the NGA in 1984,
L<https://earth-info.nga.mil/index.php?dir=wgs84&action=wgs84#tab_egm84>.

=item B<EGM96>

An earth gravity model published by the NGA in 1996,
L<https://earth-info.nga.mil/index.php?dir=wgs84&action=wgs84#tab_egm96>.

=item B<EGM2008>

An earth gravity model published by the NGA in 2008,
L<https://earth-info.nga.mil/index.php?dir=wgs84&action=wgs84#tab_egm2008>.

=item B<WGS84>

World Geodetic System 1984,
L<https://en.wikipedia.org/wiki/WGS84>.

=item B<HAE>

Height above the WGS84 ellipsoid.

=item B<MSL>

Mean sea level, used as a convenient short hand for the geoid.
(However, typically, the geoid differs by a few meters from mean sea
level.)

=back

=head1 EXAMPLES

The height of the EGM96 geoid at Timbuktu

    echo 16:46:33N 3:00:34W | GeoidEval
    => 28.7068 -0.02e-6 -1.73e-6

The first number returned is the height of the geoid and the 2nd and 3rd
are its slopes in the northerly and easterly directions.

Convert a point in UTM zone 18n from MSL to HAE

   echo 531595 4468135 23 | GeoidEval --msltohae -z 18n
   => 531595 4468135 -10.842

=head1 SEE ALSO

GeoConvert(1), Gravity(1), geographiclib-get-geoids(8).

An online version of this utility is availbable at
L<https://geographiclib.sourceforge.io/cgi-bin/GeoidEval>.

=head1 AUTHOR

B<GeoidEval> was written by Charles Karney.

=head1 HISTORY

B<GeoidEval> was added to GeographicLib,
L<https://geographiclib.sourceforge.io>, in 2009-09.

\name{MatDecomp}
\alias{MatDecomp}
\encoding{UTF-8}
\title{Matrix decomposition}
\description{
 Matrix decomposition.
}
\usage{
MatDecomp(mtx, method)
}
\arguments{
 \item{mtx}{numeric; a square positive or semipositive definite matrix.}
 \item{method}{string; the type of matrix decomposition. 
 Two possible choices: \code{cholesky} and \code{svd}.}
}

\value{Return a matrix decomposition}

\details{ Decomposition of a square positive or positive semidefinite matrix.} 

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}



\keyword{Composite}

\name{CheckST}
\alias{CheckST}
\encoding{UTF-8}
\title{Checking SpaceTime covariance models}
\description{
 The procedure control if the correlation model is spacetime.
}
\usage{
CheckST(numbermodel)
}
\arguments{
 \item{numbermodel}{numeric; the number associated to a given correlation model.}
}

\details{ The function check if the correlation model is spacetime.} 

\value{Returns TRUE or FALSE depending if the correlation model is spacetime or not.} 



\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}



\examples{

library(GeoModels)
CheckST(CkCorrModel("gneiting"))

}

\keyword{Composite}

=head1 NAME

RhumbSolve -- perform rhumb line calculations

=head1 SYNOPSIS

B<RhumbSolve> [ B<-i> | B<-L> I<lat1> I<lon1> I<azi12> ]
[ B<-e> I<a> I<f> ] [ B<-u> ]
[ B<-d> | B<-:> ] [ B<-w> ] [ B<-p> I<prec> ] [ B<-E> ]
[ B<--comment-delimiter> I<commentdelim> ]
[ B<--version> | B<-h> | B<--help> ]
[ B<--input-file> I<infile> | B<--input-string> I<instring> ]
[ B<--line-separator> I<linesep> ]
[ B<--output-file> I<outfile> ]

=head1 DESCRIPTION

The path with constant heading between two points on the ellipsoid at
(I<lat1>, I<lon1>) and (I<lat2>, I<lon2>) is called the rhumb line or
loxodrome.  Its length is I<s12> and the rhumb line has a forward
azimuth I<azi12> along its length.  The quantity I<S12> is the area
between the rhumb line from point 1 to point 2 and the equator; i.e.,
it is the area, measured counter-clockwise, of the geodesic
quadrilateral with corners (I<lat1>,I<lon1>), (0,I<lon1>),
(0,I<lon2>), and (I<lat2>,I<lon2>).  The longitude becomes
indeterminate when a rhumb line passes through a pole, and
B<RhumbSolve> reports NaNs for the longitude and the area in this
case.

B<NOTE:> the rhumb line is B<not> the shortest path between two points;
that is the geodesic and it is calculated by GeodSolve(1).

B<RhumbSolve> operates in one of three modes:

=over

=item 1.

By default, B<RhumbSolve> accepts lines on the standard input containing
I<lat1> I<lon1> I<azi12> I<s12> and prints I<lat2> I<lon2> I<S12> on
standard output.  This is the direct calculation.

=item 2.

With the B<-i> option, B<RhumbSolve> performs the inverse calculation.
It reads lines containing I<lat1> I<lon1> I<lat2> I<lon2> and prints
the values of I<azi12> I<s12> I<S12> for the corresponding shortest
rhumb lines.

=item 3.

Command line arguments B<-L> I<lat1> I<lon1> I<azi12> specify a rhumb
line.  B<RhumbSolve> then accepts a sequence of I<s12> values (one per
line) on standard input and prints I<lat2> I<lon2> I<S12> for each.
This generates a sequence of points on a rhumb line.

=back

=head1 OPTIONS

=over

=item B<-i>

perform an inverse calculation (see 2 above).

=item B<-L> I<lat1> I<lon1> I<azi12>

line mode (see 3 above); generate a sequence of points along the rhumb
line specified by I<lat1> I<lon1> I<azi12>.  The B<-w> flag can be used
to swap the default order of the 2 geographic coordinates, provided that
it appears before B<-L>.

=item B<-e> I<a> I<f>

specify the ellipsoid via the equatorial radius, I<a> and
the flattening, I<f>.  Setting I<f> = 0 results in a sphere.  Specify
I<f> E<lt> 0 for a prolate ellipsoid.  A simple fraction, e.g., 1/297,
is allowed for I<f>.  By default, the WGS84 ellipsoid is used, I<a> =
6378137 m, I<f> = 1/298.257223563.

=item B<-u>

unroll the longitude.  Normally, on output longitudes are reduced to lie
in [-180deg,180deg).  However with this option, the returned longitude
I<lon2> is "unrolled" so that I<lon2> - I<lon1> indicates how often and
in what sense the geodesic has encircled the earth.

=item B<-d>

output angles as degrees, minutes, seconds instead of decimal degrees.

=item B<-:>

like B<-d>, except use : as a separator instead of the d, ', and "
delimiters.

=item B<-w>

on input and output, longitude precedes latitude (except that on input
this can be overridden by a hemisphere designator, I<N>, I<S>, I<E>,
I<W>).

=item B<-p> I<prec>

set the output precision to I<prec> (default 3); I<prec> is the
precision relative to 1 m.  See L</PRECISION>.

=item B<-E>

By default, the rhumb line calculations are carried out using series
expansions valid for |I<f>| E<lt> 0.01.  If B<-E> is supplied, exact
equations for the rhumb line are used and the area integral is
computed with an accurate fit based on this exact equations; these are
valid for arbitrary eccentricities.

=item B<--comment-delimiter> I<commentdelim>

set the comment delimiter to I<commentdelim> (e.g., "#" or "//").  If
set, the input lines will be scanned for this delimiter and, if found,
the delimiter and the rest of the line will be removed prior to
processing and subsequently appended to the output line (separated by a
space).

=item B<--version>

print version and exit.

=item B<-h>

print usage and exit.

=item B<--help>

print full documentation and exit.

=item B<--input-file> I<infile>

read input from the file I<infile> instead of from standard input; a file
name of "-" stands for standard input.

=item B<--input-string> I<instring>

read input from the string I<instring> instead of from standard input.
All occurrences of the line separator character (default is a semicolon)
in I<instring> are converted to newlines before the reading begins.

=item B<--line-separator> I<linesep>

set the line separator character to I<linesep>.  By default this is a
semicolon.

=item B<--output-file> I<outfile>

write output to the file I<outfile> instead of to standard output; a
file name of "-" stands for standard output.

=back

=head1 INPUT

B<RhumbSolve> measures all angles in degrees, all lengths (I<s12>) in
meters, and all areas (I<S12>) in meters^2.  On input angles (latitude,
longitude, azimuth, arc length) can be as decimal degrees or degrees,
minutes, seconds.  For example, C<40d30>, C<40d30'>, C<40:30>, C<40.5d>,
and C<40.5> are all equivalent.  By default, latitude precedes longitude
for each point (the B<-w> flag switches this convention); however on
input either may be given first by appending (or prepending) I<N> or
I<S> to the latitude and I<E> or I<W> to the longitude.  Azimuths are
measured clockwise from north; however this may be overridden with I<E>
or I<W>.

For details on the allowed formats for angles, see the C<GEOGRAPHIC
COORDINATES> section of GeoConvert(1).

=head1 PRECISION

I<prec> gives precision of the output with I<prec> = 0 giving 1 m
precision, I<prec> = 3 giving 1 mm precision, etc.  I<prec> is the
number of digits after the decimal point for lengths.  For decimal
degrees, the number of digits after the decimal point is I<prec> + 5.
For DMS (degree, minute, seconds) output, the number of digits after the
decimal point in the seconds component is I<prec> + 1.  The minimum
value of I<prec> is 0 and the maximum is 10.

=head1 ERRORS

An illegal line of input will print an error message to standard output
beginning with C<ERROR:> and causes B<RhumbSolve> to return an exit code
of 1.  However, an error does not cause B<RhumbSolve> to terminate;
following lines will be converted.

=head1 ACCURACY

The algorithm used by B<RhumbSolve> uses either series expansions or
(if B<-E> is specified) exact formulas for computing the rhumb line
and the area.  These series are formulas are accurate for |I<f>| E<lt>
0.01 and the exact formulas apply for any value of the flattening.
The computation of rhumb lines and the area involves the ratio of
differences and, for nearly east- or west-going rhumb lines, this
might result in a large loss of accuracy.  However, this problem is
avoided by the use of divided differences. For the WGS84 ellipsoid,
the error is about 10 nanometers using either method.

=head1 EXAMPLES

Route from JFK Airport to Singapore Changi Airport:

   echo 40:38:23N 073:46:44W 01:21:33N 103:59:22E |
   RhumbSolve -i -: -p 0

   103:34:58.2 18523563 45921660958919

N.B. This is B<not> the route typically taken by aircraft because it's
considerably longer than the geodesic given by GeodSolve(1).

Waypoints on the route at intervals of 2000km:

   for ((i = 0; i <= 20; i += 2)); do echo ${i}000000;done |
   RhumbSolve -L 40:38:23N 073:46:44W 103:34:58.2 -: -p 0

   40:38:23.0N 073:46:44.0W 0
   36:24:30.3N 051:28:26.4W 9817078307821
   32:10:26.8N 030:20:57.3W 18224745682005
   27:56:13.2N 010:10:54.2W 25358020327741
   23:41:50.1N 009:12:45.5E 31321269267102
   19:27:18.7N 027:59:22.1E 36195163180159
   15:12:40.2N 046:17:01.1E 40041499143669
   10:57:55.9N 064:12:52.8E 42906570007050
   06:43:07.3N 081:53:28.8E 44823504180200
   02:28:16.2N 099:24:54.5E 45813843358737
   01:46:36.0S 116:52:59.7E 45888525219677

=head1 SEE ALSO

GeoConvert(1), GeodSolve(1).

An online version of this utility is availbable at
L<https://geographiclib.sourceforge.io/cgi-bin/RhumbSolve>.

An online version of this utility is availbable at
L<https://geographiclib.sourceforge.io/cgi-bin/RhumbSolve>.

This solution for rhumb line is described in C. F. F. Karney,
I<The area of rhumb polygons>,
Stud. Geophys. Geod. 68(3--4), 99--120 (2024);
DOI L<https://doi.org/10.1007/s11200-024-0709-z>.

The Wikipedia page, Rhumb line,
L<https://en.wikipedia.org/wiki/Rhumb_line>.

=head1 AUTHOR

B<RhumbSolve> was written by Charles Karney.

=head1 HISTORY

B<RhumbSolve> was added to GeographicLib,
L<https://geographiclib.sourceforge.io>, in version 1.37 and
substantially rewritten in version 2.2.

=head1 NAME

IntersectTool -- find the intersection of geodesics

=head1 SYNOPSIS

B<IntersectTool> [ B<-c> | B<-n> | B<-i> | B<-o> |
[ B<-R> I<maxdist> ]
[ B<-e> I<a> I<f>] [ B<-E> ]
[ B<-w> ] [ B<-p> I<prec> ]
[ B<--comment-delimiter> I<commentdelim> ]
[ B<--version> | B<-h> | B<--help> ]
[ B<--input-file> I<infile> | B<--input-string> I<instring> ]
[ B<--line-separator> I<linesep> ]
[ B<--output-file> I<outfile> ]

=head1 DESCRIPTION

B<IntersectTool> finds the intersection of two geodesics I<X> and
I<Y>.  The geodesics may either be specified as a location and an
azimuth, I<latX> I<lonX> I<aziX>, or as the shortest geodesic segment
between two locations, I<latX1> I<lonX1> and I<latX2> I<lonX2>.  The
intersection is then specified as the displacements, I<x> and I<y>,
along the geodesics I<X> and I<Y> from the starting points to the
intersection.  In the case of the intersection of geodesic segments,
the starting point is first point specified for I<X> or I<Y>.

Usually this tool returns the I<closest> intersection defined as the
one that minimizes the "L1" distance, |I<x>| + |I<y>|.  However, it is
possible to specify an "origin" I<x0> and I<y0> when determining
closeness so that the intersection which minimizes |I<x> - I<x0>| +
|I<y> - I<y0>| is returned.

In the case of intersecting segments the origin is taken to be the
midpoints of the segments; I<x0> is half the distance from I<X1> to
I<X2>.  In addition a flag is returned specifying whether the
intersection is "within" the segments.

The tool also returns a "coincidence indicator" I<c>.  This is
typically 0.  However if the geodesics lie on top of one another at
the point of intersection, then I<c> is set to 1, if they are
parallel, and -1, if they are antiparallel.

B<IntersectTool> operates in one of three modes:

=over

=item 1.

With the B<-c> option (the default), B<IntersectTool> accepts lines on
the standard input containing I<latX> I<lonX> I<aziX> I<latY> I<lonY>
I<aziY>, specifying two geodesic lines I<X> and I<Y>, and prints the
location of the closest intersection I<x> I<y> I<c> on standard output.

=item 2.

With the B<-n> option, B<IntersectTool> accepts lines on the standard
input containing I<latX> I<lonX> I<aziX> I<aziY> I<aziY>, specifying a
point where two geodesic lines I<X> and I<Y> intersect, and prints the
location of the next closest intersection I<x> I<y> I<c> on standard
output.

=item 3.

With the B<-i> option, B<IntersectTool> accepts lines on the standard
input containing I<latX1> I<lonX1> I<latX2> I<lonX2> I<latY1> I<lonY1>
I<latY2> I<lonY2>, specifying two geodesic segments I<X1>-I<X2> and
I<Y1>-I<Y2>, and prints I<x> I<y> I<c> I<k> on standard output.  Here
I<k> is a flag in [-4,4] specifying whether the intersection is within
the segments (0) or not (non-zero).  I<x> and I<y> give the distances
from I<X1> and I<Y1> respectively.  I<k> is set to 3 I<kx> + I<ky>
where I<kx> = -1 if I<x> < 0, 0 if 0 <= I<x> <= I<sx>, 1 if I<sx> <
I<x>, and similarly for I<ky>; I<sx> is the length of the segment
I<X1>-I<X2>.

=item 4.

With the B<-o> option, B<IntersectTool> accepts lines on the standard
input containing I<latX> I<lonX> I<aziX> I<latY> I<lonY> I<aziY> I<x0>
I<y0>, specifying two geodesic lines I<X> and I<Y> and two offsets,
and prints I<x> I<y> I<c> on standard output where [I<x>, I<y>] is the
intersection closest to [I<x0>, I<y0>].

=back

=head1 OPTIONS

=over

=item B<-c>

find the closest intersection (see 1 above).

=item B<-n>

find the intersection closest to a given intersection (see 2 above).

=item B<-i>

find the intersection of two geodesic segments (see 3 above).

=item B<-o>

find the closest intersection with an offset.

=item B<-R> I<maxdist>

modifies the four modes to return all the intersections within an L1
distance, I<maxdist>, of the relevant origin: [0, 0] for B<-c> and
B<-n>, the midpoints of the segments for B<-i>, and [I<x0>, I<y0>] for
B<-o>.  For each intersection, I<x> I<y> I<c> I<z> is printed on
standard output.  Here I<z> is the L1 distance of the intersection
from the origin and the intersections are sorted by the distances.  A
line "nan nan 0 nan" is added after the intersections, so that the
output can be associated with the correct lines of the input.  The
number of intersections scales as (I<maxdist>/(pi I<a>))^2.

=item B<-C>

check the intersections.  For each computed intersection, print on
B<standard error> a line I<latX> I<lonX> I<latY> I<lonY> I<sXY> giving
the computed positions of the intersections points on I<X> and I<Y>
and the distance between them.  If B<-w> is specified, the longitude
is given before the latitude.

=item B<-e> I<a> I<f>

specify the ellipsoid via the equatorial radius, I<a> and
the flattening, I<f>.  Setting I<f> = 0 results in a sphere.  Specify
I<f> E<lt> 0 for a prolate ellipsoid.  A simple fraction, e.g., 1/297,
is allowed for I<f>.  By default, the WGS84 ellipsoid is used, I<a> =
6378137 m, I<f> = 1/298.257223563.

=item B<-E>

use "exact" algorithms (based on elliptic integrals) for the geodesic
calculations.  These are more accurate than the (default) series
expansions for |I<f>| E<gt> 0.02.

=item B<-w>

on input, longitude precedes latitude (except that on input this can
be overridden by a hemisphere designator, I<N>, I<S>, I<E>, I<W>).

=item B<-p> I<prec>

set the output precision to I<prec> (default 3); I<prec> is the
precision relative to 1 m.  See L</PRECISION>.

=item B<--comment-delimiter> I<commentdelim>

set the comment delimiter to I<commentdelim> (e.g., "#" or "//").  If
set, the input lines will be scanned for this delimiter and, if found,
the delimiter and the rest of the line will be removed prior to
processing and subsequently appended to the output line (separated by a
space).

=item B<--version>

print version and exit.

=item B<-h>

print usage and exit.

=item B<--help>

print full documentation and exit.

=item B<--input-file> I<infile>

read input from the file I<infile> instead of from standard input; a file
name of "-" stands for standard input.

=item B<--input-string> I<instring>

read input from the string I<instring> instead of from standard input.
All occurrences of the line separator character (default is a semicolon)
in I<instring> are converted to newlines before the reading begins.

=item B<--line-separator> I<linesep>

set the line separator character to I<linesep>.  By default this is a
semicolon.

=item B<--output-file> I<outfile>

write output to the file I<outfile> instead of to standard output; a
file name of "-" stands for standard output.

=back

=head1 INPUT

B<IntersectTool> measures all angles in degrees and all lengths in
meters.  On input angles (latitude, longitude, azimuth) can be as
decimal degrees or degrees, minutes, seconds.  For example, C<40d30>,
C<40d30'>, C<40:30>, C<40.5d>, and C<40.5> are all equivalent.  By
default, latitude precedes longitude for each point (the B<-w> flag
switches this convention); however either may be given first by
appending (or prepending) I<N> or I<S> to the latitude and I<E> or
I<W> to the longitude.  Azimuths are measured clockwise from north;
however this may be overridden with I<E> or I<W>.

For details on the allowed formats for angles, see the C<GEOGRAPHIC
COORDINATES> section of GeoConvert(1).

=head1 PRECISION

I<prec> gives precision of the output with I<prec> = 0 giving 1 m
precision, I<prec> = 3 giving 1 mm precision, etc.  I<prec> is the
number of digits after the decimal point for lengths.  The latitude
and longitude printed to standard error with the B<-C> option are
given in decimal degrees with I<prec> + 5 digits after the decimal
point.  The minimum value of I<prec> is 0 and the maximum is 10.

=head1 ERRORS

An illegal line of input will print an error message to standard output
beginning with C<ERROR:> and causes B<IntersectTool> to return an exit code
of 1.  However, an error does not cause B<IntersectTool> to terminate;
following lines will be converted.

=head1 ACCURACY

This tool will give nearly full double precision accuracy for |I<f>|
E<lt> 0.02.  If the B<-E> option is given, full accuracy is achieved
for -1/4 E<lt> f E<lt> 1/5.  The tool had not been tested outside this
range.

=head1 EXAMPLES

A vessel leaves Plymouth 50N 4W on a geodesic path with initial
heading 147.7W.  When will it first cross the equator?

   echo 50N 4W 147.7W 0 0 90 | IntersectTool -c -p 0 -C

   6058049 -3311253 0
   0.00000 -29.74549 -0.00000 -29.74549 0

Answer: after 6058km at longitude 29.7W.  When will it cross the date
line, longitude 180E?  Here we need to use B<-R> because there a
closer intersection on the prime meridian:

   echo 50N 4W 147.7W 0 180 0 | IntersectTool -c -p 0 -C -R 2.6e7

   -494582 14052230 0 14546812
   53.69260 0.00000 53.69260 0.00000 0
   19529110 -5932344 0 25461454
   -53.51867 180.00000 -53.51867 180.00000 0
   nan nan 0 nan
   nan nan nan nan nan

We want the second result: after 19529 km at latitude 53.5S.

=head1 SEE ALSO

GeoConvert(1), GeodSolve(1).

This solution for intersections is described in C. F. F. Karney,
I<Geodesic intersections>, J. Surveying Eng. B<150>(3), 04024005:1-9
(2024), DOI: L<https://doi.org/10.1061/JSUED2.SUENG-1483>; preprint
L<https://arxiv.org/abs/2308.00495>.  It is based on the work of
S. Baseldga and J. C. Martinez-Llario, I<Intersection and
point-to-line solutions for geodesics on the ellipsoid>,
Stud. Geophys. Geod. B<62>, 353-363 (2018); DOI:
L<https://doi.org/10.1007/s11200-017-1020-z>;

=head1 AUTHOR

B<IntersectTool> was written by Charles Karney.

=head1 HISTORY

B<IntersectTool> was added to GeographicLib,
L<https://geographiclib.sourceforge.io>, in version 2.3.

\name{GeoTestIsotropy}
\alias{GeoTestIsotropy}
\title{Parametric Bootstrap Tests for Spatial Isotropy}
\description{
Performs two parametric-bootstrap tests of spatial isotropy by comparing an
isotropic model under the null hypothesis with a geometrically anisotropic
model under the alternative. The first test (MWCLT) is based on the estimated
anisotropy ratio. The second test (WCLRT) is based on a likelihood-ratio or
composite-likelihood-ratio statistic. Their null distributions are estimated
by simulation from the fitted isotropic model.
}

\usage{
GeoTestIsotropy(data, coordx,
 start, fixed,
 optimizer = "bobyqa",
 model = "Gaussian",
 corrmodel = "Matern",
 lower = NULL, upper = NULL,
 B = 1000,
 likelihood = NULL,
 type = NULL,
 copula = NULL,
 neighb = 5,
 method = c("cholesky", "TB"),
 L = 10000,
 n = 1,
 angle_starts = c(pi/4, 3*pi/4),
 batch_size = NULL,
 max_attempts = NULL,
 parallel = FALSE,
 ncores = 6,
 progress = TRUE,
 seed = NULL)
}

\arguments{
 \item{data}{Numeric vector of observations.}
 \item{coordx}{Numeric matrix with exactly two columns containing the spatial
 coordinates. The number of rows must equal \code{length(data)}.}
 \item{start}{Named list containing the covariance and marginal parameters to
 be estimated. Parameters \code{angle} and \code{ratio} must not be included,
 because they are handled internally by the isotropy test.}
 \item{fixed}{Named list containing the covariance and marginal parameters kept
 fixed. Together, \code{start} and \code{fixed} must contain exactly the
 parameters required by \code{CorrParam(corrmodel)} and
 \code{NuisParam(model)}.}
 \item{optimizer}{Optimization algorithm passed to \code{GeoFit}.}
 \item{model}{Marginal random-field model.}
 \item{corrmodel}{Spatial correlation model.}
 \item{lower, upper}{Named lists of lower and upper bounds for the parameters
 in \code{start}. Missing bounds are replaced by \code{-Inf} and \code{Inf},
 respectively.}
 \item{B}{Number of valid parametric-bootstrap replications used to calibrate
 each test. Failed fitting attempts are replaced by newly simulated samples.}
 \item{likelihood}{Likelihood type passed to \code{GeoFit}. If both
 \code{likelihood} and \code{type} are \code{NULL}, full likelihood is used for
 at most 2000 observations and marginal pairwise composite likelihood is used
 otherwise.}
 \item{type}{Likelihood type passed to \code{GeoFit}; typically
 \code{"Standard"} for full likelihood or \code{"Pairwise"} for composite
 likelihood. It must be supplied together with \code{likelihood}.}
 \item{copula}{Optional copula specification. With a supported purely spatial
 correlation model, both \code{method="cholesky"} and \code{method="TB"} are
 available.}
 \item{neighb}{Number of nearest neighbours used when
 \code{type = "Pairwise"}. Ignored otherwise.}
 \item{method}{Simulation method under the null hypothesis. Use
 \code{"cholesky"} for exact simulation or \code{"TB"} for approximate
 turning-bands simulation. When \code{copula} is supplied, the corresponding
 latent engine is selected internally by \code{GeoSimCopula}.}
 \item{L}{Number of spectral components used when \code{method = "TB"}.}
 \item{n}{Common number of trials/successes used by models that require an
 integer \code{n}, such as Binomial-type models. It can be a scalar or, where
 supported, a vector with one value per observation. For \code{"Bernoulli"}
 and \code{"Binary"}, \code{n} is internally fixed to 1.}
 \item{angle_starts}{Finite numeric vector of starting angles used for the
 anisotropic fit. All supplied angles are tried (modulo \eqn{\pi}) and the fit
 with the largest objective value is retained. The default uses two starts,
 \eqn{\pi/4} and \eqn{3\pi/4}, for both the observed fit and every bootstrap
 refit.}
 \item{batch_size}{Number of bootstrap datasets simulated and processed at one
 time. If \code{NULL}, a modest batch size is selected automatically. Batching
 limits memory use for large spatial datasets.}
 \item{max_attempts}{Maximum number of simulated bootstrap datasets allowed in
 order to collect exactly \code{B} valid replications. If \code{NULL}, the
 default is \code{max(B + 20, ceiling(1.5 * B))}. The function stops if this
 limit is reached before \code{B} valid replications are obtained.}
 \item{parallel}{Logical; if \code{TRUE}, bootstrap refits are distributed
 across multiple R sessions. With \code{method="TB"}, bootstrap simulation
 may also use multiple workers when supported by the simulation routine.}
 \item{ncores}{Positive integer or \code{NULL}. With \code{parallel=TRUE}, an
 explicit integer requests that many workers, capped by detected cores and the
 current number of jobs. Set \code{ncores=NULL} for automatic selection.}
 \item{progress}{Logical; if \code{TRUE}, display the same
 \pkg{progressr} text progress bar used by other parallel \pkg{GeoModels}
 functions. A single bar is shown for the whole bootstrap and counts valid
 bootstrap replications in both sequential and parallel execution.}
 \item{seed}{Optional finite numeric seed. The previous random-number-generator
 state is restored when the function exits.}
}

\details{
The null and alternative hypotheses are
\deqn{H_0: \mathrm{ratio}=1,}
and
\deqn{H_1: \mathrm{ratio}>1,}
with the anisotropy angle estimated under the alternative. Under \eqn{H_0},
the angle is not identifiable and the ratio is on the boundary of its
parameter space. Consequently, a standard chi-squared reference distribution
is not assumed; both tests are calibrated by parametric bootstrap.

The MWCLT statistic is the unconstrained anisotropy-ratio estimate
\eqn{\widehat\tau}. If \eqn{\widehat\tau_b^*} denotes the corresponding
estimate in bootstrap replication \eqn{b}, the upper-tail bootstrap p-value is
\deqn{\widehat p_\tau = \frac{1 + \sum_{b=1}^B I(\widehat\tau_b^* \geq \widehat\tau)}{B+1}.}

The WCLRT statistic is
\deqn{T_{\mathrm{obs}}=\max\{0,2(\ell_1-\ell_0)\},}
where \eqn{\ell_0} and \eqn{\ell_1} are the maximized likelihoods or composite
likelihoods under the isotropic and anisotropic models. Its bootstrap p-value
is
\deqn{\widehat p_T = \frac{1 + \sum_{b=1}^B I(T_b \geq T_{\mathrm{obs}})}{B+1}.}
The returned components \code{statistic} and \code{pvalue} remain aliases for
this WCLRT statistic and p-value for backward compatibility.

For every observed or bootstrap anisotropic fit, all angles in
\code{angle_starts} are used as starting values and the valid fit with the
largest objective value is retained. During bootstrap refitting, the fitted
observed anisotropy angle is also included as a starting angle. If the best
numerical anisotropic fit has an objective value below the isotropic fit by
more than numerical tolerance, the boundary solution is selected, giving
ratio 1 and WCLRT statistic 0.

For pairwise composite likelihood, the isotropic and anisotropic fits must use
the same retained pairs. The function checks this condition before computing
the WCLRT statistic.

Bootstrap datasets are generated from the fitted isotropic model in batches.
Failed bootstrap fits do not enter the empirical null distribution: new
samples are generated until exactly \code{B} valid replications have been
collected, or until \code{max_attempts} is reached. This makes the denominator
of both bootstrap p-values exactly \code{B+1}.

The argument \code{n} is propagated consistently to \code{GeoFit},
\code{GeoSim}, \code{GeoSimapprox}, and \code{GeoSimCopula}. Direct latent
Gaussian simulation of some non-Gaussian models imposes additional structural
constraints. In particular, direct \code{"Gamma"} simulation requires the
fitted \code{shape} to be a positive integer because it is the number of
squared latent Gaussian fields. For the direct Gamma model used in the latent
construction, it is therefore recommended to keep \code{shape} fixed at a
positive integer during the isotropy test. This integer restriction does not
apply merely because the marginal distribution is Gamma when a supported
copula construction is used.
}

\value{
An invisible list containing, among others:
\item{statistic, pvalue}{Backward-compatible aliases for the observed WCLRT
 statistic and its bootstrap p-value.}
\item{statistic_lrt, pvalue_lrt}{Observed WCLRT statistic and p-value.}
\item{statistic_ratio, pvalue_ratio}{Observed MWCLT anisotropy-ratio statistic
 and p-value.}
\item{pvalues}{Named vector containing the MWCLT and WCLRT p-values.}
\item{ratio_hat, angle_hat}{Estimated anisotropy ratio and angle. The angle is
 \code{NA} when the boundary solution is selected.}
\item{parallel, ncores}{Whether parallel bootstrap refitting was used and the
number of workers selected by the function.}
\item{fit_H0, fit_H1}{Fitted isotropic and best anisotropic \code{GeoFit}
 objects for the observed data.}
\item{B_rep, B_rep_lrt}{The \code{B} valid bootstrap WCLRT statistics.}
\item{B_rep_ratio}{The \code{B} valid bootstrap anisotropy-ratio estimates.}
\item{B_rep_angle}{Bootstrap angle estimates; boundary solutions are
 \code{NA}.}
\item{bootstrap_B}{Requested and obtained number of valid bootstrap
 replications.}
\item{bootstrap_total_attempts}{Total number of simulated datasets needed to
 obtain \code{B} valid replications.}
\item{bootstrap_failed_attempts}{Number of failed attempts that were replaced.}
\item{bootstrap_failures}{Data frame describing failed attempts.}
\item{bootstrap_boundary_solutions}{Number of valid bootstrap replications for
 which the boundary solution was selected.}
\item{observed_multistart_angles}{Starting angles used for the observed H1 fit.}
\item{batch_size, max_attempts}{Bootstrap batching and retry settings used by
 the function.}
}

\examples{
\dontrun{
## Isotropy tests for a Gaussian random field simulated under H0
set.seed(123)
nsite <- 500
coords <- cbind(runif(nsite), runif(nsite))

param <- list(
  mean = 0,
  nugget = 0,
  sill = 1,
  scale = 0.20,
  smooth = 0.5
)

z <- GeoSim(
  coordx = coords,
  corrmodel = "Matern",
  model = "Gaussian",
  param = param
)$data

start <- list(
  mean = 0,
  sill = 1,
  scale = 0.20,
  smooth = 0.5
)
fixed <- list(nugget = 0)

lower <- list(mean = -5, sill = 0.01, scale = 0.01, smooth = 0.05)
upper <- list(mean =  5, sill = 5,    scale = 1,    smooth = 2.5)

iso_test <- GeoTestIsotropy(
  data = z,
  coordx = coords,
  start = start,
  fixed = fixed,
  model = "Gaussian",
  corrmodel = "Matern",
  lower = lower,
  upper = upper,
  B = 99,
  likelihood = "Full",
  type = "Standard",
  method = "cholesky",
  seed = 321
)

iso_test$pvalue_ratio  # MWCLT
iso_test$pvalue_lrt    # WCLRT
}
}

\seealso{
\code{\link{GeoFit}}, \code{\link{GeoSim}},
\code{\link{GeoSimapprox}}, \code{\link{GeoVariogramDir}}
}

\author{
Moreno Bevilacqua \email{moreno.bevilacqua89@gmail.com}
\url{https://sites.google.com/view/moreno-bevilacqua/home}\cr
Victor Morales Onate \email{victor.morales@uv.cl}
\url{https://sites.google.com/site/moralesonatevictor/}\cr
Christian Caamano-Carrillo \email{chcaaman@ubiobio.cl}
\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\keyword{Bootstrap}
\keyword{Composite}
\keyword{HypothesisTest}
\keyword{Spatial}

=head1 NAME

MagneticField -- compute the earth's magnetic field

=head1 SYNOPSIS

B<MagneticField> [ B<-n> I<name> ] [ B<-d> I<dir> ]
[ B<-N> I<Nmax> ] [ B<-M> I<Mmax> ]
[ B<-t> I<time> | B<-c> I<time> I<lat> I<h> ]
[ B<-r> ] [ B<-w> ] [ B<-T> I<tguard> ] [ B<-H> I<hguard> ] [ B<-p> I<prec> ]
[ B<-v> ]
[ B<--comment-delimiter> I<commentdelim> ]
[ B<--version> | B<-h> | B<--help> ]
[ B<--input-file> I<infile> | B<--input-string> I<instring> ]
[ B<--line-separator> I<linesep> ]
[ B<--output-file> I<outfile> ]

=head1 DESCRIPTION

B<MagneticField> reads in times and positions on standard input and
prints out the geomagnetic field on standard output and, optionally, its
rate of change.

The input line is of the form I<time> I<lat> I<lon> I<h>. I<time> is a
date of the form 2012-11-25 (yyyy-mm-dd or yyyy-mm), a fractional year
such as 2012.9, or the string "now".  I<lat> and I<lon> are the
latitude and longitude expressed as decimal degrees or degrees,
minutes, and seconds; for details on the allowed formats for latitude
and longitude, see the C<GEOGRAPHIC COORDINATES> section of
GeoConvert(1).  I<h> is the height above the ellipsoid in meters; this
is optional and defaults to zero.  Alternatively, I<time> can be given
on the command line as the argument to the B<-t> option, in which case
it should not be included on the input lines.  Finally, the magnetic
field can be computed at various points on a circle of latitude
(constant I<time>, I<lat>, and I<h>) via the B<-c> option; in this
case only the longitude should be given on the input lines.

The output consists of the following 7 items:

  the declination (the direction of the horizontal component of
    the magnetic field measured clockwise from north) in degrees,
  the inclination (the direction of the magnetic field measured
    down from the horizontal) in degrees,
  the horizontal field in nanotesla (nT),
  the north component of the field in nT,
  the east component of the field in nT,
  the vertical component of the field in nT (down is positive),
  the total field in nT.

If the B<-r> option is given, a second line is printed giving the rates
of change of these quantities in degrees/yr and nT/yr.

The WGS84 ellipsoid is used, I<a> = 6378137 m, I<f> = 1/298.257223563.

=head1 OPTIONS

=over

=item B<-n> I<name>

use magnetic field model I<name> instead of the default C<wmm2025>.  See
L</MODELS>.

=item B<-d> I<dir>

read magnetic models from I<dir> instead of the default.  See
L</MODELS>.

=item B<-N> I<Nmax>

limit the degree of the model to I<Nmax>.

=item B<-M> I<Mmax>

limit the order of the model to I<Mmax>.

=item B<-t> I<time>

evaluate the field at I<time> instead of reading the time from the input
lines.

=item B<-c> I<time> I<lat> I<h>

evaluate the field on a circle of latitude given by I<time>, I<lat>,
I<h> instead of reading these quantities from the input lines.  In this
case, B<MagneticField> can calculate the field considerably more
quickly.

=item B<-r>

toggle whether to report the rates of change of the field.

=item B<-w>

toggle the longitude first flag (it starts off); if the flag is on, then
on input and output, longitude precedes latitude (except that, on input,
this can be overridden by a hemisphere designator, I<N>, I<S>, I<E>,
I<W>).

=item B<-T> I<tguard>

signal an error if I<time> lies I<tguard> years (default 50 yr) beyond
the range for the model.

=item B<-H> I<hguard>

signal an error if I<h> lies I<hguard> meters (default 500000 m) beyond
the range for the model.

=item B<-p> I<prec>

set the output precision to I<prec> (default 1).  Fields are printed
with precision with I<prec> decimal places; angles use I<prec> + 1
places.

=item B<-v>

print information about the magnetic model on standard error before
processing the input.

=item B<--comment-delimiter> I<commentdelim>

set the comment delimiter to I<commentdelim> (e.g., "#" or "//").  If
set, the input lines will be scanned for this delimiter and, if found,
the delimiter and the rest of the line will be removed prior to
processing and subsequently appended to the output line (separated by a
space).

=item B<--version>

print version and exit.

=item B<-h>

print usage, the default magnetic path and name, and exit.

=item B<--help>

print full documentation and exit.

=item B<--input-file> I<infile>

read input from the file I<infile> instead of from standard input; a file
name of "-" stands for standard input.

=item B<--input-string> I<instring>

read input from the string I<instring> instead of from standard input.
All occurrences of the line separator character (default is a semicolon)
in I<instring> are converted to newlines before the reading begins.

=item B<--line-separator> I<linesep>

set the line separator character to I<linesep>.  By default this is a
semicolon.

=item B<--output-file> I<outfile>

write output to the file I<outfile> instead of to standard output; a
file name of "-" stands for standard output.

=back

=head1 MODELS

B<MagneticField> computes the geomagnetic field using one of the
following models

    wmm2010, the World Magnetic Model 2010, which approximates the
      main magnetic field for the period 2010-2015.  See
      https://ngdc.noaa.gov/geomag/WMM/DoDWMM.shtml
    wmm2015v2, the World Magnetic Model 2015, which approximates the
      main magnetic field for the period 2015-2020.  See
      https://ngdc.noaa.gov/geomag/WMM/DoDWMM.shtml
    wmm2015, a deprecated version of wmm2015v2
    wmm2020, the World Magnetic Model 2020, which approximates the
      main magnetic field for the period 2020-2025.  See
      https://ngdc.noaa.gov/geomag/WMM/DoDWMM.shtml
    wmm2025, the World Magnetic Model 2025, which approximates the
      main magnetic field for the period 2025-2030.  See
      https://www.ncei.noaa.gov/products/world-magnetic-model
    wmmhr2025, the World Magnetic Model High Resolution 2025, a high
      resolution version of wmm2025.  See
      https://www.ncei.noaa.gov/products/world-magnetic-model-high-resolution
    igrf11, the International Geomagnetic Reference Field (11th
      generation), which approximates the main magnetic field for
      the period 1900-2015.  See
      https://ngdc.noaa.gov/IAGA/vmod/igrf.html
    igrf12, the International Geomagnetic Reference Field (12th
      generation), which approximates the main magnetic field for
      the period 1900-2020.  See
      https://ngdc.noaa.gov/IAGA/vmod/igrf.html
    igrf13, the International Geomagnetic Reference Field (13th
      generation), which approximates the main magnetic field for
      the period 1900-2025.  See
      https://ngdc.noaa.gov/IAGA/vmod/igrf.html
    igrf14, the International Geomagnetic Reference Field (14th
      generation), which approximates the main magnetic field for
      the period 1900-2030.  See
      https://www.ncei.noaa.gov/products/international-geomagnetic-reference-field
    emm2010, the Enhanced Magnetic Model 2010, which approximates
      the main and crustal magnetic fields for the period 2010-2015.
      See https://ngdc.noaa.gov/geomag/EMM/index.html
    emm2015, the Enhanced Magnetic Model 2015, which approximates
      the main and crustal magnetic fields for the period 2000-2020.
      See https://ngdc.noaa.gov/geomag/EMM/index.html
    emm2017, the Enhanced Magnetic Model 2017, which approximates
      the main and crustal magnetic fields for the period 2000-2022.
      See https://ngdc.noaa.gov/geomag/EMM/index.html

These models approximate the magnetic field due to the earth's core and
(in the case of emm20xx) its crust.  They neglect magnetic fields due to
the ionosphere, the magnetosphere, nearby magnetized materials,
electrical machinery, etc.

By default, the C<wmm2020> magnetic model is used.  This may changed by
setting the environment variable C<GEOGRAPHICLIB_MAGNETIC_NAME> or with
the B<-n> option.

The magnetic models will be loaded from a directory specified at compile
time.  This may changed by setting the environment variables
C<GEOGRAPHICLIB_MAGNETIC_PATH> or C<GEOGRAPHICLIB_DATA>, or with the
B<-d> option.  The B<-h> option prints the default magnetic path and
name.  Use the B<-v> option to ascertain the full path name of the data
file.

Instructions for downloading and installing magnetic models are
available at
L<https://geographiclib.sourceforge.io/C++/doc/magnetic.html#magneticinst>.

=head1 ENVIRONMENT

=over

=item B<GEOGRAPHICLIB_MAGNETIC_NAME>

Override the compile-time default magnetic name of C<wmm2020>.  The
B<-h> option reports the value of B<GEOGRAPHICLIB_MAGNETIC_NAME>, if
defined, otherwise it reports the compile-time value.  If the B<-n>
I<name> option is used, then I<name> takes precedence.

=item B<GEOGRAPHICLIB_MAGNETIC_PATH>

Override the compile-time default magnetic path.  This is typically
C</usr/local/share/GeographicLib/magnetic> on Unix-like systems and
C<C:/ProgramData/GeographicLib/magnetic> on Windows systems.  The B<-h>
option reports the value of B<GEOGRAPHICLIB_MAGNETIC_PATH>, if defined,
otherwise it reports the compile-time value.  If the B<-d> I<dir> option
is used, then I<dir> takes precedence.

=item B<GEOGRAPHICLIB_DATA>

Another way of overriding the compile-time default magnetic path.  If it
is set (and if B<GEOGRAPHICLIB_MAGNETIC_PATH> is not set), then
$B<GEOGRAPHICLIB_DATA>/magnetic is used.

=back

=head1 ERRORS

An illegal line of input will print an error message to standard output
beginning with C<ERROR:> and causes B<MagneticField> to return an exit
code of 1.  However, an error does not cause B<MagneticField> to
terminate; following lines will be converted.  If I<time> or I<h> are
outside the recommended ranges for the model (but inside the ranges
increase by I<tguard> and I<hguard>), a warning is printed on standard
error and the field (which may be inaccurate) is returned in the normal
way.

=head1 EXAMPLES

The magnetic field from WMM2020 in Timbuktu on 2020-12-25

    echo 2020-12-25 16:46:33N 3:00:34W 300 | MagneticField -r
    => -1.47 11.98 33994.9 33983.7 -871.7 7214.7 34752.1
       0.13 -0.02 21.9 23.9 77.9 -8.4 19.7

The first two numbers returned are the declination and inclination of
the field.  The second line gives the annual change.

=head1 SEE ALSO

GeoConvert(1), geographiclib-get-magnetic(8).

=head1 AUTHOR

B<MagneticField> was written by Charles Karney.

=head1 HISTORY

B<MagneticField> was added to GeographicLib,
L<https://geographiclib.sourceforge.io>, in version 1.15.

\name{CheckBiv}
\alias{CheckBiv}
\encoding{UTF-8}
\title{Checking Bivariate Covariance Models}
\description{
Checks whether the correlation model is bivariate.
}
\usage{
CheckBiv(numbermodel)
}
\arguments{
 \item{numbermodel}{A numeric value; the number associated with a given correlation model.}
}
\details{ 
This function checks whether the correlation model is bivariate.
}
\value{
A logical value: \code{TRUE} if the correlation model is bivariate, and \code{FALSE} otherwise.
} 
\author{
Moreno Bevilacqua \email{moreno.bevilacqua89@gmail.com} \cr
\url{https://sites.google.com/view/moreno-bevilacqua/home} \cr
Víctor Morales Oñate \email{victor.morales@uv.cl} \cr
\url{https://sites.google.com/site/moralesonatevictor/} \cr
Christian Caamaño-Carrillo \email{chcaaman@ubiobio.cl} \cr
\url{https://www.researchgate.net/profile/Christian-Caamano}
}
\examples{
library(GeoModels)
CheckBiv(CkCorrModel("Bi_matern_sep"))
}
\keyword{composite}

=head1 NAME

TransverseMercatorProj -- perform transverse Mercator projection

=head1 SYNOPSIS

B<TransverseMercatorProj> [ B<-s> | B<-t> ]
[ B<-l> I<lon0> ] [ B<-k> I<k0> ] [ B<-r> ]
[ B<-e> I<a> I<f> ] [ B<-w> ] [ B<-p> I<prec> ]
[ B<--comment-delimiter> I<commentdelim> ]
[ B<--version> | B<-h> | B<--help> ]
[ B<--input-file> I<infile> | B<--input-string> I<instring> ]
[ B<--line-separator> I<linesep> ]
[ B<--output-file> I<outfile> ]

=head1 DESCRIPTION

Perform the transverse Mercator projections.  Convert geodetic
coordinates to transverse Mercator coordinates.  The central meridian is
given by I<lon0>.  The longitude of origin is the equator.  The scale on
the central meridian is I<k0>.  By default an implementation of the
exact transverse Mercator projection is used.

Geodetic coordinates are provided on standard input as a set of lines
containing (blank separated) I<latitude> and I<longitude> (decimal
degrees or degrees, minutes, seconds); for detils on the allowed
formats for latitude and longitude, see the C<GEOGRAPHIC COORDINATES>
section of GeoConvert(1).  For each set of geodetic coordinates, the
corresponding projected easting, I<x>, and northing, I<y>, (meters) are
printed on standard output together with the meridian convergence
I<gamma> (degrees) and scale I<k>.  The meridian convergence is the
bearing of grid north (the I<y> axis) measured clockwise from true north.

=head1 OPTIONS

=over

=item B<-s>

use the sixth-order Krueger series approximation to the transverse
Mercator projection instead of the exact projection.

=item B<-t>

use the exact algorithm with the L</EXTENDED DOMAIN>.

=item B<-l> I<lon0>

specify the longitude of origin I<lon0> (degrees, default 0).

=item B<-k> I<k0>

specify the scale I<k0> on the central meridian (default 0.9996).

=item B<-r>

perform the reverse projection.  I<x> and I<y> are given on standard
input and each line of standard output gives I<latitude>, I<longitude>,
I<gamma>, and I<k>.

=item B<-e> I<a> I<f>

specify the ellipsoid via the equatorial radius, I<a> and
the flattening, I<f>.  Setting I<f> = 0 results in a sphere.  Specify
I<f> E<lt> 0 for a prolate ellipsoid.  A simple fraction, e.g., 1/297,
is allowed for I<f>.  By default, the WGS84 ellipsoid is used, I<a> =
6378137 m, I<f> = 1/298.257223563.  If the exact algorithm is used, I<f>
must be positive.

=item B<-w>

on input and output, longitude precedes latitude (except that on input
this can be overridden by a hemisphere designator, I<N>, I<S>, I<E>,
I<W>).

=item B<-p> I<prec>

set the output precision to I<prec> (default 6).  I<prec> is the number
of digits after the decimal point for lengths (in meters).  For
latitudes and longitudes (in degrees), the number of digits after the
decimal point is I<prec> + 5.  For the convergence (in degrees) and
scale, the number of digits after the decimal point is I<prec> + 6.

=item B<--comment-delimiter> I<commentdelim>

set the comment delimiter to I<commentdelim> (e.g., "#" or "//").  If
set, the input lines will be scanned for this delimiter and, if found,
the delimiter and the rest of the line will be removed prior to
processing and subsequently appended to the output line (separated by a
space).

=item B<--version>

print version and exit.

=item B<-h>

print usage and exit.

=item B<--help>

print full documentation and exit.

=item B<--input-file> I<infile>

read input from the file I<infile> instead of from standard input; a file
name of "-" stands for standard input.

=item B<--input-string> I<instring>

read input from the string I<instring> instead of from standard input.
All occurrences of the line separator character (default is a semicolon)
in I<instring> are converted to newlines before the reading begins.

=item B<--line-separator> I<linesep>

set the line separator character to I<linesep>.  By default this is a
semicolon.

=item B<--output-file> I<outfile>

write output to the file I<outfile> instead of to standard output; a
file name of "-" stands for standard output.

=back

=head1 EXTENDED DOMAIN

The exact transverse Mercator projection has a I<branch point> on the
equator at longitudes (relative to I<lon0>) of +/- (1 - I<e>) 90 =
82.636..., where I<e> is the eccentricity of the ellipsoid.  The
standard convention for handling this branch point is to map positive
(negative) latitudes into positive (negative) northings I<y>; i.e., a
branch cut is placed on the equator.  With the I<extended> domain, the
northern sheet of the projection is extended into the south hemisphere
by pushing the branch cut south from the branch points.  See the
reference below for details.

=head1 EXAMPLES

   echo 0 90 | TransverseMercatorProj
   => 25953592.84 9997964.94 90 18.40
   echo 260e5 100e5 | TransverseMercatorProj -r
   => -0.02 90.00 90.01 18.48

=head1 ERRORS

An illegal line of input will print an error message to standard output
beginning with C<ERROR:> and causes B<TransverseMercatorProj> to return an exit
code of 1.  However, an error does not cause B<TransverseMercatorProj> to
terminate; following lines will be converted.

=head1 AUTHOR

B<TransverseMercatorProj> was written by Charles Karney.

=head1 SEE ALSO

The algorithms for the transverse Mercator projection are described in
C. F. F. Karney, I<Transverse Mercator with an accuracy of a few
nanometers>, J. Geodesy B<85>(8), 475-485 (Aug. 2011); DOI
L<https://doi.org/10.1007/s00190-011-0445-3>; preprint
L<https://arxiv.org/abs/1002.1417>.  The explanation of the extended
domain of the projection with the B<-t> option is given in Section 5 of
this paper.

=head1 HISTORY

B<TransverseMercatorProj> was added to GeographicLib,
L<https://geographiclib.sourceforge.io>, in 2009-01.
Prior to version 1.9 it was called TransverseMercatorTest (and its
interface was slightly different).

=head1 NAME

GeodSolve -- perform geodesic calculations

=head1 SYNOPSIS

B<GeodSolve>
[ B<-i> | B<-L> I<lat1> I<lon1> I<azi1> |
B<-D> I<lat1> I<lon1> I<azi1> I<s13> | B<-I> I<lat1> I<lon1> I<lat3> I<lon3> ]
[ B<-a> ] [ B<-e> I<a> I<f> ] [ B<-u> ] [ B<-F> ]
[ B<-d> | B<-:> ] [ B<-w> ] [ B<-b> ] [ B<-f> ] [ B<-p> I<prec> ] [ B<-E> ]
[ B<--comment-delimiter> I<commentdelim> ]
[ B<--version> | B<-h> | B<--help> ]
[ B<--input-file> I<infile> | B<--input-string> I<instring> ]
[ B<--line-separator> I<linesep> ]
[ B<--output-file> I<outfile> ]

=head1 DESCRIPTION

The shortest path between two points on the ellipsoid at (I<lat1>,
I<lon1>) and (I<lat2>, I<lon2>) is called the geodesic.  Its length is
I<s12> and the geodesic from point 1 to point 2 has forward azimuths
I<azi1> and I<azi2> at the two end points.

B<GeodSolve> operates in one of three modes:

=over

=item 1.

By default, B<GeodSolve> accepts lines on the standard input containing
I<lat1> I<lon1> I<azi1> I<s12> and prints I<lat2> I<lon2> I<azi2>
on standard output.  This is the direct geodesic calculation.

=item 2.

With the B<-i> option, B<GeodSolve> performs the inverse geodesic
calculation.  It reads lines containing I<lat1> I<lon1> I<lat2>
I<lon2> and prints the corresponding values of I<azi1> I<azi2> I<s12>.

=item 3.

Command line arguments B<-L> I<lat1> I<lon1> I<azi1> specify a geodesic
line.  B<GeodSolve> then accepts a sequence of I<s12> values (one per
line) on standard input and prints I<lat2> I<lon2> I<azi2> for each.
This generates a sequence of points on a single geodesic.  Command line
arguments B<-D> and B<-I> work similarly with the geodesic line defined
in terms of a direct or inverse geodesic calculation, respectively.

=back

=head1 OPTIONS

=over

=item B<-i>

perform an inverse geodesic calculation (see 2 above).

=item B<-L> I<lat1> I<lon1> I<azi1>

line mode (see 3 above); generate a sequence of points along the
geodesic specified by I<lat1> I<lon1> I<azi1>.  The B<-w> flag can be
used to swap the default order of the 2 geographic coordinates, provided
that it appears before B<-L>.

=item B<-D> I<lat1> I<lon1> I<azi1> I<s13>

line mode (see 3 above); generate a sequence of points along the
geodesic specified by I<lat1> I<lon1> I<azi1> I<s13>.  The B<-w> flag
can be used to swap the default order of the 2 geographic coordinates,
provided that it appears before B<-D>.  Similarly, the B<-a> flag can be
used to change the interpretation of I<s13> to I<a13>, provided that it
appears before B<-D>.

=item B<-I> I<lat1> I<lon1> I<lat3> I<lon3>

line mode (see 3 above); generate a sequence of points along the
geodesic specified by I<lat1> I<lon1> I<lat3> I<lon3>.  The B<-w> flag
can be used to swap the default order of the 2 geographic coordinates,
provided that it appears before B<-I>.

=item B<-a>

toggle the arc mode flag (it starts off); if this flag is on, then on
input I<and> output I<s12> is replaced by I<a12> the arc length (in
degrees) on the auxiliary sphere.  See L</AUXILIARY SPHERE>.

=item B<-e> I<a> I<f>

specify the ellipsoid via the equatorial radius, I<a> and
the flattening, I<f>.  Setting I<f> = 0 results in a sphere.  Specify
I<f> E<lt> 0 for a prolate ellipsoid.  A simple fraction, e.g., 1/297,
is allowed for I<f>.  By default, the WGS84 ellipsoid is used, I<a> =
6378137 m, I<f> = 1/298.257223563.

=item B<-u>

unroll the longitude.  Normally, on output longitudes are reduced to lie
in [-180deg,180deg).  However with this option, the returned longitude
I<lon2> is "unrolled" so that I<lon2> - I<lon1> indicates how often and
in what sense the geodesic has encircled the earth.  Use the B<-f>
option, to get both longitudes printed.

=item B<-F>

fractional mode.  This only has any effect with the B<-D> and B<-I>
options (and is otherwise ignored).  The values read on standard input
are interpreted as fractional distances to point 3, i.e., as
I<s12>/I<s13> instead of I<s12>.  If arc mode is in effect, then the
values denote fractional arc length, i.e., I<a12>/I<a13>.  The
fractional distances can be entered as a simple fraction, e.g., 3/4.

=item B<-d>

output angles as degrees, minutes, seconds instead of decimal degrees.

=item B<-:>

like B<-d>, except use : as a separator instead of the d, ', and "
delimiters.

=item B<-w>

toggle the longitude first flag (it starts off); if the flag is on, then
on input and output, longitude precedes latitude (except that, on input,
this can be overridden by a hemisphere designator, I<N>, I<S>, I<E>,
I<W>).

=item B<-b>

report the I<back> azimuth at point 2 instead of the forward azimuth.

=item B<-f>

full output; each line of output consists of 12 quantities: I<lat1>
I<lon1> I<azi1> I<lat2> I<lon2> I<azi2> I<s12> I<a12> I<m12> I<M12>
I<M21> I<S12>.  I<a12> is described in L</AUXILIARY SPHERE>.  The four
quantities I<m12>, I<M12>, I<M21>, and I<S12> are described in
L</ADDITIONAL QUANTITIES>.

=item B<-p> I<prec>

set the output precision to I<prec> (default 3); I<prec> is the
precision relative to 1 m.  See L</PRECISION>.

=item B<-E>

use "exact" algorithms (based on elliptic integrals) for the geodesic
calculations.  These are more accurate than the (default) series
expansions for |I<f>| E<gt> 0.02.

=item B<--comment-delimiter> I<commentdelim>

set the comment delimiter to I<commentdelim> (e.g., "#" or "//").  If
set, the input lines will be scanned for this delimiter and, if found,
the delimiter and the rest of the line will be removed prior to
processing and subsequently appended to the output line (separated by a
space).

=item B<--version>

print version and exit.

=item B<-h>

print usage and exit.

=item B<--help>

print full documentation and exit.

=item B<--input-file> I<infile>

read input from the file I<infile> instead of from standard input; a file
name of "-" stands for standard input.

=item B<--input-string> I<instring>

read input from the string I<instring> instead of from standard input.
All occurrences of the line separator character (default is a semicolon)
in I<instring> are converted to newlines before the reading begins.

=item B<--line-separator> I<linesep>

set the line separator character to I<linesep>.  By default this is a
semicolon.

=item B<--output-file> I<outfile>

write output to the file I<outfile> instead of to standard output; a
file name of "-" stands for standard output.

=back

=head1 INPUT

B<GeodSolve> measures all angles in degrees and all lengths (I<s12>) in
meters, and all areas (I<S12>) in meters^2.  On input angles (latitude,
longitude, azimuth, arc length) can be as decimal degrees or degrees,
minutes, seconds.  For example, C<40d30>, C<40d30'>, C<40:30>, C<40.5d>,
and C<40.5> are all equivalent.  By default, latitude precedes longitude
for each point (the B<-w> flag switches this convention); however on
input either may be given first by appending (or prepending) I<N> or
I<S> to the latitude and I<E> or I<W> to the longitude.  Azimuths are
measured clockwise from north; however this may be overridden with I<E>
or I<W>.

For details on the allowed formats for angles, see the C<GEOGRAPHIC
COORDINATES> section of GeoConvert(1).

=head1 AUXILIARY SPHERE

Geodesics on the ellipsoid can be transferred to the I<auxiliary sphere>
on which the distance is measured in terms of the arc length I<a12>
(measured in degrees) instead of I<s12>.  In terms of I<a12>, 180
degrees is the distance from one equator crossing to the next or from
the minimum latitude to the maximum latitude.  Geodesics with I<a12>
E<gt> 180 degrees do not correspond to shortest paths.  With the B<-a>
flag, I<s12> (on both input and output) is replaced by I<a12>.  The
B<-a> flag does I<not> affect the full output given by the B<-f> flag
(which always includes both I<s12> and I<a12>).

=head1 ADDITIONAL QUANTITIES

The B<-f> flag reports four additional quantities.

The reduced length of the geodesic, I<m12>, is defined such that if the
initial azimuth is perturbed by dI<azi1> (radians) then the second point
is displaced by I<m12> dI<azi1> in the direction perpendicular to the
geodesic.  I<m12> is given in meters.  On a curved surface the
reduced length obeys a symmetry relation, I<m12> + I<m21> = 0.  On a
flat surface, we have I<m12> = I<s12>.

I<M12> and I<M21> are geodesic scales.  If two geodesics are parallel at
point 1 and separated by a small distance I<dt>, then they are separated
by a distance I<M12> I<dt> at point 2.  I<M21> is defined similarly
(with the geodesics being parallel to one another at point 2).  I<M12>
and I<M21> are dimensionless quantities.  On a flat surface, we have
I<M12> = I<M21> = 1.

If points 1, 2, and 3 lie on a single geodesic, then the following
addition rules hold:

   s13 = s12 + s23,
   a13 = a12 + a23,
   S13 = S12 + S23,
   m13 = m12 M23 + m23 M21,
   M13 = M12 M23 - (1 - M12 M21) m23 / m12,
   M31 = M32 M21 - (1 - M23 M32) m12 / m23.

Finally, I<S12> is the area between the geodesic from point 1 to point 2
and the equator; i.e., it is the area, measured counter-clockwise, of
the geodesic quadrilateral with corners (I<lat1>,I<lon1>), (0,I<lon1>),
(0,I<lon2>), and (I<lat2>,I<lon2>).  It is given in meters^2.

=head1 PRECISION

I<prec> gives precision of the output with I<prec> = 0 giving 1 m
precision, I<prec> = 3 giving 1 mm precision, etc.  I<prec> is the
number of digits after the decimal point for lengths.  For decimal
degrees, the number of digits after the decimal point is I<prec> + 5.
For DMS (degree, minute, seconds) output, the number of digits after the
decimal point in the seconds component is I<prec> + 1.  The minimum
value of I<prec> is 0 and the maximum is 10.

=head1 ERRORS

An illegal line of input will print an error message to standard output
beginning with C<ERROR:> and causes B<GeodSolve> to return an exit code
of 1.  However, an error does not cause B<GeodSolve> to terminate;
following lines will be converted.

=head1 ACCURACY

Using the (default) series solution, GeodSolve is accurate to about 15
nm (15 nanometers) for the WGS84 ellipsoid.  The approximate maximum
error (expressed as a distance) for an ellipsoid with the same equatorial
radius as the WGS84 ellipsoid and different values of the flattening is

   |f|     error
   0.01    25 nm
   0.02    30 nm
   0.05    10 um
   0.1    1.5 mm
   0.2    300 mm

If B<-E> is specified, GeodSolve is accurate to about 40 nm (40
nanometers) for the WGS84 ellipsoid.  The approximate maximum error
(expressed as a distance) for an ellipsoid with a quarter meridian of
10000 km and different values of the I<a/b> = 1 - I<f> is

   1-f    error (nm)
   1/128   387
   1/64    345
   1/32    269
   1/16    210
   1/8     115
   1/4      69
   1/2      36
     1      15
     2      25
     4      96
     8     318
    16     985
    32    2352
    64    6008
   128   19024

=head1 MULTIPLE SOLUTIONS

The shortest distance returned for the inverse problem is (obviously)
uniquely defined.  However, in a few special cases there are multiple
azimuths which yield the same shortest distance.  Here is a catalog of
those cases:

=over

=item I<lat1> = -I<lat2> (with neither point at a pole)

If I<azi1> = I<azi2>, the geodesic is unique.  Otherwise there are two
geodesics and the second one is obtained by setting [I<azi1>,I<azi2>] =
[I<azi2>,I<azi1>], [I<M12>,I<M21>] = [I<M21>,I<M12>], I<S12> = -I<S12>.
(This occurs when the longitude difference is near +/-180 for oblate
ellipsoids.)

=item I<lon2> = I<lon1> +/- 180 (with neither point at a pole)

If I<azi1> = 0 or +/-180, the geodesic is unique.  Otherwise there are
two geodesics and the second one is obtained by setting
[I<azi1>,I<azi2>] = [-I<azi1>,-I<azi2>], I<S12> = -I<S12>.  (This occurs
when I<lat2> is near -I<lat1> for prolate ellipsoids.)

=item Points 1 and 2 at opposite poles

There are infinitely many geodesics which can be generated by setting
[I<azi1>,I<azi2>] = [I<azi1>,I<azi2>] + [I<d>,-I<d>], for arbitrary
I<d>.  (For spheres, this prescription applies when points 1 and 2 are
antipodal.)

=item I<s12> = 0 (coincident points)

There are infinitely many geodesics which can be generated by setting
[I<azi1>,I<azi2>] = [I<azi1>,I<azi2>] + [I<d>,I<d>], for arbitrary I<d>.

=back

=head1 EXAMPLES

Route from JFK Airport to Singapore Changi Airport:

   echo 40:38:23N 073:46:44W 01:21:33N 103:59:22E |
   GeodSolve -i -: -p 0

   003:18:29.9 177:29:09.2 15347628

Equally spaced waypoints on the route:

   for ((i = 0; i <= 10; ++i)); do echo $i/10; done |
   GeodSolve -I 40:38:23N 073:46:44W 01:21:33N 103:59:22E -F -: -p 0

   40:38:23.0N 073:46:44.0W 003:18:29.9
   54:24:51.3N 072:25:39.6W 004:18:44.1
   68:07:37.7N 069:40:42.9W 006:44:25.4
   81:38:00.4N 058:37:53.9W 017:28:52.7
   83:43:26.0N 080:37:16.9E 156:26:00.4
   70:20:29.2N 097:01:29.4E 172:31:56.4
   56:38:36.0N 100:14:47.6E 175:26:10.5
   42:52:37.1N 101:43:37.2E 176:34:28.6
   29:03:57.0N 102:39:34.8E 177:07:35.2
   15:13:18.6N 103:22:08.0E 177:23:44.7
   01:21:33.0N 103:59:22.0E 177:29:09.2

=head1 SEE ALSO

GeoConvert(1).

An online version of this utility is availbable at
L<https://geographiclib.sourceforge.io/cgi-bin/GeodSolve>.

The algorithms are described in C. F. F. Karney,
I<Algorithms for geodesics>, J. Geodesy 87, 43-55 (2013); DOI:
L<https://doi.org/10.1007/s00190-012-0578-z>;
addenda: L<https://geographiclib.sourceforge.io/geod-addenda.html>.

The Wikipedia page, Geodesics on an ellipsoid,
L<https://en.wikipedia.org/wiki/Geodesics_on_an_ellipsoid>.

=head1 AUTHOR

B<GeodSolve> was written by Charles Karney.

=head1 HISTORY

B<GeodSolve> was added to GeographicLib,
L<https://geographiclib.sourceforge.io>, in 2009-03.  Prior to version
1.30, it was called B<Geod>.  (The name was changed to avoid a conflict
with the B<geod> utility in I<proj.4>.)

\name{anomalies}
\alias{anomalies}
\title{Annual Precipitation Anomalies in the U.S.}
\usage{data(anomalies)}
\description{
A numerical matrix of dimension \eqn{7252 \times 3} containing longitude, latitude, and yearly total precipitation anomalies registered at 7,352 location sites in the USA.
}
\format{
A numeric matrix with 7,252 rows and 3 columns:
\describe{
 \item{Column 1}{Longitude}
 \item{Column 2}{Latitude}
 \item{Column 3}{Annual precipitation anomaly}
}
}
\source{
Kaufman, C.G., Schervish, M.J., Nychka, D.W. (2008). Covariance tapering for likelihood-based estimation in large spatial data sets. \emph{Journal of the American Statistical Association, Theory & Methods}, \bold{103}, 1545--1555.
}
\keyword{datasets}

\name{GeoQQ}
\alias{GeoQQ}
\encoding{UTF-8}
\title{Quantile-quantile plot }
\description{
 Based on a GeoFit object, the procedure plots a quantile-quantile plot or compares the fitted density with the histogram of the data. It is useful as diagnostic tool.
}
\usage{GeoQQ(fit,type="Q",add=FALSE,ylim=c(0,1),xlim=NULL,breaks=10,\dots)}

\arguments{

  \item{fit}{An object of class \code{\link{GeoFit}}, For fitted models
 with covariates, the \code{\link{GeoFit}} object must first be transformed
 by \code{\link{GeoResiduals}}.}

\item{type}{The type of plot. If Q then a qq-plot (default) is performed. If D then a comparison between histrogram and the estimated marginal density is performed}
\item{add}{Logical; if TRUE the the estimated density ia added over an existing one} 
\item{ylim}{Numeric; a vector of length 2 used for the ylab parameter of the histogram plot.} 
\item{xlim}{Numeric; a vector of length 2 used for the xlab parameter of the histogram plot.} 
\item{breaks}{Numeric; an integer number specifyng the number of cells ofthe histogram plot if the option type=D is chosen.} 
\item{\dots}{Optional parameters passed to the plot function.}
}

\details{

This diagnostic requires an intercept-only marginal representation.  For
continuous models, fits with a nonconstant design matrix or a site-specific
fixed mean can first be converted with \code{\link{GeoResiduals}}.  For
discrete models with a nonconstant mean, use \code{\link{GeoPit}} instead,
because there is no single common marginal distribution for a conventional QQ
plot.  Likewise, Binomial and negative-binomial QQ/density plots require a
constant \code{n}; site-specific \code{n} should be assessed with
\code{GeoPit()}.
}

\value{ 
Produces a plot. No values are returned.
}


\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}


\examples{
library(GeoModels)


##################
### Example 1
##################
set.seed(21)
model="Tukeyh";tail=0.1
N=400 # number of location sites
# Set the coordinates of the points:
x = runif(N, 0, 1)
y = runif(N, 0, 1)
coords=cbind(x,y)

# regression parameters
mean = 5
mean1=0.8

X=cbind(rep(1,N),runif(N))
# correlation parameters:
corrmodel = "Wend0"
sill = 1
nugget = 0
scale = 0.3
power2=4


param=list(mean=mean,mean1=mean1, sill=sill, nugget=nugget, 
	 scale=scale,tail=tail,power2=power2)
# Simulation of the Gaussian RF:
data = GeoSim(coordx=coords, corrmodel=corrmodel, X=X,model=model,param=param)$data

start=list(mean=mean,mean1=mean1, scale=scale,tail=tail)
fixed=list(nugget=nugget,sill=sill,power2=power2)
# Maximum composite-likelihood fitting 
fit = GeoFit(data,coordx=coords, corrmodel=corrmodel,model=model,X=X,
 likelihood="Conditional",type='Pairwise',start=start,
 fixed=fixed,neighb=4)

res=GeoResiduals(fit)
GeoQQ(res,type="Q")
GeoQQ(res,type="D",lwd=2,ylim=c(0,0.5),breaks=20)


##################
### Example 2
##################
set.seed(21)
model="Weibull";shape=1.5
N=600 # number of location sites
# Set the coordinates of the points:
x = runif(N, 0, 1)
y = runif(N, 0, 1)
coords=cbind(x,y)


# regression parameters
mean = 0

# correlation parameters:
corrmodel = "Matern"
smooth=0.5
nugget = 0
scale = 0.2/3


param=list(mean=mean, sill=1, nugget=nugget, 
 scale=scale,smooth=smooth, shape=shape)
# Simulation of the Gaussian RF:
data = GeoSim(coordx=coords, corrmodel=corrmodel,model=model,param=param)$data

start=list(mean=mean, scale=scale,shape=shape)
I=Inf
lower=list(mean=-I, scale=0,shape=0)
upper=list(mean= I, scale=I,shape=I)
I=Inf
fixed=list(nugget=nugget,sill=1,smooth=smooth)
# Maximum composite-likelihood fitting 
fit = GeoFit(data,coordx=coords, corrmodel=corrmodel,model=model,
 likelihood="Conditional",type='Pairwise',start=start,
 optimizer="nlminb",lower=lower,upper=upper,
 fixed=fixed,neighb=3)
GeoQQ(fit,type="Q")
GeoQQ(fit,type="D",lwd=2,ylim=c(0,1),breaks=20)
}

\keyword{Composite}

\name{GeoWLS}
\alias{GeoWLS}
\alias{print.GeoWLS}
\encoding{UTF-8}
\title{WLS of Gaussian random fields}
\description{
 the function returns the parameter estimates of a Gaussian random field obtained by
 the weighted least squares estimator.
}
\usage{
GeoWLS(data, coordx, coordy=NULL,coordz=NULL, coordt=NULL, coordx_dyn=NULL, corrmodel, 
 distance="Eucl", fixed=NULL, grid=FALSE, maxdist=NULL,neighb=NULL,
 maxtime=NULL, optimizer='Nelder-Mead',
 numbins=NULL, radius=1, start=NULL, weighted=FALSE,optimization=TRUE,
 numbins_t=NULL)
}
\arguments{
 \item{data}{A \eqn{d}{d}-dimensional vector (a single spatial realisation) or a (\eqn{d \times d}{d x d})-matrix (a single spatial realisation on regular grid) or a
 (\eqn{t \times d}{t x d})-matrix (a single spatio-temporal realisation) or an (\eqn{d \times d \times t \times n }{d x d x t})-array
 (a single spatio-temporal realisation on regular grid). See \code{\link{GeoFit}} for details.}
 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix 
 Coordinates on a sphere for a fixed radius \code{radius} 
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordt}{A numeric vector giving one dimension of temporal coordinates. Optional argument; the default is \code{NULL}, in which case a spatial random field is expected. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{A list of \eqn{m} numeric (\eqn{d_t \times 2}{d x 2})-matrices
 containing dynamical (in time) spatial coordinates. Optional argument, the default is \code{NULL}
 } 
 \item{corrmodel}{String; the name of a correlation model, for the
 description (see \code{\link{GeoFit}}).}
 \item{distance}{String; the name of the spatial distance. The default is \code{Eucl},
 the Euclidean distance. See \code{\link{GeoFit}} for details.
		 of \code{\link{GeoFit}}.}
 \item{fixed}{A named list giving the values of the parameters that
 will be considered as known values. The listed parameters for a
 given correlation function will be not estimated, i.e. if
 \code{list(nugget=0)} the nugget effect is ignored.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data
 are interpreted as a vector or a (\eqn{n \times d}{n x d})-matrix,
 instead if \code{TRUE} then (\eqn{d \times d
 \times n}{d x d x n})-matrix is considered.}
 \item{maxdist}{A numeric value denoting the maximum distance, see
 \code{\link{GeoFit}} for details.}
 \item{neighb}{Numeric; an optional positive integer indicating the 
 order of neighborhood. See \code{\link{GeoFit}} for details}
 \item{maxtime}{Numeric; an optional positive maximum temporal lag, expressed in the same units as \code{coordt}. See \code{\link{GeoFit}} for details.}
 \item{optimizer}{String; the optimization algorithm
 (see \code{\link{optim}} for details). 'Nelder-Mead' is the
 default.}
 \item{numbins}{Historical GeoModels argument giving the number of spatial bin boundaries; \code{numbins - 1} empirical spatial classes are used. See Details.}
 \item{numbins_t}{Optional positive integer giving the number of temporal classes for irregularly spaced \code{coordt}. If \code{NULL}, \code{numbins - 1} is used. Regularly spaced times use their directly attainable temporal lags.}
 \item{radius}{Numeric; a value indicating the radius of the sphere when using the great circle distance. Default value is 1.} 
 \item{start}{A named list with the initial values of the
 parameters that are used by the numerical routines in maximization
 procedure. \code{NULL} is the default (see \code{\link{GeoFit}}).}
 \item{weighted}{Logical; if \code{TRUE} then the weighted least square
 estimator is considered. If \code{FALSE} (the
 default) then the classic least square is used.}
 \item{optimization}{Logical; if \code{TRUE} then the weighted least square
 minimization is performed. Otherwise the weighted least square function is evaluated at the starting value.} 
}

\details{
 \code{GeoWLS} is defined for Gaussian random fields and therefore does
 not expose a \code{model} argument. Its native least-squares kernels compare
 the empirical variogram with the Gaussian variogram parameterized by nugget,
 sill, and the selected correlation model.

 The historical \code{numbins} parameter gives the number of spatial bin
 boundaries, so \code{numbins - 1} spatial classes are used. For
 spatio-temporal data, regular temporal coordinates are handled from their
 common spacing without constructing all pairwise temporal differences;
 irregular temporal coordinates use \code{numbins_t} temporal classes.

 Empty empirical cells are retained in the rectangular spatial-temporal lag
 grid and are skipped by the least-squares objective through their zero pair
 count. This preserves the correspondence between empirical cells and their
 spatial/temporal lags.

 The \code{maxdist} parameter indicates the positive finite maximum distance
 below which pairs are considered in the (weighted) least squares.
}

\value{
 Returns an object of class \code{WLS}.
 An object of class \code{WLS} is a list containing
 at most the following components:

 \item{bins}{Adjacent intervals of grouped distances;}
 \item{bint}{Adjacent intervals of grouped temporal separations}
 \item{centers}{The centers of the bins;}
 \item{coordx}{The vector or matrix of spatial coordinates;}
 \item{coordy}{The vector of spatial coordinates;}
 \item{coordt}{The vector of temporal coordinates;}
 \item{convergence}{A string that denotes if convergence is reached;}
 \item{corrmodel}{The correlation model;}
 \item{data}{The vector or matrix of data;}
 \item{distance}{The type of spatial distance;}
 \item{fixed}{The vector of fixed parameters;}
 \item{iterations}{The number of iteration used by the numerical routine;}
 \item{maxdist}{The maximum spatial distance used for the calculation of the variogram used in least square estimation.
 If no spatial distance is specified then it is NULL;}
 \item{maxtime}{The maximum temporal distance used for the calculation of the variogram used in least square estimation.
 If no temporal distance is specified then it is NULL;}
 \item{numbins_t}{The requested number of temporal classes for irregular time coordinates, or \code{NULL};}
 \item{message}{Extra message passed from the numerical routines;}
 \item{numcoord}{The number of spatial coordinates;}
 \item{numtime}{The number the temporal realisations of the random field;}
 \item{param}{The vector of parameters' estimates;}
 \item{variograms}{The empirical spatial variogram;}
 \item{variogramt}{The empirical temporal variogram;}
 \item{variogramst}{The empirical spatio-temporal variogram;}
 \item{weighted}{A logical value indicating if its the weighted
 method;}
 \item{wls}{The value of the least squares at the minimum.}

}

\references{

 Cressie, N. A. C. (1993) \emph{Statistics for Spatial Data}. New York:
 Wiley.

 Gaetan, C. and Guyon, X. (2010) \emph{Spatial Statistics and Modelling}.
 Springer-Verlag, New York.

}

\seealso{\code{\link{GeoFit}}, \code{\link{optim}}}



\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

 
\examples{
library(GeoModels)


# Set the coordinates of the sites:

set.seed(211)
x <- runif(200, 0, 1)
set.seed(98)
y <- runif(200, 0, 1)
coords <- cbind(x,y)

################################################################
###
### Example 1. Least square fitting of a Gaussian random field
### with exponential correlation.
###
###############################################################

# Set the model's parameters:
corrmodel <- "Exponential"
mean <- 0
sill <- 1
nugget <- 0
scale <- 0.15/3
param <- list(mean=0,sill=sill, nugget=nugget, scale=scale)
# Simulation of the Gaussian random field:
set.seed(2)
data <- GeoSim(coordx=coords, corrmodel=corrmodel, param=param)$data

fixed=list(nugget=0,mean=mean)
start=list(scale=scale,sill=sill)
# Least square fitting of the random field:
fit <- GeoWLS(data=data,coordx=coords, corrmodel=corrmodel,
 fixed=fixed,start=start,maxdist=0.5)

# Results:
print(fit)

}

\keyword{LeastSquare}

\name{GeoKriglocWeights}
\alias{GeoKriglocWeights}
\encoding{UTF-8}
\title{
Compute validated local Gaussian kriging weights
}
\description{
Given a set of spatial locations (and possibly temporal instants), the function
computes local Gaussian kriging weights for each requested prediction task using
a neighborhood selected from the observations.  Non-Gaussian local prediction is
implemented by \code{\link{GeoKrigloc}}, not by this low-level weights helper.
}
\usage{
GeoKriglocWeights(coordx=NULL, coordy=NULL, coordz=NULL, coordt=NULL, coordx_dyn=NULL,
 corrmodel, distance="Eucl", grid=FALSE, loc, neighb=NULL,
 maxdist=NULL, maxtime=NULL, method="cholesky", model="Gaussian",
 n=1, nloc=NULL, param, anisopars=NULL,
 radius=1, sparse=FALSE, time=NULL, which=1,
 copula=NULL, X=NULL, Xloc=NULL, Mloc=NULL,parallel=FALSE,
 ncores=6, compact=FALSE)
}
\arguments{
 \item{coordx}{Numeric (\eqn{d\times 2}{d x 2}) or (\eqn{d\times 3}{d x 3}) matrix
 of spatial coordinates. Coordinates on a sphere are accepted
 (lon/lat in decimal degrees) when \code{distance = "Sphere"}.}
 \item{coordy}{Optional numeric vector giving an additional spatial coordinate
 dimension. Ignored if \code{coordx} is already a matrix.}
 \item{coordz}{Optional numeric vector giving a third spatial coordinate
 dimension.}
 \item{coordt}{Optional numeric vector containing the temporal coordinates of the observations. If missing, a purely spatial random field is assumed. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{List of one two- or three-column observation-coordinate matrix per temporal instant.  Blocks are concatenated by time; rows within each block retain their matrix order.  See \code{\link{GeoModels-spacetime-ordering}}.}
 \item{corrmodel}{Character string naming a valid correlation model.
 See \code{\link{GeoCovmatrix}} for admissible choices.}
 \item{distance}{Character string specifying the spatial distance.
 Default is \code{"Eucl"} (Euclidean). See \code{\link{GeoFit}}.}
 \item{grid}{Logical. If \code{TRUE}, coordinates are interpreted as
 defining a regular grid (see \code{\link{GeoKrig}}).}
 \item{loc}{Numeric (\eqn{n\times 2}{n x 2}) matrix of locations for which
 the kriging weights are required.}
 \item{neighb}{Numeric; an optional positive integer indicating the 
 order of the neighborhood.}
 \item{maxdist}{Numeric; an optional positive value indicating the distance in the spatial neighborhood.}
 \item{maxtime}{Numeric; an optional non-negative maximum temporal-distance threshold, expressed in the same units as \code{coordt}, used for the local temporal neighborhood.}
 \item{method}{Character string indicating the matrix factorisation
 used to solve the kriging system: \code{"cholesky"} (default) or \code{"svd"}.}
 \item{model}{Character string.  The validated public implementation currently
 requires \code{"Gaussian"} (the alias \code{"Gauss"} is accepted).}
 \item{n}{Integer. Number of trials for Binomial random fields
 (default \code{1}).}
 \item{nloc}{Integer. Number of trials for the prediction locations
 in Binomial random fields (default \code{1}).}
 \item{param}{Named list of covariance and mean parameters.
 See \code{\link{CorrParam}} and \code{\link{GeoCovmatrix}}.}
 \item{anisopars}{List with components \code{angle} and \code{ratio}
 defining geometric anisotropy (optional).}
 \item{radius}{Positive numeric value: sphere radius when
 coordinates are lon/lat (default \code{1}).}
 \item{sparse}{Logical. If \code{TRUE}, sparse‐matrix algorithms
 (package \pkg{spam}) are employed. Only effective with compactly
 supported covariance functions.}
 \item{time}{Numeric vector giving the temporal instants for which weights are required. Values need not be equally spaced and are interpreted on the same numeric time scale as \code{coordt}. Ignored if \code{coordt} is missing.}
 \item{which}{Integer (\code{1} or \code{2}) selecting the variable whose
 weights are returned in the bivariate case.}
 \item{copula}{Must be \code{NULL}.}
 \item{X}{Numeric design matrix at observation locations.  Fixed-location observations are ordered time then site; dynamic observations are ordered by the temporal blocks of \code{coordx_dyn}.}
 \item{Xloc}{Numeric design matrix at prediction tasks, ordered location then time: all requested times for the first row of \code{loc}, then all requested times for the second row, and so on.}
 \item{Mloc}{Numeric vector of known prediction means in the same location-major order as \code{Xloc}.  Use either \code{Mloc} or \code{Xloc}, not both.}
 \item{parallel}{Logical; default \code{FALSE}. Set \code{TRUE} to compute independent local kriging systems in parallel where supported.}
 \item{ncores}{Positive integer or \code{NULL}; default \code{6}. With \code{parallel=TRUE}, an explicit integer requests that many workers, capped by detected cores and the number of available jobs. Set \code{ncores=NULL} for automatic selection, capped at six workers and normally leaving one detected core free.}
 \item{compact}{Logical. If \code{FALSE} (default), each local element contains
 the full \code{GeoKrigWeights} result for backward compatibility. If
 \code{TRUE}, each non-empty element keeps only the numeric weights and
 \code{neighbor_indices}; this memory-saving representation is used internally
 by \code{GeoSimcond(local=TRUE)}.}
}
\details{

Mean inputs follow \code{\link{GeoKrigloc}}.  Coefficients \code{mean},
\code{mean1}, and so on correspond in order to the columns of \code{X}; an
intercept-only model uses only \code{mean}.  A vector \code{param$mean} is a
known observation mean and is mutually exclusive with \code{X}.  At local
prediction tasks, use either \code{Xloc} or \code{Mloc}, not both.

For every prediction task the function builds and solves the local kriging
system
\deqn{\Sigma \mathbf{w} = \boldsymbol{\sigma}_0}{Sigma w = sigma_0},
where \eqn{\Sigma}{Sigma} is the Gaussian covariance matrix among the selected
observations and \eqn{\boldsymbol{\sigma}_0}{sigma_0} contains their
covariances with the prediction location.  No actual prediction is carried out;
for model-specific local prediction use \code{\link{GeoKrigloc}}.  Local results are returned directly rather
than through temporary-file references, so their components remain available
after the parallel workers have terminated.
}
\section{Spatio-temporal ordering}{
Observation-level rows follow time-major order.  For fixed sites this is the
order of \code{c(t(data))}; for dynamic sites it is the row-binding of
\code{coordx_dyn} by list element.  Prediction tasks follow location-major
order, so rows of \code{Xloc} and elements of \code{Mloc} enumerate all
requested times for the first prediction location before moving to the next
location.  See \code{\link{GeoModels-spacetime-ordering}}.
}

\value{
An object of class \code{GeoKriglocWeights}.  Its component \code{weights} is a
list ordered by prediction task.  With \code{compact=FALSE}, each
non-\code{NULL} element is the local result returned by
\code{\link{GeoKrigWeights}}, augmented with \code{neighbor_indices}. With
\code{compact=TRUE}, each non-\code{NULL} element contains only
\code{weights} and \code{neighbor_indices}.  An element is \code{NULL} when the requested neighborhood contains no
usable observations.  For purely spatial and bivariate models there is one
list element per row of \code{loc}; for space-time models the order is location
then requested time.  The remaining components record the coordinates,
model, parameters, neighborhood settings, and function call used to construct
the local systems.
}
\references{
 Gaetan, C. and Guyon, X. (2010) \emph{Spatial Statistics and Modeling}.
 Springer-Verlag, New York.
}
\seealso{
\code{\link{GeoKrigloc}} for local kriging prediction,
  \code{\link{GeoKrigWeights}} for global kriging weights,
  \code{\link{GeoCovmatrix}} for covariance matrix construction.
}
\author{
 Moreno Bevilacqua, \email{moreno.bevilacqua@uai.cl}, 
 Víctor Morales-Oñate, \email{victor.morales@uv.cl}, 
 Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl}
}
\keyword{Kriging}

\name{MatSqrt, MatInv, MatLogDet}
\alias{MatSqrt}
\alias{MatInv}
\alias{MatLogDet}
\encoding{UTF-8}
\title{Square root, inverse and log determinant of a (semi)positive definite matrix, given a matrix decomposition.}
\description{
 Square root, inverse and log determinant of a (semi)positive definite matrix, given a matrix decomposition.
}
\usage{
MatSqrt(mat.decomp,method) 
MatInv(mtx)
MatLogDet(mat.decomp,method)
}

\value{The function returns a square root or inverse or log determinant of a (semi)positive definite matrix using the function in the FastGP package.}

\arguments{
 \item{mtx}{numeric; a squared symmetric positive definite matrix.}
 \item{mat.decomp}{numeric; a matrix decomposition.}
 \item{method}{string; the type of matrix decomposition. Two possible choices: \code{cholesky} and \code{svd}.}
}

\seealso{\code{\link{MatDecomp}}}



\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{
 library(GeoModels)
 ################################################################
 ###
 ### Example 1. Inverse of Covariance matrix associated to
 ### a Matern correlation model
 ###
 ###############################################################
 # Define the spatial-coordinates of the points:
 x <- runif(15, 0, 1)
 y <- runif(15, 0, 1)
 coords <- cbind(x,y)
 # Matern Parameters
 param=list(smooth=0.5,sill=1,scale=0.2,nugget=0)
 a=matrix <- GeoCovmatrix(coordx=coords, corrmodel="Matern", param=param)

 ## decomposition with cholesky method 
 b=MatDecomp(a$covmat,method="cholesky")
 ## inverse of covariance matrix
 inverse=MatInv(a$covmat)
}
\keyword{Composite}

\name{CheckSph}
\alias{CheckSph}
\encoding{UTF-8}
\title{Checking if a covariance is valid only on the sphere}
\description{
 Subroutine called by InitParam.
 The procedure controls if a covariance model is valid only on the sphere.
}
\usage{
CheckSph(numbermodel)}
\arguments{
 \item{numbermodel}{Numeric; the code number for the covariance model.}
}

\details{The function checks if a covariance is valid only on the sphere}
\value{Returns TRUE or FALSE} 


\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\keyword{Composite}

\name{madagascartemp}
\alias{madagascartemp}
\title{July Average Temperature of Madagascar}
\usage{data(madagascartemp)}
\description{
A \eqn{2500 \times 3}{2500 x 3} matrix containing UTM coordinates and July average temperatures at 2500 location sites in Madagascar, averaged over the period 1970--2000.\\
Data obtained using the \pkg{Geodata} package with the function \code{worldclim_country}.
}
\format{
A numerical matrix of dimension \eqn{2500 \times 3}{2500 x 3}.
}
\source{
Fick, S.E. and Hijmans, R.J. (2017).\\
WorldClim 2: new 1 km spatial resolution climate surfaces for global land areas.\\
\emph{International Journal of Climatology}, \bold{37}(12), 4302--4315.
}
\keyword{datasets}

\name{sp2Geo}
\alias{sp2Geo}
\encoding{UTF-8}
\title{Extracting information from an sp or spacetime object}
\description{Extracting information from an sp or spacetime object}
\usage{sp2Geo(spobj,spdata = NULL)}
\arguments{
 \item{spobj}{An object of class sp or spacetime}
 \item{spdata}{Character: The name of data in the sp or spacetime object}

}
\value{A list with spatio-temporal information}


\details{
The function accepts an \pkg{sp} or \pkg{spacetime} object and returns information used by GeoModels functions. Spatial objects should have a coordinate reference system (CRS) defined; if the CRS is missing, \code{sp2Geo} issues a warning because projected/geographic status cannot be determined reliably. For \code{STFDF} and \code{STIDF} objects, the current conversion represents the temporal positions by the sequential indices \code{1, 2, ..., T}; the original spacing of an irregular time index is therefore not preserved. To retain irregular temporal distances, use the explicit-coordinate interfaces of the relevant GeoModels functions and supply the numeric \code{coordt} values directly rather than passing \code{spobj}.
}
\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}
\examples{
if (requireNamespace("sp", quietly = TRUE)) {
# Define the spatial-coordinates of the points:
set.seed(3)
N <- 30 # number of location sites
x <- runif(N, 0, 1)
set.seed(6)
y <- runif(N, 0, 1)
coords <- cbind(x,y)

# Define spatial matrix covariates and regression parameters
X <- cbind(rep(1,N),runif(N))
# Define spatial matrix dependent variable
Y <- rnorm(nrow(X))

crs <- sp::CRS("EPSG:4326")
obj1 <- sp::SpatialPoints(coords, proj4string = crs)
obj2 <- sp::SpatialPointsDataFrame(
  coords, data = data.frame(X, Y), proj4string = crs
)

# sp2Geo info extraction
b <- sp2Geo(obj2,spdata = "Y")
class(b)
b
}
}

\keyword{Composite}

\name{spanish_wind}
\alias{spanish_wind}
\title{August monthly average wind speed in Spain between 1970-2000}
\usage{data(spanish_wind)}
\description{
 A (\eqn{6000x 3})-matrix containing lon/lat 
and august monthly average wind speed (2 m above the ground, meter/second) registered at 6000 location sites in the Iberian peninsula.
Data obtained from WorldClim version 2.1}
\format{A numerical matrix of dimension \eqn{6000 x 3}.}
\source{
 Fick, S.E., Hijmans, R.J. (2017) WorldClim 2: new 1km spatial resolution climate surfaces for global land areas.
 \emph{International Journal of Climatology}, \bold{37}, 4302--4315.
}

\keyword{datasets}

\name{CkModel}
\alias{CkModel}
\encoding{UTF-8}
\title{Checking random field Type}
\description{
Subroutine called by \code{InitParam}.\\
The procedure controls the type of random field inserted by the users.
}
\usage{
CkModel(model)
}
\arguments{
 \item{model}{String; the density associated with the likelihood
 objects. \code{Gaussian} is the default.}
}
\value{
The function returns a numeric positive integer, or \code{NULL} if the model is invalid.
}
\details{
The function controls the type of random field inserted by the users.
}
\seealso{\code{\link{GeoFit}}}
\author{
Moreno Bevilacqua \email{moreno.bevilacqua89@gmail.com} \url{https://sites.google.com/view/moreno-bevilacqua/home} \\
Víctor Morales Oñate \email{victor.morales@uv.cl} \url{https://sites.google.com/site/moralesonatevictor/} \\
Christian Caamaño-Carrillo \email{chcaaman@ubiobio.cl} \url{https://www.researchgate.net/profile/Christian-Caamano}
}
\keyword{Composite}

\name{GeoDoScores}
\alias{GeoDoScores}
\encoding{UTF-8}
\title{
Computation of drop-one predictive scores}
\description{
 The function computes RMSE, MAE, MAD, logarithmic score and CRPS from exact drop-one linear-prediction identities based on a \code{GeoCovmatrix} object.}
\usage{
GeoDoScores(data, method="cholesky", matrix)
}

\arguments{
 \item{data}{A \eqn{d}{d}-dimensional vector (a single spatial realisation) 
 or a a(\eqn{t \times d}{t x d})-matrix (a single spatio-temporal realisation).
 or a a(\eqn{2 \times d}{2 x d})-matrix (a single bivariate realisation).}
 \item{method}{String; matrix decomposition used for dense covariance matrices. Possible values are \code{"cholesky"} (default) and \code{"svd"}. Sparse covariance matrices use the sparse Cholesky path and therefore require \code{"cholesky"}.} 
 \item{matrix}{An object returned by \code{\link{GeoCovmatrix}}.}
}

\details{
 Let \eqn{Q=\Sigma^{-1}} and let \eqn{r} denote the data after subtraction of the model mean. The drop-one residual and conditional variance are computed without repeatedly refitting the model:
 \deqn{e_i^{(-i)} = (Qr)_i/Q_{ii}, \qquad v_i^{(-i)}=1/Q_{ii}.}
 The standardized residual is \eqn{z_i=(Qr)_i/\sqrt{Q_{ii}}}. The logarithmic score uses \eqn{\frac{1}{2}\{\log(2\pi v_i^{(-i)})+z_i^2\}}, and the Gaussian CRPS uses its standard closed-form expression with the normal density \eqn{\phi} and distribution function \eqn{\Phi}. The SVD path solves the linear system through the SVD itself rather than passing an SVD object to triangular solvers.}

\value{
 Returns a list containing the following information:
 \item{RMSE}{Root-mean-square error predictive score}
 \item{MAE}{Mean absolute drop-one prediction error.}
 \item{MAD}{Median absolute drop-one prediction error.}
 \item{LSCORE}{Mean Gaussian negative log predictive density for the drop-one predictions.}
 \item{CRPS}{Mean Gaussian continuous ranked probability score for the drop-one predictions.}
}


\references{
 Zhang H. and Wang Y. (2010). \emph{Kriging
 and cross-validation for massive spatial data}.
 Environmetrics, \bold{21}, 290--304.
Gneiting T. and Raftery A. \emph{Strictly Proper Scoring Rules, 
 Prediction, and Estimation}.
 Journal of the American Statistical Association, \bold{102}
}

\seealso{\code{\link{GeoCovmatrix}}}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{

library(GeoModels)

################################################################
######### Examples of predictive score computation ############
################################################################
set.seed(8)
 # Define the spatial-coordinates of the points:
x <- runif(500, 0, 2)
y <- runif(500, 0, 2)
coords=cbind(x,y)
matrix1 <- GeoCovmatrix(coordx=coords, corrmodel="Matern", param=list(smooth=0.5,
 sill=1,scale=0.2,nugget=0))
 
data <- GeoSim(coordx=coords, corrmodel="Matern", param=list(mean=0,smooth=0.5,
 sill=1,scale=0.2,nugget=0))$data

Pr_scores <- GeoDoScores(data,matrix=matrix1)

Pr_scores

}
\keyword{Predictive scores}

\name{MargParam}
\alias{MargParam}
\encoding{UTF-8}
\title{Lists the Nuisance Parameters of a random field}
\description{
The procedure returns a list with the nuisance parameters of a given
 random field model.
}
\usage{
MargParam(model, bivariate=FALSE,num_betas=c(1,1),copula=NULL)
}
\arguments{
 \item{model}{String; the name of a random field.}
 \item{bivariate}{Logical; if \code{FALSE} (the default) the correlation model is univariate spatial or spatio-temporal. 
 Otherwise is bivariate.}
 \item{num_betas}{Numerical; the nunber of mean parameters in the linear specification (default is 1) }
 \item{copula}{The type of copula.} 
}

\value{Return a vector string of nuisance parameters.}

\details{The function returns a list with the nuisance parameters of a given
 random field model.}
\seealso{\code{\link{GeoFit}}}


\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}


\examples{
library(GeoModels)

MargParam("Gaussian")

MargParam("Binomial")

MargParam("Weibull",num_betas=2)

MargParam("SkewGaussian", num_betas=3)

MargParam("SinhAsinh")

MargParam("Beta2",copula="Clayton")

MargParam("StudentT")
## note that in the bivariate case sill_1 e sill_2 are considered as correlation parameteres
MargParam("Gaussian", bivariate=TRUE)

}
\keyword{Composite}

\name{CkInput}
\alias{CkInput}
\encoding{UTF-8}
\title{Checking Input}
\description{
 Subroutine called by the fitting procedures.
 The procedure controls the the validity of the input inserted by the users.
}
\usage{
CkInput(coordx, coordy, coordz,coordt, coordx_dyn,corrmodel, data, distance, 
 fcall, fixed, grid,likelihood, maxdist, maxtime, 
 model, n, optimizer, param, radius,
 start, taper, tapsep, type, varest, 
 weighted,copula,X)
}
\arguments{
 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix 
 Coordinates on a sphere for a fixed radius \code{radius} 
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 
 \item{coordt}{A numeric vector assigning one dimension of temporal coordinates. The values need not be equally spaced; input validation does not impose a regular temporal grid.}
 \item{corrmodel}{String; the name of a correlation model, for the
 description see \code{\link{GeoFit}}.}
 \item{coordx_dyn}{A list of \eqn{m} numeric (\eqn{d_t \times 2}{d x 2})-matrices
 containing dynamical (in time) spatial coordinates. Optional argument, the default is \code{NULL}
 } 
 \item{data}{A numeric vector or a (\eqn{n \times d}{n x d})-matrix or (\eqn{d \times d
 \times n}{d x d x n})-matrix of observations.}
 \item{distance}{String; the name of the spatial distance. The default is \code{Eucl},
 the Euclidean distance. See Details for the accepted options.}
 \item{fcall}{String; \code{Fitting} to call the fitting procedure and
 \code{simulation} to call the simulation.}
 \item{fixed}{A named list giving the values of the parameters that
 will be considered as known values. The listed parameters for a
 given correlation function will be not estimated, i.e. if
 \code{list(nugget=0)} the nugget effect is ignored.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data
 are interpreted as a vector or a (\eqn{n \times d}{n x d})-matrix,
 instead if \code{TRUE} then (\eqn{d \times d
 \times n}{d x d x n})-matrix is considered.}
 \item{likelihood}{String; the configuration of the composite
 likelihood. \code{Marginal} is the default.}
 \item{maxdist}{Numeric; an optional positive value indicating the maximum
 spatial distance considered in the composite-likelihood
 computation.}
 \item{maxtime}{Numeric; an optional non-negative maximum temporal-distance threshold, expressed in the same units as \code{coordt}.}
 \item{radius}{Numeric; the radius of the sphere in the case of lon-lat coordinates.
 The default is 6371, the radius of the earth.} 
 \item{model}{String; the density associated to the likelihood
 objects. \code{Gaussian} is the default.}
 \item{n}{Numeric; the number of trials in a binomial random fields.
 Default is \eqn{1}.} 
 \item{optimizer}{String; the optimization algorithm
 (see \code{\link{optim}} for details). 'Nelder-Mead' is the
 default.}
 \item{param}{A numeric vector of parameters, needed only in
 simulation. See \code{\link{GeoSim}}.}
 \item{start}{A named list with the initial values of the
 parameters that are used by the numerical routines in maximization
 procedure. \code{NULL} is the default.}
 \item{taper}{String; the name of the tapered correlation
 function.}
 \item{tapsep}{Numeric; an optional value indicating the separability parameter in the space time quasi taper (see Details).}
 \item{type}{String; the type of the likelihood objects. If \code{Pairwise} (the
 default) then the marginal composite likelihood is formed by
 pairwise marginal likelihoods.}
 \item{varest}{Logical; if \code{TRUE} the estimate' variances and
 standard errors are returned.
 \code{FALSE} is the default.}
 \item{weighted}{Logical; if \code{TRUE} the likelihood objects are
 weighted. If \code{FALSE} (the
 default) the composite likelihood is not weighted.}
 \item{copula}{String; the type of copula. It can be "Clayton" or "Gaussian"}
 \item{X}{Numeric; Matrix of space-time covariates in the linear mean specification.}
}


\value{A list with the type of error associated with the input parameters.}

\details{Subroutine called by the fitting procedures.
 The procedure controls the the validity of the input inserted by the users.}
 
\seealso{\code{\link{GeoFit}}}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\keyword{Composite}

\name{GeoCovariogram}
\alias{GeoCovariogram}
\encoding{UTF-8}
\title{Computes the fitted variogram model.}
\description{
 The procedure computes and plots estimated covariance or semivariogram models of a Gaussian
 or a non-Gaussian spatial (temporal or bivariate spatial) random field. It allows to add 
 the empirical estimates in order to compare them with the fitted model.
}
\usage{
GeoCovariogram(fitted, distance="Eucl",answer.cov=FALSE,
 answer.vario=FALSE, answer.range=FALSE, fix.lags=NULL, 
 fix.lagt=NULL, show.cov=FALSE, show.vario=TRUE, 
 show.range=FALSE, add.cov=FALSE, add.vario=FALSE,
 pract.range=95, vario, invisible=FALSE, \dots)
}
\arguments{
 \item{fitted}{A fitted object obtained from the
 \code{\link{GeoFit}} or \code{\link{GeoWLS}} procedures.}
 \item{distance}{String; the name of the spatial distance. When omitted,
 the distance stored in \code{fitted} is used when available. The empirical
 lag values are taken from \code{vario}; the fitted and empirical objects
 should therefore use the same spatial metric.}
 \item{answer.cov}{Logical; if \code{TRUE} the estimated covariance
 function is returned. For a bivariate Gaussian fit, the returned components
 are \code{covariance11}, \code{covariance12}, and \code{covariance22}.}
 \item{answer.vario}{Logical; if \code{TRUE} a vector with the
 estimated variogram is returned; if \code{FALSE} (the
 default) the variogram is not returned.}
 \item{answer.range}{Logical; if \code{TRUE} the estimated
 practical range is returned; if \code{FALSE} (the
 default) the practical range is not returned.}
 \item{fix.lags}{Integer; a positive index in the empirical spatial-lag
 grid (including its zero-lag entry) used for the temporal profile. The
 theoretical surface is matched to the corresponding lag value rather than
 reusing this index on a different grid. For dynamic spatial supports, if the
 zero-lag entry is selected but no collocated pairs exist across distinct
 times, the plot automatically uses the first positive-distance space-time bin
 containing finite empirical pairs; the fitted curve is evaluated at that
 bin's center and the panel title reports the bin interval.}
 \item{fix.lagt}{Integer; a positive index in the empirical temporal-lag
 grid (including its zero-lag entry) used for the spatial profile. The
 theoretical surface is matched to the corresponding time-lag value.}
 \item{show.cov}{Logical; if \code{TRUE} the estimated
 covariance function is plotted; if \code{FALSE} (the
 default) the covariance function is not plotted.}
 \item{show.vario}{Logical; if \code{TRUE} the estimated
 variogram is plotted; if \code{FALSE} (the
 default) the variogram is not plotted.}
 \item{show.range}{Logical; if \code{TRUE} the estimated
 practical range is added on the plot; if \code{FALSE} (the
 default) the practical range is not added.}
 \item{add.cov}{Logical; if \code{TRUE} the vector of the
 estimated covariance function is added on the current plot; if \code{FALSE} (the
 default) the covariance is not added.}
 \item{add.vario}{Logical; if \code{TRUE} the vector with the
 estimated variogram is added on the current plot; if \code{FALSE} (the
 default) the correlation is not added.}
 \item{pract.range}{Numeric scalar in \code{[0,100]}; the percentage of
 the marginal variance reached by the semivariogram when computing the
 practical range.}
 \item{vario}{A \code{Variogram} object obtained from the
 \code{\link{GeoVariogram}} procedure.}
 \item{invisible}{Logical; if \code{TRUE}, compute the sum of squared
 differences between the empirical and fitted semivariograms on the empirical
 lag grid. This diagnostic is computed independently of whether the variogram
 is plotted.}
 \item{\dots}{other optional parameters which are passed to plot functions.}
}

\details{

When a fitted univariate model has a nonconstant mean, the returned
lag-only theoretical curve requires one representative marginal location.
The function uses the average fitted linear predictor \eqn{X\widehat\beta};
for a site-specific fixed mean it uses the average of that vector.  This does
not alter the fitted mean used by likelihood, residual, simulation, or
kriging calculations.
For a univariate \code{GeoFit} object fitted with \code{copula="Gaussian"},
\code{copula="Clayton"}, or \code{copula="SkewGaussian"}, the theoretical
covariance and semivariogram are computed on the observed marginal scale with
the same covariance engine used by \code{\link{GeoCovmatrix}} and
\code{\link{GeoKrig}}. The supported continuous margins are
\code{"Gaussian"}, \code{"StudentT"}, \code{"LogGaussian"},
\code{"Gamma"}, \code{"Weibull"}, \code{"Beta"}, \code{"Beta2"},
\code{"Kumaraswamy"}, \code{"Kumaraswamy2"}, \code{"Logistic"},
\code{"SkewLaplace"}, \code{"Tukeyh"}, \code{"Tukeyh2"}, and
\code{"SinhAsinh"}. For the Gaussian copula, the monotone transformed-Gaussian
Tukey and sinh--arcsinh margins use their exact model covariance. Clayton-like covariance uses cached deterministic
copula quadrature and skew-Gaussian covariance uses the cached Hermite
representation also used for linear prediction.

For \code{"Beta2"} and \code{"Kumaraswamy2"} with covariates, and more
generally whenever the marginal location varies over the observation sites,
a single lag-only covariance curve is not uniquely defined by distance alone.
The displayed curve therefore uses the representative marginal predictor
described above (the average fitted linear predictor). It is a representative
theoretical semivariogram for diagnostic comparison with the empirical one.

For non-Gaussian univariate models, the practical range is computed directly
from the fitted observed-scale covariance and semivariogram. This avoids using
a Gaussian-correlation range as a surrogate for a transformed marginal model.
If the requested percentage is not reached on the evaluated lag grid, the
returned practical range is \code{NA}. For a Gaussian model, the correlation
root calculation is retained. A bivariate practical range is not defined by
this function and is returned as \code{NA}.

For space-time profile plots, \code{fix.lags} and \code{fix.lagt} select an
empirical lag value; the closest point on the fitted 150-point surface grid is
then used. This keeps empirical and theoretical profiles aligned even though
their grids have different resolutions. For dynamic spatial supports, finite
empirical space-time bins are displayed as points rather than vertical stems.
The empirical temporal margin \eqn{\gamma(0,u)} is used when actual collocated
locations are available at distinct times. If \code{fix.lags=1} requests that
zero-lag profile but no collocated dynamic pairs exist, the plot instead uses
the first positive-distance space-time bin with finite empirical values. Thus
the panel represents a near-zero spatial band (for example \eqn{0<h<h_1})
rather than displaying an empty empirical profile; the fitted curve is
evaluated at the center of the selected spatial bin. The title reports the
actual bin interval. No zero-valued empirical temporal observations are
imputed, and the underlying \code{GeoVariogram} temporal margin remains the
true collocated \eqn{\gamma(0,u)} margin.

For a non-copula LogGaussian model, covariance and semivariogram are reported
on the observed scale, including the representative factor \eqn{\exp(2\eta)}
in the marginal variance, consistently with \code{GeoCovmatrix}.

For a non-copula \code{SinhAsinh} model, the nugget attenuation is applied to
the latent Gaussian correlation before the nonlinear sinh--arcsinh correlation
map is evaluated. This is the same ordering used by \code{GeoCovmatrix},
\code{GeoCorrFct}, and \code{GeoKrig}.

Objects returned by \code{GeoResiduals()} with
\code{residual_type = "Pearson"} are intentionally rejected.  The empirical
variogram of Pearson residuals is useful by itself, but it is not on the same
scale as the response covariance model stored in the original discrete
\code{GeoFit} object.

The function computes the fitted variogram model}

\value{
By default the function produces the requested plot invisibly. When
\code{answer.cov=TRUE} and/or \code{answer.vario=TRUE}, it invisibly returns a
list containing the lag grid and the fitted covariance and/or semivariogram.
When \code{answer.range=TRUE}, the practical range is included as
\code{range}. With \code{invisible=TRUE} and no requested answer, the
sum-of-squares diagnostic used by the function is returned invisibly. For a
bivariate Gaussian fit with \code{answer.cov=TRUE}, the list contains
\code{covariance11}, \code{covariance12}, and \code{covariance22}.
}
 

\references{
Cressie, N. A. C. (1993) \emph{Statistics for Spatial Data}. New York:
Wiley.

Gaetan, C. and Guyon, X. (2010) Spatial Statistics and Modelling.
 \emph{Springer-Verlag, New York}.
}

\seealso{\code{\link{GeoFit}}.}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}



\examples{
library(GeoModels)

################################################################
###
### Example 1. Plot of fitted covariance and fitted 
### and empirical semivariogram from a Gaussian RF 
### with Matern correlation. 
###
###############################################################
set.seed(21)
# Set the coordinates of the points:
x = runif(300, 0, 1)
y = runif(300, 0, 1)
coords=cbind(x,y)

# Set the model's parameters:
corrmodel = "Matern"
model = "Gaussian"
mean = 0
sill = 1
nugget = 0
scale = 0.2/3
smooth=0.5

param=list(mean=mean,sill=sill, nugget=nugget, scale=scale, smooth=smooth)
# Simulation of the Gaussian random field:
data = GeoSim(coordx=coords, corrmodel=corrmodel, model=model,param=param)$data
I=Inf
start=list(mean=0,scale=scale,sill=sill)
lower=list(mean=-I,scale=0,sill=0)
upper=list(mean= I,scale=I,sill=I)
fixed=list(nugget=nugget,smooth=smooth)
# Maximum composite-likelihood fitting of the Gaussian random field:
fit = GeoFit(data=data,coordx=coords, corrmodel=corrmodel,model=model,
 likelihood="Marginal",type='Pairwise',start=start,
 lower=lower,upper=upper,
 optimizer="nlminb", fixed=fixed,neighb=3)

# Empirical estimation of the variogram:
vario = GeoVariogram(data=data,coordx=coords,maxdist=0.5)

# Plot of covariance and variogram functions:
GeoCovariogram(fit,show.vario=TRUE, vario=vario,pch=20)

################################################################
###
### Example 2. Plot of fitted covariance and fitted 
### and empirical semivariogram from a Bernoulli 
### RF with Genwend correlation. 
###
###############################################################
set.seed(2111)

model="Binomial";n=1
# Set the coordinates of the points:
x = runif(500, 0, 1)
y = runif(500, 0, 1)
coords=cbind(x,y)

# Set the model's parameters:
corrmodel = "GenWend"
mean = 0
nugget = 0
scale = 0.2
smooth=0
power=4
param=list(mean=mean, nugget=nugget, scale=scale,smooth=0,power2=4)
# Simulation of the Gaussian RF:
data = GeoSim(coordx=coords, corrmodel=corrmodel, model=model,param=param,n=n)$data

start=list(mean=0,scale=scale)
fixed=list(nugget=nugget,power2=4,smooth=0)
# Maximum composite-likelihood fitting of the Binomial random field:
fit = GeoFit(data,coordx=coords, corrmodel=corrmodel,model=model,
 likelihood="Marginal",type='Pairwise',start=start,n=n,
 optimizer="BFGS", fixed=fixed,neighb=4)

# Empirical estimation of the variogram:
vario = GeoVariogram(data,coordx=coords,maxdist=0.5)

# Plot of covariance and variogram functions:
GeoCovariogram(fit, show.vario=TRUE, vario=vario,pch=20,ylim=c(0,0.3))


################################################################
###
### Example 3. Plot of fitted covariance and fitted 
### and empirical semivariogram from a Weibull RF
### with Wend0 correlation. 
###
###############################################################
set.seed(111)

model="Weibull";shape=4
# Set the coordinates of the points:
x = runif(700, 0, 1)
y = runif(700, 0, 1)
coords=cbind(x,y)

# Set the model's parameters:
corrmodel = "Wend0"
mean = 0
nugget = 0
scale = 0.4
power2=4


param=list(mean=mean, nugget=nugget, scale=scale,shape=shape,power2=power2)
# Simulation of the Gaussian RF:
data = GeoSim(coordx=coords, corrmodel=corrmodel, model=model,param=param)$data

start=list(mean=0,scale=scale,shape=shape)
I=Inf
lower=list(mean=-I,scale=0,shape=0)
upper=list(mean= I,scale=I,shape=I)
fixed=list(nugget=nugget,power2=power2)

fit = GeoFit(data,coordx=coords, corrmodel=corrmodel,model=model,
 likelihood="Marginal",type='Pairwise',start=start,
 lower=lower,upper=upper,
 optimizer="nlminb", fixed=fixed,neighb=3)

# Empirical estimation of the variogram:
vario = GeoVariogram(data,coordx=coords,maxdist=0.5)

# Plot of covariance and variogram functions:
GeoCovariogram(fit, show.vario=TRUE, vario=vario,pch=20)

################################################################
###
### Example 4. Plot of fitted and empirical semivariogram
### from a space time Gaussian random fields 
### with double Matern correlation. 
###
###############################################################
set.seed(92)
# Define the spatial-coordinates of the points:
x = runif(50, 0, 1)
y = runif(50, 0, 1)
coords=cbind(x,y)
# Define the temporal sequence:
time = seq(0, 10, 1)


param=list(mean=mean,nugget=nugget,
 smooth_s=0.5,smooth_t=0.5,scale_s=0.5/3,scale_t=2/2,sill=sill)
# Simulation of the spatio-temporal Gaussian random field:
data = GeoSim(coordx=coords, coordt=time, corrmodel="Matern_Matern",param=param)$data

fixed=list(nugget=0, mean=0, smooth_s=0.5,smooth_t=0.5)
start=list(scale_s=0.2, scale_t=0.5, sill=1)
# Maximum composite-likelihood fitting of the space-time Gaussian random field:
fit = GeoFit(data, coordx=coords, coordt=time, corrmodel="Matern_Matern", maxtime=1,
 neighb=3, likelihood="Marginal", type="Pairwise",fixed=fixed, start=start)

# Empirical estimation of spatio-temporal covariance:
vario = GeoVariogram(data,coordx=coords, coordt=time, maxtime=5,maxdist=0.5)

# Plot of the fitted space-time variogram
GeoCovariogram(fit,vario=vario,show.vario=TRUE)

# Plot of covariance, variogram and spatio and temporal profiles:
GeoCovariogram(fit,vario=vario,fix.lagt=1,fix.lags=1,show.vario=TRUE,pch=20)


################################################################
###
### Example 5. Plot of fitted and empirical semivariogram
### from a bivariate Gaussian random fields 
### with Matern correlation. 
###
###############################################################
set.seed(92)
# Define the spatial-coordinates of the points:
x <- runif(600, 0, 2)
y <- runif(600, 0, 2)
coords <- cbind(x,y)

# Simulation of a bivariate spatial Gaussian RF:
# with a Bivariate Matern
set.seed(12)
param=list(mean_1=4,mean_2=2,smooth_1=0.5,smooth_2=0.5,smooth_12=0.5,
 scale_1=0.12,scale_2=0.1,scale_12=0.15,
 sill_1=1,sill_2=1,nugget_1=0,nugget_2=0,pcol=-0.5)
data <- GeoSim(coordx=coords,corrmodel="Bi_matern",
 param=param)$data



# selecting fixed and estimated parameters
fixed=list(mean_1=4,mean_2=2,nugget_1=0,nugget_2=0,
 smooth_1=0.5,smooth_2=0.5,smooth_12=0.5)
start=list(sill_1=var(data[1,]),sill_2=var(data[2,]),
 scale_1=0.1,scale_2=0.1,scale_12=0.1,
 pcol=cor(data[1,],data[2,]))


# Maximum marginal pairwise likelihood
fitcl<- GeoFit(data=data, coordx=coords, corrmodel="Bi_Matern",
 likelihood="Marginal",type="Pairwise",
 optimizer="BFGS" , start=start,fixed=fixed,
 neighb=4)
print(fitcl)

# Empirical estimation of spatio-temporal covariance:
vario = GeoVariogram(data,coordx=coords,maxdist=0.4,bivariate=TRUE)
GeoCovariogram(fitcl,vario=vario,show.vario=TRUE,pch=20)
}

\keyword{Composite}

\name{CorrParam}
\alias{CorrParam}
\encoding{UTF-8}
\title{Lists the Parameters of a Correlation Model}
\description{
 The procedure returns a list with the names of the parameters of a given
 correlation model.
}
\usage{
CorrParam(corrmodel)
}

\arguments{
 \item{corrmodel}{String: the name associated to a given correlation model.}
}

\value{Return a vector string of correlation parameters.}

\details{The function returns a list with the Parameters of a Correlation Model}

\seealso{\code{\link{GeoCovmatrix}}}



\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{
require(GeoModels)
################################################################
###
### Example 1. Parameters of the Matern model
###
###############################################################

CorrParam("Matern")


################################################################
###
### Example 2. Parameters of the Generalized Wendland model
###
###############################################################

CorrParam("GenWend")


################################################################
###
### Example 3. Parameters of the Generalized Cauchy model
###
###############################################################

CorrParam("GenCauchy")


################################################################
###
### Example 4. Parameters of the space time Gneiting model
###
###############################################################

CorrParam("Gneiting")


################################################################
###
### Example 5. Parameters of the bi-Matern separable model.
### Note that in the bivariate case variance paramters are
### included
###############################################################

CorrParam("Bi_Matern_sep")
}
\keyword{Composite}

\name{GeoAniso}
\alias{GeoAniso}
\encoding{UTF-8}
\title{Spatial Anisotropy correction}
\description{Transforms or back-transforms a set of coordinates according to the geometric anisotropy parameters. }
\usage{GeoAniso(coords, anisopars=c(0,1), inverse = FALSE)}
\arguments{
 \item{coords}{An n x 2 matrix with the coordinates to be transformed.}
 \item{anisopars}{ A bivariate vector with the the anisotropy angle and the anisotropy ratio, respectively. The angle must be given in radians in [0,pi] and the anisotropy ratio must be greater or equal than 1.}
 \item{inverse}{Logical: Default to FALSE. If TRUE the reverse transformation is performed.}

}

\value{Returns a matrix of transformed coordinates}



\details{
Geometric anisotropy is defined by a linear tranformation from the anisotropic space to the isotropic space that is \deqn{Y = X R S}
 where X is a matrix with original coordinates (anisotropic space), and Y is a matrix with transformed coordinates (isotropic space). 
 Here R is a rotation matrix with associated anisotropy angle parameter (in \eqn{[0,pi]}) and a \eqn{S} is a shrinking matrix with associated anisotropy ratio 
 parameter (greeater or equal than one).
The two parameters are specified in the anisopars argument as a bivariate numeric vector. The case \eqn{(.,1)} corresponds to the isotropic case.
}
\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}


\keyword{Composite}

\name{GeoTests}
\alias{GeoTests}
\encoding{UTF-8}
\title{Statistical Hypothesis Tests for Nested Models}
\description{
Performs Wald and likelihood-ratio tests for successively nested
\code{GeoFit} models. Standard likelihood fits can be compared using
Wald or Wilks tests, whereas composite likelihood fits can be compared
using Wald or Satterthwaite-adjusted composite likelihood-ratio tests.
}

\usage{
GeoTests(object1, object2, \dots, statistic)
}

\arguments{
  \item{object1}{An unrestricted fitted model of class \code{GeoFit}.}
  \item{object2}{A restricted fitted model of class \code{GeoFit}, nested
  within \code{object1}.}
  \item{\dots}{Further fitted models of class \code{GeoFit}, ordered from
  less restricted to more restricted. Adjacent models are compared.}
  \item{statistic}{A character string specifying the test statistic.
  Available choices are \code{"Wald"}, \code{"Wilks"}, and
  \code{"WilksS"}. See Details.}
}

\details{
The models must be supplied from the unrestricted model to successively
more restricted models. For each adjacent pair, the tested parameters are
the parameters estimated in the unrestricted model but fixed in the
restricted model. Each tested parameter must therefore occur explicitly
in the \code{fixed} component of the restricted fit.

The function checks that adjacent models use the same data, coordinates,
design matrix, marginal model, correlation model, likelihood type, pair
construction, and other relevant fitting options. Common fixed parameters
must have the same values in the two models.

The following statistics are available:

\itemize{
  \item \code{"Wald"}: available for both standard and composite
  likelihood fits. The statistic is
  \deqn{
  W_{\mathrm{Wald}} =
  (\widehat{\boldsymbol\psi}-\boldsymbol\psi_0)^\top
  \widehat V_{\psi\psi}^{-1}
  (\widehat{\boldsymbol\psi}-\boldsymbol\psi_0),
  }
  where \eqn{\boldsymbol\psi} is the vector of tested parameters and
  \eqn{\widehat V_{\psi\psi}} is the corresponding block of the estimated
  covariance matrix of the unrestricted estimator. Under the null
  hypothesis, the statistic is asymptotically chi-squared with degrees of
  freedom equal to the number of restrictions. For composite likelihood
  fits, \code{varcov} must contain the Godambe covariance matrix, usually
  obtained by applying \code{GeoVarest} to the unrestricted fit.

  \item \code{"Wilks"}: available only for standard full-likelihood fits.
  The statistic is
  \deqn{
  W_{\mathrm{Wilks}} =
  2\{\ell(\widehat{\boldsymbol\theta})-
  \ell(\widetilde{\boldsymbol\theta})\},
  }
  where \eqn{\widehat{\boldsymbol\theta}} and
  \eqn{\widetilde{\boldsymbol\theta}} are the unrestricted and restricted
  maximum likelihood estimates, respectively. Under standard regularity
  conditions, the statistic is asymptotically chi-squared with degrees of
  freedom equal to the number of restrictions.

  \item \code{"WilksS"}: available only for composite likelihood fits,
  including marginal, conditional, and difference composite likelihoods.
  Let
  \deqn{
  W_{CL}=2\{\ell_{CL}(\widehat{\boldsymbol\theta})-
  \ell_{CL}(\widetilde{\boldsymbol\theta})\}.
  }
  Its asymptotic null distribution is generally a weighted sum
  \deqn{
  W_{CL}\ \dot{\sim}\ \sum_{j=1}^{r}\lambda_j\chi^2_{1,j},
  }
  where \eqn{r} is the number of restrictions. In the implementation,
  \eqn{\lambda_1,\ldots,\lambda_r} are the eigenvalues of
  \deqn{
  H_{\mathrm{eff}}V_{\psi\psi}, \qquad
  H_{\mathrm{eff}}=\{(H^{-1})_{\psi\psi}\}^{-1},
  }
  with \eqn{H} the sensitivity matrix and
  \eqn{V=H^{-1}JH^{-1}} the Godambe covariance matrix from the unrestricted
  fit. The Satterthwaite moment-matching constants are
  \deqn{
  \nu=\frac{(\sum_j\lambda_j)^2}{\sum_j\lambda_j^2},
  \qquad
  \kappa=\frac{\sum_j\lambda_j^2}{\sum_j\lambda_j},
  }
  and the adjusted statistic
  \deqn{
  W_S=\frac{W_{CL}}{\kappa}
  }
  is compared with a chi-squared distribution having \eqn{\nu} degrees of
  freedom. The unrestricted fit must contain both \code{sensmat} and the
  Godambe covariance matrix \code{varcov}; these can be obtained using
  \code{sensitivity = TRUE} in \code{GeoFit}, followed by \code{GeoVarest}.
}

For a composite likelihood-ratio comparison, the unrestricted and
restricted fits must be based on exactly the same retained pairs. This is
automatic for deterministic pair constructions. With stochastic thinning,
the two fits must use the same thinning realization; \code{GeoTests} checks
that their \code{rowidx} and \code{colidx} components are identical and
stops otherwise.

The matrices used by \code{"WilksS"} are evaluated at the unrestricted
fit. Under the null hypothesis, this is a consistent plug-in evaluation
because the unrestricted and restricted estimators converge to the same
parameter value.
}

\value{
A data frame containing one row for each supplied model. The first row
corresponds to the unrestricted model and contains missing test results.
Each subsequent row reports the comparison between that model and the
model in the preceding row. The columns are:

\describe{
  \item{Num.Par}{Number of estimated parameters in the model.}
  \item{Diff.Par}{Number of restrictions relative to the preceding model.}
  \item{Df}{Degrees of freedom used for the reference chi-squared
  distribution. For \code{"WilksS"}, this is the Satterthwaite effective
  degrees of freedom and need not be an integer.}
  \item{Chisq}{Observed value of the test statistic.}
  \item{Pr(>chisq)}{Upper-tail p-value from the chi-squared reference
  distribution with the reported degrees of freedom.}
}
}

\references{

Kent, J. T. (1982). Robust Properties of Likelihood Ratio Tests.
\emph{Biometrika}, \bold{69}, 19--27.

Satterthwaite, F. E. (1946). An Approximate Distribution of Estimates of
Variance Components. \emph{Biometrics Bulletin}, \bold{2}, 110--114.

Varin, C., Reid, N. and Firth, D. (2011). An Overview of Composite
Likelihood Methods. \emph{Statistica Sinica}, \bold{21}, 5--42.
}

\author{
Moreno Bevilacqua,
\email{moreno.bevilacqua89@gmail.com},
\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate,
\email{victor.morales@uv.cl},
\url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo,
\email{chcaaman@ubiobio.cl},
\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\seealso{
\code{\link{GeoFit}}, \code{\link{GeoVarest}}.
}

\examples{
library(GeoModels)

################################################################
### Example 1. Test of a regression coefficient using
### conditional pairwise composite likelihood
################################################################
set.seed(342)
model <- "Gaussian"

NN <- 1500
x <- runif(NN, 0, 1)
y <- runif(NN, 0, 1)
coords <- cbind(x, y)

mean <- 1
mean1 <- -1.25
nugget <- 0
sill <- 1

X <- cbind(1, runif(nrow(coords)))

corrmodel <- "Wend0"
power2 <- 4
c_supp <- 0.15

param <- list(
  power2 = power2, mean = mean, mean1 = mean1,
  sill = sill, scale = c_supp, nugget = nugget
)

data <- GeoSim(
  coordx = coords, corrmodel = corrmodel,
  model = model, param = param, X = X
)$data

I <- Inf

## H1: regression mean with a nonzero slope
fixed <- list(nugget = nugget, power2 = power2)
start <- list(mean = mean, mean1 = mean1, scale = c_supp, sill = sill)
lower <- list(mean = -I, mean1 = -I, scale = 0, sill = 0)
upper <- list(mean = I, mean1 = I, scale = I, sill = I)

fitH1 <- GeoFit(
  data = data, coordx = coords, corrmodel = corrmodel,
  model = model, likelihood = "Conditional", type = "Pairwise",
  sensitivity = TRUE, lower = lower, upper = upper, neighb = 3,
  optimizer = "nlminb", X = X, start = start, fixed = fixed
)

## H0: constant mean, that is, mean1 = 0
fixed0 <- list(power2 = power2, nugget = nugget, mean1 = 0)
start0 <- list(mean = mean, scale = c_supp, sill = sill)
lower0 <- list(mean = -I, scale = 0, sill = 0)
upper0 <- list(mean = I, scale = I, sill = I)

fitH0 <- GeoFit(
  data = data, coordx = coords, corrmodel = corrmodel,
  model = model, likelihood = "Conditional", type = "Pairwise",
  sensitivity = TRUE, lower = lower0, upper = upper0, neighb = 3,
  optimizer = "nlminb", X = X, start = start0, fixed = fixed0
)

\dontrun{
## GeoVarest is needed on the unrestricted fit to estimate its
## Godambe covariance matrix. Increase K for final analyses.
fitH1 <- GeoVarest(fitH1, K = 100, seed = 1234)

GeoTests(fitH1, fitH0, statistic = "Wald")
GeoTests(fitH1, fitH0, statistic = "WilksS")
}

################################################################
### Example 2. Test of Gaussianity for a Sinh--arcsinh random field
### using the standard full likelihood
################################################################
set.seed(99)
model <- "SinhAsinh"

NN <- 200
x <- runif(NN, 0, 1)
y <- runif(NN, 0, 1)
coords <- cbind(x, y)

mean <- 0
nugget <- 0
sill <- 1
skew <- 0
tail <- 1

corrmodel <- "Wend0"
power2 <- 4
c_supp <- 0.2

param <- list(
  power2 = power2, skew = skew, tail = tail,
  mean = mean, sill = sill, scale = c_supp, nugget = nugget
)

data <- GeoSim(
  coordx = coords, corrmodel = corrmodel,
  model = model, param = param
)$data

## H1: unrestricted Sinh--arcsinh model
fixed <- list(power2 = power2, nugget = nugget, mean = mean)
start <- list(scale = c_supp, skew = skew, tail = tail, sill = sill)
lower <- list(scale = 0, skew = -I, tail = 0, sill = 0)
upper <- list(scale = I, skew = I, tail = I, sill = I)

fitH1 <- GeoFit2(
  data = data, coordx = coords, corrmodel = corrmodel,
  model = model, likelihood = "Full", type = "Standard",
  varest = TRUE, lower = lower, upper = upper,
  optimizer = "nlminb", start = start, fixed = fixed
)

## H0: Gaussianity, that is, skew = 0 and tail = 1
fixed0 <- list(
  power2 = power2, nugget = nugget, mean = mean,
  tail = 1, skew = 0
)
start0 <- list(scale = c_supp, sill = sill)
lower0 <- list(scale = 0, sill = 0)
upper0 <- list(scale = 2, sill = 5)

fitH0 <- GeoFit(
  data = data, coordx = coords, corrmodel = corrmodel,
  model = model, likelihood = "Full", type = "Standard",
  varest = TRUE, lower = lower0, upper = upper0,
  optimizer = "nlminb", start = start0, fixed = fixed0
)

GeoTests(fitH1, fitH0, statistic = "Wald")
GeoTests(fitH1, fitH0, statistic = "Wilks")
}

\keyword{spatial}
\keyword{htest}

\name{SimCE}
\alias{SimCE}
\encoding{UTF-8}
\title{Circulant embedding simulation on a regular spatial grid}
\description{
Internal subroutine used by \code{GeoSimapprox} to simulate a Gaussian
random field on an exact, regular two-dimensional grid by circulant
embedding.
}
\usage{
SimCE(M, N, x, y, z = NULL, corrmodel, param,
      mean.val = 0, max.ext = 1L)
}
\arguments{
 \item{M}{Integer; number of coordinates in \code{x}.}
 \item{N}{Integer; number of coordinates in \code{y}.}
 \item{x}{A finite, strictly increasing, equally spaced numeric vector of length \code{M}.}
 \item{y}{A finite, strictly increasing, equally spaced numeric vector of length \code{N}.}
 \item{z}{Must be \code{NULL}. Three-dimensional CE grids are not implemented.}
 \item{corrmodel}{String; the name of a spatial correlation model.}
 \item{param}{A list of correlation parameters. \code{sill} is the marginal variance and \code{nugget} is the nugget proportion.}
 \item{mean.val}{A finite scalar or a vector with one value per grid point.}
 \item{max.ext}{Positive integer; maximum number of successive doubling attempts for the embedding.}
}
\details{
The simulated coordinates are exactly \code{expand.grid(x, y)}. The method
uses Euclidean distance. If the embedding still has substantial negative
eigenvalues at the maximum extension, they are truncated to zero and a
warning is issued.
}
\value{
A list containing \code{X}, the simulated values, \code{grid.points}, the
exact grid coordinates, \code{k0}, the selected extension attempt,
\code{approx.flag}, and the minimum embedding eigenvalue.
}
\seealso{\code{\link{GeoSimapprox}}}
\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},
Víctor Morales Oñate, \email{victor.morales@uv.cl},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl}}
\keyword{Simulation}

\name{WlsStart}
\alias{WlsStart}
\encoding{UTF-8}
\title{Computes Starting Values based on Weighted Least Squares}
\description{
 Subroutine called by \code{\link{GeoFit}}.
 The function returns appropriate starting values for the
 composite-likelihood fitting procedure based on weighted least squares.
}
\usage{
WlsStart(coordx, coordy, coordz, coordt, coordx_dyn, corrmodel, data,
 distance, fcall, fixed, grid, likelihood, maxdist, neighb,
 maxtime, model, n, param, parscale, paramrange, radius, start,
 taper, tapsep, type, varest, weighted, copula, X, memdist,
 nosym, p_neighb, thin_method)
}
\arguments{
 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix.
 Coordinates on a sphere for a fixed radius \code{radius} are passed in lon/lat format expressed in decimal degrees.}

 \item{coordy}{A numeric vector giving 1-dimension of spatial coordinates; optional argument, default is \code{NULL}.}

 \item{coordz}{A numeric vector giving 1-dimension of spatial coordinates; optional argument, default is \code{NULL}.}

 \item{coordt}{A numeric vector assigning one dimension of temporal coordinates. Temporal coordinates may be irregularly spaced; temporal lags used during initialization are computed from the supplied coordinate values.}

 \item{coordx_dyn}{A list of \eqn{m} numeric (\eqn{d_t \times 2}{d_t x 2})-matrices containing dynamical (in time) spatial coordinates.
 Optional argument, default is \code{NULL}.}

 \item{corrmodel}{String; the name of a correlation model.}

 \item{data}{A numeric vector or a (\eqn{n \times d}{n x d})-matrix or (\eqn{d \times d \times n}{d x d x n})-array of observations.}

 \item{distance}{String; the name of the spatial distance. Default is \code{"Eucl"} (Euclidean distance).}

 \item{fcall}{String; \code{"fitting"} to call the fitting procedure and \code{"simulation"} to call the simulation procedure.}

 \item{fixed}{A named list giving the values of the parameters that will be considered as known values.}

 \item{grid}{Logical; if \code{FALSE} (default) the data are interpreted as a vector or a (\eqn{n \times d}{n x d})-matrix;
 if \code{TRUE} then a (\eqn{d \times d \times n}{d x d x n})-array is considered.}

 \item{likelihood}{String; the configuration of the composite likelihood.}

 \item{maxdist}{Numeric; an optional positive value indicating the maximum spatial distance considered in the composite-likelihood computation.}

 \item{neighb}{Numeric; an optional positive integer indicating the order of neighborhood in the composite likelihood computation.}

 \item{maxtime}{Numeric; an optional non-negative maximum temporal separation, expressed in the same units as \code{coordt}, considered in the composite-likelihood initialization.}

 \item{model}{String; the name of the model.}

 \item{n}{Numeric; number of trials in a binomial random field.}

 \item{param}{A numeric vector of parameter values required in the simulation procedure of random fields.}

 \item{parscale}{A numeric vector with scaling values for improving the maximisation routine.}

 \item{paramrange}{A numeric vector with the range of the parameter space.}

 \item{radius}{Numeric; a value indicating the radius of the sphere when using great circle distance.
 Default value is the radius of the earth in km (i.e., \code{6371}).}

 \item{start}{A numeric vector (or list) with starting values.}

 \item{taper}{String; the name of the type of covariance matrix. It can be \code{"Standard"} (default) or
 \code{"Tapering"} for tapered covariance matrices.}

 \item{tapsep}{Numeric; an optional value indicating the separability parameter in the space-time quasi taper (see Details).}

 \item{type}{String; the type of estimation method.}

 \item{varest}{Logical; if \code{TRUE} the estimates variances and standard errors are returned. Default is \code{FALSE}.}

 \item{weighted}{Logical; if \code{TRUE} the likelihood objects are weighted; see \code{\link{GeoFit}}.}

 \item{copula}{String; the type of copula. It can be \code{"Clayton"} or \code{"Gaussian"}.}

 \item{X}{Numeric; matrix of spatio(temporal) covariates in the linear mean specification.}

 \item{memdist}{Deprecated compatibility argument. \code{WlsStart} always precomputes the selected pair structure; \code{FALSE} produces a warning and is treated as \code{TRUE}.}

 \item{nosym}{Logical; if \code{TRUE} symmetric weights are not considered.}

 \item{p_neighb}{Numeric; a value in \eqn{(0,1]} specifying the expected fraction of nearest-neighbor pairs retained through stochastic thinning.
 If \code{1} (default), no thinning is applied and all nearest-neighbor pairs are used. If \code{<1}, pairs are randomly retained using
 independent Bernoulli sampling.}

 \item{thin_method}{String; thinning scheme used when \code{p_neighb<1}. Default is \code{"bernoulli"} (independent Bernoulli thinning).}
}
\value{
 A list with a set of useful information in the estimation procedure.
}
\details{
 Internal function called by \code{\link{GeoFit}}.
}
\seealso{
 \code{\link{GeoFit}}.
}
\author{
Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com}, \url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl}, \url{https://www.researchgate.net/profile/Christian-Caamano}
}
\keyword{LeastSquare}

\name{StartParam}
\alias{StartParam}
\encoding{UTF-8}
\title{Initializes the Parameters for Estimation Procedures}
\description{
 Subroutine called by the fitting procedures.
 The procedure initializes the parameters for the fitting procedure.
}
\usage{
StartParam(coordx, coordy, coordz, coordt, coordx_dyn, corrmodel,
 data, distance, fcall, fixed, grid, likelihood,
 maxdist, neighb, maxtime, model, n, param,
 parscale, paramrange, radius, start, taper, tapsep,
 type, typereal, weighted, copula, X, memdist, nosym,
 p_neighb, thin_method)
}
\arguments{
 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix.
 Coordinates on a sphere for a fixed radius \code{radius} are passed in lon/lat format expressed in decimal degrees.}

 \item{coordy}{A numeric vector giving 1-dimension of spatial coordinates; optional argument, default is \code{NULL}.}

 \item{coordz}{A numeric vector giving 1-dimension of spatial coordinates; optional argument, default is \code{NULL}.}

 \item{coordt}{A numeric vector assigning one dimension of temporal coordinates. Temporal coordinates may be irregularly spaced; temporal lags used during initialization are computed from the supplied coordinate values.}

 \item{coordx_dyn}{A list of \eqn{m} numeric (\eqn{d_t \times 2}{d_t x 2})-matrices containing dynamical (in time) spatial coordinates.
 Optional argument, default is \code{NULL}.}

 \item{corrmodel}{String; the name of a correlation model.}

 \item{data}{A numeric vector or a (\eqn{n \times d}{n x d})-matrix or (\eqn{d \times d \times n}{d x d x n})-array of observations.}

 \item{distance}{String; the name of the spatial distance. Default is \code{"Eucl"} (Euclidean distance). See Details for the accepted options.}

 \item{fcall}{String; \code{"fitting"} to call the fitting procedure and \code{"simulation"} to call the simulation procedure.}

 \item{fixed}{A named list giving the values of the parameters that will be considered as known values.}

 \item{grid}{Logical; if \code{FALSE} (default) the data are interpreted as a vector or a (\eqn{n \times d}{n x d})-matrix;
 if \code{TRUE} then a (\eqn{d \times d \times n}{d x d x n})-array is considered.}

 \item{likelihood}{String; the configuration of the composite likelihood.}

 \item{maxdist}{Numeric; an optional positive value indicating the maximum spatial distance considered in the composite-likelihood computation.}

 \item{neighb}{Numeric; an optional positive integer indicating the order of neighborhood in the composite likelihood computation.
 See Details for more information.}

 \item{maxtime}{Numeric; an optional non-negative maximum temporal lag, expressed in the same units as \code{coordt}, considered in the composite-likelihood initialization.}

 \item{model}{String; the density associated to the likelihood objects. \code{"Gaussian"} is the default.}

 \item{n}{Numeric; number of trials for binomial random fields.}

 \item{param}{A numeric vector of parameter values required in the simulation procedure of random fields.}

 \item{parscale}{A numeric vector of scaling factors to improve the maximizing procedure; see \code{\link{optim}}.}

 \item{paramrange}{A numeric vector of parameter ranges; see \code{\link{optim}}.}

 \item{radius}{Numeric; the radius of the sphere in the case of lon-lat coordinates.
 The default is \code{6371}, the radius of the earth.}

 \item{start}{A named list with the initial values of the parameters that are used by the numerical routines in the maximization procedure.}

 \item{taper}{String; the name of the type of covariance matrix. It can be \code{"Standard"} (default) or \code{"Tapering"} for tapered covariance matrices.}

 \item{tapsep}{Numeric; an optional value indicating the separability parameter in the space-time adaptive taper (see Details).}

 \item{type}{String; the type of likelihood objects. Temporary value set to \code{"WLeastSquare"} (weighted least-square) in order to compute starting values.}

 \item{typereal}{String; the real type of likelihood objects. See \code{\link{GeoFit}}.}

 \item{weighted}{Logical; if \code{TRUE} the likelihood objects are weighted; see \code{\link{GeoFit}}.}

 \item{copula}{String; the type of copula.}

 \item{X}{Numeric; matrix of space-time covariates.}

 \item{memdist}{Low-level compatibility flag. The \code{GeoFit} and \code{WlsStart} workflows always use \code{TRUE}; direct use is not recommended.}

 \item{nosym}{Logical; if \code{TRUE} symmetric weights are not considered.}

 \item{p_neighb}{Numeric; a value in \eqn{(0,1]} specifying the expected fraction of nearest-neighbor pairs retained through stochastic thinning.
 If \code{1} (default), no thinning is applied and all nearest-neighbor pairs are used. If \code{<1}, pairs are randomly retained using independent Bernoulli sampling.}

 \item{thin_method}{String; thinning scheme used when \code{p_neighb<1}. Default is \code{"bernoulli"} (independent Bernoulli thinning).}
}
\value{
 A list with a set of useful information in the estimation procedure.
}
\details{
 Internal function called by \code{\link{WlsStart}}.
}
\author{
Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com}, \url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl}, \url{https://www.researchgate.net/profile/Christian-Caamano}
}
\keyword{Composite}

\name{GeoScores}
\alias{GeoScores}
\encoding{UTF-8}

\title{Computation of predictive scores}

\description{
 The function computes predictive scores for observed validation values from
 point predictions, kriging prediction objects, or conditional-simulation
 objects. Gaussian predictive scores are available from \code{GeoKrig}/\code{GeoKrigloc}
 through the \code{pred}/\code{mse} convention, while \code{GeoSimcond}
 objects are scored from their empirical conditional predictive distribution.
}

\usage{
GeoScores(data_to_pred,
 probject = NULL,
 pred = NULL,
 mse = NULL,
 score = c("pe", "crps", "intscore", "coverage"),
 lower95 = NULL,
 upper95 = NULL,
 threshold = 0.5,
 na.rm = TRUE)
}

\arguments{
 \item{data_to_pred}{
 Numeric vector, matrix or array containing the observed validation values.
 Values are internally coerced to a numeric vector.
 }

 \item{probject}{
 Optional object of class \code{GeoKrig}, \code{GeoKrigloc}, or \code{GeoSimcond}.
 For \code{GeoKrig}/\code{GeoKrigloc}, point predictions and prediction variances
 are extracted from \code{probject$pred} and \code{probject$mse}. Requested
 probabilistic scores are computed from the Gaussian predictive convention
 \eqn{N(\widehat Y, MSE)} for every marginal model. For a non-Gaussian model
 this is a Gaussian approximation based on the optimal linear predictor and its
 MSE; \code{GeoScores} reports it without deciding whether that approximation is
 appropriate for a particular analysis.

 For \code{GeoSimcond}, the conditional simulations in \code{probject$condsim}
 define an empirical predictive distribution. Point scores use
 \code{probject$cond_mean}; CRPS, PIT, Brier score, prediction intervals,
 interval score, and coverage are computed directly from the conditional draws,
 without a Gaussian approximation.
 }

 \item{pred}{
 Numeric vector, matrix or array of point predictions. This argument is
 required when \code{probject} is not supplied.
 }

 \item{mse}{
 Optional numeric vector, matrix or array of prediction variances. Predictive
 standard errors are computed as \code{sqrt(mse)}. This argument is not needed
 for \code{"pe"}, for interval score or coverage when \code{lower95} and
 \code{upper95} are supplied, or when \code{probject} is a
 \code{GeoSimcond} object. It is required for probabilistic scores based on the
 Gaussian \code{pred}/\code{mse} convention unless both interval bounds are
 supplied and used to infer a Gaussian standard error.
 }

 \item{score}{
 Character vector specifying which predictive scores should be computed.
 Possible values are \code{"brie"}, \code{"crps"}, \code{"lscore"},
 \code{"pit"}, \code{"pe"}, \code{"intscore"}, and \code{"coverage"}.
 Several values can be supplied.
 }

 \item{lower95}{
 Optional numeric vector, matrix or array containing the lower bounds of the
 95 percent prediction intervals. If supplied together with \code{upper95},
 these bounds are used directly for \code{"intscore"} and \code{"coverage"}.
 For a \code{GeoSimcond} object they override the default empirical 2.5 and
 97.5 percent conditional-simulation quantiles. If \code{mse} is not supplied
 in the Gaussian \code{pred}/\code{mse} path, they can also be used to infer a
 Gaussian predictive standard error.
 }

 \item{upper95}{
 Optional numeric vector, matrix or array containing the upper bounds of the
 95 percent prediction intervals. See \code{lower95}.
 }

 \item{threshold}{
 Finite numeric scalar used to compute the Brier score \code{"brie"}. The binary event is \eqn{Y > threshold}.
 }

 \item{na.rm}{
 Logical. If \code{TRUE}, invalid observations are removed separately for each requested score class. Thus an invalid predictive variance does not remove an otherwise valid observation from MAE/RMSPE, and an invalid interval does not affect point scores. If \code{FALSE}, invalid values relevant to any requested score produce an error.
 }
}

\details{
 \code{GeoScores} dispatches predictive scoring according to the information
 contained in \code{probject}.

 For a \code{GeoKrig} or \code{GeoKrigloc} object, \code{pred} is the point
 predictor and \code{mse} its mean squared prediction error. The option
 \code{"pe"} returns mean absolute error and root mean squared prediction
 error. Whenever probabilistic scores are requested, \code{GeoScores} uses the
 Gaussian predictive convention
 \deqn{
 Y_0\mid Y \ \dot\sim\ N\{\widehat Y_0, MSE_0\}.
 }
 This convention is used for every marginal model. For a Gaussian random field
 it is the natural conditional predictive distribution (conditional on fitted
 parameters). For a non-Gaussian model, \code{GeoKrig}/\code{GeoKrigloc}
 still returns an optimal linear predictor, so CRPS, LogScore, PIT, Brier score,
 and Gaussian intervals obtained this way must be interpreted as Gaussian
 predictive approximations based on the OLP and its MSE. \code{GeoScores} does
 not suppress those quantities; their scientific interpretation is left to the
 user.

 Predictions and prediction variances can equivalently be supplied directly
 through \code{pred} and \code{mse}; this uses the same Gaussian predictive
 convention.

 For a \code{GeoSimcond} object, the replications stored in
 \code{probject$condsim} are treated as an empirical conditional predictive
 distribution at each prediction location. No Gaussian approximation is used.
 The point forecast for \code{"pe"} is \code{probject$cond_mean} (or the
 empirical conditional mean reconstructed from the simulations if needed).
 The Brier probability for the event \eqn{Y>threshold} is the empirical
 conditional exceedance probability. PIT is the empirical conditional CDF at
 the validation observation. The default 95 percent predictive interval is
 given by the empirical 0.025 and 0.975 quantiles, unless explicit
 \code{lower95}/\code{upper95} values are supplied.

 The empirical CRPS from conditional draws
 \eqn{x_1,\ldots,x_B} is
 \deqn{
 \widehat{CRPS}(F,y)=
 \frac{1}{B}\sum_{b=1}^B |x_b-y|-
 \frac{1}{2B^2}\sum_{b=1}^B\sum_{c=1}^B |x_b-x_c|.
 }
 It is evaluated internally with an equivalent sorted-sample formula requiring
 \eqn{O(B\log B)} rather than \eqn{O(B^2)} work per prediction location.

 A logarithmic score is not computed automatically from a
 \code{GeoSimcond} object. Conditional draws identify the predictive
 distribution empirically but do not by themselves provide a predictive
 density evaluated exactly at the observed value. If \code{"lscore"} is
 requested from a \code{GeoSimcond} object, \code{LogScore = NA} is returned
 with a warning; no kernel-density approximation is introduced silently.

 In the Gaussian \code{pred}/\code{mse} path, the Brier score uses the
 Gaussian predictive probability of \eqn{Y>threshold}; the logarithmic score
 is the mean negative Gaussian log predictive density; PIT uses the Gaussian
 CDF; and CRPS uses the closed-form Gaussian expression.

 The 95 percent interval score is
 \deqn{
 (u-l)+\frac{2}{\alpha}(l-y)I(y<l)+
 \frac{2}{\alpha}(y-u)I(y>u), \qquad \alpha=0.05.
 }
 For \code{GeoKrig}/\code{GeoKrigloc} without explicit interval bounds,
 \eqn{l} and \eqn{u} are Gaussian intervals based on \code{pred} and
 \code{mse}. For \code{GeoSimcond}, they are empirical conditional-simulation
 quantiles. Empirical coverage is the proportion of validation observations
 lying in the corresponding intervals.
}
\value{
 A list containing only the components requested through \code{score}:
 \item{MAE}{Mean absolute error. For \code{GeoSimcond}, the empirical
 conditional mean is used as the point forecast.}
 \item{RMSPE}{Root mean squared prediction error.}
 \item{Brier}{Brier score for the event \eqn{Y>threshold}. It is Gaussian
 probability based for \code{GeoKrig}/\code{GeoKrigloc} and empirical
 conditional-simulation based for \code{GeoSimcond}.}
 \item{LogScore}{Mean negative Gaussian log predictive density for the
 \code{pred}/\code{mse} or \code{GeoKrig}/\code{GeoKrigloc} path. For a
 \code{GeoSimcond} object this component is \code{NA} if requested.}
 \item{PIT}{Gaussian PIT values for the \code{pred}/\code{mse} or kriging
 path, or empirical conditional CDF values for \code{GeoSimcond}.}
 \item{CRPS}{Mean Gaussian CRPS for the kriging/direct-prediction path, or
 mean empirical CRPS from conditional simulations for \code{GeoSimcond}.}
 \item{IS95}{Mean interval score for the 95 percent prediction intervals.}
 \item{Cvg95}{Empirical coverage of the 95 percent prediction intervals.}
}
\references{
 Gneiting, T. and Raftery, A. E. (2007).
 \emph{Strictly Proper Scoring Rules, Prediction, and Estimation}.
 Journal of the American Statistical Association, \bold{102}, 359--378.

 Heaton, M. J., Datta, A., Finley, A. O., Furrer, R., Guinness, J.,
 Guhaniyogi, R., Gerber, F., Gramacy, R. B., Hammerling, D.,
 Katzfuss, M., Lindgren, F., Nychka, D. W., Sun, F.,
 and Zammit-Mangion, A. (2019).
 \emph{A Case Study Competition Among Methods for Analyzing Large Spatial Data}.
 Journal of Agricultural, Biological, and Environmental Statistics,
 \bold{24}, 398--425.
}

\author{
Moreno Bevilacqua,
\email{moreno.bevilacqua89@gmail.com},
\url{https://sites.google.com/view/moreno-bevilacqua/home},

Víctor Morales Oñate,
\email{victor.morales@uv.cl},
\url{https://sites.google.com/site/moralesonatevictor/},

Christian Caamaño-Carrillo,
\email{chcaaman@ubiobio.cl},
\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{
library(GeoModels)

################################################################
######### Example of predictive score computation #############
################################################################

model <- "Gaussian"
set.seed(79)
N <- 1000

x <- runif(N, 0, 1)
y <- runif(N, 0, 1)
coords <- cbind(x, y)

# Set covariance parameters
corrmodel <- "GenWend"
mean <- 0
sill <- 5
nugget <- 0
scale <- 0.2
smooth <- 0
power2 <- 4

param <- list(mean = mean, sill = sill, nugget = nugget,
 scale = scale, smooth = smooth, power2 = power2)

# Simulation of the spatial Gaussian random field
data <- GeoSim(coordx = coords, corrmodel = corrmodel,
 param = param)$data

# Training and validation split
sel <- sample(1:N, N * 0.8)

coords_est <- coords[sel, ]
coords_to_pred <- coords[-sel, ]

data_est <- data[sel]
data_to_pred <- data[-sel]

# Pairwise likelihood fitting
fixed <- list(nugget = nugget, smooth = smooth, power2 = power2)

start <- list(mean = 0, scale = scale, sill = 1)

I <- Inf
lower <- list(mean = -I, scale = 0, sill = 0)
upper <- list(mean = I, scale = I, sill = I)

fit <- GeoFit(data_est, coordx = coords_est,
 corrmodel = corrmodel, model = model,
 likelihood = "Marginal", type = "Pairwise",
 neighb = 3, optimizer = "nlminb",
 lower = lower, upper = upper,
 start = start, fixed = fixed)

# Prediction at validation locations
pr <- GeoKrig(fit,loc = coords_to_pred, data = data_est, mse = TRUE)

# Predictive scores from pred and mse
Pr_scores <- GeoScores(data_to_pred, pred = pr$pred, mse = pr$mse,
 score = c("pe", "brie", "crps", "lscore",
 "pit", "intscore", "coverage"),
 threshold = 0)

Pr_scores$MAE
Pr_scores$RMSPE
Pr_scores$Brier
Pr_scores$CRPS
Pr_scores$LogScore
Pr_scores$IS95
Pr_scores$Cvg95

# Conditional-simulation objects can be scored directly. For example,
# after obtaining cs <- GeoSimcond(..., nrep = 500), use
# GeoScores(data_to_pred, cs)
# to compute point scores, empirical CRPS, empirical 95 percent interval
# score, and empirical coverage from cs$condsim.

}

\keyword{Predictive scores}

\name{plot.GeoVariogramDir}
\alias{plot.GeoVariogramDir}
\encoding{UTF-8}
\title{Plot empirical directional semi-variogram}

\description{
 Plots empirical directional semi-variograms for objects of class \code{"GeoVariogramDir"} as produced by \code{\link{GeoVariogramDir}}.
 All directions are displayed in a single plot, each with a different color and a legend indicating the direction (e.g., "0°", "45°", etc.).
}

\usage{
\method{plot}{GeoVariogramDir}(x,..., main = "Directional Empirical Semivariograms",
 pch = 20, lwd = 1, col = 1:8, ylab = "Semivariogram", xlab = "Distance")
}

\arguments{
 \item{x}{An object of class \code{"GeoVariogramDir"} as produced by \code{\link{GeoVariogramDir}}.}
 \item{main}{A main title for the plot.}
 \item{pch}{Plotting character (point type) for the points (default: 20).}
 \item{lwd}{Line width for the lines connecting points (default: 1).}
 \item{col}{A vector of colors, one for each direction (default: 1:8).}
 \item{ylab}{Label for the y-axis (default: "Semivariogram").}
 \item{xlab}{Label for the x-axis (default: "Lag").}
 \item{\dots}{Additional graphical parameters passed to \code{\link{plot}}.}
}

\details{
 This function plots all empirical directional semi-variograms in a single graph, using different colors and a legend in the top left corner that indicates the direction (e.g., "0°", "45°", etc.).
 Each direction is represented by points connected by lines.
}

\value{
 Produces a plot.
 No values are returned.
}

\seealso{
 \code{\link{GeoVariogramDir}} for directional variogram computation and examples.
}



\keyword{Variogram}

\name{GeoOutlier}
\alias{GeoOutlier}
\encoding{UTF-8}
\title{Spatial outlier detection}
\description{
Detects outliers in a univariate purely spatial data set using the selected local procedure. Space-time and bivariate outlier detection are not currently implemented.
}
\usage{
GeoOutlier(data, coordx, coordy=NULL,coordz=NULL, coordt=NULL, coordx_dyn=NULL, 
 distance="Eucl", grid=FALSE, neighb=10,alpha=0.001,
 method="Z-Median", radius=1, bivariate=FALSE,X=NULL)
}
\arguments{
 \item{data}{A numeric vector containing a single spatial realization, or a matrix representing a single spatial realization on a regular grid. Space-time data are not currently supported.}
 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix 
 Coordinates on a sphere for a fixed radius \code{radius} 
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}

 \item{coordt}{Retained for API compatibility. \code{GeoOutlier} is currently implemented only for purely spatial data and rejects space-time input.}

 \item{coordx_dyn}{Retained for API compatibility. Dynamic space-time coordinates are not currently implemented.}

 \item{distance}{String; the name of the spatial distance. The default
 is \code{Eucl}, the Euclidean distance. See \code{\link{GeoFit}} for details.}

 \item{grid}{Logical; if \code{FALSE} (the default), the data are interpreted as a spatial realization on explicit, possibly non-equispaced, spatial sites.}

 \item{neighb}{Numeric; an optional positive integer indicating the 
 order of neighborhoodused for Z-Median algorithm.}

 \item{alpha}{Numeric; a numeric value between 0 and 1 used for Z-Median algorithm.}

 \item{method}{String; The name of the algorithm for detecting spatial ouliers. Default is Z-median proposed in Chen et al. (2008)}
\item{radius}{Numeric; a value indicating the radius of the sphere when using the great 
 circle distance. Default value is 1.} 

 \item{bivariate}{Retained for API compatibility. Bivariate outlier detection is not currently implemented and \code{TRUE} is rejected.}

 \item{X}{Numeric; an optional matrix of spatial covariates with one row per observation.}
}


\value{
 Returns a matrix containing the detected spatial outliers
}

\references{
 Chen D, Lu C, Kou Y, Chen F (2008) On detecting spatial outliers. Geoinformatica 12:455–475
 
 Bevilacqua M., Caamaño C., Arellano-Valle R. B., Camilo Gomez C. (2022) A class of random fields with two-piece marginal distributions for modeling point-referenced data with spatial outliers. \emph{Test} 10.1007/s11749-021-00797-5
}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\section{Three-dimensional coordinates}{
Purely spatial irregular coordinates may be supplied as an
\eqn{N \times 3} matrix or as separate \code{coordx}, \code{coordy}, and
\code{coordz} vectors. Three-dimensional coordinates require
\code{distance = "Eucl"}. Space-time and bivariate outlier detection are
not implemented.
}

\name{GeoSimCopula}
\alias{GeoSimCopula}
\alias{print.GeoSimCopula}
\encoding{UTF-8}
\title{Simulation of Gaussian and non-Gaussian random fields using copula.}
\description{
 Simulation of Gaussian and some non-Gaussian univariate spatial and
 spatio-temporal random fields using Gaussian, skew-Gaussian, Clayton-like or
 AMH constructions. Bivariate correlation models are not supported.
 The function returns a realization of a random field for a given covariance model and covariance parameters.
 Exact simulation is available through Cholesky or SVD decomposition; for supported purely spatial correlation models, turning-bands simulation is available through \code{method="TB"}.
}
\usage{
GeoSimCopula(coordx=NULL, coordy=NULL,coordz=NULL, coordt=NULL, 
coordx_dyn=NULL, corrmodel, distance="Eucl", grid=FALSE, 
 method="cholesky", model='Gaussian', n=1, param,
 anisopars=NULL,radius=1, sparse=FALSE,
 copula="Gaussian",X=NULL,spobj=NULL,nrep=1,progress=FALSE,check.duplicates=FALSE,
 L=10000,parallel=FALSE,ncores=6)
}
\arguments{
 \item{coordx}{Optional when \code{coordx_dyn} or \code{spobj} is supplied. A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix 
 Coordinates on a sphere for a fixed radius \code{radius} 
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}

 \item{coordt}{A numeric vector giving the temporal coordinates at which the field is simulated. Optional argument; the default is \code{NULL}, in which case a spatial random field is expected. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{A list of \eqn{m} numeric (\eqn{d_t \times 2}{d x 2})-matrices
 containing dynamical (in time) spatial coordinates. Optional argument, the default is \code{NULL}
 }
 \item{corrmodel}{String; the name of a correlation model, for the
 see \code{\link{GeoCovmatrix}} for the list of implemented correlation models.}
 \item{distance}{String; the name of the spatial distance. The default
 is \code{Eucl}, the Euclidean distance. See \code{\link{GeoFit}} for details.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data
 are interpreted as spatial or spatio-temporal realisations on a set
 of non-equispaced spatial sites (irregular grid).}
 \item{method}{String; simulation engine for the latent fields. Use \code{"cholesky"} (default) or \code{"svd"} for exact simulation through \code{GeoSim}, and \code{"TB"} for turning-bands simulation through \code{GeoSimapprox}.} 
 \item{model}{String; the type of RF and therefore the densities associated to the likelihood
 objects. \code{Gaussian} is the default, see the Section
 Details.}
 \item{n}{Numeric; the number of trials for binomial random fields. The number of successes in the negative Binomial random fields. Default is \eqn{1}.} 
 \item{param}{A list of parameter values required in the
 simulation procedure of random fields, see \bold{Examples}.}
 \item{anisopars}{ A list of two elements "angle" and "ratio" i.e. the anisotropy angle and the anisotropy ratio, respectively.}
 \item{radius}{Numeric; a value indicating the radius of the sphere when using the great 
 circle distance. Default value is 1.} 
 \item{sparse}{Logical; if \code{TRUE} then cholesky decomposition is performed
 using sparse matrices algorithms (spam packake). It should be used
 with compactly supported covariance models.\code{FALSE} is the default.}
 \item{copula}{String; one of \code{"Gaussian"}, \code{"SkewGaussian"}, \code{"Clayton"}, or \code{"AMH"}.}
 \item{X}{Numeric; Matrix of space-time covariates.}
 \item{spobj}{An object of class \code{sp} or \code{spacetime}. For space-time objects, the current \code{sp2Geo()} conversion uses sequential temporal indices and does not preserve irregular original time spacing; to retain irregular temporal distances, use the explicit-coordinate interface with numeric \code{coordt}.}
 \item{nrep}{Numeric; Numbers of indipendent replicates.}
 \item{progress}{Logic; If TRUE then a progress bar is shown.}
\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
 \item{L}{Positive integer; number of turning-band lines used when \code{method="TB"}. Ignored by exact methods.}
 \item{parallel}{Logical; default \code{FALSE}. For \code{method="TB"}, set \code{TRUE} to allow \code{GeoSimapprox} to select a parallel turning-bands execution path.}
 \item{ncores}{Positive integer or \code{NULL}; default \code{6}. With \code{parallel=TRUE} and \code{method="TB"}, an explicit integer requests that many workers for the TB backend, subject to detected-core and job-count limits. Set \code{ncores=NULL} for automatic selection.}
}


\details{

Only univariate correlation models are accepted. Dynamic coordinates are
supported and the returned data preserve the list structure and row order of
\code{coordx_dyn}.

With \code{method="TB"}, \code{GeoSimCopula} keeps the copula and marginal
construction unchanged but generates each required latent field through
\code{GeoSimapprox(method="TB")}. Consequently, TB is available only for
purely spatial models supported by the turning-bands backend, requires
\code{distance="Eucl"}, and inherits the TB restrictions on correlation
families and coordinate dimension. The \code{sparse} argument is ignored by
the TB backend. Exact \code{"cholesky"} and \code{"svd"} simulations continue
to use \code{GeoSim}.

For the constructive Clayton-like simulator, \code{param$nu} must be a
positive integer and is never rounded silently. For \code{copula="SkewGaussian"},
\code{param$nu} is the bounded reflection-asymmetry parameter
\eqn{\eta\in(-1,1)} and is internally mapped to
\eqn{\gamma_\eta=\eta/\sqrt{1-\eta^2}}.  The continuous margins audited
for copula simulation and pairwise copula fitting with the Gaussian,
Clayton-like and skew--Gaussian copulas are
\code{Gaussian}, \code{StudentT}, \code{LogGaussian}, \code{Gamma},
\code{Weibull}, \code{Beta}, \code{Beta2}, \code{Kumaraswamy},
\code{Kumaraswamy2}, \code{Logistic}, and \code{SkewLaplace}.
The Gaussian copula additionally has audited simulation and pairwise fitting for
\code{Poisson}, \code{Binomial}, and \code{BinomialNeg}. Exact conditional
simulation for these discrete copula margins is not currently implemented.
For \code{model="StudentT"},
\code{param$df} is the reciprocal degrees-of-freedom parameter: the Student-t
degrees of freedom are exactly \code{1/param$df}, without integer rounding.
The same no-rounding convention is used for the Student-t degrees of freedom
of a \code{"SkewStudentT"} marginal. Because copula margins are obtained through
inverse CDFs, the integer latent-field restrictions used by direct \code{GeoSim} /
\code{GeoSimapprox} Gamma, Beta, and Student-t constructions do not apply here.
Gamma margins require a finite
positive \code{shape} and use shape \code{shape/2} and scale
\code{2/shape}, hence the multiplicative factor has unit mean. With dynamic
coordinates, \code{X} may be a row-stacked matrix or a list of matrices aligned
with \code{coordx_dyn}.  For \code{SkewStudentT}, \code{skew} is the native
\eqn{\delta\in(-1,1)} parameter and is converted to the skew-t shape
\eqn{\alpha=\delta/\sqrt{1-\delta^2}}.  The Tukey-g-and-h implementation
uses its continuous limit when the skew parameter is zero.  Kumaraswamy margins
use the inverse of the CDF employed by \code{GeoPit}.

Aliases accepted by the main modelling API, including \code{Gauss},
\code{SkewGauss}, \code{LogGauss}, and \code{TwoPieceGauss}, are canonicalized
before the marginal transform. \code{nrep} must be a positive integer. For the AMH
construction, \code{nu} is checked to be a finite scalar; no additional theoretical
parameter range is imposed by this simulator. The returned \code{param} component is
the parameter list supplied by the user.

\code{Binary} and \code{Bernoulli} are aliases of \code{Binomial} with
\code{n=1}; \code{Geom} and \code{Geometric} are aliases of
\code{BinomialNeg} with \code{n=1}.

For univariate simulation, the marginal location is
\eqn{\mu=X\beta}, with coefficients named \code{mean}, \code{mean1}, and so
on in the order of the columns of \code{X}.  If \code{X=NULL}, the model is
intercept-only.  A vector \code{param$mean} supplies a known location value at
each observation and cannot be combined with \code{X}.
}

\value{
 Returns an object of class \code{GeoSimCopula}.
 An object of class \code{GeoSimCopula} is a list containing
 at most the following components:
 
 \item{bivariate}{Always \code{FALSE}; bivariate correlation models are rejected.}
 \item{coordx}{A \eqn{d}{d}-dimensional vector of spatial coordinates;}
 \item{coordy}{A \eqn{d}{d}-dimensional vector of spatial coordinates;}
 \item{coordt}{A \eqn{t}{t}-dimensional vector of temporal coordinates;}
 \item{coordx_dyn}{A list of dynamical (in time) spatial coordinates;}
 \item{corrmodel}{The correlation model; see \code{\link{GeoCovmatrix}}.}
 \item{data}{The simulated data. Dynamic space-time output is a list with one vector per time, aligned with \code{coordx_dyn}; fixed-location space-time output preserves the matrix layout of \code{GeoSim}.}
 \item{distance}{The type of spatial distance;}
 \item{method}{The method of simulation}
 \item{model}{The type of RF, see \code{\link{GeoFit}}.}
 \item{n}{The number of trial for Binomial random fields;the number of successes in a negative Binomial random fields;} 
 \item{numcoord}{The number of spatial coordinates;}
 \item{numtime}{The number the temporal realisations of the RF;}
 \item{param}{A list of the parameters}
 \item{radius}{The radius of the sphere if coordinates are passed in lon/lat format;}
 \item{randseed}{The seed used for the random simulation;}
 \item{spacetime}{\code{TRUE} if spatio-temporal and \code{FALSE} if
 spatial RF;}
 \item{copula}{The type of copula}
}



\references{

 Bevilacqua M., Alvarado E., Caamano C. (2024) A flexible Clayton-like spatial copula with application to bounded support data. \emph{Journal of Multivariate Analysis} \bold{201}


}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}


\examples{
library(GeoModels)

################################################################
###
### Example: Simulation of a reparametrized Beta RF
### for beta regression
### with Gaussian and Clayton Copula 
### with underlying Wendland correlation.
###
###############################################################
set.seed(261)
NN=1400
x <- runif(NN);y <- runif(NN)
coords=cbind(x,y)

corrmodel="GenWend"
X=cbind(rep(1,NN),runif(NN))

NuisParam("Beta2",num_betas=2,copula="Gaussian")
CorrParam("GenWend")
#### Gaussian copula
param=list(smooth=0,power2=4, min=0,max=1,
 mean=0.1,mean1=0.1,scale=0.3,nugget=0,shape=5)

data <- GeoSimCopula(coordx=coords, corrmodel=corrmodel, model="Beta2",param=param,
 copula="Gaussian",sparse=TRUE,X=X)$data

if (requireNamespace("fields", quietly = TRUE)) fields::quilt.plot(coords,data)


#### Clayton copula
NuisParam("Beta2",num_betas=2,copula="Clayton")
CorrParam("GenWend")
param=list(smooth=0,power2=4, min=0,max=1,
 mean=0.2,mean1=0.1,scale=0.3,nugget=0,shape=6,nu=4)
data1 <- GeoSimCopula(coordx=coords, corrmodel=corrmodel, model="Beta2",param=param,
 copula="Clayton",sparse=TRUE,X=X)$data

hist(data1,freq=FALSE)
if (requireNamespace("fields", quietly = TRUE)) fields::quilt.plot(coords,data1)

}
\keyword{Simulation}

\name{CompLik}
\alias{CompLik}
\encoding{UTF-8}
\title{Optimizes the Composite Log-likelihood}
\description{
 Subroutine called by \code{GeoFit}.
 The procedure estimates the model parameters by maximization of the
 composite log-likelihood.
}
\usage{
CompLik(copula, bivariate, coordx, coordy, coordz, coordt, 
 coordx_dyn, corrmodel, data, distance, flagcorr, 
 flagnuis, fixed, grid, likelihood, lower, 
 model, n, namescorr, namesnuis, namesparam,
 numparam, numparamcorr, optimizer, 
 onlyvar, param,
 spacetime, type, upper, varest,
 weigthed, ns, X, sensitivity, MM, aniso,score)
}
\arguments{
 \item{copula}{String; the type of copula. It can be "Clayton" or "Gaussian".}
 \item{bivariate}{Logical; if \code{TRUE} then the data come from a bivariate random field,
 otherwise from a univariate random field.}
 \item{coordx}{A numeric \eqn{d \times 2} or \eqn{d \times 3} matrix.
 Coordinates on a sphere for a fixed radius \code{radius} 
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving one dimension of
 spatial coordinates; optional, default is \code{NULL}.}
 \item{coordz}{A numeric vector giving one dimension of
 spatial coordinates; optional, default is \code{NULL}.}
 \item{coordt}{A numeric vector giving one dimension of temporal coordinates; optional, default is \code{NULL} (in which case a spatial random field is assumed). Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{A list of \eqn{m} numeric \eqn{d_t \times 2} matrices
 containing dynamic (in time) spatial coordinates; optional, default is \code{NULL}.}
 \item{corrmodel}{Numeric; the ID of the correlation model.}
 \item{data}{A numeric vector, or a \eqn{n \times d} matrix, or a \eqn{d \times d \times n} array of observations.}
 \item{distance}{String; the name of the spatial distance.
 Default is \code{"Eucl"} (Euclidean distance). See \code{\link{GeoFit}} for details.}
 \item{flagcorr}{Numeric vector of binary values indicating which
 parameters of the correlation function will be estimated.}
 \item{flagnuis}{Numeric vector of binary values indicating which
 nuisance parameters will be estimated.}
 \item{fixed}{Numeric vector of parameters considered as known values.}
 \item{grid}{Logical; if \code{FALSE} (default), data are interpreted as vector or \eqn{n \times d} matrix;
 if \code{TRUE}, then a \eqn{d \times d \times n} array is considered.}
 \item{likelihood}{String; configuration of the composite likelihood (see \code{\link{GeoFit}}).}
 \item{lower}{Named list; optional lower bounds for parameters when using
 optimizers \code{L-BFGS-B}, \code{nlminb}, or \code{optimize}.
 Names must match those in the \code{start} list.}
 \item{model}{Numeric; ID of the density associated with the likelihood objects.}
 \item{n}{Numeric; number of trials in binomial random fields.}
 \item{namescorr}{Character vector; names of the correlation parameters.}
 \item{namesnuis}{Character vector; names of the nuisance parameters.}
 \item{namesparam}{Character vector; names of the parameters to be maximized.}
 \item{numparam}{Numeric; number of parameters to be maximized.}
 \item{numparamcorr}{Numeric; number of correlation parameters.}
 \item{optimizer}{String; optimization algorithm (see \code{\link{optim}}).
 Default is \code{"Nelder-Mead"}. Other options: \code{"nlm"}, \code{"BFGS"}, \code{"L-BFGS-B"}, \code{"nlminb"}.
 For \code{"L-BFGS-B"} and \code{"nlminb"} bounds can be provided.
 For 1D optimization, \code{optimize} is used.}
 \item{onlyvar}{Logical; if \code{TRUE} (and \code{varest} is \code{TRUE}),
 only the variance-covariance matrix is computed without optimizing.
 Default is \code{FALSE}.}
 \item{param}{Numeric vector of parameter values.}
 \item{spacetime}{Logical; if \code{TRUE}, the random field is spatio-temporal,
 otherwise spatial.}
 \item{type}{String; type of likelihood object.
 Default is \code{"Pairwise"} (marginal composite likelihood formed by pairwise marginal likelihoods).}
 \item{upper}{Named list; optional upper bounds for parameters when using
 optimizers \code{L-BFGS-B}, \code{nlminb}, or \code{optimize}.
 Names must match those in the \code{start} list.}
 \item{varest}{Logical; if \code{TRUE}, variance estimates and standard errors are returned.
 Default is \code{FALSE}.}
 \item{weigthed}{Logical; if \code{TRUE}, decreasing weights from a compactly supported
 correlation function with compact support \code{maxdist} (\code{maxtime}) are used.}
 \item{ns}{Numeric; number of (dynamic) temporal instants.}
 \item{X}{Numeric; matrix of space-time covariates in the linear mean specification.}
 \item{sensitivity}{Logical; if \code{TRUE}, the sensitivity matrix is computed.}
 \item{MM}{Numeric; a non-constant fixed mean.}
 \item{aniso}{Logical; whether anisotropy should be considered.}
 \item{score}{Logical; should score function be computed?}
}
\value{
 Returns a list from an \code{optim} call.
}
\details{
 Subroutine called by \code{GeoFit}.
 The procedure estimates model parameters by maximization of the composite log-likelihood.
}
\seealso{
\code{\link{GeoFit}} for fitting covariance models with composite likelihood,
  \code{\link{CompLik2}} and \code{\link{CompIndLik2}} for related composite-likelihood routines,
  \code{\link{GeoVarest}} for variance estimation.
}
\author{
 Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com}, \url{https://sites.google.com/view/moreno-bevilacqua/home}, \cr
 Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/}, \cr
 Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl}, \url{https://www.researchgate.net/profile/Christian-Caamano}
}
\keyword{Composite}

\name{plot.GeoCorrFct}
\alias{plot.GeoCorrFct}
\encoding{UTF-8}
\title{Plot Spatial and Spatio-temporal correlation or covariance of (non) Gaussian random fields}

\description{
 Plot Spatial and Spatio-temporal correlation or covariance of (non) Gaussian random fields
 for a given set of spatial or spatiotemporal distances
 \code{\link{GeoCorrFct}}.
}

\usage{
\method{plot}{GeoCorrFct}(x,type="p", \dots)
}

\arguments{
 \item{x}{an object of the class \code{"GeoCorrFct"} }
 \item{type}{The type of graphic. The possible options are "p" and "l". If "p" then a point type graphic is displayed. 
 Otherwise a lines type graphic displayed. }
 \item{\dots}{Other graphical options arguments.
 \code{\link{plot}} }
}

\details{
Plot Spatial and Spatio-temporal correlation or covariance of (non) Gaussian random fields
}

\value{
 Produces a plot.
 No values are returned.
}

\seealso{
 \code{\link{GeoCorrFct}} for examples.
}

\keyword{Correlation}

\name{GeoKrig}
\alias{GeoKrig}
\encoding{UTF-8}
\title{Spatial (bivariate) and spatio temporal optimal linear prediction 
 for Gaussian and non-Gaussian random fields.}
\description{
  For a given set of spatial location sites (and temporal instants),
the function computes optimal linear prediction and associated mean square error 
for the Gaussian and non-Gaussian case.
}
\usage{
GeoKrig(estobj=NULL,data, coordx=NULL, coordy=NULL, coordz=NULL, coordt=NULL, 
coordx_dyn=NULL, corrmodel,distance="Eucl",
 grid=FALSE, loc, 
 method="cholesky", model="Gaussian", n=1,nloc=NULL,mse=FALSE, 
 param, anisopars=NULL,radius=1, sparse=FALSE,
 time=NULL, type_krig="Simple",weigthed=TRUE,which=1,
 copula=NULL, X=NULL,Xloc=NULL,Mloc=NULL,spobj=NULL,spdata=NULL,varcov=NULL,
 progress=FALSE,check.duplicates=FALSE)
}
\arguments{
 \item{estobj}{An object of class Geofit that includes information about data, model and estimates.}
 \item{data}{A \eqn{d}{d}-dimensional vector (a single spatial realisation) or a (\eqn{d \times d}{d x d})-matrix (a single spatial realisation on regular grid) or a
 (\eqn{t \times d}{t x d})-matrix (a single spatio-temporal realisation) or an (\eqn{d \times d \times t \times n }{d x d x t})-array
 (a single spatio-temporal realisation on regular grid) giving the data used for prediction.}
\item{coordx}{A numeric \eqn{d \times 2}{d x 2} or \eqn{d \times 3}{d x 3}
coordinate matrix.  It may be omitted when \code{coordx_dyn} or \code{spobj}
is supplied.  Coordinates on a sphere for a fixed radius \code{radius} are
passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordt}{A numeric vector giving the temporal coordinates of the observations. The default is \code{NULL}, in which case a spatial random field is expected. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{For dynamic observation locations, a list with one two- or three-column coordinate matrix per element of \code{coordt}.  The rows of \code{coordx_dyn[[t]]} correspond to \code{data[[t]]} and to the same temporal block of \code{X}.  Dynamic observation locations do not alter the ordering of prediction tasks.  See \code{\link{GeoModels-spacetime-ordering}}.}
 \item{corrmodel}{String; the name of a correlation model, for the
 see \code{\link{GeoCovmatrix}} for the list of implemented correlation models.}
 \item{distance}{String; the name of the spatial distance. The default
 is \code{Eucl}, the Euclidean distance. See \code{\link{GeoFit}} for details.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data used for prediction
 are interpreted as spatial or spatio-temporal realisations on a set
 of non-equispaced spatial sites (irregular grid).}
 \item{loc}{A numeric (\eqn{n \times 2}{n x 2})-matrix (where
 \code{n} is the number of spatial sites) giving 2-dimensions of spatial coordinates to be predicted.}
 \item{method}{String; matrix decomposition used to solve the kriging
system.  The choices are \code{cholesky} (default) and \code{svd}.
Only \code{cholesky} is available when \code{sparse=TRUE}.}
 \item{n}{Positive integer size parameter. For Binomial it may be scalar or contain one value per observation. For direct Negative Binomial it is the single common number \eqn{r} of successes; Geometric corresponds to \eqn{r=1}.}
 \item{nloc}{Positive integer size at prediction tasks. For Binomial, if observation-side \code{n} is location-specific then \code{nloc} is required and is never inferred by averaging \code{n}; otherwise a scalar \code{n} is reused when \code{nloc=NULL}. For direct Negative Binomial, \code{nloc}, if supplied, must equal the common \eqn{r}; for Geometric it must equal 1.}
 \item{mse}{Logical; if \code{TRUE}, compute the MSE of the kriging predictor.
 The default is \code{FALSE}; in that case the returned \code{mse} component is
 \code{NULL}, including for fixed- and dynamic-support space-time prediction.} 
 \item{model}{String; the type of RF and therefore the densities associated to the likelihood
 objects. \code{Gaussian} is the default, see the Section
 Details.} 
 \item{param}{A list of parameter values required for the correlation model. See Details for the accepted options.}
 \item{anisopars}{ A list of two elements: "angle" and "ratio" i.e. the anisotropy angle and the anisotropy ratio, respectively.}
 \item{radius}{Numeric: the radius of the sphere if coordinates are passed in lon/lat format;Default value is 1.}
 \item{sparse}{Logical; if \code{TRUE} kriging is computed with sparse matrices algorithms 
 using spam package. Default is FALSE. It should be used with compactly supported covariances.} 
 \item{time}{A numeric vector giving the temporal instants to be predicted. Values need not be equally spaced and are interpreted on the same numeric time scale as \code{coordt}. The default is \code{NULL}, in which case only spatial prediction is performed.}
 \item{type_krig}{String; \code{Simple} (default) treats the supplied or
fitted mean coefficients as known.  \code{Universal} uses the same plug-in
coefficients and, when \code{mse=TRUE}, adds their estimation uncertainty
from \code{varcov} to the prediction MSE.}
 \item{weigthed}{Logical legacy argument retained for backward
compatibility.  It has no effect in standard kriging.}
 \item{which}{Numeric; In the case of bivariate cokriging it indicates which variable to predict.
 	 It can be 1 or 2}
 \item{copula}{String; optional copula specification. Linear prediction is implemented for \code{"Gaussian"}, the constructive Clayton-like copula \code{"Clayton"}, and \code{"SkewGaussian"}.}
 \item{X}{Numeric design matrix at the observations.  For fixed-location space-time data, rows follow \code{c(t(data))}, i.e. time then site.  For dynamic observations, use a stacked matrix ordered by temporal blocks, or a list aligned with \code{coordx_dyn}.}
 \item{Xloc}{Numeric design matrix at prediction tasks.  For space-time prediction its rows are location-major: all requested times for \code{loc[1, ]}, then all requested times for \code{loc[2, ]}, and so on.}
 \item{Mloc}{Numeric vector giving the known marginal location predictor at prediction tasks, in the same location-major order as \code{Xloc}. For Gaussian additive margins this is the mean itself; for models with link-scale parametrizations (for example Gamma/Weibull/LogGaussian or Beta2) it is on the same scale as \code{param$mean}. Use either \code{Mloc} or \code{Xloc}, not both.}
 \item{spobj}{An object of class \code{sp} or \code{spacetime}. For space-time objects, the current \code{sp2Geo()} conversion uses sequential temporal indices and does not preserve irregular original time spacing; to retain irregular temporal distances, use the explicit-coordinate interface with numeric \code{coordt}.}
 \item{spdata}{Character:The name of data in the sp or spacetime object}
 \item{varcov}{Covariance matrix of the estimated parameters.  It is
required for \code{type_krig="Universal"} when \code{mse=TRUE}.  For
composite-likelihood fits this is the sandwich covariance, i.e. the inverse
Godambe information, usually taken automatically from a \code{GeoFit}
object.  Mean coefficients absent from \code{varcov} are treated as fixed.}
 \item{progress}{Logical; if \code{TRUE}, show progress while the optimized
blocked prediction path processes multiple prediction blocks.  The default is
\code{FALSE}.  No progress bar is shown when the calculation uses a single
block or a legacy one-shot prediction path.}
\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
}

\details{

For univariate kriging, the mean at the observations is either
\eqn{X\beta}, where the entries of \eqn{\beta} are represented by
\code{mean}, \code{mean1}, and so on, or a site-specific vector supplied as
\code{param$mean}.  With \code{X=NULL}, the
model is intercept-only and uses the scalar coefficient \code{mean}.

The prediction mean is specified independently: use \code{Xloc} to obtain
\eqn{X_{loc}\beta}, or use \code{Mloc} to supply the prediction mean directly.
\code{Mloc} and \code{Xloc} are mutually exclusive.  If the observation mean
is site-specific, \code{Mloc} is required.  When \code{X} is used at the
observations, either \code{Xloc} or \code{Mloc} must be supplied; the latter
allows a known prediction mean that is not generated by the fitted design.

For \code{type_krig="Universal"}, prediction uses the same fitted mean
coefficients as \code{type_krig="Simple"}.  Thus coefficients estimated by
maximum likelihood, REML, or composite likelihood are not recomputed inside
\code{GeoKrig}.  When \code{mse=TRUE}, the MSE is augmented by
\eqn{(X_{loc}-\lambda^T X)\,Var(\widehat\beta)\,
(X_{loc}-\lambda^T X)^T}, using the mean-parameter block of \code{varcov}.
For composite likelihood, \code{varcov} should be the sandwich covariance
(the inverse Godambe information).  Coefficients absent from that matrix are
regarded as fixed and contribute zero uncertainty.  External mean vectors
supplied through \code{param$mean} or \code{Mloc} are already known, so a
requested universal prediction is treated as simple kriging with a warning.


For prediction, names beginning with \code{Gaussian_misp_} are interpreted as
estimation specifications rather than new data-generating models.  They are
automatically mapped to the corresponding response model before covariance and
prediction calculations (for example \code{Gaussian_misp_Poisson} to
\code{Poisson}).  Without a copula, prediction is available only for models with
a validated model-specific covariance/predictor implementation.  In particular,
plain \code{Logistic}, \code{Beta2}, \code{Kumaraswamy}, and
\code{Kumaraswamy2} are rejected rather than falling through an incomplete branch;
these margins are available in the validated continuous-copula prediction path.

For discrete models with nonlinear marginal means, the Universal-kriging MSE uses
the derivative of the marginal mean with respect to the regression predictor when
propagating the mean-parameter block of \code{varcov}.  The MSE is evaluated from
its diagonal directly and does not construct a prediction-by-prediction matrix.

For \code{copula="Gaussian"}, \code{copula="Clayton"}, and \code{copula="SkewGaussian"}, prediction is performed on the observed marginal scale using marginal means, marginal variances, and the copula-induced cross-covariances. The supported continuous margins are \code{"Gaussian"}, \code{"StudentT"}, \code{"LogGaussian"}, \code{"Gamma"}, \code{"Weibull"}, \code{"Beta"}, \code{"Beta2"}, \code{"Kumaraswamy"}, \code{"Kumaraswamy2"}, \code{"Logistic"}, \code{"SkewLaplace"}, \code{"Tukeyh"}, \code{"Tukeyh2"}, and \code{"SinhAsinh"}. The count margins \code{"Poisson"}, \code{"Binomial"}, and \code{"BinomialNeg"} are also supported by global \code{GeoKrig} for Gaussian, Clayton-like, and skew-Gaussian copulas; the Binomial copula path currently requires a common trial count \code{n}. For a Gaussian copula, the last three margins use their exact transformed-Gaussian covariance rather than a truncated Hermite approximation. For location-dependent margins such as \code{"Beta2"}, \code{"Kumaraswamy2"}, and the positive-scale models, the covariance between two sites uses both supplied location predictors. Thus \code{X}/\code{Xloc} or external means are propagated into the covariance consistently. The resulting predictor is the optimal linear predictor on the observed scale; except when the joint field is Gaussian it is generally different from the full conditional mean. For \code{type_krig="Universal"}, the MSE correction uses the derivative of the marginal mean with respect to the regression predictor (delta-method correction for nonlinear marginal means). For the Clayton-like and skew-Gaussian copulas, the expensive covariance ingredients are precomputed and cached. Clayton-like prediction uses deterministic copula quadrature tables, while skew-Gaussian prediction uses a cached bivariate Hermite expansion of the latent Gaussian representation; neither path performs adaptive two-dimensional integration separately for each covariance entry. For \code{copula="SkewGaussian"}, \code{param$nu} is the paper's \eqn{\eta\in(-1,1)}; when \code{nu=0} the covariance calculation is routed exactly to the Gaussian-copula implementation.

For the Clayton-like copula, direct adaptive two-dimensional integration for every covariance entry would be prohibitively expensive. GeoModels therefore uses a deterministic cached covariance engine: the Clayton-like copula density is evaluated once on a Gauss--Legendre quadrature grid over a compact grid of the underlying squared correlations, the discrete joint distributions are numerically balanced to preserve uniform margins, and covariance values are obtained by cached matrix products and interpolation. The expensive copula kernel is reused across prediction locations and across calls in the same R session. Positive-scale margins exploit their exact multiplicative scaling, while \code{"Beta2"} and \code{"Kumaraswamy2"} use an additional compact cached grid on the location-predictor scale, built only over the range required by the current kriging problem. This approximation is designed for kriging covariance construction and avoids one numerical double integral per pair.

For global prediction with a supported continuous copula, \code{GeoKrig()} constructs observation--prediction cross-covariances in bounded blocks of prediction locations. The observation covariance matrix is factorized once and the same Cholesky (or SVD, when requested) factorization is reused for every block. This changes only the evaluation order: predictions, MSE values, and kriging weights use the same covariance equations as the one-shot calculation. The historical full weights matrix is retained in the returned object, so blocking reduces the peak memory associated with the cross-covariance matrix without changing the returned results. The ordinary Gaussian model without a copula uses the same blocked path. When \code{progress=TRUE} and more than one block is required, progress is updated once per completed prediction block; the numerical calculation and block sizes are unchanged.

 Best linear unbiased predictor and associated mean square error is computed 
 for Gaussian and some non-Gaussian cases.
 Specifically, for a spatial or spatio-temporal or spatial bivariate dataset, given a set of spatial locations and
 temporal istants and a correlation model
\code{corrmodel} with some fixed parameters and given the type of RF (\code{model}) the function computes
 simple or universal kriging, for the specified spatial locations
 \code{loc} and temporal instants \code{time}, providing also the respective mean square error.
 For the choice of the spatial or spatio temporal correlation model see details in \code{\link{GeoCovmatrix}} function.
 The list \code{param} specifies mean and covariance parameters, see \code{\link{CorrParam}} and \code{\link{GeoCovmatrix}} for details. The \code{type_krig} parameter indicates the type of kriging. In the
case of simple kriging, the known mean can be specified by the parameter
 \code{mean} in the list \code{param} (See examples).
}

\section{Spatio-temporal ordering}{
Observed fixed-location data use time-major order: a \eqn{T \times N}{T x N}
matrix is vectorized as \code{c(t(data))}.  Dynamic observations are supplied
as aligned lists \code{data[[t]]} and \code{coordx_dyn[[t]]}, concatenated by
time.  The rows of \code{X} and any known observation mean use the same
observation order.

Prediction tasks use the different, location-major order
\preformatted{
loc[1, ] at time[1], ..., loc[1, ] at time[Tloc],
loc[2, ] at time[1], ..., loc[2, ] at time[Tloc], ...
}
Rows of \code{Xloc} and elements of \code{Mloc} must follow this order.  The
returned \code{pred} and \code{mse} objects are \eqn{Tloc \times Nloc}{Tloc x Nloc}
matrices with prediction times in rows and prediction locations in columns.
See \code{\link{GeoModels-spacetime-ordering}}.
}

\value{
 Returns an object of class \code{Kg}.
 An object of class \code{Kg} is a list containing
 at most the following components:
 \item{bivariate}{\code{TRUE} if spatial bivariate cokriging is performed, otherwise \code{FALSE};}
 \item{coordx}{A \eqn{d}{d}-dimensional vector of spatial coordinates used for prediction;}
 \item{coordy}{A \eqn{d}{d}-dimensional vector of spatial coordinates used for prediction;}
 \item{coordz}{A \eqn{d}{d}-dimensional vector of spatial coordinates used for prediction;}
 \item{coordt}{A \eqn{t}{t}-dimensional vector of temporal coordinates used for prediction;}
 \item{corrmodel}{String: the correlation model;}
 \item{covmatrix}{The covariance matrix. }
 \item{data}{The vector or matrix or array of data used for prediction}
 \item{distance}{String: the type of spatial distance;}
 \item{grid}{\code{TRUE} if the spatial data used for prediction are observed in a regular grid, otherwise \code{FALSE};}
 \item{loc}{A (\eqn{n \times 2}{n x 2})-matrix of spatial locations to be predicted.}
 \item{n}{The Binomial number of trials, or the common Negative-Binomial number \eqn{r} of successes.} 
 \item{nozero}{In the case of sparse simple kriging the percentage of non zero values in the covariance matrix.
 Otherwise is NULL.}
 \item{numcoord}{Numeric:he number \eqn{d}{d} of spatial coordinates used for prediction;}
 \item{numloc}{Numeric: the number \eqn{n}{n} of spatial coordinates to be predicted;}
 \item{numtime}{Numeric: the number \eqn{d}{d} of the temporal instants used for prediction;}
 \item{numt}{Numeric: the number \eqn{m}{m} of the temporal instants to be predicted;}
 \item{model}{The response model used for prediction after canonicalizing any
 misspecified-Gaussian fitting name.}
 \item{fit_model}{The model name supplied by the caller or stored in the input
 \code{GeoFit} object before prediction canonicalization.}
 \item{param}{The original parameter list supplied to prediction, before any
 internal standardization used to construct covariance matrices;}
 \item{pred}{For spatio-temporal prediction, a \eqn{m \times n}{m x n} matrix with rows corresponding to \code{time} and columns corresponding to rows of \code{loc}; for spatial prediction, a numeric vector.}
 \item{radius}{Numeric: the radius of the sphere if coordinates are pssed in lon/lat format;}
 \item{spacetime}{\code{TRUE} if spatio-temporal kriging and \code{FALSE} if
 spatial kriging;}
 \item{time}{A \eqn{m}{m}-dimensional vector of temporal coordinates to be predicted;}
 \item{type}{String: the type of kriging (Standard).}
 \item{type_krig}{String: the type of kriging (simple or universal)}
 \item{mse}{When \code{mse=TRUE}, the prediction MSE (a time-by-location matrix
 for space-time prediction and a vector for spatial prediction); otherwise
 \code{NULL}.}

}


\references{
 Gaetan, C. and Guyon, X. (2010) \emph{Spatial Statistics and Modelling}.
 Springer-Verlag, New York.
}

\seealso{
\code{\link{GeoCovmatrix}} for covariance matrix construction,
  \code{\link{GeoKrigloc}} and \code{\link{GeoKriglocWeights}} for local kriging,
  \code{\link{GeoFit}} for parameter estimation.
}


\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{

library(GeoModels)
################################################################
########### Examples of spatial kriging ############
################################################################

################################################################
###
### Example 1. Spatial kriging of a
### Gaussian random fields with Gen wendland correlation.
###
################################################################

model="Gaussian"
set.seed(79)
x = runif(300, 0, 1)
y = runif(300, 0, 1)
coords=cbind(x,y)
# Set the exponential cov parameters:
corrmodel = "GenWend"
mean=0; sill=5; nugget=0
scale=0.2;smooth=0;power2=4

param=list(mean=mean,sill=sill,nugget=nugget,scale=scale,smooth=smooth,power2=power2)

# Simulation of the spatial Gaussian random field:
data = GeoSim(coordx=coords, corrmodel=corrmodel,model=model,
 param=param)$data

## estimation with pairwise likelihood
fixed=list(nugget=nugget,smooth=0,power2=power2)
start=list(mean=0,scale=scale,sill=1)
I=Inf
lower=list(mean=-I,scale=0,sill=0)
upper=list(mean= I,scale=I,sill=I)
# Maximum pairwise likelihood fitting :
fit = GeoFit(data, coordx=coords, corrmodel=corrmodel,model=model,
 likelihood='Marginal', type='Pairwise',neighb=3,
 optimizer="nlminb", lower=lower,upper=upper,
 start=start,fixed=fixed)

# locations to predict
xx=seq(0,1,0.03)
loc_to_pred=as.matrix(expand.grid(xx,xx))

## first option
#param=append(fit$param,fit$fixed)
#pr=GeoKrig(loc=loc_to_pred,coordx=coords,corrmodel=corrmodel,
# model=model,param=param,data=data,mse=TRUE)

## second option using object GeoFit
pr=GeoKrig(fit,loc=loc_to_pred,mse=TRUE)


colour = rainbow(100)

opar=par(no.readonly = TRUE)
par(mfrow=c(1,3))
if (requireNamespace("fields", quietly = TRUE)) {
 fields::quilt.plot(coords, data, col = colour)
}
# simple kriging map prediction
if (requireNamespace("fields", quietly = TRUE)) {
 fields::image.plot(
  xx, xx, matrix(pr$pred, ncol = length(xx)), col = colour,
  xlab = "", ylab = "", main = " Kriging "
 )
}

# simple kriging MSE map prediction variance
if (requireNamespace("fields", quietly = TRUE)) {
 fields::image.plot(
  xx, xx, matrix(pr$mse, ncol = length(xx)), col = colour,
  xlab = "", ylab = "", main = "Std error"
 )
}
par(opar)

################################################################
###
### Example 2. Spatial kriging of a Skew
### Gaussian random fields with Matern correlation.
###
################################################################
model="SkewGaussian"
set.seed(79)
x = runif(300, 0, 1)
y = runif(300, 0, 1)
coords=cbind(x,y)
# Set the exponential cov parameters:
corrmodel = "Matern"
mean=0
sill=2
nugget=0
scale=0.1
smooth=0.5
skew=3
param=list(mean=mean,sill=sill,nugget=nugget,scale=scale,smooth=smooth,skew=skew)

# Simulation of the spatial skew Gaussian random field:
data = GeoSim(coordx=coords, corrmodel=corrmodel,model=model,
 param=param)$data

fixed=list(nugget=nugget,smooth=smooth)
start=list(mean=0,scale=scale,sill=1,skew=skew)
I=Inf
lower=list(mean=-I,scale=0,sill=0,skew=-I)
upper=list(mean= I,scale=I,sill=I,skew=I)
# Maximum pairwise likelihood fitting :
fit = GeoFit2(data, coordx=coords, corrmodel=corrmodel,model=model,
 likelihood='Marginal', type='Pairwise',neighb=3,
 optimizer="nlminb", lower=lower,upper=upper,
 start=start,fixed=fixed)

# locations to predict
xx=seq(0,1,0.03)
loc_to_pred=as.matrix(expand.grid(xx,xx))
## optimal linear kriging
pr=GeoKrig(fit,loc=loc_to_pred,mse=TRUE)

colour = rainbow(100)

opar=par(no.readonly = TRUE)
par(mfrow=c(1,3))
if (requireNamespace("fields", quietly = TRUE)) {
 fields::quilt.plot(coords, data, col = colour)
}
# simple kriging map prediction
if (requireNamespace("fields", quietly = TRUE)) {
 fields::image.plot(
  xx, xx, matrix(pr$pred, ncol = length(xx)), col = colour,
  xlab = "", ylab = "", main = " Kriging "
 )
}

# simple kriging MSE map prediction variance
if (requireNamespace("fields", quietly = TRUE)) {
 fields::image.plot(
  xx, xx, matrix(pr$mse, ncol = length(xx)), col = colour,
  xlab = "", ylab = "", main = "Std error"
 )
}
par(opar)

################################################################
###
### Example 3. Spatial kriging of a 
### Gamma random field with mean spatial regression
###
###############################################################
set.seed(312)
model="Gamma"
corrmodel = "GenWend" 
# Define the spatial-coordinates of the points:
NN=300
coords=cbind(runif(NN),runif(NN))
## matrix covariates
a0=rep(1,NN)
a1=runif(NN,0,1)
X=cbind(a0,a1)
##Set model parameters
shape=2
## regression parameters
mean = 1;mean1= -0.2
# correlation parameters
nugget = 0;power2=4
scale = 0.3;smooth=0 

## simulation
param=list(shape=shape,nugget=nugget,mean=mean,mean1=mean1, 
 scale=scale,power2=power2,smooth=smooth)
data = GeoSim(coordx=coords,corrmodel=corrmodel, param=param,
 model=model,X=X)$data

#####starting and fixed parameters
fixed=list(nugget=nugget,power2=power2,smooth=smooth)
start=list(mean=mean,mean1=mean1, scale=scale,shape=shape)

## estimation with pairwise likelihood
fit2 = GeoFit(data=data,coordx=coords,corrmodel=corrmodel,X=X,
 neighb=3,likelihood="Conditional",type="Pairwise",
 start=start,fixed=fixed, model = model)

# locations to predict with associated covariates
xx=seq(0,1,0.03)
loc_to_pred=as.matrix(expand.grid(xx,xx))
NP=nrow(loc_to_pred)
a0=rep(1,NP)
a1=runif(NP,0,1)
Xloc=cbind(a0,a1)

#optimal linear kriging 
pr=GeoKrig(fit2,loc=loc_to_pred,Xloc=Xloc,sparse=TRUE,mse=TRUE)

## map 
opar=par(no.readonly = TRUE)
par(mfrow=c(1,3))
if (requireNamespace("fields", quietly = TRUE)) {
 fields::quilt.plot(coords, data, main = "Data")
}
map=matrix(pr$pred,ncol=length(xx))
mapmse=matrix(pr$mse,ncol=length(xx))
if (requireNamespace("fields", quietly = TRUE)) {
 fields::image.plot(xx, xx, map, xlab = "", ylab = "", main = "Kriging ")
}

if (requireNamespace("fields", quietly = TRUE)) {
 fields::image.plot(xx, xx, mapmse, xlab = "", ylab = "", main = "MSE")
}
par(opar)


################################################################
########### Examples of spatio temporal kriging ############
################################################################

################################################################
###
### Example 4. Spatio temporal simple kriging of n locations
### sites and m temporal instants for a Gaussian random fields
### with estimated double Wendland correlation.
###
###############################################################
model="Gaussian"
# Define the spatial-coordinates of the points:
x = runif(300, 0, 1)
y = runif(300, 0, 1)
coords=cbind(x,y)
times=1:4

# Define model correlation modek and associated parameters
corrmodel="Wend0_Wend0"
param=list(nugget=0,mean=0,power2_s=4,power2_t=4,
 scale_s=0.2,scale_t=2,sill=1)

# Simulation of the space time Gaussian random field:
set.seed(31)
data=GeoSim(coordx=coords,coordt=times,corrmodel=corrmodel,sparse=TRUE,
 param=param)$data

# Maximum pairwise likelihood fitting of the space time random field:
start = list(scale_s=0.15,scale_t=2,sill=1,mean=0)
fixed = list(nugget=0,power2_s=4,power2_t=4)

fit = GeoFit(data, coordx=coords, coordt=times, model=model, corrmodel=corrmodel, 
 likelihood='Conditional', type='Pairwise',start=start,fixed=fixed,
 neighb=3,maxtime=1)

# locations to predict
xx=seq(0,1,0.04)
loc_to_pred=as.matrix(expand.grid(xx,xx))
# Define the times to predict
times_to_pred=2

pr=GeoKrig(fit,loc=loc_to_pred,time=times_to_pred,sparse=TRUE,mse=TRUE)

opar=par(no.readonly = TRUE)
par(mfrow=c(1,3))
zlim=c(-2.5,2.5)
colour = rainbow(100)


if (requireNamespace("fields", quietly = TRUE)) {
 fields::quilt.plot(
  coords, data[2, ], col = colour, main = " data at Time 2"
 )
}
if (requireNamespace("fields", quietly = TRUE)) {
 fields::image.plot(
  xx, xx, matrix(pr$pred, ncol = length(xx)), col = colour,
  main = " Kriging at Time 2", ylab = ""
 )
}
if (requireNamespace("fields", quietly = TRUE)) {
 fields::image.plot(
  xx, xx, matrix(pr$mse, ncol = length(xx)), col = colour,
  main = "Std err Time at time 2", ylab = ""
 )
}


par(opar)


################################################################
###
### Example r. Spatial bivariate simple cokriging of n locations
### sites for a bivariate Gaussian random fields
### with estimated Matern correlation.
###
###############################################################
#set.seed(6)
#NN=1500 # number of spatial locations 
#x = runif(NN, 0, 1); 
#y = runif(NN, 0, 1) 
#coords=cbind(x,y) 

## setting parameters
#mean_1 = 2; mean_2= -1
#nugget_1 =0;nugget_2=0
#sill_1 =0.5; sill_2 =1; 

### correlation parameters
#CorrParam("Bi_Matern")
#scale_1=0.2/3; scale_2=0.15/3; scale_12=0.5*(scale_2+scale_1) 
#smooth_1=smooth_2=smooth_12=0.5
#pcol = -0.4
#param= list(nugget_1=nugget_1,nugget_2=nugget_2,
# sill_1=sill_1,sill_2=sill_2,
# mean_1=mean_1,mean_2=mean_2,
# smooth_1=smooth_1, smooth_2=smooth_2,smooth_12=smooth_12,
# scale_1=scale_1, scale_2=scale_2,scale_12=scale_12,
# pcol=pcol)

## simulation
#data = GeoSim(coordx=coords, corrmodel="Bi_Matern",model=model,param=param)$data

#fixed=list(mean_1=mean_1,mean_2=mean_2, nugget_1=nugget_1,nugget_2=nugget_2, 
# smooth_1=smooth_1, smooth_2=smooth_2,smooth_12=smooth_12)

#start=list( sill_1=sill_1,sill_2=sill_2,
# scale_1=scale_1,scale_2=scale_2,scale_12=scale_12, pcol=pcol)

## estimation with maximum likelihood 
#fit = GeoFit(data=data,coordx=coords, corrmodel="Bi_Matern",
 #likelihood="Marginal",type="Pairwise",optimizer="BFGS",neighb=5,
 #start=start,fixed=fixed)

###### co-kriging for the fist component ##############
#xx=seq(0,1,0.022)
#loc_to_pred=as.matrix(expand.grid(xx,xx))
#pr1 = GeoKrig(fit,which=1,mse=TRUE,loc=loc_to_pred)

#opar=par(no.readonly = TRUE)
#par(mfrow=c(1,2))
#zlim=c(-2.5,2.5)
#colour = rainbow(100)
#fields::quilt.plot(coords,data[1,] ,col=colour,main = paste(" Fist component")) 
#fields::quilt.plot(loc_to_pred,pr1$pred,col=colour,
# main = paste(" Kriging first component"),ylab="")
#par(opar)


}
\keyword{Kriging}

\name{winds}
\alias{winds}
\title{Irish Daily Wind Speeds}
\usage{data(winds)}
\description{
 A matrix containing daily wind speeds, in
 kilometers per hour, from 1961 to 1978 at 12 sites in Ireland
}
\format{A (\eqn{6574 \times 11}{6574 x 11})-matrix containing wind speed observations.}
\source{
 Haslett, J. and Raftery, A. E. (1989),
 Space-time modelling with long-memory dependence: assessing
 Ireland's wind-power resource (with discussion),
 \emph{Applied Statistics}, 38, 1--50.
}
\keyword{datasets}

\name{GeoSimcond}
\alias{GeoSimcond}
\title{Conditional simulation of spatial Gaussian and non-Gaussian random fields}
\description{
 Performs global or nearest-neighbour local conditional simulation for univariate
 spatial random fields at explicit prediction coordinates. Cholesky simulation is
 available together with the approximate turning-bands method when supported by
 the selected correlation model.  The combination \code{local=TRUE} and
 \code{method="TB"} avoids global observation covariance matrices and is intended
 for large spatial datasets.
}
\usage{
GeoSimcond(estobj = NULL, data, coordx, coordy = NULL, coordz = NULL, coordt = NULL,
 coordx_dyn = NULL, corrmodel, distance = "Eucl", grid = FALSE, loc,
 maxdist = NULL, maxtime = NULL, method = "Cholesky", model = "Gaussian",
 n = 1, nrep = 1, local = FALSE, L = 1000, neighb = NULL,
 param, anisopars = NULL, radius = 1, sparse = FALSE, time = NULL,
 copula = NULL, X = NULL, Xloc = NULL, Mloc = NULL,
 parallel=FALSE, ncores = 6, progress=FALSE, n_iter=25L,
 check.duplicates=FALSE, nloc=NULL, mcmc_thin=1L)
}
\arguments{
 \item{estobj}{Object of class \code{GeoFit} containing model information}
 \item{data}{Numeric vector/matrix/array of observed data}
 \item{coordx}{Numeric matrix with one row per observed location and two or three spatial coordinate columns}
 \item{coordy}{Optional numeric vector of y-coordinates}
 \item{coordz}{Optional numeric vector of z-coordinates}
 \item{coordt}{Retained for API compatibility. Space-time conditional simulation is not currently implemented.}
 \item{coordx_dyn}{Retained for API compatibility. Dynamic coordinates are not currently implemented.}
 \item{corrmodel}{String specifying correlation model name}
 \item{distance}{String specifying distance metric (default: "Eucl")}
 \item{grid}{Must currently be \code{FALSE}; supply explicit coordinates.}
 \item{loc}{Numeric matrix of prediction locations (n x 2)}
 \item{maxdist}{Optional maximum distance for local kriging}
 \item{maxtime}{Optional maximum temporal distance}
 \item{method}{Unconditional simulation method. Currently \code{"Cholesky"} and \code{"TB"} are available; \code{"CE"} is retained for API compatibility but rejected by the explicit-coordinate conditional workflow.}
 \item{model}{String specifying random field type (default: "Gaussian")}
 \item{n}{For direct \code{Binomial}, the number of trials, either scalar or one positive integer per observed location. For direct \code{BinomialNeg}, the common positive integer number \eqn{r} of successes.}
 \item{nrep}{Number of retained conditional simulation replicates (default: 1). For direct count models these are successive thinned draws from the latent Gibbs chain.}
 \item{local}{Logical. If \code{FALSE} (default), global conditioning is used. If
 \code{TRUE}, each prediction location is conditioned only on the observations
 selected by \code{neighb} and/or \code{maxdist}.}
 \item{L}{Number of lines for turning bands method (default: 1000)}
 \item{neighb}{Optional positive integer giving the number of nearest observed
 locations used for local conditional simulation.  When \code{local=TRUE}, at
 least one of \code{neighb} or \code{maxdist} must be supplied.}
 \item{param}{List of parameter values}
 \item{anisopars}{List with anisotropy angle and ratio}
 \item{radius}{Radius used by spherical distance calculations (default: 1)}
 \item{sparse}{Must currently be \code{FALSE}.}
 \item{time}{Retained for API compatibility. Space-time conditional simulation is not currently implemented.}
 \item{copula}{Optional string specifying copula type}
 \item{X}{Optional design matrix for the marginal location at the observed locations.}
 \item{Xloc}{Optional design matrix for the marginal location at prediction locations. If \code{X} is supplied and the margin uses a location parameter, either \code{Xloc} or \code{Mloc} must be supplied.}
 \item{Mloc}{Optional vector of known/fitted marginal location values at prediction locations; an alternative to \code{Xloc}.}
 \item{parallel}{Logical; default \code{FALSE}. If \code{TRUE}, supported expensive stages use parallel workers. For \code{method="TB"} or \code{"CE"}, the setting is passed to \code{GeoSimapprox}; for \code{local=TRUE}, local kriging weights may also be computed in parallel. The final local substitution is evaluated in the main R process using a compact sparse weight operator and bounded batches, avoiding duplication of large simulation and weight objects across workers.}
 \item{ncores}{Positive integer or \code{NULL}; default \code{6}. With \code{parallel=TRUE}, an explicit integer requests that many workers, capped by detected cores and available jobs. Set \code{ncores=NULL} for automatic selection, capped at six workers and normally leaving one detected core free. A value of 1 is treated as serial execution.}
 \item{progress}{If TRUE then a progress bar is shown.}
 \item{n_iter}{Positive integer; number of Gibbs sweeps used by latent conditional samplers. For direct \code{Binomial} and \code{BinomialNeg}, it is the burn-in length of the count-constrained Gibbs chain; retained simulations are then separated by \code{mcmc_thin} sweeps. The default for these direct count models is 25. For the other latent Gibbs engines, omitting \code{n_iter} retains the historical default of 1000 sweeps. An explicitly supplied value is always respected. It is ignored by the Gaussian copula conditional simulator.}

\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
 \item{nloc}{Prediction-side count size. For direct \code{Binomial}, a scalar or one positive integer per prediction location; it is required when observation-side \code{n} varies by site. For direct \code{BinomialNeg}, if supplied it must equal the common \eqn{r}.}
 \item{mcmc_thin}{Positive integer thinning interval for the direct Binomial and Negative-Binomial count-constrained latent Gibbs sampler. The default is 1.}
}
\details{
The currently validated domain is univariate, purely spatial, non-dynamic and
non-grid conditional simulation. Both global and nearest-neighbour local
conditioning are available for the Gaussian latent-substitution paths described
below. Unsupported domain combinations are rejected before covariance matrices
or unconditional simulations are constructed.

Without a copula, the currently supported models are
\code{Gaussian}, \code{LogGaussian}, \code{Tukeyh}, \code{Tukeyh2},
\code{SinhAsinh}, \code{SkewGaussian}, \code{Gamma}, \code{Weibull}, \code{Binomial}, and \code{BinomialNeg}.  Direct non-copula \code{Gamma}
conditional simulation uses the finite Gaussian-square construction and therefore
requires \code{param$shape} to be a positive integer; non-integer values are
rejected rather than rounded. For direct and Gaussian-copula Tukey margins,
\code{Tukeyh} requires \code{0 <= tail < 0.5}; \code{Tukeyh2} requires
\code{0 <= tail1 < 0.5} and \code{0 <= tail2 < 0.5}, with \code{tail1}
the right-tail and \code{tail2} the left-tail parameter. Zero is valid and
recovers the Gaussian transformation on that side. \code{SinhAsinh} requires
a strictly positive \code{tail}.

For the direct \code{Binomial} and \code{BinomialNeg} random fields, conditional
simulation uses the repeated latent Gaussian threshold construction defining the
models.  At each observed site the latent Gaussian copies are updated as one
count-constrained block.  Their signs are sampled exactly from a conditional
Bernoulli distribution by a log-scale dynamic program, after which the latent
Gaussian magnitudes are sampled from the corresponding univariate truncated
normal full conditionals.  For \code{Binomial}, the constraint is
\deqn{\sum_{l=1}^{n_i} I\{Z_l(s_i)>0\}=y_i.}
For \code{BinomialNeg}, writing \eqn{t_i=y_i+r}, the constraint is
\deqn{\sum_{l=1}^{t_i-1} I\{Z_l(s_i)>0\}=r-1, \qquad Z_{t_i}(s_i)>0.}
The native sampler maintains the dense Gaussian precision update in compiled
code and directly generates the prediction-side latent fields, so the full
\eqn{L\times n_{obs}\times n_{rep}} latent array is not returned to R.
The first \code{n_iter} full site sweeps are discarded and retained conditional
realisations are separated by \code{mcmc_thin} sweeps.  This direct count path
currently requires \code{method="Cholesky"} and \code{local=FALSE}.

With \code{copula="Gaussian"}, conditional simulation is implemented for the
continuous margins \code{Gaussian}, \code{StudentT}, \code{LogGaussian},
\code{Gamma}, \code{Weibull}, \code{Beta}, \code{Beta2},
\code{Kumaraswamy}, \code{Kumaraswamy2}, \code{Logistic},
\code{SkewLaplace}, \code{Tukeyh}, \code{Tukeyh2}, and
\code{SinhAsinh}. With \code{copula="Clayton"} or
\code{copula="SkewGaussian"}, the currently validated continuous margins are
the same list except for \code{Tukeyh}, \code{Tukeyh2}, and
\code{SinhAsinh}.
For copula-based \code{Gamma}, \code{Weibull}, \code{Beta},
\code{Beta2}, \code{Kumaraswamy}, and \code{Kumaraswamy2} margins,
\code{param$sill} is not required: their marginal dispersion is determined by
the corresponding shape parameters.  The \code{sill} parameter is required
only for copula margins whose marginal scale is explicitly parameterized by it.
For the skew--Gaussian copula, \code{param$nu} is the bounded asymmetry
parameter \eqn{\eta\in(-1,1)}; the latent Gibbs sampler uses
\eqn{\gamma_\eta=\eta/\sqrt{1-\eta^2}}.

For the constructive Clayton-like copula, \code{param$nu} must be a positive
integer.  If \eqn{u_i} is the observed copula uniform, the sampler uses
\eqn{r_i=u_i^{2/\nu}} and the latent representation
\deqn{r_i=\frac{\sum_{k=1}^{\nu} Z_{ki}^2}{\sum_{k=1}^{\nu} Z_{ki}^2+W_{1i}^2+W_{2i}^2}.}
At each observed site the \eqn{\nu+2} Gaussian latent variables are updated
with a Gibbs kernel that preserves this ratio exactly.  Directional full
conditionals are von Mises--Fisher and the common latent radius is updated from
a one-dimensional log-concave full conditional.  The retained finite-sweep
state is therefore a Monte Carlo approximation to the Clayton latent
conditional distribution after \code{n_iter} sweeps.  Conditional Gaussian
simulation of the latent fields at \code{loc} is then followed by the Clayton
ratio reconstruction and the location-specific marginal quantile.

Discrete copula margins are not routed through the former mid-PIT
approximation, because that approximation is not exact conditioning on the
latent intervals.

For Gaussian fields, conditional simulations use the substitution method
\deqn{Z_c(s_0)=E\{Z(s_0)\mid Z(s)=z\}+Z^*(s_0)-WZ^*(s),}
where the unconditional residual field \eqn{Z^*} is generated with zero mean
and the covariance specified by \code{param}. The declared orientation of the
kriging weights is used explicitly.

With \code{local=TRUE}, the same substitution identity is applied separately at
each prediction location using only its selected neighborhood \eqn{N_j}:
\deqn{Z_c(s_j)=\hat Z_j+Z^*(s_j)-w_j^T Z^*_{N_j}.}
The local systems are solved once, and GeoSimcond keeps a compact map of
neighbor indices and weights.  When possible this map is assembled once as a
\code{spam} sparse matrix with \eqn{n_{loc}} rows and \eqn{n_{obs}} columns.
Conditional corrections for multiple realizations are then evaluated in bounded
matrix batches rather than by repeating the local graph reduction for every
replicate.  The sparse representation contains only the selected local weights;
GeoSimcond does not retain one local covariance matrix per prediction site.
Prediction locations with no observation inside a \code{maxdist}-only
neighborhood receive zero conditioning correction, so their local conditional
approximation reduces to the marginal mean plus the unconditional residual at
that location.

Local conditioning is currently validated for \code{Gaussian},
\code{LogGaussian}, \code{Tukeyh}, \code{Tukeyh2}, and \code{SinhAsinh}
without a copula, and for all supported continuous margins with
\code{copula="Gaussian"}.  In particular, \code{Tukeyh}, \code{Tukeyh2}
and \code{SinhAsinh} are also supported as continuous margins under the
Gaussian copula and use the same latent-Gaussian fast path.  The Gibbs-based constructive
\code{SkewGaussian}, \code{Gamma}, \code{Weibull}, \code{Binomial}, and \code{BinomialNeg} paths and the
\code{Clayton} and \code{SkewGaussian} copulas still require
\code{local=FALSE}.

For the sinh--arcsinh model, the implementation uses the monotone transformation
\deqn{Y(s)=\mu(s)+\sqrt{\sigma^2}\,\sinh\{[\operatorname{asinh}\{Z(s)\}+\eta]/\nu\},}
and conditions on the Gaussian scale using its exact inverse
\deqn{Z(s)=\sinh\{\nu\,\operatorname{asinh}[(Y(s)-\mu(s))/\sqrt{\sigma^2}]-\eta\}.}
Thus \code{LogGaussian}, \code{Tukeyh}, \code{Tukeyh2} and
\code{SinhAsinh} require only one latent Gaussian conditional simulation per
replicate; no non-Gaussian Gibbs step is introduced.

For margins with a location parameter, location-specific covariates are preserved:
if \code{X} is supplied for the observations, prediction requires either
\code{Xloc} or \code{Mloc}; the function does not silently replace a varying
prediction-side location by the intercept.

For continuous copula margins, observations are transformed by their fitted
marginal CDF to the copula scale, conditional simulation is performed on the
latent copula construction, and each realization is transformed back with the
location-specific marginal quantile.  For the Gaussian copula this is standard
Gaussian conditional simulation.  The skew--Gaussian case uses its latent Gibbs
sampler, while the Clayton-like case uses the Gaussian-square ratio Gibbs
sampler described above.  In both MCMC cases, a retained finite-sweep draw is
a Monte Carlo approximation to the target conditional distribution after
\code{n_iter} sweeps.

\code{method="TB"} makes the unconditional simulation step approximate.  For
large datasets it can be combined with \code{local=TRUE}; the turning-bands
simulation then avoids a dense covariance factorization for the unconditional
field, while local conditioning avoids the global observation covariance system.
For a fixed neighborhood size \eqn{m}, local preprocessing consists of small
\eqn{m \times m} kriging systems and the substitution correction uses only the
retained local weights. For multiple conditional realizations, the same sparse
local weight operator is reused and applied in bounded matrix batches. This
optimization is shared automatically by the one-Gaussian monotone transformations
and by Gaussian-copula margins. When \code{parallel=TRUE}, approximate
unconditional simulation and local-weight construction honor \code{ncores}; the
final sparse substitution remains in the main R process to avoid copying the
unconditional simulations and weight map to additional worker processes. Spatial circulant
embedding is not currently available
because it requires a regular grid, whereas the validated conditional workflow
combines explicit observation and prediction coordinates. The nugget, distance,
radius and anisotropy settings are propagated to the latent Gaussian simulations.
}
\value{
An object of class \code{GeoSimcond} containing \code{condsim}, a list with one
element per successful conditional replicate. Each element is a numeric vector of
length \code{numloc}, in prediction-location order. Thus, if a dense matrix is
needed for post-processing, \code{do.call(rbind, lapply(x$condsim, as.numeric))}
produces a matrix with replications in rows and prediction locations in columns.
The object also contains \code{cond_mean} and \code{cond_var}, computed across
replicates using a streaming update without materializing an additional full
replicate-by-location matrix, together with the model, parameter, conditioning
(\code{local}, \code{neighb}) and simulation-method metadata. With a single
replicate, \code{cond_var} contains \code{NA} values because a sample variance
cannot be estimated.
}
\references{
Gaetan, C. and Guyon, X. (2010) \emph{Spatial Statistics and Modelling}.
Springer Verlag, New York.

Blasi, F., Caamaño-Carrillo, C., Bevilacqua, M. and Furrer, R. (2022).
A selective view of climatological data and likelihood estimation.
\emph{Spatial Statistics}, 50, 100596.

Caamaño-Carrillo, C., Bevilacqua, M., López, C. and Morales-Oñate, V. (2024).
Nearest neighbors weighted composite likelihood based on pairs for (non-)Gaussian
massive spatial data with an application to Tukey-hh random fields estimation.
\emph{Computational Statistics and Data Analysis}, 191, 107887.
}
\seealso{
\code{\link[GeoModels]{GeoSim}}, 
\code{\link[GeoModels]{GeoKrig}}
}
\author{
Moreno Bevilacqua \email{moreno.bevilacqua89@gmail.com},\\
Víctor Morales Oñate \email{victor.morales@uv.cl},\\
Christian Caamaño-Carrillo \email{chcaaman@ubiobio.cl}
}
\keyword{spatial}
\keyword{simulation}
\examples{
library(GeoModels)

##############################################
## conditional simulation of a Gaussian rf ###
##############################################
model="Gaussian"
set.seed(79)
### conditioning locations
x = runif(250, 0, 1)
y = runif(250, 0, 1)
coords=cbind(x,y)

# Set the exponential cov parameters:
corrmodel = "GenWend"
mean=0; sill=1; nugget=0
scale=0.2;smooth=0;power2=4

param=list(mean=mean,sill=sill,nugget=nugget,scale=scale,smooth=smooth,power2=power2)

# Simulation 
data = GeoSim(coordx=coords, corrmodel=corrmodel,model=model,
 param=param)$data

## estimation with pairwise likelihood
fixed=list(nugget=nugget,smooth=smooth,power2=power2)
start=list(mean=0,scale=scale,sill=1)
I=Inf
lower=list(mean=-I,scale=0,sill=0)
upper=list(mean= I,scale=I,sill=I)
# Maximum pairwise likelihood fitting :
fit = GeoFit(data, coordx=coords, corrmodel=corrmodel,model=model,
 likelihood='Marginal', type='Pairwise',neighb=3,
 optimizer="nlminb", lower=lower,upper=upper,
 start=start,fixed=fixed)

# locations to simulate 
xx=seq(0,1,0.025)
loc_to_sim=as.matrix(expand.grid(xx,xx))

# Conditional simulation
sim_result <- GeoSimcond(fit,loc = loc_to_sim,nrep=50)

cond_mean=sim_result$cond_mean # conditional mean
cond_var =sim_result$cond_var # conditional var

# Empirical pointwise intervals from the conditional simulations.
# Rows are replications and columns are prediction locations.
sim_mat <- do.call(rbind, lapply(sim_result$condsim, as.numeric))
interval_summary <- t(apply(sim_mat, 2, quantile,
                            probs = c(0.025, 0.5, 0.975)))
colnames(interval_summary) <- c("q025", "q500", "q975")
head(interval_summary)

par(mfrow=c(1,3))
if (requireNamespace("fields", quietly = TRUE)) fields::quilt.plot(coords, data)
if (requireNamespace("fields", quietly = TRUE)) fields::quilt.plot(loc_to_sim, cond_mean)
if (requireNamespace("fields", quietly = TRUE)) fields::quilt.plot(loc_to_sim, cond_var)
par(mfrow=c(1,1))

##############################################
## conditional simulation of a LogGaussian rf 
##############################################
model="LogGaussian"
set.seed(79)
### conditioning locations
x = runif(500, 0, 1)
y = runif(500, 0, 1)
coords=cbind(x,y)

# Set the exponential cov parameters:
corrmodel = "Matern"
mean=0; sill=.1; nugget=0
scale=0.2;smooth=0.5

param=list(mean=mean,sill=sill,nugget=nugget,scale=scale,smooth=smooth)

# Simulation 
data = GeoSim(coordx=coords, corrmodel=corrmodel,model=model,
 param=param)$data

## estimation with pairwise likelihood
fixed=list(nugget=nugget,smooth=smooth)
start=list(mean=0,scale=scale,sill=1)
I=Inf
lower=list(mean=-I,scale=0,sill=0)
upper=list(mean= I,scale=I,sill=I)
# Maximum pairwise likelihood fitting :
fit = GeoFit(data, coordx=coords, corrmodel=corrmodel,model=model,
 likelihood='Marginal', type='Pairwise',neighb=3,
 optimizer="nlminb", lower=lower,upper=upper,
 start=start,fixed=fixed)

# locations to simulate 
xx=seq(0,1,0.025)
loc_to_sim=as.matrix(expand.grid(xx,xx))

# Conditional simulation
sim_result <- GeoSimcond(fit,loc = loc_to_sim,nrep=50)

cond_mean=sim_result$cond_mean # conditional mean
cond_var =sim_result$cond_var # conditional var

par(mfrow=c(1,3))
if (requireNamespace("fields", quietly = TRUE)) fields::quilt.plot(coords, data)
if (requireNamespace("fields", quietly = TRUE)) fields::quilt.plot(loc_to_sim,cond_mean)
if (requireNamespace("fields", quietly = TRUE)) fields::quilt.plot(loc_to_sim,cond_var)
par(mfrow=c(1,1))

}

\name{GeoVariogram}
\alias{GeoVariogram}
\encoding{UTF-8}
\title{Empirical semivariogram estimation}
\description{
 Computes an empirical estimate of the semivariogram for spatial, spatio-temporal,
 and bivariate random fields.
}
\usage{
GeoVariogram(data, coordx, coordy=NULL, coordz=NULL, coordt=NULL,
 coordx_dyn=NULL, cloud=FALSE, distance="Eucl",
 grid=FALSE, maxdist=NULL, neighb=NULL,
 maxtime=NULL, numbins=NULL,
 radius=1, type='variogram', bivariate=FALSE,
 subsample=1, subsample_t=1, numbins_t=NULL, directed=FALSE)
}
\arguments{
 \item{data}{
 A numeric vector of length \eqn{d}{d} (a single spatial realisation), or a
 \eqn{d \times d}{d x d} matrix (a single realisation on a regular grid), or a
 \eqn{t \times d}{t x d} matrix (a single spatio-temporal realisation), or a
 \eqn{d \times d \times t}{d x d x t} array (a single spatio-temporal realisation on a regular grid).
 See \code{\link{GeoFit}} for details.
 }
 \item{coordx}{
 Spatial coordinates. Either a numeric vector giving the first coordinate, or a
 \eqn{d \times 2}{d x 2} (or \eqn{d \times 3}{d x 3}) matrix of coordinates.
 If \code{distance} refers to great-circle distances, coordinates must be provided in
 lon/lat format (decimal degrees) and the sphere radius is set by \code{radius}.
 }
 \item{coordy}{
 A numeric vector giving the second spatial coordinate.
 Optional, default is \code{NULL}.
 }
 \item{coordz}{
 A numeric vector giving the third spatial coordinate (if needed).
 Optional, default is \code{NULL}.
 }
 \item{coordt}{A numeric vector of temporal coordinates. If \code{NULL} (default), a purely spatial random field is assumed. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{
 For dynamic locations, a list with one two- or three-column coordinate matrix per temporal instant.  The list length must match \code{length(coordt)}; \code{data[[t]]} must have one value per row of \code{coordx_dyn[[t]]}, in the same order.
 }
 \item{cloud}{
 Logical; if \code{TRUE} the semivariogram cloud is computed for a univariate
 purely spatial field. If \code{FALSE} (default), a binned empirical semivariogram
 is returned. For large data sets, combine \code{cloud=TRUE} with \code{maxdist},
 \code{neighb}, or \code{subsample} to limit the number of returned pairs.
 }
 \item{distance}{
 String specifying the spatial distance. Default is \code{"Eucl"} (Euclidean distance).
 See \code{\link{GeoFit}} for details.
 }
 \item{grid}{
 Logical; if \code{FALSE} (default) data are interpreted as observations on irregularly
 spaced locations. If \code{TRUE}, data are interpreted as observations on a regular grid.
 }
 \item{maxdist}{
 A positive finite numeric maximum spatial distance. In the bivariate case a
 scalar or a length-3 vector (first marginal, cross, second marginal) is
 accepted. Use \code{NULL} to let the function use the available spatial
 range. See Details.
 }
 \item{neighb}{
 Numeric; an optional positive integer indicating the order of neighborhood
 (useful for large datasets). Neighborhood pair selection is available for
 spatial and spatio-temporal semivariograms. In the bivariate case a length-3
 vector can be used for the first marginal, cross, and second marginal pair sets.
 See Details.
 }
 \item{maxtime}{A positive finite maximum temporal lag, expressed in the same units as \code{coordt}, to be considered for spatio-temporal semivariograms. See Details.}
 \item{numbins}{
 Historical GeoModels argument controlling the spatial bin grid. It is the
 number of spatial bin boundaries, so the number of empirical spatial classes
 is \code{numbins - 1}. The default is 13 boundaries (12 classes).
 }
 \item{numbins_t}{
 Optional positive integer giving the number of temporal classes for irregularly
 spaced \code{coordt}. If \code{NULL}, \code{numbins - 1} classes are used.
 For regularly spaced times the attainable temporal lags are generated directly
 from the common spacing and this argument is not used.
 }
 \item{directed}{
 Logical used only when \code{neighb} is supplied. The default \code{FALSE}
 removes reciprocal nearest-neighbour duplicates so that the empirical
 semivariogram is based on unordered pairs. Set \code{TRUE} to retain the
 directed nearest-neighbour graph used by pairwise composite-likelihood code.
 }
 \item{radius}{
 Numeric; radius of the sphere when using great-circle distances. Default is 1.
 }
 \item{type}{
 String; type of semivariogram. Currently available: \code{"variogram"}.
 }
 \item{bivariate}{
 Logical; if \code{FALSE} (default) data are interpreted as univariate spatial/spatio-temporal
 realisations. If \code{TRUE}, \code{data} is interpreted as a realisation from a bivariate field
 and (cross-)semivariograms are computed.
 }
 \item{subsample}{
 Numeric in \eqn{(0,1]}{(0,1]}. Proportion of spatial locations to be used to compute the semivariogram
 (useful for large datasets). Default is 1 (use all locations).
 }
 \item{subsample_t}{
 Numeric in \eqn{(0,1]}{(0,1]}. Proportion of time points to be used in spatio-temporal settings
 (when \code{coordt} is provided). Default is 1 (use all time points).
 }
}

\details{
 We report the definition of the semivariogram in the spatial case; extensions to spatio-temporal
 and bivariate settings are based on the same principles.

 For a spatial random field \eqn{Z(\cdot)}{Z(.)}, the (classical) binned semivariogram estimator is
 defined as
 \deqn{\hat{\gamma}(h) = \frac{1}{2 |N(h)|}\sum_{(x_i,x_j)\in N(h)} \{Z(x_i)-Z(x_j)\}^2,}
 where \eqn{N(h)}{N(h)} is the set of all sample pairs whose spatial distance falls within a tolerance
 region around lag \eqn{h}{h} (equally spaced intervals are used when \code{cloud=FALSE}).

 The historical \code{numbins} argument sets the number of spatial bin
 boundaries; hence \code{numbins - 1} empirical spatial classes are formed when
 \code{cloud=FALSE}.

 The \code{maxdist} argument sets a strictly positive finite maximum spatial
 distance. If no pair falls below the requested cutoff, a valid empty
 semivariogram object is returned instead of constructing decreasing bins.

 The \code{maxdist} option can be combined with \code{neighb} to reduce the number of pairs when handling
 large datasets, by restricting computations to local neighborhoods. By default
 reciprocal directed nearest-neighbour edges are deduplicated before binning;
 \code{directed=TRUE} retains the directed graph. For chordal and geodesic
 distances, neighbour ordering is obtained from three-dimensional unit-sphere
 coordinates and the reported lags are then evaluated in the requested metric.

 Spatial and temporal bins are left-closed and right-open, except for the last
 bin which is closed on the right. Thus a pair exactly at the maximum retained
 lag is not discarded.

 The \code{maxtime} argument sets the maximum temporal lag considered for
 spatio-temporal semivariograms. For regularly spaced times, attainable temporal
 lags are generated in linear memory/time from the common spacing; no
 \eqn{T \times T}{T x T} matrix of all time differences is formed. For
 irregular times a controlled grid of \code{numbins_t} temporal classes is used.
 The returned spatio-temporal surface is rectangular on \code{centers} by
 \code{centert}; cells with no valid pairs are returned as \code{NA} with count
 zero rather than being removed.

 For dynamic sites, the temporal marginal is the empirical \eqn{\gamma(0,u)}
 and therefore uses only spatially collocated locations (up to numerical
 tolerance). Nearby but non-collocated locations contribute to the positive-
 distance space-time surface, not to the temporal marginal. Consequently, in a
 fully dynamic design with no spatial locations repeated across temporal
 instants, \code{variogramt} is \code{NA} in the affected temporal classes.
 This is an expected property of the sampling design, not a failure of the
 space-time variogram: the positive-distance surface \eqn{\gamma(h,u)} remains
 empirically estimable. In this case \code{plot.GeoVariogram} labels the panel
 ``Temporal marginal unavailable (no repeated spatial locations)''. Repeated
 spatial locations across times are needed only when the empirical temporal
 marginal itself is required.

 In the bivariate case the two marginal semivariograms and the cross-semivariogram
 are evaluated on the same spatial bins. The cross-semivariogram uses the classical
 increment-product estimator on unordered positive-distance pairs. The trivial
 collocated contribution at lag zero is not mixed into the first positive spatial
 class. If two dynamic/support coordinate sets are supplied, they must be aligned
 to define this estimator unambiguously.

 The \code{subsample} and \code{subsample_t} arguments provide additional control
 for large datasets by using only a proportion of spatial locations and/or time
 points. With \code{grid=TRUE}, the grid is first normalized to explicit spatial
 locations, so full-grid and subsampled calculations follow the same path.

 Missing values \code{NA}/\code{NaN} are allowed and pairs involving them are
 skipped consistently. Infinite observations are rejected.
}

\section{Spatio-temporal ordering}{
For fixed sites, \code{data} is a \eqn{T \times N}{T x N} matrix with times in
rows and sites in columns, and is internally read in the order
\code{c(t(data))}.  For dynamic sites, \code{data} and \code{coordx_dyn} are
aligned lists; the function concatenates complete temporal blocks in list
order.  See \code{\link{GeoModels-spacetime-ordering}}.
}

\value{
 Returns an object of class \code{"GeoVariogram"}. The list contains, as applicable:

 \item{bins}{Spatial bin boundaries when \code{cloud=FALSE}; spatial pair distances when \code{cloud=TRUE}.}
 \item{bint}{Temporal lag representatives for a spatio-temporal variogram.}
 \item{bivariate}{Logical indicating a bivariate empirical variogram.}
 \item{cloud}{Logical indicating a variogram cloud.}
 \item{centers}{Spatial bin centers.}
 \item{centert}{Temporal lag representatives used by the rectangular spatio-temporal surface.}
 \item{distance}{Spatial distance type used to construct the empirical variogram.}
 \item{radius}{Sphere radius used for chordal or geodesic distances.}
 \item{grid}{Logical recording whether the original input was supplied as a grid.}
 \item{neighb}{Neighborhood order used for pair selection, or \code{NULL}.}
 \item{directed}{Whether reciprocal nearest-neighbour pairs were retained.}
 \item{lenbins}{Numbers of pairs in the spatial bins. In the bivariate case this is a two-row matrix.}
 \item{lenbinst}{Numbers of pairs in the cross/spatio-temporal bins. For space-time objects this follows the same row-major ordering as \code{variogramst}.}
 \item{lenbint}{Numbers of pairs in the temporal bins.}
 \item{maxdist}{Maximum spatial distance requested by the user.}
 \item{maxtime}{Maximum temporal lag requested by the user.}
 \item{numbins_t}{Requested number of temporal classes for irregular times, or \code{NULL}.}
 \item{regular}{Logical indicating regularly spaced temporal coordinates for a space-time object.}
 \item{time.breaks}{Internal temporal bin boundaries used for space-time binning.}
 \item{spacetime_dyn}{Logical indicating dynamic spatial coordinates.}
 \item{temporal.margin}{For space-time objects, a label indicating that the temporal margin is based on same-site/collocated pairs.}
 \item{subsample}{Spatial subsampling proportion.}
 \item{subsample_t}{Temporal subsampling proportion.}
 \item{variograms}{Empirical spatial semivariogram; a two-row matrix in the bivariate case.}
 \item{variogramst}{Empirical cross-semivariogram in the bivariate case, or the rectangular spatio-temporal surface stored in row-major order. Empty cells are \code{NA}.}
 \item{variogramt}{Empirical temporal marginal semivariogram.}
 \item{type}{Type of empirical variogram.}
}

\references{
 Cressie, N. A. C. (1993) \emph{Statistics for Spatial Data}. New York: Wiley.

 Gaetan, C. and Guyon, X. (2010) \emph{Spatial Statistics and Modeling}. Springer-Verlag, New York.
}

\seealso{
\code{\link{GeoFit}} for model fitting,
  \code{\link{GeoWLS}} for weighted least-squares estimation from empirical variograms,
  \code{\link{GeoCovariogram}} for fitted covariance and variogram values,
  \code{\link{plot.GeoVariogram}} for plotting empirical variograms.
}

\author{
 Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com}, \url{https://sites.google.com/view/moreno-bevilacqua/home},
 Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
 Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl}, \url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{
library(GeoModels)

################################################################
### Example 1. Empirical semivariogram from a spatial Gaussian
### random field with Matérn correlation.
################################################################
set.seed(514)
x = runif(200, 0, 1)
y = runif(200, 0, 1)
coords = cbind(x,y)

corrmodel = "Matern"
mean = 0
sill = 1
nugget = 0
scale = 0.3/3
smooth = 0.5

data = GeoSim(coordx=coords, corrmodel=corrmodel,
 param=list(mean=mean, smooth=smooth, sill=sill,
 nugget=nugget, scale=scale))$data

vario = GeoVariogram(coordx=coords, data=data, maxdist=0.6)
plot(vario, pch=20, ylim=c(0,1), ylab="Semivariogram", xlab="Distance")

################################################################
### Example 2. Empirical semivariogram for a spatio-temporal
### Gaussian random field with Gneiting correlation.
################################################################
set.seed(331)
x = runif(200, 0, 1)
y = runif(200, 0, 1)
coords = cbind(x,y)
times = seq(1,10,1)

data = GeoSim(coordx=coords, coordt=times, corrmodel="gneiting",
 param=list(mean=0, scale_s=0.08, scale_t=0.4, sill=1,
 nugget=0, power_s=1, power_t=1, sep=0.5))$data

vario_st = GeoVariogram(data=data, coordx=coords, coordt=times,
 maxtime=7, maxdist=0.5)
plot(vario_st, pch=20)

################################################################
### Example 3. Empirical (cross-)semivariograms for a bivariate
### Gaussian random field with Bi-Matérn covariance.
################################################################
set.seed(293)
x = runif(400, 0, 1)
y = runif(400, 0, 1)
coords = cbind(x,y)

param = list(mean_1=0, mean_2=0,
 scale_1=0.1/3, scale_2=0.15/3, scale_12=0.15/3,
 sill_1=1, sill_2=1,
 nugget_1=0, nugget_2=0,
 smooth_1=0.5, smooth_12=0.5, smooth_2=0.5,
 pcol=0.3)

data = GeoSim(coordx=coords, corrmodel="Bi_matern", param=param)$data
biv_vario = GeoVariogram(data, coordx=coords, bivariate=TRUE, maxdist=0.5)
plot(biv_vario, pch=20)
}

\keyword{Variogram}

\name{GeoScatterplot}
\alias{GeoScatterplot}
\encoding{UTF-8}
\title{Scatterplots of Spatial or Spatio-temporal Pairs and Fitted Bivariate Contours}
\description{
Produces scatterplots of observations associated with spatial or
spatio-temporal pairs selected by distance classes, nearest-neighbour order,
and temporal lag. When the first argument is a fitted \code{GeoFit} object,
the function can optionally superimpose contour lines of the fitted bivariate
density on the original response scale, the uniform
probability-integral-transform scale, or the Gaussian-score scale.
}
\usage{
GeoScatterplot(data, coordx = NULL, coordy = NULL, coordz = NULL,
  coordt = NULL, coordx_dyn = NULL, distance = "Eucl", grid = FALSE,
  maxdist = NULL, neighb = NULL, times = NULL, time.lag = NULL,
  numbins = 4, radius = 1, bivariate = FALSE,
  contour = inherits(data, "GeoFit"),
  residuals = FALSE, scale = c("Original", "Gaussian", "Uniform"),
  gaussian.range = 3, ngrid = 80, nlevels = 6, levels = NULL,
  contour.col = "red", contour.lwd = 1.5,
  contour.labels = FALSE,
  point.col = grDevices::adjustcolor("#481567FF", 0.45),
  lag.method = c("median", "mean"), ...)
}
\arguments{
  \item{data}{Either the data to be plotted or an object of class
  \code{GeoFit}. For the original interface, this can be a numeric vector,
  matrix, or array containing a spatial, spatio-temporal, or bivariate
  realization. See \code{\link{GeoFit}} for the supported data layouts.}

  \item{coordx}{A numeric coordinate vector, a two-column coordinate matrix,
  or a three-column coordinate matrix. Coordinates on a sphere are supplied
  in longitude/latitude format, in decimal degrees. This argument is not
  required when \code{data} is a \code{GeoFit} object.}

  \item{coordy}{An optional numeric vector containing the second spatial
  coordinate when \code{coordx} is supplied as a vector. The default is
  \code{NULL}.}

  \item{coordz}{An optional numeric vector containing the third spatial
  coordinate. The default is \code{NULL}.}

  \item{coordt}{An optional numeric vector containing temporal coordinates. If \code{NULL}, a purely spatial realization is assumed. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}

  \item{coordx_dyn}{For dynamic locations, a list with one two- or three-column coordinate matrix per element of \code{coordt}.  In the raw-data interface, \code{data[[t]]} must align row-by-row with \code{coordx_dyn[[t]]}.  In the \code{GeoFit} interface this layout is taken from the fitted object.}

  \item{distance}{Character string naming the spatial distance. The default is
  \code{"Eucl"}. See \code{\link{GeoFit}} for the available options. When
  \code{data} is a \code{GeoFit} object, the distance stored in the fitted
  object is used.}

  \item{grid}{Logical. If \code{FALSE}, the default, the observations are
  interpreted as values at non-equispaced sites. If \code{TRUE}, regular-grid
  coordinates are generated from the supplied coordinate vectors.}

  \item{maxdist}{A positive numeric value defining the maximum spatial
  distance. When \code{neighb = NULL}, the interval from zero to
  \code{maxdist} is divided into \code{numbins} distance classes. For a
  bivariate realization, this can have length one or two. Purely spatial
  distance-class pairs are generated in memory-bounded blocks rather than by
  allocating all \eqn{n(n-1)/2} possible pairs at once.}

  \item{neighb}{A positive integer or vector of positive integers defining
  the nearest-neighbour candidate sets used to construct the scatterplots.
  For a spatio-temporal \code{GeoFit} object, neighbours are determined for
  the selected temporal lags. At a positive temporal lag and with fixed
  spatial locations, the first nearest-neighbour candidate may be the
  observation at the same spatial location, corresponding to spatial lag
  zero. More generally, \code{neighb = m} displays the candidate set formed
  by the first \eqn{m} nearest neighbours at that temporal lag.}

  \item{times}{Optional numeric vector selecting temporal instants in the
  original data-and-coordinate interface. Entries can be values contained in
  \code{coordt} or valid one-based indices into \code{coordt}. At least two
  instants must be selected. The selected instants are put back in their
  original temporal order. If \code{NULL}, all temporal instants are used.}

  \item{time.lag}{Optional numeric vector of non-negative temporal lags used
  for spatio-temporal data. The requested values must be obtainable from the
  selected temporal coordinates. For a fitted \code{GeoFit} object, if
  \code{NULL}, all available temporal lags not exceeding the fitted
  \code{maxtime} are used. For the original data-and-coordinate interface, if
  \code{NULL}, all temporal lags generated by the selected instants are used.
  Lag zero is included. Regular temporal grids are handled without constructing
  a quadratic outer-difference matrix; irregular-grid lags are generated
  incrementally and numerically equivalent floating-point lags are merged.}

  \item{numbins}{A positive integer giving the number of distance classes when
  \code{neighb = NULL}. The default is 4.}

  \item{radius}{Numeric value giving the radius of the sphere for great-circle
  or chordal distances. The default is 1. For a \code{GeoFit} object, the
  radius stored in the fitted object is used unless this argument is supplied.}

  \item{bivariate}{Logical. If \code{TRUE}, the observations are interpreted
  as a realization of a bivariate random field. The default is \code{FALSE}.}

  \item{contour}{Logical. If \code{TRUE}, contour lines of the fitted
  bivariate density are added to each scatterplot. This option requires
  \code{data} to be a univariate spatial or spatio-temporal \code{GeoFit}
  object. The default is \code{TRUE} when \code{data} is a \code{GeoFit}
  object and
  \code{FALSE} otherwise. Set \code{contour = FALSE} to obtain only the
  scatterplots.}

  \item{residuals}{Logical. Relevant only when \code{data} is a
  \code{GeoFit} object and \code{scale = "Original"}. If \code{TRUE}, the
  scatterplot and fitted density are represented on a residual scale. For
  real-valued models, the fitted location is subtracted. For
  \code{"LogGaussian"}, \code{"Gamma"}, and \code{"Weibull"}, the
  observations are divided by the fitted multiplicative mean. For the direct
  count models \code{"Poisson"}, \code{"Binomial"},
  \code{"BinomialNeg"}/\code{"Geometric"}, and \code{"PoissonGamma"},
  residual-scale scatterplots use Pearson residuals based on the fitted marginal
  mean and variance, consistently with \code{\link{GeoResiduals}}. The default
  is \code{FALSE}.}

  \item{scale}{Character string specifying the scale used when \code{data}
  is a \code{GeoFit} object. \code{"Original"}, the default, uses the
  response scale. \code{"Uniform"} applies the fitted marginal distribution
  function to each observation. \code{"Gaussian"} additionally applies the
  standard-normal quantile function. The aliases \code{"Response"},
  \code{"PIT"}, and \code{"Normal"} are also accepted. The uniform and
  Gaussian scales require a fitted continuous marginal model. Pearson-residual
  objects can be displayed only on the original residual scale.}

  \item{gaussian.range}{Positive finite number defining the symmetric plotting
  interval \code{c(-gaussian.range, gaussian.range)} on the Gaussian-score
  scale when the user does not supply \code{xlim} and \code{ylim}. The
  default is 3, matching the Gaussian-scale copula comparison plots commonly
  used to display reflection asymmetry.}

  \item{ngrid}{Positive integer giving the number of grid points in each
  coordinate direction used to evaluate the fitted bivariate density. It must
  be at least 20. The default is 80.}

  \item{nlevels}{Positive integer controlling the number of automatically
  selected contour levels. Ignored when \code{levels} is supplied. The
  default is 6.}

  \item{levels}{Optional numeric vector of density levels at which contour
  lines are drawn. The default is \code{NULL}, in which case the levels are
  selected automatically.}

  \item{contour.col}{Colour of the fitted contour lines. The default is
  \code{"red"}.}

  \item{contour.lwd}{Line width of the fitted contour lines. The default is
  1.5.}

  \item{contour.labels}{Logical indicating whether contour labels are drawn.
  The default is \code{FALSE}.}

  \item{point.col}{Colour used for the scatterplot points in both the raw-data
  and \code{GeoFit} interfaces. The default is a semi-transparent purple. A
  graphical \code{col} supplied through \code{...} takes precedence.}

  \item{lag.method}{Character string specifying the representative spatial
  lag used for a nearest-neighbour panel. \code{"median"}, the default, uses
  the median distance of the displayed pairs; \code{"mean"} uses their mean
  distance. For distance classes, the midpoint of the class is used. In a
  spatio-temporal panel, the fitted density is evaluated at this representative
  spatial lag and at the temporal lag displayed in the panel title.}

  \item{...}{Additional graphical arguments passed to \code{plot}.}
}

\details{
An \eqn{h}-scatterplot displays paired observations associated with spatial
locations separated by a given range of distances or by a specified
nearest-neighbour order. The distance-class version requires \code{maxdist}
and \code{numbins}; the nearest-neighbour version requires \code{neighb}.
Nearest-neighbour scatterplots are generally preferable for large datasets.
Purely spatial distance-class pairs are generated in blocks with bounded
working memory. This avoids the former allocation of three vectors of length
\eqn{n(n-1)/2}; the total number of retained pairs can still be quadratic when
\code{maxdist} includes most location pairs. Distance classes use left-closed,
right-open intervals, with the final class also including \code{maxdist}.

The original interface, in which data and coordinates are supplied directly,
continues to produce scatterplots without requiring a fitted model. Supplying
a \code{GeoFit} object is an additional interface that extracts the data,
coordinates, distance, fitted parameters, and, when present, the fitted copula
from the object. If neither \code{neighb} nor \code{maxdist} is supplied, the
function attempts to use the corresponding pair-selection setting stored in
the fitted object. Nearest-neighbour selection uses the same \code{distance}
and \code{radius} convention as the fitted model (or as supplied in the raw
interface), and the lags returned by \code{GeoNeighIndex} are reused directly.

For fixed spatio-temporal locations in the original interface, \code{data}
must be a \code{length(coordt)} by \code{nrow(coordx)} matrix, following the
same time-by-site convention used by \code{GeoFit}, or a vector in the
corresponding time-block order. For dynamic locations, \code{coordx_dyn} must
contain one coordinate matrix per temporal instant and \code{data} can be a
matching list or a vector obtained by concatenating the temporal blocks.
Pairs are constructed with \code{GeoNeighIndex}; no pair-graph object is
required. The \code{times} argument can restrict the temporal instants before
pairs are generated, while \code{time.lag} selects the displayed temporal
lags.

With \code{contour = TRUE}, the bivariate density implied by the fitted model
is evaluated on a rectangular grid and superimposed on the points. For a
nearest-neighbour panel, the density is evaluated at the representative
spatial lag selected by \code{lag.method}; for a distance-class panel, it is
evaluated at the midpoint of the class.

For a fit obtained with \code{likelihood = "Marginal"} and
\code{type = "Independence"}, the theoretical contour is based only on the
fitted univariate marginal distribution. Any correlation model or copula
parameters stored in the \code{GeoFit} object are ignored because they do not
enter the independence likelihood. On the original scale the bivariate
density is therefore \eqn{f_1(y_1)f_2(y_2)}. On the uniform scale the
independence copula density is identically one, so there are no non-trivial
contour levels. On the Gaussian-score scale the density is
\eqn{\phi(z_1)\phi(z_2)}, giving the circular contours of two independent
standard normal variables.

For a spatio-temporal \code{GeoFit} object, pairs are classified by both the
spatial lag \eqn{h} and the temporal lag \eqn{u}. The fitted contour in each
panel therefore uses the fitted correlation \eqn{\rho(h,u)}. At a positive
temporal lag and with fixed spatial locations, the first nearest-neighbour
candidate is generally the observation at the same spatial location and thus
has \eqn{h=0}. Such colocated temporal pairs are retained. More generally,
\code{neighb = m} represents the candidate set containing the first \eqn{m}
nearest neighbours at the selected temporal lag. This is the same candidate
set returned by \code{GeoNeighIndex}; no additional filtering of zero spatial
lags is applied inside \code{GeoScatterplot}. Both fixed and dynamic spatial
locations are supported. When distance classes are requested, all spatial or
spatio-temporal pairs within \code{maxdist} are constructed for the selected
temporal lags and then divided into \code{numbins} classes.

When \code{scale = "Uniform"}, the plotted observations are
\eqn{U_i = F_i(Y_i)} and the fitted contour is the corresponding copula
density \eqn{c_h(u_1,u_2)}. When \code{scale = "Gaussian"}, the plotted
observations are \eqn{Z_i = \Phi^{-1}\{F_i(Y_i)\}} and the contour is
\deqn{c_h\{\Phi(z_1),\Phi(z_2)\}\phi(z_1)\phi(z_2),}
that is, the fitted copula represented with standard Gaussian margins. The
plotted observations on these two scales are obtained internally by calling
\code{\link{GeoPit}} with \code{type = "Uniform"} or
\code{type = "Gaussian"}; therefore the marginal CDF definitions are shared
by the two diagnostic functions rather than being duplicated in
\code{GeoScatterplot}.

For models fitted with \code{copula = "Gaussian"}, \code{"SkewGaussian"},
or \code{"Clayton"}, the uniform- and Gaussian-scale contours are evaluated
directly from the fitted copula density. The response marginal distribution is
not used in this contour calculation. This direct evaluation is numerically
more stable and ensures that the Gaussian-scale plot represents the same fitted
copula as the uniform-scale plot. For native non-copula random-field models,
the copula density is recovered from the fitted response and marginal
densities.

These transformations preserve the copula and are useful for displaying
reflection symmetry or asymmetry without confounding from the fitted marginal
distribution. In particular, reflection symmetry corresponds to invariance
under \eqn{(u_1,u_2) \mapsto (1-u_1,1-u_2)} on the uniform scale and under
\eqn{(z_1,z_2) \mapsto (-z_1,-z_2)} on the Gaussian scale. By default,
Gaussian-score panels use square plotting regions and the interval
\eqn{[-3,3]^2}.

Native fitted contours are available for the continuous models
\code{"Gaussian"}, \code{"SkewGaussian"}, \code{"Tukeyh"},
\code{"StudentT"}, \code{"SinhAsinh"}, \code{"LogGaussian"},
\code{"Gamma"}, and \code{"Weibull"}. Fitted contours are also available for
\code{"Gaussian"}, \code{"SkewGaussian"}, and \code{"Clayton"} copulas with
the continuous marginal models supported by the pairwise copula
implementation: \code{"Gaussian"}, \code{"StudentT"},
\code{"LogGaussian"}, \code{"Gamma"}, \code{"Weibull"}, \code{"Beta"},
\code{"Beta2"}, \code{"Kumaraswamy"}, \code{"Kumaraswamy2"},
\code{"Logistic"}, and \code{"SkewLaplace"}.

Fitted contours are restricted to univariate, non-misspecified spatial or
spatio-temporal \code{GeoFit} objects. If \code{scale = "Original"},
\code{residuals = FALSE}, and the fitted location varies among sites, a
unique original-scale bivariate contour is not defined. In that case, use
\code{residuals = TRUE} or select the uniform or Gaussian scale. The
\code{residuals} argument is not used with the uniform or Gaussian scale.
}

\section{Spatio-temporal ordering}{
For raw fixed-location input, \code{data} is a \eqn{T \times N}{T x N}
matrix with times in rows and sites in columns.  For raw dynamic input,
\code{data} and \code{coordx_dyn} are lists of equal length and are matched
within each time block.  Pair indices and selected \code{time.lag} values
refer to the resulting time-wise concatenation.  A supplied \code{GeoFit}
object uses the same conventions.  See
\code{\link{GeoModels-spacetime-ordering}}.
}

\value{
The function is primarily called for its graphical output.

With the purely spatial original data-and-coordinate interface, the matched
call is returned invisibly. With spatio-temporal raw data, an invisible list
is returned with components \code{call}, \code{spacetime}, \code{time_lags},
\code{time_values}, and \code{panels}. With a \code{GeoFit} object, an
invisible list is returned with components \code{call}, \code{model},
\code{copula}, \code{independence}, \code{contour},
\code{scale}, \code{residuals}, \code{spacetime}, \code{time_lags}, and
\code{panels}. For a marginal independence fit, \code{copula} is reported as
\code{"Independence"} and \code{independence} is \code{TRUE}. Each panel
contains the plotted paired values, representative
spatial lag, and temporal lag. When contours are requested, it also contains
the evaluation grid, density matrix, contour levels, and the fitted correlation
used by the panel when available. Nearest-neighbour panels additionally contain
the pair indices, individual spatial lags, and individual temporal lags.
}

\author{
Moreno Bevilacqua,
\email{moreno.bevilacqua89@gmail.com},
\url{https://sites.google.com/view/moreno-bevilacqua/home};
Víctor Morales Oñate,
\email{victor.morales@uv.cl},
\url{https://sites.google.com/site/moralesonatevictor/};
Christian Caamaño-Carrillo,
\email{chcaaman@ubiobio.cl},
\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{
library(GeoModels)
set.seed(514)

####################################
### example 1 : Weibull random field
####################################
NN <- 1000
coords <- cbind(runif(NN), runif(NN))

corrmodel <- "GenWend"
model <- "Weibull"

param <- list(mean = 0, shape = 8, nugget = 0,
              scale = 0.5, smooth = 0, power2 = 4)

data <- GeoSim(coordx = coords, corrmodel = corrmodel,
                     model = model,
                     param = param)$data

## Original interface: scatterplots on the response scale
GeoScatterplot(data, coords, neighb = c(1, 2))

## Fit 
fit <- GeoFit(data = data, coordx = coords,
              corrmodel = corrmodel, model = "Weibull",
              likelihood = "Marginal", type = "Pairwise",
              neighb = 4,
              start = list(mean = 0, shape = 6,
                            scale = 0.3),
              fixed = list(nugget = 0, smooth = 0,
                           power2 = 4))

GeoScatterplot(fit, neighb = c(1, 2),nlevels=8,scale = "Original")

## Gaussian-score scale: useful for assessing reflection asymmetry
GeoScatterplot(fit, neighb = c(1, 2),nlevels=8, scale = "Gaussian")


#############################################################
### example 2 : beta random field from a skew-Gaussian copula
#############################################################
NN <- 1000
coords <- cbind(runif(NN), runif(NN))

corrmodel <- "GenWend"
model <- "Beta2"
copula <- "SkewGaussian"
nu=0.95
## Main example: Beta2 marginal model with a Skew-Gaussian copula
param <- list(mean = 0, shape = 8, min = 0, max = 1,
              nu = nu, nugget = 0,
              scale = 0.5, smooth = 0, power2 = 4)

data <- GeoSimCopula(coordx = coords, corrmodel = corrmodel,
                     model = model, copula = copula,
                     param = param, sparse = TRUE)$data

## The simulated model is Beta2 with copula = "SkewGaussian"
## Original interface: scatterplots on the response scale
GeoScatterplot(data, coords, neighb = c(1, 2))

## Fit the Beta2 model with a Skew-Gaussian copula
fit <- GeoFit(data = data, coordx = coords,
              corrmodel = corrmodel, model = model,
              copula = copula,
              likelihood = "Marginal", type = "Pairwise",
              neighb = 4,
              start = list(mean = 0, shape = 6,
                            scale = 0.3),
              fixed = list(nugget = 0, smooth = 0,nu=nu,
                           power2 = 4, min = 0, max = 1),
              lower = list(mean = -Inf, shape = 0, scale = 0),
              upper = list(mean = Inf, shape = Inf, scale = Inf),
              optimizer = "nlminb")

## Response scale: marginal shape and copula are both visible
GeoScatterplot(fit, neighb = c(1, 2),
               scale = "Original")

## Gaussian-score scale: useful for assessing reflection asymmetry
GeoScatterplot(fit, neighb = c(1, 2),
               scale = "Gaussian")


#############################################################
### example 3 : spatio-temporal Gaussian random field
#############################################################

set.seed(89)
coordt <- 1:5
coords <- cbind(runif(200), runif(200))

corrmodel <- "Matern_Matern"
param <- list(mean = 0, sill = 1, nugget = 0,
              scale_s = 0.2 / 3, scale_t = 2 / 3,
              smooth_s = 0.5, smooth_t = 0.5)

data_st <- GeoSim(coordx = coords, coordt = coordt,
                  corrmodel = corrmodel,
                  model = "Gaussian", param = param)$data

fit_st <- GeoFit(data = data_st, coordx = coords, coordt = coordt,
                 corrmodel = corrmodel, model = "Gaussian",
                 likelihood = "Marginal", type = "Pairwise",
                 neighb = 3, maxtime = 1,
                 start = list(mean = 0, sill = 1,
                              scale_s = 0.1, scale_t = 0.5),
                 fixed = list(nugget = 0,
                              smooth_s = 0.5, smooth_t = 0.5))

## Four panels: two nearest-neighbour candidate sets at temporal lags 0 and 1.
## With fixed locations, the panel neighb = 1, time.lag = 1 contains the
## colocated temporal pairs and therefore has representative spatial lag h = 0.
out_st <- GeoScatterplot(fit_st, neighb = c(1, 3),
                         time.lag = c(0, 1),
                         scale = "Gaussian")

## The same space-time pair selection is available without fitting a model.
out_raw_st <- GeoScatterplot(data_st, coordx = coords, coordt = coordt,
                             neighb = c(1, 3), time.lag = c(0, 1),
                             contour = FALSE)

## Distance classes at selected temporal instants.
out_raw_bins <- GeoScatterplot(data_st, coordx = coords, coordt = coordt,
                               times = c(1, 3, 5), time.lag = c(0, 2),
                               maxdist = 0.3, numbins = 3,
                               contour = FALSE)


}
\keyword{Scatterplot}
\keyword{spatial}
\keyword{spatio-temporal}

\name{GeoCovmatrix}
\alias{GeoCovmatrix}
\encoding{UTF-8}
\title{Spatial and Spatio-temporal Covariance Matrix of (non-)Gaussian random fields}
\description{
 The function computes the covariance matrix associated to
 a spatial or spatio-temporal Gaussian or supported non-Gaussian random field
 with given underlying covariance model and a set of spatial location sites (and temporal instants).
}
\usage{
GeoCovmatrix(estobj = NULL, coordx = NULL, coordy = NULL, coordz = NULL, coordt = NULL,
 coordx_dyn = NULL, corrmodel, distance = "Eucl", grid = FALSE,
 model = "Gaussian", n = 1, param, anisopars = NULL, radius = 1,
 sparse = FALSE, copula = NULL, X = NULL, spobj = NULL,
 check.duplicates = FALSE)
}
\arguments{
 \item{estobj}{An object of class \code{Geofit} that includes information about data, model and estimates.}
 \item{coordx}{A numeric (\eqn{d \times 2})-matrix or (\eqn{d \times 3})-matrix.
 Coordinates on a sphere for a fixed radius \code{radius}
 are passed in lon/lat format expressed in decimal degrees. The default is \code{NULL};
 this is useful when coordinates are supplied through \code{coordx_dyn}, \code{estobj}, or \code{spobj}.}
 \item{coordy}{A numeric vector giving 1-dimension of spatial coordinates; optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of spatial coordinates; optional argument, the default is \code{NULL}.}
 \item{coordt}{A numeric vector giving one dimension of temporal coordinates. Optional argument; the default is \code{NULL}, in which case a spatial random field is expected. At the moment implemented only for the Gaussian case. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{For dynamic locations, a list of length \eqn{T}, where \eqn{T} is \code{length(coordt)}.  Element \code{coordx_dyn[[t]]} contains the two- or three-dimensional locations at \code{coordt[t]}; covariance-matrix rows and columns concatenate these blocks by time.}
 \item{corrmodel}{String; the name of a correlation model; see \code{\link{GeoCovmatrix}} for the list of implemented correlation models.}
 \item{distance}{String; the name of the spatial distance. The default is \code{"Eucl"}, the Euclidean distance. See \code{\link{GeoFit}}.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data
 are interpreted as spatial or spatio-temporal realisations on a set
 of non-equispaced spatial sites (irregular grid). See \code{\link{GeoFit}}.}
 \item{n}{Positive integer size parameter. For a direct Binomial random field it may be a scalar or one value per observation. For a direct Negative-Binomial random field it is the common number \eqn{r} of successes and must be scalar; the Geometric model is the special case \eqn{r=1}. Default is \eqn{1}.}
 \item{model}{String; the type of random field. See \code{\link{GeoFit}}.}
 \item{param}{A list of parameter values required for the covariance model.}
 \item{anisopars}{A list of two elements \code{"angle"} and \code{"ratio"}, i.e. the anisotropy angle and the anisotropy ratio, respectively.}
 \item{radius}{Numeric; a value indicating the radius of the sphere
 when using covariance models valid for the great circle distance. Default value is \code{1}.}
 \item{sparse}{Logical; if \code{TRUE} the function returns an object of class \code{spam}.
 This option should be used when a parametric compactly supported covariance is
 used. Default is \code{FALSE}.}
 \item{copula}{String; optional copula specification. Observed-scale covariance-matrix construction is implemented for \code{"Gaussian"}, \code{"Clayton"}, and \code{"SkewGaussian"} for the supported continuous copula margins.}
 \item{X}{Numeric design matrix used to evaluate the mean when required by the marginal covariance.  For fixed sites, rows follow time-major order; for dynamic sites, supply a stacked matrix in temporal-block order or a list aligned with \code{coordx_dyn}.}
 \item{spobj}{An object of class \code{sp} or \code{spacetime}. For space-time objects, the current \code{sp2Geo()} conversion uses sequential temporal indices and does not preserve irregular original time spacing; to retain irregular temporal distances, use the explicit-coordinate interface with numeric \code{coordt}.}
\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
}
\section{Spatio-temporal ordering}{
For fixed locations, covariance rows and columns are ordered
\preformatted{
(s[1], t[1]), ..., (s[N], t[1]),
(s[1], t[2]), ..., (s[N], t[2]), ...
}
For dynamic locations, all rows of \code{coordx_dyn[[1]]} come first, then all
rows of \code{coordx_dyn[[2]]}, and so on.  The rows of \code{X} and an
external \code{param$mean} vector use this same order.  See
\code{\link{GeoModels-spacetime-ordering}}.
}

\value{
 Returns an object of class \code{GeoCovmatrix}.
 An object of class \code{GeoCovmatrix} is a list containing
 at most the following components:
 \item{bivariate}{Logical: \code{TRUE} if the Gaussian random field is bivariate, otherwise \code{FALSE}.}
 \item{coordx}{A \eqn{d}-dimensional vector of spatial coordinates.}
 \item{coordy}{A \eqn{d}-dimensional vector of spatial coordinates.}
 \item{coordt}{A \eqn{t}-dimensional vector of temporal coordinates.}
 \item{coordx_dyn}{A list of \eqn{t} matrices of spatial coordinates.}
 \item{covmatrix}{The covariance matrix if \code{type} is \code{Standard}. An object of
 class \code{spam} if \code{type} is \code{Tapering} or \code{Standard} and
 \code{sparse} is \code{TRUE}.}
 \item{corrmodel}{String: the correlation model.}
 \item{distance}{String: the type of spatial distance.}
 \item{grid}{Logical: the value supplied by the user, indicating whether the original spatial coordinates described a regular grid.}
 \item{nozero}{In the case of tapered matrix the percentage of non-zero values in the
 covariance matrix; otherwise \code{NULL}.}
 \item{n}{The Binomial number of trials, or the common Negative-Binomial number \eqn{r} of successes.}
 \item{namescorr}{String: the names of the correlation parameters.}
 \item{numcoord}{Numeric: the number of spatial coordinates.}
 \item{numtime}{Numeric: the number of temporal coordinates.}
 \item{model}{The type of RF, see \code{\link{GeoFit}}.}
 \item{param}{Numeric: the scalar covariance and marginal parameters.}
 \item{mean_external}{For a univariate model, the optional known mean vector supplied through \code{param$mean}. It is stored separately from \code{param}, in observation order, so that covariance and prediction routines do not interpret its entries as regression coefficients.}
 \item{spacetime}{\code{TRUE} if spatio-temporal and \code{FALSE} if spatial covariance model.}
 \item{sparse}{Logical: is the returned object of class \code{spam}?}
 \item{call}{The matched call used to create the object.}
}
\details{


For plain (non-copula) bivariate correlation models, observed-scale covariance is currently validated only for Gaussian margins; unsupported non-Gaussian bivariate combinations are rejected explicitly. For direct Negative-Binomial fields, the public argument \code{n} denotes the single common integer \eqn{r} in the waiting-time construction.
For \code{copula="Gaussian"}, \code{copula="Clayton"}, and \code{copula="SkewGaussian"}, continuous-margin covariance matrices are constructed directly on the observed scale. The implementation supports \code{"Gaussian"}, \code{"StudentT"}, \code{"LogGaussian"}, \code{"Gamma"}, \code{"Weibull"}, \code{"Beta"}, \code{"Beta2"}, \code{"Kumaraswamy"}, \code{"Kumaraswamy2"}, \code{"Logistic"}, \code{"SkewLaplace"}, \code{"Tukeyh"}, \code{"Tukeyh2"}, and \code{"SinhAsinh"}. Under the Gaussian copula, covariance for the last three is evaluated with the exact covariance of the corresponding monotone transformed-Gaussian field. For location-dependent copula margins, pairwise covariances use the two site-specific marginal predictors supplied through \code{X} or an external mean vector. Clayton-like covariance uses cached deterministic copula quadrature; skew-Gaussian covariance uses a cached bivariate Hermite expansion of the latent representation, avoiding pair-by-pair adaptive two-dimensional integration. The count margins \code{"Poisson"}, \code{"Binomial"}, and \code{"BinomialNeg"} are also available for Gaussian, Clayton-like, and skew-Gaussian copula covariance in the global prediction path; Binomial currently requires a common trial count \code{n}.
The same observed-scale covariance engines are used by \code{\link{GeoCovariogram}} and \code{\link{GeoKrig}}, so covariance matrices, fitted semivariogram curves, and linear prediction are numerically consistent.


For univariate models whose covariance depends on the marginal location, the
mean is evaluated as \eqn{X\beta}, where coefficients \code{mean},
\code{mean1}, and so on correspond in order to the columns of \code{X}.  With
\code{X=NULL}, an intercept-only design is used.  A vector
\code{param$mean} may instead supply one known mean per observation and is
mutually exclusive with \code{X}.

 In the spatial case, the covariance matrix of the
 random vector \deqn{[Z(s_1),\ldots,Z(s_n)]^T}
 with a specific spatial covariance model is computed. Here \eqn{n} is the number of spatial location sites.

 In the space-time case, the covariance matrix of the
 random vector \deqn{[Z(s_1,t_1),Z(s_2,t_1),\ldots,Z(s_n,t_1),\ldots,Z(s_n,t_m)]^T}
 with a specific space-time covariance model is computed. Here \eqn{m} is the number of temporal instants.

 In the bivariate case, rows and columns are in variable-block order.
 For common locations the covariance matrix is associated with
 \deqn{[Z_1(s_1),\ldots,Z_1(s_n),Z_2(s_1),\ldots,Z_2(s_n)]^T.}
 For heterotopic locations supplied by \code{coordx_dyn=list(coords1,coords2)},
 the first block contains all observations of variable 1 at \code{coords1}
 and the second contains all observations of variable 2 at \code{coords2}.

 The location site \eqn{s_i} can be a point in the \eqn{d}-dimensional Euclidean space with \eqn{d=2} or \eqn{d=3}
 or a point (given in lon/lat degree format) on a sphere of arbitrary radius.

 A list with all implemented spatial, space-time and bivariate
 correlation models is given below.
 The argument \code{param} is a list including all the parameters of a given
 correlation model specified by the argument \code{corrmodel}.
 For each correlation model one can check the associated parameters' names using \code{\link{CorrParam}}.
 In what follows
 \eqn{\kappa>0}, \eqn{\beta>0}, \eqn{\alpha, \alpha_s, \alpha_t \in (0,2]} and \eqn{\gamma \in [0,1]}.
 The associated parameters in the argument \code{param} are
 \code{smooth}, \code{power2}, \code{power}, \code{power_s}, \code{power_t} and \code{sep} respectively.
 Moreover let \eqn{1(A)=1} when \eqn{A} is true and \eqn{0} otherwise.

 \itemize{
 \item Spatial correlation models:
 \enumerate{
 \item \code{GenCauchy} (generalised Cauchy in Gneiting and Schlather 2004) defined as:
 \deqn{R(h) = ( 1+h^{\alpha} )^{-\beta / \alpha}}
 If \eqn{h} is the geodesic distance then \eqn{\alpha \in (0,1]}.

 \item \code{Matern} defined as:
 \deqn{R(h) = 2^{1-\kappa} \Gamma(\kappa)^{-1} h^\kappa K_\kappa(h)}
 If \eqn{h} is the geodesic distance then \eqn{\kappa \in (0,0.5]}.

 \item \code{Kummer} (Kummer hypergeometric in Ma and Bhadra 2022) defined as:
 \deqn{R(h) = \Gamma(\kappa+\alpha) U(\alpha,1-\kappa,0.5 h^2 ) / \Gamma(\kappa+\alpha)}
 where \eqn{U(.,.,.)} is the Kummer hypergeometric function.
 If \eqn{h} is the geodesic distance then \eqn{\kappa \in (0,0.5]}.

 \item \code{Kummer_Matern} It is a rescaled version of the \code{Kummer} model,
 i.e. \eqn{h} must be divided by \eqn{(2(1+\alpha))^{0.5}}. When \eqn{\alpha} goes to infinity it is the Matern model.

 \item \code{Wave} defined as:
 \deqn{R(h)=\sin(h)/h}
 This model is valid only for dimensions less than or equal to 3.

 \item \code{GenWend} (Generalized Wendland in Bevilacqua et al. 2019) defined as:
 \deqn{R(h) = A (1-h^2)^{\beta+\kappa} F(\beta/2,(\beta+1)/2,2\beta+\kappa+1,1-h^2) 1(h \in [0,1])}
 where \eqn{\mu \ge 0.5(d+1)+\kappa}
 and \eqn{A=(\Gamma(\kappa)\Gamma(2\kappa+\beta+1))/(\Gamma(2\kappa)\Gamma(\beta+1-\kappa)2^{\beta+1})} and \eqn{F(.,.,.)}
 is the Gaussian hypergeometric function.
 The cases \eqn{\kappa=0,1,2} correspond to the \code{Wend0}, \code{Wend1} and \code{Wend2} models respectively.

 \item \code{GenWend_Matern} (Generalized Wendland Matern in Bevilacqua et al. 2022).
 It is defined as a rescaled version of the Generalized Wendland, i.e.
 \eqn{h} must be divided by \eqn{(\Gamma(\beta+2\kappa+1)/\Gamma(\beta))^{1/(1+2\kappa)}}.
 When \eqn{\beta} goes to infinity it is the Matern model.

 \item \code{GenWend_Matern2} (Generalized Wendland Matern second parametrisation).
 It is defined as a rescaled version of the Generalized Wendland, i.e.
 \eqn{h} must be multiplied by \eqn{\beta} and the smoothness parameter is \eqn{\kappa-0.5}.
 When \eqn{\beta} goes to infinity it is the Matern model.

 \item \code{Hypergeometric} (Hypergeometric model in Bevilacqua et al. 2025).

 \item \code{Hypergeometric_Matern} (Hypergeometric model first parametrisation).

 \item \code{Hypergeometric_Matern2} (Hypergeometric model second parametrisation).

 \item \code{Multiquadric} defined as:
 \deqn{R(h) = (1-\alpha 0.5)^{2\beta}/(1+(\alpha 0.5)^2-\alpha \cos(h))^{\beta}, \quad h \in [0,\pi]}
 This model is valid on the unit sphere and \eqn{h} is the geodesic distance.

 \item \code{Sinpower} defined as:
 \deqn{R(h) = 1-(\sin(h/2))^{\alpha},\quad h \in [0,\pi]}
 This model is valid on the unit sphere and \eqn{h} is the geodesic distance.

 \item \code{F_Sphere} (F family in Alegria et al. 2021) defined as:
 \deqn{R(h) = K F(1/\alpha,1/\alpha+0.5,2/\alpha+0.5+\kappa),\quad h \in [0,\pi]}
 where \eqn{K =(\Gamma(a)\Gamma(i))/(\Gamma(i)\Gamma(o))}.
 This model is valid on the unit sphere and \eqn{h} is the geodesic distance.
 }

 \item Spatio-temporal correlation models:
 \itemize{
 \item Non-separable models:
 \enumerate{
 \item \code{Gneiting} defined as:
 \deqn{R(h, u) = \exp(-h^{\alpha_s}/((1+u^{\alpha_t})^{0.5 \gamma \alpha_s}))/(1+u^{\alpha_t})}

 \item \code{Gneiting_GC} defined as:
 \deqn{R(h, u) = \exp(-u^{\alpha_t}/((1+h^{\alpha_s})^{0.5 \gamma \alpha_t}))/(1+h^{\alpha_s})}
 where \eqn{h} can be either Euclidean or geodesic distance.

 \item \code{Iacocesare} defined as:
 \deqn{R(h, u) = (1+h^{\alpha_s}+u^{\alpha_t})^{-\beta}}

 \item \code{Porcu} defined as:
 \deqn{R(h, u) = (0.5 (1+h^{\alpha_s})^\gamma + 0.5 (1+u^{\alpha_t})^\gamma)^{-\gamma^{-1}}}

 \item \code{Porcu1} defined as:
 \deqn{R(h, u) = \exp(-h^{\alpha_s}(1+u^{\alpha_t})^{0.5 \gamma \alpha_s})/((1+u^{\alpha_t})^{1.5})}

 \item \code{Stein} defined as:
 \deqn{R(h, u) = (h^{\psi(u)}K_{\psi(u)}(h))/(2^{\psi(u)}\Gamma(\psi(u)+1))}
 where \eqn{\psi(u)=\nu+u^{0.5\alpha_t}}.

 \item \code{Gneiting_mat_S} defined as:
 \deqn{R(h, u) = \phi(u)^{\tau_t} \mathrm{Mat}(h \phi(u)^{-\beta},\nu_s)}
 where \eqn{\phi(u)=(1+u^{0.5\alpha_t})}, \eqn{\tau_t \ge 3.5+\nu_s}, \eqn{\beta \in [0,1]}.

 \item \code{Gneiting_mat_T} defined by interchanging \eqn{h} with \eqn{u} in \code{Gneiting_mat_S}.

 \item \code{Gneiting_wen_S} defined as:
 \deqn{R(h, u) = \phi(u)^{\tau_t} \mathrm{GenWend}(h \phi(u)^{\beta},\nu_s,\mu_s)}
 where \eqn{\phi(u)=(1+u^{0.5\alpha_t})}, \eqn{\tau_t \ge 2.5+2\nu_s}, \eqn{\beta \in [0,1]}.

 \item \code{Gneiting_wen_T} defined by interchanging \eqn{h} with \eqn{u} in \code{Gneiting_wen_S}.

 \item \code{Matern_Matern_nosep} defined as:
\deqn{R(h, u) = \frac{\mathrm{Matern}(h;\nu_s)\,\mathrm{Matern}(u;\nu_t)}
 {1+\lambda\,h^{2}u^{2}/(1+\lambda)}}
with \eqn{\nu_s,\nu_t>0} and \eqn{\lambda\in[0,1]}.

\item \code{GenWend_GenWend_nosep} defined as:
\deqn{R(h, u) = \frac{\mathrm{GenWend}(h;\nu_s,\delta_s)\,\mathrm{GenWend}(u;\nu_t,\delta_t)}
 {1+\lambda\,h^{2}u^{2}/(1+\lambda)}}
with \eqn{\nu_x \geq -0.5}, \eqn{\delta_x>(d+1)/2+\nu_x} for \eqn{x=s,t}, and \eqn{\lambda\in[0,1]}.

 \item \code{Multiquadric_st} defined as:
 \deqn{R(h, u)= ((1-0.5\alpha_s)^2/(1+(0.5\alpha_s)^2-\alpha_s \psi(u) \cos(h)))^{a_s},\quad h \in [0,\pi]}
 where \eqn{\psi(u)=(1+(u/a_t)^{\alpha_t})^{-1}}.
 This model is valid on the unit sphere and \eqn{h} is the geodesic distance.

 \item \code{Sinpower_st} defined as:
 \deqn{R(h, u)=(\exp(\alpha_s \cos(h) \psi(u)/a_s)(1+\alpha_s \cos(h) \psi(u)/a_s))/k}
 where \eqn{\psi(u)=(1+(u/a_t)^{\alpha_t})^{-1}} and \eqn{k=(1+\alpha_s/a_s)\exp(\alpha_s/a_s)}, \eqn{h \in [0,\pi]}.
 This model is valid on the unit sphere and \eqn{h} is the geodesic distance.
 }

 \item Separable models:

 Space-time separable correlation models are easily obtained as the product of a spatial and a temporal correlation model, that is
 \deqn{R(h,u)=R(h) R(u)}
 Several combinations are possible:
 \enumerate{
 \item \code{Exp_Exp} defined as:
 \deqn{R(h, u) = \exp(-h)\exp(-u)}

 \item \code{Matern_Matern} defined as:
 \deqn{R(h, u) = \mathrm{Matern}(h;\kappa_s)\mathrm{Matern}(u;\kappa_t)}

 \item \code{GenWend_GenWend} defined as:
 \deqn{R(h, u) = \mathrm{GenWend}(h;\kappa_s,\mu_s)\mathrm{GenWend}(u;\kappa_t,\mu_t)}

 \item \code{Stable_Stable} defined as:
 \deqn{R(h, u) = \exp(-h^{\alpha_s})\exp(-u^{\alpha_t})}
 }
 Note that some models are nested (e.g. \code{Exp_Exp} within \code{Matern_Matern}).
 }

 \item Spatial bivariate correlation models:
 \enumerate{
 \item \code{Bi_Matern} (Bivariate full Matern model)
 \item \code{Bi_Matern_contr} (Bivariate Matern model with constraints)
 \item \code{Bi_Matern_sep} (Bivariate separable Matern model)
 \item \code{Bi_LMC} (Bivariate linear model of coregionalization)
 \item \code{Bi_LMC_contr} (Bivariate LMC with constraints)
 \item \code{Bi_Wendx} (Bivariate full Wendland model)
 \item \code{Bi_Wendx_contr} (Bivariate Wendland model with constraints)
 \item \code{Bi_Wendx_sep} (Bivariate separable Wendland model)
 \item \code{Bi_F_Sphere} (Bivariate full F model on the unit sphere)
 }
 }

 \bold{Remarks}:\cr
 In what follows we assume
 \eqn{\sigma^2,\sigma_1^2,\sigma_2^2,\tau^2,\tau_1^2,\tau_2^2,a,a_s,a_t,a_{11},a_{22},a_{12},\kappa_{11},\kappa_{22},\kappa_{12},f_{11},f_{12},f_{21},f_{22}}
 positive.

 The associated names of the parameters in \code{param} are
 \code{sill}, \code{sill_1}, \code{sill_2},
 \code{nugget}, \code{nugget_1}, \code{nugget_2},
 \code{scale}, \code{scale_s}, \code{scale_t}, \code{scale_1}, \code{scale_2}, \code{scale_12},
 \code{smooth_1}, \code{smooth_2}, \code{smooth_12}, \code{a_1}, \code{a_12}, \code{a_21}, \code{a_2}
 respectively.

 Let \eqn{R(h)} be a spatial correlation model given in standard notation.
 Then the covariance model applied with arbitrary variance, nugget and scale equals to \eqn{\sigma^2} if \eqn{h=0} and
 \deqn{C(h)=\sigma^2(1-\tau^2)R(h/a,\ldots), \quad h>0}
 with nugget parameter \eqn{\tau^2} between 0 and 1.

 Similarly, if \eqn{R(h,u)} is a spatio-temporal correlation model given in standard notation,
 then the covariance model is \eqn{\sigma^2} if \eqn{h=0} and \eqn{u=0} and
 \deqn{C(h,u)=\sigma^2(1-\tau^2)R(h/a_s,u/a_t,\ldots), \quad h>0, u>0}

 Here \sQuote{\ldots} stands for additional parameters.

 The bivariate models implemented are the following:
 \enumerate{
 \item \code{Bi_Matern} defined as:
 \deqn{C_{ij}(h)=\rho_{ij}(\sigma_i\sigma_j+\tau_i^2 1(i=j,h=0))\mathrm{Matern}(h/a_{ij},\kappa_{ij}), \quad i,j=1,2,\; h\ge 0}
 where \eqn{\rho=\rho_{12}=\rho_{21}} is the colocated correlation parameter and \eqn{\rho_{ii}=1}.
 The model \code{Bi_Matern_sep} (separable Matern) is a special case when
 \eqn{a=a_{11}=a_{12}=a_{22}} and \eqn{\kappa=\kappa_{11}=\kappa_{12}=\kappa_{22}}.
 The model \code{Bi_Matern_contr} (constrained Matern)
 is a special case when \eqn{a_{12}=0.5(a_{11}+a_{22})} and \eqn{\kappa_{12}=0.5(\kappa_{11}+\kappa_{22})}.

 \item \code{Bi_GenWend} defined as:
 \deqn{C_{ij}(h)=\rho_{ij}(\sigma_i\sigma_j+\tau_i^2 1(i=j,h=0))\mathrm{GenWend}(h/a_{ij},\nu_{ij},\kappa_{ij}), \quad i,j=1,2,\; h\ge 0}
 where \eqn{\rho=\rho_{12}=\rho_{21}} is the colocated correlation parameter and \eqn{\rho_{ii}=1}.
 The model \code{Bi_GenWend_sep} (separable GenWendland) is a special case when
 \eqn{a=a_{11}=a_{12}=a_{22}} and \eqn{\mu=\mu_{11}=\mu_{12}=\mu_{22}}.
 The model \code{Bi_GenWend_contr} (constrained GenWendland)
 is a special case when \eqn{a_{12}=0.5(a_{11}+a_{22})} and \eqn{\mu_{12}=0.5(\mu_{11}+\mu_{22})}.

 \item \code{Bi_LMC} defined as:
 \deqn{C_{ij}(h)=\sum_{k=1}^{2}(f_{ik}f_{jk}+\tau_i^2 1(i=j,h=0))R(h/a_k)}
 where \eqn{R(h)} is a correlation model.
 The model \code{Bi_LMC_contr}
 is a special case when \eqn{f=f_{12}=f_{21}}.
 Bivariate LMC models, in the current version of the package, are obtained with \eqn{R(h)} equal to the exponential correlation model.
 }
}
\references{
 Alegria, A., Cuevas-Pacheco, F., Diggle, P. and Porcu, E. (2021).
 The F-family of covariance functions: A Matérn analogue for modeling random fields on spheres.
 \emph{Spatial Statistics} 43, 100512.

 Bevilacqua, M., Faouzi, T., Furrer, R. and Porcu, E. (2019).
 Estimation and prediction using generalized Wendland functions under fixed domain asymptotics.
 \emph{Annals of Statistics} 47(2), 828--856.

 Bevilacqua, M., Caamaño-Carrillo, C. and Porcu, E. (2022).
 Unifying compactly supported and Matérn covariance functions in spatial statistics.
 \emph{Journal of Multivariate Analysis} 189, 104949.

 Daley, D. J., Porcu, E. and Bevilacqua, M. (2015).
 Classes of compactly supported covariance functions for multivariate random fields.
 \emph{Stochastic Environmental Research and Risk Assessment} 29(4), 1249--1263.

 Emery, X. and Alegria, A. (2022).
 The Gauss hypergeometric covariance kernel for modeling second-order stationary random fields in Euclidean spaces: its compact support, properties and spectral representation.
 \emph{Stochastic Environmental Research and Risk Assessment} 36, 2819--2834.

 Gneiting, T. (2002).
 Nonseparable, stationary covariance functions for space-time data.
 \emph{Journal of the American Statistical Association} 97, 590--600.

 Gneiting, T., Kleiber, W. and Schlather, M. (2010).
 Matérn cross-covariance functions for multivariate random fields.
 \emph{Journal of the American Statistical Association} 105, 1167--1177.

 Ma, P. and Bhadra, A. (2022).
 Beyond Matérn: on a class of interpretable confluent hypergeometric covariance functions.
 \emph{Journal of the American Statistical Association}, 1--14.

 Porcu, E., Bevilacqua, M. and Genton, M. (2015).
 Spatio-temporal covariance and cross-covariance functions of the great circle distance on a sphere.
 \emph{Journal of the American Statistical Association}.
 DOI: 10.1080/01621459.2015.1072541.

 Gneiting, T. and Schlather, M. (2004).
 Stochastic models that separate fractal dimension and the Hurst effect.
 \emph{SIAM Review} 46, 269--282.
}
\seealso{
\code{\link{GeoFit}} for parameter estimation,
  \code{\link{GeoSim}} for simulation,
  \code{\link{GeoKrig}} for prediction,
  \code{\link{CorrParam}} for correlation-parameter names.
}
\author{
 Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com}, \url{https://sites.google.com/view/moreno-bevilacqua/home}

 Víctor Morales-Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/}

 Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl}, \url{https://www.researchgate.net/profile/Christian-Caamano}
}
\examples{
library(GeoModels)
################################################################
###
### Example 1. Estimated spatial covariance matrix associated to
### a WCL estimates using the Matern correlation model
###
###############################################################

set.seed(3)
N <- 300 # number of location sites
x <- runif(N, 0, 1)
y <- runif(N, 0, 1)
coords <- cbind(x, y)
# Set the covariance model's parameters:
corrmodel <- "Matern"
mean <- 0.5
sill <- 1
nugget <- 0
scale <- 0.2/3
smooth <- 0.5

param <- list(mean = mean, sill = sill, nugget = nugget, scale = scale, smooth = smooth)
data <- GeoSim(coordx = coords, corrmodel = corrmodel, param = param)$data
fixed <- list(nugget = nugget, smooth = smooth)
start <- list(mean = mean, scale = scale, sill = sill)

fit0 <- GeoFit(data = data, coordx = coords, corrmodel = corrmodel,
 neighb = 3, likelihood = "Conditional", optimizer = "BFGS",
 type = "Pairwise", start = start, fixed = fixed)
print(fit0)

# estimated covariance matrix using Geofit object
mm <- GeoCovmatrix(fit0)$covmatrix
# estimated covariance matrix
mm1 <- GeoCovmatrix(coordx = coords, corrmodel = corrmodel,
 param = c(fit0$param, fit0$fixed))$covmatrix
sum(mm - mm1)

################################################################
###
### Example 2. Spatial covariance matrix associated to
### the Generalized Wendland-Matern correlation model
###
###############################################################

# Correlation Parameters for Gen Wendland model
CorrParam("GenWend_Matern")
# Gen Wendland Parameters
param <- list(sill = 1, scale = 0.04, nugget = 0, smooth = 0, power2 = 1/1.5)

matrix2 <- GeoCovmatrix(coordx = coords, corrmodel = "GenWend_Matern",
 param = param, sparse = TRUE)

# Percentage of non-zero values
matrix2$nozero

################################################################
###
### Example 3. Spatial covariance matrix associated to
### the Kummer correlation model
###
###############################################################

# Correlation Parameters for Kummer model
CorrParam("Kummer")
param <- list(sill = 1, scale = 0.2, nugget = 0, smooth = 0.5, power2 = 1)

matrix3 <- GeoCovmatrix(coordx = coords, corrmodel = "Kummer", param = param)

matrix3$covmatrix[1:4, 1:4]

################################################################
###
### Example 4. Covariance matrix associated to
### the space-time double Matern correlation model
###
###############################################################

# Define the temporal coordinates:
times <- seq(1, 4, 1)

# Correlation Parameters for double Matern model
CorrParam("Matern_Matern")

# Define covariance parameters
param <- list(scale_s = 0.3, scale_t = 0.5, sill = 1, smooth_s = 0.5, smooth_t = 0.5)

# Simulation of a spatial Gaussian random field:
matrix4 <- GeoCovmatrix(coordx = coords, coordt = times,
 corrmodel = "Matern_Matern", param = param)

dim(matrix4$covmatrix)

################################################################
###
### Example 5. Spatial covariance matrix associated to
### a skew Gaussian RF with Matern correlation model
###
###############################################################

param <- list(sill = 1, scale = 0.3/3, nugget = 0, skew = 4, smooth = 0.5)
# Simulation of a spatial Gaussian random field:
matrix5 <- GeoCovmatrix(coordx = coords, corrmodel = "Matern", param = param,
 model = "SkewGaussian")

# covariance matrix
matrix5$covmatrix[1:4, 1:4]

################################################################
###
### Example 6. Spatial covariance matrix associated to
### a Weibull RF with GenWend correlation model
###
###############################################################

param <- list(scale = 0.3, nugget = 0, shape = 4, mean = 0, smooth = 1, power2 = 5)
# Simulation of a spatial Gaussian random field:
matrix6 <- GeoCovmatrix(coordx = coords, corrmodel = "GenWend", param = param,
 sparse = TRUE, model = "Weibull")

# Percentage of non-zero values
matrix6$nozero

################################################################
###
### Example 7. Spatial covariance matrix associated to
### a binomial Gaussian RF with Generalized Wendland correlation model
###
###############################################################

param <- list(mean = 0.2, scale = 0.2, nugget = 0, power2 = 4, smooth = 0)
# Simulation of a spatial Gaussian random field:
matrix7 <- GeoCovmatrix(coordx = coords, corrmodel = "GenWend", param = param,
 n = 5, sparse = TRUE, model = "Binomial")

as.matrix(matrix7$covmatrix)[1:4, 1:4]

################################################################
###
### Example 8. Covariance matrix associated to
### a bivariate Matern exponential correlation model
###
###############################################################

set.seed(8)
# Define the spatial coordinates of the points:
x <- runif(4, 0, 1)
y <- runif(4, 0, 1)
coords <- cbind(x, y)

# Parameters
param <- list(mean_1 = 0, mean_2 = 0, sill_1 = 1, sill_2 = 2,
 scale_1 = 0.1, scale_2 = 0.1, scale_12 = 0.1,
 smooth_1 = 0.5, smooth_2 = 0.5, smooth_12 = 0.5,
 nugget_1 = 0, nugget_2 = 0, pcol = -0.25)

# Covariance matrix
matrix8 <- GeoCovmatrix(coordx = coords, corrmodel = "Bi_matern", param = param)$covmatrix

matrix8
}

\name{GeoCV}
\alias{GeoCV}
\encoding{UTF-8}
\title{Repeated holdout kriging cross-validation}
\description{
The procedure uses the \code{\link{GeoKrig}} or \code{\link{GeoKrigloc}}
function to compute repeated holdout kriging cross-validation using information from a
\code{\link{GeoFit}} object. The function returns prediction scores.
}

\usage{
GeoCV(
  fit, K = 100, estimation = TRUE,
  optimizer = NULL, lower = NULL, upper = NULL,
  n.fold = 0.05, local = FALSE,
  neighb = NULL, maxdist = NULL, maxtime = NULL,
  sparse = FALSE, type_krig = "Simple", which = 1,
  parallel = FALSE, ncores = 6, progress = TRUE,
  seed = NULL
)
}

\arguments{
 \item{fit}{An object of class \code{\link{GeoFit}}.}

 \item{K}{Positive integer greater than or equal to 2 giving the number of cross-validation iterations. Non-integer values are rejected rather than rounded.}

 \item{estimation}{Logical; if \code{TRUE}, the model is re-estimated on the training observations at each iteration and the fold-specific estimates are used for prediction. If \code{FALSE}, the parameter estimates in \code{fit} are reused; this is therefore a conditional predictive assessment given parameters estimated from the complete dataset, not a fully refitted out-of-sample cross-validation.}

 \item{optimizer}{The type of optimization algorithm if \code{estimation} is
 \code{TRUE}. See \code{\link{GeoFit}} for details. If \code{NULL}, then the
 optimization algorithm stored in \code{fit} is used.}

 \item{lower}{An optional named list giving the values for the lower bounds of
 the parameters when bounded optimization is used and \code{estimation} is
 \code{TRUE}.}

 \item{upper}{An optional named list giving the values for the upper bounds of
 the parameters when bounded optimization is used and \code{estimation} is
 \code{TRUE}.}

 \item{n.fold}{Numeric; the fraction of observations randomly deleted and
 predicted in each cross-validation iteration. At least one training and one
 prediction observation are retained. In the space-time case, sampling also
 retains at least one training observation at every observed time.}

 \item{local}{Logical; if \code{TRUE}, local kriging is performed. The default
 is \code{FALSE}.}

 \item{neighb}{Numeric; an optional positive integer indicating the order of
 neighborhood if local kriging is performed.}

 \item{maxdist}{Numeric; an optional positive value indicating the spatial
 neighborhood distance if local kriging is performed.}

 \item{maxtime}{Numeric; an optional non-negative temporal-distance threshold, expressed in the same units as the fitted \code{coordt}, when local kriging is performed.}

 \item{sparse}{Logical; if \code{TRUE}, kriging and simulation are computed with
 sparse matrix algorithms using the \pkg{spam} package. The default is
 \code{FALSE}. It should be used with compactly supported covariance models.}

 \item{type_krig}{String; the type of kriging. If \code{"Simple"}, the default, the fitted mean coefficients are treated as plug-in values. If \code{"Universal"}, the prediction MSE is additionally corrected using the covariance matrix of the estimated mean coefficients. With \code{estimation = TRUE}, Universal cross-validation is supported for full-likelihood refits, for which \code{varest = TRUE} is requested automatically in each fold. It is deliberately rejected for composite-likelihood refits because each fold would require its own Godambe covariance matrix. With \code{estimation = FALSE}, a stored \code{fit$varcov} may be used; for composite likelihood it can be obtained with \code{GeoVarest}.}

 \item{which}{Numeric; in the case of bivariate cokriging, it indicates which
 variable to predict. It can be \code{1} or \code{2}.}

 \item{parallel}{Logical; default \code{FALSE}. Set \code{TRUE} to evaluate the cross-validation iterations in parallel.}

 \item{ncores}{Positive integer or \code{NULL}; default \code{6}. With \code{parallel=TRUE}, an explicit integer requests that many workers, capped by detected cores and the number of cross-validation jobs. Set \code{ncores=NULL} for automatic selection, capped at six workers and normally leaving one detected core free. Non-integer values are rejected.}

 \item{progress}{Logical; if \code{TRUE}, a progress bar is shown.}

 \item{seed}{Optional finite integer seed used to make the random selection of folds and any stochastic refitting step reproducible. Non-integer values are rejected. If \code{NULL}, the current random number generator state is left unchanged. When supplied, the previous \code{.Random.seed} is restored on exit.}
}

\details{
For a spatio-temporal \code{GeoFit} object, the stored temporal coordinates are reused without imposing equal spacing. Local temporal neighborhoods interpret \code{maxtime} as a distance threshold in the same units as those coordinates.


The function randomly removes a fraction \code{n.fold} of the observations at
each iteration, predicts the removed observations using kriging, and computes a
set of predictive scores.

If \code{estimation = TRUE}, the model is re-estimated at each cross-validation iteration before prediction. Refits preserve the fitted likelihood settings, including pair weighting, pair thinning, distance-memory handling and anisotropy. For misspecified estimators, the response model stored in \code{fit$model} remains distinct from the working likelihood stored in \code{fit$estimation_model}; the latter is used for each refit. Estimated anisotropy parameters are passed once through \code{anisopars}, avoiding duplicate \code{angle}/\code{ratio} entries in the starting or fixed parameter lists.

For Universal kriging with fold-specific re-estimation, full-likelihood fits request \code{varest = TRUE} within each fold. Composite-likelihood refits are not assigned the Godambe covariance matrix from the full dataset: \code{GeoCV} instead rejects \code{estimation = TRUE, type_krig = "Universal"} for composite likelihood, because a statistically coherent analysis would require a fold-specific \code{GeoVarest} calculation.

If \code{estimation = FALSE}, the parameter estimates from the original fit are reused after removing the validation observations from the conditioning data. Consequently, the reported errors assess prediction conditional on parameters that were estimated using the complete dataset and may be more optimistic than a fully refitted out-of-sample cross-validation.

For \code{Poisson}, \code{Binomial}, and \code{BinomialNeg} with \code{copula="SkewGaussian"}, global cross-validation (\code{local=FALSE}) supports both fold refitting and fixed-parameter prediction. Discrete Clayton-like copula models still require \code{estimation=FALSE} until their pairwise likelihood is implemented.

Regular-grid fits are converted to explicit point coordinates after observations
are removed, because each training sample is no longer a complete Cartesian
grid. Space-time covariates may be stored either as one matrix in observation
order or as a list containing one matrix per time.

For Gaussian fitted models all documented scores are computed from the Gaussian
predictive mean and MSE. For non-Gaussian fitted models, only RMSE, MAE and MAD
are returned; \code{brie}, \code{crps}, \code{lscore}, \code{pit},
\code{intscore} and \code{coverage} are set to \code{NA}, because an exact
non-Gaussian predictive distribution is not available from only a mean and MSE.

When \code{seed} is supplied, the cross-validation samples are reproducible. The
function preserves and restores the user's random number generator state.
}

\value{
Returns a list containing the following components:
 \item{rmse}{The vector of root mean squared errors.}
 \item{mae}{The vector of mean absolute errors.}
 \item{mad}{The vector of median absolute errors.}
 \item{brie}{The vector of Brier scores, or \code{NA} for non-Gaussian fits.}
 \item{crps}{The vector of continuous ranked probability scores, or
 \code{NA} for non-Gaussian fits.}
 \item{lscore}{The vector of log-scores, or \code{NA} for non-Gaussian fits.}
 \item{pit_mean}{The vector of mean probability integral transform values within each holdout fold, or \code{NA} for non-Gaussian fits. This is a descriptive PIT summary, not a calibration score: a mean near 0.5 does not by itself imply a uniform PIT distribution.}
 \item{pit}{Backward-compatible alias of \code{pit_mean}.}
 \item{intscore}{The vector of interval scores, or \code{NA} for
 non-Gaussian fits.}
 \item{coverage}{The vector of empirical coverage values, or \code{NA} for
 non-Gaussian fits.}
 \item{seed}{The seed used for reproducibility, or \code{NULL}.}
}

\seealso{\code{\link{GeoKrig}}, \code{\link{GeoKrigloc}}, \code{\link{GeoFit}}.}

\author{
Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},
\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl},
\url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},
\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{
\donttest{
library(GeoModels)

###############################################################
### Example of spatial kriging cross-validation
###############################################################

model <- "Gaussian"
set.seed(79)
x <- runif(400, 0, 1)
y <- runif(400, 0, 1)
coords <- cbind(x, y)

corrmodel <- "GenWend"
mean <- 0
sill <- 5
nugget <- 0
scale <- 0.2
smooth <- 0
power2 <- 4

param <- list(
  mean = mean, sill = sill, nugget = nugget,
  scale = scale, smooth = smooth, power2 = power2
)

data <- GeoSim(coordx = coords, corrmodel = corrmodel,
               param = param)$data

fixed <- list(nugget = nugget, smooth = 0, power2 = power2)
start <- list(mean = 0, scale = scale, sill = 1)
I <- Inf
lower <- list(mean = -I, scale = 0, sill = 0)
upper <- list(mean = I, scale = I, sill = I)

fit <- GeoFit(
  data, coordx = coords, corrmodel = corrmodel,
  model = model, likelihood = "Marginal", type = "Pairwise",
  neighb = 3, optimizer = "nlminb", lower = lower,
  upper = upper, start = start, fixed = fixed
)

#a <- GeoCV(fit, K = 100, estimation = TRUE,
#           parallel = TRUE, seed = 123)
#mean(a$rmse)
}
}

\keyword{Composite}

\name{CheckDistance}
\alias{CheckDistance}
\encoding{UTF-8}
\title{Checking Distance Type}
\description{
Checks the validity and type of the specified distance.
}
\usage{
CheckDistance(distance)
}
\arguments{
 \item{distance}{A character string indicating the type of distance. 
 Available options are: \code{"Eucl"} (Euclidean), \code{"Geod"} (Geodesic), and \code{"Chor"} (Chordal).
 See also \code{\link{GeoCovmatrix}}.}
}
\details{
This function checks whether the specified distance type is valid.
}
\value{
An integer:
\itemize{
 \item \code{0} for Euclidean distance
 \item \code{1} for Geodesic distance
 \item \code{2} for Chordal distance
}
If the input is not recognized, the function returns \code{NULL}.
}
\author{
Moreno Bevilacqua \email{moreno.bevilacqua89@gmail.com} \cr
\url{https://sites.google.com/view/moreno-bevilacqua/home} \cr
Víctor Morales Oñate \email{victor.morales@uv.cl} \cr
\url{https://sites.google.com/site/moralesonatevictor/} \cr
Christian Caamaño-Carrillo \email{chcaaman@ubiobio.cl} \cr
\url{https://www.researchgate.net/profile/Christian-Caamano}
}
\keyword{composite}

\name{rainNLD}
\alias{rainNLD}
\docType{data}
\title{April Precipitation over the Netherlands with Distance-to-Coast Covariate}
\usage{data(rainNLD)}
\description{
A numeric matrix containing gridded April total precipitation (in mm) over the Netherlands,
together with projected coordinates (UTM, km) and a distance-to-coast covariate (km).
The precipitation field is derived from the WorldClim v2.1 monthly climatology for April
(i.e., long-term monthly averages rather than a specific year).
}
\format{
A numeric matrix with 71,401 rows and 4 columns:
\describe{
 \item{Column 1}{Easting coordinate (UTM zone 31N), in kilometers.}
 \item{Column 2}{Northing coordinate (UTM zone 31N), in kilometers.}
 \item{Column 3}{Distance to the coastline, in kilometers (computed from Natural Earth coastline geometry, transformed to UTM).}
 \item{Column 4}{April total precipitation, in millimeters.}
}
}
\details{
Coordinates were projected from longitude/latitude (WGS84) to UTM zone 31N and rescaled to kilometers.
The distance-to-coast covariate was computed as the Euclidean distance (in UTM) from each grid cell
to the nearest coastline.
}
\source{
Fick, S.E., Hijmans, R.J. (2017) WorldClim 2: new 1km spatial resolution climate surfaces for global land areas.
\emph{International Journal of Climatology}, \bold{37}, 4302--4315.
}
\keyword{datasets}

\name{CkLikelihood}
\alias{CkLikelihood}
\encoding{UTF-8}
\title{Checking Composite-likelihood Type}
\description{
 Subroutine called by InitParam.
 The procedure controls the type of the composite-likelihood inserted
 by the users.
}
\usage{
CkLikelihood(likelihood)
}
\arguments{
 \item{likelihood}{String; the configuration of the composite
 likelihood. \code{Marginal} is the default.}
}

\value{The function returns a numeric positive integer, or NULL if the likelihood is invalid.}

\details{ The function controls the type of the composite-likelihood inserted
 by the users.}
\seealso{\code{\link{GeoFit}}}



\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\keyword{Composite}

\name{GeoNA}
\alias{GeoNA}
\encoding{UTF-8}
\title{Deleting NA values (missing values) from a spatial or spatio-temporal dataset.}
\description{The function deletes NA values from a spatial or spatio-temporal dataset }
\usage{GeoNA(data, coordx, coordy=NULL,coordz=NULL, coordt=NULL,
coordx_dyn=NULL, grid=FALSE, X=NULL, setting="spatial")}
\arguments{
 \item{data}{A \eqn{d}{d}-dimensional vector (a single spatial realisation) or a (\eqn{d \times d}{d x d})-matrix (a single spatial realisation on regular grid) or a
 (\eqn{t \times d}{t x d})-matrix (a single spatio-temporal realisation) or an (\eqn{d \times d \times t \times n }{d x d x t})-array
 (a single spatio-temporal realisation on regular grid) giving the data.}
 \item{coordx}{A numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix 
 Coordinates on a sphere for a fixed radius \code{radius} 
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordt}{A numeric vector giving one dimension of temporal coordinates; the default is \code{NULL}, in which case a spatial random field is expected. Temporal coordinates may be irregularly spaced and their supplied numeric values are preserved.}
 \item{coordx_dyn}{For dynamic locations, a list with one two- or three-column coordinate matrix per time.  It must be aligned with the list \code{data}; missing responses remove the corresponding rows from both objects.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data 
 are interpreted as spatial or spatio-temporal realisations on a set
 of non-equispaced spatial sites (irregular grid).}
 \item{X}{Covariates aligned with \code{data}.  For fixed-coordinate space-time data, use an \eqn{TN \times p}{TN x p} matrix ordered as \code{c(t(data))}.  For dynamic data, \code{X} must be a list with \code{X[[t]]} having one row per element of \code{data[[t]]} and per row of \code{coordx_dyn[[t]]}.}
 \item{setting}{String; are data spatial, spatio-temporal or spatial bivariate
 (respectively \code{spatial}, \code{spacetime}, \code{bivariate}) }

}

\section{Spatio-temporal ordering}{
For fixed sites, \code{data} has times in rows and sites in columns; rows of
\code{X} follow time-major order.  Removing a site removes the corresponding
column of \code{data}, coordinate row, and its row from every temporal block
of \code{X}.  For dynamic sites, \code{data}, \code{coordx_dyn}, and
\code{X} are aligned lists and missing values are removed independently
within each temporal block.  See \code{\link{GeoModels-spacetime-ordering}}.
}

\value{
 Returns a list containing the following components:
 \item{coordx}{A \eqn{d}{d}-dimensional vector of spatial coordinates;}
 \item{coordy}{A \eqn{d}{d}-dimensional vector of spatial coordinates;}
 \item{coordt}{A \eqn{t}{t}-dimensional vector of temporal coordinates;}
 \item{data}{The data without NAvalues }
 \item{grid}{\code{TRUE} if the spatial data are observed in a regular grid, otherwise \code{FALSE};}
 \item{perc}{The percentage of NA values.}
 \item{setting}{Are data of spatial or spatio-temporal or spatial bivariate type}
 \item{X}{The covariate matrix after deleting the rows corresponding to removed observations. For fixed-coordinate spatio-temporal data, deleting a site removes its row from every temporal block.} 
}


\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{

library(GeoModels)

# Define the spatial-coordinates of the points:
set.seed(79)
x = runif(200, 0, 1)
y = runif(200, 0, 1)
coords=cbind(x,y)
# Set the exponential cov parameters:
corrmodel = "Matern"
mean=0
sill=1
nugget=0
scale=0.3/3
smooth=0.5
param=list(mean=mean,sill=sill,nugget=nugget,scale=scale,smooth=smooth)

# Simulation of the spatial Gaussian random field:
data = GeoSim(coordx=coords, corrmodel=corrmodel,
 param=param)$data


data[1:100]=NA
# removing NA
a=GeoNA(data,coordx=coords)
a$perc # percentage of NA values 
#a$coordx# spatial coordinates without missing values
#a$data # data without missinng values


}
\keyword{Composite}

\name{GeoKrigloc}
\alias{GeoKrigloc}
\encoding{UTF-8}
\title{Spatial (bivariate) and spatio temporal optimal linear local prediction 
 for Gaussian and non-Gaussian random fields.}
\description{
For a given set of spatial location sites (and temporal instants),
the function computes optimal local linear prediction and the associated mean squared error 
for the Gaussian and non-Gaussian case using a spatial (temporal) neighborhood
computed using the function \code{\link{GeoNeighborhood}}
}
\usage{
GeoKrigloc(estobj=NULL,data, coordx=NULL, coordy=NULL, coordz=NULL,coordt=NULL,
 coordx_dyn=NULL, corrmodel, distance="Eucl", grid=FALSE, 
 loc, neighb=NULL, maxdist=NULL, 
 maxtime=NULL, method="cholesky",
 model="Gaussian", n=1,nloc=NULL, mse=FALSE, 
 param, anisopars=NULL,radius=1,
 sparse=FALSE, time=NULL, type="Standard",
 type_krig="Simple",weigthed=TRUE, 
 which=1, copula=NULL,X=NULL,Xloc=NULL,Mloc=NULL,varcov=NULL,
 spobj=NULL,spdata=NULL,parallel=FALSE,ncores=6,progress=TRUE,
 check.duplicates=FALSE)
}
\arguments{
 \item{estobj}{An object of class Geofit that includes information about data, model and estimates.}
 \item{data}{A \eqn{d}{d}-dimensional vector (a single spatial realisation) or a (\eqn{d \times d}{d x d})-matrix (a single spatial realisation on regular grid) or a
 (\eqn{t \times d}{t x d})-matrix (a single spatio-temporal realisation) or an (\eqn{d \times d \times t \times n }{d x d x t})-array
 (a single spatio-temporal realisation on regular grid) giving the data used for prediction.}
 \item{coordx}{A numeric \eqn{d \times 2}{d x 2} or \eqn{d \times 3}{d x 3}
coordinate matrix.  It may be omitted when \code{coordx_dyn} or \code{spobj}
is supplied.  Coordinates on a sphere for a fixed radius \code{radius} are
passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}

 \item{coordt}{A numeric vector giving the temporal coordinates of the observations. The default is \code{NULL}, in which case a spatial random field is expected. Temporal coordinates may be irregularly spaced; temporal lags are computed from the supplied coordinate values.}
 \item{coordx_dyn}{For dynamic observation locations, a list with one two- or three-column coordinate matrix per element of \code{coordt}.  The rows of \code{coordx_dyn[[t]]} correspond to \code{data[[t]]} and to the same temporal block of \code{X}.  Dynamic observation locations do not alter the ordering of prediction tasks.  See \code{\link{GeoModels-spacetime-ordering}}.}
 \item{corrmodel}{String; the name of a correlation model, for the
 see \code{\link{GeoCovmatrix}} for the list of implemented correlation models.}
 \item{distance}{String; the name of the spatial distance. The default
 is \code{Eucl}, the Euclidean distance. See \code{\link{GeoFit}} for details.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data used for prediction
 are interpreted as spatial or spatio-temporal realisations on a set
 of non-equispaced spatial sites (irregular grid).}
 \item{loc}{A numeric (\eqn{n \times 2}{n x 2})-matrix (where
 \code{n} is the number of spatial sites) giving 2-dimensions of spatial coordinates to be predicted.}
 \item{neighb}{Numeric; an optional positive integer indicating the 
 order of the neighborhood.}
 \item{maxdist}{Numeric; an optional positive value indicating the distance in the spatial neighborhood.}
 \item{maxtime}{Numeric; an optional non-negative maximum temporal-distance threshold, expressed in the same units as \code{coordt}, used for the local temporal neighborhood.}
 \item{method}{String; matrix decomposition used to solve each local
kriging system.  The choices are \code{cholesky} (default) and \code{svd}.
Only \code{cholesky} is available when \code{sparse=TRUE}.}
 \item{n}{Positive integer size parameter. For Binomial it may be scalar or contain one value per observation. For direct Negative Binomial it is the single common number \eqn{r} of successes; Geometric corresponds to \eqn{r=1}. Default is \eqn{1}.}
 \item{nloc}{Positive integer size at prediction tasks. For Binomial, site-specific observation sizes require explicit prediction sizes. For direct Negative Binomial, \code{nloc}, if supplied, must equal the common \eqn{r}; for Geometric it must equal 1. No prediction size is inferred by averaging observation-side values.} 
 \item{mse}{Logical; if \code{TRUE} (the default) MSE of the kriging predictor is computed} 
 \item{model}{String; the type of RF and therefore the densities associated to the likelihood
 objects. \code{Gaussian} is the default, see the Section
 Details.} 
 \item{param}{A list of parameter values required for the correlation model. See Details for the accepted options.}
 \item{anisopars}{ A list of two elements: "angle" and "ratio" i.e. the anisotropy angle and the anisotropy ratio, respectively.}
 \item{radius}{Numeric: the radius of the sphere if coordinates are passed in lon/lat format;Default value is 1.}
 \item{sparse}{Logical; if \code{TRUE} kriging is computed with sparse matrices algorithms 
 using spam package. Default is FALSE. It should be used with compactly supported covariances.} 
 \item{time}{A numeric vector giving the temporal instants to be predicted. Values need not be equally spaced and are interpreted on the same numeric time scale as \code{coordt}. The default is \code{NULL}, in which case only spatial prediction is performed.}
 \item{type}{String; currently only \code{Standard} is computed.
Other legacy values produce a warning and are treated as \code{Standard}.}
 \item{type_krig}{String; \code{Simple} (default) treats fitted mean
coefficients as known.  \code{Universal} uses the same coefficients and adds
their uncertainty from \code{varcov} to each local prediction MSE.}
 \item{weigthed}{Logical legacy argument retained for backward
compatibility.  It has no effect for \code{type="Standard"}.}
 \item{which}{Numeric; In the case of bivariate (tapered) cokriging it indicates which variable to predict.
 It can be 1 or 2}
 \item{copula}{String; optional copula specification. Local linear prediction is implemented for \code{"Gaussian"}, \code{"Clayton"}, and \code{"SkewGaussian"}.}
 \item{X}{Numeric design matrix at the observations.  For fixed-location space-time data, rows follow \code{c(t(data))}, i.e. time then site.  For dynamic observations, use a stacked matrix ordered by temporal blocks, or a list aligned with \code{coordx_dyn}.}
 \item{Xloc}{Numeric design matrix at prediction tasks.  For space-time prediction its rows are location-major: all requested times for \code{loc[1, ]}, then all requested times for \code{loc[2, ]}, and so on.}
 \item{Mloc}{Numeric vector giving the known marginal location predictor at prediction tasks, in the same location-major order as \code{Xloc}. For Gaussian additive margins this is the mean itself; for models with link-scale parametrizations (for example Gamma/Weibull/LogGaussian or Beta2) it is on the same scale as \code{param$mean}. Use either \code{Mloc} or \code{Xloc}, not both.}
 \item{varcov}{Covariance matrix of the estimated parameters, required
for \code{type_krig="Universal"} when \code{mse=TRUE}.  With a composite
fit it should be the sandwich covariance (inverse Godambe information).
It is normally inherited from \code{estobj}.}
 \item{spobj}{An object of class \code{sp} or \code{spacetime}. For space-time objects, the current \code{sp2Geo()} conversion uses sequential temporal indices and does not preserve irregular original time spacing; to retain irregular temporal distances, use the explicit-coordinate interface with numeric \code{coordt}.}
 \item{spdata}{Character:The name of data in the sp or spacetime object}
 \item{parallel}{Logical; default \code{FALSE}. If \code{TRUE}, the
 independent local \code{GeoKrig} calls are evaluated with
 \pkg{future.apply}.}
 \item{ncores}{Numeric or \code{NULL}; default \code{6}. With
 \code{parallel=TRUE}, a numeric value requests that many worker processes,
 capped only by the number of prediction tasks and detected cores. Thus the
 default parallel execution uses up to six workers. Set \code{ncores=NULL} to
 request the automatic/safe mode, in which the package-wide automatic cap of
 six workers and the internal RAM safety check may further reduce the worker
 count.}
 \item{progress}{If TRUE then a progress bar is shown.}
\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
}

\details{

The univariate mean convention is the same as in \code{\link{GeoKrig}}:
\eqn{\mu=X\beta} with coefficients named \code{mean}, \code{mean1}, and so
on, or an external observation mean vector in \code{param$mean}.  The
intercept-only case is equivalent to a one-column matrix of ones.
At prediction tasks, use either \code{Xloc} for \eqn{X_{loc}\beta} or
\code{Mloc} for a directly supplied local mean.  The rows of \code{Xloc} and
the entries of \code{Mloc} follow the local-task order; for fixed-location
space-time kriging this is location first and prediction time second.

For \code{type_krig="Universal"}, each local prediction uses the mean
coefficients already supplied or estimated by \code{GeoFit}; they are not
re-estimated by GLS inside each neighborhood.  If \code{mse=TRUE}, the local
MSE includes the uncertainty of those coefficients using the corresponding
block of \code{varcov}.  For composite likelihood this is the sandwich
covariance (inverse Godambe information).  Mean coefficients missing from
\code{varcov} are treated as fixed.  External mean vectors are already
known, so a requested universal prediction is treated as simple kriging with
a warning.

For \code{copula="Gaussian"}, \code{copula="Clayton"}, and
\code{copula="SkewGaussian"}, each local prediction uses the same
observed-scale copula covariance, marginal-mean calculations, solver, and MSE
implementation as \code{\link{GeoKrig}}. The local predictor is an optimal
linear predictor, not in general the full conditional mean.

\code{GeoKrigloc} uses a single computational path. First,
\code{\link{GeoNeighborhood}} constructs the requested neighborhood for each
prediction task. Then \code{\link{GeoKrig}} is called on each local data set.
This keeps covariance, copula, mean, numerical-solver, and MSE logic centralized
in \code{GeoKrig} rather than duplicating those calculations inside
\code{GeoKrigloc}.

With \code{parallel=TRUE}, independent local \code{GeoKrig} calls are
submitted through \pkg{future.apply}. The default \code{ncores=6} requests
up to six workers, subject only to the number of prediction tasks and detected
cores. Any other numeric \code{ncores} value is treated in the same way and is
not reduced by the RAM safety heuristic. Set \code{ncores=NULL} explicitly to
select the automatic/safe mode, where the package-wide worker resolver and the
internal RAM safety check may reduce the worker count. On platforms where
\pkg{future} reports that forked workers are safe, the automatic backend uses
\code{multicore}; otherwise it uses \code{multisession}. Parallel and serial
execution use the same local neighborhoods and the same \code{GeoKrig}
calculations.

When geometric anisotropy is supplied, observation coordinates, prediction
locations, and neighborhood selection are transformed by the same metric.
Thus the selected local neighbors are the nearest points under the covariance
metric, not under the original isotropic distance.

This function uses \code{\link{GeoKrig}} with a
spatial or spatio-temporal neighborhood computed using \code{\link{GeoNeighborhood}}.
Each local prediction task is evaluated through the same \code{GeoKrig}
implementation used by the non-local prediction interface.
The neighborhood is specified with \code{neighb}, \code{maxdist}, and
\code{maxtime}.  Regular-grid data are internally vectorized in the same
order as \code{expand.grid}; fixed-location space-time observations remain
in time-major order.}

\section{Spatio-temporal ordering}{
Observed fixed-location data use time-major order: a \eqn{T \times N}{T x N}
matrix is vectorized as \code{c(t(data))}.  Dynamic observations are supplied
as aligned lists \code{data[[t]]} and \code{coordx_dyn[[t]]}, concatenated by
time.  The rows of \code{X} and any known observation mean use the same
observation order.

Prediction tasks use the different, location-major order
\preformatted{
loc[1, ] at time[1], ..., loc[1, ] at time[Tloc],
loc[2, ] at time[1], ..., loc[2, ] at time[Tloc], ...
}
Rows of \code{Xloc} and elements of \code{Mloc} must follow this order.  The
returned \code{pred} and \code{mse} objects are \eqn{Tloc \times Nloc}{Tloc x Nloc}
matrices with prediction times in rows and prediction locations in columns.
See \code{\link{GeoModels-spacetime-ordering}}.
}

\value{
 Returns an object of class \code{Kg}.
 An object of class \code{Kg} is a list containing
 at most the following components:
 \item{bivariate}{\code{TRUE} if spatial bivariate cokriging is performed, otherwise \code{FALSE};}
 \item{coordx}{A \eqn{d}{d}-dimensional vector of spatial coordinates used for prediction;}
 \item{coordy}{A \eqn{d}{d}-dimensional vector of spatial coordinates used for prediction;}
 \item{coordz}{A \eqn{d}{d}-dimensional vector of spatial coordinates used for prediction;}
 \item{coordt}{A \eqn{t}{t}-dimensional vector of temporal coordinates used for prediction;}
 \item{corrmodel}{String: the correlation model;}
 \item{covmatrix}{The covariance matrix if \code{type} is \code{Standard}. An object of class spam if \code{type} is \code{Tapering} }
 \item{data}{The vector or matrix or array of data used for prediction}
 \item{distance}{String: the type of spatial distance;}
 \item{grid}{\code{TRUE} if the spatial data used for prediction are observed in a regular grid, otherwise \code{FALSE};}
 \item{loc}{A (\eqn{n \times 2}{n x 2})-matrix of spatial locations to be predicted.}
 \item{n}{The Binomial number of trials, or the common Negative-Binomial number \eqn{r} of successes.} 
 \item{nozero}{In the case of tapered simple kriging the percentage of non zero values in the covariance matrix.
 Otherwise is NULL.}
 \item{numcoord}{Numeric:he number \eqn{d}{d} of spatial coordinates used for prediction;}
 \item{numloc}{Numeric: the number \eqn{n}{n} of spatial coordinates to be predicted;}
 \item{numtime}{Numeric: the number \eqn{d}{d} of the temporal instants used for prediction;}
 \item{numt}{Numeric: the number \eqn{m}{m} of the temporal instants to be predicted;}
 \item{model}{The response model used for prediction after canonicalizing any
 misspecified-Gaussian fitting name.}
 \item{fit_model}{The model name supplied by the caller or stored in the input
 \code{GeoFit} object before prediction canonicalization.}
 \item{param}{The parameter list used for prediction;}
 \item{pred}{For spatio-temporal prediction, a \eqn{m \times n}{m x n} matrix with rows corresponding to \code{time} and columns corresponding to rows of \code{loc}; for spatial prediction, a numeric vector.}
 \item{radius}{Numeric: the radius of the sphere if coordinates are pssed in lon/lat format;}
 \item{spacetime}{\code{TRUE} if spatio-temporal kriging and \code{FALSE} if
 spatial kriging;}
 \item{tapmod}{String: the taper model if \code{type} is \code{Tapering}. Otherwise is NULL.}
 \item{time}{A \eqn{m}{m}-dimensional vector of temporal coordinates to be predicted;}
 \item{type}{String: the type of kriging (Standard or Tapering).}
 \item{type_krig}{String: the type of kriging: Simple or Universal}
 \item{mse}{When \code{mse=TRUE}, the prediction MSE; otherwise \code{NULL}.}

}


\references{
 Gaetan, C. and Guyon, X. (2010) \emph{Spatial Statistics and Modelling}.
 Springer-Verlag, New York.
 Furrer R., Genton, M.G. and Nychka D. (2006). \emph{Covariance
 Tapering for Interpolation of Large Spatial Datasets}.
 Journal of Computational and Graphical Statistics, \bold{15-3}, 502--523.


}

\seealso{\code{\link{GeoCovmatrix}}}



\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}

\examples{

################################################################
############### Examples of Spatial local kriging #############
################################################################
require(GeoModels)
####
model="Gaussian"

# Define the spatial-coordinates of the points:
set.seed(759)
x = runif(1000, 0, 1)
y = runif(1000, 0, 1)
coords=cbind(x,y)
# Set the exponential cov parameters:
corrmodel = "GenWend"
mean=0; sill=1
nugget=0; scale=0.2
param=list(mean=mean,sill=sill,nugget=nugget,smooth=0,
scale=scale,power2=4)

# Simulation of the spatial Gaussian random field:
data = GeoSim(coordx=coords, corrmodel=corrmodel,
 param=param)$data

# Maximum pairwise likelihood fitting of the space time random field:

start=list(scale=scale,sill=sill,mean=mean)
fixed=list(power2=4,smooth=0,nugget=0)
fit = GeoFit(data, coordx=coords, corrmodel=corrmodel, 
 start=start,fixed=fixed,
 likelihood='Conditional', type='Pairwise',
 neighb=3)

# locations to predict
loc_to_pred=matrix(runif(8),4,2)
################################################################
###
### Example 1. Comparing spatial kriging with local kriging for
### a Gaussian random field with GenWend correlation.
### 
###############################################################
param=append(fit$param,fit$fixed)
pr=GeoKrig(fit,loc=loc_to_pred,mse=TRUE)

pr_loc=GeoKrigloc(fit,loc=loc_to_pred,neighb=100,mse=TRUE)

pr$pred;
pr_loc$pred


############################################################
#### Example: spatio temporal Gaussian local kriging ######
############################################################


require(GeoModels)
set.seed(78)
coords=cbind(runif(100),runif(100))
coordt=seq(0,5,0.25)
corrmodel="Matern_Matern"
param=list(nugget=0,mean=0,scale_s=0.2/3,scale_t=0.25/3,sill=2,
 smooth_s=0.5,smooth_t=0.5)

data = GeoSim(coordx=coords, coordt=coordt,
 corrmodel=corrmodel, param=param)$data


# Maximum pairwise likelihood fitting of the space time random field:
start = list(scale_s=0.2/3,scale_t=0.25,sill=2,mean=0)
fixed = list(smooth_s=0.5,smooth_t=0.5,nugget=0)
I=Inf
lower=list(scale_s=0,scale_t=0,sill=0,mean=-I)
upper=list(scale_s=I,scale_t=I,sill=I,mean=I)
fit = GeoFit(data, coordx=coords, coordt=coordt, model=model, corrmodel=corrmodel, 
 likelihood='Conditional', type='Pairwise',start=start,fixed=fixed,
 optimizer="nlminb",lower=lower,upper=upper,
 neighb=3,maxtime=1)

## four location to predict
loc_to_pred=matrix(runif(8),4,2)
## three temporal instants to predict
time=c(0.5,1.5,3.5)


pr=GeoKrig(fit,loc=loc_to_pred,time=time,mse=TRUE)
pr_loc=GeoKrigloc(fit,loc=loc_to_pred,time=time,
 neigh=25,maxtime=1, mse=TRUE)

## full and local prediction 
pr$pred
pr_loc$pred



############################################################
#### Example: spatio bivariate Gaussian local cokriging ######
############################################################
#set.seed(6)
#NN=1500 # number of spatial locations 
#x = runif(NN, 0, 1); 
#y = runif(NN, 0, 1) 
#coords=cbind(x,y) 

## setting parameters
#mean_1 = 2; mean_2= -1
#nugget_1 =0;nugget_2=0
#sill_1 =0.5; sill_2 =1; 

### correlation parameters
#CorrParam("Bi_Matern")
#scale_1=0.2/3; scale_2=0.15/3; scale_12=0.5*(scale_2+scale_1) 
#smooth_1=smooth_2=smooth_12=0.5
#pcol = -0.4
#param= list(nugget_1=nugget_1,nugget_2=nugget_2,
# sill_1=sill_1,sill_2=sill_2,
# mean_1=mean_1,mean_2=mean_2,
# smooth_1=smooth_1, smooth_2=smooth_2,smooth_12=smooth_12,
# scale_1=scale_1, scale_2=scale_2,scale_12=scale_12,
# pcol=pcol)

## simulation
#data = GeoSim(coordx=coords, corrmodel="Bi_Matern",model=model,param=param)$data

#fixed=list(mean_1=mean_1,mean_2=mean_2, nugget_1=nugget_1,nugget_2=nugget_2, 
# smooth_1=smooth_1, smooth_2=smooth_2,smooth_12=smooth_12)

#start=list( sill_1=sill_1,sill_2=sill_2,
# scale_1=scale_1,scale_2=scale_2,scale_12=scale_12, pcol=pcol)

## estimation with maximum likelihood 
#fit = GeoFit(data=data,coordx=coords, corrmodel="Bi_Matern",
 # likelihood="Marginal",type="Pairwise",optimizer="BFGS",neighb=5,
 #start=start,fixed=fixed)

###### co-kriging for the fist component ##############
#xx=seq(0,1,0.022)
#loc_to_pred=as.matrix(expand.grid(xx,xx))
#pr1 = GeoKrigloc(fit,which=1,mse=TRUE,loc=loc_to_pred,neighb=100)

#opar=par(no.readonly = TRUE)
#par(mfrow=c(1,2))
#zlim=c(-2.5,2.5)
#colour = rainbow(100)
#fields::quilt.plot(coords,data[1,] ,col=colour,main = paste(" Fist component")) 
#fields::quilt.plot(loc_to_pred,pr1$pred,col=colour,
# main = paste(" Kriging first component"),ylab="")
#par(opar)

}
\keyword{Composite}

\name{GeoSimapprox}
\alias{GeoSimapprox}
\encoding{UTF-8}
\title{Fast simulation of Gaussian and non-Gaussian random fields.}
\description{
 Simulation of Gaussian and some non-Gaussian spatial, spatio-temporal and spatial bivariate 
 random fields using two approximate methods of simulation: circulant embeeding and spectral turning band. (see Examples). 
}
\usage{
GeoSimapprox(coordx=NULL, coordy=NULL, coordz=NULL,coordt=NULL, 
coordx_dyn=NULL,corrmodel, distance="Eucl",
grid=FALSE, max.ext=1,
method="TB", L=10000,model='Gaussian',parallel=FALSE,ncores=6,
n=1,param,anisopars=NULL, radius=6371,X=NULL,spobj=NULL,
nrep=1,progress=TRUE,check.duplicates=FALSE)
}
\arguments{
 \item{coordx}{May be \code{NULL} when \code{coordx_dyn} or \code{spobj} is supplied. Otherwise, a numeric (\eqn{d \times 2}{d x 2})-matrix or (\eqn{d \times 3}{d x 3})-matrix 
 Coordinates on a sphere for a fixed radius \code{radius} 
 are passed in lon/lat format expressed in decimal degrees.}
 \item{coordy}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}
 \item{coordz}{A numeric vector giving 1-dimension of
 spatial coordinates; Optional argument, the default is \code{NULL}.}

 \item{coordt}{A numeric vector giving the temporal coordinates. Optional argument; the default is \code{NULL}, in which case a spatial random field is expected. For space-time \code{method = "CE"}, \code{coordt} must be finite, strictly increasing, and equally spaced; space-time turning-bands simulation is not implemented.}
 \item{coordx_dyn}{For dynamic simulation sites, a list with one two- or three-column coordinate matrix per element of \code{coordt}.  The output at time \eqn{t} follows the row order of \code{coordx_dyn[[t]]}.  See \code{\link{GeoModels-spacetime-ordering}}.}
 \item{corrmodel}{String; the name of a correlation model, for the
 see \code{\link{GeoCovmatrix}} for the list of implemented correlation models.}
 \item{parallel}{Logical; default \code{FALSE}. For \code{method = "TB"}, \code{TRUE} enables an
 adaptive scheduler that chooses among serial execution, parallelization over
 spatial chunks within each turning-bands simulation, and parallelization over
 independent replicates. Small workloads can remain serial when worker overhead
 is expected to dominate.}
 \item{ncores}{Positive integer or \code{NULL}; default \code{6}. With \code{parallel=TRUE}, an explicit integer requests that many workers, capped by detected cores and the available TB jobs. Set \code{ncores=NULL} for automatic selection, capped at six workers and normally leaving one detected core free.}
 \item{distance}{String; the name of the spatial distance. The default
 is \code{Eucl}, the Euclidean distance. Turning-bands simulation requires
 \code{distance="Eucl"}. See \code{\link{GeoFit}} for details.}
 \item{grid}{Logical; if \code{FALSE} (the default) the data
 are interpreted as spatial or spatio-temporal realisations on a set
 of non-equispaced spatial sites (irregular grid).}
 \item{max.ext}{Positive integer; maximum number of successive doubling attempts used by the spatial or temporal CE embedding.}
 \item{method}{String; the approximation method. The default is \code{TB}
 (turning bands). The alternative \code{CE} uses circulant embedding. Spatial
 CE requires an exact regular two-dimensional grid, Euclidean distance and no
 \code{anisopars}. Spatial TB accepts explicit irregular two- or three-dimensional
 Euclidean coordinates for univariate fields. Spatio-temporal CE requires fixed
 spatial locations, an equally spaced time vector and a supported separable
 correlation model.} 

 \item{L}{Numeric; the number of lines in the turning band method.} 
 \item{model}{String; the type of RF and therefore the densities associated to the likelihood
 objects. \code{Gaussian} is the default, see the Section
 Details.}
 \item{n}{Positive integer size parameter. For Binomial it may be scalar or site-specific; for direct Negative Binomial it is the single common number \eqn{r} of successes, with Geometric corresponding to \eqn{r=1}.} 
 \item{param}{A list of parameter values required in the
 simulation procedure of random fields, see \bold{Examples}.}
 \item{anisopars}{ A list of two elements "angle" and "ratio" i.e. the anisotropy angle and the anisotropy ratio, respectively.}
 \item{radius}{Numeric; a value indicating the radius of the sphere when using the great 
 circle distance. Default value is the radius of the earth in Km (i.e. 6371)} 
 \item{X}{Numeric design matrix for the mean.  For fixed locations, rows are ordered by time blocks, with all sites at the first time followed by all sites at the second time.  For dynamic locations, supply either a stacked matrix in temporal-block order or a list with \code{X[[t]]} aligned with \code{coordx_dyn[[t]]}.}
 \item{spobj}{An object of class \code{sp} or \code{spacetime}. For space-time objects, the current \code{sp2Geo()} conversion uses sequential temporal indices and does not preserve irregular original time spacing; to retain irregular temporal distances, use the explicit-coordinate interface with numeric \code{coordt}.}
 \item{nrep}{Numeric; Numbers of indipendent replicates.}
 \item{progress}{Logic; If TRUE then a progress bar is shown.}
\item{check.duplicates}{Logical. If \code{TRUE}, perform a fast exact scan for duplicated observation locations at this user entry point. The default is \code{FALSE}, so no duplicate-location scan is imposed. Internal bootstrap, cross-validation, and refitting calls do not repeat the scan.}
}


\details{

For \code{method = "CE"}, spatial simulations require \code{grid = TRUE},
separate equally spaced axes in \code{coordx} and \code{coordy},
\code{distance = "Eucl"}, and \code{anisopars = NULL}. The simulated grid is
exactly the grid supplied by the user. The \code{sill} and \code{nugget}
parameters are applied once to the latent Gaussian field.

Spatio-temporal CE is a separable hybrid method: Cholesky decomposition is
used for the spatial correlation matrix and circulant embedding/FFT for the
temporal correlation. It requires fixed spatial locations and a finite,
strictly increasing, equally spaced \code{coordt}. Dynamic coordinates supplied
through \code{coordx_dyn} are not supported by the approximate methods; use
\code{GeoSim(..., method = "cholesky")} instead. A single temporal instant is
handled as a spatial draw at the supplied locations. Bivariate approximate
simulation is not available with \code{method = "CE"}; use \code{method = "TB"}
or \code{GeoSim()}.  Turning bands are implemented only for purely spatial
models; \code{method = "TB"} is rejected for all spatio-temporal correlation
models and requires \code{distance="Eucl"}. Purely spatial univariate TB is restricted to the Matern, generalized-Wendland /
hypergeometric, and Kummer families implemented by the spectral sampler; unsupported
correlation families are rejected before entering the TB kernel. \code{L} must be a
positive integer.

Bivariate approximate simulation is restricted to \code{model="Gaussian"},
\code{corrmodel="Bi_matern"}, common spatial support for the two variables, and
\code{method="TB"}. The two marginal sills and nugget effects are applied to the
standardized bivariate TB output before the component means are added. Non-Gaussian
bivariate approximate simulation is rejected explicitly.

\code{Binary} and \code{Bernoulli} are aliases of a binomial field with
\code{n = 1}; \code{Geom} and \code{Geometric} are aliases of a
negative-binomial field with \code{n = 1}. Direct simulation of
\code{model = "Beta2"} is not implemented by \code{GeoSimapprox}; use
\code{\link{GeoSimCopula}} with an explicit copula instead. The function
stops explicitly rather than returning a latent Gaussian draw. Inferential
\code{Gaussian_misp_*} model names and any other model without an implemented direct
simulator are also rejected explicitly.


For the Tukey transformed-Gaussian models, \code{Tukeyh} requires
\code{0 <= tail < 0.5}, while \code{Tukeyh2} requires both
\code{0 <= tail1 < 0.5} and \code{0 <= tail2 < 0.5}.  In \code{Tukeyh2},
\code{tail1} is the right-tail parameter and \code{tail2} is the left-tail
parameter. Zero is an allowed boundary and recovers the Gaussian transformation
on the corresponding side. \code{SinhAsinh} requires a strictly positive
\code{tail}.

The integer latent-field restrictions of \code{GeoSim} apply unchanged:
\code{Gamma} requires integer \code{shape}; \code{Beta} requires integer
\code{shape1} and \code{shape2}; \code{1/df} must be an integer at least 3 for
the direct Student-t constructions; and \code{2*shape} must be integer for the direct
Poisson-Gamma constructions. No such parameter is rounded silently. Binomial and
negative-binomial \code{n} values must be positive integers. The negative-binomial
simulator uses per-location counters rather than storing the complete Bernoulli history,
and count-process simulators accumulate counts without a growing event-indicator matrix.
\code{nrep} must be a positive integer. The returned \code{param} component preserves
the parameter list supplied by the user.

For univariate simulation, the location parameter is
\eqn{\mu=X\beta}, with coefficients named \code{mean}, \code{mean1}, and so
on in the order of the columns of \code{X}.  If \code{X=NULL}, the simulation
is intercept-only and uses scalar \code{mean}.  Alternatively,
\code{param$mean} may be a vector with one value per simulated observation;
this external mean is mutually exclusive with \code{X}.
}

\section{Spatio-temporal ordering}{
With fixed locations, a simulated space-time realization is returned as a
\eqn{T \times N}{T x N} matrix: rows correspond to \code{coordt} and columns
to rows of \code{coordx}.  The corresponding internal and \code{X} row order
is \code{c(t(data))}, i.e. time then site.

With dynamic locations, the simulated realization is a list of length
\eqn{T}.  Element \code{data[[t]]} has one value per row of
\code{coordx_dyn[[t]]}, in the same row order.  Replicates, when requested,
contain objects with this same fixed or dynamic layout.  See
\code{\link{GeoModels-spacetime-ordering}}.
}

\value{
 Returns an object of class \code{GeoSim}.
 An object of class \code{GeoSim} is a list containing
 at most the following components:
 
 \item{bivariate}{Logical:\code{TRUE} if the Gaussian RF is bivariate, otherwise \code{FALSE};}
 \item{coordx}{A \eqn{d}{d}-dimensional vector of spatial coordinates;}
 \item{coordy}{A \eqn{d}{d}-dimensional vector of spatial coordinates;}
 \item{coordt}{A \eqn{t}{t}-dimensional vector of temporal coordinates;}
 \item{coordx_dyn}{A list of dynamical (in time) spatial coordinates;}
 \item{corrmodel}{The correlation model; see \code{\link{GeoCovmatrix}}.}
 \item{data}{The simulated data.  For fixed-location space-time simulation this is a matrix with times in rows and sites in columns; for dynamic locations it is a list with one vector per time, aligned with \code{coordx_dyn}.  See \code{\link{GeoModels-spacetime-ordering}}.}
 \item{distance}{The type of spatial distance;}
 \item{method}{The method of simulation}
 \item{model}{The type of RF, see \code{\link{GeoFit}}.}
 \item{n}{The Binomial number of trials; for direct Negative Binomial, the common number \eqn{r} of successes.} 
 \item{numcoord}{The number of spatial coordinates;}
 \item{numtime}{The number the temporal realisations of the RF;}
 \item{param}{The parameter list supplied to the simulation call;}
 \item{radius}{The radius of the sphere if coordinates are passed in lon/lat format;}
 \item{spacetime}{\code{TRUE} if spatio-temporal and \code{FALSE} if
 spatial RF;}
 \item{nrep}{The number of indipendent replicates;}
}

\references{

T. Gneiting, H. Sevcikova, D. B. Percival, M. Schlather and Y. Jiang (2006)
Fast and Exact Simulation of Large Gaussian Lattice Systems in R2: Exploring the Limits
\emph{Journal of Computational and Graphical Statistics} 15 (3)

M. Bevilacqua, X. Emery, F. Cuevas Pacheco (2025) Fast simulation of Gaussian random fields with
flexible correlation models in Euclidean spaces \emph{arxiv} 


}

\author{Moreno Bevilacqua, \email{moreno.bevilacqua89@gmail.com},\url{https://sites.google.com/view/moreno-bevilacqua/home},
Víctor Morales Oñate, \email{victor.morales@uv.cl}, \url{https://sites.google.com/site/moralesonatevictor/},
Christian Caamaño-Carrillo, \email{chcaaman@ubiobio.cl},\url{https://www.researchgate.net/profile/Christian-Caamano}
}



\section{Three-dimensional coordinates}{
For a purely spatial univariate field, \code{method = "TB"} supports explicit
irregular \eqn{N \times 3} Euclidean coordinates with \code{grid = FALSE} and
\code{anisopars = NULL}. Three-dimensional TB is not implemented for
spatio-temporal or bivariate models. CE remains restricted to regular
two-dimensional grids. For unsupported three-dimensional cases, use
\code{GeoSim(..., method = "cholesky")} instead.
}
\examples{
library(GeoModels)


################################################################
###
### Example 1. Simulation of a large spatial Gaussian RF 
### with Matern covariance model
### using circulant embeeding method
### It works only for regular grid
###############################################################
set.seed(68)
x = seq(0,1,0.005)
y = seq(0,1,0.005)
param=list(smooth=1.5,mean=0,sill=1,scale=0.2/3,nugget=0)
# Simulation of a spatial Gaussian RF with Matern correlation function
data1 <- GeoSimapprox(coordx=x,coordy=y, grid=TRUE,corrmodel="Matern", model="Gaussian",
 method="CE",param=param)$data
if (requireNamespace("fields", quietly = TRUE)) {
  fields::image.plot(matrix(data1, length(x), length(y), byrow = TRUE))
}

################################################################
###
### Example 2. Simulation of a large spatial Tukey-h RF 
### with Matern covariance model
### using spectral Turning band method
### It works for (ir)regular grid
###############################################################
set.seed(68)
x = runif(50000)
y = runif(50000)
coords=cbind(x,y)
param=list(smooth=0.5,mean=0,sill=1,scale=0.06,nugget=0,tail=0.15)
# Simulation of a spatial Gaussian RF with Matern correlation function
data1 <- GeoSimapprox(coords, corrmodel="Matern", model="Tukeyh",
                      method="TB",L=1000,param=param)$data
if (requireNamespace("fields", quietly = TRUE)) fields::quilt.plot(coords,data1)


################################################################
###
### Example 3. Simulation of a large spacetime Gaussian RF 
### with separable matern covariance model
### using Circular embeeding method
### It works for (large) regular time grid
###############################################################
set.seed(68)
coordt <- (0:100)
coords <- cbind( runif(100, 0 ,1), runif(100, 0 ,1))
param <- list(mean = 0, sill = 1, nugget = 0.25,
 scale_s = 0.05, scale_t = 2, 
 smooth_s = 0.5, smooth_t = 0.5)
# Simulation of a spatial Gaussian RF with Matern correlation function
param<-list(nugget=0,mean=0,scale_s=0.2/3,scale_t=2/3,sill=1,smooth_s=0.5,smooth_t=0.5)

data <- GeoSimapprox(coordx=coords, coordt=coordt, corrmodel="Matern_Matern",
 model="Gaussian",method="CE",param=param)$data
dim(data)

################################################################
###
### Example 4. Simulation of a large spacetime Gaussian RF 
### with separable GenWend covariance model
### using Circular embeeding method in time
###############################################################
set.seed(68)
# Simulation of a spatial Gaussian RF with Matern correlation function
param<-list(nugget=0,mean=0,scale_s=0.2,scale_t=3,sill=1,
 smooth_s=0,smooth_t=0, power2_s=4,power2_t=4)

data <- GeoSimapprox(coordx=coords, coordt=coordt, corrmodel="GenWend_GenWend",
 model="Gaussian",method="CE",param=param)$data
dim(data)


################################################################
###
### Example 6. Simulation of a large bivariate Gaussian RF
### with bivariate Matern correlation model 
### using spectral Turning band method
###############################################################

# Define the spatial-coordinates of the points:
#x <- runif(20000, 0, 2)
#y <- runif(20000, 0, 2)
#coords <- cbind(x,y)

# Simulation of a bivariate spatial Gaussian RF:
# with a Bivariate Matern
#set.seed(12)
#param=list(mean_1=4,mean_2=2,smooth_1=0.5,smooth_2=0.5,smooth_12=0.5,
# scale_1=0.12,scale_2=0.1,scale_12=0.15,
# sill_1=1,sill_2=1,nugget_1=0,nugget_2=0,pcol=0.5)
#data <- GeoSimapprox(coordx=coords,corrmodel="Bi_matern",
# param=param,method="TB",L=1000)$data
#opar=par(no.readonly = TRUE)
#par(mfrow=c(1,2))
#fields::quilt.plot(coords,data[1,],col=terrain.colors(100),main="1",xlab="",ylab="")
#fields::quilt.plot(coords,data[2,],col=terrain.colors(100),main="2",xlab="",ylab="")

}
\keyword{Simulation}
